/*

   Simple application lister/icon viewer app
   c 1999 GEC and released to public domain

   This is a simple application written for the Pilot, designed to find all the
applications on the device, put them in an on-screen list, and display the icon
of whatever app is selected in the list.  This program is my first ever
programming project for the Pilot - I'm writing it as a starting point for
writing a general-purpose (free!) Icon Editor application.

*/


#include <Pilot.h>
#include "Callbacks.h"

#include <CharAttr.h>
#include <SysEvtMgr.h>

#include "Resources.h"
#include "AppView.h"
#include "ProfilerHandler.h"
#include "aboutbox.h"

/* Can't have this static or outside modules can't access it. */
/* AppView is gradually becoming non-'static'.  Is this bad? */
globalsType global = {0};


/* appSortFun() - called by Sys sort utilities to sort the app list */
static Int appSortFun(VoidPtr a, VoidPtr b, Long other) {
  return StrCaselessCompare(((AppEnumPtr) a)->appName, ((AppEnumPtr) b)->appName);
}


/* initData() - called by app code on startup to initialize data structures. */
static void initData() {
  UInt card;
  CharPtr stringBase;
  AppEnumPtr apps = 0;
  global.numApps = 0;

  global.enumAppList = MemHandleNew(sizeof(AppEnumType));

  /* Initialize the enumerated list of applications.  This is used for finding
     apps based on their index number in the selection list */
  for (card = 0; card < 2; card++) {
    UInt dbCount = DmNumDatabases(card);
    UInt dbnum;

    MemHandleResize(global.enumAppList, sizeof(AppEnumType) * (global.numApps + dbCount));
    apps = MemHandleLock(global.enumAppList);

    for (dbnum = 0; dbnum < dbCount; dbnum++) {
      ULong dbType;
      LocalID localID = DmGetDatabase(card, dbnum);
      DmDatabaseInfo(card, localID, 0, 0, &(apps[global.numApps].iVer), 0, 0, 0, 0, 0, 0, &dbType, &(apps[global.numApps].appID));
      if (dbType == (ULong)'appl') {
	apps[global.numApps].cardNum = card;
	apps[global.numApps].localID = localID;
	global.numApps++;
      }
    }

    /* Now initialize the string list from each app's tAIN resource */
    for (dbnum = 0; dbnum < global.numApps; dbnum++) {
      DmOpenRef hDB;

      StrCopy(apps[dbnum].appName, "[unknown]");
      hDB = DmOpenDatabase(apps[dbnum].cardNum, apps[dbnum].localID, dmModeReadOnly);
      if (hDB) {
	VoidHand hRes = DmGet1Resource((ULong)'tAIN', 1000);
	if (hRes) {
	  CharPtr appName = MemHandleLock(hRes);
	  if (appName) {
	    StrNCopy(apps[dbnum].appName, appName, appNameLength);
	    apps[dbnum].appName[appNameLength-1] = 0;
	    MemHandleUnlock(hRes);
	  }
	  DmReleaseResource(hRes);
	} else {
	  DmDatabaseInfo(apps[dbnum].cardNum, apps[dbnum].localID, apps[dbnum].appName, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
	}
	DmCloseDatabase(hDB);
      }
    }

    MemHandleUnlock(global.enumAppList);
  }
}


/* StartApplication() - Called by app code when event handler starts */
static Err StartApplication(void) {
  CharPtr *nameArray;
  AppEnumPtr apps;

  FrmGotoForm(MainForm);
  initData();
  apps = MemHandleLock(global.enumAppList);
  if (apps != 0) {
    SysQSort(apps, global.numApps, sizeof(AppEnumType), appSortFun, 0);
    MemHandleUnlock(global.enumAppList);
  }

  return 0;
}


/* StopApplication() - called by app code when app stops */
static void StopApplication(void) {
  FrmCloseAllForms();
  if (global.enumAppList) {
    MemHandleFree(global.enumAppList);
  }
}


/* fldSetFromHandle - utility routine for setting text in a field */
FieldPtr fldSetFromHandle(Word fieldID, VoidHand txtH) {
  VoidHand oldTxtH;
  FormPtr frm = FrmGetActiveForm();
  FieldPtr fldP;

  // get the field and the field's current text handle
  fldP = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID));
  ErrNonFatalDisplayIf(!fldP, "Field not in form");
  oldTxtH = (VoidHand)FldGetTextHandle(fldP);

  // set the field's text to the new text
  FldSetTextHandle(fldP, (Handle)txtH);
  FldDrawField(fldP);

  // free the old field text handle
  if (oldTxtH) {
    MemHandleFree(oldTxtH);
  }

  return fldP;
}


/* fldCopyFromHandle() - copies the text in the handle and sets it to the field. */
FieldPtr fldCopyFromHandle(Word fieldID, VoidHand txtH) {
  VoidHand newTxtH;
  CharPtr temp;

  temp = MemHandleLock(txtH);
  if (temp) {
    newTxtH = MemHandleNew(sizeof(char) * (StrLen(temp) + 1));
    StrCopy(MemHandleLock(newTxtH), temp);
    MemHandleUnlock(newTxtH);
    MemHandleUnlock(txtH);
  }

  return fldSetFromHandle(fieldID, newTxtH);
}


/* drawIconGadget() - draw the icons on-screen */
static void drawIconGadget(FormPtr frm, Word gadget) {
  RectangleType bounds;
  RectangleType oldClip;
  Word gadgetIndex = FrmGetObjectIndex(frm, gadget);
  VoidHand data = FrmGetGadgetData(frm, gadgetIndex);

  WinGetClip(&oldClip);
  FrmGetObjectBounds(frm, gadgetIndex, &bounds);
  WinSetClip(&bounds);

  WinEraseRectangle(&bounds, 0);
  if (data) {
    BitmapPtr bmp = MemHandleLock(data);
    WinDrawBitmap(bmp, bounds.topLeft.x, bounds.topLeft.y);
    MemHandleUnlock(data);
  }
  WinSetClip(&oldClip);
}


void updatePopup() {
  UInt selection = global.currentApp;
  FormPtr frm = FrmGetActiveForm();
  ListPtr list;
  ControlPtr popup;
  AppEnumPtr apps = MemHandleLock(global.enumAppList);

  list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ApplicationList));
  LstMakeItemVisible(list, selection);
  LstSetSelection(list, selection);
  popup = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, AppPopup));
  CtlSetLabel(popup, apps[selection].appName);
  CtlDrawControl(popup);

  MemHandleUnlock(global.enumAppList);
}


static void deInitViewData(Int selection) {
  VoidHand h;
  FormPtr frm = FrmGetActiveForm();
  FieldPtr nameField = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, AppNameField));
  FieldPtr versField = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, AppVersionField));
  DmOpenRef db;
  AppEnumPtr apps = MemHandleLock(global.enumAppList);

  if (FldDirty(nameField)) {
    /* open the database read/write, alter the app-name resource */
    FldSetDirty(nameField, false);
    db = DmOpenDatabase(apps[selection].cardNum, apps[selection].localID, dmModeReadWrite);
    if (db) {
      VoidHand newNameH;
      CharPtr newName;
      Err error = dmErrIndexOutOfRange;

      newName = 0;
      newNameH = FldGetTextHandle(nameField);
      if (newNameH) {
	newName = MemHandleLock(newNameH);
      }

      h = DmGet1Resource(ainRsc, ainID);
      if (h) {
	MemHandleResize(h, StrLen(newName) + 1);
      } else {
	h = DmNewResource(db, ainRsc, ainID, StrLen(newName)+1);
      }

      if (newName && StrLen(newName) > 0) {
	DmStrCopy(MemHandleLock(h), 0, newName);
	StrCopy(apps[selection].appName, newName);
	MemHandleUnlock(h);
	DmReleaseResource(h);
	MemHandleUnlock(newNameH);
      } else {
	/* remove the app name resource */
	Int idx;

	idx = DmFindResource(db, 0, 0, h);
	DmReleaseResource(h);
	DmRemoveResource(db, idx);
	/* set app list name data from app's DB name */
	DmDatabaseInfo(apps[selection].cardNum, apps[selection].localID,
		       apps[selection].appName, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
      }

      DmCloseDatabase(db);
    }
  }


  /* Update app Version, if necessary */
  if (FldDirty(versField)) {
    /* open the database read/write, alter the app-version resource */
    FldSetDirty(versField, false);
    db = DmOpenDatabase(apps[selection].cardNum, apps[selection].localID, dmModeReadWrite);
    if (db) {
      CharPtr newVers;
      Err error = dmErrIndexOutOfRange;

      newVers = FldGetTextPtr(versField);

      h = DmGet1Resource(verRsc, appVersionID);
      if (h) {
	MemHandleResize(h, StrLen(newVers) + 1);
      } else {
	h = DmNewResource(db, verRsc, appVersionID, StrLen(newVers)+1);
      }

      if (newVers && StrLen(newVers) > 0) {
	DmStrCopy(MemHandleLock(h), 0, newVers);
	MemHandleUnlock(h);
	DmReleaseResource(h);
      } else {
	/* remove the app version resource */
	Int idx;

	idx = DmFindResource(db, 0, 0, h);
	DmReleaseResource(h);
	DmRemoveResource(db, idx);
      }

      DmCloseDatabase(db);
    }
  }

  MemHandleUnlock(global.enumAppList);

  h = FrmGetGadgetData(frm, FrmGetObjectIndex(frm, BigIconView));
  if (h) DmReleaseResource(h);
  
  h = FrmGetGadgetData(frm, FrmGetObjectIndex(frm, LilIconView));
  if (h) DmReleaseResource(h);
}


static void displayAppData(Word selection, Word prevSelection) {
  VoidHand tempH;
  FormPtr frm = FrmGetActiveForm();
  ListPtr list;
  FieldPtr dbNameFldP, appNameFldP, appVersionFldP;
  CharPtr tempP;
  AppEnumPtr apps = MemHandleLock(global.enumAppList);
  DmOpenRef hDB = DmOpenDatabase(apps[selection].cardNum, apps[selection].localID, dmModeReadOnly);
  UInt dbAttrib;
  ULong dbSize;
  ControlPtr checkbox, popup;
  Boolean readOnly;


  global.currentApp = selection;

  /*Set Read-only checkbox from attribute */
  DmDatabaseInfo(apps[selection].cardNum, apps[selection].localID, 0, &dbAttrib, 0, 0, 0, 0, 0, 0, 0, 0, 0);
  checkbox = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, AppReadOnlyCheck));
  readOnly = dbAttrib & dmHdrAttrReadOnly;
  CtlSetValue(checkbox, readOnly);
  CtlDrawControl(checkbox);

  /*Same for copy-protect checkbox */
  checkbox = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, AppCopyProtectCheck));
  CtlSetValue(checkbox, (dbAttrib & dmHdrAttrCopyPrevention));
  CtlDrawControl(checkbox);

  /*Same for dirty checkbox */
  checkbox = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, AppDirtyCheck));
  CtlSetValue(checkbox, (dbAttrib & dmHdrAttrAppInfoDirty));
  CtlDrawControl(checkbox);

  /*Same for backup checkbox */
  checkbox = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, AppBackupCheck));
  CtlSetValue(checkbox, (dbAttrib & dmHdrAttrBackup));
  CtlDrawControl(checkbox);


  /* Get the icons, set the gadget data */
  if (selection != prevSelection) {
    deInitViewData(prevSelection);

    tempH = DmGet1Resource(iconType, 1000);
    FrmSetGadgetData(frm, FrmGetObjectIndex(frm, BigIconView), tempH);

    tempH = DmGet1Resource(iconType, 1001);
    FrmSetGadgetData(frm, FrmGetObjectIndex(frm, LilIconView), tempH);

    drawIconGadget(frm, BigIconView);
    drawIconGadget(frm, LilIconView);
  }

  /* Set field for DB name */

  tempH = MemHandleNew(dmDBNameLength * sizeof(char));

  if (tempH) {
    CharPtr DBName = MemHandleLock(tempH);
    DmDatabaseInfo(apps[selection].cardNum, apps[selection].localID, DBName, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
    MemHandleUnlock(tempH);
  }
  fldSetFromHandle(DBNameField, tempH);

  /* Set field attributes for tAIN app name */
  {
    FieldAttrType fieldAttr;
    FieldPtr nameField = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, AppNameField));
    FldGetAttributes(nameField, &fieldAttr);
    if (readOnly) {
      fieldAttr.editable = 0;
      fieldAttr.underlined = noUnderline;
      FldReleaseFocus(nameField);
    } else {
      fieldAttr.editable = 1;
      fieldAttr.underlined = grayUnderline;
    }
    FldSetAttributes(nameField, &fieldAttr);
    FldSetDirty(nameField, false);
  }
  /* set field data for tAIN resource field */
  tempH = DmGet1Resource(ainRsc, ainID);
  if (tempH) {
    fldCopyFromHandle(AppNameField, tempH);
    DmReleaseResource(tempH);
  } else {
    fldSetFromHandle(AppNameField, NULL);
  }

  /* Set field for creator ID information */
  tempH = MemHandleNew(sizeof(char) * 5);
  if (tempH) {
    CharPtr tempStr;
    tempStr = MemHandleLock(tempH);
    if (tempStr) {
      *((ULong *)tempStr) = apps[selection].appID;
      tempStr[4] = 0;
    }
    MemHandleUnlock(tempH);
    fldSetFromHandle(CreatorIDField, tempH);
  } else {
    fldSetFromHandle(CreatorIDField, NULL);
  }


  /* Set field attributes for app Version */
  {
    FieldAttrType fieldAttr;
    FieldPtr versField = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, AppVersionField));
    FldGetAttributes(versField, &fieldAttr);
    if (readOnly) {
      fieldAttr.editable = 0;
      fieldAttr.underlined = noUnderline;
      FldReleaseFocus(versField);
    } else {
      fieldAttr.editable = 1;
      fieldAttr.underlined = grayUnderline;
    }
    FldSetAttributes(versField, &fieldAttr);
    FldSetDirty(versField, false);
  }
  /* Set field for tver Version resource */
  tempH = DmGet1Resource(verRsc, appVersionID);
  if (!tempH) {
    tempH = DmGet1Resource(verRsc, 1000);
  }
  if (tempH) {
    fldCopyFromHandle(AppVersionField, tempH);
    DmReleaseResource(tempH);
  } else {
    fldSetFromHandle(AppVersionField, NULL);
  }


  /* Set DB Size field */
  tempH = MemHandleNew(sizeof(char) * 16);
  tempP = MemHandleLock(tempH);
  DmDatabaseSize(apps[selection].cardNum, apps[selection].localID, 0, &dbSize, 0);
  StrIToA(tempP, dbSize);
  MemHandleUnlock(tempH);
  fldSetFromHandle(AppSizeField, tempH);


  MemHandleUnlock(global.enumAppList);
  DmCloseDatabase(hDB);
}


static Boolean setDBAttrib(int idx, Boolean value, Word flags) {
  /* If possible, set the given bits on the DB stored in the
     apps array to the new value. */
  UInt dbAttrib;
  Boolean readOnly, oldVal, success;
  AppEnumPtr apps;

  /* Lock the apps array */
  apps = MemHandleLock(global.enumAppList);

  /* Get read-only attribute */

  DmDatabaseInfo(apps[idx].cardNum, apps[idx].localID, 0, &dbAttrib, 0, 0, 0, 0, 0, 0, 0, 0, 0);
  oldVal = dbAttrib & flags;
  readOnly = dbAttrib & dmHdrAttrReadOnly;

  /* If we have any hope of setting the copy protection bit */
  if (!readOnly) {
    /* If oldVal and value don't represent the same boolean value */
    if ((oldVal && !value) || (!oldVal && value)) {
      /* Set the bit, and save the new attributes */
      if (value) {
	dbAttrib = dbAttrib | flags;
      } else {
	dbAttrib = dbAttrib & ~flags;
      }
      DmSetDatabaseInfo(apps[idx].cardNum, apps[idx].localID, 0, &dbAttrib, 0, 0, 0, 0, 0, 0, 0, 0, 0);
    }
    success = true;
  } else {
    success = false;
  }

  MemHandleUnlock(global.enumAppList);
  return success;
}


void appListDrawFun(UInt itemNum, RectanglePtr bounds, CharPtr *data) {
  AppEnumPtr apps;
  DmOpenRef hDB;
  CALLBACK_PROLOGUE
  

  apps = MemHandleLock(global.enumAppList);
  if (apps) {
    CharPtr s = apps[itemNum].appName;
    WinDrawChars(s, StrLen(s), bounds->topLeft.x + 16, bounds->topLeft.y);

    hDB = DmOpenDatabase(apps[itemNum].cardNum, apps[itemNum].localID, dmModeReadOnly);
    if (hDB) {
      VoidHand tempH = DmGet1Resource(iconType, 1001);
      if (!tempH) {
	tempH = DmGetResource(iconType, 10001);
      }

      if (tempH) {
	BitmapPtr bmp = MemHandleLock(tempH);
	WinDrawBitmap(bmp, bounds->topLeft.x, bounds->topLeft.y + 1);
	MemHandleUnlock(tempH);
      }
      DmReleaseResource(tempH);
      DmCloseDatabase(hDB);
    }

    MemHandleUnlock(global.enumAppList);
  }

  CALLBACK_EPILOGUE
}


Boolean CommonHandleEvent(EventPtr event) {
  FormPtr frm;
  Word formID;
  Boolean handled;

  handled = false;
  /* only called form app code, so no prologue needed */
  switch (event->eType) {
  case menuEvent:
    switch (event->data.menu.itemID) {
    case MenuAbout:
      doAbout();
      handled = true;
      break;

    case MenuViewGeneral:
    case MenuViewProfile:
      frm = FrmGetActiveForm();
      switch (event->data.menu.itemID) {
      case MenuViewGeneral:
	formID = MainForm;
	break;
      case MenuViewProfile:
	formID = ResDBProfileForm;
	break;
      }
      if (formID != FrmGetFormId(frm)) {
	FrmGotoForm(formID);
	handled = true;
      }
	
    }
    break;

  case popSelectEvent:
    switch (event->data.popSelect.controlID) {
    case AppPopup:
      global.currentApp = event->data.popSelect.selection;
      updatePopup();
      handled = false;
      break;
    }
    break;

  case keyDownEvent:
    switch (event->data.keyDown.chr) {
    case pageUpChr:
      global.prevApp = global.currentApp;
      global.currentApp = (global.currentApp + global.numApps - 1) % global.numApps;
      updatePopup();
      handled = false;

      break;

    case pageDownChr:
      global.prevApp = global.currentApp;
      global.currentApp = (global.currentApp + 1) % global.numApps;
      updatePopup();
      handled = false;
      break;
    }
    break;

  }
  return handled;
}


static Boolean MainHandleEvent(EventPtr event) {
  Boolean handled;
  FormPtr frm;
  ListPtr list;
  VoidHand h;
  CharPtr listSelection;


  CALLBACK_PROLOGUE


  handled = false;
  switch (event->eType) {

  case frmOpenEvent:
    frm = FrmGetActiveForm();
    list = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ApplicationList));
    LstSetListChoices(list, NULL, global.numApps);
    LstSetDrawFunction(list, appListDrawFun);

    FrmSetGadgetData(frm, FrmGetObjectIndex(frm, BigIconView), 0);
    FrmSetGadgetData(frm, FrmGetObjectIndex(frm, LilIconView), 0);

    FrmDrawForm(frm);

    updatePopup();
    displayAppData(global.currentApp, global.currentApp+1);
    handled = true;
    break;

  case frmCloseEvent:
    frm = FrmGetActiveForm();

    deInitViewData(global.currentApp);

    /* It's important to return false - otherwise the form won't get cleaned up by
       the default handler */
    handled = false;

    break;

  case ctlSelectEvent:
    switch (event->data.ctlSelect.controlID) {
    case AppCopyProtectCheck:
      if (!setDBAttrib(global.currentApp, CtlGetValue(event->data.ctlSelect.pControl), 
		       dmHdrAttrCopyPrevention)) {
	FrmAlert(CantChangeAttributesAlert);
      }
      displayAppData(global.currentApp, global.currentApp);
      /* Redisplay checkbox */
      handled = true;
      break;

    case AppDirtyCheck:
      if (!setDBAttrib(global.currentApp, CtlGetValue(event->data.ctlSelect.pControl),
		       dmHdrAttrAppInfoDirty)) {
	FrmAlert(CantChangeAttributesAlert);
      }
      displayAppData(global.currentApp, global.currentApp);
      /* Redisplay checkbox */
      handled = true;
      break;

    case AppBackupCheck:
      if (!setDBAttrib(global.currentApp, CtlGetValue(event->data.ctlSelect.pControl),
		       dmHdrAttrBackup)) {
	FrmAlert(CantChangeAttributesAlert);
      }
      displayAppData(global.currentApp, global.currentApp);
      /* Redisplay checkbox */
      handled = true;
      break;
    }
    break;

  case popSelectEvent:
    displayAppData(event->data.popSelect.selection, event->data.popSelect.priorSelection);
    handled = true;
    break;

  case keyDownEvent:
    switch (event->data.keyDown.chr) {
    case pageUpChr:
      displayAppData(global.currentApp, global.prevApp);
      handled = true;
      break;

    case pageDownChr:
      displayAppData(global.currentApp, global.prevApp);
      handled = true;
      break;
    }
    break;

  }


  CALLBACK_EPILOGUE

  return handled;
}


static Boolean ApplicationHandleEvent(EventPtr event) {
  FormPtr frm;
  Int formId;
  Boolean handled = false;
  
  if (event->eType == frmLoadEvent) {
    //Load the form resource specified in the event then activate it
    formId = event->data.frmLoad.formID;
    frm = FrmInitForm(formId);
    FrmSetActiveForm(frm);
    
    //Set the event handler for the form
    switch (formId) {
    case MainForm:
      FrmSetEventHandler(frm, MainHandleEvent);
      break;

    case ResDBProfileForm:
      FrmSetEventHandler(frm, ProfilerHandleEvent);
      break;
    }
    handled = true;
  }
  
  return handled;
}


static void EventLoop(void) {
  EventType event;
  Word error;
  
  do {
    EvtGetEvent(&event, evtWaitForever);
    if (!SysHandleEvent(&event))
      if (!MenuHandleEvent(0, &event, &error))
	if (!ApplicationHandleEvent(&event))
	  if (!CommonHandleEvent(&event))
	    FrmDispatchEvent(&event);
  } while (event.eType != appStopEvent);
}


DWord PilotMain(Word launchCode, Ptr cmdPBP, Word launchFlags) {
  Err err = 0;
  
  if (launchCode == sysAppLaunchCmdNormalLaunch) {
    if ((err = StartApplication()) == 0) {
      EventLoop();
      StopApplication();
    }
  }
  
  return err;
}
