/* Implementation for Rogue About Box */

#include <Pilot.h>
#include "Callbacks.h"

#include "aboutbox.h"
#include "Resources.h"



/**** Put app-spectific code here ****/
/* doCustomTextBox - accepts text ID, bitmap ID, and a string handle-
   sets gadget data w/ customGadgetDataType appropriate to AppView logo 
   The coordinates here are tweaked, and very application-specific.
*/
Word doCustomTextBox(Word textID, Word bitmap1ID, Word bitmap2ID, const Char *string) {
  Word result;
  /* Allocate a CustomGadgetData with space for three elements - 2 bitmap and 1 string */
  VoidHand customH = MemHandleNew(sizeof(CustomGadgetDataType) + 
				2*sizeof(CustomGadgetElementType));
  VoidHand stringH = MemHandleNew(StrLen(string) + 1);
  VoidHand textH = DmGet1Resource(strRsc, textID);
  //  VoidHand bitmapH = DmGet1Resource(bitmapRsc, bitmapID);

  CustomGadgetDataType *customGadgetData = MemHandleLock(customH);

  CharPtr stringCopy = MemHandleLock(stringH);
  if (stringCopy) {
    StrCopy(stringCopy, string);
    MemHandleUnlock(stringH);
  }


  if (customGadgetData) {
    int x = 0;
    customGadgetData->elementCount = 3;
    customGadgetData->element[x].x = 1;
    customGadgetData->element[x].y = 1;
    customGadgetData->element[x].type = bitmap_Rsc;
    customGadgetData->element[x].data.resID = bitmap1ID;
    x = 1;
    customGadgetData->element[x].x = 44;
    customGadgetData->element[x].y = 5;
    customGadgetData->element[x].type = bitmap_Rsc;
    customGadgetData->element[x].data.resID = bitmap2ID;
    x = 2;
    customGadgetData->element[x].x = 85;
    customGadgetData->element[x].y = 28;
    customGadgetData->element[x].type = string_H;
    customGadgetData->element[x].data.dataH = stringH;
    MemHandleUnlock(customH);
  }

  result = doTextBox(AboutTextForm, textH, customH, customGadgetHandler);

  MemHandleFree(customH);
  MemHandleFree(stringH);
  DmReleaseResource(textH);
  //  DmReleaseResource(bitmapH);

  return result;
}



/* Normal top-level call to About dialogs. */
void doAbout() {
  FormPtr previousForm = FrmGetActiveForm();
  Word result = 0;
  Boolean done = false;

  while (!done) {
    done = true;
    switch (doAboutMain()) {
    case AboutButtonTitle:
      result = doAboutApp();
      done = false;
      break;

    case AboutButtonRogue:
      result = doAboutRogue();
      done = false;
      break;
      
    case AboutButtonHome:
      result = doAboutHome();
      done = false;
      break;
      
    case AboutButtonTeam:
      result = doAboutTeam();
      done = false;
      break;
      
    case AboutButtonLicense:
      result = doAboutLicense();
      done = false;
      break;

    }
    if (!done) {
      done = (result == 0);
    }
  }

  if (previousForm) {
    FrmSetActiveForm(previousForm);
  }
}


/* High level call for About box menu dialog  - FrmDoDialog() behavior with initialization */
Word doAboutMain() {
  FormPtr frm = FrmInitForm(AboutForm);
  Word hitButton;
  
  FieldPtr fieldP = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, AboutMiscField));
  VoidHand text = DmGet1Resource('tSTR', AboutBuildInfoString);
  
  FldSetTextHandle(fieldP, (Handle)text);
  FrmSetActiveForm(frm);
  hitButton = FrmDoDialog(frm);
  
  FldSetTextHandle(fieldP, 0);
  DmReleaseResource(text);

  FrmDeleteForm(frm);
  return hitButton;
}


/* high-level calls for individual About text dialogs */
Word doAboutApp() {
  return doCustomTextBox(AboutAppString, AppLogoBmp, AppTitleBmp, "ABOUT");
}


Word doAboutRogue() {
  return doTypicalTextBox(AboutRogueString, RogueTitleBmp);
}


Word doAboutHome() {
  return doTypicalTextBox(AboutHomeString, HomeTitleBmp);
}


Word doAboutTeam() {
  return doCustomTextBox(AboutTeamString, AppLogoBmp, AppTitleBmp, "DEVELOPERS");
}


Word doAboutLicense() {
  return doCustomTextBox(AboutLicenseString, AppLogoBmp, AppTitleBmp, "LICENSE");
}


/* Wrapper call for doTextBox -
   Accepts resource IDs for a string and bitmap resource, then applies them
   to the AboutTextForm.
   Returns when a button is pressed with the ID of the button.
*/
Word doTypicalTextBox(Word textID, Word bitmapID) {
  VoidHand text = DmGet1Resource(strRsc, textID);
  VoidHand bitmap = DmGet1Resource(bitmapRsc, bitmapID);
  Word result;

  result = doTextBox(AboutTextForm, text, bitmap, bitmapGadgetHandler);

  DmReleaseResource(text);
  DmReleaseResource(bitmap);
  return result;
}


/* Modal dialog with private event loop -
   Since we can't draw a gadget within a call to FrmDoDialog, we have a custom loop
   here emulating a FrmDoDialog call, but custom-made for our form.
   This function does not do any cleanup on handles passed to it, but does
   clean up all resources it allocates itself.
   Returns when a button is pressed with the ID of the button.
*/
Word doTextBox(Word formID, VoidHand text, VoidHand gadgetData, GadgetHandler *drawfunc) {
  Boolean done;
  EventType event;
  Word error, buttonID;
  FormPtr frm, oldFrm;
  FieldPtr fieldP;

  oldFrm = FrmGetActiveForm();

  //  FrmPopupForm(AboutTextForm);
  frm = FrmInitForm(formID);
  
  FrmSetActiveForm(frm);
  FrmSetEventHandler(frm, handleTextBoxEvt);

  fieldP = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, AboutTextField));
  FldSetTextHandle(fieldP, (Handle)text);
  if (gadgetData) {
    FrmSetGadgetData(frm, FrmGetObjectIndex(frm, AboutTextGadget), gadgetData);
  }

  FrmDrawForm(frm);
  if (drawfunc) {
    drawfunc(AboutTextGadget);
  }

  done = false;
  buttonID = 0;
  do {
    EvtGetEvent(&event, evtWaitForever);
    if (!SysHandleEvent(&event)) {
      if (!MenuHandleEvent(0, &event, &error)) {
      	if (!FrmDispatchEvent(&event)) {
	  /* No one else has grabbed the event - now, if it's a ctlSelect on a button,
	     return the button ID - emulating FrmDoDialog() behavior.  */
	  if ((event.eType == ctlSelectEvent) && (((ControlPtr)event.data.ctlSelect.pControl)->style == buttonCtl)) {
	    done = true;
	    buttonID = event.data.ctlSelect.controlID;
	  }
	}
      }
    }

    /* If the event is an appStopEvent, we need to exit and add it back to the queue
       so our real event queue will get it. */
    if (event.eType == appStopEvent) {
      EvtAddEventToQueue(&event);
      buttonID = 0;
      done = true;
    }    
  } while (!done);
      
  FldSetTextHandle(fieldP, 0);
  FrmSetGadgetData(frm, FrmGetObjectIndex(frm, AboutTextGadget), 0);

  /* Life lesson in PalmOS Programming - FrmEraseForm() -
     Since we're doing all this modal dialog stuff ourselves, it's important to call FrmEraseForm()
     before deleting the form.  This will restore the screen from "behind" the dialog, so our other
     (non-modal) form will appear normally again when we're done here. */
  FrmEraseForm(frm);
  FrmDeleteForm(frm);
  if (oldFrm) {
    FrmSetActiveForm(oldFrm);
  }
  return buttonID;
  
}  


void nullGadgetHandler(Word gadgetID) {
}


void bitmapGadgetHandler(Word gadgetID) {
  FormPtr frm = FrmGetActiveForm();
  RectangleType bounds, oldClip;
  Word gadgetIndex = FrmGetObjectIndex(frm, gadgetID);
  VoidHand data = FrmGetGadgetData(frm, gadgetIndex);

  FrmGetObjectBounds(frm, gadgetIndex, &bounds);
  WinGetClip(&oldClip);
  WinSetClip(&bounds);

  if (data) {
    BitmapPtr dataP = MemHandleLock(data);
    if (dataP) {
      WinDrawBitmap(dataP, bounds.topLeft.x, bounds.topLeft.y);
      MemHandleUnlock(data);
    }
  }
  WinSetClip(&oldClip);
}


void customGadgetHandler(Word gadgetID) {
  FormPtr frm = FrmGetActiveForm();
  RectangleType bounds, oldClip;
  Word gadgetIndex = FrmGetObjectIndex(frm, gadgetID);
  VoidHand data = FrmGetGadgetData(frm, gadgetIndex);

  FrmGetObjectBounds(frm, gadgetIndex, &bounds);
  WinGetClip(&oldClip);
  WinSetClip(&bounds);

  if (data) {
    CustomGadgetDataType *customP = MemHandleLock(data);
    if (customP) {
      int cc;
      VoidHand displayDataH;
      BitmapPtr bitmapP;
      CharPtr charP;
      int x, y;
      for (cc = 0; cc < customP->elementCount; cc++) {
	x = customP->element[cc].x + bounds.topLeft.x;
	y = customP->element[cc].y + bounds.topLeft.y;
	switch (customP->element[cc].type) {
	case bitmap_Rsc:
	  displayDataH = DmGet1Resource(bitmapRsc, customP->element[cc].data.resID);
	  bitmapP = MemHandleLock(displayDataH);
	  if (bitmapP) {
	    WinDrawBitmap(bitmapP, x, y);
	    MemHandleUnlock(displayDataH);
	  } else {
	    WinDrawLine(x, y, x+10, y+10);
	    WinDrawLine(x+10, y, x, y+10);
	  }
	  DmReleaseResource(displayDataH);
	  break;

	case bitmap_H:
	  displayDataH = customP->element[cc].data.dataH;
	  bitmapP = MemHandleLock(displayDataH);
	  if (bitmapP) {
	    WinDrawBitmap(bitmapP, x, y);
	    MemHandleUnlock(displayDataH);
	  } else {
	    WinDrawLine(x, y, x+10, y+10);
	    WinDrawLine(x+10, y, x, y+10);
	  }
	  break;

	case string_Rsc:
	  displayDataH = DmGet1Resource(strRsc, customP->element[cc].data.resID);
	  charP = MemHandleLock(displayDataH);
	  WinDrawChars(charP, StrLen(charP), x, y);
	  MemHandleUnlock(displayDataH);
	  DmReleaseResource(displayDataH);
	  break;

	case string_H:
	  displayDataH = customP->element[cc].data.dataH;
	  charP = MemHandleLock(displayDataH);
	  if (charP) {
	    WinDrawChars(charP, StrLen(charP), x, y);
	    MemHandleUnlock(displayDataH);
	  } else {
	    WinDrawLine(x, y, x-10, y+30);
	    WinDrawLine(x-10, y, x, y+30);
	  }
	  break;
	}

      }
      MemHandleUnlock(data);
    }
  }
  WinSetClip(&oldClip);
}


/* fieldPageScroll - scroll the fieldID in the current form 1 page in given direction. */
void fieldPageScroll(Word fieldID, DirectionType direction) {
  FormPtr frm = FrmGetActiveForm();
  FieldPtr field;

  field = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID));
  if (FldScrollable(field, direction)) {
    int linesToScroll = FldGetVisibleLines(field) - 1;

    FldScrollField(field, linesToScroll, direction);
  }
}


/* Standard Event Handler for the AboutTextForm dialog */
Boolean handleTextBoxEvt(EventPtr evt) {
  EventType myEvt;
  Boolean handled;
  FormPtr frm;
  VoidHand h;


  CALLBACK_PROLOGUE

  handled = false;

  frm = FrmGetActiveForm();


  switch(evt->eType) {
    //  case frmOpenEvent:
    //    FrmDrawForm(frm);
    //    drawTextGadget(frm, AboutTextGadget);
    //handled = true;
    //break;

  case ctlRepeatEvent:
    /* Someone's pressing our up/down button.  Don't intercept the event (with handled = true)
       or the system won't re-draw the control correctly */
    switch (evt->data.ctlRepeat.controlID) {
    case AboutTextUpBtn:
      fieldPageScroll(AboutTextField, up);
      break;

    case AboutTextDownBtn:
      fieldPageScroll(AboutTextField, down);
      break;
    }

  case keyDownEvent:
    /* Hardware key was pressed.  If it's up or down, scroll the text */
    /* Don't return true or we won't hear system keypress sound */
    switch (evt->data.keyDown.chr) {
    case pageUpChr:
      fieldPageScroll(AboutTextField, up);
      break;

    case pageDownChr:
      fieldPageScroll(AboutTextField, down);
      break;
    }
    break;
    
  }

  CALLBACK_EPILOGUE

  return handled;
}




