/* Header file for Rogue About Box common code */

#ifndef ABOUTBOX_H
#define ABOUTBOX_H

/* type for gadget handler function pointers */
typedef void GadgetHandler(Word gadgetID);
/**** CustomGadgets ***/
/* CustomGadgets are normal gadgets with an attached data type specified here.
   the CustomGadget data allows you to define the look of the gadget with multiple bitmaps, strings,
   etc. at different offsets from the topleft of the gadget. */
/* type for elements in customGadget data */
typedef enum { bitmap_Rsc, bitmap_H, string_Rsc, string_H } CustomElementEnum;
typedef struct CustomElement_tag {
  CustomElementEnum type;
  UInt x, y;
  union {
    Word resID;
    VoidHand dataH;
  } data;
} CustomGadgetElementType;

/* type for customGadget data */
typedef struct CustomGadgetData_tag {
  UInt elementCount;
  CustomGadgetElementType element[1];
} CustomGadgetDataType;


/* high level */
/* Normal call to initiate About dialogs */
void doAbout();
/* Call About box menu dialog - FrmDoDialog() behavior */
Word doAboutMain();
/* Calls to individual About text dialogs - emulate FrmDoDialog() behavior. */
Word doAboutApp();
Word doAboutRogue();
Word doAboutHome();
Word doAboutTeam();
Word doAboutLicense();

/* medium level */
/* Wrapper call for doTextBox() - Accepts resource IDs for text and bitmap */
Word doTypicalTextBox(Word textID, Word bitmapID);
/* start About text dialog with supplied text and gadget data handles
   and supplied gadget draw function*/
Word doTextBox(Word formID, VoidHand textID, VoidHand gadgetData, GadgetHandler *drawfunc);

/* low level */
/* no-op gadget handler */
void nullGadgetHandler(Word);
/* Gadget handler for gadget with attached bitmap */
void bitmapGadgetHandler(Word);
/* Gadget handler for gadget with attached custom multi-element data structure */
void customGadgetHandler(Word);
/* Scroll the given field by 1 page in given direction */
void fieldPageScroll(Word fieldID, DirectionType direction);
/* Event handler for About text box */
Boolean handleTextBoxEvt(EventPtr evt);

#endif

