/*
* Big Two - A popular poker game for Palm OS.
* Copyright (C) 2000  Woo Kok Tong
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
* Contact the author at onehy@magix.com.sg or visit http://come.to/palmos 
*/

#include <PalmOS.h>
//#ifdef __GNUC__
//#include "Callbacks.h"
//#endif

#include "BigTwo.h"
#include "TypeDef.h"

static char *ranks = "3456789TJQKA2";
static char *suits = "\215\216\217\220";

static int select[13];
static int selectcount=0;
static char scorefield[8][6];  // fields of score form
static char leftfield[3][9];  // fields for card left
static char overallscore[6];	// overall score of human player
static char highscore[5][6];	// fields for high score
static Boolean enableColor=false;  // Is it color device?
static Boolean os35=false;  // Is it OS3.5

//char debugstring1[20];
//char debugstring2[20];
//StrIToA(debugstring1, rc);
//ErrDisplayFileLineMsg(debugstring1, 0, debugstring2);

static Err StartApplication(void);
static void EventLoop(void);
static Boolean ApplicationHandleEvent(EventPtr event);
static void StopApplication(void);
static Boolean BigTwoHandleEvent(EventPtr event);
static Boolean BigTwoHandleMenu(UInt16 menuID);
static Boolean BigTwoHandleButton(UInt16 buttonID);
static void ClearSelection(FormPtr form);
static void SwitchSortOrder(FormPtr form);
static void HandleBots(FormPtr form);
static Boolean CanPlay();
static void ShowControl(FormPtr form);
static void HideControl(FormPtr form);
static void GameOver();
static char *PlayerName(int p);
static Err ReadPrefs();
static void SavePrefs();
static void Clear(int x, int y, int w, int h);
static void DrawCard(FormPtr form, int id, int card, int x, int y, 
	int w, int h, Boolean played);
static void DrawHumanPlayedCards(FormPtr form, Boolean showpass);
static void DrawHumanPlayer(FormPtr form);
static void DrawPalmPlayer(FormPtr form, short playerIndex, Boolean delay);
static void DrawField(FormPtr form, int field, char *title, short value);
static Boolean HighScore();
static void InsertHighScore(const char* name);
static UInt16 DisplayScoreForm(short* hand);
static UInt16 DisplayPrefsForm();
static Boolean DisplayGameOverForm();
static UInt16 DisplayHighScoreForm();

UInt32 PilotMain(UInt16 launchCode, MemPtr cmdPBP, UInt16 launchFlags)
{
	Err err = 0;    

	if (launchCode == sysAppLaunchCmdNormalLaunch) 
	{ 
		if ((err = StartApplication()) == 0) 
		{
			EventLoop();
			StopApplication();
		}
	}

	return err;
}

static Err StartApplication(void)
{
	Err		error;
	UInt32	romversion;

	if ((error = ReadPrefs()) != 0)
		return error;

	// Check OS version
	if ((error = FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romversion)) != 0)
		return error;
	if (romversion >= 0x03503000)
		os35 = true;
	
	// Check for color support
	if (os35)
		WinScreenMode(winScreenModeGetSupportsColor, 0,	0, 0,	&enableColor);

	FrmGotoForm(formID_bigtwo);
	return 0;
}

static void StopApplication(void)
{
	SavePrefs();

   // Close all open forms,  this will force any unsaved data to 
   // be written to the database.
   FrmCloseAllForms();
}

static Boolean ApplicationHandleEvent(EventPtr event)
{
	FormPtr frm;
	Int16   formId;
	Boolean handled = false;

	if (event->eType == frmLoadEvent) 
	{
		formId = event->data.frmLoad.formID;
		frm = FrmInitForm(formId);
		FrmSetActiveForm(frm);

		switch(formId) 
		{
		case formID_bigtwo:
			FrmSetEventHandler(frm, BigTwoHandleEvent);
			break;
		}
		handled = true;
	}
	return handled;
}

static void EventLoop(void)
{
	EventType event;
	UInt16 error;
  	Boolean tapDelay=false;

	do 
	{
    	EvtGetEvent(&event, (tapDelay ? 5 : evtWaitForever));
      if (tapDelay && event.eType == nilEvent)
      {
         tapDelay = false;
         continue;
      }
		else if (event.eType == ctlSelectEvent) 
   	{
			if (tapDelay) 
    		{
      		// The user tapped two buttons just real fast.  This is
      	 	// almost certainly an accidental double-tap.
      		continue;
      	}
      	tapDelay = true;
   	}
		//EvtGetEvent(&event, evtWaitForever);
		if (! SysHandleEvent(&event))
			if (! MenuHandleEvent(0, &event, &error))
				if (! ApplicationHandleEvent(&event))
					FrmDispatchEvent(&event);
	} while (event.eType != appStopEvent);
}

static Boolean BigTwoHandleEvent(EventPtr event)
{
	FormPtr	form=FrmGetActiveForm();
	Boolean	handled=false;

	switch (event->eType) 
	{
	case frmOpenEvent:
		FrmHideObject(form, FrmGetObjectIndex(form, labelID_p2pass));
		FrmHideObject(form, FrmGetObjectIndex(form, labelID_p3pass));
		FrmHideObject(form, FrmGetObjectIndex(form, labelID_p4pass));
		FrmHideObject(form, FrmGetObjectIndex(form, labelID_youpass));
		FrmHideObject(form, FrmGetObjectIndex(form, labelID_autopass));
		//CtlHideControl(FrmGetObjectPtr(form, FrmGetObjectIndex(form, buttonID_pass)));
		//CtlHideControl(FrmGetObjectPtr(form, FrmGetObjectIndex(form, buttonID_play)));
		CtlHideControl(FrmGetObjectPtr(form, FrmGetObjectIndex(form, buttonID_lead)));
		FrmDrawForm(form);
		if (s.turn == HUMAN)
			ShowControl(form);
		else
			HideControl(form);
		DrawHumanPlayer(form);
		DrawHumanPlayedCards(form, false);
		DrawPalmPlayer(form, 1, false);
		DrawPalmPlayer(form, 2, false);
		DrawPalmPlayer(form, 3, false);

		HandleBots(form);
	 	selectcount = 0;
		handled = true;
		break;

	case ctlSelectEvent:  // A control button was pressed and released.
   	handled = BigTwoHandleButton(event->data.ctlEnter.controlID);
		break;

   case menuEvent:
   	handled = BigTwoHandleMenu(event->data.menu.itemID);
      break;
   }

   return handled;
}

static Boolean BigTwoHandleMenu(UInt16 menuID)
{
	Boolean 	handled=false;
	UInt16 		rc;
	short		hand[4]={0, 0, 0, 0};
	
	switch (menuID)
	{
	case menuitemID_howtoplay:
	  	FrmHelp(helpStr_bigtwo);
		handled = true;
		break;

	case menuitemID_about:
		FrmAlert(alertID_about);
		handled = true;
		break;

	case menuitemID_new:
		rc = FrmAlert(alertID_new);
		if (rc == 0) 
		{
			NewSet();
			FrmGotoForm(formID_bigtwo);
		}
		handled = true;
		break;

	case menuitemID_score:
		rc = DisplayScoreForm(hand);
		if (rc != buttonID_ok) 
		{
			rc = FrmAlert(alertID_new);
			if (rc == 0) 
			{
				NewSet();
				FrmGotoForm(formID_bigtwo);
			}
		}
		handled = true;
		break;

	case menuitemID_prefs:
		DisplayPrefsForm();
		handled = true;
		break;

	case menuitemID_reset:
		rc = FrmAlert(alertID_resethigh);
		if (rc == 0) 
		{
			ResetHighScore();
			//FrmGotoForm(formID_bigtwo);
		}
		handled = true;
		break;

	case menuitemID_high:
		DisplayHighScoreForm();
		handled = true;
		break;

	}  // switch (menuID)

	return handled;
}

static Boolean BigTwoHandleButton(UInt16 buttonID)
{
	Boolean	handled=false;
	FormPtr	form=FrmGetActiveForm();
	UInt16 		rc;
	short		status;
	int		i, j;
	char		ca[6];
	
	switch (buttonID)
	{
	case buttonID_pass:
		if (s.lastplayer == s.turn)  // cannot pass
		{
			FrmAlert(alertID_invalidplay);
			ClearSelection(form);
			handled = true;
			break;
		}
		if (selectcount > 0)  // did player accidentally press it?
		{
			rc = FrmAlert(alertID_pass);
			if (rc == 1)  // button is NO
			{
				handled = true;
				break;
			}
		}
		s.p[s.turn].numplayed = 0;
		s.turn = Next(s.turn);
		ClearSelection(form);
		Clear(CARDLEFT, CARDTOP+3*(cardHeight+1), 4*(CARDHALFWIDTH+1)+cardWidth, cardHeight);
		HideControl(form);
		FrmShowObject(form, FrmGetObjectIndex(form, labelID_youpass));
		
		HandleBots(form);
		handled = true;
		break;

	case buttonID_lead:
	case buttonID_play:
		if (selectcount < 1 || selectcount == 4 || selectcount > 5)
		{
			// Invalid play
			FrmAlert(alertID_invalidplay);
			ClearSelection(form);
			handled = true;
			break;
		}
		for (i=0; i<selectcount; i++)
		{
			ca[i] = s.p[s.turn].card[cardIndex(select[i])];
		}
		SortCards(ca, selectcount, TRUE);
		if (!VerifyPlay(ca, selectcount))
		{
			// Invalid play
			FrmAlert(alertID_invalidplay);
			ClearSelection(form);
			handled = true;
			break;
		}
		// Update state info
		//if (s.prefs.sort == SUIT)
		//	SortCards(ca, selectcount, FALSE);
		UpdateState(ca, selectcount);
		//if (s.prefs.sort == SUIT)
			// Sort the played cards so that VerifyPlay works correctly
		//	SortCards(s.p[HUMAN].played, s.p[HUMAN].numplayed, TRUE);

		ClearSelection(form);
		HideControl(form);
		DrawHumanPlayer(form);
		DrawHumanPlayedCards(form, true);
		if (s.p[s.turn].left == 0)
		{
			GameOver();
			handled = true;
			break;
		}
		s.turn = Next(s.turn);
		
		HandleBots(form);
		handled = true;
		break;

	case buttonID_toggle:
		s.prefs.sort = (s.prefs.sort+1)%2;
		SwitchSortOrder(form);
   	handled = true;
   	break;

	case buttonID_clear:
		ClearSelection(form);
   	handled = true;
   	break;

	default:  // card buttons
		status = FrmGetControlValue(form, FrmGetObjectIndex(form, buttonID));
		if (status)  // add selected card
		{
			select[selectcount] = buttonID;
			selectcount++;
		}
		else  // remove selected card
		{
			for (i=0; i<selectcount; i++)
			{
				if (select[i] == buttonID)
				{
					for (j=i; j<selectcount-1; j++)
					{
						select[j] = select[j+1];
					}
					selectcount --;
					break;
				}
			}
		}
		handled = true;
	}
	return handled;
}

static void ClearSelection(FormPtr form)
{
	short i;
	
	for (i=0; i<selectcount; i++)
 	{
     	FrmSetControlValue(form, FrmGetObjectIndex(form, select[i]), 0);
 	}
 	selectcount = 0;
}

static void SwitchSortOrder(FormPtr form)
{
	char ca[13];
	short i, j, count=selectcount;
	
	for (i=0; i<selectcount; i++)
	{
		ca[i] = s.p[HUMAN].card[cardIndex(select[i])];
	}
	ClearSelection(form);
  	SortCards(s.p[HUMAN].card, s.p[HUMAN].left, s.prefs.sort==RANK);
  	SortCards(ca, count, s.prefs.sort==RANK);
  	if (s.prefs.sort == RANK)
  	{
		FrmHideObject(form, FrmGetObjectIndex(form, bitmapID_rank));
		FrmShowObject(form, FrmGetObjectIndex(form, bitmapID_suit));
	}
	else
  	{
		FrmHideObject(form, FrmGetObjectIndex(form, bitmapID_suit));
		FrmShowObject(form, FrmGetObjectIndex(form, bitmapID_rank));
	}
  	DrawHumanPlayer(form);
  	j = 0;
  	for (i=0; i<s.p[HUMAN].left; i++)
  	{
  		if (j >= count) break;
  		if (s.p[HUMAN].card[i] == ca[j])
  		{
  			select[j] = cardID(i);
     		FrmSetControlValue(form, FrmGetObjectIndex(form, select[j]), 1);
     		selectcount++;
     		j++;
     	}
  	}
}

static void HandleBots(FormPtr form)
{
	short	factor=4;
	Boolean humanturn=false;
  	
  	if (s.prefs.speed == SLOW)
  		factor = 2;
  	else if (s.prefs.speed == FAST)
  		factor = 8;
	if (s.turn == HUMAN) 
		humanturn = true;

	while (!humanturn)
	{
   	PalmPlay();
   	
   	SysTaskDelay((Int32)(SysTicksPerSecond() / factor));
   	DrawPalmPlayer(form, s.turn, true);
   	if (s.p[s.turn].left == 0)
   	{
			GameOver();
			return;
   	}
   	s.turn = Next(s.turn);
   	if (s.turn == HUMAN)
   	{
	   	if (!s.prefs.autopass)
   		{
      		if (s.p[HUMAN].numplayed == 0)
					Clear(CARDLEFT, CARDTOP+3*(cardHeight+1), 4*(CARDHALFWIDTH+1)+cardWidth, cardHeight);
				ShowControl(form);
   			break;
   		}
   		else  // Autopass
   		{
   			if (CanPlay())
      		{
      			if (s.p[HUMAN].numplayed == 0)
						Clear(CARDLEFT, CARDTOP+3*(cardHeight+1), 4*(CARDHALFWIDTH+1)+cardWidth, cardHeight);
					ShowControl(form);
      			break;
      		}
      		else
      		{
					Clear(CARDLEFT, CARDTOP+3*(cardHeight+1), 4*(CARDHALFWIDTH+1)+cardWidth, cardHeight);
     	   	   FrmShowObject(form, FrmGetObjectIndex(form, labelID_autopass));
			    	SysTaskDelay((Int32)(SysTicksPerSecond() / factor));
    	   	   s.p[HUMAN].numplayed = 0;
      			s.turn = Next(s.turn);
      			continue;
      		}
   		}
   	}
   }  // while (!humanturn)
}

static void ShowControl(FormPtr form)
{
	short	i;
	
	//Clear(CARDLEFT, CARDTOP+3*(cardHeight+1), 4*(CARDHALFWIDTH+1)+cardWidth, cardHeight);
	FrmHideObject(form, FrmGetObjectIndex(form, labelID_you));
	if (s.prefs.sort == RANK)
	{
		FrmHideObject(form, FrmGetObjectIndex(form, bitmapID_rank));
		FrmShowObject(form, FrmGetObjectIndex(form, bitmapID_suit));
	}
	else
	{
		FrmHideObject(form, FrmGetObjectIndex(form, bitmapID_suit));
		FrmShowObject(form, FrmGetObjectIndex(form, bitmapID_rank));
	}
	FrmShowObject(form, FrmGetObjectIndex(form, labelID_yourplay1));
	FrmShowObject(form, FrmGetObjectIndex(form, labelID_yourplay2));
	CtlSetEnabled(FrmGetObjectPtr(form, FrmGetObjectIndex(form, buttonID_toggle)), true);
	CtlSetEnabled(FrmGetObjectPtr(form, FrmGetObjectIndex(form, buttonID_clear)), true);
	if (s.lastplayer == HUMAN)
	{
		CtlHideControl(FrmGetObjectPtr(form, FrmGetObjectIndex(form, buttonID_play)));
		CtlHideControl(FrmGetObjectPtr(form, FrmGetObjectIndex(form, buttonID_pass)));
		CtlShowControl(FrmGetObjectPtr(form, FrmGetObjectIndex(form, buttonID_lead)));
		CtlSetEnabled(FrmGetObjectPtr(form, FrmGetObjectIndex(form, buttonID_lead)), true);
	}
	else
	{
		CtlHideControl(FrmGetObjectPtr(form, FrmGetObjectIndex(form, buttonID_lead)));
		CtlShowControl(FrmGetObjectPtr(form, FrmGetObjectIndex(form, buttonID_play)));
		CtlShowControl(FrmGetObjectPtr(form, FrmGetObjectIndex(form, buttonID_pass)));
		CtlSetEnabled(FrmGetObjectPtr(form, FrmGetObjectIndex(form, buttonID_play)), true);
		CtlSetEnabled(FrmGetObjectPtr(form, FrmGetObjectIndex(form, buttonID_pass)), true);
	}
	for (i=0; i<13; i++)
		CtlSetEnabled(FrmGetObjectPtr(form, FrmGetObjectIndex(form, cardID(i))), true);
}

static void HideControl(FormPtr form)
{
	short	i;
	
	FrmHideObject(form, FrmGetObjectIndex(form, labelID_yourplay1));
	FrmHideObject(form, FrmGetObjectIndex(form, labelID_yourplay2));
	FrmHideObject(form, FrmGetObjectIndex(form, labelID_youpass));
	FrmHideObject(form, FrmGetObjectIndex(form, labelID_autopass));
	CtlSetEnabled(FrmGetObjectPtr(form, FrmGetObjectIndex(form, buttonID_pass)), false);
	CtlSetEnabled(FrmGetObjectPtr(form, FrmGetObjectIndex(form, buttonID_play)), false);
	CtlSetEnabled(FrmGetObjectPtr(form, FrmGetObjectIndex(form, buttonID_lead)), false);
	CtlSetEnabled(FrmGetObjectPtr(form, FrmGetObjectIndex(form, buttonID_toggle)), false);
	CtlSetEnabled(FrmGetObjectPtr(form, FrmGetObjectIndex(form, buttonID_clear)), false);
	if (s.prefs.sort == RANK)
	{
		FrmHideObject(form, FrmGetObjectIndex(form, bitmapID_rank));
		FrmShowObject(form, FrmGetObjectIndex(form, bitmapID_suit));
	}
	else
	{
		FrmHideObject(form, FrmGetObjectIndex(form, bitmapID_suit));
		FrmShowObject(form, FrmGetObjectIndex(form, bitmapID_rank));
	}
	for (i=0; i<13; i++)
		CtlSetEnabled(FrmGetObjectPtr(form, FrmGetObjectIndex(form, cardID(i))), false);
	FrmShowObject(form, FrmGetObjectIndex(form, labelID_you));
}

// Can human play? i.e. is his card bigger than what is played?
static Boolean CanPlay()
{
	char tmp[13];
	short i;
	
	if (s.lastplayer == HUMAN) 
		return true;
	if (s.p[s.lastplayer].numplayed > s.p[HUMAN].left)
		return false;
	// No autopass for five card play
	if (s.p[s.lastplayer].numplayed == 5)
		return true;

	for (i=0; i<s.p[HUMAN].left; i++)
		tmp[i] = s.p[HUMAN].card[i];
	SortCards(tmp, s.p[HUMAN].left, TRUE);
	switch (s.p[s.lastplayer].numplayed)
	{
	case 1:
		if (Rank(tmp[s.p[HUMAN].left-1]) > Rank(s.p[s.lastplayer].played[0]))
			return true;
		else if (Rank(tmp[s.p[HUMAN].left-1]) == Rank(s.p[s.lastplayer].played[0]) &&
			Suit(tmp[s.p[HUMAN].left-1]) > Suit(s.p[s.lastplayer].played[0]))
			return true;
		else
			return false;
		break;
	case 2:
		for (i=s.p[HUMAN].left-1; i>0; i--)
		{
			if (Rank(tmp[i]) == Rank(tmp[i-1]))
			{
      		if (Rank(tmp[i]) > Rank(s.p[s.lastplayer].played[1]))
      			return true;
      		else if (Rank(tmp[i]) == Rank(s.p[s.lastplayer].played[1]) &&
      			Suit(tmp[i]) > Suit(s.p[s.lastplayer].played[1]))
      			return true;
      		else
      			return false;
			}
		}
		return false;
		break;
	case 3:
		for (i=s.p[HUMAN].left-1; i>1; i--)
		{
			if ((Rank(tmp[i]) == Rank(tmp[i-1])) && (Rank(tmp[i-1]) == Rank(tmp[i-2])))
			{
      		if (Rank(tmp[i]) > Rank(s.p[s.lastplayer].played[1]))
      			return true;
      		else
      			return false;
			}
		}
		return false;
		break;
	default:
		return true;
	}
}

static void GameOver()
{
	short i, j, maxscore, hand[4];
	UInt16 rc;
	Boolean gamewon=false, gameset=false;
	
  	SysTaskDelay((Int32)(SysTicksPerSecond()/2));
  	UpdateScore(hand);
   rc = DisplayScoreForm(hand);
	if (rc != buttonID_ok) 
   {
   	rc = FrmAlert(alertID_new);
   	if (rc == 0) 
   	{
   		NewSet();
      	FrmGotoForm(formID_bigtwo);
      	return;
      }
  	}
  	if (s.round >= MAX_ROUND)
  	{
  		// Find winner
	  	maxscore = s.p[0].score;
	  	j = 0;
   	for (i=0; i<4; i++)
   	{
   		if (s.p[i].score > maxscore)
   		{
   			maxscore = s.p[i].score;
   			j = i;
   		}
   	}
   	gamewon = true;
  	}
  	if (gamewon)
  	{
  		if (DisplayGameOverForm())
	  		DisplayHighScoreForm();
  		//FrmCustomAlert(alertID_gameover, PlayerName(j), j==HUMAN?"have":"has", NULL);
		//s.p[j].games ++;
  		// If any player exceeds MAX_GAMES, new set of games
		for (i=0; i<4; i++)
		{
			s.p[i].score = 0;
  			//if (s.p[i].games >= MAX_GAMES)
  			//	gameset = true;
  		}
		if (gameset)
			NewSet();
		else
			NewGame();
		FrmGotoForm(formID_bigtwo);
	}
	else
	{
		NewGame();
		FrmGotoForm(formID_bigtwo);
	}
}

static char *
PlayerName(int p)
{
  switch (p) {
    case 0:  return "Well done, you";
    case 1:  return "Player 2";
    case 2:  return "Player 3";
    default: return "Player 4";
  }
}

static Err ReadPrefs()
{
  UInt16 size;
  Int16 rc;

  size = sizeof(s);
  rc = PrefGetAppPreferences(BigTwoAppID, 0, &s, &size, true);
  if (rc == noPreferenceFound || size != sizeof(s)) {
    bzero(&s, sizeof(s));
    NewSet();
    ResetHighScore();
  }
  return 0;
}

static void SavePrefs()
{
  PrefSetAppPreferences(BigTwoAppID, 0, 1, &s, sizeof(s), true);
}

static void Clear(int x, int y, int w, int h)
{
  RectangleType rect;

  rect.topLeft.x = x;
  rect.topLeft.y = y;
  rect.extent.x = w;
  rect.extent.y = h;
  WinEraseRectangle(&rect, 0);
}

static void DrawCard(FormPtr form, int id, int card, int x, int y, int w, int h, Boolean played)
{
	int index;
	RectangleType rect;
	FontID	oldFont=FntSetFont(0);

	rect.topLeft.x = x + 1;
	rect.topLeft.y = y + 1;
	rect.extent.x = w - 2;
	rect.extent.y = h - 2;
	if (id >= 0 && played == false) 
	{
		index = FrmGetObjectIndex(form, cardID(id));
		FrmSetObjectPosition(form, index, rect.topLeft.x, rect.topLeft.y);
		FrmSetObjectBounds(form, index, &rect);
		FrmShowObject(form, index);
	}

	if (!enableColor)
  		WinDrawRectangleFrame(rectangleFrame, &rect);
	else
	{
		WinPushDrawState();
		WinSetForeColor(255);
  		WinDrawRectangleFrame(rectangleFrame, &rect);
		WinSetForeColor(6);
		WinPaintRectangle (&rect, 0);
		WinSetBackColor(6);
	}

	WinDrawChars(&ranks[Rank(card)], 1, x+4, y+2);
	if (!enableColor || Suit(card) == 1 || Suit(card) == 3)
	{
		WinDrawChars(&suits[Suit(card)], 1, x+2, y+13);
	}
	else
	{
		RectangleType	arect;
		UInt16			index;

		arect.topLeft.x = x+2;
		arect.topLeft.y = y+13;
		arect.extent.x = BMPWIDTH;
		arect.extent.y = BMPHEIGHT;

		if (Suit(card) == 0)
			index = FrmGetObjectIndex(form, bitmapID_diamond);
		else
			index = FrmGetObjectIndex(form, bitmapID_heart);
		FrmSetObjectPosition(form, index, arect.topLeft.x, arect.topLeft.y);
		FrmSetObjectBounds(form, index, &arect);
		FrmShowObject(form, index);

	}

	if (enableColor) WinPopDrawState();
	FntSetFont(oldFont);
}

static void DrawHumanPlayedCards(FormPtr form, Boolean showpass)
{
	short i, n=s.p[HUMAN].numplayed;
	
	// Draw played cards
	Clear(CARDLEFT, CARDTOP+3*(cardHeight+1), 4*(CARDHALFWIDTH+1)+cardWidth, cardHeight);
	if (n == 0 && showpass)
		FrmShowObject(form, FrmGetObjectIndex(form, labelID_youpass));
	else if (n > 0)
	{
		for (i=0; i<n; i++)
		{
			DrawCard(form, 0, s.p[HUMAN].played[i], CARDLEFT+i*CARDHALFWIDTH, CARDTOP+3*(cardHeight+1), i==n-1?cardWidth:CARDHALFWIDTH+1, cardHeight, true);
		}
	}
}

static void DrawHumanPlayer(FormPtr form)
{
	short i, n=s.p[HUMAN].left;

	Clear(0, 160 - cardHeight, 160, cardHeight);
	for (i = 0; i < n + s.p[HUMAN].numplayed; i++)
		FrmHideObject(form, FrmGetObjectIndex(form, cardID(i)));
	for (i=0; i<n; i++)
	{
		DrawCard(form, i, s.p[HUMAN].card[i], 6*(13-n)+i*CARDHALFWIDTH, 160-cardHeight, i==n-1?cardWidth:CARDHALFWIDTH+1, cardHeight, false);
	}
	
	//DrawHumanPlayedCards(form);
}

static void DrawPalmPlayer(FormPtr form, short playerIndex, Boolean delay)
{
	short i, factor=4;
	short c=s.p[playerIndex].numplayed;
	
	if (playerIndex > 3 || playerIndex < 1)
		return;
	Clear(CARDLEFT, CARDTOP+(playerIndex-1)*(cardHeight+1), 4*(CARDHALFWIDTH+1)+cardWidth, cardHeight);
	if (delay)
	{
   	if (s.prefs.speed == SLOW)
   		factor = 2;
   	else if (s.prefs.speed == FAST)
   		factor = 8;
  		SysTaskDelay((Int32)(SysTicksPerSecond() / factor));
	}
	
	DrawField(form, fieldID_p2left + playerIndex - 1, "left: ", s.p[playerIndex].left);
	if (c == 0)
		FrmShowObject(form, FrmGetObjectIndex(form, passID(playerIndex)));
	else if (c > 0)
	{
		for (i=0; i<c; i++)
		{
			DrawCard(form, 0, s.p[playerIndex].played[i], CARDLEFT+i*CARDHALFWIDTH, CARDTOP+(playerIndex-1)*(cardHeight+1), i==c-1?cardWidth:CARDHALFWIDTH+1, cardHeight, true);
		}
	}
}

static void DrawField(FormPtr form, int field, char *title, short value)
{
	char 		v[4];
	FieldPtr ptr;

	StrCopy(leftfield[field-fieldID_p2left], title);
	if (value >= 0)
      StrIToA(v, (int)value);
   StrNCat(leftfield[field-fieldID_p2left], v, 9);
	ptr = FrmGetObjectPtr(form, FrmGetObjectIndex(form, field));
	FldEraseField(ptr);
	FldSetTextPtr(ptr, leftfield[field-fieldID_p2left]);
	FldDrawField(ptr);
}

// returns object ID of hit button
static UInt16 DisplayScoreForm(short* hand)
{
   FormPtr 	previousForm = FrmGetActiveForm();
   FormPtr 	frm = FrmInitForm(formID_score);
   UInt16    	hitButton;
	int 		i;
	FieldPtr ptr;
	char 		title[20];
    
   FrmSetActiveForm(frm);
    
	// Initialize any form objects in the form
	for (i=0; i<8; i++)
	{
		if (i < 4)
	   	StrIToA(scorefield[i], (int)hand[i]);
		else if (i < 8)
	   	StrIToA(scorefield[i], (int)s.p[i-4].score);
	   //else
	   //	StrIToA(scorefield[i], (int)s.p[i-8].games);
   	ptr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID_youhand+i));
   	FldSetTextPtr(ptr, scorefield[i]);
   }
	StrPrintF(title, "Score - round %d", s.round);
	FrmSetTitle(frm, title);

   hitButton = FrmDoDialog(frm);
    
   // read any values from the form objects here
   // before the form is deleted
    
   if (previousForm)
   	FrmSetActiveForm(previousForm);
   FrmDeleteForm(frm);
   return hitButton;
}

static UInt16 DisplayPrefsForm()
{
   FormPtr 		previousForm = FrmGetActiveForm();
   FormPtr 		frm = FrmInitForm(formID_prefs);
   UInt16    		hitButton;
   char			oldsort = s.prefs.sort;
   int			i;
   
   FrmSetActiveForm(frm);
    
	// Initialize any form objects in the form
   FrmSetControlGroupSelection(frm, 1, buttonID_medium + s.prefs.speed);
	FrmSetControlGroupSelection(frm, 2, buttonID_rank + s.prefs.sort);
   FrmSetControlValue(frm, FrmGetObjectIndex(frm, checkboxID_autopass), s.prefs.autopass);
   hitButton = FrmDoDialog(frm);
    
   // read any values from the form objects here
   // before the form is deleted
   if (FrmGetControlValue(frm, FrmGetObjectIndex(frm, buttonID_slow)) == 1)
   	s.prefs.speed = SLOW;
   else if (FrmGetControlValue(frm, FrmGetObjectIndex(frm, buttonID_medium)) == 1)
		s.prefs.speed = MEDIUM;
	else
		s.prefs.speed = FAST;
   
   if (FrmGetControlValue(frm, FrmGetObjectIndex(frm, buttonID_rank)) == 1)
		s.prefs.sort = RANK;
	else
		s.prefs.sort = SUIT;

	s.prefs.autopass = FrmGetControlValue(frm, FrmGetObjectIndex(frm, checkboxID_autopass));
   	
   if (previousForm)
   	FrmSetActiveForm(previousForm);
   if (s.prefs.sort != oldsort)
   {
		SwitchSortOrder(previousForm);
   }	
   FrmDeleteForm(frm);
   return hitButton;
}

static Boolean HighScore()
{
	short		i, lowest=4;
	
	// Determine lowest high score
	for (i=0; i<5; i++)
	{
		if (s.highscore[i] == 0)
		{
			lowest = (i-1 < 0) ? 0 : i-1;
			break;
		}
	}
	
	if (s.p[HUMAN].score > s.highscore[lowest])
		return true;
	return false;
}

static void InsertHighScore(const char* name)
{
	short i, j;
	
	for (i=0; i<5; i++)
	{
		if (s.p[HUMAN].score > s.highscore[i])
		{
			// shift high score down
			for (j=4; j>=i+1; j--)
			{
				StrNCopy(s.highname[j], s.highname[j-1], MAX_NAME+1);
				s.highscore[j] = s.highscore[j-1];
			}
			s.highscore[i] = s.p[HUMAN].score;
			StrNCopy(s.highname[i], name, MAX_NAME);
			s.highname[i][MAX_NAME] = '\0';
			break;
		}
	}
}

// returns true if it is a new high score
static Boolean DisplayGameOverForm()
{
   FormPtr 	previousForm = FrmGetActiveForm();
   FormPtr 	frm = FrmInitForm(formID_gameover);
   FieldPtr	fptr;
   Char		*cptr;
   Boolean	newhigh=HighScore();
    
   FrmSetActiveForm(frm);
    
	// Initialize any form objects in the form
	fptr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID_overallscore));
	StrIToA(overallscore, (int)s.p[HUMAN].score);
	FldSetTextPtr(fptr, overallscore);
	// Display message according to score
	if (s.p[HUMAN].score <= 0)
	{
		FrmHideObject(frm, FrmGetObjectIndex(frm, labelID_good1));
		FrmHideObject(frm, FrmGetObjectIndex(frm, labelID_good2));
		FrmHideObject(frm, FrmGetObjectIndex(frm, labelID_excellent1));
		FrmHideObject(frm, FrmGetObjectIndex(frm, labelID_excellent2));
	}
	else if (s.p[HUMAN].score < PRO_SCORE)
	{
		FrmHideObject(frm, FrmGetObjectIndex(frm, labelID_bad1));
		FrmHideObject(frm, FrmGetObjectIndex(frm, labelID_bad2));
		FrmHideObject(frm, FrmGetObjectIndex(frm, labelID_excellent1));
		FrmHideObject(frm, FrmGetObjectIndex(frm, labelID_excellent2));
	}
	else
	{
		FrmHideObject(frm, FrmGetObjectIndex(frm, labelID_bad1));
		FrmHideObject(frm, FrmGetObjectIndex(frm, labelID_bad2));
		FrmHideObject(frm, FrmGetObjectIndex(frm, labelID_good1));
		FrmHideObject(frm, FrmGetObjectIndex(frm, labelID_good2));
	}
	// If score is higher than lowest of high scores, prompt for name
	if (!newhigh)
	{
		FrmHideObject(frm, FrmGetObjectIndex(frm, labelID_newhigh));
		FrmHideObject(frm, FrmGetObjectIndex(frm, labelID_entername));
		FrmHideObject(frm, FrmGetObjectIndex(frm, fieldID_name));
	}
	else
	{
		FrmSetFocus(frm, FrmGetObjectIndex(frm, fieldID_name));
	}

   //hitButton = FrmDoDialog(frm);
   FrmDoDialog(frm);
 
   // read any values from the form objects here
   // before the form is deleted
   if (newhigh)
   {
		cptr = FldGetTextPtr(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID_name)));
		if (cptr == NULL)
		{
			InsertHighScore(" ");
		}
		else
		{
			InsertHighScore(cptr);
		}
	}
    
   if (previousForm)
   	FrmSetActiveForm(previousForm);
   FrmDeleteForm(frm);
   return newhigh;
}

// returns object ID of hit button
static UInt16 DisplayHighScoreForm()
{
   FormPtr 	previousForm = FrmGetActiveForm();
   FormPtr 	frm = FrmInitForm(formID_highscore);
   UInt16    	hitButton;
   FieldPtr	fptr;
   short		i;
    
   FrmSetActiveForm(frm);
    
	// Initialize any form objects in the form
	for (i=0; i<5; i++)
	{
		if (s.highscore[i] != INIT_SCORE)
		{
			fptr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID_name1+i));
			FldSetTextPtr(fptr, s.highname[i]);
   		fptr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID_score1+i));
   		StrIToA(highscore[i], (int)s.highscore[i]);
   		FldSetTextPtr(fptr, highscore[i]);
   	}
   	else
   	{
			fptr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID_name1+i));
			FldSetTextPtr(fptr, "-----");
   		fptr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID_score1+i));
   		FldSetTextPtr(fptr, "---");
   	}
	}

   hitButton = FrmDoDialog(frm);
    
   // read any values from the form objects here
   // before the form is deleted
    
   if (previousForm)
   	FrmSetActiveForm(previousForm);
   FrmDeleteForm(frm);
   return hitButton;
}


