/*
* Big Two - A popular poker game for Palm OS.
* Copyright (C) 2000  Woo Kok Tong
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
* Contact the author at onehy@magix.com.sg or visit http://come.to/palmos 
*/

#include "TypeDef.h"

//char startgame=FALSE;	// First play must include 3 of diamond
static char nearwin=FALSE;  	// Panic mode, play cautiously
static char lastcard=FALSE;	// Left one card, throw biggest card
char fivecard[4]={TRUE, FALSE, TRUE, FALSE};	// Play five cards or pair

void bzero(void *ptr, btI count)
{
	btI i;
	char *s = (char*)ptr;

	for (i = 0; i < count; i++)
		s[i] = 0;
}

void StartGame()
{
	btS pp;

	s.startgame = TRUE;
	nearwin = FALSE;
	lastcard = FALSE;
	// Decide who to start
	for (pp=0; pp<4; pp++)
	{
		if (s.p[pp].card[0] == 0)  // 3 of diamond
		{
			s.turn = pp;
			s.lastplayer = pp;
			break;
		}
	}
}

void PalmPlay()
{
	char played=TRUE;
	
	if (!nearwin)
	{
		if (s.p[0].left < 3 || s.p[1].left < 3 ||
			s.p[2].left < 3 || s.p[3].left < 3 )
		//if (s.p[HUMAN].left < 3)
			nearwin = TRUE;
	}
	if (!lastcard)
	{
		if (s.p[0].left == 1 || s.p[1].left == 1 ||
			s.p[2].left == 1 || s.p[3].left == 1 )
		//if (s.p[HUMAN].left < 3)
			lastcard = TRUE;
	}
	if (s.lastplayer == s.turn)  // this player has the say
	{
		if (s.p[HUMAN].left == 2)  // human has 2 cards left
		{
			if (!PlayFiveCards(s.turn) &&
				!PlayThreeCards(s.turn) &&
				!PlayOneCard(s.turn)	&&
				!PlayTwoCards(s.turn))
				played = FALSE;
		}
		else
		{
			// Play five card if possible
			if (!PlayFiveCards(s.turn) &&
				!PlayThreeCards(s.turn) &&
				!PlayTwoCards(s.turn)	&&
				!PlayOneCard(s.turn))
				played = FALSE;
		}
	}
	else  // follow what is being played
	{
		switch (s.p[s.lastplayer].numplayed)
		{
		case 5:
			if (!PlayFiveCards(s.turn)) played = FALSE;
			break;
		case 3:
			if (!PlayThreeCards(s.turn)) played = FALSE;
			break;
		case 2:
			if (!PlayTwoCards(s.turn)) played = FALSE;
			break;
		case 1:
			if (!PlayOneCard(s.turn)) played = FALSE;
			break;
		default:
#ifdef WIN32
			printf("Error at PalmPlay\n");
#endif
			return;
		}
	}

	if (!played) s.p[s.turn].numplayed = 0;
}

void UpdateScore(btS* hand)
{
	btS i, j, count, total=0, winner;
	
	for (i=0; i<4; i++)
	{
		count = 0;
		for (j=0; j<s.p[i].left; j++)
		{
			count += (Rank(s.p[i].card[j]) + 1);
		}
		if (count == 0) winner = i;
		// If 10 or more cards left, double the score
		if (s.p[i].left >= 10)
			count = 2 * count;
		hand[i] = 0 - count;
		s.p[i].score += hand[i];
		total += count;
	}
	// Add total to player who has won
	s.p[winner].score += total;
	hand[winner] = total;
}

// Update state info, playedcards & s.p.card need not be sorted similarly
void UpdateState(const char* playedcards, const btS count)
{
	btS i, j, k;
	
	s.lastplayer = s.turn;
	//for (i=0; i<count; i++)
	//	s.p[s.turn].played[i] = playedcards[i];
	i = 0;
	j = 0;
	while (i < s.p[s.turn].left && j < count)
	{
		if (s.p[s.turn].card[i] == playedcards[j])
		{
			s.p[s.turn].played[j] = playedcards[j];
			// Remove card and move others forward
			for (k=i; k<s.p[s.turn].left-1; k++)
			{
				s.p[s.turn].card[k] = s.p[s.turn].card[k+1];
			}
			s.p[s.turn].left --;
			j++;
			i = 0;
		}
		else
			i++;
	}
	s.p[s.turn].numplayed = j;
	SortCards(s.p[s.turn].played, j, TRUE);
}

// Return TRUE if five cards is played
char PlayFiveCards(const btS playerIndex)
{
	char tmp[5], ca[13], ca2[13];
	btS i, count=s.p[playerIndex].left, count2=s.p[playerIndex].left;
	
	if (s.p[playerIndex].left < 5) return FALSE;  // less than 5 cards left
	
	for (i=0; i<count; i++)
	{
		ca[i] = s.p[playerIndex].card[i];
		ca2[i] = s.p[playerIndex].card[i];
	}
	if (!fivecard[playerIndex])
		StripPair(ca, &count);
	StripFiveCard(ca2, &count2, fivecard[playerIndex], TRUE);
	if (s.lastplayer == s.turn)  // this player has the say
	{
		i = s.p[playerIndex].left < 8 ? s.p[playerIndex].left : 8;
			// Look for straight flush
		if (!LookForStraight(s.p[playerIndex].card, tmp,	
			s.p[playerIndex].left, TRUE, TRUE) &&
			// Look for straight in first 8 cards
			//!LookForStraight(s.p[playerIndex].card, tmp, 
			//i, FALSE, TRUE) &&
			// Look for straight
			!LookForStraight(s.p[playerIndex].card, tmp, 
			s.p[playerIndex].left, FALSE, TRUE) &&
			// Look for four of a kind
			!LookForFourOfAKind(s.p[playerIndex].card, tmp,
				s.p[playerIndex].left, TRUE) &&
			// Look for fullhouse
			!LookForFullHouse(s.p[playerIndex].card, tmp,
			s.p[playerIndex].left, TRUE) &&
			// Look for flush
			!LookForFlush(ca, tmp, count, TRUE))
				return FALSE;
	}
	else  // follow what is being played
	{
		i = s.p[playerIndex].left < 8 ? s.p[playerIndex].left : 8;
		switch (s.fivecardtype)
		{
		case LOWSTRAIGHT:
		case STRAIGHT:
			//if (LookForStraight(s.p[playerIndex].card, tmp,	
			//	i, FALSE, TRUE))
			//	break;
			if (LookForStraight(s.p[playerIndex].card, tmp,	
				s.p[playerIndex].left, FALSE, TRUE))
				break;
			if (LookForStraight(s.p[playerIndex].card, tmp,	
				s.p[playerIndex].left, TRUE, TRUE))
				break;
			if (LookForFourOfAKind(s.p[playerIndex].card, tmp,
				s.p[playerIndex].left, TRUE))
				break;
			if (LookForFullHouse(s.p[playerIndex].card, tmp,
				s.p[playerIndex].left, TRUE))
				break;
			if (LookForFlush(ca, tmp,	count, TRUE))
				break;
			return FALSE;
		case FLUSH:
			if (LookForStraight(s.p[playerIndex].card, tmp,	
				s.p[playerIndex].left, TRUE, TRUE))
				break;
			if (LookForFourOfAKind(ca2, tmp,	count2, TRUE))
				break;
			if (LookForFullHouse(ca2, tmp, count2, TRUE))
				break;
			if (LookForFlush(ca2, tmp,	count2, TRUE))
				break;
			if (LookForFourOfAKind(s.p[playerIndex].card, tmp,
				s.p[playerIndex].left, TRUE))
				break;
			if (LookForFullHouse(s.p[playerIndex].card, tmp,
				s.p[playerIndex].left, TRUE))
				break;
			if (LookForFlush(ca, tmp,	count, TRUE))
				break;
			return FALSE;
		case FULLHOUSE:
			if (LookForStraight(s.p[playerIndex].card, tmp,	
				s.p[playerIndex].left, TRUE, TRUE))
				break;
			if (LookForFourOfAKind(ca2, tmp,	count2, TRUE))
				break;
			if (LookForFullHouse(ca2, tmp, count2, TRUE))
				break;
			if (LookForFourOfAKind(s.p[playerIndex].card, tmp,
				s.p[playerIndex].left, TRUE))
				break;
			if (LookForFullHouse(s.p[playerIndex].card, tmp,
				s.p[playerIndex].left, TRUE))
				break;
			return FALSE;
		case FOUROFAKIND:
			if (LookForStraight(s.p[playerIndex].card, tmp,	
				s.p[playerIndex].left, TRUE, TRUE))
				break;
			if (LookForFourOfAKind(ca2, tmp,	count2, TRUE))
				break;
			if (LookForFourOfAKind(s.p[playerIndex].card, tmp,
				s.p[playerIndex].left, TRUE))
				break;
			return FALSE;
		case LOWSTRAIGHTFLUSH:
		case STRAIGHTFLUSH:
			if (LookForStraight(s.p[playerIndex].card, tmp,	
				s.p[playerIndex].left, TRUE, TRUE))
				break;
			else
				return FALSE;
		default:
			// This is an error
#ifdef WIN32
			printf("Error at PlayFiveCards\n");
#endif
			return FALSE;
		}
	}
	
	// Update state info
	UpdateState(tmp, 5);
	return TRUE;
}

// Return TRUE if three cards is played
char PlayThreeCards(const btS playerIndex)
{
	char found=FALSE, tmp[13];
	btS i, j, count;

	if (s.p[playerIndex].left < 3) return FALSE;  // less than 3 cards left
	
	// Don't play threes of 2
	for (i=0; i<s.p[playerIndex].left; i++)
		tmp[i] = s.p[playerIndex].card[i];
	count = s.p[playerIndex].left;
	if (s.p[playerIndex].left > 4)
		StripRank(tmp, &count, 12);

	j = 0;
	while (j < count-2)
	{
		if (Rank(tmp[j]) == Rank(tmp[j+2]) && Rank(tmp[j]) == Rank(tmp[j+1]))
		{
			if (!VerifyPlay(&tmp[j], 3))
			{
				j += 3;
				continue;
			}
			else
			{
				found = TRUE;
				break;
			}
		}
		j++;
	}
	
	if (!found) return FALSE;
	// Update state info
	UpdateState(&tmp[j], 3);
	return TRUE;
}

// Return TRUE if two cards is played
char PlayTwoCards(const btS playerIndex)
{
	char found=FALSE, tmp[13];
	btS i, j, count=s.p[s.turn].left;

	if (s.p[playerIndex].left < 2) return FALSE;  // less than 2 cards left
	
	for (i=0; i<s.p[s.turn].left; i++)
		tmp[i] = s.p[s.turn].card[i];

	if (!nearwin) 
	{
		// Don't play pair of 2
		if (s.p[playerIndex].left > 2)
			StripRank(tmp, &count, 12);
		// Strip five card play
		//if (s.turn != s.lastplayer && fivecard[playerIndex])
		if (s.turn != s.lastplayer)
			StripFiveCard(tmp, &count, fivecard[playerIndex], FALSE);

		j = 0;
		while (j < count-1)
		{
			// Assume sorted
			if (Rank(tmp[j]) == Rank(tmp[j+1]))
			{
				if (!VerifyPlay(&(tmp[j]), 2))
				{
					j += 2;
					continue;
				}
				else
				{
					found = TRUE;
					// Reset card index
					/*
					for (i=0; i<s.p[playerIndex].left; i++)
					{
						if (s.p[s.turn].card[i] == tmp[j])
						{
							j = i;
							break;
						}
					}
					*/
					break;
				}
			}
			else
				j++;
		}
	}
	else  // near win
	{
		// Throw biggest pair
		j = count-2;
		while (j >= 0)
		{
			// Assume sorted
			if (Rank(tmp[j]) == Rank(tmp[j+1]))
			{
				if (!VerifyPlay(&(tmp[j]), 2))
				{
					found = FALSE;
					break;
					//j -= 2;
					//continue;
				}
				else
				{
					found = TRUE;
					break;
				}
			}
			else
				j--;
		}
	}
	
	if (!found) return FALSE;
	// Update state info
	UpdateState(&(tmp[j]), 2);
	return TRUE;
}

// Return largest cards
char LargestCard(const char* card, const btS count)
{
	btS i, largest=0;
	//char largest=card[0];

	for (i=1; i<count; i++)
	{
		if (Larger(card[i], card[largest]))
			largest = i;
	}
	return card[largest];
}

// Return TRUE if first card is higher than second card
char Larger(const char first, const char second)
{
	if (Rank(first) > Rank(second) ||
		(Rank(first) == Rank(second) && Suit(first) > Suit(second)))
		return TRUE;
	
	return FALSE;
}

// Return TRUE if first card is higher than second card
char AltLarger(const char first, const char second)
{
	if (AltRank(first) > AltRank(second) ||
		(AltRank(first) == AltRank(second) && Suit(first) > Suit(second)))
		return TRUE;
	
	return FALSE;
}

// Return TRUE if one card is played
char PlayOneCard(const btS playerIndex)
{
	char tmp[13];
	btS i, j=0, count;

	if (s.p[playerIndex].left < 1) return FALSE;  // less than 1 card left
	
	if (s.p[playerIndex].left <= 2 && 
		Larger(s.p[playerIndex].card[s.p[playerIndex].left - 1], LargestCard(s.p[Next(playerIndex)].card, s.p[Next(playerIndex)].left)) &&
		Larger(s.p[playerIndex].card[s.p[playerIndex].left - 1], LargestCard(s.p[Next(playerIndex+1)].card, s.p[Next(playerIndex+1)].left)) &&
		Larger(s.p[playerIndex].card[s.p[playerIndex].left - 1], LargestCard(s.p[Next(playerIndex+2)].card, s.p[Next(playerIndex+2)].left)))
	{
		if (PlayBiggestCard(playerIndex))
		{
			// Update state info
			UpdateState(&(s.p[playerIndex].card[s.p[playerIndex].left - 1]), 1);
			return TRUE;
		}
	}

	if (!lastcard)
	{
		// cheat a bit, let other computer players have the say
		if (s.lastplayer != HUMAN && s.turn != s.lastplayer && 
			Rank(s.p[s.lastplayer].played[0]) > 9)
			return FALSE;
   	
		for (i=0; i<s.p[playerIndex].left; i++)
			tmp[i] = s.p[playerIndex].card[i];
		count = s.p[playerIndex].left;
		// Strip five card play
		if (s.turn != s.lastplayer && s.p[HUMAN].left > 3)
		{
			StripFiveCard(tmp, &count, fivecard[playerIndex], FALSE);
		//if (!fivecard[playerIndex])
			StripPair(tmp, &count);
		}

		if (PlaySmallestCard(tmp, count, &j))
		{
			// Reset card index
			for (i=0; i<s.p[playerIndex].left; i++)
			{
				if (s.p[s.turn].card[i] == tmp[j])
				{
      			// Update state info
      			UpdateState(&(s.p[s.turn].card[i]), 1);
					break;
				}
			}
			return TRUE;
		}
  		else
  			return FALSE;
	}
	else  // Last card
	{
		// If next player has 1 card left, throw biggest card
		i = Next(playerIndex);
		if (s.p[i].left == 1 || 
			(s.p[HUMAN].left == 1 && Larger(s.p[HUMAN].card[0], s.p[i].card[s.p[i].left-1])))
		{
			if (PlayBiggestCard(playerIndex))
			{
				// Update state info
				UpdateState(&(s.p[s.turn].card[s.p[s.turn].left - 1]), 1);
				return TRUE;
			}
			else
				return FALSE;
		}
		else  // throw smallest card
		{
			if (PlaySmallestCard(s.p[playerIndex].card, s.p[playerIndex].left, &j))
			{
				// Update state info
				UpdateState(&(s.p[s.turn].card[j]), 1);
				return TRUE;
			}
			else
				return FALSE;
		}
	}

	return FALSE;
}

char PlaySmallestCard(const char* tmp, const btS count, btS* index)
{
	btS j = 0;
	char found=FALSE;
	
	while (j < count)
	{
		if (!VerifyPlay(&(tmp[j]), 1))
		{
			j ++;
			continue;
		}
		else
		{
			found = TRUE;
			*index = j;
			break;
		}
	}  // while (j<count)
	
	return found;
}

char PlayBiggestCard(const btS playerIndex)
{
	if (!VerifyPlay(&(s.p[playerIndex].card[s.p[playerIndex].left - 1]), 1))
	{
		return FALSE;
	}
	else
	{
		return TRUE;
	}
}

// Return the no. of pair, threes and fours
void CountPair(const char* card, const btS count, btS* paircount,
					btS* threescount, btS* fourscount)
{
	btS i;

	*paircount = *threescount = *fourscount = 0;
	i = 0;
	while (i < count-1)
	{
		if (Rank(card[i]) == Rank(card[i+1]))
		{
			if (i+2 < count && Rank(card[i]) == Rank(card[i+2]))
			{
				if (i+3 < count && Rank(card[i]) == Rank(card[i+3]))
				{
					(*fourscount) ++;
					i += 4;
				}
				else
				{
					(*threescount) ++;
					i += 3;
				}
			}
			else
			{
				(*paircount) ++;
				i += 2;
			}
		}
		else
			i++;
	}
}

// Return TRUE if straight is good, card & straight must be similarly sorted
char EvaluateStraight(const char* card, const btS count, const char* straight, 
							 const char fivecard)
{
	btS i, first=0, last=0;
	char low=FALSE;
	btS pair, threes, fours;

	// It is a low straight?
	if (Rank(straight[0]) >= 11)
		low = TRUE;
	for (i=0; i<count; i++)
	{
		if (!low && Rank(card[i]) == Rank(straight[0]))
		{
			first = i;
			break;
		}
		else if (low && Rank(card[i]) == 0)
		{
			first = i;
			break;
		}
	}
	for (i=count-1; i>first; i--)
	{
		if (Rank(card[i]) == Rank(straight[4]))
		{
			last = i;
			break;
		}
	}
	CountPair(&card[first], (btS)(last-first+1), &pair, &threes, &fours);
	if (!fivecard)
	{
		if (!low)
		{
			if (Rank(straight[0]) < 2 && 
				(pair >= 3 || threes >= 2 || fours >= 1 || pair+threes+fours >= 2))
				return FALSE;
			// There are 2 pairs or more
			if (pair >= 2 || threes >= 2 || fours >= 1 || pair+threes+fours >= 2) 
				return FALSE;
			if (Rank(straight[0]) > 5 && (pair >= 1 || threes >= 1))
				return FALSE;
		}
		else
		{
			if (pair >= 2 || threes >= 1 || fours >= 1)
				return FALSE;
		}
	}
	else
	{
		// There are 3 pairs or more
		if (!low && (pair >= 3 || threes >= 2 || fours >= 1 || pair+threes+fours >= 2)) 
			return FALSE;
		if (low && (pair >= 2 || threes >= 1 || fours >= 1)) return FALSE;
	}

	return TRUE;
}

// Return TRUE if straight is found & return straight (sorted by rank) by card
char FindStraight(char* card, const btS count, 
						const char flush, const char play, const char altsort)
{
	btS i, j, a=count;
	char tmp[13];

	if (count < 5 || count > 13) return FALSE;

	if (altsort)
	{
		if (!flush)
		{
			AltSortCards(card, a, TRUE);
			for (i=0; i<a; i++)
				tmp[i] = card[i];
			StripDuplicateRank(tmp, &a);
		}
		else  // for straight flush, sort by suit
		{
			AltSortCards(card, a, FALSE);
			for (i=0; i<a; i++)
				tmp[i] = card[i];
		}
	}
	else
	{
		for (i=0; i<a; i++)
			tmp[i] = card[i];
		if (!flush)
			StripDuplicateRank(tmp, &a);
		else  // for straight flush, sort by suit
			SortCards(tmp, a, FALSE);
	}
	i = 0;
	while (i < a-4)
	{
		if (!altsort)
		{
			if (Rank(tmp[i+1]) != Rank(tmp[i])+1)
			{
				i++;
				continue;
			}
			if (Rank(tmp[i+2]) != Rank(tmp[i+1])+1)
			{
				i += 2;
				continue;
			}
			if (Rank(tmp[i+3]) != Rank(tmp[i+2])+1)
			{
				i += 3;
				continue;
			}
			// Last card cannot be 2
			if (Rank(tmp[i+4]) != Rank(tmp[i+3])+1 || Rank(tmp[i+4]) == 12)
			{
				i += 4;
				continue;
			}
		}
		else  // alt sort
		{
			if (AltRank(tmp[i+1]) != AltRank(tmp[i])+1)
			{
				i++;
				continue;
			}
			if (AltRank(tmp[i+2]) != AltRank(tmp[i+1])+1)
			{
				i += 2;
				continue;
			}
			if (AltRank(tmp[i+3]) != AltRank(tmp[i+2])+1)
			{
				i += 3;
				continue;
			}
			if (AltRank(tmp[i+4]) != AltRank(tmp[i+3])+1)
			{
				i += 4;
				continue;
			}
		}

		// Found a straight, is it a straight flush?
		if (Suit(tmp[i+4]) == Suit(tmp[i+3]) &&
			Suit(tmp[i+3]) == Suit(tmp[i+2]) &&
			Suit(tmp[i+2]) == Suit(tmp[i+1]) &&
			Suit(tmp[i+1]) == Suit(tmp[i]))
		{
			if (altsort)
				SortCards(&tmp[i], 5, TRUE);
			if (play && !VerifyPlay(&tmp[i], 5))
			{
				// Continue to look for straight
				if (altsort)
					AltSortCards(&tmp[i], 5, (char)!flush);
				i++;
				continue;
			}
			else
			{
				for (j=0; j<5; j++)
					card[j] = tmp[i+j];
				return TRUE;
			}
		}
		else  // Not straight flush
		{
			if (!EvaluateStraight(card, count, &tmp[i], fivecard[s.turn]))
			{
				i++;
				continue;
			}
			if (!flush)
			{
				if (altsort)
					SortCards(&tmp[i], 5, TRUE);
				if (play && !VerifyPlay(&tmp[i], 5))
				{
					// Continue to look for straight
					if (altsort)
						AltSortCards(&tmp[i], 5, (char)!flush);
					i++;
					continue;
				}
				else
				{
					for (j=0; j<5; j++)
						card[j] = tmp[i+j];
					return TRUE;
				}
			}
			else
			{
				i++;
				continue;
			}
		}

	}  // while loop

	return FALSE;
}

// Return TRUE if straight is found and return the 
// cards that form the straight
char LookForStraight(const char* card, char* straight, 
							const btS count, const char flush, const char play)
{
	char tmp[13];
	btS i;
	
	if (count < 5 || count > 13) return FALSE;
	//if (count > 13) count = 13;

	for (i=0; i<count; i++)
		tmp[i] = card[i];
	
	// Find straight first
	if (!FindStraight(tmp, count, flush, play, FALSE) &&
		 !FindStraight(tmp, count, flush, play, TRUE))
		return FALSE;
	
	for (i=0; i<5; i++)
		straight[i] = tmp[i];
	return TRUE;
}

// Return TRUE if flush is found and return the 
// cards that form the flush
char LookForFlush(const char* card, char* flush, const btS count,
						const char play)
{
	char tmp[13], found=FALSE;
	btS a=count, i, j;
	
	if (count < 5 || count > 13) return FALSE;
	//if (count > 13) count = 13;

	for (i=0; i<a; i++)
		tmp[i] = card[i];

	// Sort card by suit
	SortCards(tmp, a, FALSE);	
	
	i = 0;
	while (i < a-4)
	{
		if (Suit(tmp[i+4]) != Suit(tmp[i]))
		{
			i++;
			continue;
		}

		// Found a flush, verify play
		if (play && !VerifyPlay(&tmp[i], 5))
		{
			// Continue to look for flush
			i++;
			continue;
		}
		else
		{
			found = TRUE;
			for (j=0; j<5; j++)
				flush[j] = tmp[i+j];
			break;
		}
	}

	if (!found) return FALSE;
	return TRUE;
}

// Return TRUE if fullhouse is found and return the 
// cards that form the fullhouse
char LookForFullHouse(const char* card, char* fullhouse, const btS count,
							 const char play)
{
	char tmp[13], pair=FALSE, threes=FALSE;
	btS a=count, i, j, pi, ti;
	
	if (count < 5 || count > 13) return FALSE;
	//if (count > 13) count = 13;

	for (i=0; i<a; i++)
		tmp[i] = card[i];
	
	// Don't play threes of Ace & 2
	if (!nearwin) 
	{
		StripRank(tmp, &a, 12);
		StripRank(tmp, &a, 11);
	}

	i = 0;
	while (i < a-1)
	{
		if (Rank(tmp[i+1]) != Rank(tmp[i]))
		{
			i++;
			continue;
		}
		if (!pair && !threes)  // pair & threes not found yet
		{
			if (i > a-5) break;
			if (Rank(tmp[i+2]) != Rank(tmp[i]))
			{
				pair = TRUE;
				pi = i;
				i += 2;
				continue;
			}
			else
			{
				threes = TRUE;
				ti = i;
				i += 3;
				continue;
			}
		}
		else if (!threes) // three is not found, pair is found
		{
			if (i > a-3) break;
			if (Rank(tmp[i+2]) != Rank(tmp[i]))
			{
				i += 2;
				continue;
			}
		}

		// Found a fullhouse, verify play
		if (!pair)
		{
			pair = TRUE;
			pi = i;
		}
		else
		{
			threes = TRUE;
			ti = i;
		}
		if (pi < ti)
		{
			for (j=0; j<2; j++)
				fullhouse[j] = tmp[pi+j];
			for (j=0; j<3; j++)
				fullhouse[j+2] = tmp[ti+j];
		}
		else
		{
			for (j=0; j<3; j++)
				fullhouse[j] = tmp[ti+j];
			for (j=0; j<2; j++)
				fullhouse[j+3] = tmp[pi+j];
		}
		if (play && !VerifyPlay(fullhouse, 5))
		{
			// Continue to look for fullhouse
			if (pi > ti)
			{
				// Use the threes as pair
				j = pi;
				pi = ti;
				threes = FALSE;
				i = j;
			}
			else
			{
				threes = FALSE;
				i += 3;
			}
			continue;
		}
		else
		{
			break;
		}

	}  // while (i < a-1)

	if (!pair || !threes) return FALSE;
	return TRUE;
}

// Return TRUE if fourofakind is found and return the 
// cards that form the fourofakind
char LookForFourOfAKind(const char* card, char* fourofakind, 
								const btS count, const char play)
{
	char tmp[14], found=FALSE;
	btS a=count, i, j;
	
	if (count < 5 || count > 13) return FALSE;
	//if (count > 13) count = 13;

	for (i=0; i<a; i++)
		tmp[i] = card[i];
	
	if (!nearwin)
	{
		// Don't play fours of 2 & Ace
		StripRank(tmp, &a, 12);
		StripRank(tmp, &a, 11);
	}

	i = 0;
	while (i < a-3)
	{
		if (Rank(tmp[i+3]) != Rank(tmp[i]))
		{
			i++;
			continue;
		}

		// Found a fourofakind, verify play
		if (i == 0)
		{
			fourofakind[4] = tmp[4];
			for (j=0; j<4; j++)
				fourofakind[j] = tmp[i+j];
		}
		else
		{
			fourofakind[0] = tmp[0];
			for (j=0; j<4; j++)
				fourofakind[j+1] = tmp[i+j];
		}
		if (play && !VerifyPlay(&fourofakind[0], 5))
		{
			// Continue to look for fourofakind
			i += 4;
			continue;
		}
		else
		{
			found = TRUE;
			break;
		}
	}

	if (!found) return FALSE;
	return TRUE;
}

// Return stripped cards and number of cards after stripping
void StripPair(char* card, btS* count)
{
	btS i=0, j;
	
	while (i < (*count)-1)
	{
		// Do not strip King, Ace and 2 pairs
		if (Rank(card[i]) > 9)
		{
			i++;
			continue;
		}

		if (Rank(card[i]) == Rank(card[i+1]))
		{
			if (i < (*count-2) && Rank(card[i]) == Rank(card[i+2]))
			{
				if (i < (*count-3) && Rank(card[i]) == Rank(card[i+3]))
				{
					// strip fours
					for (j=i; j<(*count)-1; j++)
					{
						card[j] = card[j+4];
					}
					(*count) -= 4;
				}
				else
				{
					// strip threes
					for (j=i; j<(*count)-1; j++)
					{
						card[j] = card[j+3];
					}
					(*count) -= 3;
				}
			}
			else 
			{
				// strip pair
				for (j=i; j<(*count)-1; j++)
				{
					card[j] = card[j+2];
				}
				(*count) -= 2;
			}
		}
		else
			i++;
	}
}

// Return stripped cards and number of cards after stripping
void StripDuplicateRank(char* card, btS* count)
{
	btS i=0, j;
	
	while (i < (*count)-1)
	{
		if (Rank(card[i]) == Rank(card[i+1]))
		{
			// strip the larger card
			for (j=i+1; j<(*count)-1; j++)
			{
				card[j] = card[j+1];
			}
			(*count)--;
		}
		else
			i++;
	}
}

// Return stripped cards and number of cards after stripping
void StripRank(char* card, btS* count, btS rank)
{
	btS i=0, j;
	
	while (i < *count)
	{
		if (Rank(card[i]) == rank)
		{
			// strip the card
			for (j=i; j<(*count)-1; j++)
			{
				card[j] = card[j+1];
			}
			(*count)--;
		}
		else
			i++;
	}
}

// Return stripped cards and number of cards after stripping
void StripFiveCard(char* card, btS* count, const char fivecard, const char straightonly)
{
	char tmp[5];
	btS i=0, j, k;
	
	while (1)
	{
		if (*count < 5) return;
		//i = *count < 8 ? *count : 8;
		if (!fivecard || straightonly)
		{
			// Look for straight only
			if (!LookForStraight(card, tmp, *count, TRUE, FALSE) &&
				!LookForStraight(card, tmp, *count, FALSE, FALSE))
				return ;
		}
		else
		{
			if (!LookForStraight(card, tmp, *count, TRUE, FALSE) &&
				//!LookForStraight(card, tmp, i, FALSE, FALSE) &&
				!LookForStraight(card, tmp, *count, FALSE, FALSE) &&
				!LookForFourOfAKind(card, tmp, *count, FALSE) &&
				!LookForFullHouse(card, tmp, *count, FALSE) &&
				!LookForFlush(card, tmp, *count, FALSE))
				return ;
		}

		k = 0;
		i = 0;
		while (i < *count && k < 5)
		{
			if (card[i] == tmp[k])
			{
				// strip the card
				for (j=i; j<(*count)-1; j++)
				{
					card[j] = card[j+1];
				}
				(*count)--;
				k++;
			}
			else
				i++;
		}
	}  // while (1)
}

// Verify if play is valid, card & lastplayer.played must be sorted by rank
char VerifyPlay(const char* card, const btS count)
{
	char newround=FALSE, type, fh1, fh2;
	char tmp1[5], tmp2[5];
	btS i;

	if (s.turn == s.lastplayer)  // this player has the say
		newround = TRUE;
	// Number of cards played is not followed
	if (!newround && count != s.p[s.lastplayer].numplayed)
		return FALSE;
	// Number of cards played is not valid
	if (count > 5 || count < 0 || count == 4)
		return FALSE;
	// Didn't start with 3 of diamond
	if (s.startgame && card[0] != 0)
		return FALSE;


	switch (count)
	{
	case 1:
		// Single card
   	if (s.startgame)
   	{
   		s.startgame = FALSE;
   		return TRUE;
   	}
   	if (newround) return TRUE;
		if (Larger(card[0], s.p[s.lastplayer].played[0]))
			return TRUE;
		else
			return FALSE;
		break;
	
	case 2:
		// Pair
		if (Rank(card[0]) != Rank(card[1]))
			return FALSE;
		if (s.startgame)
		{
			s.startgame = FALSE;
			return TRUE;
		}
		if (newround) return TRUE;
		if (Larger(card[1], s.p[s.lastplayer].played[1]))
			return TRUE;
		else
			return FALSE;
		break;

	case 3:
		// Three card
		if (Rank(card[0]) != Rank(card[1]) || 
			Rank(card[1]) != Rank(card[2]))
			return FALSE;
		if (s.startgame)
		{
			s.startgame = FALSE;
			return TRUE;
		}
		if (newround) return TRUE;
		if (Rank(card[0]) < Rank(s.p[s.lastplayer].played[0]))
			return FALSE;
		else
			return TRUE;
		break;

	case 5:
		// Five card
		VerifyFiveCards(card, &type);
		if (type == NOTFIVECARD) return FALSE;
		if (s.startgame)
		{
			s.startgame = FALSE;
			s.fivecardtype = type;
			return TRUE;
		}
		if (newround)
		{
			s.fivecardtype = type;
			return TRUE;
		}
		else
		{
			switch (s.fivecardtype)
			{
			case LOWSTRAIGHT:
				if (type > LOWSTRAIGHT)
				{
					s.fivecardtype = type;
					return TRUE;
				}
				if (type == LOWSTRAIGHT)
				{
					for (i=0; i<5; i++)
					{
						tmp1[i] = card[i];
						tmp2[i] = s.p[s.lastplayer].played[i];
					}
					AltSortCards(tmp1, 5, TRUE);
					AltSortCards(tmp2, 5, TRUE);
					if (AltLarger(tmp1[4], tmp2[4]))
					{
						s.fivecardtype = type;
						return TRUE;
					}
					else
						return FALSE;
				}
				return FALSE;
				break;
			case STRAIGHT:
				if (type > STRAIGHT)
				{
					s.fivecardtype = type;
					return TRUE;
				}
				if (type == STRAIGHT)
				{
					if (Larger(card[4], s.p[s.lastplayer].played[4]))
					{
						s.fivecardtype = type;
						return TRUE;
					}
					else
						return FALSE;
				}
				return FALSE;
				break;
			case FLUSH:
				if (type > FLUSH)
				{
					s.fivecardtype = type;
					return TRUE;
				}
				if (type == FLUSH)
				{
					if (Larger(card[4], s.p[s.lastplayer].played[4]))
					{
						s.fivecardtype = type;
						return TRUE;
					}
					else
						return FALSE;
				}
				return FALSE;
				break;
			case LOWSTRAIGHTFLUSH:
				if (type > LOWSTRAIGHTFLUSH)
				{
					s.fivecardtype = type;
					return TRUE;
				}
				if (type == LOWSTRAIGHTFLUSH)
				{
					for (i=0; i<5; i++)
					{
						tmp1[i] = card[i];
						tmp2[i] = s.p[s.lastplayer].played[i];
					}
					AltSortCards(tmp1, 5, TRUE);
					AltSortCards(tmp2, 5, TRUE);
					if (AltLarger(tmp1[4], tmp2[4]))
					{
						s.fivecardtype = type;
						return TRUE;
					}
					else
						return FALSE;
				}
				return FALSE;
				break;
			case STRAIGHTFLUSH:
				if (type == STRAIGHTFLUSH)
				{
					if (Larger(card[4], s.p[s.lastplayer].played[4]))
					{
						s.fivecardtype = type;
						return TRUE;
					}
					else
						return FALSE;
				}
				return FALSE;
				break;
			case FULLHOUSE:
				if (type > FULLHOUSE)
				{
					s.fivecardtype = type;
					return TRUE;
				}
				else if (type == FULLHOUSE)
				{
					// Determine which is the threes
					if (Rank(card[1]) == Rank(card[2]))
						fh1 = card[2];
					else
						fh1 = card[4];
					if (Rank(s.p[s.lastplayer].played[1]) == 
						Rank(s.p[s.lastplayer].played[2]))
						fh2 = s.p[s.lastplayer].played[2];
					else
						fh2 = s.p[s.lastplayer].played[4];
					if (Rank(fh1) < Rank(fh2))
						return FALSE;
					else
					{
						s.fivecardtype = type;
						return TRUE;
					}
				}
				return FALSE;
				break;
			case FOUROFAKIND:
				if (type > FOUROFAKIND)
				{
					s.fivecardtype = type;
					return TRUE;
				}
				if (type == FOUROFAKIND)
				{
					// Determine which is the fours
					if (Rank(card[0]) == Rank(card[1]))
						fh1 = card[0];
					else
						fh1 = card[4];
					if (Rank(s.p[s.lastplayer].played[0]) == 
						Rank(s.p[s.lastplayer].played[1]))
						fh2 = s.p[s.lastplayer].played[0];
					else
						fh2 = s.p[s.lastplayer].played[4];
					if (Rank(fh1) < Rank(fh2))
						return FALSE;
					else
					{
						s.fivecardtype = type;
						return TRUE;
					}
				}
				return FALSE;
				break;
			default:
				// This is an error
#ifdef WIN32
				printf("Error at VerifyPlay\n");
#endif
				return FALSE;
			}
		}
		break;
	}

	return FALSE;
}

// Verifty five card play, return thru type, card must be sorted by rank
void VerifyFiveCards(const char* card, char* type)
{
	char tmp[5];
	btS i;

	*type = NOTFIVECARD;
	// Full house
	if (Rank(card[0]) == Rank(card[1]) && 
		Rank(card[3]) == Rank(card[4]))
	{
		if (Rank(card[2]) == Rank(card[0]) || 
			Rank(card[2]) == Rank(card[4]))
		{
			*type = FULLHOUSE;
			return;
		}
		else
			return;
	}
	
	// Four of a kind
	if (Rank(card[1]) == Rank(card[2]))
	{
		if ((Rank(card[0]) == Rank(card[1]) &&
			Rank(card[2]) == Rank(card[3])) ||
			(Rank(card[2]) == Rank(card[3]) &&
			Rank(card[3]) == Rank(card[4])))
		{
			*type = FOUROFAKIND;
			return;
		}
		else
			return;
	}

	// Straight or Straigh flush
	if (Rank(card[4]) != 12 &&
		Rank(card[4]) == Rank(card[3])+1 &&
		Rank(card[3]) == Rank(card[2])+1 &&
		Rank(card[2]) == Rank(card[1])+1 &&
		Rank(card[1]) == Rank(card[0])+1)
	{
		if (Suit(card[0]) == Suit(card[1]) &&
			Suit(card[1]) == Suit(card[2]) &&
			Suit(card[2]) == Suit(card[3]) &&
			Suit(card[3]) == Suit(card[4]))
		{
			*type = STRAIGHTFLUSH;
			return;
		}
		else
		{
			*type = STRAIGHT;
			return;
		}
	}

	// Low Straight or Low Straigh flush
	for (i=0; i<5; i++)
		tmp[i] = card[i];
	AltSortCards(tmp, 5, TRUE);
	if (AltRank(tmp[4]) == AltRank(tmp[3])+1 &&
		AltRank(tmp[3]) == AltRank(tmp[2])+1 &&
		AltRank(tmp[2]) == AltRank(tmp[1])+1 &&
		AltRank(tmp[1]) == AltRank(tmp[0])+1)
	{
		if (Suit(tmp[0]) == Suit(tmp[1]) &&
			Suit(tmp[1]) == Suit(tmp[2]) &&
			Suit(tmp[2]) == Suit(tmp[3]) &&
			Suit(tmp[3]) == Suit(tmp[4]))
		{
			*type = LOWSTRAIGHTFLUSH;
			return;
		}
		else
		{
			*type = LOWSTRAIGHT;
			return;
		}
	}

	// Flush
	if (Suit(card[0]) == Suit(card[1]) &&
		Suit(card[1]) == Suit(card[2]) &&
		Suit(card[2]) == Suit(card[3]) &&
		Suit(card[3]) == Suit(card[4]))
	{
		*type = FLUSH;
		return;
	}

}

void NewHand()
{
	btS i, j, p, c;
	char card[52];

	// Shuffle and deal.
	for (i = 0; i < 52; i++)
		card[i] = (char)i;

#ifdef WIN32
	srand((unsigned)time(NULL));
#else  // Palm
	SysRandom(TimGetTicks());	// seed random generator
#endif

	for (i = 0, p = 0; p < 4; p++)
	{
		for (c = 0; c < 13; c++, i++) 
		{
#ifdef WIN32
			j = i + (rand() % (52 - i));
#else  // Palm
			j = i + (SysRandom(0) % (52 - i));
#endif
			s.p[p].card[c] = card[j];
			card[j] = card[i];
		}
	}

	// Determine play five cards or pair
	for (i=0; i<4; i++)
	{
#ifdef WIN32
		fivecard[i] = rand() % 2;
#else  // Palm
		fivecard[i] = SysRandom(0) % 2;
#endif
	}

	SortCards(s.p[0].card, 13, TRUE);
	SortCards(s.p[1].card, 13, TRUE);
	SortCards(s.p[2].card, 13, TRUE);
	SortCards(s.p[3].card, 13, TRUE);
	if (s.prefs.sort == SUIT)
		SortCards(s.p[HUMAN].card, 13, FALSE);
}

void SortCards(char *card, const btS count, const char ranksort)
{
	btS i, j;
	char tmp, ff, ss;

	// Use bubble sort
	if (ranksort)  // Sort by rank
	{
		for (i=count-1; i>0; i--)
		{
			for (j=0; j<i; j++)
			{
				ff = card[j];
				ss = card[j+1];
				if (Larger(ff, ss))
				{
					card[j] = ss;
					card[j+1] = ff;
				}
			}
		}
	}
	else	// Sort by suit
	{
   	for (i=count-1; i>0; i--)
   	{
   		for (j=0; j<i; j++)
   		{
   			if (card[j] > card[j+1])
   			{
   				tmp = card[j];
   				card[j] = card[j+1];
   				card[j+1] = tmp;
   			}
   		}
   	}
   }
}

// Sort cards from Ace to King
void AltSortCards(char *card, const btS count, const char ranksort)
{
	btS i, j;
	char ff, ss;

	// Use bubble sort
	if (ranksort)  // Sort by rank
	{
		for (i=count-1; i>0; i--)
		{
			for (j=0; j<i; j++)
			{
				ff = card[j];
				ss = card[j+1];
				if (AltLarger(ff, ss))
				{
					card[j] = ss;
					card[j+1] = ff;
				}
			}
		}
	}
	else	// Sort by suit
	{
   	for (i=count-1; i>0; i--)
   	{
   		for (j=0; j<i; j++)
   		{
				ff = card[j];
				ss = card[j+1];
				if (Suit(ff) > Suit(ss))
				{
					card[j] = ss;
					card[j+1] = ff;
				}
				else if (Suit(ff) == Suit(ss) && AltRank(ff) > AltRank(ss))
				{
					card[j] = ss;
					card[j+1] = ff;
				}
   		}
   	}
   }
}

void NewSet()
{
	btS i;
	
   //bzero(&s, sizeof(s));
   for (i=0; i<4; i++)
   {
		s.p[i].score = 0;
		//s.p[i].games = 0;
   }
   s.round = 0;
   NewGame();
}

void NewGame()
{
	btS i;

	for (i = 0; i < 4; i++)
	{
		s.p[i].left = 13;
		s.p[i].numplayed = -1;
		//s.p[i].played[0] = -1;
	}
	s.fivecardtype = NOTFIVECARD;
	s.round ++;
	s.round = (s.round > MAX_ROUND) ? 1 : s.round;
	
	NewHand();
	StartGame();
}

void ResetHighScore()
{
	btS i;
	
	for (i=0; i<5; i++)
	{
		s.highscore[i] = INIT_SCORE;
	}
}