/*
* Big Two - A popular poker game for Palm OS.
* Copyright (C) 2000  Woo Kok Tong
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*
* Contact the author at onehy@magix.com.sg or visit http://come.to/palmos 
*/

#ifndef typedef_h
#define typedef_h

#ifdef WIN32
	#include <stdio.h>
	#include <string.h>
	#include <stdlib.h>
	#include <time.h>
#else  // Palm
	//#pragma pack(2)
	//#include <Common.h>
	//#include <System/SysAll.h>
	#include <PalmOS.h>
#endif

#define cardID(n)	(cardID_0 + (n))
#define passID(n)	(labelID_youpass + (n))
#define cardIndex(id)	((id) - cardID_0)

#define Rank(id)	((int)(id) % 13)
#define Suit(id)	((int)(id) / 13)
#define AltRank(id)	(((int)(id) + 2) % 13)
#define Next(id)	(((int)(id) + 1) % 4)

#define BigTwoAppID	'wkt1'
#define BigTwoDBType	'Data'
#define HUMAN	0	// human player index
#define CARDLEFT			28  // played card left coordinate
#define CARDTOP			16  // played card top coordinate
#define CARDHALFWIDTH	12  // played card half width
#define BMPWIDTH			9
#define BMPHEIGHT			9

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

// Five card play
#define NOTFIVECARD			0
#define LOWSTRAIGHT			1
#define STRAIGHT				2
#define FLUSH					3
#define FULLHOUSE				4
#define FOUROFAKIND			5
#define LOWSTRAIGHTFLUSH	6
#define STRAIGHTFLUSH		7

// Preferences
#define MEDIUM		0
#define SLOW		1
#define FAST		2

#define RANK		0
#define SUIT		1

#define MAX_SCORE			100
#define MAX_GAMES			999
#define MAX_ROUND			10
#define MAX_NAME			10
#define INIT_SCORE		-10000
#define PRO_SCORE			300

typedef short				btS;
typedef unsigned short	btUS;
typedef int					btI;
typedef unsigned int		btUI;
typedef long				btL;
typedef unsigned long	btUL;

struct State 
{
  	// Player information
  	struct Player {
    	btS score;	// total score for this game
    	//btS games;	// no. of games won
    	btS left;	// no. of cards left
    	btS numplayed;	// number of cards played
    	char card[13];	// player's cards
		char played[5];	// cards that are played
	} p[4];
	btS lastplayer;		// last player who played
	btS turn;		// next player's turn
	char fivecardtype;	// Type of five card play
	btS round;		// Keep track of which round this is
	btS highscore[5];		// Keep track of 5 highest score
	char highname[5][MAX_NAME+1];// Corresponding names for high score
	char startgame;	// First play must include 3 of diamond
	
	// Preferences
	struct Prefs
	{
		char autopass;	// Auto pass when human player's card is smaller
		char speed;		// Game speed
		char sort;		// Sort cards by
	} prefs;
} s;

extern void bzero(void *ptr, btI count);
extern void NewHand();
extern void SortCards(char *card, const btS count, const char ranksort);
extern void AltSortCards(char *card, const btS count, const char ranksort);
extern void NewSet();
extern void NewGame();
extern void StartGame();
extern void PalmPlay();
extern void UpdateScore(btS* hand);
extern void UpdateState(const char* playedcards, const btS count);
extern char PlayFiveCards(const btS playerIndex);
extern char PlayThreeCards(const btS playerIndex);
extern char PlayTwoCards(const btS playerIndex);
extern char PlayOneCard(const btS playerIndex);
extern char LargestCard(const char* card, const btS count);
extern char Larger(const char first, const char second);
extern char AltLarger(const char first, const char second);
extern char PlaySmallestCard(const char* tmp, const btS count, btS* index);
extern char PlayBiggestCard(const btS playerIndex);
extern void CountPair(const char* card, const btS count, btS* paircount, 
					btS* threescount, btS* fourscount);
extern char EvaluateStraight(const char* card, const btS count, const char* straight, const char fivecard);
extern char LookForStraight(const char* card, char* straight, const btS count, const char flush, const char play);
extern char LookForFlush(const char* card, char* flush, const btS count, const char play);
extern char LookForFullHouse(const char* card, char* fullhouse, const btS count, const char play);
extern char LookForFourOfAKind(const char* card, char* fourofakind, const btS count, const char play);
extern void StripPair(char* card, btS* count);
extern void StripDuplicateRank(char* card, btS* count);
extern void StripRank(char* card, btS* count, btS rank);
extern void StripFiveCard(char* card, btS* count, const char fivecard, const char straightonly);
extern char VerifyPlay(const char* card, const btS count);
extern void VerifyFiveCards(const char* card, char* type);
extern void ResetHighScore();

#endif  // typedef_h