/* Main code for BOFH */

#include <Pilot.h>
#include "callback.h"
#include "bofh.h"
#include "excuses.h"

DateType today;
ULong secs;
int currentExcuse;

typedef struct {
	Word day;
	int version;
	Word lastExcuse;
} Prefs;

static FieldPtr SetFieldText(Word fid, CharPtr s) {
	Handle hdl, oldhdl;
	FieldPtr fld;
	FormPtr frm;

	hdl = (Handle) MemHandleNew(StrLen(s) + 1);
	if (!hdl)
		return NULL;
	StrCopy(MemHandleLock((VoidHand) hdl), s);
	MemHandleUnlock((VoidHand) hdl);
	frm = FrmGetActiveForm();
	fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fid));
	ErrNonFatalDisplayIf(!fld, "Missing Field!");
	oldhdl = FldGetTextHandle(fld);
	FldSetTextHandle(fld, hdl);
	FldDrawField(fld);
	if (oldhdl)
		MemHandleFree((VoidHand) oldhdl);
	return fld;
}

static Boolean MainFormHandleEvent(EventPtr e) {
	Boolean handled = false;
    FormPtr frm;
    
#ifdef __GNUC__
    CALLBACK_PROLOGUE;
#endif /* __GNUC__ */

    switch (e->eType) {
	    case frmOpenEvent:
			frm = FrmGetActiveForm();
			FrmDrawForm(frm);
			if (currentExcuse >= NUMEXCUSES)
				currentExcuse = 0;
			SetFieldText(FIELD_EXCUSE, excuses[currentExcuse]);
			handled = true;
			break;

	    case menuEvent:
			MenuEraseStatus(NULL);

			switch(e->data.menu.itemID) {
				case MENU_ABOOT:
					FrmAlert(ALERT_ABOOT);
					handled = true;
					break;
			}
	    	handled = true;
			break;

	    case ctlSelectEvent:
			switch(e->data.ctlSelect.controlID) {
				case BUTTON_NEW:
					++currentExcuse;
					if (currentExcuse >= NUMEXCUSES) currentExcuse = 0;
					SetFieldText(FIELD_EXCUSE, excuses[currentExcuse]);
					handled = true;
			}
			break;

	    default:
    	    break;
    }

#ifdef __GNUC__
    CALLBACK_EPILOGUE;
#endif /* __GNUC__ */

    return handled;
}

static Boolean ApplicationHandleEvent(EventPtr e) {
    FormPtr frm;
    Word    formId;
    Boolean handled = false;

    if (e->eType == frmLoadEvent) {
		formId = e->data.frmLoad.formID;
		frm = FrmInitForm(formId);
		FrmSetActiveForm(frm);

		switch(formId) {
			case FORM_MAIN:
			    FrmSetEventHandler(frm, MainFormHandleEvent);
			    break;
		}
		handled = true;
    }

    return handled;
}

/* Get preferences, open (or create) app database */
static Word StartApplication(void) {
	Prefs p;
	
	secs = TimGetSeconds();
	DateSecondsToDate(secs, &today);
	if (!PrefGetAppPreferencesV10('Bofh', 0x00, &p, sizeof(p))) {
		FrmAlert(ALERT_ABOOT);
		p.lastExcuse = currentExcuse = SysRandom(0) % NUMEXCUSES;
		p.version = 1;
		p.day = today.day;
		PrefSetAppPreferencesV10('Bofh', 0x00, &p, sizeof(p));
	} else {
		currentExcuse = p.lastExcuse;
		if (p.day != today.day)
				currentExcuse++;
	}
    FrmGotoForm(FORM_MAIN);
    return 0;
}

/* Save preferences, close forms, close app database */
static void StopApplication(void) {
	Prefs p;
	
	p.lastExcuse = currentExcuse;
	p.version = 1;
	p.day = today.day;
	PrefSetAppPreferencesV10('Bofh', 0x00, &p, sizeof(p));
    FrmSaveAllForms();
    FrmCloseAllForms();
}

/* The main event loop */
static void EventLoop(void) {
    Word err;
    EventType e;

    do {
		EvtGetEvent(&e, evtWaitForever);
		if (!SysHandleEvent (&e))
	    	if (!MenuHandleEvent (NULL, &e, &err))
				if (!ApplicationHandleEvent (&e))
				    FrmDispatchEvent (&e);
    } while (e.eType != appStopEvent);
}

/* Main entry point; it is unlikely you will need to change this except to
   handle other launch command codes */
DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags) {
    Word err;

    if (cmd == sysAppLaunchCmdNormalLaunch) {
		err = StartApplication();
		if (err) return err;
		EventLoop();
		StopApplication();
	} else {
			return sysErrParamErr;
	}
    return 0;
}
