;****************************************************************
;*               Calc - Basic Calculator                        *
;*                                                              *
;*                                                              *
;*  Written by:  Gary T. Desrosiers                             *
;*               desrosi@pcnet.com                              *
;*               http://www.pcnet.com/~desrosi                  *
;*                                                              *
;*  Date:	 March 28th, 1997                               *
;*                                                              *
;*  Description: Calc is a basic calculator that implements     *
;*               add, subtract, multiply, divide, and percentage*
;*               functions. The basic idea for this code was to *
;*               demonstrate the usage of floating point        *
;*               functions within pila.                         *
;****************************************************************

        Appl    "SimpleCalc", 'ScLc'
        include "Pilot.inc"
        include "Startup.inc"

kidbAIB         equ             $7FFE
kidrPREF        equ             1
kidrTVER        equ             1
kidrTAIB        equ             1000
kidfMain        equ             1000
kidmMain        equ             1000
kidmAbout       equ             1002
kidrAboutAlert  equ             1000
kidrDivideZero  equ             1001
kidrFPConvError	equ		1002
kidb0		equ		48
kidb1		equ		49
kidb2		equ		50
kidb3		equ		51
kidb4		equ		52
kidb5		equ		53
kidb6		equ		54
kidb7		equ		55
kidb8		equ		56
kidb9		equ		57
kidbdot		equ		46
kidbadd		equ		43
kidbsub		equ		45
kidbmul		equ		42
kidbdiv		equ		47
kidbclr		equ		67
kidbce		equ		69
kidbequal	equ		61
kidbpercent	equ		37
kidcbackspace	equ		2001

struct FloatType
man.l                           ;unsigned 32 bit
exp.w                           ;signed 16 bit
sign.b                          ;+/- 1	
endstruct

        data
 	align 4
global Op1.FloatType		;Previously input operand
 	align 4
global Op2.FloatType		
 	align 4
global fpnResult.FloatType		
	align 4
global fpnTemp.FloatType
global NumOps.b			;Number of operands
global DecimalFlag.b		;Decimal Point Present Yet?
global LastInput.b		;Last input 
global LastOp.b			;Last operation + - x /
global inchar.2			;temp input keypress
global ReadOut.20		;What's displayed

        code

proc PilotMain(cmd.w, cmdPBP.l, launchFlags.w)
local   err.w
local   evt.EventType
beginproc
        tst.w   cmd(a6)                 ;sysAppLaunchCmdNormalLaunch is 0
        bne     Return                  ;not a normal launch, bag out
	call	Initialization()
        systrap FrmGotoForm(#kidfMain.w)
        systrap FplInit()		;initialize floating point environment
EvtLoop
	systrap EvtGetEvent(&evt(a6), #evtWaitForever.w)
        systrap SysHandleEvent(&evt(a6))
        tst.b   d0                      ;handled?
        bne.s   EvtDone                ;yep
        systrap MenuHandleEvent(&0, &evt(a6), &err(a6))
        tst.b   d0                      ;handled?
        bne.s   EvtDone                 ;yep
        call    MainFormHandleEvent(&evt(a6))
        tst.b   d0                      ;handled?
        bne.s   EvtDone                 ;yep
        systrap FrmGetActiveForm()
        systrap FrmHandleEvent(a0.l, &evt(a6))
EvtDone cmpi.w  #appStopEvent,evt+EventType.eType(a6) ;time to stop?
        bne.s   EvtLoop                 ;nope, loop until it is
        systrap FplFree()		;clean up floating point env.
Return  moveq   #0,d0
endproc

 
proc MainFormHandleEvent(pevt.l)
local rect.RectangleType
beginproc
        movem.l a3-a4,-(a7)             ;save regs
        movea.l pevt(a6),a0		;a0 = pevt
        move.w  EventType.eType(a0),d0

        cmp.w	#frmLoadEvent,d0 	;frmLoadEvent?
        bne.s   MFH0
        systrap FrmInitForm(EventType.data+frmLoad.formID(a0).w)
        systrap FrmSetActiveForm(a0.l)
	bra	MFHEat

MFH0	cmp.w   #frmOpenEvent,d0
        bne.s	NotOpen
        systrap FrmGetActiveForm()
        systrap FrmDrawForm(a0.l)
	call	Initialization()
	move.w	#10,rect.topLeft.x(a6)
	move.w	#25,rect.topLeft.y(a6)
	move.w	#140,rect.extent.x(a6)
	move.w	#25,rect.extent.y(a6)
	systrap	WinDrawRectangleFrame(#5.w,&rect(a6))
	call	UpdateDisplay()
        bra     MFHEat

NotOpen cmp.w   #menuEvent,d0
        bne     MFH1
        cmp.w   #kidmAbout,EventType.data+menu.itemID(a0)
        bne     MFHNotHandled
        systrap FrmAlert(#kidrAboutAlert.w)
        bra     MFHEat
MFH1	cmp.w	#ctlEnterEvent,d0
	bne.s	MFH2
	cmp.w	#kidcbackspace,EventType.data+ctlSelect.controlID(a0)
	bne	MFHNotHandled
	tst.b	ReadOut(a5)		;is the display empty?
	beq	beep
	systrap	StrLen(&ReadOut(a5))
	subq.l	#1,d0
	cmp.b	#'.',ReadOut(a5,d0)
	bne.s	bs1
	clr.b	DecimalFlag(a5)
bs1	clr.b	ReadOut(a5,d0)
	call	UpdateDisplay()
	clr.b	LastInput(a5)		;set last input to num
	systrap SndPlaySystemSound(#sndClick.b)
	bra	MFHEat

MFH2	cmpi.w	#keyDownEvent,d0
	bne.s	MFH3
	cmpi.w	#'.',EventType.data+keyDown.chr(a0)
	beq.s	DoNum
	cmpi.w	#'0',EventType.data+keyDown.chr(a0)
	blt	MFHNotHandled
	cmpi.w	#'9',EventType.data+keyDown.chr(a0)
	ble.s	DoNum
	bra	MFHNotHandled


MFH3	cmp.w   #ctlSelectEvent,d0
        bne     MFHNotHandled
	cmp.w   #kidbdot,EventType.data+ctlSelect.controlID(a0)
	beq.s 	DoNum
        cmp.w   #kidb0,EventType.data+ctlSelect.controlID(a0)
	blt.s	NoDigit
        cmp.w   #kidb9,EventType.data+ctlSelect.controlID(a0)
	bgt.s 	NoDigit
DoNum	tst.b	LastInput(a5)
	beq.s	isNum
	clr.b	ReadOut(a5)
	clr.b	DecimalFlag(a5)
isNum   
	call	CountDigits(&ReadOut(a5))
	cmp.b	#8,d0		;max 8 digits (Pilot restriction)
	bge.s	beep
	movea.l	pevt(a6),a0	;restore trashed reg
	cmp.w   #kidbdot,EventType.data+ctlSelect.controlID(a0)
	bne.s 	NumInp
	tst.b	DecimalFlag(a5)
	beq.s	NoPrvDP
beep
	systrap SndPlaySystemSound(#sndError.b)
	bra	MFHNotHandled
NoPrvDP	move.b	#1,DecimalFlag(a5)
NumInp	move.w	EventType.data+ctlSelect.controlID(a0),d0
	move.b	d0,inchar(a5)
	clr.b   inchar+1(a5)
	systrap	StrCat(&ReadOut(a5),&inchar(a5))
	clr.b	LastInput(a5)		;set last input to num
	call	UpdateDisplay()
	bra	MFHEat
	
NoDigit cmp.w   #kidbclr,EventType.data+ctlSelect.controlID(a0)
	bne.s 	NotClr
	call	Initialization()
	call	UpdateDisplay()
	bra	MFHDone

NotClr  cmp.w   #kidbpercent,EventType.data+ctlSelect.controlID(a0)
	bne.s 	Notpct
;	Op1=Op1*ReadOut/100
        systrap FplAToF(&fpnTemp(a5),&ReadOut(a5))
	move.b	#$01,-(a7)	 ;this is 
	move.w	#$ffe7,-(a7)	 ; 100
	move.l	#$c8000000,-(a7) ;  in floating point
	move.b	FloatType.sign+fpnTemp(a5),-(a7)
	move.w	FloatType.exp+fpnTemp(a5),-(a7)
	move.l	FloatType.man+fpnTemp(a5),-(a7)
	pea.l	fpnResult(a5)
	trap	#15
	dc.w	sysTrapFplDiv
	lea.l	20(a7),a7
	move.b	FloatType.sign+fpnResult(a5),-(a7)
	move.w	FloatType.exp+fpnResult(a5),-(a7)
	move.l	FloatType.man+fpnResult(a5),-(a7)
	move.b	FloatType.sign+Op1(a5),-(a7)
	move.w	FloatType.exp+Op1(a5),-(a7)
	move.l	FloatType.man+Op1(a5),-(a7)
	pea.l	fpnResult(a5)
	trap	#15
	dc.w	sysTrapFplMul
	lea.l	20(a7),a7
	call	Conv(&fpnResult(a5),&ReadOut(a5))
	move.b	#1,DecimalFlag(a5)
	call	UpdateDisplay()
	bra	MFHEat

Notpct  cmp.w   #kidbce,EventType.data+ctlSelect.controlID(a0)
	bne.s 	NotCE
	tst.b	LastInput(a5)		;Num or Op?
	bne.s	CEOp
	clr.b	DecimalFlag(a5)		;clear decimal flag
	clr.b	LastInput(a5)		; = num
	call	Conv(&Op1(a5),&ReadOut(a5))
	call	UpdateDisplay()
	clr.b	ReadOut(a5)
	bra	MFHNotHandled
CEOp	clr.b	LastOp(a5)		;no last op
	clr.b	NumOps(a5)		;no ops yet
	bra	MFHDone
;
; 	Operator Click
;

NotCE	tst.b	LastInput(a5)		;LastInput was num?
	bne.s	LNotNum
	addq.b	#1,NumOps(a5)		;increments number of args
LNotNum	cmp.b	#1,NumOps(a5)	
	bne.s	ChkTwo
        systrap FplAToF(&Op1(a5),&ReadOut(a5))	;op1 = last value entered
	bra	MFHDone

ChkTwo	cmp.b	#2,NumOps(a5)
	bne	MFHOpX
        systrap FplAToF(&Op2(a5),&ReadOut(a5))
	cmp.b   #kidbadd,LastOp(a5)
	bne.s 	NotAdd
	move.b	FloatType.sign+Op1(a5),-(a7)
	move.w	FloatType.exp+Op1(a5),-(a7)
	move.l	FloatType.man+Op1(a5),-(a7)
	move.b	FloatType.sign+Op2(a5),-(a7)
	move.w	FloatType.exp+Op2(a5),-(a7)
	move.l	FloatType.man+Op2(a5),-(a7)
	pea.l	fpnResult(a5)
	trap	#15
	dc.w	sysTrapFplAdd
	lea.l	20(a7),a7
	bra	ASMD

NotAdd	cmp.b   #kidbsub,LastOp(a5)
	bne.s	NotSub
	move.b	FloatType.sign+Op2(a5),-(a7)
	move.w	FloatType.exp+Op2(a5),-(a7)
	move.l	FloatType.man+Op2(a5),-(a7)
	move.b	FloatType.sign+Op1(a5),-(a7)
	move.w	FloatType.exp+Op1(a5),-(a7)
	move.l	FloatType.man+Op1(a5),-(a7)
	pea.l	fpnResult(a5)
	trap	#15
	dc.w	sysTrapFplSub
	lea.l	20(a7),a7
	bra	ASMD

NotSub	cmp.b   #kidbmul,LastOp(a5)
	bne.s	NotMul
	move.b	FloatType.sign+Op1(a5),-(a7)
	move.w	FloatType.exp+Op1(a5),-(a7)
	move.l	FloatType.man+Op1(a5),-(a7)
	move.b	FloatType.sign+Op2(a5),-(a7)
	move.w	FloatType.exp+Op2(a5),-(a7)
	move.l	FloatType.man+Op2(a5),-(a7)
	pea.l	fpnResult(a5)
	trap	#15
	dc.w	sysTrapFplMul
	lea.l	20(a7),a7
	bra	ASMD

NotMul	cmp.b   #kidbdiv,LastOp(a5)
	bne.s	NotDiv
	call	CheckDBZ(&Op2(a5))
	tst.b	d0
	bne	MFHOpX
	move.b	FloatType.sign+Op2(a5),-(a7)
	move.w	FloatType.exp+Op2(a5),-(a7)
	move.l	FloatType.man+Op2(a5),-(a7)
	move.b	FloatType.sign+Op1(a5),-(a7)
	move.w	FloatType.exp+Op1(a5),-(a7)
	move.l	FloatType.man+Op1(a5),-(a7)
	pea.l	fpnResult(a5)
	trap	#15
	dc.w	sysTrapFplDiv
	lea.l	20(a7),a7
	bra	ASMD

NotDiv	cmp.b   #kidbequal,LastOp(a5)
	bne.s	MFHNotHandled
	move.l	FloatType.man+Op2(a5),FloatType.man+Op1(a5)
	move.w	FloatType.exp+Op2(a5),FloatType.exp+Op1(a5)
	move.b	FloatType.sign+Op2(a5),FloatType.sign+Op1(a5)
	bra.s	MFHOpX

ASMD	move.l	FloatType.man+fpnResult(a5),FloatType.man+Op1(a5)
	move.w	FloatType.exp+fpnResult(a5),FloatType.exp+Op1(a5)
	move.b	FloatType.sign+fpnResult(a5),FloatType.sign+Op1(a5)
MFHOpX	move.b	#1,NumOps(a5)
	call	Conv(&Op1(a5),&ReadOut(a5))
MFHDone movea.l pevt(a6),a0		;a0 = pevt
	move.w	EventType.data+ctlSelect.controlID(a0),d0
	move.b	d0,LastOp(a5)
	move.b	#1,LastInput(a5)	;was Ops
	call	UpdateDisplay()
MFHEat	moveq.l	#1,d0
	bra.s	MFHReturn
MFHNotHandled
        clr.b   d0
MFHReturn
        movem.l (a7)+,a3-a4		;restore regs
endproc

proc CheckDBZ(fpn.l)			
beginproc
	movea.l	fpn(a6),a0
	clr.l	d0			;assume not zero
	cmp.l	#0,FloatType.man(a0)
	bne.s	DBZx
	cmp.w	#0,FloatType.exp(a0)
	bne.s	DBZx
	cmp.b	#1,FloatType.sign(a0)
	bne.s	DBZx
        systrap FrmAlert(#kidrDivideZero.w)
	move.l	#1,d0			;return true, is DBZ
DBZx
endproc

proc CountDigits(pStr.l)
local count.l
beginproc
	clr.l	count(a6)		;clear counter	
	movea.l	pStr(a6),a0		;get string pointer
CntLoop	move.b	(a0)+,d0		;get next digit
	beq.s	CntDone			;complete on zero terminator
	cmp.b	#'0',d0		
	blt.s	CntLoop
	cmp.b	#'9',d0	
	bgt.s	CntLoop
	addq.l	#1,count(a6)		;increment num of digits
	bra.s	CntLoop
CntDone	move.l	count(a6),d0	
endproc

proc UpdateDisplay()
local TempDisp.2
local rect.RectangleType
beginproc
	movem.l	d1-d3/a3,-(a7)
	move.w	#11,rect.topLeft.x(a6)
	move.w	#27,rect.topLeft.y(a6)
	move.w	#132,rect.extent.x(a6)
	move.w	#20,rect.extent.y(a6)
	systrap	WinEraseRectangle(&rect(a6),#0.w)
	systrap	FntSetFont(#ledFont.b)
	lea	ReadOut(a5),a3
	tst.b	ReadOut(a5)		;anything to display?
	bne.s	UD1
	move.b	#'0',TempDisp(a6)
	clr.b	TempDisp+1(a6)
	lea	TempDisp(a6),a3
UD1
	systrap	StrLen(a3.l)
	move.l	d0,d3
	systrap	FntCharsWidth(a3.l,d0.w)
	move.l	#130,d2
	sub.l	d0,d2
	systrap	WinDrawChars(a3.l,d3.w,d2.w,#29.w)
	tst.b	LastOp(a5)
	beq.s	UDDone
	cmp.b	#60,LastOp(a5)
	bgt.s	UDDone
	systrap	FntSetFont(#stdFont.b)
	systrap WinDrawChars(&LastOp(a5),#1.w,#12.w,#26.w)
UDDone
	movem.l	(a7)+,d1-d3/a3
endproc
	
proc Initialization()
beginproc
	clr.b	ReadOut(a5)		;null string readout
	clr.b	DecimalFlag(a5)		;clear decimal flag
	clr.b	LastOp(a5)		;no last op
	clr.b	NumOps(a5)		;no ops yet
	clr.b	LastInput(a5)		; = "none"
	clr.l	FloatType.man+Op1(a5)
	clr.w	FloatType.exp+Op1(a5)
	move.b	#1,FloatType.sign+Op1(a5)	;set all floats to zero
	clr.l	FloatType.man+Op2(a5)
	clr.w	FloatType.exp+Op2(a5)
	move.b	#1,FloatType.sign+Op2(a5)
endproc

proc Conv(pFpn.l,pResult.l)
local expNegFlag.w
local strtemp.20
beginproc
	move.l	pResult(a6),-(a7)
	movea.l	pFpn(a6),a0
	move.b	FloatType.sign(a0),-(a7)
	move.w	FloatType.exp(a0),-(a7)
	move.l	FloatType.man(a0),-(a7)
	trap 	#15
	dc.w	sysTrapFplFToA
	lea.l	12(a7),a7
	tst.l	d0
	beq.s	fplOK
        systrap FrmAlert(#kidrFPConvError.w)
	bra	ConvXit
;The result will be [-]x[.]yyyyyyyy[e[-]zz] or 0
fplOK	movea.l	pResult(a6),a3		;a3 -> source field
	lea	strtemp(a6),a4		;a4 -> destination string
	cmp.b	#'0',(a3)		;result zero?
	beq	ConvXit			;yes, then just display it
	clr.b	expNegFlag(a6)		;clear exponent flag
	cmp.b	#'-',(a3)		;neg mantissa result?
	bne 	NotNeg
	move.b	(a3)+,(a4)+		;negative result
NotNeg  move.b	#47,9(a3)		;font 6 strangeness
	cmp.b	#'-',10(a3)		;is exponent negative?
	bne	expPos
	move.b	#'-',expNegFlag(a6)	;save sign of exponent
	pea	11(a3)
	bra.s	doExpCv
expPos	pea	10(a3)
doExpCv trap	#15			;convert exponent to int in d0
	dc.w	sysTrapStrAToI
	addq.l	#4,a7
	cmp.l	#7,d0			;only convert +/- 7 exponents
	bgt	ConvXit			;exponent out of range, just display
manPos	tst.b	expNegFlag(a6)		;exponent neg?
	beq.s	expIsP
	move.b	#'0',(a4)+		;put in the decimal point
	move.b	#'.',(a4)+
	subq.l	#1,d0			;exp-1 digits left 
neLoop	tst.l	d0			
	beq.s	moveMan
	move.b	#'0',(a4)+		;move leading zeros
	subq.l	#1,d0
	bra.s	neLoop
expIsP  cmp.b	#'.',(a3)		;don't move the dec pt in the input stream
	bne.s	expIsPm
	lea	1(a3),a3		;inc input stream ptr
	bra.s	expIsP			;continue...
expIsPm	move.b	(a3)+,(a4)+		;move input stream to output
	dbeq	d0,expIsP		;continue
	move.b	#'.',(a4)+		;here's where the dp goes
moveMan	cmp.b	#'.',(a3)		;is it the dec pt on input?
	bne.s	moveMn2			;yes, then skip
	lea	1(a3),a3
	bra.s	moveMan
moveMn2	cmp.b	#47,(a3)		;end?
	beq.s	moveEnd
	move.b	(a3)+,(a4)+		;move input to output
	bra.s	moveMan
moveEnd	cmp.b	#'0',-(a4)		;Righthand zero suppress
	beq.s	moveEnd
	cmp.b	#'.',(a4)		;if x.0 form
	bne.s	dotZero
	lea	1(a4),a4		; then keep 1 trailing zero
	cmp.b	#'0',(a4)		;make sure '0' trails
	beq.s	dotZero
	move.b	#'0',(a4)
dotZero	lea	1(a4),a4		;non-zero terminating char
Reslt	clr.b	(a4)			;c-string type terminate string
	movea.l	pResult(a6),a0
	systrap	StrCopy(a0.l,&strtemp(a6))
ConvXit
endproc


;
; Resources -----------------------------------------------------------------
;

        res 'WBMP', kidbAIB, "calc.bmp"
        res 'pref', kidrPREF
        dc.w    sysAppLaunchFlagNewStack|sysAppLaunchFlagNewGlobals|sysAppLaunchFlagUIApp|sysAppLaunchFlagSubCall
        dc.l    $1000                           ; stack size
        dc.l    $1000                           ; heap size
        res 'tFRM', kidfMain, "tfrm03e8.bin"
        res 'tver', kidrTVER
        dc.b    '1.0', 0
        res 'MBAR', kidmMain, "mbar03e8.bin"
        res 'Talt', kidrAboutAlert, "talt03e8.bin"
        res 'Talt', kidrDivideZero, "talt03e9.bin"
        res 'Talt', kidrFPConvError, "talt03ea.bin"

