

#include "pfiltermanager.h"
#include "pdatabase.h"

#include "pfilter.h"

#include "plistmanager.h"

#include "global.h"

#include "pbottle.h"
#include "pcontext.h"
#include "pcollection.h"

#include "pfilter.h"

#include "cprotos.h"



PCollection	    PFilterManager::filtersName_(0, 0, 0);
PFilter		    PFilterManager::currentFilter_;
short			PFilterManager::currentFilterIndex_ = -1;
char**			PFilterManager::filtersList_ = 0;


void PFilterManager::destroy()
{
	MESSAGE("clearAndDestroy");
	filtersName_.clearAndDestroy();
	
	MESSAGE("MemPtrRecoverHandle");
	VoidHandle handle = MemPtrRecoverHandle(filtersList_);
	MemHandleFree(handle);
}

Boolean PFilterManager::init()
{
	currentFilter_.construct();

	filtersName_.construct(removeString, 0);

	VoidHandle handle = MemHandleNew(10);
	filtersList_ = (char**)MemHandleLock(handle);
	return true;
}

void PFilterManager::createFilteredList()
{
	PContext::getFilteredCollection().clear();

	PRecord& rec = PContext::getTheRecord();
	PDatabase& db = PContext::getBottlesDatabase();
	PString str;
	for (short i=0; i<db.getNumRecords(); i++)
	{
		str.print("get record %d", i);
		MESSAGE(str.data());
		if (!db.getRecord(i, rec))
		{
			MessageBox("Can't get the record from createFiltered");
		}
		else
		{
			if (matchFilter(rec))
			{
				PContext::getFilteredCollection().add( (void*)(i) );
			}
			str.print("release record %d", i);
			MESSAGE(str.data());
			db.releaseRecord(i);
			MESSAGE("Release done");
		}
	}
	MESSAGE("Create done");
}

Boolean PFilterManager::matchFilter(PRecord& rec)
{
	MESSAGE("matchFilter");
	if (currentFilter_.getName() == "")
	{
		MESSAGE("getName == \"\"");
		return true;
	}

	PString str;

    PFilter& filter = getCurrentFilter();
	PBottle& bottle = PContext::getTheBottle();
	bottle.load(rec);

	if (filter.getType() != "")
	{
		if (bottle.getType() != filter.getType())
		{
			return false;
		}
	}

	if (filter.getCapacite() != "")
	{
		if (bottle.getContenance() != filter.getCapacite())
		{
			return false;
		}
	}

	if (filter.getMillesime() != 0)
	{
		if (filter.getOlder())	// Older
		{
			if ( bottle.getMillesime() > filter.getMillesime() )
			{
				return false;
			}
		}
		else	// Younger
		{
			if ( bottle.getMillesime() < filter.getMillesime() )
			{
				return false;
			}
		}
	}

	if (filter.getToDrinkMax() != 0)
	{
		if (bottle.getToDrinkMin() == 0)
			return false;

		if ( bottle.getToDrinkMin() > filter.getToDrinkMax() )
		{
			return false;
		}
	}

	if (filter.getAtLeastNumber() != 0)
	{
		if ( bottle.getQuantite() < filter.getAtLeastNumber() )
		{
			return false;
		}
	}

	if (filter.getSearchString() != "")
	{
		if ( !bottle.findString(filter.getSearchString()) )
		{
			return false;
		}
	}

	return true;
}

//----------------------------------------------------------------------------
Boolean PFilterManager::getRecordWithCurrentFilter(PRecord& rec)
{
	if (currentFilterIndex_ < 1 || currentFilterIndex_ > getNbFilters() )
		return false;

	return PContext::getFiltersDatabase().getRecord(currentFilterIndex_, rec);
}

//----------------------------------------------------------------------------
Boolean PFilterManager::releaseCurrentFilterRecord()
{
	if (currentFilterIndex_ < 0)
		return false;

	return PContext::getFiltersDatabase().releaseRecord(currentFilterIndex_);
}

//----------------------------------------------------------------------------
short PFilterManager::getNbFilters()
{
	if (PContext::getFiltersDatabase().getNumRecords() == 0)
	{
		MessageBox("The preferences record is not created");
		return 0;
	}
	
	return (PContext::getFiltersDatabase().getNumRecords()-1);
}
//----------------------------------------------------------------------------
char* PFilterManager::getCurrentFilterName()
{
	static PString name;

	if (currentFilterIndex_ == -1)
		return PContext::getAllString();
	
	name = currentFilter_.getName().data();
	return name.data();
}

//----------------------------------------------------------------------------
Boolean PFilterManager::saveCurrentFilter()
{
	if (currentFilterIndex_ < 1)
	{
		MessageBox("saveCurrentFilter: Invalid filter index");
		return false;
	}

	PRecord rec;
	if (!getRecordWithCurrentFilter(rec))
	{
		currentFilterIndex_ = -1;
		currentFilter_.setName("");
		MessageBox("Can't get the current filter record");
		return false;
	}

	currentFilter_.save(rec);
	if (!releaseCurrentFilterRecord())
	{
		currentFilterIndex_ = -1;
		currentFilter_.setName("");
		MessageBox("Can't release the current filter record");
		return false;
	}

	return true;
}

//----------------------------------------------------------------------------
Boolean PFilterManager::addCurrentFilter()
{
	PRecord rec;
	short index = PContext::getFiltersDatabase().createRecord(0, rec);
	if (index < 0)
		return false;

	currentFilter_.save(rec);
	currentFilterIndex_ = index;
	
	return releaseCurrentFilterRecord();
}

//----------------------------------------------------------------------------
Boolean PFilterManager::setCurrentFilterIndex(short index)
{
	if (index < 0)
	{
		currentFilterIndex_ = -1;
		currentFilter_.setName("");
		return true;
	}

	if (index < 0 || index >= getNbFilters())
	{
		currentFilterIndex_ = -1;
		currentFilter_.setName("");
		//MessageBox("Invalid index to SET");
		return false;
	}

	index++;
	currentFilterIndex_ = index;

	PRecord rec;
	if (!getRecordWithCurrentFilter(rec))
	{
		currentFilterIndex_ = -1;
		currentFilter_.setName("");
		MessageBox("Can't get the current filter record");
		return false;
	}

	currentFilter_.load(rec);
	if (!releaseCurrentFilterRecord())
	{
		currentFilterIndex_ = -1;
		currentFilter_.setName("");
		MessageBox("Can't release the current filter record");
		return false;
	}
	return true;
}

//----------------------------------------------------------------------------
Boolean PFilterManager::removeCurrentFilter()
{
	if (currentFilterIndex_ < 0)
		return false;

	if (!PContext::getFiltersDatabase().removeRecord(currentFilterIndex_))
		return false;

	currentFilterIndex_ = -1;
	currentFilter_.setName("");
	return true;
}

//----------------------------------------------------------------------------
void* resizePointer(void* ptr, long size);

//----------------------------------------------------------------------------
char** PFilterManager::getFiltersList(Boolean withAll)
{
	// Create the filters name list
	PDatabase& db = PContext::getFiltersDatabase();
	PRecord rec;
	PFilter filter;

	MESSAGE("clearAndDest");
	filtersName_.clearAndDestroy();

	short i;
	for (i=0;i<getNbFilters();i++)
	{
		if (!db.getRecord(i+1, rec))
		{
			MessageBox("Can't get the record");
		}
		else
		{
			MESSAGE("load filter");
			filter.load(rec);
			MESSAGE("add name");
			filtersName_.add( new PString( filter.getName() ) );
			MESSAGE("releasing");
			db.releaseRecord(i+1);
			MESSAGE("record released");
		}
	}

	MESSAGE("delete filterslist_");
	filtersList_ = (char**)resizePointer(filtersList_, (getNbFilters()+1)*sizeof(char*) );

	short index = 0;
	if (withAll)
	{
		filtersList_[index] = PContext::getAllString();
		index++;
	}

	PString temp;
	PString* str;
	for (i=0;i<getNbFilters();i++)
	{
		temp.print("getAt '%d'", i);
		MESSAGE(temp.data());
		str = (PString*)filtersName_.getAt(i);
		if (str == 0)
		{
			MessageBox("NULL string!!!");
		}
		else
		{
			MESSAGE("assigning");
			filtersList_[i+index] = str->data();
		}
		MESSAGE("ok done");
	}

	MESSAGE("returning");
	return filtersList_;
}
