
#include "global.h"

#include "pstring.h"


#include "precord.h"

#include "cprotos.h"


#include <stdio.h>
#include <stdarg.h>





// <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>



void removeString(void* element)
{
	delete (PString*)element;
}


// <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
//	String Private Functions
// <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>

//--------------------------------------------------------------------------
PString::PString (void)
	:data_(0),
	allocSize_(0)
{
	construct();
    //allocString("abc\0");
#ifndef CAVIST_MFC
	if (data_ == 0)
		MessageBox("data_ == 0 in PString(void)");
#endif
}

//--------------------------------------------------------------------------
void PString::construct()
{
    data_ = 0;
	initString();
}
//--------------------------------------------------------------------------
void PString::construct(const char* str)
{
    construct();
	allocString(str);
}

//--------------------------------------------------------------------------
PString::PString (const char* str)
	:data_(0),
	allocSize_(0)
{
	construct();
#ifndef CAVIST_MFC
	if (data_ == 0)
		MessageBox("data_ == 0 before allocString");
#endif
	allocString(str);
}

//--------------------------------------------------------------------------
PString::PString(const PString& str)
	:data_(0),
	allocSize_(0)
{
    construct();
    allocString( str.data() );
}

//--------------------------------------------------------------------------
PString::~PString (void)
{
	resetString();
}







//-------------------------------------------------------------------
void PString::initString()
{
	resetString();

#ifdef CAVIST_MFC

    data_ = new char[STRING_MAX_BUF+10];
	data_[0] = 0;
	allocSize_ = STRING_MAX_BUF;

#else
	
	VoidHandle handle = MemHandleNew(STRING_MAX_BUF+1);
	if (handle == 0)
		MessageBox("Invalid Handle in resetString");
	allocSize_ = MemHandleSize(handle)-1;
	data_ = (char*)MemHandleLock(handle);
	if (data_ == 0)
		MessageBox("data_ == 0 after MemHandleLock");
	//data_[0] = 0;
	char a = 0;
	DmWrite((void*)data_, 0, (void*)&a, 1);
	if (data_ == 0)
		MessageBox("data_ == 0 after DmWrite");

#endif
}





//-------------------------------------------------------------------
void PString::resetString()
{
#ifdef CAVIST_MFC

	delete [] data_;

#else

	if (data_ != 0)
	{
		VoidHandle handle = MemPtrRecoverHandle(data_);
		if (handle == 0)
			MessageBox("Invalid Handle in resetString");
		MemHandleUnlock(handle);
		MemHandleFree(handle);
	}

#endif

    data_ = 0;
	allocSize_ = 0;
}

//-------------------------------------------------------------------
void PString::allocString(const char* str)
{
    if (str != (char *) NULL)
    {

#ifdef CAVIST_MFC
        long len = strlen(str);
#else
        long len = StrLen(str);
#endif

		allocNewLen(len+2);

#ifdef CAVIST_MFC
        strcpy(data_, str);
#else
        //StrCopy(data_, str);
		DmWrite((void*)data_, 0, (void*)str, len+1);
#endif
    }
	else
	{
		char a = 0;
		//data_[0] = 0;
#ifdef CAVIST_MFC
		data_[0] = 0;
#else
		DmWrite((void*)data_, 0, (void*)&a, 1);
#endif
	}
}

//-------------------------------------------------------------------
void PString::allocNewLen(long newLen)
{
#ifdef CAVIST_MFC

	if (newLen > (allocSize_-2) )
	{
		char* temp = data_;
		allocSize_ = newLen + STRING_MAX_BUF;
		data_ = new char[allocSize_+10];
		data_[0] = 0;
		if (temp != 0)
		{
			strcpy(data_, temp);
			delete [] temp;
		}
	}

#else

	if (newLen > (allocSize_-2) )
	{

		allocSize_ = newLen + STRING_MAX_BUF;
		
		if (data_ == 0)
			MessageBox("data_ == 0");
		VoidHandle handle = MemPtrRecoverHandle(data_);
		if (handle == 0)
			MessageBox("Invalid Handle in allocNewLen");

		MemHandleUnlock(handle);
		MemHandleResize(handle, allocSize_+10);
		allocSize_ = MemHandleSize(handle)-10;
		data_ = (char*)MemHandleLock(handle);
        if (data_ == 0)
            MessageBox("data_ == 0 in allocNewLen");
	}

#endif
}

//-------------------------------------------------------------------
void PString::appendString (const char *str)
{
#ifdef CAVIST_MFC
    long len = strlen(str);
#else
    long len = StrLen(str);
#endif

	allocNewLen(Length()+len+2);

#ifdef CAVIST_MFC
	strcat(data_, str);
#else
	//StrCat(data_, str);
	DmWrite((void*)data_, Length(), (void*)str, len+1);
#endif

}

//-------------------------------------------------------------------
Boolean PString::findSubString(const PString& str, bool caseSensitive)
{
#ifdef CAVIST_MFC
  	if (strstr(data(), str.data()) == NULL)
    {
      return false;
    }
#else
    if (caseSensitive)
    {
	    if (StrStr(data(), str.data()) == NULL)
      {
    		return false;
      }
    }
    else
    {
	    if (StrStr(data(), str.data()) == NULL)
      {
    		return false;
      }
    }
#endif

	return true;
}

//-------------------------------------------------------------------
long PString::CompareNoCase(const PString& other) const
{
#ifdef CAVIST_MFC
	return strcmp(data_, other.data_);
#else
	return StrCaselessCompare(data_, other.data_);
#endif
}

//-------------------------------------------------------------------
long PString::Compare(const PString& other) const
{
#ifdef CAVIST_MFC
	return strcmp(data_, other.data_);
#else
	return StrCompare(data_, other.data_);
#endif
}

//-------------------------------------------------------------------
bool CompareEqual (const char *x, const char *y)
{
    bool rv = FALSE;

    if (x == (char *) NULL && y == (char *) NULL)
        rv = TRUE;
    else if (x == (char *) NULL || y == (char *) NULL)
        rv = FALSE;
    else
#ifdef CAVIST_MFC
        rv = (strcmp(x, y) == 0);
#else
        rv = (StrCompare(x, y) == 0);
#endif
    return (rv);
}


// <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
//	String Public Functions
// <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>


//--------------------------------------------------------------------------
const PString& PString::convert(long number)
{
	char buf[STRING_MAX_BUF];
#ifdef CAVIST_MFC
	itoa(number, buf, 10);
#else
	StrIToA(buf, number);
#endif
	
    allocString(buf);
	return *this;
}




// Operators for (==, !=, =)

//--------------------------------------------------------------------------
bool operator== (const PString& x, const PString& y)
{
    bool rv = FALSE;

    rv = CompareEqual (x.data_, y.data_);
    return (rv);
}

//--------------------------------------------------------------------------
bool operator== (const PString& x, const char* y)
{
    bool rv = FALSE;

    rv = CompareEqual (x.data_, y);
    return (rv);
}

//--------------------------------------------------------------------------
bool operator!= (const PString& x, const PString& y)
{
    bool rv = FALSE;

    rv = CompareEqual (x.data_, y.data_);
    return (!rv);
}

//--------------------------------------------------------------------------
bool operator!= (const PString& x, const char* y)
{
    bool rv = FALSE;

    rv = CompareEqual (x.data_, y);
    return (!rv);
}

//--------------------------------------------------------------------------
PString& PString::operator=(const PString& other)
{
#ifndef CAVIST_MFC
    if (data_ == 0)
		MessageBox("data_ == 0 in operator=(PString&)");
#endif
    allocString(other.data_);
    return (*this);
}

//--------------------------------------------------------------------------
PString& PString::operator=(const char* other)
{
#ifndef CAVIST_MFC
	if (data_ == 0)
		MessageBox("data_ == 0 in operator=(char*)");
#endif
    allocString(other);
    return (*this);
}



// Operators for (+, +=)

//--------------------------------------------------------------------------
PString& PString::operator+(const PString& other)
{
	appendString (other.data_);
    return (*this);
}

//--------------------------------------------------------------------------
PString&  PString::operator+(const char* y)
{
    appendString(y);
    return (*this);
}

//--------------------------------------------------------------------------
PString& PString::operator+=(const PString& other)
{
    appendString(other.data_);
    return (*this);
}


// Indexing

//--------------------------------------------------------------------------
void PString::setAt(long i, char c)
{
	allocNewLen(i+1);
	data_[i] = c;
}

//--------------------------------------------------------------------------
char& PString::operator[] (long i)
{
    // ASSERT (i<0 || cllength_<i)
    return (data_[i]);
}

//--------------------------------------------------------------------------
const char& PString::operator[] (long i) const
{
    // ASSERT (i<0 || cllength_<i)
    return (data_[i]);
}

//--------------------------------------------------------------------------
PString& PString::print(const char *format, ...)
{
    va_list arglist;
    va_start(arglist, format);

	char buf[1024];buf[1023] = 0;
#ifdef CAVIST_MFC
	vsprintf(buf, format, arglist);
#else
	StrVPrintF(buf, format, arglist);
#endif
	operator=(buf);
#ifndef CAVIST_MFC
    if (Length() > 500)
		MessageBox("WARNING: print() overflow");
#endif

	va_end (arglist);
    return *this;
}

//--------------------------------------------------------------------------
long PString::Length() const
{
#ifdef CAVIST_MFC
	return strlen(data_);
#else
	short len = StrLen(data_);
	return len;
#endif
}

/*
//--------------------------------------------------------------------------
void PString::Fill (const char ch, int length)
{
    char	text[2], buf[STRING_MAX_BUF];
    int		ix, maxlength = length;

    if (length <= Length ())
        return;

    if (maxlength > STRING_MAX_BUF)
        maxlength = STRING_MAX_BUF;

    text[0] = ch;
    text[1] = '\0';

    StrCopy (buf, "");
    for (ix=Length(); ix<maxlength; ix++)
        strncat(buf, text, 1);
    appendString (buf);
}
*/

//--------------------------------------------------------------------------
char* PString::data() const
{
	return data_;
}






//--------------------------------------------------------------------------
PRecord& operator<<(PRecord& buff, const PString& str)
{
	buff.append(str.data(), str.Length()+1);
	return buff;
}

//--------------------------------------------------------------------------
PRecord& operator>>(PRecord& buff, PString& str)
{
	buff.get(str);

	return buff;
}

//--------------------------------------------------------------------------
PString operator+(const char* y, const PString& str)
{
	PString s;
	s = y;
	s += str;
	return s;
}
