
#include "global.h"
#include "ids.h"


#include "cprotos.h"

#include "PContext.h"

#include "FloatMgr.h"

#include "PDatabaseManager.h"
#include "PFilterManager.h"
#include "PListManager.h"
#include "PYearQuality.h"




//---------------------------------------------------
void about()
{
	FormPtr form = FrmInitForm(SPLASH_FORM_ID);
#ifdef DEMO_VERSION
	ControlPtr ctl;
	ctl = (ControlType*)FrmGetObjectPtr(form, FrmGetObjectIndex(form, VERSION_LBL_ID) );
	CtlSetLabel(ctl, "Unregistered version");
#else
	ControlPtr ctl;
	ctl = (ControlType*)FrmGetObjectPtr(form, FrmGetObjectIndex(form, REGISTER_LBL_ID) );
	CtlSetLabel(ctl, "Thanks for registering");
	ctl = (ControlType*)FrmGetObjectPtr(form, FrmGetObjectIndex(form, PGEAR_LBL_ID) );
	CtlSetLabel(ctl, "               Cavist");
#endif
	FrmDoDialog(form);
	FrmDeleteForm(form);
}





//---------------------------------------------------
long getFreeHeapSize(const char* msg = 0)
{
#ifndef DEBUG
    return 0;
#endif

    ULong freeBytes = 0;
    for (UInt cardNo=0;cardNo<MemNumCards();cardNo++)
    {
        UInt nbHeaps = MemNumHeaps(cardNo);
        ULong max = 0, free = 0;
        for (UInt heapNo=0;heapNo<nbHeaps;heapNo++)
        {
            UInt id = MemHeapID(cardNo, heapNo);
            if (MemHeapDynamic(id))
            {
                MemHeapFreeBytes(id, &free, &max);
                freeBytes += free;
            }
        }
    }
    if (msg != 0)
    {
        PString str;
        str.print("%s\n'%ld'", msg, freeBytes);
        MessageBox(str.data());
    }

    return freeBytes;
}





//---------------------------------------------------
void help()
{
	FrmHelp(HELP_STR_ID);
	return;
}

//---------------------------------------------------
void caveIdeale()
{
	FrmHelp(CAVE_IDEALE_STR_ID);
	return;
}






//---------------------------------------------------
DWord PilotMain (Word cmd, Ptr cmdPBP, Word launchFlags)
{
	
	if (cmd == sysAppLaunchCmdNormalLaunch)		// Command test, could also use "if (!cmd)"
	{
        long beginFree = getFreeHeapSize();

#ifndef DEBUG
#ifdef DEMO_VERSION
		about();
#endif
#endif

		if (!StartApplication())				// Application start code
			return FALSE;

		EventLoop();                           // Event loop

		StopApplication ();                    // Application stop code

        long endFree = getFreeHeapSize();
#ifdef DEBUG
        if (endFree - beginFree > 0)
        {
            PString str;
            str.print("Memory leak: '%ld'", endFree - beginFree);
            MessageBox(str.data());
        }
#endif
	}

	return 0;
}








//---------------------------------------------------
static Word StartApplication()
{
	if (0 != FplInit())
	{
		return false;
	}

    /*
    getFreeHeapSize("Before allocating");
    VoidHandle h = MemHandleNew(5000);
    getFreeHeapSize("After allocating");
    MemHandleFree(h);
    getFreeHeapSize("After deallocating");
    */

    //getFreeHeapSize("Beginning");

    MESSAGE("Init strings");
    initStrings();
    //getFreeHeapSize("initStrings");


    MESSAGE("PContext::init");
	if (!PContext::init())
	{
		return false;
	}
    //getFreeHeapSize("PContext");

    MESSAGE("PListManager::init");
	if (!PListManager::init())
	{
		return false;
	}
    //getFreeHeapSize("PListManager");

    MESSAGE("PFilterManager::init");
	if (!PFilterManager::init())
	{
		return false;
	}
    //getFreeHeapSize("PFilterManager");

    MESSAGE("PDatabaseManager::init");
	if (!PDatabaseManager::init())
	{
		return false;
	}
    //getFreeHeapSize("PDatabaseManager");

    MESSAGE("PYearQuality::init");
	if (!PYearQuality::init())
	{
		return false;
	}
    //getFreeHeapSize("PYearQuality (end)");

	return true;
}











//---------------------------------------------------
static void EventLoop(void)
{
	Word      err;
	int        formID;
	FormPtr    form;
	EventType  event;


	//FrmGotoForm(WINE_QUALITY_FORM_ID);
	//FrmGotoForm(EDIT_FORM_ID);
	FrmGotoForm(MAIN_FORM_ID);
	do
	{

		EvtGetEvent(&event, 2000);

		if (SysHandleEvent(&event))
			continue;

		if (MenuHandleEvent((MenuBarPtr)0, &event, &err))
			continue;

		if (event.eType == frmLoadEvent)
		{
			formID = event.data.frmLoad.formID;
			form = FrmInitForm(formID);
			FrmSetActiveForm(form);
			switch (formID) 
			{
			case MAIN_FORM_ID:
				FrmSetEventHandler(form, mainEventHandler);
				break;
			
			case TABLE_EDIT_FORM_ID:
			case EDIT_FORM_ID:
				FrmSetEventHandler(form, editEventHandler);
				break;

			case FILTER_FORM_ID:
				FrmSetEventHandler(form, filterEventHandler);
				break;

			case WINE_QUALITY_FORM_ID:
				FrmSetEventHandler(form, wineQualityEventHandler);
				break;

			case PREFS_FORM_ID:
				FrmSetEventHandler(form, preferencesEventHandler);
				break;

			default:
				MessageBox( Form_Invalide );
				break;
			}
		}

		FrmDispatchEvent(&event);

	} while(event.eType != appStopEvent);

}












//---------------------------------------------------
static void StopApplication()
{
	PDatabaseManager::destroy();

	PFilterManager::destroy();
	PListManager::destroy();
	PContext::destroy();
	
	FplFree();
}







void _exit(int i)
{
	//exit(i);
}

