

#include "pbottle.h"

#include "cprotos.h"

#include "plistmanager.h"

#include "pcollection.h"




#ifdef CAVIST_MFC
void logLine(const char* str);
#endif




//--------------------------------------------------------------------------
void removeBottle(void* element)
{
	if (element != 0)
	{
		PBottle* elem = (PBottle*)element;
		delete elem;
	}
}

//--------------------------------------------------------------------------

//--------------------------------------------------------------------------
void removeJournalEvent(void* element)
{
	if (element != 0)
	{
		PJournalEvent* elem = (PJournalEvent*)element;
		delete elem;
	}
}

//--------------------------------------------------------------------------
PJournalEvent::PJournalEvent(const PJournalEvent& event)
{
    copy(event);
}

//--------------------------------------------------------------------------
PJournalEvent& PJournalEvent::copy(const PJournalEvent& event)
{
    note_ = event.note_;
	date_ = event.date_;
    return *this;
}

//--------------------------------------------------------------------------
PString PJournalEvent::image()
{
    PString str;

    getDate().toString(str);
    str += ": ";
    str += getNote();
    str += "\n";

    return str;
}

//--------------------------------------------------------------------------









//--------------------------------------------------------------------------

char PBottle::currentVersion_ = 1;



//--------------------------------------------------------------------------
PBottle::PBottle()
    :journal_(removeJournalEvent, 0)
{
    construct();
#ifndef CAVIST_MFC
    contenance_ = Bouteille;
    type_ = Rouge;
    quantite_ = 1;
#endif
}
void PBottle::construct()
{
   	journal_.construct(removeJournalEvent, 0);

	region_.construct();
	appelation_.construct();
	cru_.construct();
	type_.construct();
	contenance_.construct();
	cepage_.construct();
	dateAchat_ = PDate();
	lieuAchat_.construct();
	producteur_.construct();
	appreciation_.construct();
	commentaire_.construct();
	prix_.construct();
	casier_.construct();
	classe_.construct();

	journal_.clearAndDestroy();
}
PBottle::PBottle(const PBottle& other)
    :journal_(removeJournalEvent, 0)
{
    construct();
	copy(other);
}
PBottle::~PBottle()
{
    //getJournal().clearAndDestroy();
}

//--------------------------------------------------------------------------
void PBottle::resetData()
{
	quantite_ = 0;
	region_ = "";
	appelation_ = "";
	cru_ = "";
	type_ = "";
	millesime_ = 0;
	contenance_ = "";
	toDrinkMin_ = 0;
	toDrinkMax_ = 0;
	cepage_ = "";
	dateAchat_ = PDate();
	lieuAchat_ = "";
	producteur_ = "";
	appreciation_ = "";
	commentaire_ = "";
	prix_ = "";
	casier_ = "";
	classe_ = "";

	journal_.clearAndDestroy();
}

//--------------------------------------------------------------------------
bool PBottle::isAllDataNull() const
{
	if (quantite_ != 0)
		return false;

	if (region_ != "")
		return false;

	if (appelation_ != "")
		return false;

	if (cru_ != "")
		return false;

	if (type_ != "")
		return false;

	if (millesime_ != 0)
		return false;

	if (contenance_ != "")
		return false;

	if (toDrinkMin_ != 0)
		return false;

	if (toDrinkMax_ != 0)
		return false;

	if (cepage_ != "")
		return false;

	if ( !dateAchat_.isNull() )
		return false;

	if (producteur_ != "")
		return false;

	if (appreciation_ != "")
		return false;

	if (commentaire_ != "")
		return false;

	if (prix_ != "")
		return false;

	if (casier_ != "")
		return false;
	
	if (lieuAchat_ != "")
		return false;

	if (classe_ != "")
		return false;

	return true;
}

//--------------------------------------------------------------------------
bool PBottle::isAllMandatorySet() const
{
	PString str;
	if (quantite_ == 0)
    {
#ifndef CAVIST_MFC
        MessageBox( No_Quantity );
#endif
		return false;
    }

	if (appelation_ == "")
    {
#ifndef CAVIST_MFC
        MessageBox( No_Appelation );
#endif
		return false;
    }

	if (type_ == "")
    {
#ifndef CAVIST_MFC
        MessageBox( No_Type );
#endif
		return false;
    }

	if (millesime_ == 0)
    {
#ifndef CAVIST_MFC
        MessageBox( No_Millesime );
#endif
		return false;
    }

	if (contenance_ == "")
    {
#ifndef CAVIST_MFC
        MessageBox( No_Contenance );
#endif
		return false;
    }

	return true;
}

//--------------------------------------------------------------------------
Boolean PBottle::findString(const PString& toSearch)
{
    PString str;

    // Numbers
    str.convert(quantite_);
    if (str.findSubString(toSearch))
        return true;

    str.convert(millesime_);
    if (str.findSubString(toSearch))
        return true;

    // Strings
	if (appelation_.findSubString(toSearch))
		return true;

	if (type_.findSubString(toSearch))
		return true;

    if (contenance_.findSubString(toSearch))
		return true;

    if (region_.findSubString(toSearch))
		return true;

    if (appreciation_.findSubString(toSearch))
		return true;
    
    if (cru_.findSubString(toSearch))
		return true;

	if (classe_.findSubString(toSearch))
		return true;

    // Numbers
    str.convert(toDrinkMin_);
    if (str.findSubString(toSearch))
        return true;

    str.convert(toDrinkMax_);
    if (str.findSubString(toSearch))
        return true;

    // Strings
	if (producteur_.findSubString(toSearch))
		return true;

	if (cepage_.findSubString(toSearch))
		return true;

    dateAchat_.toString(str);
	if (str.findSubString(toSearch))
		return true;

	if (lieuAchat_.findSubString(toSearch))
		return true;

	if (prix_.findSubString(toSearch))
		return true;

	if (casier_.findSubString(toSearch))
		return true;

	if (commentaire_.findSubString(toSearch))
		return true;

    return false;
}

//--------------------------------------------------------------------------
Boolean PBottle::testVersion(char version)
{
	if (version != getCurrentVersion())
	{
		PString str;
		str.print("Invalid version (%d) loaded while waiting for (%d)", version, getCurrentVersion());
#ifdef CAVIST_MFC
		logLine( str.data() );
#else
		MessageBox( str.data() );
#endif
		return false;
	}

	return true;
}

//--------------------------------------------------------------------------
void PBottle::save(PRecord& buffer)
{
	buffer.lock();
	buffer.reset();

	buffer << *this;

	buffer.unlock();
}

//--------------------------------------------------------------------------
Boolean PBottle::load(PRecord& buffer)
{
	buffer.lock();

	char version;
	buffer >> version;
	
	if (!testVersion(version))
	{
		buffer.unlock();
		return false;
	}

	loadBegin(buffer);
	loadEnd(buffer);

	buffer.unlock();
	return true;
}

//--------------------------------------------------------------------------
Boolean PBottle::loadSmall(PRecord& buffer)
{
	buffer.lock();

	char version;
	buffer >> version;
	if (!testVersion(version))
	{
		buffer.unlock();
		return false;
	}

	loadBegin(buffer);

	buffer.unlock();
	return true;
}

//--------------------------------------------------------------------------
Boolean PBottle::loadWithEvents(PRecord& buffer)
{
	buffer.lock();

	char version;
	buffer >> version;
	if (!testVersion(version))
	{
		buffer.unlock();
		return false;
	}

	loadBegin(buffer);
	loadEnd(buffer);
	loadEvents(buffer);

	buffer.unlock();
	return true;
}

//--------------------------------------------------------------------------
void PBottle::addEvent(const char* str)
{
    PJournalEvent* event = new PJournalEvent;
    event->setDate( PDate::today() );
    event->setNote(str);
    addEvent(event);
}

//--------------------------------------------------------------------------
void PBottle::addEvent(const PString& str)
{
    addEvent(str.data());
}

//--------------------------------------------------------------------------
void PBottle::addEvent(PJournalEvent* event)
{
    journal_.add(event);
}

//--------------------------------------------------------------------------
PBottle& PBottle::copy(const PBottle& other)
{
	quantite_ = other.quantite_;
	appelation_ = other.appelation_;
	type_ = other.type_;
	millesime_ = other.millesime_;
	contenance_ = other.contenance_;
	toDrinkMin_ = other.toDrinkMin_;
	toDrinkMax_ = other.toDrinkMax_;
	region_ = other.region_;
	producteur_ = other.producteur_;
	appreciation_ = other.appreciation_;
	cepage_ = other.cepage_;
	dateAchat_ = other.dateAchat_ ;
	prix_ = other.prix_;
	commentaire_ = other.commentaire_;
	casier_ = other.casier_;
	cru_ = other.cru_;
	classe_ = other.classe_;
	lieuAchat_ = other.lieuAchat_;

    PJournalEvent* event;
    PJournalEvent* oevent;
	for (long i=0;i<other.journal_.getSize();i++)
	{
        event = new PJournalEvent;
        oevent = (PJournalEvent*)other.journal_.getAt(i);
        if (event != 0 && oevent != 0)
        {
            event->copy(*oevent);
    		journal_.add(event);
        }
        else
		{
#ifndef CAVIST_MFC
			MessageBox("Event NULL");
#endif
		}
	}

	return *this;
}



//--------------------------------------------------------------------------
bool operator!=(const PBottle& b1, const PBottle& b2)
{
    if (b1 == b2)
        return false;
    return true;
}

//--------------------------------------------------------------------------
bool operator==(const PBottle& b1, const PBottle& b2)
{
    if (b1.quantite_ != b1.quantite_)
    {
        //MessageBox("quantite !=");
        return false;
    }

    if (b1.millesime_ != b2.millesime_)
    {
        //MessageBox("millesime_ !=");
        return false;
    }

	if (b1.appelation_ != b2.appelation_)
    {
        //MessageBox("appelation_ !=");
        return false;
    }

	if (b1.type_ != b2.type_)
    {
        //MessageBox("type_ !=");
        return false;
    }

    if (b1.contenance_ != b2.contenance_)
    {
        //MessageBox("contenance_ !=");
        return false;
    }

    if (b1.region_ != b2.region_)
    {
        //MessageBox("region_ !=");
        return false;
    }

    if (b1.appreciation_ != b2.appreciation_)
    {
        //MessageBox("appreciation_ !=");
        return false;
    }
    
    if (b1.cru_ != b2.cru_)
    {
        //MessageBox("cru_ !=");
        return false;
    }

	if (b1.classe_ != b2.classe_)
    {
        //MessageBox("classe_ !=");
        return false;
    }

    if (b1.toDrinkMin_ != b2.toDrinkMin_)
    {
        //MessageBox("toDrinkMin_ !=");
        return false;
    }

    if (b1.toDrinkMax_ != b2.toDrinkMax_)
    {
        //MessageBox("toDrinkMax_ !=");
        return false;
    }

	if (b1.producteur_ != b2.producteur_)
    {
        //MessageBox("producteur_ !=");
        return false;
    }

	if (b1.cepage_ != b2.cepage_)
    {
        //MessageBox("cepage_ !=");
        return false;
    }

    if (b1.dateAchat_ != b2.dateAchat_)
    {
        //MessageBox("dateAchat_ !=");
        return false;
    }

	if (b1.lieuAchat_ != b2.lieuAchat_)
    {
        //MessageBox("lieuAchat_ !=");
        return false;
    }

	if (b1.prix_ != b2.prix_)
    {
        //MessageBox("prix_ !=");
        return false;
    }

	if (b1.casier_ != b2.casier_)
    {
        //MessageBox("casier_ !=");
        return false;
    }

	if (b1.commentaire_ != b2.commentaire_)
    {
        //MessageBox("commentaire_ !=");
        return false;
    }

    return true;
}

//--------------------------------------------------------------------------

PRecord& operator<<(PRecord& buff, const PBottle& bottle)
{
	buff << PBottle::getCurrentVersion();

	buff << bottle.quantite_;
	buff << bottle.millesime_;
	buff << bottle.appelation_;
	buff << bottle.type_;
	buff << bottle.contenance_;

	buff << bottle.cru_;
	buff << bottle.region_;

	buff << bottle.appreciation_;
	buff << bottle.classe_;
	
	buff << bottle.toDrinkMin_;
	buff << bottle.toDrinkMax_;

	buff << bottle.producteur_;
	buff << bottle.cepage_;

	buff << bottle.dateAchat_;
	buff << bottle.lieuAchat_;
	buff << bottle.prix_;
	
	buff << bottle.casier_;
	buff << bottle.commentaire_;

	// Journal
	char nbItems = bottle.getJournal().getSize();
	buff << nbItems;
    PJournalEvent* event;
    for (long i=0;i<nbItems;i++)
    {
        event = (PJournalEvent*)bottle.getJournal().getAt(i);
        buff << *event;
    }

	return buff;

}

void PBottle::loadBegin(PRecord& buff)
{

	buff >> quantite_;
	buff >> millesime_;
	buff >> appelation_;
	buff >> type_;
	buff >> contenance_;

}

void PBottle::loadEnd(PRecord& buff)
{
	buff >> cru_;
	buff >> region_;

	buff >> appreciation_;
	buff >> classe_;
	
	buff >> toDrinkMin_;
	buff >> toDrinkMax_;

	buff >> producteur_;
	buff >> cepage_;

	buff >> dateAchat_;
	buff >> lieuAchat_;
	buff >> prix_;
	
	buff >> casier_;
	buff >> commentaire_;

}

void PBottle::loadEvents(PRecord& buff)
{
	// Journal
	char nbItems = (char)getJournal().getSize();
	buff >> nbItems;

    PString str;
    PJournalEvent* event;
    PString temp;
    for (long i=0;i<nbItems;i++)
    {
        event = new PJournalEvent;
        buff >> *event;
        journal_.add(event);
    }
}

//--------------------------------------------------------------------------
PRecord& operator<<(PRecord& buff, const PJournalEvent& event)
{
    buff << event.getDate();
    buff << event.getNote();
	return buff;
}

//--------------------------------------------------------------------------
PRecord& operator>>(PRecord& buff, PJournalEvent& event)
{
    buff >> event.date_;
    buff >> event.note_;
	return buff;
}


