
#ifndef PBottleINCLUDED
#define PBottleINCLUDED


#include "global.h"

#include "pstring.h"
//#include "precord.h"
#include "pdate.h"
#include "pcollection.h"

class PRecord;



//------------------------------------------------------------------------

enum e_Capacite { NOT_VALID_CAPA = -1, QUART = 187, DEMI_POT = 250, DEMI = 375, POT = 500, BOUTEILLE = 750, LITRE = 1000, MAGNUM = 1500, JEROBOAM = 3000, REHOBOAM = 4300, MATHUSALEM = 6000, SALMANAZAR = 9000, BALTHASAR = 12000, NABUCHODONOSOR = 15000 };


void removeJournalEvent(void* element);

//------------------------------------------------------
class PJournalEvent
{

public:

	PJournalEvent() {};
	PJournalEvent(const PJournalEvent& event);
	~PJournalEvent() {};

	const PString& getNote() const      { return note_; };
	const PDate& getDate() const        { return date_; };

	void setNote(const PString& note) { note_ = note; };
	void setDate(const PDate& date) { date_ = date; };

    friend PRecord& operator<<(PRecord& buff, const PJournalEvent& event);
    friend PRecord& operator>>(PRecord& buff, PJournalEvent& event);

    PJournalEvent& copy(const PJournalEvent& event);
    PString image();


private:

	PString note_;
	PDate date_;

};




//------------------------------------------------------------------------


void removeBottle(void* element);




class PBottle
{

public:

	PBottle();
	PBottle(const PBottle& other);
	void construct();
	~PBottle();

	static Boolean testVersion(char version);

	Boolean loadSmall(PRecord& buffer);
	Boolean load(PRecord& buffer);
	Boolean loadWithEvents(PRecord& buffer);

	void save(PRecord& buffer);

	void resetData();
	bool isAllDataNull() const;
	bool isAllMandatorySet() const;


	Boolean findString(const PString& toSearch);


	PBottle& operator=(const PBottle& other) { return copy(other); };
	PBottle& copy(const PBottle& other);



	void loadBegin(PRecord& buffer);
	void loadEnd(PRecord& buffer);
	void loadEvents(PRecord& buffer);


private:

	char quantite_;
	short millesime_;
	PString appelation_;
	PString type_;
	PString contenance_;
	PString region_;
	PString appreciation_;
	PString cru_;
	PString classe_;
	short toDrinkMin_;
	short toDrinkMax_;
	PString producteur_;
	PString cepage_;
	PDate dateAchat_;
	PString lieuAchat_;
	PString prix_;
	PString casier_;
	PString commentaire_;


public:

	PCollection journal_;

	static char getCurrentVersion()			{ return 1; };//currentVersion_; }

	
	char getQuantite() const				{ return quantite_; };
	const PString& getRegion() const		{ return region_; };
	const PString& getCru() const			{ return cru_; };
	const PString& getAppelation() const	{ return appelation_; };
	const PString& getType() const			{ return type_; };
	short getMillesime() const				{ return millesime_; };
	const PString& getContenance() const	{ return contenance_; };
	short getToDrinkMin() const				{ return toDrinkMin_; };
	short getToDrinkMax() const				{ return toDrinkMax_; };
	const PString& getCepage() const		{ return cepage_; };
	const PDate& getDateAchat() const		{ return dateAchat_; };
	const PString& getLieuAchat() const		{ return lieuAchat_; };
	const PString& getProducteur() const	{ return producteur_; };
	const PString& getAppreciation() const	{ return appreciation_; };
	const PString& getCommentaire() const	{ return commentaire_; };
	const PString& getCasier() const		{ return casier_; };
	const PString& getPrix() const			{ return prix_; };
	const PString& getClasse() const		{ return classe_; };

	const PCollection& getJournal() const   { return journal_; }





	void setQuantite(char nb)					{ quantite_ = nb; };
	void setRegion(const PString& pays)			{ region_ = pays; };
	void setCru(const PString& cru)				{ cru_ = cru; };
	void setAppelation(const PString& app)		{ appelation_ = app; };
	void setType(const PString& type)			{ type_ = type; };
	void setMillesime(short mil)					{ if (mil < 100) mil += 1900; millesime_ = mil; };
	void setContenance(const PString& capa)		{ contenance_ = capa; };
	void setToDrinkMin(short date)				{ toDrinkMin_ = date; };
	void setToDrinkMax(short date)				{ toDrinkMax_ = date; };
	void setCepage(const PString& str)			{ cepage_ = str; };
	void setDateAchat(const PDate& date)		{ dateAchat_ = date; };
	void setLieuAchat(const PString& str)		{ lieuAchat_ = str; };
	void setProducteur(const PString& prod)		{ producteur_ = prod; };
	void setAppreciation(const PString& note)	{ appreciation_ = note; };
	void setCommentaire(const PString& com)		{ commentaire_ = com; };
	void setCasier(const PString& cas)			{ casier_ = cas; };
	void setPrix(const PString& prix)			{ prix_ = prix; };
	void setClasse(const PString& str)			{ classe_ = str; };
	
	void addEvent(PJournalEvent* event);
	void addEvent(const char* str);
	void addEvent(const PString& str);


	friend PRecord& operator<<(PRecord& buff, const PBottle& b);
	friend PRecord& operator>>(PRecord& buff, PBottle& b);

	friend bool operator!=(const PBottle& b1, const PBottle& b2);
	friend bool operator==(const PBottle& b1, const PBottle& b2);


private:

	static char currentVersion_;

};

#endif

