

#include "pcollection.h"

#include "pstring.h"
#include "cprotos.h"


//#include "precord.h"

#include "pbottle.h"


#ifdef CAVIST_MFC
#include <memory.h>
#endif


//-----------------------------------------------------------
PCollection::PCollection(removeElementPtr func, void* def, long limit)
	:list_(0),
	nbUsedElements_(0),
	allocated_(0),
	removeFunction_(func),
	limit_(limit),
	default_(def)
{
	resize(20);
}

//-----------------------------------------------------------
void PCollection::construct(removeElementPtr func, void* def, long limit)
{
	list_ = 0;
	nbUsedElements_ = 0;
	allocated_ = 0;

	removeFunction_ = func;
	default_ = def;
	limit_ = limit;

#ifndef CAVIST_MFC
	if (list_ != 0)
		MessageBox("list_ != 0 !?!?!?");
#endif

	resize(20L);
}

//-----------------------------------------------------------
PCollection::~PCollection()
{
	clearAndDestroy();
}

//-----------------------------------------------------------
void PCollection::sort(compareElementPtr more, Boolean ascending)
{
	long size = getSize();
	void* coll1, *coll2;
	void* max;
	long index;
	for (long i=0;i<size;i++)
	{
		coll1 = getAt(i);
		max = coll1;
		index = i;
		for (long j=i+1;j<size;j++)
		{
			coll2 = getAt(j);
			if (ascending)
			{
				if ( more(coll2, max) )
				{
					index = j;
					max = coll2;
				}
			}
			else
			{
				if ( more(max, coll2) )
				{
					index = j;
					max = coll2;
				}
			}
		}
		setAt(index, coll1);
		setAt(i, max);
	}
}

//-----------------------------------------------------------
long PCollection::add(void* coll)
{
	PString str;
	str.print("In add\nAll:%ld nbUsed:%ld", allocated_, nbUsedElements_);
	//MessageBox(str.data());

	resize(nbUsedElements_+1L);

#ifdef DEMO_VERSION
	if (limit_ != 0 && limit_ <= getSize())
	{
		MessageBox( Limite_Atteinte );
		return -1;
	}
#endif

	long pos = nbUsedElements_;
	list_[pos] = coll;
	nbUsedElements_++;
	return pos;
}

//-----------------------------------------------------------
void PCollection::remove(void* coll)
{
	if (coll == 0)
		return;

	void* item;
	for (long i=0;i<getSize();i++)
	{
		item = getAt(i);
		if (item == coll)
		{
			removeFunction_(item);
			list_[i] = 0;

			for (long j=i+1;j<getSize();j++)
			{
				list_[j-1] = list_[j];
			}
			list_[getSize()-1] = 0;

			nbUsedElements_--;
			return;
		}
	}

#ifdef DEBUG
	MessageBox("Can't remove: Element not found");
#endif
}

//-----------------------------------------------------------
long PCollection::getIndexOf(void* coll) const
{
	if (coll == default_)
		return -1;

	for (long i=0;i<getSize();i++)
	{
		if (getAt(i) == coll)
			return i;
	}
	return -1;
}

//-----------------------------------------------------------
Boolean PCollection::setAt(long index, void* coll)
{
	if (index >= nbUsedElements_)
		return false;

	list_[index] = coll;

	return true;
}

//-----------------------------------------------------------
void* PCollection::getAt(long index)
{
	if (index <0 || index >= nbUsedElements_)
	{
#ifndef CAVIST_MFC
		MessageBox("Returning default");
#endif
		return default_;
	}

	return list_[index];
}

//-----------------------------------------------------------
const void* PCollection::getAt(long index) const
{
	if (index <0 || index >= nbUsedElements_)
	{
#ifndef CAVIST_MFC
		MessageBox("Returning default");
#endif
		return default_;
	}

	return list_[index];
}

//-----------------------------------------------------------
void PCollection::clear()
{
#ifdef CAVIST_MFC
	memset( list_, 0, allocated_*sizeof(voidPtr));
#else
	MemSet( list_, allocated_*sizeof(voidPtr), 0);
#endif
	nbUsedElements_ = 0;
}

//-----------------------------------------------------------
void PCollection::clearAndDestroy()
{
	for (long i=0;i<getSize();i++)
	{
		if (getAt(i) != 0)
		{
			removeFunction_(getAt(i));
		}
	}
	clear();
}

//-----------------------------------------------------------
void* resizePointer(void* ptr, long size);
//-----------------------------------------------------------
void PCollection::resize(long newSize)
{
	if (newSize - allocated_ < 0)
	{
		//MESSAGE("End of resize without");
		return;
	}

	if (newSize - allocated_ < 10)
		newSize = allocated_+20;

#ifdef CAVIST_MFC
	voidPtr* temp = new voidPtr[newSize];
	memset( temp, 0, newSize*sizeof(voidPtr));

	if (list_ != 0)
	{
		memcpy( temp, list_, allocated_*sizeof(voidPtr) );
		delete [] list_;
	}

	list_ = temp;

#else
	VoidHandle handle;
	if (list_ == 0)
	{
		MESSAGE("calling MemHandleNew");
		handle = MemHandleNew( newSize*sizeof(voidPtr) );
		if (handle == 0)
			MessageBox("Handle == 0");
		MESSAGE("calling MemHandleLock");
		list_ = (voidPtr*)MemHandleLock(handle);
	}
	else
	{
		MESSAGE("calling resizePointer");
		list_ = (voidPtr*)resizePointer(list_, newSize*sizeof(voidPtr) );
	}

#endif

	allocated_ = newSize;
	MESSAGE("End of resize");
}

