

#include "pdatabase.h"
#include "cprotos.h"

#include "precord.h"



//-----------------------------------------------------------
PDatabase::PDatabase(short maxRecords)
{
	ref_ = 0;
	localId_ = 0;
	maxRecords_ = maxRecords;
}

//-----------------------------------------------------------
void PDatabase::construct(short maxRecords)
{
	ref_ = 0;
	localId_ = 0;
	maxRecords_ = maxRecords;
	name_.construct();
}

//-----------------------------------------------------------
PDatabase::~PDatabase()
{
	close();
}

//-----------------------------------------------------------
Boolean PDatabase::open(const PString& str)
{
	name_ = str;

	LocalID id = DmFindDatabase(0, str.data());
	if (id == 0)
		return false;

	ref_ = DmOpenDatabase(0, id, dmModeReadWrite);
	return (ref_ != 0);
}

//-----------------------------------------------------------
Boolean PDatabase::openOrCreate(long type, const PString& str)
{
	if (!open(str))
	{
		if (0 != DmCreateDatabase(0, str.data(), CREATOR_ID, type, false))
			return false;

		if (!open(str))
			return false;
	}

	return true;
}

//-----------------------------------------------------------
void PDatabase::close()
{
    LocalID id = DmFindDatabase(0, name_.data());
    if (id != 0)
    {
        UInt attr;
        if (0 == DmDatabaseInfo( 0, id, NULL, &attr, NULL, NULL, NULL,
                        NULL, NULL, NULL, NULL, NULL, NULL))
        {
            attr = attr | dmHdrAttrBackup;
            DmSetDatabaseInfo( 0, id, NULL, &attr, NULL, NULL, NULL,
                        NULL, NULL, NULL, NULL, NULL, NULL);
        }
    }

	if (ref_ != 0)
	{
		// TODO: Set backup bit
		DmCloseDatabase(ref_);
	}
	ref_ = 0;
}

Boolean PDatabase::remove()
{
	LocalID id = DmFindDatabase(0, name_.data());
	if (id == 0)
		return false;
	
	return (0 == DmDeleteDatabase(0, id));
}

//-----------------------------------------------------------
short PDatabase::getNumRecords()
{
	if (ref_ == 0)
		return -1;

	return DmNumRecords(ref_);
}

//---------------------------------------------------------------
short PDatabase::createRecord(long size, PRecord& rec)
{
	if (size == 0)
		size = 16;

	unsigned short in = getNumRecords();
#ifdef DEMO_VERSION
	if (maxRecords_ != 0 && in >= maxRecords_)
	{
		MessageBox( Demo_10_seulement );
		return -1;
	}
#endif
	
    PString str;
    str.print("CreateRec:ref_db:%ld\nindex:%d\nsize:%ld", ref_, in, size);
    MESSAGE2(str.data());
	VoidHandle h = DmNewRecord(ref_, &in, size);
	if (h==0)
    {
        MESSAGE2("Failed.Record not created");
		return -1;
    }

    MESSAGE2("OK.Record created");
	rec.setHandle(h);
	return in;
}

//----------------------------------------------------------------
Boolean PDatabase::removeRecord(short index)
{
	if (index < 0 || index >= getNumRecords() || ref_ == 0)
	{
		MessageBox("Record Index not valid");
		return false;
	}
	else
	{
		if (0 != DmRemoveRecord(ref_, index))
		{
			MessageBox("Can't remove the record");
			return false;
		}
	}
	return true;
}

/*
VoidHandle PDatabase::detachRecord(long index)
{
	Handle handle = 0;
	if (0 == DmDetachRecord(ref_, (short)index, &handle))
	{
		handle = 0;
	}
	return (VoidHandle)handle;
}

long PDatabase::getIndex(long id)
{
	long index;
	if (0 != DmFindRecordByID(ref_, id, &index))
	{
		MessageBox("Can't remove the record");
		return -1;
	}

	return index;
}

Boolean PDatabase::attachRecord(VoidHandle handle)
{
	Handle h;
	return (0 == DmAttachRecord(ref_, 0, (Handle)handle, &h));
}
*/
Boolean PDatabase::getRecord(short index, PRecord& rec)
{
	short nb = getNumRecords();
	if (index < 0 || index >= nb)
	{
		MessageBox("Index not in bounds");
		return false;
	}

	rec.setHandle( DmGetRecord(ref_, index) );
	return true;
}

Boolean PDatabase::releaseRecord(short index)
{
	return (0 == DmReleaseRecord(ref_, (short)index, false));
}

Boolean PDatabase::quickSort(DmComparF* func, Boolean asc)
{
	if (0 != DmQuickSort(ref_, func, asc))
		return false;

	return true;
}

