
#include "global.h"

#include "pdatabasemanager.h"
#include "pdatabase.h"

#include "cprotos.h"

#include "precord.h"

#include "pcontext.h"

#include "ppreferences.h"
#include "pfiltermanager.h"
#include "plistmanager.h"
#include "pbottle.h"




Boolean PDatabaseManager::init()
{
	MESSAGE("loadprefs");
	if (!loadPreferences())
		MessageBox("Can't load preferences");

	MESSAGE("loadfilters");
	if (!loadFilters())
		MessageBox("Can't load filters");

	MESSAGE("loadbottles");
	if (!loadBottles())
		MessageBox("Can't load bottles");

	MESSAGE("setprefs");
	PPreferences::setPrefs();
	MESSAGE("end of init DB");

	return true;
}

void PDatabaseManager::destroy()
{
	MESSAGE("retrieve");
	PPreferences::retrieve();
	MESSAGE("savePreferences");
	savePreferences();

	MESSAGE("saveFilters");
	saveFilters();

	MESSAGE("saveBottles");
	saveBottles();
}

Boolean PDatabaseManager::loadPreferences()
{
	Boolean ret = false;

	PDatabase& db = PContext::getFiltersDatabase();
	PRecord& rec = PContext::getTheRecord();
	short index = 0;
	if (db.getNumRecords() == 0)
	{
		index = db.createRecord(0, rec);
		if (index == -1)
		{
			MessageBox("Can't create prefs record");
			return false;
		}
		PPreferences::save(rec);
		ret = true;
	}
	else
	{
		if (!db.getRecord(index, rec))
			return false;
		ret = PPreferences::load(rec);
	}
	db.releaseRecord(index);
	return ret;

}

Boolean PDatabaseManager::loadBottles()
{

	if (PContext::getBottlesDatabase().getNumRecords() > 0)
		return true;

//#ifndef DEBUG
#ifdef DEMO_VERSION

	//MessageBox("No DB: Create the bottle list by default");

	PRecord& rec = PContext::getTheRecord();
	PBottle& bottle = PContext::getTheBottle();
	PDatabase& db = PContext::getBottlesDatabase();
	short index;
	
	// Chateau d'Allaman
	bottle.setQuantite(12);
	bottle.setMillesime(1987);
	bottle.setAppelation("Chateau d'Allaman tres long et plus");
	bottle.setCru("Haut medoc");
#ifdef ML_FRENCH
	bottle.setContenance("Bouteille(0.75)");
	bottle.setType("Rouge");
#endif
#ifdef ML_ENGLISH
	bottle.setContenance("Bottle(0.75)");
	bottle.setType("Red");
#endif
	bottle.setAppreciation("***");
	bottle.setDateAchat( PDate(22, 3, 1974) );
	bottle.setPrix("12.50Frs");
	bottle.setCepage("Chasselas");
	bottle.setProducteur("Rochat Marcel, La Valle");
	bottle.setRegion("Aubonne");
	bottle.setCasier("A/34");
	bottle.setCommentaire("Excellent vin de la Cte, un peu brut sur les bords mais agrable en bouche. Un peu tannique.");


	index = db.createRecord(0, rec);
	if (index == -1)
	{
		MessageBox("Can't create record");
		return false;
	}
	bottle.save(rec);
	db.releaseRecord(index);


	// Chateau Iquem
	bottle.resetData();
	bottle.setQuantite(2);
	bottle.setAppelation("Chateau IQUEM");
	bottle.setCru("Haut medoc");
	bottle.setMillesime(1990);
	bottle.setPrix("24.50Frs");
#ifdef ML_FRENCH
	bottle.setContenance("Bouteille(0.75)");
	bottle.setType("Mousseux");
#endif
#ifdef ML_ENGLISH
	bottle.setContenance("Bottle(0.75)");
	bottle.setType("Sparkling");
#endif
	bottle.setAppreciation("****");
	index = db.createRecord(0, rec);
	if (index == -1)
	{
		MessageBox("Can't create record");
		return false;
	}
	bottle.save(rec);
	db.releaseRecord(index);



	// Villette
	bottle.setQuantite(6);
	bottle.setAppelation("Villette");
	bottle.setCru("Bas rhin");
	bottle.setMillesime(1996);
	bottle.setPrix("7.80Frs");
#ifdef ML_FRENCH
	bottle.setContenance("Bouteille(0.75)");
	bottle.setType("Blanc");
#endif
#ifdef ML_ENGLISH
	bottle.setContenance("Bottle(0.75)");
	bottle.setType("White");
#endif

	index = db.createRecord(0, rec);
	if (index == -1)
	{
		MessageBox("Can't create record");
		return false;
	}
	bottle.save(rec);
	db.releaseRecord(index);

	
	
	// Champagne
	bottle.setQuantite(1);
	bottle.setAppelation("Veuve Cliquot");
	bottle.setCru("Rien du tout");
	bottle.setMillesime(1993);
	bottle.setPrix("24.50FF");
#ifdef ML_FRENCH
	bottle.setContenance("Bouteille(0.75)");
	bottle.setType("Champagne");
#endif
#ifdef ML_ENGLISH
	bottle.setContenance("Bottle(0.75)");
	bottle.setType("Champagne");
#endif

	index = db.createRecord(0, rec);
	if (index == -1)
	{
		MessageBox("Can't create record");
		return false;
	}
	bottle.save(rec);
	db.releaseRecord(index);

#endif
//#endif


	return true;
}

Boolean PDatabaseManager::loadFilters()
{
	return true;
}

void PDatabaseManager::savePreferences()
{
    MESSAGE("getTheRecord");
	PRecord& rec = PContext::getTheRecord();
    MESSAGE("getFiltersDatabase");
	PDatabase& db = PContext::getFiltersDatabase();
	short index = 0;
    MESSAGE("getNumRecords");
	if (db.getNumRecords() == 0)
	{
        MESSAGE("Creating prefs record");
		index = db.createRecord(0, rec);
        MESSAGE("OK. Prefs record created");
		return;
	}
	else
    {
        MESSAGE("getRecord");
		if (!db.getRecord(index, rec))
            return;
    }

    MESSAGE("PPreferences::save");
	PPreferences::save(rec);
    MESSAGE("db.releaseRecord");
	db.releaseRecord(index);
    MESSAGE("savePreferences done.");
}

void PDatabaseManager::saveBottles()
{
	// Nothing todo
}

void PDatabaseManager::saveFilters()
{
	// Nothing todo
}





