
#include "pdate.h"
#include "precord.h"
#include "pstring.h"

#include "cprotos.h"


#ifdef CAVIST_MFC
char* Date_Str = "Date";
#endif

//--------------------------------------------------------------------------
PDate::PDate()
{
	date_.day = 0;
	date_.month = 0;
	date_.year = 0;
}

//--------------------------------------------------------------------------
PDate::PDate(char jour, char mois, long annee)
{
	setDay(jour);
	setMonth(mois);
	setYear(annee);
}

//--------------------------------------------------------------------------
PDate::PDate(const PDate& other)
{
	operator=(other);
}

void PDate::setYear(short annee)
{
	if (annee < 100)
    {
		date_.year = annee;
    }
    else
    {
		date_.year = (annee-1900);
    }

	if (date_.year < 0)
		date_.year = 0;

}

//--------------------------------------------------------------------------
PDate::~PDate()
{
}

//--------------------------------------------------------------------------
PDate& PDate::operator=(const PDate& other)
{
	date_.day = other.date_.day;
	date_.month = other.date_.month;
	date_.year = other.date_.year;

	return *this;
}

//--------------------------------------------------------------------------
bool operator!=(const PDate& date1, const PDate& date2)
{
	return FALSE;
}

//--------------------------------------------------------------------------
void PDate::toString(PString& str) const
{
	if (isNull())
	{
		str = Date_Str;
		return;
	}
	char data[100];
#ifdef CAVIST_MFC
	tm time;
	time.tm_mday = date_.day;
	time.tm_mon = date_.month-1;
	time.tm_year = date_.year;
	strftime(data, 100, "%d-%m-%Y", &time);
#else
	DateToAscii(date_.month, date_.day, date_.year, dfDMYWithDashes, data );
#endif
	str = data;
}

//--------------------------------------------------------------------------
PDate PDate::today()
{
    PDate date;

#ifndef CAVIST_MFC
    
    DateTimeType datetime;

    ULong secs = TimGetSeconds();

    TimSecondsToDateTime(secs, &datetime);

    date.setDay(datetime.day);
    date.setMonth(datetime.month);
    date.setYear(datetime.year);

#endif

    return date;
}


//--------------------------------------------------------------------------

PRecord& operator<<(PRecord& buff, const PDate& date)
{
	buff.append( (char*)&date.date_, 2);	// 2 bytes
	BIG_TO_LITTLE_SHORT(*(short*)&date.date_);
	return buff;
}

//--------------------------------------------------------------------------
PRecord& operator>>(PRecord& buff, PDate& date)
{
	buff.get( (char*)&date.date_, 2 );
	BIG_TO_LITTLE_SHORT(*(short*)&date.date_);
	return buff;
}

//--------------------------------------------------------------------------









#ifdef CAVIST_MFC

const char dateSeparators_[]    = {'/','.','-' ,0 };// Authorized date separators

Boolean PDate::fromString(const PString& str)
{
    
    const char* p;              // Parsing pointer
    const char* p1;             // First element 
    const char* s1;             // First separator 
    const char* p2;             // Second element
    const char* s2;             // Second separator 
    const char* p3;             // Third element 
    const char* separators;     // Authorized separators   


    // Init of the parsing pointer
    p = str.data();

    // Ignoring spaces
    while(isspace(*p)) p++;
    if(*p == 0)
    {
        return false;
    }

    // Beginning of the first value
    p1=p;

    // Jump over the first value
    while(isdigit(*p)) p++;
    if(*p == 0 || isalnum(*p))
    {
        return false;
    }
    
    // First separator
    s1=p;
    // Looking for separator ignoring spaces

    Boolean found=false; // Allowing only one separator

    while (!isalnum(*p) && *p != 0)
    {
        separators=dateSeparators_;

        while(*separators!=0)
        {
            if (*p == *separators++)
            {
                if(!found)
                {
                    s1=p;
                    found=true;
                }
                else
                {
                    return false;
                }
            }
        }
        p++;
    }
    if (found)
    {
        if (isalnum(*s1) || !*p)
        {
            return false;
        }
    }
    // Beginning of second value
    p2=p;
    // Jump over the second value
    if (isdigit(*p))
    {
        while(isdigit(*p)) p++;
    }
    else
    {
        while(isalpha(*p)) p++;
    }
    if(*p == 0 || isalnum(*p))
    {
        return false;
    }
    // Second separator
    s2=p;

    found=false;
    // Looking for second separator ignoring spaces
    while (!isalnum(*p) && *p != 0)
    {
        separators=dateSeparators_;
        while(*separators!=0)
        {
            if (*p == *separators++)
            {
                if(!found)
                {
                    s2=p;
                    found=true;
                }
                else
                {
                    return false;
                }
            }
        }
        p++;
    }
    if (found)
    {
        if (isalnum(*s1) || !*p)
        {
            return false;
        }
    }
    // Compare the two separators
    if (*s2 != *s1)
    {
        return false;
    }
    // Beginning of third value
    p3=p;
    
    if(*p3 == '\0')
    {
        return false;
    }
    // Jump over the third value
    while(isdigit(*p)) p++;
    // Looking for additonal caracters except spaces
    while(isspace(*p)) p++;
    if(*p != 0)//!p*
    {
        return false;
    }
    // We have our 5 elements. Let's convert them

    long day=0;
    long month=0;
    long year=0;

    if (*s1 == '/')
    {
        // AngloSaxon
        while(isdigit(*p1))
        {
            month*=10;
            month+=(int(*p1++)-int('0'));
        }
        while(isdigit(*p2))
        {
            day*=10;
            day+=(int(*p2++)-int('0'));
        }
    }
    else
    {
        // Standard
        while(isdigit(*p1))
        {
            day*=10;
            day+=(int(*p1++)-int('0'));
        }
        while(isdigit(*p2))
        {
            month*=10;
            month+=(int(*p2++)-int('0'));
        }
    }
    while(isdigit(*p3))
        {
            year*=10;
            year+=(int(*p3++)-int('0'));
        }
    // Year 2k correction
    if (year< 100)
    {
        if (year<50)
        {
            year+=2000;
        }
        else
        {
            year+=1900;
        }
    }
    // we have the day, month and year in int let's create the Tm_Date
	setDay(day);
	setMonth(month);
	setYear(year);
	/*
    date_.month = (short)month;
	date_.day = (short)day;
	date_.year = (short)year;
	*/

	return true;
}

#endif
