
#include "pfilter.h"
#include "cprotos.h"
#include "pstring.h"





char PFilter::currentFilterVersion_ = 10;


// To remove a filter in a collection
void removeFilter(void* element)
{
	delete (PFilter*)element;
}


//------------------------------------------------------------------
PFilter::PFilter()
	:millesime_(0),
	atLeastNumber_(0),
	toDrinkMax_(0),
	older_(0)
{
}

//------------------------------------------------------------------
PFilter::PFilter(const PString& name)
	:name_(name),
	millesime_(0),
	atLeastNumber_(0),
	toDrinkMax_(0),
	older_(0)
{
	reset();
}

//------------------------------------------------------------------
void PFilter::construct()
{
	millesime_ = 0;
	atLeastNumber_ = 0;
	toDrinkMax_ = 0;
	older_ = 0;
    name_.construct();
	type_.construct();
	capacite_.construct();
    searchString_.construct();

    reset();
}

//------------------------------------------------------------------
PFilter::~PFilter()
{
}

//------------------------------------------------------------------
void PFilter::reset()
{
	type_ = "";
	capacite_ = "";
	millesime_ = 0;
	atLeastNumber_ = 0;
	searchString_ = "";
}

//------------------------------------------------------------------
Boolean PFilter::load(PRecord& buffer)
{
	buffer.lock();

	char version;
	buffer >> version;
	if (!testVersion(version))
	{
		PString str;
		str.print("Wrong filter version '%d'", version);
		MessageBox(str.data());
		return false;
	}

	buffer >> *this;

	buffer.unlock();
	return true;
}

//------------------------------------------------------------------
void PFilter::save(PRecord& buffer)
{
	buffer.lock();
	buffer.reset();

	buffer << *this;

	buffer.unlock();
}
//------------------------------------------------------------------
char PFilter::getCurrentFilterVersion()
{
	return currentFilterVersion_;
}

//------------------------------------------------------------------
Boolean PFilter::testVersion(char version)
{
	if (version != getCurrentFilterVersion())
		return false;

	return true;
}

//------------------------------------------------------------------
PRecord& operator<<(PRecord& buffer, const PFilter& filter)
{
	buffer << PFilter::getCurrentFilterVersion();

	buffer << filter.name_;

	buffer << filter.type_;
	buffer << filter.capacite_;
	buffer << filter.millesime_;
	buffer << filter.older_;
	buffer << filter.atLeastNumber_;
	buffer << filter.searchString_;
	buffer << filter.toDrinkMax_;

	return buffer;
}

PRecord& operator>>(PRecord& buffer, PFilter& filter)
{
	buffer >> filter.name_;

	buffer >> filter.type_;
	buffer >> filter.capacite_;
	buffer >> filter.millesime_;
	buffer >> filter.older_;
	buffer >> filter.atLeastNumber_;
	buffer >> filter.searchString_;
	buffer >> filter.toDrinkMax_;

	return buffer;
}
