
#include "global.h"

#include "precord.h"

#include "cprotos.h"

#include "pstring.h"


#ifdef CAV_PILOT
//#include <string.h>
#endif




//------------------------------------------------------------------------
PRecord::PRecord()
{
    construct();
}
void PRecord::construct()
{
    toLock_ = true;
	data_ = 0;
	allocSize_ = 0;
	handle_ = 0;
	index_ = 0;

	reset();
}
PRecord::PRecord(VoidHandle handle)
{
    construct();
    handle_ = handle;
}

//------------------------------------------------------------------------
PRecord::~PRecord()
{
	reset();
	unlock();
}

//------------------------------------------------------------------------
void PRecord::reset()
{
	index_ = 0;
	allocSize_ = 0;
}

//------------------------------------------------------------------------
void PRecord::allocNewLen(long newLen)
{
	if (newLen <= allocSize_)
		return;

	if (newLen-allocSize_ < 128)
		newLen += 128;

#ifdef CAVIST_MFC

	char* old = data_;
	long oldLen = allocSize_;
	
	data_ = new char[newLen];
	allocSize_ = newLen;
	memcpy(data_, old, oldLen);
	
	delete [] old;

#else

	MemHandleUnlock(handle_);
	MemHandleResize(handle_, newLen);
	allocSize_ = MemHandleSize(handle_);
	data_ = (char*)MemHandleLock(handle_);

#endif

}

//-------------------------------------------------------------------------
void PRecord::lock()
{
#ifndef CAVIST_MFC
    if (!toLock_)
        return;

    if (data_ != 0)
	{
		MessageBox("Already locked");
		return;
	}

	data_ = (char*)MemHandleLock(handle_);
	if (data_ == 0)
	{
		MessageBox("Fatal error: data_ == 0");
	}

	allocSize_ = MemHandleSize(handle_);
#endif
	index_ = 0;
}

//-------------------------------------------------------------------------
void PRecord::unlock()
{
#ifndef CAVIST_MFC
    if (!toLock_)
        return;

	if (data_ == 0)
		return;

	MemHandleUnlock(handle_);
	data_ = 0;
#endif
}

//-------------------------------------------------------------------------
void PRecord::append(const char* data, long len)
{
    MESSAGE2("PRecord::append -> allocNewLen");
	allocNewLen(index_ + len);

	if (len == 0)
	{
#ifdef CAVIST_MFC
		len = strlen(data);
#else
		len = StrLen(data);
#endif
	}

#ifdef CAVIST_MFC
	memcpy(&data_[index_], (void*)data, len);
#else
    
    PString str;
    str.print("MemMove(%ld[%ld], %ld, %ld)", data_, data, index_, len);
    MESSAGE2(str.data());

	//MemMove(&data_[index_], (void*)data, len);
    DmWrite(data_, index_, (void*)data, len);
#endif
	index_ += len;
}

//-------------------------------------------------------------------------
Boolean PRecord::hasStillData()
{
	if ( (index_) >= allocSize_)
		return false;

	return true;
}

//-------------------------------------------------------------------------
void PRecord::get(char* data, long len)
{
	if ( (index_+len) > allocSize_)
	{
#ifdef DEBUG
		PString str;
		str.print("Out of bounds: WhTG_:%ld, LEN_:%ld WANT:%ld", index_, allocSize_, len);
		MessageBox(str.data());
#endif
		return;
	}

#ifdef CAVIST_MFC
	memcpy(data, &data_[index_], len);
#else
	MemMove(data, &data_[index_], len);
#endif
	index_ += len;
}

//-------------------------------------------------------------------------
void PRecord::get(PString& str)
{
	str = "";
	long len=0;
#ifdef CAVIST_MFC
	len = strlen( (char*)&data_[index_] );
#else
	len = StrLen( (char*)&data_[index_] );
#endif
	str = (char*)&data_[index_];
	index_ += (len + 1);
}






//-------------------------------------------------------------------------

PRecord& operator>>(PRecord& buff, long& val)
{
	buff.get( (char*)&val, 4 );
#ifdef CAVIST_MFC
	BIG_TO_LITTLE_LONG(val)
#endif
	return buff;
}

PRecord& operator<<(PRecord& buff, long val)
{
#ifdef CAVIST_MFC
	BIG_TO_LITTLE_LONG(val)
#endif
	buff.append((char*)&val, 4);
	return buff;
}



PRecord& operator>>(PRecord& buff, short& val)
{
	buff.get( (char*)&val, 2 );
#ifdef CAVIST_MFC
	BIG_TO_LITTLE_SHORT(val)
#endif
	return buff;
}

PRecord& operator<<(PRecord& buff, short val)
{
#ifdef CAVIST_MFC
	BIG_TO_LITTLE_SHORT(val)
#endif
	buff.append((char*)&val, 2);
	return buff;
}



PRecord& operator>>(PRecord& buff, float& val)
{
	buff.get( (char*)&val, 4 );
	return buff;
}

PRecord& operator<<(PRecord& buff, float f)
{
	buff.append((char*)&f, 4);
	return buff;
}





PRecord& operator>>(PRecord& buff, char& c)
{
	buff.get( (char*)&c, 1);
	return buff;
}

PRecord& operator<<(PRecord& buff, char c)
{
	buff.append(&c, 1);
	return buff;
}
