/*
 *  NAME: string.h
 *
 *  AUTHOR: joseph
 *
 *  DESCRIPTION: Defines a string object.
 *
 */

#ifndef PStringINCLUDED
#define PStringINCLUDED


// Includes
#ifdef CAVIST_MFC
#include <stdlib.h>
#else
#include <stringmgr.h>
#endif


#include "global.h"

//#include "precord.h"

class PRecord;


// Defines
#define STRING_NULL_LENGTH	0
#define STRING_NULL_SIZE	1
#define STRING_MAX_BUF		8




void removeString(void* element);




// <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
//    String
// <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>

class PString
{
public:

	friend PString operator+(const char* y, const PString& str);


    PString();
    PString(const char* str);		// string x = "abc"
    PString(const PString& str);	// string x = string
    void construct();
    void construct(const char* str);

	~PString();


	Boolean findSubString(const PString& str, bool caseSensitive = false);

    
    // Operators for (==, !=, =)
    PString& operator= (const PString& other);
    PString& operator= (const char* other);

    // Operators for (+, +=)
    PString& operator+ (const PString& other);
    PString& operator+ (const char* y);
    PString& operator+= (const PString& other);

    // Indexing
    char& operator[] (long i);
    const char& operator[] (long i) const;


    // Inline functions
    long Length() const;

    char* data() const;

#ifdef CAVIST_MFC
    long IntOf (void)
        { return ((long) atoi(data_)); };
#else
    long IntOf (void)
        { return ((long) StrAToI(data_)); };
#endif

    
	// Num to char
	const PString& convert(long number);
	//PString(FlpDouble number);

    
	// Print internals
    //void Dump();

	PString& print(const char *format, ...);
	
		
	friend bool CompareEqual (const char *x, const char *y);
	long CompareNoCase(const PString& other) const;
	long Compare(const PString& other) const;

	friend bool operator== (const PString& x, const PString& y);
	friend bool operator== (const PString& x, const char* y);
	friend bool operator!= (const PString& x, const PString& y);
	friend bool operator!= (const PString& x, const char* y);

	void setAt(long i, char c);

	friend PRecord& operator<<(PRecord& buff, const PString& b);
	friend PRecord& operator>>(PRecord& buff, PString& b);


protected:

    void allocNewLen(long newLen);
    void initString();
    void resetString();
    void allocString(const char *str);
    void appendString(const char *str);


private:

    char* data_;
    long allocSize_;

};

#endif /* _STRING_H_ */

