#ifndef   __DATABASE_H__
#define   __DATABASE_H__

#include "Main.h"

#define cardNo				0
#define cryptoPadName		"CryptoPadDB"
#define cryptoFontName		"CryptoFontDB"
#define cryptoPadType		'Data'
#define cryptoPadCreator	'CrPd'

#define version			0x03003000	/* PalmOS 3.0 version number */
#define OS35			0x03503000

#define memoPadName			"MemoDB"
#define memoPadType			'DATA'
#define memoPadCreator		'memo'

#define memoPadDB				0
#define cryptoPadDB				1

#define dbPrefIDV32				4
#define dbPrefIDV31				3
#define dbPrefID				6
//#define dbPrefID				7

#define dbVersion				36
#define dbFontVersion			32
#define dbGlobalPWVersion		31

#define noRecordSelected		-1
#define noRecordSelectedID		-1

#define catDecrypted			3
#define catEncrypted			2
#define recEncrypted			1
#define recDecrypted			0

#define globalPasswordChange	2
#define queryGlobalPassword		1
#define globalPasswordNotSet 	0

#define soManual				0
#define soAlphabetic			1

#define roAll					0
#define roAllUnfiled			1
#define roCategoriesUnfiled		2

#define newMemoSize  			64
#define maxGlobalPasswordLength 56

#define digitsForRecordPosition	6

#define updateCategoryChanged	0x01
#define updateTable				0x02
#define updateFontChanged		0x04
#define updateRedrawAll			0x08

#define titleReplacementChar	'#'

#define linesShown		 		1
#define linesHidden 			0

#define singleFont				1
#define separateFont 			0

#define recMove					1
#define recNotMove	 			0

#define setFont					0
#define getFont					1
#define fontTable				0
#define fontField				1

typedef char globalPasswordType[maxGlobalPasswordLength];

typedef struct {
	UInt16	renamedCategories;
	char	categoryLabels[dmRecNumCategories][dmCategoryLength];
	UInt8	categoryUniqIDs[dmRecNumCategories];
	UInt8	lastUniqID;

	UInt8	reserved1;
	UInt16	reserved2;
	globalPasswordType	prefsGlobalPassword;
} ApplicationInfoType;

typedef ApplicationInfoType* ApplicationInfoPtr;

typedef struct {
   	UInt16	renamedCategories;		// bitfield of categories with a different name
	char	categoryLabels[dmRecNumCategories][dmCategoryLength];
	UInt8	categoryUniqIDs[dmRecNumCategories];
	UInt8	lastUniqID;			// Uniq IDs generated by the device are between
									// 0 - 127.  Those from the PC are 128 - 255.
	UInt8	reserved1;			// from the compiler word aligning things
	UInt16	reserved2;
	UInt16	reserved3;
	UInt8	sortOrder;
	UInt8	reserved;
} MemoApplicationInfoType;

typedef MemoApplicationInfoType* MemoApplicationInfoPtr;

typedef struct {
	char	flag;
	char	passwdLen;
	UInt16	textLen;
	char	dataP[1];	/* beginning of the null-terminated string */
} PackedRecordType;

typedef PackedRecordType* PackedRecordPtr;

typedef struct {
	char	flag;
	char	passwdLen;
	UInt16	textLen;
	Char *	passwordP;
	Char *	textP;
} RecordType;

typedef RecordType* RecordPtr;

typedef struct {
	Boolean 	db;
	UInt32		uniqueID;
	FontID		tableFont;
	FontID		fieldFont;
} FontRecordType;

typedef FontRecordType* FontRecordPtr;


/* The application info structure intended to contain
 * the category information.
 * Category information is always placed in the beginning
 * where the category.c code expects it.
 */

typedef struct {
	UInt16	currentForm;
	UInt16	currentCategory;
	FontID	currentFieldFont;
	UInt8 reserved1;
	UInt16	currentRecord;
	UInt16	topRecord;
	UInt16	currentSortOrder;
	UInt16 	editScrollPosition;	
	Boolean	currentDB;
	Boolean	topDB;
	Boolean	showLines;
	Boolean	saveBackup;
	Boolean showAllCategories;
	UInt8 reserved2;
	UInt32	currentRecordID;
} ApplicationPreferenceTypeV22;

typedef struct {
	UInt16	currentForm;
	UInt16	currentCategory;
	FontID	currentFieldFont;
	UInt16	currentRecord;
	UInt16	topRecord;
	UInt16	currentSortOrder;
	FontID 	currentTableFont;
	UInt16 	editScrollPosition;	
	Boolean	currentDB;
	Boolean	topDB;
	Boolean	showLines;
	Boolean	saveBackup;
	Boolean	showAllCategories;
} ApplicationPreferenceTypeV31;

typedef struct {
	UInt16	currentForm;
	UInt16	currentCategory;
	FontID	currentFieldFont;
	UInt16	currentRecord;
	UInt16	topRecord;
	UInt16	currentSortOrder;
	FontID 	currentTableFont;
	UInt16 	editScrollPosition;	
	Boolean	currentDB;
	Boolean	topDB;
	Boolean	showLines;
	Boolean	saveBackup;
	Boolean showAllCategories;
	Boolean	useSingleFont;
} ApplicationPreferenceTypeV32;

typedef struct {
	UInt16	currentForm;
	UInt16	currentCategory;
	FontID	currentFieldFont;
UInt8 currentRotateOrder;

	UInt16	currentRecord;
	Boolean	topRecord;
	UInt16	currentSortOrder;
	FontID 	currentTableFont;
UInt8 moveRecordOnTop;

	UInt16 	editScrollPosition;
	Boolean	currentDB;
	Boolean	topDB;
	Boolean	showLines;
	Boolean	saveBackup;
	Boolean showAllCategories;
	Boolean	useSingleFont;
	UInt32	currentRecordID;
} ApplicationPreferenceType;

typedef ApplicationPreferenceType* ApplicationPreferencePtr;

typedef enum
{
	justLeft,
	justCenter,
	justRight
} JustificationEnum;


extern DmOpenRef dbCryptoPadRef;	/* Database */
extern DmOpenRef dbMemoPadRef;	/* Database */
extern DmOpenRef dbCryptoFontRef;	/* Database */
			
#define currentDBRef	((currentDB == cryptoPadDB) ? dbCryptoPadRef : dbMemoPadRef)
#define recordDBRef	((recordDB == cryptoPadDB) ? dbCryptoPadRef : dbMemoPadRef)
#define selectedDBRef	((selectedDB == cryptoPadDB) ? dbCryptoPadRef : dbMemoPadRef)

extern UInt16   currentForm;
extern UInt16   currentCategory;
extern FontID 	currentFieldFont;
extern FontID 	currentTableFont;
extern Boolean 	useSingleFont;
extern UInt16  	currentRecord;
extern UInt16	topRecord;
extern Boolean 	topDB;
extern UInt16 	currentSortOrder;
extern Boolean 	currentDB;
extern Boolean 	selectedDB;
extern Boolean 	showAllCategories;
extern Boolean 	showLines;
extern Boolean 	saveBackup;
extern UInt16 	editScrollPosition;
extern UInt8 	currentRotateOrder;
extern Boolean 	moveRecordOnTop;


extern Char 	categoryName[];
extern Char * 	currentPasswordP;
extern Char * 	globalPasswordP;
extern Char * 	prefsGlobalPasswordP;
extern UInt16	memosInCategory;
extern privateRecordViewEnum PrivateRecordVisualStatus;
extern UInt32 OSVersion;

DmOpenRef OpenMemoDatabase(UInt16);
DmOpenRef OpenOrCreateCryptoDatabase(UInt16);
DmOpenRef OpenOrCreateFontDatabase(void);

void CreateFontRecord(DmOpenRef, UInt32);
Boolean SetOrGetRecordFont(DmOpenRef, UInt32, Boolean, Boolean);
void SynchroniseTableFontForCurrentCategory(void);
void GetTableFontOfCategory(UInt16);

void RemoveCurrentPassword(void);
Boolean CreateRecord(void);
Boolean FontCharsInWidth(Char *, UInt16*, UInt16*);

void SynchroniseCategories(void);
void SynchroniseFont(void);
void ChangeCategory(UInt16);

void GoToItem (GoToParamsPtr, Boolean );
void Search (FindParamsPtr, Boolean);

Int16 CompareCryptoPadRecords(PackedRecordPtr, PackedRecordPtr, Int16);
Int16 CompareMemoPadRecords(MemPtr, MemPtr, Int16);

void UnpackRecord(RecordPtr, PackedRecordPtr);
Err PackRecord(RecordPtr, MemHandle);

void GetPrefsGlobalPassword (void);
void SetPrefsGlobalPassword(DmOpenRef);

UInt8 GetMemoPadSortOrder (DmOpenRef);
Err SetMemoPadSortOrder(DmOpenRef, Boolean);

#endif /* __DATABASE_H__ */
