#include "MainForm.h"
#include "Database.h"
#include "Utils.h"
#include "blowfish.h"

#include <SysEvtMgr.h>


static Boolean MainFormSeekRecord (UInt16* , Int16, Int16);
static WinHandle MainFormInvertMoveIndicator(RectanglePtr, WinHandle, Boolean);
static void MainFormDragMemo(EventPtr);
static void MainFormSelectTableItem (Boolean, TablePtr, Int16, Int16, RectangleType *r);
static void MainFormEnterTableItem(void);
static Boolean TableFindRowOfSelectedRecord(TablePtr, UInt16, UInt16*);
static void DrawInversionEffect (const RectangleType *rP, UInt16);

static void MainFormInit(void);

static void MainFormDrawTableItem(void *, Int16, Int16, RectanglePtr);
static void DrawCryptoPadRecordText(UInt16, UInt16, RectanglePtr);
static void DrawMemoPadRecordText(UInt16, UInt16, RectanglePtr);
static void MainFormDisplayMask (RectanglePtr);
static Int16 MainFormNumberOfRows (TablePtr);

static void MainFormLoadTable(void);
static void MainFormLoadRecords(void);
static void MainFormShowFullDisplay(void);
static MemHandle MainFormGetNextRecord(Int16*);
static MemHandle MainFormGetPreviousRecord(Int16*);

static void EncryptCategory();
static void DecryptCategory();
static ProgressInfoPtr ProgressFormInitAndDraw (FormPtr frmP);
static Boolean ProgressCallback (void* progressInfoP, ProgrType* progressP, Char* messageP);
static void ProgressUpdateGauge (RectangleType*	gaugeRP, UInt32 gaugeLevel);
static UInt32 ProgressHandleEvent (Int32 timeout);

static void MainFormSelectCategory(void);
static UInt16 MainFormGetNextCategory(UInt16);
static void MainFormNextCategory(void);

static void MainFormUpdateScrollBar(void);
static Boolean MainFormUpdateDisplay(UInt16);
static void MainFormScrollLines(Int16);
static void MainFormScrollPage(WinDirectionType);

static Boolean MainFormSeekRecord (UInt16* , Int16, Int16);
static UInt16 MainFormGotFirstCryptoPadRow(TablePtr);
static void MainFormDragMemo(EventPtr);

static void MainFormHandleMenuEvent(UInt16);

static UInt16 recordDB;

/***************************************
 * MainFormSeekRecord:
 * utility routine for MainFormDragMemo
 ***************************************
 */

static Boolean MainFormSeekRecord (UInt16* indexP, Int16 offset, Int16 direction)
{
	DmSeekRecordInCategory (selectedDBRef, indexP, offset, 
		direction, currentCategory);

	if (DmGetLastErr()) return (false);
	
	return (true);
}


/***************************************
 * MainFormGotFirstCryptoPadRow:
 * Search the number of the first
 * row with CryptoPad memos.
 ***************************************
 */

static UInt16 MainFormGotFirstCryptoPadRow(TablePtr tableP)
{
UInt16 nRow;
UInt16 numRows;

	numRows = TblGetNumberOfRows(tableP);
	
	for (nRow = 0; nRow < numRows; nRow++) {
		if (TblGetItemInt(tableP, nRow, kRecordColumn) == cryptoPadDB)
			return (nRow);
	}

	return (- 1);
}

/***************************************
 * MainFormInvertMoveIndicator:
 * Saves the screen behind the move
 * indicator while dragging a record.
 ***************************************
 */

static WinHandle MainFormInvertMoveIndicator(RectanglePtr itemR,
			WinHandle savedBits, Boolean draw)
{
	WinHandle winH = 0;
	RectangleType indictatorR;

	indictatorR.topLeft.x = itemR->topLeft.x;
	indictatorR.topLeft.y = itemR->topLeft.y + itemR->extent.y - 2;
	indictatorR.extent.x = itemR->extent.x;
	indictatorR.extent.y = 2;
				
	if (draw) {
		UInt16 i;
		UInt16 err;
		CustomPatternType pattern;
		CustomPatternType savedPattern;

		WinGetPattern ((CustomPatternType *) &savedPattern);

		for (i = 0; i < 8; i++) {
			pattern[i]= 0xAA;
			i++;
			pattern[i]= 0x55;
		}
	
		WinSetPattern ((const CustomPatternType *) &pattern);
	
		winH = WinSaveBits (&indictatorR, &err);
	
		WinFillRectangle (&indictatorR, 0);
	
		WinSetPattern ((const CustomPatternType *) &savedPattern);
	
	} else WinRestoreBits (savedBits, indictatorR.topLeft.x, indictatorR.topLeft.y);
	
	return (winH);
}

/**********************************************************
 * MainFormSelectTableItem
 * This routine either selects or unselects the specified
 * table item.				
 * While a table item is "on the move," having it be selected
 * can cause the Table code grief.  So, instead of using
 * the TblSelectItem()/TblUnhighlightSelect() calls, we now
 * manually select/unselect the table's row.  Before color,
 * only WinInvertRectangle() was called, so this is now in line
 * again with the way things used to work.  Sigh.
 **********************************************************/

static void MainFormSelectTableItem (Boolean selected, TablePtr table, Int16 row, Int16 column, RectangleType *r)
{
	// Get the item's rectangle.
	TblGetItemBounds (table, row, column, r);
	
	if (OSVersion < OS35)
		WinInvertRectangle (r, 0);
	else {

		// Set up the drawing state the way we want it.
		WinPushDrawState();
		WinSetBackColor(UIColorGetTableEntryIndex(UIFieldBackground));
		WinSetForeColor(UIColorGetTableEntryIndex(UIObjectForeground));
		WinSetTextColor(UIColorGetTableEntryIndex(UIObjectForeground));

		// Erase and (re)draw the item.
		WinEraseRectangle(r, 0);
		MainFormDrawTableItem(table, row, column, r);
	
		// If selected, make it look that way.
		if (selected)
			DrawInversionEffect(r, 0);
		
		// Restore the previous drawing state.
		WinPopDrawState();
	}
}

/**********************************************************
 * DrawInversionEffect
 * This routine does an inversion effect by swapping colors
 * this is NOT undoable by calling it a second time, rather
 * it just applies a selected look on top of already
 * rendered data.  (It's kind of a hack.)
 **********************************************************/
static void DrawInversionEffect (const RectangleType *rP, UInt16 cornerDiam) 
{
	WinPushDrawState();
	WinSetDrawMode(winSwap);
	WinSetPatternType (blackPattern);
	WinSetBackColor(UIColorGetTableEntryIndex(UIFieldBackground));
	WinSetForeColor(UIColorGetTableEntryIndex(UIObjectSelectedFill));
	WinPaintRectangle(rP, cornerDiam);
	
	if (UIColorGetTableEntryIndex(UIObjectSelectedFill) != UIColorGetTableEntryIndex(UIObjectForeground)) {
		WinSetBackColor(UIColorGetTableEntryIndex(UIObjectForeground));
		WinSetForeColor(UIColorGetTableEntryIndex(UIObjectSelectedForeground));
		WinPaintRectangle(rP, cornerDiam);
	}
	WinPopDrawState();
}

/*********************************
 * TableFindRowOfSelectedRecord:
 * Find the row which corresponds
 * to the selected record.
 *********************************/
static Boolean TableFindRowOfSelectedRecord(TablePtr tableP, 
			UInt16 selectedRecord, UInt16* selectedRow)

{
	UInt16 row;

	for (row = 0; row < TblGetNumberOfRows (tableP); row++) {
		if ((TblGetItemInt(tableP, row, kRecordColumn) == selectedDB) 
				&& TblRowUsable (tableP, row)) {
			if (TblGetRowID(tableP, row) == selectedRecord) {
				*selectedRow = row;
				return (true);
			}
		}
	}
	return (false);
}

/***************************************
 * MainFormDragMemo:
 * Performs the drag and drop operation
 * implementing Manual sorting.
 ***************************************
 */

static void MainFormDragMemo(EventPtr eventP)
{
	Int16 sortOrder;
	UInt16 selectedRow;
	UInt16 recordNum;
	UInt16 selectedRecord;
	UInt16 recAttributes;

	Int16 x, y;
	Boolean penDown = true;
	Boolean moving = false;
	Boolean selected = true;

	TablePtr tableP;
	Int16 row;
	UInt16 column;

	WinHandle savedBits;
	RectangleType rec;

	sortOrder = currentSortOrder;

	tableP = eventP->data.tblSelect.pTable;
	row = eventP->data.tblSelect.row;
	column = eventP->data.tblSelect.column;

	/* Highlight the selected record */

	selectedRecord = TblGetRowID(tableP, row);
	selectedDB = TblGetItemInt(tableP, row, kRecordColumn);

	MainFormSelectTableItem(selected, tableP, row, column, &rec);

	/* Trace the pen until it moves enough to constitute
	 * a move operation or until the pen it released.
	 */

	while (true) {

		PenGetPoint (&x, &y, &penDown);
		if (!penDown) break;
		
		if (!moving) {

			if (sortOrder != soAlphabetic) {

				/* Is the pen still within the bounds of the selected item?
				 * if not, draw the move indicator.
				 */

				if (!RctPtInRectangle (x, y, &rec)) {
					moving = true;
					TblGetItemBounds (tableP, row, column, &rec);
					savedBits = MainFormInvertMoveIndicator (&rec, 0, true);
				}

			} else selected = RctPtInRectangle(x, y, &rec);

		} else if (!RctPtInRectangle (x, y, &rec)) {

			/* Above the first item? */
			if (row < 0) {
				if (y >= rec.topLeft.y) {
					row++;
					MainFormInvertMoveIndicator(&rec, savedBits, false);
					rec.topLeft.y -= rec.extent.y;
					savedBits = MainFormInvertMoveIndicator (&rec, 0, true);
				}					
			}

			/* Move up */
			else if (y < rec.topLeft.y) {
				recordNum = TblGetRowID(tableP, row);
				recordDB = TblGetItemInt(tableP, row, kRecordColumn);

				if (MainFormSeekRecord (&recordNum, 1, dmSeekBackward) 
				&& (recordDB == selectedDB)) 
					{
					MainFormInvertMoveIndicator(&rec, savedBits, false);
					if (row) row--;
					else {				
						MainFormScrollLines(-1);
						if (TableFindRowOfSelectedRecord(tableP, selectedRecord, &selectedRow))
							MainFormSelectTableItem(selected, tableP, selectedRow, column, &rec);
					} 
					TblGetItemBounds (tableP, row, column, &rec);
					savedBits = MainFormInvertMoveIndicator (&rec, 0, true);

				} else if (((row == 0) && (recordDB == selectedDB)) 
					|| (MainFormGotFirstCryptoPadRow(tableP) == row)) {
					row--;
					MainFormInvertMoveIndicator (&rec, savedBits, false);
					rec.topLeft.y -= rec.extent.y;
					savedBits = MainFormInvertMoveIndicator (&rec, 0, true);
				}
			}

			/* Move down */
			else {
				recordNum = TblGetRowID(tableP, row);
				recordDB = TblGetItemInt(tableP, row, kRecordColumn);

				if (MainFormSeekRecord(&recordNum, 1, dmSeekForward)
				|| ((recordDB == memoPadDB) && (selectedDB == cryptoPadDB))) 
					{
					MainFormInvertMoveIndicator (&rec, savedBits, false);				
					if (row < TblGetLastUsableRow (tableP)) row++;
					else {
						MainFormScrollLines(1);
						if (TableFindRowOfSelectedRecord(tableP, selectedRecord, &selectedRow))
							MainFormSelectTableItem(selected, tableP, selectedRow, column, &rec);
					}

					TblGetItemBounds (tableP, row, column, &rec);
					savedBits = MainFormInvertMoveIndicator (&rec, 0, true);
				}
			}
		}
	}

	/* Turn off the move indicator, if it is on */

	if (moving) savedBits = MainFormInvertMoveIndicator (&rec, savedBits, false);

	/* If the highlighted item is visible, unhighlight it */
		if (TableFindRowOfSelectedRecord(tableP, selectedRecord, &selectedRow))
			MainFormSelectTableItem(false, tableP, selectedRow, column, &rec);

	if (moving) {
		if (row >= 0) {
			recordNum = TblGetRowID (tableP, row);
			recordDB = TblGetItemInt(tableP, row, kRecordColumn);

			if ((selectedRecord == recordNum) && (selectedDB == recordDB))
				return;
			
			recordNum++;

		} else recordNum = TblGetRowID(tableP, 0);

		if (recordDB != selectedDB) {
			if (selectedDB == cryptoPadDB) recordNum = 0;
			else recordNum = DmNumRecordsInCategory (dbMemoPadRef, currentCategory);
		}

		DmMoveRecord (selectedDBRef, selectedRecord, recordNum);

		MainFormLoadRecords();
		TblRedrawTable (tableP);
	}
		
	/* If we didn't move the item then it's been selected for editing,
	 * go to the Edit field.
	 */

	else if (sortOrder != soAlphabetic || selected) {
		currentRecord = TblGetRowID(tableP, row);
		currentDB = TblGetItemInt(tableP, row, kRecordColumn);

		editScrollPosition = 0;

		// Get the category and secret attribute of the current record.
		DmRecordInfo (currentDBRef, currentRecord, &recAttributes, NULL, NULL);

		// If this is a "private" record, then determine what is to be shown.
		if (recAttributes & dmRecAttrSecret) {
			switch (PrivateRecordVisualStatus) {
				case showPrivateRecords:
					MainFormEnterTableItem();
					break;
					
				case maskPrivateRecords:
					if (SecVerifyPW (showPrivateRecords) == true) {
						PrivateRecordVisualStatus = showPrivateRecords;
						MainFormEnterTableItem();
					}
					break;
			}
		} else MainFormEnterTableItem();
	}
}

/******************************************
 * MainFormEnterTableItem:
 * A item in the table of the main form
 * has been selected.
 ******************************************
 */

static void MainFormEnterTableItem(void)

{
	if (currentDB == cryptoPadDB)  {
		if (isEncrypted(currentRecord)) {
			EventType event;

			if ((prefsGlobalPasswordP != 0) && (globalPasswordP == 0)) {
				event.eType = globalPasswordStatusEvent;
				event.data.generic.datum[1] = queryGlobalPassword;
				event.data.generic.datum[2] = recEncrypted;

				FrmPopupForm(frmQueryGlobalPasswordForm);
				EvtAddEventToQueue(&event);
				return;
			}

		/* The record is encrypted.
		 * Pops the password form up.
		 */

			event.eType = memoStatusEvent;
			event.data.generic.datum[1] = recEncrypted;

			FrmPopupForm(frmPasswordForm);
			EvtAddEventToQueue(&event);
		} else FrmGotoForm(frmEditForm);

	} else FrmGotoForm(frmEditForm);
}

/******************************************
 * MainFormInit:
 * Initialize the main form and sets up the
 * table drawing callback functions.
 ******************************************
 */

static void MainFormInit(void)
{
	ControlPtr triggerP;
	UInt16 category;

	triggerP = GetActiveFormObjectPtr(popMainCategory);

	if (showAllCategories)
		category = dmAllCategories;
	else 
		category = currentCategory;

	/* Loads the record in the table */
	MainFormLoadRecords();

	/* Updates the currently selected category */
	CategoryGetName(dbMemoPadRef, category, categoryName);
	CategorySetTriggerLabel(triggerP, categoryName);

	currentForm = frmMainForm;
}

/****************************************
 * DrawStringWithEllipsis:
 * This routine draws the title of the specified memo. 
 ****************************************
 */

void DrawStringWithEllipsis (Char * textP, Int16 x, 
			Int16 y, Int16 width, JustificationEnum just)
{
// ***************** Code for PalmOS 3.0 to 3.2 *************//
/*	Boolean fits;
	Char ellipsisCharP[2];
	UInt16		strWidth = width;
	UInt16		maxStrWidth = width;
	UInt16		ellipsWidth = 0;
	UInt16 textLength;

	fits = FontCharsInWidth(textP, &strWidth, &textLength);

	// If the string fits, draw it entirely
	if (!fits) {
		MemSet(ellipsisCharP , 2, 0);
		ChrHorizEllipsis (&(ellipsisCharP[0]));

		// Take out the width of an ellipsis char
		ellipsWidth = FntCharWidth ((ellipsisCharP[0]));
		maxStrWidth -= ellipsWidth;

		// Keep removing last char until name fits
		while ((textLength > 0) && (strWidth > maxStrWidth)) {
			textLength--;
			strWidth -= FntCharWidth (textP[textLength]);
		}
	}

	if (just == justCenter)
		x += ((width - strWidth - ellipsWidth) / 2);

	// Draw the string
	WinDrawChars(textP, textLength, x, y);

	// Draw the ellipsis, if necessary
	if (ellipsWidth != 0) {
		x += strWidth;
		WinDrawChars (ellipsisCharP, 1, x, y);
	}
*/
// ***************** End of code for PalmOS 3.0 to 3.2 *************//

	Char * ptr = StrChr (textP, linefeedChr);
	Int16 strWidth;
	
	UInt16 titleLen = (ptr == NULL ? StrLen (textP) : (UInt16) (ptr - textP));
	if (FntWidthToOffset (textP, titleLen, width, NULL, &strWidth) == titleLen) {
		if (just == justCenter)
			x += ((width - strWidth) / 2);
		WinDrawChars (textP, titleLen, x, y);
	} else {
		Int16 ellipsisWidth;

		ellipsisWidth = FntCharWidth (chrEllipsis);
		titleLen = FntWidthToOffset (textP, titleLen, width - ellipsisWidth, NULL, &strWidth);
		if (just == justCenter)
			x += ((width - strWidth - ellipsisWidth) / 2);

		WinDrawChars (textP, titleLen, x, y);
		WinDrawChar (chrEllipsis, x + strWidth, y);
	}
}


/*******************************************
 * DrawMemoPadRecordText:
 * Draw each usable table cell
 * that contains a MemoPad record.
 *******************************************
 */

static void DrawMemoPadRecordText(UInt16 recordNum, UInt16 x, RectanglePtr bounds)
{
	MemHandle handle;
	Char * textP;

	Int16 cellWidth;

	handle = DmGetRecord(dbMemoPadRef, recordNum);
	textP = MemHandleLock(handle);

	// Determine the length of text that will fit within the bounds.
	cellWidth = bounds->extent.x - x;

	DrawStringWithEllipsis (textP, x, bounds->topLeft.y, cellWidth, justLeft);
	
	// Unlock the handle to the record.
	MemHandleUnlock(handle);
	
	// Release the record, not dirty.
	DmReleaseRecord(dbMemoPadRef, recordNum, false);	
}


 /*******************************************
 * DrawCryptoPadRecordText:
 * Draw each usable table cell
 * that contains a CryptoPad record.
 *******************************************
 */

 static void DrawCryptoPadRecordText(UInt16 recordNum, UInt16 x, RectanglePtr bounds)
{
	MemHandle handle;
	PackedRecordPtr packedP;
	RecordType record;

	UInt16 right;
	Int16 cellWidth;

	/* Adjust spacing with magic values! */
	right = 7;


	handle = DmGetRecord(dbCryptoPadRef, recordNum);
	packedP = MemHandleLock(handle);
	UnpackRecord(&record, packedP);

	// Determine the length of text that will fit within the bounds.	
	cellWidth = bounds->extent.x - x - right;

	DrawStringWithEllipsis (record.textP, x, bounds->topLeft.y, cellWidth, justLeft);

	MemHandleUnlock(handle);
	DmReleaseRecord(dbCryptoPadRef, recordNum, false);

	
	/* Draw the little icon */

	if (isEncrypted(recordNum)) {
		BitmapPtr bitmapP;

		handle = DmGetResource('Tbmp', bmpMainKeyBitmap);
		bitmapP = MemHandleLock(handle);
		WinDrawBitmap(bitmapP, bounds->extent.x - right, bounds->topLeft.y);
		MemHandleUnlock(handle);
		DmReleaseResource(handle);
	}
}
 
 /*******************************************
 * MainFormDrawTableItem:
 * Callback to draw each usable table cell
 *******************************************
 */

static void MainFormDrawTableItem(void * tableP,
				Int16 row, Int16 column, RectanglePtr bounds)
{
	UInt16 recordNum;
	UInt16 recordDB;
	UInt16 posInCategory;

	char posStr [6];
	UInt16 length;
	UInt16 x, y;
	UInt16 category;
	UInt16 recAttributes;


	/* Draw the record number */
	recordNum = TblGetRowID(tableP, row);
	recordDB  = TblGetItemInt(tableP, row, kRecordColumn);

	if (showAllCategories) category = dmAllCategories;
	else category = currentCategory;

	if (recordDB == memoPadDB) {
		posInCategory = DmPositionInCategory(dbMemoPadRef, recordNum, category);
	} else {
		posInCategory = DmNumRecordsInCategory(dbMemoPadRef, category);
		posInCategory += DmPositionInCategory(dbCryptoPadRef, recordNum, category);
	}

	x = bounds->topLeft.x + 1;
	y = bounds->topLeft.y;

	if (useSingleFont == separateFont)
		FntSetFont(TblGetItemFont (tableP, row, kRecordColumn));
	else FntSetFont(currentTableFont);

	/* Format the memo's postion, within its category, 
	* an draw it.
	*/
	StrIToA (posStr, posInCategory+1);
	length = StrLen(posStr);
	ErrNonFatalDisplayIf(length > sizeof(posStr) - 2, "Too many records");
	posStr[length++] = '.';
	posStr[length] = 0;
	
	if (length < 3) x += FntCharWidth ('1');
	WinDrawChars (posStr, length, x, y);
	x += FntCharsWidth (posStr, length) + 4;	

	// If we are here then we either we either mask the memo out or display the
	// memo title.
	DmRecordInfo (recordDBRef, recordNum, &recAttributes, NULL, NULL);
   	if (((recAttributes & dmRecAttrSecret) 
   	&& PrivateRecordVisualStatus == maskPrivateRecords)) {
		RectangleType maskRectangle;

		MemMove (&maskRectangle, bounds, sizeof (RectangleType));
		maskRectangle.topLeft.x = x;
		maskRectangle.extent.x = bounds->extent.x - x;
		MainFormDisplayMask (&maskRectangle);

	} else {
		/* Draw the record text */
		if (recordDB == cryptoPadDB) DrawCryptoPadRecordText(recordNum, x, bounds);
		else  DrawMemoPadRecordText(recordNum, x, bounds);
	}
}

/*******************************************
 * MainFormDisplayMask
 * Draws the masked display for the record.
  ******************************************/
static void MainFormDisplayMask (RectanglePtr bounds)
{
	RectangleType tempRect;
	CustomPatternType origPattern;
	MemHandle	bitmapH;
	BitmapType * bitmapP;
	
	CustomPatternType pattern = {0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55};
	
	MemMove (&tempRect, bounds, sizeof (RectangleType));
	// Make sure it fits nicely into the display.
	tempRect.topLeft.y++;
	tempRect.extent.y --;
	tempRect.extent.x -= SecLockWidth + 1;

	WinGetPattern((CustomPatternType *) &origPattern);
	WinSetPattern ((const CustomPatternType *) &pattern);
	WinFillRectangle (&tempRect, 0);
	WinSetPattern((const CustomPatternType *) &origPattern);
	
	//draw lock icon
	bitmapH = DmGetResource (bitmapRsc, SecLockBitmap);
	if (bitmapH) {
		bitmapP = MemHandleLock (bitmapH);
		WinDrawBitmap (bitmapP, tempRect.topLeft.x + tempRect.extent.x + 1, 
			tempRect.topLeft.y + ((tempRect.extent.y - SecLockHeight) / 2));
		MemPtrUnlock (bitmapP);
	}
}

/********************************************************
 * MainFormNumberOfRows
 * This routine return the maximun number of visible rows,
 * with the current list view font setting.
 ********************************************************/
static Int16 MainFormNumberOfRows (TablePtr tableP)
{
	Int16	rowsInTable;
	Int16 nRows = 0;
	UInt32 recordID;

	UInt16 lineHeight;
	UInt16 tableHeight;
	RectangleType	r;

	rowsInTable = TblGetNumberOfRows (tableP);

	TblGetBounds (tableP, &r);
	tableHeight = r.extent.y;

	if ((useSingleFont == separateFont) 
	&& (currentCategory == dmAllCategories)) {
		UInt16	recordNum;
		MemHandle nextRecord;
		UInt16 dataHeight = 0;

		recordNum = topRecord;
		currentDB = topDB;

		nextRecord = DmQueryRecord(currentDBRef, recordNum);

		if (nextRecord) {
			DmRecordInfo (currentDBRef, recordNum, 0, &recordID, 0);
			SetOrGetRecordFont(currentDBRef, recordID, getFont, fontTable);
			FntSetFont (currentTableFont);
			lineHeight = FntLineHeight ();
			dataHeight += lineHeight;
			nRows++;
		}

		while (dataHeight <= tableHeight) {
			nextRecord = MainFormGetNextRecord(&recordNum);
			if (nextRecord == 0) break;
				DmRecordInfo (currentDBRef, recordNum, 0, &recordID, 0);
				SetOrGetRecordFont(currentDBRef, recordID, getFont, fontTable);
				FntSetFont (currentTableFont);
				lineHeight = FntLineHeight ();
				dataHeight += lineHeight;
				nRows++;
		}

		if (nextRecord)
			nRows--;
	} else {
		if (useSingleFont == separateFont)
			GetTableFontOfCategory(currentCategory);
		FntSetFont (currentTableFont);
		nRows = tableHeight / FntLineHeight ();
	}

	if (nRows <= rowsInTable) 
		return (nRows);
	else
		return (rowsInTable);
}


/******************************************
 * MainFormGetNextRecord:
 * Get the record index of
 * the next record in the
 * current category.
 ******************************************
 */

static MemHandle MainFormGetNextRecord(Int16* recordNum)
{
	MemHandle record;

	(*recordNum)++;

	/* Seek the item after the recordNum index in MemoPad database.*/
	if (currentDB == memoPadDB) {
		record = DmQueryNextInCategory(dbMemoPadRef, recordNum, currentCategory);

		/* If record is null, set the currentDB to CryptoPadDB
		* to continue the search in the CryptoPad database from
		* the record 0.
		*/
		if (record == 0) {
			currentDB = cryptoPadDB;
			*recordNum = 0;
		}
	}

	/* Seek the item after the recordNum index in CryptoPad database.*/
	if (currentDB == cryptoPadDB) {
		record = DmQueryNextInCategory(dbCryptoPadRef, recordNum, currentCategory);
	}

	return (record);
}


/******************************************
 * MainFormGetPreviousRecord:
 * Get the record index of
 * the previous record in the
 * current category.
 ******************************************
 */

static MemHandle MainFormGetPreviousRecord(Int16* recordNum)
{
	Int16 offset = 1;
	Err err;

	/* Seek the item befor the recordNum index in CryptoPad database.*/
	if (currentDB == cryptoPadDB) {
		if (*recordNum == dmMaxRecordIndex)
			offset = 0;
		err = DmSeekRecordInCategory (dbCryptoPadRef, recordNum, 
				offset, dmSeekBackward, currentCategory);

		/* If record is null, set the currentDB to CryptoPadDB
		* to continue the search in the CryptoPad database from
		* the record 0.
		*/
		if ((err == dmErrSeekFailed) &&
		(DmNumRecordsInCategory (dbMemoPadRef, currentCategory) != 0)) {
			currentDB = memoPadDB;
			*recordNum = dmMaxRecordIndex;
			offset = 0;
		}
	}

	/* Seek the item before the recordNum index in MemoPad database.*/
	if (currentDB == memoPadDB) {
		err = DmSeekRecordInCategory (dbMemoPadRef, recordNum, 
				offset, dmSeekBackward, currentCategory);
	}

	if (!err) return (DmQueryRecord (currentDBRef, *recordNum));
	else return (0);
}

/******************************************
 * MainFormLoadTable:
 * Load the table with the records.
 ******************************************
 */

static void MainFormLoadTable(void)
{	
	TablePtr tableP;
	Int16 recordNum;
	Int16 numRows;

	MemHandle record;
	Int16 nRow;
	UInt16 lineHeight;
	UInt16 dataHeight;
	UInt16 tableHeight;
	RectangleType	r;

	tableP = GetActiveFormObjectPtr(tblMainRecordList);
	numRows = TblGetNumberOfRows(tableP);

	/* Set the correct topRecord and topDB
	 * according to the category.
	 */
	currentDB = topDB;

	topRecord--;
	MainFormGetNextRecord(&topRecord);
	
	topDB = currentDB;
	recordNum = topRecord;


	TblGetBounds (tableP, &r);
	tableHeight = r.extent.y;
	dataHeight = 0;

	/* For each row in the table, store the 
	* record number and database in the 
	* table item that will display the record.
	*/
	for (nRow = 0; nRow < numRows; nRow++) {

		if  (!nRow) recordNum--;
		record = MainFormGetNextRecord(&recordNum);

		if (record && (useSingleFont == separateFont) 
		&& (currentCategory == dmAllCategories)) {
			UInt32 recordID;

			DmRecordInfo (currentDBRef, recordNum, 0, &recordID, 0);
			SetOrGetRecordFont(currentDBRef, recordID, getFont, fontTable);
			FntSetFont (currentTableFont);
		}

		lineHeight = FntLineHeight ();

		if (record && (tableHeight >= dataHeight + lineHeight)) {
			TblSetItemStyle(tableP, nRow, kRecordColumn, customTableItem);
			TblSetItemFont (tableP, nRow, kRecordColumn, currentTableFont);
			TblSetRowHeight (tableP, nRow, lineHeight);
			TblSetRowID(tableP, nRow, recordNum);
			TblSetItemInt(tableP, nRow, kRecordColumn, currentDB);
			TblSetRowUsable(tableP, nRow, true);

		/* Marks the row invalid so that is is redrawn */
		TblMarkRowInvalid(tableP, nRow);
				
		} else {

			// Set the row height - when scrolling down, the heights of the last rows of
			// the table are used to determine how far to scroll.  As rows are deleted
			// from the top of the table, formerly unused rows scroll into view, and the
			// height is used before the next call to ListViewLoadTable (which would set
			// the height correctly).
			TblSetRowHeight (tableP, nRow, lineHeight);
			TblSetRowUsable(tableP, nRow, false);
		}

	dataHeight += lineHeight;
	}

	MainFormUpdateScrollBar();
}

/******************************************
 * MainFormLoadRecords:
 * Load the records that will be displayed
 * in the table.
 ******************************************
 */

static void MainFormLoadRecords(void)
{
	TablePtr tableP;

	MemHandle record;

	if (showAllCategories) currentCategory = dmAllCategories;

	tableP = GetActiveFormObjectPtr(tblMainRecordList);

	/* Try to show a full display of records 
	 *  and perform positioning on the table.
	 */
	MainFormShowFullDisplay();

	/* Load each row of the table with a record.
	 * If no record could be found, mark the row
	 * unusable.
	 */
	MainFormLoadTable();

	/* Set the table's custom draw procedure */
	TblSetCustomDrawProcedure(tableP, kRecordColumn, MainFormDrawTableItem);

	/* Mark the columns as usable otherwise
	 * the table won't be displayed by default.
	 */

	TblSetColumnUsable(tableP, kRecordColumn, true);

	/* Updates the scrollbars */
	MainFormUpdateScrollBar();
}

/******************************************
 * MainFormShowFullDisplay:
 * Show a full diplay in the main form.
 ******************************************
 */

static void MainFormShowFullDisplay(void)
{
	UInt16 recordNum;
	MemHandle previousRecord;
	TablePtr tableP;
	Int16 nRow, numRows = 0;
	UInt32 recordID;

	UInt16 lineHeight;
	UInt16 dataHeight = 0;
	UInt16 tableHeight;
	RectangleType	r;

	tableP = GetActiveFormObjectPtr(tblMainRecordList);

	TblGetBounds (tableP, &r);
	tableHeight = r.extent.y;

	if (currentRecord != noRecordSelected) {
	/* There is a current record. Make it visible.
	 * If the current record index is smaller than the top index
	 * of table, update the top index.
	 * Otherwise, find the record displayed at the bottom of the table
	 * If the record is before the current record then the 
	 * current record isn't displayed.  In that case make
	 * the current record the first record displayed in the table.
	 */

		if (currentDB == memoPadDB) {
			if (topRecord > currentRecord)
				topRecord = currentRecord;
			else {
				numRows = MainFormNumberOfRows (tableP);
				recordNum = topRecord;

				for (nRow = 1; nRow < numRows; nRow++)
					MainFormGetNextRecord(&recordNum);

				if ((currentDB == memoPadDB)
						&& (recordNum < currentRecord))
					topRecord = currentRecord;
			}
		}
	}

	/* Try to show a full display of records.  Starting at the last
	 * record and working backwards, find the record displayed at the 
	 * top of the table.  If the record is before the TopVisibleRecord
	 * then TopVisibleRecord is set too far down the list of records.
	 * Set TopVisibleRecord to the record one screen full from the end.
	 */

	if ((useSingleFont == separateFont)
		&& (currentCategory == dmAllCategories)) {

				recordNum = dmMaxRecordIndex;
				currentDB = cryptoPadDB;

				while (dataHeight <= tableHeight) {
					previousRecord = MainFormGetPreviousRecord(&recordNum);

					if (previousRecord == 0) break;

					if (useSingleFont == separateFont) {
						DmRecordInfo (currentDBRef, recordNum, 0, &recordID, 0);
						SetOrGetRecordFont(currentDBRef, recordID, getFont, fontTable);
					}
					FntSetFont (currentTableFont);
					lineHeight = FntLineHeight ();
					dataHeight += lineHeight;
				}

			if (previousRecord)
				MainFormGetNextRecord(&recordNum);
		
	} else {
		recordNum = dmMaxRecordIndex;
		currentDB = cryptoPadDB;

		if (!numRows)
			numRows = MainFormNumberOfRows (tableP);
		
		previousRecord = MainFormGetPreviousRecord(&recordNum);

		for (nRow = 1; nRow < numRows; nRow++) {
			previousRecord = MainFormGetPreviousRecord(&recordNum);

			if (previousRecord == 0) break;
		}
	}

	if (topDB == currentDB)
		topRecord = min(topRecord, recordNum);
	else if (topDB == cryptoPadDB) {
		topRecord = recordNum;
		topDB = currentDB;
	}
}

/******************************************
 * DecryptRecord:
 * Decrypt the current record.
 ******************************************
 */

void DecryptRecord(UInt16 recordNum, Char * passwordP)
{
	MemHandle handle;
	PackedRecordPtr packedP;
	RecordType record;

	UInt16 nOffset;


	handle = DmGetRecord(dbCryptoPadRef, recordNum);
	packedP = MemHandleLock(handle);
	UnpackRecord(&record, packedP);

	/* Finds the offset to start decryption */

	nOffset = 0;
	while (nOffset <= record.textLen) {
		if (record.textP[nOffset++] == '\n') break;
	}

	if (nOffset < record.textLen) {
		RecordType newRecord;

		/* Set the flag to reflect the status of the record */

		newRecord.flag      = recDecrypted;
		newRecord.passwdLen = 0;
		newRecord.textLen   = record.textLen;

		/* Get rid of the stored password */

		newRecord.passwordP = MemPtrNew(1);
		newRecord.passwordP[0] = '\0';

		/* Store the decrypted text in the record */

		newRecord.textP = MemPtrNew(record.textLen + 1);
		StrNCopy(newRecord.textP, record.textP, nOffset);
		Blowfish(record.textP + nOffset, newRecord.textP + nOffset,
					passwordP, newRecord.textLen - nOffset,
					BF_DECRYPT);

		/* Replaces the record with the new one */

		PackRecord(&newRecord, handle);

		/* Deallocates memory */

		MemPtrFree(newRecord.passwordP);
		MemPtrFree(newRecord.textP);
	}

	/* Release the record to the data manager */

	MemHandleUnlock(handle);
	DmReleaseRecord(dbCryptoPadRef, recordNum, true);
}

/******************************************
 * EncryptCategory:
 * Encrypt the current category.
 ******************************************
 */

static void EncryptCategory()
{
	MemHandle memoPadHandle;

	UInt16 recMemoAttributes, recCryptoAttributes;
	UInt16 i, numRecInCategory;

	Char * textP;
	UInt16 nOffset;

	FormPtr formP;
    ProgressInfoPtr progrInfoP;
    ProgrType progr;
	UInt16 progrinc;

	Char * strProgressP;
	MemHandle strProgressH;


	numRecInCategory = DmNumRecordsInCategory(dbMemoPadRef, currentCategory);
	if (numRecInCategory == 0) {
		FrmAlert(altCategoryAlreadyEncryptedAlert);
		RemoveCurrentPassword();
		return;
	}

	formP = FrmInitForm(frmProgressForm);

	// Display the title of the progress form.
	strProgressH = DmGetResource('tSTR', strProgressEncryption);
	strProgressP = MemHandleLock(strProgressH);
		
	FrmSetTitle (formP, strProgressP);

	MemHandleUnlock(strProgressH);   
	DmReleaseResource(strProgressH); 

	FrmDrawForm(formP);

	currentRecord = 0;


	progrInfoP = ProgressFormInitAndDraw (formP);

	for (i = 0; i < numRecInCategory; i++) {
		progr.major.soFar = 0;
		progr.major.total = numRecInCategory;
		progr.minor.soFar = 0;
		progr.minor.total = 0;
		progrinc = i ;
		progr.major.soFar = progrinc;

		/* Seek the item after the currentRecord index in MemoPad database.*/
		memoPadHandle = DmQueryNextInCategory(dbMemoPadRef, &currentRecord, currentCategory);
		textP = MemHandleLock(memoPadHandle);

		/* Validate the format of the record
		 * in checking if there is 2 lines.
		 */
		nOffset = 0;
		while (nOffset <= StrLen(textP)) {
			if (textP[nOffset++] == '\n') break;
		}

		progrInfoP->statusFunc (progrInfoP, &progr, textP);

		if (nOffset < StrLen(textP)) {
			MemHandle cryptoPadHandle;
			RecordType newRecord;
			UInt16 index;
			UInt32 memoRecordID;
			UInt16 textLength, passwordLength;

			/* Creates a new record at the end of the database */
			index = dmMaxRecordIndex;

			cryptoPadHandle = DmNewRecord(dbCryptoPadRef, &index, sizeof(PackedRecordType));
			ErrFatalDisplayIf(cryptoPadHandle == 0, "Unable to add record to the database.");

			/* Current record now contains the record index.
			 * Sets an initial value for the record.
			 */
	
			passwordLength = StrLen(currentPasswordP);
			textLength = StrLen(textP);

			newRecord.flag      = recDecrypted;
			newRecord.passwdLen =  passwordLength;
			newRecord.textLen   = textLength;
			newRecord.passwordP = currentPasswordP;
			newRecord.textP = textP;

			PackRecord(&newRecord, cryptoPadHandle);

			DmReleaseRecord(dbCryptoPadRef, index, true);

			/* Set the category of the record */
			DmRecordInfo(dbMemoPadRef, currentRecord, &recMemoAttributes, &memoRecordID, 0);

			DmSetRecordInfo(dbCryptoPadRef, index, &recMemoAttributes, 0);

	if (useSingleFont == separateFont) {
		UInt32 cryptoRecordID;

		DmRecordInfo (dbCryptoPadRef, index, 0, &cryptoRecordID, 0);
		GetTableFontOfCategory(recMemoAttributes & dmRecAttrCategoryMask);
		CreateFontRecord(dbCryptoPadRef, cryptoRecordID);
		SetOrGetRecordFont(dbMemoPadRef, memoRecordID, getFont, fontField);
		SetOrGetRecordFont(dbCryptoPadRef, cryptoRecordID, setFont, fontField);
	}

			/* Delete the record from MemoPadDB */
			DmDeleteRecord (dbMemoPadRef, currentRecord);
			DmMoveRecord (dbMemoPadRef, currentRecord, DmNumRecords (dbMemoPadRef));

			/* Encrypt the record */
			EncryptRecord(index, currentPasswordP);
		} else {
			currentRecord++;
			MemHandleUnlock(memoPadHandle);
		}
	}

	progr.major.soFar = progr.major.total;
	progrInfoP->statusFunc (progrInfoP,&progr,NULL);

	RemoveCurrentPassword();

	if (currentSortOrder == soAlphabetic) {
		DmQuickSort(dbCryptoPadRef, 
			(DmComparF*) &CompareCryptoPadRecords, (Int16) currentSortOrder);
		DmQuickSort(dbMemoPadRef, 
			(DmComparF*) &CompareMemoPadRecords, (Int16) currentSortOrder);
	}

	currentRecord = 0;
	currentDB = cryptoPadDB;

	FrmDeleteForm(formP);
	FrmGotoForm(frmMainForm);
}

/******************************************
 * DecryptCategory:
 * Decrypt the current category.
 ******************************************
 */

static void DecryptCategory()
{
	UInt16 i, numRecInCategory;

	FormPtr formP;
    ProgressInfoPtr progrInfoP;
    ProgrType progr;
	UInt16 progrinc;

	Char * strProgressP;
	MemHandle strProgressH;


	numRecInCategory = DmNumRecordsInCategory(dbCryptoPadRef, currentCategory);
	if (numRecInCategory == 0) {
		FrmAlert(altCategoryAlreadyDecryptedAlert);
		RemoveCurrentPassword();
		return;
	}

	formP = FrmInitForm(frmProgressForm);

	// Display the title of the progress form.
	strProgressH = DmGetResource('tSTR', strProgressDecryption);
	strProgressP = MemHandleLock(strProgressH);
		
	FrmSetTitle (formP, strProgressP);

	MemHandleUnlock(strProgressH);   
	DmReleaseResource(strProgressH); 


	FrmDrawForm(formP);


	currentRecord = 0;

	progrInfoP = ProgressFormInitAndDraw (formP);

	for (i = 0; i < numRecInCategory; i++) {

		progr.major.soFar = 0;
		progr.major.total = numRecInCategory;
		progr.minor.soFar = 0;
		progr.minor.total = 0;
		progrinc = i ;
		progr.major.soFar = progrinc;

		progrInfoP->statusFunc (progrInfoP,&progr,NULL);

		/* Seek the item after the currentRecord index in MemoPad database.*/
		DmSeekRecordInCategory (dbCryptoPadRef, &currentRecord, 0, dmSeekForward, currentCategory);

		if (ValidatePassword(currentPasswordP)) {
			MemHandle cryptoPadHandle;
			UInt16 recMemoAttributes, recCryptoAttributes;
			UInt32  cryptoRecordID;
			UInt16 index;
			UInt16 nLength;

			MemHandle memoPadHandle;
			PackedRecordPtr packedP;
			RecordType record;
			MemPtr		p;

			if (isEncrypted(currentRecord))
				DecryptRecord(currentRecord, currentPasswordP);

			cryptoPadHandle = DmGetRecord(dbCryptoPadRef, currentRecord);

			packedP = MemHandleLock(cryptoPadHandle);
			UnpackRecord(&record, packedP);
			MemHandleUnlock(cryptoPadHandle);
			DmReleaseRecord(dbCryptoPadRef, currentRecord, false);

			progrInfoP->statusFunc (progrInfoP,&progr,record.textP);

			/* Create a new record in the MemoPad database */
			index = dmMaxRecordIndex;
			nLength = StrLen(record.textP);
	
			memoPadHandle = DmNewRecord(dbMemoPadRef, &index, nLength + 1);
			ErrFatalDisplayIf(memoPadHandle == 0, "Unable to add record to the database.");

			/* Lock down the block containing the new record */
			p = MemHandleLock(memoPadHandle);

			/* Write the field's text to the database record */
			DmWrite(p, 0, record.textP, StrLen(record.textP) + 1);

			/* Unlock the block of the new record */
			MemPtrUnlock(p);
			
			DmReleaseRecord(dbMemoPadRef, index, true);
			
			/* Set the category of the record */
			DmRecordInfo(dbCryptoPadRef, currentRecord, &recCryptoAttributes, &cryptoRecordID, 0);

			DmSetRecordInfo(dbMemoPadRef, index, &recCryptoAttributes, 0);

	if (useSingleFont == separateFont) {
		UInt32 memoRecordID;

		DmRecordInfo (dbMemoPadRef, index, 0, &memoRecordID, 0);
		GetTableFontOfCategory(recCryptoAttributes & dmRecAttrCategoryMask);
		CreateFontRecord(dbMemoPadRef, memoRecordID);
		SetOrGetRecordFont(dbCryptoPadRef, cryptoRecordID, getFont, fontField);
		SetOrGetRecordFont(dbMemoPadRef, memoRecordID, setFont, fontField);

		// Delete the record in the Font database.
		DeleteFontRecord(dbCryptoPadRef, cryptoRecordID);
	}

			/* Delete the record from CryptoPadDB */
			DmRemoveRecord(dbCryptoPadRef, currentRecord);
		} else {
			currentRecord++;
		}
	}

	progr.major.soFar = progr.major.total;
	progrInfoP->statusFunc (progrInfoP,&progr,NULL);

	RemoveCurrentPassword();

	if (currentSortOrder == soAlphabetic) {
		DmQuickSort(dbCryptoPadRef, 
			(DmComparF*) &CompareCryptoPadRecords, (Int16) currentSortOrder);
		DmQuickSort(dbMemoPadRef, 
			(DmComparF*) &CompareMemoPadRecords, (Int16) currentSortOrder);
	}

	currentRecord = noRecordSelected;
	currentDB = memoPadDB;

	FrmDeleteForm(formP);
	FrmGotoForm(frmMainForm);
}

/*****************************************
 * ProgressFormInitAndDraw:
 * Init and draw the progress bar
 * for encryption/decryption of
 * the category. 
 *****************************************
 */
static ProgressInfoPtr ProgressFormInitAndDraw (FormPtr frmP)
{
	ProgressInfoPtr infoP = NULL;

	ErrNonFatalDisplayIf (!frmP, "nil form");


	// Allocate and initialize the Progress Info structure
	infoP = MemPtrNew (sizeof (ProgressInfoType));
	ErrNonFatalDisplayIf (!infoP, "alloc failed");
	if (!infoP)
		return (NULL);

	MemSet (infoP, sizeof (*infoP), 0);

	// Get bounds of the progress message and gauge objects once
	FrmGetObjectBounds (frmP, FrmGetObjectIndex (frmP, gadProgressMessage),
			&infoP->messageR);

	WinDrawGrayRectangleFrame(simpleFrame, &infoP->messageR);

	FrmGetObjectBounds (frmP, FrmGetObjectIndex (frmP, gadProgressBar),
					  &infoP->gaugeR);


	// Init the rest of the progress info structure
	infoP->progressForm = frmP;
	infoP->gaugeWidth = infoP->gaugeR.extent.x;
	infoP->gaugeLevel = 0;
	infoP->statusFunc = ProgressCallback;

	// Draw the form
	FrmDrawForm (frmP);

	// Initialize the gauge
	ProgressUpdateGauge (&infoP->gaugeR, 0);

	return (infoP);
}

/***************************************
 * ProgressCallback:
 * callback function for
 * the progress bar.
 ***************************************
 */
static Boolean ProgressCallback (void* progressInfoP, 
		ProgrType* progressP, Char* messageP)
{
	UInt16		gaugeLevel = 0;
	ProgressInfoPtr infoP = (ProgressInfoPtr) progressInfoP;
	UInt16		itemWidth = 0;


	// Process any pending events, such as those for our Stop button
 
	// Reset the auto-off timer to make sure we don't fall asleep in the
	//  middle of the update
	EvtResetAutoOffTimer ();

	while (ProgressHandleEvent (0 /*timeout*/) != nilEvent)
		{/*intentionally left blank*/}


	// Update the progress gauge
	if (progressP) {
		// Update the progress gauge counts
		ErrNonFatalDisplayIf (!progressP->major.total, "major denominator is zero");

		if (progressP->major.total)
			gaugeLevel = (UInt16) ((progressP->major.soFar * infoP->gaugeWidth)
								  / progressP->major.total);

		if (progressP->minor.soFar) {
			ErrNonFatalDisplayIf (!progressP->minor.total, "minor denominator is zero");

			if (progressP->major.total)
				itemWidth = (UInt16)(infoP->gaugeWidth / progressP->major.total);
			if (itemWidth && progressP->minor.total)
				gaugeLevel += (UInt16) ((progressP->minor.soFar * itemWidth)
									  / progressP->minor.total);
		}

		if (gaugeLevel != infoP->gaugeLevel) {
			infoP->gaugeLevel = gaugeLevel;
		  
		ProgressUpdateGauge (&infoP->gaugeR, gaugeLevel);
	}

} 


	// Update the progress message
	if (messageP) {
		FontID	currFont;

		// Switch to bold, saving the current font
		currFont = FntSetFont (stdFont);

		WinEraseRectangle (&infoP->messageR, 0 /*cornerDiam*/);

		DrawStringWithEllipsis (messageP, infoP->messageR.topLeft.x, 
			infoP->messageR.topLeft.y, infoP->messageR.extent.x, justCenter);

		// Restore font
		FntSetFont (currFont);
	}

	return (false);

} 

/**************************************
 * ProgressUpdateGauge:
 * Update the progress indicator.
 **************************************
 */
static void ProgressUpdateGauge (RectangleType*	gaugeRP, 
										UInt32 gaugeLevel)
{
	RectangleType 	eraseR;
	RectangleType 	completeR;

  
	ErrNonFatalDisplayIf (!gaugeRP, "null arg");


	// Draw the frame of gadget
	WinDrawRectangleFrame (simpleFrame, gaugeRP);
  

	// Fill in the gauge.
	if (gaugeLevel > (UInt32)gaugeRP->extent.x) {
		ErrDisplay ("gaugeLevel out of bounds");
		gaugeLevel = gaugeRP->extent.x;
	}

	completeR.topLeft.x = gaugeRP->topLeft.x;
	completeR.topLeft.y = gaugeRP->topLeft.y;
	completeR.extent.x = (UInt16)gaugeLevel;
	completeR.extent.y = gaugeRP->extent.y;
	WinDrawRectangle (&completeR, 0);
  

	// Erase the empty portion of the gauge (not really necessary).
	eraseR.topLeft.x = (UInt16)(gaugeRP->topLeft.x + gaugeLevel);
	eraseR.topLeft.y = gaugeRP->topLeft.y;
	eraseR.extent.x = (UInt16)(gaugeRP->extent.x - gaugeLevel);
	eraseR.extent.y = gaugeRP->extent.y;
	WinEraseRectangle (&eraseR, 0);

	return;
}

/******************************************
 * MainFormGetNextCategory:
 * find the next category where there is
 * either a record from MemoPadDB or
 * from CryptoPad
 ******************************************
 */

static UInt16 MainFormGetNextCategory(UInt16 category)
{
	UInt16 nextCryptoPadDBCategory;
	UInt16 nextMemoPadDBCategory;
	
	Char nextCryptoPadCategoryName[dmCategoryLength];
	Char nextMemoPadCategoryName[dmCategoryLength];

	Int16 cmpNextMemoAndCryptoCategory;
	Boolean goToUnfiled = (((DmNumRecordsInCategory (dbMemoPadRef, dmUnfiledCategory)
		+ DmNumRecordsInCategory (dbCryptoPadRef, dmUnfiledCategory)) != 0)
		&& (category != dmUnfiledCategory));

	nextCryptoPadDBCategory = CategoryGetNext(dbCryptoPadRef, category);
	nextMemoPadDBCategory = CategoryGetNext(dbMemoPadRef, category);

	CategoryGetName(dbCryptoPadRef, nextCryptoPadDBCategory, nextCryptoPadCategoryName);
	CategoryGetName(dbMemoPadRef, nextMemoPadDBCategory, nextMemoPadCategoryName);

	if (nextCryptoPadDBCategory == dmAllCategories)	{
		if (currentRotateOrder != roAll) {
		/* The rotate order is either Categories-Unfiled
		 *  or All-Unfiled.
		 */
			if (nextMemoPadDBCategory == dmAllCategories) {
				if (goToUnfiled) return dmUnfiledCategory;
		
				// The rotate order is Categories-Unfiled.
				if ((currentRotateOrder == roCategoriesUnfiled)
					&& !(goToUnfiled))
						return MainFormGetNextCategory(dmAllCategories);
			}
		}
		
		return nextMemoPadDBCategory;
	}

	if (nextMemoPadDBCategory == dmAllCategories) 
		return nextCryptoPadDBCategory;

	cmpNextMemoAndCryptoCategory = StrCompare(nextCryptoPadCategoryName, 
		nextMemoPadCategoryName);

	if (cmpNextMemoAndCryptoCategory < 0) return nextCryptoPadDBCategory;
	else return nextMemoPadDBCategory;
}


/******************************************
 * MainFormNextCategory:
 * switches to the next category view
 ******************************************
 */

static void MainFormNextCategory(void)
{
	UInt16 category;

	category = MainFormGetNextCategory(currentCategory);

	if (category != currentCategory) {
		TablePtr tableP;
		ControlPtr controlP;
		FormPtr formP;

		if (category == dmAllCategories) showAllCategories = true;
		else showAllCategories = false;

		ChangeCategory(category);		

		/* Set the label of the category trigger */
		controlP = GetActiveFormObjectPtr(popMainCategory);
		CategoryGetName(dbMemoPadRef, currentCategory, categoryName);
		CategorySetTriggerLabel(controlP, categoryName);

		/* Display the new category */
		formP = FrmGetActiveForm ();
		if (useSingleFont == separateFont) {
			GetTableFontOfCategory(currentCategory);
			FntSetFont (currentTableFont);
		}

		MainFormLoadTable ();
		tableP = GetActiveFormObjectPtr(tblMainRecordList);
		TblEraseTable (tableP);
		TblDrawTable (tableP);
	}
}

/******************************************
 * MainFormSelectCategory:
 * Invokes the standard category UI
 ******************************************
 */

static void MainFormSelectCategory(void)
{
	Boolean categoryUpdated;
	UInt16 category;

	FormPtr formP;
	TablePtr tableP;

	// Process the category popup list. 
	category = currentCategory;
	
	formP = FrmGetActiveForm();
	categoryUpdated = CategorySelect(dbMemoPadRef, formP,
			popMainCategory, lstMainCategory, true,
			&category, categoryName, 1, 0);

	if (category == dmAllCategories) showAllCategories = true;
	else showAllCategories = false;

	if (categoryUpdated || (category != currentCategory)) {
		/* The category was changed, renamed or merged to
		 * another category.
		 */
		ChangeCategory(category);

		/* Display the new category */
		FrmUpdateForm(frmMainForm, updateRedrawAll);	
	}
}

/******************************************
 * MainFormUpdateDisplay:
 * Updates the display of the main form.
 ******************************************
 */

static Boolean MainFormUpdateDisplay(UInt16 updateCode)
{	
	TablePtr tableP;

	if (updateCode & (updateTable | 
			updateFontChanged | updateRedrawAll)) {
		if (updateCode & updateTable) {
			topRecord = 0;
			topDB = memoPadDB;
		} 

		if (updateCode & updateFontChanged) {
			if (useSingleFont == separateFont)
				SynchroniseTableFontForCurrentCategory();
		}
		
		MainFormLoadRecords();
		tableP = GetActiveFormObjectPtr(tblMainRecordList);
		TblEraseTable(tableP);
		TblDrawTable(tableP);		

		return (true);
	}
		
	return (false);
}

/******************************************
 * MainFormUpdateScrollBar:
 * Updates the size and position of the
 * scroll bar object.
 ******************************************
 */

static void MainFormUpdateScrollBar(void)
{
	FormPtr formP;
	TablePtr tableP;
	ScrollBarPtr scrollBarP;
	Int16 nRows;
	UInt16 nPosition, maxValue;

	tableP = GetActiveFormObjectPtr(tblMainRecordList);
	scrollBarP = GetActiveFormObjectPtr(sclMainScrollBar);

	nRows = MainFormNumberOfRows(tableP);

	if (memosInCategory < nRows) {
		maxValue = 0;
		nPosition = 0;
	}
	else {
		if (topDB == cryptoPadDB) {
			nPosition = (DmNumRecordsInCategory(dbMemoPadRef, currentCategory)
				+ DmPositionInCategory(dbCryptoPadRef, topRecord, currentCategory));
		} else {
			nPosition = DmPositionInCategory(dbMemoPadRef, topRecord, currentCategory);
		}

		maxValue = memosInCategory - nRows;
	}

	SclSetScrollBar(scrollBarP, nPosition, 0, maxValue, nRows);
}

/******************************************
 * MainFormScrollLines:
 * Scrolls the table nLines up (negative)
 * or down (positive).
 ******************************************
 */

static void MainFormScrollLines(Int16 nLines)
{
	TablePtr tableP;
	Int16 direction;
	Int16 iLines;

	UInt16 nRow;
	Int16 recordNum;
	MemHandle record;

	if (nLines < 0) {
		nLines = -nLines;
		direction = dmSeekBackward;
	} else direction = dmSeekForward;
	
	recordNum = topRecord;
	currentDB = topDB;

	for (nRow = 0; nRow < nLines; nRow++) {
		if (direction == dmSeekForward) 
			record = MainFormGetNextRecord(&recordNum);
		else record = MainFormGetPreviousRecord(&recordNum);

		if (record == 0) break;
	}

	topRecord = recordNum;
	topDB = currentDB;

	currentRecord = noRecordSelected;

	tableP = GetActiveFormObjectPtr(tblMainRecordList);
	MainFormLoadRecords();

	TblRedrawTable(tableP);
}


/******************************************
 * MainFormScrollPage:
 * Scrolls the table one full page in
 * the given direction.
 ******************************************
 */

static void MainFormScrollPage(WinDirectionType direction)
{
	TablePtr tableP;
	Int16 nRows;

	tableP = GetActiveFormObjectPtr(tblMainRecordList);
	nRows = MainFormNumberOfRows(tableP);

	if (direction == winUp)
		MainFormScrollLines(-nRows + 1);
	else MainFormScrollLines(nRows - 1);
}


/******************************************
 * MainFormHandleMenuEvent:
 * Process menu commands.
 ******************************************
 */

static void MainFormHandleMenuEvent(UInt16 itemID)
{
	UInt32 currentRecordID;
	Boolean wasHiding;
	UInt16 mode;
	EventType event;
	FormPtr formP;

Int16 index = 0, i, recAttributes;
USES_ALERT;
Err err;

	switch (itemID) {
		case itemRecordEncryptCategory:
		case itemRecordDecryptCategory:
			if ((prefsGlobalPasswordP != 0) && (globalPasswordP == 0)) {
				event.eType = globalPasswordStatusEvent;
				event.data.generic.datum[1] = queryGlobalPassword;
				if (itemID == itemRecordEncryptCategory)
					event.data.generic.datum[2] = catDecrypted;
				else event.data.generic.datum[2] = catEncrypted;

				FrmPopupForm(frmQueryGlobalPasswordForm);
				EvtAddEventToQueue(&event);
			} else {
				event.eType = memoStatusEvent;
				if (itemID == itemRecordEncryptCategory)
					event.data.generic.datum[1] = catDecrypted;
				else event.data.generic.datum[1] = catEncrypted;

				FrmPopupForm(frmPasswordForm);
				EvtAddEventToQueue(&event);
			}
			break;

		case itemOptionsFont:
			if (useSingleFont == separateFont) {
				if (currentCategory == dmAllCategories) {
					FrmAlert(altChangeTableFontOfCategoryAllAlert);
					break;
				} else GetTableFontOfCategory(currentCategory);
			}

			currentTableFont = SelectFont (currentTableFont);
			break;

		case itemRecordBeamCategory:
			MemoSendCategory(dbMemoPadRef, currentCategory);
			break;

		case itemOptionsPrefs:
			FrmPopupForm(frmPrefsForm);
			break;

		case itemOptionsSecurity:
			wasHiding = (PrivateRecordVisualStatus == hidePrivateRecords);
			 
			PrivateRecordVisualStatus = SecSelectViewStatus();
			
			if (wasHiding ^ (PrivateRecordVisualStatus == hidePrivateRecords)) 
			//xor on two logical values - mode to open DB has changed
				{
				// Close the application's data file.
				SetApplicationPrefs(0);
				DmCloseDatabase(dbMemoPadRef);
				DmCloseDatabase(dbCryptoPadRef);
			
				mode = (PrivateRecordVisualStatus == hidePrivateRecords) ?
					dmModeReadWrite : (dmModeReadWrite | dmModeShowSecret);
				
				dbCryptoPadRef = OpenOrCreateCryptoDatabase(mode);
				dbMemoPadRef = OpenMemoDatabase(mode);

				// Read the preferences.
				GetApplicationPrefs(&currentRecordID);
			}
			
			//For safety, simply reset the currentRecord
			currentRecord = noRecordSelected;
			MainFormUpdateDisplay(updateTable);
			break;

		case itemOptionsAbout:
			formP = FrmInitForm(frmAboutForm);
			FrmDoDialog(formP);
			FrmDeleteForm(formP);
			break;

	}
}

/*************************************************
* ProgressHandleEvent:
* Get event from the progress
* bar form.
**************************************************
*/
static UInt32 ProgressHandleEvent (Int32 timeout)
{
	EventType event;
	UInt16 error;

	EvtGetEvent (&event, timeout);

	return (event.eType);
}

Boolean MainFormHandleEvent(EventPtr eventP)
{
	Boolean handled = false;

	switch (eventP->eType) {
		/* A request to decrypt the record has
		 * been made. Swith to the Edit form
		 * to allow modification of the memo.
		 */
		case memoProcessEvent:
			ErrFatalDisplayIf(currentPasswordP == 0, "Password shouldn't be NULL");
			DecryptRecord(currentRecord, currentPasswordP);
			FrmGotoForm(frmEditForm);
			handled = true;
			break;

		case catProcessEvent:
			ErrFatalDisplayIf(currentPasswordP == 0, "Password shouldn't be NULL");
			if (eventP->data.generic.datum[1] == catDecrypted)
				EncryptCategory();
			else DecryptCategory();
			handled = true;
			break;

		/* We need to reload the records because a
		 * record has been deleted as a result of
		 * a forgotten password.
		 */
		case frmUpdateEvent:
			handled = MainFormUpdateDisplay (eventP->data.frmUpdate.updateCode);
			break;

		case ctlSelectEvent:
			switch (eventP->data.ctlSelect.controlID) {
				/* Creates a new memo and switch to the edit form */
				case btnMainNewMemo:
					if (CreateRecord())
						FrmGotoForm(frmEditForm);
					handled = true;
					break;

				/* Changes the currently selected category */
				case popMainCategory:
					MainFormSelectCategory();
					handled = true;
					break;
			}
			break;

		/* A record has been tapped in the table 
		* or drag the memo with the stylus 
		*/
		case tblEnterEvent:
			MainFormDragMemo(eventP);
			handled = true;
			break;

		/* Scrolls the records */
		case sclRepeatEvent:
			MainFormScrollLines(eventP->data.sclRepeat.newValue -
						eventP->data.sclRepeat.value);
			/* Let the system repeat the event */
			break;


		case keyDownEvent:
			// Memo button pressed?
			if (TxtCharIsHardKey(eventP->data.keyDown.modifiers, eventP->data.keyDown.chr)) {
				if (!(eventP->data.keyDown.modifiers & poweredOnKeyMask))
					MainFormNextCategory ();
				handled = true;
			}

			else if (EvtKeydownIsVirtual(eventP)) {
				// Scroll up key presed?
				if (eventP->data.keyDown.chr == vchrPageUp) {
					MainFormScrollPage(winUp);
					handled = true;
				}
	
				// Scroll down key presed?
				else if (eventP->data.keyDown.chr == vchrPageDown) {
					MainFormScrollPage(winDown);
					handled = true;
				}
	
				// Send Data key presed?
				else if (eventP->data.keyDown.chr == vchrSendData) {
					MainFormHandleMenuEvent(itemRecordBeamCategory);
					handled = true;
				}
			}
		
			// If printable character, create a new record.
			else if (TxtCharIsPrint (eventP->data.keyDown.chr)) {
				if (CreateRecord ()) {
					WChar c;

					c =	eventP->data.keyDown.chr;
					if (c >= 'a' && c <= 'z')
						c -= ('a' - 'A');

					eventP->data.keyDown.chr = c;

					FrmGotoForm (frmEditForm);
					EvtAddEventToQueue (eventP);
				}
				handled = true;
			}
			break;

		case menuEvent:
			MainFormHandleMenuEvent(eventP->data.menu.itemID);
			return (true);
			break;

		case menuOpenEvent:
			if (OSVersion >= OS35) {
				MemHandle menuItemH;
				Char * menuItemP;

				menuItemH = DmGetResource('tSTR', strMenuSecurityItem);
				menuItemP = MemHandleLock(menuItemH);

				MenuAddItem (itemOptionsPrefs, itemOptionsSecurity, 'S', menuItemP);

				MemHandleUnlock(menuItemH);
				DmReleaseResource(menuItemH);
			}
			break;

		case menuCmdBarOpenEvent:
			MenuCmdBarAddButton(menuCmdBarOnLeft, BarSecureBitmap, menuCmdBarResultMenuItem, itemOptionsSecurity, 0);
			MenuCmdBarAddButton(menuCmdBarOnLeft, bmpBarFontBitmap, menuCmdBarResultMenuItem, itemOptionsFont, 0);

			// tell the field package to not add buttons automatically; we've done it all ourselves.
			eventP->data.menuCmdBarOpen.preventFieldButtons = true;
		
			// don't set handled to true; this event must fall through to the system.
			break;

		case frmOpenEvent:
			MainFormInit();
			FrmDrawForm(FrmGetActiveForm());
			handled = true;
			break;
	}

	return handled;
}
