#include "PasswordForm.h"
#include "Database.h"
#include "Utils.h"
#include "blowfish.h"

static void PasswordApplyChanges(Char *);

static Boolean alreadyTriedWithGlobalPassword;
static Int16 encryptedStatus;

static Char * charP;

/******************************************
 * ValidatePassword:
 * Checks that the encrypted password
 * of the current record matches the
 * given string.
 ******************************************
 */

Boolean ValidatePassword(Char * textP)
{
	MemHandle handle;
	PackedRecordPtr packedP;
	RecordType record;

	Boolean bResult;
	Char * passwordP;

	if (currentRecord == noRecordSelected) return;

	/* Get the record */

	handle = DmGetRecord(currentDBRef, currentRecord);
	packedP = MemHandleLock(handle);
	UnpackRecord(&record, packedP);

	/* Decrypts the password using textP as the key,
	 * and checks that we find textP itself!
	 */

	passwordP = MemPtrNew(record.passwdLen + 1);
	Blowfish(record.passwordP, passwordP, textP, record.passwdLen, BF_DECRYPT);

	bResult = (StrCompare(passwordP, textP) == 0);

	/* Release the record to the data manager */

	MemHandleUnlock(handle);
	DmReleaseRecord(currentDBRef, currentRecord, false);

	return bResult;
}


/******************************************
 * PasswordApplyChanges:
 * Performs the appropriate encryption/
 * decryption handling according to the
 * encryptedStatus.
 ******************************************
 */

static void PasswordApplyChanges(Char * textP)
{
	EventType event;
	FormPtr formP;
	UInt16 fieldIndex;

	if ((encryptedStatus == recEncrypted) && !ValidatePassword(textP)) {
		if ((globalPasswordP != 0) && 
			!alreadyTriedWithGlobalPassword) {
				FrmAlert(altUseNoGlobalPasswordAlert);
				alreadyTriedWithGlobalPassword = true;
				formP = FrmGetActiveForm();
				fieldIndex = FrmGetObjectIndex(formP, fldPasswordText);

				FrmDrawForm(formP);
				FrmSetFocus(formP, fieldIndex);
		} 
		else FrmAlert(altInvalidPasswordAlert);
		return;
	}

	if (globalPasswordP != 0) {
		if (encryptedStatus == recDecrypted)
			FrmAlert(altEncryptRecordWithGlobalPasswordAlert);
	
		if (encryptedStatus == catDecrypted)
			FrmAlert(altEncryptCategoryWithGlobalPasswordAlert);
	}

	/* Sets the currentPasswordP variable and
	 * return to the calling form. Notify the
	 * calling form to immediately process the
	 * record by sending a memoProcess event.
	 */

	currentPasswordP = MemPtrNew(StrLen(textP) + 1);
	StrCopy(currentPasswordP, textP);

	if ((encryptedStatus == recEncrypted) || 
					(encryptedStatus == recDecrypted))
		event.eType = memoProcessEvent;
	else {
		event.eType = catProcessEvent;
		event.data.generic.datum[1] = encryptedStatus;
	}

	FrmReturnToForm(0);
	EvtAddEventToQueue(&event);
}


Boolean PasswordFormHandleEvent(EventPtr eventP)
{
	Boolean handled = false;

	FormPtr formP;
	FieldPtr fieldP;
	UInt16 fieldIndex;

	Char * textP;

	formP = FrmGetActiveForm();
	fieldIndex = FrmGetObjectIndex(formP, fldPasswordText);
	fieldP = FrmGetObjectPtr(formP, fieldIndex);

	switch (eventP->eType) {
		case memoStatusEvent:
			encryptedStatus = eventP->data.generic.datum[1];
			alreadyTriedWithGlobalPassword = false;

			if (globalPasswordP == 0) {
				/* Hide the 'forget...' button if the memo
				 * is not already encrypted.
				 */
				if (encryptedStatus != recEncrypted) {
					FrmHideObject(formP,
						FrmGetObjectIndex(formP, btnPasswordForget));
				}
			} else PasswordApplyChanges(globalPasswordP);

			handled = true;
			break;

		case ctlSelectEvent:
			switch (eventP->data.ctlSelect.controlID) {
				/* Apply changes to the password form */
				case btnPasswordOK:
					textP = FldGetTextPtr(fieldP);

					if (textP != 0 && *textP != '\0')
						PasswordApplyChanges(textP);
					else FrmAlert(altInvalidPasswordAlert);
					handled = true;
					break;

				/* Simply return to the calling form */
				case btnPasswordCancel:
					FrmReturnToForm(0);
					handled = true;
					break;

				/* Without the password, we cannot restore
				 * a memo. Deletes the memo and ask the
				 * Main form to reload the table.
				 */

				case btnPasswordForget:
					if (FrmAlert(altConfirmDeleteAlert) == 0 /* OK button */) {
						if (useSingleFont == separateFont) {
							UInt32 recordID;

							DmRecordInfo (currentDBRef, currentRecord, 0, &recordID, 0);
							DeleteFontRecord(currentDBRef, recordID);
						}
						DmRemoveRecord(currentDBRef, currentRecord);
						FrmReturnToForm(frmMainForm);
						FrmUpdateForm(frmMainForm, updateTable);
					}
					handled = true;
					break;
			}
			break;

		case frmOpenEvent:
			if (globalPasswordP == 0) {
				FrmDrawForm(FrmGetActiveForm());
				FrmSetFocus(formP, fieldIndex);
			}
			handled = true;
			break;
	}

	return handled;
}

