#include <Pilot.h>
#include "beam.h"
#include "diddle.h"

extern data d;
extern pref p;

void BeamReceive(ExgSocketPtr sockP, Boolean appIsActive) {
    DmOpenRef dbR;
    UInt num_records, pos, record_size;
    UInt bytes_to_get, bytes_got;
    VoidHand t;
    VoidPtr ptr;
    Char buf[BEAM_BUF_SIZE];
    Err error;

    /* Open the database */
    if (sockP->appData) {
        /* named sketch... */
        dbR = DmOpenDatabaseByTypeCreator(DBIType, AppType, dmModeReadWrite);
        if (!dbR) {
            FrmCustomAlert(GenericError,
                            "Diddle Could not open Indexed Database", "", "");
            return;
        }
        record_size = 3360;
    }
    else {
        /* scratch sketch... */
        dbR = DmOpenDatabaseByTypeCreator(DBType, AppType, dmModeReadWrite);
        if (!dbR) {
            FrmCustomAlert(GenericError,
                            "Diddle Could not open Scratch Database", "", "");
            return;
        }
        record_size = 3200;
    }

    /* Create the new record */
    num_records = dmMaxRecordIndex;
    t = DmNewRecord(dbR, &num_records, record_size);
    if (!t) FrmCustomAlert(GenericError, "Diddle could not create new record",
                            "", "");
    if (!t) return;
    ptr = MemHandleLock(t);

    DmSet(ptr,0,3200,0);
    
    /* Open the socket and start receiving */
    error = ExgAccept(sockP);
    if (!error) {
        pos = 0;
        while (1) {
            bytes_to_get = (((record_size-pos) < BEAM_BUF_SIZE) ?
                                record_size-pos : BEAM_BUF_SIZE);
	    if( !bytes_to_get ) break;
            bytes_got = ExgReceive(sockP, buf, bytes_to_get, &error);

            if (error) break;

            DmWrite(ptr, pos, buf, bytes_got);
            pos += bytes_got;
        }
        ExgDisconnect(sockP, error);
    }
    else {
        FrmCustomAlert(GenericError, "Diddle could not accept the socket",
                        "", "");
    }

    /* Cleanup */
    MemHandleUnlock(t);
    DmReleaseRecord(dbR, num_records, true);
    DmCloseDatabase(dbR);
}

/* 
** BeamSendSketch - do the actual beaming for BeamSend()
*/
void BeamSendSketch(ExgSocketPtr sockP) {
    Char buf[BEAM_BUF_SIZE];
    UInt pos, bytes_sent, bytes_to_send, record_size;
    VoidHand t;
    VoidPtr ptr;
    short dbI;
    DmOpenRef dbR;
    Err error;

    /* Use the current sketch */
    dbI = (p.onscratch ? p.dbI : p.dbII);
    dbR = (p.onscratch ? d.dbR : d.dbRI);
    record_size = sockP->length;
    /* Open the record for reading */
    t = DmQueryRecord(dbR, dbI);
    ptr = MemHandleLock(t);

    /* Open the socket and start sending */
    error = ExgPut(sockP);
    if (!error) {
        pos = 0;
        while (1) {
            bytes_to_send = (((record_size-pos) < BEAM_BUF_SIZE) ?
                                record_size-pos : BEAM_BUF_SIZE);
	    if( !bytes_to_send ) break;
            MemMove(buf, ptr+pos, bytes_to_send);
            bytes_sent = ExgSend(sockP, buf, bytes_to_send, &error);
            if (error) break;
            pos += bytes_sent;
        }
        ExgDisconnect(sockP, error);
    }

    /* Cleanup */
    MemHandleUnlock(t);
}

/*
** BeamSend - Beam the current sketch to another Palm
*/
void BeamSend(void) {
    ExgSocketType exgSock;
    VoidHand descriptionH, nameH, t;
    CharPtr descriptionP, nameP, ptr;
    DWord romversion = 0;

    /* Check if we have IR capability */
    FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romversion);
    if (romversion < 0x03003000) {
        FrmAlert(NoBeamForm);
        return;
    }

    /* Zero the exchange socket */
    MemSet(&exgSock, sizeof(exgSock), 0);

    /* Set the target */
    exgSock.target = AppType;
    /* Set the length */
    if (p.onscratch) exgSock.length = 3200;
    else exgSock.length = 3360;
    /* Set the appData (0 = scratch, 1 = indexed) */
    if (p.onscratch) exgSock.appData = 0;
    else exgSock.appData = 1;
    /* Set the description */
    descriptionH = MemHandleNew(14);
    if (descriptionH) {
        descriptionP = MemHandleLock(descriptionH);
        MemMove(descriptionP, "Diddle Sketch", 13);
        descriptionP[13] = nullChr;
        exgSock.description = descriptionP;
    }
    /* Set the name */
    nameH = NULL;
    if (p.onscratch) {
        nameH = MemHandleNew(8);
        if (nameH) {
            nameP = MemHandleLock(nameH);
            MemMove(nameP, "scratch", 7);
            nameP[7] = nullChr;
            exgSock.name = nameP;
        }
    }
    else {
        t = DmQueryRecord(d.dbRI, p.dbII);
        if (t) {
            ptr = MemHandleLock(t);
                if (ptr+3200) {
                    if (StrLen(ptr+3200)) {
                        nameH = MemHandleNew(StrLen(ptr+3200)+1);
                        if (nameH) {
                            nameP = MemHandleLock(nameH);
                            MemMove(nameP, ptr+3200, StrLen(ptr+3200));
                            nameP[StrLen(ptr+3200)] = nullChr;
                            exgSock.name = nameP;
                        }
                        MemHandleUnlock(t);
                    }
                    else MemHandleUnlock(t);
                }
                else MemHandleUnlock(t);
        }
    }
    if (!nameH) {
        nameH = MemHandleNew(9);
        nameP = MemHandleLock(nameH);
        MemMove(nameP, "Untitled", 8);
        nameP[8] = nullChr;
        exgSock.name = nameP;
    }

    /* Send the actual data */
    BeamSendSketch(&exgSock);

    /* Clean up */
    if (descriptionH) {
        MemHandleUnlock(descriptionH);
        MemHandleFree(descriptionH);
    }
    if (nameH) {
        MemHandleUnlock(nameH);
        MemHandleFree(nameH);
    }
}

