/*
** diddle - a scribble clone for pilot, based on Doodle
** Copyright (c) 1999, Mitch Blevins <mblevin@debian.org>.
** Licensed under the GNU GPL, version 2 or later
** See file "COPYING" that you should have received with this program
** or visit http://www.gnu.org/copyleft/gpl.html
** 
** Doodle is Copyright 1997, Roger E Critchlow Jr., San Francisco, California.
** All rights reserved.  Fair use permitted.  Caveat emptor.  No warranty.
*/

#include "diddle.h"
#include "callback.h"
#include "linkaware.h"
#include "find.h"

/* beam.c */

extern void BeamSend();
extern void BeamReceive(ExgSocketPtr sockP, Boolean appIsActive);


/* shape.c */
extern void DoShapePopup(void);
extern void ClearOldShape(void);
extern Boolean DoShape(EventPtr e);
extern void EndShape(EventPtr e);

data d;					/* initialized to zero */

pref p = {
    PrefVersion,
    1,
    0,
    false,
    true,
    0,
    DiddleTForm,
    { 0, 16, 160, 144 },
    "p", "=", ".", "8",
    4,
    "\x88",
    { 0xffff, 0xffff, 0xffff, 0xffff },
    stdFont,
    TXT_BLACK,
    0,
    0,
    0,
    false,
    false,
    false,  /* disable_button_drop */
    "@"
};

/*
** Default penpixel set.
*/
char penpixFine[] =		"\x88";
char penpixMedium[] =		"\x88\x98\x89\x99";
char penpixBold[] =         "\x77\x87\x97\x78\x98\x79\x89\x99";
char penpixBroad[] =		"\x66\x76\x86\x96\xA6\x67\xA7\x68\xA8\x69\xA9\x6A\x7A\x8A\x9A\xAA";
char penpixFineItalic[] =	"\x78\x88\x87\x97";
char penpixMediumItalic[] =	"\x69\x79\x78\x88\x87\x97\x96\xA6";
char penpixBroadItalic[] =	"\x5A\x6A\x69\x79\x78\x88\x87\x97\x96\xA6\xA5\xB5";
RectangleType penr = { 42+2*14+2*4, 1, 14, 10 };

/*
** Default ink pattern set.
*/
unsigned short patternWhite[] =  { 0x0000, 0x0000, 0x0000, 0x0000 };
unsigned short patternDith12[] = { 0x1100, 0x4400, 0x1100, 0x4400 };
unsigned short patternDith25[] = { 0x2211, 0x4488, 0x2211, 0x4488 };
unsigned short patternDith37[] = { 0xc84c, 0x2331, 0x8cc4, 0x3213 };
unsigned short patternDith50[] = { 0xaa55, 0xaa55, 0xaa55, 0xaa55 };
unsigned short patternDith62[] = {~0xc84c,~0x2331,~0x8cc4,~0x3213 };
unsigned short patternDith75[] = {~0x2211,~0x4488,~0x2211,~0x4488 };
unsigned short patternDith87[] = {~0x1100,~0x4400,~0x1100,~0x4400 };
unsigned short patternBlack[] =  {~0x0000,~0x0000,~0x0000,~0x0000 };
RectangleType inkr = { 42+3*14+3*4, 1, 14, 10 };

/*
** GetObjectPointer - return an object pointer given a form and objID
*/
static void * GetObjectPointer(FormPtr frm, Word objID) {
     return FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, objID));
}

/*
** GetListWidth
*/
SWord GetListWidth(ListPtr list) {
    Word num_items, i;
    CharPtr list_textP;
    FontID oldfont;
    SWord text_width=0, list_width=0;

    if (!list) return 10;  /* real skinny list indicates an error */

    num_items = LstGetNumberOfItems(list);
    oldfont = FntSetFont(stdFont);
    for (i=0; i<num_items; i++) {
        list_textP = LstGetSelectionText(list, i);
        if (list_textP) {
            text_width = FntLineWidth(list_textP, StrLen(list_textP));
            if (text_width > list_width)
                list_width = text_width;
        }
    }
    FntSetFont(oldfont);
    list_width += 8;

    return list_width;
}


/*
** SetDatabaseWriteMode - Set the write mode of d.dbR
*/
static void SetDatabaseWriteMode(Boolean write, Boolean is_scratch) {
    UInt mode;

    /* Close the current database */
    if (is_scratch) DmCloseDatabase(d.dbR);
    else DmCloseDatabase(d.dbRI);

    /* Open the new one */
    if (write) mode = dmModeReadWrite;
    else mode = dmModeReadOnly;
    if (is_scratch) d.dbR = DmOpenDatabase(0, d.scratch_dbID, mode);
    else d.dbRI = DmOpenDatabase(0, d.indexed_dbID, mode);
    if ((is_scratch && !d.dbR) || (!is_scratch && !d.dbRI))
        abort("Could not reopen database!");
}

/*
** SetLabel - set the label of a control on the current form.
*/
static void SetLabel(Int objectID, char *label) {
    FormPtr frm;

    frm = FrmGetActiveForm();
    CtlSetLabel(GetObjectPointer(frm, objectID), label);
}

/*
** LabelImage - put the program state in the title bar
*/
static void LabelImage(void) {
    if (p.formID == DiddleTForm) {
        char temp[8];
        char temp2[4];
        short i, x, y;
        /* Show mode label */
        SetLabel(ModeButton, p.mode_label);
        /* Show filter label */
        SetLabel(SmoothButton, p.filt_label);
        /* Show current pen */
        WinSetPattern(patternBlack);
        WinEraseRectangle(&penr, 4);
        for (i = 0; i < NPENPIX && p.penpix[i] != 0; i += 1) {
            x = 42+2*14+2*4+6+dxFromPenpix(p.penpix[i]);
            y = 1+4+dyFromPenpix(p.penpix[i]);
            WinFillLine(x, y, x, y);
        }
        /* Show current ink */
        WinSetPattern(p.inkpat);
        WinFillRectangle(&inkr, 4);
        /* Show current page number */
        if (p.onscratch) {
            StrIToA(temp, p.dbI);
            StrNCat(temp, "/", 7);
            StrIToA(temp2, DmNumRecords(d.dbR)-1);
            StrNCat(temp, temp2, 7);
            SetLabel(PageButton, temp);
        }
        else {
            SetLabel(TitleButton, "T");
            SetLabel(IndexButton, "I");
        }
    }
}

/*
** AllocImage - create a new image record.
*/
static void AllocImage(void) {
    VoidHand t;

    SetDatabaseWriteMode(true, p.onscratch);
    if(p.onscratch) {
        t = DmNewRecord(d.dbR, &p.dbI, 3200);
        if (!t) abort("out of memory?");
        if (0 == p.dbI) {
            DmSet(MemHandleLock(t), 0, 3200, 0);
            MemHandleUnlock(t);
        }
        DmReleaseRecord(d.dbR, p.dbI, true);
        /* Clipboard? */
        if (0 == p.dbI) {
            p.dbI++;
            t = DmNewRecord(d.dbR, &p.dbI, 3200);
            if (!t) abort("out of memory?");
            DmReleaseRecord(d.dbR, p.dbI, true);
        }
    } else {
        t = DmNewRecord(d.dbRI, &p.dbII, 3360);
        if (!t) abort("out of memory?");
        DmSet(MemHandleLock(t), 0, 3360, 0);
        MemHandleUnlock(t);
        DmReleaseRecord(d.dbRI, p.dbII, true);
    }
    SetDatabaseWriteMode(false, p.onscratch);
}

/*
** ClearImage - clear a new image record.
*/
inline static void ClearImage(void) {
    VoidHand t;

    SetDatabaseWriteMode(true, p.onscratch);
    if (p.onscratch) {
        t = DmGetRecord(d.dbR, p.dbI);
        if (!t) abort("ClearImage failed to Get");
        DmSet(MemHandleLock(t), 0, 3200, 0);
        MemHandleUnlock(t);
        DmReleaseRecord(d.dbR, p.dbI, true);
    } else {
        t = DmGetRecord(d.dbRI, p.dbII);
        if (!t) abort("ClearImage failed to Get");
        DmSet(MemHandleLock(t), 0, 3200, 0);
        MemHandleUnlock(t);
        DmReleaseRecord(d.dbRI, p.dbII, true);
    }
    SetDatabaseWriteMode(false, p.onscratch);
}

/*
** MakeImage - create a new empty image record.
*/
static void MakeImage(void) {
    AllocImage();
    ClearImage();
}

/*
** NewScratchImage - create a new scratch image or go to last empty one
*/
static void NewScratchImage(void) {
    VoidHand t;
    VoidPtr ptr;
    UInt i, j;
    Byte buf[PASTE_BUFFER_SIZE];
    Boolean still_empty;

    /* Get the highest numbered scratch sketch */
    p.onscratch = true;
    p.dbI = DmNumRecords(d.dbR) - 1;
    if (p.dbI > 0)
        t = DmQueryRecord(d.dbR, p.dbI);
    else {
        MakeImage();
        t = DmQueryRecord(d.dbR, p.dbI);
    }
    if (!t) abort("NewScratchImage failed to Get");
    ptr = MemHandleLock(t);

    /* Check to see if the current sketch is empty */
    still_empty = true;
    for ( i=0; i < 3200; i+=PASTE_BUFFER_SIZE ) {
        MemMove(buf, ptr+i, PASTE_BUFFER_SIZE);
        for ( j=0; j < PASTE_BUFFER_SIZE; j++ ) {
            if (buf[j]) {
                still_empty = false;
                break;
            }
        }
        if (!still_empty) break;
    }
    MemHandleUnlock(t);

    /* Create a new sketch if current one has black pixels */
    if (!still_empty) {
        p.dbI += 1;
        MakeImage();
    }
}

/*
** FillImage - fill a new image record
*/
static void FillImage(void) {
    VoidHand t;
    VoidPtr ptr;
    RectangleType rect;
    DmOpenRef dbR;
    short dbI;
    
    /* Setup */
    SetDatabaseWriteMode(true, p.onscratch);
    dbI = p.onscratch ? p.dbI : p.dbII;
    dbR = p.onscratch ? d.dbR : d.dbRI;
    rect.topLeft.x = p.r.topLeft.x;
    rect.topLeft.y = p.r.topLeft.y;
    rect.extent.x = p.r.extent.x;
    rect.extent.y = 80;

    /* Fill half the screen with the pattern */
    WinFillRectangle(&rect, 0);
    /* Write to disk (twice) */
    t = DmGetRecord(dbR, dbI);
    ptr = MemHandleLock(t);
    if (!t) abort("FillImage failed to Get");
    DmWrite(ptr, 0,
            WinGetWindowPointer(d.winM)->displayAddrV20+p.r.topLeft.y*20, 1600);
    DmWrite(ptr, 1600,
            WinGetWindowPointer(d.winM)->displayAddrV20+p.r.topLeft.y*20, 1600);
    /* Write it back to the screen */
    MemMove(WinGetWindowPointer(d.winM)->displayAddrV20+p.r.topLeft.y*20,
            ptr, p.r.extent.y*20);

    /* Clean up */
    MemHandleUnlock(t);
    DmReleaseRecord(dbR, dbI, true);
    SetDatabaseWriteMode(false, p.onscratch);
}

/*
** InvertImage - XOR the current image, done on DB, rather than window
**               because window is sometimes obscured by titlebar
*/
static void InvertImage(void) {
    VoidHand t;
    VoidPtr ptr;
    UInt i, j;
    DmOpenRef dbR;
    short dbI;
    Byte buf[PASTE_BUFFER_SIZE];

    /* Setup */
    SetDatabaseWriteMode(true, p.onscratch);
    dbI = p.onscratch ? p.dbI : p.dbII;
    dbR = p.onscratch ? d.dbR : d.dbRI;

    t = DmGetRecord(dbR, dbI);
    ptr = MemHandleLock(t);
    if (!t) abort("InvertImage failed to Get");

    for ( i=0; i < 3200; i+=PASTE_BUFFER_SIZE ) {
        MemMove(buf, ptr+i, PASTE_BUFFER_SIZE);

        for ( j=0; j < PASTE_BUFFER_SIZE; j++ ) {
            buf[j] = ~buf[j];
        }
        DmWrite(ptr, i, buf, PASTE_BUFFER_SIZE);
    }

    /* Cleanup */
    DmReleaseRecord(dbR, dbI, true);
    MemHandleUnlock(t);
    SetDatabaseWriteMode(false, p.onscratch);
}


/*
** LoadImage - copy the database record p.dbI onto the current display.
*/
static void LoadImage(void) {
    VoidHand t;
    VoidPtr ptr;
    Word error;
    short dbI;
    DmOpenRef dbR;

    dbI = (p.onscratch ? p.dbI : p.dbII);
    dbR = (p.onscratch ? d.dbR : d.dbRI);

    t = DmQueryRecord(dbR, dbI);
    if (!t) {
        MakeImage();
        dbI = (p.onscratch ? p.dbI : p.dbII); /* MakeImage() may change p.dbI */
        t = DmQueryRecord(dbR, dbI);
        if (!t) abort("LoadImage failed to load");
    }
    ptr = MemHandleLock(t);
    MemMove(WinGetWindowPointer(d.winM)->displayAddrV20+p.r.topLeft.y*20,
            ptr, p.r.extent.y*20);

    /* Set global locked variable if a named sketch */
    if (!p.onscratch) MemMove(&(d.locked), ptr+3359, 1);
    d.lock_count = 0;

    MemHandleUnlock(t);
}

/*
** SaveImage - copy the current display into the database record p.dbI[I].
*/
static void SaveImage(void) {
    VoidHand t;
    DmOpenRef dbR;
    short dbI;

    SetDatabaseWriteMode(true, p.onscratch);
    dbI = (p.onscratch ? p.dbI : p.dbII);
    dbR = (p.onscratch ? d.dbR : d.dbRI);

    MenuEraseStatus(NULL);
    if ((dbI+1) > DmNumRecords(dbR)) return; /* index out of range */
    t = DmGetRecord(dbR, dbI);
    if (!t) abort("SaveImage failed to Get");
    DmWrite(MemHandleLock(t), 0,
            WinGetWindowPointer(d.winM)->displayAddrV20+p.r.topLeft.y*20,
            p.r.extent.y*20);
    MemHandleUnlock(t);
    DmReleaseRecord(dbR, dbI, true);
    SetDatabaseWriteMode(false, p.onscratch);
}

/*
** SaveScratchToIndex - copy contents of current scratch buffer to index
*/
static void SaveScratchToIndex(void) {
    VoidHand ts, ti;

    SetDatabaseWriteMode(true, false);
    ts = DmQueryRecord(d.dbR, p.dbI);
    if (!ts) abort("SaveScratchToIndex failed to Get scratch");
    ti = DmGetRecord(d.dbRI, p.dbII);
    if (!ti) abort("SaveScratchToIndex failed to Get index");
    DmWrite(MemHandleLock(ti), 0, MemHandleLock(ts), 3200);
    MemHandleUnlock(ts);
    MemHandleUnlock(ti);
    DmReleaseRecord(d.dbRI, p.dbII, true);
    SetDatabaseWriteMode(false, false);
}

/*
** CopyImage - create a new copy of the current record.
*/
static void CopyImage(short oldIndex) {
    VoidHand h_old, h_new;
    /* Only called from scratch buffer */
    AllocImage();
    SetDatabaseWriteMode(true, true);
    h_old = DmQueryRecord(d.dbR, oldIndex);
    if (!h_old) abort("CopyImage failed to Get old record");
    h_new = DmGetRecord(d.dbR, p.dbI);
    if (!h_new) abort("CopyImage failed to Get new record");
    DmWrite(MemHandleLock(h_new), 0, MemHandleLock(h_old), 3200);
    MemHandleUnlock(h_old);
    MemHandleUnlock(h_new);
    DmReleaseRecord(d.dbR, p.dbI, true);
    SetDatabaseWriteMode(false, true);
}

/*
** EditCopyImage - Copy current image to diddle clipboard (index 0)
*/
static void EditCopyImage(void) {
    VoidHand tcurrent, tclip;
    DmOpenRef dbR;
    short dbI;

    /* Use the current sketch */
    SetDatabaseWriteMode(true, true); /* scratch database */
    if (!p.onscratch) SetDatabaseWriteMode(true, false);
    dbI = (p.onscratch ? p.dbI : p.dbII);
    dbR = (p.onscratch ? d.dbR : d.dbRI);

    /* Save the current sketch */
    tcurrent = DmGetRecord(dbR, dbI);
    if (!tcurrent) abort("EditCopyImage failed to Get current record");
    MenuEraseStatus(NULL);
    DmWrite(MemHandleLock(tcurrent), 0,
            WinGetWindowPointer(d.winM)->displayAddrV20+p.r.topLeft.y*20,
            p.r.extent.y*20);
    MemHandleUnlock(tcurrent);

    /* Open the clipboard and start writing */
    tclip = DmGetRecord(d.dbR, 0);
    if (!tclip) abort("EditCopyImage failed to Get clipboard");
    DmWrite(MemHandleLock(tclip), 0, MemHandleLock(tcurrent), 3200);

    /* Cleanup */
    MemHandleUnlock(tclip);
    MemHandleUnlock(tcurrent);
    DmReleaseRecord(d.dbR, 0, true);
    DmReleaseRecord(dbR, dbI, true);
    SetDatabaseWriteMode(false, true); /* scratch database */
    if (!p.onscratch) SetDatabaseWriteMode(false, false);

}

/*
** EditPasteImage - write clipboard image to current record
*/
static void EditPasteImage(FormPtr frm) {
    Word baction, waction;
    VoidHand tclip, tcurrent;
    VoidPtr pclip, pcurr;
    UInt i, j;
    DmOpenRef dbR;
    short dbI;
    ListPtr list;
    Byte bb[PASTE_BUFFER_SIZE], wb, cb[PASTE_BUFFER_SIZE];

    /* Get black and white action selections from lists */
    list = GetObjectPointer(frm, PasteBlackList);
    baction = LstGetSelection(list);
    list = GetObjectPointer(frm, PasteWhiteList);
    waction = LstGetSelection(list);

    /* Use the current sketch */
    SetDatabaseWriteMode(true, p.onscratch);
    dbI = (p.onscratch ? p.dbI : p.dbII);
    dbR = (p.onscratch ? d.dbR : d.dbRI);

    /* Get pointer to memory areas */
    tcurrent = DmGetRecord(dbR, dbI);
    if (!tcurrent) abort("EditPasteImage could not get current record");
    pcurr = MemHandleLock(tcurrent);
    tclip = DmQueryRecord(d.dbR, 0);
    if (!tclip) abort("EditPasteImage could not get clip record");
    pclip = MemHandleLock(tclip);
    
    /* Check for the most common case */
    if ((PASTE_NORMAL == baction) && (PASTE_NORMAL == waction)) {
        DmWrite(pcurr, 0, pclip, 3200);
    } else { /* Oh well... do it the long way */
        for ( i=0; i < 3200; i+=PASTE_BUFFER_SIZE ) {
            MemMove(bb, pclip+i, PASTE_BUFFER_SIZE);
            MemMove(cb, pcurr+i, PASTE_BUFFER_SIZE);

            for ( j=0; j < PASTE_BUFFER_SIZE; j++ ) {
                wb = ~(bb[j]);
                if (baction == PASTE_NORMAL) cb[j] |= bb[j];
                else if (baction == PASTE_INVERSE) cb[j] &= wb;
                else if (baction == PASTE_XOR) cb[j] ^= bb[j];

                if (waction == PASTE_NORMAL) cb[j] &= bb[j];
                else if (waction == PASTE_INVERSE) cb[j] |= wb;
                else if (waction == PASTE_XOR) cb[j] ^= wb;
            }
            DmWrite(pcurr, i, cb, PASTE_BUFFER_SIZE);
        }
    }

    /* Clean up */
    MemHandleUnlock(tcurrent);
    MemHandleUnlock(tclip);
    DmReleaseRecord(dbR, dbI, true);
    SetDatabaseWriteMode(false, p.onscratch);
}

/*
** LoadName - Write text from Properties form into recrod
*/
void LoadName(FormPtr frm) {
    CharPtr pTitle, pKeyWords;
    FieldPtr fld;
    VoidHand t;
    Byte islocked;
    FormBitmapType *bm;

    /* Get the Record */
    SetDatabaseWriteMode(true, false); /*indexed database */
    t = DmGetRecord(d.dbRI, p.dbII);
    if (!t) abort("LoadName failed to Get");

    /* Get the text from the fields */
    fld = GetObjectPointer(frm, PropNameField);
    pTitle = FldGetTextPtr(fld);
    fld = GetObjectPointer(frm, PropKeyField);
    pKeyWords = FldGetTextPtr(fld);

    /* Put the text into the Record */
    if (pTitle) {
        if ((StrLen(pTitle) > 0) && (StrLen(pTitle) < 32)) {
            DmWrite(MemHandleLock(t), 3200, pTitle, StrLen(pTitle)+1);
            MemHandleUnlock(t);
        }
    }
    if (pKeyWords) {
        if ((StrLen(pKeyWords) > 0) && (StrLen(pKeyWords) < 127)) {
            DmWrite(MemHandleLock(t), 3232, pKeyWords, StrLen(pKeyWords)+1);
            MemHandleUnlock(t);
        }
    }

    /* Store the locked state */
    bm = GetObjectPointer(frm, LockBitmap);
    if (bm->attr.usable) islocked = true;
    else islocked = false;
    DmSet(MemHandleLock(t), 3359, 1, islocked);
    MemHandleUnlock(t);

    /* Release the Record */
    DmReleaseRecord(d.dbRI, p.dbII, true);
    SetDatabaseWriteMode(false, false);
}

/*
** ShowName - Show diddle title in popup list
*/
void ShowName(FormPtr frm) {
    ListPtr list;
    CharPtr name;
    VoidHand h, t;
    VoidPtr ptr;

    h = MemHandleNew(32);
    name = MemHandleLock(h);

    t = DmQueryRecord(d.dbRI, p.dbII);
    ptr = MemHandleLock(t);

    if (StrLen(ptr+3200)) StrNCopy(name, (const CharPtr)(ptr+3200), 32);
    else StrCopy(name, "Untitled");
    MemHandleUnlock(t);

    list = GetObjectPointer(frm, PopNameList);

    LstSetHeight(list, 1);
    LstSetListChoices(list, &name, 1);
    if (-1 != LstPopupList(list)) {
        SaveImage();
        FrmGotoForm(PropForm);
    }

    MemHandleUnlock(h);
    MemHandleFree(h);
}

/*
** CompareTitles - Sort the index records
*/
static Int CompareTitles(void *rec1, void *rec2, Int o, SortRecordInfoPtr pSR1,
                         SortRecordInfoPtr pSR2, VoidHand AppInfoH) {
    CharPtr t1, t2;
    Int ret;

    t1 = rec1+3200;
    t2 = rec2+3200;

    if (!StrLen(t1)) {
        if (!StrLen(t2)) return 0;
        return 1;
    }
    if (!StrLen(t2)) return -1;

    ret = StrCompare(t1, t2);
    return ret;
}

/*
** DoPrefPopup - Popup the preferences form and load the results
*/
static void DoPrefPopup(void) {
    FormPtr frm;
    ControlPtr ctl;
    Word pushbutton;

    frm = FrmInitForm(PrefForm);
    /* Set the start-on-index checkbox */
    ctl = GetObjectPointer(frm, PrefStartIndexCheck);
    CtlSetValue(ctl, p.start_on_index);
    /* Set the disable-titlebar checkbox */
    ctl = GetObjectPointer(frm, PrefDisableTitleCheck);
    CtlSetValue(ctl, p.disable_title_sc);
    /* Set the disable button-drop checkbox */
    ctl = GetObjectPointer(frm, PrefDisableTitlebarDroplists);
    CtlSetValue(ctl, p.disable_button_drop);
    /* Set the hardware pushbutton selection */
    switch (p.hard_button) {
       case hard1Chr:
            pushbutton = PrefHardware1Push;
            break;
       case hard2Chr:
            pushbutton = PrefHardware2Push;
            break;
        case hard3Chr:
            pushbutton = PrefHardware3Push;
            break;
        case hard4Chr:
            pushbutton = PrefHardware4Push;
            break;
        case 0:
        default:
            pushbutton = PrefHardwareNonePush;
            break;
    }
    ctl = GetObjectPointer(frm, pushbutton);
    CtlSetValue(ctl, 1);

    /* Show the form */
    FrmDoDialog(frm);

    /* Set the values from the form */
    ctl = GetObjectPointer(frm, PrefStartIndexCheck);
    p.start_on_index = CtlGetValue(ctl);
    ctl = GetObjectPointer(frm, PrefDisableTitleCheck);
    p.disable_title_sc = CtlGetValue(ctl);
    ctl = GetObjectPointer(frm, PrefDisableTitlebarDroplists);
    p.disable_button_drop = CtlGetValue(ctl);
    /* Set the hardware pushbutton value */
    ctl = GetObjectPointer(frm, PrefHardwareNonePush);
    if (CtlGetValue(ctl)) p.hard_button = 0;
    ctl = GetObjectPointer(frm, PrefHardware1Push);
    if (CtlGetValue(ctl)) p.hard_button = hard1Chr;
    ctl = GetObjectPointer(frm, PrefHardware2Push);
    if (CtlGetValue(ctl)) p.hard_button = hard2Chr;
    ctl = GetObjectPointer(frm, PrefHardware3Push);
    if (CtlGetValue(ctl)) p.hard_button = hard3Chr;
    ctl = GetObjectPointer(frm, PrefHardware4Push);
    if (CtlGetValue(ctl)) p.hard_button = hard4Chr;

    /* Cleanup */
    FrmDeleteForm(frm);
}

/*
** DoTextPopup - Pop up the Insert text form
*/
static void DoTextPopup(void) {
    FormPtr frm;
    Word foi;     /* font pushbutton object index */
    ListPtr list;
    ControlPtr ctl;
    CharPtr str;
    Word button_pushed;

    frm = FrmInitForm(TextForm);

    /* Set the font pushbutton */
    switch (p.txt_font) {
        case boldFont:
            foi = TextBoldFontPush;
            break;
        case largeFont:
            foi = TextLargeFontPush;
            break;
        case stdFont: /* fall thru */
        default:
            foi = TextStdFontPush;
            break;
    }
    CtlSetValue(GetObjectPointer(frm, foi), true);
    /* Set the dropdown list selection */
    list = GetObjectPointer(frm, TextModeList);
    LstSetSelection(list, p.txt_mode);
    str = LstGetSelectionText(list, p.txt_mode);
    CtlSetLabel(GetObjectPointer(frm, TextModePop), str);
    /* Set the focus to the text field */
    FrmSetFocus(frm, FrmGetObjectIndex(frm, TextField));

    button_pushed = FrmDoDialog(frm);

    if (TextInsertButton == button_pushed) {
        /* Get the font to use */
        ctl = GetObjectPointer(frm, TextStdFontPush);
        if (CtlGetValue(ctl)) p.txt_font = stdFont;
        ctl = GetObjectPointer(frm, TextBoldFontPush);
        if (CtlGetValue(ctl)) p.txt_font = boldFont;
        ctl = GetObjectPointer(frm, TextLargeFontPush);
        if (CtlGetValue(ctl)) p.txt_font = largeFont;

        /* Get the text insertion mode */
        list = GetObjectPointer(frm, TextModeList);
        p.txt_mode = LstGetSelection(list);

        /* Copy text to insert into d.InsertionText */
        str = 
          FldGetTextPtr(GetObjectPointer(frm, TextField));
        d.InsertionText[0] = 0; /* just to make sure... */
        if (str) {
            StrNCopy(d.InsertionText, str, 31);
            /* Signal the new mode for the Pen Handler */
            if (StrLen(d.InsertionText) > 0) {
                d.drawmode = MODE_TEXT;
                d.last_insert = d.drawmode;
            }
            d.txt_oldfont = FntSetFont(p.txt_font);
        }
        /* We rely on the Pen Queue being empty here */
        EvtFlushPenQueue(); 
    }

    /* Cleanup */
    FrmDeleteForm(frm);
}

/*
** DoPastePopup - pop up the paste form
*/
static void DoPastePopup(void) {
    FormPtr frm;
    Word button_pressed;

    frm = FrmInitForm(PasteForm);

    button_pressed = FrmDoDialog(frm);

    if (PasteOK == button_pressed) {
        EditPasteImage(frm);
        LoadImage();
    }
    FrmDeleteForm(frm);
}

/*
** Load a pen pattern
*/
inline static void LoadPen(UInt chr, unsigned char *pen) {
    StrCopy(p.penpix, pen);
    if (p.mode_label[0] == modeErase)
        p.erase_pen_label[0] = chr;
    else
        p.pen_label[0] = chr;
}

/*
** Random integer in the range 0..(n-1).
*/
inline static UInt RandomLessThan(UInt n) {
    return SysRandom(0) / (1+sysRandomMax/n);
}

/*
** Load an ink pattern
*/
static void LoadInk(UInt chr, unsigned short *pattern) {
    if (pattern != NULL) {
        /* Load the specified ink pattern */
        p.ink_label[0] = chr;
        ((long *)(p.inkpat))[0] = ((long *)pattern)[0];
        ((long *)(p.inkpat))[1] = ((long *)pattern)[1];
    } else {
        unsigned i, id, im, ib, j, jd, jm, jb;

        SysRandom(((long *)p.inkpat)[0]+((long *)p.inkpat)[1]);
        for (i = 0; i < 64; i += 1) {
            j = i + RandomLessThan(64-i);
            ib = p.inkpat[id = i/16] & (im = 1<<(i%16));
            jb = p.inkpat[jd = j/16] & (jm = 1<<(j%16));
            if (ib)
                p.inkpat[jd] |= jm;
            else
                p.inkpat[jd] &= ~jm;
            if (jb)
                p.inkpat[id] |= im;
            else
                p.inkpat[id] &= ~im;
        }
    }
    WinSetPattern(p.inkpat);
    p.mode_label[0] = modePaint;
}

/*
** LockAsk - Ask the user if the sketch should be unlocked
*/
static Boolean LockAsk(void) {
    VoidHand t;

    if (p.onscratch || !d.locked) return false;

    if (FrmAlert(LockAskForm)) {
        SetDatabaseWriteMode(true, false);
        t = DmGetRecord(d.dbRI, p.dbII);
        DmSet(MemHandleLock(t), 3359, 1, 0);
        MemHandleUnlock(t);
        DmReleaseRecord(d.dbRI, p.dbII, true);
        d.locked = false;
        SetDatabaseWriteMode(false, false);
    }
    d.lock_count = 0;
    return true;
}

/*
** LockWarn - Beep at the user and increment the lock count
*/
static void LockWarn(void) {
    if (d.lock_count < 2) {
        SndPlaySystemSound(sndWarning);
        d.lock_count++;
    }
    else {
        LockAsk();
        EvtFlushPenQueue();  /* Leftover penMoves and penUps */
    }
}

/*
** ShortCutPop - handle selection of shortcut menu
*/
static void ShortCutPop(enum sc_locations sc_location) {
    ListPtr list;
    FormPtr frm;
    short selection;
    
    frm = FrmGetFormPtr(p.formID);
    if (!frm) return;
    list = GetObjectPointer(frm, PopShortcutList);

    list->bounds.extent.x = d.shortcut_width;
    if (sc_location == low) {
        LstSetPosition(list, 157-d.shortcut_width, 97);
    }
    else LstSetPosition(list, 3, 15);
    LstSetSelection(list, -1);

    selection = LstPopupList(list);
    if (-1 == selection) return;

    switch (selection) {
        case SCUT_LINE:
            EvtFlushPenQueue();
            d.drawmode = MODE_LINE;
            d.last_insert = d.drawmode;
            return;
            break;
        case SCUT_RECTANGLE:
            p.shape_type = SHAPE_RECTANGLE;
            break;
        case SCUT_ROUNDRECT:
            p.shape_type = SHAPE_ROUNDRECT;
            break;
        case SCUT_OVAL:
            p.shape_type = SHAPE_OVAL;
            break;
        case SCUT_CIRCLE:
            p.shape_type = SHAPE_CIRCLE;
            break;
        case SCUT_TEXT:
            DoTextPopup();
            return;
            break;
        default:
            break;
    }
    /* Handle the common code for shape events here */
    EvtFlushPenQueue();  /* Pen queue must be empty */
    d.drawmode = MODE_SHAPE;
    d.last_insert = d.drawmode;
}

/*
** KeyDownIndex - act on key or menu event specific to indexed buffer
*/
static Boolean KeyDownIndex(UInt chr) {
    Boolean handled = false;
    FormPtr frm;

    switch (chr) {
    case cmdDetails:
        SaveImage();
        FrmGotoForm(PropForm);
        handled = true;
        break;

    case cmdRemove:
        if (LockAsk()) break;
        if (0 == FrmAlert(ConfirmDeleteForm)) {
            SetDatabaseWriteMode(true, false);
            DmRemoveRecord(d.dbRI, p.dbII);
            SetDatabaseWriteMode(false, false);
            if (p.dbII > 0) p.dbII -=1;
            else p.dbII = 0;
            if (DmNumRecords(d.dbRI) > 0) LoadImage();
            FrmGotoForm(IndexForm);
        }
        handled = true;
        break;

    case cmdIndex:
        SaveImage();
        d.noselect = false;
        FrmGotoForm(IndexForm);
        handled = true;
        break;

    case cmdLink:
        SaveImage();
        LinkPublish(LinkDiddleNew(&d, &p));
        handled = true;
        break;

    case cmdScratch:
        frm = FrmGetActiveForm();
        SaveImage();
        p.onscratch = true;
        FrmSetMenu(frm, DiddleMenu);
        CtlHideControl(GetObjectPointer(frm, TitleButton));
        CtlHideControl(GetObjectPointer(frm, IndexButton));
        CtlShowControl(GetObjectPointer(frm, PageButton));
        SndPlaySystemSound(sndClick);
        LoadImage();
        LabelImage();
        handled = true;
        break;

    }

    LabelImage();
    return handled;
}

/*
** KeyDownScratch - act on key or menu event specific to scratch buffer
*/
static Boolean KeyDownScratch(UInt chr) {
    Boolean handled = false;

    switch (chr) {
    case cmdNew:
        SndPlaySystemSound(sndClick);
        SaveImage();
        p.dbI += 1;
        MakeImage();
        LoadImage();
        handled = true;
        break;

    case cmdDuplicate:
        SndPlaySystemSound(sndClick);
        SaveImage();
        p.dbI += 1;
        CopyImage(p.dbI-1);
        LoadImage();
        handled = true;
        break;

    case cmdIndex:
        SaveImage();
        d.noselect = true;
        FrmGotoForm(IndexForm);
        handled = true;
        break;

    case cmdRemove:
        SetDatabaseWriteMode(true, true);
        SndPlaySystemSound(sndClick);
        DmRemoveRecord(d.dbR, p.dbI);
        if (p.dbI > 1) p.dbI -= 1;
        else if (p.dbI == DmNumRecords(d.dbR)) MakeImage();
        SetDatabaseWriteMode(false, true);
        LoadImage();
        handled = true;
        break;

    case cmdPrevious:
    case prevFieldChr:
    case pageUpChr: {
            int n = DmNumRecords(d.dbR);
            SndPlaySystemSound(sndClick);
            SaveImage();
            if (p.dbI <= 1) p.dbI = n-1;
            else p.dbI--;
            LoadImage();
            handled = true;
            break;
        }

    case cmdNext:
    case nextFieldChr:
    case pageDownChr:
        SndPlaySystemSound(sndClick);
        SaveImage();
        if (p.dbI >= (DmNumRecords(d.dbR)-1)) p.dbI = 1;
        else p.dbI++;
        LoadImage();
        handled = true;
        break;

    case cmdSave:
        SaveImage();
        d.newdiddle = true;
        d.savediddle = true;
        FrmGotoForm(PropForm);
        handled = true;
        break;

    }

    LabelImage();
    return handled;
}


/*
** Act on a key (or menu event) received.
*/
static Boolean KeyDown(UInt chr) {
    DmOpenRef dbR;
    short dbI;

    /* Map command character to lower case */
    if (chr >= 'A' && chr <= 'Z')
        chr += 'a'-'A';

    /* Check for mode-specific key presses first */
    if (p.onscratch) {
        if (KeyDownScratch(chr)) return true;
    } else {
        if (KeyDownIndex(chr)) return true;
    }

    /* Dispatch on character */
    switch (chr) {

    case penFine:         LoadPen(chr, penpixFine);         break;
    case penMedium:	  LoadPen(chr, penpixMedium);       break;
    case penBold:         LoadPen(chr, penpixBold);         break;
    case penBroad:	  LoadPen(chr, penpixBroad);        break;
    case penFineItalic:	  LoadPen(chr, penpixFineItalic);   break;
    case penMediumItalic: LoadPen(chr, penpixMediumItalic); break;
    case penBroadItalic:  LoadPen(chr, penpixBroadItalic);  break;

    case inkWhite:  LoadInk(chr, patternWhite);  break;
    case inkDith12: LoadInk(chr, patternDith12); break;
    case inkDith25: LoadInk(chr, patternDith25); break;
    case inkDith37: LoadInk(chr, patternDith37); break;
    case inkDith50: LoadInk(chr, patternDith50); break;
    case inkDith62: LoadInk(chr, patternDith62); break;
    case inkDith75: LoadInk(chr, patternDith75); break;
    case inkDith87: LoadInk(chr, patternDith87); break;
    case inkBlack:  LoadInk(chr, patternBlack); break;
    case inkRandom: LoadInk(chr, NULL); break;

    case modePaint:
    case modeSmear:
    case modeErase:
        p.mode_label[0] = chr;
        if (modeErase == chr)
            KeyDown(p.erase_pen_label[0]);
        else
            KeyDown(p.pen_label[0]);
        break;

    case cmdRough:
        p.filt = 1;
        p.filt_label[0] = chr;
        break;

    case cmdSmooth:
        p.filt = 4;
        p.filt_label[0] = chr;
        break;

    case cmdSmoother:
        p.filt = 8;
        p.filt_label[0] = chr;
        break;

    case cmdAbout:
        FrmAlert(AboutForm);
        break;

    case cmdPref:
        DoPrefPopup();
        break;

    case cmdTitle:
        SaveImage();
        FrmGotoForm(p.formID == DiddleForm ? DiddleTForm : DiddleForm);
        break;

    case cmdClear:
        if (LockAsk()) break;
        MenuEraseStatus(NULL);
        ClearImage();
        LoadImage();
        break;

    case cmdFill:
        if (LockAsk()) break;
        MenuEraseStatus(NULL);
        FillImage();
        LoadImage();
        break;

    case cmdInvert:
        if (LockAsk()) break;
        MenuEraseStatus(NULL);
        SaveImage();
        InvertImage();
        LoadImage();
        break;

    case cmdCopy:
        SndPlaySystemSound(sndClick);
        EditCopyImage();
        break;

    case cmdPaste:
        if (LockAsk()) break;
        DoPastePopup();
        break;

    case cmdText:
        DoTextPopup();
        break;

    case cmdLine:
        SndPlaySystemSound(sndClick);
        EvtFlushPenQueue();
        d.drawmode = MODE_LINE;
        d.last_insert = d.drawmode;
        break;

    case cmdShape:
        DoShapePopup();
        break;

    case cmdInsertLast:
        EvtFlushPenQueue();
        d.drawmode = d.last_insert;
        if (d.drawmode == MODE_TEXT) d.txt_oldfont = FntSetFont(p.txt_font);
        SndPlaySystemSound(sndClick);
        break;

    case sendDataChr:
    case cmdBeam:
        BeamSend();
        break;

    }
    LabelImage();
    return true;
}

/*
** ButtonDrop - show drop-down list for titlebar buttons
*/
Boolean ButtonDrop(Word ctlID) {
    FormPtr frm;
    ListPtr list;
    Word list_position, n1, n2;
    UInt button_num;
    short list_selection;
    unsigned char sneaky_string[11];
    unsigned char sneaky_char;
    unsigned char * chrP;
    SWord list_right_edge;

    frm = FrmGetActiveForm();
    switch (ctlID) {
        case ModeButton:
            list = GetObjectPointer(frm, MainModeList);
            StrCopy(sneaky_string, "pme");
            sneaky_char = p.mode_label[0];
            button_num = 0;
            list_right_edge = (42+14);
            break;
        case SmoothButton:
            list = GetObjectPointer(frm, MainSmoothList);
            StrCopy(sneaky_string, "<=>");
            sneaky_char = p.filt_label[0];
            button_num = 1;
            list_right_edge = (42+14+18);
            break;
        case PenButton:
            list = GetObjectPointer(frm, MainPenList);
            StrCopy(sneaky_string, ".@,/;");
            if (p.mode_label[0] == modeErase)
                sneaky_char = p.erase_pen_label[0];
            else
                sneaky_char = p.pen_label[0];
            button_num = 2;
            list_right_edge = (42+14+18+18);
            break;
        case InkButton:
            list = GetObjectPointer(frm, MainInkList);
            StrCopy(sneaky_string, "012345678");
            sneaky_char = p.ink_label[0];
            button_num = 3;
            list_right_edge = (42+14+18+18+18);
            break;
        default:
            return false;
            break;
    }
    /* Set the list width and location */
    if (!d.drop_list_width[button_num])
        d.drop_list_width[button_num] = GetListWidth(list);
    list->bounds.extent.x = d.drop_list_width[button_num];
    if (d.drop_list_width[button_num] > list_right_edge)
        d.drop_list_width[button_num] = list_right_edge;
    LstSetPosition(list, list_right_edge-d.drop_list_width[button_num], 13);
    /* Set the list selection */
    n1 = StrLen(sneaky_string);
    chrP = StrChr(sneaky_string, sneaky_char);
    n2 = StrLen(chrP);
    list_position = n1 - n2;
    LstSetSelection(list, list_position);
    /* Popup the list and handle the result */
    list_selection = LstPopupList(list);
    if (list_selection != -1)
        return KeyDown(sneaky_string[list_selection]);

    return false;
}

/*
** Clear the pen state.
*/
static void DiddlePenClear(void) {
    d.nwts = d.wx = d.wy = 0;
    MemSet(d.wts, sizeof(d.wts), 0);
}

/*
** Act on a diddle pen event.
*/
static void DiddlePenEvent(enum events etype, short nx, short ny) {
    if (etype != penDownEvent) {
        short i, dx, dy;
        /* Draw the lines defined by the pen */
        for (i = 0; i < NPENPIX && p.penpix[i] != 0; i += 1) {
            dx = dxFromPenpix(p.penpix[i]);
            dy = dyFromPenpix(p.penpix[i]);
            if ( ! rctPtInRectangle(d.ox+dx, d.oy+dy, &p.r))
                continue;
            if ( ! rctPtInRectangle(nx+dx, ny+dy, &p.r))
                continue;
            if (p.mode_label[0] == modeSmear)
                WinInvertLine(d.ox+dx, d.oy+dy, nx+dx, ny+dy);
            else
                WinFillLine(d.ox+dx, d.oy+dy, nx+dx, ny+dy);
        }
    }
    d.ox = nx;
    d.oy = ny;
}

/*
** Choose the successor to the current choice on some menu.
*/
static unsigned int ChooseSuccessor(int now, unsigned char *choices) {
    unsigned char *cp = StrChr(choices, now);
    return (cp == NULL || cp[1] == 0) ? choices[0] : cp[1];
}

/*
** Fill the Index listbox
*/
static void IndexDrawFunc(UInt item, RectanglePtr b, CharPtr* text) {
    VoidHand h, rh;
    UInt rec;
    Ptr t, rt;
    Byte islocked;

    CALLBACK_PROLOGUE;

    rec = 0;
    if (DmSeekRecordInCategory(d.dbRI, &rec, item, dmSeekForward, p.category))
        ErrDisplay("Could not Seek record");
    h = DmQueryRecord(d.dbRI, rec);
    t = MemHandleLock(h);

    /* Draw the padlock bitmap */
    MemMove(&islocked, t+3359, 1);
    if (islocked) rh = DmGetResource(DiddleBitmapType, LockSmallBitmap);
    else rh = DmGetResource(DiddleBitmapType, UnlockSmallBitmap);
    rt = MemHandleLock(rh);
    WinDrawBitmap((BitmapPtr)rt, b->topLeft.x, b->topLeft.y+1);

    /* Draw the Chars */
    if (0 == StrLen(t+3200))
        WinDrawChars("Untitled", 8, b->topLeft.x+18, b->topLeft.y);
    else
        WinDrawChars(t+3200, StrLen(t+3200), b->topLeft.x+18, b->topLeft.y);

    /* Clean up */
    MemHandleUnlock(h);
    MemHandleUnlock(rh);
    DmReleaseResource(rh);

    CALLBACK_EPILOGUE;
}

/*
** Initialize the Index listbox
*/
void InitList(FormPtr frm) {
    ListPtr list;
    UInt num_items;
    UInt attr;

    list = GetObjectPointer(frm, IndexList);
    num_items = DmNumRecordsInCategory(d.dbRI, p.category);
    LstSetListChoices(list, NULL, num_items);
    if ((!d.noselect) && (num_items)) {
        if(DmRecordInfo(d.dbRI, p.dbII, &attr, NULL, NULL))
            ErrDisplay("Could not get record info in InitList");
        if( (p.category == dmAllCategories) || 
            (p.category == attr&dmRecAttrCategoryMask) ) {
        LstSetSelection(list, DmPositionInCategory(d.dbRI,p.dbII,p.category));
        }
        else {
            LstSetSelection(list, -1);
        }
    }
    else
        LstSetSelection(list, -1);
    if (num_items)
        LstSetDrawFunction(list, IndexDrawFunc);
}

/*
** Initialize the category trigger
*/
void InitCategoryTrigger(FormPtr frm) {
    ListPtr list;
    ControlPtr ctl;
    UInt attr;

    /* Set p.category to the category of the current record if that record
       exists and is different from p.category */
    if (DmNumRecords(d.dbRI) > 0) {
        if(DmRecordInfo(d.dbRI, p.dbII, &attr, NULL, NULL))
            ErrDisplay("Could not get record info in InitCategoryTrigger");
        if( (p.category != dmAllCategories) &&
            (p.category != attr&dmRecAttrCategoryMask) ) {
                p.category = attr&dmRecAttrCategoryMask;
        }
    }
       
    ctl = GetObjectPointer(frm, CategoryPopup);
    CategoryGetName(d.dbRI, p.category, d.CategoryName);
    CategorySetTriggerLabel(ctl, d.CategoryName);
}

/*
** ScrollIndex - Scroll the index list
*/
static void ScrollIndex(enum directions dir) {
    ListPtr list;
    FormPtr frm = FrmGetActiveForm();

    list = GetObjectPointer(frm, IndexList);
    LstScrollList(list, dir, 10);
}

/*
** SetRecordCategory - store the global record category info in the record
*/
static void SetRecordCategory(void) {
    UInt attr;

    SetDatabaseWriteMode(true, false);
    DmRecordInfo(d.dbRI, p.dbII, &attr, NULL, NULL);
    attr &= ~dmRecAttrCategoryMask;
    if (d.RecordCategory == dmAllCategories)
        attr |= dmUnfiledCategory;
    else
        attr |= d.RecordCategory;
    DmSetRecordInfo(d.dbRI, p.dbII, &attr, NULL);
    SetDatabaseWriteMode(false, false);
           
}

/* 
** ForceMenuDown - forces the menu to erase
*/
static void ForceMenuDown(void) {
    MenuBarPtr menuBarP;
    EventType me;
    Err error;

    menuBarP = MenuGetActiveMenu();
    if (menuBarP && menuBarP->attr.visible) {
        me.eType = keyDownEvent;
        me.data.keyDown.chr = menuChr;
        MenuHandleEvent(menuBarP, &me, &error);
    }
}

/*
** SetProperties - Bind the diddle locked state when
**                 the Properties form is opened
*/
void SetProperties(FormPtr frm) {
    VoidHand t;
    VoidPtr ptr;
    Byte islocked;
    Word bmid_toshow, bmid_toerase;
    FormBitmapType *bmP;
    FieldPtr fld;
    VoidHand textH;

    t = DmQueryRecord(d.dbRI, p.dbII);

    /* Set the locked state */
    ptr = MemHandleLock(t);
    MemMove(&islocked, ptr+3359, 1);

    /* Set the title */
    fld = GetObjectPointer(frm, PropNameField);
    FldFreeMemory(fld);
    textH = MemHandleNew(32);
    StrNCopy(MemHandleLock(textH), ptr+3200, 31);
    MemHandleUnlock(textH);
    FldSetTextHandle(fld, (Handle)textH);

    /* Set the keywords */
    fld = GetObjectPointer(frm, PropKeyField);
    FldFreeMemory(fld);
    textH = MemHandleNew(127);
    StrNCopy(MemHandleLock(textH), ptr+3232, 126);
    MemHandleUnlock(textH);
    FldSetTextHandle(fld, (Handle)textH);

    MemHandleUnlock(t);

    if (islocked) {
        bmid_toshow = LockBitmap;
        bmid_toerase = UnlockBitmap;
    }
    else {
        bmid_toshow = UnlockBitmap;
        bmid_toerase = LockBitmap;
    }
    bmP = GetObjectPointer(frm, bmid_toerase);
    bmP->attr.usable = false;
    bmP = GetObjectPointer(frm, bmid_toshow);
    bmP->attr.usable = true;
}

/*
** DoStroke - a starting or continuing pen stroke while doodling
*/
static Boolean DoStroke(EventPtr e) {
    unsigned short n;

    if(!rctPtInRectangle(e->screenX, e->screenY, &p.r)) return false;

    /* Update sum of coordinates */
    n = (d.nwts-p.filt)%NPENWTS;
    d.wx += e->screenX - d.wts[n].x;
    d.wy += e->screenY - d.wts[n].y;

    /* Update array of coordinates */
    n = d.nwts%NPENWTS;
    d.wts[n].x = e->screenX;
    d.wts[n].y = e->screenY;

    /* Send averaged coordinate */
    n = (d.nwts >= p.filt) ? p.filt : d.nwts+1;
    DiddlePenEvent(d.nwts ? e->eType : penDownEvent, d.wx/n, d.wy/n);

    /* Update count of coordinates */
    d.nwts += 1;

    return true;
}

/*
** EndStroke - end a pen stroke while doodling
*/
static void EndStroke(void) {
    short i;
    unsigned short n;

    /* Drain queue */
    for (i = (d.nwts < p.filt) ? d.nwts : p.filt; --i > 0; ) {
        n = (d.nwts-i-1)%NPENWTS;
        d.wx -= d.wts[n].x;
        d.wy -= d.wts[n].y;
        DiddlePenEvent(penMoveEvent, d.wx/i, d.wy/i);
    }
    /* Clear state */
    DiddlePenClear();
}

/*
** ClearOldText() - clean up the textual doggie-doos left while placing text
*/
inline static void ClearOldText(void) {
    WinInvertChars(d.InsertionText, StrLen(d.InsertionText), d.ox, d.oy);
}

/*
** DoTextInsert - a starting or continuing pen stroke while dragging text
*/
static Boolean DoTextInsert(EventPtr e) {

    /* Return immediately if out of bounds */
    if(!rctPtInRectangle(e->screenX, e->screenY, &p.r)) return false;
    WinInvertChars(d.InsertionText, StrLen(d.InsertionText),
                    e->screenX, e->screenY);
    d.ox = e->screenX;
    d.oy = e->screenY;
    return true;
}

/*
** EndTextInsert - Place the final text when user lifts pen
*/
static void EndTextInsert(EventPtr e) {
     RectangleType rect;

     /* Size the rectangle */
     rect.topLeft.x = e->screenX - 3;
     rect.topLeft.y = e->screenY;
     rect.extent.x = FntCharsWidth(d.InsertionText, StrLen(d.InsertionText))
                        +6;
     rect.extent.y = FntCharHeight();


    /* Draw the text background first */
    switch (p.txt_mode) {
        case TXT_BLACKONWHITE:
            WinEraseRectangle(&rect, 3);
            break;
        case TXT_WHITEONBLACK:
            WinDrawRectangle(&rect, 3);
            break;
        case TXT_BLACK:
        case TXT_XOR:
        default:
            break; /* No background for TXT_BLACK or TXT_XOR */
    }

    /* Now draw the text */
    if (TXT_BLACK == p.txt_mode) {
        /* For some reason WinDrawChars was erasing the background... */
        /* This is a prime example of a kludge */
        WinEraseChars(d.InsertionText, StrLen(d.InsertionText),
                        e->screenX, e->screenY);
        WinInvertChars(d.InsertionText, StrLen(d.InsertionText),
                        e->screenX, e->screenY);
    }
    else {
        WinInvertChars(d.InsertionText, StrLen(d.InsertionText),
                        e->screenX, e->screenY);
    }

    /* Clean up */
    d.drawmode = MODE_DOODLE;
    FntSetFont(d.txt_oldfont);
}

/*
** ClearOldLine - Remove the xor'd line for rubber-banding
*/
inline static void ClearOldLine(void) {
    WinInvertLine(d.anchor.x, d.anchor.y, d.ox, d.oy);
}

/*
** DoLine - Draw rubber band line
*/
static Boolean DoLine(EventPtr e) {

    /* Return immediately if out of bounds */
    if(!rctPtInRectangle(e->screenX, e->screenY, &p.r)) return false;

    /* Draw the line */
    WinInvertLine(d.anchor.x, d.anchor.y, e->screenX, e->screenY);

    d.ox = e->screenX;
    d.oy = e->screenY;

    return true;
}

/*
** EndLine - Place the final line on the drawing
*/
static void EndLine(EventPtr e) {
    short dx, dy, i;

    /* Draw the lines defined by the pen */
    for (i = 0; i < NPENPIX && p.penpix[i] != 0; i += 1) {
        dx = dxFromPenpix(p.penpix[i]);
        dy = dyFromPenpix(p.penpix[i]);
        if ( ! rctPtInRectangle(e->screenX+dx, e->screenY+dy, &p.r))
            continue;
        if ( ! rctPtInRectangle(d.anchor.x+dx, d.anchor.y+dy, &p.r))
            continue;
        if (p.mode_label[0] == modeSmear)
            WinInvertLine(d.anchor.x+dx, d.anchor.y+dy,
                            e->screenX+dx, e->screenY+dy);
        else if (p.mode_label[0] == modeErase)
            WinEraseLine(d.anchor.x+dx, d.anchor.y+dy,
                            e->screenX+dx, e->screenY+dy);
        else
            WinFillLine(d.anchor.x+dx, d.anchor.y+dy,
                            e->screenX+dx, e->screenY+dy);
    }

    d.drawmode = MODE_DOODLE; /* Clean up */
}

/*
** ToggleLockButton - toggle the lock button on the Details form
*/
static void ToggleLockButton(void) {
    FormPtr frm;
    FormBitmapType *bm_locked, *bm_unlocked;
    Word bm_locked_indx, bm_unlocked_indx;

    frm = FrmGetActiveForm();
    bm_locked_indx = FrmGetObjectIndex(frm, LockBitmap);
    bm_locked = FrmGetObjectPtr(frm, bm_locked_indx);
    bm_unlocked_indx = FrmGetObjectIndex(frm, UnlockBitmap);
    bm_unlocked = FrmGetObjectPtr(frm, bm_unlocked_indx);

    if (bm_unlocked->attr.usable) {
        FrmHideObject(frm, bm_unlocked_indx);
        FrmShowObject(frm, bm_locked_indx);
        bm_unlocked->attr.usable = false;
        bm_locked->attr.usable = true;
    }
    else {
        FrmHideObject(frm, bm_locked_indx);
        FrmShowObject(frm, bm_unlocked_indx);
        bm_locked->attr.usable = false;
        bm_unlocked->attr.usable = true;
    }
}

/*
** PropEvent - handle events on the Properties form
*/
static Boolean PropEvent(EventPtr e) {
    Boolean handled = false;
    FormPtr frm;
    ControlPtr ctl;
    Char new_catname[dmCategoryLength];
    UInt attr;
    Word index1, index2;
    ULong unique;

    switch (e->eType) {

    case frmOpenEvent:
        frm = FrmGetActiveForm();
        ctl = GetObjectPointer(frm, PropCatPopup);
        if (d.newdiddle) {
            CtlHideControl(GetObjectPointer(frm, PropDone));
            /* Set trigger to category of current category */
            if (p.category == dmAllCategories)
                d.RecordCategory = dmUnfiledCategory;
            else
                d.RecordCategory = p.category;
        }
        else {
            CtlHideControl(GetObjectPointer(frm, PropOK));
            CtlHideControl(GetObjectPointer(frm,PropCancel));
            /* Find the category of the current record and set trigger */
            if (DmNumRecords(d.dbRI) > 0) {
                DmRecordInfo(d.dbRI, p.dbII, &attr, NULL, NULL);
                d.RecordCategory = attr & dmRecAttrCategoryMask;
            }
            else
                if (p.category == dmAllCategories)
                    d.RecordCategory = dmUnfiledCategory;
                else
                    d.RecordCategory = p.category;
            /* Show correct lock/unlock icon */
            SetProperties(frm);
        }
        CategoryGetName(d.dbRI, d.RecordCategory, d.RecordCategoryName);
        CategorySetTriggerLabel(ctl, d.RecordCategoryName);
        FrmDrawForm(frm);
        FrmSetFocus(frm,FrmGetObjectIndex(frm,PropNameField));
        handled = true;
        break;

    case appStopEvent:
    case frmCloseEvent:
        d.savediddle = false;
        d.newdiddle = false;
        break;

    case frmGotoEvent:
        frm = FrmGetActiveForm();
        if (e->data.frmGoto.matchCustom) {
            p.dbII = e->data.frmGoto.recordNum;
            p.onscratch = false;
        }
        else {
            p.dbI = e->data.frmGoto.recordNum;
            p.onscratch = true;
        }

        FrmGotoForm(p.formID == DiddleForm ? DiddleForm : DiddleTForm);
        d.drawmode = MODE_DOODLE;
        handled = true;
        break;

    case keyDownEvent:
        switch (e->data.keyDown.chr) {
            case prevFieldChr:
                SndPlaySystemSound(sndClick);
            case nextFieldChr:
                frm = FrmGetActiveForm();
                index1 = FrmGetObjectIndex(frm, PropNameField);
                index2 = FrmGetObjectIndex(frm, PropKeyField);
                if (index1 == FrmGetFocus(frm))
                    FrmSetFocus(frm, index2);
                else
                    FrmSetFocus(frm, index1);
                handled = true;
                break;
        }
        break;

    case ctlSelectEvent:
        switch (e->data.ctlEnter.controlID) {
        case PropOK:
            p.onscratch = false;
            MakeImage();
            if (d.savediddle) SaveScratchToIndex();
            /* fall thru */
        case PropDone:
            LoadName(FrmGetActiveForm());
            SetRecordCategory();
            /* Find the unique ID of the current record */
            DmRecordInfo(d.dbRI, p.dbII, NULL, &unique, NULL);
            SetDatabaseWriteMode(true, false);
            DmInsertionSort(d.dbRI, (DmComparF*)CompareTitles, 0);
            SetDatabaseWriteMode(false, false);
            /* Since we sorted, relocate the current record index */
            DmFindRecordByID(d.dbRI, unique, &(p.dbII));
            FrmGotoForm(p.formID == DiddleForm ? DiddleForm : DiddleTForm);
            handled = true;
            break;

        case PropCancel:
            if (d.savediddle) FrmGotoForm(
                            p.formID == DiddleForm ? DiddleForm : DiddleTForm
            );
            else FrmGotoForm(IndexForm);
            handled = true;
            break;

        case PropLockButton:
            ToggleLockButton();
            handled = true;
            break;

        case PropCatPopup:
            frm = FrmGetActiveForm();
            SetDatabaseWriteMode(true, false); /* indexed */
            CategorySelect(d.dbRI, frm, PropCatPopup, PropCatList, false,
                               &(d.RecordCategory), d.RecordCategoryName, 1, 0);
            SetDatabaseWriteMode(false, false); /* indexed */
            /* Store Immediately if an existing record */
            if (!d.newdiddle) {
                DmRecordInfo(d.dbRI, p.dbII, &attr, NULL, NULL);
                if (d.RecordCategory != (attr&dmRecAttrCategoryMask))
                    SetRecordCategory();
            }
            handled = true;
            break;
        }
        break;
    }
    return handled;
}

/*
** IndexEvent - handle events on the Index form
*/
static Boolean IndexEvent(EventPtr e) {
    Boolean handled = false;
    FormPtr frm;
    Word old_category;
    Boolean m;
    ListPtr list;

    CALLBACK_PROLOGUE;

    switch (e->eType) {

    case frmOpenEvent: {
        frm = FrmGetActiveForm();
        InitCategoryTrigger(frm);
        InitList(frm);
        FrmDrawForm(frm);
        p.onindex = true;
        handled = true;
        break;
        }

    case appStopEvent:
    case frmCloseEvent:
        d.noselect = true;
        break;

    case menuEvent:
        switch (e->data.menu.itemID-DiddleMenu) {
            case cmdAbout:
                FrmAlert(AboutForm);
                break;
            case cmdPref:
                DoPrefPopup();
                break;
        }
        handled = true;
        break;

    case frmGotoEvent:
        if (e->data.frmGoto.matchCustom) {
            p.dbII = e->data.frmGoto.recordNum;
            p.onscratch = false;
        }
        else {
            p.dbI = e->data.frmGoto.recordNum;
            p.onscratch = true;
        }
        FrmGotoForm(p.formID == DiddleForm ? DiddleForm : DiddleTForm);
        handled = true;
        break;

    case ctlSelectEvent:
        switch (e->data.ctlEnter.controlID) {
        case ScratchButton:
            handled = true;
            p.onscratch = true;
            FrmGotoForm(p.formID == DiddleForm ? DiddleForm : DiddleTForm);
            break;
        case NewIndexButton:
            p.onscratch = false;
            d.newdiddle = true;
            FrmGotoForm(PropForm);
            handled = true;
            break;
        case NewScratchButton:
            NewScratchImage();
            d.newdiddle = false; /* this is true only for named sketches */
            FrmGotoForm(p.formID == DiddleForm ? DiddleForm : DiddleTForm);
            handled = true;
            break;
        case CategoryPopup:
            frm = FrmGetActiveForm();
            old_category = p.category;
            SetDatabaseWriteMode(true, false); /* indexed */
            m = CategorySelect(d.dbRI, frm, CategoryPopup, CategoryList, true,
                                &(p.category), d.CategoryName, 1, 0);
            SetDatabaseWriteMode(false, false); /* indexed */
            if (m || (old_category != p.category)) {
                InitList(frm);
                list = GetObjectPointer(frm, IndexList);
                LstEraseList(list);
                LstDrawList(list);
            }
            handled = true;
            break;
        }
        break;

    case lstSelectEvent:
        p.dbII = 0;
        if (DmSeekRecordInCategory(d.dbRI, &(p.dbII),
                e->data.lstSelect.selection, dmSeekForward, p.category)) {
            ErrDisplay("Could not seek in category");
        }
        FrmGotoForm(p.formID == DiddleForm ? DiddleForm : DiddleTForm);
        handled = true;
        p.onscratch = false;
        break;

    case keyDownEvent:
        switch (e->data.keyDown.chr) {
            case pageUpChr:
                ScrollIndex(up);
                handled = true;
                break;

            case pageDownChr:
                ScrollIndex(down);
                handled = true;
                break;
        }
        break;

    }

    CALLBACK_EPILOGUE;
    return handled;
}

/*
** DiddleEvent - handle events on the doodling form.
*/
static Boolean DiddleEvent(EventPtr e) {
    Boolean handled = false;
    FormPtr frm;
    UInt chr;

    /*
    ** We only set the inkpat here because it was being mysteriously
    ** reset by menu handling.  But having done that, it makes a perfect
    ** place to implement erase.
    */
    WinSetPattern(p.mode_label[0] == modeErase ? patternWhite : p.inkpat);

    switch (e->eType) {

    case frmOpenEvent:
        frm = FrmGetActiveForm();
        d.winM = FrmGetWindowHandle(frm);
        if (p.onscratch) FrmSetMenu(frm, DiddleMenu);
        else FrmSetMenu(frm, DiddleIMenu);
        if (p.formID == DiddleForm)
            RctSetRectangle(&p.r, 0, 0, 160, 160);
        else {
            RctSetRectangle(&p.r, 0, 16, 160, 144);
            if (p.onscratch) {
                CtlHideControl(GetObjectPointer(frm, TitleButton));
                CtlHideControl(GetObjectPointer(frm, IndexButton));
            }
            else {
                CtlHideControl(GetObjectPointer(frm, PageButton));
            }
        }
        FrmDrawForm(frm);
        LoadImage();
        LabelImage();
        p.onindex = false;
        handled = true;
        break;

    case appStopEvent:
        ForceMenuDown();
    case frmCloseEvent:
        SaveImage();
        break;

    case frmGotoEvent:
        d.drawmode = MODE_DOODLE;
        SaveImage();
        if (e->data.frmGoto.matchCustom) {
            p.dbII = e->data.frmGoto.recordNum;
            p.onscratch = false;
        }
        else {
            p.dbI = e->data.frmGoto.recordNum;
            p.onscratch = true;
        }
        LoadImage();
        LabelImage();
        handled = true;
        break;
        break;

    case menuEvent:
        d.drawmode = MODE_DOODLE;
        handled = KeyDown(e->data.menu.itemID-DiddleMenu);
        break;

    case keyDownEvent:
        d.drawmode = MODE_DOODLE;
        handled = KeyDown(e->data.keyDown.chr);
        break;

    case ctlSelectEvent:
        d.drawmode = MODE_DOODLE;
        switch (e->data.ctlEnter.controlID) {
        case ModeButton:
            if (p.disable_button_drop)
                handled = KeyDown(ChooseSuccessor(p.mode_label[0], "pme"));
            else
                handled = ButtonDrop(ModeButton);
            break;
        case SmoothButton:
            if (p.disable_button_drop)
                handled = KeyDown(ChooseSuccessor(p.filt_label[0], "<=>"));
            else
                handled = ButtonDrop(SmoothButton);
            break;
        case PenButton:
            if (p.disable_button_drop)
                if (p.mode_label[0] == modeErase)
                    handled = KeyDown(ChooseSuccessor(p.erase_pen_label[0],
                                                        ".@,/;"));
                else
                    handled = KeyDown(ChooseSuccessor(p.pen_label[0],
                                                        ".@,/;"));
            else
                handled = ButtonDrop(PenButton);
            break;
        case InkButton:
            if (p.disable_button_drop)
                handled = KeyDown(ChooseSuccessor(p.ink_label[0], "012345678"));            else
                handled = ButtonDrop(InkButton);
            break;
        case PageButton:
            handled = KeyDown(cmdNext);
            break;
        case TitleButton:
            ShowName(FrmGetActiveForm());
            handled = true;
            break;
        case IndexButton:
            SaveImage();
            d.noselect = false;
            FrmGotoForm(IndexForm);
            handled = true;
            break;
        }
        break;

    case penDownEvent:
        DiddlePenClear();
        if (d.locked && !p.onscratch &&
                rctPtInRectangle(e->screenX, e->screenY, &p.r)) {
            LockWarn();
            handled = true;
            break;
        }
        switch (d.drawmode) {
            case MODE_TEXT:
                if (DoTextInsert(e)) handled = true;
                else { d.drawmode = MODE_DOODLE; FntSetFont(d.txt_oldfont); }
                break;
            case MODE_LINE:
                d.anchor.x = e->screenX;
                d.anchor.y = e->screenY;
                if (DoLine(e)) handled = true;
                else d.drawmode = MODE_DOODLE;
                break;
            case MODE_SHAPE:
                d.anchor.x = e->screenX;
                d.anchor.y = e->screenY;
                if (DoShape(e)) handled = true;
                else d.drawmode = MODE_DOODLE;
                break;
            case MODE_DOODLE: /* fall thru */
            default:
                if(DoStroke(e)) handled = true;
                else EndStroke();
                break;
        }
        break;

    case penMoveEvent:
        if (d.locked && !p.onscratch &&
                rctPtInRectangle(e->screenX, e->screenY, &p.r)) {
            handled = true;
            break;
        }
        switch (d.drawmode) {
            case MODE_TEXT:
                ClearOldText();
                if (DoTextInsert(e)) handled = true;
                else { d.drawmode = MODE_DOODLE; FntSetFont(d.txt_oldfont); }
                break;
            case MODE_LINE:
                ClearOldLine();
                if (DoLine(e)) handled = true;
                else d.drawmode = MODE_DOODLE;
                break;
            case MODE_SHAPE:
                ClearOldShape();
                if (DoShape(e)) handled = true;
                else d.drawmode = MODE_DOODLE;
                break;
            case MODE_DOODLE: /* fall thru */
            default:
                if(DoStroke(e)) handled = true;
                else EndStroke();
                break;
        }
        break;

    case penUpEvent:
        if (d.locked && !p.onscratch &&
                rctPtInRectangle(e->screenX, e->screenY, &p.r)) {
            handled = true;
            break;
        }
        switch (d.drawmode) {
            case MODE_TEXT:
                ClearOldText();
                if (rctPtInRectangle(e->screenX, e->screenY, &p.r)) {
                    EndTextInsert(e);
                    handled = true;
                }
                else { d.drawmode = MODE_DOODLE; FntSetFont(d.txt_oldfont); }
                break;
            case MODE_LINE:
                ClearOldLine();
                if (rctPtInRectangle(e->screenX, e->screenY, &p.r)) {
                     EndLine(e);
                     handled = true;
                }
                else d.drawmode = MODE_DOODLE;
                break;
            case MODE_SHAPE:
                ClearOldShape();
                if (rctPtInRectangle(e->screenX, e->screenY, &p.r)) {
                    EndShape(e);
                    handled = true;
                }
                else d.drawmode = MODE_DOODLE;
                break;
            case MODE_DOODLE: /* fall thru */
            default:
                handled = DoStroke(e);
                EndStroke();
                break;
        }
        break;

    case frmTitleSelectEvent:
        if (p.disable_title_sc) return false;
        ShortCutPop(high);
        handled = true;
        break;

    }

    return handled;

}

/*
** SpecialKeyDown - Act on a hardware or special key
*/
static Boolean SpecialKeyDown(EventPtr e) {
    Word chr;
    FormPtr frm;

    if (e->eType != keyDownEvent) return false;

    if (e->data.keyDown.modifiers&commandKeyMask) {
        chr = e->data.keyDown.chr;

        if (e->data.keyDown.modifiers&autoRepeatKeyMask)
            return false;
        if (e->data.keyDown.modifiers&poweredOnKeyMask)
            return false;

        if (chr == p.hard_button) {
            d.drawmode = MODE_DOODLE;
            ForceMenuDown();
            switch (d.formID) {
            case DiddleForm:
            case DiddleTForm:
                SaveImage();
                frm = FrmGetActiveForm();
                FrmSetMenu(frm, DiddleMenu);
                NewScratchImage();
                LoadImage();
                LabelImage();
                break;
            case IndexForm:
                NewScratchImage();
                FrmGotoForm(p.formID == DiddleForm ? DiddleForm : DiddleTForm);
                break;
            }
            return true;
            EvtFlushPenQueue();
        }

        /* SoftKeys only work on main doodling form */
        if ((d.formID != DiddleForm) && (d.formID != DiddleTForm))
            return false;

        if (keyboardAlphaChr == chr) {
            d.drawmode = MODE_DOODLE;
            ForceMenuDown();
            SaveImage();
            FrmGotoForm(p.formID == DiddleForm ? DiddleTForm : DiddleForm);
            return true;
        }
        else if (keyboardNumericChr == chr) {
            ForceMenuDown();
            d.drawmode = MODE_DOODLE;
            ShortCutPop(low);
            return true;
        }
    }
    return false;
}

/*
** SendGotoEvent - queue a frmGotoEvent in response to a Find command
*/
static void SendGotoEvent(GoToParamsPtr pgoto) {
    EventType event;

    event.eType = frmGotoEvent;
    event.data.frmGoto.recordNum = pgoto->recordNum;
    event.data.frmGoto.formID = FrmGetActiveFormID();
    event.data.frmGoto.matchCustom = pgoto->matchCustom;
    EvtAddEventToQueue(&event);
}

/*
** CloseDatabases - you know what it does
*/
static void CloseDatabases(void) {
    DmCloseDatabase(d.dbR);
    DmCloseDatabase(d.dbRI);
}


/*
** OpenDatabases - Open databases and set backup bit
*/
Err OpenDatabases(void) {
    Err err;
    VoidHand h;
    LocalID dbID;
    LocalID appInfoID;
    UInt dbAttrs;
    UInt cardNo;
    AppInfoPtr appInfoP;
    SWord actualSize;
    pref tempPref;
    Word prefsSize = sizeof(p);
    FormPtr frm;
    ListPtr list;
    DWord romversion;

    /* Check for at least PalmOS 2.0 */
    FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romversion);
    if (romversion < 0x02003000) {
        FrmCustomAlert(GenericError, "Need PalmOS 2.0 or greater", "", "");
        return 1;
    }

    /* Open scratch database and get preferences */
#ifdef _DOODLE_COMPAT
    if ((d.dbR = DmOpenDatabaseByTypeCreator(DBType, DoodleType,
                     dmModeReadWrite)) != NULL) {
#else   /* Normal Mode */
    if ((d.dbR = DmOpenDatabaseByTypeCreator(DBType, AppType,
                     dmModeReadWrite)) != NULL) {
#endif  /* DOODLE_COMPAT */
        actualSize = PrefGetAppPreferences(AppType, 0, &tempPref, &prefsSize,
                        true);
        if (actualSize == noPreferenceFound || prefsSize != sizeof(p) ||
                tempPref.version != PrefVersion) {
            /* Do nothing */
        } else {
            PrefGetAppPreferences(AppType, 0, &p, &prefsSize, true);
        }
    } else {
#ifdef _DOODLE_COMPAT
        if (err = DmCreateDatabase(0, DOODName, DoodleType, DBType, false))
            return err;
        d.dbR = DmOpenDatabaseByTypeCreator(DBType,DoodleType,dmModeReadWrite);
#else
        if (err = DmCreateDatabase(0, DBName, AppType, DBType, false))
            return err;
        d.dbR = DmOpenDatabaseByTypeCreator(DBType, AppType, dmModeReadWrite);
#endif  /* DOODLE_COMPAT */
    }

    /* Open indexed database */
    if ((d.dbRI = DmOpenDatabaseByTypeCreator(DBIType, AppType,
                      dmModeReadWrite)) == NULL) {
        if (err = DmCreateDatabase(0, DBIName, AppType, DBIType, false))
            return err;
        d.dbRI = DmOpenDatabaseByTypeCreator(DBIType, AppType, dmModeReadWrite);
    }

    /* Set backup bits and initialize categories */
    if (err = DmOpenDatabaseInfo(d.dbR, &dbID, NULL, NULL, &cardNo, NULL)) {
        CloseDatabases();
        return err;
    }
    if (err = DmDatabaseInfo(0, dbID, NULL, &dbAttrs, NULL, NULL, NULL,
                                 NULL, NULL, NULL, NULL, NULL, NULL)) {
        CloseDatabases();
        return err;
    }
    dbAttrs |= dmHdrAttrBackup;
    if (err = DmSetDatabaseInfo(0, dbID, NULL, &dbAttrs, NULL, NULL,
                                    NULL, NULL, NULL, NULL, NULL, NULL,
                                    NULL)) {
        CloseDatabases();
        return err;
    }
    d.scratch_dbID = dbID;
    if (err = DmOpenDatabaseInfo(d.dbRI, &dbID, NULL, NULL, &cardNo, NULL)) {
        CloseDatabases();
        return err;
    }
    if (err = DmDatabaseInfo(0, dbID, NULL, &dbAttrs, NULL, NULL, NULL,
                                 NULL, NULL, &appInfoID, NULL, NULL, NULL)) {
        CloseDatabases();
        return err;
    }
    d.indexed_dbID = dbID;
    /* Set backup bits */
    dbAttrs |= dmHdrAttrBackup;
    /* Initialize categories if needed */
    if (appInfoID == NULL) {
        h = DmNewHandle(d.dbRI, sizeof(AppInfoType));
        if (!h) return dmErrMemError;
        appInfoID = MemHandleToLocalID(h);
        DmSetDatabaseInfo(0, dbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
                            &appInfoID, NULL, NULL, NULL);
        appInfoP = MemLocalIDToLockedPtr(appInfoID, 0);
        DmSet(appInfoP, 0, sizeof(AppInfoType), 0);
        CategoryInitialize(appInfoP, CategoryString);
        MemPtrUnlock(appInfoP);
        /* For some reason the resource string is not working.  Oh well...*/
        CategorySetName(d.dbRI, 0, "Unfiled");
        CategorySetName(d.dbRI, 1, "Personal");
        CategorySetName(d.dbRI, 2, "Business");
    }
    if (err = DmSetDatabaseInfo(0, dbID, NULL, &dbAttrs, NULL, NULL,
                                    NULL, NULL, NULL, NULL, NULL, NULL,
                                    NULL)) {
        CloseDatabases();
        return err;
    }
    /* Stay read-only as much as possible */
    SetDatabaseWriteMode(false, true); /* scratch */
    SetDatabaseWriteMode(false, false); /* indexed */

    /* Set the ShortCut popup list width */
    frm = FrmInitForm(DiddleTForm);
    list = GetObjectPointer(frm, PopShortcutList);
    d.shortcut_width = GetListWidth(list);
    FrmDeleteForm(frm);
}


/*
** LinkSetGlobals - set global variables according to passed LinkInfo
*/
UInt LinkSetGlobals(LinkDiddlePtr linkinfo) {
    if (DmFindRecordByID(d.dbRI, linkinfo->record_id, &(p.dbII))) {
        FrmAlert(NoRecordForm);
        return 1;
    }
    p.onscratch = false;
    d.newdiddle = false;
    d.savediddle = false;
    d.noselect = true;
    return 0;
}


/*
** EventLoop - handle events
*/
static void EventLoop(void) {
    EventType e;
    Word error;
    FormPtr frm;

    do {
        EvtGetEvent(&e, evtWaitForever);

        if (SpecialKeyDown(&e)) continue;

        if (SysHandleEvent(&e)) continue;

        if (MenuHandleEvent(NULL, &e, &error)) continue;

        if (e.eType == frmLoadEvent) {
            frm = FrmInitForm(d.formID = e.data.frmLoad.formID);
            FrmSetActiveForm(frm);
            switch (d.formID) {
                case DiddleForm:
                case DiddleTForm:
                    p.formID = d.formID;
                    FrmSetEventHandler(frm, DiddleEvent);
                    break;
                case IndexForm:
                    FrmSetEventHandler(frm, IndexEvent);
                    break;
                case PropForm:
                    FrmSetEventHandler(frm, PropEvent);
                    break;
            }
            continue;
        }

        FrmDispatchEvent(&e);

    } while (e.eType != appStopEvent);
}

/*
** StopApplication
*/
static void StopApplication(void) {
    PrefSetAppPreferencesV10(AppType, PrefVersion, &p, sizeof(p));
    CloseDatabases();
}


/*
** PilotMain - handle the toplevel command and event dispatch.
*/
DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags) {
    Err err;
    GoToParamsPtr pgoto;

    switch (cmd) {
    case sysAppLaunchCmdNormalLaunch:
        if(err = OpenDatabases()) return err;

        d.newdiddle = false; /* Always jump to existing diddle */
        d.savediddle = false;
        d.noselect = true;
        p.dbII = 0;
        if (p.onindex || p.start_on_index) FrmGotoForm(IndexForm);
        else FrmGotoForm(p.formID);

        EventLoop();
        StopApplication();
        break;
    case linkAppLaunchRequestLink:
        if (err = OpenDatabases()) return err;

        d.newdiddle = false;
        d.savediddle = false;
        d.noselect = true;
        p.dbII = 0;
        FrmGotoForm(IndexForm);

        EventLoop();
        StopApplication();
        break;
    case linkAppLaunchFollowLink:
        if (err = OpenDatabases()) return err;

        if(!LinkSetGlobals((LinkDiddlePtr)cmdPBP)) {
            FrmGotoForm(p.formID);
            EventLoop();
        }
        StopApplication();
        break;

    case sysAppLaunchCmdSaveData:
        /*if (p.formID == d.formID)
            SaveImage(); */
        break;

    case sysAppLaunchCmdGoTo:
        pgoto = (GoToParamsPtr)cmdPBP;
        if (launchFlags&sysAppLaunchFlagNewGlobals) {
            if (err = OpenDatabases()) return err;
            p.onindex = false;
            d.newdiddle = false;
            d.savediddle = false;
            d.noselect = true;
            if (pgoto->matchCustom) {
                p.dbII = pgoto->recordNum;
                p.onscratch = false;
            }
            else {
                p.dbI = pgoto->recordNum;
                p.onscratch = true;
            }
            FrmGotoForm(p.formID);

            EventLoop();
            StopApplication();
        } else {
            SendGotoEvent(pgoto);
        }
        break;

    case sysAppLaunchCmdFind:
        DiddleFind((FindParamsPtr)cmdPBP);
        break;

    case sysAppLaunchCmdExgReceiveData:
        if(launchFlags&sysAppLaunchFlagNewGlobals)
                BeamReceive((ExgSocketPtr)cmdPBP, false);
        else
                BeamReceive((ExgSocketPtr)cmdPBP, true);
        break;
    }

    return 0;

}

