#ifndef __diddle_h
#define __diddle_h

#include "diddleRsc.h"
#include <Pilot.h>
#include <System/SysEvtMgr.h>
#include <System/ExgMgr.h>

/* Application type identifier */
#ifndef AppType
#define AppType	0x4449444C	/* 'DIDL' */
#endif

/* Type for original Doodle */
#ifndef DoodleType
#define DoodleType	0x7263444F	/* 'rcDO' */
#endif

/* Preferences version - Incremented everytime pref format changes */
/* Diddle versions 1.0-1.3 = 1 */
/* Diddle versions 1.4 = 2     */
/* Diddle versions 1.5 = 3     */
/* Diddle versions 1.6-1.7 = 4 */
/* Diddle versions 1.8 = 5  */
/* Diddle versions 1.20-up = 6 */
#ifndef PrefVersion
#define PrefVersion 6
#endif

/* Database type */
#ifndef DBType
#define DBType	0x44617461	/* 'Data' */
#endif

/* Indexed Database type */
#ifndef DBIType
#define DBIType 0x496e6478 /* 'Indx' */
#endif

/* BitmapType */
#ifndef DiddleBitmapType
#define DiddleBitmapType 0x54626d70  /* Tbmp */
#endif

/* Database name */
#ifndef DBName
#define DBName "DiddleDB"
#endif

/* Indexed Database name */
#ifndef DBIName
#define DBIName "DiddleIDB"
#endif

/* Doodle compatibility DB name */
#ifndef DOODName
#define DOODName "DoodleDB"
#endif

/* Number of pixels allowed in a pen */
#ifndef NPENPIX
#define NPENPIX	32
#endif

/* Number of pen events which may be weighted together */
#ifndef NPENWTS
#define NPENWTS 8		/* must be power of two */
#endif

/* Pasting actions */
#define PASTE_NORMAL    0
#define PASTE_INVERSE   1
#define PASTE_TRANS     2
#define PASTE_XOR       3

#define PASTE_BUFFER_SIZE   128

/* Drawing modes */
#define MODE_DOODLE         0
#define MODE_TEXT           1
#define MODE_LINE           2
#define MODE_ARROW          3
#define MODE_SHAPE          4

/* Text insertion modes */
#define TXT_BLACK           0
#define TXT_BLACKONWHITE    1
#define TXT_WHITEONBLACK    2
#define TXT_XOR             3

/* Shape type codes */
#define SHAPE_RECTANGLE     0
#define SHAPE_ROUNDRECT     1
#define SHAPE_OVAL          2
#define SHAPE_CIRCLE        3

/* Shape fill codes */
#define SHFILL_OUTLINE      0
#define SHFILL_FILL         1
#define SHFILL_XOR          2

/* Shortcut codes */
#define SCUT_LINE           0
#define SCUT_RECTANGLE      1
#define SCUT_ROUNDRECT      2
#define SCUT_OVAL           3
#define SCUT_CIRCLE         4
#define SCUT_TEXT           5


/* Operations */
#define penFine		'.'
#define penMedium	bulletChr
#define penBold     degreeChr
#define penBroad	'@'
#define penFineItalic	','
#define penMediumItalic	'/'
#define penBroadItalic	';'

#define inkWhite	'0'
#define inkDith12	'1'
#define inkDith25	'2'
#define inkDith37	'3'
#define inkDith50	'4'
#define inkDith62	'5'
#define inkDith75	'6'
#define inkDith87	'7'
#define inkBlack	'8'
#define inkRandom	'9'

#define modeErase	'e'
#define modePaint	'p'
#define modeSmear	'm'
#define cmdRough	'<'
#define cmdSmooth	'='
#define cmdSmoother	'>'

#define cmdAbout	'a'
#define cmdPref     'h'
#define cmdClear	'x'
#define cmdFill		'f'
#define cmdInvert   'w'
#define cmdRemove	'r'
#define cmdTitle	't'
#define cmdIndex    'i'
#define cmdSave     's'
#define cmdScratch  's'
#define cmdDetails  'd'
#define cmdNew		'n'
#define cmdDuplicate	'd'
#define cmdPrevious	'-'
#define cmdNext		'+'
#define cmdLink		'g'
#define cmdCopy     'c'
#define cmdPaste    'v'
#define cmdText     'z'
#define cmdLine     'k'
#define cmdShape    'o'
#define cmdBeam     'b'
#define cmdInsertLast 'l'

/* Error messages */
#if 0
#define abort(msg)	ErrDisplayFileLineMsg(__FILE__, __LINE__, msg)
#else
#define abort(msg)	ErrDisplayFileLineMsg(__FILE__, __LINE__, "")
#endif

/*
** Extract the pen pixel offsets from the penpex array bytes.
** Each nibble is treated as a 2's complement integer.
*/
#define dxFromPenpix(p)	(((int)(p>>4))-8)
#define dyFromPenpix(p)	(((int)(p&0xF))-8)

/*
** Clipping macro -- PalmOS doesn't clip at screen bottom.
*/
#define rctPtInRectangle(tx,ty,rp) \
(   ((unsigned) ((tx)-(rp)->topLeft.x) < (rp)->extent.x) \
    && ((unsigned) ((ty)-(rp)->topLeft.y) < (rp)->extent.y) )

/*
** Global data structure.
*/
typedef struct {
    DmOpenRef dbR;			/* database handle for scratch buffer */
    DmOpenRef dbRI;			/* database handle for saved diddles */
    WinHandle winM;			/* main window */
    unsigned int formID;    /* Current formID (including index) */
	Boolean newdiddle;		/* For Properties form to know new from old */
	Boolean noselect;		/* Suppress automatic highlist in index */
    Boolean savediddle;     /* Copy p.dbI data to d.dbII */
    Byte locked;            /* Record is locked */
    short lock_count;       /* Number of times a locked screen has been tap */
    struct wts {
        unsigned short x, y;    /* incoming pen coordinates */
    } wts[NPENWTS];
    unsigned short nwts;        /* pen event coordinates */
    unsigned short wx, wy;      /* sum of pwt pen coordinates */
    unsigned short ox, oy;      /* last synthesized pen coordinates */
    short drawmode;             /* MODE_DOODLE, MODE_TEXT, MODE_SHAPE */
    short last_insert;          /* Memory of d.drawmode for Last Insert */
    Char CategoryName[dmCategoryLength];        /* Current category name */
    Word RecordCategory;                        /* Category for record */
    Char RecordCategoryName[dmCategoryLength];  /* Record category name */
    Char InsertionText[32];     /* Text buffer for the "Insert Text" mode */
    FontID txt_oldfont;       /* Old text font to restore after insert */
    PointType anchor;           /* Anchor used for line, arrow, and shapes */
    LocalID scratch_dbID;       /* dbID of the scratch database */
    LocalID indexed_dbID;       /* dbID of the indexed database */
    SWord shortcut_width;       /* Width of the shortcut popup list */
    SWord drop_list_width[4];   /* Width of the titlebar droplists */
} data;

typedef struct {
    unsigned int version;           /* version number */
    short dbI;                      /* page record for scratch buffer */
	short dbII;                     /* index record for named buffer */
    Boolean onindex;                /* On the sketch index */
	Boolean onscratch;              /* Using scratch or named db */
    Word category;                  /* Current category */
    unsigned int formID;            /* titled or untitled */
    RectangleType r;                /* current drawing window */
    unsigned char mode_label[2];    /* mode label for title and increment */
    unsigned char filt_label[2];    /* filter lable for title and increment */
    unsigned char pen_label[2];     /* pen label for increment */
    unsigned char ink_label[2];     /* ink label for increment */
    unsigned short filt;            /* pen event weighting */
    unsigned char penpix[NPENPIX];  /* pen coordinate offsets */
    unsigned short inkpat[4];       /* current ink pattern */
    FontID txt_font;                /* Font for insertion of text */
    short txt_mode;                 /* Text insertion mode */
    short shape_type;               /* shape mode: circle, rectangle, etc. */
    short shape_fill;               /* Whether shapes are filled or not */
    Word hard_button;               /* Hardware button */
    Boolean start_on_index;         /* Application starup on sketch index */
    Boolean disable_title_sc;       /* disable the titlebar shortcut menu */
    Boolean disable_button_drop;    /* Drop down lists from titlebar buttons */
    unsigned char erase_pen_label[2];  /* Pen for eraser */
} pref;

typedef data *DataPtr;
typedef pref *PrefPtr;

/* Shortcut menu locations */
enum sc_locations {high, low};

#endif /* __diddle_h */
