/* linkaware.h
   This file released to the public domain by Ben Darnell.
   You may use this file however you wish to make
   LinkMaster-compatible applications.
*/
#ifndef _linkaware_h
#define _linkaware_h

#include "diddle.h"

/* All link-aware applications should include a resource of type
   'Link', number 1, containing a character indicating what version
   of the LinkMaster protocol the application implements.  For this
   version, that character should be 'a'.

   If you are using the GNU tools, you can create this resource by
   executing "echo a > Link0001.bin" and including Link0001.bin
   in your build-prc command.
*/

typedef struct {
   DWord creator;
   Char data[60];
} LinkInfoType;
typedef LinkInfoType* LinkInfoPtr;


#define linkAppLaunchRequestLink 0x8901
/* Sending: An application should send this launch code to LinkMaster to
   request a link.  The parameter block consists of one LinkInfoType
   structure, which will be returned with the link.

   Recieving: LinkMaster will send this launch code to an application
   when the user selects the application via the "New Link" command.
   No parameters will be sent.  The application is not required to
   behave specially in response to this launch code, but may for instance
   open in an index mode instead of viewing the previously edited record.
*/

#define linkAppLaunchReturnLink 0x8902
/* Sending: An application should send this launch code to LinkMaster after
   the user has selected a record to link to.  This may be done at any time,
   not just after recieving a linkAppLaunchRequestLink.  If there are no
   pending link requests, the link will be added to a list within
   LinkMaster.  The parameter block consists of one LinkInfoType structure
   plus an optional null-terminated string, which may be used to identify
   the link.

   Recieving: LinkMaster will send this launch code to an application which
   has previously requested a link.  The parameter block consists of two
   LinkInfoType structures - the one initially passed with
   linkAppLaunchRequestLink, and the one selected by the user - and a null-
   terminated string which identifies the link..
*/

#define linkAppLaunchFollowLink 0x8903
/* Recieving: This code is sent to applications when the user follows
   a link.  The parameter block consist of the LinkInfoType structure
   that defines the link.  The application should display the record
   indicated by the link.
*/

typedef struct {
   DWord creator;
   ULong record_id;
} LinkDiddleType;
typedef LinkDiddleType* LinkDiddlePtr;

LinkDiddlePtr LinkDiddleNew(DataPtr d, PrefPtr p);

void LinkPublish(LinkDiddlePtr linkinfo);

#endif /* _linkaware_h */
