//=============================================================================
// $Id: dstpanel.c,v 1.2 2000/04/18 20:59:28 unknown Exp unknown $
//=============================================================================

/*
Copyright (C) 2000 Remo Hofer <cocos.nucifera@gmx.net>
For more information see http://www.geocities.com/SiliconValley/Cable/5206/

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

// include header

#include <PalmOS.h>
#include "callback.h"
#include "rsrc.h"
#include "rsrcfix.h"

/*
 *	DEBUG >= 2	Include code for silent errorcheck. Does only interact with 
 *				the user in the case of an error. 
 *	DEBUG >= 4	Include code for silent logging to the emulator log.
 *	DEBUG >= 6	Include code that may interact with (display info to) the 
 *				user without making the program unusable.
 *	DEBUG >= 8	Include code that may make the program almost unuseable.
 */
#define DEBUG 0

#define Creator				'DSTp'
#define PrefsId				0
#define PrefsVersion		0
#define MinRequiredVersion	sysMakeROMVersion(2,0,0,sysROMStageRelease,0)
#define FirstWeek			0
#define LastWeek			4

#define GracePeriod			300 /* 5 Minutes */
#define OneHourInSecs		3600
#define OneHourInMins		60

// mask out lowest bit of seconds to store start or stop event type
#define SecondsEventMask	0xfffffffe
#define StartDstEvent		0x00000000
#define EndDstEvent			0x00000001

// defaults for Europe	
#define DefaultStartHour	2
#define DefaultStartMinute	0
#define DefaultStartDay		sunday
#define DefaultStartWeek	LastWeek
#define DefaultStartMonth	(march-1) /* prefs month begin with 0 (not 1)! */
#define DefaultEndHour		3
#define DefaultEndMinute	0
#define DefaultEndDay		sunday
#define DefaultEndWeek		LastWeek
#define DefaultEndMonth		(october-1) /* prefs month begin with 0 (not 1)! */
// defaults for USA
#define DefaultStartWeekUS	FirstWeek
#define DefaultStartMonthUS	(april-1) /* prefs month begin with 0 (not 1)! */
#define DefaultEndHourUS	2

typedef struct {
	UInt8	startHour;
	UInt8	startMinute;
	UInt8	startDay;
	UInt8	startWeek;
	UInt8	startMonth;
	UInt8	endHour;
	UInt8	endMinute;
	UInt8	endDay;
	UInt8	endWeek;
	UInt8	endMonth;
	Boolean	endDstAmbigousTime;
} PrefsType;

// globals

static PrefsType			prefs;
static Boolean				prefsChanged = false;
static Boolean				calledFromApp = false;
static Char					startStr[timeStringLength];
static Char					endStr[timeStringLength];
static SysDBListItemType	*panelIDsP = NULL;	// ptr to panel pick list items

//-----------------------------------------------------------------------------
#if DEBUG >= 4
#include <stdarg.h>
#include "gccHostControl.h"

void EmulatorLog(Char * format,...) {
	va_list	args;
	UInt32	dummy;
	UInt32	romVersion;
	Char	s[64];
	
	va_start(args, format);
	StrVPrintF(s, format, args);
	va_end(args);

	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
	if ((romVersion >= sysMakeROMVersion(3,0,0,sysROMStageRelease,0)) &&
	  (FtrGet('pose', 0, &dummy) == 0)) {
		HostFPutS(s, HostLogFile());
	}
	
#if DEBUG >= 8
	WinDrawChars(s, StrLen(s), 0, 0);
#endif

}
#endif

//-----------------------------------------------------------------------------
static void * FormGetObjectPtr(FormType * formP, UInt16 objId) {
	return (FrmGetObjectPtr(formP, FrmGetObjectIndex(formP, objId)));
}

//-----------------------------------------------------------------------------
static void FormShowObject(FormType * formP, UInt16 objId) {
	FrmShowObject(formP, FrmGetObjectIndex(formP, objId));
}

//-----------------------------------------------------------------------------
static void PanelFreeData(void) {
		
#if DEBUG >= 4
	EmulatorLog("PanelFreeData");
#endif

	if (panelIDsP) {
		MemPtrFree(panelIDsP);
		panelIDsP = NULL;
	}
}
			
//-----------------------------------------------------------------------------
static void PanelDone(void) {
	EventType		newEvent;
	
#if DEBUG >= 4
	EmulatorLog("PanelDone");
#endif

	// by simply exiting we return to the prior which called us
	MemSet(&newEvent, sizeof(EventType), 0);
	newEvent.eType = appStopEvent;
	EvtAddEventToQueue(&newEvent);
}

//-----------------------------------------------------------------------------
static void PanelSelect(UInt16 index) {
  	UInt16		currentAppCardNo;
	LocalID		currentAppDBID;
	
#if DEBUG >= 4
	EmulatorLog("PanelSelect");
#endif

	if (panelIDsP) {
		
		// Switch to the panel if it isn't this one
		SysCurAppDatabase(&currentAppCardNo, &currentAppDBID);

		if ((panelIDsP[index].dbID != currentAppDBID) ||
		  (panelIDsP[index].cardNo != currentAppCardNo)) {
			SysUIAppSwitch(
			  panelIDsP[index].cardNo, panelIDsP[index].dbID, 
			  sysAppLaunchCmdNormalLaunch, 0);
		}
	}
}

//-----------------------------------------------------------------------------
static void PanelDrawItem(Int16 itemNum, RectanglePtr bounds, Char ** dummy) {
	Char * itemText;
	
	CALLBACK_PROLOGUE
	
#if DEBUG >= 4
	EmulatorLog("PanelDrawItem");
#endif

	if (panelIDsP) {
		itemText = panelIDsP[itemNum].name;
		WinDrawChars(itemText, StrLen(itemText), 
		  bounds->topLeft.x, bounds->topLeft.y);
	}

	CALLBACK_EPILOGUE
}

//-----------------------------------------------------------------------------
// Create a list of panels available and select the current one.
// panelIDsP is set.
static void PanelInit(Boolean calledFromApp, FormPtr f, 
  UInt16 idPopup, UInt16 idList, UInt16 idLabel, UInt16 idDone) {

#if DEBUG >= 4
	EmulatorLog("PanelInit");
#endif

	if (calledFromApp) {
		// hide panel popup and show done button
		
		FormShowObject(f, idLabel);
		FormShowObject(f, idDone);
		
	} else {
		// create panel list and show popup trigger
		
		ListPtr		listP;
	  	UInt16		currentAppCardNo;
		LocalID		currentAppDBID;
		MemHandle	panelIDsH;	// MemHandle to pick list items
		UInt16		panelCount, i;
		PrefActivePanelParamsType prefs;
		
		// create the panel list data
		if (SysCreatePanelList(&panelCount, &panelIDsH)) {
	
			panelIDsP = MemHandleLock(panelIDsH);
	
			listP = FormGetObjectPtr(f, idList);
		
			// Now set the list to hold the number of panels found.  There
			// is no array of text to use.
			LstSetListChoices(listP, NULL, panelCount);
	
			// Now resize the list to the number of panel found
			LstSetHeight(listP, panelCount);
	
			// Because there is no array of text to use, we need a function
			// to interpret the panelIDsP list and draw the list items.
			LstSetDrawFunction(listP, PanelDrawItem);
	
			// Now we should select the item in the list which matches the 
			// current app.
			SysCurAppDatabase(&currentAppCardNo, &currentAppDBID);
			for (i = 0; i < panelCount; i++) {
				if (panelIDsP[i].dbID == currentAppDBID &&
				  panelIDsP[i].cardNo == currentAppCardNo) {
					LstSetSelection(listP, i);
					CtlSetLabel(
					  FormGetObjectPtr(f, idPopup), 
					  panelIDsP[i].name);
					break;
				}
			}
			  
			// draw the popup trigger 
			FormShowObject(f, idPopup);

			// set this panel as the active one next time the 
			// Prefs app is opened
			prefs.activePanel = Creator;
			AppCallWithCommand(sysFileCPreferences, 
			  prefAppLaunchCmdSetActivePanel, &prefs);
		}
	}
}

//-----------------------------------------------------------------------------
static void UpdatePopup(FormPtr f, UInt16 popupId, UInt16 listId, 
  UInt16 index) {

#if DEBUG >= 4
	EmulatorLog("UpdatePopup");
#endif

	CtlSetLabel(
	  FormGetObjectPtr(f, popupId), 
	  LstGetSelectionText(FormGetObjectPtr(f, listId), index));
}  
 	
//-----------------------------------------------------------------------------
static void UpdateTime(FormPtr f, UInt16 id, UInt8 hour, UInt8 min, Char * s) {
	
#if DEBUG >= 4
	EmulatorLog("UpdateTime");
#endif

	TimeToAscii(hour, min, PrefGetPreference(prefTimeFormat), s);
	CtlSetLabel(FormGetObjectPtr(f, id), s);	   
}

//-----------------------------------------------------------------------------
static void UpdateForm(FormPtr f) {

#if DEBUG >= 4
	EmulatorLog("UpdateForm");
#endif

	UpdateTime(f, idStartTimeSelector, 
	  prefs.startHour, prefs.startMinute, startStr);
	UpdatePopup(f, idStartWeekPopup, idWeekList, prefs.startWeek);
	UpdatePopup(f, idStartDayPopup, idDayList, prefs.startDay);
	UpdatePopup(f, idStartMonthPopup, idMonthList, prefs.startMonth);
	UpdateTime(f, idEndTimeSelector, 
	  prefs.endHour, prefs.endMinute, endStr);
	UpdatePopup(f, idEndWeekPopup, idWeekList, prefs.endWeek);
	UpdatePopup(f, idEndDayPopup, idDayList, prefs.endDay);
	UpdatePopup(f, idEndMonthPopup, idMonthList, prefs.endMonth);
}

//-----------------------------------------------------------------------------
static void PopupList(UInt8 * value, 
  FormPtr f, UInt16 triggerId, UInt16 listId) {
	Int16	newi, oldi;
	ListPtr	list;
	Coord	x, y;

#if DEBUG >= 4
	EmulatorLog("PopupList");
#endif

	list = FormGetObjectPtr(f, listId);
	FrmGetObjectPosition(f, FrmGetObjectIndex(f, triggerId), &x, &y);
	LstSetPosition(list, x, y);
	oldi = (Int16) *value;
	LstSetSelection(list, oldi);
	newi = LstPopupList(list);
	if ((newi >= 0) && (newi != oldi)) {
		*value = (UInt8) newi;
		UpdatePopup(f, triggerId, listId, *value);
		prefsChanged = true;
	}
}

//-----------------------------------------------------------------------------
static void TimeSelector(UInt8 * hour, UInt8 * min, FormPtr f, UInt16 id,
  Char * s) {
	TimeType	time, dummy;
	UInt32		sysRomVersion;
	Int16		hour16, min16;
	MemHandle	h;
	Char		*title;

#if DEBUG >= 4
	EmulatorLog("TimeSelector");
#endif

  	h = DmGetResource(strRsc, idSelectTimeStr);
  	title = MemHandleLock(h);

	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &sysRomVersion);
	if (sysRomVersion < sysMakeROMVersion(3,1,0,sysROMStageRelease,0)) {
		time.hours = *hour;
		time.minutes = *min;
		dummy = time;
		if (SelectTimeV33(&time, &dummy, false, title, 0)) {
			if (*(Int16*)&time != noTime) {
				*hour = time.hours;
				*min = time.minutes;
				prefsChanged = true;
			}
		}
	} else {
		hour16 = *hour;	// implicit type cast
		min16 = *min;
		if (SelectOneTime(&hour16, &min16, title)) {
			*hour = hour16;	// implicit type cast
			*min = min16;
			prefsChanged = true;
		}
	}

	MemHandleUnlock(h);
  	DmReleaseResource(h);

	UpdateTime(f, id, *hour, *min, s);
}

//-----------------------------------------------------------------------------
// no globals
static void SavePrefs(PrefsType * p) {

#if DEBUG >= 4
	EmulatorLog("SavePrefs");
#endif

	PrefSetAppPreferences(Creator, PrefsId, PrefsVersion, 
	  p, sizeof(PrefsType), true);	
}
		
//-----------------------------------------------------------------------------
// no globals
static void InitPrefs(PrefsType * p) {
	
#if DEBUG >= 4
	EmulatorLog("InitPrefs");
#endif

	p->startHour = DefaultStartHour;
	p->startMinute = DefaultStartMinute;
	p->startDay = DefaultStartDay;
	p->startWeek = DefaultStartWeek;
	p->startMonth =  DefaultStartMonth;

	p->endHour = DefaultEndHour;
	p->endMinute = DefaultEndMinute;
	p->endDay = DefaultEndDay;
	p->endWeek = DefaultEndWeek;
	p->endMonth =  DefaultEndMonth;
	
	p->endDstAmbigousTime = false;
}

//-----------------------------------------------------------------------------
// no globals
static void LoadPrefs(PrefsType * p) {
	UInt16		size;
	UInt16		version;

#if DEBUG >= 4
	EmulatorLog("LoadPrefs");
#endif

	size = sizeof(PrefsType);
	version = PrefGetAppPreferences(Creator, PrefsId,  p, &size, true);
	if ((version != PrefsVersion) || (size != sizeof(PrefsType))) 
	  InitPrefs(p);
}

//-----------------------------------------------------------------------------
// no globals
static UInt32 PrefTimeToSeconds(UInt8 hour, UInt8 min, UInt8 weekday, UInt8 week, 
 UInt8 month, Int16 year) {
 	DateTimeType	date;
 	Int16			day;

#if DEBUG >= 4
	EmulatorLog("PrefTimeToSeconds");
#endif

	month++; // prefs month begin with 0 (not 1)!

	// get the day which matches the rule (eg. last Sunday)
 	if (week == LastWeek) {
 		day = DaysInMonth(month, year);
 		while (DayOfWeek(month, day, year) != weekday) day--;
 	} else {
 		day = 1 + (7 * week);
 		while (DayOfWeek(month, day, year) != weekday) day++;
 	}
 	
 	date.second = 0;
 	date.minute = min;
 	date.hour = hour;
 	date.day = day;
 	date.month = month;
 	date.year = year;
 	date.weekDay = weekday;
 	
 	return (TimDateTimeToSeconds(&date));
}

//-----------------------------------------------------------------------------
// no globals
static Err ResetAlarm(void) {
	UInt32			time, alarmTime, startTime, endTime, type = 0;
	DateTimeType	date;
	UInt16			cardNo;
	LocalID			dbId;
	PrefsType		prefs;
	DmSearchStateType	state;

#if DEBUG >= 4
	EmulatorLog("ResetAlarm");
#endif

	LoadPrefs(&prefs);
	
	do { // make sure alarm is in the future
		
		alarmTime = 0;
		time = TimGetSeconds();
		TimSecondsToDateTime(time, &date);
	
		while (alarmTime < time) {

			// convert date rule to seconds
			startTime = PrefTimeToSeconds(prefs.startHour, prefs.startMinute, 
			  prefs.startDay, prefs.startWeek, prefs.startMonth, date.year);
			endTime = PrefTimeToSeconds(prefs.endHour, prefs.endMinute, 
			  prefs.endDay, prefs.endWeek, prefs.endMonth, date.year);
	  
			// get the next rule date
			if (((startTime < endTime) && (startTime > time)) ||
			  ((startTime > endTime) && (endTime < time))) { 
				alarmTime = startTime;
				type = StartDstEvent;
			} else {
				alarmTime = endTime;
				type = EndDstEvent;
			}
		
			// if both start and end date are passed try next year
			date.year++;
	
		}
	
		if (((type != EndDstEvent) || 
		  (time < alarmTime - OneHourInSecs - GracePeriod)) &&
		  prefs.endDstAmbigousTime) {
			// we are definitly NOT anymore in the ambigous time period
			// maybe we just jumped out of it with a user time change
			prefs.endDstAmbigousTime = false;
		  	SavePrefs(&prefs);
		}
		
		// SysCurAppDatabase doesn't seem to work during reset on PalmOS 2
		// SysCurAppDatabase(&cardNo, &dbId);
		DmGetNextDatabaseByTypeCreator(true, &state, sysFileTPanel, Creator,
		  true, &cardNo, &dbId);

#if DEBUG >= 4
		{
			Char			datestr[longDateStrLength];
			Char			timestr[timeStringLength];
	 		DateTimeType	date;
		
   			TimSecondsToDateTime(alarmTime, &date);
			DateToAscii(date.month, date.day, date.year, dfDMYLongWithDot, 
			  datestr);
			TimeToAscii(date.hour, date.minute, tfColon24h, timestr);
			EmulatorLog("Alarm Time: %s %s", datestr, timestr);
		}
#endif

	// if alarmTime passes during calculation, recalc next alarm
	} while (alarmTime <= TimGetSeconds());

	// set the alarm and return the error code
	return (AlmSetAlarm(cardNo, dbId, 
	  ((alarmTime & SecondsEventMask) | type), 
	  alarmTime, true));
}

//-----------------------------------------------------------------------------
// no globals
static void ChangeTime(UInt32 ref) {
	PrefsType		p;
	Int32			minutesWestOfGMT, deltaWestOfGMT=0;
	UInt32			time, alarmTime, type;

	type = ref & (~SecondsEventMask);

#if DEBUG >= 4
	EmulatorLog("ChangeTime: %lu", type);
#endif

	alarmTime = ref & SecondsEventMask;
	time = TimGetSeconds();

	// only process if we are close to the scheduled time
	if((time > alarmTime - GracePeriod) && (time < alarmTime + GracePeriod)) {
	
  		// change time
  		if (type == StartDstEvent) {
 			TimSetSeconds(TimGetSeconds() + OneHourInSecs);
			deltaWestOfGMT = OneHourInMins;
		} else {
			LoadPrefs(&p);
	 		if (p.endDstAmbigousTime) {
  				// we are leaving the ambigous time period, so
  				// we don't change time (did it already one hour ago)
				p.endDstAmbigousTime = false;
			} else {
				TimSetSeconds(TimGetSeconds() - OneHourInSecs);
				deltaWestOfGMT = - OneHourInMins;
				// since we changed the time backwards, the endDST time will
				// reoccur in one hour!
				p.endDstAmbigousTime = true;
			}
		  	SavePrefs(&p);
		}
	  		  	
  		// adjust "timezone"
  		minutesWestOfGMT = (Int32)PrefGetPreference(prefMinutesWestOfGMT);
		PrefSetPreference(prefMinutesWestOfGMT, 
		  minutesWestOfGMT + deltaWestOfGMT);
	}
}

//-----------------------------------------------------------------------------
// no globals
static void RegisterNotify(void) {
	UInt16			cardNo;
	LocalID			dbId;
	UInt32			value;
	Err				error;
	
#if DEBUG >= 4
	EmulatorLog("RegisterNotify");
#endif

	// test if the notification manager is available
	error = FtrGet(sysFtrCreator, sysFtrNumNotifyMgrVersion, &value);
	if ((error == 0) && (value != 0)) {
		SysCurAppDatabase(&cardNo, &dbId);
		// we want to know when the time is changed by the user
		SysNotifyRegister(cardNo, dbId, sysNotifyTimeChangeEvent,
		  NULL, sysNotifyNormalPriority, NULL);
	}
}
//-----------------------------------------------------------------------------
static Boolean MainFormHandleEvent(EventPtr event) {
	FormPtr			formPtr;
	MemHandle		h;
	Boolean			handled = false;
	
	CALLBACK_PROLOGUE
	
#if DEBUG >= 4
	EmulatorLog("MainFormHandleEvent: %u", event->eType);
#endif

	formPtr = FrmGetFormPtr(idMainForm);

	switch (event->eType) {

	  case frmOpenEvent:
		// get the dst rule
		LoadPrefs(&prefs);
		// init all form elements
		UpdateForm(formPtr);
		// setup the version string
  		h = DmGetResource(verRsc, idVersion);
		FrmCopyLabel(formPtr, idVersionLabel, MemHandleLock(h));
  		MemHandleUnlock(h);
  		DmReleaseResource(h);
 		// draw the form
		FrmDrawForm(formPtr);
		
#if DEBUG >= 6
		{
			Char			s[64];
			UInt32			dummy;
			UInt16			cardNo;
			LocalID			dbId;
 		 	DateTimeType	date;
		
 		 	SysCurAppDatabase(&cardNo, &dbId);
  			TimSecondsToDateTime(AlmGetAlarm(cardNo, dbId, &dummy), &date);
			DateToAscii(date.month, date.day, date.year, dfDMYLongWithDot, s);
			StrCat(s, " ");
			TimeToAscii(date.hour, date.minute, tfColon24h, s+StrLen(s));
			WinDrawChars(s, StrLen(s), 0, 148);
		}
#endif

		// Now that the form is displayed, generate the panel pick list.
		// By occuring here after everything is displayed the time required
		// isn't noticed by the user.
		PanelInit(calledFromApp, formPtr, idPanelPopup, idPanelList, 
		  idPanelLabel, idDoneButton);
		handled = true;
		break;
		
	  case frmCloseEvent:
		PanelFreeData();
		if (prefsChanged) SavePrefs(&prefs);
		// reset the alarm even if the prefs haven't changed, 
		// because it may be the first time the program is called.
		// e.g. it was beamed to the device (no sysAppLaunchCmdSyncNotify)
		if (ResetAlarm()) FrmAlert(idErrSetAlarmAlert);
		// handled = false; // give std form handler a chance to cleanup
		break;
		
	  case menuEvent:
		switch (event->data.menu.itemID) {
		  case idDefaultMenu:
		  case idDefaultUSMenu:
			// reset prefs to defaults
			InitPrefs(&prefs);
			if (event->data.menu.itemID == idDefaultUSMenu) {
				// reset prefs to defaults for USA
				prefs.startWeek = DefaultStartWeekUS;
				prefs.startMonth =  DefaultStartMonthUS;
				prefs.endHour = DefaultEndHourUS;
			}
			prefsChanged = true;
			UpdateForm(formPtr);
			handled = true;
			break;
		  case idThanksMenu:
			// show about form
			FrmHelp(idThanksStr);
			handled = true;
			break;
		  case idAboutMenu:
			// show about form
			FrmHelp(idAboutStr);
			handled = true;
			break;
		}
		break;

	  case ctlSelectEvent:
		// one of the controls (popup/selector trigger) is selected
		switch (event->data.ctlSelect.controlID) {
		  case idStartTimeSelector:
			TimeSelector(&prefs.startHour, &prefs.startMinute, 
			  formPtr, idStartTimeSelector, startStr);
			handled = true;
			break;
		  case idStartWeekPopup:
			PopupList(&prefs.startWeek, formPtr, idStartWeekPopup, idWeekList);
			handled = true;
			break;
		  case idStartDayPopup:
			PopupList(&prefs.startDay, formPtr, idStartDayPopup, idDayList);
			handled = true;
			break;
		  case idStartMonthPopup:
			PopupList(&prefs.startMonth, formPtr,idStartMonthPopup,idMonthList);
			handled = true;
			break;
		  case idEndTimeSelector:
			TimeSelector(&prefs.endHour, &prefs.endMinute,
			  formPtr, idEndTimeSelector, endStr);
			handled = true;
			break;
		  case idEndWeekPopup:
			PopupList(&prefs.endWeek, formPtr, idEndWeekPopup, idWeekList);
			handled = true;
			break;
		  case idEndDayPopup:
			PopupList(&prefs.endDay, formPtr, idEndDayPopup, idDayList);
			handled = true;
			break;
		  case idEndMonthPopup:
			PopupList(&prefs.endMonth, formPtr,idEndMonthPopup,idMonthList);
			handled = true;
			break;
		  case idDoneButton:
			// if calledFromApp this is the way to end the program
			PanelDone();
			handled = true;
			break;
		}
		break;

	  case popSelectEvent:
		switch (event->data.popSelect.listID) {
		  case idPanelList:
			// a new panel has been selected
			PanelSelect(event->data.popSelect.selection);
			handled = true;
			break;
		}
		break;
				
	  default:
	}
	
	CALLBACK_EPILOGUE
	return (handled);
}

//-----------------------------------------------------------------------------
static Boolean ApplicationHandleEvent(EventPtr event) {
	FormPtr	formPtr;
	
#if DEBUG >= 4
	EmulatorLog("ApplicationHandleEvent");
#endif

	if (event->eType == frmLoadEvent) {

		// we have only one form!
		// load the form resource and set the event handler
		formPtr = FrmInitForm(event->data.frmLoad.formID);
		FrmSetActiveForm(formPtr);		
		FrmSetEventHandler(formPtr, MainFormHandleEvent);

		return (true);
		
	}
			
	return (false);

}

//-----------------------------------------------------------------------------
UInt32 PilotMain(UInt16 cmd, void * cmdPBP, UInt16 launchFlags) {
	UInt32 romVersion;
	EventType	event;
	Err			error;
		
#if DEBUG >= 4
	EmulatorLog("PilotMain: %u", cmd);
#endif

	// see if we're on in minimum required version of the ROM or later
	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
	if (romVersion < MinRequiredVersion) {
		if ((launchFlags & (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp))
		  == (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) {
		  	FrmAlert(idErrOldRomAlert);
		}
		return (sysErrRomIncompatible);
	}
	
	// handle launch codes
	switch (cmd) {
		
	  case sysAppLaunchCmdPanelCalledFromApp:
		// we are called directly from an App (not by Prefs)
		calledFromApp = true;
	  case sysAppLaunchCmdNormalLaunch:
		// normal call from Prefs (or other panel)
	  case sysAppLaunchCmdReturnFromPanel:
		// does this happen if we switch a lot between panels?
		
		// we want to be notified about time changes
		RegisterNotify();
		// go to the initial form
		FrmGotoForm(idMainForm);
		// go into the event loop
		do {
			EvtGetEvent(&event, evtWaitForever);
			if (! SysHandleEvent(&event)) {
				if (! MenuHandleEvent(NULL, &event, &error)) {
					if (! ApplicationHandleEvent(&event)) {
						FrmDispatchEvent(&event);
					}
				}
			}
		} while (event.eType != appStopEvent);	
		// send a frmCloseEvent to active form handlers
		FrmCloseAllForms();
		break;
		
	  case sysAppLaunchCmdSyncNotify:
		// we are freshly installed by a HotSync
	  case sysAppLaunchCmdSystemReset:
		// the unit has been reset
		// so we need to tell the system about time change notifications 
		RegisterNotify();
	  case sysAppLaunchCmdTimeChange:
		// time has changed (old fashioned global launch code)
	  case sysAppLaunchCmdNotify:
		// time has changed (new notification manager)
		// so rescedule the alarm time
		ResetAlarm();
		break;
		
	  case sysAppLaunchCmdAlarmTriggered:
		// it's time to begin or end DST
		// we don't need a sysAppLaunchCmdDisplayAlarm code
		((SysAlarmTriggeredParamType *)cmdPBP)->purgeAlarm = true;
		// change the time
		ChangeTime(((SysAlarmTriggeredParamType *)cmdPBP)->ref);
		// schedule next alarm
		ResetAlarm();
		break;
		
	}

	return (0);
}

//=============================================================================
