/*   
 * $Id: calc.c,v 1.6 2000/11/25 14:50:58 ondrap Exp $
 * 
 * Scientific Calculator for Palms and PilmPilots.
 *   Copyright (C) 1999,2000 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
 * 
 *  Antonio Fiol Bonnn" <Antonio.FIOL@enst-bretagne.fr> submitted patches 
 *   for the financial part.
*/

#include <PalmOS.h>
#include <StringMgr.h>
#include <stdio.h>

#include "defuns.h"
#include "MathLib.h"
#include "calc.h"
#include "calcDB.h"
#include "konvert.h"
#include "mathem.h"
#include "stack.h"
#include "funcs.h"
#include "result.h"
#include "ansops.h"
#include "history.h"
#include "defmgr.h"
#include "calcrsc.h"    
#include "graph.h"
#include "finance.h"
#include "varmgr.h"
#include "prefs.h"
#include "fp.h"
#include "memo.h"

#include "pane.h"

#ifdef GRAPHS_ENABLED
 
#include "grprefs.h"
#include "grsetup.h"
#include "graph.h"
#include "grtable.h"

#endif /* GRAPH_ENABLED */

Boolean palmOS35=false;
Boolean colorDisplay=false;
static void default_form_update(void);

void *GetObjectPtr(UInt16 objectID)
{
	FormPtr frm;
	
	frm=FrmGetActiveForm();
	return (FrmGetObjectPtr(frm,FrmGetObjectIndex(frm,objectID)));
}

void
gadget_bounds(Int16 gadget,RectangleType *bounds)
{
	FormPtr frm = FrmGetActiveForm();
	UInt16 index = FrmGetObjectIndex(frm,gadget);
	
	FrmGetObjectBounds(frm,index,bounds);
}

static char *historylist[MAX_HISTORY];

static void
PushHistory(char *text)
{
	int i;
	
	for (i=MAX_HISTORY-1;i>0;i--)
	  StrCopy(historylist[i],historylist[i-1]);
	StrCopy(historylist[0],text);
}

static void 
calc_init_application()
{
	int i;

	prefs_read_preferences();
/* Initialize the field for the list resource */
	for (i=0;i<MAX_HISTORY;i++)
	  historylist[i]=calcPrefs.history[i];
}

/* Change to the next screen */
void
GotHardKey(void)
{
	int formId;
	
	formId=(FrmGetActiveForm()->formId)+1;
#ifdef GRAPHS_ENABLED
	if (formId>tdFormGraph)	
	  formId=tdFormPane;
#else
	if (formId>tdFormInt)
	  formId=tdFormPane;
#endif
	ChangeForm(formId);
}

/* Generic button action - actions for buttons  on all screens */
Int16
chooseForm(Int16 button)
{
	Boolean handled = true;

	switch (button) {
	 case tdHelp2:
		FrmHelp(tdHelp2);
		break;
	 case tdAbout:
		FrmAlert(tdAlertAbout);
		break;
	 case tdBInt:
		ChangeForm(tdFormInt);
		break;
	 case tdBBasic:
		ChangeForm(tdFormBasic);
		break;
#ifdef GRAPHS_ENABLED		
	 case tdBGra:
		ChangeForm(tdFormGraph);
		break;
#endif		
	 case tdBSci:
		ChangeForm(tdFormScientific);
		break;
	 default:
		handled = false;
	}
	return handled;
}

static Err 
StartApplication(void)
{
	UInt32 osversion;
	Err error;

	/* Get version of PalmOS */
	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &osversion);
	if (sysGetROMVerMajor(osversion)==3 &&
		sysGetROMVerMinor(osversion)>=5)
	  palmOS35 = true;
	else
	  palmOS35 = false;
	
	if (palmOS35)
	  WinScreenMode(winScreenModeGet,NULL,NULL,NULL,&colorDisplay);
	
	/* Initialize all modules */
	mlib_init_mathlib();
	
	error = db_open();
	ErrFatalDisplayIf(error, "Can't open CalcDB");
	
	error = history_open();
	ErrFatalDisplayIf(error, "Can't open History DB");
	
	calc_init_application();
	
	fp_setup_flpoint();

#ifdef GRAPHS_ENABLED	
	grpref_init();
#endif
	
	pane_init();
	
	FrmGotoForm(calcPrefs.form);
	
	return 0;
}

static void 
StopApplication(void)
{	
	FrmSaveAllForms();
	FrmCloseAllForms();

	/* close all modules */
	result_destroy();
	
	mlib_close_mathlib();
	
	prefs_save_preferences();
	
	db_close();
	
	history_close();
#ifdef GRAPHS_ENABLED	
	grpref_close();
#endif	
}

char *
print_error(CError err)
{
	char *res;

	res = MemPtrNew(MAX_RSCLEN);
	SysStringByIndex(strErrCodes,err,res,MAX_RSCLEN-1);
	return res;
}

static void
ReplaceEnters(char *text)
{
	for (;*text;text++)
	  if (*text=='\n')
	    *text=';';
}
	  
Int16
calc_input_exec(void)
{
	char *inp;
	CError err;
	FieldPtr pole;
	int inpsize;
	Trpn vysledek;
	CodeStack *stack;
	
	inp = FldGetTextPtr(GetObjectPtr(tdInput));
	if (!inp)
	  return 1;
	inpsize=StrLen(inp);
	
	ReplaceEnters(inp); 
	
	stack = text_to_stack(inp,&err);
	if (!err) {
		err = stack_compute(stack);
		if (!err) {
			vysledek = stack_pop(stack);
			set_ans_var(vysledek);
			result_set(vysledek);
			rpn_delete(vysledek);
		}
		stack_delete(stack);
	}
	if (err) 
	  result_error(err);

	result_draw();

	pole=GetObjectPtr(tdInput);
	if (!err) {
		FldSetSelection(pole,0,inpsize);
		FrmSetFocus(FrmGetActiveForm(),FrmGetObjectIndex(FrmGetActiveForm(),tdInput));
	}
	pole=GetObjectPtr(tdInput);
	if (FldGetTextPtr(pole))
	  PushHistory(FldGetTextPtr(pole));
	return err;
}

struct {
	UInt16 id;
	char *string;
	Boolean operator; /* Prepend Ans when in the beggining of line */
	Boolean parenth;  /* Add ')' and backspace one position */
}tdButStr[]={	
	{tdBut0,"0",false},
	{tdBut1,"1",false},
	{tdBut2,"2",false},
	{tdBut3,"3",false},
	{tdBut4,"4",false},
	{tdBut5,"5",false},
	{tdBut6,"6",false},
	{tdBut7,"7",false},
	{tdBut8,"8",false},
	{tdBut9,"9",false},
	{tdButA,"A",false},
	{tdButB,"B",false},
	{tdButC,"C",false},
	{tdButD,"D",false},
	{tdButE,"E",false},
	{tdButF,"F",false},
	{tdButAnd,"&",true},
	{tdButOr,"|",true},
	{tdButShl,"<<",true},
	{tdButShr,">>",true},
	{tdButPlus,"+",true},
	{tdButMinus,"-",true},
	{tdButMult,"*",true},
	{tdButDivide,"/",true},
	{0,NULL,false}
};

void ChangeForm(Int16 formID)
{
	Int16 activefrm = FrmGetActiveForm()->formId;
	
	if (activefrm==formID)
	  return;
	if (formID == tdFormBasic || formID==tdFormScientific) {
		if (activefrm != tdFormPane)
		  FrmGotoForm(tdFormPane);
		if (formID == tdFormBasic)
		  calcPrefs.lastPane = 'OpBa';
		else
		  calcPrefs.lastPane = 'OpSc';
		if (activefrm == tdFormPane)
		  pane_goto_pane(calcPrefs.lastPane,true,paneGadget);
		return;
	}
	FrmGotoForm(formID);       
}

static void 
default_form_init(int formId)
{
	FieldPtr pole;
	FormPtr frm;
	
	frm = FrmGetActiveForm();

	result_init(resultGadget);
	result_draw();
	
	pole=GetObjectPtr(tdInput);
	FldInsert(pole,calcPrefs.input,StrLen(calcPrefs.input));
	FldSetInsPtPosition(pole,calcPrefs.insertPos);
	FrmSetFocus(frm,FrmGetObjectIndex(frm,tdInput));
	FldSetSelection(pole,calcPrefs.selPosStart,
					calcPrefs.selPosEnd);	
	
	LstSetListChoices(GetObjectPtr(tdHistList),historylist,
					  MAX_HISTORY);			
}

static void default_form_update(void)
{
	FormPtr frm;
	Tbase base;
	
	frm=FrmGetActiveForm();
	switch (frm->formId) {
	 case tdFormInt:
		base=dispPrefs.base;
		CtlSetValue(GetObjectPtr(tdIntDec),base==disp_decimal);
		CtlSetValue(GetObjectPtr(tdIntOct),base==disp_octal);
		CtlSetValue(GetObjectPtr(tdIntBin),base==disp_binary);
		CtlSetValue(GetObjectPtr(tdIntHex),base==disp_hexa);
		break;
	}
	result_draw();
}

void
wait_draw(void)
{
	RectangleType rec,frame;

	frame.topLeft.x = 50;
	frame.topLeft.y = 50;
	frame.extent.x = 60;
	frame.extent.y = 20;
	
	rec.topLeft.x = 48;
	rec.topLeft.y = 48;
	rec.extent.x = 64;
	rec.extent.y = 24;
	
	WinEraseRectangle(&rec,2);
	WinDrawRectangleFrame(boldRoundFrame,&frame);
	WinDrawChars("Please wait...",14,55,55);
						  
}

void
wait_erase(void)
{
	RectangleType rec;
		
	rec.topLeft.x = 48;
	rec.topLeft.y = 48;
	rec.extent.x = 64;
	rec.extent.y = 24;
	WinEraseRectangle(&rec,2);
}

static Boolean 
MainFormHandleEvent(EventPtr event)
{
	Boolean    handled = false;
	FormPtr     frm;
	FieldPtr pole;
	ListPtr  history;
	ControlPtr knoflik;
	char *text;
	int formId;
	int controlID;
	int i;
	
	  switch (event->eType) {
	   case popSelectEvent:
		  if (event->data.popSelect.controlID==tdHistory) {
			  history=event->data.popSelect.listP;
			  pole=GetObjectPtr(tdInput);
			  text=LstGetSelectionText(history,
						       event->data.popSelect.selection);
			  if (text==NULL) /* I didn't know this might happen */
				break;
			  if (FldGetTextPtr(pole))
				FldDelete(pole,0,StrLen(FldGetTextPtr(pole)));
			  FldInsert(pole,text,StrLen(text));
			  FrmSetFocus(FrmGetActiveForm(),FrmGetObjectIndex(FrmGetActiveForm(),tdInput));
			  handled=true;
		  }		
		  break;
	   case penDownEvent:
		  handled=result_track(event->screenX,event->screenY);
		  if (!handled && calcPrefs.form == tdFormPane)
			handled = pane_handle_pendown(event->screenX,event->screenY,paneGadget);
		  break;
	   case keyDownEvent:
		  if (TxtCharIsHardKey(event->data.keyDown.modifiers,event->data.keyDown.chr)) {
			  GotHardKey();
			  handled=true;
		  }
		  else if (event->data.keyDown.chr=='\n') {
			  calc_input_exec();
			  handled=true;
		  }
		  break;
	   case frmCloseEvent:		 
		  formId=event->data.frmClose.formID;
		  
		  pole=GetObjectPtr(tdInput);
		  calcPrefs.insertPos=FldGetInsPtPosition(pole);
		  FldGetSelection(pole,&calcPrefs.selPosStart,
				  &calcPrefs.selPosEnd);
		  if (FldGetTextPtr(pole))
			StrCopy(calcPrefs.input,FldGetTextPtr(pole));
		  
		  if (formId == tdFormPane)
			pane_destroy_gadget(paneGadget);
		  
		  handled=false; /* This is important - let the
				  * FrmHandleEvent handle this
				  * itself */
		  break;
	   case frmUpdateEvent:
		  if (event->data.frmUpdate.updateCode == frmUpdateVars) {
			  default_form_update();
			  if (calcPrefs.form == tdFormPane)
				pane_draw_pane(paneGadget);
			  handled = true;
		  }
		  if (event->data.frmUpdate.updateCode == frmUpdateResult) {
			  result_draw();
			  handled = true;
		  }
		  if (event->data.frmUpdate.updateCode == frmRedrawUpdateCode) {
			  FrmDrawForm(FrmGetActiveForm());
			  result_draw();
			  if (calcPrefs.form == tdFormPane)
				pane_draw_pane(paneGadget);
			  handled = true;
		  }
		  break;
	   case frmOpenEvent:		  
		  frm=FrmGetActiveForm();
		  formId=frm->formId;
		  
		  calcPrefs.form=formId;
		  
		  FrmDrawForm(frm);
		  
		  default_form_init(formId);
		  default_form_update();
		  
		  switch (formId) {
		   case tdFormPane:
			  if (calcPrefs.lastPane == 'OpBa')
				knoflik=GetObjectPtr(tdBBasic);
			  else
				knoflik=GetObjectPtr(tdBSci);
			  break;
		   case tdFormInt:
			  switch (dispPrefs.base) {
			   case disp_decimal:
				  i=tdIntDec;
				  break;
			   case disp_octal:
				  i=tdIntOct;
				  break;
			   case disp_binary:
				  i=tdIntBin;
				  break;
			   case disp_hexa:
				  i=tdIntHex;
				  break;
			   default:
				  i=0;
				  break;
			  }
			  if (i)
			    CtlSetValue(GetObjectPtr(i),1);
			  
			  knoflik=GetObjectPtr(tdBInt);
			  break;
		   default:
			  knoflik=NULL;
			  break;
		  }
		  if (knoflik)
			CtlSetValue(knoflik,1);
		  
		  if (formId==tdFormPane) {
			  pane_init_gadget(paneGadget);
			  pane_goto_pane(calcPrefs.lastPane,true,paneGadget);
		  }
		  
		  handled = true;
		  break;
	   case ctlSelectEvent:
		  controlID=event->data.ctlSelect.controlID;

		  if (chooseForm(controlID)) {
			  handled=true;
			  break;
		  }
		  
		  handled=true;
		  switch (controlID) {
		   case calcResPopupBut: 
			  result_popup();
			  break;
		   case varFuncListButton:
		   case varBuiltinListButton:
			  if (controlID == varFuncListButton)
				text = varmgr_popup(function);
			  else
				text = varmgr_popup_builtin();
			  if (text) {
				  pole = GetObjectPtr(tdInput);
				  FldInsert(pole,text,StrLen(text));
				  FldInsert(pole,"(",1);
				  if (calcPrefs.matchParenth) {
				      if (FldInsert(pole,")",1))
						FldSetInsPtPosition(pole,FldGetInsPtPosition(pole)-1);
				  }
				  MemPtrFree(text);
			  }
			  break;
		   case varListButton:
		   case histListButton:
			  if (controlID==varListButton)
				text = varmgr_popup(variable);
			  else 
				text = history_popup();
			  if (text) {
				  pole = GetObjectPtr(tdInput);
				  FldInsert(pole,text,StrLen(text));
				  MemPtrFree(text);
			   }
			  break;
		   case tdIntBin:
		   case tdIntHex:
		   case tdIntDec:
		   case tdIntOct:
			  switch (controlID) {
			   case tdIntBin:
				  i=disp_binary;
				  break;
			   case tdIntHex:
				  i=disp_hexa;
				  break;
			   case tdIntOct:
				  i=disp_octal;
				  break;
			   case tdIntDec:
			   default:
				  i=disp_decimal;
				  break;
			  }
			  dispPrefs.base = i;
			  fp_reread_prefs();
			  ans_redisplay("ans");
			  break;
		   case tdDoit:
			  calc_input_exec();
			  break;
		   default:
			  handled=false;
		  }
		  if (handled)
		    break;
		  
		  for (i=0;tdButStr[i].id;i++) 
		    if (tdButStr[i].id==event->data.ctlSelect.controlID) {
			    FieldPtr pole;
			    UInt16 startp, endp;

			    pole=GetObjectPtr(tdInput);
			    /* If we are on the beggining and 
			     * the user pressed operator, add Ans to 
			     * the beggining */
			    FldGetSelection(pole,&startp,&endp);
			    if (tdButStr[i].operator && 
				(FldGetInsPtPosition(pole)==0 || 
				(startp!=endp && startp==0)))
			      FldInsert(pole,"ans",StrLen("ans"));
			    /* If it is kind of function and
			     * '()' matching is enabled, then
			     * add the ')' and 
			     * backspace one position
			     */
			    if (FldInsert(pole,tdButStr[i].string,
					  StrLen(tdButStr[i].string))) 
			      if (calcPrefs.matchParenth && tdButStr[i].parenth) {
				      if (FldInsert(pole,")",1))
					FldSetInsPtPosition(pole,FldGetInsPtPosition(pole)-1);
			      }
			    
			    FrmSetFocus(FrmGetActiveForm(),
					FrmGetObjectIndex(FrmGetActiveForm(),
							  tdInput));
			    
			    handled=true;
			    break;
		    }		  
		  break;
	   case menuEvent:		  
		  if (chooseForm(event->data.menu.itemID)) {
			  handled=true;
			  break;
		  }
		  switch (event->data.menu.itemID) {
		   case tdDefMgr:
			  FrmPopupForm(defForm);
			  handled = true;
			  break;
		   case tdFinCalc:
			  FrmPopupForm(finForm);
			  handled = true;
			  break;
	 	   case tdBPref:
			  FrmPopupForm(prefForm);
			  handled=true;
			  break;
		   case tdClearHistory:
			  history_shrink(0);
			  handled = true;
			  break;
		   case tdMemoExport:
			  switch (FrmAlert(memoExportQuestion)) {
			   case 0:
				  memo_dump(0);
				  break;
			   case 1:
				  memo_dump(function);
				  break;
			   case 2:
				  memo_dump(variable);
				  break;
			   default:
				  break;
			  }
			  handled = true;
			  break;
		   case tdMemoImport:
			  FrmPopupForm(memoImportForm);
			  handled = true;
			  break;
		  }
		  break;		    
	  }
	
	  return handled;
}

static Boolean 
ApplicationHandleEvent(EventPtr event)
{
	FormPtr  frm;
	Int16    formId;
	Boolean  handled = false;
	
	if (event->eType == frmLoadEvent) {
		//Load the form resource specified in the event then activate it
		formId = event->data.frmLoad.formID;
		frm = FrmInitForm(formId);
		FrmSetActiveForm(frm);
		// Set the event handler for the form.  The handler of the currently
		// active form is called by FrmDispatchEvent each time it is called
		switch (formId) {
#ifdef GRAPHS_ENABLED			
		 case grSetupForm:
			FrmSetEventHandler(frm,GraphSetupHandleEvent);
			break;
		 case grPrefForm:
			FrmSetEventHandler(frm,GraphPrefsHandleEvent);
			break;
		 case grTableForm:
			FrmSetEventHandler(frm,GraphTableHandleEvent);
			break;
		 case tdFormGraph:
			FrmSetEventHandler(frm,GraphFormHandleEvent);
			break;
#endif /* GRAPHS_ENABLED */
		 case prefForm:
			FrmSetEventHandler(frm,PreferencesHandleEvent);
			break;
		 case tdFormInt:
		 case tdFormPane:
			FrmSetEventHandler(frm, MainFormHandleEvent);
			break;
		 case defForm:
			FrmSetEventHandler(frm, DefmgrHandleEvent);
			break;
		 case finForm:
			FrmSetEventHandler(frm, FinHandleEvent);
			break;
		 case varEntryForm:
			FrmSetEventHandler(frm,VarmgrEntryHandleEvent);
			break;
		 case memoImportForm:
			FrmSetEventHandler(frm,MemoImportHandleEvent);
			break;
		}	
		handled = true;	       
	}
	
	return handled;
}

Int32 evtTimeout = evtWaitForever;

void
calc_nil_timeout(Int32 timeout)
{
	evtTimeout = timeout;
}

static void 
EventLoop(void)
{
	EventType  event;
	UInt16      error;
	
	do {
		EvtGetEvent(&event, evtTimeout);
		if (! SysHandleEvent(&event))
		  if (! MenuHandleEvent(0, &event, &error))
		    if (! ApplicationHandleEvent(&event))
		      FrmDispatchEvent(&event);
	} while (event.eType != appStopEvent);
}

UInt32
PilotMain(UInt16 launchCode, MemPtr cmdPBP, UInt16 launchFlags)
{
	Err err=0;
	
	if (launchCode == sysAppLaunchCmdNormalLaunch) { 
		if ((err = StartApplication()) == 0) {
			EventLoop();
			StopApplication();
		}
	}
	else
	  return sysErrParamErr;
	
	return err;
}
