/*
 *   $Id: grtable.c,v 1.2 2000/11/14 23:15:44 ondrap Exp $
 * 
 *   Scientific Calculator for Palms and PilmPilots.
 *   Copyright (C) 1999,2000 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
*/

#include <PalmOS.h>
#include "calcrsc.h"

#include "segment.h"

#include "calc.h"
#include "display.h"
#include "konvert.h"
#include "calcDB.h"
#include "stack.h"
#include "fp.h"
#include "varmgr.h"

#include "grtable.h"
#include "grsetup.h"
#include "grprefs.h"
#include "graph.h"


static double firstVisible;
static double rowStep;
static Int16 firstCol;

static TdispPrefs grtabDPrefs = { 5,true,true,disp_normal,disp_decimal,false,false,false};

/***********************************************************************
 *
 * FUNCTION:     grtable_tbl_contents
 *
 * DESCRIPTION:  Draw a contents of cells in the table, called as a
 *               callback function. This funtion puts in the first 
 *               row function names and in the first column
 *               parameter values.
 *
 * PARAMETERS:   See PalmOS reference
 *
 * RETURN:       None
 *      
 ***********************************************************************/
static void
grtable_tbl_contents(void *table,Int16 row,Int16 column, 
					 RectangleType *bounds) IFACE;
static void
grtable_tbl_contents(void *table,Int16 row,Int16 column, 
					 RectangleType *bounds)
{
	FontID oldfont;
	char *text;
	const char *text2;
	double val;
	RectangleType rp;
	CodeStack *stack;
	
	WinGetClip(&rp);
	WinSetClip(bounds);
	
	if (row==0) {
		oldfont = FntSetFont(boldFont);
		
		if (column == 0) {
			if (graphPrefs.functype == graph_func)
			  text = "x";
			else if (graphPrefs.functype == graph_polar)
			  text = "r";
			else 
			  text = "T";
			WinDrawChars(text,1,bounds->topLeft.x,bounds->topLeft.y);
	    }
		else {
			text2 = grsetup_fn_descr(column+firstCol-1);
			WinDrawChars(text2,StrLen(text2),bounds->topLeft.x,
						 bounds->topLeft.y);
			
		}		  
		FntSetFont(oldfont);
	}
	else if (column == 0) {
		text = display_real(firstVisible+rowStep*(row-1));
		WinDrawChars(text,StrLen(text),bounds->topLeft.x,
					 bounds->topLeft.y);
		MemPtrFree(text);
	}
	else  {
		if (graphPrefs.functype == graph_param) {
			if ((column+firstCol-1) % 2) 
			  stack = graphCurves[(column+firstCol-1)/2].stack2;
			else
			  stack = graphCurves[(column+firstCol-1)/2].stack1;
		}
		else /* func and polar */
		  stack = graphCurves[column+firstCol-1].stack1;
		if (stack){
			val = graph_get_value(stack,firstVisible+rowStep*(row-1));
			text = display_real(val);
			WinDrawChars(text,StrLen(text),bounds->topLeft.x,bounds->topLeft.y);
			MemPtrFree(text);
		}
	}
	WinSetClip(&rp);
}

/***********************************************************************
 *
 * FUNCTION:     grtable_init_tbl
 *
 * DESCRIPTION:  Initialize the table, called once from FrmOpenEvent
 *
 * PARAMETERS:   Nothing
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
static void grtable_init_tbl(void) IFACE;
static void
grtable_init_tbl(void)
{
	Int16 i;
	TablePtr table;
	UInt16 rows;
	
	table = GetObjectPtr(grTableTable);
	rows = TblGetNumberOfRows(table);
	
	for (i=0;i<rows;i++) {
		TblSetItemStyle(table,i,0,customTableItem);
		TblSetItemStyle(table,i,1,customTableItem);
		TblSetItemStyle(table,i,2,customTableItem);
	}
	TblSetColumnUsable(table,0,true);
	TblSetColumnUsable(table,1,true);
	TblSetColumnUsable(table,2,true);
	
	TblSetCustomDrawProcedure(table,0,grtable_tbl_contents);
	TblSetCustomDrawProcedure(table,1,grtable_tbl_contents);
	TblSetCustomDrawProcedure(table,2,grtable_tbl_contents);
	
	if (graphPrefs.functype == graph_func) {
		firstVisible = (graphPrefs.xmax+graphPrefs.xmin)/2;
		rowStep = (graphPrefs.xmax - graphPrefs.xmin)/140.0;
	}
	else if (graphPrefs.functype == graph_polar) {
		firstVisible = graphPrefs.fimin;
		rowStep = graphPrefs.fistep;
	}
	else {
		firstVisible = graphPrefs.tmin;
		rowStep = graphPrefs.tstep;
	}
	firstCol = 0;
}

/***********************************************************************
 *
 * FUNCTION:     grtable_scroll
 *
 * DESCRIPTION:  Scroll a table according to parameters and redraw
 *               the resulting table
 *
 * PARAMETERS:   dir - winUp or winDown - which direction to scroll
 *               count - how many lines to scroll
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
void grtable_scroll(WinDirectionType dir,Int16 count) IFACE;
void 
grtable_scroll(WinDirectionType dir,Int16 count)
{
	TablePtr table;
	
	if (dir==winDown) 
	  firstVisible+=rowStep*count;
	else
	  firstVisible-=rowStep*count;
	
	table = GetObjectPtr(grTableTable);
	TblMarkTableInvalid(table);
	TblRedrawTable(table);
}

/***********************************************************************
 *
 * FUNCTION:     GraphTableHandleEvent
 * 
 * DESCRIPTION:  Event handler for graphic graph Table view
 *
 * PARAMETERS:   event
 *
 * RETURN:       handled
 *      
 ***********************************************************************/
Boolean 
GraphTableHandleEvent(EventPtr event)
{
	Boolean    handled = false;
	FormPtr    frm=FrmGetActiveForm();
	Int16 formId;
	Int16 controlId;
	static TdispPrefs oldprefs;
	double position;

	switch (event->eType) {
	 case tblEnterEvent:
		handled = true;
		break;
	 case keyDownEvent:
		switch (event->data.keyDown.chr) {
		 case pageDownChr:
			grtable_scroll(winDown,8);
			handled = true;
			break;
		 case pageUpChr:
			grtable_scroll(winUp,8);
			handled = true;
			break;
		}
		break;
	 case ctlSelectEvent:
		controlId=event->data.ctlSelect.controlID;
		switch (controlId) {
		 case grTableDone:
			FrmReturnToForm(0);
			dispPrefs = oldprefs;
			fp_reread_prefs();
			handled = true;			
			break;
		 case grTableGoto:
			if (varmgr_get_double(&position,"Go directly to value")) {
				firstVisible = position;
				TblMarkTableInvalid(GetObjectPtr(grTableTable));
				TblRedrawTable(GetObjectPtr(grTableTable));
			}
			handled = true;
			break;
		 case grTableLeft:
			firstCol--;
			if (firstCol>=0) {
				TblMarkTableInvalid(GetObjectPtr(grTableTable));
				TblRedrawTable(GetObjectPtr(grTableTable));
			}
			else
			  firstCol++;
			handled = true;
			break;
		 case grTableRight:
			firstCol++;
			if ((firstCol<MAX_GRFUNCS*2-1 && graphPrefs.functype==graph_param)
				|| firstCol<MAX_GRFUNCS-1) {
				TblMarkTableInvalid(GetObjectPtr(grTableTable));
				TblRedrawTable(GetObjectPtr(grTableTable));
			}
			else
			  firstCol--;
			handled = true;
			break;
/*			
		 case grSetupDown:
			grsetup_scroll(winDown);
			handled = true;
			break;
		 case grSetupUp:
			grsetup_scroll(winUp);
            handled = true;
			break;
*/
		}
		break;
	 case frmUpdateEvent:
		break;
	 case frmCloseEvent:
		dispPrefs = oldprefs;
		fp_reread_prefs();
		handled = false;
		break;
	 case frmOpenEvent:
		formId=frm->formId;
		
		oldprefs = dispPrefs;
		dispPrefs = grtabDPrefs;
		fp_reread_prefs();
		
		grtable_init_tbl();

		FrmDrawForm(frm);
		handled=true;
		break;
	}
	  
	return handled;
}
