/*
 *   $Id: grtaps.c,v 1.3 2000/11/14 23:15:44 ondrap Exp $
 * 
 *   Scientific Calculator for Palms and PilmPilots.
 *   Copyright (C) 1999,2000 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
*/

#include <PalmOS.h>
#include "MathLib.h"
#include "grtaps.h"
#include "calcrsc.h"
#include "grprefs.h"
#include "graph.h"
#include "calc.h"
#include "mathem.h"
#include "fp.h"
#include "grsetup.h"

Int16 trackSelected=0;
double lastParam=0.0;
static Coord crossX=-1,crossY=-1;

/***********************************************************************
 *
 * FUNCTION:     grtaps_rectangle
 * 
 * DESCRIPTION:  Draws a rectangle for use with 'Zoom in' operation
 *
 * PARAMETERS:   corners of the rectangle
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
static void
grtaps_rectangle(Int16 x1,Int16 y1,Int16 x2,Int16 y2)
{
	if (x1<0 || x2<0 || y1<0 || y2<0)
	  return;
	
	WinInvertLine(x1,y1,x2,y1);
	WinInvertLine(x1,y2,x2,y2);
	WinInvertLine(x1,y1,x1,y2);
	WinInvertLine(x2,y1,x2,y2);
	WinInvertLine(x1,y1,x2,y2);
	WinInvertLine(x2,y1,x1,y2);
}

/***********************************************************************
 *
 * FUNCTION:     grtaps_clear_edges
 * 
 * DESCRIPTION:  Used when moving screen, clears the edges, that should 
 *               be left empty, when the contents of the screen is moved
 *
 * PARAMETERS:   bounds - bounds of draw area
 *               mx,my - coordinates, where the contents is moved
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
static void
grtaps_clear_edges(Coord mx,Coord my,RectangleType *bounds)
{
	RectangleType tmp,tmp2;
	
	tmp = *bounds;
	tmp.topLeft.x+=mx;
	tmp.topLeft.y+=my;
	WinClipRectangle(&tmp);

	tmp2 = *bounds;
	tmp2.extent.x = tmp.topLeft.x - bounds->topLeft.x;
	WinEraseRectangle(&tmp2,0);

	tmp2 = *bounds;
	tmp2.extent.y = tmp.topLeft.y - bounds->topLeft.y;
	WinEraseRectangle(&tmp2,0);
	
	tmp2 = *bounds;
	tmp2.topLeft.x = tmp.topLeft.x+tmp.extent.x;
	WinEraseRectangle(&tmp2,0);
	
	tmp2 = *bounds;
	tmp2.topLeft.y = tmp.topLeft.y+tmp.extent.y;
	WinEraseRectangle(&tmp2,0); 
}

/***********************************************************************
 *
 * FUNCTION:     grtaps_move
 * 
 * DESCRIPTION:  Started to fetch pen moves and move the drawing screen
 *               contents accordingly
 *
 * PARAMETERS:   begx,begy - the original user tap
 *               bounds - bounds of the drawing area
 *
 * RETURN:       true/false - if the user ended within the bounds
 *               and the screen should be moved
 *      
 ***********************************************************************/
static Boolean
grtaps_move(Coord begx,Coord begy,RectangleType *bounds)
{
	Boolean penDown;
	Coord x,y;
	Coord oldx=-1,oldy=-1;
	WinHandle orgwindow;
	UInt16 err;
	RectangleType orgbounds,rs;
	
	orgwindow = WinCreateOffscreenWindow(bounds->extent.x,bounds->extent.y,
										 screenFormat,&err);
	ErrFatalDisplayIf(err,"Error while creating tmp window.");
	WinCopyRectangle(NULL,orgwindow,bounds,0,0,winPaint);
	WinGetClip(&rs);
	WinSetClip(bounds);
	orgbounds = *bounds;
	orgbounds.topLeft.x = orgbounds.topLeft.y = 0;
	
	graph_setcolor(-1);
	do {
		EvtGetPen(&x,&y,&penDown);
		if ((oldx!=x || oldy!=y) && RctPtInRectangle(x,y,bounds)) {
			WinCopyRectangle(orgwindow,NULL,&orgbounds,
							 bounds->topLeft.x + (x-begx),
							 bounds->topLeft.y + (y-begy),winPaint);
			grtaps_clear_edges(x-begx,y-begy,bounds);
			oldx=x;oldy=y;
		}
	}while(penDown);
	graph_unsetcolor();
	
	if (!RctPtInRectangle(x,y,bounds)) 
	  WinCopyRectangle(orgwindow,NULL,&orgbounds,bounds->topLeft.x,
					   bounds->topLeft.y,winPaint);
	else {
		graphPrefs.xmax=graph_xscr2gr(bounds->topLeft.x+bounds->extent.x-
									  (x-begx));
		graphPrefs.xmin=graph_xscr2gr(bounds->topLeft.x-(x-begx));		
		graphPrefs.ymax=graph_yscr2gr(bounds->topLeft.y-(y-begy));
		graphPrefs.ymin=graph_yscr2gr(bounds->topLeft.y+bounds->extent.y-
									  (y-begy));
	}
	
	WinSetClip(&rs);
	WinDeleteWindow(orgwindow,false);
	
	if (!RctPtInRectangle(x,y,bounds))
	  return false;
	return true;
}

/***********************************************************************
 *
 * FUNCTION:     grtaps_zoom_in
 * 
 * DESCRIPTION:  Fetch pen moves, draw a rectangle on the screen
 *               according to pen moves and zoom in, if the user
 *               lift the pen within bounds
 *
 * PARAMETERS:   begx,begy - first tap, one of the corners
 *               bounds - bounds of drawing area
 *
 * RETURN:       true - zoom ok
 *               false - did not zoom
 *      
 ***********************************************************************/
static Boolean
grtaps_zoom_in(Coord begx,Coord begy,RectangleType *bounds)
{
	Boolean penDown;
	Coord x,y;
	Coord oldx=-1,oldy=-1;
	double tmp;
	
	do {
		EvtGetPen(&x,&y,&penDown);
		if ((oldx!=x || oldy!=y) && RctPtInRectangle(x,y,bounds)) {
			grtaps_rectangle(begx,begy,x,y);
			grtaps_rectangle(begx,begy,oldx,oldy);
			
			oldx=x;oldy=y;
		}
	}while(penDown);
	grtaps_rectangle(begx,begy,oldx,oldy); /* Remove the rectangle from screen */
	if (!RctPtInRectangle(x,y,bounds)) /* User wanted to cancel the zoom */
	  return false;
	
	if (oldx==begx || oldy==begy)
	  return false;
	
	if (begx>oldx) {
		tmp=begx;begx=oldx;oldx=tmp;
	}
	/* The Y axis is inversed on the screen */
	if (begy<oldy) {
		tmp=begy;begy=oldy;oldy=tmp;
	}       	
	/* begx - Left corner of the box, oldx - right */
	/* begy - bottom corner, oldy - top corner */
	
	graphPrefs.xmax=graph_xscr2gr(oldx);
	graphPrefs.xmin=graph_xscr2gr(begx);

	graphPrefs.ymax=graph_yscr2gr(oldy);
	graphPrefs.ymin=graph_yscr2gr(begy);

	return true;
}

/***********************************************************************
 *
 * FUNCTION:     grtaps_track_but
 * 
 * DESCRIPTION:  The handler for the 'T' button, popups a list of 
 *               functions, that can be tracked
 *
 * PARAMETERS:   Nothing
 *
 * RETURN:       true - user selected the func, tracking should begin
 *               false - user canceled operation
 *      
 ***********************************************************************/
Boolean
grtaps_track_but(void)
{
	ListPtr list;
	char *descr[MAX_GRFUNCS];
	Int16 nums[MAX_GRFUNCS];
	Int16 count,i;

	for (i=0,count=0;i<MAX_GRFUNCS;i++) {
		if (graphPrefs.functype == graph_param &&
			StrLen(grsetup_get_fname(i*2)) &&
			StrLen(grsetup_get_fname(i*2+1))) {

			descr[count] = (char *)grsetup_fn_descr(i*2);
			nums[count] = i;
			count++;
		}
		else if (graphPrefs.functype != graph_param &&
				 StrLen(grsetup_get_fname(i))) {
			descr[count] = (char *)grsetup_fn_descr(i);
			nums[count] = i;
			count++;
		}
	}
	if (count==0)
	  return false;
	list = GetObjectPtr(graphTrackList);
	LstSetListChoices(list,descr,count);
	LstSetHeight(list,count);
	LstSetSelection(list,trackSelected);
	i = LstPopupList(list);
	if (i == noListSelection)
	  return false;
	trackSelected = nums[i];
	
	return true;
}

/***********************************************************************
 *
 * FUNCTION:     grtaps_cross
 * 
 * DESCRIPTION:  Draw a tracking cross on the screen
 *
 * PARAMETERS:   x,y - coordinates of the center of cross
 *               bounds - bounds of drawing area
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
static void
grtaps_cross(Int16 x,Int16 y,RectangleType *bounds)
{
	if (x!=-1) 
	  WinInvertLine(x,bounds->topLeft.y,x,
					bounds->topLeft.y+bounds->extent.y-1);
	if (y!=-1) 
	  WinInvertLine(bounds->topLeft.x,y,
					bounds->topLeft.x+bounds->extent.x-1,
					y);	
}

/***********************************************************************
 *
 * FUNCTION:     grtaps_print_val
 * 
 * DESCRIPTION:  Print a specified 'double' on the location, clear
 *               the location before printing
 *
 * PARAMETERS:   name - text, that should precede the number
 *               value - number to be drawn
 *               x,y - coordinates
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
static void
grtaps_print_val(char *name, double value, Coord x, Coord y)
{
	char text[MAX_FP_NUMBER+10];
	RectangleType bounds;
	
	StrCopy(text,name);
	fp_print_double(text+StrLen(text),value);
	
	bounds.topLeft.x = x;
	bounds.topLeft.y = y;
	bounds.extent.x = 68;
	bounds.extent.y = 12;
	WinEraseRectangle(&bounds,0);
	WinDrawChars(text,StrLen(text),x,y);
}

/***********************************************************************
 *
 * FUNCTION:     grtaps_print_pol, grtaps_print_param, grtaps_print_coords
 * 
 * DESCRIPTION:  Prints r & angle or x & y respectively on
 *               given places of the screen
 *
 * PARAMETERS:   r,angle,x,y - values
 *               bounds - bounds of drawing area
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
static void
grtaps_print_pol(double r, double angle,RectangleType *bounds)
{
	grtaps_print_val("r:",r,bounds->topLeft.x+1,
					 bounds->topLeft.y+bounds->extent.y-12);
	
	grtaps_print_val("fi:",angle,bounds->topLeft.x+bounds->extent.x/2+2,
					 bounds->topLeft.y+bounds->extent.y-12);
}

static void
grtaps_print_param(double t,RectangleType *bounds)
{
	grtaps_print_val("T:",t,bounds->topLeft.x+1,
					 bounds->topLeft.y+bounds->extent.y-12);
}

static void
grtaps_print_coords(double realx, double realy,RectangleType *bounds)
{
	grtaps_print_val("x:",realx,bounds->topLeft.x+1,bounds->topLeft.y);
	
	grtaps_print_val("y:",realy,bounds->topLeft.x+bounds->extent.x/2+2,
					 bounds->topLeft.y);
}

/***********************************************************************
 *
 * FUNCTION:     grtaps_track_manual
 *
 * DESCRIPTION:  manually (not by pen-tap) set the tracking cross
 *
 * PARAMETERS:   value - value to which set the cross
 *               action - track_reset - the screen was redrawn,
 *                                      reset the coordinates so
 *                                      the next track will not clear
 *                                      the lastly drawn cross
 *                        track_set - set a cross to a value and
 *                                    draw it
 *                        track_add - moves a cross by value
 * 
 * RETURN:       Nothing
 *      
 ***********************************************************************/
void
grtaps_track_manual(double value,TtrackAction action)
{
	double realx,realy,r;
	RectangleType bounds;
	
	if (action==track_reset) {
		lastParam = value;
		crossX = crossY = -1;
		return;
	}
	else {
		gadget_bounds(graphGadget,&bounds);
		
		if (action==track_set)
		  lastParam=value;
		else /* track_add */
		  lastParam+=value;
		
		grtaps_cross(crossX,crossY,&bounds);
		if (graph_get_vals(trackSelected,lastParam,&realx,&realy)) {
			crossX = graph_xgr2scr(realx);
			crossY = graph_ygr2scr(realy);
			if (!RctPtInRectangle(crossX,crossY,&bounds))
			  crossX = crossY = -1;
			else
			  grtaps_cross(crossX,crossY,&bounds);
			grtaps_print_coords(realx,realy,&bounds);
			if (graphPrefs.functype == graph_param)
			  grtaps_print_param(lastParam,&bounds);
			else if (graphPrefs.functype == graph_polar) {
				r = graph_get_value(graphCurves[trackSelected].stack1,lastParam);
				grtaps_print_pol(r,lastParam,&bounds);
			}
		}
		else
		  crossX = crossY = -1;
	}
}

/***********************************************************************
 *
 * FUNCTION:     grtaps_track_pol
 * 
 * DESCRIPTION:  Tracking of a polar graph
 *
 * PARAMETERS:   x,y - first tap
 *               bounds
 *
 * RETURN:       true - user lifted pen within bounds
 *               false - outside bounds
 *      
 ***********************************************************************/
static Boolean
grtaps_track_pol(Coord x,Coord y,RectangleType *bounds)
{
	double realx,realy;
	double angle,r;	
	Coord oldx,oldy,dx,dy;
	CodeStack *stack;
	Boolean penDown;	

	stack = graphCurves[trackSelected].stack1;
	if (!stack)
	  return false;

	oldx = oldy = -1;
	realx = realy = r = angle = NaN;
	do {
		EvtGetPen(&x,&y,&penDown);
		if ((oldx!=x || oldy!=y) && RctPtInRectangle(x,y,bounds)) {
			realx = graph_xscr2gr(x);
			realy = graph_yscr2gr(y);
			angle = NaN;
			r = hypot(realx,realy);
			if (r!=0.0) {
				if (realx>=0.0 && realy>=0.0)
				  angle = asin(realy/r);
				else if (realx<=0.0)
				  angle = M_PIl - asin(realy/r);
				else
				  angle = 2*M_PIl + asin(realy/r);

				r = graph_get_value(stack,angle);
				
				dx = graph_xgr2scr(r*cos(angle));
				dy = graph_ygr2scr(r*sin(angle));
				if (!finite(angle) || !RctPtInRectangle(dx,dy,bounds)) {
					dy = -1;				
					dx = -1;
				}				  
			}
			else {
				dx = dy = -1;
				realx = realy = NaN;
			}
			grtaps_cross(dx,dy,bounds);
			grtaps_cross(crossX,crossY,bounds);
			grtaps_print_coords(realx,realy,bounds);
			grtaps_print_pol(r,angle,bounds);
			oldx = x;oldy = y;
			crossX = dx;crossY=dy;
		}		
	}while(penDown);	   	
	
	if (RctPtInRectangle(crossX,crossY,bounds)) {
		lastParam = angle;
		return true;
	}
	else
	  return false;
}

/***********************************************************************
 *
 * FUNCTION:     grtaps_track_func
 * 
 * DESCRIPTION:  Tracking of a normal function
 *
 * PARAMETERS:   x,y - first tap
 *               bounds
 *
 * RETURN:       true/false - user lifted inside/outside of bounds
 *      
 ***********************************************************************/
static Boolean
grtaps_track_func(Coord x,Coord y,RectangleType *bounds)
{
	double realx,realy;
	CodeStack *stack;
	Boolean penDown;	

	stack = graphCurves[trackSelected].stack1;
	if (!stack)
	  return false;

	realx = realy = NaN;
	do {
		EvtGetPen(&x,&y,&penDown);
		if (crossX!=x && RctPtInRectangle(x,y,bounds)) {
			realx = graph_xscr2gr(x);
			realy = graph_get_value(stack,realx);
			y = graph_ygr2scr(realy);
			if (!finite(realy) || !RctPtInRectangle(x,y,bounds))
			  y = -1;
			grtaps_cross(x,y,bounds);
			grtaps_cross(crossX,crossY,bounds);
			grtaps_print_coords(realx,realy,bounds);
			crossX = x;crossY = y;
		}
	}while(penDown);	   	
	
	if (RctPtInRectangle(x,y,bounds)) {
		lastParam = realx;
		return true;
	}
	else
	  return false;
}

/***********************************************************************
 *
 * FUNCTION:     grtaps_tap
 * 
 * DESCRIPTION:  Generic handler started when user taps on the drawing 
 *               gadget. 
 *
 * PARAMETERS:   x,y - coords, where user tapped
 *
 * RETURN:       true - handled, user tapped inside gadget
 *               false - should be handled by somone else
 *      
 ***********************************************************************/
Boolean
grtaps_tap(Int16 x,Int16 y)
{	
	RectangleType bounds;
	
	gadget_bounds(graphGadget,&bounds);
	
	if (RctPtInRectangle(x,y,&bounds)) {
		/* Is zoom-in checked? */
		if (CtlGetValue(GetObjectPtr(graphZoomIn))) {
			if (grtaps_zoom_in(x,y,&bounds)) {
				CtlSetValue(GetObjectPtr(graphZoomIn),0);
				FrmUpdateForm(tdFormGraph,frmUpdateVars);
			}			
		}	
		/* Is tracking checked? */
		else if (CtlGetValue(GetObjectPtr(graphTrack))) {
			if (graphPrefs.functype==graph_func)
			  grtaps_track_func(x,y,&bounds);
			else if (graphPrefs.functype == graph_polar)
			  grtaps_track_pol(x,y,&bounds);
		}
		else /* Move the graph */
		  if (grtaps_move(x,y,&bounds)) 
			FrmUpdateForm(tdFormGraph,frmUpdateVars);
		return true;
	}	
	return false;
}

void 
grtaps_track_start(void)
{
	FormPtr frm = FrmGetActiveForm();
	
	grtaps_track_manual(0,track_add);
	FrmShowObject(frm,FrmGetObjectIndex(frm,graphTrackGoto));
}

void
grtaps_track_pause(void)
{
	FormPtr frm = FrmGetActiveForm();
	
	CtlSetValue(GetObjectPtr(graphTrack),track_reset);
	FrmHideObject(frm,FrmGetObjectIndex(frm,graphTrackGoto));
}

void
grtaps_track_reset(double start)
{
	FormPtr frm = FrmGetActiveForm();
	
	grtaps_track_manual(start,track_reset);
	CtlSetValue(GetObjectPtr(graphTrack),false);
	FrmHideObject(frm,FrmGetObjectIndex(frm,graphTrackGoto));
}
