#!/usr/bin/python

"""
panegen.py    $Id: panegen.py,v 1.1.1.1 2000/10/27 17:17:04 ondrap Exp $
Pane generator for EasyCalc, works in a similar way as pilrc does.

Copyright 2000 by Ondrej Palkovsky <ondrap@penguin.cz>

Distributable under the GNU General Public License Version 2 or newer.
"""

import struct
import os
import sys
import string
import re

idcount = 1
panes = []
path = ""

##
BUTTONTEXT = 1
BUTTONBITMAP = 2
###
NOFRAME = 0
NORMALFRAME = 1
BOLDFRAME = 2
###
CMD_STOP = 0
CMD_INSERT = 1
CMD_EXEC_INP = 2
CMD_CLEAR_INP = 3
CMD_POPUP_PANE = 4
CMD_PRINT = 5

class PointType:
    def __init__(self,x,y):
        self.x = x
        self.y = y
    

class RectangleType:
    def __init__(self,x,y,xx,yy):
        self.topLeft = PointType(x,y)
        self.extent = PointType(xx,yy)

class Pane:
    def __init__(self,name,paneid,width,height,inlist=1):
        global idcount
        
        self.name = name
        self.paneid = paneid
        self.width = width
        self.height = height
        self.items = []
        self.id = idcount
        self.attrs = 0
        if inlist:
            self.attrs = self.attrs | 128
        idcount = idcount + 1
        
    def set_startup_offset(self,offset):
        self.startup_offset = offset
    def set_draw_offset(self,offset):
        self.draw_offset = offset
        
    def get_startup_code(self):
        return "\0\0"
    def get_draw_code(self):
        return "\0\0"
        
    def add_item(self,item):
        self.items.append(item)
        
    def gen_rcp(self):
        encitems = ""
        rcp = []
        code = ""
        data = ""

        mypane = os.path.join(path,self.paneid)+".pane"

        # Set local offsets
        self.set_startup_offset(len(code))
        code = code + self.get_startup_code()
        self.set_draw_offset(len(code))
        code = code + self.get_draw_code()

        # Gather all code & data for this pane
        for item in self.items:
            item.set_offsets(len(data),len(code))
            code = code + item.get_code()
            data = data + item.get_data()
	    encitems = encitems + item.pack()

        # Code is directly behind the data module, and MUST
        # begin on an even address

        data = data + '\0'*((len(encitems)+len(data))%2)

        headerfmt = ">16s4chhhHHHHh"
        headerlen = struct.calcsize(headerfmt)
	head = struct.pack(headerfmt,
                           self.name,
                           self.paneid[0],self.paneid[1],
                           self.paneid[2],self.paneid[3],
                           self.width,self.height,
                           self.attrs,
                           len(encitems)+headerlen, # dataoffset
                           len(encitems)+len(data)+headerlen, #codeoffset
                           self.startup_offset,self.draw_offset,
                           len(self.items))
	    
        f = open(mypane,"w")
        f.write(head)
        f.write(encitems)
        f.write(data)
	f.write(code)
        f.close()

        rcp.append('DATA "pane" ID '+str(self.id)+' "'+mypane+'"\n')
        for item in self.items:
            rcp.append(item.gen_rcp())
            
        return string.join(rcp,'')
        

class Code:
    def __init__(self,lines):
        insre = re.compile('INSERT (FUNC|OPER|NOBRACK)?\s*(FUNC|OPER|NOBRACK)?\s*(FUNC|OPER|NOBRACK)?\s*"(.*)"',re.I)        

        popupre = re.compile('POPUP\s+(STICKY)?\s*"(\w\w\w\w)"\s*AT\s*\\((\d+)\s+(\d+)\\)',re.I)
        if not lines:
            self.code = "\0\0"
            return

        printre = re.compile('PRINT\s+"(.*)[^\\\\]"\s+AT\s*\\((\d+)\s+(\d+)\\)\s*(.*)$')
        
        self.code = ""
            
        for text in lines:
            res = insre.search(text)
            respop = popupre.search(text)
            resprint = printre.search(text)
            if res:
                attr = 0
                cattrs = [ res.group(1), res.group(2), res.group(3) ]
                if "FUNC" in cattrs:
                    attr = attr | 64
                if "OPER" in cattrs:
                    attr = attr | 128
                if "NOBRACK" in cattrs:
                    attr = attr | 32
                data = struct.pack(">b%ds" % (len(res.group(4))+1),
                                   attr,res.group(4))
                # Head has 4 bytes, align data
                data = data + "\0"*(len(data) % 2)
                
                head = struct.pack(">bxh",1,len(data)+4) # 4->len(head)

                self.code = self.code + head + data
            elif respop:
                name = respop.group(2)
                if (respop.group(1)):
                    attr = 128
                else:
                    attr = 0
                data = struct.pack(">bxhcccchhbx",CMD_POPUP_PANE,
                                   14,
                                   name[0],name[1],name[2],name[3],
                                   int(respop.group(3)),int(respop.group(4)),
                                   attr)
                self.code = self.code + data
                                   
                                   
            elif re.search("CLEARINPUT",text,re.I):
                self.code = self.code + struct.pack(">bxh",CMD_CLEAR_INP,4)
            elif re.search("EXECINPUT",text,re.I):
                self.code = self.code + struct.pack(">bxh",CMD_EXEC_INP,4)
                
        self.code = self.code + "\0\0"
        
    def compile(self):
        return self.code

class Item:
    def __init__(self,bounds,code):
        global idcount
        self.bounds = bounds
        self.id = idcount
        idcount = idcount + 1
        self.code = code

    def gen_rcp(self):
        return ""
    
    def get_data(self):
        return ""
    
    def get_code(self):
	return self.code.compile()
    
    def set_offsets(self,data,code):
        self.data_offset = data
        self.code_offset = code

    def pack(self):
        return struct.pack(">hhhhhbxH",self.id,self.bounds.topLeft.x,
                           self.bounds.topLeft.y,self.bounds.extent.x,
                           self.bounds.extent.y,
                           self.type,self.code_offset)

class Button(Item):
    def __init__(self,bounds,frame,code):
        Item.__init__(self,bounds,code)
        self.frame = frame
    def pack(self):
        attr = self.frame
        return Item.pack(self)+struct.pack(">b",attr)
        

class ButtonText(Button):
    def __init__(self,bounds,text,font,frame,code):
        Button.__init__(self,bounds,frame,code)
        self.text = text
        self.type = BUTTONTEXT
        self.font = font
        
    def get_data(self):
        return self.text+"\0"
    
    def pack(self):
        return Button.pack(self)+struct.pack(">bH",self.font,
                                             self.data_offset)

class ButtonBitmap(Button):
    def __init__(self,bounds,bitmap,frame,code):
        Button.__init__(self,bounds,frame,code)
        self.bitmap = bitmap
        self.type = BUTTONBITMAP
    def pack(self):
        return Button.pack(self)+struct.pack(">xh",self.id)
    def gen_rcp(self):
        return 'BITMAP ID %d "%s"\n' % (self.id,
                                      os.path.join(path,self.bitmap))

def main(argv):
    global path
    
    if (len(argv)!=3):
        print "Usage: argv[0] <infile> <outfile>"
        sys.exit(1)    
    f = open(argv[1],"r")    
    path = argv[2]

    # Compile regexp
    panere = re.compile('PANE\s+"(\w\w\w\w)"\s+(\d+)\s+(\d+)\s+"(.*[^\\\\])"',
                        re.I)
    buttextre = re.compile('BUTTON\s+"(.*[^\\\\])"\s+AT\s+\\(\s*(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s*\\)\s*(.*)$',re.I)
    butbmpre = re.compile('BUTTON\s+BMP\s+"(.*[^\\\\])"\s+AT\s+\\(\s*(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s*\\)\s*(.*)$',re.I)

    while 1:
        line = f.readline()
        if not line:
            break
        line = string.strip(line)
        
        res = panere.match(line)
        if res:
            panes.append(Pane(name=res.group(4),paneid=res.group(1),
                              width=int(res.group(2)),height=int(res.group(3))))
            continue

        ####### Buttons, both bitmap & normal
        textres = buttextre.match(line)
        bmpres = butbmpre.match(line)

        if textres or bmpres:
            if bmpres:
                res = bmpres
            else:
                res = textres
            bounds = RectangleType(int(res.group(2)),
                                   int(res.group(3)),
                                   int(res.group(4)),
                                   int(res.group(5)))
            rest = res.group(6)
            font = 0
            frame = NORMALFRAME
            fre = re.search("FONT (\d+)",rest,re.I)
            if fre:
                font = int(fre.group(1))
            if re.search("BOLDFRAME",rest,re.I):
                frame = BOLDFRAME
            elif re.search("NOFRAME",rest,re.I):
                frame = NOFRAME
            if re.search("\$NM",rest,re.I):
                code = Code(['INSERT "%s"' % res.group(1)])
            elif re.search("\$NFM",rest,re.I):
                code = Code(['INSERT FUNC "%s"' % res.group(1)])
            elif re.search("\$NOM",rest,re.I):
                code = Code(['INSERT OPER "%s"' % res.group(1)])
            elif re.search("CODE",rest,re.I):
                clines = []
                while 1:
                    cline = f.readline()
                    if not cline:
                        break
                    if re.search("ENDCODE",cline,re.I):
                        break
                    clines.append(string.strip(cline))
                code = Code(clines)
            else:
                code = Code([])
            if textres:
                panes[-1].add_item(ButtonText(bounds,res.group(1),font=font,
                                              frame=frame,code=code))
            else:
                panes[-1].add_item(ButtonBitmap(bounds,res.group(1),
                                                frame=frame,code=code))
        

    rcp = open(argv[1]+'.rcp',"w")
    for pane in panes:
        rcp.write(pane.gen_rcp())
    rcp.close()
    f.close()
    


if __name__ == "__main__":    
    main(sys.argv)
