/*   
 *   $Id: mathem.h,v 1.1.1.1 2000/10/27 17:17:59 ondrap Exp $
 * 
 *   Scientific Calculator for Palms and PilmPilots.
 *   Copyright (C) 1999,2000 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
*/

#ifndef _INTOPS_H
#define _INTOPS_H

#include "segment.h"

/* These are helper constants to allow one function serve
 * more types of operation. They needn't be unique, but must
 * be unique inside the function, that operates this type of
 * operation.
 */

#define MATH_SIN   0
#define MATH_COS   1
#define MATH_TAN   2
#define MATH_SINH  3
#define MATH_COSH  4
#define MATH_TANH  5
/* */
#define MATH_ASIN  6
#define MATH_ACOS  7
#define MATH_ATAN  8
#define MATH_ASINH 9
#define MATH_ACOSH 10
#define MATH_ATANH 11
/* */
#define MATH_EXP   12
#define MATH_LN    13
#define MATH_LOG   14
#define MATH_SQRT  15
#define MATH_CBRT  16
#define MATH_ABS   17
#define MATH_LOG2  18
#define MATH_TRUNC 19
#define MATH_ROUND 20
#define MATH_FPART 21
#define MATH_REAL  22
/* */
#define MATH_FACT  19
#define MATH_GAMMA 20
/* Combinatorics */
#define MATH_PERM  23
#define MATH_COMB  24
/* */
#define MATH_FZERO 25
/* */
#define MATH_RTOPR  26
#define MATH_RTOPD  27
#define MATH_PTORX  28
#define MATH_PTORY  29

#ifdef SPECFUN_ENABLED

/* We'll reserve values 100 and up for other special functions */
#define MATH_BETA    100	/* Beta function */
#define MATH_IGAMMA  101	/* incomplete Gamma function */
#define MATH_IBETA   102	/* incomplete Beta function */
#define MATH_ERF     103	/* error function */
#define MATH_ERFC    104	/* complementary error function */

#define MATH_BESSELJ 110	/* Bessel functions of the first kind */
#define MATH_BESSELY 111	/* Bessel functions of the second kind */
#define MATH_BESSELI 112	/* Modified Bessel functions, first kind */
#define MATH_BESSELK 113	/* Modified Bessel functions, second kind */
#define MATH_AIRY    114	/* Airy function, Ai */
#define MATH_BIRY    115	/* Airy function, Bi */

#define MATH_ELLC1   120	/* Complete Elliptic Integral, 1st kind */
#define MATH_ELLC2   121	/* Complete Elliptic Integral, 2nd kind */
#define MATH_ELLI1   122	/* Incomplete Elliptic Integral, 1st kind */
#define MATH_ELLI2   123	/* Incomplete Elliptic Integral, 1st kind */
#define MATH_SN      124	/* Jacobian elliptic function sn */
#define MATH_CN      125	/* Jacobian elliptic function cn */
#define MATH_DN      126	/* Jacobian elliptic function dn */

#define MATH_HYPGG   130	/* Gaussian hypergeometric function F */
#define MATH_HYPGC   131	/* Confluent hypergeometric function M */

#define MATH_ALGNDRF 140	/* associated Legendre functions */
#define MATH_SPHHARM 141	/* spherical harmonics */

#endif

/* Mathematical declarations */
#ifndef M_PIl
#define M_PIl 3.14159265358979323846
#endif
#define M_PI_SQRT 1.77245385091
/* Precision for solving functions */
#define F_IS_ZERO(x)   (fabs(x)<1E-10)

CError math_oper(Int16 funcnum,CodeStack *stack) MLIB;
CError math_neg(Int16 funcnum,CodeStack *stack) MLIB;
CError math_trigo(Int16 funcnum,CodeStack *stack) MLIB;
CError math_arctrigo(Int16 funcnum,CodeStack *stack) MLIB;
CError math_math(Int16 funcnum,CodeStack *stack) MLIB;
CError math_bin(Int16 funcnum,CodeStack *stack) MLIB; 
CError math_fact(Int16 funcnum,CodeStack *stack) MLIB;
CError math_gamma(Int16 funcnum,CodeStack *stack) MLIB;
CError math_int(Int16 funcnum,CodeStack *stack) MLIB;
CError math_comb(Int16 funcnum,CodeStack *stack) MLIB;
CError math_fsolve(Int16 funcnum,CodeStack *stack) MLIB;
CError math_convcoord(Int16 funcnum,CodeStack *stack) MLIB;
#endif
