/*
 *   $Id: display.c,v 1.1.1.1 2000/10/27 17:16:32 ondrap Exp $
 * 
 *   Scientific Calculator for Palm's and PilmPilot's.
 *   Copyright (C) 2000 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
*/

#include <PalmOS.h>

#include "konvert.h"
#include "mathem.h"
#include "display.h"
#include "defuns.h"
#include "funcs.h"
#include "prefs.h"
#include "fp.h"

TdispPrefs dispPrefs;

/***********************************************************************
 *
 * FUNCTION:     display_complex
 *
 * DESCRIPTION:  Return a textualy formated complex number
 * 
 * NOTE;         Returns a dynamically allocated structure,
 *               it must be freed after use
 *
 * PARAMETERS:   number
 *
 * RETURN:       textualy formated number
 *      
 ***********************************************************************/
char *
display_complex(Complex number)
{
	char *result;
	
	result = MemPtrNew(dispPrefs.decPoints*2+45);
	result[0] = '\0';
	
	if (number.real != 0.0 || number.imag==0.0) {
		fp_print_double(result,number.real);
		if (number.imag > 0.0)
		  StrCat(result,"+");
	}
	if (number.imag != 0.0) {
		fp_print_double(result+StrLen(result),number.imag);
		StrCat(result,"i");
	}
	
	return result;
}

/***********************************************************************
 *
 * FUNCTION:     double_real
 *
 * DESCRIPTION:  Return a nicely formatted real number
 *
 * NOTE:         Result is a dynamically allocated text!
 * 
 * PARAMETERS:   number
 *
 * RETURN:       formatted text
 *      
 ***********************************************************************/
char *
display_real(double number)
{
	char *result;
		 
	result = MemPtrNew(dispPrefs.decPoints+20); /* The ipart shouldn't be more then 10 */
	fp_print_double(result,number);
	
	return result;
}

char *
display_integer(UInt32 number,Tbase mode)
{
	Int16 base;
	Int16 i,j;
	char *result,tmp;
	
	result = MemPtrNew(33); /* 32 bits + '\0' */
	
	switch (mode) {
	 case disp_binary:
		base=2;
		break;
	 case disp_octal:
		base=8;
		break;
	 case disp_hexa:
		base=16;
		break;
	 case disp_decimal:
	 default:
		base=10;
		break;
	}
        for (i=0;number;i++,number/=base)
	  if ((number % base)<10)
	    result[i]='0'+(number % base);
	else
	  result[i]='A'+(number % base) - 10;
	
	result[i]='\0';		
	/* Revert the string */
	for (j=0,i--;j<i;j++,i--) {
		tmp=result[i];
		result[i]=result[j];
		result[j]=tmp;
	}
		
	if (StrLen(result)==0)
	  StrCopy(result,"0");
	
	return result;
}

/***********************************************************************
 *
 * FUNCTION:     display_default
 *
 * DESCRIPTION:  Default displaying routine, automatically detects
 *               type of input and formats it accordingly
 * 
 * NOTE:         Result is a dynamically allocated, muste be freed after use.
 *
 * PARAMETERS:   rpn 
 *
 * RETURN:       formatted rpn suitable for output
 *      
 ***********************************************************************/
char *
display_default(Trpn rpn)
{
	char *result;
	CError err;

	if (rpn.type==variable) {
		err=rpn_eval_variable(&rpn,rpn);
		if (err) {
			result = print_error(err);			
			return result;
		}
	}
	
	switch (rpn.type) {
	 case integer:
		result = display_integer(rpn.u.intval,dispPrefs.base);
		break;
	 case complex: 
		result = display_complex(*rpn.u.cplxval);
		break;
	 case real:
		result = display_real(rpn.u.realval);
		break;
	 case string:	       
		result = MemPtrNew(StrLen(rpn.u.stringval)+1);
		StrCopy(result,rpn.u.stringval);
		break;
	 case funcempty:
		result = MemPtrNew(StrLen(rpn.u.funcname)+5);
		StrCopy(result,rpn.u.funcname);
		StrCat(result,"()");
		break;
	 default:
		result = MemPtrNew(16);
		StrCopy(result,"Unknown type.");
		break;
	}	
	return result;
}
