/*
 *   $Id: fp.c,v 1.1.1.1 2000/10/27 17:16:33 ondrap Exp $
 * 
 *   Scientific Calculator for Palms and PilmPilots.
 *   Copyright (C) 1999 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
*/


#ifdef UNIX
#define StrChr(x,y)	strchr(x,y)
#define StrCopy(x,y)	strcpy(x,y)
#define StrLen(x)	strlen(x)

#include <stdlib.h>
#include <string.h>
#else
#include <PalmOS.h>
#include <StringMgr.h>
#include <MathLib.h>
#endif
#include "defuns.h"
#include "display.h"
#include "fp.h"

static double SCI_LIMIT_MAX;   /* Above this number show it as xEy */
static double SCI_LIMIT_MIN;   /* Under this number show as xE-y */
static double DISP_MIN_LIMIT = 1E-15; /* Show numbers as sin(pi) as 1E-16,
				       it's 0 */
static double DISP_MAX_LIMIT = 1E300; /* Don't show bigger numbers, it
				       * makes problems */
static double BASE = 10.0;
static char FLPOINT_CHAR = '.';
 
/* If the number is smaller, reduce the number of 
 * decimal points, so we do not get something like
 * 9.999999987E-10
 */
static double REDUCE_PREC = 1E-6;
#define REDUCE_TO 6

void fp_setup_flpoint(void) MLIB;
void fp_setup_flpoint(void)
{
	/* Set the floating 'point' to '.' or ',' */	
	switch (PrefGetPreference(prefNumberFormat)) {
	 case nfCommaPeriod:
	 case nfApostrophePeriod:
		FLPOINT_CHAR='.';
		break;
	 case nfPeriodComma:
	 case nfSpaceComma:
	 case nfApostropheComma:
	 default:
		FLPOINT_CHAR=',';
		break;
	}
}

/***********************************************************************
 *
 * FUNCTION:     fp_reread_config
 * 
 * DESCRIPTION:  Setups some maximum/miniums according to displayPrefs
 *
 * PARAMETERS:   Nothing
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
void fp_reread_prefs(void) MLIB;
void fp_reread_prefs(void)
{
	BASE = dispPrefs.base;
	switch (dispPrefs.mode) {
	 case disp_normal:
//		SCI_LIMIT_MAX=1E10;
		SCI_LIMIT_MAX=pow(BASE,10);
//		SCI_LIMIT_MIN=1E-5;
		SCI_LIMIT_MIN=pow(BASE,-5);
		break;
	 case disp_sci:
	 case disp_eng:		
		SCI_LIMIT_MAX=1.00;
		SCI_LIMIT_MIN=1.01;
		break;	
	}
	if (dispPrefs.reducePrecision) {
		DISP_MIN_LIMIT = 1E-15;
		REDUCE_PREC = 1E-6;
	} else {
		DISP_MIN_LIMIT = 1E-100;
		REDUCE_PREC = 1E-90;
	}
}

/***********************************************************************
 *
 * FUNCTION:     cvt_fltoa
 * 
 * DESCRIPTION:  Convert a truncated float to ascii
 *
 * PARAMETERS:   value
 *
 * RETURN:       strP - converted value
 *      
 ***********************************************************************/
static void cvt_fltoa(char *strP,double value) MLIB;
static void
cvt_fltoa(char *strP,double value)
{
	char *cP;
	char tmpstr[MAX_FP_NUMBER+3];
	Int16 i,ichar;	

	if (value==0.0) {
		StrCopy(strP,"0");
		return;
	}
	/* We convert the number backwards */
	for (i=MAX_FP_NUMBER-2;i>=0;i--) {
		ichar = (Int16)fmod(value,BASE) + '0';
		tmpstr[i] = ichar>'9'?ichar-'9'-1+'A':ichar;
		value /= BASE;
		value=trunc(value);
	}
	
	tmpstr[MAX_FP_NUMBER-1] = '\0';
	
	for( cP = tmpstr; *cP && (*cP == '0'); )
	  ++cP;
	
	StrCopy(strP,cP);
}

/***********************************************************************
 *
 * FUNCTION:     fp_print
 * 
 * DESCRIPTION:  Prints a double to string
 *
 * PARAMETERS:   value 
 *
 * RETURN:       strP - converted value
 *      
 ***********************************************************************/
static void
fp_print(char *strP,double value) MLIB;
static void
fp_print(char *strP,double value)
{
	Int16 i,myfrac;
	double ipart;
	double fpart;
	double limit;
	char str[MAX_FP_NUMBER*2];
	
	if (isinf(value)) {
		if (isinf(value)>0) 
		  StrCopy(strP,"+Inf");
		else
		  StrCopy(strP,"-Inf");
		return;
	}
	if (!finite(value)) {
		StrCopy(strP,"NaN");
		return;
	}

    if( value < 0.0 ) {
		*strP++ = '-';
		value = -value;
	}
	myfrac = dispPrefs.decPoints;
/*	
	if (value<REDUCE_PREC)
	  myfrac = REDUCE_TO;
	else
	  myfrac = dispPrefs.decPoints;
*/
    ipart = trunc(value);
    value -= ipart;
	/* recover frac part */
    for (limit = 1.0,i=myfrac; i> 0;i--) {
		value *= BASE;
		limit *= BASE;
	}
	
    value += 0.5;
    if( value >= limit ) {
		fpart = 0.0;				/* overflowed */
		ipart++;
    }
    else
	  fpart = value;

    cvt_fltoa(strP,ipart);
   
    if (myfrac) {
		cvt_fltoa(str,fpart);
		strP += StrLen(strP);
		
		if (dispPrefs.stripZeros) {
			/* remove trailing 0's */
			for (i = StrLen(str)-1;i>=0 && str[i]=='0';i--) {
				str[i]='\0';
				myfrac--;
			}	      
		}      
		if (StrLen(str)>0) {
			*strP++ = FLPOINT_CHAR;
			for(i=StrLen(str); i<myfrac;i++ )
			  *strP++ = '0';			/* need some padding at the beginning */
			StrCopy(strP,str);			/* and now the value */
		}
    }
}

static const char expNames[] = {'f','p','n',181,'m',0,'K','M','G','T'};

static void fp_print_exp(char *str,Int16 expon) MLIB;
static void
fp_print_exp(char *str,Int16 expon)
{
	Int16 i;
	
	if (dispPrefs.base == disp_decimal) {
		if (dispPrefs.mode == disp_eng && dispPrefs.cvtUnits) {
			i = expon/3 + 5;
			if (i<0 || i>9)
			  StrPrintF(str,"E%d",expon);
			else 
			  StrPrintF(str,"%c",expNames[i]);
		}
		else
		  StrPrintF(str,"E%d",expon);	
	}
	else 
	  StrPrintF(str,"*%d^%d",(Int16)dispPrefs.base,expon);
}

/***********************************************************************
 *
 * FUNCTION:     fp_print_double
 * 
 * DESCRIPTION:  Function used for general number formatting according
 *               all rules from dispPrefs, adds the 'E+-' things
 *               for too big numbers, scientific & engineer purposes
 *
 * PARAMETERS:   value
 *
 * RETURN:       strP - converted value
 *      
 ***********************************************************************/
void fp_print_double(char *strP,double value) MLIB;
void 
fp_print_double(char *strP, double value)
{
	Int16 ex;
	char tmpstr[32];
	     
	
	if (!finite(value)) {
		/* Handle the NaN, +Inf & -Inf cases */
		fp_print(strP,value);
	}
	else if (fabs(value)>DISP_MAX_LIMIT) {
		StrCopy(strP,"It's too big");
	}
	else if (fabs(value)<DISP_MIN_LIMIT) { 
		/* Don't show too small numbers */
		fp_print(strP,0);
		if (dispPrefs.mode!=disp_normal)
		  fp_print_exp(strP+StrLen(strP),0);
	}
	else if (fabs(value)>=1 && dispPrefs.mode==disp_eng) {
		/* Engineer mode selected */
		for (ex=0;fabs(value)>=pow(BASE,3) || (ex % 3);ex++,value/=BASE)
		  ;
		fp_print(strP,value);
		fp_print_exp(tmpstr,ex);
		StrCat(strP,tmpstr);
	}
	else if (fabs(value)<1.0 && dispPrefs.mode==disp_eng) {
		/* Engineer mode selected */
		for (ex=0;fabs(value)<0.99999999 || (ex % 3);ex++,value*=BASE)
		  ;
		fp_print(strP,value);

		fp_print_exp(tmpstr,-ex);
		StrCat(strP,tmpstr);
	}
	else if (fabs(value)>SCI_LIMIT_MAX) {
		/* Normal/Scientific mode */
		for (ex=0;fabs(value)>=BASE;value/=BASE,ex++)
		  ;
		fp_print(strP,value);
		fp_print_exp(tmpstr,ex);
		StrCat(strP,tmpstr);
	}
	else if(fabs(value)<SCI_LIMIT_MIN && fabs(value)>0) {
		/* Normal/Scientific mode */
		for (ex=0;fabs(value)<0.99999999;value*=BASE,ex--)
		  ;
		fp_print(strP,value);
		fp_print_exp(tmpstr,ex);
		StrCat(strP,tmpstr);
	}
	else {
		/* No special formatting required */
		fp_print(strP,value);
	}       
}
