/*
 *   $Id: pane.c,v 1.3 2000/11/19 18:39:28 ondrap Exp $
 * 
 *   Scientific Calculator for Palms and PilmPilots.
 *   Copyright (C) 2000 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
*/

#include <PalmOS.h>

#include "pane.h"
#include "panectrl.h"
#include "calc.h"

#define xor(x,y) (((x) && !(y)) || (!(x) && (y)))

static struct {
	char name[MAX_PANE_NAME];
	DmOpenRef dbref;
	Int32 paneid;
	UInt16 index;
	DmResID resId;
}paneList[MAX_PANES];

static Int16 paneCount = 0;

typedef struct {
	struct {
		Int16 index;
		Coord x,y;
		Boolean sticky;		
		Boolean popup;
		WinHandle save;
	}panes[MAX_ACTIVE_PANES];
	Int16 active; /* index to paneList array */
}TpaneData;

static void *
gadget_data(Int16 gadget) IFACE;
static void *
gadget_data(Int16 gadget)
{
	FormPtr frm = FrmGetActiveForm();
	UInt16 index = FrmGetObjectIndex(frm,gadget);
	
	return FrmGetGadgetData(frm,index);
}


/***********************************************************************
 *
 * FUNCTION:     pane_import_panes
 *
 * DESCRIPTION:  Import panes (resources with type='pane' 
 *               from passed resource database, add them
 *               to paneList structure
 *
 * PARAMETERS:   dbref - pointer to database from which to import
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
static void
pane_import_panes(DmOpenRef dbref) IFACE;
static void
pane_import_panes(DmOpenRef dbref)
{
	UInt16 i;
	UInt16 index;
	MemHandle handle;
	Tpane *pane;
	
	i=0;
	while ((index=DmFindResourceType(dbref,PANE_TYPE,i++))!=-1) {
		handle = DmGetResourceIndex(dbref,index);
		pane = MemHandleLock(handle);
		
		StrCopy(paneList[paneCount].name,pane->name);
		paneList[paneCount].paneid = pane->paneid;
		/* Recover the db ID of a resource */
		paneList[paneCount].index = index;
		paneList[paneCount].dbref = dbref;
		DmResourceInfo(dbref,index,NULL,&paneList[paneCount].resId,
					   NULL);
		paneCount++;

		MemHandleUnlock(handle);
		DmReleaseResource(handle);
	}
}

/***********************************************************************
 *
 * FUNCTION:     pane_find_pane
 *
 * DESCRIPTION:  Find a pane with the given 4-byte ID
 *
 * PARAMETERS:   paneid - id of pane to find
 *
 * RETURN:       index to paneList structure or '-1'
 *      
 ***********************************************************************/
Int16
pane_find_pane(Int32 paneid) IFACE;
Int16
pane_find_pane(Int32 paneid)
{
	Int16 i;
	
	for (i=0;i<paneCount;i++)
	  if (paneList[i].paneid == paneid)
		return i;
	return -1;
}

/***********************************************************************
 *
 * FUNCTION:     pane_draw_button
 *
 * DESCRIPTION:  Draw a button (text or bitmap) on a pane, 
 *               should look like standard PalmOS buttons
 *
 * PARAMETERS:   item - pointer to an item structure on a pane
 *               data - pointer to additional data (where the
 *                      dynamic length fields are stored)
 *               bounds - bounds of a pane gadget, where the button
 *                        resides
 *               dbref - database pointer, from which to get additional
 *                       data (bitmaps)
 * 
 * RETURN:       Nothing
 *      
 ***********************************************************************/
static void
pane_draw_button(Tpane_item *item,char *data,
				 RectangleType *bounds,
				 DmOpenRef dbref) IFACE;
static void
pane_draw_button(Tpane_item *item,char *data,
				 RectangleType *bounds,
				 DmOpenRef dbref)
{
	RectangleType rs,itembounds;
	Int16 chrwidth;
	char *text;
	FontID oldfont;
	
	WinGetClip(&rs);

	itembounds = item->bounds;
	itembounds.topLeft.x+=bounds->topLeft.x;
	itembounds.topLeft.y+=bounds->topLeft.y;
	
	if (colorDisplay) {
		WinPushDrawState();
		WinSetBackColor(UIColorGetTableEntryIndex(UIObjectFill));
		WinSetForeColor(UIColorGetTableEntryIndex(UIObjectFrame));
		WinEraseRectangle(&itembounds,2);
	}
	if (item->d.buttonText.attr & NORMALFRAME)
	  WinDrawRectangleFrame(roundFrame,&itembounds);
	else if (item->d.buttonText.attr & BOLDFRAME)
	  WinDrawRectangleFrame(boldRoundFrame,&itembounds);
	
	/* Draw everything with default font, but try to center
	 * it a little */
	text = data + item->d.buttonText.textOffset;

	WinSetClip(&itembounds); /* Draw only IN the button */
	
	if (item->type == button_text) {
		oldfont = FntSetFont(item->d.buttonText.font);
		chrwidth = FntCharsWidth(text,StrLen(text));
		if (chrwidth>itembounds.extent.x)
		  chrwidth = itembounds.extent.x;
		
		WinDrawChars(text,StrLen(text),
					 itembounds.topLeft.x+itembounds.extent.x/2-chrwidth/2,
					 itembounds.topLeft.y+(itembounds.extent.y-FntCharHeight())/2);
		
		FntSetFont(oldfont);
	} 
	else { /* Bitmap */
		MemHandle bmpRes;
		BitmapType *bmpPtr;
		UInt16 index;
		
		index = DmFindResource(dbref,bitmapRsc,item->d.buttonBitmap.bitmap,
							   NULL);
		bmpRes = DmGetResourceIndex(dbref,index);
		bmpPtr = MemHandleLock(bmpRes);
		WinDrawBitmap(bmpPtr,itembounds.topLeft.x+1,itembounds.topLeft.y+1);
		MemHandleUnlock(bmpRes);
		DmReleaseResource(bmpRes);
	}
	WinSetClip(&rs);
	
	if (colorDisplay)
	  WinPopDrawState();
}

/***********************************************************************
 *
 * FUNCTION:     pane_draw_pane
 *
 * DESCRIPTION:  Draw a pane and all items on this pane
 *
 * PARAMETERS:   gadget - id of a pane gadget
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
void
pane_draw_pane(Int16 gadget) IFACE;
void
pane_draw_pane(Int16 gadget)
{
	Int16 i;
	Tpane *pane;
	TpaneData *paneData = gadget_data(gadget);
	Int16 active = paneData->panes[paneData->active].index;
	char *data;
	MemHandle handle;
	RectangleType bounds;
	
	gadget_bounds(gadget,&bounds);
	/* Open pane */
	handle = DmGetResourceIndex(paneList[active].dbref,paneList[active].index);
	pane = MemHandleLock(handle);
	/* Data is stored behind the pane */
	data = ((void *) pane) + pane->dataOffset;
	
	/* Set the title of a pane form */
	if (! paneData->panes[paneData->active].popup)
	  FrmCopyTitle(FrmGetActiveForm(),pane->name);
	
	/* Set bounds to real bounds of a pane */
	gadget_bounds(gadget,&bounds);
	bounds.topLeft.x = paneData->panes[paneData->active].x+bounds.topLeft.x;
	bounds.topLeft.y = paneData->panes[paneData->active].y+bounds.topLeft.y;
	bounds.extent.x = pane->width;
	bounds.extent.y = pane->height;
	
	WinEraseRectangle(&bounds,0);
	if (paneData->panes[paneData->active].popup)
	  WinDrawRectangleFrame(rectangleFrame,&bounds);
	
	for (i=0;i < pane->itemCount;i++) {
		if (pane->items[i].type == button_text ||
			pane->items[i].type == button_bitmap) 
		  pane_draw_button(&pane->items[i],data,
						   &bounds,paneList[active].dbref);
	} 
	MemHandleUnlock(handle);
	DmReleaseResource(handle);
}


/***********************************************************************
 *
 * FUNCTION:     pane_goto_pane
 *
 * DESCRIPTION:  Closes the current pane and changes to requested pane
 *
 * PARAMETERS:   paneid - 4-byte id of a new pane
 *               base - if true, the new pane should be the base pane,
 *                      close all popups first
 *               gadget - id of a pane gadget
 *
 * RETURN:       true - pane changed
 *               false - pane doesn't exist
 *      
 ***********************************************************************/
Boolean
pane_goto_pane(Int32 paneid,Boolean base,Int16 gadget) IFACE;
Boolean
pane_goto_pane(Int32 paneid,Boolean base,Int16 gadget)
{
	TpaneData *paneData = gadget_data(gadget);
	Int16 index;
	
	index = pane_find_pane(paneid);
	if (index==-1)
	  return false;
	
	if (paneData->active == -1)
	  paneData->active++;
	
	if (base)
	  paneData->active = 0;
	
	paneData->panes[paneData->active].index = index;
	paneData->panes[paneData->active].x = 0;
	paneData->panes[paneData->active].y = 0;
	paneData->panes[paneData->active].popup = false;
	paneData->panes[paneData->active].sticky =false;
	paneData->panes[paneData->active].save = NULL;
	pane_draw_pane(gadget);

	return true;
}

static void
pane_save(TpaneData *paneData,Tpane *pane,Int16 gadget) IFACE;
static void
pane_save(TpaneData *paneData,Tpane *pane,Int16 gadget)
{
	RectangleType bounds;
	Int16 err;
	
	gadget_bounds(gadget,&bounds);
	bounds.topLeft.x+=paneData->panes[paneData->active].x-1;
	bounds.topLeft.y+=paneData->panes[paneData->active].y-1;
	
	bounds.extent.x=pane->width+2;
	bounds.extent.y=pane->height+2;

	paneData->panes[paneData->active].save = WinSaveBits(&bounds,&err);
	if (err)
	  paneData->panes[paneData->active].save = NULL;
}

static Boolean
pane_restore(Coord x,Coord y,TpaneData *pane,Int16 gadget) IFACE;
static Boolean
pane_restore(Coord x,Coord y,TpaneData *pane,Int16 gadget)
{
	if (pane->panes[pane->active].save) {
		WinRestoreBits(pane->panes[pane->active].save,x-1,y-1);
		pane->panes[pane->active].save = NULL;
		return true;
	}
	return false;
}

/***********************************************************************
 *
 * FUNCTION:     pane_hide_pane
 *
 * DESCRIPTION:  Hides a usually popup pane, tries to restore
 *               the saved screen of the bottom pane, if it doesn't
 *               work calls pane_draw_pane
 *
 * PARAMETERS:   gadget - id of a pane gadget
 *               bounds - bounds of an old pane
 *
 * RETURN:       
 *      
 ***********************************************************************/
static void
pane_hide_pane(Int16 gadget,RectangleType *bounds) IFACE;
static void
pane_hide_pane(Int16 gadget,RectangleType *bounds)
{
	TpaneData *paneData = gadget_data(gadget);
	Boolean restored;
	
	if (paneData->active>0) {		
		restored = pane_restore(bounds->topLeft.x,bounds->topLeft.y,
								paneData,gadget);		
		paneData->active--;
		if (!restored) {
			WinEraseRectangle(bounds,0);
			WinEraseRectangleFrame(rectangleFrame,bounds);
			pane_draw_pane(gadget);
		}
	}
}

/***********************************************************************
 *
 * FUNCTION:     pane_popup_pane
 *
 * DESCRIPTION:  Popups a new pane above an existing one
 *
 * PARAMETERS:   paneid - id of a new pane
 *               x,y - coordinates where the topLeft of a new pane
 *                     should be
 *               sticky - false - new pane should close on a button press
 *                                on it
 *                        true - pane remains active until somone taps
 *                               outside of its bounds
 *               gadget - id of a pane gadget
 *
 * RETURN:       true - pane exists & displayed
 *               false - pane doesn't exist
 *      
 ***********************************************************************/
Boolean
pane_popup_pane(Int32 paneid,Coord x,Coord y,
				Boolean sticky,Int16 gadget) IFACE;
Boolean
pane_popup_pane(Int32 paneid,Coord x,Coord y,
				Boolean sticky,Int16 gadget)
{
	TpaneData *paneData = gadget_data(gadget);
	Int16 index;
	Int16 active;
	Tpane *pane;
	MemHandle handle;
	
	index = pane_find_pane(paneid);
	if (index==-1)
	  return false;

	paneData->active++;
	paneData->panes[paneData->active].index = index;
	paneData->panes[paneData->active].x = x;
	paneData->panes[paneData->active].y = y;
	paneData->panes[paneData->active].popup = true;
	paneData->panes[paneData->active].sticky =sticky;
	
	active = paneData->panes[paneData->active].index;
	handle = DmGetResourceIndex(paneList[active].dbref,paneList[active].index);
	pane = MemHandleLock(handle);		
	/* save behind */
	pane_save(paneData,pane,gadget);
	MemHandleUnlock(handle);
	DmReleaseResource(handle);
	
	pane_draw_pane(gadget);
	
	return true;
}

/***********************************************************************
 *
 * FUNCTION:     pane_destroy_gadget
 *
 * DESCRIPTION:  Destroy data structures allocated to the pane, should
 *               be called for every active gadget on frmCloseEvent
 *
 * PARAMETERS:   gadget - id of a pane gadget
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
void
pane_destroy_gadget(Int16 gadget) IFACE;
void
pane_destroy_gadget(Int16 gadget)
{
	TpaneData *paneData = gadget_data(gadget);
	int i;
	
	/* Free the saved backgrounds */
	for (i=paneData->active;i>0;i--)
	  if (paneData->panes[paneData->active].save)
		WinDeleteWindow(paneData->panes[paneData->active].save,false);
	
	MemPtrFree(paneData);	
}

/***********************************************************************
 *
 * FUNCTION:     pane_init_gadget
 *
 * DESCRIPTION:  Allocates base data structures corresponding to a
 *               pane, should be called for each gadget on screen 
 *               in frmOpenEvent
 *
 * PARAMETERS:   gadget - id of a pane gadget
 *
 * RETURN:       Nothing
 *      
 ***********************************************************************/
void
pane_init_gadget(Int16 gadget) IFACE;
void
pane_init_gadget(Int16 gadget)
{
	FormPtr frm = FrmGetActiveForm();
	UInt16 index = FrmGetObjectIndex(frm,gadget);
	TpaneData *paneData;

	paneData = MemPtrNew(sizeof(*paneData));
	paneData->active = -1;
	
	FrmSetGadgetData(frm,index,paneData);	
}

void
pane_init(void) IFACE;
void
pane_init(void)
{
	pane_import_panes(DmNextOpenResDatabase(NULL));
}

void
pane_close(void) IFACE;
void
pane_close(void)
{
	/*  */
}

/***********************************************************************
 *
 * FUNCTION:     pane_track_button
 *
 * DESCRIPTION:  This function gets called when a user taps a button on
 *               pane, emulate the default behaviour, if the user
 *               lifts the pen on a button execute an action
 *               associated with the button
 *
 * PARAMETERS:   frame - bounds of a button
 *               frmattr - attributes of frame (no,normal,bold)
 *
 * RETURN:       true - button pressed
 *               false - button NOT pressed
 *      
 ***********************************************************************/
Boolean
pane_track_button(RectangleType *frame,Int8 frmattr) IFACE;
Boolean
pane_track_button(RectangleType *frame,Int8 frmattr)
{
	RectangleType bounds;
	Coord oldx = frame->topLeft.x;
	Coord oldy = frame->topLeft.y;
	Coord x,y;
	Boolean penDown;
	
	bounds = *frame;
	bounds.topLeft.x-=1;
	bounds.topLeft.y-=1;
	bounds.extent.x+=2;
	bounds.extent.y+=2;

	#define INVERT_BTN { WinInvertRectangle(frame,1);\
                     if (frmattr & NORMALFRAME) \
                         WinDrawRectangleFrame(roundFrame,frame);\
                     else if (frmattr & BOLDFRAME)\
	                     WinDrawRectangleFrame(boldRoundFrame,frame); \
    }
	
	INVERT_BTN;
	
	do {
		EvtGetPen(&x,&y,&penDown);
		if (xor(RctPtInRectangle(x,y,&bounds),
				RctPtInRectangle(oldx,oldy,&bounds)))
		  INVERT_BTN;
		oldx = x;
		oldy = y;
	}while(penDown);
	if (RctPtInRectangle(oldx,oldy,&bounds)) {
		INVERT_BTN;
		return true;
	}
	return false;
}

/***********************************************************************
 *
 * FUNCTION:     pane_handle_pendown
 *
 * DESCRIPTION:  Handle a tap and subsequent move of a pen, execute
 *               requested actions etc.
 *
 * PARAMETERS:   x,y - coordinates of a penDown event
 *               gadget - id of a pane gadget
 *
 * RETURN:       true - tap handled
 *               false - tap not handled, should be handled by other routines
 *      
 ***********************************************************************/
Boolean
pane_handle_pendown(Coord x,Coord y,Int16 gadget) IFACE;
Boolean
pane_handle_pendown(Coord x,Coord y,Int16 gadget)
{
	RectangleType bounds,realbounds;
	TpaneData *paneData = gadget_data(gadget);
	Int16 active = paneData->panes[paneData->active].index;
	Tpane *pane;
	MemHandle handle;
	Tpane_item *item;
	Int16 i;
	Boolean pressed=false;
	Coord realtopx, realtopy;
	
	gadget_bounds(gadget,&bounds);
	if (!RctPtInRectangle(x,y,&bounds)) 
	  return false;

	handle = DmGetResourceIndex(paneList[active].dbref,paneList[active].index);
	pane = MemHandleLock(handle);		

	/* Check if the user clicked outside the pane */
	realtopx = paneData->panes[paneData->active].x + bounds.topLeft.x;
	realtopy = paneData->panes[paneData->active].y + bounds.topLeft.y;
	bounds.topLeft.x = realtopx;
	bounds.topLeft.y = realtopy;
	bounds.extent.x = pane->width;
	bounds.extent.y = pane->height;	
	if (paneData->panes[paneData->active].popup &&
		!RctPtInRectangle(x,y,&bounds)) {
		/* Discard the topmost form */
		MemHandleUnlock(handle);
		DmReleaseResource(handle);
		
		SndPlaySystemSound(sndWarning);
		pane_hide_pane(gadget,&bounds);
		return true;
	}
	
	for (i=0;i < pane->itemCount;i++) {
		if (RctPtInRectangle(x-realtopx,
							 y-realtopy,&pane->items[i].bounds)) {
			item = &pane->items[i];
			realbounds = item->bounds;
			realbounds.topLeft.x+=realtopx;
			realbounds.topLeft.y+=realtopy;
			if ((item->type == button_text || item->type == button_bitmap))
			  if (pane_track_button(&realbounds,item->d.buttonText.attr)) {
				  pressed = true;
				  break;
			  }
		}
	}
	if (pressed) {
		/* Find & execute the code bound to the given object */
		Tcommand *command;
		
		/* Hide pane if it is not sticky */
		/* Do it before executing the button */
		if (!paneData->panes[paneData->active].sticky &&
			paneData->panes[paneData->active].popup)
		  pane_hide_pane(gadget,&bounds);
		
		/* make gcc happy */
		item = &pane->items[i];
		/* Execute the code bound to this object */
		command = ((void *) pane) + pane->codeOffset + item->codeOffset;
		
		SndPlaySystemSound(sndClick);
		pnctl_exec(command);
	}
	MemHandleUnlock(handle);
	DmReleaseResource(handle);
	return true;
}
