/*
 *   $Id: panectrl.c,v 1.2 2000/10/29 22:35:43 ondrap Exp $
 * 
 *   Scientific Calculator for Palms and PilmPilots.
 *   Copyright (C) 2000 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
*/

#include <PalmOS.h>

#include "panectrl.h"
#include "pane.h"
#include "calc.h"
#include "calcrsc.h"
#include "prefs.h"
#include "konvert.h"
#include "stack.h"
#include "calcDB.h"

static void
pnctl_insert(char *text,Boolean func,Boolean oper,Boolean nobrack)
{
	FieldPtr field = GetObjectPtr(tdInput);
	UInt16 startp,endp;
	
	FldGetSelection(field,&startp,&endp);
	
	if (oper && !func) {
		/* Insert 'ans' if on the beginning */
		if (FldGetInsPtPosition(field)==0 ||
			(startp!=endp && startp==0))
		  FldInsert(field,"ans",3);
		FldInsert(field,text,StrLen(text));
	}
	else if (func) {
		/* if selected, insert in front of selection and
		 * put brackets around selection */
		if (startp!=endp) {
			FldSetSelection(field,0,0);
			FldSetInsPtPosition(field,startp);
		}
		if (FldGetInsPtPosition(field)==0 && oper) {
			FldInsert(field,"ans",3);
			if (startp!=endp)
			  endp+=3;
		}
		FldInsert(field,text,StrLen(text));
		if (!nobrack)
		  FldInsert(field,"(",1);
		if (startp!=endp) {
			startp+=StrLen(text)+1;
			endp+=StrLen(text)+1;
			FldSetInsPtPosition(field,endp);
			FldInsert(field,")",1);
			FldSetSelection(field,startp,endp);
		}
		else if (calcPrefs.matchParenth) { /* nothing selected */
			FldInsert(field,")",1);
			FldSetInsPtPosition(field,FldGetInsPtPosition(field)-1);
		}
	}
	else 
	  FldInsert(field,text,StrLen(text));
}

static void
pcntl_clear(void)
{
	FieldPtr field;
	
	field = GetObjectPtr(tdInput);
	FldDelete(field,0,FldGetTextLength(field));
}

static char *
pcntl_exec_textf(Tcommand *command)
{
	char *result = NULL;
	CodeStack *stack;
	CError err;
	Trpn item;
	
	switch (command->type) {
	 case cmd_text:
		result = MemPtrNew(StrLen(command->d.data.text)+1);
		StrCopy(result,command->d.data.text);
		break;
	 case cmd_expr:
		stack = text_to_stack(command->d.data.text,&err);
		if (!err) {
			err = stack_compute(stack);
			if (!err) {
				item = stack_pop(stack);
				result = display_default(item);
				rpn_delete(item);
			}
			stack_delete(stack);
		}
		break;
	 case cmd_func:
		err = db_func_description(command->d.data.text,&result,NULL);
		if (err)
		  result = NULL;
		break;		
	}
	return result;
}

static void
pane_print(Tcommand *command,Coord x,Coord y,FontID font)
{
}

static Boolean
pane_cmd_expr(char *text)
{
	CodeStack *stack;
	CError err;
	
	stack = text_to_stack(text,&err);
	if (!err) {
		err = stack_compute(stack);
		stack_delete(stack);
	}
	if (err)
	  return false;
	return true;
}

static Boolean
pcntl_exec_boolean(Tcommand *command)
{
	Boolean result = false;
	
	switch (command->type) {
	 case cmd_insert:
		pnctl_insert(command->d.insert.text,
					 command->d.insert.func,
					 command->d.insert.oper,
					 command->d.insert.nobrack);
		result = true;
		break;
	 case cmd_clear_inp:
		pcntl_clear();
		result = true;
		break;
	 case cmd_exec_inp:
		calc_input_exec();
		result = true;
		break;
	 case cmd_popup_pane:
		result = pane_popup_pane(command->d.popup.paneid,
								 command->d.popup.x,
								 command->d.popup.y,
								 command->d.popup.sticky,
								 paneGadget);
	 case cmd_print:
		pane_print((Tcommand *)&command->d.print.expr[0],
				   command->d.print.x,command->d.print.y,
				   command->d.print.font);
		result = true;
		break;
	 case cmd_expr:
		result = pane_cmd_expr(command->d.data.text);
		break;
	 default:
		break;
	}
	return result;
}

void
pnctl_exec(Tcommand *command)
{
	while (command->type != cmd_stop) {
		if (!pcntl_exec_boolean(command))
		  break;
		command = ((void *)command) + command->len;
	}
}
