 /*   
 * $Id: prefs.c,v 1.1.1.1 2000/10/27 17:15:59 ondrap Exp $
 * 
 *   Scientific Calculator for Palms and PilmPilots.
 *   Copyright (C) 1999,2000 Ondrej Palkovsky
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 * 
 *  You can contact me at 'ondrap@penguin.cz'.
*/

#include <PalmOS.h>

#include "calcrsc.h"
#include "calc.h"
#include "prefs.h"
#include "calcDB.h"
#include "fp.h"

tPrefs calcPrefs;

void
prefs_read_preferences()
{
	int i;
	UInt16 size;
	
	size=sizeof(calcPrefs);
	if (PrefGetAppPreferences(APP_ID,PREF_DEFAULT,&calcPrefs,&size,
				  true)!=PREF_VERSION) {
		calcPrefs.input[0]='\0';
		calcPrefs.form=tdFormPane;
		calcPrefs.lastPane = 'OpBa';
		calcPrefs.insertPos=0;
		calcPrefs.selPosStart=calcPrefs.selPosEnd=0;
		calcPrefs.finBegin=false;
		calcPrefs.matchParenth=false;
		for (i=0;i<MAX_HISTORY;i++) 
		  calcPrefs.history[i][0]='\0';
		calcPrefs.trigo_mode=radian;
		
		calcPrefs.dispPrefs.forceComplex=false;
		calcPrefs.dispPrefs.forceInteger = false;		
		calcPrefs.dispPrefs.decPoints = 9;
		calcPrefs.dispPrefs.stripZeros = true;
		calcPrefs.dispPrefs.reducePrecision = true;
		calcPrefs.dispPrefs.base = disp_decimal;
		calcPrefs.dispPrefs.mode = disp_normal;
		calcPrefs.dispPrefs.cvtUnits = true;
		
		FrmAlert(tdAlertAbout);
		FrmHelp(tdHelp2);
		dispPrefs = calcPrefs.dispPrefs; /* db_recompile needs dispPrefs */
		db_recompile_all();
	}
	dispPrefs = calcPrefs.dispPrefs;
	fp_reread_prefs();
}
	
void 
prefs_save_preferences()
{
	calcPrefs.dispPrefs = dispPrefs;
	PrefSetAppPreferences(APP_ID,PREF_DEFAULT,PREF_VERSION,&calcPrefs,
			      sizeof(calcPrefs),true);
}

Boolean
PreferencesHandleEvent(EventPtr event)
{
	Boolean handled=false;
	int controlID;
	FormPtr frm;
	int formId;
	ControlPtr knoflik;
	ListPtr list;
	
	switch (event->eType) {
	 case frmOpenEvent:
		frm=FrmGetActiveForm();		
		formId=frm->formId;
		FrmDrawForm(frm);
		
		switch (dispPrefs.base) {
		 default:
		 case disp_decimal:
			knoflik=GetObjectPtr(prefDecimal);
			break;
		 case disp_binary:
			knoflik=GetObjectPtr(prefBinary);
			break;
		 case disp_octal:
			knoflik=GetObjectPtr(prefOctal);
			break;
		 case disp_hexa:
			knoflik=GetObjectPtr(prefHexa);
			break;
		}
		CtlSetValue(knoflik,1);
		
		switch (calcPrefs.trigo_mode) {
		 case degree:
			knoflik=GetObjectPtr(prefDegree);
			break;
		 case grad:
			knoflik=GetObjectPtr(prefGrad);
			break;
		 case radian:
		 default:
			knoflik=GetObjectPtr(prefRadian);
			break;
		}
		CtlSetValue(knoflik,1);
		
		
		switch (dispPrefs.mode) {
		 case disp_normal:
			knoflik=GetObjectPtr(prefNormal);
			break;
		 case disp_sci:		
			knoflik=GetObjectPtr(prefSci);
			break;
		 case disp_eng:
		 default:
			knoflik=GetObjectPtr(prefEng);
			break;
		}
		CtlSetValue(knoflik,1);
/* Set the precision */
		list = GetObjectPtr(prefPrecList);
		LstSetSelection(list,dispPrefs.decPoints);
		  {
			  static char tmp[4]; /* It's simpler than MemPtrNew/Free.. */
			  StrPrintF(tmp,"%d",dispPrefs.decPoints);
			  CtlSetLabel(GetObjectPtr(prefPrecPopup),tmp);
		  }
		/* Set the topItem of the list */
		LstMakeItemVisible(list,dispPrefs.decPoints);
/* Set the checkboxes */	
		CtlSetValue( GetObjectPtr(prefReducePrecision),
					dispPrefs.reducePrecision);		
		CtlSetValue(GetObjectPtr(prefStrip),dispPrefs.stripZeros);		
		CtlSetValue(GetObjectPtr(prefForceComplex),dispPrefs.forceComplex);
		CtlSetValue(GetObjectPtr(prefIntInput),dispPrefs.forceInteger);
		CtlSetValue(GetObjectPtr(prefCvtUnits),dispPrefs.cvtUnits);
		CtlSetValue(GetObjectPtr(prefParenth),calcPrefs.matchParenth);
		
		handled=true;
		break;
	 case ctlSelectEvent:
		controlID=event->data.ctlSelect.controlID;  
		switch (controlID) {
		 case prefForceComplex:
			CtlSetValue(GetObjectPtr(prefIntInput),0);
			break;
		 case prefIntInput:
			CtlSetValue(GetObjectPtr(prefForceComplex),0);
			break;
		 case prefOK:
			if (CtlGetValue(GetObjectPtr(prefNormal)))
			  dispPrefs.mode=disp_normal;
			else if (CtlGetValue(GetObjectPtr(prefSci)))
			  dispPrefs.mode=disp_sci;
			else if (CtlGetValue(GetObjectPtr(prefEng)))
			  dispPrefs.mode=disp_eng;
			
			if (CtlGetValue(GetObjectPtr(prefDegree)))
			  calcPrefs.trigo_mode = degree;
			else if (CtlGetValue(GetObjectPtr(prefRadian)))
			  calcPrefs.trigo_mode = radian;
			else if (CtlGetValue(GetObjectPtr(prefGrad)))
			  calcPrefs.trigo_mode = grad;
			
			if (CtlGetValue(GetObjectPtr(prefDecimal)))
			  dispPrefs.base = disp_decimal;
			else if (CtlGetValue(GetObjectPtr(prefBinary)))
			  dispPrefs.base = disp_binary;
			else if (CtlGetValue(GetObjectPtr(prefOctal)))
			  dispPrefs.base = disp_octal;
			else if (CtlGetValue(GetObjectPtr(prefHexa)))
			  dispPrefs.base = disp_hexa;
			
			calcPrefs.matchParenth=CtlGetValue(GetObjectPtr(prefParenth));
			
			dispPrefs.stripZeros=CtlGetValue(GetObjectPtr(prefStrip));
			dispPrefs.decPoints = LstGetSelection(GetObjectPtr(prefPrecList));
			dispPrefs.reducePrecision=CtlGetValue(GetObjectPtr(prefReducePrecision));
			dispPrefs.forceComplex=CtlGetValue(GetObjectPtr(prefForceComplex));
			dispPrefs.forceInteger=CtlGetValue(GetObjectPtr(prefIntInput));
			dispPrefs.cvtUnits=CtlGetValue(GetObjectPtr(prefCvtUnits));

			fp_reread_prefs();
			FrmReturnToForm(0);
			FrmUpdateForm(calcPrefs.form,frmUpdateVars);
			handled=true;
			break;
		 case prefCancel:
			FrmReturnToForm(0);
			handled=true;
			break;
		}
	}	
	  
	return handled;
}

