/* Main code for GEORDI DB Installer */

#include <PalmOS.h>
#include <PalmCompatibility.h>
#include <SysEvtMgr.h>
#include <StringMgr.h>
#include <DataMgr.h>
#include <System/DLServer.h>
#include <sys_types.h>
#include <stdarg.h>
#include "geordi.h"
#include "installdbRsc.h"

#define INSTALLBTN 2001

BuildCmdObject navbco[9] =
{
{ 901,0,902,0,0,0,2,0,{"cd"},{"0"} },
{ 902,901,903,1,0,0,2,17,{"cp"},{"abdfiHlLpPrRsSuvx"} },
{ 903,902,904,2,0,0,2,8,{"ln"},{"bdfFnisSv"} },
{ 904,903,905,3,0,0,2,39,{"ls"},{"aAbBcCdDfFgGhHiIklLmnNopqrRsStTuUvwxX1"} },
{ 905,904,906,4,0,0,2,3,{"mkdir"},{"mpv"} },
{ 906,905,907,5,0,0,2,5,{"mv"},{"bfSuv"} },
{ 907,906,908,6,0,0,2,4,{"rm"},{"dfrv"} },
{ 908,907,909,7,0,0,2,2,{"rmdir"},{"pv"} },
{ 909,908,0,8,0,0,2,4,{"wc"},{"clLw"} }
};

BuildCmdObject adminbco[10] = 
{
{921,0,922,0,0,2,7,0,{"adduser"},{"0"}},
{922,921,923,1,0,2,8,0,{"addgroup"},{"0"}},
{923,922,924,2,0,2,6,0,{"passwd"},{"0"}},
{924,923,925,3,0,2,6,0,{"chroot"},{"0"}},
{925,924,926,4,0,2,4,0,{"cron"},{"0"}},
{926,925,927,5,0,2,7,0,{"delgroup"},{"0"}},
{927,926,928,6,0,2,5,0,{"dmesg"},{"0"}},
{928,927,929,7,0,2,2,12,{"df"},{"ahHiklmPtTxv"} },
{929,928,930,8,0,2,5,0,{"eject"},{"0"}},
{930,929,931,9,0,2,7,0,{"groupadd"},{"0"}}
};

BuildCmdObject netbco[7] = 
{
{911,0,912,0,1,0,8,1,{"ifconfig"},{"a"}},
{912,911,913,1,1,1,6,7,{"ifdown"},{"hVaivn"}},
{913,912,914,2,1,1,4,7,{"ifup"},{"hVaivn"}},
{914,913,915,3,1,1,7,16,{"netstat"},{"rgiMsvnAceopLaFC"}},
{915,914,916,4,1,1,4,12,{"ping"},{"dfnqrvRcilps"}},
{916,915,917,5,1,1,5,6,{"route"},{"AFCvne"}},
{917,916,0,6,1,1,10,18,{"traceroute"},{"dFIlnrvxfgimpqstwz"}},
};

DmOpenRef TrkDB;

void PalmOS_printf(char * pszFormat, ...)
{
   static int iOffset = 20;
   const int iDelta = 10;
   char szBuf[200];
   char *szBlanks = "                                                                             ";
   va_list args;

   va_start(args,pszFormat);
   StrVPrintF(szBuf,pszFormat,args);
   va_end(args);

   WinDrawChars(szBlanks,StrLen(szBlanks),0,iOffset);
   WinDrawChars(szBuf,StrLen(szBuf),0,iOffset);


   iOffset += iDelta;

   if (iOffset > 135) {
      iOffset = 15;
   }

}

void writedb() 
{
 BuildCmdObject *bcoptr; 
 MemHandle newbcoH;
 UInt16	   i;
 UInt16	   bcoIndex = 0;
 Err err;
 
   TrkDB=DmOpenDatabaseByTypeCreator(TrkDBType,TrkDBCreatorID,dmModeReadWrite|dmModeExclusive);
   if(!TrkDB) {
     PalmOS_printf("Creating TrkDB");
     err = DmCreateDatabase(0,TrkDBName,TrkDBCreatorID,TrkDBType,false);
     TrkDB=DmOpenDatabaseByTypeCreator(TrkDBType,TrkDBCreatorID,dmModeReadWrite|dmModeExclusive); 
     }
	   PalmOS_printf("Opened TrkDB !!"); 
          
           /*err = TrackAppInfoInit(TrkDB); */
 
           for(i = 0;i < 9;i++) {
	        newbcoH = DmNewRecord(TrkDB, &bcoIndex,sizeof(BuildCmdObject));
		bcoptr = MemHandleLock(newbcoH);
		DmWrite(bcoptr,0,&navbco[i],sizeof(BuildCmdObject));
		MemHandleUnlock(newbcoH);
		DmReleaseRecord(TrkDB,bcoIndex,true); 
		bcoIndex++;
	   }
           PalmOS_printf("Wrote nav BCO");   
	        bcoIndex++;
           for(i = 0;i < 10;i++) {
	        newbcoH = DmNewRecord(TrkDB, &bcoIndex,sizeof(BuildCmdObject));
		bcoptr = MemHandleLock(newbcoH);
		DmWrite(bcoptr,0,&adminbco[i],sizeof(bcoptr));
		MemHandleUnlock(newbcoH);
		DmReleaseRecord(TrkDB,bcoIndex,true); 
		bcoIndex++;
	   }
           PalmOS_printf("Wrote admin BCO");   
		bcoIndex++;
           for(i = 0;i < 7;i++) {
	        newbcoH = DmNewRecord(TrkDB, &bcoIndex,sizeof(BuildCmdObject));
		bcoptr = MemHandleLock(newbcoH);
		DmWrite(bcoptr,0,&netbco[i],sizeof(bcoptr));
		MemHandleUnlock(newbcoH);
		DmReleaseRecord(TrkDB,bcoIndex,true); 
		bcoIndex++;
	   }
           PalmOS_printf("Wrote net BCO");
		
	   DmCloseDatabase(TrkDB);
	   PalmOS_printf("Done!");	
 	   
}     

Err TrackAppInfoInit(DmOpenRef TrkDB) 

{
 UInt16 cardNo;
 LocalID TrkDBID,appInfoID;
 MemHandle h;
 CmdBuilderLayout *cbl;

 if (DmOpenDatabaseInfo(TrkDB,&TrkDBID,NULL,NULL,&cardNo,NULL))
	return dmErrInvalidParam;
 if (DmDatabaseInfo(cardNo,TrkDBID,NULL,NULL,NULL,NULL,NULL,NULL,NULL,&appInfoID,NULL,NULL,NULL))
	return dmErrInvalidParam;

 if (appInfoID == NULL) {
	h = DmNewHandle(TrkDB,sizeof(CmdBuilderLayout));
        if (!h) return dmErrMemError;
        appInfoID = MemHandleToLocalID(h);
	DmSetDatabaseInfo(cardNo,TrkDBID,NULL,NULL,NULL,NULL,NULL,NULL,NULL,&appInfoID,NULL,NULL,NULL);
 }
 
 cbl = MemLocalIDToLockedPtr(appInfoID,cardNo);
 
 DmSet(cbl,0,sizeof(CmdBuilderLayout), 0);

 cbl->ActiveSection = 0;
 cbl->TopTrack = 0;
 cbl->MiddleTrack = 1;
 cbl->BottomTrack = 2;

 MemPtrUnlock(cbl);
 return 0;

}
 
static Boolean MainFormHandleEvent (EventPtr e)
{
    Boolean handled = false;
    FormPtr frm;
    
    switch (e->eType) {
    case frmOpenEvent:
	frm = FrmGetActiveForm();
	FrmDrawForm(frm);
	handled = true;
	break;

    case menuEvent:
	MenuEraseStatus(NULL);

	switch(e->data.menu.itemID) {
	}

    	handled = true;
	break;

    case ctlSelectEvent:
	switch(e->data.ctlSelect.controlID) {

	case INSTALLBTN:
		writedb();
		handled = true;
		break;
	}
	break;

    default:
        break;
    }

    return handled;
}

static Boolean ApplicationHandleEvent(EventPtr e)
{
    FormPtr frm;
    Word    formId;
    Boolean handled = false;

    if (e->eType == frmLoadEvent) {
	formId = e->data.frmLoad.formID;
	frm = FrmInitForm(formId);
	FrmSetActiveForm(frm);

	switch(formId) {
	case MainForm:
	    FrmSetEventHandler(frm, MainFormHandleEvent);
	    break;
	}
	handled = true;
    }

    return handled;
}

/* Get preferences, open (or create) app database */
static Word StartApplication(void)
{
    FrmGotoForm(MainForm);
    return 0;
}

/* Save preferences, close forms, close app database */
static void StopApplication(void)
{
    FrmSaveAllForms();
    FrmCloseAllForms();
}

/* The main event loop */
static void EventLoop(void)
{
    Word err;
    EventType e;

    do {
	EvtGetEvent(&e, evtWaitForever);
	if (! SysHandleEvent (&e))
	    if (! MenuHandleEvent (NULL, &e, &err))
		if (! ApplicationHandleEvent (&e))
		    FrmDispatchEvent (&e);
    } while (e.eType != appStopEvent);
}

/* Main entry point; it is unlikely you will need to change this except to
   handle other launch command codes */
DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
    Word err;

    if (cmd == sysAppLaunchCmdNormalLaunch) {

	err = StartApplication();
	if (err) return err;

	EventLoop();
	StopApplication();

    } else {
	return sysErrParamErr;
    }

    return 0;
}
