#include <arpa/telnet.h>

typedef struct {       /*A Task Object describes a command on the main screen*/
	int TaskID;    /*The TaskID is used to identify each task on the main list                     */
 	int ChildID;   /*ChildID is the GUI Rsc ID of the dialog that pops up as a result of tapping on*/ 
	int ChildType; /*TaskID. ChildType is the kind of dialog that pops up                          */
	int taskxpos;  /*X position within the TaskList. Altered when the TaskList is scrolled	       */
	int taskypos;  /*Y "		"	"	 "	 "	  "	   "  "		       */
        UInt16 TaskTxtLen; /*Length of the text of the command				               */
	Char *TaskTxt;     /*pointer to the command text itself					       */
} TaskObject;

typedef struct {
	int BuildID; /*Similar to a TaskID, but used for commands on the Command Track of the Builder  */
	int LeftID;  /*ID of the Build command to the left of BuildID				       */
	int RightID; /*"  "      "	"	"     right of BuildID				       */
        int trackpos;/*position w/in the Command Track 						       */
	int track;   /*Track this command is on							       */
	int section; /*Section(Nav,Net,Admin or User) that this command is in 			       */
        UInt16 CmdTxtLen; /*Length of the text of the command					       */
	UInt16 FlagTxtLen;									       
	Char CmdText[10];    /*pointer to the command text itself				       */
        Char FlagText[50];  /*pointer to an array of flags(options) associated w/ this command	       */
} BuildCmdObject;

typedef struct {
	int numtracks;
	int db_numrecs;
	int track_numrecs[3];
} bcoDB_hdr;


typedef struct {
	UInt16 ctID;
	Char ctname[40];
	Char *CtText;
} Cocktail;


typedef struct {
       TaskObject to; 
       BuildCmdObject bo;
       UInt16 attribs; /*Universal attributes and permissions as a bitfield*/
       
} CmdObject;


typedef struct {
	FieldType *CLIPtr; 
        ControlType *BtnPtr;

}CLIObj;
typedef struct {
	Char user[10];
	Char pid[6];
	Char mem[6];
	Char cpu[6];
	Char cmd[41];
} psentry;

typedef struct {
	Char volname[40];
	Char device[25];
	Char type[10];
	Char size[10];
	Char used[10];
	Char avail[10];
	Char percent[5];
} dfentry;

typedef struct {
	Char size[11];
	Char path[80];
} duentry;

typedef struct {
	Char addr[15];
	Char mask[15];
	Char bcast[15];
	Char mac[18];
} ifentry;

typedef struct {
	int ActiveSection;
	int TopTrack;
	int MiddleTrack;
	int BottomTrack;
} CmdBuilderLayout;

typedef struct {
	Char user[16];
	Char fullname[35];
        Char homedir[30];
	Char shell[20];
} userentry;

typedef struct {
    int state;
    int *dead;
    FieldPtr msgfield;
} ErrP;

typedef struct {
    char *buf;
    int size;   /* total size of buffer */
    int head;   /* next input character goes here */
    int tail;   /* next output character comes from here */
    int count;  /* chars presently stored in buffer */
    // The buffer is empty when head==tail.

    int marked;   /* this character is marked */
} ringbuf; 

/*Functional prototypes */
#ifdef GEORDI_MAIN
void PalmOS_printf(char * pszFormat, ...);
void Err_Decode(UInt16 AlertID, Err errcode,Char* msg);
void Geordi_EraseProgress();
void Geordi_UpdateProgress();
void Geordi_ModemCtrl(int state);
UInt16 RandomNum(UInt n);
UInt16 Geordi_GetNetIF(Word libRefnum);
Err Geordi_Connect(Word librefnum,DWord creator_id,Word iface);
Err Geordi_Disconnect();
Char* GS_GetBytes(int sockd);
Char* GS_GetNBytes(int sockd,int numbytes);
Char* GS_GetLine(Int32 gst);
Int16 GS_SendBytes(int sockd,NetSocketAddrINType sockaddr,int buflen,unsigned char* GS_buf);
UInt16 GS_SendCommand(Char *CmdStr);

Err GS_TelOptChat(NetSocketRef sockd,NetSocketAddrINType sockaddr);
Int16 GS_UnameChat();
Err GS_LearnHostPrompt(NetSocketRef sockd,NetSocketAddrINType sockaddr);
UInt16 Geordi_GetNetIF(Word libRefnum);
int Geordi_ConnectMgr();
void DecrementActiveSelector(ControlType *ActiveCtl, int Ctlx,int Ctly,int ctrfloor,int ctrceil);
void IncrementActiveSelector(ControlType *ActiveCtl, int Ctlx,int Ctly,int ctrfloor,int ctrceil);

extern Boolean LoginFormHandleEvent (EventPtr e);
extern Boolean SignalFormHandleEvent (EventPtr e);
extern Boolean CocktailActFormHandleEvent (EventPtr e);
extern Boolean PaletteFormHandleEvent (EventPtr e); 
extern Boolean BuilderSaveFormHandleEvent (EventPtr e);
extern Boolean DskFormHandleEvent (EventPtr e);
extern Boolean DskDuFormHandleEvent (EventPtr e);
extern Boolean DskDuSelectFormHandleEvent (EventPtr e);
extern Boolean DskDetailsFormHandleEvent (EventPtr e);
extern Boolean IfaceListFormHandleEvent (EventPtr e);
extern Boolean CommandFormHandleEvent (EventPtr e);
extern Boolean FilterFormHandleEvent (EventPtr e);
extern Boolean LogFormHandleEvent (EventPtr e);
extern Boolean LogSelectFormHandleEvent (EventPtr e);
extern Boolean CliFormHandleEvent (EventPtr e);
extern Boolean UIDFormHandleEvent (EventPtr e);
extern Boolean FindFormHandleEvent (EventPtr e);

static Boolean isnumeric(Char c);
static Boolean isalpha(Char c);
Char* get_field(Char* wholestr, Char sep, int fieldpos);
Char* Geordi_AppendArgToCmdStr(Char *basestr, Char *param);
Char* GetOctetFromAddr(int part, Char *addr);
void GetAddrFromIface(FormType *frm, ListType *iflist);
Char *GetAddrFromForm(FormType *frm,int whichaddr);
int PopulateBtn(UInt16 BtnID, int TrkIndex);
int PopulateBtnTrack(FormType *frm,UInt16 PBtnIndex, UInt16 TrkDBIndex);
Err SetListSelectionByName(FormType *frm,ListType *LstPtr,Char *LstTargetText);
void DisableField(UInt16 DisFieldID);
void PopulateBtnRowByCmd(int CmdBtnIndex,int OptBtnIndex,Char *CmdBtnText);
void PopulateField(Char *NewFldText, Int16 FieldID);
void AppendBtnTxtToBuildField(ControlType *CmdBtn, FieldType *CmdFld);
void NudgeObject(FormType *frm,int refObjID,int nudgeObjID);

static Err LoadUID(void *table,Int16 row,Int16 column, Boolean editable,MemHandle *dataH,Int16 *dataOffset, Int16 *dataSize, FieldType *field);

static Err LoadPID(void *table,Int16 row,Int16 column, Boolean editable,MemHandle *dataH,Int16 *dataOffset, Int16 *dataSize, FieldType *field);

static Err LoadMem(void *table,Int16 row,Int16 column, Boolean editable,MemHandle *dataH,Int16 *dataOffset, Int16 *dataSize, FieldType *field);

static Err LoadCPU(void *table,Int16 row,Int16 column, Boolean editable,MemHandle *dataH,Int16 *dataOffset, Int16 *dataSize, FieldType *field);

static Err LoadCmd(void *table,Int16 row,Int16 column, Boolean editable,MemHandle *dataH,Int16 *dataOffset, Int16 *dataSize, FieldType *field);

static Err LoadVol(void *table,Int16 row,Int16 column, Boolean editable,MemHandle *dataH,Int16 *dataOffset, Int16 *dataSize, FieldType *field);

static Err LoadSize(void *table,Int16 row,Int16 column, Boolean editable,MemHandle *dataH,Int16 *dataOffset, Int16 *dataSize, FieldType *field);

static Err LoadPct(void *table,Int16 row,Int16 column, Boolean editable,MemHandle *dataH,Int16 *dataOffset, Int16 *dataSize, FieldType *field);

void ScrollTrackLeft(int TrkGroup, UInt16 TrkIndex,int TrkBase);
void ScrollTrackRight(int TrkGroup, UInt16 TrkIndex,int TrkBase);

Int16 Geordi_DoPSCmd();
Int16 Geordi_DoDF();
Int16 Geordi_DoDU(Char *DuCmdStr,Int32 numlines);
int Geordi_DoUmount(Char *FsPtr);
int Geordi_DoMount();

void Geordi_ParseResults(Char* parsebuf,Char* formstr,int cmdtype,int linenum,int buflen);

void Geordi_PsTableUpdate();
void Geordi_PsTableFlush();
void Geordi_DfTableUpdate();
void Geordi_DfTableFlush();
void AuthRenewalHandler(Word AlmProcCmd,SysAlarmTriggeredParamType *paramP);

UInt16 GS_GetResults(int Type, Char *kw,Char *kw2, Char *cmdFormStr,UInt16 KeywordLen, int RLL_Threshhold,int maxresults);

UInt8 GS_GetTelOptByte(int sockd);

void Init_dfTable(FormType *frm);
void Init_CBTable(FormType *frm);
void Init_psTable(FormType *frm);

void SetLoginLabel(Char *LogLabel);

#endif
#define TrkDBName "TrkDB"
#define TrkDBType 'DATA'
#define TrkDBCreatorID 'GEOR'
#define TrfDBName "TrfDB"
#define TrfDBType "DATA"
#define TrfDBCreatorID "GEOR"

#define CMD_PS 101
#define CMD_IF	 102
#define CMD_DF 103
#define CMD_NS 104
#define CMD_TD 105
#define CMD_DU 106

#ifndef __ERF_H__
#define __ERF_H__

typedef void (*Erf)(int, char *, char *, char *, int, void *);

#endif
