/*
 * Structures and functions used by GutenPalm
 */

#ifndef GUTENPALM_H
#define GUTENPALM_H

#include <PalmOS.h>
#include <PalmTypes.h>
#include "zlib-palmos.h"


/* Constants for GutenPalm */
#define GPLM_VERSION	0x0100

/*  Databases of type 'zTMP'.  They hold the uncompressed data  */
#define ZTMP_TYPE_ID	"zTMP"

/*  Default size of document list  */
#define MIN_DOCLIST_SIZE        (sizeof(struct DocListType) + \
                                (sizeof(struct DocEntryType) * 9))


/* An individual zTXT entry in the DocList */
typedef struct DocEntryType {
  Boolean       cache;
  Char          name[dmDBNameLength];
} DocEntry;

/* This is a list of all zTXT's available at runtime */
typedef struct DocListType {
  UInt16        numDocuments;
  DocEntry      documents[1];
} DocList;

/* A structure for the currently displayed zTXT */
typedef struct DocumentType {
  /* Info from record #0 */
  UInt8         version;
  UInt32        size;
  UInt16        numRecords;
  UInt16        recordSize;
  UInt16        numBookmarks;
  UInt16        bookmarkRecord;
  /* Bookmark record info */
  MemHandle     bmrkRecordHandle;
  MemPtr        bmrkRecordPtr;
  /* Currently open record info */
  UInt16        curRecordIndex;
  MemHandle     curRecordHandle;
  MemPtr        curRecordPtr;
  /* Currently displayed text state */
  UInt32        curPos;
  Boolean       startOfDoc;
  Boolean       endOfDoc;
  struct {
    UInt16      offset;
    UInt16      recIndex;
  } curPage;
  struct {
    UInt16      offset;
    UInt16      recIndex;
  } nextPage;
  /* zTXT database info */
  UInt16        dbIndex;
  Char          dbName[dmDBNameLength];
  LocalID       dbID;
  DmOpenRef     dbRef;
  /* Temp storage DB for decompressed text */
  Char          tempName[dmDBNameLength];
  LocalID       tempID;
  DmOpenRef     tempRef;
  UInt16        numTmpRecords;
  /* Buffers and streams - only needed during decompression */
  z_stream      zstream;
  MemPtr        zBuffer;
} Document;



/* Event handlers */
extern Boolean  MainFormHandleEvent(EventPtr);
extern Boolean  DocViewFormHandleEvent(EventPtr);

/* Helper functions */
extern UInt32   id_to_int(Char *);
extern void     SetFieldText(UInt16, Char *);
extern void     SetFieldTextAndDraw(UInt16, Char *);
extern void     OpenTmpRecord(UInt16);
#ifdef GPLM_DEBUG
extern void     DebugMessage(Char *);
#endif


/* GutenPalm globals */
extern MemHandle gDoclistHandle;
extern UInt16    gNumDocuments;
extern Document  gCurrentDoc;
extern UInt16    gCurrentForm;


#endif
