/*
 * GutenPalm: A zTXT format document reader for the Palm Handheld Organizer
 * Copyright (C) 2000 John Gruenenfelder
 *   johng@as.arizona.edu
 *   http://gutenpalm.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *      Free Software Foundation, Inc.
 *      59 Temple Place
 *      Suite 330
 *      Boston, MA 02111-1307
 *      USA
 *
 */


#include <PalmOS.h>
#include <PalmTypes.h>
#include <DataMgr.h>

#include "gutenpalm.h"



/* Sets the text in a field with a given string */
void
SetFieldText(UInt16 fieldID, Char *newStr)
{
  MemHandle     oldTxtHandle;
  MemHandle     txtHandle;
  FormPtr       frm;
  FieldPtr      field;

  txtHandle = MemHandleNew(StrLen(newStr) + 1);
  StrCopy(MemHandleLock(txtHandle), newStr);
  MemHandleUnlock(txtHandle);

  frm = FrmGetActiveForm();
  field = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID));
  oldTxtHandle = FldGetTextHandle(field);

  FldSetTextHandle(field, txtHandle);
  if (oldTxtHandle)
    MemHandleFree(oldTxtHandle);
}


/* Sets the text in a field with a given string and redraw the field */
void
SetFieldTextAndDraw(UInt16 fieldID, Char *newStr)
{
  FormPtr       frm;

  SetFieldText(fieldID, newStr);
  frm = FrmGetActiveForm();
  FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID)));
}


/* Open a record in the temp database */
void
OpenTmpRecord(UInt16 index)
{
  if ((index != gCurrentDoc.curRecordIndex) &&
      (index < gCurrentDoc.numTmpRecords))
    {
      /* Close old record */
      MemHandleUnlock(gCurrentDoc.curRecordHandle);
      DmReleaseRecord(gCurrentDoc.tempRef, gCurrentDoc.curRecordIndex, false);
      /* Open new record */
      gCurrentDoc.curRecordIndex = index;
      gCurrentDoc.curRecordHandle = DmGetRecord(gCurrentDoc.tempRef, index);
      gCurrentDoc.curRecordPtr = MemHandleLock(gCurrentDoc.curRecordHandle);
    }
}


/*  Converts a 4 byte string Palm ID into a UInt32 */
UInt32
id_to_int(Char *idstr)
{
  if (StrLen(idstr) != 4)
    return 0;

  return (0x1000000*(UInt32)idstr[0] +
          0x10000*(UInt32)idstr[1] +
          0x100*(UInt32)idstr[2] +
          (UInt32)idstr[3]);
}


/* Prints a debug message at the bottom of the screen and scrolls
 * everything else up one line
 */
#ifdef GPLM_DEBUG
void
DebugMessage(Char *message)
{
  static RectangleType bottomRec = {{0,148}, {160,11}};
  RectanglePtr rectPtr = NULL;

  WinGetWindowBounds(rectPtr);
  WinScrollRectangle(rectPtr, winUp, 11, NULL);
  WinEraseRectangle(&bottomRec, 0);
  WinDrawChars(message, StrLen(message), 0, 148);
}
#endif
