/*
 * Simple http (web) server.
 * Jim Rees University of Michigan
 * Robin Kirkman
 * Original version April 1998
 */

#pragma pack(2)

#if (__GNUC_MINOR__ > 7)
#include <PalmOS.h>
#else
#include <Common.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>
#endif
#include <System/DataMgr.h>

#include <System/Unix/unix_stdio.h>
#include <System/Unix/unix_stdlib.h>
#include <System/Unix/unix_string.h>

#include "httpd.h"

#define NCATS 16

static char appname[] = "Memos";

int send_memo_index(XFILE *f, char *url)
{
    int i, cat;
    UInt16 n, oldn;
    DmOpenRef MemoDB;
    char title[42], *mp, *doctype, catname[dmCategoryLength], *u, ub[80];
    void *hp;

    MemoDB = DmOpenDatabaseByTypeCreator('DATA', 'memo', dmModeReadOnly);
    if (!MemoDB) {
	send_404(f, "/memo/");
	return;
    }

    wrstr(f, html0, "text/html");
    wrstr(f, "<title>Directory listing for /memo/%s</title><h1>Directory listing for /memo/%s</h1>\n", url, url);

    for (cat = n = 0; cat < NCATS; ) {
	oldn = n;
	hp = DmQueryNextInCategory(MemoDB, &n, cat);
	if (!hp) {
	    if (n)
		xputs(f, "<p>\n");
	    n = 0;
	    cat++;
	    continue;
	}
	if (oldn == 0) {
	    CategoryGetName(MemoDB, cat, catname);
	    wrstr(f, "<b>Files in category <i>%s</i>:</b><p>\n", catname);
	}

	mp = (char *) MemHandleLock(hp);

	get_memo_title(mp, title);

	MemPtrUnlock((void *) mp);

	if(!strncmp(title,url,strlen(url)) || strlen(url)==0){
	    mp=title+strlen(url);
	    while(*mp && *mp != '/' ) {
		mp++;
	    }

	    strcpy(ub,title);
	    u=ub;
	  
	    while(*++u)
		*u=(*u==' ')?'+':*u;
	    u=ub+strlen(url);
	    while(*u && *u != '/' ) ++u;
	    if( *mp!='/' ) {
		wrstr(f, "<a href=\"/memo/%s\">%s</a><br>\n", ub, title+strlen(url));
	    } else {
		*++mp=0; *++u=0;
		wrstr(f, "<a href=\"/memo/%s\">%s</a><br>\n",ub,title+strlen(url));
	    }
#ifdef DEBUG
	} else {
	    wrstr(f, "Rejected file |%s| for url |%s|<br>\n",title,url);
#endif
	}
	n++;
    }
    DmCloseDatabase(MemoDB);
    xputs(f, html_go_home);
}

send_memo_file(XFILE *f, char *url)
{
    int cat;
    DmOpenRef MemoDB;
    char doctype[20], subtype[41], *cp, title[41], *mp, mext[20], mtype[40];
    void *hp=0, *chp=0, *mime=0;
    UInt16 n, oldn,i=0;

    title[40]=0;

    MemoDB = DmOpenDatabaseByTypeCreator('DATA', 'memo', dmModeReadOnly);
    if (!MemoDB)
	goto notfound;
    
    strcpy(subtype,url+6);

    mp=subtype;
    while(*mp) mp++;
    mp--;
    if(*mp == '/') {
	return send_memo_index(f,subtype);
    }

    for (cat = n = 0; cat < NCATS; ) {
	oldn = n;
	hp = DmQueryNextInCategory(MemoDB, &n, cat);
	if (!hp) {
	    n = 0;
	    cat++;
	    continue;
	}

	mp = (char *) MemHandleLock(hp);

	get_memo_title(mp, title);

	if(!strcmp(subtype,title)) {
	    chp=hp;
	}

	MemPtrUnlock((void *) mp);

	if(!strcmp(title,"MIME_TYPES.HTTPD"))
	    mime=hp;

	n++;
    }



    hp = chp;

    if (!hp)
	goto notfound;

    /* mime type info is encoded in the url for now */

    mp = cp = (char *) MemHandleLock(hp);

    while(*cp && *cp != '\n' && *cp != '\r' && i<40) {
	++cp; ++i;
    }

    if (mime == NULL) {
	/* Hardcoded types, if we can't find the mime file... */
	if(!strncmp(cp-5,".html",5)) {
	    sprintf(doctype, "text/html\n\n");
	} else {
	    sprintf(doctype, "text/plain\n\n");
	}
    } else {
	/* This is the good stuff... */
	char *mf,*mti;
#ifdef DEBUG
	printf("Using MIME file...\n");
#endif
	mti=mf=MemHandleLock(mime);
	while(*mf++!='\n');
	strcpy(doctype,"text/plain"); /* Default type... */
	while(sscanf(mf,"%s\t%s\n",mext,mtype)==2) {
#ifdef DEBUG
	    printf("%s -> %s (%s)\n",mext,mtype);
#endif
	    if(!strncmp(cp-strlen(mext),mext,strlen(mext))) {
		strcpy(doctype,mtype);
#ifdef DEBUG
		printf("Match!\n");
#endif
	    }
	    while(*mf++!='\n');
	}
	MemPtrUnlock((void *) mti);
    }

    wrstr(f, html0, doctype);


    cp++;
    xputs(f, cp);

    MemPtrUnlock((void *) mp);
    xputs(f, "\n");
    DmCloseDatabase(MemoDB);
    return;

 notfound:
    if (MemoDB)
	DmCloseDatabase(MemoDB);
    send_404(f, url);
    return;
}

get_memo_title(char *mp, char *title)
{
    char *cp0, *cp1;

    /* copy memo title, first line or up to 40 chars */
    for (cp0 = mp, cp1 = title; *cp0 && *cp0 != '\n' && *cp0 != '\r' && cp0 - mp < 40; )
	*cp1++ = *cp0++;
    *cp1 = '\0';
}

/*
copyright 1998, 2000
the regents of the university of michigan
all rights reserved

permission is granted to use, copy, create derivative works
and redistribute this software and such derivative works
for any purpose, so long as the name of the university of
michigan is not used in any advertising or publicity
pertaining to the use or distribution of this software
without specific, written prior authorization.  if the
above copyright notice or any other identification of the
university of michigan is included in any copy of any
portion of this software, then the disclaimer below must
also be included.

this software is provided as is, without representation
from the university of michigan as to its fitness for any
purpose, and without warranty by the university of
michigan of any kind, either express or implied, including
without limitation the implied warranties of
merchantability and fitness for a particular purpose. the
regents of the university of michigan shall not be liable
for any damages, including special, indirect, incidental, or
consequential damages, with respect to any claim arising
out of or in connection with the use of the software, even
if it has been or is hereafter advised of the possibility of
such damages.
*/
