#include <Pilot.h>
#include "iLarnRsc.h"
#include "Globals.h"

/***********************************************************************
 * This form is the select-thing-from-inventory form.
 * It is used when you list your inventory, and in the trading post.
 *
 * When player selects an item, buttons are enabled or disabled (and may
 * also be re-labeled) to let him/her do the actions appropriate to that
 * type of item and to his/her current location.
 * There is also a cancel button.
 * If the player hits a variably-labeled button, we call a routine that
 * figures out (based on item type) what the action is supposed to be.
 **********************************************************************/

extern Short in_store;
Short inventory_item; // (a to z) - 'a' .. index into env->iven, etc.
Boolean InvSelect_Form_HandleEvent(EventPtr e)
{
  Boolean handled = false;
  ListPtr lst;
  FormPtr frm;
  CharPtr label;
  Short selected_btn, k;
  Boolean took_time;

  switch (e->eType) {

  case frmOpenEvent:
    frm = FrmGetActiveForm();
    /* ...init form... */
    init_inventory_select(frm, (in_store==STORE_TPOST));
    // incorporates FrmDrawForm!
    inventory_item = -1;
    handled = true;
    break;

  case lstSelectEvent:
    frm = FrmGetActiveForm();
    lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, list_i));
    label = LstGetSelectionText(lst, e->data.lstSelect.selection);
    if (label) inventory_item = label[0] - 'a';
    else inventory_item = -1; // bug if this happens
    //    inventory_item = e->data.lstSelect.selection;
    specialize_labels(frm, inventory_item, (in_store==STORE_TPOST));
    handled = true;
    break;

  case keyDownEvent:
    // hardware button -- or else graffiti.
    frm = FrmGetActiveForm();
    lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, list_i));
    switch(e->data.keyDown.chr) {
    case pageUpChr:
      if (LstScrollList(lst, up, 11)) {// there are 11 items visible at a time
	LstSetSelection(lst, -1);
	inventory_item = -1;
      }
      handled = true;
      break;
    case pageDownChr:
      if (LstScrollList(lst, down, 11)) {
	LstSetSelection(lst, -1);
	inventory_item = -1;
      }
      handled = true;
      break;
    }
    break;

  case ctlSelectEvent:
    frm = FrmGetActiveForm();
    switch(e->data.ctlSelect.controlID) {
    case btn_if_frob:
      if (in_store == STORE_TPOST && inventory_item != -1) {
	// pop up something to sell it..
	// then, we have to redo the darned inventory, probably
	if (tpost_sell_item(inventory_item)) {
	  lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, list_i));
	  k = LstGetSelection(lst);
	  remove1_inventory_select(frm, k);
	}
	break;
      }
      // if we're not in the TPOST then we just fall through.
    case btn_if_drop:
    case btn_if_wield:
      selected_btn = e->data.ctlSelect.controlID - btn_if_frob;
      free_inventory_select(frm);
      LeaveForm(); // LeaveForm(true);
      /* It is important to do 'stuff' AFTER LeaveForm for msg to display */
      message_clear();
      if (inventory_item != -1) {
	// Do The Right Thing
	took_time = deduce_task(inventory_item, selected_btn);
	if (took_time)
	  start_of_Play_Level_loop();
	//else if (env->pending_splash)
	//	  show_messages();  // FrmPopupForm(SplashForm);
      }
      handled = true;
      break;
    case btn_i_cancel:
      free_inventory_select(frm);
      LeaveForm();
      if (in_store == STORE_TPOST)
	print_stats(); // in case you spent money
      handled = true;
      break;
    }

  default:
    break;
  }

  return handled;
}
