/*

    WheelDB - Creates Wheel of Fortune Databases for the PalmPilot
    Version 1.00

    Copyright 1997 Scott C. Duensing
    (scott@jaegertech.com)

    1997 Jaeger Technologies
    (www.jaegertech.com)



    Based on zvert - a brutal unix port of R. Bram's ZIL->PDB program
    by Brian J. Swetland <swetland@uiuc.edu>

*/

//---------------------------------------------------------------------------
// Includes
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

//---------------------------------------------------------------------------
// Remove crap from a line of text
void CleanUp (char *what)
{
  int x;

  // Clean tail end
  x = strlen (what);
  while (x > 0)
  {
    if ((what[x] == 0)  ||    // NULL
        (what[x] == 10) ||    // LF
        (what[x] == 13) ||    // CR
        (what[x] == 32) ||    // Space
        (what[x] == 9))       // TAB
      x--;
    else
      break;
  }
  what[x + 1] = 0;

  // Clean front end
  x = 0;
  while (x < strlen (what))
  {
    if ((what[x] == 0)  ||    // NULL
        (what[x] == 10) ||    // LF
        (what[x] == 13) ||    // CR
        (what[x] == 32) ||    // Space
        (what[x] == 9))       // TAB
      x++;
    else
      break;
  }
  if (x > 0)
    memmove (what, &what[x], strlen (what) - x + 1);
}

//---------------------------------------------------------------------------
// Reverse byte orders on longs (32 bits)
long htonl (long motorola)
{
  char *cp1;
  char *cp2;
  long lv;

  cp1 = (char *)&motorola;
  cp2 = (char *)&lv;

  cp2[0] = cp1[3];
  cp2[1] = cp1[2];
  cp2[2] = cp1[1];
  cp2[3] = cp1[0];

  return lv;
}

//---------------------------------------------------------------------------
// Reverse byte orders on ints (16 bits)
int htons (int motorola)
{
  char *cp1;
  char *cp2;
  int iv;

  cp1 = (char *)&motorola;
  cp2 = (char *)&iv;

  cp2[0] = cp1[1];
  cp2[1] = cp1[0];

  return iv;
}

//---------------------------------------------------------------------------
// Main Program
void main (int argc, char *argv[])
{
  int x;
  int y;
  int line;
  int puzzleline;
  int records;
  char c;
  char linecheck[5];
  char buffer[256];
  char puzzle[5][15];
  char buff[32];
  char name[32];
  FILE *infile;
  FILE *outfile;

  unsigned long size, w, w2;
  unsigned long flags = 0xAE44D106;
  unsigned int  z, z2, numRecs;
  unsigned char head[4];

  printf ("\n\n"
          "=================================\n"
          "Wheel of Fortune Database Creator\n"
          "Version 1.00\n\n"
          "Copyright 1997 Scott C. Duensing\n"
          "(scott@jaegertech.com)\n\n"
          "1997 Jaeger Technologies\n"
          "(www.jaegertech.com)\n"
          "=================================\n\n");

  if (argc != 4)
  {
    printf ("USAGE:  WheelDB [inputfile.wdb] [outputfile.pdb] [\"Description\"]\n\n");
    return;
  }

  memset (name, 0, 23);
  if ((strlen (argv[3]) < 1) || (strlen (argv[3]) > 31))
  {
    printf ("ERROR:  Description is invalid.\n\n");
    return;
  }
  strcpy (name, argv[3]);

  infile = fopen (argv[1], "rt");
  if (infile == NULL)
  {
    printf ("ERROR:  Unable to open %s!\n\n", argv[1]);
    return;
  }

  outfile = fopen ("wheeldb.tmp", "wb");
  if (outfile == NULL)
  {
    fclose (infile);
    printf ("ERROR:  Unable to open wheeldb.tmp!\n\n");
    return;
  }

  printf ("Processing...  ");

  line = 0;
  records = 0;
  linecheck[0] = linecheck[1] = linecheck[2] = linecheck[3] = linecheck[4] = 0;
  memset (puzzle, 0, sizeof (puzzle));
  while (!feof (infile))
  {
    fgets (buffer, 255, infile);
    if (feof (infile))
      continue;

    line++;

    CleanUp (buffer);

    // Check for comments
    if ((buffer[0] == '/') && (buffer[1] == '/'))
      continue;

    // Check for blanks
    if (buffer[0] == 0)
      continue;

    // Check for a valid puzzle line
    puzzleline = (int)(buffer[0] - '0');
    if ((puzzleline < 1) || (puzzleline > 5))
    {
      printf ("\n\nERROR:  Invalid line number on line %d!\n", line);
      break;
    }

    // Make us 0-4 instead of 1-5
    puzzleline--;

    // Check for a valid puzzle line
    if (puzzleline < 4)
      if (strlen (buffer) != 15)
      {
        printf ("\n\nERROR:  Invalid line length on line %d!\n", line);
        break;
      }

    // Do we already have this line?
    if (linecheck[puzzleline] != 0)
    {
      printf ("\n\nERROR:  Duplicate puzzle line on line %d!\n", line);
      break;
    }

    // Store puzzle line
    linecheck[puzzleline] = 1;
    strcpy (puzzle[puzzleline], &buffer[1]);

    // Did we get a whole puzzle?
    if ((linecheck[0] + linecheck[1] + linecheck[2] + linecheck[3] + linecheck[4]) == 5)
    {
      // Clean up puzzle data for Pilot
      for (x=0; x<14; x++)
        for (y=0; y<4; y++)
          if (puzzle[y][x] == '*')
            puzzle[y][x] = ' ';
          else
            puzzle[y][x] = tolower (puzzle[y][x]);
      puzzle[0][0] = puzzle[0][13] = puzzle[3][0] = puzzle[3][13] = '*';

      // Write puzzle data to temp file
      fwrite (puzzle, sizeof (puzzle), 1, outfile);

      // Get ready for next puzzle
      linecheck[0] = linecheck[1] = linecheck[2] = linecheck[3] = linecheck[4] = 0;
      memset (puzzle, 0, sizeof (puzzle));
      records++;
    }
  }

  fclose (outfile);
  fclose (infile);

  memset (buff, 0, 23);

  infile = fopen ("wheeldb.tmp", "rb");
  if (infile == NULL)
  {
    printf ("\n\nERROR:  Unable to open wheeldb.tmp!\n\n");
    return;
  }

  outfile = fopen (argv[2], "wb");
  if (outfile == NULL)
  {
    fclose (infile);
    printf ("\n\nERROR:  Unable to open %s!\n\n", argv[2]);
    return;
  }

  // --- Database Header (78 bytes) ---

  fwrite (name, 32, 1, outfile); // DB Name
  fwrite (buff, 2, 1, outfile);  // Flags
  fwrite (buff, 2, 1, outfile);  // Version

  flags = htonl (flags);
  fwrite (&flags, 4, 1, outfile); // Creation Time
  fwrite (&flags, 4, 1, outfile); // Modificaton Time

  fwrite (buff, 4, 1, outfile); // Backup Time
  fwrite (buff, 4, 1, outfile); // Modification Number
  fwrite (buff, 4, 1, outfile); // App Info Offset
  fwrite (buff, 4, 1, outfile); // Sort Info Offset

  fwrite ("JTpd", 4, 1, outfile); // Type
  fwrite ("JTpd", 4, 1, outfile); // Creator (was JTwf)

  fwrite (buff, 4, 1, outfile); // unique ID seed
  fwrite (buff, 4, 1, outfile); // next record list id

  z = records + 1;
  z2 = htons (z);
  fwrite (&z2, 2, 1, outfile); // Number of records for this header

  // --- Record Entry Header (8 bytes) ---

  w = 0x50L + (long)z * 8;

  head[0] = 0x40;  // (Record Attributes)
  head[1] = 0x6f;  // (Unique ID)
  head[2] = 0x80;  // (Unique ID)
  head[3] = 0;     // (Unique ID)

  // Write out index for header records
  w2 = htonl (w);
  fwrite (&w2, 4, 1, outfile);  // Offset
  fwrite (head, 4, 1, outfile); // Record attributes (1) & unique ID (3)
  w += 8;
  head[3]++;
  z--;

  // Write out index for puzzle records
  while (z--)
  {
    w2 = htonl (w);
    fwrite (&w2, 4, 1, outfile);  // Offset
    fwrite (head, 4, 1, outfile); // Record attributes (1) & unique ID (3)
    w += sizeof (puzzle);
    head[3]++;
  }

  // --- End of Headers ---

  head[0] = 0;
  head[1] = 0;
  fwrite (head, 2, 1, outfile); // Unknown

  // --- Write out Header Record ---

  z = 1;
  z2 = htons (z);
  fwrite (&z2, 2, 1, outfile);  // Major Version

  z = 0;
  z2 = htons (z);
  fwrite (&z2, 2, 1, outfile);  // Minor Version

  z = records;
  z2 = htons (z);
  fwrite (&z2, 2, 1, outfile);  // Number of Puzzles

  z = 1;
  z2 = htons (z);
  fwrite (&z2, 2, 1, outfile);  // Current Puzzle Number

  // --- Write out Puzzle Records ---

  while (!feof (infile))
  {
    fread (&c, 1, 1, infile);
    if (!feof (infile))
      fwrite (&c, 1, 1, outfile);
  }

  z = ftell (outfile);

  fclose (outfile);
  fclose (infile);

  printf ("Finished!\n\n");
  printf ("- %d puzzles written.\n", records);
  printf ("- %d bytes each.\n", sizeof (puzzle));
  printf ("- %d puzzle bytes.\n", sizeof (puzzle) * records);
  printf ("- %lu total bytes.\n\n", (long)z);
}
