//
//  "JT WOF" for the PalmPilot
//  Version 1.20
//
//  Copyright 1997 Scott C. Duensing
//  (scott@jaegertech.com)
//
//  1997 Jaeger Technologies
//  (www.jaegertech.com)
//
//
//  Changes in this version:
//
//   - Multiple rounds added
//   - Wheel values change depending on round
//   - Starting player changes depending on round
//   - Game actually ends after four rounds
//   - No longer able to spin when there are no more consonants
//   - No longer able to buy when there are no more vowels
//   - Must solve when all letters are displayed
//   - Letter Board added to show available letters
//   - High score table modifications started, not ready yet
//   - Editable player names modifications started, not ready yet
//
// --------------------------------------------------------------------------

#pragma pack(2)

// --------------------------------------------------------------------------

#include <Common.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>

// --------------------------------------------------------------------------

#include "wheel.h"

// --------------------------------------------------------------------------

static void ShowLetterBoard (void);
static void NextRound (void);
static void DrawDBBMP (int index, int x, int y);
static void NewGame (void);
static void RestoreScreen (void);
static void LoadPuzzle (void);
static void CheckForSolve (void);
static void HandleSolveLetter (UInt letter);
static void GetNextUnknown (int position);
static void BuyVowel (void);
static void NextPlayer (void);
static int IsVowel (UInt letter);
static int PickLetter (UInt letter);
static void Clear (int where);
static void SomeoneSays (int who, char *message);
static void VannaSays (char *message);
static void PatSays (char *message);
static void AnimationStart (int which);
static void DrawBitmap (int handle, int x, int y);
static void DrawPuzzle (void);
static void DrawScores (int which);
static void AnimationRun (void);
static Boolean ApplicationHandleEvent (EventPtr event);
static void StartApplication (void);
static void StopApplication (void);
static void EventLoop (void);
DWord PilotMain (Word cmd, Ptr cmdPBP, Word launchFlags);

// --------------------------------------------------------------------------

#define AppCreator   0x4A547766  // JTwf
#define DataCreator  0x4A547064  // JTpd
#define ImageCreator 0x4A546964  // JTid

#define AppVersion   120

#define gstAtTitle            0
#define gstPatSaysRound       1
#define gstWaitToDrawButtons  2
#define gstSpinBuySolve       3
#define gstPickLetter         4
#define gstBuyVowel           5
#define gstDoingNothing       6
#define gstWaitingForHost     7
#define gstSolving            8
#define gstEndOfRound         9
#define gstPatSaysPlayer     10
#define gstPatSaysScore      11
#define gstEndOfGame         12

#define dscRoundScores 0
#define dscTotalScores 1

#define clrLL 0
#define clrLR 1
#define clrUL 2

#define aniWheel      1
#define aniSlideValue 2
#define aniBlinkSpin  3
#define aniBlinkBuy   4
#define aniBlinkSolve 5
#define aniSolving    6

#define TotalWheelValues    26
#define TotalSpecialSpaces   2
#define ValueSlot            0

#define valBankrupt    0
#define valLoseATurn  -1

// --------------------------------------------------------------------------

typedef struct {
  int MajorVersion;
  int MinorVersion;
  int Records;
  int Current;
} PuzzleDBHeader;

// --------------------------------------------------------------------------

static int WheelValues[TotalWheelValues] = 
{
  1000,  // This is 'ValueSlot'
  250, 
  300,
  400,
  valLoseATurn,
  500,
  550,
  600,
  700,
  750,
  valBankrupt,
  900,
  950,
  250, 
  300,
  valLoseATurn,
  400,
  500,
  550,
  600,
  700,
  750,
  valBankrupt,
  900,
  950,
  250  // Should be a Free Spin
};

static Char SpecialSpaces[TotalSpecialSpaces][32] = {
  "Bankrupt!",
  "Lose a Turn"
};

static FormPtr CurrentForm;

typedef struct {
  char name[11];
  int score;
} highscores_T;

// The following variables need to be saved between sessions
typedef struct {
  int VowelsLeft;
  int ConsonantsLeft;
  char Remaining[27];

  char Puzzle[5][15];
  char TempPuzzle[5][15];

  int TotalScores[3];
  int Scores[3];
  int CurrentPlayer;
  int NumberOfPlayers;
  char PlayerNames[3][11];

  char GameState;

  char WhichAni;
  int Speed;
  int RandomValue;
  int LoopCount;
  RectangleType WheelValueBox;
  Boolean Running;
  Boolean JustSpun;

  int LastWheelFrame;
  int LastWheelValue;

  int SolvingLetter;
  int SolvingX;
  int SolvingY;

  int Round;

  char WindowBuffer[1600];

  highscores_T Scoreboard[10];
} Persistant_T;

static Persistant_T Persistant;

// --------------------------------------------------------------------------

static void DrawDBBMP (int index, int x, int y)
{
  ULong DBType = ImageCreator;
  ULong DBCreator = ImageCreator;
  DmOpenRef db;
  char temp[64];
  VoidHand rechandle;
  VoidPtr dataptr;

  db = DmOpenDatabaseByTypeCreator (DBType, DBCreator, dmModeReadOnly);
  if (db == NULL)
  {
    StrCopy (temp, "Error opening image database.");
    WinDrawChars (temp, StrLen (temp), 0, 0);
    return;
  }

  rechandle = DmGetRecord (db, index);
  if (rechandle == 0)
  {
    StrCopy (temp, "Error reading image record.");
    WinDrawChars (temp, StrLen (temp), 0, 0);
    DmCloseDatabase (db);
    return;
  }

  dataptr = MemHandleLock (rechandle);
  WinDrawBitmap (dataptr, x, y);
  MemHandleUnlock (rechandle);

  DmReleaseRecord (db, index, false);

  DmCloseDatabase (db);
}

// --------------------------------------------------------------------------
static void LoadPuzzle (void)
{
  ULong DBType = DataCreator;
  ULong DBCreator = DataCreator;
  DmOpenRef db;
  char temp[128];
  VoidHand rechandle;
  VoidHand headhandle;
  PuzzleDBHeader Header;
  int last;

  db = DmOpenDatabaseByTypeCreator (DBType, DBCreator, dmModeReadWrite);
  if (db == NULL)
  {
    StrCopy (temp, "Error opening puzzle database.");
    WinDrawChars (temp, StrLen (temp), 0, 0);
    return;
  }

  headhandle = DmGetRecord (db, 0);
  if (headhandle == 0)
  {
    StrCopy (temp, "Error reading header record.");
    WinDrawChars (temp, StrLen (temp), 0, 0);
    DmCloseDatabase (db);
    return;
  }

  MemMove (&Header, MemHandleLock (headhandle), sizeof (PuzzleDBHeader));
  MemHandleUnlock (headhandle);

  rechandle = DmGetRecord (db, Header.Current);
  if (rechandle == 0)
  {
    StrCopy (temp, "Error reading puzzle record.");
    WinDrawChars (temp, StrLen (temp), 0, 0);
    DmReleaseRecord (db, 0, false);
    DmCloseDatabase (db);
    return;
  }

  MemMove (Persistant.Puzzle, MemHandleLock (rechandle), sizeof (Persistant.Puzzle));
  MemHandleUnlock (rechandle);

  last = Header.Current;
  Header.Current++;
  if (Header.Current > Header.Records)
    Header.Current = 1;

  DmWrite (MemHandleLock (headhandle), 0, &Header, sizeof (Header));
  MemHandleUnlock (headhandle);

  DmReleaseRecord (db, 0, false);
  DmReleaseRecord (db, last, false);

  DmCloseDatabase (db);
}

// --------------------------------------------------------------------------

static void NextPlayer (void)
{
  Persistant.CurrentPlayer++;
  if (Persistant.CurrentPlayer > (Persistant.NumberOfPlayers - 1))
    Persistant.CurrentPlayer = 0;
}

// --------------------------------------------------------------------------

static void Clear (int where)
{
  RectangleType rect;

  switch (where)
  {
    case clrLL:
      rect.topLeft.x = 0;
      rect.topLeft.y = 80;
      break;

    case clrLR:
      rect.topLeft.x = 80;
      rect.topLeft.y = 80;
      break;

    case clrUL:
      rect.topLeft.x = 0;
      rect.topLeft.y = 0;
      break;
  }

  rect.extent.x  = 80;
  rect.extent.y  = 80;

  WinEraseRectangle (&rect, 0);
}

// --------------------------------------------------------------------------

static void SomeoneSays (int who, char *message)
{
  Char temp[256];
  Char buffer1[64];
  Char buffer2[64];
  Char backup[64];
  int w;
  int OldFont;
  int x;
  int y;
  Boolean checking;

  DrawDBBMP (who, 0, 80);

  if (!Persistant.JustSpun)
    Clear (clrLR);

  Persistant.JustSpun = false;

  OldFont = FntGetFont ();
  FntSetFont (boldFont);

  StrCopy (temp, message);

  y = 159 - 12;
  while (StrLen (temp) > 0)
  {
    buffer1[0] = 0;
    checking = true;
    do {
      x = StrLen (temp);
      while ((x > 0) && (temp[x] != ' '))
        x--;
      StrCopy (backup, buffer1);
      StrCopy (buffer2, &temp[x]);
      StrCat (buffer2, buffer1);
      StrCopy (buffer1, buffer2);
      w = FntCharsWidth (buffer1, StrLen (buffer1));    
      if (w > 75)
      {
        checking = false;
        StrCopy (buffer1, backup);
      } else
        temp[x] = 0;
      if (StrLen (temp) == 0)
        checking = false;
    } while (checking);

    w = FntCharsWidth (buffer1, StrLen (buffer1));    
    x = 80 + (40 - (w / 2));
    WinDrawChars (buffer1, StrLen (buffer1), x, y);
    y -= 12;
  }
      
  FntSetFont (OldFont);
}

// --------------------------------------------------------------------------

static void VannaSays (char *message)
{
  SomeoneSays (dbbVanna, message);
}

// --------------------------------------------------------------------------

static void PatSays (char *message)
{
  SomeoneSays (dbbPat, message);
}

// --------------------------------------------------------------------------

static void DrawBitmap (int handle, int x, int y)
{
  Handle resH;
  BitmapPtr resP;

  resH = (Handle)DmGetResource (bitmapRsc, handle);
  resP = (BitmapPtr)MemHandleLock ((VoidHand)resH);
  WinDrawBitmap (resP, x, y);
  MemPtrUnlock (resP);
  DmReleaseResource ((VoidHand)resH);
}

// --------------------------------------------------------------------------

static void CheckForSolve (void)
{
  int y;
  Char temp1[64];
  Char temp2[64];

  StrCopy (temp1, Persistant.Puzzle[0]);
  for (y=1; y<4; y++)
    StrCat (temp1, Persistant.Puzzle[y]);

  StrCopy (temp2, Persistant.TempPuzzle[0]);
  for (y=1; y<4; y++)
    StrCat (temp2, Persistant.TempPuzzle[y]);

  Persistant.Running = false;

  if (StrCaselessCompare (temp1, temp2) == 0)
  {
    PatSays ("Way to go!  You got it!");
    Persistant.GameState = gstEndOfRound;
  } else {
    Persistant.GameState = gstWaitingForHost;
    for (y=0; y<4; y++)
      StrCopy (Persistant.Puzzle[y], Persistant.TempPuzzle[y]);
    DrawPuzzle ();
    PatSays ("Nope, I'm sorry, that is incorrect.");
    NextPlayer ();
  }
}

// --------------------------------------------------------------------------

static void HandleSolveLetter (UInt letter)
{
  Char temp1[2];

  // Force lower case
  if (letter >= 'A' && letter <= 'Z')
    letter += 'a'-'A';

  // Was it even a letter?
  if (!(letter >= 'a' && letter <= 'z'))
    return;

  // Make an uppercase version
  temp1[0] = letter - ('a'-'A');

  // Make an index into the remaining letters array
  temp1[1] = letter - 'a';

  // Is this letter used?
  if (Persistant.Remaining[(int)temp1[1]] == ' ')
    return;

  // Add letter to puzzle board
  Persistant.Puzzle[Persistant.SolvingY][Persistant.SolvingX] = temp1[0];

  // Get ready to blink next piece
  GetNextUnknown (Persistant.SolvingLetter);

  DrawPuzzle ();

  // No more letters to fill in - check it.
  if (Persistant.SolvingLetter < 0)
    CheckForSolve ();
}

// --------------------------------------------------------------------------

static void ShowLetterBoard (void)
{
  int OldFont;
  int w;
  int x;
  int y;
  int z;
  char c[128];

  OldFont = FntGetFont ();

  Persistant.GameState = gstWaitToDrawButtons;

  Clear (clrLL);
  Clear (clrLR);

  FntSetFont (boldFont);

  StrCopy (c, "Letters Remaining:");
  x = 80 - (FntCharsWidth (c, StrLen (c)) / 2);
  WinDrawChars (c, StrLen (c), x, 85);

  FntSetFont (stdFont);

  c[1] = 0;
  y = 105;
  x = 15;
  for (z=0; z<26; z++)
  {
    c[0] = Persistant.Remaining[z] - ('a'-'A');
    if ((Persistant.Remaining[z] >= 'a') && (Persistant.Remaining[z] <= 'z'))
    {
      w = 5 - (FntCharWidth (c[0]) / 2);
      WinDrawChars (c, 1, x, y);
    }

    x += 15;
    if (x > 135)
    {
      x = 15;
      y += 15;
    }
  }


  FntSetFont (OldFont);
}

// --------------------------------------------------------------------------

static void DrawPuzzle (void)
{
  int OldFont;
  int x;
  int y;
  int w;
  RectangleType rect;

  OldFont = FntGetFont ();
  FntSetFont (stdFont);

  DrawDBBMP (dbbBoard, 0, 0);

  Persistant.ConsonantsLeft = 0;
  Persistant.VowelsLeft = 0;

  rect.extent.x  = 9;
  rect.extent.y  = 13;
  for (x=0; x<14; x++)
    for (y=0; y<4; y++)
    {
      rect.topLeft.x = (x * 10) + 9;
      rect.topLeft.y = (y * 14) + 11;

      if ((Persistant.Puzzle[y][x] > (Char)96) && (Persistant.Puzzle[y][x] != '*') && (Persistant.Puzzle[y][x] != ' '))
      {
        WinDrawRectangle (&rect, 0);
        if (IsVowel (Persistant.Puzzle[y][x]))
          Persistant.VowelsLeft++;
        else
          Persistant.ConsonantsLeft++;
      }

      if ((Persistant.Puzzle[y][x] < (char)97) && (Persistant.Puzzle[y][x] != '*') && (Persistant.Puzzle[y][x] != ' '))
      {
        WinEraseRectangle (&rect, 0);
        w = 5 - (FntCharWidth (Persistant.Puzzle[y][x]) / 2);
        WinDrawChars (&Persistant.Puzzle[y][x], 1, rect.topLeft.x + w, rect.topLeft.y);
      }
    }

  FntSetFont (OldFont);
}

// --------------------------------------------------------------------------

static void GetNextUnknown (int position)
{
  int x;
  int y;
  int p;

  p = 0;
  Persistant.SolvingLetter = -1;
  for (y=0; y<4; y++)
    for (x=0; x<14; x++)
    {
      if (p >= position)
        if ((Persistant.Puzzle[y][x] > (Char)96) && (Persistant.Puzzle[y][x] != '*') && (Persistant.Puzzle[y][x] != ' '))
          if (Persistant.SolvingLetter < 0)
          {
            Persistant.SolvingLetter = p;
            Persistant.SolvingX = x;
            Persistant.SolvingY = y;
          }
      p++;
    }
}

// --------------------------------------------------------------------------

static void DrawScores (int which)
{
  int OldFont;
  int p;
  int y;
  int yoff;
  Char temp1[32];
  Char temp2[32];

  if (which == dscRoundScores)
  {
    Clear (clrLL);
    yoff = 85;
  } else {
    Clear (clrUL);
    yoff = 5;
  }

  OldFont = FntGetFont ();

  FntSetFont (boldFont);

  for (p=0; p<Persistant.NumberOfPlayers; p++)
  {
    StrCopy (temp1, Persistant.PlayerNames[p]);
    StrCat (temp1, ":");
    y = yoff + (24 * p);
    WinDrawChars (temp1, StrLen (temp1), 15, y);
  }

  y = yoff + (24 * Persistant.CurrentPlayer);
  temp1[0] = 4;
  temp1[1] = 0;
  FntSetFont (symbolFont);
  WinDrawChars (temp1, 1, 1, y);

  FntSetFont (stdFont);

  for (p=0; p<Persistant.NumberOfPlayers; p++)
  {
    StrCopy (temp1, "$");
    if (which == dscRoundScores)
      StrIToA (temp2, Persistant.Scores[p]);
    else
      StrIToA (temp2, Persistant.TotalScores[p]);
    StrCat (temp1, temp2);
    y = 11 + yoff + (24 * p);
    WinDrawChars (temp1, StrLen (temp1), 20, y);
  }

  FntSetFont (OldFont);
}

// --------------------------------------------------------------------------

static void AnimationRun (void)
{
  int OldFont;
  int w;
  int x;
  int y;
  Char temp[64];
  Char temp2[64];
  RectangleType rect;

  if (Persistant.WhichAni == aniSolving)
  {
    w = 0;
    for (y=0; y<4; y++)
      for (x=0; x<14; x++)
      {
        if (w == Persistant.SolvingLetter)
        {
          rect.extent.x  = 9;
          rect.extent.y  = 13;
          rect.topLeft.x = (x * 10) + 9;
          rect.topLeft.y = (y * 14) + 11;
          WinInvertRectangle (&rect, 0);
        }
        w++;
      }
  }

  if (Persistant.WhichAni == aniBlinkSpin)
  {
    rect.topLeft.x = 82;
    rect.topLeft.y = 82;
    rect.extent.x  = 76;
    rect.extent.y  = 24;
    WinInvertRectangle (&rect, 0);
    Persistant.LoopCount++;
    if (Persistant.LoopCount > 11)
      if (Persistant.ConsonantsLeft > 0)
        AnimationStart (aniWheel);
      else {
        Persistant.Running = false;
        if (Persistant.VowelsLeft > 0)
          StrCopy (temp, "No consonants.  Buy a vowel or solve.");
        else
          StrCopy (temp, "No consonants or vowels.  Solve.");
        PatSays (temp);
        Persistant.GameState = gstWaitingForHost;
      }
  }

  if (Persistant.WhichAni == aniBlinkBuy)
  {
    rect.topLeft.x = 82;
    rect.topLeft.y = 108;
    rect.extent.x  = 76;
    rect.extent.y  = 24;
    WinInvertRectangle (&rect, 0);
    Persistant.LoopCount++;
    if (Persistant.LoopCount > 11)
    {
      Persistant.Running = false;
      if (Persistant.VowelsLeft > 0)
        BuyVowel ();
      else {
        if (Persistant.ConsonantsLeft > 0)
          StrCopy (temp, "No vowels left.  Spin or solve.");
        else
          StrCopy (temp, "No vowels or consonants.  Solve.");
        PatSays (temp);
        Persistant.GameState = gstWaitingForHost;
      }
    }
  }

  if (Persistant.WhichAni == aniBlinkSolve)
  {
    rect.topLeft.x = 82;
    rect.topLeft.y = 134;
    rect.extent.x  = 76;
    rect.extent.y  = 24;
    WinInvertRectangle (&rect, 0);
    Persistant.LoopCount++;
    if (Persistant.LoopCount > 11)
      if ((Persistant.ConsonantsLeft == 0) && (Persistant.VowelsLeft == 0))
      {
        Persistant.Running = false;
        PatSays ("Yeah, that's it.");
        Persistant.GameState = gstEndOfRound;
      } else {
        Persistant.Running = false;
        StrCopy (temp, "$");
        StrIToA (temp2, Persistant.Scores[Persistant.CurrentPlayer]);
        StrCat (temp, temp2);
        x = FrmCustomAlert (altSolve, temp, NULL, NULL);
        if (x == 0)
        {
          Persistant.GameState = gstSolving;
          AnimationStart (aniSolving);
        } else
          Persistant.GameState = gstWaitingForHost;
      }
  }

  if (Persistant.WhichAni == aniWheel)
  {
    DrawBitmap (bmpWheel1 + Persistant.LastWheelFrame, 84, 83);

    SndPlaySystemSound (sndClick);

    Persistant.LastWheelFrame++;
    if (Persistant.LastWheelFrame > 2)
    {
      Persistant.LastWheelFrame = 0;
      Persistant.Speed++;
    }
    Persistant.LastWheelValue++;
    if (Persistant.LastWheelValue == TotalWheelValues)
    {
      Persistant.LastWheelValue = 0;
    }
    Persistant.LoopCount++;
    if (Persistant.LoopCount > (30 + Persistant.RandomValue))
    {
      OldFont = FntGetFont ();
      FntSetFont (largeFont);

      if (WheelValues[Persistant.LastWheelValue] > 0)
      {
        StrCopy (temp, "$");
        StrIToA (temp2, WheelValues[Persistant.LastWheelValue]);
        StrCat (temp, temp2);
      } else
        StrCopy (temp, SpecialSpaces[-WheelValues[Persistant.LastWheelValue]]);

      w = FntCharsWidth (temp, StrLen (temp));

      rect.topLeft.x = 80 + (40 - (w / 2)) - 4;
      rect.topLeft.y = 108;
      rect.extent.x  = w + 8;
      rect.extent.y  = 22;
      Persistant.WheelValueBox = rect;

      FntSetFont (OldFont);

      AnimationStart (aniSlideValue);
    }
  }

  if (Persistant.WhichAni == aniSlideValue)
  {
    if (Persistant.WheelValueBox.topLeft.y > 80)
    {
      OldFont = FntGetFont ();
      FntSetFont (largeFont);

      if (WheelValues[Persistant.LastWheelValue] > 0)
      {
        StrCopy (temp, "$");
        StrIToA (temp2, WheelValues[Persistant.LastWheelValue]);
        StrCat (temp, temp2);
      } else
        StrCopy (temp, SpecialSpaces[-WheelValues[Persistant.LastWheelValue]]);

      w = FntCharsWidth (temp, StrLen (temp));

      Persistant.WheelValueBox.topLeft.y--;

      Clear (clrLR);

      if (Persistant.WheelValueBox.topLeft.y > 80)
        DrawBitmap (bmpWheel1 + Persistant.LastWheelFrame, 84, 83);

      rect = Persistant.WheelValueBox;
      WinDrawRectangle (&rect, 0);

      rect.topLeft.x += 2;
      rect.topLeft.y += 2;
      rect.extent.x  -= 4;
      rect.extent.y  -= 4;
      WinEraseRectangle (&rect, 0);

      WinDrawChars (temp, StrLen (temp), rect.topLeft.x + 2, rect.topLeft.y + 2);

      FntSetFont (OldFont);
    } else {
      Persistant.Running = false;
      if (WheelValues[Persistant.LastWheelValue] > 0)
      {
        PatSays ("Please pick a letter.");
        Persistant.GameState = gstPickLetter;
      } else {
        if (WheelValues[Persistant.LastWheelValue] == valBankrupt)
        {
          StrCopy (temp, "Oh!  I'm sorry, the bankrupt got you!");
          Persistant.Scores[Persistant.CurrentPlayer] = 0;
        }
        if (WheelValues[Persistant.LastWheelValue] == valLoseATurn)
        {
          StrCopy (temp, "Whoops!  Landed on Lose a Turn - sorry!");
        }
        PatSays (temp);
        NextPlayer ();
        Persistant.GameState = gstWaitingForHost;
      }
    }
  }
}

// --------------------------------------------------------------------------

static void AnimationStart (int which)
{
  Persistant.WhichAni = which;

  if ((Persistant.WhichAni == aniBlinkSpin) ||
      (Persistant.WhichAni == aniBlinkBuy)  ||
      (Persistant.WhichAni == aniBlinkSolve))
  {
    Persistant.Speed = 2;
    Persistant.LoopCount = 0;
    Persistant.Running = true;
  }

  if (Persistant.WhichAni == aniWheel)
  {
    Persistant.RandomValue = (int)(SysRandom (0) * 10 / sysRandomMax);
    Persistant.Speed = 3;
    Persistant.LoopCount = 0;
    Persistant.Running = true;
    Clear (clrLR);
    Persistant.JustSpun = true;
  }

  if (Persistant.WhichAni == aniSlideValue)
  {
    Persistant.Speed = 1;
    Persistant.Running = true;
  }

  if (Persistant.WhichAni == aniSolving)
  {
    PatSays ("Fill in the missing letters to complete the puzzle.");
    Persistant.Speed = 10;
    GetNextUnknown (0);
    Persistant.Running = true;
  }

  AnimationRun ();
}

// --------------------------------------------------------------------------

static void BuyVowel (void)
{
  if (Persistant.Scores[Persistant.CurrentPlayer] > 249)
  {
    if ((Persistant.Remaining['A'-'A'] != ' ') ||
        (Persistant.Remaining['E'-'A'] != ' ') ||
        (Persistant.Remaining['I'-'A'] != ' ') ||
        (Persistant.Remaining['O'-'A'] != ' ') ||
        (Persistant.Remaining['U'-'A'] != ' '))
    {
      PatSays ("Okay, which vowel would you like?");
      Persistant.Scores[Persistant.CurrentPlayer] -= 250;
      Persistant.GameState = gstBuyVowel;
    } else {
      PatSays ("All the vowels are gone.  Spin or Solve.");
      Persistant.GameState = gstWaitingForHost;
    }
  } else {
    PatSays ("You need $250 to buy a vowel.  Spin or Solve.");
    Persistant.GameState = gstWaitingForHost;
  }
}

// --------------------------------------------------------------------------

static int PickLetter (UInt letter)
{
  int found;
  int x;
  int y;
  Char temp1[64];
  Char temp2[16];
  Boolean onboard;
  Boolean used;

  // Force lower case
  if (letter >= 'A' && letter <= 'Z')
    letter += 'a'-'A';

  // Was it even a letter?
  if (!(letter >= 'a' && letter <= 'z'))
    return -1;

  found = 0;
  onboard = false;
  used = false;
  temp1[0] = letter - ('a'-'A');
  temp1[1] = letter - 'a';

  for (x=0; x<14; x++)
    for (y=0; y<4; y++)
    {
      if (Persistant.Puzzle[y][x] == temp1[0])
        onboard = true;
      if (Persistant.Puzzle[y][x] == (char)letter)
      {
        Persistant.Puzzle[y][x] -= 'a'-'A';
        found++;
      }
    }

  if (found > 0)
    DrawPuzzle ();

  if (Persistant.Remaining[(int)temp1[1]] != ' ')
    Persistant.Remaining[(int)temp1[1]] = ' ';
  else
    if (!onboard)
      used = true;

  if (used)
  {
    StrCopy (temp1, "The '");
    temp2[0] = letter;
    temp2[1] = 0;
    StrCat (temp1, temp2);
    StrCat (temp1, "' has already been chosen.");
  } else
    if (onboard)
    {
      StrCopy (temp1, "The '");
      temp2[0] = letter;
      temp2[1] = 0;
      StrCat (temp1, temp2);
      StrCat (temp1, "' is already on the board.");
    } else
      if (found == 0)
      {
        StrCopy (temp1, "Sorry, there is no '");
        temp2[0] = letter;
        temp2[1] = 0;
        StrCat (temp1, temp2);
        StrCat (temp1, "'.");
      } else {
        StrCopy (temp1, "There ");
        if (found == 1)
          StrCat (temp1, "is ");
        else
          StrCat (temp1, "are ");
        StrIToA (temp2, found);
        StrCat (temp1, temp2);
        StrCat (temp1, " '");
        temp2[0] = letter;
        temp2[1] = 0;
        StrCat (temp1, temp2);
        StrCat (temp1, "'");
        if (found > 1)
          StrCat (temp1, "s");
        StrCat (temp1, ".");
      }

  VannaSays (temp1);

  Persistant.GameState = gstWaitingForHost;

  return found;
}

// --------------------------------------------------------------------------

static int IsVowel (UInt letter)
{
  // Force lower case
  if (letter >= 'A' && letter <= 'Z')
    letter += 'a'-'A';

  // Was it even a letter?
  if (!(letter >= 'a' && letter <= 'z'))
    return -1;

  if ((letter == 'a') ||
      (letter == 'e') ||
      (letter == 'i') ||
      (letter == 'o') ||
      (letter == 'u'))
    return 1;
  else
    return 0;
}

// --------------------------------------------------------------------------

static Boolean ApplicationHandleEvent (EventPtr event)
{
  int r;
  Char temp[256];
  Char temp2[256];

  if (event->eType == frmLoadEvent)
  {
    CurrentForm = FrmInitForm (event->data.frmLoad.formID);
    FrmSetActiveForm (CurrentForm);
    return true;
  }

  if (event->eType == frmOpenEvent)
  {
    FrmDrawForm (FrmGetActiveForm ());
    switch (Persistant.GameState)
    {
      case gstAtTitle:
        DrawDBBMP (dbbTitle, 0, 0);
        break;

      case gstEndOfGame:
        DrawScores (dscTotalScores);
        DrawDBBMP (dbbVanna, 80, 0);
        RestoreScreen ();
        break;

      default:
        DrawPuzzle ();
        RestoreScreen ();
        break;
    }
    return true;
  }

  if (Persistant.GameState == gstEndOfRound)
    if (event->eType == penDownEvent)
    {
      StrCopy (temp, "The winner for this round is ");
      StrCat (temp, Persistant.PlayerNames[Persistant.CurrentPlayer]);
      StrCat (temp, ".");
      PatSays (temp);
      Persistant.GameState = gstPatSaysScore;
      return true;
    }

  if (Persistant.GameState == gstPatSaysScore)
    if (event->eType == penDownEvent)
    {      
      if (Persistant.Scores[Persistant.CurrentPlayer] < 500)
      {
        r = 1;
        Persistant.Scores[Persistant.CurrentPlayer] = 500;
      } else
        r = 0;
      StrCopy (temp, "You won $");
      StrIToA (temp2, Persistant.Scores[Persistant.CurrentPlayer]);
      StrCat (temp, temp2);
      if (r == 1)
        StrCat (temp, " (the house minimum)");
      StrCat (temp, ".  Your total so far is $");
      r = Persistant.CurrentPlayer;
      NextRound ();
      StrIToA (temp2, Persistant.TotalScores[r]);
      StrCat (temp, temp2);
      StrCat (temp, ".");
      if (Persistant.GameState != gstEndOfGame)
      {
        PatSays (temp);
        Persistant.GameState = gstPatSaysRound;
      }
      return true;
    }

  if (Persistant.GameState == gstAtTitle)
    if (event->eType == penDownEvent)
    {
      DrawPuzzle ();
      PatSays ("Hello players!  Let's play some Wheel!");
      Persistant.GameState = gstPatSaysRound;
      return true;
    }

  if (Persistant.GameState == gstPatSaysRound)
    if (event->eType == penDownEvent)
    {
      StrCopy (temp, "Here's our ");
      switch (Persistant.Round)
      {
        case 1:
          StrCat (temp, "first");
          WheelValues[ValueSlot] = 1000;
          break;

        case 2:
          StrCat (temp, "second");
          WheelValues[ValueSlot] = 2500;
          break;

        case 3:
          StrCat (temp, "third");
          WheelValues[ValueSlot] = 3500;
          break;

        case 4:
          StrCat (temp, "fourth and last");
          WheelValues[ValueSlot] = 5000;
          break;
      }
      StrCat (temp, " puzzle.  '");
      StrCat (temp, Persistant.Puzzle[4]);
      StrCat (temp, "' is the catagory.");
      PatSays (temp);
      Persistant.GameState = gstPatSaysPlayer;
      DrawPuzzle ();
      return true;
    }

  if (Persistant.GameState == gstPatSaysPlayer)
    if (event->eType == penDownEvent)
    {
      StrCopy (temp, Persistant.PlayerNames[Persistant.CurrentPlayer]);
      StrCat (temp, ", you will start this round.  There's a $");
      StrIToA (temp2, WheelValues[ValueSlot]);
      StrCat (temp, temp2);
      StrCat (temp, " space on the wheel.");
      PatSays (temp);
      Persistant.GameState = gstWaitToDrawButtons;
      return true;
    }

  if (Persistant.GameState == gstWaitToDrawButtons)
    if (event->eType == penDownEvent)
    {
      Persistant.GameState = gstSpinBuySolve;
      DrawScores (dscRoundScores);
      DrawBitmap (bmpButtons, 80, 80);
      return true;
    }

  if (Persistant.GameState == gstSpinBuySolve)
    if (event->eType == penDownEvent)
    {
      // Spin
      if ((event->screenX > 81) && (event->screenX < 158) && (event->screenY > 81) && (event->screenY < 81 + 24))
      {
        AnimationStart (aniBlinkSpin);
        Persistant.GameState = gstDoingNothing;
        return true;
      }
      // Buy a vowel
      if ((event->screenX > 81) && (event->screenX < 158) && (event->screenY > 108) && (event->screenY < 108 + 24))
      {
        AnimationStart (aniBlinkBuy);
        Persistant.GameState = gstDoingNothing;
        return true;
      }
      // Solve Puzzle
      if ((event->screenX > 81) && (event->screenX < 158) && (event->screenY > 134) && (event->screenY < 134 + 24))
      {
        for (r=0; r<5; r++)
          StrCopy (Persistant.TempPuzzle[r], Persistant.Puzzle[r]);
        AnimationStart (aniBlinkSolve);
        Persistant.GameState = gstDoingNothing;
        return true;
      }
    }

  if (Persistant.GameState == gstPickLetter)
    if (event->eType == keyDownEvent)
    {
      r = IsVowel (event->data.keyDown.chr);
      if (r == 0)
      {
        r = PickLetter (event->data.keyDown.chr);
        if (r > 0)
          Persistant.Scores[Persistant.CurrentPlayer] += (WheelValues[Persistant.LastWheelValue] * r);
        if (r == 0)
          NextPlayer ();
        return true;
      }
      if (r == 1)
      {
        StrCopy (temp, "I'm sorry, but '");
        temp[254] = event->data.keyDown.chr;
        if (temp[254] >= 'A' && temp[254] <= 'Z')
          temp[254] += 'a'-'A';
        temp[255] = 0;
        StrCat (temp, &temp[254]);
        StrCat (temp, "' is a vowel.");
        PatSays (temp);
        NextPlayer ();
        Persistant.GameState = gstWaitToDrawButtons;
        return true;
      }
      return true;
    }

  if (Persistant.GameState == gstBuyVowel)
    if (event->eType == keyDownEvent)
    {
      r = IsVowel (event->data.keyDown.chr);
      if (r == 1)
      {
        r = PickLetter (event->data.keyDown.chr);
        if (r == 0)
          NextPlayer ();
        return true;
      }
      if (r == 0)
      {
        StrCopy (temp, "I'm sorry, but '");
        temp[254] = event->data.keyDown.chr;
        if (temp[254] >= 'A' && temp[254] <= 'Z')
          temp[254] += 'a'-'A';
        temp[255] = 0;
        StrCat (temp, &temp[254]);
        StrCat (temp, "' is a consonant.");
        PatSays (temp);
        NextPlayer ();
        Persistant.GameState = gstWaitToDrawButtons;
        return true;
      }
      return true;
    }

  if (Persistant.GameState == gstSolving)
    if (event->eType == keyDownEvent)
    {
      HandleSolveLetter (event->data.keyDown.chr);
      return true;
    }

  if (Persistant.GameState == gstWaitingForHost)
    if (event->eType == penDownEvent)
    {
      DrawScores (dscRoundScores);
      Clear (clrLR);
      DrawBitmap (bmpButtons, 80, 80);
      Persistant.GameState = gstSpinBuySolve;
      return true;
    }

  if (event->eType == menuEvent)
  {
    if (event->data.menu.itemID == mnuAbout)
    {
      FrmAlert (altAbout);
      return true;
    }
    if (event->data.menu.itemID == mnuNewGame)
    {
      if (FrmAlert (altNewGame) == 0)
      {
        DrawDBBMP (dbbTitle, 0, 0);
        NewGame ();
      }
      return true;
    }
    if (event->data.menu.itemID == mnuShowLetters)
    {
      if (Persistant.GameState == gstSpinBuySolve)
        ShowLetterBoard ();
      else
        FrmAlert (altNoBoardNow);
      return true;
    }
  }

  if (Persistant.Running && event->eType == nilEvent)
  {
    AnimationRun ();
    return true;
  }

  return false;
}

// --------------------------------------------------------------------------

static void StartApplication (void)
{
  if (!PrefGetAppPreferences (AppCreator, AppVersion, &Persistant, sizeof (Persistant)))
    NewGame ();

  FrmGotoForm (frmMain);
}

// --------------------------------------------------------------------------

static void NextRound (void)
{
  int x;
  char temp[256];
  RectangleType rect;

  Persistant.Round++;

  for (x=0; x<3; x++)
  {
    Persistant.TotalScores[x] += Persistant.Scores[x];
    Persistant.Scores[x] = 0;
  }

  if (Persistant.Round < 4)
    Persistant.CurrentPlayer = Persistant.Round - 1;
  else
    Persistant.CurrentPlayer = 0;

  Persistant.JustSpun = false;
  Persistant.Running = false; 

  StrCopy (Persistant.Remaining, "abcdefghijklmnopqrstuvwxyz");

  if (Persistant.Round > 4)
  {
    // Find Winner
    Persistant.CurrentPlayer = 0;
    if (Persistant.NumberOfPlayers > 1)
      for (x=1; x<Persistant.NumberOfPlayers; x++)
        if (Persistant.TotalScores[x] > Persistant.TotalScores[Persistant.CurrentPlayer])
          Persistant.CurrentPlayer = x;

    rect.topLeft.x = 0;
    rect.topLeft.y = 0;
    rect.extent.x  = 160;
    rect.extent.y  = 160;
    WinEraseRectangle (&rect, 0);

    DrawScores (dscTotalScores);
    DrawDBBMP (dbbVanna, 80, 0);

    StrCopy (temp, "Congrat- ulations ");
    StrCat (temp, Persistant.PlayerNames[Persistant.CurrentPlayer]);
    StrCat (temp, "!  You're our big winner!");
    PatSays (temp);

    Persistant.GameState = gstEndOfGame;
  } else {
    Persistant.GameState = gstAtTitle;
    LoadPuzzle ();
  }
}

// --------------------------------------------------------------------------

static void NewGame (void)
{
  int x;

  for (x=0; x<3; x++)
    Persistant.TotalScores[x] = 0;

  Persistant.Round = 0;
  Persistant.CurrentPlayer = 0;
  Persistant.NumberOfPlayers = 3;

  StrCopy (Persistant.PlayerNames[0], "Player 1");
  StrCopy (Persistant.PlayerNames[1], "Player 2");
  StrCopy (Persistant.PlayerNames[2], "Player 3");

  NextRound ();
}

// --------------------------------------------------------------------------

static void StopApplication (void)
{
  WinPtr Window;  
  int Length;
  int Offset;

  Length = 1600; // 80 * 20 (20 = 160/8)
  Offset = Length;
  Window = WinGetWindowPointer (FrmGetWindowHandle (FrmGetActiveForm ()));
  MemMove (Persistant.WindowBuffer, Window->displayAddr + Offset, Length);

  PrefSetAppPreferences (AppCreator, AppVersion, &Persistant, sizeof (Persistant));
}

// --------------------------------------------------------------------------

static void RestoreScreen (void)
{
  WinPtr Window;  
  int Length;
  int Offset;

  Length = 1600; // 80 * 20 (20 = 160/8)
  Offset = Length;
  Window = WinGetWindowPointer (FrmGetWindowHandle (FrmGetActiveForm ()));
  MemMove (Window->displayAddr + Offset, Persistant.WindowBuffer, Length);
}

// --------------------------------------------------------------------------

static void EventLoop (void)
{
  Boolean handled;
  EventType event;
  Word error;

  do {

    EvtGetEvent (&event, (Persistant.Running ? Persistant.Speed : -1));

    handled = SysHandleEvent (&event);

    if (!handled)
      if (Persistant.GameState != gstAtTitle)
        if (!Persistant.Running)
          handled = MenuHandleEvent (NULL, &event, &error);

    if (!handled)
      handled = ApplicationHandleEvent (&event);

    if (!handled)
      FrmHandleEvent (FrmGetActiveForm (), &event);

  } while (event.eType != appStopEvent);
}

// --------------------------------------------------------------------------

DWord PilotMain (Word cmd, Ptr cmdPBP, Word launchFlags)
{
  if (cmd == sysAppLaunchCmdNormalLaunch)
  {
    StartApplication ();
    EventLoop ();
    StopApplication ();
  }

  return 0;
}

// --------------------------------------------------------------------------

