/*********************************************************************
 * iLarn - Larn adapted for the PalmPilot.                           *
 * Copyright (C) 2000 Bridget Spitznagel                             *
 * iLarn is derived from Ularn by Phil Cordier                       *
 * which in turn was based on Larn by Noah Morgan                    *
 *                                                                   *
 * This program is free software; you can redistribute it and/or     *
 * modify it under the terms of the GNU General Public License       *
 * as published by the Free Software Foundation; either version 2    *
 * of the License, or (at your option) any later version.            *
 *                                                                   *
 * This program is distributed in the hope that it will be useful,   *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
 * GNU General Public License for more details.                      *
 *                                                                   *
 * You should have received a copy of the GNU General Public License *
 * along with this program; if not, write to                         *
 * The Free Software Foundation, Inc.,                               *
 * 59 Temple Place - Suite 330,                                      *
 * Boston, MA  02111-1307, USA.                                      *
 *********************************************************************/
#include <Pilot.h>
//#include <Globals.h>
#include "fix_db.h"

/*****************************************************************************
 *                                                                           *
 *                      canned levels ...                                    *
 *                                                                           *
 *****************************************************************************/

// there are 20 levels,  0-19
#define CAN_LEV 20
#define CAN_ROW 15
#define CAN_COL 33

// canned could be more compressed than this, but octal seems painful

Char canned[CAN_LEV*CAN_ROW][CAN_COL] = {
 {0, 0, 0, 0, 0, 2, 0, 64, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 2, 0, 0, 2, 0, 0, 2, 0, 4, 2},
 {0, 0, 0, 0, 0, 5, 0, 0, 0, 0, 0, 5, 0, 64, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 0, 0, 0, 5, 4, 0, 2},
 {34, 82, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 32, 2, 0, 0, 2, 0, 0, 2, 34, 82, 34},
 {0, 0, 0, 0, 0, 50, 0, 0, 0, 0, 0, 0, 0, 0, 2, 64, 0, 2, 0, 0, 0, 32, 34, 34, 34, 34, 34, 34, 34, 34, 4, 0, 66},
 {0, 0, 2, 34, 34, 34, 2, 34, 34, 34, 32, 34, 34, 34, 34, 34, 34, 5, 2, 34, 32, 32, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 2},
 {4, 68, 2, 66, 0, 2, 2, 0, 32, 64, 32, 32, 0, 0, 0, 0, 0, 2, 2, 34, 32, 32, 34, 34, 34, 34, 34, 34, 2, 2, 34, 82, 34},
 {0, 0, 2, 66, 2, 2, 2, 32, 32, 0, 32, 32, 34, 34, 34, 34, 34, 34, 2, 34, 32, 32, 35, 0, 32, 0, 0, 2, 2, 2, 64, 0, 2},
 {0, 64, 2, 2, 2, 2, 2, 32, 35, 0, 32, 32, 32, 0, 3, 0, 0, 5, 2, 34, 32, 32, 32, 64, 80, 0, 2, 2, 66, 2, 4, 68, 2},
 {0, 0, 2, 2, 66, 2, 0, 0, 32, 0, 32, 32, 32, 0, 64, 64, 0, 2, 2, 0, 0, 32, 32, 0, 32, 0, 2, 2, 50, 2, 6, 65, 2},
 {34, 82, 34, 2, 34, 2, 34, 34, 34, 82, 32, 32, 34, 34, 34, 34, 34, 34, 2, 34, 34, 32, 34, 34, 34, 34, 34, 2, 34, 2, 34, 34, 34},
 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 32, 0, 0, 0, 0, 0, 4, 32, 0, 0, 0, 2, 0, 0, 0, 68, 68, 64, 0, 0, 0, 68, 66},
 {34, 82, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 82, 34},
 {0, 64, 2, 68, 68, 66, 0, 0, 2, 0, 0, 2, 0, 0, 2, 0, 0, 50, 0, 0, 2, 0, 0, 2, 0, 0, 2, 0, 0, 2, 0, 0, 2},
 {0, 0, 0, 68, 68, 64, 0, 4, 0, 0, 0, 5, 0, 0, 0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 0, 0, 5, 0, 0, 5, 64, 0, 2},
 {0, 0, 2, 68, 68, 66, 0, 0, 2, 0, 0, 2, 0, 0, 2, 0, 0, 2, 0, 0, 2, 0, 0, 66, 0, 0, 2, 0, 0, 2, 0, 0, 2},
 {2, 0, 0, 0, 0, 0, 0, 4, 3, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 0, 0, 0, 0, 0, 0, 64, 0, 2},
 {2, 82, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 32, 66},
 {2, 2, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 48, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 32, 34},
 {66, 66, 82, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 0, 36, 2},
 {2, 2, 2, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 64, 0, 0, 0, 0, 0, 2, 0, 34, 50},
 {2, 50, 2, 82, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 2, 0, 32, 66},
 {2, 2, 2, 2, 4, 0, 0, 2, 3, 0, 0, 2, 0, 0, 0, 2, 0, 64, 0, 50, 0, 64, 0, 2, 68, 68, 68, 66, 18, 2, 0, 32, 34},
 {2, 2, 2, 2, 2, 34, 2, 34, 2, 34, 2, 34, 66, 34, 2, 34, 2, 34, 2, 34, 2, 34, 2, 34, 66, 34, 66, 34, 98, 2, 0, 36, 2},
 {66, 2, 2, 4, 0, 2, 4, 0, 64, 2, 0, 0, 64, 2, 0, 0, 48, 2, 4, 4, 0, 2, 0, 4, 68, 66, 68, 68, 66, 2, 0, 34, 2},
 {2, 66, 2, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 2, 0, 32, 66},
 {2, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 68, 0, 0, 0, 0, 0, 0, 64, 0, 0, 0, 0, 0, 2, 64, 35, 34},
 {2, 2, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 0, 36, 2},
 {2, 0, 0, 0, 0, 68, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 0, 0, 0, 0, 64, 0, 0, 0, 64, 64, 34, 2},
 {50, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 32, 2},
 {64, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 2},
 {34, 0, 2, 0, 0, 34, 0, 36, 68, 32, 0, 0, 0, 0, 0, 64, 4, 4, 0, 64, 0, 0, 32, 2, 0, 32, 2, 32, 0, 0, 0, 66, 34},
 {2, 32, 0, 0, 32, 2, 32, 0, 32, 0, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 32, 32, 2, 64, 0, 34, 0, 34, 34, 37, 37, 2},
 {5, 34, 34, 34, 34, 0, 34, 0, 34, 34, 32, 0, 32, 0, 5, 5, 0, 0, 0, 32, 0, 32, 32, 2, 0, 2, 32, 2, 32, 52, 66, 32, 2},
 {2, 2, 36, 67, 2, 0, 2, 32, 0, 0, 32, 0, 32, 0, 34, 2, 32, 4, 0, 32, 32, 32, 32, 2, 0, 34, 32, 2, 34, 0, 34, 32, 2},
 {2, 0, 34, 52, 66, 0, 0, 34, 4, 0, 34, 0, 32, 2, 35, 64, 34, 4, 0, 32, 32, 32, 32, 0, 2, 32, 32, 2, 52, 50, 32, 32, 2},
 {2, 0, 2, 32, 50, 34, 34, 2, 32, 0, 32, 4, 0, 34, 4, 3, 2, 32, 0, 32, 32, 32, 32, 0, 34, 0, 32, 2, 68, 34, 0, 32, 2},
 {66, 34, 32, 34, 82, 35, 66, 0, 34, 0, 32, 4, 2, 32, 68, 68, 67, 34, 0, 32, 32, 32, 36, 66, 32, 3, 34, 34, 2, 32, 0, 32, 2},
 {2, 0, 32, 85, 2, 35, 69, 0, 2, 32, 32, 68, 34, 34, 34, 82, 34, 34, 32, 32, 32, 32, 36, 85, 0, 34, 32, 5, 34, 0, 34, 32, 2},
 {2, 66, 32, 34, 2, 34, 34, 0, 34, 2, 34, 0, 0, 32, 2, 82, 0, 2, 0, 0, 32, 0, 36, 66, 32, 0, 32, 0, 2, 36, 68, 32, 2},
 {2, 66, 2, 36, 66, 32, 0, 2, 32, 2, 0, 0, 32, 0, 34, 66, 32, 0, 2, 34, 34, 34, 32, 0, 34, 0, 34, 0, 34, 34, 68, 32, 2},
 {50, 64, 34, 68, 66, 32, 0, 34, 64, 2, 34, 34, 32, 34, 32, 67, 34, 0, 0, 32, 3, 0, 32, 0, 2, 32, 2, 32, 0, 2, 32, 32, 2},
 {2, 2, 36, 66, 66, 32, 2, 36, 2, 34, 0, 32, 32, 34, 67, 64, 66, 32, 0, 32, 0, 34, 34, 34, 0, 34, 0, 34, 34, 0, 34, 84, 2},
 {2, 34, 2, 34, 34, 0, 34, 0, 0, 0, 0, 4, 34, 33, 64, 64, 70, 34, 64, 0, 34, 0, 32, 0, 0, 2, 32, 2, 32, 0, 50, 32, 2},
 {2, 32, 34, 0, 0, 2, 34, 34, 34, 34, 34, 32, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 32, 2, 34, 0, 34, 0, 34, 34, 0, 34, 2},
 {34, 0, 0, 2, 0, 5, 0, 0, 64, 64, 0, 4, 0, 0, 0, 48, 0, 0, 0, 0, 0, 0, 68, 0, 0, 0, 2, 36, 68, 68, 0, 34, 34},
 {64, 2, 0, 2, 0, 0, 0, 32, 0, 64, 0, 0, 0, 0, 0, 2, 0, 0, 2, 0, 32, 0, 0, 0, 34, 0, 0, 34, 34, 0, 0, 68, 2},
 {2, 2, 2, 2, 34, 34, 32, 32, 2, 0, 0, 64, 0, 0, 0, 64, 64, 2, 0, 0, 0, 64, 32, 0, 0, 4, 32, 4, 2, 36, 0, 68, 2},
 {2, 64, 2, 0, 64, 0, 0, 64, 2, 0, 32, 64, 32, 0, 0, 37, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 2, 82, 32, 3, 2},
 {2, 34, 34, 34, 34, 34, 34, 34, 34, 0, 0, 32, 0, 0, 2, 36, 34, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 68, 34, 0, 2},
 {2, 0, 0, 4, 32, 0, 64, 0, 2, 0, 2, 34, 0, 0, 34, 64, 66, 32, 2, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 0, 2, 32, 2},
 {2, 0, 48, 4, 80, 0, 64, 48, 2, 0, 32, 48, 32, 2, 32, 2, 0, 34, 0, 2, 64, 2, 0, 2, 0, 2, 0, 2, 0, 2, 32, 34, 2},
 {2, 34, 34, 34, 32, 4, 64, 0, 2, 0, 0, 0, 0, 34, 0, 34, 32, 2, 32, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 2, 32, 2, 34},
 {5, 64, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 2, 32, 2, 33, 34, 0, 34, 0, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 4, 64, 34},
 {2, 34, 34, 34, 34, 34, 34, 34, 34, 4, 0, 0, 34, 0, 34, 70, 66, 32, 2, 34, 34, 2, 2, 2, 2, 2, 2, 2, 34, 34, 4, 68, 2},
 {5, 64, 0, 0, 32, 0, 0, 4, 2, 0, 0, 2, 36, 2, 32, 66, 64, 34, 0, 34, 4, 66, 2, 2, 66, 2, 2, 66, 0, 0, 64, 52, 2},
 {2, 4, 48, 64, 36, 0, 0, 48, 2, 0, 0, 34, 0, 34, 4, 66, 68, 2, 32, 2, 32, 2, 2, 2, 50, 2, 2, 2, 0, 0, 34, 34, 34},
 {2, 0, 0, 0, 84, 0, 0, 0, 2, 0, 2, 32, 2, 32, 2, 34, 34, 48, 34, 0, 34, 0, 2, 2, 2, 2, 2, 2, 0, 36, 32, 0, 2},
 {2, 34, 34, 34, 34, 34, 34, 37, 34, 0, 34, 0, 34, 4, 64, 68, 64, 68, 2, 32, 2, 34, 34, 2, 66, 2, 2, 2, 32, 32, 0, 48, 34},
 {2, 64, 0, 0, 64, 0, 0, 0, 2, 2, 32, 2, 34, 34, 34, 37, 34, 34, 34, 34, 0, 34, 4, 2, 2, 2, 2, 2, 0, 32, 32, 4, 2},
 {2, 0, 0, 3, 0, 0, 0, 4, 2, 34, 0, 0, 3, 0, 0, 64, 64, 0, 0, 0, 0, 2, 36, 0, 2, 0, 2, 0, 2, 34, 32, 32, 2},
 {2, 4, 48, 64, 36, 0, 0, 48, 2, 0, 0, 34, 0, 34, 4, 66, 64, 2, 32, 2, 32, 2, 2, 2, 50, 2, 2, 2, 0, 0, 34, 34, 34},
 {2, 0, 0, 0, 84, 0, 0, 0, 2, 0, 2, 32, 2, 32, 2, 34, 34, 48, 34, 0, 34, 4, 0, 2, 2, 2, 2, 2, 0, 36, 32, 0, 2},
 {2, 34, 34, 34, 34, 34, 34, 37, 34, 0, 34, 0, 34, 4, 0, 0, 0, 68, 2, 32, 2, 34, 34, 2, 66, 2, 2, 2, 32, 32, 0, 48, 34},
 {64, 2, 0, 2, 0, 0, 0, 32, 0, 64, 0, 0, 0, 0, 0, 2, 34, 34, 2, 0, 32, 0, 0, 0, 2, 0, 0, 34, 34, 0, 0, 68, 2},
 {2, 82, 2, 2, 34, 34, 32, 32, 2, 0, 0, 64, 0, 0, 0, 64, 64, 2, 0, 0, 0, 64, 32, 0, 0, 4, 32, 4, 2, 36, 0, 68, 2},
 {2, 64, 2, 0, 64, 0, 0, 64, 2, 0, 32, 64, 32, 0, 0, 37, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 0, 82, 32, 3, 2},
 {2, 34, 34, 34, 34, 34, 34, 34, 34, 4, 0, 0, 34, 0, 34, 64, 66, 32, 2, 34, 34, 2, 2, 2, 2, 2, 2, 2, 34, 34, 4, 68, 2},
 {5, 64, 0, 0, 32, 0, 0, 4, 2, 0, 0, 2, 36, 2, 32, 66, 64, 34, 0, 34, 4, 66, 2, 2, 66, 2, 2, 66, 0, 0, 64, 52, 2},
 {2, 64, 0, 0, 64, 0, 0, 0, 2, 2, 32, 2, 34, 34, 34, 37, 34, 34, 34, 34, 0, 34, 4, 2, 2, 2, 2, 2, 0, 32, 32, 4, 2},
 {2, 34, 34, 34, 34, 34, 34, 34, 34, 0, 0, 32, 0, 0, 2, 36, 34, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 68, 34, 0, 2},
 {2, 0, 0, 4, 32, 0, 64, 0, 2, 0, 34, 34, 0, 0, 34, 64, 66, 37, 82, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 0, 2, 32, 2},
 {2, 0, 48, 4, 80, 0, 64, 48, 2, 0, 32, 50, 32, 2, 32, 2, 0, 34, 0, 2, 64, 2, 0, 2, 0, 2, 64, 2, 0, 2, 32, 34, 2},
 {2, 34, 34, 34, 32, 4, 64, 0, 2, 0, 32, 2, 0, 34, 4, 34, 32, 66, 32, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 2, 32, 2, 34},
 {5, 64, 0, 0, 0, 0, 0, 0, 2, 0, 0, 2, 2, 32, 2, 33, 34, 0, 34, 0, 2, 0, 2, 2, 2, 2, 2, 2, 0, 0, 4, 64, 34},
 {2, 0, 0, 3, 0, 0, 0, 4, 2, 34, 0, 0, 50, 32, 64, 70, 64, 64, 34, 0, 0, 2, 34, 64, 2, 0, 2, 0, 2, 34, 32, 32, 2},
 {68, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 5, 0, 32, 0, 0, 0, 0, 0, 0, 0, 4, 0, 32, 0, 0, 0, 0, 0, 0, 2, 48, 2},
 {34, 34, 34, 34, 34, 34, 2, 34, 34, 34, 34, 2, 2, 32, 34, 32, 34, 34, 32, 34, 2, 34, 32, 34, 34, 34, 2, 34, 34, 34, 2, 34, 2},
 {66, 18, 98, 2, 0, 2, 0, 0, 0, 4, 50, 2, 2, 48, 0, 32, 0, 2, 0, 2, 2, 3, 32, 32, 0, 0, 2, 0, 0, 0, 0, 0, 2},
 {2, 2, 66, 0, 64, 2, 0, 2, 34, 34, 34, 2, 2, 0, 0, 32, 0, 2, 0, 2, 2, 0, 32, 32, 0, 0, 2, 0, 0, 0, 2, 34, 34},
 {2, 4, 66, 2, 34, 34, 0, 2, 0, 0, 2, 2, 2, 34, 32, 32, 0, 2, 32, 34, 2, 32, 32, 34, 34, 34, 2, 34, 34, 34, 2, 0, 2},
 {3, 4, 69, 2, 0, 5, 0, 2, 0, 4, 5, 2, 2, 0, 32, 36, 34, 34, 32, 34, 2, 32, 32, 36, 2, 66, 2, 68, 32, 2, 2, 34, 2},
 {34, 34, 34, 2, 34, 34, 34, 2, 34, 0, 2, 2, 2, 0, 0, 32, 32, 0, 0, 2, 5, 0, 32, 80, 5, 0, 2, 68, 80, 2, 0, 0, 2},
 {48, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 2, 2, 34, 32, 32, 34, 34, 34, 2, 2, 32, 32, 36, 2, 0, 2, 68, 32, 0, 2, 34, 34},
 {34, 2, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 48, 32, 32, 32, 0, 2, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 32, 2, 0, 2},
 {0, 4, 68, 0, 0, 0, 2, 0, 4, 2, 0, 0, 2, 68, 32, 34, 32, 0, 2, 0, 3, 4, 64, 4, 0, 0, 0, 0, 0, 36, 2, 34, 2},
 {2, 34, 34, 34, 32, 0, 2, 0, 0, 2, 0, 0, 2, 0, 32, 34, 32, 0, 2, 34, 34, 34, 34, 34, 34, 34, 34, 34, 32, 32, 2, 0, 2},
 {0, 0, 0, 0, 32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 34, 32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2},
 {34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 2},
 {48, 0, 0, 32, 0, 0, 0, 0, 0, 0, 32, 0, 5, 0, 0, 34, 32, 0, 2, 0, 0, 0, 0, 0, 32, 0, 0, 0, 0, 0, 2, 0, 2},
 {0, 64, 0, 0, 0, 0, 32, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 32, 5, 0, 2},
 {0, 68, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2},
 {2, 34, 34, 34, 34, 34, 34, 32, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 2, 2},
 {0, 0, 0, 0, 2, 0, 0, 32, 32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 4, 64, 0, 2, 2, 2},
 {0, 0, 0, 0, 2, 82, 32, 32, 32, 34, 34, 34, 34, 34, 34, 82, 34, 34, 34, 34, 34, 34, 34, 34, 2, 34, 34, 34, 34, 34, 2, 2, 2},
 {34, 34, 34, 34, 2, 48, 32, 32, 32, 35, 0, 0, 80, 0, 2, 2, 6, 0, 32, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 2, 2, 2, 2},
 {0, 0, 0, 2, 2, 0, 32, 32, 32, 32, 0, 34, 32, 32, 2, 2, 0, 0, 80, 0, 48, 0, 0, 2, 2, 2, 34, 34, 34, 2, 2, 2, 2},
 {0, 68, 68, 2, 2, 34, 32, 32, 32, 32, 0, 33, 32, 32, 2, 2, 34, 34, 32, 0, 0, 4, 64, 2, 66, 2, 0, 0, 2, 2, 2, 2, 2},
 {0, 68, 68, 2, 0, 0, 0, 32, 32, 32, 34, 32, 32, 32, 2, 2, 0, 0, 34, 34, 34, 34, 37, 34, 34, 2, 2, 34, 2, 2, 2, 2, 2},
 {0, 68, 68, 2, 34, 34, 34, 32, 32, 32, 80, 0, 32, 35, 2, 2, 68, 0, 32, 0, 0, 0, 68, 66, 66, 2, 2, 66, 2, 2, 2, 2, 2},
 {0, 0, 0, 2, 0, 0, 0, 0, 32, 34, 32, 0, 32, 34, 34, 2, 67, 0, 80, 0, 48, 0, 0, 2, 66, 2, 66, 66, 2, 2, 2, 2, 2},
 {34, 34, 85, 34, 2, 34, 34, 34, 32, 32, 0, 0, 32, 0, 5, 5, 64, 0, 32, 0, 0, 0, 0, 2, 2, 2, 68, 66, 2, 2, 2, 2, 2},
 {2, 0, 4, 66, 2, 0, 0, 32, 32, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 2, 2, 34, 34, 2, 2, 2, 2, 2},
 {4, 68, 68, 66, 2, 0, 32, 0, 32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 2, 2, 2, 2, 2},
 {2, 34, 34, 34, 2, 34, 34, 32, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 2, 34, 34, 34, 34, 2, 2, 34, 2},
 {0, 0, 0, 0, 0, 0, 0, 4, 84, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 2},
 {0, 0, 0, 0, 0, 0, 0, 0, 0, 34, 0, 0, 0, 0, 34, 0, 0, 0, 0, 0, 0, 34, 32, 0, 0, 0, 0, 0, 0, 0, 2, 32, 2},
 {0, 34, 34, 32, 0, 0, 0, 0, 0, 2, 32, 68, 48, 2, 32, 0, 0, 0, 0, 0, 2, 36, 34, 0, 0, 0, 0, 0, 0, 0, 34, 0, 2},
 {0, 32, 16, 34, 0, 0, 0, 0, 0, 0, 34, 4, 0, 34, 0, 0, 0, 0, 0, 0, 34, 0, 66, 32, 0, 4, 64, 0, 0, 2, 32, 0, 2},
 {0, 36, 68, 66, 34, 34, 32, 0, 0, 0, 2, 32, 2, 32, 0, 0, 0, 0, 0, 2, 32, 4, 0, 34, 0, 0, 0, 0, 0, 34, 0, 0, 2},
 {0, 32, 48, 2, 0, 0, 34, 0, 0, 0, 0, 34, 82, 0, 0, 0, 0, 0, 0, 34, 0, 64, 0, 2, 32, 0, 0, 0, 2, 32, 0, 0, 2},
 {0, 34, 34, 37, 0, 96, 2, 34, 34, 34, 2, 34, 68, 68, 68, 68, 0, 0, 2, 32, 68, 64, 0, 0, 34, 0, 0, 0, 34, 0, 0, 0, 2},
 {0, 32, 0, 2, 0, 0, 2, 32, 0, 0, 0, 2, 32, 0, 68, 68, 0, 0, 34, 0, 64, 0, 0, 0, 2, 32, 0, 2, 32, 0, 0, 0, 2},
 {0, 36, 0, 2, 34, 34, 34, 32, 0, 0, 0, 0, 34, 0, 0, 0, 0, 2, 32, 0, 64, 0, 0, 48, 2, 34, 37, 34, 34, 0, 3, 0, 2},
 {0, 36, 48, 2, 68, 66, 32, 34, 0, 4, 68, 0, 2, 32, 0, 0, 0, 2, 34, 0, 68, 68, 68, 0, 34, 0, 0, 0, 2, 32, 0, 0, 2},
 {0, 36, 0, 2, 68, 34, 0, 2, 32, 0, 0, 0, 0, 34, 34, 34, 34, 32, 2, 32, 0, 4, 68, 2, 32, 0, 34, 32, 0, 34, 4, 64, 2},
 {0, 36, 64, 2, 66, 32, 0, 0, 34, 2, 32, 48, 2, 32, 4, 0, 34, 34, 0, 34, 0, 0, 0, 34, 0, 2, 34, 34, 0, 2, 32, 0, 2},
 {0, 34, 34, 37, 34, 0, 0, 0, 0, 0, 34, 0, 34, 0, 0, 2, 34, 34, 32, 2, 32, 0, 2, 32, 0, 34, 34, 34, 32, 0, 34, 0, 2},
 {0, 80, 3, 66, 32, 0, 0, 0, 0, 0, 2, 34, 32, 0, 0, 34, 34, 34, 34, 0, 34, 85, 34, 0, 0, 34, 34, 34, 34, 0, 2, 80, 2},
 {0, 34, 34, 34, 0, 4, 68, 0, 0, 0, 0, 34, 0, 0, 2, 34, 34, 34, 34, 32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 34, 2},
 {34, 32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 0, 0, 64, 0, 2, 34},
 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2},
 {0, 0, 2, 34, 36, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 2, 32, 34, 34, 34, 34, 34},
 {0, 0, 2, 0, 0, 0, 36, 36, 36, 32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 36, 64, 36, 0, 2, 0, 0, 0, 0, 0, 2, 0, 0, 2},
 {34, 34, 34, 0, 0, 32, 32, 32, 32, 32, 0, 0, 0, 34, 34, 32, 0, 0, 0, 33, 0, 32, 0, 2, 34, 34, 34, 34, 34, 2, 2, 34, 2},
 {0, 0, 2, 0, 0, 32, 32, 32, 32, 32, 0, 0, 2, 36, 68, 34, 0, 0, 0, 32, 0, 32, 0, 2, 0, 0, 0, 0, 2, 2, 2, 50, 2},
 {4, 67, 5, 0, 0, 32, 32, 32, 32, 32, 0, 0, 34, 4, 0, 2, 32, 0, 0, 34, 34, 82, 34, 34, 0, 68, 2, 34, 2, 0, 2, 2, 2},
 {0, 0, 2, 0, 0, 32, 0, 0, 0, 32, 0, 2, 36, 0, 6, 4, 34, 0, 0, 32, 0, 32, 0, 2, 34, 0, 2, 2, 2, 34, 34, 2, 2},
 {34, 34, 34, 34, 34, 32, 32, 32, 32, 32, 0, 34, 0, 64, 64, 0, 66, 32, 0, 36, 68, 32, 68, 2, 66, 0, 2, 2, 0, 0, 0, 2, 2},
 {0, 68, 0, 0, 4, 32, 32, 32, 32, 32, 0, 34, 0, 0, 3, 0, 2, 32, 0, 32, 0, 32, 0, 5, 69, 0, 2, 2, 34, 34, 34, 34, 2},
 {0, 0, 0, 64, 0, 80, 32, 32, 32, 32, 0, 2, 36, 68, 68, 68, 34, 0, 0, 34, 34, 82, 34, 34, 66, 0, 2, 0, 0, 0, 0, 0, 2},
 {0, 3, 4, 0, 0, 32, 32, 32, 32, 32, 0, 0, 34, 0, 0, 2, 32, 0, 0, 32, 64, 32, 0, 2, 34, 0, 2, 34, 34, 34, 34, 34, 2},
 {34, 34, 34, 34, 34, 32, 36, 36, 32, 32, 0, 0, 2, 34, 82, 34, 0, 0, 0, 32, 64, 36, 68, 66, 0, 0, 0, 0, 0, 0, 0, 2, 2},
 {0, 0, 0, 0, 0, 0, 32, 35, 32, 80, 0, 0, 0, 0, 0, 0, 0, 68, 0, 32, 4, 32, 0, 2, 34, 68, 68, 34, 0, 34, 50, 34, 2},
 {34, 32, 0, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 2, 34, 34, 32, 0, 2, 34, 0, 2},
 {0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2},
 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2},
 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 37, 34, 34, 34, 32, 2, 32, 2},
 {0, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 36, 68, 68, 64, 34, 0, 2, 32, 2},
 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 34, 34, 34, 34, 82, 34, 34, 32, 2, 0, 2, 1, 34, 68, 68, 2, 32, 0, 2, 32, 2},
 {0, 2, 34, 34, 34, 34, 34, 32, 0, 2, 32, 2, 32, 0, 4, 68, 0, 0, 0, 32, 2, 0, 2, 68, 66, 32, 96, 34, 2, 34, 34, 32, 2},
 {0, 2, 0, 0, 0, 0, 0, 32, 0, 2, 0, 0, 32, 68, 68, 68, 0, 0, 0, 32, 2, 0, 2, 4, 64, 34, 34, 32, 0, 0, 2, 32, 2},
 {0, 2, 0, 0, 48, 68, 0, 32, 0, 2, 0, 0, 32, 0, 34, 34, 34, 34, 34, 32, 2, 0, 2, 0, 64, 2, 34, 34, 34, 34, 2, 32, 2},
 {0, 2, 34, 34, 0, 34, 34, 32, 0, 2, 0, 0, 32, 0, 32, 0, 0, 0, 0, 0, 2, 0, 2, 68, 64, 0, 34, 0, 0, 0, 2, 32, 2},
 {0, 0, 0, 2, 0, 32, 0, 0, 0, 2, 34, 34, 34, 34, 32, 2, 34, 34, 34, 32, 2, 34, 34, 34, 37, 82, 34, 32, 34, 34, 34, 32, 2},
 {0, 0, 0, 2, 0, 32, 4, 68, 64, 2, 0, 0, 0, 0, 32, 2, 0, 0, 0, 0, 0, 0, 2, 0, 0, 2, 0, 34, 0, 0, 2, 32, 2},
 {0, 2, 34, 34, 0, 32, 0, 0, 0, 2, 0, 0, 68, 68, 32, 2, 0, 34, 34, 34, 34, 32, 2, 0, 0, 2, 0, 2, 34, 34, 2, 32, 2},
 {0, 2, 48, 0, 2, 34, 34, 37, 34, 34, 34, 34, 34, 68, 32, 2, 0, 34, 34, 34, 34, 32, 2, 0, 48, 2, 68, 0, 34, 0, 2, 32, 2},
 {0, 2, 34, 34, 0, 32, 0, 0, 0, 0, 0, 0, 0, 0, 32, 2, 0, 0, 0, 0, 0, 0, 2, 68, 64, 2, 68, 68, 66, 32, 0, 0, 2},
 {0, 0, 0, 2, 0, 34, 34, 34, 34, 34, 34, 34, 34, 34, 32, 2, 34, 34, 34, 34, 34, 34, 34, 68, 64, 2, 34, 32, 0, 32, 34, 34, 34},
 {0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2},
 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2},
 {34, 34, 34, 34, 34, 2, 34, 34, 0, 0, 0, 34, 34, 32, 0, 0, 34, 34, 37, 34, 34, 0, 0, 34, 34, 32, 0, 0, 0, 34, 34, 34, 2},
 {68, 0, 0, 0, 2, 2, 52, 69, 2, 34, 34, 32, 0, 34, 0, 0, 32, 34, 0, 0, 2, 0, 2, 32, 68, 34, 0, 0, 34, 32, 34, 34, 2},
 {0, 34, 34, 2, 34, 2, 34, 34, 2, 0, 0, 0, 0, 2, 32, 0, 32, 2, 34, 34, 2, 0, 34, 2, 34, 2, 32, 0, 32, 0, 32, 0, 2},
 {0, 32, 2, 2, 0, 0, 0, 2, 2, 2, 34, 34, 2, 0, 34, 0, 32, 0, 0, 2, 2, 34, 32, 2, 2, 68, 34, 34, 32, 48, 34, 32, 2},
 {0, 32, 2, 2, 34, 34, 34, 2, 34, 2, 68, 66, 2, 0, 2, 34, 34, 34, 32, 2, 0, 0, 0, 2, 2, 100, 0, 4, 68, 64, 0, 32, 2},
 {0, 32, 2, 0, 0, 0, 2, 0, 0, 2, 68, 66, 2, 0, 0, 0, 0, 0, 32, 2, 34, 34, 34, 34, 2, 34, 34, 34, 34, 34, 34, 32, 2},
 {0, 32, 2, 34, 34, 32, 2, 34, 34, 34, 0, 2, 2, 34, 0, 2, 34, 0, 32, 0, 0, 0, 68, 0, 0, 0, 0, 0, 0, 0, 0, 32, 2},
 {0, 32, 0, 0, 0, 32, 0, 0, 0, 0, 3, 2, 2, 50, 34, 34, 18, 0, 32, 32, 0, 0, 2, 34, 32, 2, 32, 34, 34, 0, 0, 32, 2},
 {0, 34, 34, 34, 0, 32, 0, 34, 34, 34, 34, 34, 2, 68, 4, 68, 66, 0, 32, 34, 34, 34, 34, 0, 32, 2, 0, 32, 2, 32, 0, 34, 2},
 {0, 0, 3, 66, 0, 34, 34, 32, 0, 0, 0, 0, 2, 34, 36, 68, 34, 0, 32, 0, 4, 0, 66, 0, 36, 66, 0, 32, 0, 34, 0, 2, 2},
 {0, 0, 34, 34, 0, 0, 0, 0, 34, 34, 34, 0, 0, 0, 36, 34, 32, 0, 34, 34, 36, 34, 34, 0, 34, 34, 0, 34, 32, 2, 32, 2, 2},
 {0, 0, 35, 0, 0, 34, 34, 34, 36, 68, 66, 34, 32, 0, 68, 68, 68, 0, 0, 0, 36, 32, 0, 0, 0, 64, 0, 0, 32, 0, 34, 2, 2},
 {0, 0, 34, 34, 34, 32, 0, 0, 0, 0, 0, 0, 34, 34, 32, 34, 34, 34, 34, 34, 36, 34, 34, 34, 34, 34, 34, 34, 32, 0, 2, 34, 2},
 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2},
 {0, 2, 0, 0, 2, 48, 0, 64, 0, 0, 2, 0, 2, 2, 2, 2, 2, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2},
 {2, 2, 0, 0, 2, 34, 34, 66, 34, 34, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 34, 34, 34, 34, 34, 82, 34, 34, 34, 34, 34, 2},
 {2, 2, 34, 34, 0, 0, 2, 2, 2, 34, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 34, 0, 0, 2, 68, 68, 0, 0, 32, 2, 0, 2, 2},
 {2, 0, 0, 2, 0, 0, 2, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 2, 65, 64, 52, 64, 36, 34, 66, 2, 2},
 {2, 4, 3, 2, 0, 0, 2, 2, 34, 34, 34, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 2, 34, 34, 34, 34, 32, 34, 2, 0, 2},
 {66, 34, 34, 34, 0, 0, 2, 0, 0, 0, 0, 2, 0, 0, 0, 5, 0, 0, 2, 0, 0, 0, 0, 64, 0, 0, 0, 0, 0, 0, 2, 2, 2},
 {64, 0, 0, 50, 0, 0, 2, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 2},
 {0, 2, 34, 34, 0, 0, 5, 64, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 0, 2},
 {0, 32, 0, 0, 0, 0, 34, 32, 2, 34, 82, 34, 2, 34, 2, 34, 2, 34, 82, 34, 2, 34, 2, 34, 2, 34, 82, 34, 2, 34, 66, 34, 2},
 {0, 32, 34, 34, 34, 0, 34, 32, 66, 0, 0, 66, 2, 64, 0, 2, 2, 0, 0, 2, 2, 0, 0, 2, 2, 0, 0, 2, 2, 4, 0, 2, 2},
 {0, 34, 32, 32, 50, 0, 34, 32, 66, 64, 48, 66, 2, 0, 0, 2, 2, 68, 67, 2, 2, 4, 64, 2, 2, 0, 52, 66, 2, 4, 0, 2, 2},
 {0, 0, 0, 32, 2, 0, 34, 32, 2, 64, 96, 66, 2, 0, 0, 2, 2, 0, 0, 2, 2, 0, 0, 2, 2, 4, 64, 66, 2, 0, 0, 2, 2},
 {66, 34, 34, 32, 2, 0, 34, 32, 2, 0, 0, 2, 2, 68, 68, 66, 2, 0, 0, 2, 2, 0, 0, 2, 2, 0, 0, 2, 2, 0, 0, 2, 2},
 {2, 0, 0, 0, 34, 34, 34, 32, 2, 34, 34, 34, 2, 34, 34, 34, 2, 34, 34, 34, 2, 34, 34, 34, 2, 34, 34, 34, 2, 34, 34, 34, 2},
 {2, 4, 68, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 64, 0, 0, 0, 0, 0, 0, 0, 0, 2},
 {0, 0, 0, 0, 0, 2, 0, 64, 0, 0, 0, 2, 0, 0, 2, 34, 0, 0, 0, 0, 2, 0, 0, 2, 0, 0, 2, 0, 0, 2, 0, 4, 2},
 {0, 0, 0, 0, 0, 5, 0, 0, 0, 0, 0, 5, 0, 64, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 64, 0, 0, 0, 5, 4, 0, 2},
 {34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 2, 0, 0, 2, 0, 0, 2, 34, 82, 34},
 {0, 0, 0, 0, 0, 50, 0, 0, 0, 0, 0, 0, 0, 0, 2, 64, 0, 2, 0, 0, 0, 32, 34, 34, 34, 34, 34, 34, 34, 34, 4, 0, 66},
 {0, 0, 2, 34, 34, 34, 82, 34, 34, 34, 32, 34, 34, 34, 34, 34, 34, 2, 2, 34, 32, 32, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 2},
 {4, 68, 2, 64, 0, 2, 2, 22, 0, 64, 32, 32, 0, 0, 0, 0, 0, 2, 2, 51, 32, 32, 34, 34, 34, 34, 34, 34, 2, 34, 34, 82, 34},
 {0, 0, 2, 66, 2, 2, 2, 32, 32, 0, 32, 32, 34, 34, 34, 34, 34, 34, 2, 51, 32, 32, 35, 0, 32, 0, 0, 2, 2, 2, 64, 0, 2},
 {0, 64, 2, 2, 50, 2, 2, 32, 35, 0, 32, 32, 32, 0, 3, 0, 0, 5, 2, 32, 32, 32, 32, 64, 80, 34, 34, 2, 66, 2, 4, 68, 2},
 {0, 0, 2, 2, 50, 2, 0, 0, 32, 0, 32, 32, 32, 0, 64, 64, 0, 2, 2, 0, 0, 32, 32, 0, 32, 0, 2, 2, 50, 2, 3, 67, 2},
 {34, 82, 34, 2, 34, 2, 34, 34, 34, 2, 32, 32, 34, 34, 34, 34, 34, 34, 2, 34, 34, 32, 34, 34, 34, 34, 34, 2, 34, 2, 34, 82, 34},
 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 32, 0, 0, 0, 0, 0, 0, 64, 0, 0, 0, 0, 0, 0, 0, 68, 68, 64, 0, 0, 0, 68, 66},
 {34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 0, 0, 0, 0, 2, 2, 34, 34, 34, 34, 34, 34, 34, 82, 34},
 {0, 64, 2, 68, 68, 66, 0, 0, 2, 0, 0, 2, 0, 0, 2, 0, 0, 50, 34, 34, 34, 34, 34, 2, 0, 0, 2, 0, 0, 2, 0, 0, 2},
 {0, 0, 0, 68, 68, 64, 0, 4, 0, 0, 0, 5, 0, 0, 0, 0, 0, 5, 0, 0, 0, 0, 2, 0, 0, 0, 5, 0, 0, 5, 64, 0, 2},
 {0, 0, 2, 68, 68, 66, 0, 0, 2, 0, 0, 2, 0, 0, 2, 0, 0, 2, 0, 0, 0, 0, 0, 66, 0, 0, 2, 0, 0, 2, 0, 0, 2},
 {68, 4, 0, 0, 0, 2, 0, 0, 0, 0, 0, 5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 2, 48, 2},
 {34, 2, 34, 34, 34, 34, 2, 34, 34, 34, 34, 2, 2, 32, 34, 32, 34, 34, 32, 34, 2, 34, 34, 34, 34, 34, 2, 34, 34, 34, 2, 34, 2},
 {66, 2, 0, 2, 0, 2, 0, 0, 0, 4, 50, 2, 2, 48, 0, 32, 0, 2, 0, 2, 2, 3, 32, 32, 0, 0, 2, 0, 0, 0, 0, 0, 2},
 {2, 2, 66, 0, 64, 2, 0, 2, 34, 34, 34, 2, 2, 0, 0, 32, 0, 2, 0, 2, 2, 0, 32, 32, 0, 0, 2, 0, 0, 0, 2, 34, 34},
 {2, 4, 66, 34, 34, 34, 0, 2, 0, 0, 2, 2, 2, 34, 32, 32, 0, 2, 32, 34, 2, 33, 32, 34, 34, 34, 2, 34, 34, 34, 2, 0, 2},
 {3, 4, 69, 2, 0, 5, 0, 2, 0, 4, 5, 2, 2, 0, 32, 36, 34, 34, 32, 34, 2, 32, 32, 36, 2, 66, 2, 68, 32, 2, 2, 34, 2},
 {34, 34, 34, 2, 34, 34, 34, 2, 34, 0, 2, 2, 2, 0, 0, 32, 32, 0, 0, 2, 5, 0, 32, 80, 5, 0, 2, 68, 80, 2, 0, 0, 2},
 {48, 0, 0, 0, 0, 0, 0, 0, 2, 0, 2, 2, 2, 34, 32, 32, 34, 34, 34, 2, 2, 32, 32, 36, 2, 0, 2, 68, 32, 0, 2, 34, 34},
 {34, 2, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 48, 32, 32, 32, 0, 2, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 32, 2, 0, 2},
 {0, 4, 68, 0, 0, 0, 2, 0, 4, 0, 0, 0, 2, 68, 32, 34, 32, 0, 2, 0, 3, 4, 64, 4, 0, 0, 0, 0, 0, 36, 2, 34, 2},
 {0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 2, 0, 32, 35, 32, 0, 2, 34, 34, 34, 34, 34, 34, 34, 34, 34, 32, 32, 2, 0, 2},
 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2},
 {34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 35, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 2},
 {52, 68, 68, 0, 0, 0, 0, 0, 0, 2, 2, 34, 37, 0, 0, 34, 32, 0, 2, 0, 0, 0, 0, 0, 32, 0, 0, 0, 0, 0, 2, 0, 2},
 {100, 68, 68, 0, 0, 0, 0, 0, 0, 0, 32, 0, 2, 0, 0, 0, 0, 0, 0, 32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 32, 5, 0, 2},
 {0, 68, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2},
 {2, 34, 34, 34, 34, 34, 34, 32, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 2, 2},
 {0, 0, 0, 0, 2, 0, 0, 32, 32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 4, 64, 0, 2, 2, 2},
 {0, 0, 0, 0, 2, 82, 32, 32, 32, 34, 34, 34, 34, 34, 34, 82, 34, 34, 34, 34, 34, 34, 34, 34, 2, 34, 34, 34, 34, 34, 2, 2, 2},
 {34, 34, 34, 34, 2, 48, 32, 32, 32, 35, 0, 0, 80, 0, 2, 2, 65, 68, 32, 0, 0, 0, 0, 2, 2, 0, 0, 0, 0, 2, 2, 2, 2},
 {0, 2, 0, 2, 2, 0, 32, 32, 32, 32, 0, 34, 32, 34, 2, 2, 68, 68, 80, 0, 48, 0, 0, 2, 2, 2, 34, 34, 34, 2, 2, 2, 2},
 {0, 68, 68, 2, 2, 34, 32, 32, 32, 32, 0, 38, 32, 32, 2, 2, 34, 34, 32, 0, 0, 4, 64, 2, 66, 2, 0, 0, 2, 2, 2, 2, 2},
 {34, 68, 68, 34, 0, 0, 0, 32, 32, 32, 34, 36, 32, 32, 2, 0, 0, 0, 34, 34, 34, 34, 37, 34, 2, 2, 2, 34, 2, 2, 2, 2, 2},
 {0, 68, 68, 2, 34, 34, 34, 32, 32, 32, 80, 4, 32, 35, 2, 0, 68, 0, 32, 0, 0, 0, 68, 66, 66, 2, 2, 66, 2, 2, 2, 2, 2},
 {0, 2, 0, 34, 0, 0, 0, 0, 32, 34, 32, 4, 32, 34, 34, 0, 67, 0, 80, 0, 48, 0, 0, 2, 66, 2, 66, 66, 2, 2, 2, 2, 2},
 {34, 34, 85, 34, 2, 34, 34, 34, 32, 36, 68, 4, 32, 0, 0, 0, 64, 0, 32, 0, 0, 0, 0, 2, 2, 2, 66, 64, 2, 2, 2, 2, 2},
 {2, 0, 4, 66, 2, 0, 0, 0, 32, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 2, 2, 2, 34, 34, 2, 2, 2, 2},
 {4, 68, 68, 66, 2, 0, 32, 32, 32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 0, 2, 2, 2, 2},
 {2, 34, 34, 34, 2, 34, 34, 32, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 82, 2, 34, 34, 34, 34, 2, 34, 2},
 {0, 0, 0, 0, 0, 0, 0, 4, 84, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 52, 50},
 {0, 0, 0, 0, 0, 0, 0, 0, 0, 34, 0, 0, 0, 0, 34, 0, 0, 0, 0, 2, 32, 0, 0, 0, 2, 32, 0, 0, 0, 0, 2, 32, 2},
 {0, 34, 34, 34, 34, 34, 34, 34, 0, 2, 32, 68, 48, 2, 32, 0, 0, 0, 0, 2, 32, 4, 0, 0, 2, 32, 0, 0, 0, 0, 34, 0, 2},
 {0, 32, 0, 2, 0, 0, 0, 2, 32, 0, 34, 4, 0, 34, 0, 0, 0, 0, 0, 2, 32, 0, 64, 0, 2, 36, 64, 0, 0, 2, 32, 0, 2},
 {0, 36, 68, 66, 34, 34, 32, 0, 34, 0, 2, 32, 2, 34, 34, 34, 32, 34, 34, 34, 32, 4, 0, 0, 2, 32, 0, 0, 0, 34, 0, 0, 2},
 {0, 32, 48, 2, 0, 0, 34, 0, 2, 32, 0, 34, 82, 0, 0, 0, 0, 0, 0, 34, 0, 64, 0, 0, 2, 32, 0, 0, 2, 32, 0, 0, 2},
 {0, 32, 0, 0, 0, 0, 5, 34, 34, 34, 2, 34, 68, 68, 68, 68, 0, 0, 2, 32, 68, 64, 0, 0, 2, 32, 0, 0, 34, 34, 34, 0, 2},
 {0, 32, 0, 2, 0, 0, 2, 32, 0, 0, 0, 2, 32, 0, 68, 68, 0, 0, 34, 0, 64, 0, 0, 0, 2, 32, 0, 2, 32, 3, 2, 0, 2},
 {0, 36, 0, 2, 34, 34, 34, 32, 0, 0, 0, 0, 34, 0, 0, 0, 0, 2, 32, 0, 64, 0, 0, 48, 2, 34, 37, 34, 34, 68, 5, 0, 2},
 {0, 36, 48, 2, 68, 66, 32, 34, 0, 4, 68, 0, 2, 32, 0, 0, 0, 2, 34, 0, 68, 68, 68, 0, 34, 0, 0, 0, 2, 34, 34, 0, 2},
 {0, 36, 0, 2, 68, 34, 0, 2, 32, 0, 0, 0, 0, 34, 34, 34, 34, 32, 2, 32, 0, 4, 68, 2, 32, 0, 34, 32, 0, 34, 4, 64, 2},
 {0, 36, 64, 2, 66, 32, 0, 0, 34, 2, 32, 48, 2, 32, 4, 0, 34, 34, 0, 34, 0, 0, 0, 34, 0, 2, 34, 34, 0, 2, 32, 0, 2},
 {0, 34, 34, 37, 34, 34, 34, 32, 2, 0, 34, 0, 34, 0, 0, 2, 36, 66, 32, 2, 32, 0, 2, 32, 0, 34, 0, 2, 32, 0, 34, 0, 2},
 {0, 80, 3, 66, 32, 0, 0, 32, 2, 0, 2, 34, 32, 0, 0, 34, 67, 67, 34, 0, 34, 85, 34, 0, 0, 34, 32, 34, 34, 0, 2, 32, 2},
 {0, 34, 34, 34, 64, 4, 68, 32, 34, 34, 0, 34, 0, 0, 2, 34, 32, 34, 34, 32, 0, 0, 0, 0, 0, 0, 0, 2, 32, 0, 0, 34, 82},
 {34, 38, 20, 68, 64, 0, 0, 80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 0, 0, 64, 0, 0, 2},
 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2},
 {0, 0, 2, 34, 36, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 2, 32, 34, 34, 34, 34, 34},
 {0, 0, 2, 0, 0, 0, 36, 36, 36, 32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 36, 64, 4, 0, 2, 0, 0, 0, 0, 0, 2, 0, 0, 2},
 {34, 34, 34, 0, 0, 32, 32, 32, 32, 32, 0, 34, 34, 34, 34, 34, 34, 32, 0, 32, 0, 0, 0, 2, 34, 34, 34, 34, 34, 2, 2, 34, 2},
 {0, 0, 2, 0, 0, 32, 32, 32, 32, 32, 0, 3, 50, 36, 68, 34, 51, 0, 0, 32, 0, 0, 0, 2, 0, 0, 0, 0, 2, 2, 2, 50, 2},
 {4, 67, 5, 0, 0, 32, 32, 32, 32, 32, 0, 35, 34, 4, 0, 2, 35, 32, 0, 34, 34, 82, 34, 34, 0, 68, 2, 34, 2, 0, 2, 2, 2},
 {0, 0, 2, 0, 0, 32, 0, 0, 0, 32, 0, 34, 36, 0, 0, 4, 34, 32, 0, 32, 96, 32, 0, 2, 34, 0, 2, 34, 2, 34, 34, 2, 2},
 {34, 34, 34, 34, 34, 32, 32, 32, 32, 32, 0, 34, 0, 64, 64, 0, 66, 32, 0, 36, 68, 32, 68, 2, 66, 0, 0, 0, 0, 0, 0, 2, 2},
 {0, 68, 0, 0, 4, 32, 32, 32, 32, 32, 0, 34, 0, 0, 3, 0, 2, 32, 0, 32, 0, 32, 0, 5, 64, 0, 2, 2, 34, 34, 34, 34, 2},
 {0, 0, 0, 64, 0, 80, 32, 32, 32, 32, 0, 34, 36, 68, 68, 68, 34, 32, 0, 34, 34, 82, 34, 34, 66, 0, 2, 0, 0, 0, 0, 0, 2},
 {0, 3, 4, 0, 0, 32, 32, 32, 32, 32, 0, 3, 34, 0, 0, 2, 35, 0, 0, 32, 64, 0, 0, 2, 34, 0, 2, 34, 34, 34, 34, 34, 2},
 {34, 34, 34, 34, 34, 32, 36, 36, 32, 32, 0, 34, 34, 34, 82, 34, 34, 32, 0, 32, 64, 4, 68, 66, 0, 0, 0, 0, 0, 0, 0, 2, 2},
 {0, 0, 0, 0, 0, 0, 33, 35, 32, 32, 0, 0, 0, 0, 0, 0, 0, 68, 0, 32, 4, 0, 0, 2, 34, 68, 68, 34, 0, 34, 50, 34, 2},
 {34, 32, 0, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 0, 34, 34, 34, 34, 34, 2, 34, 34, 32, 0, 2, 34, 0, 2},
 {0, 80, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2},
 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2},
 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 34, 34, 34, 34, 34, 34, 34, 37, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 82, 0, 2},
 {0, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 36, 68, 68, 64, 0, 0, 2, 0, 2},
 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 34, 34, 34, 34, 82, 34, 34, 32, 2, 0, 2, 0, 34, 68, 68, 0, 0, 0, 2, 0, 2},
 {0, 2, 34, 34, 34, 34, 34, 34, 34, 2, 32, 2, 32, 0, 4, 68, 0, 0, 0, 32, 2, 0, 2, 68, 66, 32, 0, 0, 2, 34, 34, 0, 2},
 {0, 2, 0, 0, 96, 0, 0, 0, 2, 2, 0, 0, 33, 68, 68, 68, 0, 0, 0, 32, 2, 0, 2, 4, 64, 34, 34, 32, 0, 0, 2, 0, 2},
 {0, 2, 0, 0, 48, 68, 0, 0, 2, 2, 0, 0, 32, 0, 34, 34, 34, 34, 34, 32, 2, 0, 2, 0, 64, 2, 34, 34, 34, 34, 2, 0, 2},
 {0, 2, 34, 34, 0, 34, 34, 34, 34, 2, 0, 0, 32, 0, 32, 0, 0, 0, 0, 32, 2, 0, 2, 68, 64, 0, 34, 0, 0, 0, 2, 0, 2},
 {0, 0, 64, 2, 0, 32, 0, 0, 2, 2, 34, 34, 34, 34, 32, 2, 34, 34, 34, 32, 2, 34, 34, 34, 37, 2, 34, 32, 34, 34, 34, 0, 2},
 {0, 0, 64, 2, 0, 32, 4, 68, 66, 2, 0, 0, 0, 0, 32, 2, 0, 0, 0, 0, 0, 0, 2, 0, 0, 2, 0, 34, 0, 0, 2, 0, 2},
 {0, 2, 34, 34, 0, 32, 34, 34, 34, 2, 0, 0, 68, 68, 32, 2, 0, 34, 34, 34, 34, 32, 2, 0, 0, 2, 0, 2, 34, 34, 82, 0, 2},
 {0, 2, 48, 0, 2, 32, 32, 0, 0, 2, 34, 34, 34, 68, 32, 2, 0, 34, 34, 34, 34, 32, 2, 0, 48, 2, 68, 0, 0, 0, 0, 0, 2},
 {0, 2, 34, 34, 0, 32, 0, 0, 0, 0, 0, 0, 0, 0, 32, 2, 0, 0, 0, 0, 0, 0, 2, 68, 64, 2, 68, 68, 64, 0, 0, 0, 2},
 {0, 0, 0, 2, 0, 34, 34, 34, 34, 34, 34, 34, 34, 34, 32, 2, 34, 34, 34, 34, 34, 34, 34, 68, 64, 2, 34, 32, 0, 0, 34, 34, 34},
 {0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2},
 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2},
 {34, 34, 34, 34, 34, 2, 34, 34, 0, 0, 0, 32, 0, 0, 0, 0, 34, 34, 37, 34, 34, 0, 0, 34, 34, 32, 0, 0, 0, 34, 34, 34, 2},
 {68, 0, 0, 0, 2, 2, 100, 69, 2, 34, 34, 32, 0, 34, 0, 0, 32, 34, 0, 0, 2, 0, 2, 32, 68, 34, 0, 0, 34, 32, 34, 34, 2},
 {0, 34, 34, 2, 34, 2, 34, 34, 2, 0, 0, 0, 0, 2, 32, 0, 32, 2, 34, 34, 2, 0, 34, 2, 34, 2, 32, 0, 32, 0, 32, 0, 2},
 {0, 32, 2, 2, 0, 0, 0, 2, 2, 2, 34, 34, 2, 0, 34, 0, 32, 0, 0, 2, 2, 34, 32, 2, 2, 68, 34, 34, 32, 48, 34, 32, 2},
 {0, 32, 2, 2, 34, 34, 34, 2, 34, 2, 68, 66, 2, 0, 2, 34, 34, 34, 32, 2, 0, 0, 0, 2, 2, 20, 0, 4, 68, 64, 0, 32, 2},
 {0, 32, 2, 0, 0, 0, 2, 0, 0, 2, 68, 66, 2, 0, 0, 0, 0, 0, 32, 2, 34, 34, 34, 34, 2, 34, 34, 34, 34, 34, 34, 32, 2},
 {0, 32, 2, 34, 34, 32, 2, 34, 34, 34, 0, 2, 2, 34, 0, 2, 34, 0, 32, 0, 0, 0, 68, 0, 0, 0, 0, 0, 0, 0, 0, 32, 2},
 {0, 32, 0, 0, 0, 32, 0, 0, 0, 0, 3, 2, 2, 48, 0, 2, 2, 0, 32, 0, 0, 0, 2, 34, 34, 34, 32, 34, 34, 0, 0, 32, 2},
 {0, 32, 0, 0, 0, 32, 0, 34, 34, 34, 34, 34, 2, 68, 4, 68, 66, 0, 32, 34, 34, 34, 34, 0, 0, 0, 0, 32, 2, 32, 0, 34, 2},
 {0, 32, 3, 64, 0, 34, 34, 32, 0, 0, 0, 0, 2, 34, 36, 68, 34, 0, 32, 0, 4, 0, 66, 0, 4, 64, 0, 32, 0, 0, 0, 2, 2},
 {0, 32, 0, 0, 0, 0, 0, 0, 34, 34, 34, 0, 0, 0, 34, 34, 32, 0, 34, 34, 36, 34, 34, 0, 0, 0, 0, 34, 34, 34, 32, 2, 2},
 {0, 32, 3, 0, 0, 34, 34, 34, 36, 68, 66, 34, 32, 0, 68, 68, 68, 0, 0, 0, 36, 32, 0, 0, 0, 64, 0, 0, 32, 0, 34, 2, 2},
 {0, 34, 34, 34, 34, 32, 0, 0, 0, 0, 0, 0, 34, 34, 32, 34, 34, 34, 34, 34, 36, 34, 34, 34, 34, 34, 34, 34, 32, 0, 2, 34, 2},
 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2},
 {5, 5, 0, 0, 2, 52, 68, 68, 68, 68, 66, 0, 0, 2, 0, 2, 0, 2, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2},
 {2, 2, 0, 0, 2, 34, 34, 66, 34, 34, 66, 2, 0, 2, 0, 2, 0, 2, 2, 2, 34, 34, 34, 34, 34, 34, 82, 34, 34, 34, 34, 34, 2},
 {2, 2, 34, 34, 0, 36, 50, 66, 98, 34, 66, 2, 0, 2, 0, 2, 0, 2, 2, 2, 34, 0, 0, 0, 68, 68, 0, 0, 32, 2, 0, 2, 2},
 {2, 4, 0, 2, 2, 82, 34, 68, 68, 68, 66, 2, 0, 2, 0, 2, 0, 2, 2, 2, 0, 0, 0, 0, 64, 64, 52, 64, 36, 2, 66, 2, 2},
 {2, 4, 3, 2, 2, 36, 2, 66, 34, 34, 34, 2, 0, 2, 0, 2, 0, 2, 2, 2, 34, 34, 34, 34, 34, 34, 34, 34, 32, 0, 2, 0, 2},
 {82, 34, 34, 34, 2, 0, 34, 64, 0, 0, 0, 2, 0, 0, 0, 5, 0, 0, 2, 0, 0, 0, 0, 64, 0, 0, 0, 0, 0, 0, 2, 2, 2},
 {64, 80, 80, 50, 2, 0, 2, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 34, 2},
 {0, 34, 34, 34, 2, 32, 0, 64, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 64, 0, 2},
 {0, 32, 0, 5, 0, 0, 34, 32, 2, 34, 82, 34, 82, 34, 34, 34, 2, 34, 82, 34, 82, 34, 34, 34, 2, 34, 82, 34, 82, 34, 34, 34, 2},
 {0, 32, 34, 34, 34, 0, 32, 32, 66, 0, 0, 66, 2, 64, 0, 2, 2, 0, 0, 2, 2, 0, 0, 2, 2, 0, 0, 2, 2, 4, 0, 2, 2},
 {0, 34, 32, 0, 50, 0, 32, 32, 66, 64, 48, 66, 2, 0, 0, 2, 2, 68, 67, 2, 2, 4, 64, 2, 2, 0, 52, 66, 2, 4, 0, 2, 2},
 {85, 32, 0, 0, 2, 0, 32, 32, 2, 64, 0, 66, 2, 0, 0, 2, 2, 0, 0, 2, 2, 0, 0, 2, 2, 4, 64, 66, 2, 0, 0, 2, 2},
 {66, 34, 34, 32, 2, 0, 0, 32, 2, 0, 0, 2, 2, 68, 68, 66, 2, 0, 0, 2, 2, 0, 0, 2, 2, 1, 0, 2, 2, 0, 0, 2, 2},
 {2, 0, 0, 0, 34, 34, 34, 32, 2, 34, 34, 34, 2, 34, 2, 34, 82, 34, 34, 34, 2, 34, 2, 34, 82, 34, 34, 34, 2, 34, 2, 34, 82},
 {2, 4, 68, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 64, 0, 0, 0, 0, 0, 0, 0, 0, 2}
};

/*
Char canned[CAN_LEV * CAN_ROW][CAN_COL*2] = {
"           #  .        #                 #     #     #     #   . ",
"           D           D  .                       .        D .   ",
"##D##########################################  #     #     ###D##",
"          -#                 #.    #      # ################ .  .",
"     ####### ######## ############ D #### #                #     ",
" ... #.#   # #  # . # #            # #### # ############ # ###D##",
"     #.# # # ## #   # # ############ #### # #-  #      # # #.    ",
"  .  # # # # ## #-  # # #    -     D #### # # . D    # #.# # ... ",
"     # #.# #    #   # # #   . .    # #    # #   #    # #-# # ~.! ",
"##D### ### #######D## # ############ ###### ########## ### ######",
"                    #            .#        #      .....       ...",
"##D###########################################################D##",
"  .  #.....#     #     #     #    -#     #     #     #     #     ",
"      .....    .       D           D                 D     D.    ",
"     #.....#     #     #     #     #     #    .#     #     #     ",
//
" #             . - .                          .             .    ",
" #D############################################################ .",
" # #       .                          -                     . # #",
".#.#D#######################################################  #. ",
" # # #   .                     .              .            #  ##-",
" #-# #D################################################### #  # .",
" # # # # .     # -     #       #  .   -#  .    #.......#!# #  # #",
" # # # # ### ### ### ###.### ### ### ### ### ###.###.###~# #  #. ",
".# # # .   # .  .  #    .  #    -  # . .   #   ....#.....# #  ## ",
" #.# ##################################################### #  # .",
" # #                            ..            .            #. #-#",
" # #########################################################  #. ",
" #        ..     -                      .         .       . . ## ",
"-##############################################################  ",
".                                                              . ",
//
"##   #    ##  #...#           .  . .  .     #  #  #  ##       .##",
" ##     #  ##   #   ####################### #  #.   ##  #####D#D ",
" D########  ##  #####   #    D D      #   # #  #   ##  ## -..##  ",
" # ##..- #   ##     #   #   ## ##  .  # # # #  #  ###  ###  ###  ",
" #  ##-..#    ## .  ##  #  ##-. ## .  # # # #    ## #  #-.-## #  ",
" #   ## -##### ##   #  .  ## . - ##   # # # #   ##  #  #..##  #  ",
".#### ##D##-.#  ##  #  . ## .....-##  # # # #..##  -#### ##   #  ",
" #  # DD ##-.D   ## # ..######D###### # # # #.DD  ###  D##  ###  ",
" #.## ## #####  ## ###    #  #D#   #    #   #..##   #    ##...#  ",
" #.# ##..##    ##  #    #   ##.##    ########   ##  ##  ####..#  ",
"-#. ##...##   ##.  ###### ### .-##    #  -  #    ##  ##    ## #  ",
" # ##..#.##  ##. ###  # # ##.-. .##   #   ######  ##  ####  ##D. ",
" ### #####  ##         .###!. . .~##.   ##  #      ##  ##   -##  ",
" ## ##     ############ #####################  ###  ##  ####  ## ",
"##     #   D    . .    .      -             ..       ##.....  ###",
//
".  #   #      #   .            #     #  #       ##    ####    .. ",
" # # # ###### #  #    .       . .  #      . #      .#  . ##.  .. ",
" #.  #  .     .  #  # . #     #D######################## #D##  - ",
" #################    #      ##.## . .                   #..##   ",
" #     .#   .    #   ###    ##. .##  #####################   ##  ",
" #  -  .D   . -  #  # - #  ##  #  ##   #.  #   #   #   #   ## ## ",
" ########  ..    #        ##  ###  ##  # # # # # # # # #   ##  ##",
" D.              #       ##  ##!##  ##   # # # # # # # #     .. #",
" ################# .    ##  ##.~.##  ##### # # # # # # ##### ... ",
" D.     #      . #     ##. ## .#. ##  ## ..# # #.# # #.#    . -. ",
" # .- . #.    -  #    ##  ## ..#.. ##  ##  # # #-# # # #    #####",
" #      D.       #   ##  ##  #####- ##  ##   # # # # # #  #.#    ",
" ##############D##  ##  ## .. ... .. ##  ##### #.# # # ## #   - #",
" #.     .        # ##  ########D########  ## . # # # # #  # #  . ",
" #     -       . ###     -    . .          ##.   #   #   #### #  ",
//
" # .- . #.    -  #    ##  ## ..#.  ##  ##  # # #-# # # #    #####",
" #      D.       #   ##  ##  #####- ##  ## .   # # # # #  #.#    ",
" ##############D##  ##  ## .      .. ##  ##### #.# # # ## #   - #",
".  #   #      #   .            ##### #  #        #    ####    .. ",
" #D# # ###### #  #    .       . .  #      . #      .#  . ##.  .. ",
" #.  #  .     .  #  # . #     #D########################  D##  - ",
" ################# .    ##  ##. .##  ##### # # # # # # ##### ... ",
" D.     #      . #     ##. ## .#. ##  ## ..# # #.# # #.#    . -. ",
" #.     .        # ##  ########D########  ## . # # # # #  # #  . ",
" #################    #      ##.## . .                   #..##   ",
" #     .#   .    #  ####    ##. .##DD#####################   ##  ",
" #  -  .D   . -  #  # -##  ##  #  ##   #.  #   #   #.  #   ## ## ",
" ########  ..    #  #  #  ## .### .##  # # # # # # # # #   ##  ##",
" D.              #     # ##  ##!##  ##   #   # # # # # #     .. #",
" #     -       . ###    -## . .~. . ##     ###.  #   #   #### #  ",
//
".. .                   D  #                .  #              #-  ",
"############ ######### # ## ### ##### ## #### ###### ####### ### ",
".#!#~# #   #       .-# # #-   #    #   # # -# #      #           ",
" # #.#  .  #   ####### # #    #    #   # #  # #      #       ####",
" # ..# #####   #     # # #### #    ## ## ## # ###### ####### #   ",
" - ..D #   D   #   . D # #  # #.##### ## ## # #. #.# #..#  # ### ",
"###### ####### ###   # # #    # #      # D  # D  D   #..D  #     ",
"-                #   # # #### # ###### # ## # #. #   #..#    ####",
"## #######################- # # #    ######################  #   ",
"   ...       #   . #     #..# ###    #   - ..  .          #. ### ",
" ########    #     #     #  # ###    #################### #  #   ",
"        #                     ###                                ",
"################################################################ ",
"-     #             #    D    ###    #          #            #   ",
"  .         #            #            #                   #  D   ",
//
"  ..                                                             ",
" ############## ############################################## # ",
"         #    # #                                #     ..    # # ",
"         #D## # # ############D################# ########### # # ",
"######## #- # # # #-    D    # # ~  #          # #         # # # ",
"       # #  # # # #   ### #  # #    D   -      # # ####### # # # ",
"  .... # #### # # #   #!# #  # ######      ..  #.# #     # # # # ",
"  .... #      # # # ### # #  # #    #########D#### # ### # # # # ",
"  .... ######## # # D   # #- # #..  #       ...#.# # #.# # # # # ",
"       #        # ###   # #### #.-  D   -      #.# #.#.# # # # # ",
"####DD## ######## #     #    D D.   #          # # #...# # # # # ",
" #   ..# #    # # ############################## # ##### # # # # ",
" ......# #  #   #                                #       # # # # ",
" ####### ###### ################################ ######### # ### ",
"               .D.                                         #     ",
//
"                  ##        ##            ###                ##  ",
"  #####            ## ..-  ##            ##.##              ##   ",
"  # ! ##            ## .  ##            ##  .##    ..      ##    ",
"  #....######        ##  ##            ##  .  ##          ##     ",
"  # -  #    ##        ##D#            ##  .    ##        ##      ",
"  #####D  ~  ####### ###........     ## ...     ##      ##       ",
"  #    #     ##        ##   ....    ##  .        ##    ##        ",
"  #.   ########         ##         ##   .     -  ####D####   -   ",
"  #.-  #...## ##   ...   ##        ###  ......  ##       ##      ",
"  #.   #..##   ##         #########  ##    ... ##   ###   ## ..  ",
"  #..  #.##     ## ## -  ##  .  ####  ##      ##   #####   ##    ",
"  #####D##          ##  ##     ######  ##    ##   #######   ##   ",
"  D  -.##            ####     ########  ##DD##    ########   #D  ",
"  ######   ...        ##     ##########                       ## ",
"###                                                 .     .    ##",
//
"                                                                 ",
"     ####.########################################## ## #########",
"     #      #.#.#.#                   #.. #.   #           #     ",
"######    # # # # #       #####       #!  #    ########### # ### ",
"     #    # # # # #      ##...##      #   #    #         # # #-# ",
" ..- D    # # # # #     ## .   ##     ####D#####  .. ### #   # # ",
"     #    #       #    ##.   ~ .##    #   #    ###   # # ##### # ",
"########### # # # #   ##  . .   .##   #...# .. #.#   # #       # ",
"  ..     .# # # # #   ##     -   ##   #   #    D.D   # ######### ",
"      .   D # # # #    ##.......##    ####D#####.#   #           ",
"   - .    # # # # #     ##     ##     # . #    ###   ########### ",
"########### #.#.# #      ###D###      # . #....#               # ",
"            # #-# D               ..  #  .#    ###....##  ##-### ",
"###   ########################################## ######    ###   ",
"         .                                                       ",
//
"                                                                 ",
"                        ###########################D#######  ##  ",
"  #######################                    # ##...... ##   ##  ",
"                        ########D######  #   # !##.... ##    ##  ",
"   ############    ##  ##    ...      #  #   #...## ~ ## ######  ",
"   #          #    #    # ......      #  #   # .. #####      ##  ",
"   #    - ..  #    #    #   ###########  #   #  .  ######### ##  ",
"   #####  #####    #    #   #            #   #...   ##       ##  ",
"       #  #        ##########  ########  ########DD#### #######  ",
"       #  #  ....  #        #  #             #     #  ##     ##  ",
"   #####  #        #    ....#  #  #########  #     #   ##### ##  ",
"   #-    ######D##########..#  #  #########  #  -  #..  ##   ##  ",
"   #####  #                 #  #             #...  #.....##      ",
"       #  ###################  ###############...  ####   # #####",
"       #                                                         ",
//
"                                                                 ",
"########## #####      #####     #####D####    #####       ###### ",
"..       # #-..D ######   ##    # ##     #   ## ..##    ### #### ",
"  #### ### ##### #         ##   #  ##### #  ## ### ##   #   #    ",
"  #  # #       # # ##### #  ##  #      # ####  # #..##### - ###  ",
"  #  # ####### ### #...# #   ########  #       # #~.   ....   #  ",
"  #  #       #     #...# #          #  ######### ##############  ",
"  #  ######  #######   # ###   ###  #       ..                #  ",
"  #       #          - # #-#####!#  # #      ####  ## ####    #  ",
"  ######  #   ########## #.. ....#  # ########  #  #  #  ##   ## ",
"     -.#  #####          ####...##  #    .  .#  #..#  #   ##   # ",
"    ####        ######      #.###   #####.####  ####  ###  ##  # ",
"    #-    #######....####   ......      #.#       .     #   ## # ",
"    #######             ##### ###########.###############    ### ",
"                                         .                       ",
//
"   #     #-   .      #   # # # # # #   #                         ",
" # #     #####.##### # # # # # # # # # # ###########D########### ",
" # #####     # # ### # # # # # # # # # ###     #....    #  #   # ",
" #     #     #       # # # # # # # # # #       #.!. -.. #.##.# # ",
" # . - #     # ####### # # # # # # # # #       ########## ## #   ",
".#######     #         #       D     #        .              # # ",
".     -#     ################################################### ",
"   #####     D.                                             .    ",
"  #         ###  ###D### ### ### ###D### ### ### ###D### ###.### ",
"  # ######  ### .#    .# #.    # #     # #     # #     # # .   # ",
"  ### # -#  ### .#. - .# #     # #...- # # ..  # #  -..# # .   # ",
"      #  #  ###  #. ~ .# #     # #     # #     # # .. .# #     # ",
".######  #  ###  #     # #.....# #     # #     # #     # #     # ",
" #      #######  ####### ####### ####### ####### ####### ####### ",
" # ...                                       ..                  ",
//
"           #  .        #     ###         #     #     #     #   . ",
"           D           D  .        #              .        D .   ",
"############################################## #     #     ###D##",
"          -#                 #.    #      # ################ .  .",
"     #######D######## ############ # #### #                #     ",
" ... #.    # #!~  . # #            # #--# # ############ #####D##",
"     #.# # # ## #   # # ############ #--# # #-  #      # # #.    ",
"  .  # #-# # ## #-  # # #    -     D ## # # # . D #### #.# # ... ",
"     # #-# #    #   # # #   . .    # #    # #   #    # #-# # -.- ",
"##D### ### ####### ## # ############ ###### ########## ### ###D##",
"                    #             .               .....       ...",
"####################################         # ###############D##",
"  .  #.....#     #     #     #    -########### #     #     #     ",
"      .....    .       D           D         #       D     D.    ",
"     #.....#     #     #     #     #          .#     #     #     ",
//
".. .       #           D                   .                 #-  ",
"## ######### ######### # ## ### ##### ## ########### ####### ### ",
".# #   #   #       .-# # #-   #    #   # # -# #      #           ",
" # #.#  .  #   ####### # #    #    #   # #  # #      #       ####",
" # ..#######   #     # # #### #    ## ## ##!# ###### ####### #   ",
" - ..D #   D   #   . D # #  # #.##### ## ## # #. #.# #..#  # ### ",
"###### ####### ###   # # #    # #      # D  # D  D   #..D  #     ",
"-                #   # # #### # ###### # ## # #. #   #..#    ####",
"## #######################- # # #    ######################  #   ",
"   ...       #   .       #..# ###    #   - ..  .          #. ### ",
"             #           #  # #-#    #################### #  #   ",
"                         #      #                                ",
"###############################-################################ ",
"-.....             # ####D    ###    #          #            #   ",
"~.....              #    #            #                   #  D   ",
//
"  ..                                                             ",
" ############## ############################################## # ",
"         #    # #                                #     ..    # # ",
"         #D## # # ############D################# ########### # # ",
"######## #- # # # #-    D    # #.!..#          # #         # # # ",
"   #   # #  # # # #   ### ## # #....D   -      # # ####### # # # ",
"  .... # #### # # #   #~# #  # ######      ..  #.# #     # # # # ",
"##....##      # # # ###.# #  #      #########D## # # ### # # # # ",
"  .... ######## # # D  .# #- #  ..  #       ...#.# # #.# # # # # ",
"   #  ##        # ###  .# ####  .-  D   -      #.# #.#.# # # # # ",
"####DD## ######## #... .#       .   #          # # #.#.  # # # # ",
" #   ..# #      # ############################## # # ##### # # # ",
" ......# #  # # #                                # #       # # # ",
" ####### ###### ################################D# ######### ### ",
"               .D.                                           #-.-",
//
"                  ##        ##         ##        ##          ##  ",
"  ##############   ## ..-  ##          ##  .     ##         ##   ",
"  #    #       ##   ## .  ##           ##   .    ##..      ##    ",
"  #....######   ##   ##  ######## #######  .     ##       ##     ",
"  # -  #    ##   ##   ##D#            ##  .      ##      ##      ",
"  #          D###### ###........     ## ...      ##     ######   ",
"  #    #     ##        ##   ....    ##  .        ##    ##  - #   ",
"  #.   ########         ##         ##   .     -  ####D####.. D   ",
"  #.-  #...## ##   ...   ##        ###  ......  ##       #####   ",
"  #.   #..##   ##         #########  ##    ... ##   ###   ## ..  ",
"  #..  #.##     ## ## -  ##  .  ####  ##      ##   #####   ##    ",
"  #####D#######  #  ##  ##     ##..##  ##    ##   ##   ##   ##   ",
"  D  -.##     #  #   ####     ##.-.-##  ##DD##    ### ####   ##  ",
"  ######.  ...# ####  ##     #### #####                ##     ##D",
"###~!....     D                                     .     .      ",
//
"                                                                 ",
"     ####.########################################## ## #########",
"     #      #.#.#.#                   #..  .   #           #     ",
"######    # # # # #   #############   #        ########### # ### ",
"     #    # # # # #    --##...##--    #        #         # # #-# ",
" ..- D    # # # # #   #-## .   ##-#   ####D#####  .. ### #   # # ",
"     #    #       #   ###.     .###   # ~ #    ###   ### ##### # ",
"########### # # # #   ##  . .   .##   #...# .. #.#             # ",
"  ..     .# # # # #   ##     -   ##   #   #    D.    # ######### ",
"      .   D # # # #   ###.......###   ####D#####.#   #           ",
"   - .    # # # # #    -##     ##-    # .      ###   ########### ",
"########### #.#.# #   ######D######   # .  ....#               # ",
"            #!#-# #               ..  #  .     ###....##  ##-### ",
"###   ##############################  ########## ######    ###   ",
"  D      .                                                       ",
//
"                                                                 ",
"                        ###############D####################D#   ",
"  #######################                    # ##......      #   ",
"                        ########D######  #   #  ##....       #   ",
"   ############### ##  ##    ...      #  #   #...##      #####   ",
"   #    ~        # #    #!......      #  #   # .. #####      #   ",
"   #    - ..     # #    #   ###########  #   #  .  ######### #   ",
"   #####  ######## #    #   #         #  #   #...   ##       #   ",
"    .  #  #      # ##########  ########  ########D #### ######   ",
"    .  #  #  ....# #        #  #             #     #  ##     #   ",
"   #####  # ###### #    ....#  #  #########  #     #   #####D#   ",
"   #-    ## #      #######..#  #  #########  #  -  #..           ",
"   #####  #                 #  #             #...  #.....        ",
"       #  ###################  ###############...  ####     #####",
"       #                                                         ",
//
"                      #                                          ",
"########## #####      #         #####D####    #####       ###### ",
"..       # #~..D ######   ##    # ##     #   ## ..##    ### #### ",
"  #### ### ##### #         ##   #  ##### #  ## ### ##   #   #    ",
"  #  # #       # # ##### #  ##  #      # ####  # #..##### - ###  ",
"  #  # ####### ### #...# #   ########  #       # #!.   ....   #  ",
"  #  #       #     #...# #          #  ######### ##############  ",
"  #  ######  #######   # ###   ###  #       ..                #  ",
"  #       #          - # #-    # #  #        ######## ####    #  ",
"  #       #   ########## #.. ....#  # ########        #  ##   ## ",
"  #  -.   #####          ####...##  #    .  .#   ..   #        # ",
"  #             ######      #####   #####.####        #######  # ",
"  #  -    #######....####   ......      #.#       .     #   ## # ",
"  #########             ##### ###########.###############    ### ",
"                                         .                       ",
//
" D D     #-..........#     #   #   #   #                         ",
" # #     #####.#####.# #   #   #   # # #############D########### ",
" # #####  #.-#.#~###.# #   #   #   # # ###      ....    #  #   # ",
" # .   # #D###.......# #   #   #   # # #        . . -.. #. #.# # ",
" # . - # ##. #.####### #   #   #   # # ##################    #   ",
"D####### #  ##.        #       D     #        .              # # ",
". D D -# #   ################################################### ",
"  ###### ##   .                                             .    ",
"  #    D    ###  ###D###D####### ###D###D####### ###D###D####### ",
"  # ######  # # .#    .# #.    # #     # #     # #     # # .   # ",
"  ###   -#  # # .#. - .# #     # #...- # # ..  # #  -..# # .   # ",
"DD#      #  # #  #.   .# #     # #     # #     # # .. .# #     # ",
".######  #    #  #     # #.....# #     # #     # # !   # #     # ",
" #      #######  ####### ### ###D####### ### ###D####### ### ###D",
" # ...                                       ..                  "
};
*/

void make_canned_levels(DmOpenRef iLarnDB)
{
  Err error;
  VoidHand fooRec;
  Ptr p;
  Int i, offset, size, k, row;
  UInt index;
  ULong uniqueID = REC_CANNED;

  index = 1; // DNDSTORE_RECORD == 0
  size = sizeof(Char) * CAN_ROW * CAN_COL;

  for (k = 0, row = 0 ; k < CAN_LEV; k++) {
    fooRec = DmNewRecord(iLarnDB, &index, size);
    p = MemHandleLock(fooRec);
    for (i = 0, offset = 0 ; i < CAN_ROW ; i++, row++) {
      error = DmWrite(p, offset, canned[row], sizeof(Char) * CAN_COL);
      offset += sizeof(Char) * CAN_COL;
    }
    //  error = DmWrite(p, offset, &itm, sizeof(itm));
    MemPtrUnlock(p);
    DmReleaseRecord(iLarnDB, index, true);
    DmSetRecordInfo(iLarnDB, index, NULL, &uniqueID); // set its uid!
    uniqueID++;
    index++;
  }
}
