/*********************************************************************
 * kMoria - Moria adapted for the PalmPilot.                         *
 * Copyright (C) 2000 Bridget Spitznagel                             *
 *                                                                   *
 * This program is derived from Umoria 5.5.2.  Moria is              *
 *   Copyright (c) 1989-94 James E. Wilson, Robert A. Koeneke        *
 *                                                                   *
 * This program is free software; you can redistribute it and/or     *
 * modify it under the terms of the GNU General Public License       *
 * as published by the Free Software Foundation; either version 2    *
 * of the License, or (at your option) any later version.            *
 *                                                                   *
 * This program is distributed in the hope that it will be useful,   *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
 * GNU General Public License for more details.                      *
 *                                                                   *
 * You should have received a copy of the GNU General Public License *
 * along with this program; if not, write to                         *
 * The Free Software Foundation, Inc.,                               *
 * 59 Temple Place - Suite 330,                                      *
 * Boston, MA  02111-1307, USA.                                      *
 *********************************************************************/
#include <Pilot.h>
#include <System/Globals.h> // for font foo
#include "kMoriaRsc.h"
#include "kMoria.h"
#include "lock-externs.h"
// these live in lock.c
extern Short cur_height, cur_width;

extern MoriaPreferenceType my_prefs;

static void put_char_at(Short row, Short col, Char ch, Boolean bold) SEC_6;
static void refresh_2() SEC_6; // refresh top 2 lines only
static void print_stats_large(Boolean siwdcc);
static void print_stats_itsy(Boolean siwdcc);
static void print_stats_real(UChar which_stats, Boolean large) SEC_5;
static void print_stats_hp(Short width_used) SEC_5;
void try_print_status() SEC_5;
static void draw_circle(Short radius, Short x0, Short y0) SEC_5;

// Ok.  I have how much screen?
// Top line sometimes taken by messages
// Bottom two lines taken

/* The screen can have 15 rows, but I took 2 for stats */
//#define visible_h 11
#define visible_h 13
/* You can put about 19.9 m's on a memo line.. so.. we'll have 20 cols */
#define visible_w 20
/* Height in pixels of one character ... in this case an 'M' */
#define visible_char_h 10
/* Width in pixels of one character ... 160 pixels / 20 col = 8 pix/col */
#define visible_char_w 8

// itsy is a smaller font
//#define visible_h_itsy 17
#define visible_h_itsy 22
#define visible_w_itsy 32
#define visible_char_h_itsy 6
#define visible_char_w_itsy 5

Boolean itsy_on = false;

/* This is the offset to convert character-positions in
   the "visible" part of the screen represented by 'terminal'
   to the larger "real" dungeon represented by 'buffer'. */
Short visible_x = 0, visible_y = 0;

// From Ularn's constant.h, these are 6*11 down by 6*33 across.
#define DROWS MAX_HEIGHT
#define DCOLS MAX_WIDTH

// used in print_stats_hp, should agree with print_stats_itsy
#define HP_AT_X 109
#define MANA_AT_X 66

Char terminal[visible_h_itsy][visible_w_itsy];/*relative: model of the screen*/
Char buffer[DROWS][DCOLS]; /* absolute - a model of the dungeon */
Boolean screen_dirty, stats_dirty = true;
Boolean lines_dirty[DROWS]; /* absolute */
Boolean screen_recalced = false;


Boolean words_dirty = true;
Short words_dirty_x[2];
Boolean pending_messages;
Char *old_messages[SAVED_MSGS];
Short last_old_message_shown;
extern UChar need_print_stats;

/*
 * Switch between the small font and the regular font.
 * (Do NOT allow the small font if the SDK that I compiled with
 * is inappropriate for your operating system version.)
 */
Boolean toggle_itsy()
{
  DWord version;
  FtrGet(sysFtrCreator, sysFtrNumROMVersion, &version);
#ifdef I_AM_OS_2
  if (version < 0x03000000l) {
    itsy_on = !itsy_on;
    return true;
  } else itsy_on = false;
#else
  if (version >= 0x03000000l) {
    itsy_on = !itsy_on;
    return true;
  } else itsy_on = false;
#endif
  return false;
}

Boolean stats_bar_on_bottom = true;
UChar DunTopY = 0; // 0 or <DunLineY+1>
UChar DunBotY = 0; // <DunLineY-1> or 160
UChar DunLineY = 133; // where to draw The Line: 133 or 160-133= 27
UChar StatsTopY = 134; // add 4 to be where large stats start... hm
UChar StatsBotY = 160; // add -12 to start itsy stats: 160 or <DunLineY-1>
// note: the 'word' stats will still be printed at DunTopY.

// caller should decide whether to redraw screen.
void set_statsbar_loc()
{
  if (!my_prefs.stats_at_top) {
    DunTopY = 0;
    DunLineY = 133;
    DunBotY = DunLineY-1;
    StatsTopY = DunLineY+1;
    StatsBotY = 160;
  } else {
    StatsTopY = 0;
    DunLineY = 160-133;
    StatsBotY = DunLineY-1;
    DunTopY = DunLineY+1;
    DunBotY = 160;
  }
  stats_bar_on_bottom = !my_prefs.stats_at_top;
}


// I also need a routine to translate a tap on the screen
// into absolute-dungeon coordinates.
// This seems to work.
void where_in_dungeon(Short scr_x, Short scr_y, Short *dun_x, Short *dun_y)
{
  Short vc_w = itsy_on ? visible_char_w_itsy : visible_char_w;
  Short vc_h = itsy_on ? visible_char_h_itsy : visible_char_h;
  if (!stats_bar_on_bottom)
    scr_y -= DunTopY; // need a slight vertical shift
  *dun_x = scr_x / vc_w;
  *dun_y = scr_y / vc_h;
  *dun_x += visible_x;
  *dun_y += visible_y;
  return;
}

// Draw 1 cell on screen (here row,col are NOT screen-relative.)
// For making animatedness. If c = 0, it will set cell to 'original' value
void animate_char(Short y, Short x, Char c, Boolean bold)
{
  Short row, col, ch;
  Short v_h = itsy_on ? visible_h_itsy : visible_h;
  Short v_w = itsy_on ? visible_w_itsy : visible_w;

  if (y < visible_y || x < visible_x ||
      y >= visible_y + v_h || x >= visible_x + v_w)
    return; // the cursor position is not visible on the screen!

  if (c != 0 && FrmGetActiveFormID() == MainForm)
    lines_dirty[y] = screen_dirty = true;// worst case, refresh will clean up

  row = y - visible_y;
  col = x - visible_x;
  
  ch = terminal[row][col];
  if (c == 0) c = ch;

#ifdef I_AM_OS_2
  if (itsy_on)
    FntSetFont(ledFont);
#else
  if (itsy_on)
    FntSetFont(129);
#endif
  put_char_at(row, col, c, bold);
#ifdef I_AM_OS_2
  if (itsy_on)
    FntSetFont(stdFont);
#else
  if (itsy_on)
    FntSetFont(stdFont);
#endif

  terminal[row][col] = ch;
}


/* 
 *  This will clear the part of the display that belongs to
 *  "the dungeon".  It also clears 'terminal' since that is
 *  a representation of the display.
 *  This should be called before the offset for 'terminal' is
 *  moved to show a different part of 'buffer'.
 */
// Actually, the Moria equivalent of this clears the WHOLE screen.
void clear_screen() 
{
  Short i, j;
  Short v_h = itsy_on ? visible_h_itsy : visible_h;
  Short v_w = itsy_on ? visible_w_itsy : visible_w;
  //  RectangleType r;
  if (FrmGetActiveFormID() != MainForm) return;

  /* Clear the physical screen. */
  //  RctSetRectangle(&r, 0, 0, 160, 134-1);
  //  WinEraseRectangle(&r, 0);

  /* Update 'terminal' to represent the cleared screen. */
  screen_dirty = true;
  for (i = 0 ; i < v_h; i++) {
    lines_dirty[i+visible_y] = true;
    for (j = 0; j < v_w; j++) {
      terminal[i][j] = ' ';
    }
  }
}

void print(Short ch, Short row, Short col) {
  buffer[row][col] = ch;
  lines_dirty[row] = true;
  screen_dirty = true;
}



/*
 * Draw the whole screen, as the player knows it.
 * On second thought, maybe update but DON'T draw.
 */
void recalc_screen()
{
  Short row, col;// absolute
  if (FrmGetActiveFormID() != MainForm) return;
  clear_screen();
  for (row = 0 ; row < cur_height ; row++)
    for (col = 0 ; col < cur_width ; col++)
      buffer[row][col] = loc_symbol(row, col);
  //      buffer[row][col] = '^';
  //  for (row = visible_y, vr = 0 ; vr < visible_h_itsy ; row++, vr++)
  //    for (col = visible_x, vc = 0 ; vc < visible_w_itsy ; col++, vc++)
  //      terminal[vr][vc] = buffer[row][col];
  /*refresh();*/
  screen_recalced = true;
}
/*
 * The funtion to actually draw a character ch at a col,row position
 * (row, col are RELATIVE i.e. this is used in model-of-screen)
 * note the row+1 offset is so that we have a line for msgs.
 * The character will be drawn in either the small or normal font.
 */
static void put_char_at(Short row, Short col, Char ch, Boolean bold)
{
  //  Short cheat, vcheat = itsy_on ? 14 : 5; // center the dungeon vertically
  //Short cheat, vcheat = itsy_on ? 0 : 0; // center the dungeon vertically
  Short cheat, vcheat = stats_bar_on_bottom ? 0 : DunTopY;
  /* the space has already been cleared with clear_line */
  /* on second thought, no it hasn't. */
  RectangleType r;
  Short vc_w = itsy_on ? visible_char_w_itsy : visible_char_w;
  Short vc_h = itsy_on ? visible_char_h_itsy : visible_char_h;

  RctSetRectangle(&r, col * vc_w, row*vc_h+vcheat,
		  vc_w, vc_h); /* Left, Top, W, H */
  WinEraseRectangle(&r, 0); /* 0 for square corners */
    
  // calculate pixel position of "row, col" and put char there
  cheat = vc_w - FntCharWidth(ch); // center the variable width characters

  if (cheat <= 1)   cheat = 0;
  else              cheat /= 2;
  if (ch != ' ') {
    if (!itsy_on && (ch== 'g' || ch== 'j' || ch== 'p' ||ch == 'q' ||ch == 'y'))
      // unfortunately, letters with dangling bits are a pain.
      WinDrawChars(&ch, 1, col * vc_w + cheat, row * vc_h+vcheat-1);
    else
      WinDrawChars(&ch, 1, col * vc_w + cheat, row * vc_h+vcheat);
    if (bold)  WinInvertRectangle(&r, 0); /* 0 for square corners */
  }
  terminal[row][col] = ch;

}

/*
 *  Actually redraw the screen.  (if it is 'dirty'.)
 */
static void refresh_tail() SEC_5;
extern Char last_command; // use this to un-show "detect <foo>" creatures 
void refresh(Boolean clearall)
{
  Short col, line, absline;
  Short v_h = itsy_on ? visible_h_itsy : visible_h;
  Short v_w = itsy_on ? visible_w_itsy : visible_w;
  Short vc_h = itsy_on ? visible_char_h_itsy : visible_char_h;
  Short vcheat = stats_bar_on_bottom ? 0 : DunTopY;
  RectangleType r;
  if (last_command == 'I') return; // don't overwrite nifty detected creatures
  if (FrmGetActiveFormID() != MainForm) return;
  if (!screen_dirty) return;
  screen_dirty = false;
  if (screen_recalced) { recalc_and_refresh(); return; }

#ifdef I_AM_OS_2
  if (itsy_on)
    FntSetFont(ledFont);
#else
  if (itsy_on)
    FntSetFont(129);
#endif
  v_w = min(v_w, cur_width  - visible_x);
  v_h = min(v_h, cur_height - visible_y);

  for (line = 0 ; line < v_h ; line++) {
    /* line is RELATIVE, line+visible_y is ABSOLUTE. */
    absline = line + visible_y;
    if (!lines_dirty[absline]) continue;
    //// this rectangle's y coords must match with put_char_at's...
    if (clearall) {
      RctSetRectangle(&r, 0, line*vc_h+vcheat, 160, vc_h);
      WinEraseRectangle(&r, 0);
    }
    //// ...ok maybe that worked
    lines_dirty[absline] = false;
    for (col = 0; col < v_w; col++) {
      /* col is RELATIVE, col+visible_x is ABSOLUTE. */
      if (col+visible_x >= cur_width) break;
      if (buffer[absline][col+visible_x] != terminal[line][col]) {
	put_char_at(line, col, buffer[absline][col+visible_x], false);
      }
    }
  }
  refresh_tail();
}
// This is an "optimized" (redundant bits taken out) combo of recalc + refresh
// to spped up the transition from screen to screen I hope
void recalc_and_refresh()
{
  Short col, line, absline;
  Short v_h = itsy_on ? visible_h_itsy : visible_h;
  Short v_w = itsy_on ? visible_w_itsy : visible_w;
  //  Short vc_h = itsy_on ? visible_char_h_itsy : visible_char_h;
  //  RectangleType r;
  if (FrmGetActiveFormID() != MainForm &&
      FrmGetActiveFormID() != LookForm) return;
  screen_recalced = false;
  screen_dirty = false;

#ifdef I_AM_OS_2
  if (itsy_on)
    FntSetFont(ledFont);
#else
  if (itsy_on)
    FntSetFont(129);
#endif
  if (!itsy_on) {
    // erase that last 1-2 pixels of turds
    RectangleType r;
    RctSetRectangle(&r, 0, DunBotY-2, 160, 2);
    WinEraseRectangle(&r, 0);
  }

  v_w = min(v_w, cur_width  - visible_x);
  v_h = min(v_h, cur_height - visible_y);

  for (line = 0 ; line < v_h ; line++) {
    /* line is RELATIVE, line+visible_y is ABSOLUTE. */
    absline = line + visible_y;
    //// this rectangle's y coords must match with put_char_at's...
    //    RctSetRectangle(&r, 0, line*vc_h, 160, vc_h);
    //    WinEraseRectangle(&r, 0);
    //// ...ok cleared
    lines_dirty[absline] = false; // keep this
    for (col = 0; col < v_w; col++) {
      /* col is RELATIVE, col+visible_x is ABSOLUTE. */
      if (col+visible_x >= cur_width) break;
      buffer[absline][col+visible_x] = loc_symbol(absline, col+visible_x);
      put_char_at(line, col, buffer[absline][col+visible_x], false);
      // (put_char_at takes care of setting terminal[][].)
    }
  }
  refresh_tail();
}

// clear EVERYTHING, and redraw dungeon and stats
void nuke_display()
{
  RectangleType r;
  RctSetRectangle(&r, 0, 0, 160, 160);
  WinEraseRectangle(&r, 0);
  get_panel(char_row, char_col, true); // FORCE recalculation of panel
  recalc_and_refresh(); //	recalc_screen(); refresh(true);
  stats_dirty = true;
  print_stats_real(STATS_QUO, true);
}

static void refresh_tail()
{
  //  WinDrawLine(0,12,160,12);
  //  spell_stats(env); // don't forget the spell-status on right hand side.
#ifdef I_AM_OS_2
  if (itsy_on)
    FntSetFont(stdFont);
#else
  if (itsy_on)
    FntSetFont(stdFont);
#endif
  /* I might eventually have a pref to draw run/walk circles. This code works*/
  if (my_prefs.overlay_circles && !my_prefs.relative_move) {
    Short center_x = 80, center_y = 80;
    /*
    if (my_prefs.relative_move) {
      // eh, probably this will be right.  Center it around the '@'
      center_x = (char_col - visible_x) * (itsy_on ? visible_char_w_itsy :
					   visible_char_w);
      center_y = (char_row - visible_y) * (itsy_on ? visible_char_h_itsy :
					   visible_char_h);
      // Come to think of it this will make the screen look like utter crap.
      // Ok - I will not allow circle overlay && relative_move until I
      // figure out how to resolve this sad little issue.
      // Also rewrite this to use relativize_move.
    }
    */
    draw_circle(my_prefs.walk_center_border * 10, center_x, center_y);
    if (my_prefs.run)
      draw_circle(my_prefs.run_walk_border * 10, center_x, center_y);
  }
  // oh yeah, this might be a good idea.  testing........... 
  try_print_status();
}

static void refresh_2()
{
  Short col, line, absline;
  Short v_w = itsy_on ? visible_w_itsy : visible_w;
  if (FrmGetActiveFormID() != MainForm) return;

#ifdef I_AM_OS_2
  if (itsy_on)
    FntSetFont(ledFont);
#else
  if (itsy_on)
    FntSetFont(129);
#endif
  v_w = min(v_w, cur_width  - visible_x);

  for (line = 0 ; line < 2 ; line++) {
    /* line is RELATIVE, line+visible_y is ABSOLUTE. */
    absline = line + visible_y;
    lines_dirty[absline] = false;
    // Draw only from the edge of the status words.  Ha.
    for (col = words_dirty_x[line]; col < v_w; col++) {
      /* col is RELATIVE, col+visible_x is ABSOLUTE. */
      if (col+visible_x >= cur_width) break;
      // if (buffer[absline][col+visible_x] != terminal[line][col]) // no..
      put_char_at(line, col, buffer[absline][col+visible_x], false);
    }
  }
#ifdef I_AM_OS_2
  if (itsy_on)
    FntSetFont(stdFont);
#else
  if (itsy_on)
    FntSetFont(stdFont);
#endif
  try_print_status();
}

/*
 * Given input SCREEN coordinates x,y, with center of screen at 0,0,
 * Translate x,y such that the onscreen location of '@' is mapped to 0,0
 */
void relativize_move(Short *x, Short *y)
{
  Short center_x, center_y, vcheat = stats_bar_on_bottom ? 0 : DunTopY;
  center_x = (char_col - visible_x) * (itsy_on ? visible_char_w_itsy :
				       visible_char_w);
  center_y = (char_row - visible_y) * (itsy_on ? visible_char_h_itsy :
				       visible_char_h) + vcheat;
  *x = (*x + 80) - center_x;
  *y = (*y + 80) - center_y;
}


// y,x are ABSOLUTE row,col.
void change_visible_xy(Short x, Short y)
{
  y = max(y, 0);
  x = max(x, 0);
  y = min(y, cur_height - (itsy_on ? visible_h_itsy : visible_h));
  x = min(x, cur_width - (itsy_on ? visible_w_itsy : visible_w));
  if (visible_y == y && visible_x == x) return;  // no change!
  visible_y = y;
  visible_x = x;
  recalc_and_refresh(); //  recalc_screen(); refresh(true);

  // this is like panel_bounds()
  panel_row_min = visible_y;
  panel_row_max = visible_y + (itsy_on ? visible_h_itsy : visible_h) - 1;
  panel_col_min = visible_x;
  panel_col_max = visible_x + (itsy_on ? visible_w_itsy : visible_w) - 1;

}
// x and y can be -1, 0, 1.
static Short dirx[9] = { -1,  0,  1, -1,  0,  1, -1,  0,  1 };
static Short diry[9] = {  1,  1,  1,  0,  0,  0, -1, -1, -1 };
void slide_dir(Short dir)
{
  Short x, y;
  dir = max(min(dir,9),1); // keep in bounds [1,9]
  x = dirx[dir-1];
  y = diry[dir-1];
  y *= (itsy_on ? visible_h_itsy : visible_h);
  x *= (itsy_on ? visible_w_itsy : visible_w);
  change_visible_xy(x+visible_x,y+visible_y);
}


Boolean get_panel(Short y, Short x, Boolean force)
{
  Short screen_width = itsy_on ? visible_w_itsy : visible_w; // unit = cells
  Short screen_height = itsy_on ? visible_h_itsy : visible_h; // unit = cells
  Short epsilon_x = 3, epsilon_y = 2; // # of border cells you "can't get to"
  Short new_visible_x, new_visible_y;
  Boolean in_new_panel = true;
  // cur_height and cur_width are the # of cells in CURRENT floor's dimensions
  if (!(force ||
	(x < visible_x + epsilon_x) ||
	(x >= visible_x + screen_width - epsilon_x) ||
	(y < visible_y + epsilon_y) ||
	(y >= visible_y + screen_height - epsilon_y)))
    return false; // no need to change panels

  // CALCULATE X   of top left corner of the panel to move to
  // This is a little ugly in order to deal with possibly odd screen_width.
  new_visible_x = ((x - epsilon_x)/screen_width) * screen_width;
  panel_col = ((x - epsilon_x)/screen_width) * 2; // panels overlap by width/2
  if ((x - epsilon_x) % screen_width >= screen_width/2) {
    new_visible_x += screen_width/2;
    panel_col++;
  }
  // Boundary conditions.
  if (new_visible_x > cur_width - screen_width)
    new_visible_x = cur_width - screen_width;
  if (new_visible_x < 0)
    new_visible_x = 0;
  // CALCULATE Y   of top left corner of the panel to move to
  // This is a little ugly in order to deal with possibly odd screen_width.
  new_visible_y = ((y - epsilon_y)/screen_height) * screen_height;
  panel_row = ((y - epsilon_y)/screen_height) * 2; // panels overlap by h/2
  if ((y - epsilon_y) % screen_height >= screen_height/2) {
    new_visible_y += screen_height/2;
    panel_row++;
  }
  // Boundary conditions.
  if (new_visible_y > cur_height - screen_height)
    new_visible_y = cur_height - screen_height;
  if (new_visible_y < 0)
    new_visible_y = 0;
  
  if (visible_x == new_visible_x && visible_y == new_visible_y)
    in_new_panel = false; // Ha!  There was no change.  Silly rabbit.

  visible_x = new_visible_x;
  visible_y = new_visible_y;
  panel_row_min = visible_y;
  panel_row_max = visible_y + screen_height - 1;
  panel_col_min = visible_x;
  panel_col_max = visible_x + screen_width - 1;

  if (in_new_panel)
    if (find_bound)
      end_find(); // stop running, etc.

  // Caller decides whether to recalc_screen & refresh.

  return in_new_panel;
}


/*
 * Set the buffer value to what it "should" be "now"
 * but it will get actually drawn when the screen is refreshed
 */
//void update_screen_cell(Short j, Short i, struct everything *env)


/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
// Here is the code formerly known as stats.c.
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////

/*
  S00 I00 W00 D00 C00 Ch00  0000 feet
  L 00 xp 0123456 mana 000 hp 000/000
  AC 00 Gold ...
*/

// HEY - revise this so that it just queues up the messages,
// and have another routine print them at the end of the turn
// (with --more-- as necessary)
void count_message(Char *buf)
{
  Short i;
  i = command_count;
  message(buf);
  command_count = i;
}
void message(Char *buf)
{
  Word curfrm = FrmGetActiveFormID();
  Short i;
  Char *tmp;
  if (curfrm == MainForm ||
      curfrm == CharSummaryForm) {    // Log a message for later display
    if (!buf) return;
    tmp = old_messages[0];
    for (i = 0 ; i < SAVED_MSGS-1 ; i++)
      old_messages[i] = old_messages[i+1];
    StrNCopy(tmp, buf, SAVED_MSG_LEN-2); /* leave room for \n.. */
    tmp[SAVED_MSG_LEN-2] = '\0'; /* making sure it's terminated.. */
    StrCat(tmp, "\n");
    old_messages[SAVED_MSGS-1] = tmp;
    // this part is for "--more--" ability:
    if ((command_count <= 0) ||
	(0 != StrCompare(old_messages[SAVED_MSGS-1],
			 old_messages[SAVED_MSGS-2])))
      // don't --more-- a count_message if it's the same as the previous one
      last_old_message_shown--;
    if (last_old_message_shown < 0) // "none of these msgs have been shown"
      last_old_message_shown = -1;
    // Remember that we have something to display
    pending_messages = true;
    command_count = 0;  // (lame hack!)
  } else if (curfrm == InvForm || curfrm == EquForm) {
    // These messages should be printed straightaway.
    // well, really I should be logging them, too.  in most cases.
    RectangleType r;
    RctSetRectangle(&r,0,128,156,11); /* left,top, width and height */
    WinEraseRectangle(&r, 0); /* 0 for square corners */
    if (buf)
      WinDrawChars(buf, StrLen(buf), 1, 128);
  }
}
// a mainform message that we require to be displayed in "real time"
void alert_message(Char *buf)
{
  RectangleType r;
  if (FrmGetActiveFormID() != MainForm || !buf) return;
  //  RctSetRectangle(&r,0,134,160,160-134); /* left,top, width and height */
  RctSetRectangle(&r,0,StatsTopY,160,StatsBotY-StatsTopY); /* x,y, W, and H */
  WinEraseRectangle(&r, 0); /* 0 for square corners */
  //  if (buf)
  //    WinDrawChars(buf, StrLen(buf), 1, 134);
  WinDrawChars(buf, StrLen(buf), 1, StatsTopY);
}
// Clear the message area
Boolean message_clear()
{
  RectangleType r;
  //  RctSetRectangle(&r,0,134,160,160-134); /* left,top, width and height */
  RctSetRectangle(&r,0,StatsTopY,160,StatsBotY-StatsTopY); /* x,y, W, and H */
  if (false)
    WinEraseRectangle(&r, 0);
  //  WinDrawLine(0,133,160,133);
  WinDrawLine(0,DunLineY,160,DunLineY);
  //print_stats(STATS_QUO);
  if (last_old_message_shown < SAVED_MSGS-1)  return true;
  else return false;
}

/*
 * Blow away the message queue
 * (Use this when you die)
 */
void preempt_messages()
{
  pending_messages = false;
  last_old_message_shown = SAVED_MSGS - 1;  
}

// Print up to 2 lines of messages.  make sure you word-wrap.
// 1. print first message
// 2. if necessary, word wrap
// 3. else if second message exists + does not need word wrap, print it too.
// 4. else if second message does not exist, print teensy stats.
// 5. print "--more--" if any more messages are in the queue
void show_messages()
{
  Short lines_avail = 2, len_line, len_str, lines_used = 0;
  Short str_width = 160;
  Char *msg;

  RectangleType r;
  //RctSetRectangle(&r,0,134,160,160-134); /* left,top, width and height */
  RctSetRectangle(&r,0,StatsTopY,160,StatsBotY-StatsTopY); /* x,y, W, and H */

  // if there's nothing to print, just print stats.
  if (last_old_message_shown >= SAVED_MSGS-1) {
    need_print_stats &= ~STATS_WORD;
    print_stats_real(need_print_stats, true);
    need_print_stats = STATS_QUO;
  } else {
    WinEraseRectangle(&r, 0);
    //    WinDrawLine(0,133,160,133);
    WinDrawLine(0,DunLineY,160,DunLineY);
    while (lines_avail > 0 && last_old_message_shown < SAVED_MSGS-1) {
      msg = old_messages[last_old_message_shown+1];
      len_str = StrLen(msg);
      if (lines_avail <= 1 && FntCharsWidth(msg, len_str-1) > SCR_WIDTH)
	break; // 1 line avail., message is 2 lines.  Do not print partial msg.
      while (lines_avail > 0) {
	len_str = StrLen(msg);
	len_line = FntWordWrap(msg, SCR_WIDTH);
	if (len_line >= len_str) { // we have finished the string!
	  //	  WinDrawChars(msg, len_str-1, 0, 134 + 11*lines_used);
	  WinDrawChars(msg, len_str-1, 0, StatsTopY + 11*lines_used);
	  str_width = FntCharsWidth(msg, len_str-1); // remember for print..hp!
	  lines_avail--; lines_used++;
	  break;
	}
	//	WinDrawChars(msg, len_line, 0, 134 + 11*lines_used);
	WinDrawChars(msg, len_line, 0, StatsTopY + 11*lines_used);
	str_width = FntCharsWidth(msg, len_line); // remember for print_s._hp!
	lines_avail--; lines_used++;
	msg += len_line;
      }    // we finished the string, or we ran out of space.
      last_old_message_shown++;
    } // finished all messages or ran out of space.
    if ((last_old_message_shown < SAVED_MSGS-1) || (last_command == 'I'))
      show_a_more(lines_used); // prompt for the Any Key with "--more--"
    else {
      if (lines_used > 1)	// see if there was room for hit points
	print_stats_hp(str_width); // this will decide whether to print them
    }
    if (lines_used < 2) {
      need_print_stats &= ~STATS_WORD;
      print_stats_real(need_print_stats, false);
    }
    else stats_dirty = true;
  }
  pending_messages = false;
  need_print_stats = STATS_QUO;
}
// This is what prints the MORE in lower right corner of screen.
// A couple non-display.c routines also would like to use it..
void show_a_more(Short lines_used)
{
  DWord version;
  Boolean large = false;
  FtrGet(sysFtrCreator, sysFtrNumROMVersion, &version);
#ifdef I_AM_OS_2
  if (version < 0x03000000l) FntSetFont(ledFont);
  else large = true;
#else
  if (version >= 0x03000000l) FntSetFont(129);
  else large = true;
#endif
  //    WinDrawChars("--more--", 8, 120, 134 + 11*lines_used);
  //WinDrawChars("MORE", 8, (large ? 130 : 140), 134 + 11*lines_used - 3);
  WinDrawChars("MORE", 8, (large ? 130 : 140), StatsTopY + 11*lines_used - 3);
#ifdef I_AM_OS_2
  FntSetFont(stdFont);
#else
  FntSetFont(stdFont);
#endif
}

// call alloc_message_log initially when all stuff is being allocated..
// it has to be dynamic so we can play games with pointers in log_message.
void alloc_message_log()
{
  Short i;
  for (i = 0 ; i < SAVED_MSGS ; i++) {
    old_messages[i] = (Char *) md_malloc(sizeof(Char) * SAVED_MSG_LEN);
    old_messages[i][0] = '\0';
  }
  // the log is all null strings.. so (we pretend) all of it has been shown
  last_old_message_shown = SAVED_MSGS - 1;
}
void clear_message_log()
{
  Short i;
  for (i = 0 ; i < SAVED_MSGS ; i++)
    old_messages[i][0] = '\0';
  // the log is all null strings.. so (we pretend) all of it has been shown
  last_old_message_shown = SAVED_MSGS - 1;
}


/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////
static Boolean verbing() SEC_6;
static Boolean verbing()
{
  return ((pyflags.paralysis > 1) ||
	  (PY_REST & pyflags.status) ||
	  (command_count > 0) ||
	  (PY_SEARCH & pyflags.status));
}

// See prt_stat_block() in misc3.c .. there are other things I should print
void print_stats(UChar which_stats)
{
  need_print_stats = which_stats;
  if (which_stats & STATS_WORD) words_dirty = true;
  stats_dirty = true;
  return; // let's see if this sorta works.
}
static void print_stats_real(UChar which_stats, Boolean large)
{
  static Boolean was_siwdcc = true;
  Boolean change = true;
  /*  Short voteStrInt = 0, voteGoldAc = 0; */
  if (FrmGetActiveFormID() != MainForm) return;
  /*
  // decide what to show for first line of stats
  // I can add more if statements and stuff, should we have a 3rd stats option
  if (which_stats & STATS_STR_ETC)
    voteStrInt++; //   was_siwdcc = true;
  if (which_stats & STATS_GOLD_AC)
    voteGoldAc++; //   was_siwdcc = false;    
  if (which_stats & STATS_TOGGLE) {
    if (!was_siwdcc) voteStrInt++;
    else             voteGoldAc++;
  }
  if (which_stats & STATS_QUO)
    ; // we only care about refreshing the bottom line

  // Ok, the vote is taken.  You must have a majority to win....
  if (voteStrInt > voteGoldAc)
    was_siwdcc = true;
  else if (voteStrInt > voteGoldAc)
    was_siwdcc = false;
  else
    ;  // Here I add a bias for sameness: if there is a tie, do not change!
  */
  if (words_dirty || verbing()) {
    update_status_words(); // refresh the top 2 lines, then (re)print words 
    words_dirty = false;
  }
  if ((which_stats & STATS_STR_ETC) && (which_stats & STATS_GOLD_AC))
    was_siwdcc = !was_siwdcc; // it will get the user's attention, I reckon
  else if (which_stats & STATS_STR_ETC)
    was_siwdcc = true;
  else if (which_stats & STATS_GOLD_AC)
    was_siwdcc = false;
  else if (which_stats == STATS_TOGGLE)
    was_siwdcc = !was_siwdcc;
  else if (which_stats & STATS_HP)
      ; // don't toggle
  else change = false;
  if (change) stats_dirty = true; // better force it to redraw

  if (large) {
    print_stats_large(was_siwdcc);
  } else
    print_stats_itsy(was_siwdcc);
}

/*
 * Print two lines of status at the bottom of screen
 */
Short statwidth_large[6] = { 5,7,8,6,5,10 };// also, width of "%2d " is 12
static void print_stats_large(Boolean siwdcc)
{
  Char buf[80];
  Short hi, x, i, tmp, w, depth;
  RectangleType r;

  if (FrmGetActiveFormID() != MainForm) return;
  if (!stats_dirty) return;
  //  RctSetRectangle(&r,0,134,160,160-134); /* left,top, width and height */
  RctSetRectangle(&r,0,StatsTopY,160,StatsBotY-StatsTopY); /* x,y, W, and H */
  WinEraseRectangle(&r, 0); /* 0 for square corners */
  //  WinDrawLine(0,133,160,133);
  WinDrawLine(0,DunLineY,160,DunLineY);

  // Tap bottom line of screen to toggle what the first line says
  
  //hi = 138;
  hi = StatsTopY + 4;

  x = 1;
  if (siwdcc) { // I need to do this differently.
    // need to print each number at a fixed position, also
    // subtract 18 if > 18 (and print '00' if 100) and invert the display.
    //    StrPrintF(buf, "S00 In00 W00 D00 C00 Ch00");
    // '\x80' is the whitespace with width 5 (same as a digit).
    //    StrPrintF(buf,
    //"S\x80\x80 In\x80\x80 W\x80\x80 D\x80\x80 C\x80\x80 Ch\x80\x80 ");
    StrPrintF(buf, "S      In      W      D      C      Ch     ");
    WinDrawChars(buf, StrLen(buf), x, hi);
    for (i = 0, w = x ; i < 6 ; i++, w+=12) {
      w += statwidth_large[i];
      tmp = pystats.use_stat[i];
      if (tmp > 18) {
	tmp -= 18;
	if (tmp > 99) tmp = 0;
	StrPrintF(buf, (tmp > 9 ? "%d" : "0%d"), tmp);
	WinDrawInvertedChars(buf, StrLen(buf), w, hi);
      } else {
	StrPrintF(buf, "%d", tmp);
	WinDrawChars(buf, StrLen(buf), (tmp > 9) ? w : w+5, hi);
      }
    }
  } else {
    StrPrintF(buf, "AC %d  $ %ld", pymisc.dis_ac, pymisc.au);
    WinDrawChars(buf, StrLen(buf), x, hi);
  }

  x += 115;
  depth = dun_level * 50;
  if (depth > 0)
    StrPrintF(buf, "%d feet", depth);
  else 
    StrPrintF(buf, "Town");
  WinDrawChars(buf, StrLen(buf), x, hi);
  // done with first line
  x = 1;
  hi += 11;

  // player level 12
  StrPrintF(buf, "L %d", (Short) pymisc.lev);
  WinDrawChars(buf, StrLen(buf), x, hi);
  x += 20;
  // player experience 1234567
  StrPrintF(buf, "x %ld", pymisc.exp);
  WinDrawChars(buf, StrLen(buf), x, hi);
  x += 45;
  // mana points 123
  StrPrintF(buf, "mana %d", pymisc.cmana);
  WinDrawChars(buf, StrLen(buf), x, hi);
  x += 43;
  // hit points 123/123
  //  StrPrintF(buf, "hp %d/%d", pymisc.mhp, pymisc.chp);
  StrPrintF(buf, "hp %d/%d", pymisc.chp, pymisc.mhp);
  if (pymisc.chp < pymisc.mhp / 4) // might want to make this tweakable
    WinDrawInvertedChars(buf, StrLen(buf), x, hi);
  else
    WinDrawChars(buf, StrLen(buf), x, hi);

  // xxxxxxxxxxx also need FLAGS: hungry/weak, blind, confused, fear, poisoned,
  // and also search/rest/speed/study!
  stats_dirty = false;
  return;
}

static void print_stats_itsy(Boolean siwdcc)
{
  Char buf[80];
  Short hi, x, depth;
  Short i, w, tmp;

  if (FrmGetActiveFormID() != MainForm) return;

#ifdef I_AM_OS_2
  FntSetFont(ledFont);
#else
  FntSetFont(129);
#endif
  
  //  hi = 160-12;
  hi = StatsBotY - 12;

  x = 1;
  if (siwdcc) {
    // Dude!  Fix this!  Do that same inverse-video stats thing too.
    /*
    StrPrintF(buf, "S%d In%d W%d D%d C%d Ch%d", 
	      pystats.max_stat[0],
	      pystats.max_stat[1],
	      pystats.max_stat[2],
	      pystats.max_stat[3],
	      pystats.max_stat[4],
	      pystats.max_stat[5]);
    WinDrawChars(buf, StrLen(buf), x, hi);
    */
    StrPrintF(buf, "S      In      W      D      C      Ch     ");
    //    StrPrintF(buf, "S00 In00 W00 D00 C00 Ch00");
    // '\x80' is the whitespace with width 5 (same as a digit).
    //StrPrintF(buf,
    //     "S\x80\x80 In\x80\x80 W\x80\x80 D\x80\x80 C\x80\x80 Ch\x80\x80 ");
    WinDrawChars(buf, StrLen(buf), x, hi);
    for (i = 0, w = 5 ; i < 6 ; i++, w+=3*5+2) {
      // each character in this font is 5 pixels wide.  well, except ' ' is 2.
      if (i == 1 || i == 5) w += 5; // extra letter for In and Ch
      tmp = pystats.use_stat[i];
      if (tmp > 18) {
	tmp -= 18;
	if (tmp > 99) tmp = 0;
	StrPrintF(buf, (tmp > 9 ? "%d" : "0%d"), tmp);
	WinDrawInvertedChars(buf, StrLen(buf), w, hi);
      } else {
	StrPrintF(buf, "%d", tmp);
	WinDrawChars(buf, StrLen(buf), (tmp > 9) ? w : w+5, hi);
      }
    }
  } else {
    StrPrintF(buf, "AC %d  $ %ld", pymisc.dis_ac, pymisc.au);
    WinDrawChars(buf, StrLen(buf), x, hi);
  }

  x += 115;
  depth = dun_level * 50;
  if (depth > 0)
    StrPrintF(buf, "%d feet", depth);
  else 
    StrPrintF(buf, "Town");
  WinDrawChars(buf, StrLen(buf), x, hi);
  // done with first line

  x = 1;
  // hi = 160-6;
  hi = StatsBotY - 6;

  // player level 12
  StrPrintF(buf, "L %d", (Short) pymisc.lev);
  WinDrawChars(buf, StrLen(buf), x, hi);
  x += 20;
  // player experience 1234567
  StrPrintF(buf, "/ %ld", pymisc.exp);
  WinDrawChars(buf, StrLen(buf), x, hi);
  x = MANA_AT_X; // same as: x += 45;
  // mana points 123
  StrPrintF(buf, "Mana %d", pymisc.cmana);
  WinDrawChars(buf, StrLen(buf), x, hi);
  x = HP_AT_X; // same as: x += 43;
  // hit points 123/123
  //  StrPrintF(buf, "HP %d/%d", pymisc.mhp, pymisc.chp);
  StrPrintF(buf, "HP %d/%d", pymisc.chp, pymisc.mhp);
  if (pymisc.chp < pymisc.mhp / 4) // might want to make this tweakable
    WinDrawInvertedChars(buf, StrLen(buf), x, hi);
  else
    WinDrawChars(buf, StrLen(buf), x, hi);

  // this is a test
  //  try_print_status();

#ifdef I_AM_OS_2
  FntSetFont(stdFont);
#else
  FntSetFont(stdFont);
#endif

  stats_dirty = true;
  return;
}

// This duplicates part of the itsy-stats functionality.  "plugh"
// It will print mana if possible, and as much of hp as it can
// (at the very least it will print the current hp number).
// Inefficient (but slowness will not matter because we only do this
// once and then wait for user action) and/or inelegant at present.
static void print_stats_hp(Short width_used)
{
  Short x, w, len, hi = StatsBotY - 6; //160-6;
  Boolean low; 
  Char buf[20];
  if (FrmGetActiveFormID() != MainForm) return;
  //if (width_used > HP_AT_X - 2) return;
#ifdef I_AM_OS_2
  FntSetFont(ledFont);
#else
  FntSetFont(129);
#endif

  low = pymisc.chp < pymisc.mhp / 4; // might want to make this tweakable

  if (width_used <= MANA_AT_X - 2) {
    StrPrintF(buf, "Mana %d", pymisc.cmana);
    WinDrawChars(buf, StrLen(buf), MANA_AT_X, hi);
  }
  StrPrintF(buf, "HP %d/%d", pymisc.chp, pymisc.mhp);
  len = StrLen(buf);
  w = FntCharsWidth(buf, len);
  // If possible print at usual position, otherwise try: full string right-jus;
  // full string partly stacked; just numbers stacked.
  if (width_used < HP_AT_X - 2)
    x = HP_AT_X;
  else if (width_used < 160-(w+2))
    x = 160 - w; // print full string but right justified
  else {
    StrPrintF(buf, "%d/%d", pymisc.chp, pymisc.mhp);
    len = StrLen(buf);
    w = FntCharsWidth(buf, len);
    if (width_used < 160-(w+2)) {
      x = 160 - w; // print both numbers but stack 'HP' above
      WinDrawChars("HP", 2, x, hi);
    } else {
      StrPrintF(buf, "%d", pymisc.chp);  // stack cur above max. no 'HP'.
      len = StrLen(buf);
      x = 160 - 5*len;//==160-FntCharsWidth(buf, len);
      if (low)
	WinDrawInvertedChars(buf, len, x, hi-6);
      else
	WinDrawChars(buf, len, x, hi-6);
      StrPrintF(buf, "%d", pymisc.mhp);
      len = StrLen(buf);
      x = 160 - 5*len;//==160-FntCharsWidth(buf, len);
    }
  }
  if (low)
    WinDrawInvertedChars(buf, len, x, hi);
  else
    WinDrawChars(buf, len, x, hi);


#ifdef I_AM_OS_2
  FntSetFont(stdFont);
#else
  FntSetFont(stdFont);
#endif
}

// Currently this is called by refresh()
// and currently it prints at the very top of the screen.
// these are the one-word status states that were on the bottom line in moria.
// Note:
// If something other than 'refresh' calls this, it would be a good
// idea to re-print the top 2 lines of the dungeon to 'clean up' FIRST.
// in case there's a word we are no longer printing.
// (actually it could set those 2 lines to 'dirty' and refresh())
void update_status_words()
{
  try_print_status();
  refresh_2();
}
void try_print_status()
{
  Char buf[80];
  Short hi, x, i, maxlen=160;//maxlen=120;
  //  Short absline;

  if (FrmGetActiveFormID() != MainForm) return;

#ifdef I_AM_OS_2
  FntSetFont(ledFont);
#else
  FntSetFont(129);
#endif
  // Test of stats flags:
  // PoisonedSearchingVslowStudy
  // HungryBlindConfusedAfraid
  //  hi = 0;
  hi = DunTopY;
  x = 0;
  words_dirty_x[0] = words_dirty_x[1] = 0;
  // hunger
  if (PY_WEAK & pyflags.status) {
    StrPrintF(buf, "Weak");
    WinDrawInvertedChars(buf, StrLen(buf), x, hi);
    x += 20;
  } else if (PY_HUNGRY & pyflags.status) {
    StrPrintF(buf, "Hungry"); // 30
    WinDrawInvertedChars(buf, StrLen(buf), x, hi);
    x += 30;
  }
  // blindness
  if (PY_BLIND & pyflags.status) {
    StrPrintF(buf, "Blind"); // 25
    WinDrawInvertedChars(buf, StrLen(buf), x, hi);
    x += 25;
  }
  if (PY_CONFUSED & pyflags.status) {
    StrPrintF(buf, "Confused"); // 40
    WinDrawInvertedChars(buf, StrLen(buf), x, hi);
    x += 40;
    if (x > maxlen) { words_dirty_x[0] = x; x = 0; hi += 6; }
  }
  if (PY_FEAR & pyflags.status) {
    StrPrintF(buf, "Afraid"); // 30
    WinDrawInvertedChars(buf, StrLen(buf), x, hi);
    x += 30;
    if (x > maxlen) { words_dirty_x[0] = x; x = 0; hi += 6; }
  }
  if (PY_POISONED & pyflags.status) {
    StrPrintF(buf, "Poisoned"); // 40
    WinDrawInvertedChars(buf, StrLen(buf), x, hi);
    x += 40;
    if (x > maxlen) { words_dirty_x[0] = x; x = 0; hi += 6; }
  }
  // this needs to print Searching, Resting, Paralysis, or 'count' status :
  {
    Boolean display_counts = true;
    buf[0] = '\0';
    if (pyflags.paralysis > 1)
      StrPrintF(buf, "Paralysed"); // 45
    else if (PY_REST & pyflags.status) {
      if (pyflags.rest < 0)
	StrPrintF(buf, "Rest *");
      else if (display_counts)
      	StrPrintF(buf, "Rest%s%d", (pyflags.rest>9?" ":" \t"), pyflags.rest);
      else
	StrPrintF(buf, "Rest");
    } else if (command_count > 0) {
      if (display_counts)
      	StrPrintF(buf, "Repeat%s%d",(command_count>9?" ":" \t"),command_count);
      else
      StrPrintF(buf, "Repeat");
      // pyflags.status |= PY_REPEAT; // ?
      if (PY_SEARCH & pyflags.status)
	StrCat(buf, " Search");
    } else if (PY_SEARCH & pyflags.status)
      StrPrintF(buf, "Searching"); // 45
    if (buf[0]) {
      WinDrawInvertedChars(buf, StrLen(buf), x, hi);
      x += StrLen(buf) * visible_char_w_itsy;
      if (x > maxlen) { words_dirty_x[0] = x; x = 0; hi += 6; }
    }
  }
  // speed
  i = pyflags.speed;
  buf[0] = '\0';
  if (PY_SEARCH & pyflags.status) i--;   /* Search mode. */
  if (i > 1)        StrPrintF(buf,"SLOW");
  else if (i == 1)  StrPrintF(buf,"Slow");
  else if (i == 0)  ;
  else if (i == -1) StrPrintF(buf,"Fast");
  else              StrPrintF(buf,"FAST");
  if (buf[0] != '\0') {
    WinDrawInvertedChars(buf, StrLen(buf), x, hi);
    x += StrLen(buf) * 5;
    if (x > maxlen) { words_dirty_x[0] = x; x = 0; hi += 6; }
  }
  // this needs fixed:
  pyflags.status &= ~PY_STUDY;
  if (pyflags.new_spells != 0) {
    StrPrintF(buf, "Study"); // 25
    WinDrawInvertedChars(buf, StrLen(buf), x, hi);
    x += 25;
    if (x > maxlen) { words_dirty_x[0] = x; x = 0; hi += 6; }
  }

  if (hi > 0) words_dirty_x[1] = x / visible_char_w_itsy; 
  words_dirty_x[0] /= visible_char_w_itsy; 
  

#ifdef I_AM_OS_2
  FntSetFont(stdFont);
#else
  FntSetFont(stdFont);
#endif

  // this may help us to tidy up later.  nope, doesn't seem to help.  hmmm.
  /*
  absline = visible_y;
  if (x > 0 || hi > 0) {
    screen_dirty = true;
    lines_dirty[absline] = true;
    if (x > 0 && hi > 0) lines_dirty[absline+1] = true;
  }
  */
}
/* The various FLAGS:
   hunger:   Weak | Hungry   23,0
   blind:    Blind           23,7
   confused: Confused        23,13
   afraid:   Afraid          23,22
   poisoned: Poisoned        23,29
   state:    Paralysed | Rest %-5d | Repeat %-3d | Search | Searching   23,38
   speed:    Very Slow | Slow | Fast | Very Fast                        23,49
   study:    Study           23,59
   depth:    Town level  was 23,65
Weak   Blind Confused Afraid Poisoned Paralysed  Very Slow Study
Hungry Blind Confused Afraid Poisoned Repeat 123 Fast      Study
 */


// Draw a circle of radius radius (pixels) without using "sqrt".
// this is a sad sad little function.  but it doesn't look all that bad really.
static void draw_circle(Short radius, Short x0, Short y0)
{
  Short x, y, gap=2;
  if (radius >= 80) return;

  for (x = 0, y = radius ; x <= y ; x += gap) {
    while (x*x + y*y > radius*radius)
      y--;
    // draw horizontalish parts
    WinDrawLine( x + x0, y + y0, x + x0, y + y0);
    WinDrawLine( x + x0,-y + y0, x + x0,-y + y0);
    WinDrawLine(-x + x0, y + y0,-x + x0, y + y0);
    WinDrawLine(-x + x0,-y + y0,-x + x0,-y + y0);
    // draw verticalish parts
    WinDrawLine( y + x0, x + y0, y + x0, x + y0);
    WinDrawLine( y + x0,-x + y0, y + x0,-x + y0);
    WinDrawLine(-y + x0, x + y0,-y + x0, x + y0);
    WinDrawLine(-y + x0,-x + y0,-y + x0,-x + y0);
  }
  // really I shouldn't draw the part that falls into the message area oh well.
}
