/*********************************************************************
 * kMoria - Moria adapted for the PalmPilot.                         *
 * Copyright (C) 2000 Bridget Spitznagel                             *
 *                                                                   *
 * This program is derived from Umoria 5.5.2.  Moria is              *
 *   Copyright (c) 1989-94 James E. Wilson, Robert A. Koeneke        *
 *                                                                   *
 * This program is free software; you can redistribute it and/or     *
 * modify it under the terms of the GNU General Public License       *
 * as published by the Free Software Foundation; either version 2    *
 * of the License, or (at your option) any later version.            *
 *                                                                   *
 * This program is distributed in the hope that it will be useful,   *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
 * GNU General Public License for more details.                      *
 *                                                                   *
 * You should have received a copy of the GNU General Public License *
 * along with this program; if not, write to                         *
 * The Free Software Foundation, Inc.,                               *
 * 59 Temple Place - Suite 330,                                      *
 * Boston, MA  02111-1307, USA.                                      *
 *********************************************************************/
#include <Pilot.h>
#include "kMoria.h"
#include "lock-externs.h"
#include "kMoriaRsc.h"
#include <Preferences.h> // for date formats

static void draw_tombstone() SEC_5;
static void windraw_xcentered(Char *buf, Short, Short, Short *) SEC_5;
// [6][40]
static DmOpenRef NPOpenDB(Char *name) SEC_5;
static void prtstat(Short i, Char *stat) SEC_5;
static Char *likert(Short x, Short y) SEC_5;
static void snarf_history_block(Char *history_block) SEC_5;
static void death_record_memo() SEC_5;

static void init_highscore() SEC_6;
static void frob_topscores() SEC_6;

extern Boolean came_from_generation;
extern Char *player_title; // from lock.c
Long death_date;
Boolean Death_Form_HandleEvent(EventPtr e)
{
  Boolean handled = false;
  FormPtr frm;
    
  switch (e->eType) {

  case frmOpenEvent:
    death_date = TimGetSeconds(); // record when character is "pronounced dead"
    frm = FrmGetActiveForm();
    FrmDrawForm(frm);
    draw_tombstone();
    handled = true;
    break;

  case ctlSelectEvent:
    switch(e->data.ctlSelect.controlID) {
    case btn_death_wok:
      death_record_memo();      
      // fall through
    case btn_death_ok:
      // note: at THIS point I should write the new scorefile record,
      // then I can have a global boolean checked in ScoresForm to say
      // how to behave on exit, and it will be like totally viewable in-game!
      frob_topscores();
      antisave_char(); // delete the save file.
      re_init();
      came_from_generation = true;
      LeaveForm();
      FrmGotoForm(ScoresForm);
      /* then scoresform will: FrmGotoForm(CharRaceForm); 
	 ...'death' will not be set false until character generation */
      handled = true;
      break;
    }
    break;

  default:
    break;
  }

  return handled;
}

// S_WIDTH < 160, because this is a bordered screen.
#define S_WIDTH 156
static void draw_tombstone()
{
  Char buf[80];
  Boolean total_winner = false;
  Short x, x2, dx, y = 12, maxw = 0, topw, midw;

  y += 11;
  StrPrintF(buf, "%s", pymisc.name);
  windraw_xcentered(buf, S_WIDTH, y, &maxw);
  topw = maxw; // remember width at top of tombstone
  maxw += 16; // to make sure it's not just a boring rectangle

  y += 11;
  StrPrintF(buf, "the");
  windraw_xcentered(buf, S_WIDTH, y, &maxw);

  y += 11;
  if (!total_winner) // from player.c
    //    StrPrintF(buf, "player_title[%d][%d]", pymisc.pclass, pymisc.lev-1);
    StrPrintF(buf, "%s", 
	      &player_title[( (pymisc.pclass*MAX_PLAYER_LEVEL)+(pymisc.lev-1) )
			   * MAXLEN_PTITLE]);
  else
    StrPrintF(buf, "Magnificent");
  windraw_xcentered(buf, S_WIDTH, y, &maxw);
  midw = maxw; // remember width at ~middle of tombstone

  y += 11;
  y += 11;
  if (!total_winner)
    StrPrintF(buf, "%s, Level %d", class[pymisc.pclass].title, pymisc.lev);
  else
    StrPrintF(buf, "%s, Level %d", ((pymisc.male)?"King":"Queen"), pymisc.lev);
  windraw_xcentered(buf, S_WIDTH, y, &maxw);

  y += 11;
  StrPrintF(buf, "%ld Exp %ld Au", pymisc.exp, pymisc.au);
  windraw_xcentered(buf, S_WIDTH, y, &maxw);

  y += 11;
  StrPrintF(buf, "Died on Level: %d", dun_level);
  windraw_xcentered(buf, S_WIDTH, y, &maxw);

  y += 11;
  StrPrintF(buf, "killed by");
  windraw_xcentered(buf, S_WIDTH, y, &maxw);

  y += 11;
  StrPrintF(buf, "%s.", died_from); // "an Ancient Multi-Hued Dragon" = widest
  windraw_xcentered(buf, S_WIDTH, y, &maxw);

  y += 11;
  {
    DateTimeType datetime;
    Char datestr[dateStringLength];
    TimSecondsToDateTime(death_date, &datetime); // see DateTime.h
    DateToAscii(datetime.month, datetime.day, datetime.year,
		PrefGetPreference(prefDateFormat), datestr);
    StrPrintF(buf, "%s", datestr);
  }
  windraw_xcentered(buf, S_WIDTH, y, &maxw);
  // maxw is the max width of the tombstone

  // Now!  Draw a tombstone sort of shape around it.  If possible.
  // Draw a line from (bottom, S_WIDTH/2 +- maxw/2) to (mid, same x).
  y += 11;
  x = S_WIDTH/2 - topw/2;
  WinDrawLine(x, 17, x+topw, 17);// top horizontal of tombstone.

  dx = (maxw/2 + topw/2); // find the mean
  if (midw > dx) dx = midw;
  // left upper diagonal
  x = S_WIDTH/2 - topw/2;
  x2 = S_WIDTH/2 - dx/2;
  if (x2 < 0) x2 = 0;
  WinDrawLine(x, 17, x2, 17+11);
  // left middle diagonal
  x = x2;
  x2 = S_WIDTH/2 - (maxw/2 + 3);
  if (x2 < 0) x2 = 0;
  WinDrawLine(x, 17+11, x2, y-66);
  // left lower vertical
  x = x2;
  WinDrawLine(x, y, x, y-66);
  // now the right side...
  // right upper diagonal
  x = S_WIDTH/2 + topw/2;
  x2 = S_WIDTH/2 + dx/2;
  if (x2 > S_WIDTH) x2 = S_WIDTH;
  WinDrawLine(x, 17, x2, 17+11);
  // right middle diagonal
  x = x2;
  x2 = S_WIDTH/2 + (maxw/2 + 3);
  if (x2 > S_WIDTH) x2 = S_WIDTH;
  WinDrawLine(x, 17+11, x2, y-66);
  // right lower vertical
  x = x2;
  WinDrawLine(x, y, x, y-66);

}
// buf must be null-terminated string.  S_WIDTH is 160 or less pixels.
// y is the y location to draw it at.  w is for remembering the max width.
static void windraw_xcentered(Char *buf, Short s_width, Short y, Short *w)
{
  Short x, width, len;
  len = StrLen(buf);
  width = FntCharsWidth(buf, len);
  x = (s_width - width) / 2;
  if (x < 0) x = 0;
  WinDrawChars(buf, len, x, y);
  if (width > *w) *w = width;
}

/**************************************************************************
 **************************************************************************
 * The rest of this file is dedicated to printing the standard 80-column
 * dead Character Record to a MemoPad record.  It does not look too hot
 * when viewed in memopad but should be spiff if you sync it to a computer.
 * Dedicated to the people who post 'em on rec.games.roguelike.moria .
 **************************************************************************
 **************************************************************************/

/* Let's also have something that can record to a memo... 
   I wonder how one does that.
   Make the title of the memo be "kMoria <date> <time>"     */

/*
 * Props to Matthias Jordan for YANoff (which is GPL and has an Export To Memo
 * feature that shows it really is easy).
 */

//
// Opens database by name. If it doesn't exist: error
//
// returns 0 in case of error
// returns DmOpenRef > 0 in case of success
//
static DmOpenRef NPOpenDB(Char *name)
{
  LocalID lid;
  //  Err err;
  DmOpenRef db;

  lid = DmFindDatabase(0, name);
  if (lid == 0) {
    // not found - so create it
    //    err = DmCreateDatabase(0, name, CREATOR, TYPE, 0);
    //    if (err) return 0;
    //    lid = DmFindDatabase(0, name);
    return 0;
  }
  db = DmOpenDatabase(0, lid, dmModeReadWrite);
  return db;
} /* NPOpenDB */


// similar fn in display.c
static void prtstat(Short i, Char *stat)
{
  Short tmp = pystats.use_stat[i];
  if (tmp > 18) {
    tmp -= 18;
    if (tmp > 99) tmp = 0;
    StrPrintF(stat, (tmp > 9 ? "18/%d" : "18/0%d"), tmp);
  } else {
    StrPrintF(stat, "%d", tmp);
  }
}
// similar fn in form_chargen.c
static Char *likert(Short x, Short y)
{
  // I hope this still DTRT
  switch((x/y)) {
  case -3: case -2: case -1: return("Very Bad");
  case 0: case 1:            return("Bad");
  case 2:                    return("Poor");
  case 3: case 4:            return("Fair");
  case  5:                   return("Good");
  case 6:                    return("Very Good");
  case 7: case 8:            return("Excellent");
  default:                   return("Superb");
  }
}

static void snarf_history_block(Char *history_block)
{
  ULong uniqueID = REC_PHISTORY;
  UInt index;
  static VoidHand vh;
  VoidPtr p = NULL;
  // lock, copy, unlock record
  if (0!=DmFindRecordByID(MoriaSaveDB, uniqueID, &index)) return;
  vh = DmQueryRecord(MoriaSaveDB, index);
  p = MemHandleLock(vh);
  StrNCopy(history_block, (Char *) p, 239);
  MemHandleUnlock(vh);
  history_block[239] = '\0';//I forget if StrNCopy always terminates it
}

// A ridiculously long function.  It prints character record to memopad memo.
static void death_record_memo()
{
  DmOpenRef mdb;
  Char buf[200], stat[16]; // 160 + 24 < 200 so ok
  UInt index, memosize = 2000 * sizeof(Char); // xxxx MAKE THE SIZE BETTER
  VoidHand vh;
  DateTimeType datetime;
  Char datestr[dateStringLength], timestr[timeStringLength];
  Char *p;
  Short i, offset = 0;

  if (0 != (mdb = NPOpenDB("MemoDB"))) {
    if (0 != (vh = DmNewRecord(mdb, &index, memosize))) {
      if (0 != (p = MemHandleLock(vh))) {
	DmSet(p, 0, memosize, ' '); // will this work?
	TimSecondsToDateTime(death_date, &datetime); // see DateTime.h
	DateToAscii(datetime.month, datetime.day, datetime.year,
		    PrefGetPreference(prefDateFormat), datestr);
	TimeToAscii(datetime.hour, datetime.minute,
		    PrefGetPreference(prefTimeFormat), timestr);
	StrPrintF(buf, "kMoria %s %s\n\n", datestr, timestr);
	DmStrCopy(p, offset, buf);
	offset += StrLen(buf);

	////////////////////////////////// Section 1

	StrPrintF(buf,  " Name        : %s",                pymisc.name);
	for (i = StrLen(buf) ; i < 39 ; i++) buf[i] = ' ';
	StrPrintF(buf+39, "Age          :       ");
	StrPrintF(stat,   "%d",                             pymisc.age);
	StrPrintF(buf+60-StrLen(stat), "%s   STR :       ", stat);
	prtstat(0, stat);
	StrPrintF(buf+75-StrLen(stat), "%s\n", stat);
	DmStrCopy(p, offset, buf);
	offset += StrLen(buf);

	StrPrintF(buf,  " Race        : %s",        race[pymisc.prace].trace);
	for (i = StrLen(buf) ; i < 39 ; i++) buf[i] = ' ';
	StrPrintF(buf+39, "Height       :       ");
	StrPrintF(stat,   "%d",                             pymisc.ht);
	StrPrintF(buf+60-StrLen(stat), "%s   INT :       ", stat);
	prtstat(1, stat);
	StrPrintF(buf+75-StrLen(stat), "%s\n", stat);
	DmStrCopy(p, offset, buf);
	offset += StrLen(buf);

	StrPrintF(buf,  " Sex         : %s",    pymisc.male?"Male":"Female");
	for (i = StrLen(buf) ; i < 39 ; i++) buf[i] = ' ';
	StrPrintF(buf+39, "Weight       :       ");
	StrPrintF(stat,   "%d",                             pymisc.wt);
	StrPrintF(buf+60-StrLen(stat), "%s   WIS :       ", stat);
	prtstat(2, stat);
	StrPrintF(buf+75-StrLen(stat), "%s\n", stat);
	DmStrCopy(p, offset, buf);
	offset += StrLen(buf);

	StrPrintF(buf,  " Class       : %s",    class[pymisc.pclass].title);
	for (i = StrLen(buf) ; i < 39 ; i++) buf[i] = ' ';
	StrPrintF(buf+39, "Social Class :       ");
	StrPrintF(stat,   "%d",                             pymisc.sc);
	StrPrintF(buf+60-StrLen(stat), "%s   DEX :       ", stat);
	prtstat(3, stat);
	StrPrintF(buf+75-StrLen(stat), "%s\n", stat);
	DmStrCopy(p, offset, buf);
	offset += StrLen(buf);

	StrPrintF(buf,  " Title       : %s", ((total_winner) ? "Magnificent" :
	      (&player_title[((pymisc.pclass*MAX_PLAYER_LEVEL)+(pymisc.lev-1))
			    * MAXLEN_PTITLE]) ));
	for (i = StrLen(buf) ; i < 63 ; i++) buf[i] = ' ';
	StrPrintF(buf+63, "CON :       ", stat);
	prtstat(4, stat);
	StrPrintF(buf+75-StrLen(stat), "%s\n", stat);
	DmStrCopy(p, offset, buf);
	offset += StrLen(buf);

	for (i = 0 ; i < 63 ; i++) buf[i] = ' ';
	StrPrintF(buf+63, "CHR :       ", stat);
	prtstat(5, stat);
	StrPrintF(buf+75-StrLen(stat), "%s\n\n", stat);
	DmStrCopy(p, offset, buf);
	offset += StrLen(buf);

	////////////////////////////////// Section 2

	StrPrintF(buf, " + To Hit    :       ");
	StrPrintF(stat,   "%d",                             pymisc.dis_th);
	StrPrintF(buf+21-StrLen(stat), "%s       Level      :        ", stat);
	StrPrintF(stat,   "%d",                             pymisc.lev);
	StrPrintF(buf+48-StrLen(stat), "%s    Max Hit Points :      ", stat);
	StrPrintF(stat,   "%d",                             pymisc.mhp);
	StrPrintF(buf+75-StrLen(stat), "%s\n", stat);
	DmStrCopy(p, offset, buf);
	offset += StrLen(buf);

	StrPrintF(buf, " + To Damage :       ");
	StrPrintF(stat,   "%d",                             pymisc.dis_td);
	StrPrintF(buf+21-StrLen(stat), "%s       Experience :        ", stat);
	StrPrintF(stat,   "%d",                             pymisc.exp);
	StrPrintF(buf+48-StrLen(stat), "%s    Cur Hit Points :      ", stat);
	StrPrintF(stat,   "%d",                             pymisc.chp);
	StrPrintF(buf+75-StrLen(stat), "%s\n", stat);
	DmStrCopy(p, offset, buf);
	offset += StrLen(buf);

	StrPrintF(buf, " + To AC     :       ");
	StrPrintF(stat,   "%d",                             pymisc.dis_tac);
	StrPrintF(buf+21-StrLen(stat), "%s       Max Exp    :        ", stat);
	StrPrintF(stat,   "%d",                             pymisc.max_exp);
	StrPrintF(buf+48-StrLen(stat), "%s    Max Mana       :      ", stat);
	StrPrintF(stat,   "%d",                             pymisc.mana);
	StrPrintF(buf+75-StrLen(stat), "%s\n", stat);
	DmStrCopy(p, offset, buf);
	offset += StrLen(buf);

	StrPrintF(buf, "   Total AC  :       ");
	StrPrintF(stat,   "%d",                             pymisc.dis_ac);
	StrPrintF(buf+21-StrLen(stat), "%s       Exp to Adv :        ", stat);
	// xxx this might be bogus, check for division truncation problems
	StrPrintF(stat, "%ld", (Long) (player_exp[pymisc.lev-1]
				       * pymisc.expfact / 100));//form_chargen.
	StrPrintF(buf+48-StrLen(stat), "%s    Cur Mana       :      ", stat);
	StrPrintF(stat,   "%d",                             pymisc.cmana);
	StrPrintF(buf+75-StrLen(stat), "%s\n", stat);
	DmStrCopy(p, offset, buf);
	offset += StrLen(buf);

	for (i = 0; i < 28 ; i++) buf[i] = ' ';
	StrPrintF(buf+28, "Gold       :        ");
	StrPrintF(stat,   "%ld",                             pymisc.au);
	StrPrintF(buf+48-StrLen(stat), "%s\n", stat);
	DmStrCopy(p, offset, buf);
	offset += StrLen(buf);

	////////////////////////////////// Section 3
	StrPrintF(buf, "\n(Miscellaneous Abilities)\n\n");
	DmStrCopy(p, offset, buf);
	offset += StrLen(buf);

	{
	  // From form_chargen.c - would it save space as a separate fn? maybe
	  // Calculate the miscellaneous abilities;
	  Short xbth, xbthb, xfos, xsrh, xstl, xdis, xsave, xdev;

	  xbth  = pymisc.bth + pymisc.ptohit*BTH_PLUS_ADJ
	    + (class_level_adj[pymisc.pclass][CLA_BTH] * pymisc.lev);
	  /* this results in a range from 0 to 9 */
	  xstl  = pymisc.stl + 1;
	  xfos  = 40 - pymisc.fos;
	  if (xfos < 0)  xfos = 0;

	  StrPrintF(buf, " Fighting    : %s", likert(xbth, 12));
	  for (i = StrLen(buf) ; i < 28 ; i++) buf[i] = ' ';
	  StrPrintF(buf+28, "Stealth     : %s", likert(xstl, 1));
	  for (i = StrLen(buf) ; i < 55; i++) buf[i] = ' ';
	  StrPrintF(buf+55, "Perception  : %s\n", likert(xfos, 3));
	  DmStrCopy(p, offset, buf);
	  offset += StrLen(buf);
	
	  xbthb = pymisc.bthb + pymisc.ptohit*BTH_PLUS_ADJ
	    + (class_level_adj[pymisc.pclass][CLA_BTHB] * pymisc.lev);
	  xdis  = pymisc.disarm + 2*todis_adj() + stat_adj(A_INT)
	    + (class_level_adj[pymisc.pclass][CLA_DISARM] * pymisc.lev / 3);
	  /* this results in a range from 0 to 29 */
	  xsrh  = pymisc.srh;

	  StrPrintF(buf, " Bows/Throw  : %s", likert(xbthb,12));
	  for (i = StrLen(buf) ; i < 28 ; i++) buf[i] = ' ';
	  StrPrintF(buf+28, "Disarming   : %s", likert(xdis, 8));
	  for (i = StrLen(buf) ; i < 55; i++) buf[i] = ' ';
	  StrPrintF(buf+55, "Searching   : %s\n", likert(xsrh, 6));
	  DmStrCopy(p, offset, buf);
	  offset += StrLen(buf);
  
	  xsave = pymisc.save + stat_adj(A_WIS)
	    + (class_level_adj[pymisc.pclass][CLA_SAVE] * pymisc.lev / 3);
	  xdev  = pymisc.save + stat_adj(A_INT)
	    + (class_level_adj[pymisc.pclass][CLA_DEVICE] * pymisc.lev / 3);

	  StrPrintF(buf, " Saving Throw: %s", likert(xsave, 6));
	  for (i = StrLen(buf) ; i < 28 ; i++) buf[i] = ' ';
	  StrPrintF(buf+28, "Magic Device: %s", likert(xdev, 6));
	  for (i = StrLen(buf) ; i < 55; i++) buf[i] = ' ';
	  StrPrintF(buf+55, "Infra-Vision: %d feet\n\n", pyflags.see_infra*10);
	  DmStrCopy(p, offset, buf);
	  offset += StrLen(buf);
	}
	
	////////////////////////////////// Section 4
	StrPrintF(buf, "\nCharacter Background\n");
	DmStrCopy(p, offset, buf);
	offset += StrLen(buf);

	// copy the REC_PHISTORY string!
	{
	  Char history_block[240];
	  Short j, k, len;

	  snarf_history_block(history_block);
	  // word-wrap the history block!
	  len = StrLen(history_block);
	  j = k = 0;
	  // this (60) is the line length used in moria create.c get_history().
#define LINELEN 60
	  while (j < len) {
	    //Step left from the end of the (desired) line to find a whitespace
	    for (j = k + LINELEN ; j > k ; j--)
	      if (history_block[j] == ' ')
		break;
	    if (j == k) // this case (really long word) should never happen:
	      j = k + LINELEN;
	    history_block[j] = '\n';
	    k = j;
	  }
	  history_block[239] = '\0'; // bride of paranoia
#undef LINELEN
	  // ok now copy the word-wrapped string to the memopad record
	  DmStrCopy(p, offset, history_block);
	  offset += StrLen(history_block);
	}		

	////////////////////////////////// Section 5
	StrPrintF(buf, "\n\n\n  [Character's Equipment List]\n\n");
	DmStrCopy(p, offset, buf);
	offset += StrLen(buf);

	StrPrintF(buf, "buffalo\n"); // XXX
	if (equip_ctr == 0) {
	  StrPrintF(buf, "  Character has no equipment in use.\n\n");
	  DmStrCopy(p, offset, buf);
	  offset += StrLen(buf);
	} else {
	  inven_type *i_ptr;
	  Short i, j=0, k, len;
	  for (i = INVEN_WIELD; i < INVEN_ARRAY_SIZE; i++) {
	    i_ptr = &inventory[i];
	    // THERE IS A BUG SOMEWHERE AFTER HERE.
	    if (i_ptr->tval != TV_NOTHING) {
	      StrPrintF(buf, "  %c) ", (Char) (j+'a'));
	      j++;
	      switch (i) {
	      case INVEN_WIELD:	StrPrintF(buf+5, "You are wielding");	break;
	      case INVEN_HEAD:	StrPrintF(buf+5, "Worn on head");	break;
	      case INVEN_NECK:	StrPrintF(buf+5, "Worn around neck");	break;
	      case INVEN_BODY:	StrPrintF(buf+5, "Worn on body");	break;
	      case INVEN_ARM:	StrPrintF(buf+5, "Worn on shield arm"); break;
	      case INVEN_HANDS:	StrPrintF(buf+5, "Worn on hands");	break;
	      case INVEN_RIGHT:	StrPrintF(buf+5, "Right ring finger");  break;
	      case INVEN_LEFT:	StrPrintF(buf+5, "Left  ring finger");  break;
	      case INVEN_FEET:	StrPrintF(buf+5, "Worn on feet");	break;
	      case INVEN_OUTER:	StrPrintF(buf+5, "Worn about body");	break;
	      case INVEN_LIGHT:	StrPrintF(buf+5, "Light source is");	break;
	      case INVEN_AUX:	StrPrintF(buf+5, "Secondary weapon");	break;
	      default: StrPrintF(buf+5, "*Unknown value*");     break;
	      }
	      for (k = StrLen(buf) ; k < 26 ; k++) buf[k] = ' ';
	      buf[24] = ':'; // (and 25 is already a space)
	      objdes(buf+26, &inventory[i], true); // print object descrip
	      len = StrLen(buf); // tack on the newline...
	      buf[len] = '\n';
	      buf[len+1] = '\0'; // done...
	      DmStrCopy(p, offset, buf);
	      offset += StrLen(buf);
	    }
	  }
	}

	////////////////////////////////// Section 6
	StrPrintF(buf, "\n  [General Inventory List]\n\n");
	DmStrCopy(p, offset, buf);
	offset += StrLen(buf);

	if (inven_ctr == 0) {
	  StrPrintF(buf, "  Character has no objects in inventory.\n\n");
	  DmStrCopy(p, offset, buf);
	  offset += StrLen(buf);
	} else {
	  Short i, len;
	  for (i = 0; i < inven_ctr; i++) {
	    StrPrintF(buf, "%c) ", (Char) (i+'a'));
	    objdes(buf+3, &inventory[i], true); // print object description
	    len = StrLen(buf); // tack on the newline...
	    buf[len] = '\n';
	    buf[len+1] = '\0'; // done...
	    DmStrCopy(p, offset, buf);
	    offset += StrLen(buf);
	  }
	}
	// all done!!!!!

	// clean up now
	MemPtrUnlock(p);
	DmReleaseRecord(mdb, index, true);
	DmCloseDatabase(mdb);
	return;
      }
    }
    DmCloseDatabase(mdb);    
  }
  // Report the failure... make the button go away and do a popup alert?

}



/************************************************************************
 ************************************************************************
 *******************   Record the highscore thing   *********************
 ************************************************************************
 ************************************************************************/

static void init_highscore(high_scores *score)
{
  Char *p;
  UChar c;  // c = 0rrrCCCm

  c = (((pymisc.prace << 3) | pymisc.pclass) << 1) | (pymisc.male ? 0x1 : 0x0);
  score->points = total_points();
  score->birth_date = birth_date;
  //  score->mhp = pymisc.mhp;
  //  score->chp = pymisc.chp;
  //  score->dun_level = dun_level;
  //  score->max_dlv = pymisc.max_dlv;
  score->lev = pymisc.lev;
  score->race_class_sex = c;
  StrCopy(score->name, pymisc.name);
  p = died_from;
  // skip "an " or "a "
  if (*p == 'a') {
    p++;
    if (*p == 'n') p++;
    while (*p == ' ') p++;
  }
  StrCopy(score->died_from, p);
}


// was in death.c ...
/* Calculates the total number of points earned		-JWT-	 */
Long total_points()
{
  Long total;
  Short i;

  total = pymisc.max_exp + (100 * pymisc.max_dlv);
  total += pymisc.au / 100;
  for (i = 0; i < INVEN_ARRAY_SIZE; i++)
    total += item_value(&inventory[i]);
  total += dun_level * 50;

  /* Don't ever let the score decrease from one save to the next.  */
  if (max_score > total)
    return max_score; // max_score is set/saved as my_save_info.total_points.

  return total;
}


/*****************************************
 * Ok, here is my outline of what to do.

 Write player info to a struct on the stack.
 Create a new record.
 _If_ there is already a REC_TOPSCORES record,
      figure out what index the player info will appear at (or replace),
      copy up to that index from old record to new record.
 Write the player info from the struct to the new record.
 _If_ there is already a REC_TOPSCORES record,
      copy the remainder (skip a spot if replacing index) to the new record.
      delete the REC_TOPSCORES record.
 Release the new record.
 Set its uid to REC_TOPSCORES.
 Congratulate self on doing things the easy way this time (sorta).

 *************************************/

/* Add the player's data to the top-score list, perhaps replacing an entry */
static void frob_topscores()
{
  ULong uniqueID;
  UInt index_old, index;
  VoidHand vh, vh_old = 0;
  VoidPtr p, p_old;
  high_scores score, *scores_old = 0;
  Short num = 1, offset = 0, i = 0, j = 0; // i counts old entries, j the new.
  Short *nump;
  Boolean exists;
  ULong size;
  Char buf[40];

  init_highscore(&score);

  // Does the top scores list exist.
  uniqueID = REC_TOPSCORES;
  exists = (0 == DmFindRecordByID(MoriaSaveDB, uniqueID, &index_old));
  if (exists) {
    // it's a Short (n), followed by n records.
    vh_old = DmQueryRecord(MoriaSaveDB, index_old);
    p_old = MemHandleLock(vh_old);
    nump = (Short *) p_old;
    num += nump[0];
    scores_old = (high_scores *) &(nump[1]);
  }

  if (num > SCOREFILE_SIZE) num = SCOREFILE_SIZE;
  // we will always replace the last entry even if we are at max size.

  // Open a new record.
  index = DmNumRecords(MoriaSaveDB);
  size = sizeof(Short) + num * sizeof(high_scores);
  vh = DmNewRecord(MoriaSaveDB, &index, size);
  p = MemHandleLock(vh);
  if (p) {
    DmWrite(p, offset, &num, sizeof(Short)); // write the # of items in set
    offset += sizeof(Short);
    if (exists && scores_old) { // testing
      // Write records until we find one that we are better than (or
      // have the same birthday as).  We will pick up here (if we have
      // the same birthday, ++ the counter so that we'll skip it) after.
      // Oh, stop if we run out of records or space.
      while ((i < num-1) && (j < SCOREFILE_SIZE-1)) {
	if (score.points >= scores_old[i].points) break; // found our rank
	if (score.birth_date != scores_old[i].birth_date) {
	  DmWrite(p, offset, &scores_old[i], sizeof(high_scores));
	  offset += sizeof(high_scores);
	  j++;
	} // else we are a clone of i; we'll delete i cause we're newer
	i++;
      }
    }
    DmWrite(p, offset, &score, sizeof(high_scores));
    offset += sizeof(high_scores);
    j++;
    if (exists && scores_old) {
      // Write the remaining records.  Stop after num-1.
      while ((i < num-1) && (j < SCOREFILE_SIZE)) {
	if (score.birth_date != scores_old[i].birth_date) {
	  DmWrite(p, offset, &scores_old[i], sizeof(high_scores));
	  offset += sizeof(high_scores);
	  j++;
	} // else we are a clone of i; we'll delete i cause we're newer
	i++;
      }
      MemHandleUnlock(vh_old);
    }
  }

  exists = (0 == DmFindRecordByID(MoriaSaveDB, uniqueID, &index_old));

  // in with the new
  MemPtrUnlock(p);
  DmReleaseRecord(MoriaSaveDB, index, true);
  DmSetRecordInfo(MoriaSaveDB, index, NULL, &uniqueID); // set its uid!
  if (exists)
    DmRemoveRecord(MoriaSaveDB, index_old);      // EX TERM IN ATE !

}
