/*********************************************************************
 * kMoria - Moria adapted for the PalmPilot.                         *
 * Copyright (C) 2000 Bridget Spitznagel                             *
 *                                                                   *
 * This program is derived from Umoria 5.5.2.  Moria is              *
 *   Copyright (c) 1989-94 James E. Wilson, Robert A. Koeneke        *
 *                                                                   *
 * This program is free software; you can redistribute it and/or     *
 * modify it under the terms of the GNU General Public License       *
 * as published by the Free Software Foundation; either version 2    *
 * of the License, or (at your option) any later version.            *
 *                                                                   *
 * This program is distributed in the hope that it will be useful,   *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
 * GNU General Public License for more details.                      *
 *                                                                   *
 * You should have received a copy of the GNU General Public License *
 * along with this program; if not, write to                         *
 * The Free Software Foundation, Inc.,                               *
 * 59 Temple Place - Suite 330,                                      *
 * Boston, MA  02111-1307, USA.                                      *
 *********************************************************************/
#include <Pilot.h>
#include "kMoria.h"
#include "lock-externs.h"
#include "kMoriaRsc.h"

static void screen_map() SEC_6;
static void screen_map1() SEC_6;
static void screen_map2() SEC_6;

  /* Display highest priority object in the RATIO_W by RATIO_H area */
#define	RATIO_W 4
#define	RATIO_H 3
// we can make ratio_w 4 by drawing the map in two sections.

#define visible_w_itsy 32
#define vc_h 6

Boolean left_side = true;
Boolean map_mode_chars = true; // perhaps should be a pref.  nahhh.

Boolean Map_Form_HandleEvent(EventPtr e)
{
  Boolean handled = false;
  FormPtr frm;
  Short bot_y;

  switch (e->eType) {

  case frmOpenEvent:
    frm = FrmGetActiveForm();
    FrmDrawForm(frm);
    left_side = (char_col < MAX_WIDTH/2);
    /* this must be AFTER draw form */
    screen_map();
    handled = true;
    break;

  case penDownEvent:
    if (map_mode_chars) bot_y = (MAX_HEIGHT/RATIO_H + 1) * vc_h;
    else                bot_y = MAX_HEIGHT*2+1;
    if (e->screenY > bot_y) {
      if (e->screenX > 80)
	LeaveForm();
      else {
	map_mode_chars = !map_mode_chars;
	screen_map();
      }
    } else if ((left_side && e->screenX > 80) ||
	       (!left_side && e->screenX < 80)) {
	left_side = !left_side;
	screen_map();
      }
    handled = true;
    break;

  default:
    break;
  }

  return handled;
}


/*
  The dungeon is, at maximum, 66 high by 198 wide.

  The screen is 160 by 160.  In tiny font chars, it's 160/6 high by 160/5 wide.
  that's 26 high by 32 wide.  so if I use characters it's
  3 cells high per character and *7* cells wide.  gack.

     if I use _2_ screens wide, then it's 160/6 high by 320/5 wide.
     that's 26 high by 64 wide.  so if I use 2 screens of characters it's
     3 cells high by (just barely more than 3, round up to) 4 cells wide.

  so, width-wise, we must represent at least 5 cells in each 4 pixels.
  if we try to keep the same proportions as in the dungeon,
  what do we get?  h:w = 5:4 or 6:5.  so, 3 cells per 2 pixels (6 in 4).

  in a 4 x 4 block of pixels,  we are representing 6 cells high by 5 cells wide
  that's 11 4x4 blocks of height or a mere 44 pixels, 1/4 of the screen. hmmm
  and 39.6 ~= 40 blocks of width or the total 160 pixels. hmmm

  It would be better to allow 2 (rectangle) or 4 (square) screens.

.... .... .... .... .... ....
***. ***. ..*. *... .... *...
***. *.*. .*.. .*.. .... *...
***. *.*. *... ..*. .... .*..
***. ***. .*.. .*.. .*.. ..*.



+------------------------------------------------------------------+
|>########4############                                            |
|#..#2###.###...###...#                                            |
|#..#####.###...###...#                                            |
|#..............5##..@#                                            |
|#..####..6##...#1#..p#                                            |
|#..####..###...###...#                                            |
|#..###3........###...#                                            |
|######################                                            |
|                                                                  |
|                                                                  |
|                                                                  |
|                                                                  |
|                                                                  |
|                                                                  |
|                                                                  |
|                                                                  |
|                                                                  |
|                                                                  |
|                                                                  |
|                                                                  |
|                                                                  |
|                                                                  |
+----------------------Hit any key to continue---------------------+

             ##################################################################
Human        #................................................................#
Paladin      #>.........................4######...............................#
Gallant      #..........##2########.....#######...........#########...........#
             #..........###########.....#######...........#########...........#
STR :     18 #..........###########.....#######...........#########...........#
INT :      5 #..........###########.....#######...........#########...........#
WIS :     18 #..........................#######...........#########...........#
DEX :     11 #..........................#######...........#########...........#
CON :     13 #............................................5########...........#
CHR :     17 #.............................................................@..#
             #................................................................#
LEV :      1 #............................................#####1##........p...#
EXP :      0 #..........##########........6####...........########............#
MANA:      2 #..........##########........#####...........########............#
MHP :     16 #..........##########........#####...........########............#
CHP :     16 #..........##########........#####...........########............#
             #..........##########........#####...........########............#
AC  :      0 #..........########3#........................########............#
GOLD:    459 #................................................................#
             #................................................................#
             ##################################################################
                                                                 Town level

+------------------------------------------------------------------+
|                                                                  |
|              ###########                                         |
|              ####'######                                         |
|               ####    #######                                    |
|            #########  #########                                  |
|            #.......#     '....#######<                           |
|            ##s###+##     #############>##                  #     |
|                                     ######                 #     |
|                                     #....# ##          ######    |
|                                     #....# ###         '##~.#    |
|                                     #s########         ## ###    |
|                                    #########           ####@#    |
|                                    '...##              ##...#    |
|                                    ######           #########    |
|                                      #              ###:######   |
|                                      #        #####   ###'..'#   |
|                                      '#########...###### #..##   |
|                                               ########   ###'#   |
|                                                                  |
|                                                                  |
|                                                                  |
|                                                                  |
+----------------------Hit any key to continue---------------------+



*/

/* definitions used by screen_map() */
/* index into border character array */
#define TL 0	/* top left */
#define TR 1
#define BL 2
#define BR 3
#define HE 4	/* horizontal edge */
#define VE 5
#   define CH(x)	(screen_border[x])

static void screen_map() {
  Short y;
  RectangleType r;  
  RctSetRectangle(&r, 0, 0, 160, 160);
  WinEraseRectangle(&r, 0);
  if (map_mode_chars) y = (MAX_HEIGHT/RATIO_H + 1) * vc_h;
  else                y = MAX_HEIGHT*2+1;
  WinDrawLine(0, y, 160, y);
  WinDrawLine(80, y, 80, 160);
  WinDrawChars("Mode", 4, 30, 160-11);
  WinDrawChars("Exit", 4, 30+80, 160-11);
  if (map_mode_chars)
    screen_map1();
  else
    screen_map2();
  if (left_side)
    WinDrawChars(">>", 2, 150, y-5);  
  else
    WinDrawChars("<<", 2, 1, y-5);  
}



static void screen_map1()
{
  //Short i, j, vc_h = 6, x=9, y;// set x so that the map is centered horizntly
  Short i, j, x=0, y; // no, have some horizontal overlap.
  //  UChar screen_border[6] = {
  //    '+', '+', '+', '+', '-', '|'	/* screen border chars */
  //  };
  UChar map[MAX_WIDTH / RATIO_W + 1];
  UChar tmp;
  //  Short priority[256];
  Char priority[256];
  Short row, orow, col; // , myrow, mycol = 0;
  //  Short prntscrnbuf[80];
  Short overlappish = (MAX_WIDTH / RATIO_W + 1) - visible_w_itsy;

#ifdef I_AM_OS_2
  FntSetFont(ledFont);
#else
  FntSetFont(129);
#endif

  for (i = 0; i < 256; i++)
    priority[i] = 0;
  priority['\t'] = -15; //this glyph is an em-space. use instead of ' '
  priority[' '] = -15;
  priority['.'] = -10;
  priority['\''] = -3;
  priority['#'] = -5;
  priority['<'] = 5;
  priority['>'] = 5;
  priority['@'] = 10;

  orow = -1;
  map[MAX_WIDTH / RATIO_W] = '\0';
  for (i = 0; i < MAX_HEIGHT; i++) {
    row = i / RATIO_H;
    if (row != orow) {
      if (orow >= 0) {
	y = (orow + 1) * vc_h;
	if (left_side)
	  WinDrawChars(map, visible_w_itsy, x, y);
	else
	  WinDrawChars(map+overlappish, StrLen(map+overlappish), x, y);
	//	StrPrintF(prntscrnbuf,"%s", map);
	//	mvaddstr(orow+1, 0, prntscrnbuf);
      }
      for (j = 0 ; j < MAX_WIDTH / RATIO_W ; j++)
	map[j] = '\t'; // use the em-width space.
      orow = row;
    }
    for (j = 0; j < MAX_WIDTH; j++) {
      col = j / RATIO_W;
      tmp = loc_symbol(i, j);
      if (tmp == ' ') tmp = '\t'; // use the em-width space.
      if (priority[map[col]] < priority[tmp])
	map[col] = tmp;
      if (map[col] == '@') {
	//	mycol = col + 1; /* account for border */
	//	myrow = row + 1;
      }
    }
  }
  if (orow >= 0) {
    y = (orow + 1) * vc_h;
    if (left_side)
      WinDrawChars(map, visible_w_itsy, x, y);
    else
      WinDrawChars(map+overlappish, StrLen(map+overlappish), x, y);
    //    StrPrintF(prntscrnbuf,"%s", map);
    //    mvaddstr(orow+1, 0, prntscrnbuf);
  }

#ifdef I_AM_OS_2
  FntSetFont(stdFont);
#else
  FntSetFont(stdFont);
#endif

}


static void screen_map2()
{
  Short i, j; // row, col; or y, x
  Short j_Offset;
  UChar tmp;

  /*
    #ifdef I_AM_OS_2
    FntSetFont(ledFont);
    #else
    FntSetFont(129);
    #endif
  */

  //  j_Offset = (left_side ? 0 : MAX_WIDTH / 2); // no fear, MAX_WIDTH is even
  //  for (j = 0; j < MAX_WIDTH/2; j++)
  j_Offset = (left_side ? 0 : MAX_WIDTH - 160);
  for (j = 0; j < 160; j++) {
    for (i = 0; i < MAX_HEIGHT; i++) {
      tmp = loc_symbol(i, j + j_Offset);
      if (tmp == '@') {
	WinDrawLine(j, i*2+1, j, i*2+1); // draw 1 measly pixel
	// WinDrawChars("@", 1, j, MAX_HEIGHT*2+1); // draw @ in space below
	// WinDrawChars("@", 1, j-2, i*2-2); // center the itsy character
      }
      if (tmp == '#' || tmp == '%') // wall
	WinDrawLine(j, i*2, j, i*2+1); // draw 2 measly pixels
    }
  }
  /*
    #ifdef I_AM_OS_2
    FntSetFont(stdFont);
    #else
    FntSetFont(stdFont);
    #endif
  */

}
