/*********************************************************************
 * kMoria - Moria adapted for the PalmPilot.                         *
 * Copyright (C) 2000 Bridget Spitznagel                             *
 *                                                                   *
 * This program is derived from Umoria 5.5.2.  Moria is              *
 *   Copyright (c) 1989-94 James E. Wilson, Robert A. Koeneke        *
 *                                                                   *
 * This program is free software; you can redistribute it and/or     *
 * modify it under the terms of the GNU General Public License       *
 * as published by the Free Software Foundation; either version 2    *
 * of the License, or (at your option) any later version.            *
 *                                                                   *
 * This program is distributed in the hope that it will be useful,   *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
 * GNU General Public License for more details.                      *
 *                                                                   *
 * You should have received a copy of the GNU General Public License *
 * along with this program; if not, write to                         *
 * The Free Software Foundation, Inc.,                               *
 * 59 Temple Place - Suite 330,                                      *
 * Boston, MA  02111-1307, USA.                                      *
 *********************************************************************/
#include <Pilot.h>
//#include "kMoriaRsc.h"
#include "kMoria.h"
#include "lock-externs.h"
#include "kMoriaRsc.h"

/* Form for inscribing an already selected object. 
   Note that inscribing should not take a turn whether cancelled or not. */
extern Short needed_direction_item; // this is the selected item

static void init_scribe(Short item_val) SEC_6;
static void init_scribe_fld(FormPtr frm) SEC_6;
static void commit_scribe_fld(FormPtr frm, Short item_val) SEC_6;

Boolean Scribe_Form_HandleEvent(EventPtr e)
{
  Boolean handled = false;
  FormPtr frm;

  switch (e->eType) {

  case frmOpenEvent:
    frm = FrmGetActiveForm();
    /* ...init form... */
    init_scribe_fld(frm);
    FrmDrawForm(frm);
    init_scribe(needed_direction_item);
    handled = true;
    break;

  case ctlSelectEvent:
    frm = FrmGetActiveForm();
    switch(e->data.ctlSelect.controlID) {
    case btn_sb_ok:
      commit_scribe_fld(frm, needed_direction_item);
      LeaveForm();
      needed_direction_item = -1;
      handled = true;
      break;
    case btn_sb_cancel:
      LeaveForm();
      needed_direction_item = -1;
      handled = true;
      break;
    }

  default:
    break;
  }
  return handled;
}


static void init_scribe(Short item_val)
{
  Char buf[160], *p;
  Word wwlen, wwpixels=156;
  Short x=2, y=15, lines = 0;

  // Print up to 3 lines of "Inscribing <original object>"
  StrPrintF(buf, "Inscribing ");
  p = buf + StrLen(buf);
  objdes(p, &inventory[item_val], true);
  p = buf;
  wwlen = 0;
  wwpixels -= 2*x; // leave margins
  while (lines < 3 && StrLen(p) > 0) {
    wwlen = FntWordWrap(p, wwpixels);
    WinDrawChars(p, wwlen, x, y+11*lines);
    p += wwlen;
    lines++;
  }
  // print old inscription if any
  if (inventory[item_val].inscrip[0] != '\0') {
    StrPrintF(buf, "Replace %s", inventory[item_val].inscrip);
    WinDrawChars(buf, StrLen(buf), x, y+3*11);
    WinDrawChars("New", 3,  3, y+4*11);
  }
}


static void init_scribe_fld(FormPtr frm)
{
  VoidHand h;
  VoidPtr p;
  FieldPtr fld;
  // allocate memory for field, I hope.
  h = MemHandleNew(sizeof(Char) * INSCRIP_SIZE);
  p = MemHandleLock(h);
  ((Char *)p)[0] = '\0';
  MemPtrUnlock(p);
  fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, field_sb));
  FldSetTextHandle(fld, (Handle) h);
}

static void commit_scribe_fld(FormPtr frm, Short item_val)
{
  FieldPtr fld;
  CharPtr textp;
  // Copy contents of field
  fld = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, field_sb));
  textp = FldGetTextPtr(fld); // if (textP) ...
  textp[INSCRIP_SIZE - 1] = '\0';
  DmStrCopy(inventory, item_val*sizeof(inven_type)+I_inscrip, textp);
  // Get rid of the field
  FldReleaseFocus(fld);
  FldSetSelection(fld, 0, 0);
  FldFreeMemory(fld);
}
