/*********************************************************************
 * kMoria - Moria adapted for the PalmPilot.                         *
 * Copyright (C) 2000 Bridget Spitznagel                             *
 *                                                                   *
 * This program is derived from Umoria 5.5.2.  Moria is              *
 *   Copyright (c) 1989-94 James E. Wilson, Robert A. Koeneke        *
 *                                                                   *
 * This program is free software; you can redistribute it and/or     *
 * modify it under the terms of the GNU General Public License       *
 * as published by the Free Software Foundation; either version 2    *
 * of the License, or (at your option) any later version.            *
 *                                                                   *
 * This program is distributed in the hope that it will be useful,   *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
 * GNU General Public License for more details.                      *
 *                                                                   *
 * You should have received a copy of the GNU General Public License *
 * along with this program; if not, write to                         *
 * The Free Software Foundation, Inc.,                               *
 * 59 Temple Place - Suite 330,                                      *
 * Boston, MA  02111-1307, USA.                                      *
 *********************************************************************/
#include <Pilot.h>
#include "kMoria.h"
#include "lock-externs.h"
#include "kMoriaRsc.h"
#include <Preferences.h> // for date formats


static void draw_topscores_header() SEC_6;
static Short draw_topscores(Boolean reset) SEC_6;

extern Boolean came_from_generation;
extern Long death_date;
Boolean Scores_Form_HandleEvent(EventPtr e)
{
  Boolean handled = false;
  FormPtr frm;
    
  switch (e->eType) {

  case frmOpenEvent:
    frm = FrmGetActiveForm();
    FrmDrawForm(frm);
    if (!death) {
      ControlPtr btn;
      btn = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, btn_scores_ok));
      CtlSetLabel(btn, "- Done -");
      CtlShowControl(btn);
    }
    draw_topscores(true);
    handled = true;
    break;

  case keyDownEvent:
    // hardware button -- or else graffiti.
    switch(e->data.keyDown.chr) {
    case pageUpChr:
      draw_topscores(true);
      handled = true;
      break;
    case pageDownChr:
      draw_topscores(false);
      handled = true;
      break;
    default:
      break;
    }
    break;

  case ctlSelectEvent:
    switch(e->data.ctlSelect.controlID) {
    case btn_scores_ok:
      LeaveForm();
      if (death)
	FrmGotoForm(CharRaceForm);
      // else we're still alive and were viewing a popup on top of MainForm!
      handled = true;
      break;
    case btn_scores_next:
      draw_topscores(false);
      handled = true;
      break;
    }
    break;

  default:
    break;
  }

  return handled;
}


// was:  Rank Points Name Sex Race Class Lvl Killed By
// now:  Rank Score [Sex] Race Class Level
//        [Name]
//         [Killed by]
extern high_scores score;


static void draw_topscores_header()
{
  Char buf[80];
  Short x, y;
  RectangleType r;
  RctSetRectangle(&r, 0, 15, 160, 146-15); // left, top, width, height
  WinEraseRectangle(&r, 0);

  ///////////////////////////////////////////////////////////////
  //     v---right justified    v---right justified
  // 0  +17  +42  +10+42  +32  +15
  y = 15;
  x = 0;
  StrPrintF(buf, "Rank"); // left-justify
  WinDrawChars(buf, StrLen(buf), x, y);
  x += 17 + 42; // width of "<rank> <points> "
  StrPrintF(buf, "Points"); // right-justify to 17+42
  WinDrawChars(buf, StrLen(buf), x - 27, y); // was "Score ", width 25
  StrPrintF(buf, "S."); // left-justify
  WinDrawChars(buf, StrLen(buf), x+1, y);//S looks better if we sorta center it
  x += 10; // width of "M "
  WinDrawChars(buf, StrLen(buf), x, y);
  StrPrintF(buf, "Race"); // left-justify
  WinDrawChars(buf, StrLen(buf), x, y);
  x += 42; // width of "Half-Troll "
  StrPrintF(buf, "Class"); // left-justify
  WinDrawChars(buf, StrLen(buf), x, y);
  x = 160;
  StrPrintF(buf, "Level"); // right-justify   ("Lvl" is 12 pixels but looks
  WinDrawChars(buf, StrLen(buf), x - 22, y); // really weird in this font.)
  WinDrawLine(0, y+10, 160, y+10);
}

// Returns index of the next one to draw..
static Short draw_topscores(Boolean reset)
{
  static Short i = 0;
  ULong uniqueID = REC_TOPSCORES;
  UInt index;
  VoidHand vh;
  VoidPtr p = NULL;
  Short scores_ctr, *tmp;
  high_scores *scores;
  //
  Char buf[80];
  UChar prace, pclass;
  Short x, y, diedlen;
  Boolean terse;
  //
  if (0 != DmFindRecordByID(MoriaSaveDB, uniqueID, &index)) {
    WinDrawChars("None", 4, 20, 20);
    return -1;
  }
  vh = DmQueryRecord(MoriaSaveDB, index);
  p = MemHandleLock(vh);
  tmp = (Short *) p;
  scores_ctr = tmp[0];
  scores = (high_scores *) &(tmp[1]);

  if (scores_ctr <= 0) {
    WinDrawChars("None", 4, 20, 20);
    return -1;
  }
  draw_topscores_header();
  StrPrintF(buf, "n/%d", scores_ctr);
  WinDrawChars(buf, StrLen(buf), 160-25, 148);

  y = 15 + 12;
  x = 0;

  if (reset || (i >= scores_ctr)) i = 0;

  for ( ; i < scores_ctr ; i++) {
    diedlen = FntCharsWidth(scores[i].died_from, StrLen(scores[i].died_from));
    terse = false;
    if (6 + diedlen + 21 + 2 + FntCharsWidth(scores[i].name,
					     StrLen(scores[i].name)) < 160)
      terse = true; // add "died: " (21) line 2, or "killed by: " (38), line3.
    // Check whether we will have enough room.
    // We will take 22 height if terse, 33 height if !terse.  Button at 147.
    if ((y + (terse ? 22 : 33)) > 147)
      break; // when we go to the next page, start with *this* value of i..
    // Ok, we have enough space to print this record.
    // Line 1: Numbers
    // Skip to line 2: Name
    y += 10;
    x = 6;
    WinDrawChars(scores[i].name, StrLen(scores[i].name), x, y);
    if (terse) {
      StrPrintF(buf, "died: %s", scores[i].died_from);
      WinDrawChars(buf, StrLen(buf), 160-(diedlen+21), y);
    }
    y -= 10;
    // Skip back to line 1
    x = 0;
    StrPrintF(buf, "%d", i+1); // Rank
    WinDrawChars(buf, StrLen(buf), x, y);
    x += 17 + 42; // for right-justified column.  width of "000 00000000 "
    StrPrintF(buf, "%ld", scores[i].points); // Score
    // Width of a [0-9]+ string in the standard font = 5 * StrLen. ' ' is 2 pix
    WinDrawChars(buf, StrLen(buf), x-(5*StrLen(buf)+2), y);
    buf[0] = ((scores[i].race_class_sex & 0x1) ? 'M' : 'F'); // Gender
    WinDrawChars(buf, 1, x, y);
    x += 10; // width of "M "
    prace = (scores[i].race_class_sex >> 4) & 0x7;;
    StrPrintF(buf, "%s", race[prace].trace); // Race
    WinDrawChars(buf, StrLen(buf), x, y);
    x += 42; // width of "Half-Troll "
    pclass = (scores[i].race_class_sex >> 1) & 0x7;
    StrPrintF(buf, "%s", class[pclass].title); // Class  
    WinDrawChars(buf, StrLen(buf), x, y);
    x = 160; // for right-justified column.  at right edge of screen.
    StrPrintF(buf, "%d", scores[i].lev); // Level (experience, not dungeon)
    WinDrawChars(buf, StrLen(buf), x-5*StrLen(buf), y);
    // Skip over line 2:
    y += 10;
    // Line 3: COD
    if (!terse) {
      y += 11;
      x = 13;
      StrPrintF(buf, "killed by: %s", scores[i].died_from);
      WinDrawChars(buf, StrLen(buf), x, y);
    }
    WinDrawGrayLine(0, y+11, 160, y+11);  
    y += 12;
  }

  MemHandleUnlock(vh);  
  return i;
}


