/*********************************************************************
 * kMoria - Moria adapted for the PalmPilot.                         *
 * Copyright (C) 2000 Bridget Spitznagel                             *
 *                                                                   *
 * This program is derived from Umoria 5.5.2.  Moria is              *
 *   Copyright (c) 1989-94 James E. Wilson, Robert A. Koeneke        *
 *                                                                   *
 * This program is free software; you can redistribute it and/or     *
 * modify it under the terms of the GNU General Public License       *
 * as published by the Free Software Foundation; either version 2    *
 * of the License, or (at your option) any later version.            *
 *                                                                   *
 * This program is distributed in the hope that it will be useful,   *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of    *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     *
 * GNU General Public License for more details.                      *
 *                                                                   *
 * You should have received a copy of the GNU General Public License *
 * along with this program; if not, write to                         *
 * The Free Software Foundation, Inc.,                               *
 * 59 Temple Place - Suite 330,                                      *
 * Boston, MA  02111-1307, USA.                                      *
 *********************************************************************/
#include <Pilot.h>
#include "kMoria.h"
#include "lock-externs.h"

static void create_save_structs() SEC_5;
static Boolean create_py_rec() SEC_5;
static Boolean read_py_rec() SEC_5;
static void read_save_structs() SEC_5;

Boolean save_char()
{
  disturb(1, 0);		/* Turn off resting and searching. */
  change_speed(-pack_heavy);	/* Fix the speed */
  pack_heavy = 0;

  create_save_structs();
  // maybe need to write bytes:  major version, minor version, patchlevel
  return create_py_rec(); // yay.
}
Boolean load_char()
{
  if (!read_py_rec()) return false; // saved char does not exist..
  read_save_structs();
  return true;
}
// Dude.  I get a fatal error when the record does not exist.  WHY.
void antisave_char()
{
  ULong uniqueID;
  UInt index;

  uniqueID = REC_PY;
  if (0 == DmFindRecordByID(MoriaSaveDB, uniqueID, &index))
    DmRemoveRecord(MoriaSaveDB, index);      // EX TERM IN ATE !
}

typedef struct my_save_struct
{
  ULong spell_learned;
  ULong spell_worked;
  ULong spell_forgotten;
  ULong town_seed; // skip randes_seed
  ULong current_time; // use TimGetWhatever
  ULong birth_date; // in create.c
  Long turn;
  Long total_points; // call total_points() to calc this
  Short missile_ctr;
  Short inven_ctr;
  Short inven_weight;
  Short equip_ctr;
  Short tcptr;
  Short mfptr;
  Boolean total_winner;
  Boolean death;
  // Skipping: panic_save, noscore.
  Char died_from[25];
  // Also save the previous-message-foo!
} my_save;
my_save my_save_info;
typedef struct floor_save_struct
{
  Short dun_level;
  Short char_row;
  Short char_col;
  Short mon_tot_mult;
  Short cur_height;
  Short cur_width;
  Short max_panel_rows;
  Short max_panel_cols;
} floor_save;
floor_save floor_save_info;


static void create_save_structs()
{
  my_save_info.spell_learned = spell_learned;     
  my_save_info.spell_worked = spell_worked;
  my_save_info.spell_forgotten = spell_forgotten;
  my_save_info.town_seed = town_seed;
  my_save_info.current_time = TimGetSeconds();
  my_save_info.birth_date = birth_date;
  my_save_info.turn =  turn;
  my_save_info.total_points = total_points();
  my_save_info.missile_ctr = missile_ctr;
  my_save_info.inven_ctr = inven_ctr;
  my_save_info.inven_weight = inven_weight;
  my_save_info.equip_ctr = equip_ctr;
  my_save_info.tcptr = tcptr;
  my_save_info.mfptr = mfptr;
  my_save_info.total_winner = total_winner;
  my_save_info.death = death;  // Skipping: panic_save, noscore.
  StrNCopy(my_save_info.died_from, died_from, 25);
  floor_save_info.dun_level = dun_level;
  floor_save_info.char_row = char_row;
  floor_save_info.char_col = char_col;
  floor_save_info.mon_tot_mult = mon_tot_mult;
  floor_save_info.cur_height = cur_height;
  floor_save_info.cur_width = cur_width;
  floor_save_info.max_panel_rows = max_panel_rows;
  floor_save_info.max_panel_cols = max_panel_cols;
}
static void read_save_structs()
{
  spell_learned   = my_save_info.spell_learned;     
  spell_worked    = my_save_info.spell_worked;
  spell_forgotten = my_save_info.spell_forgotten;
  town_seed       = my_save_info.town_seed;
  //  current_time    = my_save_info.TimGetSeconds();
  birth_date      = my_save_info.birth_date;
  turn            = my_save_info.turn;
  max_score       = my_save_info.total_points;
  missile_ctr     = my_save_info.missile_ctr;
  inven_ctr       = my_save_info.inven_ctr;
  inven_weight    = my_save_info.inven_weight;
  equip_ctr       = my_save_info.equip_ctr;
  tcptr       = my_save_info.tcptr;
  mfptr       = my_save_info.mfptr;
  total_winner    = my_save_info.total_winner;
  death           = my_save_info.death;
  StrNCopy(died_from, my_save_info.died_from, 25);
  dun_level      = floor_save_info.dun_level;
  char_row       = floor_save_info.char_row;
  char_col       = floor_save_info.char_col;
  mon_tot_mult   = floor_save_info.mon_tot_mult;
  cur_height     = floor_save_info.cur_height;
  cur_width      = floor_save_info.cur_width;
  max_panel_rows = floor_save_info.max_panel_rows;
  max_panel_cols = floor_save_info.max_panel_cols;
}
// Create the py storage record (will delete it if it already exists)
static Boolean create_py_rec()
{
  Boolean ok = false;
  ULong uniqueID;
  UInt index;
  VoidHand vh;
  VoidPtr p;
  ULong size = sizeof(struct flags) + sizeof(struct stats);
  size += sizeof(struct misc);
  size += sizeof(UChar) * (MAX_SPELLS+1);
  size += sizeof(Short) * MAX_PLAYER_LEVEL;
  size += sizeof(my_save) + sizeof(floor_save);
  //  size += sizeof(Char) * 240;
  // Delete existing py record
  uniqueID = REC_PY;
  if (0 == DmFindRecordByID(MoriaSaveDB, uniqueID, &index))
    DmRemoveRecord(MoriaSaveDB, index);      // EX TERM IN ATE !
  index = DmNumRecords(MoriaSaveDB);
  vh = DmNewRecord(MoriaSaveDB, &index, size);
  p = MemHandleLock(vh);
  if (p) {
    Short offset = 0;
    // Write things.
    DmWrite(p, offset, &pymisc, sizeof(struct misc));
    offset += sizeof(struct misc);
    DmWrite(p, offset, &pystats, sizeof(struct stats));
    offset += sizeof(struct stats);
    DmWrite(p, offset, &pyflags, sizeof(struct flags));
    offset += sizeof(struct flags);
    //  player_hp;
    DmWrite(p, offset, player_hp, sizeof(Short) * MAX_PLAYER_LEVEL);
    offset += sizeof(Short) * MAX_PLAYER_LEVEL;
    //  spell_order;
    DmWrite(p, offset, spell_order, sizeof(UChar) * (MAX_SPELLS+1));
    offset += sizeof(UChar) * (MAX_SPELLS+1);
    //  my_save;
    DmWrite(p, offset, &my_save_info, sizeof(my_save));
    offset += sizeof(my_save);
    //  floor_save;
    DmWrite(p, offset, &floor_save_info, sizeof(floor_save));
    offset += sizeof(floor_save);
    //  don't forget history_block..
    //    DmWrite(p, offset, &pyflags, sizeof(struct flags));
    //    offset += sizeof(struct flags);
    //    DmStrCopy(p, offset, history_block); // hm, hope that works
    ok = true;
  }
  MemPtrUnlock(p);
  DmReleaseRecord(MoriaSaveDB, index, true);
  DmSetRecordInfo(MoriaSaveDB, index, NULL, &uniqueID); // set its uid!
  return ok;
}
static Boolean read_py_rec()
{
  UInt recindex;
  VoidHand vh;
  VoidPtr p;
  //  Char *hist;
  Err err;
  ULong uniqueID;
  //
  uniqueID = REC_PY;
  err = DmFindRecordByID(MoriaSaveDB, uniqueID, &recindex);
  if (err) return false; // didn't find it
  vh = DmQueryRecord(MoriaSaveDB, recindex);
  p = MemHandleLock(vh);
  if (!p) return false;   // we're in some kind of trouble

  // Ok now read a whole bunch of stuff.
  err = MemMove(&pymisc, p, sizeof(struct misc));
  if (err) { MemHandleUnlock(vh); return false; }
  p += sizeof(struct misc);
  err = MemMove(&pystats, p, sizeof(struct stats));
  if (err) { MemHandleUnlock(vh); return false; }
  p += sizeof(struct stats);
  err = MemMove(&pyflags, p, sizeof(struct flags));
  if (err) { MemHandleUnlock(vh); return false; }
  p += sizeof(struct flags);
  err = MemMove(player_hp, p, sizeof(Short) * MAX_PLAYER_LEVEL);
  if (err) { MemHandleUnlock(vh); return false; }
  p += sizeof(Short) * MAX_PLAYER_LEVEL;
  err = MemMove(spell_order, p, sizeof(UChar) * (MAX_SPELLS+1));
  if (err) { MemHandleUnlock(vh); return false; }
  p += sizeof(UChar) * (MAX_SPELLS+1);
  err = MemMove(&my_save_info, p, sizeof(my_save));
  if (err) { MemHandleUnlock(vh); return false; }
  p += sizeof(my_save);
  err = MemMove(&floor_save_info, p, sizeof(floor_save));
  if (err) { MemHandleUnlock(vh); return false; }
  //  p += sizeof(floor_save);
  //  hist = (Char *) p;
  //  StrNCopy(history_block, p, 240);
  MemHandleUnlock(vh);
  return true;
}
/*
// Create the miscsave storage record (will delete it if it already exists)
static Boolean create_miscsave_rec()
{
  Boolean ok = false;
  ULong uniqueID;
  UInt index;
  VoidHand vh;
  VoidPtr p;
  ULong 
  // Delete existing miscsave record
  uniqueID = REC_MISCSAVE;
  if (0 == DmFindRecordByID(MoriaSaveDB, uniqueID, &index))
    DmRemoveRecord(MoriaSaveDB, index);      // EX TERM IN ATE !
  index = DmNumRecords(MoriaSaveDB);
  vh = DmNewRecord(MoriaSaveDB, &index, size);
  p = MemHandleLock(vh);
  if (p) {
    Short offset = 0;
    // Write things.
    ok = true;
  }
  MemPtrUnlock(p);
  DmReleaseRecord(MoriaSaveDB, index, true);
  DmSetRecordInfo(MoriaSaveDB, index, NULL, &uniqueID); // set its uid!
  return ok;
}

*/
