/*
 * File: format.c
 * Author: Holger Klawitter
 * Date: 28.07.1997
 *
 * Printing Floating point numbers.
 *
 * Changes:
 * 05.10.1999 support for multiple bases
 */

/* Derived from libc.0.1.2 */

#include "view.h"

#ifdef pc
#  define Int int
#  define CharPtr char*
#  define ULong unsigned long
#  define EXPCHAR 'e' /* not the one from pilot, but better for debugging */
#  include "format.h"
#else
#  include "format.h"
#  include "stk.h"
#endif

#define PSH(X) (*(st++)=(X))
#define PSHDigit(d) *(st++) = ((d)>=10 ? (d)+'A'-10 : (d)+'0')

void ftoanrm(CharPtr st, double f, Int16 flags, Int16 prec);
void ftoaexp(CharPtr st, double f, Int16 flags, Int16 prec);

void hkftoa(CharPtr st, double f, Int16 flags)
{
  Int16 prec = flags&15, i;
  double up = 0.5, tmp;
  union { double f; UInt32 i,j; } caster;

  caster.f = f;
  if( (caster.i & 0x7ff00000) == 0x7ff00000 ) {
    PSH('X');
    PSH(0);
    return;
  }

  flags &= 240;
  if (f < 0) {
    PSH('-');
    f = -f;
  }

  if( (baseMode == 10) &&
			((flags & (FMT_EXP|FMT_ENG)) || (f && (f<1.0e-9 || f>1.0e9)))
	)	{
    if( f ) {
      tmp = f;
      while( tmp>10.0 ) { tmp/=10; up*=10; }
      while( tmp<1.0 ) { tmp*=10; up/=10; }
      for(i=0; i<prec; i++) { up/=10; }
      f += up;
    }
    ftoaexp(st,f,flags,prec);
  } else {
    for(i=0;i<prec;i++) { up/=baseMode; }
    f += up;
    ftoanrm(st,f,flags,prec);
  }

}

void ftoanrm(CharPtr st, double f, Int16 flags, Int16 prec)
{
  Int16 i;
  Int16 z;
  Int16 dot = 0; /* position of dot (behind index 0) */
  Int16 zero = 0; /* leading zeros */

  if (f) {
    while (f < 1) {
      f *=10;
      zero++;
    }

    while (f >= baseMode) {
      f /= baseMode;
      dot++;
    }
  }

  if( zero-- ) {
    PSH( '0' );
  } else {
    for(i=0;i<=dot && i<LINESIZE-3;i++) {
      z = f;
      PSHDigit(z);
      f -= z;
      f *= baseMode;
    }
		if( i==LINESIZE-3 ) PSH('.');
  }
  PSH('.');
	if( baseMode == 10 ) {
		for (i=0;i<prec;i++) {
			if( i<zero ) {
				PSH('0');
			} else {
				z = f;
				PSHDigit(z);
				f -= z;
				f *= baseMode;
			}
		}
	}

  PSH(0);
}

void ftoaexp(CharPtr st, double f, Int16 flags, Int16 prec)
{
  Int16 i;
  Int16 z;
  Int16 dot = 0; /* dot begin char index 0 */
  Int16 exp = 0;

  if (f) {
    while (f < 1) {
      f *=10;
      exp--;
    }

    while (f >= 10) {
      f /=10;
      exp++;
    }

    if( flags & FMT_ENG ) {
      if( exp%3 ) {
	if( exp<0 ) {
	  dot += (exp%3+3); /* -1 => 2, -2 => 1, -3 => 0, 0 won't enter here */
	  exp -= (exp%3+3);
	} else {
	  dot += (exp%3);
	  exp -= (exp%3);
	}
      }
    }
  }

  if( prec<=dot ) { prec=dot+1; } /* at least to the dot */

  for (i=0;i<prec;i++) {
    z = f;
    PSH('0'+z);
    f -= z;
    f *= 10;
    if( i==dot && i+1<prec ) PSH('.');
  }

  PSH(EXPCHAR);
  if (exp < 0) {
    PSH('-');
    exp = -exp;
  }

  PSH('0'+exp/10);
  exp -= (exp/10) * 10;
  PSH('0'+exp);

  PSH(0);
}

#ifdef pc

#include <stdio.h>
#include <stdlib.h>

Int main( Int16 args, CharPtr argv[] ) {
  double f1,f2;
  Int16 flags = 0;
  char res[40];
  if( args!=4 ) {
    fprintf( stderr, "format-2 f1 f2 flags\n" );
    return 1;
  }
  flags = atoi(argv[3]);
  f1 = atof(argv[1]);
  f2 = atof(argv[2]);

  hkftoa( res, f1+f2, flags );
  fprintf( stdout, "%f => %s\n", f1+f2, res );

  return 0;
}

#endif
