/*
 * File: kalk.c
 * Author: Holger Klawitter
 * Datum: 14.07.1997
 *
 * Changes:
 * 05.10.1999 support for different bases
 * 15.04.2000 copy and paste added
 */

#include <PalmOS.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>
#include <UI/UIColor.h>

#include "kalk.h"
#include "stk.h"
#include "model.h"
#include "view.h"

/***********************************************************************
 * Helper functions
 ***********************************************************************/

float angleConvert;
double lastx;

Boolean colorDevice;

UInt16 DispID[] = { OptAll, OptFixed, OptSci, OptEng, };
UInt16 AngID[] = { OptRad, OptDeg, };

VoidPtr GetObjectPtr( UInt16 objectId ) {
  FormPtr frm;
  frm = FrmGetActiveForm();
  return FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, objectId ) );
}

UInt16 HELP( UInt16 alertID ) {
  return FrmAlert( alertID );
}

void DrawButton( UInt16 id, Boolean lift ) {
	FormPtr frm = FrmGetActiveForm();
	MemHandle h;
	VoidPtr data;
	Int16 idx;
	Coord x,y;
	IndexedColorType oldFore;
	IndexedColorType oldBack;

	if( !colorDevice ) return;

	h = DmGetResource( 'Tbmp', id+2000 );
	if( !h ) return;
	idx = FrmGetObjectIndex( frm, id+2000 );

	if( idx != -1 ) {
		if( lift ) {
			FrmShowObject( frm, idx );
		} else {
			data = MemHandleLock( h );
			FrmGetObjectPosition( frm, idx, &x, &y );
			oldFore = WinSetForeColor
				( UIColorGetTableEntryIndex(UIObjectSelectedForeground) );
			oldBack = WinSetBackColor
				( UIColorGetTableEntryIndex(UIObjectSelectedFill) );
			WinDrawBitmap( data, x, y );
			WinSetForeColor( oldFore );
			WinSetBackColor( oldBack );
			MemHandleUnlock( h );
		}
	}
	DmReleaseResource( h );
}

/***********************************************************************
 * Event Handler for the MathForm
 ***********************************************************************/

Boolean MathFormsHandleEvent( EventPtr eptr )
{
  Int16 ctl;

  switch( eptr->eType ) {
  case ctlSelectEvent:
		DrawButton( eptr->data.ctlSelect.controlID, true );
    switch( ctl=eptr->data.ctlSelect.controlID ) {

      /* Normal non-nary, unary operations */
    case MtRez: case MtSqr: case MtSqrt:
    case MtLog: case Mt10pw:case MtLn:  case MtExp:
    case MtSin: case MtCos: case MtTan: case MtPi:
    case MtASin:case MtACos:case MtATan:
    case MtSinH: case MtCosH: case MtTanH:
    case MtASinH:case MtACosH:case MtATanH:
    case StXAv: case StYAv: case StXVa: case StYVa:
    case StToX: case StToY: case Stn: case StClear:
    case StR: case StBXY: case StBYX: case MtProz: case MtFak:
      
      FrmReturnToForm( MainForm );
      DoOperate( ctl );
      return true;

      /* binary operations */
    case StAdd: case StDel: case MtPow:
      FrmReturnToForm( MainForm );
      DoOperate( ctl-1000 );
      return true;


      /* Leaving current form */
    case MtBack: case StBack:
      FrmReturnToForm( MainForm );
      return true;
    }
    return false;

	case ctlExitEvent:
		DrawButton( eptr->data.ctlSelect.controlID, true );
		break;

  case keyDownEvent:
    switch( ctl=eptr->data.keyDown.chr ) {
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
			if( ctl - '0' >= baseMode ) return false;
			/* fall through */
    case '.':
      FrmReturnToForm( MainForm );
      DoAdd( ctl );
      return true;
		case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
			if( baseMode == 16 ) {
				FrmReturnToForm( MainForm );
				DoAdd( ctl - 'a' + 'A' );
				return true;
			} else if( ctl == 'e' ) {
				FrmReturnToForm( MainForm );
				DoAdd( 'E' );
				return true;
			}
			break;
    case '\n':
      FrmReturnToForm( MainForm );
      DoEnter();
      return true;
    case '\010':
      FrmReturnToForm( MainForm );
      DoBackspace();
      return true;
    }
    break;

  case frmOpenEvent:
    FrmDrawForm(FrmGetActiveForm());
    return true;

  default:
  }

  return false;
}

/***********************************************************************
 * Event handler for the about form.
 ***********************************************************************/

Boolean AboutFormHandleEvent( EventPtr eptr )
{
  switch( eptr->eType ) {
  case ctlSelectEvent:
    if( eptr->data.ctlSelect.controlID == AboutOk ) {
      FrmReturnToForm( MainForm );
      return true;
    }
    return false;
  case frmOpenEvent:
    FrmDrawForm(FrmGetActiveForm());
    return true;
  default:
  }

  return false;
}

/***********************************************************************
 * Event handler for the memory form.
 ***********************************************************************/

double registers[MEMSIZE] =
{ 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
Int16 openingControl = 0;

Boolean MemoryFormHandleEvent( EventPtr eptr )
{
  Int16 i,j;
  CharPtr lblstr;
  ControlPtr button;

  switch( eptr->eType ) {
  case ctlSelectEvent:
    i = eptr->data.ctlSelect.controlID;
    if( i == MemClear ) {
      if( FrmCustomAlert( AreYouSure,
													"Do you really want to flush all registers?",
													NULL, NULL ) == 1 ) {
				for( j=0; j<MEMSIZE; j++ ) {
					registers[j] = 0.0;
				}
      }
    }
    FrmReturnToForm( MainForm );
    if( i >= MemClear /* == MemExit or MemClear */  ) return true;

    switch( openingControl ) {
    case BtRcl:
      if( StkSize==0 && state==enterState ) {
      } else {
				StkDup();
      }      
      ViewFltToStr(registers[i-MemButton],StkLine);
      StkSetTop(registers[i-MemButton]);
      state = opState;
      ViewUScore = false;
      ViewUpdateAll();     
      break;

    case BtSto:
      if( StkSize==0 && state==enterState ) {
				FrmAlert( NotEnoughArgs );
      } else {
				StkSetTop( registers[i-MemButton] = StkTop() );
				state=opState;
      }
      break;

    default:
    }
    return true;
  case frmOpenEvent:
    FrmDrawForm(FrmGetActiveForm());
    for( i=0; i<MEMSIZE; i++ ) {
      Int16 j;
      button = GetObjectPtr( MemButton+i );
      if( StrLen( lblstr=CtlGetLabel( button ) )==0 ) {
				/* we need memory for labels */
				lblstr=MemPtrNew( sizeof(Char)*20 );
      }
      ViewFltToStr( registers[i], lblstr );

      /* label font has normal exp char, we have to convert here.
       * ?? clumsy, we should convert for the irrgular font, not for the
       * reglar
       */
      for(j=StrLen(lblstr)-1;j>=0;j--) {
				if( lblstr[j]==EXPCHAR ) { lblstr[j]='e'; }
      }

      CtlSetLabel( button, lblstr );
    }
    return true;
  default:
  }

  return false;
}

/***********************************************************************
 * Event handler for options form.
 ***********************************************************************/

void HandleDigits( UInt16 localMode ) {
  FormPtr frm = FrmGetActiveForm();
  UInt16 precIdx = FrmGetObjectIndex( frm, OptPrec ); 
  UInt16 labelIdx = FrmGetObjectIndex( frm, OptPrecLabel ); 
  if( localMode != OptAll ) {
    FrmShowObject( frm, precIdx );
    FrmShowObject( frm, labelIdx );
  } else {
    FrmHideObject( frm, precIdx );
    FrmHideObject( frm, labelIdx );
  }
}

/* Reads the global variables and puts content into the corresponding
 * option fields.
 */
void SetOptionWidgets() {
  FormPtr frm = FrmGetActiveForm();
  CharPtr text;
  MemHandle hand;
  FieldPtr fld;
	UInt16 n;

  FrmSetControlGroupSelection( frm, OptDisplayGroup, DispID[displayMode] );
  FrmSetControlGroupSelection( frm, OptAngleGroup, AngID[angleMode] );

	switch( baseMode ) {
	case 2: n = OptBin; break;
	case 8: n = OptOct; break;
	case 16: n = OptHex; break;
	default:
		n = OptDec;
	}
  FrmSetControlGroupSelection( frm, OptBaseGroup, n );
	
  fld = GetObjectPtr( OptPrec );
  hand = (MemHandle)FldGetTextHandle( fld );
  if( !hand ) {
    FldSetTextHandle( fld, (VoidPtr)(hand=MemHandleNew(2)) );
  }
  text = MemHandleLock( hand );
  *text++ = places + '0';
  *text = 0;
  FldSetTextHandle( fld, hand ); /* Notify field about change. */
  MemHandleUnlock( hand );

  HandleDigits( DispID[displayMode] );
}

Boolean OptFormHandleEvent( EventPtr eptr )
{
  FormPtr frm = FrmGetActiveForm();
  UInt16 id;
  MemHandle hand;
  CharPtr text;
	UInt8 newBaseMode;
  
  switch( eptr->eType ) {
  case ctlSelectEvent:
    switch( (id=eptr->data.ctlSelect.controlID) ) {
    case OptAll:
    case OptFixed:
    case OptSci:
    case OptEng:
      HandleDigits( id );
      return true;
    case OptDefault:
      displayMode = fixedMode;
      places = 4;
      angleMode = radMode;
      SetOptionWidgets();
      return true;
    case OptOk:
      angleMode = FrmGetObjectId
				( frm, FrmGetControlGroupSelection(frm,OptAngleGroup) ) == OptDeg;
      angleConvert = (angleMode==0) ? 1.0 : 1.74532925199e-2;
      id = FrmGetObjectId( frm, FrmGetControlGroupSelection
													 ( frm, OptDisplayGroup ));
      switch( id ) {
      case OptAll: displayMode = freeMode; break;
      case OptFixed: displayMode = fixedMode; break;
      case OptSci: displayMode = sciMode; break;
      case OptEng: displayMode = engMode; break;
      }
			id = FrmGetObjectId( frm, FrmGetControlGroupSelection
													 (frm, OptBaseGroup ));
			switch( id ) {
			case OptBin: newBaseMode = 2; break;
			case OptOct: newBaseMode = 8; break;
			case OptHex: newBaseMode = 16; break;
			case OptDec:
			default:
				newBaseMode = 10;
			}
      hand = FldGetTextHandle( GetObjectPtr( OptPrec ) );
      if( hand ) {
				text = MemHandleLock( hand );
				if( text && *text>='1' && *text<='9' ) {
					places = *text - '0';
				}
				MemHandleUnlock( hand );
      } 
      FrmReturnToForm( MainForm );
			if( baseMode != newBaseMode ) {
				DoEnter();
				baseMode = newBaseMode;
			}
      DoUpdateLine();
      ViewUpdateAll();
      return true;
    case OptCancel:
      FrmReturnToForm( MainForm );
      return true;
    }
    return false;
  case frmOpenEvent:
    SetOptionWidgets();
    FrmDrawForm(frm);
    FldSetInsPtPosition( GetObjectPtr( OptPrec ), 1 );
    FrmSetFocus( frm, FrmGetObjectIndex( frm, OptPrec ));
    return true;
  default:
  }

  return false;
}

/***********************************************************************
 * Event Handler for the MainForm
 ***********************************************************************/

Boolean MainFormHandleEvent( EventPtr eptr )
{
  Int16 ctl;
  switch( eptr->eType ) {

  case ctlSelectEvent:
		DrawButton( eptr->data.ctlSelect.controlID, true );
    switch( ctl=eptr->data.ctlSelect.controlID ) {
    case BtMath:
      FrmPopupForm( MathForm );
      return true;
    case BtStat:
      FrmPopupForm( StatForm );
      return true;

      /* entering parts of number */
    case Bt0:case Bt1:case Bt2:case Bt3:case Bt4:
    case Bt5:case Bt6:case Bt7:case Bt8:case Bt9:
			if( baseMode > ctl - Bt0 ) {
				DoAdd( ctl + '0' - Bt0 );
				return true;
			}
			break;
    case BtDot: DoAdd('.'); return true;
    case BtE: DoAdd('E'); return true;
    case BtEnt: DoEnter(); return true;
    case BtDel: DoBackspace(); return true;
    case BtChs: DoChangeSign(); return true;

      /* unary operations */
    case BtRez: case BtSqr: case BtSqrt: case BtLastX:
      DoOperate( ctl ); return true;

      /* binary operations */
    case BtAdd: case BtSub: case BtMult: case BtDiv: case BtSwp: case BtPow:
      DoOperate( ctl - 1000 ); return true;

      /* Memory */
    case BtSto: case BtRcl:
      openingControl = ctl;
      FrmPopupForm( MemForm ); return true;

    }
	case ctlExitEvent:
		DrawButton( eptr->data.ctlSelect.controlID, true );
		break;

  case keyDownEvent:
    switch( ctl=eptr->data.keyDown.chr ) {
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
			if( ctl - '0' >= baseMode ) return false;
			/* fall through */
    case '.': 
      DoAdd( ctl );
      return true;
		case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
			if( baseMode == 16 ) {
				DoAdd( ctl - 'a' + 'A' );
				return true;
			} else if( ctl == 'e' ) {
				DoAdd( 'E' );
				return true;
			}
			break;
    case '\n':
      DoEnter();
      return true;
    case '\010':
      DoBackspace();
      return true;
    case pageUpChr:
      ViewUp();
      return true;
    case pageDownChr:
      ViewDown();
      return true;
    }
  case menuEvent:
    switch( eptr->data.menu.itemID ) {
    case MenuOptions:
      FrmPopupForm( OptForm );
      return true;
    case MenuAbout:
      FrmPopupForm( AboutForm );
      return true;
    case MenuClear:
      StkSize=*StkLine=0;
      state=enterState;
      ViewUScore=false;
      ViewUpdateAll();
      return true;
		case MenuCopy:
			DoCopy();
			return true;
		case MenuPaste:
			DoPaste();
			return true;
    }
    return false;
  case frmOpenEvent:
    FrmDrawForm(FrmGetActiveForm());
    ViewUpdateAll();
    return true;
  default:
  }

  return false;
}

/***********************************************************************
 * MAIN PROGRAM STUFF
 ***********************************************************************/

typedef struct {
  double stk[STACKSIZE];
  double regs[MEMSIZE];
  Int16 StkSize;
  Char StkLine[LINESIZE];
  Int16 places;
  AngleMode angleMode;
  ViewMode displayMode;
	UInt8 baseMode;
  StackState state;
  Statistics S;
  double stacktop;
	double lastx;
  Boolean ViewUScore;
} KalkPrefs;

/***********************************************************************
 * Starting and ending the application
 **********************************************************************/
static UInt16 StartApplication()
{
  KalkPrefs prefs;
  Int16 i;
	UInt32 romVersion;
	UInt16 osVersion;
	UInt32 dummy;

  if( PrefGetAppPreferencesV10( KALKTYPE, KALKVERSION,
																&prefs, sizeof(KalkPrefs) ) ) {
    for( i=0; i<STACKSIZE; i++ ) stk[i] = prefs.stk[i]; 
    for( i=0; i<LINESIZE; i++ ) StkLine[i] = prefs.StkLine[i]; 
    for( i=0; i<MEMSIZE; i++ ) registers[i] = prefs.regs[i];
    places = prefs.places;
    StkSize = prefs.StkSize;
    angleMode = prefs.angleMode;
    displayMode = prefs.displayMode;
		baseMode =  prefs.baseMode;
    state = prefs.state;
    S = prefs.S;
    stacktop = prefs.stacktop;
    ViewUScore = prefs.ViewUScore;
		lastx = prefs.lastx;
  } else {
		baseMode = 10;
	}
  angleConvert = (angleMode == OptRad) ? 1.0 : 1.74532925199e-2;

	FtrGet( sysFtrCreator, sysFtrNumROMVersion, &romVersion );
	osVersion = romVersion >> 20;
	if( osVersion >= 0x035 ) {
		WinScreenMode( winScreenModeGetDefaults,
									 &dummy, &dummy, &dummy, &colorDevice );
	}
  return 0;
}

static void StopApplication()
{
  KalkPrefs prefs;
  Int16 i;
  for( i=0; i<STACKSIZE; i++ ) prefs.stk[i] = stk[i];
  for( i=0; i<LINESIZE; i++ ) prefs.StkLine[i] = StkLine[i];
  for( i=0; i<MEMSIZE; i++ ) prefs.regs[i] = registers[i];
  prefs.places = places;
  prefs.StkSize = StkSize;
  prefs.angleMode = angleMode;
  prefs.displayMode = displayMode;
  prefs.baseMode = baseMode;
  prefs.state = state;
  prefs.stacktop = stacktop;
  prefs.S = S;
  prefs.ViewUScore = ViewUScore;
	prefs.lastx = lastx;
  PrefSetAppPreferencesV10( KALKTYPE, KALKVERSION, &prefs, sizeof(KalkPrefs) );
}

/***********************************************************************
 * The master Event handler stuff
 ***********************************************************************/
static Boolean ApplicationHandleEvent( EventPtr eptr )
{
  FormPtr frm;
  Int16 formId;
  Boolean handled = false;

  if( eptr->eType == frmLoadEvent ) {
    formId = eptr->data.frmLoad.formID;

    /* Why not FrmGotoForm ?? */
    frm = FrmInitForm( formId );
    FrmSetActiveForm( frm );

    switch( formId ) {
    case MainForm:
      FrmSetEventHandler( frm, (MainFormHandleEvent) );
      break;
    case AboutForm:
      FrmSetEventHandler( frm, (AboutFormHandleEvent) );
      break;
    case StatForm:
    case MathForm:
      FrmSetEventHandler( frm, (MathFormsHandleEvent) );
      break;
    case OptForm:
      FrmSetEventHandler( frm, (OptFormHandleEvent) );
      break;
    case MemForm:
      FrmSetEventHandler( frm, (MemoryFormHandleEvent) );
      break;
    }
    handled = true;
  }  
  return handled;
}

static void EventLoop()
{
  EventType e;
  UInt16 err;

  do {
    EvtGetEvent( &e, evtWaitForever );
    if( !SysHandleEvent( &e ) &&
				!MenuHandleEvent( NULL, &e, &err ) &&
				!ApplicationHandleEvent( &e )
				) {
      FrmDispatchEvent( &e );
    }
  } while( e.eType != appStopEvent );
}

/***********************************************************************
 * The Main loop
 ***********************************************************************/
UInt32 PilotMain(UInt16 cmd, VoidPtr cmdPBP, UInt16 launchFlags)
{
  UInt16 err;

  if( cmd == sysAppLaunchCmdNormalLaunch ) {
    err = StartApplication();
    if( err ) {
      return err;
    }
    FrmGotoForm( MainForm );
    EventLoop();
    StopApplication();
  }

  return 0;
}
