/*
 * File: model.c
 * Author: Holger Klawitter
 * Date: 15.07.1997
 *
 * The model controlling the upn calculator.
 * This is somehow the real main program :)
 *
 * Changes:
 * 05.10.1999 support for hex digits
 * 06.03.2000 lastx support
 * 15.04.2000 copy and paste support
 */

#include "stk.h"
#include "kalk.h"
#include "view.h"
#include "model.h"
#include "ops.h"

#include <System/SysAll.h>
#include <UI/UIAll.h>

/**********************************************************************
 * public variables.
 **********************************************************************/

AngleMode angleMode = radMode;

Statistics S = { 0.0, 0.0, 0.0, 0.0, 0.0, 0 };

/*********************************************************************
 * Functions
 **********************************************************************/

void DoEnter() {
		if( StkSize==0 && state==enterState ) return; /* nothing to push here */
		StkDup();
		
		/* We have to update the StkLine, but the number is already pushed.
		 * This means that StkSize already has changed.
		 */
		ViewFltToStr( stk[StkSize-1], StkLine );

		ViewUScore = false;
		state = enterState;
		ViewUpdateAll();
}

void DoAdd( Char ch ) {
		CharPtr s;
		Int16 len;

		if( baseMode != 16 && ch=='E' ) {
			ch = EXPCHAR; /* internally the exp sign is not 'e' */
		}
		switch( state ) {
		case opState:
				StkDup();
				/* no break */
		case enterState:
				*StkLine = 0;
				break;
		case editState:
				/* just continue */
		}

		s = StkLine;
		len = StrLen(s);

		if( len>13 ) return; /* Brute force limit to protect end of string. */
		if( len==0 ) {
				/* Give proper start values when entering special characters */
				if( ch=='.' ) s[len++] = '0';
				else if( ch==EXPCHAR ) s[len++] = '1';
		} else {
				if( ch=='.' && StrChr( s, '.' ) ) return;
				if( ch==EXPCHAR && StrChr( s, EXPCHAR ) ) return;
				if( ch=='.' && StrChr( s, EXPCHAR ) ) return;
				if( ch>='A' && ch<='F' && StrChr( s, EXPCHAR ) ) return;
		}

		s[len] = ch;
		s[len+1] = 0;
		ViewUScore = true;
		if( state==opState ) {
				ViewUpdateAll();
		} else {
				ViewLine();
		}
		state=editState;
};

/* Inserts one char at position pos, assumes enough space! */
void SignInsert( CharPtr s ) {
	Int16 i;
	s[StrLen(s)+1] = 0;
	for( i=StrLen(s); i>=0; i-- ) {
		s[i]=s[i-1];
	}
	*s = '-';
}

void SignDelete( CharPtr s ) {
	Int16 i,len = StrLen(s);
	for( i=0;i<len; i++ ) {
		s[i]=s[i+1];
	}
}
  
void DoChangeSign() {
		CharPtr s = StkLine;
		CharPtr epos;

		if( state != editState ) {
				ViewFltToStr( stacktop=-StkTop(), StkLine );
		} else {
				epos = StrChr( s, EXPCHAR );
				if( !epos ) {
						/* no exponent */
						if( s[0]=='-' ) {
								SignDelete( s );
						} else {
								SignInsert( s );
						}
				} else {
						/* chs in exponent */
						epos++;
						if( epos[0]=='-' ) {
								SignDelete( epos );
						} else {
								SignInsert( epos );
						}
				}
		}
		ViewLine();
}

void DoBackspace() {
	Int16 len = StrLen(StkLine);

	if( state == editState && len>1 ) {
		StkLine[--len] = 0;
		ViewLine();
		return;
	} 
	state = (StkSize==0) ? enterState : opState;
	StkPop();
	ViewUScore = false;
	ViewUpdateAll();
}

void DoUpdateLine() {
	ViewFltToStr( StkTop(), StkLine );
}

/*
 * This dispatcher gets the 'id' of a function to be applied.
 * It
 * - Ensures that the last entered number is interpreted as a double
 * - determines the number of neccessary arguments,
 * - checks whether there are enough
 * - gets the cossesponding function pointer
 * - applies the function to the operands
 * - updates the display with the result
 *
 * This method produces quite a lot of overhaed as the parameter passing is
 * being defined for each call. So I am fiddling around here by passing
 * stack contents to local variables.
 */
void DoOperate( Int16 id ) {
	double arg1 = StkTop();
	double arg2;
	
	
	if( id == MtPi || id == BtLastX || ( id >= StXAv && id <= Stn ) ) {
		/***** without operands */
		if( StkSize==0 && state==enterState ) {
		} else {
			StkDup();
		}
		arg1 = DoNonaryOp( id );
		lastx = arg1;
		
	} else if( id == StClear ) {
		if( FrmCustomAlert( AreYouSure,
												"You you want to remove all statistical data?",
												NULL, NULL ) == 1 ) {
			S.N = 0;
			S.X = S.Y = S.X2 = S.Y2 = S.XY = 0.0;
		}
		return;
		
	} else if( id<1000 || id==MtProz ) {
		lastx = arg1;
		/***************** binary operations */
		if( id < 1000 ) { id += 1000; }
		if( StkSize<1 ) {
			if( state!=enterState && (id == StAdd || id == StDel) ) {
				arg2 = 0.0;
			} else {
				FrmAlert( NotEnoughArgs );
				return;
			}
		} else {
			arg2 = stk[StkSize-1];
			if( id != MtProz ) StkPop();
			}
		
		switch( id ) {
			
		case BtAdd:
			arg1 = arg2 + arg1;
			break;
		case BtSub:
			arg1 = arg2 - arg1;
			break;
		case BtMult:
			arg1 = arg2 * arg1;
			break;
		case BtDiv:
			arg1 = arg2 / arg1;
			break;
			
		case BtSwp:
			stacktop = arg1;
			state = opState; /* !! This makes StkDup to use stacktop in favour of
												* !! StkLine. */
			StkDup();
			arg1 = arg2;
			break;
			
		case StAdd:
			S.X += arg1;
			S.Y += arg2;
			S.X2 += arg1*arg1;
			S.Y2 += arg2*arg2;
			S.XY += arg1*arg2;
			S.N++;
			state = (StkSize == 0 ? enterState : opState);
			StkPop();
			goto finishWithoutTosReplace;
			break;
			
		case StDel:
			S.X -= arg1;
			S.Y -= arg2;
			S.X2 -= arg1*arg1;
			S.Y2 -= arg2*arg2;
			S.XY -= arg1*arg2;
			S.N--;
			state = (StkSize == 0 ? enterState : opState);
			StkPop();
			goto finishWithoutTosReplace;
			break;
			
		case BtPow:
			arg1 = pow( arg2, arg1 );
			break;
			
		case MtProz:
			/* note that we didn't pop for MtProz */
			arg1 = arg2 * arg1 / 100.0;
			break;
			
		}
	} else {
		/*********************************************** unary operations  */
		lastx = arg1;
		if( StkSize==0 && state==enterState ) {
			FrmAlert( NotEnoughArgs );
			return;
		}
		arg1 = DoUnaryOp( id, arg1 );
	}
	ViewFltToStr(arg1,StkLine);
	StkSetTop(arg1);
	state = opState;
 finishWithoutTosReplace:
	ViewUScore = false;
	ViewUpdateAll();
}

void DoCopy() {
	ClipboardAddItem( clipboardText, StkLine, StrLen(StkLine) );
}

void DoPaste() {
	MemHandle handle;
	UInt16   len;
	CharPtr  str;
	Int16     i;
	Boolean  space = true;
	Boolean  wasnumber = true;
	Char     c;

	handle = ClipboardGetItem( clipboardText, &len );
	if( !handle ) return;

	str = MemHandleLock( handle );

	for( i=0; i<len; i++ ) {
		c = str[i];
		if( c==' ' || c=='\t' || c=='\n') {
			space = true;
		} else {
			switch( c ) {
			case '0':	case '1':	case '2':	case '3':	case '4':
			case '5':	case '6':	case '7':	case '8':	case '9':
				if( c-'0' >= baseMode ) break;
			case '.':
				if( space && wasnumber ) DoEnter();
				DoAdd( c );
				wasnumber = true;
				break;
			case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
				c = c + 'A' - 'a';
			case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
				if( baseMode == 16 ) {
					if( space && wasnumber ) DoEnter();
					DoAdd( c );
					wasnumber = true;
				} else if( c == 'E' ) {
					if( space && wasnumber ) DoEnter();
					DoAdd( c );
					wasnumber = true;
				}
				break;
			case '+':
				DoOperate( BtAdd-1000 );
				wasnumber = false;
				break;
			case '-':
				DoOperate( BtSub-1000 );
				wasnumber = false;
				break;
			case '*':
				DoOperate( BtMult-1000 );
				wasnumber = false;
				break;
			case '/':
				DoOperate( BtDiv-1000 );
				wasnumber = false;
				break;
			case '^':
				DoOperate( BtPow-1000 );
				wasnumber = false;
				break;
			}
			space = false;
		}
	}

	MemHandleUnlock( handle );
}
