#ifndef _model_h_
#define _model_h_

/*
 * File  : model.h
 * Author: Holger Klawitter
 * Date  : 15.04.2000
 *
 * The model controlling the upn calculator.
 * This is somehow the real main program :)
 */

#include <PalmOS.h>
#include "util.h"
#include "ops.h"

/**********************************************************************
 * Public variables and DS
 **********************************************************************/

typedef enum { radMode, degMode } AngleMode;

extern Statistics S;
/* Statistical registers. */

extern AngleMode angleMode;

/**********************************************************************
 * Public functions
 **********************************************************************/

void DoEnter();
/* Handling of the enter event */

void DoAdd( Char ch );
/* Adding a character which may be part of the number (except '-').
 * Expected characters are 0, ..., 9, '.' and 'e'
 * Illegal input (like two dots) will be ignored.
 */
  
void DoChangeSign();
/* Handles pressing the "+/-" key. */

void DoBackspace();
/* Handles the DEL key. */

void DoOperate( Int16 id );
/* Handles mathematical functions of the calculator.
 * Checks for legal number of arguments and so on.
 * Some functiones are taken from the floatlib which does not handle
 * illegal values properly, so you may get garbage.
 *
 * The ids are usually the ObjectIds from the corresponding button.
 * Ids for binary operations are smaller than 1000!
 */

void DoUpdateLine();
/* Recalculates the contents of StkLine, if in proper mode */

void DoCopy();
/* Copies the stak top to the clipboard */

void DoPaste();
/* Reads the conents of the clipboard as a command */

#endif /* _model_h_ */
