/*
 * File: ops.c
 * Author: Holger Klawitter
 * Date: 14.7.1997
 *
 * Wrapper for all operators
 */

#include "ops.h"
#include "kalk.h"
#include "model.h"

#include <System/SysAll.h>
#include <UI/UIAll.h>

static double fak( double d );

/*********************************************************************
 * Functions
 **********************************************************************/

/* This function handles all no nary functions, meaning function which
 * take no stack elements from stack.
 */
double DoNonaryOp( Int16 id ) {
  double res=0.0;
  switch( id ) {
  case MtPi:
    res = 3.14159265358979;
    break;
  case StXAv:
    res = S.X / S.N;
    break;
  case StYAv:
    res = S.Y / S.N;
    break;
  case StXVa:
    res = sqrt(( S.X2 - S.X*S.X / S.N ) / ( S.N-1 ));
    break;
  case StYVa:
    res = sqrt(( S.Y2 - S.Y*S.Y / S.N ) / ( S.N-1 ));
    break;
  case StR:
    res = ( S.XY - S.X * S.Y / S.N ) / sqrt
      ( ( S.X2 - S.X*S.X / S.N ) * ( S.Y2 - S.Y*S.Y / S.N ) );
    break;
  case StBYX:
    res = ( S.XY - S.X * S.Y / S.N ) / ( S.X2 - S.X*S.X / S.N );
    break;
  case StBXY:
    res = ( S.XY - S.X * S.Y / S.N ) / ( S.Y2 - S.Y*S.Y / S.N );
    break;
  case Stn:
    res = S.N;
    break;
	case BtLastX:
		res = lastx;
		break;
  default:
  }
  
  return res;
}

/* Wrapper for all unary functions. */
double DoUnaryOp( Int16 id, double arg1 ) {
  switch(id) {
    
  case BtRez:
    arg1 = 1.0/arg1;         break;
  case BtSqr:
    arg1 = arg1*arg1;      break;
  case BtSqrt:
    arg1 = sqrt(arg1);     break;
    
  case MtLog:
    arg1 = log10(arg1);    break;
  case Mt10pw:
    arg1 = exp10(arg1);    break;
  case MtLn:
    arg1 = log(arg1);      break;
  case MtExp:
    arg1 = exp(arg1);      break;
    
  case MtSin:
    arg1 = sin(arg1*angleConvert);      break;
  case MtCos:
    arg1 = cos(arg1*angleConvert);      break;
  case MtTan:
    arg1 = tan(arg1*angleConvert);      break;
    
  case MtASin:
    arg1 = asin(arg1)/angleConvert;     break;
  case MtACos:
    arg1 = acos(arg1)/angleConvert;     break;
  case MtATan:
    arg1 = atan(arg1)/angleConvert;     break;
    
  case MtSinH:
    arg1 = sinh(arg1);      break;
  case MtCosH:
    arg1 = cosh(arg1);      break;
  case MtTanH:
    arg1 = tanh(arg1);      break;
    
  case MtASinH:
    arg1 = asinh(arg1);     break;
  case MtACosH:
    arg1 = acosh(arg1);     break;
  case MtATanH:
    arg1 = atanh(arg1);     break;

	case MtFak:
			arg1 = fak(arg1);     break;


    /* Statistical funcitons */
  case StToY:
    arg1 = S.Y / S.N + ( arg1 - S.X / S.N ) *
      ( ( S.XY - S.X * S.Y / S.N ) /
	( S.X2 - S.X * S.X / S.N  )
	);
    break;
  case StToX:
    arg1 = S.X / S.N + ( arg1 - S.Y / S.N ) *
      ( ( S.XY - S.X * S.Y / S.N ) /
	( S.Y2 - S.Y * S.Y / S.N  )
	);
    break;
  }
  return arg1;
}

double fak( double d ) {
		double res = 1.0;
		double i;
		if( d < 0 ) { return res; }
		for( i = 1.0; i<=d; i += 1.0 ) {
				res *= i;
		}
		return res;
}
