/*
 * File: stk.c
 * Author: Holger Klawitter
 * Date: 01.05.1997
 *
 * Handling of the upn-stack. This stack consists of a stack of floating
 * point numbers, where the TOS can be edited.
 *
 * Changes:
 * 05.10.1999 support for Hexagesimal digits.
 */

#include "stk.h"
#include "view.h"
#include <System/SysAll.h>
#include <mathf.h>

/**********************************************************************
 * Public variables
 **********************************************************************/

Char StkLine[LINESIZE] = ""; /* the tos */

StackState state = enterState;

/**********************************************************************
 * Local constants & variables
 **********************************************************************/

double stk[STACKSIZE];
/* the stack without the tos */

double stacktop = 0.0;
/* backup value for tos */

Int16 StkSize = 0;
/* Number of elements on stack, tos included */

/**********************************************************************
 * Public functions
 **********************************************************************/

void StkDup() {
  Int16 i;

  if( StkSize == STACKSIZE ) {
    /* Rolls the stack upward and lose the last element. */
    for( i=0; i<StkSize-1; i++ ) {
      stk[i] = stk[i+1];
    }
  } else {
    StkSize++;
  }
  stacktop = stk[StkSize-1] = (state==editState) ? StkFlt() : stacktop;
}

void StkPop() {
  if( StkSize > 0 ) {
    StkSize--;
    ViewFltToStr( stacktop=stk[StkSize], StkLine );
  } else {
    stacktop = 0.0;
    *StkLine = 0;
  }
}

void StkGetStr(Int16 i,CharPtr s) {
  if( i==0 ) {
    StrCopy( s, StkLine );
  } else if( i>StkSize || i<0 ) {
    *s = 0;
  } else {
    ViewFltToStr( stk[StkSize-i], s );
  }
}

Int16 StkElements() {
  return StkSize+1;
}

double StkTop() {
  if( state==editState ) return StkFlt();
  return stacktop;
}

void StkSetTop( double f ) {
  stacktop = f;
}

double StkFlt() {
  CharPtr s = StkLine;
  Int16
    pow = 0,     /* the exponent */
    ch;          /* current char (to help casting mechinism) */

  double
    res=0.0,       /* this will become the number */
    mult=1.0,      /* factor for the next fractional digit */
    sign=1.0,      /* the sign of the mantissa */
		base=baseMode, /* base *only for mantissa*, not for exponent! */
    expsign=1.0;   /* the sign of the exponent */

  if( *s == '-' ) {
    sign=-1.0;
    s++;
  }
  while( (*s>='0' && *s<='9') || (*s>='A' && *s<='F') ) {
    if( *s<='9' ) ch = *s++ - '0';
		else ch = *s++ - 'A' + 10;
    res = res * base + ch;
  }
  if( *s == '.' ) {
    s++;
    while( (*s>='0' && *s<='9') || (*s>='A' && *s<='F') ) {
			if( *s<='9' ) ch = *s++ - '0';
			else ch = *s++ - 'A' + 10;
      res = res * base + ch;
      mult = mult * base;
    }
    res /= mult;
  }
  if( *s==EXPCHAR ) {
    s++;
    if( *s=='-' ) {
      expsign = -1.0;
      s++;
    }
    while( *s>='0' && *s<='9' ) {
      pow = pow * 10 + *s++ -'0';
    }
    res = res * exp10( expsign*pow );
  }
  return sign * res;
}
