#ifndef _stk_h_
/*
 * File: stk.h
 * Author: Holger Klawitter
 * Date: 01.05.1997
 *
 * Handling of the upn-stack. This stack consists of a stack of floating
 * point numbers, where the TOS can be edited.
 *
 */

#include <PalmOS.h>
#include "util.h"

/**********************************************************************
 * Global Constants and Variables
 **********************************************************************/

#define LINESIZE 30
/* Maximal number of chars in a number, terminating 0 included. */

#define STACKSIZE 10
/* Maximal number of numbers on the stack */

#define EXPCHAR '/'
/* Char internally used for exponential sign. Different than 'e'! */

extern Char StkLine[];
/* The TOS. 
 * This variable is intended to make the TOS editable for the model and
 * to circumvent unneccessary copies.
 */

extern double stk[];
/* The stack */

extern Int16 StkSize;
/* The size of the stack = Number of Elements on stack, excluding TOS. */

typedef enum { enterState, opState, editState } StackState;

extern StackState state;
/* The current state of the stack. */
/* If state==enterState && StkSize==0 => stack is really empty */

extern double stacktop;
/* Internal backup value for TOS. */

/**********************************************************************
 * Public functions
 **********************************************************************/

void StkDup();
/* Puts the TOS on stack. */

void StkPop();
/* Removes the TOS. */

double StkTop();
/* Returns float value of TOS. Can return the full value even if
 * the dislay value is not accurately displayed.
 */

void StkGetStr(Int16 i,CharPtr s);
/* Gets the string representation of the i'th element from stack.
 * Returns empty string for illegal indices.
 * The caller has to pass LINESIZE Chars.
 */

double StkFlt();
/* Transforms StkLine to float. */

void StkSetTop( double f );
/* Sets the stack tos to 'f'. */

#endif /* _stk_h_ */
