/*
 * File: view.c
 * Author: Holger Klawitter
 * Date: 25.04.1997
 *
 * Handling of the visual part of the stack on screen.
 *
 * Changes:
 * 05.10.1999 use large font for nonstd bases
 */

#include "view.h"
#include "stk.h"
#include "format.h"
#include "kalk.h"
#include <System/SysAll.h>
#include <UI/UIAll.h>

#ifndef BMPU
#define BMPU 1113
#define BMPError 1114
#endif

/**********************************************************************
 * Local constants and variables
 **********************************************************************/

Int16 tov = 0;
/* Index of the lowermost visible line on stack. */

Boolean ViewUScore = false;

ViewMode displayMode = fixedMode;
Int16 places = 4;

#define MAXFREE 1e9
#define MINFREE 1e-8
/* Beyond these thresholds free formatted numbers will be displayed
 * in exponential form.
 */

#define VLEFT 10
#define VRIGHT 150
#define VBOT 22
#define VSIZE 17 /* height of line */
/* VTOP is 0 */

#define VISUALCHARS 20
/* maximal number of visible chars */

Int8 baseMode;

/**********************************************************************
 * Private functions
 **********************************************************************/

void ViewBmp( UInt16 id, UInt8 x, UInt8 y );
/* Draws the bitmap with id at (x,y). */

void ViewUpdateLine( Int16 idx, UInt8 row );
/* Draws the number 'idx' from stack at given display row.
 * If idx=0 and underscore set, add an underscore at the end.
 */

/**********************************************************************
 * Public functions
 **********************************************************************/

void ViewUp() {
  if( StkSize - tov <= 1 ) return;
  tov++;
  ViewUpdateAll();
}

void ViewDown() {
  if( tov == 0 ) return;
  tov--;
  ViewUpdateAll();
}

void ViewLine() {
  if( tov==0 ) {
    ViewUpdateLine( 0, VBOT );
  } else {
    tov = 0;
    ViewUpdateAll();
  }
}

void ViewFltToStr( double f, CharPtr s ) {
	if( baseMode == 10 ) {
		switch( displayMode ) {
		case sciMode:
			hkftoa( s, f, places+FMT_EXP );
			break;
		case engMode:
			hkftoa( s, f, places+FMT_ENG );
			break;
		case fixedMode:
			hkftoa( s, f, places );
			break;
		default:
			hkftoa( s, f, 9 );
		}
	} else {
		hkftoa( s, f, 0 );
	}
}

void ViewBmp( UInt16 id, UInt8 x, UInt8 y ) {
  MemHandle handle;
  VoidPtr memory;

  handle = DmGetResource( bitmapRsc, id );
  ErrFatalDisplayIf(!handle, "View: Unable to find bitmap handle");
     
  memory = MemHandleLock(handle);
  ErrFatalDisplayIf(!memory, "View: Unable to allocate bitmap memory");
  
  WinDrawBitmap(memory, x, y);
  
  MemPtrUnlock(memory);
  DmReleaseResource(handle);
}

void ViewUpdateAll() {
  UInt8 row = VBOT;
  Int16 idx;
  for( idx=0; idx<=1; idx++, row-=VSIZE+2 ) {
    ViewUpdateLine( idx+tov, row );
  }
}

void ViewUpdateLine( Int16 idx, UInt8 row ) {
  RectangleType r = { {VLEFT,row} , {VRIGHT,VSIZE} };
  Char buf[LINESIZE];
  Int16 col = VLEFT;
  FontID font;
	Int16 charsToShow;

  StkGetStr(idx,buf);
  WinEraseRectangle( &r, 0 );
  if( *buf=='X' ) {
    ViewBmp( BMPError, col, row );
  } else {
    font = FntSetFont( baseMode==10 ? ledFont : largeFont );
		charsToShow = StrLen(buf);
		if( charsToShow > VISUALCHARS ) charsToShow = VISUALCHARS;
    WinDrawChars( buf, charsToShow, col, row + (baseMode == 10 ? 0 : 4) );
    col += FntLineWidth( buf, StrLen(buf) );
    if( idx==0 && ViewUScore && col <= VRIGHT )
      ViewBmp( BMPU, col, row );
    FntSetFont( font );
  }
}
