/* Main code for PGP Key DB Manager */

#include <Pilot.h>
#include <Hardware/Hardware.h>

#include <pgplib.h>
#include <md5.h>

#include "callback.h"

#include "keymanagerRsc.h"


typedef struct {
} KeyManPrefs_t;

static DmOpenRef pubring = NULL;
static DmOpenRef secring = NULL;

enum { FRM_MAIN, FRM_GENERATE, FRM_VIEW, FRM_LAST };
static int list_keysizes[] = { 128, 256, 384, 512, 768, 1024, 2048, 4096 };
/* static int list_keysize_eta[] = { 10, 46, 77, 140, 332, 1309, 18674, -1 }; */
static char *list_keysizes_eta_approx[] = {
    "10 seconds", "45 seconds", "1.5 minutes", "2.5 minutes", "5.5 minutes",
    "25 minutes", "5 hours", "1-3 days"
};
#define KEYSIZEINDEX_DEFAULT	2 /* I.e. 384 */
#define KEYSIZEINDEX_MAX	7 /* I.e. 4096 */
static int form_ids[7+1] = {MainForm, GenerateForm, ViewForm, 0};
static int current_form = 0;

static char smbuf[101];
static char ubuf[41];

/* Makes <n> digits out of <v>, and puts the result in <buf>. Will
 * zeroterminate <buf>. If <buf> is NULL, it will return a pointer to it's own
 * little buffer, which data will be valid until the next call. */
static char *
l2hex(register char *buf, register long v, register int n)
{
    static char mybuf[10];
    register char c, *s;

    if (8 < n)
	return NULL;
    s = (NULL == buf) ? mybuf : buf;
    while (n-- > 0)
	*s++ = (c = (v >> (n << 2)) & 0xF) <= 9 ? c + '0' : c - 10 + 'A';
    *s = '\0';
    return ((NULL == buf) ? mybuf : buf);
}

static char *
PGPKEYID2str(PGPKEYID *keyid)
{
    static char buf[18];
    register char *s;

    s = buf;
    MemSet(buf, sizeof(buf), 0);
#ifndef USE_LONGLONG
    l2hex(s, *((long *)&keyid->data[0]), 8);
    l2hex(s+8, *((long *)&keyid->data[4]), 8);
#else
	#error "USE_LONGLONG handling missing!"
#endif
    return buf;
}

static void set_object_text(Int id, CharPtr new_text)
{
    FormPtr	frm;
    VoidPtr	fld;
    VoidHand	hnd;
    CharPtr	p;
    Word	i;
    FormObjectKind k;
    int		t;

    frm = FrmGetActiveForm();
    i = FrmGetObjectIndex(frm, id);
    k = FrmGetObjectType(frm, i);
    if (frmLabelObj == k) {
	FrmCopyLabel(frm, id, new_text);
    } else {
	fld = FrmGetObjectPtr(frm, i);
	FldSetSelection(fld, 0, 0);
	t = StrLen(new_text);
	hnd = MemHandleNew(t+1);
	ErrFatalDisplayIf(hnd==NULL, "MemHandleNew");
	p = MemHandleLock(hnd);
	ErrFatalDisplayIf(p==NULL, "MemHandleLock");
	StrCopy(p, new_text);
	p[t] = '\0';
	MemHandleUnlock(hnd);
	switch (k) {
	    case frmFieldObj:
		FldSetTextHandle(fld, (Handle)hnd);
		FldSetUsable(fld, true);
		FldDrawField(fld);
		break;
	    default:
		p = NULL;	/* i.e. illegal object type */
		break;
	}
	if (NULL == p) {
	    MemHandleFree(hnd);
	}
    }
}

static void set_object_visibility(Word id, int visible)
{
    FormPtr	frm;
    ControlPtr	ctr;

    frm = FrmGetActiveForm();
    ctr = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, id));
    if (visible)
	CtlShowControl(ctr);
    else 
	CtlHideControl(ctr);
}

/* Pops up a text input request, with title (up to 19 chars) in <title>
 * returning either NULL or a pointer to a string, which must be freed at some
 * later time. */
static char *
popup_textinput(char *title)
{
    FormPtr frm;
    CharPtr s, new = NULL;
    int i;

    frm = FrmInitForm(TextInputForm);
    s = title;
    if (NULL == s)
	s = "Enter text:";
    if (StrLen(s) > 19)
	s[19] = '\0';
    FrmCopyTitle(frm, s);
    FrmSetFocus(frm, FrmGetObjectIndex(frm, FldInput));
    
    i = FrmDoDialog(frm);

    if (BtnOk == i) {
	/* We don't actually do anything unless it's the ok-button */
	s = FldGetTextPtr(FrmGetObjectPtr(frm,
		    FrmGetObjectIndex(frm, FldInput)));
	if (NULL != s) {
	    i = StrLen(s);
	    new = MemPtrNew(i + 1);
	    if (NULL != new)
		StrCopy(new, s);
	    MemSet(s, i, 0); /* We _must_ destroy the data in the field! */
	}
    }
    FrmEraseForm(frm);
    FrmDeleteForm(frm);

    return new;
}

static int rand_numbits;
static unsigned char *rand_buffer;
static int rand_buflen;

/* Pop up the random source form, and get <bits> bits of hopefully random
 * source data. We use the digitizer, and it's builtin inaccuracy (it
 * wobbles). When this returns 0, the globals <rand_buffer> will contain
 * <rand_buflen> bits. Remember to free and reset rand_buflen to NULL. You can
 * call this function with a parameter of 0 to achieve this (ignore the error
 * code in that case). */
int popup_gen_getrand(int bits)
{
    if (NULL != rand_buffer) {
	MemSet(rand_buffer, (int)MemPtrSize(rand_buffer), 0);
	MemPtrFree(rand_buffer);
    }
    rand_buffer = NULL;
    rand_buflen = 0;
    if (bits <= 0 || bits > 4096)
	return -1;

    rand_numbits = ((bits >> 3) + 1) << 3;
    rand_buflen = rand_numbits >> 3;
    rand_buffer = MemPtrNew(rand_buflen);
    if (NULL == rand_buflen) {
	rand_buffer = NULL;
	rand_buflen = 0;
	return -2;
    }

    FrmPopupForm(RandGenForm);
    return 0;
}

/* Our random source handler */
static Boolean RandGenHandleEvent (EventPtr e)
{
    Boolean handled = false;
    FormPtr frm;
    int px, py;
    static int lastx, lasty;
    static ULong n = 0;
    static SWord extentX, extentY;
    static RectangleType winbounds;
    static RectangleType todorect = { {0, 50 }, {159, 20} };
    static int todo;
    static unsigned char *buf;
    char *s;
    int i, x;

    CALLBACK_PROLOGUE;

    px = (int)e->screenX;
    py = (int)e->screenY;
    switch (e->eType) {
	case frmOpenEvent:
	    lastx = -1;
	    lasty = -1;
	    todo = rand_buflen;
	    buf = rand_buffer;
	    frm = FrmGetActiveForm();
	    FrmDrawForm(frm);
	    WinGetDisplayExtent(&extentX, &extentY);
	    WinGetWindowBounds(&winbounds);
	    FntSetFont(1);	/* Bold font */
	    s = "Move the pen around at random";
	    i = StrLen(s);
	    WinDrawChars(s, i, 1, 20);
	    s = "until you're told to stop.";
	    i = StrLen(s);
	    WinDrawChars(s, i, 1, 31);
	    WinDrawRectangle(&todorect, 0);
	    handled = true;
	    break;

	case menuEvent:
	    MenuEraseStatus(NULL);
	    handled = true;
	    break;

	case penDownEvent:
	    n = 0;
	    WinEraseRectangle(&winbounds, 0);
	    WinInvertLine(0, py, extentX, py);
	    WinInvertLine(px, 0, px, extentY);
	    lastx = px;
	    lasty = py;

	    if (NULL == buf) {
		FrmCustomAlert(WarnAlert,
			"Buffer not properly initialized!", "", NULL);
		FrmReturnToForm(0);
	    }
	    handled = true;
	    break;

	case penMoveEvent:
	    /* Erase previous lines */
	    WinInvertLine(0, lasty, extentX, lasty);
	    WinInvertLine(lastx, 0, lastx, extentY);

	    /* Update counter here */
	    x = (((rand_buflen - todo) * 155) / 155) + 2;
	    WinDrawLine(x, todorect.topLeft.y + 1,
		    x, todorect.extent.y - 1);

	    /* Draw new lines */
	    WinInvertLine(0, py, extentX, py);
	    WinInvertLine(px, 0, px, extentY);

	    /* Here we record the changes, or, if you like, the randomness of
	     * the position. We xor the x and y coordinate together. There are
	     * probably other methods of doing this that are better. */
	    *buf++ = px ^ py;
	    if (0 == todo--) {
		/* We're done */
		WinEraseRectangle(&winbounds, 0);
		s = "Thanks, that's enough!";
		i = StrLen(s);
		WinDrawChars(s, i, 1, 31);
		SysTaskDelay(2 * SysTicksPerSecond());
		
		FrmReturnToForm(0);
	    }

	    lastx = px;
	    lasty = py;

	    handled = true;
	    break;

	default:
	    break;
    }

    CALLBACK_EPILOGUE;

    return handled;
}


#define GenPref_Powersave	(1<<0)
#define GenPref_Silent		(1<<1)
#define GenPref_Bench		(1<<2)

/* Pops up the prefs form for the generate form. Expects the preference bitmask
 * as input, and returns the possibly modified preference bitmask. */
long popup_gen_prefs(long prefs)
{
    FormPtr	frm;
    long	l;
    int		i;

    l = prefs;
    frm = FrmInitForm(GenPrefsForm);

    /* Set control settings */
    CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, Chk_Silent)),
	    (prefs & GenPref_Silent) ? 1 : 0);
    CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, Chk_Bench)),
	    (prefs & GenPref_Bench) ? 1 : 0);
    
    i = FrmDoDialog(frm);

    if (BtnOk == i) {
	/* We don't actually do anything unless it's the ok-button */
	l = 0;
	/* Read control settings */
	if (CtlGetValue(FrmGetObjectPtr(frm,
			FrmGetObjectIndex(frm, Chk_Silent))))
	    l |= GenPref_Silent;
	if (CtlGetValue(FrmGetObjectPtr(frm,
			FrmGetObjectIndex(frm, Chk_Bench))))
	    l |= GenPref_Bench;
    }
    FrmEraseForm(frm);
    FrmDeleteForm(frm);

    return l;
}


static Boolean GenerateFormHandleEvent (EventPtr e)
{
    Boolean handled = false;
    FormPtr frm;
    FieldPtr fld;
    ListPtr lst;
    ControlPtr ctl;
    PGPuserid_t userid;
    PGPtrust_t	trust;
    PGPKEYID	keyid;
    register CharPtr s, s2;
    register int i;
    register ULong tstart, tstop;
    PGPseckey_t *sec;		/* secret key (empty struct) */
    PGPpubkey_t *pub;		/* public key (ditto) */
    static u_char digest[16];	/* passphrase */
    static int digest_set;
    static int eta;
    static unsigned short valid;
    static long prefs;
    int keysize;
    int flags;
    unsigned char rndptr;
    int	rndlen;

    CALLBACK_PROLOGUE;

    frm = FrmGetActiveForm();
    lst = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, KeySizeList));

    switch (e->eType) {
	case frmOpenEvent:
	    digest_set = 0;
	    eta = 2;
	    valid = 0;	/* Keys are valid forever per default */
	    prefs = 0; /* i.e. neither powersave nor silent nor bench */
	    LstSetSelection(lst, KEYSIZEINDEX_DEFAULT);
	    if (NULL != (s = LstGetSelectionText(lst, KEYSIZEINDEX_DEFAULT)))
		CtlSetLabel(FrmGetObjectPtr(frm,
			    FrmGetObjectIndex(frm, KeySizeTrigger)), s);
	    FrmDrawForm(frm);
	    handled = true;
	    break;

	case menuEvent:
	    MenuEraseStatus(NULL);
	    switch(e->data.menu.itemID) {
		case itemGenPrefs:
		    prefs = popup_gen_prefs(prefs);
		    break;
	    }

	    handled = true;
	    break;

	case popSelectEvent:
	    if(KeySizeList == e->data.popSelect.listID) {
		i = e->data.popSelect.selection;
		if (KEYSIZEINDEX_MAX >= i) {
		    eta = i;
		    /* Update LblETA according to selected keysize */
		    StrPrintF(smbuf, "Estimated time: %s",
			    list_keysizes_eta_approx[eta]);
		    smbuf[32] = '\0';	/* Must be max 32 chars! */
		    FrmHideObject(frm, FrmGetObjectIndex(frm, LblETA));
		    FrmCopyLabel(frm, LblETA, smbuf);
		    FrmShowObject(frm, FrmGetObjectIndex(frm, LblETA));
		}
	    }
	    break;

	case ctlSelectEvent:
	    switch(e->data.ctlSelect.controlID) {
		case sltPass:
		    /* Get and set passphrase (i.e. digest of passphrase) */
		    /* XXX: use alternative method - pass-symbol? */
		    s = s2 = NULL;
		    s = popup_textinput("Enter passphrase:");
		    if (NULL != s) {
			/* Verify passphrase */
			s2 = popup_textinput("...and once more:");
			if (NULL != s2) {
			    /* Make sure they're identical */
			    i = StrCompare(s, s2);
			    MemSet(s2, StrLen(s2), 0);	/* Destroy passphrase */
			    MemPtrFree(s2);
			    s2 = NULL;
			    if (0 == i) {
				/* The passphrases were identical, hurrah! */
				ctl = FrmGetObjectPtr(frm,
					FrmGetObjectIndex(frm, sltPass));
				CtlSetLabel(ctl, "- Assigned -");
				CtlDrawControl(ctl);
				MD5(s, i, digest);
				digest_set = 1;
			    } else
				FrmCustomAlert(WarnAlert,
					"Passphrases not identical!",
					"", NULL);
			}
			if (NULL != s) {
			    MemSet(s, StrLen(s), 0);	/* Destroy passphrase */
			    MemPtrFree(s);
			}
		    }
		    break;
		case sltValid:
		    /* Get and set number of days the key is valid */
		    s = s2 = NULL;
		    s = popup_textinput("Number of days key is valid:");
		    if (NULL != s) {
			valid = (unsigned short)pgplib_strtol(s, NULL, 10);
			ctl = FrmGetObjectPtr(frm,
				FrmGetObjectIndex(frm, sltPass));
			CtlSetLabel(ctl, s);
			CtlDrawControl(ctl);
		    }
		    break;
		case BtnOk:
		    /* See if all required parameters are set */
		    i = FrmGetObjectIndex(frm, FldUID);
		    fld = FrmGetObjectPtr(frm, i);
		    i = FldGetTextLength(fld);
		    if (i < 1) {
			/* User-id _must_ be set! */
			FrmCustomAlert(WarnAlert, "Missing parameter:",
				"UserID must be filled in!", NULL);
			break;
		    }
		    if (! digest_set) {
			/* A passphrase _must_ be set! */
			FrmCustomAlert(WarnAlert, "Missing parameter:",
				"You must set a passphrase!", NULL);
			break;
		    }

		    keysize = LstGetSelection(lst);
		    if (KEYSIZEINDEX_MAX < keysize || 0 > keysize)
			keysize = KEYSIZEINDEX_DEFAULT;
		    keysize = list_keysizes[keysize];

		    /* See if user really want to create this key, if the size
		     * of the key is greater than or equal to 1024 bits. */
		    if (1024 <= keysize) {
			StrPrintF(smbuf, "Generating a %d-bit key finishes " \
				"approx. within %s!", keysize,
				list_keysizes_eta_approx[eta]);
			i = FrmCustomAlert(ConfAlert, smbuf,
				"You must reset to abort this!.\n\nContinue?",
				NULL);
			if (i != 0)
			    break;
		    }

		    /* Determine what flags we should use */
		    flags = GENKEY_DEFAULT;
		    if (CtlGetValue(FrmGetObjectPtr(frm,
				    FrmGetObjectIndex(frm, Chk_Powersave))))
			flags |= (GENKEY_SCREEN_OFF | GENKEY_BEEP_UPDATE);
		    if (prefs & GenPref_Silent)
			flags &= ~(GENKEY_BEEP_WHEN_DONE | GENKEY_BEEP_UPDATE);

		    sec = MemPtrNew(sizeof(PGPseckey_t));
		    if (NULL == sec) {
			FrmCustomAlert(WarnAlert, "Not enough memory for",
				"secret key!", NULL);
			break;
		    }
		    pub = MemPtrNew(sizeof(PGPpubkey_t));
		    if (NULL == pub) {
			FrmCustomAlert(WarnAlert, "Not enough memory for",
				"public key!", NULL);
			MemPtrFree(sec);
			break;
		    }
		    MemSet(sec, sizeof(PGPseckey_t), 0);
		    MemSet(pub, sizeof(PGPpubkey_t), 0);
		    
		    /* TODO: Make proper random source!  Currently the random
		     * source we're using here (i.e. the event structure, with
		     * the most random stuff being the coordinates, time,
		     * various data fields and unspecified values) sucks. */
		    /* Get random data */
		    if (0 == popup_gen_getrand(keysize)) {
			/* Oops, something went haywire. Use bad source. */
			rndptr = 
		    }
		    rndptr = rand_buffer;
		    rndlen = rand_buflen;

		    if (flags & GENKEY_SCREEN_OFF) {
			HwrBacklight(1, 0);	/* Turn off backlight */
			HwrLCDSleep(0, 0);	/* Turn off lcd */
		    }

		    tstart = TimGetTicks();
		    /* XXX: There's yet no way to abort a key generation! */
		    i = generate_keypair(sec, pub, keysize, 17, 0, digest,
			    (void *)e, sizeof(*e), flags);
		    tstop = TimGetTicks();

		    if (flags & GENKEY_SCREEN_OFF) {
			HwrLCDWake();	/* Turn lcd back on */
		    }

		    popup_gen_getrand(0);	/* Free random buffer */

		    /* See if we should show benchmarking */
		    if (prefs & GenPref_Bench) {
			StrPrintF(smbuf, "%lu ticks, (%lu seconds) "
				"for %d-bit keypair", tstop-tstart,
				(tstop-tstart)/(ULong)SysTicksPerSecond(),
				keysize);
			FrmCustomAlert(WarnAlert, "Benchmark results:",
				smbuf, NULL);
		    }

		    if (0 == i) {
			/* generate_keypair() succeeded, add sec and pub to the
			 * key database! */
			sec->valid = pub->valid = valid;
			/* Get PGPKeyid */
			i = PGPKEYID_FromBN(&keyid, sec->N);
			if (0 != i)
			    FrmCustomAlert(WarnAlert, "Failed retrieving",
				    "secret key's key id!", NULL);

			/* Store secret key */
			trust.byte = 0x87;	/* max trust */
			i = add_seckey(secring, sec, &trust);
			if (0 != i)
			    FrmCustomAlert(WarnAlert, "Failed storing",
				    "secret key!", NULL);

			trust.byte = 0x03;	/* max trust */
			userid.name = FldGetTextPtr(fld);	/* Uid */
			i = add_uid2key(secring, &keyid, &userid, &trust);
			if (0 != i)
			    FrmCustomAlert(WarnAlert, "Failed storing",
				    "secret key's userID!", NULL);
			/* We don't make any signatures of the secret key. What
			 * would be the point of that, anyway? */

			/* Store public key */
			
			trust.byte = 0x87;	/* max trust */
			i = add_pubkey(pubring, pub, &trust);
			if (0 != i)
			    FrmCustomAlert(WarnAlert, "Failed storing",
				    "public key!", NULL);

			trust.byte = 0x03;	/* max trust */
			userid.name = FldGetTextPtr(fld);	/* Uid */
			i = add_uid2key(pubring, &keyid, &userid, &trust);
			if (0 != i)
			    FrmCustomAlert(WarnAlert, "Failed storing",
				    "public key's userID!", NULL);

			/*
			 * TODO: Make signature of public key
			 */
		    }
		    else
			FrmCustomAlert(WarnAlert, "generate_keypair failed!",
				"", NULL);
		    /* We're done with the keys, so free them (sensitive data
		     * will be destroyed first) */
		    if (sec) {
			free_seckey(sec);
			MemPtrFree(sec);
		    }
		    free_pubkey(pub);
		    MemPtrFree(pub);
		    /* FALLTHROUGH */
		case BtnCancel:
		    MemSet(digest, sizeof(digest), 0);	/* Always destroy digest */

		    FrmUpdateForm(MainForm, 0);
		    FrmReturnToForm(MainForm);
		    break;
	    }
	    break;
	
	default:
	    break;
    }

    CALLBACK_EPILOGUE;

    return handled;
}


static void
ViewDrawCell(VoidPtr tbl, Word row, Word col, RectanglePtr bounds)
{
    char buf[10], *s, **uids;
    int	i, uid_nums;
    int currFont;
    ULong data;
    PGPKEYID keyid;

    CALLBACK_PROLOGUE;
    data = TblGetRowData(tbl, row);

    s = NULL;
    uids = NULL;
    switch(col) {
	case 0:			/* Cell in KeyID column */
	    StrPrintF(buf,"%lx", (long)data);
	    s = buf;
	    break;
	case 1:			/* Cell in data-column */
	    i = lookup_keyid(pubring, &keyid, NULL, &data, 1, NULL);
	    i = list_uids(pubring, &keyid, &uids, &uid_nums);
	    /* TODO: handle showing multiple userIDs */
	    s = uids[0];
	    break;
	default:
	    break;
    }
    if (NULL != s) {
	currFont = FntSetFont(0);
	WinEraseRectangle(bounds, 0);
	/* TODO: handle text larger than boundary */
	WinDrawChars(s, StrLen(s),
		bounds->topLeft.x + 1, bounds->topLeft.y);
	FntSetFont(currFont);
	if (NULL != uids)
	    free_uid_list(uids);
    }
    CALLBACK_EPILOGUE;
}

enum { VIEW_UPDATE_TABLE, VIEW_GENERATE_LIST };

static Boolean
ViewFormHandleEvent (EventPtr e)
{
    Boolean handled = false;
    FormPtr frm;
    TablePtr tbl;
    ScrollBarPtr bar;
    Word numRows, numCols, w;
    DmOpenRef dbR;
    static Word row, col;
    static ULong *shortid_list;
    static int num_keys;
    static WinHandle popwin, savewin, oldwin;
    static RectangleType winrec = { {1,17}, {157, 70} };
    static RectangleType saverec;
    Short bar_value, bar_min, bar_max, bar_pagesize;
    PGPKEYID keyid;
    PGPpubkey_t *pub;
    PGPseckey_t *sec;
    DateTimeType datetime;
    DateType date;
    ULong data, ul;
    short keysize;
    char *s;
    int i,j;

    CALLBACK_PROLOGUE;

    frm = FrmGetActiveForm();
    tbl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, KeyTable));
    bar = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, KeyScroll));

    switch (e->eType) {
	case frmOpenEvent:
	    FrmDrawForm(frm);
	    FrmUpdateForm(ViewForm, VIEW_GENERATE_LIST);
	    FrmUpdateForm(ViewForm, VIEW_UPDATE_TABLE);
	    TblHasScrollBar(tbl, 1);
	    set_object_visibility(BtnOk, 0);
	    CtlSetValue(FrmGetObjectPtr(frm,
			FrmGetObjectIndex(frm, btnPublic)), 1);
	    shortid_list = NULL;
	    handled = true;
	    break;
	    
	case frmUpdateEvent:
	    switch(e->data.frmUpdate.updateCode) {
		case VIEW_UPDATE_TABLE:
		    SclGetScrollBar(bar, &bar_value,
			    &bar_min, &bar_max, &bar_pagesize);
		    if (bar_value < num_keys - bar_pagesize)
			bar_value = 0;
		    SclSetScrollBar(bar, bar_value,
			    0, num_keys, bar_pagesize);

		    numCols = 2;
		    numRows = TblGetNumberOfRows(tbl);
		    for (i = 0; i < numRows; i++) {
			/* Set row data. */
			if (i < num_keys) {
			    TblSetRowData(tbl, i, shortid_list[i]);
			    TblSetRowUsable(tbl, i, true);
			}
			else 
			    TblSetRowUsable(tbl, i, false);

			for(j = 0; j < numCols; j++)
			    TblSetItemStyle(tbl, i, j, customTableItem);
			TblMarkRowInvalid(tbl, i);
		    }

		    for(j = 0; j < numCols; j++) {
			TblSetCustomDrawProcedure(tbl, j, ViewDrawCell);
			TblSetColumnUsable(tbl, j, true);
		    }

		    FrmDrawForm(frm);
		    WinDrawLine(0, 28, 159, 28);
		    i = 30+89;	/* bottom line */
		    WinDrawLine(43, 16, 43, i);
		    WinDrawLine(0, i, 159, i);
		    handled = true;
		    break;

		case VIEW_GENERATE_LIST:
		    if (NULL != shortid_list)
			free(shortid_list);
		    shortid_list = NULL;
		    i = CtlGetValue(FrmGetObjectPtr(frm,
				FrmGetObjectIndex(frm, btnPublic)));
		    dbR = (i) ? pubring : secring;
		    i = get_shortkeyidlist(dbR, &shortid_list, &num_keys);
		    ErrFatalDisplayIf(-1 == i, "get_shortkeyid_list failed!");
		    break;
	    }
	    break;

	case tblSelectEvent:
	    row = e->data.tblSelect.row;
	    col = e->data.tblSelect.column;
	    if (0 != col) {
		TblUnhighlightSelection(tbl);
		TblSelectItem(tbl, row, 0);
	    }
	    set_object_visibility(BtnOk, 1);
	    break;

	case tblEnterEvent:
	    row = e->data.tblSelect.row;
	    col = e->data.tblSelect.column;
	    /* See if we should pop up window */
	    if (1 == e->data.tblSelect.column) {
		popwin = oldwin = savewin = NULL;
		oldwin = WinGetDrawWindow();
		/* Save the portion of the old window we'll be overwriting */
		saverec = winrec;
		saverec.topLeft.x = 0;
		saverec.topLeft.y -= 1;
		saverec.extent.x += 3;
		saverec.extent.y += 4;
		savewin = WinSaveBits(&saverec, &w);
		if (NULL != savewin) {
		    WinEraseRectangle(&winrec, 0);
		    popwin = WinCreateWindow(&winrec, popupFrame, 0, 0, &w);
		    WinInitializeWindow(popwin);
		    WinSetDrawWindow(popwin);
		    WinDrawWindowFrame();
		    /* get and draw info here */
		    data = TblGetRowData(tbl, row);	/* get short keyid */
		    i = lookup_keyid(pubring, &keyid, NULL, &data, 1, &keysize);
		    s = i ? "failed retrieving" : PGPKEYID2str(&keyid);

		    pub = (PGPpubkey_t *)MemPtrNew(sizeof(PGPpubkey_t));
		    ErrFatalDisplayIf(NULL == pub, "Can't alloc pubkey");
		    MemSet(pub, sizeof(PGPpubkey_t), 0);
		    i = get_pubkey(pubring, &keyid, pub);
		    ErrFatalDisplayIf(0 != i, "Can't get pubkey!");
		    
		    WinDrawChars("KeyID:", 6, 1, 0);
		    WinDrawChars(s, StrLen(s), 40, 0);
		    
		    WinDrawChars("ShortID:", 8, 1, 11);
		    StrPrintF(smbuf,"%lx", (long)data);
		    WinDrawChars(smbuf, StrLen(smbuf), 40, 11);

		    WinDrawChars("Bits:", 5, 95, 11);
		    StrPrintF(smbuf,"%i", (int)keysize);
		    WinDrawChars(smbuf, StrLen(smbuf), 95+25, 11);

		    WinDrawChars("Created:", 8, 1, 22);
		    ul = pub->created + POSIX_EPOCH_KLUDGE;
		    TimSecondsToDateTime(ul, &datetime);
		    DateToAscii(datetime.month,
			    datetime.day, datetime.year,
			    dfYMDLongWithSpace, smbuf),
		    StrCat(smbuf, ", ");
		    TimeToAscii(datetime.hour, datetime.minute,
				tfColon24h, ubuf);
		    StrCat(smbuf, ubuf);
		    StrPrintF(ubuf, ":%d", datetime.second);
		    StrCat(smbuf, ubuf);
		    WinDrawChars(smbuf, StrLen(smbuf), 40, 22);

		    WinDrawChars("Valid:", 6, 1, 33);
		    if (0 == pub->valid)
			StrPrintF(smbuf, "Forever");
		    else {
			date.year = datetime.year;
			date.month = datetime.month;
			date.day = datetime.day;
		    	ul = DateToDays(date) + (unsigned long)pub->valid;
			DateDaysToDate(ul, &date);
			DateToAscii(date.month, date.day, date.year,
				dfYMDLongWithSpace, smbuf);
		    }
		    WinDrawChars(smbuf, StrLen(smbuf), 40, 33);

		    WinDrawChars("Fingerprint:", 12, 1, 46);
		    smbuf[0] = '\0';
		    pub_fingerprint(pub, smbuf);
		    WinDrawChars(smbuf, 24, 50, 46);
		    s = smbuf + 25;
		    WinDrawChars(s, StrLen(s), 50, 55);

		    WinSetDrawWindow(oldwin);
		    
		    free_pubkey(pub);
		    MemPtrFree(pub);
		    pub = NULL;

		    handled = true;
		}
	    }
	    break;
	case tblExitEvent:
	    set_object_visibility(BtnOk, 0);
	    row = col = -1;	/* i.e. no valid selection */
	    break;

	case sclRepeatEvent:
	    FrmUpdateForm(ViewForm, VIEW_UPDATE_TABLE);
	    break;
	    
	case penUpEvent:
	    if (NULL != popwin) {
		WinDeleteWindow(popwin, 1);
		if (NULL != savewin) {
		    WinRestoreBits(savewin, saverec.topLeft.x,
			    saverec.topLeft.y);
		    popwin = savewin = oldwin = NULL;
		    TblUnhighlightSelection(tbl);
		    TblSelectItem(tbl, row, 0);
		    set_object_visibility(BtnOk, 1);
		}
	    }
	    break;

	case menuEvent:
	    MenuEraseStatus(NULL);
	    switch(e->data.menu.itemID) {
	    }

	    handled = true;
	    break;

	case ctlSelectEvent:
	    switch(e->data.ctlSelect.controlID) {
		case BtnCancel:
		    if (NULL != shortid_list)
			free(shortid_list);
		    shortid_list = NULL;
		    FrmUpdateForm(MainForm, 0);
		    FrmReturnToForm(MainForm);
	    }
	    break;
	
	case popSelectEvent:
	    if(DataColList == e->data.popSelect.listID) {
		i = e->data.popSelect.selection;
	    }
	    break;

	case appStopEvent:
	    if (NULL != shortid_list)
		free(shortid_list);
	    shortid_list = NULL;
	    break;

	default:
	    break;
    }

    CALLBACK_EPILOGUE;

    return handled;
}


static Boolean MainFormHandleEvent (EventPtr e)
{
    Boolean handled = false;
    FormPtr frm;
    VoidPtr foo;
    CharPtr s;
    int i;

    CALLBACK_PROLOGUE;

    frm = FrmGetActiveForm();

    switch (e->eType) {
	case frmOpenEvent:
	    FrmDrawForm(frm);
	    FrmUpdateForm(MainForm, 0);
	    handled = true;
	    break;
	    
	case frmUpdateEvent:
	    /* Set mode-list/trigger */
	    current_form = FRM_MAIN;
	    foo = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ModeList));
	    LstSetSelection(foo, current_form);
	    s = LstGetSelectionText(foo, current_form);
	    if (NULL != s) {
		foo = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ModeTrigger));
		CtlSetLabel(foo, s);
	    }
	    break;

	case menuEvent:
	    MenuEraseStatus(NULL);
	    switch(e->data.menu.itemID) {
	    }

	    handled = true;
	    break;

	case ctlSelectEvent:
	    switch(e->data.ctlSelect.controlID) {
	    }
	    break;
	
	case popSelectEvent:
	    if(ModeList == e->data.popSelect.listID) {
		i = e->data.popSelect.selection;
		if ( -1 < i && FRM_LAST > i) {	/* Valid selection ? */
		    if (i != current_form) {
			current_form = i;
			FrmPopupForm(form_ids[current_form]);
		    }
		}
	    }
	    break;

	default:
	    break;
    }

    CALLBACK_EPILOGUE;

    return handled;
}

static Boolean ApplicationHandleEvent(EventPtr e)
{
    FormPtr frm;
    Word    formId;
    Boolean handled = false;

    if (e->eType == frmLoadEvent) {
	formId = e->data.frmLoad.formID;
	frm = FrmInitForm(formId);
	FrmSetActiveForm(frm);

	switch(formId) {
	    case RandGenForm:
		FrmSetEventHandler(frm, RandGenHandleEvent);
		break;
	    case GenerateForm:
		FrmSetEventHandler(frm, GenerateFormHandleEvent);
		break;
	    case ViewForm:
		FrmSetEventHandler(frm, ViewFormHandleEvent);
		break;
	    case MainForm:
		FrmSetEventHandler(frm, MainFormHandleEvent);
		break;
	}
	handled = true;
    }

    return handled;
}

/* Get preferences, open (or create) app database */
static Word StartApplication(void)
{
    int ret = -1;
    CALLBACK_PROLOGUE;

    /* Open public keyring */
    pubring = open_pubkey_db(NULL);	/* Open default database */
    if (NULL == pubring) {
	/* If we failed, we couldn't create the database, so fail! */
	ErrDisplay("Couldn't open pubkey database!");
	goto end;
    }

    /* Open secret keyring */
    secring = open_seckey_db(NULL);	/* Open default database */
    if (NULL == secring) {
	/* If we failed, we couldn't create the database, so fail! */
	close_db(pubring);
	ErrDisplay("Couldn't open seckey database!");
	goto end;
    }

    FrmGotoForm(MainForm);
    ret = 0;

end:
    CALLBACK_EPILOGUE;
    return 0;
}

/* Save preferences, close forms, close app database */
static void StopApplication(void)
{
    CALLBACK_PROLOGUE;
    
    FrmSaveAllForms();
    FrmCloseAllForms();
    close_db(pubring);	/* close public keyring database */
    close_db(secring);	/* close secret keyring database */

    CALLBACK_EPILOGUE;
}

/* The main event loop */
static void EventLoop(void)
{
    Word err;
    EventType e;

    do {
	EvtGetEvent(&e, evtWaitForever);
	if (! SysHandleEvent (&e))
	    if (! MenuHandleEvent (NULL, &e, &err))
		if (! ApplicationHandleEvent (&e))
		    FrmDispatchEvent (&e);
    } while (e.eType != appStopEvent);
}

/* Main entry point; it is unlikely you will need to change this except to
   handle other launch command codes */
DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
    Word err;

    if (cmd == sysAppLaunchCmdNormalLaunch) {

	err = StartApplication();
	if (err) return err;

	EventLoop();
	StopApplication();

    } else {
	return sysErrParamErr;
    }

    return 0;
}
