/*
   Win4.c  -  (c) 1999 by Till Harbaum

   four bit per pixel (16 greyscale) mulg graphics
*/

#define NON_PORTABLE
#ifdef OS35HDR
#include <PalmOS.h>
#else
#include <Pilot.h>
typedef signed   char  Int8;
typedef unsigned char  UInt8; 
typedef VoidHand MemHandle;
#endif

#include "winx.h"
#include "mulg.h"
#include "tiles.h"
#include "rsc.h"

// Define auxillary routines dependent on WinXopts settings
#if defined(WINXDRAWCHARS)
#define WINXDRAWBITMAPEX
#endif

#if defined(WINXDRAWBITMAP)
#define WINXDRAWBITMAPEX
#endif


/* defines */
#define VPW   ((unsigned char  *)0xFFFFFA05)
#define LSSA  ((unsigned long  *)0xFFFFFA00)
#define PICF  ((unsigned char  *)0xFFFFFA20)
#define LPXCD ((unsigned char  *)0xFFFFFA25)
#define CKCON ((unsigned char  *)0xFFFFFA27)
#define LBAR  ((unsigned char  *)0xFFFFFA29)
#define FRCM  ((unsigned char  *)0xFFFFFA31)
#define LGPMR ((unsigned short *)0xFFFFFA32)

/* vars */
void _WinXSwitchDisplayModeGrey();
void _WinXSwitchDisplayModeBW();

typedef struct {
  WinXColor clr;
  WinXColor clrBackground;
  UInt8 *pbGreyScreenBase, *base;
  UInt8 *pbMonoScreenBase;
  short fGreyScale;
  Boolean flip;
  UInt8 lpxcd;
} WinXType;

WinXType vWinX = {
  clrBlack,	// clr
  clrWhite,	// clrBackground
};

#define cbppGrey 4		// bits per pixel grey
#define cppbuGrey 8		// pixels per BltUnit grey
#define cbpbu 16		// bits per BltUnit
#define shppbuGrey 2		// shift count pixels->BltUnit
#define buAllOnes 0xFFFF
#define BltUnit UInt16

#define dxScreen 160
#define dyScreen 160
#define cbGreyScreen (dxScreen*cbppGrey/8*dyScreen)	
#define MonoBToBltUnit(b) ((mpNibbleGreyUInt8[(b)>>4]<<8)|mpNibbleGreyUInt8[b&0xf])
#define WriteMask(buDest,buSrc,buMask)  (buDest=((buDest)&~(buMask))|((buSrc)&(buMask)))
	
BltUnit mpclrbuRep[] = { 0x0000, 0x5555, 0xAAAA, 0xFFFF };	

int _WinXFAllocScreen() {
  vWinX.flip = true;
  vWinX.base = vWinX.pbGreyScreenBase = MemPtrNew(2*cbGreyScreen);
  return vWinX.base != NULL;
}

void _WinXFreeScreen() {
  if (vWinX.base != NULL) {
    MemPtrFree(vWinX.base);
    vWinX.base = NULL;
  }
}
	
MemHandle      TileResH;
unsigned long *TileResP;

static Boolean EZcpu(void) {
  /* if we're not using a new include file, define things appropriately */
#ifndef sysFtrNumProcessorID
#define sysFtrNumProcessorID    2
  /* Product id */
  /* 0xMMMMRRRR, where MMMM is the processor model */
  /* and RRRR is the revision. */
#define sysFtrNumProcessorMask  0xFFFF0000 // Mask to obtain processor model
#define sysFtrNumProcessor328   0x00010000 // Motorola 68328 (Dragonball)
#define sysFtrNumProcessorEZ    0x00020000 // Motorola 68EZ328 (Dragonball EZ)
#endif
  UInt32 id;

  if(!FtrGet(sysFtrCreator, sysFtrNumProcessorID, &id))
    if((id & sysFtrNumProcessorMask) == sysFtrNumProcessorEZ)
      return true;
  
  return false;
}

/*-----------------------------------------------------------------------------
|	WinXSetGreyscale
|	
|		Switches Pilots display into greyscale mode
|	
|	Returns: 0 if successful, nonzero if not
-------------------------------------------------------wesc@ricochet.net-----*/
int WinXSetGreyscale() {
  if (vWinX.fGreyScale==0) {	/* only chage if req. */

    if(!EZcpu()) {
      FrmCustomAlert(alt_err, "This program requires a Palm IIIx, Palm V or Visor! Try the standard mulg version on this machine.",0,0);
      return(1);
    }

    if (!_WinXFAllocScreen()) {
      FrmCustomAlert(alt_err, "Out of memory!",0,0);
      return(1);
    }

    vWinX.pbMonoScreenBase=(char*)*LSSA;

    MemSet(vWinX.pbGreyScreenBase, 2*cbGreyScreen, 0);

    _WinXSwitchDisplayModeGrey();
    
    *LSSA=(long)vWinX.pbGreyScreenBase + cbGreyScreen;

    vWinX.fGreyScale=1;
  }
  
  /* get access to tile data */
  TileResH = (MemHandle)DmGetResource('tile', 0);
  TileResP = MemHandleLock((MemHandle)TileResH);

  return(0);
}

/*-----------------------------------------------------------------------------
|	WinXSetMono
|	
|		Switches back to mono mode.
|
|	Returns: 0 if successful, nonzero if not
-------------------------------------------------------wesc@ricochet.net-----*/
int WinXSetMono() {
  if (vWinX.fGreyScale!=1) return 1; /* already mono */
	
  MemPtrUnlock(TileResP);
  DmReleaseResource((MemHandle)TileResH );

  *LSSA=(long)vWinX.pbMonoScreenBase;
  _WinXSwitchDisplayModeBW();
	
  _WinXFreeScreen();
  vWinX.fGreyScale=0;
  return 0;
}

/*-----------------------------------------------------------------------------
|	WinXPreFilterEvent
|	
|		Helper -- put in your main event loop to handle common cases
|	where the pilot should be put back into mono mode
|	
|	Arguments:
|		const EventType *pevt
|	
|	Returns: true if event is processed
-------------------------------------------------------wesc@ricochet.net-----*/
Boolean WinXPreFilterEvent(const EventType *pevt) {
  if (pevt->eType == keyDownEvent && (pevt->data.keyDown.modifiers & 
				      commandKeyMask)) {
      switch (pevt->data.keyDown.chr) {
      case alarmChr:                       // back to bw for alarm events
	WinXSetMono();
	break;
      case lowBatteryChr:                  // dito for low battery warnings
      case keyboardChr:
      case graffitiReferenceChr:           // popup the Graffiti reference
      case keyboardAlphaChr:               // popup the keyboard in alpha mode
      case keyboardNumericChr:             // popup the keyboard in number mode
	return 1;	                   // eat, don't switch to mono
      case hardPowerChr:
      case backlightChr:	
	// don't switch to mono, but don't eat either
	break;
      }
  }
  return 0;
}		

const unsigned char mpNibbleGreyUInt8[]={
  0x00,0x03,0x0c,0x0f,0x30,0x33,0x3c,0x3f,
  0xc0,0xc3,0xcc,0xcf,0xf0,0xf3,0xfc,0xff};

void _WinXSwitchDisplayModeGrey() {
  /* set pixel clock devider to full speed */
  vWinX.lpxcd = *LPXCD;
  *LPXCD = 0;

  *CKCON=*CKCON & 0x7F;	/* display off*/
	
  /*virtual page width now 80 bytes (160 pixels)*/
  *VPW=40;
  *PICF=*PICF | 0x02; /*switch to grayscale mode*/
  *LBAR=20; /*line buffer now 40 bytes*/
  
  /*register to control grayscale pixel oscillations*/
  *FRCM=0xB9;
  
  /*let the LCD get to a 2 new frames (40ms delay) */
  SysTaskDelay(4);
  
  /*switch LCD back on */
  *CKCON=*CKCON | 0x80;
}

void _WinXSwitchDisplayModeBW() {
  //switch off LCD update temporarily
  *CKCON=*CKCON & 0x7F;
  
  //virtual page width now 20 bytes (160 pixels)
  *VPW=10;
  *PICF=*PICF & 0xFC; //switch to black and white mode
  *LBAR=10; // line buffer now 20 bytes
  
  //let the LCD get to a new frame (20ms delay)
  SysTaskDelay(4);
  
  //switch LCD back on in new mode
  *CKCON=*CKCON | 0x80;

  // restore original pixel clock devider state
  *LPXCD = vWinX.lpxcd;
}

unsigned long colors[]={
  0x00000000,0x55555555,0xaaaaaaaa,0xffffffff
};

unsigned long mask0[]={ 
  0xffffffff,0x0fffffff,0x00ffffff,0x000fffff,
  0x0000ffff,0x00000fff,0x000000ff,0x0000000f
};

unsigned long mask1[]={ 
  0x00000000,0xf0000000,0xff000000,0xfff00000,
  0xffff0000,0xfffff000,0xffffff00,0xfffffff0
};

unsigned short tr[]={
  0x0000, 0x000f, 0x00f0, 0x00ff, 0x0f00, 0x0f0f, 0x0ff0, 0x0fff,
  0xf000, 0xf00f, 0xf0f0, 0xf0ff, 0xff00, 0xff0f, 0xfff0, 0xffff
};

void WinXDrawBitmapEx(UInt8 *pbBits, short xSrc, short ySrc, 
		     short xDest, short yDest, short dxDest, short dyDest, 
		     short rowBytes, Boolean fDrawChar) {
  unsigned short *dst = ((unsigned short *)vWinX.pbGreyScreenBase) + 
    xDest/4 + 40*yDest;
  unsigned short *src = ((unsigned short*)pbBits) + xSrc/16 + 
    (rowBytes/2)*ySrc;
  unsigned long *s, *d, mask, b, smask;

  int i, doff = (xDest&3)<<2, soff = xSrc&15;

  if(vWinX.base == 0) return;

  smask = 0xffff0000 >> dxDest;

  for(i=0;i<dyDest;i++) {
    d = (unsigned long*)dst;
    s = (unsigned long*)src;

    /* get min 16 pixels (enough for character) */
    b = (*s >> (16-soff)) & smask;

    mask = mask1[(dxDest>4)?4:dxDest] >> doff;
    *d = (*d & ~mask) | ((unsigned long)tr[(b>>12)&15])<<(16-doff); 

    /* more than 4 pixels? */
    if(dxDest>4) {
      mask = mask1[(dxDest>8)?4:(dxDest-4)] >> doff;
            d = (unsigned long*)(dst+1);
      *d = (*d & ~mask) | ((unsigned long)tr[(b>> 8)&15])<<(16-doff); 
    }

    dst += 40;  /* next row */
    src += rowBytes/2;
  }
}

void WinXFillHorizLine(short xDest, short dx, short y, WinXColor clr) {
  unsigned long *dst0, *dst1;
  int off0, off1;

  dst0 = ((unsigned long *)vWinX.pbGreyScreenBase) + (xDest/8)    + 20*y;
  dst1 = ((unsigned long *)vWinX.pbGreyScreenBase) + (xDest+dx)/8 + 20*y;

  off0 = xDest&7;
  off1 = (xDest+dx)&7;

  if (vWinX.pbGreyScreenBase == 0) return;

  /* whole line within one long word? */
  if(dst0 == dst1) {
    *dst0 = ((*dst0) & ~(mask0[off0]&mask1[off1])) | 
             (colors[clr]&(mask0[off0]&mask1[off1]));
  } else {
    *dst0++ = ((*dst0) & ~mask0[off0]) | (colors[clr] & mask0[off0]);
    while(dst0 != dst1) *dst0++ = colors[clr];
    *dst1 = ((*dst1) & ~mask1[off1]) | (colors[clr] & mask1[off1]);
  }
}

int WinXFillRect(RectangleType *prc, int clr) {
  RectangleType rc;
  int y;
  
  rc = *prc;
  for (y = 0; y < rc.extent.y; y++)
    WinXFillHorizLine(rc.topLeft.x, rc.extent.x, rc.topLeft.y+y, clr);
  return 1;			
}
	
/*-----------------------------------------------------------------------------
|	WinXSetColor
|	
|		Sets the current drawing color.  All Fill, Line and Text
|	operations will use this color
|	
|	Arguments:
|		WinXColor clr
|	
|	Returns: old color
-------------------------------------------------------wesc@ricochet.net-----*/
WinXColor WinXSetColor(WinXColor clr) {
  WinXColor clrOld;
	
  clrOld = vWinX.clr;	
  vWinX.clr = clr;
  return clrOld;
}
	
/*-----------------------------------------------------------------------------
|	WinXSetBackgroundColor
|	
|		Sets the background color for WinXDrawChars
|	
|	Arguments:
|		WinXColor clr
|	
|	Returns: old color
-------------------------------------------------------wesc@ricochet.net-----*/
WinXColor WinXSetBackgroundColor(WinXColor clr) {
  WinXColor clrOld;
	
  clrOld = vWinX.clrBackground;	
  vWinX.clrBackground = clr;
  return clrOld;
}
	
#if defined(WINXDRAWCHARS)
/*-----------------------------------------------------------------------------
|	WinXDrawChars
|	
|		Draws chars on the screen.  Uses current color and background color
|	Note that Underline mode is not supported!
|	
|	Arguments:
|		CharPtr chars
|		UInt16 len
|		SUInt16 x
|		SUInt16 y
-------------------------------------------------------wesc@ricochet.net-----*/
void WinXDrawChars(char *chars, UInt16 len, Int16 x, Int16 y) {
  char *pch;
  char *pchMac;
  FontPtr pfnt;
  UInt16 *plocTbl;
  UInt16 *ploc;
  FontCharInfoType *pfciTbl;
  FontCharInfoType *pfci;
  RectangleType rc;
  int xRight;
  
  rc.topLeft.x = x;
  rc.topLeft.y = y;
  rc.extent.x = dxScreen-x;
  rc.extent.y = FntCharHeight();
  
  pfnt = FntGetFontPtr();
  
  xRight = rc.topLeft.x+rc.extent.x;
  pchMac = chars+len;
  plocTbl = ((UInt16 *)(pfnt+1))+pfnt->rowWords*pfnt->fRectHeight;
  pfciTbl = (FontCharInfoType *) (((UInt16 *)&pfnt->owTLoc)+pfnt->owTLoc);
  pchMac = chars+len;
  
  for (pch = chars; pch < pchMac && x < xRight; pch++) {
    int ch;
    int dch;
    int dxCharImage;
    int xCharImage;	
    
    ch = *pch;
    if (ch < pfnt->firstChar || ch > pfnt->lastChar) {
Missing:		
      ch = pfnt->lastChar+1;
    }

    dch = ch-pfnt->firstChar;
    pfci = pfciTbl+dch;
    if (*(UInt16 *)pfci == 0xffff)
      goto Missing;
    ploc = plocTbl+dch;
    xCharImage = *ploc;
    dxCharImage = *(ploc+1)-xCharImage;
    if (dxCharImage > 0) {
      if (x < rc.topLeft.x) {
	// check for partial on left
	if (x + dxCharImage >= rc.topLeft.x)
	  xCharImage+=rc.topLeft.x-x;
	else
	  goto SkipChar;
      }
      if (x + dxCharImage >= xRight) {
	// partial on right
	dxCharImage = xRight-x;
      }
      WinXDrawBitmapEx((UInt8 *)(pfnt+1), xCharImage, 0,  
		       x, y, dxCharImage, pfnt->fRectHeight, 
		       pfnt->rowWords*2, 1);
    }
SkipChar:			
    x += pfci->width;
  }
}	
#endif // WINXDRAWCHARS	

/* mulg game routines */

unsigned long save_sprite[MAX_SPRITES];
int save_hx[MAX_SPRITES], save_hy[MAX_SPRITES];
void create_sprite(int no, int sprite, int hx, int hy) {
  save_hx[no]=hx; save_hy[no]=hy;
  save_sprite[no]=sprite;
}

void init_sprites(void) {
}

int disp_grey(void) {
  return(vWinX.base!=0);
}

void draw_sprite(int no, int x, int y) {
  int a,n,m;
  unsigned long *save = (unsigned long *)(vWinX.pbGreyScreenBase+ 
		        4*((x-save_hx[no])/8) + 80*(y-save_hy[no]) );

  if (vWinX.base == 0) return;

  /* draw sprite (mask, shift) */
  n=4*((x-save_hx[no])&0x07);
  m=32-n;

  for(a=0;a<16;a++) {

    save[a*20]   = (save[a*20]   & 
		    ~(TileResP[(save_sprite[no]+1)*32+2*a+0]>>n)) | 
                     (TileResP[(save_sprite[no]  )*32+2*a+0]>>n);

    save[a*20+1] = (((save[a*20+1] & 
		     ~(TileResP[(save_sprite[no]+1)*32+2*a+1]>>n)) | 
		      (TileResP[(save_sprite[no]  )*32+2*a+1]>>n)) &
		     ~(TileResP[(save_sprite[no]+1)*32+2*a+0]<<m)) | 
		      (TileResP[(save_sprite[no]  )*32+2*a+0]<<m);

    save[a*20+2] = (save[a*20+2] & 
		    ~(TileResP[(save_sprite[no]+1)*32+2*a+1]<<m)) | 
                     (TileResP[(save_sprite[no]  )*32+2*a+1]<<m);
  }
}

void draw_tile(int tile_num, int x, int y) {
  int a,i;
  unsigned long *screen=(unsigned long *)vWinX.pbGreyScreenBase;

  if (vWinX.base == 0) return;

  tile_num &= 0xff;

  for(a=0;a<16;a++) {
    *(screen + y*320 + 2*x+   a*20)=TileResP[tile_num*32+2*a  ];
    *(screen + y*320 + 2*x+1+ a*20)=TileResP[tile_num*32+2*a+1];
  }
}

/* buffer with level data */
extern unsigned char level_buffer[MAX_HEIGHT][MAX_WIDTH];

void draw_level(int xp, int yp) {
  int x,y,b;
  unsigned long *screen=(unsigned long *)vWinX.pbGreyScreenBase;
  unsigned long *d, *s;
  
  if (vWinX.pbGreyScreenBase == 0) return;

  d = screen;
  for(y=0;y<9;y++) {

    for(x=0;x<10;x++) { 
      s = &TileResP[level_buffer[y+yp][x+xp]*32];

      *d++=*s++; *d=*s++; d+=19; *d++=*s++; *d=*s++; d+=19;
      *d++=*s++; *d=*s++; d+=19; *d++=*s++; *d=*s++; d+=19;
      *d++=*s++; *d=*s++; d+=19; *d++=*s++; *d=*s++; d+=19;
      *d++=*s++; *d=*s++; d+=19; *d++=*s++; *d=*s++; d+=19;

      *d++=*s++; *d=*s++; d+=19; *d++=*s++; *d=*s++; d+=19;
      *d++=*s++; *d=*s++; d+=19; *d++=*s++; *d=*s++; d+=19;
      *d++=*s++; *d=*s++; d+=19; *d++=*s++; *d=*s++; d+=19;
      *d++=*s++; *d=*s++; d+=19; *d++=*s++; *d=*s++; d+=19;

      d -= 318;
    }
    d += 300;
  }
}

void WinFlip(void) {
  if(vWinX.flip) {
    vWinX.pbGreyScreenBase = vWinX.base;
    *LSSA = (long)(vWinX.base+cbGreyScreen);
  } else {
    vWinX.pbGreyScreenBase = vWinX.base+cbGreyScreen;
    *LSSA = (long)vWinX.base;
  }
  vWinX.flip = !vWinX.flip;
}
