/*
 PalGar - use a PALm to up-, download, edit waypoints from a GARmin GPS
 Copyright (C) 1999 Paul Nash
 Copyright (C) 2000 Ralf Meyer <ranf@users.sourceforge.net>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

//
//	File/database manipulation routines
//
//	Paul Nash
//	paul@frcs.alt.za

//
//	$Header: /cvs/PalGar/file.c,v 1.16 2000/11/26 15:01:53 ranf Exp $
//
//	$Log: file.c,v $
//	Revision 1.16  2000/11/26 15:01:53  ranf
//	- edit eTrex waypoint now with altitude
//	
//	Revision 1.15  2000/11/09 21:21:19  ranf
//	- eTrex: added Altitude to WPT
//
//	Revision 1.14  2000/08/27 19:33:02  ranf
//	- Makefile: Target veryclean renamed to ``distclean
//	- after successfully talking to a GPS (Get ID) the prefs screen
//		shows ID, version, product description
//	- reordered Menu
//
//	Revision 1.13  2000/08/24 20:48:22  ranf
//	- global struct to keep Product_Data
//	- global struct to keep Protocol_Array
//	- removed anything related to the callback.h stuff
//	- version number up to 1.7.0
//
//	Revision 1.12  2000/07/12 18:35:21  ranf
//	- bugfix: when merging on a Palm OS >= 3 it only took
//	every second record
//	- first attempt to add eTrex waypoint up- /download
//	(needs testing)
//
//	Revision 1.11  2000/07/10 19:14:00  ranf
//	Edit Wpt symbol. No bitmap yet.
//
//	Revision 1.10  2000/07/06 19:09:12  ranf
//	file.c rcs_id Problem
//	Modified Files:
//	 	file.c
//
//	Revision 1.9  2000/07/06 19:02:15  ranf
//	Options dialogue for setting the time zone
//
//	Revision 1.8  2000/07/02 13:18:30  ranf
//	- get date/time from the Palm (PalGar has to be in host mode).
//		Be careful with that: there's no local time to UTC conversion yet.
//	- turn off the Palm (again host mode required).
//
//	Revision 1.7  2000/06/26 17:55:16  ranf
//	Checksum errors when asked for GPS type
//
//	Revision 1.3  2000/03/27 18:24:31  giga
//	OS 35 compiles now
//
//	Revision 1.2  2000/03/27 18:03:29  giga
//	OS 35 Upgrader Perl ran
//
//	Revision 1.1.1.1  2000/03/23 18:18:40  giga
//	Initial import
//
//	Revision 1.14  1999/08/30 06:35:32  paul
//	Added Jaque Andre's patches
//
//	Revision 1.13  1999/05/12 08:48:05  paul
//	Ready for release as Gypsy 1.4.  Major changes include:
//	*  Rudimentary track/route buffer to transfer data between GPS
//	   and PC (a'la gps2p2pc)
//	*  Protocol fix for DLE escaping bug
//	*  Code changes from <andre@drfmc.ceng.cea.fr> to be able to
//	   compile a PalmOS2 version
//	*  Garmin protocol tries a bit harder to recover from comms
//	   errors (but is still *very* dumb)
//
//


#include <PalmOS.h>
#include "gps.h"

#include "globals.h"
#include "GarminDef.h"

//
//	Local function prototypes
//

void 	fileDrawName (MemPtr tbl, UInt16 row, UInt16 column, RectanglePtr bounds);
int 	HandleFileMenu (UInt16 menu_item);
int 	HandleFileButton (UInt16 button);
void 	DisplayFileTable (void);
void 	MergeDatabase (LocalID id, UInt16 card);
Boolean	mergeProgressCallback (PrgCallbackData *data);


//	Internal global variables and the like

// static char *rcs_id = "$Id: file.c,v 1.16 2000/11/26 15:01:53 ranf Exp $";

static LocalID 	current_file = -1;
static UInt16 	current_card = 0;
extern UInt32 	romVersion;
extern Int8		timeoffset;
extern Product_Data_Type	A000;

Boolean HandleFileListEvent (EventPtr e)
{
   	int 		handled = false;
   	FormPtr 	form;
   	TablePtr 	table;
   	UInt16 		row;

   	form = FrmGetActiveForm ();
   	switch (e->eType) {
      	case frmOpenEvent:
	 	setField (dbName, currentDBName ());
	 	FrmDrawForm (form);
	 	DisplayFileTable ();
	 	current_file = -1;
	 	handled = true;
	 	break;
      	case tblSelectEvent:
	 	table = FrmGetObjectPtr( form, FrmGetObjectIndex( form, fileTable ));
	 	row = e->data.tblSelect.row;
	 	current_file = TblGetRowData (table, row);
	 	current_card = TblGetItemInt (table, row, 2);
	 	handled = true;
	 	break;
      	case ctlSelectEvent:
        	handled =  HandleFileButton (e->data.ctlEnter.controlID);
	 	break;
      	default:
		break;
   	}
   	return handled;
}


//
//	Display a table of file names with our Application ID
//

void DisplayFileTable ()
{
   	int 		i;
   	FormPtr 	form;
   	TablePtr 	table;
   	FieldPtr 	field;
   	DmSearchStateType state;
   	Boolean 	new = true;
   	UInt16 		card;
   	UInt32 		nrecs = 42;
   	LocalID 	id = 0;
	LocalID   	current_id = -1;

   	form = FrmGetActiveForm();
   	table = FrmGetObjectPtr( form, FrmGetObjectIndex( form, fileTable ));
   	field = FrmGetObjectPtr (form, FrmGetObjectIndex (form, dbName));
   	if (waypoint_db) {
      		DmOpenDatabaseInfo (waypoint_db, &current_id, NULL, NULL, NULL, NULL);
   	}
   	FldEraseField (field);
   	setField (dbName, currentDBName ());
   	FldDrawField (field);
   	TblEraseTable (table);
   	TblSetColumnUsable (table, 0, true);
   	TblSetColumnUsable (table, 1, true);
   	TblSetColumnUsable (table, 2, true);
   	TblSetCustomDrawProcedure (table, 0, fileDrawName);
   	for (i = 0; i < TABLE_ROWS; ++i) {
      		if (DmGetNextDatabaseByTypeCreator (new, &state, DB_TYPE, DB_APPID,
			false, &card, &id)) {
	 		TblSetRowUsable (table, i, false);
      		}
    		else {
	 		new = false;
	 		if (waypoint_db && file_merge && current_id == id) {
	    			--i;
	    			continue;
	 		}
	 		TblSetRowData (table, i, id);
	 		TblSetItemStyle (table, i, 0, customTableItem);
	 		DmDatabaseSize (card, id, &nrecs, NULL, NULL);
	 		TblSetItemStyle (table, i, 1, numericTableItem);
	 		TblSetItemInt (table, i, 1, nrecs);
	 		TblSetItemStyle (table, i, 2, numericTableItem);
	 		TblSetItemInt (table, i, 2, card);
	 		TblSetRowUsable (table, i, true);
	 		TblSetRowSelectable (table, i, true);
	 		TblMarkRowInvalid (table, i);
      		}
	}
	TblDrawTable (table);
}


//
//	Custom draw procedure for the file name.  Takes the
//	local ID & gets the name.
//

void fileDrawName (MemPtr tbl, UInt16 row, UInt16 column, RectanglePtr bounds)
{
   	char 		txt[32];
   	LocalID 	id;
   	FontID 		old_font;
   	UInt16 		card;

   	id = TblGetRowData (tbl, row);
   	card = TblGetItemInt (tbl, row, 2);
   	if (DmDatabaseInfo (card, id, txt, NULL, NULL, NULL, NULL, NULL,
				      NULL, NULL, NULL, NULL, NULL)) {
      		return;
   	}
   	old_font = FntSetFont (0);
   	WinDrawChars (txt, StrLen(txt), bounds->topLeft.x, bounds->topLeft.y);
   	FntSetFont (old_font);
   	return;
}


//
//	Process the various buttons.
//

int HandleFileButton (UInt16 button)
{
   	int 		handled = false;
   	FormPtr 	form;
   	LocalID 	id = -1;
   	char 		name[32];

   	form = FrmGetActiveForm ();
   	switch (button) {
      	case gpsOkBtn:
	 	if (current_file != -1) {
	    		if (! file_merge) {
	       			if (waypoint_db) {
		  			DmCloseDatabase (waypoint_db);
		  			waypoint_db = NULL;
	       			}
	       			waypoint_db = DmOpenDatabase (current_card,
					current_file, dmModeReadWrite);
	       			num_records = DmNumRecords (waypoint_db);
	       			table_top = 0;
	       			FrmGotoForm (gpsWptListFrm);
	    		}
	    		else {
	       			MergeDatabase (current_file, current_card);
	       			num_records = DmNumRecords (waypoint_db);
	       			table_top = 0;
	    		}
	 	}
	 	else {
	    		FrmAlert (noActiveDB);
	 	}
	 	handled = true;
	 	break;
      	case gpsNokBtn:
	 	if (waypoint_db) {
	    		FrmGotoForm (gpsWptListFrm);
	 	}
	 	else {
	    		FrmAlert (noActiveDB);
	 	}
	 	handled = true;
	 	break;
      	case gpsKillBtn:
	 	if (current_file != -1) {
	    		DmDatabaseInfo (current_card, current_file, name, NULL, NULL,
			    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
	    		if (FrmCustomAlert (fileDeleteAlert, name, " ", " ") == 0) {
	       			if (waypoint_db) {
		  			DmOpenDatabaseInfo (waypoint_db, &id, NULL, NULL, NULL, NULL);
		  			if (current_file == id) {
		     				DmCloseDatabase (waypoint_db);
		     				waypoint_db = NULL;
		  			}
	       			}
	       			DmDeleteDatabase (current_card, current_file);
	    		}
	 	}
	 	current_file = -1;
	 	DisplayFileTable ();
	 	handled = true;
	 	break;
      	case gpsNewBtn:
	 	newfile_camefrom = FrmGetFormId (form);
	 	FrmPopupForm (gpsFileNewFrm);
	 	handled = true;
	 	break;
      	default:
	 	FrmAlert (gpsAlert1);
	 	handled = true;
	 	break;
   	}
   	return handled;
}


//
//	MemHandle creating a new file, including getting filename
//

Boolean HandleFileNewEvent (EventPtr e)
{
   	int handled = false;
   	UInt16 err;
   	char name[64];
   	FormPtr form;

   	form = FrmGetActiveForm ();
   	switch (e->eType) {
      	case frmOpenEvent:
	 	FrmDrawForm (form);
	 	FrmSetFocus (form, FrmGetObjectIndex (form, newFile));
	 	handled = true;
	 	break;
      	case ctlSelectEvent:
         	switch (e->data.ctlEnter.controlID) {
	    	case gpsOkBtn:
	       		getField (newFile, name);
	       		if (name[0] != '\0') {
		  		err = DmCreateDatabase (0, name, DB_APPID, DB_TYPE, false);
		  		if (! err
		      			&& newfile_camefrom == gpsWptListFrm
					&& waypoint_db) {
		     			DmCloseDatabase (waypoint_db);
		     			waypoint_db = NULL;
		  		}
		  		if (!waypoint_db) {
		     			current_file = DmFindDatabase (0, name);
		     			waypoint_db = DmOpenDatabase (0, current_file,
						   dmModeReadWrite);
		  		}
		  		FrmCloseAllForms ();
		  		if (newfile_camefrom) {
		     			FrmGotoForm (newfile_camefrom);
		  		}
		  		else {
		     			FrmGotoForm (gpsWptListFrm);
		  		}
	       		}
	       		handled = true;
	       		break;
		case gpsNokBtn:
	       		FrmReturnToForm (0);
	       		handled = true;
	       		break;
	 	}
	 	break;
      	default:
		break;
   	}
   	return handled;
}

//
//	MemHandle Options
//

Boolean HandleFileOptionsEvent (EventPtr e)
{
   	int 		handled = false;
   	FormPtr 	form;
	ControlPtr 	ctl;
	ListPtr 	list;
	char		x[5];

   	form = FrmGetActiveForm ();
   	switch (e->eType) {
      	case frmOpenEvent:
    		/* Time offset */
    		list = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
        		optUTClist));
    		ctl = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
        		optUTCtrigger));
    		LstSetSelection (list, (Int16) (timeoffset +12) );
//    		LstScrollList   (list, winDown, (Int16) (timeoffset +12) );
    		LstSetTopItem   (list, (Int16) (timeoffset +12) );
    		CtlSetLabel(ctl, LstGetSelectionText(list,
			(Int16) (timeoffset +12) ));

	 	FrmDrawForm (form);

		// product ID
		StrIToA (x, A000.product_ID);
		WinDrawChars (x, StrLen(x), 10, 80);
		// Software version
		StrIToA (x, A000.software_version);
		WinDrawChars (x, StrLen(x), 10, 90);
		// description
		WinDrawChars (A000.product_description,
				StrLen(A000.product_description), 10, 100);
		handled = true;
	 	break;
      	case ctlSelectEvent:
         	switch (e->data.ctlEnter.controlID) {
	    	case gpsOkBtn:
    			/* Time offset */
    			list = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
        			optUTClist));
    			timeoffset = (Int16) (LstGetSelection(list) -12);

     			FrmGotoForm (gpsWptListFrm);
	       		handled = true;
	       		break;
	    	case gpsNokBtn:
     			FrmGotoForm (gpsWptListFrm);
	       		handled = true;
	       		break;
	 	 }
	 	 break;
      	default:
		 break;
   	}
   	return handled;
}


//
//	Merge another database into the current waypoint db
//

void MergeDatabase (LocalID id, UInt16 card)
{
   	UInt16 		i = 0;
        UInt16		new;
   	MemHandle 	rec,
			new_rec;
   	DmOpenRef 	db;
   	waypoint 	*wpt,
			*new_wpt;
   	LocalID 	my_id = -1;
   	UInt16 		my_card;
   	char 		msg[64];
   	ProgressPtr 	progress = NULL;

   	if (!waypoint_db) {
      		FrmAlert (gpsAlert1);
      		DisplayFileTable ();
      		return;
   	}
   	DmOpenDatabaseInfo (waypoint_db, &my_id, NULL, NULL, &my_card, NULL);
   	if (my_id == id && my_card == card) {
      		FrmAlert (gpsAlert1);
      		DisplayFileTable ();
      		return;
   	}
   	if (! (db = DmOpenDatabase (card, id, dmModeReadOnly))) {
      		DisplayFileTable ();
      		return;
   	}
   	new = DmNumRecords (db);
   	StrCopy (msg, " / ");
   	StrIToA (msg + 3, new);
   	if ( romVersion >= 0x03000000) {
      		progress = PrgStartDialogV31 ("Merge waypoints", mergeProgressCallback);
   	}
   	while (true) {
      		if (! (rec = DmQueryRecord (db, i++))) {
			break;
      		}
      		if ( romVersion >= 0x03000000) {
//         		PrgUpdateDialog (progress, 0, ++i, msg, true);
         		PrgUpdateDialog (progress, 0, i, msg, true);
      		}
      		new = DmNumRecords (waypoint_db) + 1;
      		new_rec = DmNewRecord (waypoint_db, &new, sizeof (waypoint));
      		wpt = (waypoint *) MemHandleLock (rec);
      		new_wpt = (waypoint *) MemHandleLock (new_rec);
      		DmWrite (new_wpt, 0, wpt, sizeof (waypoint));
      		MemHandleUnlock (rec);
      		MemHandleUnlock (new_rec);
      		DmReleaseRecord (waypoint_db, new, true);
   	}
   	if ( romVersion >= 0x03000000) {
      		PrgStopDialog (progress, true);
   	}
   	DmCloseDatabase (db);
   	DisplayFileTable ();
   	return;
}


//
//	Callback routine for the progress dialogue box
//

Boolean mergeProgressCallback (PrgCallbackData *data)
{
   	if (data->textLen > 20) {
      		StrCopy (data->textP, "Record ");
   	}
   	else {
      		data->textP[0] = '\0';
   	}
   	StrIToA (data->textP + StrLen (data->textP), data->stage);
   	StrCat (data->textP, data->message);
   	data->textChanged = true;
   	return true;
}
