/*
 PalGar - use a PALm to up-, download, edit waypoints from a GARmin GPS
 Copyright (C) 1999 Paul Nash
 Copyright (C) 2000 Ralf Meyer <ranf@users.sourceforge.net>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

//
//	Garmin protocol support
//
//	Paul Nash
//	paul@frcs.alt.za

//
//	$Header: /cvs/PalGar/garmin.c,v 1.25 2000/11/26 15:01:53 ranf Exp $
//
//	$Log: garmin.c,v $
//	Revision 1.25  2000/11/26 15:01:53  ranf
//	- edit eTrex waypoint now with altitude
//	
//	Revision 1.24  2000/11/25 14:18:56  ranf
//	- check for D108
//
//	Revision 1.23  2000/11/19 18:57:01  ranf
//	- WPT should also work with the eTrex Summit
//
//	Revision 1.22  2000/11/12 16:28:19  ranf
//	- eTrex: display symbol (in Hex for now)
//
//	Revision 1.21  2000/11/09 21:21:20  ranf
//	- eTrex: added Altitude to WPT
//
//	Revision 1.20  2000/11/07 19:42:42  ranf
//	- eTrex WPT download at least works
//
//	Revision 1.19  2000/09/19 20:19:03  ranf
//	- fixed probs with preferences saving
//
//	Revision 1.18  2000/09/13 19:38:24  giga
//	*** empty log message ***
//
//	Revision 1.17  2000/08/27 19:33:03  ranf
//	- Makefile: Target veryclean renamed to ``distclean
//	- after successfully talking to a GPS (Get ID) the prefs screen
//		shows ID, version, product description
//	- reordered Menu
//
//	Revision 1.16  2000/08/24 20:48:22  ranf
//	- global struct to keep Product_Data
//	- global struct to keep Protocol_Array
//	- removed anything related to the callback.h stuff
//	- version number up to 1.7.0
//
//	Revision 1.15  2000/07/25 18:16:27  ranf
//	- new target test: run POSE
//
//	Revision 1.14  2000/07/24 20:45:07  ranf
//	- bugfix: Get date/time from GPS
//
//	Revision 1.13  2000/07/23 17:19:36  ranf
//	- get Protocol info works now
//
//	Revision 1.12  2000/07/19 19:59:39  ranf
//	Get ID from GPS finally works
//
//	Revision 1.11  2000/07/12 18:35:21  ranf
//	- bugfix: when merging on a Palm OS >= 3 it only took
//	every second record
//	- first attempt to add eTrex waypoint up- /download
//	(needs testing)
//
//	Revision 1.10  2000/07/10 19:14:01  ranf
//	Edit Wpt symbol. No bitmap yet.
//
//	Revision 1.9  2000/07/06 19:02:15  ranf
//	Options dialogue for setting the time zone
//
//	Revision 1.8  2000/07/02 13:18:30  ranf
//	- get date/time from the Palm (PalGar has to be in host mode).
//		Be careful with that: there's no local time to UTC conversion yet.
//	- turn off the Palm (again host mode required).
//
//	Revision 1.7  2000/06/26 17:55:16  ranf
//	Checksum errors when asked for GPS type
//
//	Revision 1.4  2000/04/24 14:20:30  ranf
//	Added small icon
//		(not very nice. But now we have one)
//
//	Revision 1.3  2000/03/27 18:24:31  giga
//	OS 35 compiles now
//
//	Revision 1.2  2000/03/27 18:03:29  giga
//	OS 35 Upgrader Perl ran
//
//	Revision 1.1.1.1  2000/03/23 18:18:40  giga
//	Initial import
//
//	Revision 1.34  1999/08/31 06:03:23  paul
//	Ready for release as Gypsy 1.6
//
//	Revision 1.33  1999/08/30 07:03:44  paul
//	added test to skip old-style tracklogs
//
//	Revision 1.32  1999/08/30 06:35:32  paul
//	Added Jaque Andre's patches
//
//	Revision 1.31  1999/06/13 17:32:06  paul
//	fixed a minor bug in the new display model :-)
//
//	Revision 1.30  1999/06/13 17:26:13  paul
//	Changed lat & long edit to use N/S E/W list (button) input, rather
//	than text.  This allows the ':' separator to use ' ' or '-'.
//
//	Revision 1.29  1999/06/12 17:30:52  paul
//	Removed "track to GPS" from menu; confirm clear track log
//
//	Revision 1.28  1999/06/12 14:38:39  paul
//	changed track log to allow cumulative logging.  Not tested yet
//
//	Revision 1.27  1999/05/12 08:48:05  paul
//	Ready for release as Gypsy 1.4.  Major changes include:
//	*  Rudimentary track/route buffer to transfer data between GPS
//	   and PC (a'la gps2p2pc)
//	*  Protocol fix for DLE escaping bug
//	*  Code changes from <andre@drfmc.ceng.cea.fr> to be able to
//	   compile a PalmOS2 version
//	*  Garmin protocol tries a bit harder to recover from comms
//	   errors (but is still *very* dumb)
//
//

#include <PalmOs.h>
#include "gps.h"
#include <System/SerialMgrOld.h>

#include "globals.h"
#include "GarminDef.h"

#define BUFLEN  256
#define TRK_NAME_LEN 20
#define TRK_SIZE 1024*15
#define TRK_DB_NAME "PalGar tracks"
//#define TRK_DB_NAME "PGtrk"   +dbName
#define ROUTE_DB_NAME "PalGar raw routes"

//
//	Private data structures
//

typedef struct  {
   	char ident[GARMIN_NAME_LEN];
   	char lat[4];
   	char lon[4];
   	long unused;
   	char cmnt[GARMIN_DESC_LEN];
   	char smbl;
   	char dspl;
}	garmin_wpt;

typedef struct {
	UInt8	wpt_class;
	UInt8	color;
	UInt8	dspl;
	UInt8	attr;
	UInt16	symbol;
	char	subclass[18];
	char	lat[4];
	char	lon[4];
	char	alt[4];
	float	depth;
	float	dist;
	char	state[2];
	char	cc[2];	                    // country code
	char	ident[GARMIN_DESC_LEN];	    // 1-51
	char 	comment[GARMIN_DESC_LEN];	// 1-51
}	D108_garmin_wpt;

typedef struct {
   	char code;
   	char len;
   	char lat[4];
   	char lon[4];
   	char time[4];
   	unsigned char flag;
}	garmin_trk;

typedef struct {
   	char 	code;
   	char 	len;
   	char 	lat[4];
   	char 	lon[4];
   	char 	time[4];
	float	alt;
	float	dpth;
   	unsigned char flag;
}	D301_garmin_trk;

typedef struct {
   	char ident[TRK_NAME_LEN];
   	long lat_min;
   	long lat_max;
   	long lon_min;
   	long lon_max;
   	long distance;	 /* not yet implemented */
   	long size;
   	int  nb_point;
   	char flagged;
   	char data[1];
	float	alt;
}	pilot_trk;

//
//	Internal function prototypes
//

static void AppendToDB (garmin_wpt *wpt);
static void Trim (char *txt);
static void ConvertToGarmin (waypoint *wpt, garmin_wpt *garmin);
static void CopyAndFill (char *dst, char *src, int len);
static Boolean garminProgressCallback (PrgCallbackData *data);
static void dumpPacket (char *txt, char *buf, int len, int err);
static DmOpenRef openFreshDB (UInt32 type);
DmOpenRef openAppendDB (UInt32 type);
static pilot_trk * init_trk(void);
static void add_trkpoint(pilot_trk * trk, garmin_trk * gtrk);
static void save_trk(pilot_trk * trk);
static void AppendToDB_D108 (D108_garmin_wpt *garmin);


//	and internal global variables

// static char *rcs_id = "$Id: garmin.c,v 1.25 2000/11/26 15:01:53 ranf Exp $";
extern UInt32 	romVersion;
extern Product_Data_Type	A000;
extern Protocol_Array_Type	A001;
extern Int8		timeoffset;

//
//	Retrieve the waypoint list from a GPS, and put all
//	waypoints into the current database.
//

int GarminGetAll (int mode, UInt16 num_wpts)
{
   	int 	err = false;
   	garmin_wpt * garmin;
   	D108_garmin_wpt * garmin108;
   	UInt8	buf[BUFLEN];
   	int 		len;
   	char 	last_ack = 0;
   	char 	cmd = 0;
   	DmOpenRef 	raw_db = NULL;
   	MemPtr 	rec;
   	MemHandle 	new_rec;
   	int 	i = 0,
	num_errs = 0;
   	UInt16 	rec_num = 0;
   	char 	msg[64];
   	ProgressPtr progress = NULL;
   	pilot_trk * trk = NULL;
   	int		lOut = false;


   	if (mode != HOST_WAYPOINT) {
      		if ((err = OpenSerialPort ())) {
			return err;
      	}
     	switch (mode) {
	case GPS_WAYPOINT:
	   	cmd = Cmnd_Transfer_Wpt;
	   	break;
	case GPS_TRACK:
	   	if (! (trk = init_trk())) {
	      		err = -1;
	   	}
	   	cmd = Cmnd_Transfer_Trk;
	   	break;
	case GPS_ROUTE:
	   	if (! (raw_db = openFreshDB (DB_ROUTE))) {
	      		err = -1;
	   	}
	   	cmd = Cmnd_Transfer_Rte;
	   	break;
	case Pid_Product_Rqst :
		cmd = Pid_Product_Rqst;
		break;
	}
	err = GarminGetID ();

     	if (err) {
	 	SerClose (serial_lib);
	 	serial_port_open = false;
     	}

     	num_wpts = 0;
     	// Garmin "send xxxx" command
     	buf[0] = Pid_Command_Data;
     	buf[1] = 2;
     	buf[2] = cmd;
     	buf[3] = 0;
     	GarminSendPacket (buf);
   }
   else if ( romVersion >= 0x03000000) {
      	StrCopy (msg, "Receiving ~    / ");
      	StrIToA (msg + StrLen (msg), num_wpts);
      	progress = PrgStartDialogV31 ("Receive from Host", garminProgressCallback);
   }

   while (! lOut) {
      len = BUFLEN;
      err = 0;
      if (GarminGetPacket (buf, &len, &err)) {
	 	if (++num_errs > 3) {
			lOut = true;
	    		FrmCustomAlert (serialAlert, "Receive packet", " ", " ");
	    		break;
	 	}
	 	SendAck (buf[1], false);
	 	continue;
      	}
      	num_errs = 0;
      	if (buf[1] != Pid_Ack_Byte && buf[1] != Pid_Nak_Byte) {
	 	if (!err) {
	    		SendAck (buf[1], true);
	    		last_ack = buf[1];
	 	}
	 	else {
	    		SendAck (buf[1], false);
	 	}
     	}
	switch (buf[1]) {
	case Pid_Records:		// # of waypoints
	    num_wpts = buf[4] & 0x00FF;
	    num_wpts = num_wpts << 8;
	    num_wpts |= buf[3] & 0x00FF;
 	    if ( romVersion >= 0x03000000) {
 	        StrCopy (msg, "Receiving ~   / ");
 	        StrIToA (msg + StrLen (msg), num_wpts);
            	progress = PrgStartDialogV31 ("Receive from GPS",
			garminProgressCallback);
 	    }
	    break;
	 case Pid_Wpt_Data:		// waypoint record
		if (( A001[4].tag == 'D' ) &&  ( A001[4].data == 108 )) {
			garmin108 = (D108_garmin_wpt *) (buf +3);
	    		AppendToDB_D108 (garmin108);
		} else {
	    	  	garmin = (garmin_wpt *) (buf + 3);
	    	  	AppendToDB (garmin);
		};
 	    if ( romVersion >= 0x03000000) {
 			PrgUpdateDialog (progress, 0, ++i, msg, true);
 	    }
	    break;
	 case Pid_Rte_Hdr:		// route ID
	 case Pid_Rte_Wpt_Data:		// route waypoint
	    if (! raw_db) {
	       raw_db = openFreshDB (DB_ROUTE);
	    }
	    if (raw_db) {
	       rec_num = DmNumRecords (raw_db) + 1;
	       new_rec = DmNewRecord (raw_db, &rec_num, buf[2] + 2);
	       rec = MemHandleLock (new_rec);
	       DmWrite (rec, 0, buf + 1, buf[2] + 2);
	       MemHandleUnlock (new_rec);
	    }
	    if ( romVersion >= 0x03000000) {
			PrgUpdateDialog (progress, 0, ++i, msg, true);
	    }
	    break;
	 case Pid_Trk_Data:		// track record
		// D301 ?
	    if (! trk) {
	       trk = init_trk();
	    }
	    if (trk) {
	       add_trkpoint(trk, (garmin_trk *) (buf + 1));
	    }
	    if ( romVersion >= 0x03000000) {
		PrgUpdateDialog (progress, 0, ++i, msg, true);
	    }
	    break;
	 case Pid_Xfer_Cmplt:		// EOF
		lOut = true;
	    	break;
	 case Pid_Ack_Byte:		// ACK
	    	break;
	 case Pid_Nak_Byte:		// NAK
	    if (last_ack) {
	       	SendAck (last_ack, true);
	    }
	    else {
	       	buf[0] = Pid_Command_Data;
	       	buf[1] = 2;
	       	buf[2] = cmd;
	       	buf[3] = 0;
	       	GarminSendPacket (buf);
	    }
	    break;
//	 case Pid_Product_Data :
	 default:
	    break;
      }
   }
   if ( romVersion >= 0x03000000) {
      PrgStopDialog (progress, true);
   }
   if (mode != HOST_WAYPOINT) {
      SerClose (serial_lib);
      serial_port_open = false;
   }
   if (raw_db) {
      DmCloseDatabase (raw_db);
   }
   if (trk) {
      MemPtrFree (trk);
      trk = NULL;
   }
   return(0);
}


//
//  Get ID from GPS
//

int GarminGetID ()
{
    int  	err;
    UInt8 	buf[4];
    UInt8 	ack[BUFLEN];
    int  	len = 0;
    int  	paklen;
    char    last_ack = 0;
    int  	num_errs = 0;
    char    msg[BUFLEN];
    int     lOut = false;
	int		lGotID = false;
	int		serial_port_already_open;
	int		i;
	char	tmp[4];
	char	data[4];

	serial_port_already_open = serial_port_open;
	if ( ! serial_port_open ) {
    		if ((err = OpenSerialPort ())) {
        		serial_port_open = false;
        		return err;
		}
	}

    	buf[0] = Pid_Product_Rqst;
    	buf[1] = 2;
    	buf[2] = 0x20;
    	buf[3] = 0;
	SerReceiveFlush( serial_lib, 0);
    	GarminSendPacket (buf);

    while (! lOut) {
        paklen = BUFLEN;
        err = 0;
        if (GarminGetPacket (ack, &paklen, &err)) {
            if (++num_errs > 3) {
		lOut = true;
		StrIToA(msg, err);
		if ( ! lGotID ) {
                	FrmCustomAlert (serialAlert, "Receive ID", msg, " ");
		}
                continue;
            }
            SendAck (ack[1], false);
            continue;
        }
        num_errs = 0;
        if (ack[1] != Pid_Ack_Byte && buf[1] != Pid_Nak_Byte) {
            if (! err) {
                SendAck (buf[1], true);
                last_ack = buf[1];
            }
            else {
                SendAck (buf[1], false);
            }
        }
        switch (ack[1]) {
        case Pid_Product_Data :
			A000.product_ID = ack[4] & 0xFF;
			A000.product_ID = A000.product_ID << 8;
			A000.product_ID |= ack[3] & 0xFF;

			A000.software_version = ack[6] & 0xFF;
			A000.software_version = A000.software_version << 8;
			A000.software_version |= ack[5] & 0xFF;
			StrCopy( A000.product_description, & ack[7] );
			len = StrLen( A000.product_description );
//			StrCopy( msg, & ack[7] );
//			len = StrLen( msg );
//			StrCat( msg, "\n" );
			StrCopy( A000.product_additional, & ack[7] +len +1 );
//			StrCat( msg, & ack[7 +len +1] );

//			len = StrLen( msg );
//			StrIToA( msg, len );
			#ifdef DEBUG
//            	FrmCustomAlert (gpsBreak1, msg, " ", " ");
			#endif
            		//lOut = true;
            		//break;
            		SendAck (ack[1], true);
			lGotID = true;
			continue;
		case Pid_Protocol_Array:
			lOut = true;
			MemSet( A001, sizeof( A001), 0 );
//			dumpPacket(msg, ack, paklen, err);
			for ( i=0; i < ((paklen -3) / 3) -1; i++ ) {
				A001[i].tag  = (char)   ( ack[3 *i +3] ) ;
				A001[i].data = (UInt16) ( ack[3 *i +5] );
				A001[i].data = A001[i].data << 8;
				A001[i].data |= ack[3 *i +4];
		}
		StrCopy(msg,"\n");
		tmp[1] = '\0';
		for ( i=0; i < ((paklen -3) / 3) -1; i++ ) {
			tmp[0] = A001[i].tag;
			StrCat(msg, tmp );
			StrIToA(data, A001[i].data);
			StrCat(msg, data );
			StrCat(msg, ", ");
		}
		#ifdef DEBUG
            		FrmCustomAlert (gpsBreak1, msg, " ", " ");
		#endif
		break;
        case Pid_Ack_Byte:     // ACK
            break;
        case Pid_Nak_Byte:     // NAK
            if (last_ack) {
                SendAck (last_ack, true);
            }
            else {
                GarminSendPacket (buf);
            }
            break;
        default:
            break;
        }
    }
	if ( ! serial_port_already_open ) {
    	SerClose (serial_lib);
    	serial_port_open = false;
	}

    return(0);
}


//
//  Get date/time from GPS
//

int GarminGetDateTime ()
{
   	int 	err;
   	UInt8	buf[BUFLEN];
   	UInt8	ack[BUFLEN];
   	int 	len;
   	char 	last_ack = 0;
   	int 	num_errs = 0;
    int     lOut = false;
   	char 	msg[64];
   	ProgressPtr progress = NULL;
	DateTimeType	dat;
   	UInt8	yearHi;
   	UInt8	yearLo;
	UInt32	seconds;
	int		serial_port_already_open;

    Char	DateAscii[longDateStrLength];
	Char	TimeAscii[timeStringLength];
	SystemPreferencesType	sysPrefs;

	serial_port_already_open = serial_port_open;
	if ( ! serial_port_open ) {
    	if ((err = OpenSerialPort ())) {
        	serial_port_open = false;
        	return err;
		}
	}


    	// Garmin "send xxxx" command
    	buf[0] = Pid_Command_Data;
    	buf[1] = 2;
    	buf[2] = Cmnd_Transfer_Time;
    	buf[3] = 0;
	SerReceiveFlush ( serial_lib, 0) ;
    	GarminSendPacket (buf);
   	if ( romVersion >= 0x03000000) {
      		StrCopy (msg, "Receiving ~    / ");
      		progress = PrgStartDialogV31 ("Receive Date/Time", garminProgressCallback);
   	}

   	while ( ! lOut ) {
      		len = BUFLEN;
      		err = 0;
      		if (GarminGetPacket (ack, &len, &err)) {
	 		if (++num_errs > 3) {
	    		FrmCustomAlert (serialAlert, "Receive Date/Time", " ", " ");
				lOut = true;
				continue;
	 		}
	 		SendAck (ack[1], false);
	 		continue;
      	}
      	num_errs = 0;
      	if (ack[1] != Pid_Ack_Byte && ack[1] != Pid_Nak_Byte) {
	 	if (!err) {
	    		SendAck (ack[1], true);
	    		last_ack = ack[1];
	 	}
	 	else {
	    		SendAck (ack[1], false);
	 	}
     	}
     	switch (ack[1]) {
	 	case Pid_Date_Time_Data :
		   	dat.month	= ack[3];
		   	dat.day		= ack[4];
		   	yearLo		= ack[5];
		   	yearHi		= ack[6];
			dat.year	= yearHi << 8;
			dat.year	|= yearLo;
		   	dat.hour	= ack[7];
		   	dat.minute	= ack[9];
		   	dat.second	= ack[10];
			// don`t forget: timeoffset
			TimAdjust(&dat, timeoffset *60 *60);

			// Get time and date formats
			// from the system preferences.
    			PrefGetPreferences (&sysPrefs);
        		DateToDOWDMFormat (dat.month,
            		dat.day, dat.year,
                	sysPrefs.longDateFormat, DateAscii);
			TimeToAscii(dat.hour,
				dat.minute,
				sysPrefs.timeFormat, TimeAscii);

			seconds = TimDateTimeToSeconds( &dat );

		#ifdef DEBUG
    			FrmCustomAlert (gpsBreak1, "Date/Time",
					DateAscii, TimeAscii);
		#else
			TimSetSeconds (seconds);
		#endif
		lOut = true;
	    	break;
	 	case Pid_Xfer_Cmplt:		// EOF
			lOut = true;
			break;
	 	case Pid_Ack_Byte:		// ACK
	    	break;
	 	case Pid_Nak_Byte:		// NAK
	    	if (last_ack) {
	       		SendAck (last_ack, true);
	    	}
	    	else {
	       		GarminSendPacket (buf);
	    	}
	    	break;
  		default:
	    	break;
      }
   	}
   		if ( romVersion >= 0x03000000) {
      		PrgStopDialog (progress, true);
   	}
    	SerClose (serial_lib);
    	serial_port_open = false;
   	return(0);
}


//
//	Send one or more waypoints to the GPS.  This
//	requires us to count the number of qualifying
//	waypoints (flagged, changed or all), then sift
//	through the database and send them.
//

int GarminSend (UInt16 command)
{
   	int num_wpts;
   	waypoint *wpt;
   	garmin_wpt *garmin;
   	MemHandle rec;
   	int rec_num, max, retries;
   	UInt8	buf[BUFLEN];
   	UInt8	ack[BUFLEN];
   	int len, err, i = 0;
   	union {
      		int i;
      		char c[2];
   	} swap;
   	char msg[64];
   	ProgressPtr progress = NULL;

   	int		serial_port_already_open;
	int		lOut = false;

	max = DmNumRecords (waypoint_db);
	if (command == menuSendAll) {
    		num_wpts = max;
	}
	else {			// find out how many records to send
    		num_wpts = 0;
    		for (rec_num = 0; rec_num < max; ++rec_num) {
	 		rec = DmQueryRecord (waypoint_db, rec_num);
	 		if (! rec) {
	    			continue;
	 		}
	 		wpt = MemHandleLock (rec);
	 		if (command == menuSendChanged && wpt->changed) {
	    			++num_wpts;
	 		}
	 		if (command == menuSendFlagged && wpt->flagged) {
	    			++num_wpts;
	 		}
	 		MemHandleUnlock (rec);
    		}
	}

   	serial_port_already_open = serial_port_open;
   	if (! serial_port_open) {
      		if ((err = OpenSerialPort ())) {
	 		return err;
      		}
   	}
   	buf[0] = Pid_Records;
   	buf[1] = 2;
   	swap.i = num_wpts;
   	buf[2] = swap.c[1];
   	buf[3] = swap.c[0];
   	SerReceiveFlush (serial_lib, 0);
    	if ( romVersion >= 0x03000000) {
       		StrCopy (msg, "Sending ~   / ");
       		StrIToA (msg + StrLen (msg), num_wpts);
       		if (serial_port_already_open) {
          		progress = PrgStartDialogV31 ("Send to Host", garminProgressCallback);
       		} else {
          		progress = PrgStartDialogV31 ("Send to GPS", garminProgressCallback);
       		}
   	}
   	GarminSendPacket (buf);		// await 'n' records
   	len = BUFLEN;
   	GarminGetPacket (ack, &len, &err);

   rec_num = 0;
   lOut = false;
   while ( (rec_num < max) || lOut) {
      rec = DmQueryRecord (waypoint_db, rec_num++);
      if (! rec) {
	 	break;
      }
      wpt = MemHandleLock (rec);
      if (command == menuSendChanged && !wpt->changed) {
	 	MemHandleUnlock (rec);
	 	continue;
      }
      if (command == menuSendFlagged && !wpt->flagged) {
	 	MemHandleUnlock (rec);
	 	continue;
      }
      buf[0] = Pid_Wpt_Data;
      buf[1] = sizeof (garmin_wpt);
      garmin = (garmin_wpt *) (buf + 2);
      ConvertToGarmin (wpt, garmin);
      MemHandleUnlock (rec);
      SerReceiveFlush (serial_lib, 0);
      if ( romVersion >= 0x03000000) {
         PrgUpdateDialog (progress, 0, ++i, msg, true);
      }
      	for (retries = 0; retries < 5; ++retries) {
	 		GarminSendPacket (buf);
	 		len = BUFLEN;
	 		GarminGetPacket (ack, &len, &err);
	 		if (ack[1] == Pid_Ack_Byte)
	    		break;
	 		if (ack[1] == Pid_Nak_Byte) {
				lOut = true;
	    		break;
      		};
    	};
		if ( retries > 3) {
			lOut = true;
			break;
		};

//      if (ack[1] != Pid_Ack_Byte) {
//	 		FrmCustomAlert (serialAlert, "Transmit packet", " ", " ");
//      }
   }
   buf[0] = Pid_Xfer_Cmplt;
   buf[1] = 2;
   buf[2] = 7;
   buf[3] = 0;
   SerReceiveFlush (serial_lib, 0);
   GarminSendPacket (buf);
   len = BUFLEN;
   GarminGetPacket (ack, &len, &err);
   if ( romVersion >= 0x03000000) {
      PrgStopDialog (progress, true);
   }
   if (! serial_port_already_open) {
      SerClose (serial_lib);
      serial_port_open = false;
   }
   return(0);
}


//
//	Convert from internal format to Garmin
//

void ConvertToGarmin (waypoint *wpt, garmin_wpt *garmin)
{
   	MemSet (garmin, sizeof (garmin_wpt), 0);
   	CopyAndFill (garmin->ident, wpt->name,  sizeof (garmin->ident));
   	CopyAndFill (garmin->cmnt,  wpt->descr, sizeof (garmin->cmnt));

	endian_convert( & wpt->latitude, sizeof(long) );
   	MemMove ( &garmin->lat, &wpt->latitude, sizeof (long));

	endian_convert( & wpt->longitude, sizeof(long) );
   	MemMove ( &garmin->lon, &wpt->longitude, sizeof (long));

	garmin->smbl = (UInt16) wpt->symbol;
	endian_convert( & garmin->smbl, sizeof(UInt16) );

   	garmin->dspl = (char) wpt->display;
//   wpt->changed = false;
}

UInt16 ConvertToGarminD108 (waypoint *wpt, D108_garmin_wpt *garmin)
{
   	UInt16	len = 0;

   	MemSet (garmin, sizeof (D108_garmin_wpt), 0);
   	StrCopy (garmin->ident, wpt->name );
   	len += StrLen(wpt->name);
//   	StrCopy (garmin->comment, wpt->descr );
//   	len += StrLen(wpt->descr) ;

	endian_convert( & wpt->latitude, sizeof(long) );
   	MemMove ( &garmin->lat, &wpt->latitude, sizeof (long));
   	len += sizeof (long);

	endian_convert( & wpt->longitude, sizeof(long) );
   	MemMove ( &garmin->lon, &wpt->longitude, sizeof (long));
   	len += sizeof (long);

	garmin->symbol = (UInt16) wpt->symbol;
   	len += sizeof (char);

	// alt
	endian_convert( & wpt->alt, sizeof(long) );
   	MemMove ( &garmin->alt, &wpt->alt, sizeof (long));
   	len += sizeof (long);

   	return (len);
//   wpt->changed = false;
}


//
//	The Garmin units have fixed-length fields, padded
//	with spaces.
//

void CopyAndFill (char *dst, char *src, int len)
{
   	int i;

   	i = StrLen (src);
   	if (i > len)
    		i = len;
   	MemMove (dst, src, i);
   	if (i < len) {
      		while (i < len) {
	 		dst[i++] = ' ';
      		}
   	}
}


//
//	Append a record to the current database
//

void AppendToDB (garmin_wpt *garmin)
{
   	MemHandle 	rec_handle;
   	waypoint 	*new_wpt, *wpt;

   	current_record = DmNumRecords (waypoint_db) + 1;
   	rec_handle = DmNewRecord (waypoint_db, &current_record,
		sizeof (waypoint));
   	num_records = DmNumRecords (waypoint_db);
   	table_top = num_records - TABLE_ROWS;
   	wpt = MemPtrNew (sizeof (waypoint));
   	MemSet (wpt, sizeof (waypoint), 0);
   	new_wpt = MemHandleLock (rec_handle);
   	StrNCopy (wpt->name, garmin->ident, sizeof (garmin->ident));
   	*(wpt->name + sizeof (wpt->name) - 1) = '\0';
   	Trim (wpt->name);
   	StrNCopy (wpt->descr, garmin->cmnt, sizeof (garmin->cmnt));
   	*(wpt->descr + sizeof (wpt->descr) - 1) = '\0';
   	Trim (wpt->descr);

   	MemMove ( &wpt->latitude, &garmin->lat, sizeof (long));
	endian_convert( & wpt->latitude, sizeof(long) );

	MemMove ( &wpt->longitude, &garmin->lon, sizeof (long));
	endian_convert( & wpt->longitude, sizeof(long) );

   	wpt->symbol = garmin->smbl;
	//endian_convert( & wpt->symbol, sizeof(UInt16) );
   	wpt->display = garmin->dspl;
   	wpt->flagged = wpt->changed = false;
   	DmWrite (new_wpt, 0, wpt, sizeof (waypoint));
   	MemPtrFree (wpt);
   	MemHandleUnlock (rec_handle);
   	DmReleaseRecord (waypoint_db, current_record, true);
}


//
//	Append a record to the current database
//
void AppendToDB_D108 (D108_garmin_wpt *garmin)
{
   	MemHandle 	rec_handle;
   	waypoint 	*new_wpt, *wpt;

   	current_record = DmNumRecords (waypoint_db) + 1;
   	rec_handle = DmNewRecord (waypoint_db, &current_record,
		sizeof (waypoint));
   	num_records = DmNumRecords (waypoint_db);
   	table_top = num_records - TABLE_ROWS;
   	wpt = MemPtrNew (sizeof (waypoint));
   	MemSet (wpt, sizeof (waypoint), 0);
   	new_wpt = MemHandleLock (rec_handle);
   	StrCopy (wpt->name, garmin->ident);

	MemMove ( &wpt->latitude, &garmin->lat, sizeof (long));
	endian_convert( & wpt->latitude, sizeof(long) );

	MemMove ( &wpt->longitude, &garmin->lon, sizeof (long));
	endian_convert( & wpt->longitude, sizeof(long) );

	MemMove( &wpt->symbol, &garmin->symbol, sizeof(UInt16));
	endian_convert( & wpt->symbol, sizeof(UInt16) );
//   wpt->dspl = garmin->display;
   	wpt->flagged = wpt->changed = false;
   	//  alt
   	MemMove ( &wpt->alt, &garmin->alt, sizeof (long));
	endian_convert( & wpt->alt, sizeof(long) );

   	DmWrite (new_wpt, 0, wpt, sizeof (waypoint));
   	MemPtrFree (wpt);
   	MemHandleUnlock (rec_handle);
   	DmReleaseRecord (waypoint_db, current_record, true);
}


//
//	Send a packet, adding the relevant headers, stuffing DLEs and
//	the like ...
//

void GarminSendPacket (UInt8 *buf)
{
   	int i, j;
   	int len = 0;
   	char checksum = 0;
   	static UInt8	snd_buf[512];
   	UInt16 err;

   	len = buf[1] +2;
   	if ( buf[1] == 0) {
		len = 0;
   	};
   	snd_buf[0] = DLE;
   	for (i = j = 0; i < len; ++i) {
      		snd_buf[++j] = buf[i];
      		checksum -= buf[i];
      		if (snd_buf[j] == DLE) {
	 		snd_buf[++j] = DLE;
      		}
   	}
   	snd_buf[++j] = checksum;
   	if (snd_buf[j] == DLE) {
      		snd_buf[++j] = DLE;
   	}
   	snd_buf[++j] = DLE;
   	snd_buf[++j] = ETX;
   	++j;
   	SerSend (serial_lib, snd_buf, j, &err);
}


//
//	Open the serial port with some suitable settings
//

int OpenSerialPort ()
{
   	UInt16 err = 0;
   	SerSettingsType settings;

   	if (serial_lib == -1) {
      		FrmAlert (gpsSerialAlert);
      		return -1;
   	}
   	err = SerOpen (serial_lib, 0, 9600L);
   	if (err == serErrAlreadyOpen) {
      		SerClose (serial_lib);
   	}
   	if (err) {
      		FrmAlert (gpsSerialAlert);
      		return -1;
   	}
   	serial_port_open = true;
   	settings.baudRate = 9600L;
   	settings.flags = serSettingsFlagStopBits1 | serSettingsFlagBitsPerChar8;
   	settings.ctsTimeout = 0; // 60*sysTicksPerSecond;
   	err = SerSetSettings (serial_lib, &settings);
   	return err;
}


//
//	Trim trailing blanks from a waypoint ...
//

void Trim (char *txt)
{
   	int j;

   	for (j = StrLen (txt) - 1; j >= 0; --j) {
      		if (txt[j] == ' ' || txt[j] == '\0') {
	 		txt[j] = '\0';
      		}
      		else {
	 		break;
      		}
   	}
}


//
//	Send an 'ACK' message
//

void SendAck (char type, int ok)
{
   char ack[10];

   if (ok)
      ack[0] = Pid_Ack_Byte;
   else
      ack[0] = Pid_Nak_Byte;
   ack[1] = 2;
   ack[2] = type;
   ack[3] = 0;
   GarminSendPacket (ack);
}


//
//	Get a single, well-formed Garmin packet
//

int GarminGetPacket (UInt8 *buf, int *len, int *proto_err)
{
   UInt16	err;
   char		last = '\0';
   int		i, l;
   char 	checksum = 0;

   *proto_err = i  = 0;
   l = *len - 2;
   MemSet (buf, *len, 0);
   while (true) {
      	EvtResetAutoOffTimer();
      	SerSendWait (serial_lib, sysTicksPerSecond/2);
      	err = SerReceiveWait (serial_lib, 1, sysTicksPerSecond);
      	if (err == serErrTimeOut) {
	 		*len = i;
	 		return err;
      	}
      	if (!err) {
	 		SerReceive (serial_lib, buf+i, 1, 0, &err);
      	}
      	if (err) {
	 		SerClearErr (serial_lib);
	 		continue;
      	}
      	if (i == 0 && buf[0] != DLE) {
	 		continue;
      	}
      	if (i == 1 && buf[1] == DLE) {
	 		i = 0;
	 		continue;
      	}
      	if (buf[i] == DLE && last == DLE) {
	 		last = '\0';
	 		continue;
      	}
      	if (i > 0 && i < l-1) {
		 	checksum += buf[i];
      	}
      	if (i == 2) {
	 		if (buf[i] + 5 <= l) {
	    		l = buf[i] + 5;
	 		}
      	}
      	if (i == l) {		// && buf[i] == ETX || last == DLE) {
	 		*len = i - 5;
	 		buf[i+1] = checksum;
	 		if (buf[i] == ETX && last == DLE && checksum == 0)
	    		*proto_err = 0;
	 		else
	    		*proto_err = 1;
	 		return 0;
      	}
      	last = buf[i];
      	if (++i >= *len) {
	 	*len = i;
	 	*proto_err = 1;
	 	return 0;
      }
   }
   return 0;
}


//
//	Display latitude & longitude in a human-readable and
//	-editable format
//

char *GarminLongToPos (long pos)
{
   static char txt[20];
   int 		c;
   float 	dms;
   long 	sec100, sec, min, deg;

   if (pos < 0) {
      pos *= -1L;
   }
   dms = (float)pos/33.1402;
   sec100 = (long) dms % 100;
   dms /= 100;
   sec = (long) dms % 60;
   dms /= 60;
   min = (long) dms % 60;
   dms /= 60;
   deg = (long) dms;
   StrIToA (txt, deg);
   StrCat (txt, ":");
   c = StrLen (txt);
   StrIToA (txt+c, min);
   StrCat (txt, ":");
   c = StrLen (txt);
   StrIToA (txt+c, sec);
   StrCat (txt, ".");
   c = StrLen (txt);
   StrIToA (txt+c, sec100);
   return txt;
}


//
//	... and convert back again!
//
//	This is actually quite sexy.  It will take input in the form
//	dd:mm:ss.ss NSWE or dd:mm.mmm NSWE or dd.ddddd NSWE or even
//	directly in semicircles.  The *only* proviso is that an entry
//	in degrees must have at least on ':' or '.' in it, as this
//	is how I decide whether the value is in DMS or SC.
//

long GarminPosToLong (char *pos)
{
   	long 	factor = 11930465;
   	long 	result;
   	char 	*b, *c;
   	char 	d;

   	//	decide whether this is in DMS or semicircles
   	if (!StrChr (pos, '.') && !StrChr (pos, ':')) {
      		return StrAToI (pos);
   	}

   	//   allow ',' or '!' or 'I' or ' ' or '-' instead of ':'
   	for (b = pos; *b != '\0'; ++b) {
      		if (*b == ',' || *b == '!' || *b == 'i' || *b == 'I'
		    || *b == ' ' || *b == '-') {
	 		*b = ':';
      		}
   	}

   //	skip through the string, converting
   	b = pos;
   	result = 0;
   	while (b != NULL && *b != '\0') {
      		d = '\0';
      		c = StrChr (b, ':');
      		if (c == NULL)
	 		c = StrChr (b, '.');
      		if (c != NULL) {
	 		d = *c;
	 		*c = '\0';
      		}
      		result += (StrAToI (b) * factor);
      		if (c == NULL || d == '\0') {
	 		break;
      		}
      		b = c + 1;
      		if (d == ':') {		 // ':' means D->M->S
	 		factor /= 60;
	 		continue;
      		}
      		if (d == '.') {		 // '.' means fractions of whatever is current
         		while (*b >= '0' && *b <= '9') {
	    			factor /= 10;
	    			result += ((*b - '0') * factor);
	    			++b;
	 		}
      		}
      		break;			// else end of numbers
   	}
   	return result;
}


//
//	Callback routine for the progress dialogue box
//

Boolean garminProgressCallback (PrgCallbackData *data)
{
   char *c;
//   EventType e;

   if (data->textLen > 20) {
      StrCopy (data->textP, data->message);
      c = StrChr (data->textP, '~');
   }
   else {
      c = data->textP;
   }
   StrIToA (c, data->stage);
   *(data->textP + StrLen (data->textP)) = ' ';
   data->textChanged = true;
//   EvtGetEvent(&e, 10);
//   FrmDispatchEvent (&e);

   return true;
}

//
//	Get the current location from a GPS
//

int GarminGetPos (long *lat, long *lon)
{
   	int 	err;
   	UInt8	buf[BUFLEN];
   	char	txt[BUFLEN];
   	int 	len;
   	int 	i = 0;
   	double 	radians;
   	char 	swap[8];
	int		serial_port_already_open;

   	*lat = *lon = 0L;

	serial_port_already_open = serial_port_open;
	if ( ! serial_port_open ) {
    	if ((err = OpenSerialPort ())) {
        	serial_port_open = false;
        	return err;
		}
	}

   	// Garmin "send position" command
   	buf[0] = Pid_Command_Data;
   	buf[1] = 2;
   	buf[2] = Cmnd_Transfer_Posn;
   	buf[3] = 0;
   	SerReceiveFlush (serial_lib, 0);
   	GarminSendPacket (buf);
   	while (true) {
      		len = BUFLEN;
      		if (GarminGetPacket (buf, &len, &err)) {
        		SerClose (serial_lib);
        		serial_port_open = false;
			StrIToA( txt, err );
	 		FrmCustomAlert (serialAlert, "Receive Position", txt, " ");
	 		return 1;
      		}
      		if (buf[1] != Pid_Ack_Byte && buf[1] != Pid_Nak_Byte) {
	 		SendAck (buf[1], true);
      		}
      		if (buf[1] == Pid_Position_Data) {
//	 dumpPacket (txt, buf, len, err);
	 		for (i = 0; i < 8; ++i) {
	    			swap[i] = buf[10 - i];
	    			MemMove (&radians, swap, sizeof (radians));
	    			*lat = (long) (radians * 683565275.576);
	 		}
	 		for (i = 0; i < 8; ++i) {
	    			swap[i] = buf[18 - i];
	    			MemMove (&radians, swap, sizeof (radians));
	    			*lon = (long) (radians * 683565275.576);
	 		}
        		SerClose (serial_lib);
        		serial_port_open = false;
//	 	FrmCustomAlert (serialAlert, txt, " ", " ");
	 		StrIToA (txt, *lon);
	 		StrIToA (buf, *lat);
	 		StrCat (txt, ":");
	 		StrCat (txt, buf);
			#ifdef DEBUG
	 			FrmCustomAlert (gpsBreak1, txt, " ", " ");
			#endif
	 		return 0;
      		}
   	}
}


//
//	dump packet contents into a buffer
//

void dumpPacket (char *txt, char *buf, int len, int err)
{
   	char tmp[16];
   	int i;

   	StrIToA (txt, len);
   	StrCat (txt, ":");
   	StrIToA (tmp, err);
   	StrCat (txt, tmp);
   	StrCat (txt, " 0x");
//   	for (i = 0; i < 30; ++i) {
	if ( len > 60 ) {
		len = 60;
	}
   	for (i = 0; i < len; ++i) {
      		if (i > 0) {
	 		StrCat (txt, " ");
      		}
      		StrIToA (tmp, buf[i]);
      		StrCat (txt, tmp );
//		StrCat(txt, buf[i] );
   	}
}


//
//	send raw buffered data (either route or track)
//

int GarminSendRaw (UInt16 command)
{
   	int 	num_wpts;
   	MemHandle rec;
   	DmOpenRef raw_db = NULL;
   	char *raw;
   	int 	rec_num, retries;
   	UInt8	buf[BUFLEN];
   	UInt8	ack[BUFLEN];
   	int 	len;
	int	err;
	int	i = 0;
   	char 	cmd = 0;
   	char 	msg[64];
   	ProgressPtr progress = NULL;
   	int 	serial_port_already_open;

   	serial_port_already_open = serial_port_open;
   	switch (command) {
      	case menuXferTrackOut:
	 	raw_db = DmOpenDatabaseByTypeCreator (DB_TRACK, DB_APPID, dmModeReadWrite);
	 	cmd = Cmnd_Transfer_Trk;
	 	break;
      	case menuXferRouteOut:
	 	raw_db = DmOpenDatabaseByTypeCreator (DB_ROUTE, DB_APPID, dmModeReadWrite);
	 	cmd = Cmnd_Transfer_Rte;
	 	break;
   	}
   	if (! raw_db) {
      		return 0;
   	}
   	num_wpts = DmNumRecords (raw_db);
   	if (! serial_port_open) {
      		if ((err = OpenSerialPort ())) {
	 		return err;
      		}
   	}
   	buf[0] = Pid_Records;
   	buf[1] = 2;
	endian_convert( &num_wpts, sizeof(int) );
	MemMove( & buf[2], & num_wpts, 2 );
   	SerReceiveFlush (serial_lib, 0);
    	if ( romVersion >= 0x03000000) {
       		StrCopy (msg, "Sending ~   / ");
       		StrIToA (msg + StrLen (msg), num_wpts);
       		if (serial_port_already_open) {
          		progress = PrgStartDialogV31 ("Send to Host", garminProgressCallback);
       		}
       		else {
          		progress = PrgStartDialogV31 ("Send to GPS", garminProgressCallback);
       		}
   	}
   	GarminSendPacket (buf);		// await 'n' records
   	len = BUFLEN;
   	GarminGetPacket (ack, &len, &err);
   	if (ack[1] != Pid_Ack_Byte) {
	 	FrmCustomAlert (serialAlert, "Transmit packet", " ", " ");
   	}

   rec_num = 0;
   while ( rec_num < num_wpts) {
      rec = DmQueryRecord (raw_db, rec_num++);
      if (! rec) {
	 	break;
      }
      raw = MemHandleLock (rec);
      MemMove (buf, raw, raw[1] + 2);
      MemHandleUnlock (rec);
      SerReceiveFlush (serial_lib, 0);
      if ( romVersion >= 0x03000000) {
         PrgUpdateDialog (progress, 0, ++i, msg, true);
      }
      for (retries = 0; retries < 5; ++retries) {
	 	GarminSendPacket (buf);
	 	len = BUFLEN;
	 	GarminGetPacket (ack, &len, &err);
	 	if (ack[1] == Pid_Ack_Byte)
	    	break;
      }
//      if (ack[1] != Pid_Ack_Byte) {
//	 FrmCustomAlert (serialAlert, "Transmit packet", " ", " ");
//      }
   }
   buf[0] = Pid_Xfer_Cmplt;
   buf[1] = 2;
   buf[2] = cmd;
   buf[3] = 0;
   SerReceiveFlush (serial_lib, 0);
   GarminSendPacket (buf);
   len = BUFLEN;
   GarminGetPacket (ack, &len, &err);
   if ( romVersion >= 0x03000000) {
      PrgStopDialog (progress, true);
   }
   if (! serial_port_already_open) {
      SerClose (serial_lib);
      serial_port_open = false;
   }
   if (raw_db) {
      DmCloseDatabase (raw_db);
   }
   return(0);
}

//
//	send track data
//

int GarminSendTrack (void)
{
   	static long lat, lon, t;
   	short dlat, dlon;
   	int num_wpts, num_recs, flag;
   	MemHandle rec;
   	DmOpenRef trk_db = NULL;
   	unsigned char *ptr;
   	int rec_num, retries;
   	UInt8	buf[BUFLEN];
   	UInt8	ack[BUFLEN];
   	int len, err, i = 0, j;
   	char 	cmd = 0;
   	char 	msg[64];
   	ProgressPtr 	progress = NULL;
   	int 	    	serial_port_already_open;
   	pilot_trk * 	trk;
   	garmin_trk 	gtrk;

   	serial_port_already_open = serial_port_open;
	trk_db = DmOpenDatabaseByTypeCreator (DB_TRACK, DB_APPID, dmModeReadWrite);
	cmd = Cmnd_Transfer_Trk;
   	if (! trk_db) {
      		return 0;
   	}
   	num_recs = DmNumRecords (trk_db);
   	num_wpts = 0;
   	for (rec_num = 0; rec_num < num_recs; ++rec_num) {
      	rec = DmQueryRecord (trk_db, rec_num);
      	if (! rec) {
	 	DmCloseDatabase (trk_db);
	 	return(0);
      	}
      	if (MemHandleSize (rec) < sizeof (pilot_trk) + 12) {
	 	continue; 		//  old style track record
      	}
      	trk = MemHandleLock (rec);
      	num_wpts += trk -> nb_point;
      	MemHandleUnlock (rec);
   }
   if (! serial_port_open) {
      if ((err = OpenSerialPort ())) {
	 	return err;
      }
   }
   buf[0] = Pid_Records;
   buf[1] = 2;
   MemMove( & buf[2], & num_wpts, sizeof(int) );
   endian_convert( & buf[2], sizeof(int) );
   SerReceiveFlush (serial_lib, 0);
    if ( romVersion >= 0x03000000) {
       StrCopy (msg, "Sending ~   / ");
       StrIToA (msg + StrLen (msg), num_wpts);
       if (serial_port_already_open) {
          progress = PrgStartDialogV31 ("Send to Host", garminProgressCallback);
       }
       else {
          progress = PrgStartDialogV31 ("Send to GPS", garminProgressCallback);
       }
   }
   GarminSendPacket (buf);
   len = BUFLEN;
   GarminGetPacket (ack, &len, &err);

   gtrk.code = 0x22;
   gtrk.len = 13;
   rec_num = 0;
   while ( rec_num < num_recs) {
      rec = DmQueryRecord (trk_db, rec_num++);
      if (! rec) {
	 	break;
      }
      if (MemHandleSize (rec) < sizeof (pilot_trk) + 12) {
	 	continue; 		//  old style track record
      }
      	trk = MemHandleLock (rec);
      	ptr = trk -> data;
      	for(j = 0; j < trk -> nb_point; j++) {
        	flag = *ptr++;
	 	if (flag > 253) {
	    		gtrk.flag = flag - 254;
            		MemMove(&lat, ptr,   4);
            		MemMove(&lon, ptr+4, 4);
            		MemMove(&t,   ptr+8, 4);
	    		ptr += 12;
	 	} else {
	    		gtrk.flag = 0;
	    		if ( flag > 127 ) {
	       			t += flag - 128;
	       			MemMove(&dlat, ptr, 2);
	       			MemMove(&dlon, ptr + 2, 2);
	       			ptr +=4;
	    		} else {
	       			t += flag;
	       			dlat = *((char *) ptr++);
	       			dlon = *((char *) ptr++);
	    		}
	    		lat += ((long) dlat << 6);
	    		lon += ((long) dlon << 6);
	 	}
	 	endian_convert( &lat, sizeof(long) );
		MemMove( & gtrk.lat, & lat, sizeof(long) );
	 	endian_convert( &lon, sizeof(long) );
		MemMove( & gtrk.lon, & lon, sizeof(long) );
		endian_convert( &t,   sizeof(long) );
		MemMove( & gtrk.time, & t, sizeof(long) );

	 	SerReceiveFlush (serial_lib, 0);
        if ( romVersion >= 0x03000000) {
            PrgUpdateDialog (progress, 0, ++i, msg, true);
         }
	 	MemMove(buf,(UInt8 *) &gtrk, 15);
	 	for (retries = 0; retries < 5; ++retries) {
	    		GarminSendPacket (buf);
	    		len = BUFLEN;
	    		GarminGetPacket (ack, &len, &err);
	    		if (ack[1] == Pid_Ack_Byte)
	       			break;
         	}
//       	if (ack[1] != Pid_Ack_Byte) {
//	    		FrmCustomAlert (serialAlert, "Transmit packet", " ", " ");
//       	}

      }
      MemHandleUnlock (rec);
   }
   buf[0] = Pid_Xfer_Cmplt;
   buf[1] = 2;
   buf[2] = cmd;
   buf[3] = 0;
   SerReceiveFlush (serial_lib, 0);
   GarminSendPacket (buf);
   len = BUFLEN;
   GarminGetPacket (ack, &len, &err);
   if ( romVersion >= 0x03000000) {
      PrgStopDialog (progress, true);
   }
   if (! serial_port_already_open) {
      SerClose (serial_lib);
      serial_port_open = false;
   }
   if (trk_db) {
      DmCloseDatabase (trk_db);
   }
   return(0);
}


//
//	Open a fresh raw database of whatever type is requested
//

DmOpenRef openFreshDB (UInt32 type)
{
   DmSearchStateType state;
   UInt16 card;
   LocalID dbID;
   char *name = TRK_DB_NAME;

   while (! DmGetNextDatabaseByTypeCreator (true, &state, type, DB_APPID,
					    false, &card, &dbID)) {
      if (DmDeleteDatabase (card, dbID)) {
	 break;
      }
   }
   if (type == DB_ROUTE) {
      name = ROUTE_DB_NAME;
   }
   DmCreateDatabase (0, name, DB_APPID, type, false);
   return DmOpenDatabaseByTypeCreator (type, DB_APPID, dmModeReadWrite);
}


//
//	Open a raw database of whatever type is requested
//

DmOpenRef openAppendDB (UInt32 type)
{
   DmSearchStateType state;
   UInt16 card;
   LocalID dbID;
   char *name = TRK_DB_NAME;

   if (type == DB_ROUTE) {
      name = ROUTE_DB_NAME;
   }
   if (DmGetNextDatabaseByTypeCreator (true, &state, type, DB_APPID,
					    false, &card, &dbID)) {
      DmCreateDatabase (0, name, DB_APPID, type, false);
   }
   return DmOpenDatabaseByTypeCreator (type, DB_APPID, dmModeReadWrite);
}


//
//	Trash any track logs
//

void GarminClearTrackLog (void)
{
   DmSearchStateType state;
   UInt16 card;
   LocalID dbID;

   while (! DmGetNextDatabaseByTypeCreator (true, &state, DB_TRACK, DB_APPID,
					    false, &card, &dbID)) {
      if (DmDeleteDatabase (card, dbID)) {
	 	break;
      }
   }
}

//
//	Initialize a new track ptr
//

static pilot_trk * init_trk(void) {
   pilot_trk * trk;

   if( (trk = MemPtrNew (sizeof (pilot_trk) + TRK_SIZE)) ) {
      trk -> nb_point = 0;
      trk -> size = 0;
      trk -> distance = 0;
      trk -> alt = 0;
   }
   return(trk);
}

//
//	add a point to the current track
//

static void add_trkpoint(pilot_trk * trk, garmin_trk * gtrk) {
   static long lat0, lon0, t0;
   unsigned char * ptr;
   long lat, lon, t, flag, dlat, dlon, dt;
   short s;

   if(trk -> size + 13 > TRK_SIZE) {
      return;
   }
	MemMove( & lat, & gtrk->lat, sizeof(long) );
	endian_convert( &lat, sizeof(long) );
	MemMove( & lon, & gtrk->lon, sizeof(long) );
	endian_convert( &lon, sizeof(long) );
	MemMove( & t, & gtrk->time, sizeof(long) );
	endian_convert( &t,   sizeof(long) );
   /*xxx*   Altitude */

   flag = gtrk -> flag;

   if ( trk -> nb_point == 0 ) {
      trk -> lat_min = lat;
      trk -> lat_max = lat;
      trk -> lon_min = lon;
      trk -> lon_max = lon;
      flag = 1;
   } else {
      if (lat < trk -> lat_min ) trk -> lat_min = lat;
      if (lat > trk -> lat_max ) trk -> lat_max = lat;
      if (lon < trk -> lon_min ) trk -> lon_min = lon;
      if (lon > trk -> lon_max ) trk -> lon_max = lon;
   }
   ptr = trk -> data + trk -> size;
   if(flag) {
      *ptr++ = 255;
      MemMove(ptr, &lat, 4);
      MemMove(ptr+4, &lon, 4);
      MemMove(ptr+8, &t, 4);
      trk -> size += 13;
   } else {
      dlat=(lat - lat0)>>6;
      dlon=(lon - lon0)>>6;
      dt=t - t0;
      if (abs(dlat) < 127 && abs(dlon) < 127 && dt < 126) {
        	*ptr++ = dt;
	 	*ptr++ = dlat;
	 	*ptr++ = dlon;
        	trk -> size += 3;
      } else if (abs(dlat) < 32767 && abs(dlon) < 32767 && dt < 126 ) {
        	*ptr++ = dt + 128;
	 	s = dlat;
        	MemMove(ptr, &s, 2);
	 	s = dlon;
        	MemMove(ptr+2, &s, 2);
        	trk -> size += 5;
      } else {
        	*ptr++ = 254;
        	MemMove(ptr, &lat, 4);
        	MemMove(ptr+4, &lon, 4);
        	MemMove(ptr+8, &t, 4);
        	trk -> size += 13;
      }
   }
   lat0 = lat;
   lon0 = lon;
   t0 = t;
   trk -> nb_point ++;
}


//
//	save the current track in tracks database
//

static void save_trk(pilot_trk * trk) {
   DmOpenRef trk_db;
   long record_size;
   MemPtr rec;
   MemHandle new_rec;
   UInt16 rec_num;

   record_size = trk -> size + sizeof (pilot_trk);
   if ( (trk_db = openAppendDB (DB_TRACK)) ) {
      rec_num = DmNumRecords (trk_db) + 1;
      new_rec = DmNewRecord (trk_db, &rec_num, record_size);
      rec = MemHandleLock (new_rec);
      DmWrite (rec, 0, trk, record_size);
      MemHandleUnlock (new_rec);
      DmCloseDatabase (trk_db);
   }
}


