/*
 PalGar - use a PALm to up-, download, edit waypoints from a GARmin GPS
 Copyright (C) 1999 Paul Nash
 Copyright (C) 2000 Ralf Meyer <ranf@users.sourceforge.net>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

//
//	Global variables, constants and data structures for the GPS editor
//
//	Paul Nash
//	paul@frcs.alt.za

//
//	$Header: /cvs/PalGar/globals.h,v 1.18 2000/11/26 15:01:53 ranf Exp $
//
//	$Log: globals.h,v $
//	Revision 1.18  2000/11/26 15:01:53  ranf
//	- edit eTrex waypoint now with altitude
//	
//	Revision 1.17  2000/11/12 16:28:19  ranf
//	- eTrex: display symbol (in Hex for now)
//
//	Revision 1.16  2000/11/07 19:42:42  ranf
//	- eTrex WPT download at least works
//
//	Revision 1.15  2000/09/19 20:19:03  ranf
//	- fixed probs with preferences saving
//
//	Revision 1.14  2000/08/24 20:48:22  ranf
//	- global struct to keep Product_Data
//	- global struct to keep Protocol_Array
//	- removed anything related to the callback.h stuff
//	- version number up to 1.7.0
//
//	Revision 1.13  2000/07/24 20:45:07  ranf
//	- bugfix: Get date/time from GPS
//
//	Revision 1.12  2000/07/19 19:59:40  ranf
//	Get ID from GPS finally works
//
//	Revision 1.11  2000/07/12 18:35:21  ranf
//	- bugfix: when merging on a Palm OS >= 3 it only took
//	every second record
//	- first attempt to add eTrex waypoint up- /download
//	(needs testing)
//
//	Revision 1.10  2000/07/10 19:14:01  ranf
//	Edit Wpt symbol. No bitmap yet.
//
//	Revision 1.9  2000/07/06 19:02:15  ranf
//	Options dialogue for setting the time zone
//
//	Revision 1.8  2000/07/02 13:18:30  ranf
//	- get date/time from the Palm (PalGar has to be in host mode).
//		Be careful with that: there's no local time to UTC conversion yet.
//	- turn off the Palm (again host mode required).
//
//	Revision 1.7  2000/06/26 17:55:16  ranf
//	Checksum errors when asked for GPS type
//
//	Revision 1.3  2000/03/27 18:24:31  giga
//	OS 35 compiles now
//
//	Revision 1.2  2000/03/27 18:03:29  giga
//	OS 35 Upgrader Perl ran
//
//	Revision 1.1.1.1  2000/03/23 18:18:40  giga
//	Initial import
//
//	Revision 1.23  1999/08/30 06:35:32  paul
//	Added Jaque Andre's patches
//
//	Revision 1.22  1999/06/13 17:26:13  paul
//	Changed lat & long edit to use N/S E/W list (button) input, rather
//	than text.  This allows the ':' separator to use ' ' or '-'.
//
//	Revision 1.21  1999/06/12 14:38:39  paul
//	changed track log to allow cumulative logging.  Not tested yet
//
//	Revision 1.20  1999/05/12 08:48:05  paul
//	Ready for release as Gypsy 1.4.  Major changes include:
//	*  Rudimentary track/route buffer to transfer data between GPS
//	   and PC (a'la gps2p2pc)
//	*  Protocol fix for DLE escaping bug
//	*  Code changes from <andre@drfmc.ceng.cea.fr> to be able to
//	   compile a PalmOS2 version
//	*  Garmin protocol tries a bit harder to recover from comms
//	   errors (but is still *very* dumb)
//
//


//
//	Clever hack to get global variables defined correctly
//

#if ! defined (EXTERN)
#   define EXTERN extern
#endif // EXTERN


//
//	Some useful constants
//

#define DB_TYPE			'Data'
#define DB_APPID		'fRCS'
#define DB_NAME			"GPS Waypoint DB"
#define DB_ROUTE		'RteD'
#define DB_TRACK		'TrkD'

#define DLE			16
#define ETX			3

#define GARMIN_NAME_LEN		6
#define GARMIN_DESC_LEN		40

#define HOST_TRACK		0
#define HOST_ROUTE		1
#define HOST_WAYPOINT		2
#define GPS_TRACK		3
#define GPS_ROUTE		4
#define GPS_WAYPOINT		5

//
//	Data structures
//

typedef struct {
   	char 	name[GARMIN_NAME_LEN + 1];
   	char 	descr[GARMIN_DESC_LEN + 1];
   	UInt16 	symbol;
   	int 	display;
   	long 	latitude;
   	long 	longitude;
   	int 	flagged;
   	int 	changed;
   	int 	record_num;

	float 	alt;
}	waypoint ;


//
//	function prototypes in "garmin.c"
//

int GarminGetAll (int mode, UInt16 num_recs);
int GarminGetID ();
int GarminGetDateTime ();
char *GarminLongToPos (long pos);
long GarminPosToLong (char *pos);
int GarminSend (UInt16 command);
int GarminGetPos (long *lat, long *lon);
int OpenSerialPort (void);
int GarminGetPacket (UInt8 *buf, int *len, int *proto_err);
void SendAck (char type, int ack_or_nak);
void GarminSendPacket (UInt8 *buf);
int GarminSendRaw (UInt16 command);
int GarminSendTrack (void);
void GarminClearTrackLog (void);

//	... in "gps.c"

void setField (UInt16 fieldname, char *contents);
void getField (UInt16 fieldname, char *contents);
char *currentDBName (void);

//	... and "file.c"

Boolean HandleFileListEvent (EventPtr e);
Boolean HandleFileNewEvent (EventPtr e);

//	... and "host.c"

Boolean HandleHostEvent (EventPtr e);
Boolean HandleFileOptionsEvent (EventPtr e);

//	... and "endian.c"

void endian_convert( void * data, int len );

//
//	The odd global variable ...
//

EXTERN DmOpenRef 	waypoint_db;
EXTERN int 		num_records;
EXTERN UInt16 		current_record;
EXTERN int 		table_top;
EXTERN MemHandle 	wpt_copy_buf;
EXTERN UInt16 		serial_lib;
EXTERN int 		serial_port_open;
EXTERN UInt16 		newfile_camefrom;
EXTERN int 		file_merge;

