/*
 PalGar - use a PALm to up-, download, edit waypoints from a GARmin GPS
 Copyright (C) 1999 Paul Nash
 Copyright (C) 2000 Ralf Meyer <ranf@users.sourceforge.net>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

//	First attempt at a GPS waypoint editor
//
//	Paul Nash
//	paul@frcs.alt.za

//
//	$Header: /cvs/PalGar/gps.c,v 1.24 2000/11/26 15:01:53 ranf Exp $
//
//	$Log: gps.c,v $
//	Revision 1.24  2000/11/26 15:01:53  ranf
//	- edit eTrex waypoint now with altitude
//	
//	Revision 1.23  2000/11/25 14:18:56  ranf
//	- check for D108
//
//	Revision 1.22  2000/11/19 18:57:01  ranf
//	- WPT should also work with the eTrex Summit
//
//	Revision 1.21  2000/11/12 16:28:19  ranf
//	- eTrex: display symbol (in Hex for now)
//
//	Revision 1.20  2000/11/11 14:19:20  ranf
//	- eTrex has another WPT edit form
//	+ now I need to find out how to handle float's
//
//	Revision 1.19  2000/11/09 21:21:20  ranf
//	- eTrex: added Altitude to WPT
//
//	Revision 1.18  2000/11/07 19:42:42  ranf
//	- eTrex WPT download at least works
//
//	Revision 1.17  2000/09/19 20:19:03  ranf
//	- fixed probs with preferences saving
//
//	Revision 1.16  2000/08/27 19:33:03  ranf
//	- Makefile: Target veryclean renamed to ``distclean''
//	- after successfully talking to a GPS (Get ID) the prefs screen
//		shows ID, version, product description
//	- reordered Menu
//
//	Revision 1.15  2000/08/24 20:48:22  ranf
//	- global struct to keep Product_Data
//	- global struct to keep Protocol_Array
//	- removed anything related to the callback.h stuff
//	- version number up to 1.7.0
//
//	Revision 1.14  2000/07/24 20:45:07  ranf
//	- bugfix: Get date/time from GPS
//
//	Revision 1.13  2000/07/19 19:59:40  ranf
//	Get ID from GPS finally works
//
//	Revision 1.12  2000/07/12 18:35:21  ranf
//	- bugfix: when merging on a Palm OS >= 3 it only took
//	every second record
//	- first attempt to add eTrex waypoint up- /download
//	(needs testing)
//
//	Revision 1.11  2000/07/10 19:14:01  ranf
//	Edit Wpt symbol. No bitmap yet.
//
//	Revision 1.10  2000/07/06 19:02:15  ranf
//	Options dialogue for setting the time zone
//
//	Revision 1.9  2000/07/02 21:52:45  ranf
//	changed preferences to include the time offset to UTC
//
//	Revision 1.8  2000/07/02 13:18:30  ranf
//	- get date/time from the Palm (PalGar has to be in host mode).
//		Be careful with that: there's no local time to UTC conversion yet.
//	- turn off the Palm (again host mode required).
//
//	Revision 1.7  2000/06/26 17:55:16  ranf
//	Checksum errors when asked for GPS type
//
//	Revision 1.6  2000/04/24 14:20:30  ranf
//	Added small icon
//		(not very nice. But now we have one)
//
//	Revision 1.5  2000/03/27 20:17:40  giga
//	html
//
//	Revision 1.4  2000/03/27 18:24:31  giga
//	OS 35 compiles now
//
//	Revision 1.3  2000/03/27 18:03:29  giga
//	OS 35 Upgrader Perl ran
//
//	Revision 1.2  2000/03/23 18:40:14  giga
//	OS3,5
//
//	Revision 1.1.1.1  2000/03/23 18:18:40  giga
//	Initial import
//
//	Revision 1.53  1999/08/30 06:35:32  paul
//	Added Jaque Andre's patches
//
//	Revision 1.52  1999/06/13 17:26:13  paul
//	Changed lat & long edit to use N/S E/W list (button) input, rather
//	than text.  This allows the ':' separator to use ' ' or '-'.
//
//	Revision 1.51  1999/06/12 17:30:52  paul
//	Removed "track to GPS" from menu; confirm clear track log
//
//	Revision 1.50  1999/06/12 14:38:39  paul
//	changed track log to allow cumulative logging.  Not tested yet
//
//	Revision 1.49  1999/05/12 08:48:05  paul
//	Ready for release as Gypsy 1.4.  Major changes include:
//	*  Rudimentary track/route buffer to transfer data between GPS
//	   and PC (a'la gps2p2pc)
//	*  Protocol fix for DLE escaping bug
//	*  Code changes from <andre@drfmc.ceng.cea.fr> to be able to
//	   compile a PalmOS2 version
//	*  Garmin protocol tries a bit harder to recover from comms
//	   errors (but is still *very* dumb)
//
//



#include <PalmOS.h>
#include "gps.h"
#include <System/SerialMgrOld.h>

//
//	Get in global vars, etc, plus reserve space for them
//

#define EXTERN
#include "globals.h"
#include "GarminDef.h"

// #include <CoreTraps.h>
// #include <System/FloatMgr.h>

typedef struct {
   	char	name[32];
   	UInt16	card;
}	prefs;

typedef struct {
	prefs	p;
	Int8	timeoffset;
	UInt16	GPS_type;
}	prefs162;

typedef struct {
   	char	name[32];
   	UInt16	card;
	Int8	timeoffset;
	Product_Data_Type	A000;
	Protocol_Array_Type	A001;
}	prefs170;

//
//	Local function prototypes
//

static int StartApp (void);
static void StopApp (void);
static void ProcessLoop (void);
static Boolean HandleWaypointListEvent (EventPtr e);
static Boolean HandleWaypointEditEvent (EventPtr e);
static void DisplayTable ();
static Err descrTableLoadName (MemPtr t, UInt16 r, UInt16 c, Boolean e,
                        MemHandle *d, UInt16 *o, UInt16 *s, FieldPtr f);
static Boolean descrTableSaveName (MemPtr t, UInt16 r, UInt16 c);
static void descrTableDrawDescr (MemPtr tbl, UInt16 row, UInt16 column,
                          RectanglePtr bounds);
static void uppercase (char *c, int strip_blanks);
static void deleteWaypoints (int cmd);
static void changeFlags (int cmd);
static int CompareWaypoints (waypoint *wpt1, waypoint *wpt2,
                      int unused, SortRecordInfoPtr un1,
		      SortRecordInfoPtr un2, MemHandle appInfo);
static int EditButtons (UInt16 button);
static int HandleEditControls (UInt16 menu_item);
static int HandleEditOpen ();
static int HandleListMenu (UInt16 menu_item);
static int HandleListButton (UInt16 button);


//
//	and internal globals

// static char *rcs_id = "$Id: gps.c,v 1.24 2000/11/26 15:01:53 ranf Exp $";

extern char *version_info;
extern char *compile_date;
UInt32 		romVersion;
Int8		timeoffset;
Product_Data_Type	A000;
Protocol_Array_Type	A001;

//
//	main routine, does very little ...
//

UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
   	int err;

   	if (cmd == sysAppLaunchCmdNormalLaunch) {
      		err = StartApp ();
      		if (err) {
	 			return err;
      		}
      		FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
      		ProcessLoop ();
      		StopApp ();
   	}
   	return 0;
}

static void InitProto( UInt16 typ)
{
	MemSet(& A000, sizeof(Product_Data_Type), 0);
	MemSet(& A001, sizeof(Protocol_Array_Type), 0);
	timeoffset	= 2;                    // currently it's DST here in Germany

	switch ( typ ) {
	case 130: // eTrex
		A000.product_ID	= 130;
		A000.software_version	= 110;
		StrCopy( A000.product_description,	"emulated eTrex" );
		A000.product_additional[0] = '\0';

		A001[0].tag	= 'P';	A001[0].data	=   0;
		A001[1].tag	= 'L';	A001[1].data	=   1;
		A001[2].tag	= 'A';	A001[2].data	=  10;
		A001[3].tag	= 'A';	A001[3].data	= 100;
		A001[4].tag	= 'D';	A001[4].data	= 108;
		A001[5].tag	= 'A';	A001[5].data	= 200;
		A001[6].tag	= 'D';	A001[6].data	= 201;
		A001[7].tag	= 'D';	A001[7].data	= 103;
		A001[8].tag	= 'A';	A001[8].data	= 300;
		A001[9].tag	= 'D';	A001[9].data	= 301;
		break;
	default:
		A000.product_ID	= 97;
		A000.software_version	= 170;
		StrCopy( A000.product_description,	"PalGar" );
		A000.product_additional[0] = '\0';

		A001[0].tag	= 'P';	A001[0].data	=   0;
		A001[1].tag	= 'L';	A001[1].data	=   1;
		A001[2].tag	= 'A';	A001[2].data	=  10;
		A001[3].tag	= 'A';	A001[3].data	= 100;
		A001[4].tag	= 'D';	A001[4].data	= 103;
		A001[5].tag	= 'A';	A001[5].data	= 200;
		A001[6].tag	= 'D';	A001[6].data	= 201;
		A001[7].tag	= 'D';	A001[7].data	= 103;
		A001[8].tag	= 'A';	A001[8].data	= 300;
		A001[9].tag	= 'D';	A001[9].data	= 300;
		break;
	}

}

//
//	Load database, set the initial form, etc
//

int StartApp (void)
{
   	int 		err = 0;
   	UInt16 		i = 0;
   	prefs 		p;
   	prefs170	p170;
   	UInt32 		creator, type;
   	LocalID 	id;
   	int	 	lOk = false;

   	FrmGotoForm (gpsWptListFrm);
   	table_top = 0;
   	wpt_copy_buf = NULL;
   	if (SysLibFind ("Serial Library", &serial_lib)) {
      		serial_lib = -1;
      		FrmAlert (gpsSerialAlert);
   	}
   	serial_port_open = false;

   	waypoint_db = NULL;
	InitProto( 130 );
   	PrefGetAppPreferences (DB_APPID, 0, NULL, &i, true);
	if ( (i = sizeof(prefs170)) ) {
		lOk = PrefGetAppPreferences (DB_APPID, 0, &p170, &i, true)
				!= noPreferenceFound;
		if ( lOk ) {
			timeoffset	= p170.timeoffset;
			A000		= p170.A000;
			MemMove( & p170.A001, & A001, sizeof(Protocol_Array_Type) );
		}
   	}

   	if (lOk) {
      		if (p.name[0] != '\0') {
	 		id = DmFindDatabase (p.card, p.name);
	 		if (id) {
	    			DmDatabaseInfo (p.card, id, NULL, NULL, NULL, NULL, NULL,
			    		NULL, NULL, NULL, NULL, &type, &creator);
	    			if (type == DB_TYPE && creator == DB_APPID) {
	       				waypoint_db = DmOpenDatabase (p.card, id, dmModeReadWrite);
	    			}
	 		}
      		}
   	}
   	if (!waypoint_db) {
      	waypoint_db = DmOpenDatabaseByTypeCreator (DB_TYPE, DB_APPID,
	  	 	dmModeReadWrite);
   	}
   	if (!waypoint_db) {
      		err = DmCreateDatabase (0, DB_NAME, DB_APPID, DB_TYPE, false);
      		if (!err) {
	 		waypoint_db = DmOpenDatabaseByTypeCreator (DB_TYPE, DB_APPID,
		    	dmModeReadWrite);
      		}
   	}
   	num_records = DmNumRecords (waypoint_db);

   	return err;
}


//
//	Shut down and go home
//

void StopApp (void)
{
   	prefs170	p170;


   	if (waypoint_db) {
      		DmOpenDatabaseInfo (waypoint_db, NULL, NULL, NULL, &p170.card, NULL);
      		StrCopy (p170.name, currentDBName() );
      		//CopyAndFill (p170.name, currentDBName(), sizeof(p170.name) );
	  	p170.timeoffset	= timeoffset;
	  	p170.A000		= A000;
	  	MemMove( & A001, & p170.A001, sizeof(Protocol_Array_Type) );
      		DmCloseDatabase (waypoint_db);
      		PrefSetAppPreferences (DB_APPID, 0, 1, &p170, sizeof (prefs170), true);
      		waypoint_db = NULL;
   	}
   	if (serial_port_open) {
      		SerClose (serial_lib);
   	}

}


//
//	Look for "load form" events, otherwise punt to form handlers
//

void ProcessLoop (void)
{
   	UInt16 		err = 0;
   	UInt16 		formID;
   	FormPtr 	form;
   	EventType 	e;
   	Int32 		timeout = evtWaitForever;

   	do {
      		EvtGetEvent(&e, timeout);
      		if (SysHandleEvent(&e))  {
	 		continue;
      		}
      		if (MenuHandleEvent((void *)0, &e, &err)) {
	 		continue;
      		}

      		if (e.eType == frmLoadEvent) {
	 		formID = e.data.frmLoad.formID;
	 		form = FrmInitForm (formID);
	 		FrmSetActiveForm (form);
	 		timeout = evtWaitForever;
			switch (formID) {
			case gpsWptListFrm:
	       			FrmSetEventHandler (form, HandleWaypointListEvent);
	       			break;
			case gpsWptEditFrm:
	       			FrmSetEventHandler (form, HandleWaypointEditEvent);
	       			break;
			case gpsWpt_D108_EditFrm:
	       			FrmSetEventHandler (form, HandleWaypointEditEvent);
	       			break;
			case gpsFileOpenFrm:
			case gpsFileMergeFrm:
			case gpsFileUtilFrm:
	       			FrmSetEventHandler (form, HandleFileListEvent);
	       			break;
			case gpsFileNewFrm:
	       			FrmSetEventHandler (form, HandleFileNewEvent);
	       			break;
			case garminHostFrm:
	       			timeout = 0L;
	       			FrmSetEventHandler (form, HandleHostEvent);
		   		break;
			case gpsOptions:
	       			FrmSetEventHandler (form, HandleFileOptionsEvent);
		   		break;
	 		}
      		}
      		FrmDispatchEvent (&e);
   	} while (e.eType != appStopEvent);
}


//
//	Waypoint list:
//
//	Allow the user to select a waypoint, then "EDIT"; or
//	Allow the user to go to "XFER"; or
//	Allow the user to create a "NEW" entry (special case of edit)
//

Boolean HandleWaypointListEvent (EventPtr e)
{
   	int handled = false;
   	FormPtr form;
   	TablePtr table;
   	UInt16 row;
   	waypoint *wpt, *new_wpt;
   	MemHandle rec_handle;

   	form = FrmGetActiveForm ();
   	switch (e->eType) {
   		case frmOpenEvent:
	 		setField (dbName, currentDBName ());
	 		FrmDrawForm (form);
	 		DisplayTable ();
	 		current_record = -1;
	 		handled = true;
	 		break;
    	case menuEvent:
         	handled = HandleListMenu (e->data.menu.itemID);
	 		break;
      	case tblSelectEvent:
	 		table = FrmGetObjectPtr( form, FrmGetObjectIndex( form, wptTable ));
	 		row = e->data.tblSelect.row;
	 		current_record = TblGetRowID (table, row);
	 		new_wpt = MemPtrNew (sizeof (waypoint));
	 		rec_handle = DmGetRecord (waypoint_db, current_record);
	 		wpt = MemHandleLock (rec_handle);
	 		MemMove (new_wpt, wpt, sizeof (waypoint));
	 		new_wpt->flagged = TblGetItemInt (table, row, 0);
	 		DmWrite (wpt, 0, new_wpt, sizeof (waypoint));
	 		MemHandleUnlock (rec_handle);
	 		MemPtrFree (new_wpt);
	 		DmReleaseRecord (waypoint_db, current_record, true);
	 		handled = true;
	 		break;
      	case ctlSelectEvent:
         	handled =  HandleListButton (e->data.ctlEnter.controlID);
	 		break;
      	case keyDownEvent:
	 		switch (e->data.keyDown.chr) {
	    		case pageUpChr:
	       			table_top -= TABLE_SCROLL;
	       			DisplayTable ();
	       			handled = true;
	       			break;
	    		case pageDownChr:
	       			table_top += TABLE_SCROLL;
	       			DisplayTable ();
	       			handled = true;
	       			break;
	 		}
         	break;
      	default:
			break;
   	}
   	return handled;
}


//
//	Waypoint edit:
//
//	Fill in fields from selected entry
//	On "OK", transfer back to list
//	On "CANCEL", just go back
//

Boolean HandleWaypointEditEvent (EventPtr e)
{
   	int handled = false;

   	switch (e->eType) {
      	case frmOpenEvent:
         	handled =  HandleEditOpen ();
	 	break;
      	case menuEvent:
         	handled = HandleEditControls (e->data.menu.itemID);
	 	break;
      	case ctlSelectEvent:
         	handled = EditButtons (e->data.ctlEnter.controlID);
	 	break;
      	default:
		break;
   	}
   	return handled;
}

//
//	Display a table of waypoints
//
//	For now, just fill in from memory, but later this will
//	read from a database file
//

void DisplayTable ()
{
   int i;
   FormPtr form;
   TablePtr table;
   UInt16 rec_num;
   MemHandle rec_handle;
   waypoint *wpt;

   // first, check that we're displaying from a logical position
   num_records = DmNumRecords (waypoint_db);
   if (table_top > num_records - TABLE_ROWS) {
      table_top = num_records - TABLE_ROWS;
   }
   if (table_top < 0) {
      table_top = 0;
   }

   form = FrmGetActiveForm();
   table = FrmGetObjectPtr( form, FrmGetObjectIndex( form, wptTable ));
   TblEraseTable (table);
   TblSetColumnUsable (table, 0, true);
   TblSetColumnUsable (table, 1, true);
   TblSetColumnUsable (table, 2, true);
   TblSetColumnUsable (table, 3, true);
   TblSetLoadDataProcedure (table, 2, descrTableLoadName);
   TblSetSaveDataProcedure (table, 2, descrTableSaveName);
   TblSetCustomDrawProcedure (table, 3, descrTableDrawDescr);
   for (i = 0; i < TABLE_ROWS; ++i) {
      rec_num = i + table_top;
      if (rec_num >= num_records) {
	 	TblSetRowUsable (table, i, false);
      }
      else {
	 	rec_handle = DmQueryRecord (waypoint_db, rec_num);
	 	wpt = MemHandleLock (rec_handle);
	 	TblSetRowID (table, i, rec_num);
	 	TblSetItemStyle (table, i, 0, checkboxTableItem);
	 	TblSetItemInt (table, i, 0, wpt->flagged);
	 	TblSetItemStyle (table, i, 1, numericTableItem);
	 	TblSetItemInt (table, i, 1, rec_num);
	 	TblSetItemStyle (table, i, 2, textTableItem);
	 	TblSetItemStyle (table, i, 3, customTableItem);
	 	MemHandleUnlock (rec_handle);
	 	TblSetRowUsable (table, i, true);
	 	TblSetRowSelectable (table, i, true);
	 	TblMarkRowInvalid (table, i);
      }
   }
   TblDrawTable (table);
}


//
//	Used to put the waypoint NAME into a table field, so
//	that it can be edited.  This is fairly simple ...
//

Err descrTableLoadName (MemPtr tbl, UInt16 row, UInt16 column,
                        Boolean editable, MemHandle *dataH, UInt16 *offset,
			UInt16 *size, FieldPtr fID)
{
   char *txt;
   waypoint *wpt;
   UInt16 rec_num = 0;
   MemHandle mem_handle,
            rec_handle;

   rec_num = TblGetRowID (tbl, row);
   rec_handle = DmQueryRecord (waypoint_db, rec_num);
   if (rec_handle == NULL) {
      *dataH = NULL;
      *offset = *size = 0;
      return 0;
   }
   wpt = MemHandleLock (rec_handle);
   mem_handle = MemHandleNew (GARMIN_NAME_LEN + 1);
   txt = MemHandleLock (mem_handle);
   StrNCopy (txt, wpt->name, GARMIN_NAME_LEN);
   txt[GARMIN_NAME_LEN] = '\0';
   *dataH = mem_handle;
   *offset = 0;
   *size = StrLen(txt);
   FldSetMaxChars (fID, GARMIN_NAME_LEN);
   MemHandleUnlock (rec_handle);
   MemHandleUnlock (mem_handle);
   return 0;
}


//
//	... and save it once it has changed
//

Boolean descrTableSaveName (MemPtr tbl, UInt16 row, UInt16 col)
{
   char *txt;
   waypoint *wpt,
		   *new_wpt;
   UInt16 rec_num = 0;
   MemHandle rec_handle;
   MemHandle mem_handle;
   FieldPtr field;

   field = TblGetCurrentField (tbl);
   if (field && FldDirty (field)) {
      rec_num = TblGetRowID (tbl, row);
      rec_handle = DmGetRecord (waypoint_db, rec_num);
      if (rec_handle == NULL) {
	 	return 0;
      }
      wpt = MemHandleLock (rec_handle);
      new_wpt = MemPtrNew (sizeof (waypoint));
      MemMove (new_wpt, wpt, sizeof (waypoint));
      mem_handle = (MemHandle) FldGetTextHandle (TblGetCurrentField (tbl));
      txt = MemHandleLock (mem_handle);
      StrNCopy (new_wpt->name, txt, GARMIN_NAME_LEN);
      new_wpt->name[GARMIN_NAME_LEN] = '\0';
      uppercase (new_wpt->name, true);
      new_wpt->changed = true;
      DmWrite (wpt, 0, new_wpt, sizeof (waypoint));
      MemHandleUnlock (rec_handle);
      MemHandleUnlock (mem_handle);
      DmReleaseRecord (waypoint_db, rec_num, true);
   }
   return 0;
}


//
//	Custom draw procedure for the description, which isn't
//	editable.  This forces it to select as a chunk.
//

void descrTableDrawDescr (MemPtr tbl, UInt16 row, UInt16 column,
	RectanglePtr bounds)
{
   char txt[16];
   waypoint *wpt;
   UInt16 rec_num = 0;
   MemHandle rec_handle;
   FontID old_font;

   rec_num = TblGetRowID (tbl, row);
   rec_handle = DmQueryRecord (waypoint_db, rec_num);
   if (rec_handle == NULL) {
      return;
   }
   wpt = MemHandleLock (rec_handle);
   StrNCopy (txt, wpt->descr, 15);
   txt[15] = '\0';
   old_font = FntSetFont (0);
   WinDrawChars (txt, StrLen(txt), bounds->topLeft.x, bounds->topLeft.y);
   FntSetFont (old_font);
   MemHandleUnlock (rec_handle);
   return;
}


//
//	Set an editable field's contents ...
//

void setField (UInt16 fieldname, char *contents)
{
   FormPtr form;
   FieldPtr field;
   MemHandle handle;
   char *txt;

   form = FrmGetActiveForm ();
   field = FrmGetObjectPtr( form, FrmGetObjectIndex( form, fieldname));
   handle = MemHandleNew (StrLen (contents) + 1);
   txt = MemHandleLock (handle);
   StrCopy (txt, contents);
   FldSetText(field, handle, 0, StrLen(contents));
   MemHandleUnlock (handle);
}

//
//	... and retrieve it
//

void getField (UInt16 fieldname, char *contents)
{
   FormPtr form;
   FieldPtr field;
   MemHandle handle;
   char *txt;

   form = FrmGetActiveForm ();
   field = FrmGetObjectPtr( form, FrmGetObjectIndex( form, fieldname));
   handle = (MemHandle) FldGetTextHandle (field);
   if (handle == NULL) {
      *contents = '\0';
   }
   else {
      txt = MemHandleLock (handle);
      StrCopy (contents, txt);
      MemHandleUnlock (handle);
   }
}


//
//	Convert a string to UPPERCASE in place, plus remove
//	non-Garmin characters.
//
//	*chk* does the Garmin support German Umlauts?

void uppercase (char *c, int strip_blanks)
{
   	for (; *c != '\0'; ++c) {
      		if (*c >= 'a' && *c <= 'z') {
	 		*c = *c - 'a' + 'A';
      		}
      		if (*c == ' ' && strip_blanks) {
	 		*c = '-';
      		}
      		if ((*c < 'A' || *c > 'Z') && (*c < '0' || *c > '9') &&
	   		*c != ' ' && *c != '-') {
	 		*c = '-';
      		}
	}
}


//
//	Delete a whole bunch of entries (either clear the lot or
//	check whether they are flagged before deleting).
//

void deleteWaypoints (int cmd)
{
   	int 		i;
   	int 		flag;
   	MemHandle 	handle;
   	waypoint 	*wpt;

   	for (i = 0; i < num_records; ++i) {
      		handle = DmQueryRecord (waypoint_db, i);
      		if (! handle)
	 		break;
      		wpt = MemHandleLock (handle);
      		flag = wpt->flagged;
      		MemHandleUnlock (handle);
      		switch (cmd) {
	 	case menuClrAll:
	    		DmRemoveRecord (waypoint_db, i--);
	    		break;
	 	case menuClrFlagged:
	    		if (flag)
	       			DmRemoveRecord (waypoint_db, i--);
	    		break;
	 	case menuClrUnflagged:
	    		if (! flag)
	       			DmRemoveRecord (waypoint_db, i--);
	    		break;
/*
		case menuGetID:
			GarminGetID();
			break;
 */
      		}
   	}
   	DisplayTable ();
}



//
//	manipulate the "flag" field of the database, either set all,
//	clear all or toggle.  A tad more complex than above, but not
//	by much ...
//

void changeFlags (int cmd)
{
   	int 		i;
   	MemHandle 	handle;
   	waypoint 	*wpt;
   	waypoint 	*new_wpt;

   	new_wpt = MemPtrNew (sizeof (waypoint));
   	for (i = 0; i < num_records; ++i) {
      		handle = DmGetRecord (waypoint_db, i);
      		wpt = MemHandleLock (handle);
      		MemMove (new_wpt, wpt, sizeof (waypoint));
      		switch (cmd) {
	 	case menuFlag:
	    		new_wpt->flagged = true;
	    		break;
	 	case menuUnflag:
	    		new_wpt->flagged = false;
	    		break;
	 	case menuToggleFlag:
	    		if (new_wpt->flagged)
	       			new_wpt->flagged = false;
	    		else
	       			new_wpt->flagged = true;
	    		break;
	 	case menuFlagChanged:
	    		if (new_wpt->changed)
	       			new_wpt->flagged = true;
	    		break;
	 	case menuClearChanged:
	    		new_wpt->changed = false;
	    		break;
      		}
      		DmWrite (wpt, 0, new_wpt, sizeof (waypoint));
      		MemHandleUnlock (handle);
      		DmReleaseRecord (waypoint_db, i, true);
   	}
   	MemPtrFree (new_wpt);
   	DisplayTable ();
}


//
//	Compare two waypoints for a sort
//

int CompareWaypoints (waypoint *wpt1, waypoint *wpt2,
                      int unused, SortRecordInfoPtr un1,
		      SortRecordInfoPtr un2, MemHandle appInfo)
{
   int result;

   result = StrCompare (wpt1->name, wpt2->name);
   if (!result) {
      result = StrCompare (wpt1->descr, wpt2->descr);
   }
   return result;
}


//
//	MemHandle edit "open form" event
//

int HandleEditOpen ()
{
   	FormPtr 	form;
   	FieldPtr 	field;
   	waypoint 	*wpt;
   	MemHandle 	rec_handle;
   	int		new = false;
       	int		grab_focus = false;
	ControlPtr 	ctl;
	ListPtr 	list;
	Char		s[80];
	float		fAlt;
	Int32		l;

   	DateTimeType	dat;
	Char		txt [longDateStrLength +timeStringLength];
	Char		TimeAscii[timeStringLength];
	SystemPreferencesType	sysPrefs;
	UInt16		symbol;

	// FlpCompFloat	fAlt;

   	form = FrmGetActiveForm ();
   	setField (dbName, currentDBName ());
   	if (current_record >= 0 && current_record < num_records) {
      		rec_handle = DmQueryRecord (waypoint_db, current_record);
      		wpt = MemHandleLock (rec_handle);
      		setField (editName, wpt->name);


/*xxx*/  //Different for D108
		if (( A001[4].tag == 'D' ) && ( A001[4].data == 108 )) {
		//if (A000.product_ID == 130) {
			// fAlt.f = wpt->alt;
			// FplFloatToLong( fAlt.f, &s);
			symbol = (Int16) (wpt->symbol);
			StrIToA( s, symbol );
      			setField (editSymbol, s);

			fAlt = wpt->alt;
			l = _f_ftoi( fAlt );
			StrIToA( s, l );
      			setField (editAltitude, s);

		} else {
    			/* Waypt Symbol */
    			list = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
        			editSymbollist));
    			ctl = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
        			editSymboltrigger));

			symbol = (Int8) (wpt->symbol) && 0x0F;
    			LstSetSelection (list, symbol );
    			LstSetTopItem   (list, symbol );
    			CtlSetLabel(ctl, LstGetSelectionText(list, symbol ));

      			setField (editDescr, wpt->descr);

		}

      		setField (editLatitude, GarminLongToPos (wpt->latitude));
      		setField (editLongitude, GarminLongToPos (wpt->longitude));
      		LstSetSelection (FrmGetObjectPtr (form, FrmGetObjectIndex (form, listNS)),
		       wpt->latitude < 0 ? 1 : 0);
      		LstSetSelection (FrmGetObjectPtr (form, FrmGetObjectIndex (form, listEW)),
			wpt->longitude < 0 ? 1 : 0);
      		if (StrLen (wpt->name) == 3 &&
	  		wpt->name[0] >= '0' && wpt->name[0] <= '9') {
	 		grab_focus = true;
      		}
      		MemHandleUnlock (rec_handle);
   	}
   	else {
      		new = grab_focus = true;
      		setField (editName, "WAYPT");

		// set description quite similar to the Garmin
		// but depending on system preferences

		//1 current date/time of the Palm
		TimSecondsToDateTime( TimGetSeconds(), &dat);

		//2 Get time and date formats
		//  from the system preferences.
    		PrefGetPreferences (&sysPrefs);
        	DateToDOWDMFormat (dat.month,
            		dat.day, dat.year,
            		sysPrefs.longDateFormat, txt);
		TimeToAscii(dat.hour,
			dat.minute,
			sysPrefs.timeFormat, TimeAscii);

		StrCat( txt, " " );
		StrCat( txt, TimeAscii );

      		setField (editDescr, txt);

      		setField (editLatitude, GarminLongToPos (0L));
      		setField (editLongitude, GarminLongToPos (0L));

    		/* Waypt Symbol */
		if (( A001[4].tag == 'D' ) && ( A001[4].data == 108 )) {
		} else {
    			list = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
        			editSymbollist));
    			ctl = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
        			editSymboltrigger));
    			LstSetSelection (list, (0) );
    			LstSetTopItem   (list, (0) );
    			CtlSetLabel(ctl, LstGetSelectionText(list, (0) ));
		}
   	}
   	FrmSetFocus (form, FrmGetObjectIndex (form, editName));
   	if (grab_focus) {
      		field = FrmGetObjectPtr( form, FrmGetObjectIndex( form, editName));
      		FldSetSelection (field, 0, 6);
   	}
   	FrmDrawForm (form);
   	return true;
}


//
//	Process edit "menu" events
//

int HandleEditControls (UInt16 menu_item)
{
   	int handled = false;
   	FormPtr form;
   	FieldPtr field;
   	UInt16 fID, start, end;

   	form = FrmGetActiveForm ();
   	fID = FrmGetFocus (form);
   	field = FrmGetObjectPtr (form, fID);
   	switch (menu_item) {
      	case menuAbout:
        	FrmCustomAlert (gpsAboutAlert, version_info, compile_date, " ");
        	handled = true;
        	break;
      	case menuDB:
	 	FrmCustomAlert (whoAmI, currentDBName(), " ", " ");
	 	handled = true;
	 	break;
      	case menuCut:
	 	FldCut (field);
	 	handled = true;
	 	break;
      	case menuCopy:
	 	FldCopy (field);
	 	handled = true;
	 	break;
      	case menuPaste:
	 	FldPaste (field);
	 	handled = true;
	 	break;
      	case menuCopySpecial:
	 	// grab all fields from the screen & internal & save
	 	handled = true;
	 	break;
      	case menuPasteSpecial:
	 	// must update waypoint + all fields from saved buffer
	 	handled = true;
	 	break;
      	case menuUndo:
	 	FldUndo (field);
	 	handled = true;
	 	break;
      	case menuSelectAll:
	 	FldSetSelection (field, 0, FldGetTextLength (field));
	 	handled = true;
	 	break;
      	case menuDelete:
	 	FldGetSelection (field, &start, &end);
	 	FldDelete (field, start, end);
	 	handled = true;
	 	break;
      	default:
	 	FrmAlert (gpsAlert1);
	 	break;
   	}
   	return handled;
}


//
//	... and process edit button taps
//

int EditButtons (UInt16 button)
{
   	int		handled = false;
   	FormPtr		form;
   	waypoint 	*wpt;
   	waypoint 	*new_wpt;
   	MemHandle 	rec_handle;
   	char 		tmp[15];
   	int 		dirty;
   	long 		lat;
	long		lon;
   	ListPtr 	list;
	Int32	i32;
	float	f;

   	form = FrmGetActiveForm ();
   	switch (button) {
      	case gpsOkBtn:
	 	dirty = false;
	 	if (current_record >= 0 && current_record < num_records) {
	    		rec_handle = DmGetRecord (waypoint_db, current_record);
	    		wpt = MemPtrNew (sizeof (waypoint));
	    		MemSet (wpt, sizeof (waypoint), 0);
	    		new_wpt = MemHandleLock (rec_handle);
	    		MemMove (wpt, new_wpt, sizeof (waypoint));
	 	}
	 	else {
	    		current_record = num_records + 1;
	    		rec_handle = DmNewRecord (waypoint_db, &current_record,
			      sizeof (waypoint));
	    		num_records = DmNumRecords (waypoint_db);
	    		table_top = num_records - TABLE_ROWS;
	    		wpt = MemPtrNew (sizeof (waypoint));
	    		MemSet (wpt, sizeof (waypoint), 0);
	    		new_wpt = MemHandleLock (rec_handle);
	 	}
	 	wpt->changed = true;
	 	getField (editName, wpt->name);

		/* symbol */
		if ( (A001[4].tag == 'D') && (A001[4].data == 108) ) {
			getField( editSymbol, tmp );
			wpt->symbol = StrAToI( tmp );
			/* Alt */
			getField( editAltitude, tmp );
			i32 = StrAToI( tmp );
			f = _f_itof( i32 );
			wpt->alt = f;
		} else {
    			list = FrmGetObjectPtr(form, FrmGetObjectIndex(form,
        			editSymbollist));
    			wpt->symbol = (Int16) (LstGetSelection(list));
	 		getField (editDescr, wpt->descr);
		}


	 	getField (editLatitude, tmp);
	 	wpt->latitude = GarminPosToLong (tmp);
	 	if (LstGetSelection (FrmGetObjectPtr (form,
		      FrmGetObjectIndex (form, listNS))) == 1) {
	    		wpt->latitude *= -1L;
	 	}
	 	getField (editLongitude, tmp);
	 	wpt->longitude = GarminPosToLong (tmp);
	 	if (LstGetSelection (FrmGetObjectPtr (form,
		      FrmGetObjectIndex (form, listEW))) == 1) {
	    		wpt->longitude *= -1L;
	 	}
	 	uppercase (wpt->name, true);
	 	uppercase (wpt->descr, false);
	 	DmWrite (new_wpt, 0, wpt, sizeof (waypoint));
	 	MemHandleUnlock (rec_handle);
	 	MemPtrFree (wpt);
	 	DmReleaseRecord (waypoint_db, current_record, true);
	 	FrmGotoForm (gpsWptListFrm);
	 	handled = true;
	 	break;
      	case gpsNokBtn:
	 	FrmGotoForm (gpsWptListFrm);
	 	handled = true;
	 	break;
      	case gpsKillBtn:
	 	if (current_record >= 0 && current_record < num_records) {
	    		DmRemoveRecord (waypoint_db, current_record);
	 	}
	 	FrmGotoForm (gpsWptListFrm);
	 	handled = true;
	 	break;
      	case gpsHereBtn:
	 	if (! GarminGetPos (&lat, &lon)) {
	    		setField (editLatitude, GarminLongToPos (lat));
	    		setField (editLongitude, GarminLongToPos (lon));
	    		LstSetSelection (FrmGetObjectPtr (form,
			     FrmGetObjectIndex (form, listNS)),
			     lat < 0 ? 1 : 0);
	    		LstSetSelection (FrmGetObjectPtr (form,
			     FrmGetObjectIndex (form, listEW)),
			     lon < 0 ? 1 : 0);
	    		FrmDrawForm (form);
	 	}
	 	break;
      	default:
//	 	FrmAlert (gpsAlert1);
//	 	handled = true;
	 	break;
   	}
   	return handled;
}


//
//	MemHandle list view menu events
//

int HandleListMenu (UInt16 menu_item)
{
   	int handled = false;
   	waypoint *wpt, *sel_wpt;
   	MemHandle record;

   	switch (menu_item) {
      	case menuOptions:
        	FrmPopupForm (gpsOptions);
        	handled = true;
        	break;
      	case menuAbout:
        	FrmCustomAlert (gpsAboutAlert, version_info, compile_date, " ");
        	handled = true;
        	break;
      	case menuDB:
        	FrmCustomAlert (whoAmI, currentDBName(), " ", " ");
        	handled = true;
        	break;
      	case menuNew:
        	newfile_camefrom = gpsWptListFrm;
        	FrmPopupForm (gpsFileNewFrm);
        	handled = true;
        	break;
      	case menuOpen:
        	file_merge = false;
        	FrmGotoForm (gpsFileOpenFrm);
        	handled = true;
        	break;
      	case menuMerge:
        	file_merge = true;
        	FrmGotoForm (gpsFileMergeFrm);
        	handled = true;
        	break;
      	case menuFileUtil:
        	file_merge = false;
        	FrmGotoForm (gpsFileUtilFrm);
        	handled = true;
        	break;
      	case menuClrAll:
      	case menuClrFlagged:
      	case menuClrUnflagged:
        	deleteWaypoints (menu_item);
        	handled = true;
        	break;
      	case menuFlag:
      	case menuFlagChanged:
      	case menuToggleFlag:
      	case menuUnflag:
      	case menuClearChanged:
        	changeFlags (menu_item);
        	handled = true;
        	break;
      	case menuSelectAll:
        	changeFlags (menuFlag);
        	handled = true;
        	break;
      	case menuSort:
        	DmQuickSort (waypoint_db, (DmComparF *) CompareWaypoints, 0);
        	current_record = table_top = 0;
        	DisplayTable ();
        	handled = true;
        	break;
      	case menuFetch:
        	GarminGetAll (GPS_WAYPOINT, 0);
        	num_records = DmNumRecords (waypoint_db);
        	table_top = num_records - TABLE_ROWS;
        	DisplayTable ();
        	handled = true;
        	break;
      	case menuXferTrackIn:
        	GarminGetAll (GPS_TRACK, 0);
        	break;
      	case menuClearTrackLog:
        	if (FrmCustomAlert (fileDeleteAlert, "Track Log", " ", " ") == 0) {
            		GarminClearTrackLog ();
        	}
        	break;
      	case menuXferRouteIn:
        	GarminGetAll (GPS_ROUTE, 0);
        	break;
      	case menuXferTrackOut:
        	GarminSendTrack ();
        	break;
      	case menuXferRouteOut:
        	GarminSendRaw (menu_item);
        	break;
      	case menuSendAll:
      	case menuSendFlagged:
      	case menuSendChanged:
        	GarminSend (menu_item);
        	handled = true;
        	break;
      	case menuGarminHost:
        	FrmGotoForm (garminHostFrm);
        	handled = true;
        	break;
	case menuGetID:
		GarminGetID();
		FrmAlert( gpsProtoAlert );
		break;
	case menuDateTime:
		GarminGetDateTime();
		break;
    	default:
        	FrmAlert (gpsAlert1);
        	break;
    	case menuDelete:
        	if (current_record >= 0 && current_record < num_records) {
            		DmRemoveRecord (waypoint_db, current_record);
        	}
        	FrmGotoForm (gpsWptListFrm);
        	handled = true;
        	break;
    	case menuCopySpecial:
        	if (current_record >= 0 && current_record < num_records) {
            		if (wpt_copy_buf == NULL) {
                		wpt_copy_buf = MemHandleNew (sizeof (waypoint));
            		}
            		wpt = MemHandleLock (wpt_copy_buf);
            		record = DmQueryRecord (waypoint_db, current_record);
            		sel_wpt = MemHandleLock (record);
            		MemMove (wpt, sel_wpt, sizeof (waypoint));
            		MemHandleFree (wpt_copy_buf);
            		MemHandleFree (record);
        	}
        	else {
            		if (wpt_copy_buf != NULL) {
            			MemHandleFree (wpt_copy_buf);
                		wpt_copy_buf = NULL;
            		}
        	}
        	handled = true;
        	break;
    	}
    	return handled;
}


//
//	List view button selections
//

int HandleListButton (UInt16 button)
{
   	int handled = false;

   	switch (button) {
      	case gpsOkBtn:
	 	if (current_record >= 0 && current_record < num_records) {
			if ( A000.product_ID == 130) {
	    			FrmGotoForm (gpsWpt_D108_EditFrm);
			} else {
	    			FrmGotoForm (gpsWptEditFrm);
			}
		};
	 	handled = true;
	 	break;
      	case gpsNewBtn:
	 	current_record = -1;
	 	FrmGotoForm (gpsWptEditFrm);
	 	handled = true;
	 	break;
      	case gpsKillBtn:
	 	if (current_record >= 0 && current_record < num_records) {
	    		DmRemoveRecord (waypoint_db, current_record);
	    		DisplayTable ();
		}
	 	handled = true;
	 	break;
      	case gpsSortBtn:
	 	DmQuickSort (waypoint_db, (DmComparF *) CompareWaypoints, 0);
	 		current_record = table_top = 0;
	 	DisplayTable ();
	 	handled = true;
	 	break;
      	default:
	 	FrmAlert (gpsAlert1);
	 	break;
   	}
   	return handled;
}


//
//	Get the name of the currently active database
//

char * currentDBName (void)
{
   LocalID id;
   UInt16 card;
   static char name[32];

   name[0] = '\0';
   if (waypoint_db) {
      DmOpenDatabaseInfo (waypoint_db, &id, NULL, NULL, &card, NULL);
      DmDatabaseInfo (card, id, name, NULL, NULL, NULL, NULL, NULL,
				      NULL, NULL, NULL, NULL, NULL);
   }
   return name;
}
