/*
 PalGar - use a PALm to up-, download, edit waypoints from a GARmin GPS
 Copyright (C) 1999 Paul Nash
 Copyright (C) 2000 Ralf Meyer <ranf@users.sourceforge.net>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

//
//	manifest constants for GPS editor
//
//	Paul Nash
//	paul@frcs.alt.za
//

//
//	$Header: /cvs/PalGar/gps.h,v 1.15 2000/11/26 15:01:53 ranf Exp $
//
//	$Log: gps.h,v $
//	Revision 1.15  2000/11/26 15:01:53  ranf
//	- edit eTrex waypoint now with altitude
//	
//	Revision 1.14  2000/11/09 21:21:20  ranf
//	- eTrex: added Altitude to WPT
//
//	Revision 1.13  2000/09/19 20:19:03  ranf
//	- fixed probs with preferences saving
//
//	Revision 1.12  2000/08/24 20:48:22  ranf
//	- global struct to keep Product_Data
//	- global struct to keep Protocol_Array
//	- removed anything related to the callback.h stuff
//	- version number up to 1.7.0
//
//	Revision 1.11  2000/07/19 19:59:40  ranf
//	Get ID from GPS finally works
//
//	Revision 1.10  2000/07/10 19:14:01  ranf
//	Edit Wpt symbol. No bitmap yet.
//
//	Revision 1.9  2000/07/06 19:02:15  ranf
//	Options dialogue for setting the time zone
//
//	Revision 1.8  2000/07/02 21:52:45  ranf
//	changed preferences to include the time offset to UTC
//
//	Revision 1.7  2000/06/26 17:55:16  ranf
//	Checksum errors when asked for GPS type
//
//	Revision 1.1.1.1  2000/03/23 18:18:40  giga
//	Initial import
//
//	Revision 1.31  1999/06/13 17:26:13  paul
//	Changed lat & long edit to use N/S E/W list (button) input, rather
//	than text.  This allows the ':' separator to use ' ' or '-'.
//
//	Revision 1.30  1999/06/12 14:38:39  paul
//	changed track log to allow cumulative logging.  Not tested yet
//
//	Revision 1.29  1999/05/12 08:48:05  paul
//	Ready for release as Gypsy 1.4.  Major changes include:
//	*  Rudimentary track/route buffer to transfer data between GPS
//	   and PC (a'la gps2p2pc)
//	*  Protocol fix for DLE escaping bug
//	*  Code changes from <andre@drfmc.ceng.cea.fr> to be able to
//	   compile a PalmOS2 version
//	*  Garmin protocol tries a bit harder to recover from comms
//	   errors (but is still *very* dumb)
//
//


//
//	Various system constants
//

#define TABLE_ROWS		10
#define TABLE_SCROLL		 9


//
//	Interface definitions (for gps.rcp)
//

//	forms

#define gpsWptListFrm		1000
#define gpsWptEditFrm		1001
#define gpsXfrFrm		1002
#define garminHostFrm		1003
#define gpsFileListFrm		1004
#define gpsFileNewFrm		1005
#define gpsFileOpenFrm		1006
#define gpsFileMergeFrm		1007
#define gpsFileUtilFrm		1008
#define gpsWpt_D108_EditFrm	1009

//	buttons

#define gpsOkBtn		1100
#define gpsNewBtn		1101
#define gpsDnlBtn		1103
#define gpsUplCngBtn		1104
#define gpsUplAllBtn		1105
#define gpsKillBtn		1105
#define gpsHereBtn		1106
#define gpsSortBtn		1107
#define gpsUplFlagBtn		1108
#define gpsNokBtn		1109

//	tables and sets of fields masquerading as tables

#define wptTable		1200
#define fileTable		1201

//	normal fields and labels

#define editName		1700
#define editDescr		1701
#define editLatitude 		1702
#define editLongitude 		1703
#define garminMessage		1704
#define newFile			1705
#define dbName			1706
#define dbNameLbl		1707
#define hostAction		1708
#define optUTClabel		1709
#define optUTClist		1710
#define optUTCtrigger		1711
#define optUTChours		1712

#define optWPTlabel		1721
#define optWPTlist		1722
#define optWPTtrigger		1723


#define editSymbollabel		1713
#define editSymboltrigger	1714
#define editSymbollist		1715

#define editSymbol		1740
#define editAltLabel		1751
#define editAltitude		1741


//	alerts

#define gpsAlert1		1300
#define gpsAlert2		1301
#define gpsAlert3		1302
#define gpsAboutAlert		1303
#define gpsError		1304
#define gpsBreak1		1305
#define gpsBreak2		1306
#define gpsBreak3		1307
#define gpsSerialAlert		1308
#define serialAlert		1309
#define fileDeleteAlert		1310
#define whoAmI			1311
#define noActiveDB		1312
#define gpsProtoAlert		1313

//	help screens

#define gpsHelp1		1500
#define gpsHelp2		1501

#define gpsGpl			1502
#define gpsOptions		1503

//	menus

#define menuList		1400
#define menuEdit		1401

//	... and menu items

#define menuOpen		1600
#define menuAbout		1601
#define menuUndo		1602
#define menuCut			1603
#define menuCopy		1604
#define menuPaste		1605
#define menuKbd			1606
#define menuFlag		1607
#define menuUnflag		1608
#define menuFlagChanged		1609
#define menuSort		1610
#define menuFetch		1611
#define menuSendAll		1612
#define menuSendFlagged		1613
#define menuSendChanged		1614
#define menuClrAll		1615
#define menuClrFlagged		1616
#define menuClrUnflagged	1617
#define menuToggleFlag		1618
#define menuSelectAll		1619
#define menuDelete		1620
#define menuNew			1621
#define menuMerge		1622
#define menuClearChanged	1623
#define menuPasteSpecial	1624
#define menuCopySpecial		1625
#define menuDB			1626
#define menuFileUtil		1627
#define menuGarminHost		1628
#define menuXferTrackIn		1629
#define menuXferTrackOut	1630
#define menuXferRouteIn		1631
#define	menuXferRouteOut	1632
#define menuClearTrackLog	1633
#define menuOptions		1634

#define menuGetID		1635
#define menuDateTime		1636

//	lists

#define listNS			1800
#define listEW			1801
