/*
 PalGar - use a PALm to up-, download, edit waypoints from a GARmin GPS
 Copyright (C) 1999 Paul Nash
 Copyright (C) 2000 Ralf Meyer <ranf@users.sourceforge.net>

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

//
//	Garmin host-mode protocol engine.
//
//	Paul Nash
//	paul@frcs.alt.za

//
//	$Header: /cvs/PalGar/host.c,v 1.19 2000/11/26 15:01:53 ranf Exp $
//
//	$Log: host.c,v $
//	Revision 1.19  2000/11/26 15:01:53  ranf
//	- edit eTrex waypoint now with altitude
//	
//	Revision 1.18  2000/11/19 18:57:01  ranf
//	- WPT should also work with the eTrex Summit
//
//	Revision 1.17  2000/11/09 21:21:20  ranf
//	- eTrex: added Altitude to WPT
//
//	Revision 1.16  2000/09/13 19:38:24  giga
//	*** empty log message ***
//
//	Revision 1.15  2000/08/24 20:48:22  ranf
//	- global struct to keep Product_Data
//	- global struct to keep Protocol_Array
//	- removed anything related to the callback.h stuff
//	- version number up to 1.7.0
//
//	Revision 1.14  2000/07/19 19:59:40  ranf
//	Get ID from GPS finally works
//
//	Revision 1.13  2000/07/12 18:35:21  ranf
//	- bugfix: when merging on a Palm OS >= 3 it only took
//	every second record
//	- first attempt to add eTrex waypoint up- /download
//	(needs testing)
//
//	Revision 1.12  2000/07/10 19:14:01  ranf
//	Edit Wpt symbol. No bitmap yet.
//
//	Revision 1.11  2000/07/06 19:02:16  ranf
//	Options dialogue for setting the time zone
//
//	Revision 1.10  2000/07/02 21:52:45  ranf
//	changed preferences to include the time offset to UTC
//
//	Revision 1.9  2000/07/02 13:18:30  ranf
//	- get date/time from the Palm (PalGar has to be in host mode).
//		Be careful with that: there's no local time to UTC conversion yet.
//	- turn off the Palm (again host mode required).
//
//	Revision 1.8  2000/06/30 18:36:46  ranf
//	- Send date/time in host mode (not yet UTC. But local time)
//	- identify a Cmnd_Turn_Off_pwr (how the heck do I turn the Palm off?)
//
//	Revision 1.7  2000/06/26 17:55:16  ranf
//	Checksum errors when asked for GPS type
//
//	Revision 1.4  2000/04/24 14:20:30  ranf
//	Added small icon
//		(not very nice. But now we have one)
//
//	Revision 1.3  2000/03/27 18:24:31  giga
//	OS 35 compiles now
//
//	Revision 1.2  2000/03/27 18:03:29  giga
//	OS 35 Upgrader Perl ran
//
//	Revision 1.1.1.1  2000/03/23 18:18:40  giga
//	Initial import
//
//	Revision 1.9  1999/08/30 06:35:32  paul
//	Added Jaque Andre's patches
//
//	Revision 1.8  1999/05/13 10:25:13  paul
//	Still trying to make GPSy work.
//
//	Revision 1.7  1999/05/12 08:48:05  paul
//	Ready for release as Gypsy 1.4.  Major changes include:
//	*  Rudimentary track/route buffer to transfer data between GPS
//	   and PC (a'la gps2p2pc)
//	*  Protocol fix for DLE escaping bug
//	*  Code changes from <andre@drfmc.ceng.cea.fr> to be able to
//	   compile a PalmOS2 version
//	*  Garmin protocol tries a bit harder to recover from comms
//	   errors (but is still *very* dumb)
//
//


#include <PalmOS.h>
#include "gps.h"
#include <System/SerialMgrOld.h>

#include "globals.h"
#include "GarminDef.h"

extern Int8	timeoffset;

//
//	Local function prototypes
//

void GetHostCommand (void);
void displayPacket (char *msg, UInt8 *buf);
void showLog (char *contents);

//	Internal global variables and the like

// static char *rcs_id = "$Id: host.c,v 1.19 2000/11/26 15:01:53 ranf Exp $";

#define BUFLEN 256

extern Product_Data_Type	A000;
extern Protocol_Array_Type	A001;

//
//	Primary form event handler
//

Boolean HandleHostEvent (EventPtr e)
{
   int handled = false;
   FormPtr form;
   Int32 num_bytes = 0L;

   	form = FrmGetActiveForm ();
   	switch (e->eType) {
	case frmOpenEvent:
		setField (dbName, currentDBName ());
	 	showLog ("Initialising");
	 	FrmDrawForm (form);
	 	if (OpenSerialPort ()) {
	    		FrmGotoForm (gpsWptListFrm);
	 	}
	 	handled = true;
	 	break;
      	case ctlSelectEvent:
        	switch (e->data.ctlEnter.controlID) {
	    	case gpsNokBtn:
	       		if (serial_port_open) {
		  		SerClose (serial_lib);
		  		serial_port_open = false;
	       		}
	       		FrmGotoForm (gpsWptListFrm);
	       		handled = true;
	       		break;
		}
	 	break;
      	case nilEvent:
	 	SerReceiveCheck(serial_lib, &num_bytes);
	 	if (num_bytes != 0L) {
//	    		showLog ("Checking");
	    		GetHostCommand ();
	 	}
	 	handled = true;
	 	break;
      	default:
		break;
   	}
   	FrmDrawForm (form);
   	return handled;
}


//
// put a virtual keydown into the event queue
// Rustin Nesse
//

static void AddKeyDownEvent( UInt16 character )
{
	EventType 	theEvent;	           // Create the key down event

	MemSet( &theEvent, sizeof(EventType), 0);
	theEvent.eType = keyDownEvent;
	theEvent.data.keyDown.chr = character;
	theEvent.data.keyDown.keyCode = 0;
	theEvent.data.keyDown.modifiers = commandKeyMask;
	EvtAddEventToQueue(&theEvent); // Add the event to the event queue
}

//
//	Get a Garmin host type command and act on it
//

void GetHostCommand (void)
{
   	UInt8	buf[BUFLEN];
   	int 	len, err;
   	UInt16 	num_wpts;
	UInt16	data;
   	DateTimeType dat;
   	UInt8	yearHi;
   	UInt8	yearLo;
   	int 	i = 0;
   	int 	nCnt = 0;

   	len = BUFLEN;
   	err = 0;
   	if (GarminGetPacket (buf, &len, &err)) {
      		if (buf[1] != Pid_Ack_Byte && buf[1] != Pid_Nak_Byte) {
	 		SendAck (buf[1], false);
      		}
      		displayPacket ("G:", buf);
      		return;
   	}
   	if (buf[1] != Pid_Ack_Byte && buf[1] != Pid_Nak_Byte) {
      		SendAck (buf[1], true);
   	}
	switch (buf[1]) {
      	case Pid_Ack_Byte:
      	case Pid_Nak_Byte:
	 	break;
      	case Pid_Command_Data:
	 	switch (buf[3]) {
		case Cmnd_Transfer_Wpt:
	       		showLog ("Send waypoints");
	       		GarminSend (menuSendAll);
	       		break;
		case Cmnd_Transfer_Time:
	       		showLog ("Time");
	       		MemSet (buf, BUFLEN, 0);
		   		TimSecondsToDateTime( TimGetSeconds(), &dat);
		   		TimAdjust (&dat, -timeoffset *60 *60);  // make it UTC
		   		yearLo = dat.year & 0xFF;
		   		yearHi = dat.year >> 8;

	       		buf[0] = Pid_Date_Time_Data;
	       		buf[1] = 10;
			buf[2] = dat.month;
			buf[3] = dat.day;
			buf[4] = yearLo;
			buf[5] = yearHi;
			buf[6] = dat.hour;
			buf[7] = 0;
			buf[8] = dat.minute;
			buf[9] = dat.second;
	       		GarminSendPacket (buf);
	       		break;
		case Cmnd_Start_Pvt_Data:
	       		showLog ("PVT data");
	       		MemSet (buf, BUFLEN, 0);
	       		buf[0] = Pid_Pvt_Data;
	       		buf[1] = 63;
	       		buf[18] = 1;
	       		GarminSendPacket (buf);
	       		break;
		case Cmnd_Transfer_Posn:
	       		showLog ("Position");
	       		break;
		case Cmnd_Stop_Pvt_Data:
	       		showLog ("stop PVT");
	       		break;
		case Cmnd_Transfer_Trk:
	       		showLog ("Send track");
	       		GarminSendTrack ();
	       		break;
		case Cmnd_Transfer_Rte:
	       		showLog ("Send routes");
	       		GarminSendRaw (menuXferRouteOut);
	       		break;
		case Cmnd_Turn_Off_Pwr:
			showLog ("Turn off");
			AddKeyDownEvent (hardPowerChr);
			break;
		default:
	       		displayPacket ("U:", buf);
	       		break;
		}
		break;
	case Pid_Xfer_Cmplt	:
	 	displayPacket ("EOF", buf);
	 	break;
      	case Pid_Records		:
	 	showLog ("Rx waypoint/track/route");
	 	num_wpts = buf[4];
	 	num_wpts = num_wpts << 8;
	 	num_wpts |= buf[3];
	 	num_wpts &= 0x00FFFFFF;
	 	GarminGetAll (HOST_WAYPOINT, num_wpts);
	 	break;
      	case Pid_Product_Rqst	:
        	showLog ("GPS type");
        	buf[0] = Pid_Product_Data;
        	buf[1] = StrLen(A000.product_description) +4;
		if (A000.product_ID == 0) {
			A000.product_ID = 97;  		//   GPS II plus
		};
        	buf[2] = (unsigned char) (A000.product_ID);
        	buf[3] = 0;

		A000.software_version = (UInt16) A000.software_version;
		endian_convert( & A000.software_version, sizeof(UInt16) );
		MemMove( &buf[4], & A000.software_version, sizeof(UInt16) );

        	StrCopy (buf + 6, A000.product_description);
        	GarminSendPacket (buf);
		// wait for ACk
	 	len = BUFLEN;

//	 	GarminGetPacket (buf, &len, &err);
		// then send Proto array
//		lOut = true;
		MemSet(buf, sizeof(buf), 0 );
        	buf[0] = Pid_Protocol_Array;
		// find length
		for ( i=0; i < sizeof(A001) /3; i++) {
			if ( A001[i].tag == (char) "\0") {
				break;
			}
		};
		nCnt = i;
        	buf[1] = nCnt *3 +4;
		for ( i=0; i < nCnt; i++ ) {
			buf[3 *i +3] = (char) A001[i].tag;
			data = (UInt16) A001[i].data;
			endian_convert( & data, sizeof(UInt16) );
			MemMove( & buf[3 *i +4], & data, sizeof(UInt16) );
		};
        	GarminSendPacket (buf);
        	break;
      	default:
        	displayPacket ("U:", buf);
        	break;
   	}
}



//
//	Display the first few bytes of a Garmin packet
//

void displayPacket (char *msg, UInt8 *buf)
{
   	char txt[256];
   	char tmp[16];
   	int i;

   	StrCopy (txt, msg);
   	StrCat (txt, " 0x");
   	for (i = 1; i < 5; ++i) {
      		if (i > 1)
	 		StrCat (txt, " ");
      		StrIToH (tmp, buf[i]);
      		StrCat (txt, tmp + 6);
   	}
   	showLog (txt);
}


//
//	display the last ten lines or so ,,,
//

void showLog (char *contents)
{
#define LOG 150
   	FormPtr 	form;
   	FieldPtr 	field;
   	MemHandle 	handle;
   	static char 	log[LOG + 1] = {'\0'};
   	char 		*txt;
   	int 		i;

   	form = FrmGetActiveForm ();
   	field = FrmGetObjectPtr( form, FrmGetObjectIndex( form, hostAction));
   	i = (StrLen (log) + StrLen (contents) + 2) - LOG;
   	if (i > 0) {
      		StrCopy (log, log + i);
   	}
   	StrCat (log, "; ");
   	StrCat (log, contents);
   	handle = MemHandleNew (StrLen (log) + 1);
   	txt = MemHandleLock (handle);
   	StrCopy (txt, log);
   	FldSetText(field, handle, 0, StrLen(log));
   	MemHandleUnlock (handle);
   	FldDrawField (field);
}
