;**************************************************************************
;*  emumem.asm
;*			Memory emulation routines for PalmBoy
;*
;*  2000-09-10	Bodo Wenzel	Got from runmode.asm
;*  2000-09-17	Bodo Wenzel	Now nearly full screen emulation
;*  2000-10-19	Bodo Wenzel	Interrupts and timer
;*  2000-11-15	Bodo Wenzel	Support of color devices
;**************************************************************************
;
;  (c)2000 Bodo Wenzel
;
;  This program is free software; you can redistribute it and/or modify
;  it under the terms of the GNU General Public License as published by
;  the Free Software Foundation; either version 2 of the License, or
;  (at your option) any later version.
; 
;  This program is distributed in the hope that it will be useful,
;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;  GNU General Public License for more details.
; 
;  You should have received a copy of the GNU General Public License
;  along with this program; if not, write to the Free Software
;  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
;**************************************************************************

;	IMPORTANT!
;	Please make sure that any referenced resource is included!
;	Unfortunately there is no automatic!

;=== Global variables =====================================================

	data

struct	MemRamType
	vid8000.$2000			;video RAM
	extA000.$2000			;external RAM, if none in database
	intC000.$2000			;internal RAM
	oamFe00.$0100			;OAM
	cpuFf80.$0080			;CPU RAM
endstruct
memRamSize	equ	sizeof(MemRamType)
global	MemRam.l			;address of allocated MemRamType

;NOTE: as GB addresses are used as offsets, the addresses for Palm memory
;blocks are offset by the base addresses.
global	MemRomPages.128		;32*l	;addresses of ROM pages
global	MemRamPages.16		;4*l	;addresses of RAM pages

;=== Program code =========================================================

	code

;--------------------------------------------------------------------------
;Boolean MemInit(DmOpenRef dbP)
;------------------------------
;	Get memory (in the given database). If successful return true
;	otherwise false.
proc	MemInit(dbP.l)
local	recId.w
beginproc
	clr.l	MemRam(a5)

	move.w	#dmMaxRecordIndex,recId(a6)
	systrap	DmNewRecord(dbP(a6).l,&recId(a6),#memRamSize.l)
	move.l	a0,d0
	beq.s	MiReturnError

	systrap	MemHandleLock(a0.l)
	move.l	a0,MemRam(a5)

	moveq	#true,d0
	bra.s	MiReturn

MiReturnError
	moveq	#false,d0
MiReturn
endproc

;--------------------------------------------------------------------------
;void MemExit(void)
;------------------
;	Release the used memory.
proc	MemExit()
beginproc
	move.l	MemRam(a5),d0
	beq.s	MeReturn

	systrap	MemPtrUnlock(d0.l)
	clr.l	MemRam(a5)
MeReturn
endproc

;--------------------------------------------------------------------------
;void MemRomSelect(bank.b)
;-------------------------
;	Switch the ROM bank.
MemRomSelect
	and.w	#$1f,d0			;only 32 banks emulated (512 KB)
	add.w	d0,d0
	add.w	d0,d0
	lea	MemRomPages(a5),a3
	move.l	(a3,d0.w),BASE_ROM_1(a4)
	jmp	(a6)

;--------------------------------------------------------------------------
;void MemRamSelect(bank.b)
;-------------------------
;	Switch the RAM bank.
MemRamSelect
	and.w	#$03,d0			;only 4 banks emulated (32 KB)
	add.w	d0,d0
	add.w	d0,d0
	lea	MemRamPages(a5),a3
	move.l	(a3,d0.w),BASE_RAM_EXT(a4)
	jmp	(a6)

;--------------------------------------------------------------------------
;void MemRamNone(data.b,addr.w)
;------------------------------
;	Write to RAM without watching.
MemRamNone
	move.b	d0,(a2,d1.w)
	jmp	(a6)

;--------------------------------------------------------------------------
;void MemRamTileBg(data.b,addr.w)
;--------------------------------
;	Write into tile table for background.
MemRamTileBg
	cmp.b	(a2,d1.w),d0
	beq.s	MrtbQuit		;no change?

	move.w	d3,-(sp)

	move.b	d0,(a2,d1.w)		;store pattern in tile table
MrtbStart
	and.w	#$fffe,d1
	lea	(a2,d1.w),a3
	ror.w	#4,d1
	move.w	d1,d0
	and.w	#$ff,d1
	move.w	d1,d3
	ror.w	#3,d0
	and.w	#7*$100*4,d0
	add.w	d1,d1
	add.w	d1,d1
	add.w	d1,d0			;prepare some values

MrtbCheck
	move.w	VidBgScxPixel(a5),d2
	move.l	VidBgUsed(a5),a2
	cmp.b	(a2,d3.w),d2
	bne.s	MrtbExit		;tile not used?

	move.l	VidBgTile(a5),a2
	add.w	d0,a2

	move.l	a0,-(sp)
	moveq	#0,d0
	moveq	#0,d1
	move.l	VidBgWiColorL(a5),a0
	jsr	(a0)
	swap	d0
	move.l	VidBgWiColorH(a5),a0
	jsr	(a0)
	add.l	d0,d0
	move.l	GrayCopyTable(a5),a0
	move.w	(a0,d0.w),d1
	add.w	d1,d1
	swap	d0
	or.w	(a0,d0.w),d1
	lsl.l	d2,d1
	move.l	d1,(a2)
	moveq	#0,d1
	move.b	(a3)+,d1
	or.b	(a3),d1
	add.w	d1,d1
	move.w	(a0,d1.w),d1
	move.w	d1,d0
	add.w	d0,d0
	or.w	d0,d1
	lsl.l	d2,d1
	move.l	d1,8*$100*4(a2)
	move.l	(sp)+,a0		;convert one line
MrtbExit
	move.w	(sp)+,d3
MrtbQuit
	jmp	(a6)

;--------------------------------------------------------------------------
;void MemRamTileBgWi(data.b,addr.w)
;----------------------------------
;	Write into tile table for background and window.
MemRamTileBgWi
	cmp.b	(a2,d1.w),d0
	beq.s	MrtbQuit		;no change?

	move.w	d3,-(sp)

	move.b	d0,(a2,d1.w)		;store pattern in tile table
MrtbwStart
	and.w	#$fffe,d1
	lea	(a2,d1.w),a3
	ror.w	#4,d1
	move.w	d1,d0
	and.w	#$ff,d1
	move.w	d1,d3
	ror.w	#3,d0
	and.w	#7*$100*4,d0
	add.w	d1,d1
	add.w	d1,d1
	add.w	d1,d0			;prepare some values

	move.w	VidWiWxPixel(a5),d2
	move.l	VidWiUsed(a5),a2
	cmp.b	(a2,d3.w),d2
	bne.s	MrtbwQuit		;tile not used in window?

	move.l	VidWiTile(a5),a2
	add.w	d0,a2

	move.l	a0,-(sp)
	move.w	d0,-(sp)
	moveq	#0,d0
	moveq	#0,d1
	move.l	VidBgWiColorL(a5),a0
	jsr	(a0)
	swap	d0
	move.l	VidBgWiColorH(a5),a0
	jsr	(a0)
	add.l	d0,d0
	move.l	GrayCopyTable(a5),a0
	move.w	(a0,d0.w),d1
	add.w	d1,d1
	swap	d0
	or.w	(a0,d0.w),d1
	lsl.l	d2,d1
	move.l	d1,(a2)
	moveq	#0,d1
	move.b	(a3),d1
	or.b	1(a3),d1
	add.w	d1,d1
	move.w	(a0,d1.w),d1
	move.w	d1,d0
	add.w	d0,d0
	or.w	d0,d1
	lsl.l	d2,d1
	move.l	d1,8*$100*4(a2)
	move.w	(sp)+,d0
	move.l	(sp)+,a0		;convert one line (wi)
MrtbwQuit
	bra	MrtbCheck		;now check for background

;--------------------------------------------------------------------------
;void MemRamTileOb(data.b,addr.w)
;--------------------------------
;	Write into tile table for objects.
MemRamTileOb
	cmp.b	(a2,d1.w),d0
	beq.s	MrtoQuit		;no change?

	move.b	d0,(a2,d1.w)		;store pattern in tile table

	move.w	d3,-(sp)
	bsr	VidObTileCheck
	move.w	(sp)+,d3
MrtoQuit
	jmp	(a6)

;--------------------------------------------------------------------------
;void MemRamTileBgOb(data.b,addr.w)
;----------------------------------
;	Write into tile table for background and objects.
MemRamTileBgOb
	cmp.b	(a2,d1.w),d0
	beq.s	MrtoQuit		;no change?

	move.b	d0,(a2,d1.w)		;store pattern in tile table

	move.w	d3,-(sp)
	bsr	VidObTileCheck
	bra	MrtbStart

;--------------------------------------------------------------------------
;void MemRamTileBgWiOb(data.b,addr.w)
;------------------------------------
;	Write into tile table for background, window, and objects.
MemRamTileBgWiOb
	cmp.b	(a2,d1.w),d0
	beq.s	MrtoQuit		;no change?

	move.b	d0,(a2,d1.w)		;store pattern in tile table

	move.w	d3,-(sp)
	bsr	VidObTileCheck
	bra	MrtbwStart

;--------------------------------------------------------------------------
;void MemRamMapNone(data.b,addr.w)
;---------------------------------
;	write into tile map without watching
MemRamMapNone
	cmp.b	(a2,d1.w),d0
	beq.s	MrmnQuit		;no change?

	move.b	d0,(a2,d1.w)		;store tile number in map

	add.w	d1,d1
	and.w	#2*MAP_WRAP*MAP_WRAP*2-1,d1
	move.l	VidMapIndices(a5),a2
	moveq	#0,d2
	move.b	d0,d2
	add.w	d2,d2
	add.w	d2,d2
	move.w	d2,(a2,d1.w)		;store precalculated index
MrmnQuit
	jmp	(a6)

;--------------------------------------------------------------------------
;void MemRamMapBg(data.b,addr.w)
;-------------------------------
;	Write into background tile map.
MemRamMapBg
	cmp.b	(a2,d1.w),d0
	beq.s	MrmbQuit		;no change?

	move.b	d0,(a2,d1.w)		;store tile number in map

	add.w	d1,d1
	and.w	#2*MAP_WRAP*MAP_WRAP*2-1,d1
	move.l	VidMapIndices(a5),a2
	moveq	#0,d2
	move.b	d0,d2
	move.w	d2,d0
	add.w	d2,d2
	add.w	d2,d2
	move.w	d2,(a2,d1.w)		;store precalculated index

	move.w	d2,d1
	move.w	VidBgScxPixel(a5),d2
	move.l	VidBgUsed(a5),a2
	cmp.b	(a2,d0.w),d2
	beq.s	MrmbQuit		;already converted?
	move.b	d2,(a2,d0.w)

	move.l	VidBgWiTileLow(a5),a3
	tst.b	d0
	bpl.s	MrmbTileOK
	move.l	VidBgWiTileHigh(a5),a3
MrmbTileOK
	move.l	VidBgTile(a5),a2
	add.w	d1,a2
	add.w	d1,d1
	add.w	d1,d1
	add.w	d1,a3
	movem.l	d3/a0-a1/a4,-(sp)
	move.l	VidBgWiColorL(a5),a0
	move.l	VidBgWiColorH(a5),a1
	move.l	GrayCopyTable(a5),a4
	bsr	VidConvertBgWi
	movem.l	(sp)+,d3/a0-a1/a4

MrmbQuit
	jmp	(a6)

;--------------------------------------------------------------------------
;void MemRamMapWi(data.b,addr.w)
;-------------------------------
;	Write into window tile map.
MemRamMapWi
	cmp.b	(a2,d1.w),d0
	beq.s	MrmwQuit		;no change?

	move.b	d0,(a2,d1.w)		;store tile number in map

	add.w	d1,d1
	and.w	#2*MAP_WRAP*MAP_WRAP*2-1,d1
	move.l	VidMapIndices(a5),a2
	moveq	#0,d2
	move.b	d0,d2
	move.w	d2,d0
	add.w	d2,d2
	add.w	d2,d2
	move.w	d2,(a2,d1.w)		;store precalculated index

	move.w	d2,d1
	move.w	VidWiWxPixel(a5),d2
	move.l	VidWiUsed(a5),a2
	cmp.b	(a2,d0.w),d2
	beq.s	MrmwQuit		;already converted?
	move.b	d2,(a2,d0.w)

	move.l	VidBgWiTileLow(a5),a3
	tst.b	d0
	bpl.s	MrmwTileOK
	move.l	VidBgWiTileHigh(a5),a3
MrmwTileOK
	move.l	VidWiTile(a5),a2
	add.w	d1,a2
	add.w	d1,d1
	add.w	d1,d1
	add.w	d1,a3
	movem.l	d3/a0-a1/a4,-(sp)
	move.l	VidBgWiColorL(a5),a0
	move.l	VidBgWiColorH(a5),a1
	move.l	GrayCopyTable(a5),a4
	bsr	VidConvertBgWi
	movem.l	(sp)+,d3/a0-a1/a4

MrmwQuit
	jmp	(a6)

;--------------------------------------------------------------------------
;void MemRamMapBgWi(data.b,addr.w)
;---------------------------------
;	Write into background and window tile map.
MemRamMapBgWi
	cmp.b	(a2,d1.w),d0
	beq	MrmbwWiQuit		;no change?

	move.l	VidBgWiTileLow(a5),a3
	move.b	d0,(a2,d1.w)		;store tile number in map
	bpl.s	MrmbwTileOK
	move.l	VidBgWiTileHigh(a5),a3
MrmbwTileOK
	add.w	d1,d1
	and.w	#2*MAP_WRAP*MAP_WRAP*2-1,d1
	move.l	VidMapIndices(a5),a2
	moveq	#0,d2
	move.b	d0,d2
	move.w	d2,d0
	add.w	d2,d2
	add.w	d2,d2
	move.w	d2,(a2,d1.w)		;store precalculated index

	move.w	d0,-(sp)
	move.w	d2,d1
	add.w	d2,d2
	add.w	d2,d2
	add.w	d2,a3			;prepare some values

	move.w	VidBgScxPixel(a5),d2
	move.l	VidBgUsed(a5),a2
	cmp.b	(a2,d0.w),d2
	beq.s	MrmbwBgQuit		;already converted?
	move.b	d2,(a2,d0.w)

	move.l	VidBgTile(a5),a2
	add.w	d1,a2
	movem.l	d3/a0-a1/a4,-(sp)
	move.l	VidBgWiColorL(a5),a0
	move.l	VidBgWiColorH(a5),a1
	move.l	GrayCopyTable(a5),a4
	bsr	VidConvertBgWi		;convert for background
	lea	-8*2(a3),a3
	movem.l	(sp)+,d3/a0-a1/a4
MrmbwBgQuit
	move.w	(sp)+,d0

	move.w	VidWiWxPixel(a5),d2
	move.l	VidWiUsed(a5),a2
	cmp.b	(a2,d0.w),d2
	beq.s	MrmbwWiQuit		;already converted?
	move.b	d2,(a2,d0.w)

	add.w	d0,d0
	add.w	d0,d0
	move.l	VidWiTile(a5),a2
	add.w	d0,a2
	movem.l	d3/a0-a1/a4,-(sp)
	move.l	VidBgWiColorL(a5),a0
	move.l	VidBgWiColorH(a5),a1
	move.l	GrayCopyTable(a5),a4
	bsr	VidConvertBgWi		;convert for window
	movem.l	(sp)+,d3/a0-a1/a4
MrmbwWiQuit
	jmp	(a6)

;--------------------------------------------------------------------------
;void MemRamOam(data.b,addr.w)
;-----------------------------
;	Write into object attribute memory (oam).
MemRamOam
	pea	(a6)			;save return address

	moveq	#$fc-$100,d2
	and.w	d1,d2
	lea	(a2,d2.w),a3
	move.l	(a3),d2
	move.b	d0,(a2,d1.w)		;store byte in memory
	move.l	(a3),d0

	cmp.b	#40<<2,d1
	bcc.s	MroeQuit		;object # too big?

	;fall through to MemRamOamEntry

;--------------------------------------------------------------------------
;void MemRamOamEntry(olddata.l,newdata.l)
;----------------------------------------
;	New entry in oam.
MemRamOamEntry
mroeRegs	reg	d3-d5/a0/a1/a4
;d0.l	new value (already stored)
;d1.b	offset into oam
;d2.l	old value
;a2	(scratch)
;a3	pointer into oam
;a6	(scratch)

	eor.l	d0,d2
	beq.s	MroeQuit		;no change?

	movem.l	mroeRegs,-(sp)
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	check changes in y or x
	move.l	#$ffff7f,d5
	and.l	d2,d5			;save changes in x/tile/flags

	move.l	VidObFirst(a5),a6
	move.w	#$00fc,d0
	and.w	d1,d0
	lsl.w	#4-1,d0
	move.w	d0,d1
	lsl.w	#3,d0
	sub.w	d1,d0
	add.w	#SCR_HEIGHT*8*2*2,d0
	lea	(a6,d0.w),a2		;prepare some values

	eor.w	d5,d2
	bne.s	MroeXYLink
	swap	d2
	tst.w	d2
	beq.s	MroeXYQuit		;no change?
MroeXYLink
	bsr	VidObLink		;re-link object
MroeXYQuit
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	check changes in x, tile #, or flags
	tst.l	d5
	beq.s	MroeXTFQuit		;no change?

	move.w	(a3)+,d4
	move.w	VidObTilemask(a5),d3
	and.b	(a3)+,d3
	move.b	(a3)+,d0
	lsl.w	#3+1,d3
	move.l	BASE_RAM_VID(a4),a3
	lea	$8000-$10000(a3),a3
	add.w	d3,a3
	move.w	VidObCount(a5),d3
	moveq	#$10,d2
	and.b	d0,d2
	bne.s	MroeXTFObp1
	move.l	VidObColor0L(a5),a0
	move.l	VidObColor0H(a5),a1
	bra.s	MroeXTFConv
MroeXTFObp1
	move.l	VidObColor1L(a5),a0
	move.l	VidObColor1H(a5),a1
MroeXTFConv
	move.l	GrayCopyTable(a5),a4
	bsr	VidConvertOb
MroeXTFQuit
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	movem.l	(sp)+,mroeRegs
MroeQuit
	rts

;=== The end... ===========================================================
