;**************************************************************************
;*  gray.asm
;*			Gray display routines for PalmBoy
;*
;*  2000-05-05	Bodo Wenzel	Got from full source v.0.4
;*  2000-07-26	Bodo Wenzel	Longer off time while mode switching
;*  2000-09-13	Bodo Wenzel	Now nearly full screen emulation
;*  2000-11-01	Till Harbaum	Added support for the Palm IIIc and friends
;*  2000-11-19	Bodo Wenzel	Clean-up of support for color devices
;**************************************************************************
;
;  (c)2000 Bodo Wenzel
;
;  This program is free software; you can redistribute it and/or modify
;  it under the terms of the GNU General Public License as published by
;  the Free Software Foundation; either version 2 of the License, or
;  (at your option) any later version.
; 
;  This program is distributed in the hope that it will be useful,
;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;  GNU General Public License for more details.
; 
;  You should have received a copy of the GNU General Public License
;  along with this program; if not, write to the Free Software
;  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
;**************************************************************************

;	IMPORTANT!
;	Please make sure that any referenced resource is included!
;	Unfortunately there is no automatic!

;=== NOTES ================================================================
;	This routines base on given conditions by the GB hardware and the
;	grayscale screen mode.
;	Pay attention: some code doesn't depend on the equates!

;=== Constants ============================================================

grayBytesPerBuffer	equ	SCR_HEIGHT*8*SCR_WIDTH*8/4

;=== I/O addresses ========================================================

LGPMR328	equ	$fffffa32	;gray palette mapping (68328)
LGPMREZ		equ	$fffffa33	;gray palette mapping (68EZ328)

;=== Global variables =====================================================

	data

global	GrayDisplayPtr.l		;pointer to display memory
global	GrayDisplayWidth.w		;offset between two pixel lines

global	GrayBufferPtr.l			;pointer to allocated memory
global	GrayCopyTable.l			;pointer to conversion table
grayCopyTableSize	equ	$100*2

;=== Program code =========================================================

	code

;--------------------------------------------------------------------------
;Boolean GraySwitchOn(void)
;--------------------------
;	Try to switch into grayscale mode and set up some things. Return
;	success.
proc	GraySwitchOn()
local	depth.l
local	romV.l
local	extentX.w
local	extentY.w
beginproc
	move.l	#2,depth(a6)
	systrap	WinScreenMode(#winScreenModeSet.b,&NULL,&NULL,&depth(a6),&NULL)
	tst.w	d0
	bne	GsoReturnError

	systrap	FtrGet(#sysFtrCreator.l,#sysFtrNumROMVersion.w,&romV(a6))
	systrap	WinGetDrawWindow()
	cmp.l	#$03500000,romV(a6)
	bcs.s	GsoPreV35
	systrap	WinGetBitmap(a0.l)
	systrap	BmpGetBits(a0.l)
	move.l	a0,-(sp)
	systrap	WinGetDisplayExtent(&extentX(a6),&extentY(a6))
	move.l	(sp)+,a0
	move.w	extentX(a6),d0
	bra.s	GsoStoreAddress
GsoPreV35
	move.w	WindowType.displayWidth(a0),d0
	move.l	WindowType.displayAddr(a0),a0
GsoStoreAddress
	move.l	a0,GrayDisplayPtr(a5)
	lsr.w	#3-1,d0
	move.w	d0,GrayDisplayWidth(a5)

	call	GetFtrNumProcessor()
	cmp.l	#sysFtrNumProcessor328,d0
	beq.s	Gso328
	cmp.l	#sysFtrNumProcessorEZ,d0
	bne.s	GsoReturnOK
	move.b	#$75,LGPMREZ.w		;tested on Palm IIIx
	bra.s	GsoReturnOK
Gso328
	move.w	#$2073,LGPMR328.w	;tested on Palm III

GsoReturnOK
	moveq	#true,d0
	bra.s	GsoReturn

GsoReturnError
	moveq	#false,d0
GsoReturn
endproc

;--------------------------------------------------------------------------
;void GraySwitchOff(void)
;------------------------
;	Switch back to default screen mode.
proc	GraySwitchOff()
beginproc
	systrap	WinScreenMode(#winScreenModeSetToDefaults.b,&NULL,&NULL,&NULL,&NULL)
endproc

;--------------------------------------------------------------------------
;Boolean GrayInit(DmOpenRef dbP)
;-------------------------------
;	Get memory for the conversion table and for the GB screen (in the
;	given database). If successful create the table. Then return true
;	otherwise false.
proc	GrayInit(dbP.l)
local	recId.w
beginproc
	clr.l	GrayCopyTable(a5)
	clr.l	GrayBufferPtr(a5)

	move.w	#dmMaxRecordIndex,recId(a6)
	systrap	DmNewRecord(dbP(a6).l,&recId(a6),#grayCopyTableSize.l)
	move.l	a0,d0
	beq	GiReturnError

	systrap	MemHandleLock(a0.l)
	move.l	a0,GrayCopyTable(a5)

	systrap	MemSemaphoreReserve(#true.b)

	move.l	GrayCopyTable(a5),a0
	clr.b	d0
GiLoop
	move.b	d0,d1
	clr.w	d2
	add.b	d1,d1
	roxl.w	#2,d2
	add.b	d1,d1
	roxl.w	#2,d2
	add.b	d1,d1
	roxl.w	#2,d2
	add.b	d1,d1
	roxl.w	#2,d2
	add.b	d1,d1
	roxl.w	#2,d2
	add.b	d1,d1
	roxl.w	#2,d2
	add.b	d1,d1
	roxl.w	#2,d2
	add.b	d1,d1
	addx.w	d2,d2
	move.w	d2,(a0)+
	addq.b	#1,d0
	bne.s	GiLoop

	systrap	MemSemaphoreRelease(#true.b)

	move.w	#dmMaxRecordIndex,recId(a6)
	systrap	DmNewRecord(dbP(a6).l,&recId(a6),#grayBytesPerBuffer.l)
	move.l	a0,d0
	beq	GiReturnError

	systrap	MemHandleLock(a0.l)
	move.l	a0,GrayBufferPtr(a5)

	moveq	#true,d0
	bra.s	GiReturn

GiReturnError
	moveq	#false,d0
GiReturn
endproc

;--------------------------------------------------------------------------
;void GrayExit(void)
;-------------------
;	Release the used memory.
proc	GrayExit()
beginproc
	move.l	GrayCopyTable(a5),d0
	beq.s	GeNoCopyTable

	systrap	MemPtrUnlock(d0.l)
	clr.l	GrayCopyTable(a5)
GeNoCopyTable

	move.l	GrayBufferPtr(a5),d0
	beq.s	GeNoGrayBuffer

	systrap	MemPtrUnlock(d0.l)
	clr.l	GrayBufferPtr(a5)
GeNoGrayBuffer
endproc

;--------------------------------------------------------------------------
;void GrayClearScreen(void)
;--------------------------
;	Clear the screen buffer.
GrayClearScreen
;ATTENTION! Look carefully at the register usage!
	move.l	GrayBufferPtr(a5),a2
	move.w	#SCR_HEIGHT*8-1,d0
	move.w	GrayDisplayWidth(a5),d1
	sub.w	#SCR_WIDTH*8/4,d1
	moveq	#0,d2
GcsLoop
	move.l	d2,(a2)+
	move.l	d2,(a2)+
	move.l	d2,(a2)+
	move.l	d2,(a2)+
	move.l	d2,(a2)+
	move.l	d2,(a2)+
	move.l	d2,(a2)+
	move.l	d2,(a2)+
	move.l	d2,(a2)+
	move.l	d2,(a2)+
	add.w	d1,a2
	dbra	d0,GcsLoop
	rts

;--------------------------------------------------------------------------
;ptr GrayCopyBuffer(void)
;------------------------
;	Copies the buffer content into the display memory and returns the
;	address of the buffer.
GrayCopyBuffer
;ATTENTION! Look carefully at the register usage! Returns "ptr" in a2.
gcbRegs		reg	d3-d7/a0-a1/a4
gcbCopyRegs	reg	d2-d7/a2-a4/a6	;these have to be 10 registers
	movem.l	gcbRegs,-(sp)

	move.l	GrayBufferPtr(a5),a0
	move.l	GrayDisplayPtr(a5),a1
	move.w	#SCR_HEIGHT*8-1,d0
	move.w	GrayDisplayWidth(a5),d1
GcbLoop
	movem.l	(a0)+,gcbCopyRegs
	movem.l	gcbCopyRegs,(a1)
	add.w	d1,a1
	dbra	d0,GcbLoop

	move.l	GrayBufferPtr(a5),a2
	movem.l	(sp)+,gcbRegs
	rts

;=== The end... ===========================================================
