;**************************************************************************
;*  listform.asm
;*			List form routines for PalmBoy
;*
;*  2000-05-01	Bodo Wenzel	Got from full source v.0.4
;*  2000-05-06	Bodo Wenzel	First working version with built-in games
;*  2000-05-07	Bodo Wenzel	First working version
;*  2000-11-03	Bodo Wenzel	Removed bug if no games present
;*  2000-11-14	Bodo Wenzel	Support of color devices
;**************************************************************************
;
;  (c)2000 Bodo Wenzel
;
;  This program is free software; you can redistribute it and/or modify
;  it under the terms of the GNU General Public License as published by
;  the Free Software Foundation; either version 2 of the License, or
;  (at your option) any later version.
; 
;  This program is distributed in the hope that it will be useful,
;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;  GNU General Public License for more details.
; 
;  You should have received a copy of the GNU General Public License
;  along with this program; if not, write to the Free Software
;  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
;**************************************************************************

;	IMPORTANT!
;	Please make sure that any referenced resource is included!
;	Unfortunately there is no automatic!

;=== Constants ============================================================

listGameType	equ	'GAME'

;=== Global variables =====================================================

	data

global	ListDbList.l
global	ListAllRows.w
global	ListVisibleRows.w
global	ListTopRow.w

;=== Program code =========================================================

	code

;--------------------------------------------------------------------------
;void ListInit(void)
;-------------------
;	Summarize a list of games on the device. It's no error if none is
;	found. While looking for and sorting show a "wait" frame. The
;	memory of the list is locked.
;	At last switch to the list form.
proc	ListInit()
;a4	formP.l
local	dbCount.w
local	listH.l
beginproc
	clr.w	ListAllRows(a5)
	clr.l	ListDbList(a5)

	systrap	FrmInitForm(#formIdListSummarize.w)
	move.l	a0,a4
	systrap	FrmDrawForm(a4.l)

	systrap	SysCreateDataBaseList(#listGameType.l,#creator.l,&dbCount(a6),&listH(a6),#true.b)
	tst.b	d0
	beq.s	LiNoGames

	move.w	dbCount(a6),ListAllRows(a5)
	systrap	MemHandleLock(listH(a6).l)
	move.l	a0,ListDbList(a5)
LiNoGames

	clr.w	ListTopRow(a5)
	systrap	FrmEraseForm(a4.l)
	systrap	FrmGotoForm(#formIdList.w)
endproc

;--------------------------------------------------------------------------
;void ListExit(void)
;-------------------
;	If a list of games was built, release and free it.
proc	ListExit()
;a4	listH.l
beginproc
	move.l	ListDbList(a5),d0
	beq.s	LeReturn

	systrap	MemPtrRecoverHandle(ListDbList(a5).l)
	move.l	a0,a4
	systrap	MemHandleUnlock(a4.l)
	systrap	MemHandleFree(a4.l)
LeReturn
endproc

;--------------------------------------------------------------------------
;Boolean ListFormHandleEvent(EventPtr pEvent)
;--------------------------------------------
;	Process all user events in the list form. These are: set up the
;	form (build the list), manage scrolling, manage menus, and react on
;	commands. The list is made by a table because the Palm provided
;	lists always have borders!
proc	ListFormHandleEvent(pEvent.l)
;d5	formP.l
;d6	tableP.l
;d7	row.w,ret.b
local	r.RectangleType
local	min.w
local	max.w
local	size.w
beginproc
	move.l	pEvent(a6),a0
	moveq	#false,d0

	move.w	EventType.eType(a0),d1
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	cmp.w	#frmOpenEvent,d1
	bne	LfheTable

	systrap	FrmGetActiveForm()
	move.l	a0,d5
	systrap	FrmGetObjectIndex(d5.l,#tableIdList.w)
	systrap	FrmGetObjectPtr(d5.l,d0.w)
	move.l	a0,d6

	systrap	TblGetNumberOfRows(d6.l)
	move.w	d0,ListVisibleRows(a5)
	move.w	d0,d7
	bra.s	LfheFormInitGoon
LfheFormInitLoop
	systrap	TblSetItemStyle(d6.l,d7.w,#0.w,#customTableItem.b)
LfheFormInitGoon
	dbra	d7,LfheFormInitLoop

	systrap	TblSetCustomDrawProcedure(d6.l,#0.w,&ListDrawOneTitle(pc))
	systrap	TblSetColumnUsable(d6.l,#0.w,#true.b)

	call	ListUpdate(d5.l,d6.l)

	systrap	FrmSetControlGroupSelection(d5.l,#groupIdList.b,#pushbuttonIdRun.w)
	systrap	FrmDrawForm(d5.l)
	bra	LfheReturnOK
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
LfheTable
	cmp.w	#tblSelectEvent,d1
	bne	LfheScroll

	move.w	EventType.data+tblSelect.row(a0),d7
	add.w	ListTopRow(a5),d7

	systrap	FrmGetActiveForm()
	move.l	a0,d5
	systrap	FrmGetControlGroupSelection(d5.l,#groupIdList.b)
	and.w	#$ff,d0
	systrap	FrmGetObjectId(d5.l,d0.w)

	cmp.w	#pushbuttonIdRun,d0
	bne.s	LfheTable1

	call	RunGame(d7.w)
	bra.s	LfheTableOK

LfheTable1
	cmp.w	#pushbuttonIdInfo,d0
	bne.s	LfheTable2

	call	ListInfoShow(d7.w)

LfheTableOK
	move.l	pEvent(a6),a0
	move.l	EventType.data+tblSelect.pTable(a0),a0
	systrap	TblUnhighlightSelection(a0.l)
	bra	LfheReturnOK

LfheTable2
	cmp.w	#pushbuttonIdDelete,d0
	bne.s	LfheTableOK

	move.l	d5,-(sp)
	call	ListDeleteGame(d7.w)
	move.l	(sp)+,d5
	move.b	d0,d7
	move.l	pEvent(a6),a0
	move.l	EventType.data+tblSelect.pTable(a0),d6
	systrap	TblUnhighlightSelection(d6.l)
	tst.b	d7
	beq	LfheReturnOK

	call	ListUpdate(d5.l,d6.l)
	bra	LfheReturnOK
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
LfheScroll
	cmp.w	#sclRepeatEvent,d1
	bne.s	LfheKey

	move.w	EventType.data+sclRepeat.newValue(a0),ListTopRow(a5)
	moveq	#false,d7
LfheRedraw
	systrap	FrmGetActiveForm()
	move.l	a0,d5
	systrap	FrmGetObjectIndex(d5.l,#tableIdList.w)
	systrap	FrmGetObjectPtr(d5.l,d0.w)
	move.l	a0,d6
	systrap	TblMarkTableInvalid(d6.l)
	systrap	TblRedrawTable(d6.l)

	move.b	d7,d0
	bra	LfheReturn
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
LfheKey
	cmp.w	#keyDownEvent,d1
	bne	LfheMenu

	move.w	EventType.data+keyDown.chr(a0),d1
	move.w	ListTopRow(a5),d7

	cmp.w	#pageUpChr,d1
	bne.s	LfheKey1

	sub.w	ListVisibleRows(a5),d7
	bcc.s	LfheKeyOK
	clr.w	d7
	bra.s	LfheKeyOK
LfheKey1
	cmp.w	#pageDownChr,d1
	bne	LfheReturn

	add.w	ListVisibleRows(a5),d7
	move.w	ListAllRows(a5),d0
	sub.w	ListVisibleRows(a5),d0
	bcc.s	LfheKey2
	clr.w	d0
LfheKey2
	cmp.w	d0,d7
	bls.s	LfheKeyOK
	move.w	d0,d7
LfheKeyOK
	move.w	d7,ListTopRow(a5)

	systrap	FrmGetActiveForm()
	move.l	a0,d5
	systrap	FrmGetObjectIndex(d5.l,#scrollbarIdList.w)
	systrap	FrmGetObjectPtr(d5.l,d0.w)
	move.l	a0,d6
	systrap	SclGetScrollBar(d6.l,&r(a6),&min(a6),&max(a6),&size(a6))
	systrap	SclSetScrollBar(d6.l,d7.w,min(a6).w,max(a6).w,size(a6).w)

	moveq	#true,d7
	bra	LfheRedraw
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
LfheMenu
	cmp.w	#menuEvent,d1
	bne.s	LfheReturn

	move.w	EventType.data+menu.itemID(a0),d1
	cmp.w	#menuIdLOptionsAbout,d1
	bne.s	LfheReturn

	systrap	FrmAlert(#alertIdAbout.w)
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
LfheReturnOK
	moveq	#true,d0
LfheReturn
endproc

;--------------------------------------------------------------------------
;void ListInfoShow(Word no)
;--------------------------
;	The user likes to know more about the tapped game. So we give him
;	some information...
proc	ListInfoShow(no.w)
;a4	recH.l
;d4	itemP.l
;d5	dbId.l
;d6.msw	romSize.w
;d6.lsw	ramSize.w
local	sRom.4
local	sRam.4
beginproc
	call	ListGetItem(no(a6).w)
	move.l	a0,d4
	beq	LisReturn

	systrap	DmOpenDatabase(SysDBListItemType.cardNo(a0).w,SysDBListItemType.dbID(a0).l,#dmModeReadOnly.w)
	move.l	a0,d5
	beq	LisReturn

	systrap	DmQueryRecord(d5.l,#0.w)
	move.l	a0,a4
	systrap	MemHandleLock(a4.l)
	clr.l	d6
	move.b	$0148(a0),d6
	swap	d6
	move.b	$0149(a0),d6
	systrap	MemHandleUnlock(a4.l)
	systrap	DmCloseDatabase(d5.l)

	move.b	bLisRamSizes(pc,d6.w),d6
	systrap	StrPrintF(&sRam(a6),&szLisFormat(pc),d6.w)
	swap	d6
	move.w	#32,d0
	asl.w	d6,d0
	systrap	StrPrintF(&sRom(a6),&szLisFormat(pc),d0.w)
	move.l	d4,a0
	systrap	FrmCustomAlert(#alertIdInfo.w,&SysDBListItemType.name(a0),&sRom(a6),&sRam(a6))
LisReturn
endproc

bLisRamSizes
	dc.b	0,2,8,32
szLisFormat
	dc.b	'%3d',0
	align	2

;--------------------------------------------------------------------------
;Boolean ListDeleteGame(Word no)
;-------------------------------
;	No more playing this game? OK, but confirm it first! Return false
;	if the user cancels (used for updates) or errors occur. If the user
;	confirms, delete the game (return false on fail). Otherwise move
;	down the other entries in the list and update the variables.
proc	ListDeleteGame(no.w)
;a4	entryP.l
beginproc
	call	ListGetItem(no(a6).w)
	move.l	a0,a4
	move.l	a0,d0
	beq.s	LdgReturnNo

	systrap	FrmCustomAlert(#alertIdDelete.w,&SysDBListItemType.name(a4),&LdgDummy(pc),&LdgDummy(pc))
	tst.w	d0
	bne.s	LdgReturnNo

	systrap	DmDeleteDatabase(SysDBListItemType.cardNo(a4).w,SysDBListItemType.dbID(a4).l)
	tst.l	d0
	bne.s	LdgReturnNo		;errors silently ignored

	subq.w	#1,ListAllRows(a5)
	beq.s	LdgReturnOK

	move.w	ListAllRows(a5),d0
	sub.w	no(a6),d0
	beq.s	LdgReturnOK

	move.w	#sizeof(SysDBListItemType),d1
	mulu	d1,d0
	lea	(a4,d1.w),a0
	systrap	MemMove(a4.l,a0.l,d0.l)

LdgReturnOK
	moveq	#true,d0
	bra.s	LdgReturn
LdgReturnNo
	moveq	#false,d0
LdgReturn
endproc

LdgDummy
	dc.b	' ',0
	align	2

;--------------------------------------------------------------------------
;void ListUpdate(FormPtr formP,TablePtr tableP)
;----------------------------------------------
;	Update the scrollbar and the table.
proc	ListUpdate(formP.l,tableP.l)
;d7	row.w
beginproc
	systrap	FrmGetObjectIndex(formP(a6).l,#scrollbarIdList.w)
	systrap	FrmGetObjectPtr(formP(a6).l,d0.w)

	move.w	ListAllRows(a5),d0
	sub.w	ListVisibleRows(a5),d0
	bcc.s	LuMaxOK
	clr.w	d0			;'max' value for scrollbar
LuMaxOK

	move.w	ListTopRow(a5),d1
	sub.w	d0,d1
	bls.s	LuTopOK
	move.w	d0,ListTopRow(a5)	;top index for scrollbar
LuTopOK
	systrap	SclSetScrollBar(a0.l,ListTopRow(a5).w,#0.w,d0.w,ListVisibleRows(a5).w)

	systrap	TblMarkTableInvalid(tableP(a6).l)

	move.w	ListVisibleRows(a5),d7
	bra.s	LuSetupRowsNext
LuSetupRowsLoop
	cmp.w	ListAllRows(a5),d7
	scs	d1
	systrap	TblSetRowUsable(tableP(a6).l,d7.w,d1.b)
LuSetupRowsNext
	dbra	d7,LuSetupRowsLoop

	systrap	TblRedrawTable(tableP(a6).l)
endproc

;--------------------------------------------------------------------------
;SysDBListItemTypePtr ListGetItem(Word no)
;-----------------------------------------
;	Return a locked pointer to the given entry in the game list, NULL
;	if error.
proc	ListGetItem(no.w)
beginproc
	clr.l	d0
	move.w	ListAllRows(a5),d0
	beq.s	LgiReturn
	move.l	ListDbList(a5),d0
	beq.s	LgiReturn

	move.w	no(a6),d1
	mulu	#sizeof(SysDBListItemType),d1
	add.l	d1,d0
LgiReturn
	move.l	d0,a0
endproc

;--------------------------------------------------------------------------
;void ListDrawOneTitle(VoidPtr table,Word row,Word col,RectanglePtr bounds)
;--------------------------------------------------------------------------
;	Depending on the row parameter the correct string is found and
;	displayed. This function will not be called for unused rows!
proc	ListDrawOneTitle(table.l,row.w,col.w,bounds.l)
beginproc
	move.w	row(a6),d0
	add.w	ListTopRow(a5),d0
	call	ListGetItem(d0.w)
	move.l	a0,d0
	beq.s	LdotReturn

	lea	SysDBListItemType.name(a0),a0
	move.l	a0,-(sp)
	systrap	StrLen(a0.l)
	move.l	(sp)+,a0
	move.l	bounds(a6),a1
	systrap	WinDrawChars(a0.l,d0.w,RectangleType.topLeft+PointType.x(a1).w,RectangleType.topLeft+PointType.y(a1).w)
LdotReturn
endproc

;=== The end... ===========================================================
