;**************************************************************************
;*  palmboy.asm
;*			Nintendo Gameboy Emulator for
;*			Palm Computing Platform
;*
;*  2000-03-28	Bodo Wenzel	v.0.0, Creation of user interface without
;*				emulator mock-up and gray image
;*  2000-04-15	Bodo Wenzel	v.0.1, Creation of user interface without
;*				gray image
;*  2000-04-19	Bodo Wenzel	v.0.2, Creation of user interface
;*  2000-04-25	Bodo Wenzel	v.0.3, Errors fixed with Palm IIIC, V
;*  2000-05-01	Bodo Wenzel	v.0.4, Quick'n'dirty screen conversion,
;*				splitting in modules, and usage of more
;*				registers
;*  2000-05-06	Bodo Wenzel	v.0.9, First working version with built-in
;*				games
;*  2000-05-07	Bodo Wenzel	v.1.0, First working version
;*  2000-09-11	Bodo Wenzel	v.2.0, Nearly full screen emulation
;*  2000-09-30	Bodo Wenzel	v.2.1, Scroller in GB code
;*  2000-10-16	Bodo Wenzel	v.3.0, Interrupts and timer
;*  2000-11-03	Bodo Wenzel	v.3.0a, No more "Visor crash"
;*  2000-11-18	Bodo Wenzel	v.3.1, Support of color devices
;**************************************************************************
;
;  (c)2000 Bodo Wenzel
;
;  This program is free software; you can redistribute it and/or modify
;  it under the terms of the GNU General Public License as published by
;  the Free Software Foundation; either version 2 of the License, or
;  (at your option) any later version.
; 
;  This program is distributed in the hope that it will be useful,
;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;  GNU General Public License for more details.
; 
;  You should have received a copy of the GNU General Public License
;  along with this program; if not, write to the Free Software
;  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
;**************************************************************************

;	IMPORTANT!
;	Please make sure that any referenced resource is included!
;	Unfortunately there is no automatic!

;=== The name of the application ==========================================

creator	equ	'PBoy'
	appl	"PalmBoy",creator

;=== Includes =============================================================

	include	"pilot.inc"
	include	"startup.inc"	;standards Pila

	include	"resource.inc"	;resource IDs
	include	"constant.s"	;emulator constants

	include	"listform.asm"	;list form routines
	include	"runmode.asm"	;run mode and CPU emulation
	include	"emuvideo.asm"	;video emulation
	include	"emumem.asm"	;memory emulation
	include	"gray.asm"	;gray display routines
	include	"resource.asm"	;UI resource code

;=== Debugging ============================================================
;	Yes, I can't program correctly in the first try. The delivered
;	program will not contain this code.

;	include	"debug.asm"

;=== Global variables =====================================================

	data

global	MemoryDbId.l
global	GameNo.w

;=== Program code =========================================================

	code

;--------------------------------------------------------------------------
;DWord PilotMain(Word cmd,Ptr cmdPBP,Word launchflags)
;-----------------------------------------------------
;	Check launch code, and ROM version. Get memory needed.
;	Initialize everything.
;	Then the standard event loop will run until "appStopEvent".
proc	PilotMain(cmd.w,cmdPBP.l,launchFlags.w)
local	event.EventType
local	err.w
beginproc
	cmp.w	#sysAppLaunchCmdNormalLaunch,cmd(a6)
	bne	PmReturnOK

	call	RomVersionCompatible(#$03003000.l,launchFlags(a6).w)
	tst.w	d0
	bne	PmReturn

	call	GraySwitchOn()
	tst.b	d0
	bne.s	PmGoon1

	systrap	FrmAlert(#alertIdIncompatibleScreen.w)
	move.w	#sysErrParamErr,d0
	bra	PmReturn
PmGoon1

	call	AllMemoryInit()
	tst.b	d0
	bne.s	PmGoon2

	systrap	FrmAlert(#alertIdNoFreeRAM.w)
	move.w	#sysErrNoFreeRAM,d0
	bra	PmReturn
PmGoon2

	call	ListInit()
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
PmEventLoop
	systrap	EvtGetEvent(&event(a6),#evtWaitForever.l)

	systrap	SysHandleEvent(&event(a6))
	tst.b	d0
	bne.s	PmEventDone

	systrap	MenuHandleEvent(&NULL,&event(a6),&err(a6))
	tst.b	d0
	bne.s	PmEventDone

	call	ApplicationHandleEvent(&event(a6))
	tst.b	d0
	bne.s	PmEventDone

	systrap	FrmDispatchEvent(&event(a6))
PmEventDone
	cmp.w	#appStopEvent,event+EventType.eType(a6)
	bne.s	PmEventLoop
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	call	ListExit()
	call	AllMemoryExit()

	call	GraySwitchOff()

PmReturnOK
	moveq	#0,d0
PmReturn
endproc

;--------------------------------------------------------------------------
;Boolean ApplicationHandleEvent(EventPtr pEvent)
;-----------------------------------------------
;	Process events for form initializing. When a new form is loaded,
;	set up the appropriate event handler.
proc	ApplicationHandleEvent(pEvent.l)
;d4	formId.w
;d5	formP.l
beginproc
	move.l	pEvent(a6),a0

	moveq	#false,d0
	cmp.w	#frmLoadEvent,EventType.eType(a0)
	bne.s	AheReturn

	move.w	EventType.data+frmLoad.formID(a0),d4
	systrap	FrmInitForm(d4.w)
	move.l	a0,d5
	systrap	FrmSetActiveForm(d5.l)

	lea	ListFormHandleEvent(pc),a1
	cmp.w	#formIdList,d4
	beq.s	AheSetEventHandler

	lea	RunFormHandleEvent(pc),a1
	cmp.w	#formIdRun,d4
	bne.s	AheReturnOK

AheSetEventHandler
	systrap	FrmSetEventHandler(d5.l,a1.l)

AheReturnOK
	moveq	#true,d0
AheReturn
endproc

;--------------------------------------------------------------------------
;Word RomVersionCompatible(DWord requiredVersion,Word launchflags)
;-----------------------------------------------------------------
;	Compare the actual ROM version with the required one. If it's not
;	good enough, return with error code. But if it's OS version 2.0
;	start the default application - otherwise this app will be started
;	over and over again!
proc	RomVersionCompatible(requiredVersion.l,launchFlags.w)
local	romV.l
local	sStt.DmSearchStateType
local	cardNo.w
local	dbID.l
beginproc
	systrap	FtrGet(#sysFtrCreator.l,#sysFtrNumROMVersion.w,&romV(a6))
	moveq	#0,d0
	move.l	romV(a6),d1
	cmp.l	requiredVersion(a6),d1
	bcc.s	RvcReturn

	move.w	launchFlags(a6),d0
	and.w	#sysAppLaunchFlagNewGlobals|sysAppLaunchFlagUIApp,d0
	cmp.w	#sysAppLaunchFlagNewGlobals|sysAppLaunchFlagUIApp,d0
	bne.s	RvcIncompatible

	systrap	FrmAlert(#alertIdIncompatibleRom.w)

	cmp.l	#$02000000,romV(a6)
	bcc.s	RvcIncompatible

	systrap	DmGetNextDatabaseByTypeCreator(#true.b,&sStt(a6),#sysFileTApplication.l,#sysFileCDefaultApp.l,#true.b,&cardNo(a6),&dbID(a6))
	tst.l	dbID(a6)
	beq.s	RvcIncompatible

	systrap	SysUIAppSwitch(cardNo(a6).w,dbID(a6).l,#sysAppLaunchCmdNormalLaunch.w,&NULL)

RvcIncompatible
	move.w	#sysErrRomIncompatible,d0
RvcReturn
endproc

;--------------------------------------------------------------------------
;DWord GetFtrNumProcessor(void)
;------------------------------
;	Check for OS version 3.1 first - earlier ones use always an 68328.
;	Otherwise get the actual processor feature number.
proc	GetFtrNumProcessor()
local	ftr.l
beginproc
	systrap	FtrGet(#sysFtrCreator.l,#sysFtrNumROMVersion.w,&ftr(a6))
	move.l	#sysFtrNumProcessor328,d0
	cmp.l	#$03103000,ftr(a6)
	bcs.s	GfnpQuit

	systrap	FtrGet(#sysFtrCreator.l,#sysFtrNumProcessorID.w,&ftr(a6))
	move.l	ftr(a6),d0
GfnpQuit
	and.l	#sysFtrNumProcessorMask,d0
endproc

;--------------------------------------------------------------------------
;	Memory database constants

DbName
	dc.b	'PalmBoy-Database',0
	align	2
dbType	equ	'DATA'
;	NOTE:
;	The database is created on card #0. For the moment I'll do no
;	effort to change this behaviour.
dbCard	equ	0

;--------------------------------------------------------------------------
;Boolean AllMemoryInit(void)
;------------------------
;	Create the database for all bigger memory requirements. If it
;	exists already, the creation error is silently ignored.
;	Call the initializing routines (they save the locked memory
;	pointers to the records). If one of them fails, return false.
proc	AllMemoryInit()
beginproc
	systrap	DmCreateDatabase(#dbCard.w,&DbName(pc),#creator.l,#dbType.l,#false.b)
	systrap	DmFindDatabase(#dbCard.w,&DbName(pc))
	tst.l	d0
	beq.s	AmiDbError
	systrap	DmOpenDatabase(#dbCard.w,d0.l,#dmModeReadWrite.w)
	move.l	a0,MemoryDbId(a5)
	beq.s	AmiDbError

	call	GrayInit(MemoryDbId(a5).l)
	move.b	d0,-(sp)
	call	VidInit(MemoryDbId(a5).l)
	and.b	d0,(sp)
	call	MemInit(MemoryDbId(a5).l)
	and.b	(sp)+,d0
	bne.s	AmiReturn

	call	AllMemoryExit()
AmiDbError
	moveq	#false,d0
AmiReturn
endproc

;--------------------------------------------------------------------------
;void AllMemoryExit(void)
;---------------------
;	Call all release routines. Then close and delete the database.
proc	AllMemoryExit()
local	locId.l
beginproc
	call	MemExit()
	call	VidExit()
	call	GrayExit()

	systrap	DmOpenDatabaseInfo(MemoryDbId(a5).l,&locId(a6),&NULL,&NULL,&NULL,&NULL)
	systrap	DmCloseDatabase(MemoryDbId(a5).l)
	systrap	DmDeleteDatabase(#dbCard.w,locId(a6).l)
endproc

;=== System resources =====================================================

	res	'pref',1
	dc.w	sysAppLaunchFlagNewStack|sysAppLaunchFlagNewGlobals|sysAppLaunchFlagUIApp|sysAppLaunchFlagSubCall
	dc.l	$1000		;stack size
	dc.l	$1000		;heap size

	res	'tver',1
	dc.b	'3.1',0

	res	'tAIB',1000,"tAIB03e8.bin"	;ICON
	res	'tAIB',1001,"tAIB03e9.bin"	;SMALLICON

;=== The end... ===========================================================
	end
