;**************************************************************************
;*  runmode.asm
;*			Run mode routines for PalmBoy
;*
;*  2000-05-01	Bodo Wenzel	Got from full source v.0.4
;*  2000-05-06	Bodo Wenzel	First working version with built-in games
;*  2000-05-07	Bodo Wenzel	First working version
;*  2000-08-02	Bodo Wenzel	Now nearly full screen emulation
;*  2000-09-13	Bodo Wenzel	Some things in separated files
;*  2000-09-30	Bodo Wenzel	Scroller in GB code
;*  2000-10-22	Bodo Wenzel	Interrupts and timer
;*  2000-11-03	Bodo Wenzel	Request of EasterEgg corrected
;*  2000-11-21	Bodo Wenzel	Support of color devices and
;*				better joypad emulation
;**************************************************************************
;
;  (c)2000 Bodo Wenzel
;
;  This program is free software; you can redistribute it and/or modify
;  it under the terms of the GNU General Public License as published by
;  the Free Software Foundation; either version 2 of the License, or
;  (at your option) any later version.
; 
;  This program is distributed in the hope that it will be useful,
;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;  GNU General Public License for more details.
; 
;  You should have received a copy of the GNU General Public License
;  along with this program; if not, write to the Free Software
;  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
;**************************************************************************

;	IMPORTANT!
;	Please make sure that any referenced resource is included!
;	Unfortunately there is no automatic!

;=== Constants ============================================================

runLineBegX	equ	0		;coordinates for separator
runLineEndX	equ	159		;coordinates for separator
runLinePosY	equ	144

;	Key bits are inverted: released/pressed = 1/0
gbCursorMask	equ	$ef		;bits: 1 1 1 0 down up left right
gbControlMask	equ	$df		;bits: 1 1 0 1 start select b a

;=== Global variables =====================================================

	data

global	RunForm.l			;form variables
global	RunSelectButton.w
global	RunSelectBounds.RectangleType
global	RunStartButton.w
global	RunStartBounds.RectangleType

global	RunEmulator.l			;address of emulator kernel

global	RunResult.l			;CPU state
global	RunCpuPc.w
global	RunCpuCode.6
global	RunCpuSp.w
global	RunCpuStack.6
global	RunCpuIE.b
	align	2
global	RunCpuAF.w
global	RunCpuBC.w
global	RunCpuDE.w
global	RunCpuHL.w

;=== Emulator code ========================================================
;	HERE IT IS! This is the emulator! The code is generated externally
;	by "STemBoy" (Atari ST based emulator) and filled in a resource.
;	It's nearly 64 KByte big...

rsrcCode	equ	'code'
rsrcIdEmu	equ	1000

	res	rsrcCode,rsrcIdEmu,"gbemu.bin"

;=== Program code =========================================================

	code

;--------------------------------------------------------------------------
;void RunGame(Word no)
;---------------------
;	Just store the game number and go to run form. This form provides
;	two buttons for "SELECT" and "START" below the GB screen, but it's
;	just a dummy to be used for detection and copy.
proc	RunGame(no.w)
beginproc
	move.w	no(a6),GameNo(a5)

	systrap	FrmGotoForm(#formIdRun.w)
endproc

;--------------------------------------------------------------------------
;Boolean RunFormHandleEvent(EventPtr pEvent)
;-------------------------------------------
;	When the run form is set up, we just wait for the "frmOpenEvent".
;	Then we'll draw the form and set up everything for the emulation.
;	After the emulation returns, we restore the original state and
;	go back to the list form.
proc	RunFormHandleEvent(pEvent.l)
;d0-d7/a0-a4/a6 used also by emulator...
;a4	(scratch).l
;d4	rsrcP.l
;d5	bitmapP.l, memP.l
;d6	y.w, offset.l, dirty.b
;d7	ticks.l, index.w
local	easterEgg.w
local	keyOldMask.l
local	emuRsrcP.l
local	dbP.l
local	numRecs.w
local	recH.128		;32*l	;same size as MemRomPages
beginproc
RunMyKeys	equ	keyBitPageUp|keyBitPageDown|keyBitHard1|keyBitHard2|keyBitHard3|keyBitHard4
RunMyKeyMask	equ	keyBitsAll&~RunMyKeys
	move.l	pEvent(a6),a0
	moveq	#false,d0

	cmp.w	#frmOpenEvent,EventType.eType(a0)
	bne	RfheQuit

	systrap	PrefGetPreference(#prefAllowEasterEggs.b)
	tst.l	d0
	sne	d0
	move.b	d0,easterEgg(a6)

	systrap	EvtEnableGraffiti(#false.b)
	systrap	KeySetMask(#RunMyKeyMask.l)
	move.l	d0,keyOldMask(a6)

	systrap	FrmGetActiveForm()
	move.l	a0,RunForm(a5)
	systrap	FrmDrawForm(a0.l)

	systrap	WinDrawLine(#runLineBegX.w,#runLinePosY.w,#runLineEndX.w,#runLinePosY.w)

	systrap	FrmGetObjectIndex(RunForm(a5).l,#buttonIdSelect.w)
	move.w	d0,RunSelectButton(a5)
	systrap	FrmGetObjectBounds(RunForm(a5).l,d0.w,&RunSelectBounds(a5))
	systrap	FrmGetObjectIndex(RunForm(a5).l,#buttonIdStart.w)
	move.w	d0,RunStartButton(a5)
	systrap	FrmGetObjectBounds(RunForm(a5).l,d0.w,&RunStartBounds(a5))
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	move.l	#USER_STOP,RunResult(a5)	;prevent writing a note

	call	ListGetItem(GameNo(a5).w)
	move.l	a0,d0
	beq	RfheReturn

	systrap	DmOpenDatabase(SysDBListItemType.cardNo(a0).w,SysDBListItemType.dbID(a0).l,#dmModeReadWrite.w)
	move.l	a0,dbP(a6)
	beq	RfheReturn

	systrap	DmNumRecords(dbP(a6).l)
	move.w	d0,numRecs(a6)
	beq	RfheCloseDb
	cmp.w	#32,d0
	bhi	RfheCloseDb

	systrap	DmGet1Resource(#rsrcCode.l,#rsrcIdEmu.w)
	move.l	a0,emuRsrcP(a6)
	beq	RfheNoEmulator

	systrap	MemHandleLock(emuRsrcP(a6).l)
	move.l	#$8000,d0
	add.l	d0,a0
	move.l	a0,RunEmulator(a5)

	clr.w	d7
	lea	recH(a6),a4
RfheGetRomBanks
	systrap	DmGetRecord(dbP(a6).l,d7.w)
	move.l	a0,(a4)+
	addq.w	#1,d7
	cmp.w	numRecs(a6),d7
	bne.s	RfheGetRomBanks		;the sequence is important!

	lea	recH(a6),a4
	move.l	(a4)+,d0
	systrap	MemHandleLock(d0.l)
	lea	MemRomPages(a5),a1
	move.l	a0,(a1)+
	move.l	a0,(a1)+

	move.w	numRecs(a6),d7
	addq.w	#1,d7
	and.w	#$fffe,d7		;round up to next even count
	move.w	#2,d6
RfheLockRomBanks
	cmp.w	d7,d6
	beq.s	RfheLockRomBanksOK
	move.l	a1,-(sp)
	move.l	(a4)+,d0
	systrap	MemHandleLock(d0.l)
	move.l	(sp)+,a1
	lea	-$4000(a0),a0
	move.l	a0,(a1)+
	addq.w	#1,d6
	bra.s	RfheLockRomBanks
RfheLockRomBanksOK
	lea	MemRomPages(a5),a0
RfheMirrorRomBanks
	cmp.w	#32,d6
	beq.s	RfheMirrorRomBanksOK
	move.l	(a0)+,(a1)+
	addq.w	#1,d6
	bra.s	RfheMirrorRomBanks
RfheMirrorRomBanksOK

	clr.l	d6
	move.l	MemRam(a5),d5
	add.l	#MemRamType.extA000,d5

	move.w	numRecs(a6),d7
	btst	#0,d7
	bne.s	RfheSetRamBanks

	move.l	(a4),d0
	systrap	MemHandleLock(d0.l)
	move.l	a0,d5
	move.l	MemRomPages(a5),a2
	cmp.b	#3,$0149(a2)
	bne.s	RfheSetRamBanks		;only "3" needs banking

	move.l	#$2000,d6
RfheSetRamBanks
	add.l	#($10000-$a000),d5
	lea	MemRamPages(a5),a1
	move.l	d5,(a1)+
	add.l	d6,d5
	move.l	d5,(a1)+
	add.l	d6,d5
	move.l	d5,(a1)+
	add.l	d6,d5
	move.l	d5,(a1)+
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	systrap	DmSet(MemRam(a5).l,&0,#memRamSize.l,#0.b)
	systrap	DmSet(MemRam(a5).l,#MemRamType.vid8000+$1800.l,#$800.l,#$ff.b)

	systrap	MemSemaphoreReserve(#true.b)

	bsr	GrayClearScreen

	move.l	RunEmulator(a5),a4

	move.l	MemRam(a5),a0
	move.l	#MemRamType.vid8000+($10000-$8000),d0
	lea	(a0,d0.l),a1
	move.l	a1,BASE_RAM_VID(a4)
	move.l	MemRamPages(a5),BASE_RAM_EXT(a4)
	move.l	#MemRamType.intC000+($10000-$c000),d0
	lea	(a0,d0.l),a1
	move.l	a1,BASE_RAM_INT(a4)
	move.l	#MemRamType.oamFe00+($10000-$fe00),d0
	lea	(a0,d0.l),a1
	move.l	a1,BASE_RAM_OAM(a4)
	move.l	#MemRamType.cpuFf80+($10000-$ff80),d0
	lea	(a0,d0.l),a1
	move.l	a1,BASE_RAM_CPU(a4)

	lea	RunOpcodeStop(pc),a1
	move.l	a1,V_OPCODE_STOP+2(a4)
	move.b	#$80,LCD_LCDC(a4)
	move.w	#-1, LCD_I_LY(a4)
	move.w	#0,  LCD_I_WY(a4)
	move.b	#$00,LCD_BGP(a4)
	move.b	#$00,LCD_OBP0(a4)
	move.b	#$00,LCD_OBP1(a4)
	lea	VidIoWrLcdc(pc),a1
	move.l	a1,V_IO_WR_LCDC+2(a4)
	lea	VidIoRdStat(pc),a1
	move.l	a1,V_IO_RD_STAT+2(a4)
	lea	VidIoWrStat(pc),a1
	move.l	a1,V_IO_WR_STAT+2(a4)
	lea	VidIoWrScy(pc),a1
	move.l	a1,V_IO_WR_SCY+2(a4)
	lea	VidIoWrScx(pc),a1
	move.l	a1,V_IO_WR_SCX+2(a4)
	lea	VidIoWrLyc(pc),a1
	move.l	a1,V_IO_WR_LYC+2(a4)
	lea	VidIoWrDma(pc),a1
	move.l	a1,V_IO_WR_DMA+2(a4)
	lea	VidIoWrBgp(pc),a1
	move.l	a1,V_IO_WR_BGP+2(a4)
	lea	VidIoWrObp0(pc),a1
	move.l	a1,V_IO_WR_OBP0+2(a4)
	lea	VidIoWrObp1(pc),a1
	move.l	a1,V_IO_WR_OBP1+2(a4)
	lea	VidIoWrWy(pc),a1
	move.l	a1,V_IO_WR_WY+2(a4)
	lea	VidIoWrWx(pc),a1
	move.l	a1,V_IO_WR_WX+2(a4)
	lea	MemRomSelect(pc),a1
	move.l	a1,V_ROM_SELECT+2(a4)
	lea	MemRamSelect(pc),a1
	move.l	a1,V_RAM_SELECT+2(a4)

	move.l	a6,-(sp)		;save pointer to local variables

	lea	CODE_RESET(a4),a0
	move.l	a0,BASE_ROM_0(a4)
	move.l	a0,CPU_STATE_PC_B(a4)
	move.w	#32000,d6		;(more than enough)
	jsr	(a4)			;call the emulator for reset code

	clr.w	d6			;immediate "support" of emulator
	move.b	d6,INT_IME(a4)
	move.b	d6,INT_IE(a4)
	move.b	d6,INT_IF(a4)
	move.b	d6,INT_IM(a4)		;interrupts disabled

	move.b	#gbCursorMask&gbControlMask,JOYPAD_BOTH(a4)
	move.b	#gbControlMask,JOYPAD_CMD(a4)
	move.b	#gbCursorMask,JOYPAD_DIR(a4)
	move.b	#$ff,JOYPAD_NONE(a4)

	move.l	(sp),a6
	tst.b	easterEgg(a6)
	beq.s	RfheNoScroller		;no scroller?

	lea	CODE_SCROLLER(a4),a0
	jsr	(a4)			;call the emulator for scroller
	cmp.w	#USER_STOP,d0
	beq.s	RfheEmulatorStop
	cmp.w	#UNKNOWN,d0
	bne.s	RfheEmulatorStop
RfheNoScroller

	move.l	MemRomPages(a5),a0
	move.l	a0,BASE_ROM_0(a4)
	move.l	a0,BASE_ROM_1(a4)
	move.l	a0,CPU_STATE_PC_B(a4)
	lea	$100(a0),a0		;start address of GB program

; *** HERE IS IT! THE EMULATOR STARTS EXECUTION... ************************
	jsr	(a4)
; *************************************************************************

RfheEmulatorStop
	move.l	d0,RunResult(a5)

	move.l	a0,d0
	sub.l	CPU_STATE_PC_B(a4),d0
	move.w	d0,RunCpuPc(a5)

	move.l	a1,d0
	sub.l	CPU_STATE_SP_B(a4),d0
	move.w	d0,RunCpuSp(a5)

	move.b	INT_IME(a4),RunCpuIE(a5)
	move.b	d7,RunCpuAF+0(a5)
	swap	d6
	move.b	d6,RunCpuAF+1(a5)
	move.b	d5,RunCpuBC+1(a5)
	swap	d5
	move.b	d5,RunCpuBC+0(a5)
	move.b	d4,RunCpuDE+1(a5)
	swap	d4
	move.b	d4,RunCpuDE+0(a5)
	move.b	d3,RunCpuHL+1(a5)
	swap	d3
	move.b	d3,RunCpuHL+0(a5)	;store CPU state

	lea	RunCpuCode(a5),a2
	subq.l	#3,a0
	move.b	(a0)+,(a2)+
	move.b	(a0)+,(a2)+
	move.b	(a0)+,(a2)+
	move.b	(a0)+,(a2)+
	move.b	(a0)+,(a2)+
	move.b	(a0)+,(a2)+		;store code surroundings

	lea	RunCpuStack(a5),a2
	subq.l	#2,a1
	move.b	(a1)+,(a2)+
	move.b	(a1)+,(a2)+
	move.b	(a1)+,(a2)+
	move.b	(a1)+,(a2)+
	move.b	(a1)+,(a2)+
	move.b	(a1)+,(a2)+		;store stack surroundings

	move.l	(sp)+,a6		;restore pointer to local variables

	systrap	MemSemaphoreRelease(#true.b)
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	move.w	numRecs(a6),d7
	moveq	#false,d6
	btst	#0,d7
	bne.s	RfheReleaseRomBanks
	moveq	#true,d6
RfheReleaseRomBanks
	subq.w	#1,d7
	lea	recH(a6),a4
RfheReleaseRecords
	move.l	(a4)+,d0
	systrap	MemHandleUnlock(d0.l)
	systrap	DmReleaseRecord(dbP(a6).l,d7.w,d6.b)
	moveq	#false,d6
	dbra	d7,RfheReleaseRecords

	systrap	MemHandleUnlock(emuRsrcP(a6).l)
	systrap	DmReleaseResource(emuRsrcP(a6).l)

	bra.s	RfheCloseDb

RfheNoEmulator
	systrap	FrmAlert(#alertIdRunNoEmulator.w)

RfheCloseDb
	systrap	DmCloseDatabase(dbP(a6).l)

	move.l	RunResult(a5),d0
	cmp.w	#USER_STOP,d0
	beq.s	RfheReturn

	neg.w	d0
	call	RunErrorNote(d0.w,GameNo(a5).w)

RfheReturn
	systrap	KeySetMask(keyOldMask(a6).l)
	systrap	EvtEnableGraffiti(#true.b)

	systrap	FrmGotoForm(#formIdList.w)
	moveq	#true,d0
RfheQuit
endproc

;--------------------------------------------------------------------------
;Boolean RunUserPoll(void)
;-------------------------
;	This function is called to poll events. If the emulation has to
;	quit it returns true else false.
;
;	The hardkeys are scanned and converted. If no events are available
;	nothing else happens.
;	Pen events are checked against the bounds of the "SELECT" and
;	"START" buttons. The system gets this events only for the area
;	below the display. If the user moves from one button to the other
;	the first button remains selected.

proc	RunUserPoll()
local	jpCursor.b
local	jpControl.b
local	event.EventType
local	x.w
local	y.w
beginproc
	systrap	MemSemaphoreRelease(#true.b)
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	move.b	JOYPAD_CMD(a4),jpControl(a6)

	systrap	KeyCurrentState()
	clr.b	d1
	lsr.l	#2,d0			;up	= up
	roxr.b	#1,d1
	lsr.l	#1,d0			;down	= down
	roxr.b	#1,d1
	rol.b	#2,d1
	lsr.l	#1,d0			;hard1	= left
	roxl.b	#1,d1
	lsr.l	#1,d0			;hard2	= right
	roxl.b	#1,d1
	eor.b	#gbCursorMask,d1
	move.b	d1,jpCursor(a6)
	clr.b	d1
	lsr.l	#1,d0			;hard3	= b
	roxl.b	#1,d1
	lsr.l	#1,d0			;hard4	= a
	roxl.b	#1,d1
	or.b	#$03,jpControl(a6)
	eor.b	d1,jpControl(a6)

	systrap	EvtEventAvail()
	tst.b	d0
	beq	RupInputHandled
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	systrap	EvtGetEvent(&event(a6),#0.l)

	move.w	event+EventType.eType(a6),d1

	cmp.w	#nilEvent,d1
	beq	RupInputHandled
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	cmp.w	#penDownEvent,d1
	beq.s	RupPenDown
	cmp.w	#penMoveEvent,d1
	beq.s	RupPenDown
	cmp.w	#penUpEvent,d1
	bne	RupKey

	or.b	#$0c,jpControl(a6)
	systrap	FrmSetControlValue(RunForm(a5).l,RunSelectButton(a5).w,#false.w)
	systrap	FrmSetControlValue(RunForm(a5).l,RunStartButton(a5).w,#false.w)
	bra.s	RupPenQuit

RupPenDown
	move.w	event+EventType.screenX(a6),x(a6)
	move.w	event+EventType.screenY(a6),y(a6)
	systrap	WinDisplayToWindowPt(&x(a6),&y(a6))

	systrap	RctPtInRectangle(x(a6).w,y(a6).w,&RunSelectBounds(a5))
	tst.b	d0
	beq.s	RupPenDown2
	and.b	#~$04,jpControl(a6)
	systrap	FrmSetControlValue(RunForm(a5).l,RunSelectButton(a5).w,#true.w)
RupPenDown2
	systrap	RctPtInRectangle(x(a6).w,y(a6).w,&RunStartBounds(a5))
	tst.b	d0
	beq.s	RupPenQuit
	and.b	#~$08,jpControl(a6)
	systrap	FrmSetControlValue(RunForm(a5).l,RunStartButton(a5).w,#true.w)

RupPenQuit
	cmp.w	#160,event+EventType.screenY(a6)
	bge	RupSystemHandle

	bra	RupInputHandled
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
RupKey
	cmp.w	#keyDownEvent,d1
	bne.s	RupCheckAppStop

	move.w	event+EventType.data+keyDown.modifiers(a6),d1
	and.w	#commandKeyMask,d1
	beq.s	RupSystemHandle

	move.w	event+EventType.data+keyDown.chr(a6),d1

	cmp.w	#findChr,d1
	beq.s	RupSnapshot		;special meaning for "Find"

	cmp.w	#menuChr,d1		;special meaning for "Menu"
	bne.s	RupSystemHandle

	systrap	SndPlaySystemSound(#sndClick.b)
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
RupExitEmulation
	systrap	MemSemaphoreReserve(#true.b)

	moveq	#true,d0
	bra.s	RupQuit
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	Check for application stop. Then put the event back in the queue
;	and leave the emulation!
RupCheckAppStop
	cmp.w	#appStopEvent,d1
	bne.s	RupInputHandled

	systrap	EvtAddEventToQueue(&event(a6))
	bra.s	RupExitEmulation
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
RupSnapshot
	;no action yet...
	systrap	SndPlaySystemSound(#sndAlarm.b)
	bra.s	RupInputHandled
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	The system gets a chance to handle the event but the emulation
;	goes on.
RupSystemHandle
	systrap	SysHandleEvent(&event(a6))
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	Copy joypad values and check for interrupt.
RupInputHandled
	systrap	MemSemaphoreReserve(#true.b)

	move.l	JOYPAD_PTR(a4),a0
	move.b	(a0),d3

	move.b	jpCursor(a6),d0
	move.b	d0,JOYPAD_DIR(a4)
	move.b	jpControl(a6),d1
	move.b	d1,JOYPAD_CMD(a4)
	and.b	d1,d0
	move.b	d0,JOYPAD_BOTH(a4)

	move.b	(a0),d0
	not.b	d0
	and.b	d0,d3
	sne	d1
	and.b	#$10,d1
	or.b	d1,INT_IF(a4)

	moveq	#false,d0
RupQuit
endproc

;--------------------------------------------------------------------------
;void RunOpcodeStop(void)
;------------------------
;	Emulation of opcode $10 $00 (STOP).
RunOpcodeStop
	bsr	GrayClearScreen

	lea	INT_IM(a4),a2
	move.b	(a2),-(sp)
	move.b	#$00,(a2)
	and.b	#$ef,INT_IF(a4)
	pea	RosAbort(pc)		;set up for waiting
RosWait
	bsr	VidSupport
	btst	#4,INT_IF(a4)
	beq.s	RosWait			;wait for user

	addq.w	#4,sp
	move.b	(sp)+,INT_IM(a4)
	rts				;back to emulation
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
;	emulation has to stop
RosAbort
	move.b	(sp)+,INT_IM(a4)
	subq.w	#2,a0			;restore state for STOP

	addq.w	#4,sp			;ignore return to emulation
	rts

;--------------------------------------------------------------------------
;void RunErrorNote(Word errorNo,Word gameNo)
;-------------------------------------------
proc	RunErrorNote(errorNo.w,gameNo.w)
;d4	rsrcP.l,dbId.l
;d5	len.l
;d6	recordH.l
local	note.500			;should be enough space...
local	dt.DateTimeType
local	l1.52
local	l2.32
local	card.w
local	locId.l
local	attr.w
local	idx.w
beginproc
	systrap	TimGetSeconds()
	systrap	TimSecondsToDateTime(d0.l,&dt(a6))
	lea	dt+DateTimeType.day(a6),a0
	move.w	(a0)+,d0
	move.w	(a0)+,d1
	move.w	(a0),d2
	systrap	DateToAscii(d1.b,d0.b,d2.w,#dfYMDWithDashes.b,&l1(a6))
	lea	dt+DateTimeType.minute(a6),a0
	move.w	(a0)+,d0
	move.w	(a0),d1
	systrap	TimeToAscii(d1.b,d0.b,#tfColon24h.b,&l2(a6))

	move.l	#'tSTR',d0
	systrap	DmGet1Resource(d0.l,#stringIdRunNoteHeader.w)
	move.l	a0,d4
	systrap	MemHandleLock(d4.l)
	systrap	StrPrintF(&note(a6),a0.l,&l1(a6),&l2(a6))
	systrap	MemHandleUnlock(d4.l)
	systrap	DmReleaseResource(d4.l)

	move.l	#'tver',d0
	systrap	DmGet1Resource(d0.l,#1.w)
	move.l	a0,d4
	beq.s	RenNoVersion
	systrap	MemHandleLock(d4.l)
	systrap	StrCat(&note(a6),a0.l)
	systrap	MemHandleUnlock(d4.l)
	systrap	DmReleaseResource(d4.l)
	bra.s	RenVersionOK
RenNoVersion
	lea	RenUnknown(pc),a0
	systrap	StrCat(&note(a6),a0.l)
RenVersionOK
	systrap	StrCat(&note(a6),&RenNl(pc))

	call	ListGetItem(gameNo(a6).w)
	lea	SysDBListItemType.name(a0),a1
	move.l	a0,d0
	bne.s	RenNameOK
	lea	RenUnknown(pc),a1
RenNameOK
	systrap	StrCat(&note(a6),a1.l)
	systrap	StrCat(&note(a6),&RenNl(pc))

	tst.w	no(a6)
	beq.s	RenMessage0
	systrap	SysStringByIndex(#strtblIdRunNoteReasons.w,no(a6).w,&l1(a6),#51.w)
	systrap	StrLen(&l1(a6))
	tst.w	d0
	bne.s	RenMessageOK
RenMessage0
	systrap	SysStringByIndex(#strtblIdRunNoteReasons.w,#0.w,&l2(a6),#31.w)
	systrap	StrPrintF(&l1(a6),&l2(a6),no(a6).w)
RenMessageOK
	systrap	StrCat(&note(a6),&l1(a6))
	systrap	StrCat(&note(a6),&RenNl(pc))

	move.w	RunCpuAF(a5),d0
	tst.b	RunCpuIE(a5)
	seq	d1
	and.b	#'i'-'I',d1
	add.b	#'I',d1
	move.w	RunCpuBC(a5),d2
	move.w	RunCpuDE(a5),d3
	move.w	RunCpuHL(a5),d4
	systrap	StrPrintF(&l2(a6),&RenHexLine(pc),d0.w,d1.w,d2.w,d3.w,d4.w)
	systrap	StrCat(&note(a6),&l2(a6))
	systrap	StrCat(&note(a6),&RenNl(pc))
	move.w	RunCpuPc(a5),d0
	move.b	#':',d1
	lea	RunCpuCode(a5),a0
	move.w	(a0)+,d2
	move.w	(a0)+,d3
	move.w	(a0)+,d4
	systrap	StrPrintF(&l2(a6),&RenHexLine(pc),d0.w,d1.w,d2.w,d3.w,d4.w)
	systrap	StrCat(&note(a6),&l2(a6))
	systrap	StrCat(&note(a6),&RenNl(pc))
	move.w	RunCpuSp(a5),d0
	move.b	#':',d1
	lea	RunCpuStack(a5),a0
	move.w	(a0)+,d2
	move.w	(a0)+,d3
	move.w	(a0)+,d4
	systrap	StrPrintF(&l2(a6),&RenHexLine(pc),d0.w,d1.w,d2.w,d3.w,d4.w)
	systrap	StrCat(&note(a6),&l2(a6))
	systrap	StrCat(&note(a6),&RenNl(pc))

	systrap	StrLen(&note(a6))
	addq.w	#1,d0
	ext.l	d0
	move.l	d0,d5
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	move.l	#'DATA',d0
	systrap	DmOpenDatabaseByTypeCreator(d0.l,#sysFileCMemo.l,#dmModeReadWrite.w)
	move.l	a0,d4
	beq	RenNoDb

	clr.w	idx(a6)
	systrap	DmNewRecord(d4.l,&idx(a6),d5.l)
	move.l	a0,d6
	beq	RenNoMem

	systrap	DmOpenDatabaseInfo(d4.l,&locId(a6),&NULL,&NULL,&card(a6),&NULL)
	systrap	DmDatabaseInfo(card(a6).w,locId(a6).l,&NULL,&attr(a6),&NULL,&NULL,&NULL,&NULL,&NULL,&NULL,&NULL,&NULL,&NULL)
	or.w	#dmHdrAttrBackup,attr(a6)
	systrap	DmSetDatabaseInfo(card(a6).w,locId(a6).l,&NULL,&attr(a6),&NULL,&NULL,&NULL,&NULL,&NULL,&NULL,&NULL,&NULL,&NULL)

	systrap	MemHandleLock(d6.l)
	systrap	DmWrite(a0.l,#0.l,&note(a6),d5.l)
	systrap	MemHandleUnlock(d6.l)
	systrap	DmReleaseRecord(d4.l,idx(a6).w,#true.b)

	systrap	DmCloseDatabase(d4.l)

	systrap	FrmAlert(#alertIdRunErrorWritten.w)
	bra.s	RenQuit

RenNoMem
	systrap	DmCloseDatabase(d4.l)
RenNoDb
	systrap	FrmCustomAlert(#alertIdRunErrorNotWritten.w,&l1(a6),&RenDummy(pc),&RenDummy(pc))
RenQuit
endproc
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
RenUnknown
	dc.b	'-- unknown --',0
RenNl
	dc.b	10,0
RenHexLine
	dc.b	'%hX%c%hX%hX%hX',0
RenDummy
	dc.b	' ',0
	align	2

;--------------------------------------------------------------------------
;	String table with error messages:
;	The length of each message must be shorter than 50 chars!

	res	'tSTL',strtblIdRunNoteReasons
	dc.b	0		;no prefix
	dc.b	0,27		;number of strings
	dc.b	'Unknown error %d.',0		;0
	dc.b	'Debug check failed.',0		;-1  ERR_DEBUG
	dc.b	'Single stepped.',0		;-2  SINGLE_STEP
	dc.b	'User stop.',0			;-3  USER_STOP
	dc.b	0				;-4
	dc.b	0				;-5
	dc.b	0				;-6
	dc.b	0				;-7
	dc.b	0				;-8
	dc.b	0				;-9
	dc.b	0				;-10
	dc.b	'Opcode STOP.',0		;-11 OPCODE_STOP
	dc.b	0				;-12
	dc.b	0				;-13
	dc.b	0				;-14
	dc.b	0				;-15
	dc.b	0				;-16
	dc.b	0				;-17
	dc.b	0				;-18
	dc.b	0				;-19
	dc.b	0				;-20
	dc.b	'Opcode not implemented.',0	;-21 OP_NOT_IMPL
	dc.b	'Unknown opcode.',0		;-22 UNKNOWN
	dc.b	'No memory.',0			;-23 NO_MEM
	dc.b	'No input port.',0		;-24 NO_INP
	dc.b	'No output port.',0		;-25 NO_OUT
	dc.b	'Feature not implemented.',0	;-26 FTR_NOT_IMPL

;=== The end... ===========================================================
