/**	
	$Author: dizzy $
	$Revision: 1.17 $
	Last updated: $Date: 2000/11/19 19:19:44 $
	
	$Log: AccountsDB.c,v $
	Revision 1.17  2000/11/19 19:19:44  dizzy
	The way the tables are loaded was re-written. The account table works perfectly now
	but the transaction table still has a few problems.
	
	Revision 1.16  2000/11/16 18:21:59  dizzy
	Added new fields to the Account structure.
	
	Revision 1.15  2000/11/15 18:58:42  dizzy
	New balance adjust option in transaction form menu.
	Help updated in Transaction form and Accounts form.
	Merged ShowScrollArrows functions.
	Updated a few variable and function names.
	
	Revision 1.14  2000/11/12 19:10:11  dizzy
	Structures and functions modified to accomodate memorising the cleared atribute as a category
	rather then a member in the Transaction structure.
	
	Revision 1.13  2000/11/10 01:25:55  chaese
	merged New Account and Account Details forms
	
	Revision 1.12  2000/10/27 22:29:20  dizzy
	All records are sorted now; fake indentation in account view
	
	Revision 1.11  2000/10/21 14:35:29  dizzy
	added delete functionality (for accounts and transactions)
	
	Revision 1.10  2000/10/12 17:17:24  dizzy
	Email address update; lots of bugfixes
	
	Revision 1.9  2000/10/08 15:40:46  dizzy
	Fixed a load of bugs related to transaction support.
	Displaying i/e accounts in all view is optional.
	
	Revision 1.8  2000/09/15 10:10:48  dizzy
	First useful version
	
	Copyright (C) 2000 Vlad Mereuta <dizzy@users.sourceforge.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//define this to ignore the external definition of the vars below in the header file
#define __ACCOUNTS_DB_C__

#include "TransactionsDB.h"


//the variables below have to be accessed in different object files

DmOpenRef	AccountsDB_ref;
UInt32		AccountsDB_num_cat;
UInt16		AccountsDB_current_cat=dmAllCategories;
Int16		AccountsDB_current_record;
Boolean		AccountsDB_record_is_new;
totals 		AccountsDB_totals={true,0,0};

void	AccountsDB_UpdateTotals()
{
	UInt16					recordNo,i;
	MemHandle				rech;
	packedAccountStructure*	pas;
	accountStructure 		as;
		
	if(AccountsDB_totals.dirty) {
		//we have to update the totals
		//scan through the database adding the balances from each account
		
		AccountsDB_totals.dirty=false;
		AccountsDB_totals.total=AccountsDB_totals.expected_total=0;
		
	  	for(i=0,recordNo=0;i<AccountsTable_total_items;i++,recordNo++) {
			//fetch record
			rech=DmQueryNextInCategory(AccountsDB_ref,&recordNo,AccountsDB_current_cat);		
			ErrFatalDisplayIf(!rech, "rech is 0 - aborting");
	
			//dont display income/expense accounts
			if( AccountsDB_current_cat!=dmAllCategories || coins_pref.disp_ie_acc || \
			  ( db_GetRecordCategory(AccountsDB_ref,recordNo)!=CategoryFind(AccountsDB_ref,"Income") && \
			  db_GetRecordCategory(AccountsDB_ref,recordNo)!=CategoryFind(AccountsDB_ref,"Expense"))) {
				pas=(packedAccountStructure*) MemHandleLock(rech);
				ErrFatalDisplayIf(!pas, "pas is 0 - aborting");
				if(!pas) return;
				AccountsDB_UnpackAccount(&as,pas);
		
				AccountsDB_totals.total+=as.balance;
				AccountsDB_totals.expected_total+=as.expected_balance;
					
				MemHandleUnlock(rech);
			} else i--;
		}
	}
}

MemHandle AccountsDB_GetUnpackedRecord(UInt32 id,accountStructure* as)
{
	Int16						record;
	MemHandle					rech;
	packedAccountStructure* 	pas;

	DmFindRecordByID(AccountsDB_ref,id,&record);
	
	rech=DmQueryRecord(AccountsDB_ref,record);
	if(!rech) return NULL;
	ErrFatalDisplayIf(!rech, "rech is 0 - aborting");

	pas=(packedAccountStructure*) MemHandleLock(rech);
	ErrFatalDisplayIf(!pas, "pas is 0 - aborting");
	if(!pas) return NULL;

	AccountsDB_UnpackAccount(as,pas);
	MemHandleUnlock(rech);
	return rech;
}


Int32	AccountsDB_GetBalance(UInt32 id)
{
	Int16						record;
	MemHandle					rech;
	packedAccountStructure* 	pas;
	Int32						ret_val;

	DmFindRecordByID(AccountsDB_ref,id,&record);

	rech=DmQueryRecord(AccountsDB_ref,record);
	ErrFatalDisplayIf(!rech, "rech is 0 - aborting");

	pas=(packedAccountStructure*) MemHandleLock(rech);
	ErrFatalDisplayIf(!pas, "pas is 0 - aborting");

	ret_val=pas->balance;
	MemHandleUnlock(rech);
	return ret_val;
}

Int32	AccountsDB_GetExpectedBalance(UInt32 id)
{
	Int16						record;
	MemHandle					rech;
	packedAccountStructure* 	pas;
	Int32						ret_val;

	DmFindRecordByID(AccountsDB_ref,id,&record);

	rech=DmQueryRecord(AccountsDB_ref,record);
	ErrFatalDisplayIf(!rech, "rech is 0 - aborting");

	pas=(packedAccountStructure*) MemHandleLock(rech);
	ErrFatalDisplayIf(!pas, "pas is 0 - aborting");

	ret_val=pas->expected_balance;
	MemHandleUnlock(rech);
	return ret_val;
}

void AccountsDB_SetRecordCategory(UInt16 index, UInt16 cat)
{
	db_SetRecordCategory(AccountsDB_ref,index,cat);
}

void AccountsDB_UnpackAccount (accountStructure* account, const packedAccountStructure* packedAccount)
{
	Char*	s = packedAccount->description;
	
	account->balance = packedAccount->balance;
	account->expected_balance = packedAccount->expected_balance;
	account->credit_limit = packedAccount->credit_limit;
	account->parent_uid = packedAccount->parent_uid;
	account->children_total = packedAccount->children_total;
	account->children_expected_total = packedAccount->children_expected_total;
	account->depth = packedAccount->depth;
	account->children_no = packedAccount->children_no;
	account->collapsed = packedAccount->collapsed;
			
	account->description = s;
	s += StrLen(s)+1;
	account->acname = s;
}

void AccountsDB_PackAccount (accountStructure* account, MemHandle accountDBEntry)
{
	// figure out necessary size
	UInt32	length = 0;
	MemPtr	s;
	UInt32	offset = 0;
	length = sizeof(account->balance) + sizeof(account->expected_balance) + sizeof(account->credit_limit) + 
		sizeof(account->parent_uid) + sizeof(account->children_total) +  sizeof(account->children_expected_total) + 
		sizeof(account->depth) + sizeof(account->children_no) + sizeof(account->collapsed) +	
		StrLen(account->description) + StrLen(account->acname) + 2;  // for 2 string terminators

	// resize the VoidHand
	if (MemHandleResize(accountDBEntry, length) == 0) {
		// copy the fields
		s = MemHandleLock(accountDBEntry);
		offset = 0;
		DmWrite(s, offset, (MemPtr) &account->balance,sizeof(account->balance));
		offset += sizeof(account->balance);
		DmWrite(s, offset, (MemPtr) &account->expected_balance,sizeof(account->expected_balance));
		offset += sizeof(account->expected_balance);
		DmWrite(s, offset, (MemPtr) &account->credit_limit,sizeof(account->credit_limit));
		offset += sizeof(account->credit_limit);
		DmWrite(s, offset, (MemPtr) &account->parent_uid,sizeof(account->parent_uid));
		offset += sizeof(account->parent_uid);
		DmWrite(s, offset, (MemPtr) &account->children_total,sizeof(account->children_total));
		offset += sizeof(account->children_total);
		DmWrite(s, offset, (MemPtr) &account->children_expected_total,sizeof(account->children_expected_total));
		offset += sizeof(account->children_expected_total);
		DmWrite(s, offset, (MemPtr) &account->depth,sizeof(account->depth));
		offset += sizeof(account->depth);
		DmWrite(s, offset, (MemPtr) &account->children_no,sizeof(account->children_no));
		offset += sizeof(account->children_no);
		DmWrite(s, offset, (MemPtr) &account->collapsed,sizeof(account->collapsed));
		offset += sizeof(account->collapsed);
		
		DmStrCopy(s, offset, (Char*) account->description);
		offset += StrLen(account->description)+1;
		DmStrCopy(s, offset, (Char*) account->acname);
		
		MemHandleUnlock(accountDBEntry);
	}
	
}

Err		AccountsDB_Init()
{
	UInt32	mode = dmModeReadWrite;
	Err	err = 0;
	MemPtr	c;
	Boolean	created;

	palmCoinsAppInfoType* 	appInfoP;

	// Find the database.  If it doesn't exist, create it.
	db_OpenOrCreate(&AccountsDB_ref, ACCOUNTSDB_TYPE,PALM_COINS_CREATOR, mode,0, ACCOUNTSDB_NAME, &created);
	//if the db was created put in some sample values
	//have to initialise the application info
	c = AccountsDB_GetLockedAppInfo();
	if (created) AccountsDB_InitializeAccounts();

	MemPtrUnlock(c);
	appInfoP = (palmCoinsAppInfoType*) c;
	
	return err;
}

void	AccountsDB_Close()
{
	DmCloseDatabase(AccountsDB_ref);
}


MemPtr	AccountsDB_GetLockedAppInfo()
{
	UInt16	cardNo;
	LocalID	dbID;
	LocalID	appInfoID;
	Err	err;

	err = DmOpenDatabaseInfo(AccountsDB_ref, &dbID, NULL, NULL,&cardNo, NULL);
	ErrFatalDisplayIf (err,  "Get getting application info block");
	err = DmDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL,NULL, NULL, &appInfoID, NULL, NULL, NULL);
	ErrFatalDisplayIf (err,  "Get getting application info block");
	
	if(appInfoID==NULL) return CoinsAppInfoInit(dbID,cardNo);
	//a bit useless - to be removed unless code will be inserted here
	return MemLocalIDToLockedPtr(appInfoID, cardNo);
}

void	AccountsDB_InitializeAccounts()
{
	UInt16 i;
	
	//** sample accounts
//	accountStructure a1 = {20000,15000,0,"JDe","Natwest","Test acc"};
//	accountStructure a4 = {2200,1500,0,"John Doe","Galactic Bank-West","Test1 acc"};
	
	//*** income accounts
	accountStructure a3 = {0,0,0,0,0,0,0,0,0,"n/a","Misc"};
	accountStructure a5 = {0,0,0,0,0,0,0,0,0,"n/a","Balance update"};
	accountStructure a6 = {0,0,0,0,0,0,0,0,0,"n/a","Balance update"};

	accountStructure a7 = {0,0,0,0,0,0,0,0,0,"n/a","Interest"};	
	accountStructure a8 = {0,0,0,0,0,0,0,0,0,"n/a","Salary"};
	accountStructure a9 = {0,0,0,0,0,0,0,0,0,"n/a","Gift"};
	accountStructure a10 = {0,0,0,0,0,0,0,0,0,"n/a","Allowance"};
	
	//*** expense accounts
	accountStructure a2 = {0,0,0,0,0,0,0,0,0,"n/a","Misc"};	
	accountStructure a51 = {0,0,0,0,0,0,0,0,0,"n/a","Utilities"};
	accountStructure a52 = {0,0,0,0,0,0,1,0,0,"n/a","U:Gas"};
	accountStructure a53 = {0,0,0,0,0,0,1,0,0,"n/a","U:TV"};
	accountStructure a54 = {0,0,0,0,0,0,1,0,0,"n/a","U:Phone"};
	accountStructure a55 = {0,0,0,0,0,0,1,0,0,"n/a","U:Electricity"};
	accountStructure a56 = {0,0,0,0,0,0,1,0,0,"n/a","U:Rent"};

	accountStructure a57 = {0,0,0,0,0,0,0,0,0,"n/a","Travel"};
	accountStructure a58 = {0,0,0,0,0,0,1,0,0,"n/a","T:Land"};
	accountStructure a59 = {0,0,0,0,0,0,1,0,0,"n/a","T:Sea"};
	accountStructure a60 = {0,0,0,0,0,0,1,0,0,"n/a","T:Air"};
	accountStructure a61 = {0,0,0,0,0,0,1,0,0,"n/a","T:Meals"};

	accountStructure a62 = {0,0,0,0,0,0,0,0,0,"n/a","Food"};
	accountStructure a63 = {0,0,0,0,0,0,1,0,0,"n/a","F:Dining"};
	accountStructure a64 = {0,0,0,0,0,0,0,0,0,"n/a","Clothing"};
	accountStructure a65 = {0,0,0,0,0,0,0,0,0,"n/a","Gifts"};
	accountStructure a66 = {0,0,0,0,0,0,0,0,0,"n/a","Computer"};
	accountStructure a67 = {0,0,0,0,0,0,0,0,0,"n/a","Books"};
	accountStructure a68 = {0,0,0,0,0,0,0,0,0,"n/a","Entretainment"};
	
	//do this to force setting the category to income before the 1st transaction is to be recorded
	AccountsDB_current_cat=CategoryFind(AccountsDB_ref,"Income");
	i=AccountsDB_NewRecord(&a3);

	AccountsDB_current_cat=dmAllCategories;
	
	//default test accounts
//	i=AccountsDB_NewRecord(&a1);
//	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Bank"));
//	i=AccountsDB_NewRecord(&a4);
//	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Bank"));
	
	//misc and balance update - used by the program
	i=AccountsDB_NewRecord(&a2);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Expense"));
	i=AccountsDB_NewRecord(&a5);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Income"));
	i=AccountsDB_NewRecord(&a6);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Expense"));

	//other income accounts - modify as you see fit
	i=AccountsDB_NewRecord(&a7);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Income"));
	i=AccountsDB_NewRecord(&a8);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Income"));
	i=AccountsDB_NewRecord(&a9);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Income"));
	i=AccountsDB_NewRecord(&a10);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Income"));

	//other expense accounts
	i=AccountsDB_NewRecord(&a51);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Expense"));
	a52.parent_uid=db_GetRecordUid(AccountsDB_ref,i);
	a53.parent_uid=a52.parent_uid;
	a54.parent_uid=a52.parent_uid;
	a55.parent_uid=a52.parent_uid;
	a56.parent_uid=a52.parent_uid;
	
	i=AccountsDB_NewRecord(&a52);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Expense"));
	i=AccountsDB_NewRecord(&a53);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Expense"));
	i=AccountsDB_NewRecord(&a54);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Expense"));
	i=AccountsDB_NewRecord(&a55);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Expense"));
	i=AccountsDB_NewRecord(&a56);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Expense"));

	i=AccountsDB_NewRecord(&a57);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Expense"));
	a58.parent_uid=db_GetRecordUid(AccountsDB_ref,i);
	a59.parent_uid=a58.parent_uid;
	a60.parent_uid=a58.parent_uid;
	a61.parent_uid=a58.parent_uid;
	
	i=AccountsDB_NewRecord(&a58);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Expense"));
	i=AccountsDB_NewRecord(&a59);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Expense"));
	i=AccountsDB_NewRecord(&a60);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Expense"));
	i=AccountsDB_NewRecord(&a61);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Expense"));
	
	i=AccountsDB_NewRecord(&a62);
	a63.parent_uid=db_GetRecordUid(AccountsDB_ref,i);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Expense"));
	i=AccountsDB_NewRecord(&a63);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Expense"));
	
	i=AccountsDB_NewRecord(&a64);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Expense"));
	i=AccountsDB_NewRecord(&a65);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Expense"));
	i=AccountsDB_NewRecord(&a66);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Expense"));
	i=AccountsDB_NewRecord(&a67);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Expense"));
	i=AccountsDB_NewRecord(&a68);
	AccountsDB_SetRecordCategory(i,CategoryFind(AccountsDB_ref,"Expense"));
			
}

void AccountsDB_UpdateBalance(UInt32 uid,Int32 newbal)
{
	accountStructure		a, parent;
	MemHandle				parent_h;
	transactionStructure	t;

	AccountsDB_GetUnpackedRecord(uid,&a);
	
	if (a.balance < newbal) { //adjust by adding money
		//record the opening account transaction
		t.source_acc=AccountsDB_FindAccountByName("Balance update",CategoryFind(AccountsDB_ref,"Income"));
		ErrFatalDisplayIf(!t.source_acc,"not found");
		t.dest_acc=uid;
		t.ammount=newbal-a.balance;
	} else { //adjust by taking away money
		t.dest_acc=AccountsDB_FindAccountByName("Balance update",CategoryFind(AccountsDB_ref,"Expense"));
		ErrFatalDisplayIf(!t.dest_acc,"not found");
		t.source_acc=uid;
		t.ammount=a.balance-newbal;
	}

	if(a.depth!=0) { //update parent information
		parent_h=AccountsDB_GetUnpackedRecord(a.parent_uid,&parent);
		parent.children_total+=newbal-a.balance;;
		AccountsDB_PackAccount(&parent,parent_h);
	}

	t.payee="Balance update";
	DateSecondsToDate(TimGetSeconds(),& t.date);
	t.method=pm_transfer;
	t.note="";	
	TransactionsDB_NewRecord(&t,true);	//cleared by default
}	


UInt32	AccountsDB_FindAccountByName(char* n, UInt16 categ)
{
	UInt16				recordNo;
	UInt32				uniqueID;
	accountStructure	as;

  	for(recordNo=0;;recordNo++) {
		if(!DmQueryNextInCategory(AccountsDB_ref,&recordNo,categ)) return 0;		
		DmRecordInfo(AccountsDB_ref,recordNo,NULL,&uniqueID,NULL);

		AccountsDB_GetUnpackedRecord(uniqueID,&as);

		if(!StrCaselessCompare(n,as.acname)) return uniqueID;
	}
	
}

UInt16	AccountsDB_NewRecord(accountStructure* s)
{
	accountStructure		parent;
	MemHandle				h,p,parent_h;
	UInt16					index;
	transactionStructure	ts;
	Int32					bal;
	
	//clear the balance since it is maintained via the transaction
	bal=s->balance;
	s->balance=s->expected_balance=0;
	s->collapsed=1;
	if(s->depth!=0) { //update parent information
		parent_h=AccountsDB_GetUnpackedRecord(s->parent_uid,&parent);
		parent.children_total+=s->balance;
		parent.children_expected_total+=s->expected_balance;
		parent.children_no++;
		//overwrite the depth
		s->depth=parent.depth+1;
		
		AccountsDB_PackAccount(&parent,parent_h);
	}

	//set size to 1 - will be resized by packAccount
	h=DmNewHandle(AccountsDB_ref,1);
	
	ErrFatalDisplayIf(!h,"Creating new account failed. No space left on the device?");

	AccountsDB_PackAccount(s,h);

	p=MemHandleLock(h);
	index=DmFindSortPosition(AccountsDB_ref,p, 0, (DmComparF *) AccountsDB_CompareRecords, 0);
	MemPtrUnlock(p);
	
	DmAttachRecord(AccountsDB_ref,&index,h,0);
	DmReleaseRecord(AccountsDB_ref, index, true); //true sets the dirty bit	

	if(AccountsDB_current_cat!=dmAllCategories) AccountsDB_SetRecordCategory(index,AccountsDB_current_cat);

	//record the opening account transaction
	ts.source_acc=AccountsDB_FindAccountByName("Misc",CategoryFind(AccountsDB_ref,"Income"));
	ErrFatalDisplayIf(!ts.source_acc,"Misc account not found");
	ts.payee="Account opened";
	//save uniques id of the new record

	DmRecordInfo(AccountsDB_ref,index,NULL,&ts.dest_acc,NULL);
	DateSecondsToDate(TimGetSeconds(),& ts.date);
	ts.ammount=bal;
	ts.method=pm_transfer;
	ts.note="";	

	TransactionsDB_NewRecord(&ts,true);
	return index;	
}

void	AccountsDB_DeleteRecord(UInt16	id)
{
	UInt16					recordNo,peer_id,acc_cat,old_trans_cl;
	UInt32					trans_uid,acc_uid,peer_uid;
	transactionStructure	ts;

	//get the account uid
	DmRecordInfo(AccountsDB_ref,id,NULL,&acc_uid,NULL);
	
	for(recordNo=0;;recordNo++) {
		if(!DmQueryNextInCategory(TransactionsDB_ref,&recordNo,dmAllCategories)) break;				
		DmRecordInfo(TransactionsDB_ref,recordNo,NULL,&trans_uid,NULL);
		TransactionsDB_GetUnpackedRecord(trans_uid,&ts);
		old_trans_cl=(db_GetRecordCategory(TransactionsDB_ref,recordNo)==CategoryFind(TransactionsDB_ref,"Cleared"));
		

		//is the current transaction related to this account?
		peer_uid=0;
		if(ts.source_acc == acc_uid)
			peer_uid=ts.dest_acc;
		else if	(ts.dest_acc == acc_uid) 
			peer_uid=ts.source_acc;
		
		//if the trans is related to our account
		if(peer_uid) {
			//get its id
			ErrFatalDisplayIf(DmFindRecordByID(AccountsDB_ref,peer_uid,&peer_id),"rec not found");
			//if the other acc is in income/expense just remove the transaction
			acc_cat=db_GetRecordCategory(AccountsDB_ref,peer_id);
			if((acc_cat==CategoryFind(AccountsDB_ref,"Income")) || (acc_cat==CategoryFind(AccountsDB_ref,"Expense")))
					TransactionsDB_DeleteRecord(recordNo);
			else { //otherwise change the acc to expense/income
				if(peer_uid==ts.source_acc) 				//income
					ts.dest_acc=AccountsDB_FindAccountByName("Misc",CategoryFind(AccountsDB_ref,"Expense"));
				else ts.source_acc=AccountsDB_FindAccountByName("Misc",CategoryFind(AccountsDB_ref,"Income"));

				TransactionsDB_DeleteRecord(recordNo);
				TransactionsDB_NewRecord(&ts,old_trans_cl);
			}
		}
	}
	DmRemoveRecord(AccountsDB_ref,id);
	AccountsDB_totals.dirty=true;	
}

MemPtr CoinsAppInfoInit(LocalID dbID,UInt16 cardNo)
{
	UInt16					wordValue;
	LocalID					appInfoID;
	palmCoinsAppInfoType*	nilP = 0;
	palmCoinsAppInfoType* 	appInfoP;

	MemHandle h;

	
	h = DmNewHandle(AccountsDB_ref, sizeof (palmCoinsAppInfoType));
	ErrFatalDisplayIf (!h,  "Cannot allocate app info block");

	//get the local id of the newly allocated appinfo record
	appInfoID = MemHandleToLocalID (h);
	//record the id
	DmSetDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL,NULL, &appInfoID, NULL, NULL, NULL);
	
	appInfoP = MemLocalIDToLockedPtr(appInfoID, cardNo);

	// Clear the app info block.
	DmSet (appInfoP, 0, sizeof(palmCoinsAppInfoType), 0);

	// Initialize the categories.
	CategoryInitialize ((AppInfoPtr) appInfoP, PALM_COINS_CATEGORIES);

	// I don't know what this field is used for.
	wordValue = 0xFFFF;
	DmWrite (appInfoP, (UInt32)&nilP->dirtyAppInfo, &wordValue,
		sizeof(appInfoP->dirtyAppInfo));

	// Initialize the sort order.
	DmSet (appInfoP, (UInt32)&nilP->sortOrder, sizeof(appInfoP->sortOrder), 
		soName);

	return appInfoP;
}

Int16	AccountsDB_CompareRecords(packedAccountStructure* rec1,packedAccountStructure* rec2,Int16 other,SortRecordInfoPtr sr1,SortRecordInfoPtr sr2,MemHandle app_info)
{
	Int16 rez;
	accountStructure ac1,ac2;

	AccountsDB_UnpackAccount(&ac1,rec1);
	AccountsDB_UnpackAccount(&ac2,rec2);

	rez=CompareAccountNames(ac1.acname, ac2.acname);
	
	return rez;
}
