/**	
	Code to access the database containing the account information

	$Author: dizzy $
	$Revision: 1.13 $
	Last updated: $Date: 2000/11/16 18:21:59 $
	
	$Log: AccountsDB.h,v $
	Revision 1.13  2000/11/16 18:21:59  dizzy
	Added new fields to the Account structure.
	
	Revision 1.12  2000/11/15 18:58:42  dizzy
	New balance adjust option in transaction form menu.
	Help updated in Transaction form and Accounts form.
	Merged ShowScrollArrows functions.
	Updated a few variable and function names.
	
	Revision 1.11  2000/11/10 01:25:55  chaese
	merged New Account and Account Details forms
	
	Revision 1.10  2000/10/27 22:29:20  dizzy
	All records are sorted now; fake indentation in account view
	
	Revision 1.9  2000/10/21 14:35:29  dizzy
	added delete functionality (for accounts and transactions)
	
	Revision 1.8  2000/10/12 17:17:24  dizzy
	Email address update; lots of bugfixes
	
	Revision 1.7  2000/09/15 10:10:48  dizzy
	First useful version
	
	Revision 1.6  2000/09/13 11:50:27  dizzy
	The transaction module is fully functional
	
	Revision 1.5  2000/09/08 10:41:10  dizzy
	Transactiondatabas initial support and column display preferences support
	
	Revision 1.4  2000/09/04 14:23:07  dizzy
	General code cleanup
	
	Revision 1.2  2000/08/23 12:08:29  dizzy
	Added details dialog for accounts
	
	Revision 1.1  2000/08/21 08:59:30  dizzy
	Reorganised files

	Copyright (C) 2000 Vlad Mereuta <dizzy@users.sourceforge.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
	
*/

#ifndef __ACCOUNTS_DB__
#define __ACCOUNTS_DB__

#include <PalmOS.h>
#include <DateTime.h>
#include <FeatureMgr.h>

#include "PalmCoinsRsc.h"
#include "AccountsTable.h"
#include "MiscFunc.h"

/// type of the account database
#define ACCOUNTSDB_TYPE		'DATA'
/// db name
#define ACCOUNTSDB_NAME	"PalmCoins-Accounts"

///this structure contains the overall totals for all the selected accounts
typedef struct {
	///the totals have to be recalculated
	Boolean	dirty;
	///current balance total
	Int32	total;
	///end of month total
	Int32	expected_total;
} totals;


//define this variables as extern if the file is not included from AccountsDB.c
#ifndef __ACCOUNTS_DB_C__

/// reference to the database
extern DmOpenRef		AccountsDB_ref;
/// total number of categories
extern UInt32			AccountsDB_num_cat;
/// the currenct category
extern UInt16			AccountsDB_current_cat;
/// the current record (to be displayed/edited/etc)
extern Int16			AccountsDB_current_record;
/// indicates that a new account is being created
extern Boolean			AccountsDB_record_is_new;
/// ths structure holds the running totals for the current category
extern totals			AccountsDB_totals;
#endif


/// Application Info Block
typedef struct {
	UInt16	renamedCategories;					// bitfield of categories with a different name
	Char	categoryLabels [dmRecNumCategories] [dmCategoryLength];
	UInt8	categoryUniqIDs [dmRecNumCategories];
	UInt8	lastUniqID;							// Uniq IDs generated by the device are between
												// 0 - 127.  Those from the PC are 128 - 255.
	UInt8	reserved1;							// from the compiler word aligning things
	UInt16	reserved2;	
	UInt16	dirtyAppInfo;
	UInt8	sortOrder;
	UInt8	reserved3;
} palmCoinsAppInfoType;


/// sort order type
typedef enum { soName, soAmmount, soDate } sortOrder;

///the packed structure of an account - see the unpacked struct for explanations
typedef struct {
	Int32		balance;
	Int32		expected_balance;
	Int32		credit_limit;
	UInt32		parent_uid;
	Int32		children_total;
	Int32		children_expected_total;
	UInt8		depth;
	UInt8		children_no;
	UInt8		collapsed;	
	Char		description[1]; // actually may be longer than 1
} packedAccountStructure;

///unpacked structure of an account
typedef struct {
	/// account balance
	Int32		balance;
	/**
		expected balance in the account at a set date (after transactions cleared and 
		standing orders/direct debits which were due went out)
	*/
	Int32		expected_balance;
	///the credit limit for creditcards/overdraft for other accounts
	Int32		credit_limit;	
	/// uid of the parent of this account. if the account is a root one this field should be ignored
	UInt32		parent_uid;
	/// overall balance of the children of this account
	Int32		children_total;
	/// overall expected balance of the children of this account
	Int32		children_expected_total;
	/// depth at which this account is situated (0 for root accounts, >0 for the others)
	UInt8		depth;
	/// total number of children owned by this account. this should facilitate things like sorting
	UInt8		children_no;	
	/// indicates whether the account is collapsed or not
	UInt8		collapsed;
	///account desciption
	Char* 		description;
	///the name of the account - this will show up throughout the program
	Char*		acname;
} accountStructure;


// ***** FUNCTIONS

/** 
	sets the category of a given record
	@param index	the index of the record
	@param cat		the record category
*/
void	AccountsDB_SetRecordCategory(UInt16 index, UInt16 cat);

/// updates the totals structure (if necessary)
void	AccountsDB_UpdateTotals();

/**
	@param n name of the account to be found
	@param categ category of the account
	@return unique id of the account or 0 if the account could not be found
*/
UInt32	AccountsDB_FindAccountByName(char* n, UInt16 categ);

/// unpacks an account record
void	AccountsDB_UnpackAccount (accountStructure* account, const packedAccountStructure* packedAccount);

/// packs an account record
void	AccountsDB_PackAccount (accountStructure* account, MemHandle accountDBEntry);

/// initialize the databases
 Err	AccountsDB_Init(void);

/// closes the database
void 	AccountsDB_Close(void);

/**
	Update the balance for an account
	@param uid unique id
	@param newbal the new balance
*/
void AccountsDB_UpdateBalance(UInt32 uid,Int32 newbal);


/**
	get the unpacked strucure of a given record
	@param id	unique id of the account record
	@param as 	pointer in which the unpacked data will be saved
	@return pointer to the record handler
*/
 MemHandle AccountsDB_GetUnpackedRecord(UInt32 id,accountStructure* as);

/**
	fetch the balance of a given account
	@param id	the unique id of the account record
	@return the balance in pence
*/
Int32	AccountsDB_GetBalance(UInt32 id);

/**
	fetch the expected balance of a given account
	@param id	the unique id of the account record
	@return the expected balance in pence
*/
Int32	AccountsDB_GetExpectedBalance(UInt32 id);

//init the Accounts DB with some default records
void	AccountsDB_InitializeAccounts(void);

///returns a pointer to the app info record
MemPtr	AccountsDB_GetLockedAppInfo();

///used by the func above to initialise the app info rec if it is empty
MemPtr CoinsAppInfoInit(LocalID dbID,UInt16 cardNo);

/**
	packs and writes the data from s into the database
	note: after calling this function the record has to be released

	@return the index to the record
*/
UInt16	AccountsDB_NewRecord(accountStructure* s);

/**
	Deletes an account given by its id
	The function scans through all the transactions and removes the ones which
	have as source/destinations the account to be deleted and as dest/src an account
	in the Income or Expense categories.
	
	All the transaction involving user defined account will be updated so that the 
	reference to the account to be deleted is replaced with a reference to a Income/Expense
	account
	
	@param id the id of the account to be deleted
*/
void	AccountsDB_DeleteRecord(UInt16	id);

/**
	Compares two records (by the account name)
*/
Int16	AccountsDB_CompareRecords(packedAccountStructure* rec1,packedAccountStructure* rec2,Int16 other,SortRecordInfoPtr sr1,SortRecordInfoPtr sr2,MemHandle app_info);


#endif
