/**	
	$Author: dizzy $
	$Revision: 1.17 $
	Last updated: $Date: 2000/11/19 19:19:44 $
	
	$Log: AccountsForm.c,v $
	Revision 1.17  2000/11/19 19:19:44  dizzy
	The way the tables are loaded was re-written. The account table works perfectly now
	but the transaction table still has a few problems.
	
	Revision 1.16  2000/11/15 18:58:42  dizzy
	New balance adjust option in transaction form menu.
	Help updated in Transaction form and Accounts form.
	Merged ShowScrollArrows functions.
	Updated a few variable and function names.
	
	Revision 1.15  2000/11/10 01:25:55  chaese
	merged New Account and Account Details forms
	
	Revision 1.14  2000/11/06 22:27:02  dizzy
	Numbers are now localised and alligned to the right; the dates are displayed using the default format (these changes were submitted by Carsten Haese
	
	Revision 1.13  2000/10/11 21:12:23  dizzy
	*** empty log message ***
	
	Revision 1.12  2000/10/11 20:46:57  dizzy
	Few fixes in the menu and versioning; initial help
	
	Revision 1.11  2000/10/08 15:40:46  dizzy
	Fixed a load of bugs related to transaction support.
	Displaying i/e accounts in all view is optional.
	

	Copyright (C) 2000 Vlad Mereuta <dizzy@users.sourceforge.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
	
*/


#include "AccountsForm.h"

/// name of the current category
static Char cat_name [dmCategoryLength];
///this form
static FormPtr	frm;

void	AccountsForm_InitCat()
{
	ControlPtr	ctl;
	
	// Set the label of the category trigger.
	ctl = GetObjectPtr (ACC_CAT_TRIGGER);
	CategoryGetName (AccountsDB_ref, AccountsDB_current_cat, cat_name);
	CategorySetTriggerLabel(ctl, cat_name);
}

void	AccountsForm_SelectCat()
{
	UInt16		category=AccountsDB_current_cat;
	Boolean 	categoryEdited;

	//have to do this due to some odd behaviour 
	frm = FrmGetActiveForm();
	
	categoryEdited=CategorySelect(AccountsDB_ref,frm, ACC_CAT_TRIGGER,ACC_CAT_LIST,true,&category,\
		cat_name,1,categoryDefaultEditCategoryString);
	if(categoryEdited || category!=AccountsDB_current_cat) {
		AccountsDB_current_cat=category;	
		//recalculate the totals
		AccountsDB_totals.dirty=true;			
		AccountsTable_top_visible_item=0;		
	}	
	
	//remember the current category
	coins_pref.opening_category=category;
	Prefs_Save();
}

void	AccountsForm_DisplayTotals()
{
	Char* 			s;
	RectangleType	rp;
		
	RctSetRectangle(&rp,70,129,90,11);
	WinEraseRectangle(&rp,0);

	AccountsDB_UpdateTotals();
	
	s=Pence2String(AccountsDB_totals.total);
	RctSetRectangle(&rp, 115, 129, 45, 11);
	DrawCharsRightAlign(s,&rp);
	MemPtrFree(s);
	
	s=Pence2String(AccountsDB_totals.expected_total);
	RctSetRectangle(&rp, 70, 129, 45, 11);
	DrawCharsRightAlign(s,&rp);
	MemPtrFree(s);	

	//add some lines at the bottom
	WinDrawLine(0,129,160,129);
	WinDrawLine(0,140,160,140);	
}

Boolean AccountsForm_HandleEvent( EventPtr event )
{
	FormPtr 	frm; //this form
	FormPtr		f;//about form
	Boolean 	handled = true;
	ControlPtr	ptr;
	
	// Get the form pointer
	frm = FrmGetActiveForm();

	// Key Events
	switch(event->eType) {

	// Form Open
	case frmOpenEvent:

		AccountsDB_current_cat=coins_pref.opening_category;
		
		AccountsTable_Init();
		//display the category name
		AccountsForm_InitCat();
		//draw form
		FrmDrawForm(frm);
		
		//display the running totals
		AccountsForm_DisplayTotals();
		break;

	case keyDownEvent:
		switch( event->data.keyDown.chr ) {
			case pageUpChr:			// Scroll up key presed?
				//link the hard button functionality to the soft up/down controls
				ptr = GetObjectPtr(ACC_PGUP);
				CtlHitControl(ptr);
				handled = true;
				break;								
			case pageDownChr:		// Scroll down key presed?						
				ptr = GetObjectPtr(ACC_PGDN);
				CtlHitControl(ptr);
				handled = true;
				break;
			default:
				handled = false;
		};
		break;
	
	// Controls
	case ctlSelectEvent:
		switch( event->data.ctlSelect.controlID ) {
			case ACC_NEW:
				AccountsDB_record_is_new = true;
				FrmPopupForm(EDIT_ACC_FORM);				
				FrmUpdateForm(ACCOUNTS_MAIN_FORM,0);
				handled = true;
				break;
			case ACC_CAT_TRIGGER:		//Select a category
				AccountsForm_SelectCat();
				FrmUpdateForm(ACCOUNTS_MAIN_FORM,0);
				handled = true;
				break;
			case ACC_PGUP:
				if(AccountsTable_top_visible_item) {
					if(AccountsTable_top_visible_item>=AccountsTable_rows) 
						AccountsTable_top_visible_item-=AccountsTable_rows-1;
					else AccountsTable_top_visible_item=0;
					FrmUpdateForm(ACCOUNTS_MAIN_FORM,0);
				}
				handled = true;
				break;								
			case ACC_PGDN:		// Scroll down control presed?	
				if(AccountsTable_total_items>AccountsTable_rows) {					
					if(AccountsTable_top_visible_item+AccountsTable_rows+1<AccountsTable_total_items)
						AccountsTable_top_visible_item+=AccountsTable_rows-1;
					else 
						AccountsTable_top_visible_item=AccountsTable_total_items-AccountsTable_rows;
			
					FrmUpdateForm(ACCOUNTS_MAIN_FORM,0);
				}
				handled = true;
				break;
				
			default:
				handled = false;
		}
		break;
	
	// a item in the table was selected
	// should probably be moved in the table files
	case tblSelectEvent:		
		//save the record on the current row into DetailRecord variable
		DmFindRecordByID(AccountsDB_ref,TblGetRowData(GetObjectPtr(ACC_TABLE),event->data.tblSelect.row),&AccountsDB_current_record);

		if(event->data.tblSelect.column==AccountsTable_name_col) {
			AccountsDB_record_is_new = false;
			FrmPopupForm(EDIT_ACC_FORM);
			FrmUpdateForm(ACCOUNTS_MAIN_FORM,0);
		}
		else { //opens the transaction list
			if((event->data.tblSelect).column==AccountsTable_balance_col) {
				//we have to set the current_account
				
				TransactionsTable_top_visible_item=0;		
				curr_account=TblGetRowData(GetObjectPtr(ACC_TABLE),event->data.tblSelect.row);
				FrmPopupForm(TRANSACTIONS_FORM);
				FrmUpdateForm(ACCOUNTS_MAIN_FORM,0);
				AccountsTable_Unhilight();
			}
			else {
				FrmAlert(NOT_IMPLEMENTED_ALERT);
				FrmUpdateForm(ACCOUNTS_MAIN_FORM,0);
				// to be implemented
			}
		}
		handled=true;
		break;
			
	// Menus
	case menuEvent:
		switch (event->data.menu.itemID) {
			case ACC_MENU_NEW:
				AccountsDB_record_is_new = true;
				FrmPopupForm(EDIT_ACC_FORM);
				FrmUpdateForm(ACCOUNTS_MAIN_FORM,0);
				handled=true;
				break;
			case ACC_MENU_PREF:
				FrmPopupForm(PREF_FORM);
				FrmUpdateForm(ACCOUNTS_MAIN_FORM,0);
				handled=true;
				break;
			case ACC_MENU_ABOUT:
			    f = FrmInitForm(ABOUT_FORM);
    
				FrmSetActiveForm(f);
    			FrmDoDialog(f);
        
			    if (frm) {
			        FrmSetActiveForm(frm);
				    FrmDeleteForm(f);
				}

				handled=true;
				break;
			case ACC_MENU_HELP:
				FrmHelp(ACCOUNTS_HELP);

				handled=true;
				break;
				
			default:
				handled = false;
		}
		break;
	
	// Form Close
	case frmCloseEvent:
		handled = false;
		break;
	
	// Form Update
	case frmUpdateEvent:
		AccountsTable_Init();
		//display the category name
		AccountsForm_InitCat();
		//draw form
		FrmDrawForm(frm);
		//display the running totals
		AccountsForm_DisplayTotals();

		handled = true;
		break;
	
	// Everything else...
	default:
		handled = false;
	}	
	return handled;
}
