/**	
	$Author: dizzy $
	$Revision: 1.15 $
	Last updated: $Date: 2000/11/19 19:19:44 $
	
	$Log: AccountsTable.c,v $
	Revision 1.15  2000/11/19 19:19:44  dizzy
	The way the tables are loaded was re-written. The account table works perfectly now
	but the transaction table still has a few problems.
	
	Revision 1.14  2000/11/16 18:21:59  dizzy
	Added new fields to the Account structure.
	
	Revision 1.13  2000/11/15 18:58:42  dizzy
	New balance adjust option in transaction form menu.
	Help updated in Transaction form and Accounts form.
	Merged ShowScrollArrows functions.
	Updated a few variable and function names.
	
	Revision 1.12  2000/11/06 22:27:02  dizzy
	Numbers are now localised and alligned to the right; the dates are displayed using the default format (these changes were submitted by Carsten Haese
	
	Revision 1.11  2000/10/27 22:29:20  dizzy
	All records are sorted now; fake indentation in account view
	
	Revision 1.10  2000/10/12 17:17:24  dizzy
	Email address update; lots of bugfixes
	
	Revision 1.9  2000/10/08 15:40:46  dizzy
	Fixed a load of bugs related to transaction support.
	Displaying i/e accounts in all view is optional.
	

	Copyright (C) 2000 Vlad Mereuta <dizzy@users.sourceforge.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
	
*/

#define __ACCOUNTS_TABLE_C__

#include "AccountsTable.h"

// the uniqueid for top visible items
UInt16		AccountsTable_top_visible_item=0;
// the max no of rows that can be displayed
UInt16		AccountsTable_rows;
// no of elements that could be loaded in the table
UInt16		AccountsTable_total_items;

//name all the columns
UInt16		AccountsTable_name_col=0;
UInt16		AccountsTable_expected_balance_col=1;
UInt16		AccountsTable_balance_col=2;

// the table
static TablePtr		table;


void AccountsTable_DrawCurrency(void* table, Int16 row, Int16 column, RectangleType* bounds)
{
	Char	*s;

	if(column==AccountsTable_expected_balance_col) s=Pence2String(AccountsDB_GetExpectedBalance(TblGetRowData(table,row)));
		else s=Pence2String(AccountsDB_GetBalance(TblGetRowData(table,row)));

	DrawCharsRightAlign(s, bounds);

	//free memory allocated by Pence2String
	MemPtrFree(s);
}

void AccountsTable_DrawText(void* table, Int16 row, Int16 column, RectangleType* bounds)
{
	accountStructure 	as;
	Char				acn[30];

	AccountsDB_GetUnpackedRecord(TblGetRowData(table,row),&as);
	
	WinDrawChars(FakeAccountNameIndent(as.acname,acn),StrLen(as.acname),bounds->topLeft.x,bounds->topLeft.y);
}

void AccountsTable_SetRowsUnusable(UInt16 start)
{
	UInt16 i;
	for(i=start;i<AccountsTable_rows;i++)
		TblSetRowUsable(table,i,false);		
}

void AccountsTable_LoadData()
{
	UInt16		recordNo=0,row,target;

	AccountsTable_Unhilight();
	
	AccountsTable_total_items=DmNumRecordsInCategory(AccountsDB_ref,AccountsDB_current_cat);
	if(AccountsDB_current_cat==dmAllCategories && !coins_pref.disp_ie_acc) { //dont display income/expense accounts
		AccountsTable_total_items-=DmNumRecordsInCategory(AccountsDB_ref,CategoryFind(AccountsDB_ref,"Income"));
		AccountsTable_total_items-=DmNumRecordsInCategory(AccountsDB_ref,CategoryFind(AccountsDB_ref,"Expense"));
	}
	

	//show the scrolling arrows if necessary
	ShowScrollArrows(ACC_PGUP,ACC_PGDN,AccountsTable_total_items,AccountsTable_rows,AccountsTable_top_visible_item);
	
	//position on the 1st record in our category
	if(!coins_pref.disp_ie_acc && AccountsDB_current_cat==dmAllCategories) { //some records will have to be ignored in All
		//first record in all
		if(DmSeekRecordInCategory(AccountsDB_ref,&recordNo,0,dmSeekForward,AccountsDB_current_cat)!=errNone) {
			//if no records abort
			AccountsTable_SetRowsUnusable(0);
			return;
		} 
		//if in the last seek we did not find a valid record we have to add one
		if (db_GetRecordCategory(AccountsDB_ref,recordNo)==CategoryFind(AccountsDB_ref,"Income") || db_GetRecordCategory(AccountsDB_ref,recordNo)==CategoryFind(AccountsDB_ref,"Expense") ) {
			target=AccountsTable_top_visible_item+1;
		} else { 
			target=AccountsTable_top_visible_item;
		}

		//now find the top record
		while(target>0) { //if we run out of records before reaching the rec we have to display at the top, abort
			if(DmSeekRecordInCategory(AccountsDB_ref,&recordNo,1,dmSeekForward,AccountsDB_current_cat)!=errNone) {
				AccountsTable_SetRowsUnusable(0);
				return;
			} else 	if (db_GetRecordCategory(AccountsDB_ref,recordNo)!=CategoryFind(AccountsDB_ref,"Income") && db_GetRecordCategory(AccountsDB_ref,recordNo)!=CategoryFind(AccountsDB_ref,"Expense") ) 
						target--;
		}
	} else //we can display all records - a simple seek would position us in the rigt place
		if(DmSeekRecordInCategory(AccountsDB_ref,&recordNo,AccountsTable_top_visible_item,dmSeekForward,AccountsDB_current_cat)!=errNone) {
			//no records in this category - invalidate all rows and return
			AccountsTable_SetRowsUnusable(0);
			return;
		} 		

	//set row 0 with the record we found above
	TblSetRowData(table,0,db_GetRecordUid(AccountsDB_ref,recordNo));					
	TblSetRowUsable(table,0,true);	
	
  	for(row=1;row<AccountsTable_rows;row++) {
		//try and fetch next record
		if(!coins_pref.disp_ie_acc && AccountsDB_current_cat==dmAllCategories) {
			//cycle through the records until we find an acceptable one
			do {
				if(DmSeekRecordInCategory(AccountsDB_ref,&recordNo,1,dmSeekForward,AccountsDB_current_cat)!=errNone) {
					//if we run out of items to display just set the rows to unusable		
					AccountsTable_SetRowsUnusable(row);
					return;
				}
			} while ( db_GetRecordCategory(AccountsDB_ref,recordNo)==CategoryFind(AccountsDB_ref,"Income") || db_GetRecordCategory(AccountsDB_ref,recordNo)==CategoryFind(AccountsDB_ref,"Expense") );					
		} else 
			if(DmSeekRecordInCategory(AccountsDB_ref,&recordNo,1,dmSeekForward,AccountsDB_current_cat)!=errNone) {
				//if we run out of items to display just set the rows to unusable		
				AccountsTable_SetRowsUnusable(row);
				break;
			}
	
		TblSetRowData(table,row,db_GetRecordUid(AccountsDB_ref,recordNo));								
		TblSetRowUsable(table,row,true);
	}
}	
		

void AccountsTable_Unhilight()
{
	TblUnhighlightSelection(table);

}

void AccountsTable_Init()
{
	UInt16		row;
	
	//init globals
	table=(TablePtr) GetObjectPtr(ACC_TABLE);
	
	AccountsTable_Unhilight();

	ErrFatalDisplayIf(!table, "table is null!!!");

	//the numbers are drawn with our custom procedure
	TblSetCustomDrawProcedure(table, AccountsTable_name_col, AccountsTable_DrawText);
	TblSetCustomDrawProcedure(table, AccountsTable_expected_balance_col, AccountsTable_DrawCurrency);
	TblSetCustomDrawProcedure(table, AccountsTable_balance_col, AccountsTable_DrawCurrency);

	//get max number of visible rows
	AccountsTable_rows = TblGetNumberOfRows(table);

	//set style for all cells
	for (row = 0; row < AccountsTable_rows; row++) {
		TblSetItemStyle(table, row, AccountsTable_name_col, customTableItem);
		TblSetItemStyle(table, row, AccountsTable_expected_balance_col, customTableItem);
		TblSetItemStyle(table, row, AccountsTable_balance_col, customTableItem);
	}
	
	//set all columns as usable
	TblSetColumnUsable(table, AccountsTable_name_col, true);
	if(coins_pref.disp_expected_col) {
		TblSetColumnUsable(table, AccountsTable_expected_balance_col,true);
		TblSetColumnWidth(table,AccountsTable_name_col,66);
	}
	else {
		//dont display the column
		TblSetColumnUsable(table, AccountsTable_expected_balance_col,false);
		//make the anme column bigger
		TblSetColumnWidth(table,AccountsTable_name_col,111);
	}
		
	TblSetColumnUsable(table, AccountsTable_balance_col, true);
	
	AccountsTable_LoadData();
}
