/**
    $Author: dizzy $
    $Revision: 1.6 $
    Last updated: $Date: 2000/11/16 18:21:59 $

    $Log: EditAccountForm.c,v $
    Revision 1.6  2000/11/16 18:21:59  dizzy
    Added new fields to the Account structure.

    Revision 1.5  2000/11/15 18:58:42  dizzy
    New balance adjust option in transaction form menu.
    Help updated in Transaction form and Accounts form.
    Merged ShowScrollArrows functions.
    Updated a few variable and function names.

    Revision 1.4  2000/11/14 22:29:15  chaese
    small change in frmUpdateForm handler

    Revision 1.3  2000/11/11 23:48:50  chaese
    Eliminated flashing in form initialization

    Revision 1.2  2000/11/10 02:24:59  chaese
    changed order in which the account form is drawn: form first, then contents

    Revision 1.1  2000/11/10 01:25:55  chaese
    merged New Account and Account Details forms


    Copyright (C) 2000 Vlad Mereuta <dizzy@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


*/

#include "EditAccountForm.h"

/// the selected category
static UInt16	selected_cat=dmUnfiledCategory;
/// selected category name
static Char		cat_name [dmCategoryLength];
/// this form
static FormPtr	frm;


Boolean EditAcc_Save()
{	
	accountStructure		 	a;
	packedAccountStructure* 	pas;
	MemHandle					rech = 0;
	Int32						newbal;
	UInt32						uid;
	UInt16						index;

	if (!AccountsDB_record_is_new) {
		rech=DmGetRecord(AccountsDB_ref,AccountsDB_current_record);
		ErrFatalDisplayIf(!rech,"rech null");
		pas=(packedAccountStructure*) MemHandleLock(rech);
		AccountsDB_UnpackAccount(&a,pas);
		MemHandleUnlock(rech);
	}

	//for the moment make all accounts root by default
	a.depth=0;
	
    //if the balance is invalid it will be 0
   	newbal=String2Pence((Char*) GetTextPtr(EDIT_ACC_BALANCE));

   	a.credit_limit=String2Pence((Char*) GetTextPtr(EDIT_ACC_CREDIT));

   	a.description=(Char*) GetTextPtr(EDIT_ACC_DESCRIPTION);
   	if(!a.description) return false;
   	a.acname=(Char*) GetTextPtr(EDIT_ACC_NAME);
   	if(!a.acname) return false;

	if (AccountsDB_record_is_new) {
    	a.balance=newbal;
    	a.expected_balance=newbal;
    	index=AccountsDB_NewRecord(&a);
    	AccountsDB_SetRecordCategory(index,selected_cat);

    	//release record and set the dirty bit
    	DmReleaseRecord(AccountsDB_ref, index, true);

    	//recalculate the totals
    	AccountsDB_totals.dirty=true;

    	return true;
	} else {
		//write the data
		AccountsDB_PackAccount(&a,rech);

		AccountsDB_SetRecordCategory(AccountsDB_current_record,selected_cat);
		//release record and set the dirty bit
		DmReleaseRecord(AccountsDB_ref, AccountsDB_current_record, true); 

		if (newbal!=a.balance) { 
			DmRecordInfo(AccountsDB_ref,AccountsDB_current_record,NULL,&uid,NULL);
			AccountsDB_UpdateBalance(uid,newbal);
		}
		//sort the records in the db - just in case the name changed
		DmInsertionSort(AccountsDB_ref,(DmComparF *) AccountsDB_CompareRecords,0);
		return true;
	}
}

void EditAcc_Init()
{
	ControlPtr					ctl;
	MemHandle					rech;
	packedAccountStructure* 	pas;
	accountStructure 			as;
	UInt16						attr;
	Char*						s;
	

	if (AccountsDB_record_is_new) {
		FrmCopyTitle(frm, "Create New Account");

	    // Set the label of the category trigger. it defaults to the current
		// category unless the current one is all (in which case it defaults
		// to unfiled)
		if(AccountsDB_current_cat==dmAllCategories)
			selected_cat=dmUnfiledCategory;
        else
			selected_cat=AccountsDB_current_cat;

	    ctl = GetObjectPtrByForm (frm,EDIT_ACC_TYPE_TRIGGER);
    	CategoryGetName (AccountsDB_ref,selected_cat, cat_name);
    	CategorySetTriggerLabel(ctl, cat_name);
	}
	else {
		FrmCopyTitle(frm, "Account Details");
		FrmShowObject(frm, FrmGetObjectIndex(frm, EDIT_ACC_DELETE));
		rech=DmQueryRecord(AccountsDB_ref,AccountsDB_current_record);
		pas=(packedAccountStructure*) MemHandleLock(rech);

		if(!pas) return;

		AccountsDB_UnpackAccount(&as,pas);

		// init the category
		DmRecordInfo(AccountsDB_ref,AccountsDB_current_record,&attr,NULL,NULL);
		attr&=dmRecAttrCategoryMask;	
		selected_cat=attr;

		ctl = GetObjectPtrByForm (frm,EDIT_ACC_TYPE_TRIGGER);
		CategoryGetName (AccountsDB_ref,selected_cat, cat_name);
		CategorySetTriggerLabel(ctl, cat_name);

		SetFieldTextFromStr(EDIT_ACC_DESCRIPTION,as.description);
		SetFieldTextFromStr(EDIT_ACC_NAME,as.acname);

		s=Pence2String(as.balance);
		SetFieldTextFromStr(EDIT_ACC_BALANCE,s);
		s=Pence2String(as.credit_limit);
		SetFieldTextFromStr(EDIT_ACC_CREDIT,s);

		MemHandleUnlock(rech);

		//free memory allocated by Pence2String
		MemPtrFree(s);
	}
}

void	EditAcc_SelectCat()
{
	UInt16		category;
	Boolean 	categoryEdited;
	
	// Process the category popup list.  
	category = selected_cat;

	categoryEdited=CategorySelect(AccountsDB_ref,frm, EDIT_ACC_TYPE_TRIGGER,EDIT_ACC_TYPE_LIST,\
		false,&category,cat_name,1,categoryDefaultEditCategoryString);
	selected_cat=category;
}

Boolean EditAcc_HandleEvent(EventPtr event)
{
	Boolean 	handled = true;

	// Get the form pointer
	frm = FrmGetActiveForm();

	// Key Events
	switch(event->eType) {

	case frmUpdateEvent:
		if (event->data.frmUpdate.updateCode==frmRedrawUpdateCode) {
			handled = false;
		}			
		break;
	
	// Form Open
	case frmOpenEvent:		
		//draw form
		FrmDrawForm(frm);
		EditAcc_Init();
		break;
	
	// Controls
	case ctlSelectEvent:
		switch( event->data.ctlSelect.controlID ) {
			case EDIT_ACC_TYPE_TRIGGER:		//Select a category
				EditAcc_SelectCat();
				handled = true;
				break;
			case EDIT_ACC_OK:
				handled = true;
				//if something went wrong do nothing
				if(! EditAcc_Save()) break;
				//update the main form
				FrmUpdateForm(ACCOUNTS_MAIN_FORM,0);				
				FrmReturnToForm(ACCOUNTS_MAIN_FORM);
				break;
			
			case EDIT_ACC_CANCEL:
				//cancel button was pressed;
				//simply return to the main form
				handled = true;
				FrmUpdateForm(ACCOUNTS_MAIN_FORM,0);				
				FrmReturnToForm(ACCOUNTS_MAIN_FORM);
				break;

			case EDIT_ACC_DELETE:
				//delete button
				if (AccountsDB_record_is_new) {
					// we shouldn't be here, the button is invisible.
					// Just to be safe, though...
					handled = false;
				}
				else {
					handled = true;
					AccountsDB_DeleteRecord(AccountsDB_current_record);
					FrmUpdateForm(ACCOUNTS_MAIN_FORM,0);				
					FrmReturnToForm(ACCOUNTS_MAIN_FORM);
				}
				break;
				
			default:
				handled = false;
		};
		break;	
	// Everything else...
	default:
		handled = false;
	}	
	return handled;
}
