/**	
	$Author: chaese $
	$Revision: 1.1 $
	Last updated: $Date: 2000/11/16 02:38:32 $
	
	Copyright (C) 2000 Vlad Mereuta <dizzy@users.sourceforge.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	$Log: EditTransactionForm.c,v $
	Revision 1.1  2000/11/16 02:38:32  chaese
	Merged New Transaction and Transaction Details forms
	
	
*/

#include "EditTransactionForm.h"

//current form
static FormPtr	frm;
//stores the choosen date
static Int16	day,month,year;
//the ascii text coresponding to the date
static Char		date_txt[dateStringLength];	

//populated with account names to be displayed in the popup
static Char*	list_choices[100];
//no of accounts
static UInt16	choices_no;
//true if the current transaction is a paymen
static Boolean	is_debit;
//uid of the current account
static UInt32	cur_acc;

static transactionStructure t;
static MemHandle			rech;
static Char					p_acc[30];



Boolean EditTrans_Save()
{	
	UInt32	peer_acc;
	UInt16	rec_index;	

	//if the balance is invalid it will be 0
	t.ammount=String2Pence((Char*) GetTextPtr(EDIT_TRANS_AMMOUNT));
	if(t.ammount==0) return false; //ammount has to be != 0
	t.payee=(Char*) GetTextPtr(EDIT_TRANS_PAYEE);
	if(!t.payee) t.payee="n/a";
	
	t.date.day=day;
	t.date.year=year-1904;
	t.date.month=month;
	
	peer_acc=AccountsDB_FindAccountByName(LstGetSelectionText(GetObjectPtr(EDIT_TRANS_ACC_LIST),\
				LstGetSelection(GetObjectPtr(EDIT_TRANS_ACC_LIST))),dmAllCategories);
	t.method=TransactionsDB_GetTypeIDByName(LstGetSelectionText(GetObjectPtr(EDIT_TRANS_TYPE_LIST),\
			LstGetSelection(GetObjectPtr(EDIT_TRANS_TYPE_LIST))));
	if(is_debit) {
		t.source_acc=cur_acc;
		t.dest_acc=peer_acc;
	}
	else {
		t.source_acc=peer_acc;
		t.dest_acc=cur_acc;		
	}

	if (!TransactionsDB_record_is_new)
		TransactionsDB_DeleteRecord(TransactionsDB_current_record);

	if(FrmGetControlValue(frm,FrmGetObjectIndex(frm,EDIT_TRANS_CLEARED_CBX)))	//cleared
		rec_index=TransactionsDB_NewRecord(&t,true);
	else //not cleared
		rec_index=TransactionsDB_NewRecord(&t,false);	
		
	return true;
}

void EditTrans_InitAccList(Boolean use_acc_name)
{
	UInt16					recordNo,i,select=0;
	MemHandle				rh;
	packedAccountStructure*	pas;
	accountStructure 		as;
	Char					cn[dmCategoryLength];
	UInt32					uid;
	ControlPtr				ctl = GetObjectPtr(EDIT_TRANS_ACC_TRIGGER);
	

		
	choices_no=0;
  	for(i=0,recordNo=0;;i++,recordNo++) {
		//fetch record

		rh=DmQueryNextInCategory(AccountsDB_ref,&recordNo,dmAllCategories);		
		if(!rh) break;

		pas=(packedAccountStructure*) MemHandleLock(rh);
		ErrFatalDisplayIf(!pas, "pas is 0 - aborting");
		if(!pas) return;
		AccountsDB_UnpackAccount(&as,pas);

		//get the uid
		DmRecordInfo(AccountsDB_ref,recordNo,NULL,&uid,NULL);
		if(uid==cur_acc) {
			MemHandleUnlock(rh);
			continue; //do not show the current account as destination
		}
	
		CategoryGetName(AccountsDB_ref,db_GetRecordCategory(AccountsDB_ref,recordNo),cn);
		if (!TransactionsDB_record_is_new) {
			if(!use_acc_name && (t.source_acc==uid || t.dest_acc==uid)) {
						select=choices_no;
						StrCopy(p_acc,as.acname);
			}
		}
				
		if (is_debit) { //it is a payment
			if(StrCaselessCompare(cn,"Income")) {//accept everything except the accounts in income
				if(use_acc_name && !StrCompare(p_acc,as.acname)) select=choices_no;				
				list_choices[choices_no++]=as.acname;					
			}
		} else	if(StrCaselessCompare(cn,"Expense")) {
					if(use_acc_name && !StrCompare(p_acc,as.acname)) select=choices_no;				
					list_choices[choices_no++]=as.acname;					
				}
		MemHandleUnlock(rh);
	}
	LstSetListChoices(GetObjectPtr(EDIT_TRANS_ACC_LIST),list_choices,choices_no);
	LstSetSelection(GetObjectPtr(EDIT_TRANS_ACC_LIST),select);
	CategorySetTriggerLabel(ctl, p_acc);	
}

void EditTrans_SetDefaults()
{
	ControlPtr	ctl;
	Char*		s;

	if(TransactionsDB_record_is_new || t.source_acc==cur_acc) {
		FrmSetControlGroupSelection(frm,1,EDIT_TRANS_PAYMENT);
		is_debit=true;
	}
	else {
		FrmSetControlGroupSelection(frm,1,EDIT_TRANS_DEPOSIT);
		is_debit=false;		
	}
	ctl = GetObjectPtr(EDIT_TRANS_TYPE_TRIGGER);
	if (!TransactionsDB_record_is_new) {
		LstSetSelection(GetObjectPtr(EDIT_TRANS_TYPE_LIST),t.method);

		CategorySetTriggerLabel(ctl, TransactionsDB_types[t.method]);
		if(db_GetRecordCategory(TransactionsDB_ref,TransactionsDB_current_record)==CategoryFind(TransactionsDB_ref,"Cleared"))
			FrmSetControlValue(frm,FrmGetObjectIndex(frm,EDIT_TRANS_CLEARED_CBX),1);
		else 
			FrmSetControlValue(frm,FrmGetObjectIndex(frm,EDIT_TRANS_CLEARED_CBX),0);

		SetFieldTextFromStr(EDIT_TRANS_PAYEE,t.payee);
		s=Pence2String(t.ammount);
		SetFieldTextFromStr(EDIT_TRANS_AMMOUNT,s);

		//free memory allocated by Pence2String
		MemPtrFree(s);
	}
	else {
		CategorySetTriggerLabel(ctl, TransactionsDB_types[0]);
	}
}

void EditTrans_InitMethods()
{
	LstSetListChoices(GetObjectPtr(EDIT_TRANS_TYPE_LIST),TransactionsDB_types,TransactionsDB_types_no);
}
void EditTrans_UpdateDateField()
{
	DateFormatType dateFormat;
	dateFormat = PrefGetPreference(prefDateFormat);
	DateToAscii(month, day, year, dateFormat, date_txt);
	SetFieldTextFromStr(EDIT_TRANS_DATE,date_txt);
	
}
 
void EditTrans_InitDate()
{
	if (TransactionsDB_record_is_new)
	    DateSecondsToDate(TimGetSeconds(),& t.date);
	day=t.date.day;
	month=t.date.month;
	year=t.date.year+1904;
	EditTrans_UpdateDateField();	
}

 
void EditTrans_ChooseDate()
{
	Boolean c;

	c=SelectDay(selectDayByDay,&month,&day,&year,"Choose transaction date");
	if(c) EditTrans_UpdateDateField();
}

Boolean EditTrans_HandleEvent(EventPtr event)
{
	Boolean 					handled = true;
	packedTransactionStructure* pts;

	// Get the form pointer
	frm = FrmGetActiveForm();

	//get the uid of the current account
	DmRecordInfo(AccountsDB_ref,AccountsDB_current_record,NULL,&cur_acc,NULL);

		
	// Key Events
	switch(event->eType) {

	// Form Update
	case frmUpdateEvent:
        if (event->data.frmUpdate.updateCode==frmRedrawUpdateCode) {
            handled = false;
        }
        break;
  	
	// Form Open
	case frmOpenEvent:		
		//draw form
		FrmDrawForm(frm);
		if (!TransactionsDB_record_is_new) {
        	FrmShowObject(frm, FrmGetObjectIndex(frm, EDIT_TRANS_DELETE));
			FrmCopyTitle(frm, "Transaction Details");
			rech=DmQueryRecord(TransactionsDB_ref,TransactionsDB_current_record);
			pts=(packedTransactionStructure*) MemHandleLock(rech);
			TransactionsDB_UnpackTransaction(&t,pts);
			MemHandleUnlock(rech);
			is_debit=(t.source_acc==cur_acc);
			EditTrans_InitAccList(false);
		}
		else {
			FrmCopyTitle(frm, "Create New Transaction");
			StrCopy(p_acc, "Misc");
			EditTrans_InitAccList(true);
 		}
	
		EditTrans_InitDate();
		EditTrans_InitMethods();
		EditTrans_SetDefaults();
		break;
	
	// Controls
	case ctlSelectEvent:
		switch( event->data.ctlSelect.controlID ) {
			case EDIT_TRANS_OK:
				//if something went wrong do nothing
				if(! EditTrans_Save()) break;
				//update the main form
				FrmUpdateForm(TRANSACTIONS_FORM,0);				
				FrmReturnToForm(TRANSACTIONS_FORM);
				break;
			case EDIT_TRANS_DEPOSIT: //the radio buttons
				if(!is_debit) break;
				is_debit=false;
				StrCopy(p_acc, "Misc");		
				EditTrans_InitAccList(true);
				break;
			case EDIT_TRANS_PAYMENT:
				if(is_debit) break;
				is_debit=true;
				StrCopy(p_acc, "Misc");		
				EditTrans_InitAccList(true);
				break;				
			case EDIT_TRANS_CHOOSE_DATE:
				EditTrans_ChooseDate();
				break;
			case EDIT_TRANS_CANCEL:
				//cancel button was pressed;
				//simply return to the main form
				FrmUpdateForm(TRANSACTIONS_FORM,0);
				FrmReturnToForm(TRANSACTIONS_FORM);
				break;
			case EDIT_TRANS_DELETE:
				//delete button was pressed;
				//simply return to the main form
				// We shouldn't get here if the transactions is new,
				// but just to be safe...
				if (TransactionsDB_record_is_new) break;
				TransactionsDB_DeleteRecord(TransactionsDB_current_record);
				FrmUpdateForm(TRANSACTIONS_FORM,0);
				FrmReturnToForm(TRANSACTIONS_FORM);
				break;
				
			default:
				handled = false;
		};
		break;	
	// Everything else...
	default:
		handled = false;
	}	
	return handled;
}
