/**	
	$Author: dizzy $
	$Revision: 1.13 $
	Last updated: $Date: 2000/11/16 18:21:59 $
	
	$Log: MiscFunc.c,v $
	Revision 1.13  2000/11/16 18:21:59  dizzy
	Added new fields to the Account structure.
	
	Revision 1.12  2000/11/15 18:58:42  dizzy
	New balance adjust option in transaction form menu.
	Help updated in Transaction form and Accounts form.
	Merged ShowScrollArrows functions.
	Updated a few variable and function names.
	
	Revision 1.11  2000/11/06 22:27:02  dizzy
	Numbers are now localised and alligned to the right; the dates are displayed using the default format (these changes were submitted by Carsten Haese
	
	Revision 1.10  2000/10/27 22:29:20  dizzy
	All records are sorted now; fake indentation in account view
	
	Revision 1.9  2000/10/09 21:30:19  dizzy
	Fixed a bug which prevented having the - sign displayed for ammounts under 1
	

	Copyright (C) 2000 Vlad Mereuta <dizzy@users.sourceforge.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA	
*/

#include "MiscFunc.h"

Err	db_OpenOrCreate(DmOpenRef *dbP, UInt32 type, UInt32 creator, UInt16 mode, UInt16 cardNo, Char* name, Boolean* created)
{
	Err	err=0;

	*created = false;
	
	if (! DmFindDatabase(cardNo,name) ) { //database does not exist
		err = DmCreateDatabase(0, name, creator, type, false);
		if (err) { 
			return err;
			ErrFatalDisplayIf(err,"create db failed db");
		}
		*created = true; //set flag
	}

	//open database (has to exist at this point
	*dbP = DmOpenDatabase(cardNo,DmFindDatabase(cardNo,name),mode);
	if (! *dbP) return DmGetLastErr();

	return err;
}

void db_SetRecordCategory(DmOpenRef db, UInt16 index, UInt16 cat)
{
	UInt16	attr;

	DmRecordInfo(db,index,&attr,NULL,NULL);
	attr &= ~dmRecAttrCategoryMask;	
	attr|=cat;
	DmSetRecordInfo(db,index,&attr,NULL);			
}

UInt16 db_GetRecordCategory(DmOpenRef db, UInt16 index)
{
	UInt16	attr;

	DmRecordInfo(db,index,&attr,NULL,NULL);
	attr &= dmRecAttrCategoryMask;	
	return attr;
}

UInt32 db_GetRecordUid(DmOpenRef db, UInt16 id)
{
	UInt32 acc_uid;
	
	DmRecordInfo(db,id,NULL,&acc_uid,NULL);
	return acc_uid;
}

UInt16 db_GetRecordId(DmOpenRef db, UInt32 uid)
{	
	UInt16	acc_id;
	
	DmFindRecordByID(db,uid,&acc_id);
	return acc_id;
}

Char*	Pence2String(Int32 pence)
{
	Char* s=MemPtrNew(30);
	Char* t=MemPtrNew(30);
	Char cThousandSep, cDecimalSep;

	StrCopy(s,"");
	if(pence<0 && pence > -100) StrCopy(s,"-");
	StrIToA(t,pence/100);
	StrCat(s,t);
	StrCat(s,".");
	pence%=100; if(pence<0) pence*=-1;
	StrIToA(t,pence);
	if(pence<10) StrCat(s,"0");
	StrCat(s,t);
	MemPtrFree(t);
	LocGetNumberSeparators(
		PrefGetPreference(prefNumberFormat),
		&cThousandSep, &cDecimalSep);
	StrLocalizeNumber(s, cThousandSep, cDecimalSep);
	return s;
}

void DrawCharsRightAlign(const Char* s, const RectangleType* bounds)
{
	Int16 offset;
	RectangleType cliprect;

	WinGetClip(&cliprect);
	WinSetClip(bounds);
	offset = bounds->extent.x - FntCharsWidth(s, StrLen(s)) - 1;
	if (offset<0) offset = 0;
	WinDrawChars(s,StrLen(s),bounds->topLeft.x+offset, bounds->topLeft.y);
	WinSetClip(&cliprect);
}
 
Int32	String2Pence(Char* sum)
{
	Int32	res,p;
	Char*	pence;
	Char cThousandSep, cDecimalSep;

	//returns 0 if the pointer is 0
	if(!sum) return 0;

	LocGetNumberSeparators(
		PrefGetPreference(prefNumberFormat),
		&cThousandSep, &cDecimalSep);
	StrDelocalizeNumber(sum, cThousandSep, cDecimalSep);
	pence = StrChr(sum, '.');	
	if(pence) {
		//split the pence part from the main sum part
		pence[0]=0; //overwrite the dot
		pence++;
	}
	p=StrAToI(pence);
	
	
	res=StrAToI(sum);
	
	if(p<10 && StrLen(pence)!=2) p*=10;

	res*=100;
	if(res<0) res-=p;
		else res+=p;
	return res;
}

void*	GetObjectPtr(UInt16 objID)
{
	FormPtr frm;
	frm=FrmGetActiveForm();
	return (FrmGetObjectPtr (frm, FrmGetObjectIndex (frm,objID)));
}

void*	GetObjectPtrByForm(FormPtr frm,UInt16 objID)
{
	return (FrmGetObjectPtr (frm, FrmGetObjectIndex (frm,objID)));
}

void*	GetTextPtr(UInt16 objID)
{
	FieldPtr fp=(FieldPtr) GetObjectPtr(objID);
	
	return FldGetTextPtr(fp);
}

FieldPtr SetFieldTextFromHandle(Int16 fieldID, MemHandle txtH)
{
   MemHandle	oldTxtH;
   FormPtr	frm = FrmGetActiveForm();
   FieldPtr	fldP;


   // get the field and the field's current text handle.
   fldP = GetObjectPtrByForm(frm, fieldID);
   ErrNonFatalDisplayIf(!fldP, "missing field");
   oldTxtH = FldGetTextHandle(fldP);
   
   // set the field's text to the new text.
   FldSetTextHandle(fldP, txtH);
   FldDrawField(fldP);

   // free the handle AFTER we call FldSetTextHandle().
   if (oldTxtH) 
      MemHandleFree(oldTxtH);
   
   return fldP;
} 

FieldPtr SetFieldTextFromStr(Int16 fieldID, const Char* str)
{
   MemHandle      txtH;
   
   // get some space in which to stash the string.
   txtH  = MemHandleNew(StrLen(str) + 1);
   if (!txtH)
      return NULL;

   // copy the string to the locked handle.
   StrCopy(MemHandleLock(txtH), str);

   // unlock the string handle.
   MemHandleUnlock(txtH);
   
   // set the field to the handle
   return SetFieldTextFromHandle(fieldID, txtH);
} 

//clears the text from a field:
void  ClearFieldText(Int16 fieldID)
{
   SetFieldTextFromHandle(fieldID, NULL);
}

Int16	CompareAccountNames(Char* n1, Char* n2)
{
	Int16	len = (StrLen(n1)<StrLen(n2)) ? StrLen(n1):StrLen(n2);	
	Int16	i;
	
	for(i=0;i<len;i++)
		if(n1[i]<n2[i]) {
				if(n1[i]==':') return 1;
				else return -1;
		} else if(n1[i]>n2[i]) {
				if(n2[i]==':') return -1;
				else return 1;
			}
				
	if(StrLen(n1)<StrLen(n2)) return -1;
	else	if(StrLen(n1)<StrLen(n2)) return 1;
	return 0;
}

Char*	FakeAccountNameIndent(Char* src, Char* dst)
{
	UInt16 i;
	StrCopy(dst,src);
	for(i=0;i<StrLen(dst);i+=2)
		if(dst[i+1]==':') dst[i]=dst[i+1]=' ';
	return dst;
}

void ShowScrollArrows(UInt32 obj_up, UInt32 obj_dn, UInt32 total_items, UInt32 rows_in_table, UInt32 top_item)
{
    ControlPtr	ptrup, ptrdown;
    Boolean		enable;
	char 		*pageupchr="\001", *pagedownchr="\002";
	char 		*pageupgreychr="\003", *pagedowngreychr="\004";

    ptrup = GetObjectPtr(obj_up);
    ptrdown = GetObjectPtr(obj_dn);

    if ((total_items > rows_in_table) || top_item) {
    	// first element is not 0, hence we can scroll up
        if (top_item) {
            CtlSetLabel(ptrup, pageupchr);
            enable = true;
        } else {
            CtlSetLabel(ptrup, pageupgreychr);
            enable = false;
        }
        
		CtlShowControl(ptrup);
        CtlSetEnabled(ptrup, enable);
	
		// enable scrolling down if there are enough elements
        if ((top_item+rows_in_table) < total_items) {
            CtlSetLabel(ptrdown, pagedownchr);
            enable = true;
        } else {
            CtlSetLabel(ptrdown, pagedowngreychr);
            enable = false;
        }
        CtlShowControl(ptrdown);
        CtlSetEnabled(ptrdown, enable);
    } else {
        CtlHideControl(ptrup);
        CtlHideControl(ptrdown);
		CtlSetEnabled(ptrdown,false);
		CtlSetEnabled(ptrup,false);		
    }
}
