/**	
	$Author: dizzy $
	$Revision: 1.12 $
	Last updated: $Date: 2000/11/16 18:21:59 $
	
	$Log: MiscFunc.h,v $
	Revision 1.12  2000/11/16 18:21:59  dizzy
	Added new fields to the Account structure.
	
	Revision 1.11  2000/11/15 18:58:42  dizzy
	New balance adjust option in transaction form menu.
	Help updated in Transaction form and Accounts form.
	Merged ShowScrollArrows functions.
	Updated a few variable and function names.
	
	Revision 1.10  2000/11/06 22:27:02  dizzy
	Numbers are now localised and alligned to the right; the dates are displayed using the default format (these changes were submitted by Carsten Haese
	
	Revision 1.9  2000/10/27 22:29:20  dizzy
	All records are sorted now; fake indentation in account view
	
	Revision 1.8  2000/10/12 17:17:24  dizzy
	Email address update; lots of bugfixes
	
	Revision 1.7  2000/09/15 10:10:48  dizzy
	First useful version
	
	Revision 1.6  2000/09/08 10:39:52  dizzy
	Databases are now opened by name.
	
	Revision 1.5  2000/09/04 14:23:07  dizzy
	General code cleanup
	
	Revision 1.3  2000/08/23 12:08:30  dizzy
	Added details dialog for accounts
	
	Revision 1.2  2000/08/21 15:04:20  dizzy
	Fixed more bugs. Added scrolling support
	
	Revision 1.1  2000/08/21 08:59:30  dizzy
	Reorganised files

	Copyright (C) 2000 Vlad Mereuta <dizzy@users.sourceforge.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
	
*/

#ifndef __MISC_FUNC__
#define __MISC_FUNC__

#include <PalmOS.h>

/// application creator
#define PALM_COINS_CREATOR	'DyCo'


/**
	Sets the category of a given record
	@param db		database to seek the record in
	@param index	the index of the record
	@param cat		the (new) category
*/
void db_SetRecordCategory(DmOpenRef db, UInt16 index, UInt16 cat);

/**
	Returns the category of a given record
	@param db		database to seek the record in
	@param index	the index of the record
	@return	the record category
*/

UInt16 db_GetRecordCategory(DmOpenRef db, UInt16 index);


/**
	opens or creates a database
	@param dbP	reference to the db
	@param type database type
	@param creator database creator
	@param mode	mode in which to open the database
	@param name db name
	@param created on return is set to true if the db had to be created
*/

Err	db_OpenOrCreate(DmOpenRef *dbP, UInt32 type, UInt32 creator, UInt16 mode, UInt16 cardNo, Char* name, Boolean* created);

/**
	Fetch the uid of a record
	@param db	reference to the database which contains the record
	@param id	the record id
	
	@return the record uid
*/
UInt32 db_GetRecordUid(DmOpenRef db, UInt16 id);

/**
	Fetch the uid of a record
	@param db	reference to the database which contains the record
	@param uid	the record uid
	
	@return the record id
*/
UInt16 db_GetRecordId(DmOpenRef db, UInt32 uid);


/**
	Get a 32bit int containg a sum of money in pence and return a pointer
	to a proper formatted string.
	
	the pointer has to be deallocated after use with MemPtrFree
	
*/
Char* Pence2String(Int32 pence);

/**
	draw right aligned text
	@param s string to be drawn
	@param bounds rectangle inside which the text will be displayed
*/
void DrawCharsRightAlign(const Char* s, const RectangleType* bounds);

/**
	Get a string containing a sum of money and convert it to pence
	@param sum the sting containing the formatted ammount
	@return	the ammount of pence in the sum
*/
Int32 String2Pence(Char* sum);

/**
	returns a pointer to a given object in the current form
*/
void* GetObjectPtr(UInt16 objID);

/**
	returns a pointer to a given object in a given form
*/
void*	GetObjectPtrByForm(FormPtr frm,UInt16 objID);

/**
	returns pointer to the text in a given field in the current form
*/
void*	GetTextPtr(UInt16 objID);

FieldPtr SetFieldTextFromHandle(Int16 fieldID, MemHandle txtH);

/// Allocates new handle and copies incoming string
FieldPtr SetFieldTextFromStr(Int16 fieldID, const Char* str);

///clears the text from a field:
void  ClearFieldText(Int16 fieldID);

/**
	the function is similar to StrCompare with the only diference that
	':'> text chars
*/
Int16	CompareAccountNames(Char* n1, Char* n2);

/**
	the function replaces all the tokens of the form '?:' with '  '
*/
Char*	FakeAccountNameIndent(Char* src, Char* dst);

/**
	displays scroll arrows for a table (if needed)
	@param obj_up id of the object containing the up arrow
	@param obj_dn id of the object containing the down arrow
	@param total_items total no of items available to place in the table
	@param rows_in_table no of rows in the table
	@param top_item the no of item currently at the top of the table
	
*/
void ShowScrollArrows(UInt32 obj_up, UInt32 obj_dn, UInt32 total_items, UInt32 rows_in_table, UInt32 top_item);

#endif
