/**	
	$Author: chaese $
	$Revision: 1.11 $
	Last updated: $Date: 2000/11/16 02:38:32 $
	
	$Log: PalmCoins.c,v $
	Revision 1.11  2000/11/16 02:38:32  chaese
	Merged New Transaction and Transaction Details forms
	
	Revision 1.10  2000/11/10 01:25:56  chaese
	merged New Account and Account Details forms
	
	Revision 1.9  2000/10/12 17:17:24  dizzy
	Email address update; lots of bugfixes
	
	Revision 1.8  2000/09/15 10:10:48  dizzy
	First useful version
	
	Revision 1.7  2000/09/13 11:50:27  dizzy
	The transaction module is fully functional
	
	Revision 1.6  2000/09/08 10:41:10  dizzy
	Transactiondatabas initial support and column display preferences support
	
	Revision 1.5  2000/09/04 14:23:07  dizzy
	General code cleanup
	
	Revision 1.2  2000/08/23 12:08:30  dizzy
	Added details dialog for accounts
	
	Revision 1.1  2000/08/21 08:59:30  dizzy
	Reorganised files

	Copyright (C) 2000 Vlad Mereuta <dizzy@users.sourceforge.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
	
*/

#include "PalmCoins.h"

static Boolean ApplicationHandleEvent( EventPtr event )
{
	Boolean	eventHandled = true;

	if (event->eType == frmLoadEvent) {
		// Load the form resource.		
		UInt16	formID	= event->data.frmLoad.formID;
		FormPtr	frm = FrmInitForm(formID);

		FrmSetActiveForm(frm);		

		// set the event handler for the form. the handler of the currently
		// active form is called by FrmHandleEvent each time is receives an
		// event.
		//
		// for this app, all forms are handled by the same routine.
			
		switch( formID ){
			case ACCOUNTS_MAIN_FORM:
				FrmSetEventHandler(frm, AccountsForm_HandleEvent);
				break;
			case EDIT_ACC_FORM:
				FrmSetEventHandler(frm, EditAcc_HandleEvent);
				break;
			case PREF_FORM:
				FrmSetEventHandler(frm, PrefForm_HandleEvent);
				break;
			case TRANSACTIONS_FORM:
				FrmSetEventHandler(frm, TransactionsForm_HandleEvent);
				break;
			case EDIT_TRANS_FORM:
				FrmSetEventHandler(frm, EditTrans_HandleEvent);
				break;
			// if you add more forms to the app, put a case in for each new
			// form...
			default:
				eventHandled = false;
				break;				
		}
	} else {
		eventHandled = false;
	}
	
	return eventHandled;
}


static void ApplicationEventLoop(void)
{
	UInt16 error;
	EventType event;

	do {
		EvtGetEvent(&event, evtWaitForever);
		if(!SysHandleEvent(&event)) //not system event
			if(!MenuHandleEvent(NULL,&event,&error))
				if(!ApplicationHandleEvent(&event)) //our event handling function
					FrmDispatchEvent(&event);
	} while (event.eType != appStopEvent);
}

static UInt16 StartApplication( void )
{
	//load the application preferences
	Prefs_Get();
	
	//init the databases
	TransactionsDB_Init();

	return 	AccountsDB_Init();
}


static void StopApplication( void )
{
	// Close all open forms,  this will force any unsaved data to
	// be written to the database.
	FrmCloseAllForms();
	
	// cleanup the DB.
	
	AccountsDB_Close();
	TransactionsDB_Close();	
}

/*
 * 	Application entry point
 */
UInt32	PilotMain( UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags )
{
	
	// If normal launch
	if ( cmd == sysAppLaunchCmdNormalLaunch )
	{
		UInt16 error;

		//we can start the application
		error = StartApplication();
		if ( error ) return error;

		// first form.
		FrmGotoForm(ACCOUNTS_MAIN_FORM);
		
		// run the event loop. returns to exit.		
		ApplicationEventLoop();
		StopApplication();
	}
	return 0;
}	

