/**	
	$Author: chaese $
	$Revision: 1.11 $
	Last updated: $Date: 2000/11/16 02:38:32 $
	
	Copyright (C) 2000 Vlad Mereuta <dizzy@users.sourceforge.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	$Log: TransactionsDB.c,v $
	Revision 1.11  2000/11/16 02:38:32  chaese
	Merged New Transaction and Transaction Details forms
	
	Revision 1.10  2000/11/15 18:58:42  dizzy
	New balance adjust option in transaction form menu.
	Help updated in Transaction form and Accounts form.
	Merged ShowScrollArrows functions.
	Updated a few variable and function names.
	
	Revision 1.9  2000/11/12 19:10:11  dizzy
	Structures and functions modified to accomodate memorising the cleared atribute as a category
	rather then a member in the Transaction structure.
	
	Revision 1.8  2000/11/12 14:34:11  dizzy
	Initial transaction category support
	
	Revision 1.7  2000/11/06 22:27:02  dizzy
	Numbers are now localised and alligned to the right; the dates are displayed using the default format (these changes were submitted by Carsten Haese
	
	Revision 1.6  2000/10/27 22:29:20  dizzy
	All records are sorted now; fake indentation in account view
	
	Revision 1.5  2000/10/21 14:35:29  dizzy
	added delete functionality (for accounts and transactions)
	
	Revision 1.4  2000/10/12 17:17:24  dizzy
	Email address update; lots of bugfixes
	
	Revision 1.3  2000/09/15 10:10:48  dizzy
	First useful version
	
	Revision 1.2  2000/09/13 11:50:27  dizzy
	The transaction module is fully functional
*/

//define this to ignore the external definition of the vars below in the header file
#define __TRANSACTIONS_DB_C__

#include "TransactionsDB.h"


//the variables below have to be accessed in different object files
//see the .h files for further explanations

UInt16			TransactionsDB_types_no=6;
DmOpenRef		TransactionsDB_ref;
UInt32			TransactionsDB_num_cat;
UInt16			TransactionsDB_current_cat;
Int16			TransactionsDB_current_record;
Boolean			TransactionsDB_record_is_new;
Char*			TransactionsDB_types[6]={"Card","Cash","Cheque","Standing Order","Direct Debit","Transfer"};

payment_method	TransactionsDB_GetTypeIDByName(Char* name)
{
	if(!StrCompare(name,"Card")) return pm_card;
	if(!StrCompare(name,"Cash")) return pm_cash;
	if(!StrCompare(name,"Cheque")) return pm_cheque;
	if(!StrCompare(name,"Standing Order")) return pm_so;
	if(!StrCompare(name,"Direct Debit")) return pm_dd;
	if(!StrCompare(name,"Transfer")) return pm_transfer;
	//top stop the compiler from complaining ;)
	ErrFatalDisplay("Name not correct");
	return 0;
}

Int32	TransactionsDB_RecordGetAmmount(UInt32 id)
{
	Int16						record;
	MemHandle					rech;
	packedTransactionStructure*	pts;
	transactionStructure		ts;

	DmFindRecordByID(TransactionsDB_ref,id,&record);
	
	rech=DmQueryRecord(TransactionsDB_ref,record);
	ErrFatalDisplayIf(!rech, "rech is 0 - aborting");

	pts=(packedTransactionStructure *) MemHandleLock(rech);
	ErrFatalDisplayIf(!pts, "pts is 0 - aborting");

	TransactionsDB_UnpackTransaction(&ts,pts);
	MemHandleUnlock(rech);
	return ts.ammount;
}

void	TransactionsDB_RecordGetDate(UInt32 id, DateType * pDate)
{
	Int16						record;
	MemHandle					rech;
	packedTransactionStructure* pts;

	DmFindRecordByID(TransactionsDB_ref,id,&record);

	rech=DmQueryRecord(TransactionsDB_ref,record);
	ErrFatalDisplayIf(!rech, "rech is 0 - aborting");

	pts=(packedTransactionStructure*) MemHandleLock(rech);
	ErrFatalDisplayIf(!pts, "pts is 0 - aborting");

	pDate->month = pts->date.month;	
	pDate->day = pts->date.day;	
	pDate->year = pts->date.year;	

	MemHandleUnlock(rech);
}


void	TransactionsDB_GetUnpackedRecord(UInt32 id,transactionStructure* ts)
{
	Int16						record;
	MemHandle					rech;
	packedTransactionStructure*	pts;

	DmFindRecordByID(TransactionsDB_ref,id,&record);
	
	rech=DmQueryRecord(TransactionsDB_ref,record);
	ErrFatalDisplayIf(!rech, "rech is 0 - aborting");

	pts=(packedTransactionStructure *) MemHandleLock(rech);
	ErrFatalDisplayIf(!pts, "pts is 0 - aborting");

	TransactionsDB_UnpackTransaction(ts,pts);
	MemHandleUnlock(rech);
}


void	TransactionsDB_RecordGetDestAcc(UInt32 id,accountStructure* as)
{
	Int16						record;
	MemHandle					rech;
	packedTransactionStructure*	pts;


	DmFindRecordByID(TransactionsDB_ref,id,&record);

	rech=DmQueryRecord(TransactionsDB_ref,record);
	ErrFatalDisplayIf(!rech, "rech is 0 - aborting");

	pts=(packedTransactionStructure*) MemHandleLock(rech);
	ErrFatalDisplayIf(!pts, "pts is 0 - aborting");

	AccountsDB_GetUnpackedRecord(pts->dest_acc,as);
	MemHandleUnlock(rech);
	
}

void	TransactionsDB_RecordGetSrcAcc(UInt32 id,accountStructure* as)
{
	Int16						record;
	MemHandle					rech;
	packedTransactionStructure*	pts;


	DmFindRecordByID(TransactionsDB_ref,id,&record);

	rech=DmQueryRecord(TransactionsDB_ref,record);
	ErrFatalDisplayIf(!rech, "rech is 0 - aborting");

	pts=(packedTransactionStructure*) MemHandleLock(rech);
	ErrFatalDisplayIf(!pts, "pts is 0 - aborting");

	AccountsDB_GetUnpackedRecord(pts->source_acc,as);
	MemHandleUnlock(rech);
	
}

void TransactionsDB_SetRecordCategory(UInt16 index, UInt16 cat)
{
	db_SetRecordCategory(TransactionsDB_ref,index,cat);
}

void TransactionsDB_UnpackTransaction (transactionStructure* trans, const packedTransactionStructure* packed_trans)
{
	Char*	s = packed_trans->payee;
	
	trans->source_acc = packed_trans->source_acc;
	trans->dest_acc = packed_trans->dest_acc;
	trans->date = packed_trans->date;
	trans->ammount = packed_trans->ammount;
	trans->method = packed_trans->method;

	trans->payee = s;
	s += StrLen(s) + 1;
	trans->note = s;
}

void TransactionsDB_PackTransaction (transactionStructure* trans, MemHandle db_entry)
{
	// figure out necessary size
	UInt32	length = 0;
	MemPtr	s;
	UInt32	offset = 0;

	//for some reason cleared seems to have the size 1 byte bigger then what sizeof reports
	length = sizeof(trans->source_acc) + sizeof(trans->dest_acc) + sizeof(trans->date) +  \
		 sizeof(trans->ammount) + sizeof(trans->method) +StrLen(trans->payee) + StrLen(trans->note) + 3;  // 4 for 2 string terminators

	// resize the VoidHand
	if (MemHandleResize(db_entry, length) == 0) {
		// copy the fields
		s = MemHandleLock(db_entry);
		offset = 0;
		DmWrite(s, offset, (MemPtr) &trans->source_acc,sizeof(trans->source_acc));
		offset += sizeof(trans->source_acc);
		DmWrite(s, offset, (MemPtr) &trans->dest_acc,sizeof(trans->dest_acc));
		offset += sizeof(trans->dest_acc);
		DmWrite(s, offset, (MemPtr) &trans->date,sizeof(DateType));
		offset += sizeof(DateType);;
		DmWrite(s, offset, (MemPtr) &trans->ammount,sizeof(trans->ammount));
		offset += sizeof(trans->ammount);
		DmWrite(s, offset, (MemPtr) &trans->method,sizeof(trans->method));
		offset += sizeof(trans->method);
		
		DmStrCopy(s, offset, (Char*) trans->payee);
		offset += StrLen(trans->payee) + 1;
		DmStrCopy(s, offset, (Char*) trans->note);
				
		MemHandleUnlock(db_entry);
	}
}

Err		TransactionsDB_Init()
{
	UInt32	mode = dmModeReadWrite;
	Err	err = 0;
	Boolean created;
	MemPtr c;

	// Find the database.  If it doesn't exist, create it.
	err=db_OpenOrCreate(&TransactionsDB_ref, TRANSACTIONSDB_TYPE,PALM_COINS_CREATOR, mode,0, TRANSACTIONSDB_NAME, &created);

	//have to initialise the application info
	c = TransactionsDB_GetLockedAppInfo();
	MemPtrUnlock(c);
	
	//set the default category to Cl/Uncl
	TransactionsDB_current_cat=CategoryFind(TransactionsDB_ref,"Cl/Uncl");
	return err;
}

MemPtr TransactionsDB_AppInfoInit(LocalID dbID,UInt16 cardNo)
{
	UInt16					wordValue;
	LocalID					appInfoID;
	palmCoinsAppInfoType*	nilP = 0;
	palmCoinsAppInfoType* 	appInfoP;

	MemHandle h;

	
	h = DmNewHandle(TransactionsDB_ref, sizeof (palmCoinsAppInfoType));
	ErrFatalDisplayIf (!h,  "Cannot allocate app info block");

	//get the local id of the newly allocated appinfo record
	appInfoID = MemHandleToLocalID (h);
	//record the id
	DmSetDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL, NULL,NULL, &appInfoID, NULL, NULL, NULL);
	
	appInfoP = MemLocalIDToLockedPtr(appInfoID, cardNo);

	// Clear the app info block.
	DmSet (appInfoP, 0, sizeof(palmCoinsAppInfoType), 0);

	// Initialize the categories.
	CategoryInitialize ((AppInfoPtr) appInfoP, TRANSACTIONS_CATEGORIES);

	// I don't know what this field is used for.
	wordValue = 0xFFFF;
	DmWrite (appInfoP, (UInt32)&nilP->dirtyAppInfo, &wordValue,
		sizeof(appInfoP->dirtyAppInfo));

	// Initialize the sort order.
	DmSet (appInfoP, (UInt32)&nilP->sortOrder, sizeof(appInfoP->sortOrder), 
		soName);

	return appInfoP;
}

MemPtr	TransactionsDB_GetLockedAppInfo()
{
	UInt16	cardNo;
	LocalID	dbID;
	LocalID	appInfoID;
	Err	err;

	err = DmOpenDatabaseInfo(TransactionsDB_ref, &dbID, NULL, NULL,&cardNo, NULL);
	ErrFatalDisplayIf (err,  "Get getting application info block");
	err = DmDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL,NULL, NULL, &appInfoID, NULL, NULL, NULL);
	ErrFatalDisplayIf (err,  "Get getting application info block");
	
	if(appInfoID==NULL) return TransactionsDB_AppInfoInit(dbID,cardNo);
	//a bit useless - to be removed unless code will be inserted here
	return MemLocalIDToLockedPtr(appInfoID, cardNo);
}

void	TransactionsDB_Close()
{
	DmCloseDatabase(TransactionsDB_ref);
}

UInt16	TransactionsDB_NewRecord(transactionStructure* s, Boolean cleared)
{
	MemHandle					h,p;
	UInt16						index=0;
	accountStructure		 	a;
	MemHandle					rech;
	
	h=DmNewHandle(TransactionsDB_ref,1);
	ErrFatalDisplayIf(!h,"Cannot create new transaction. No space left?");

	TransactionsDB_PackTransaction(s,h);
	p=MemHandleLock(h);
	index=DmFindSortPosition(TransactionsDB_ref,p, 0, (DmComparF *) TransactionsDB_CompareRecords, 0);
	//have no clue whatsoever why I have to use Ptr and not handle, but otherwise it does not work
	MemPtrUnlock(p);
	
	
	DmAttachRecord(TransactionsDB_ref,&index,h,0);	
	DmReleaseRecord(TransactionsDB_ref, index, true); //true sets the dirty bit			


	//update the balances for the accounts involved in the transfer
	rech=AccountsDB_GetUnpackedRecord(s->source_acc,&a);

	if(cleared) a.balance-=s->ammount;
	a.expected_balance-=s->ammount;

	//write the data
	AccountsDB_PackAccount(&a,rech);

	rech=AccountsDB_GetUnpackedRecord(s->dest_acc,&a);

	if(cleared) a.balance+=s->ammount;
	a.expected_balance+=s->ammount;

	//write the data
	AccountsDB_PackAccount(&a,rech);

	AccountsDB_totals.dirty=true;

	//set the category
	if(cleared) TransactionsDB_SetRecordCategory(index,CategoryFind(TransactionsDB_ref,"Cleared"));
	else TransactionsDB_SetRecordCategory(index,CategoryFind(TransactionsDB_ref,"Uncleared"));

	return index;
}

void TransactionsDB_DeleteRecord(UInt16 id)
{
	accountStructure			a;
	packedTransactionStructure*	pts;
	transactionStructure		t;
	MemHandle					rech;

	rech=DmQueryRecord(TransactionsDB_ref,id);
	pts=(packedTransactionStructure*) MemHandleLock(rech);
	TransactionsDB_UnpackTransaction(&t,pts);
	MemHandleUnlock(rech);
		
	//cancel the old transfer - source
	rech=AccountsDB_GetUnpackedRecord(t.source_acc,&a); 

	if (db_GetRecordCategory(TransactionsDB_ref,id)==CategoryFind(TransactionsDB_ref,"Cleared"))
		a.balance+=t.ammount;
	a.expected_balance+=t.ammount;

	//write the data
	AccountsDB_PackAccount(&a,rech);

	//cancel old transaction - dest acc
	rech=AccountsDB_GetUnpackedRecord(t.dest_acc,&a); 

	if (db_GetRecordCategory(TransactionsDB_ref,id)==CategoryFind(TransactionsDB_ref,"Cleared"))
		a.balance-=t.ammount;
	a.expected_balance-=t.ammount;

	//write the data
	AccountsDB_PackAccount(&a,rech);

	DmRemoveRecord(TransactionsDB_ref,id);				
}

Int16	TransactionsDB_CompareRecords(packedTransactionStructure* rec1,packedTransactionStructure* rec2,Int16 other,SortRecordInfoPtr sr1,SortRecordInfoPtr sr2,MemHandle app_info)
{
	if(rec1->date.year<rec2->date.year) return -1;
	else	if (rec1->date.year>rec2->date.year) return 1;
			else	if(rec1->date.month<rec2->date.month) return -1;
					else	if(rec1->date.month>rec2->date.month) return 1;
							else	if(rec1->date.day>rec2->date.day) return -1;
									else	if(rec1->date.day<rec2->date.day) return 1;
											else return 0;
}	
