/**	
	$Author: dizzy $
	$Revision: 1.11 $
	Last updated: $Date: 2000/11/19 19:19:44 $
	
	$Log: TransactionsForm.c,v $
	Revision 1.11  2000/11/19 19:19:44  dizzy
	The way the tables are loaded was re-written. The account table works perfectly now
	but the transaction table still has a few problems.
	
	Revision 1.10  2000/11/16 18:21:59  dizzy
	Added new fields to the Account structure.
	
	Revision 1.9  2000/11/16 02:38:32  chaese
	Merged New Transaction and Transaction Details forms
	
	Revision 1.8  2000/11/15 19:20:03  dizzy
	Finally fixed bug which prevented displaying all the transactions
	
	Revision 1.7  2000/11/15 18:58:42  dizzy
	New balance adjust option in transaction form menu.
	Help updated in Transaction form and Accounts form.
	Merged ShowScrollArrows functions.
	Updated a few variable and function names.
	
	Revision 1.6  2000/11/13 18:16:14  dizzy
	Fixed scrolling in transaction form. Added help option in menu
	
	Revision 1.5  2000/11/12 14:34:12  dizzy
	Initial transaction category support
	
	Revision 1.4  2000/11/06 22:27:02  dizzy
	Numbers are now localised and alligned to the right; the dates are displayed using the default format (these changes were submitted by Carsten Haese
	
	Revision 1.3  2000/11/02 19:05:17  dizzy
	Few bugfixes
	
	Revision 1.2  2000/10/12 17:17:24  dizzy
	Email address update; lots of bugfixes
	
	Revision 1.1  2000/10/11 21:12:23  dizzy
	

	Copyright (C) 2000 Vlad Mereuta <dizzy@users.sourceforge.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#define __TRANSACTIONS_FORM_C__

#include "TransactionsForm.h"

/// name of the current category
//static Char cat_name [dmCategoryLength];
///this form
static	FormPtr	frm;
static	Boolean item_selected;

UInt32	curr_account;
UInt32	curr_transaction;
Boolean	curr_transaction_exists;
static Char cat_name [dmCategoryLength];

void	TransactionsForm_InitCat()
{
	ControlPtr	ctl;	
	// Set the label of the category trigger.
	ctl = GetObjectPtr (TRANS_CAT_TRIGGER);
	CategoryGetName (TransactionsDB_ref, TransactionsDB_current_cat, cat_name);
	CategorySetTriggerLabel(ctl, cat_name);
}

void	TransactionsForm_SelectCat()
{
 	UInt16		category=TransactionsDB_current_cat;
	Boolean 	categoryEdited;

	//have to do this due to some odd behaviour 
	frm = FrmGetActiveForm();
	
	categoryEdited=CategorySelect(TransactionsDB_ref,frm, TRANS_CAT_TRIGGER,TRANS_CAT_LIST,false,&category,
		cat_name,4,categoryHideEditCategory);
	if(categoryEdited || category!=TransactionsDB_current_cat) {
		TransactionsDB_current_cat=category;	
		//start from the beginning
		TransactionsTable_top_visible_item=0;
	}
}

void	TransactionsForm_AdjustAccBalance()
{
	Char	r[9];
	UInt16	rez;
	
	rez=FrmCustomResponseAlert(ADJUST_BALANCE_ALERT,NULL,NULL,NULL,r,9,NULL);
	if(!rez) AccountsDB_UpdateBalance(curr_account,String2Pence(r));			
}

void	TransactionsForm_DisplayTotals()
{
	Char* 			s;
	RectangleType	rp;
		
	RctSetRectangle(&rp,70,129,90,11);
	WinEraseRectangle(&rp,0);

	
	s=Pence2String(AccountsDB_GetBalance(curr_account));
	RctSetRectangle(&rp,115,129,45,11);
	DrawCharsRightAlign(s, &rp);
	MemPtrFree(s);
	
	s=Pence2String(AccountsDB_GetExpectedBalance(curr_account));
	RctSetRectangle(&rp,70,129,45,11);
	DrawCharsRightAlign(s, &rp);
	MemPtrFree(s);	

	//add some lines at the bottom
	WinDrawLine(0,129,160,129);
	WinDrawLine(0,140,160,140);	
}

Boolean TransactionsForm_HandleEvent( EventPtr event )
{
	Boolean 			handled = true;
	ControlPtr			ptr;
	accountStructure	as;
	FormPtr				f;

	// Get the form pointer
	frm = FrmGetActiveForm();


	// Key Events
	switch(event->eType) {

	// Form Open
	case frmOpenEvent:

		//set the form title to the name of the account
		AccountsDB_GetUnpackedRecord(curr_account,&as);
		FrmSetTitle(frm,as.acname);
		if(coins_pref.disp_payee) 
			FrmSetControlGroupSelection(frm,1,TRANS_OPT_PAYEE);
		else FrmSetControlGroupSelection(frm,1,TRANS_OPT_ACCOUNT);
		curr_transaction_exists=false;
		TransactionsTable_Init();
		item_selected=false;

		//display the category name
		TransactionsForm_InitCat();
		//draw form

		FrmDrawForm(frm);
		
		//display the running totals
		TransactionsForm_DisplayTotals();
		break;

	case keyDownEvent:
		switch( event->data.keyDown.chr ) {
			case pageUpChr:			// Scroll up key presed?
				//link the hard button functionality to the soft up/down controls
				ptr = GetObjectPtr(TRANS_PGUP);
				CtlHitControl(ptr);
				handled = true;
				break;								
			case pageDownChr:		// Scroll down key presed?						
				ptr = GetObjectPtr(TRANS_PGDN);
				CtlHitControl(ptr);
				handled = true;
				break;
			default:
				handled = false;
		};
		break;
	
	// Controls
	case ctlSelectEvent:
		switch( event->data.ctlSelect.controlID ) {
			case TRANS_NEW:
				item_selected=false;
				TransactionsDB_record_is_new = true;
				FrmPopupForm(EDIT_TRANS_FORM);				
				FrmUpdateForm(TRANSACTIONS_FORM,0);
				//cancel the current selection
				handled = true;
				break;
			case TRANS_DETAILS:
				if(!item_selected) {
					FrmAlert(NO_SELECTED_TRANS);
					break;
				}
					
				TransactionsDB_record_is_new = false;
				FrmPopupForm(EDIT_TRANS_FORM);		
				item_selected=false;						
				FrmUpdateForm(TRANSACTIONS_FORM,0);
				handled = true;
				//cancel the current selection
				break;
			case TRANS_ACCTS:
				//cancel the current selection
				item_selected=false;
				//erase the up/down arrows 
				CtlSetEnabled(GetObjectPtr(TRANS_PGDN),false);
				CtlSetEnabled(GetObjectPtr(TRANS_PGDN),false);
		        CtlHideControl(GetObjectPtr(TRANS_PGUP));
		        CtlHideControl(GetObjectPtr(TRANS_PGDN));
				FrmUpdateForm(ACCOUNTS_MAIN_FORM,0);				
				FrmReturnToForm(ACCOUNTS_MAIN_FORM);
				handled = true;
				break;				
			case TRANS_CAT_TRIGGER:		//Select a category
				TransactionsForm_SelectCat();
				FrmUpdateForm(TRANSACTIONS_FORM,0);
				handled = true;
				break;
			case TRANS_OPT_PAYEE: //switch to payee vieew
				if(coins_pref.disp_payee) break;
				else coins_pref.disp_payee=true;
				Prefs_Save();
				FrmUpdateForm(TRANSACTIONS_FORM,0);
				handled = true;
				break;
			case TRANS_OPT_ACCOUNT: //switch to account view
				if(!coins_pref.disp_payee) break;
				else coins_pref.disp_payee=false;
				Prefs_Save();
				FrmUpdateForm(TRANSACTIONS_FORM,0);
				handled = true;
				break;
				
			case TRANS_PGUP:
				if(TransactionsTable_top_visible_item) {
					if(TransactionsTable_top_visible_item>=TransactionsTable_rows) 
						TransactionsTable_top_visible_item-=TransactionsTable_rows-1;
					else
						TransactionsTable_top_visible_item=0;
					FrmUpdateForm(TRANSACTIONS_FORM,0);
				}
				handled = true;
				break;								
			case TRANS_PGDN:		// Scroll down control presed?	
				if(TransactionsTable_total_items>TransactionsTable_rows) {					
					if(TransactionsTable_top_visible_item+TransactionsTable_rows+1<TransactionsTable_total_items)
						TransactionsTable_top_visible_item+=TransactionsTable_rows-1;
					else
						TransactionsTable_top_visible_item=TransactionsTable_total_items-TransactionsTable_rows;
				
					FrmUpdateForm(TRANSACTIONS_FORM,0);
				}
				handled = true;
				break;
				
			default:
				handled = false;
		}
		break;
	
	// a item in the table was selected
	// should probably be moved in the table files
	case tblSelectEvent:		
		//save the record on the current row into DetailRecord variable
		DmFindRecordByID(TransactionsDB_ref,TblGetRowData(GetObjectPtr(TRANS_TABLE),event->data.tblSelect.row),&TransactionsDB_current_record);
		item_selected=true;
//		if(event->data.tblSelect.column==TransactionsTable_name_col) {
//			FrmPopupForm(ACC_DETAILS_FORM);
//			FrmUpdateForm(ACCOUNTS_MAIN_FORM,0);
//		}
//		else {
//				FrmAlert(NOT_IMPLEMENTED_ALERT);
//				FrmUpdateForm(ACCOUNTS_MAIN_FORM,0);
//				// to be implemented
//		}
//		handled=true;
//		break;
			
	// Menus
	case menuEvent:
		switch (event->data.menu.itemID) {
			case TRANS_MENU_NEW:
				//cancel the current selection
				item_selected=false;
				TransactionsDB_record_is_new = true;
				FrmPopupForm(EDIT_TRANS_FORM);				
				FrmUpdateForm(TRANSACTIONS_FORM,0);
				handled = true;
				break;
			case TRANS_MENU_DETAILS:
				if(!item_selected) {
					FrmAlert(NO_SELECTED_TRANS);
					break;
				}
				TransactionsDB_record_is_new = false;
				FrmPopupForm(EDIT_TRANS_FORM);				
				FrmUpdateForm(TRANSACTIONS_FORM,0);
				//cancel the current selection
				item_selected=false;
				handled = true;
				break;		
			case TRANS_MENU_ADJST_BAL:
				TransactionsForm_AdjustAccBalance();
				FrmUpdateForm(TRANSACTIONS_FORM,0);
				//cancel the current selection
				item_selected=false;
				handled = true;
				break;		
						
			case TRANS_MENU_ABOUT:
			    f = FrmInitForm(ABOUT_FORM);
    
				FrmSetActiveForm(f);
    			FrmDoDialog(f);
        
			    if (frm) {
			        FrmSetActiveForm(frm);
				    FrmDeleteForm(f);
				}

				handled=true;
				break;
			case TRANS_MENU_HELP:
				FrmHelp(TRANS_HELP);

				handled=true;
				break;
				
			default:
				handled = false;
		}
		break;
	
	// Form Close
	case frmCloseEvent:
		CtlSetEnabled(GetObjectPtr(TRANS_PGDN),false);
		CtlSetEnabled(GetObjectPtr(TRANS_PGDN),false);
        CtlHideControl(GetObjectPtr(TRANS_PGUP));
        CtlHideControl(GetObjectPtr(TRANS_PGDN));
		handled = true;
		break;
	
	// Form Update
	case frmUpdateEvent:
		TransactionsTable_Init();
		//display the category name
		TransactionsForm_InitCat();
		//draw form
		FrmDrawForm(frm);
		//display the running totals
		TransactionsForm_DisplayTotals();
		
		handled = true;
		break;
	
	// Everything else...
	default:
		handled = false;
	}	
	return handled;
}
