/**	
	$Author: dizzy $
	$Revision: 1.9 $
	Last updated: $Date: 2000/11/19 19:19:44 $
	
	$Log: TransactionsTable.c,v $
	Revision 1.9  2000/11/19 19:19:44  dizzy
	The way the tables are loaded was re-written. The account table works perfectly now
	but the transaction table still has a few problems.
	
	Revision 1.8  2000/11/15 19:20:03  dizzy
	Finally fixed bug which prevented displaying all the transactions
	
	Revision 1.7  2000/11/15 18:58:42  dizzy
	New balance adjust option in transaction form menu.
	Help updated in Transaction form and Accounts form.
	Merged ShowScrollArrows functions.
	Updated a few variable and function names.
	
	Revision 1.6  2000/11/13 18:16:14  dizzy
	Fixed scrolling in transaction form. Added help option in menu
	
	Revision 1.5  2000/11/12 14:34:12  dizzy
	Initial transaction category support
	
	Revision 1.4  2000/11/06 22:27:02  dizzy
	Numbers are now localised and alligned to the right; the dates are displayed using
	the default format (these changes were submitted by Carsten Haese
	
	Revision 1.3  2000/10/12 17:17:24  dizzy
	Email address update; lots of bugfixes
	
	Revision 1.2  2000/09/15 10:10:49  dizzy
	First useful version
	
	Revision 1.1  2000/09/13 11:50:27  dizzy
	The transaction module is fully functional

	Copyright (C) 2000 Vlad Mereuta <dizzy@users.sourceforge.net>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
	
*/

#define __TRANSACTIONS_TABLE_C__

#include "TransactionsTable.h"

// the uniqueid for top visible items
UInt16		TransactionsTable_top_visible_item=0;
// the max no of rows that can be displayed
UInt16		TransactionsTable_rows;
// no of elements that could be loaded in the table
UInt16		TransactionsTable_total_items;

//name all the columns
UInt16		TransactionsTable_date_col=0;
UInt16		TransactionsTable_dest_col=1;
UInt16		TransactionsTable_ammount_col=2;

// the table
static TablePtr		table;
static UInt32		cur_acc; //uniqueid of the current record in accounts db

// bitmasks that determine relevant features of the date format preference
const UInt32 DF_MASK_DAY_FIRST =
	((1L<<dfDMYWithSlashes)|(1L<<dfDMYWithDots)|(1L<<dfDMYWithDashes));
const UInt32 DF_MASK_SLASH =
	((1L<<dfMDYWithSlashes)|(1L<<dfYMDWithSlashes)|(1L<<dfDMYWithSlashes));
const UInt32 DF_MASK_DASH = ((1L<<dfYMDWithDashes)|(1L<<dfDMYWithDashes));
const UInt32 DF_MASK_DOT = ((1L<<dfYMDWithDots)|(1L<<dfDMYWithDots));


void TransactionsTable_DrawCurrency(void* table, Int16 row, Int16 column, RectangleType* bounds)
{
	Char					*s;
	transactionStructure 	ts;

	TransactionsDB_GetUnpackedRecord(TblGetRowData(table,row),&ts);
	if(ts.dest_acc != cur_acc)  //debit show a minus, otherwise 
		s=Pence2String(-ts.ammount);
	else s=Pence2String(ts.ammount);

	DrawCharsRightAlign(s, bounds);

	//free memory allocated by Pence2String
	MemPtrFree(s);
}

void TransactionsTable_DrawDate(void* table, Int16 row, Int16 column, RectangleType* bounds)
{
	Char		s[6];
	UInt32 datePrefBit;
	DateType d;
	UInt8 m_index, d_index;
	
	StrCopy(s, "00/00");
	datePrefBit = 1L << PrefGetPreference(prefDateFormat);
	if (datePrefBit & DF_MASK_DAY_FIRST) {
		d_index = 0; m_index = 3;
	}
	else {
		d_index = 3; m_index = 0;
	}
	if (datePrefBit & DF_MASK_DASH) s[2] = '-';
	if (datePrefBit & DF_MASK_DOT) s[2] = '.';
	TransactionsDB_RecordGetDate(TblGetRowData(table,row), &d);
	s[d_index  ] = '0' + (d.day / 10);
	s[d_index+1] = '0' + (d.day % 10);
	s[m_index  ] = '0' + (d.month / 10);
	s[m_index+1] = '0' + (d.month % 10);

	WinDrawChars(s,5,bounds->topLeft.x,bounds->topLeft.y);
}


void TransactionsTable_DrawAcc(void* table, Int16 row, Int16 column, RectangleType* bounds)
{
	accountStructure		as;
	transactionStructure 	ts;

	TransactionsDB_GetUnpackedRecord(TblGetRowData(table,row),&ts);
	
	if(coins_pref.disp_payee) // display the payee
		WinDrawChars(ts.payee,StrLen(ts.payee),bounds->topLeft.x,bounds->topLeft.y);
	else { //display the peer account
		if(ts.dest_acc != cur_acc) //payment
			TransactionsDB_RecordGetDestAcc(TblGetRowData(table,row),&as);
		else 
			TransactionsDB_RecordGetSrcAcc(TblGetRowData(table,row),&as);

		WinDrawChars(as.acname,StrLen(as.acname),bounds->topLeft.x,bounds->topLeft.y);
	}
}

void TransactionsTable_SetRowsUnusable(UInt16 start)
{
	UInt16 i;
	for(i=start;i<TransactionsTable_rows;i++)
		TblSetRowUsable(table,i,false);		
}

Boolean TransactionsTable_RecordIsValid(UInt16 id)
{
	transactionStructure	ts;
	
	//check the category
	if(TransactionsDB_current_cat==CategoryFind(TransactionsDB_ref,"Cl/Uncl") && ( \
		db_GetRecordCategory(TransactionsDB_ref,id)!=CategoryFind(TransactionsDB_ref,"Cleared") && \
		db_GetRecordCategory(TransactionsDB_ref,id)!=CategoryFind(TransactionsDB_ref,"Uncleared"))) return false; 

	//check source/dest accounts
	TransactionsDB_GetUnpackedRecord(db_GetRecordUid(TransactionsDB_ref,id),&ts);

	if((ts.source_acc != cur_acc) && (ts.dest_acc != cur_acc)) 
		return false;
	
	return true;
}

void TransactionsTable_LoadData()
{
	UInt16					recordNo=0,row,target;
	Boolean					is_cl_uncl,records_left=true;
	
	TblUnhighlightSelection(table);
	//start from 0. we have to count all the elements that could go in the table in order to be
	//able to show the scroll arrows
	TransactionsTable_total_items=TransactionsTable_top_visible_item;
	
	//the cleared/uncleared category has to be treated in a special way
	is_cl_uncl=(TransactionsDB_current_cat==CategoryFind(TransactionsDB_ref,"Cl/Uncl"));

	//position on the 1st record in our category
	
	//first record in our category
	if(DmSeekRecordInCategory(TransactionsDB_ref,&recordNo,0,dmSeekForward,(is_cl_uncl) ? dmAllCategories:TransactionsDB_current_cat)!=errNone) {
		//if no records abort
		TransactionsTable_SetRowsUnusable(0);
		return;
	} 
	//if in the last seek we did not find a valid record we have to add one
	if (! TransactionsTable_RecordIsValid(recordNo)) 
		target=TransactionsTable_top_visible_item+1;
	else 
		target=TransactionsTable_top_visible_item;
	
	//now find the top record
	while(target>0) { //if we run out of records before reaching the rec we have to display at the top, abort
		if(DmSeekRecordInCategory(TransactionsDB_ref,&recordNo,1,dmSeekForward,(is_cl_uncl) ? dmAllCategories:TransactionsDB_current_cat)!=errNone) {
			TransactionsTable_SetRowsUnusable(0);
			return;
		} else 	if (TransactionsTable_RecordIsValid(recordNo)) 
					target--;
	}

	//set row 0 with the record we found above
	TblSetRowData(table,0,db_GetRecordUid(TransactionsDB_ref,recordNo));					
	TblSetRowUsable(table,0,true);	
	
			
	//the loop runs until there are no records left(for counting purposes) and all the rows are set
	for (row=1; records_left || (row<TransactionsTable_rows); row++) {
		//cycle through the records until we find an acceptable one
		do {
			if(DmSeekRecordInCategory(TransactionsDB_ref,&recordNo,1,dmSeekForward,(is_cl_uncl) ? dmAllCategories:TransactionsDB_current_cat)!=errNone) {
				//if we run out of items to display just set the rows to unusable		
				TransactionsTable_SetRowsUnusable(row);
				records_left=false;
				break;
			}
		} while ( ! TransactionsTable_RecordIsValid(recordNo));					
		if(!records_left) break;			

		TransactionsTable_total_items++;
		
		if(row<TransactionsTable_rows) { //only execute the statements below if the row is valid			
			TblSetRowUsable(table,row,true);
			TblSetRowData(table,row,db_GetRecordUid(TransactionsDB_ref,recordNo));										
		}
	}
	
	//show the scrolling arrows if necessary
	ShowScrollArrows(TRANS_PGUP,TRANS_PGDN,TransactionsTable_total_items,TransactionsTable_rows,TransactionsTable_top_visible_item);
}

void TransactionsTable_Unhilight()
{
	TblUnhighlightSelection(table);

}

void TransactionsTable_Init()
{
	UInt16		row;
	
	//init globals
	table=(TablePtr) GetObjectPtr(TRANS_TABLE);
	
	//set the uid for the account described by this table
	DmRecordInfo(AccountsDB_ref,AccountsDB_current_record,NULL,&cur_acc,NULL);
	TransactionsTable_Unhilight();

	ErrFatalDisplayIf(!table, "table is null!!!");

	//the numbers are drawn with our custom procedure
	TblSetCustomDrawProcedure(table, TransactionsTable_date_col, TransactionsTable_DrawDate);
	TblSetCustomDrawProcedure(table, TransactionsTable_dest_col, TransactionsTable_DrawAcc);
	TblSetCustomDrawProcedure(table, TransactionsTable_ammount_col, TransactionsTable_DrawCurrency);

	//get max number of visible rows
	TransactionsTable_rows = TblGetNumberOfRows(table);

	//set style for all cells
	for (row = 0; row < TransactionsTable_rows; row++) {
		TblSetItemStyle(table, row, TransactionsTable_date_col, customTableItem);
		TblSetItemStyle(table, row, TransactionsTable_dest_col, customTableItem);
		TblSetItemStyle(table, row, TransactionsTable_ammount_col, customTableItem);
	}
	
/** 	//set all columns as usable
	TblSetColumnUsable(table, TransactionsTable_name_col, true);
	if(coins_pref.disp_expected_col) {
		TblSetColumnUsable(table, TransactionsTable_expected_balance_col,true);
		TblSetColumnWidth(table,TransactionsTable_name_col,66);
	}
	else {
		//dont display the column
		TblSetColumnUsable(table, TransactionsTable_expected_balance_col,false);
		//make the anme column bigger
		TblSetColumnWidth(table,TransactionsTable_name_col,111);
	}
 */		
	TblSetColumnUsable(table, TransactionsTable_date_col, true);
 	TblSetColumnUsable(table,TransactionsTable_dest_col, true);
	TblSetColumnUsable(table, TransactionsTable_ammount_col, true);
	
	TransactionsTable_LoadData();
}
