/* $Id: pcount.c,v 1.9 2000/10/07 16:27:14 bodo Exp $ */
/*
 * Copyright (C) 1998 Bodo Bellut
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   Authors:  Bodo Bellut (bodo@bellut.net)
 */
/*
 * $Log: pcount.c,v $
 * Revision 1.9  2000/10/07 16:27:14  bodo
 * new email address
 *
 * Revision 1.8  2000/10/07 16:09:56  bodo
 * fix hardware-button select
 *
 * Revision 1.7  1998/09/24 15:51:56  bodo
 * do memory leak checking based on new switch DEBUG_MEMORA_LEAK
 *
 * Revision 1.6  1998/09/24 00:47:01  bodo
 * options now get applied when option dialog closeb with OK
 * now only changed controls in options dialog get redrawn
 * changed redraw strategy for mainform back to FldEraseField/FldDrawField
 * main form only gets redrawn if counter names changed
 * added reset all counters menuitem and associated code
 *
 * Revision 1.5  1998/09/23 22:32:59  bodo
 * fixed small bug re setting counter name #4
 * added handling of user editing counter fields
 *
 * Revision 1.4  1998/09/18 03:43:18  bodo
 * added GPL messages to files
 *
 * Revision 1.3  1998/09/17 23:06:01  bodo
 * setting checkboxes in options form now works
 *
 * Revision 1.2  1998/09/16 01:43:04  bodo
 * implemented changeable label texts
 * implemented changeable hardware buttons <-> counter
 * some clean up
 *
 * Revision 1.1  1998/09/10 22:25:56  bodo
 * Initial revision
 *
 */

/* Main code for PalmCounter */

#include <Pilot.h>
#include "callback.h"

#include "pcount.h"
#include "pcountRsc.h"

/* begin globals */

PreferenceType prefs = { "Counter1  ",
                         "Counter2  ",
                         "Counter3  ",
                         "Counter4  ",
                         1,
                         1,
                         1,
                         1,
                         0,
                         0,
                         0,
                         0,
                         1 };
PreferenceType newprefs;

/* FormPtr MainFormPtr = NULL; */
FormPtr OptionsFormPtr = NULL;
Long    aktList = 0;
Long    aktLabel = 0;

/* end globals */

VoidPtr GetObjectPtr(Int objectID)
{
    FormPtr pfrm;
    
    pfrm = FrmGetActiveForm();
    return FrmGetObjectPtr(pfrm, FrmGetObjectIndex(pfrm, objectID));
}

void increaseField(Int objectID, int delta)
{
    Handle	textH;
    CharPtr	textP;
    Long	tmpCounter;
    FieldPtr	fld = GetObjectPtr(objectID);

    textH = FldGetTextHandle(fld);
    if (textH) {
    	textP = MemHandleLock((VoidHand)(textH));
    	tmpCounter = StrAToI(textP);
    	tmpCounter += delta;
    	StrIToA(textP, tmpCounter);
    	FldSetTextHandle(fld, textH);
	MemHandleUnlock((VoidHand)(textH));

	FldEraseField(fld);
	FldDrawField(fld);

/*	FldRecalculateField(fld, true); */
/*
	if (MainFormPtr)
	   FrmDrawForm(MainFormPtr);
*/
	switch (objectID) {
	case fCounter1: prefs.counter1 = tmpCounter; break;
	case fCounter2: prefs.counter2 = tmpCounter; break;
	case fCounter3: prefs.counter3 = tmpCounter; break;
	case fCounter4: prefs.counter4 = tmpCounter; break;
	default: ErrNonFatalDisplayIf(true, "invalif objectID in increaseField()"); break;
	}
    }
}

void ResetOneCounter(FieldPtr fld)
{
    Handle	textH;
    CharPtr	textP;

    textH = FldGetTextHandle(fld);
    if (textH) {
    	textP = MemHandleLock((VoidHand)(textH));
    	StrIToA(textP, 0);
    	FldSetTextHandle(fld, textH);
	MemHandleUnlock((VoidHand)(textH));

	FldEraseField(fld);
	FldDrawField(fld);
    }
}

void ResetAllCounters(void)
{
    prefs.counter1 = 0;
    prefs.counter2 = 0;
    prefs.counter3 = 0;
    prefs.counter4 = 0;

    ResetOneCounter(GetObjectPtr(fCounter1));
    ResetOneCounter(GetObjectPtr(fCounter2));
    ResetOneCounter(GetObjectPtr(fCounter3));
    ResetOneCounter(GetObjectPtr(fCounter4));
}

static Boolean MainFormHandleEvent (EventPtr e)
{
    Boolean handled = false;
    FormPtr frm = FrmGetActiveForm();

    
    CALLBACK_PROLOGUE

    switch (e->eType) {
    case frmOpenEvent: {
	FieldPtr fld;
	Char    tmpText[10];

	fld = GetObjectPtr(fCounter1);
	StrIToA(tmpText, prefs.counter1);
	FldInsert(fld, tmpText, StrLen(tmpText));

	fld = GetObjectPtr(fCounter2);
	StrIToA(tmpText, prefs.counter2);
	FldInsert(fld, tmpText, StrLen(tmpText));

	fld = GetObjectPtr(fCounter3);
	StrIToA(tmpText, prefs.counter3);
	FldInsert(fld, tmpText, StrLen(tmpText));

	fld = GetObjectPtr(fCounter4);
	StrIToA(tmpText, prefs.counter4);
	FldInsert(fld, tmpText, StrLen(tmpText));
	
	CtlSetLabel(GetObjectPtr(lCounter1), prefs.label1);
	CtlSetLabel(GetObjectPtr(lCounter2), prefs.label2);
	CtlSetLabel(GetObjectPtr(lCounter3), prefs.label3);
	CtlSetLabel(GetObjectPtr(lCounter4), prefs.label4);

	FrmDrawForm(frm);
/*	FldSetInsertionPoint(fld, FldGetTextLength(fld)); */
	handled = true;
	break;
    }

    case menuEvent:
	MenuEraseStatus(NULL);

	switch(e->data.menu.itemID) {
	case mOptions:
		FrmPopupForm(OptionsForm);
		break;
	case mReset:
		if (FrmAlert(aReset) == 0)
		    ResetAllCounters();
		break;
	case mAbout:
		FrmAlert(aAbout);
		break;
	}

    	handled = true;
	break;

    case ctlSelectEvent:
	switch(e->data.ctlSelect.controlID) {
	case bCounter1u:
		increaseField(fCounter1, prefs.inc1);
		handled = true;
		break;
	case bCounter1d:
		increaseField(fCounter1, -prefs.inc1);
		handled = true;
		break;
	case bCounter2u:
		increaseField(fCounter2, prefs.inc2);
		handled = true;
		break;
	case bCounter2d:
		increaseField(fCounter2, -prefs.inc2);
		handled = true;
		break;
	case bCounter3u:
		increaseField(fCounter3, prefs.inc3);
		handled = true;
		break;
	case bCounter3d:
		increaseField(fCounter3, -prefs.inc3);
		handled = true;
		break;
	case bCounter4u:
		increaseField(fCounter4, prefs.inc4);
		handled = true;
		break;
	case bCounter4d:
		increaseField(fCounter4, -prefs.inc4);
		handled = true;
		break;
	}
	break;

    case keyDownEvent:
	switch (e->data.keyDown.chr) {
	case pageUpChr:
	case pageDownChr:
		switch (prefs.hardware) {
		case 1: increaseField(fCounter1, (e->data.keyDown.chr == pageUpChr) ? prefs.inc1 : -prefs.inc1); break;
		case 2: increaseField(fCounter2, (e->data.keyDown.chr == pageUpChr) ? prefs.inc2 : -prefs.inc2); break;
		case 3: increaseField(fCounter3, (e->data.keyDown.chr == pageUpChr) ? prefs.inc3 : -prefs.inc3); break;
		case 4: increaseField(fCounter4, (e->data.keyDown.chr == pageUpChr) ? prefs.inc4 : -prefs.inc4); break;
		default: ErrNonFatalDisplayIf(true, "invalid prefs.hardware setting"); break;
		}
		handled = true;
		break;
        }
        break;

    case fldChangedEvent: {
    	switch (e->data.fldChanged.fieldID) {
    	case fCounter1: 
    	case fCounter2: 
    	case fCounter3: 
    	case fCounter4: {
		Handle		textH;
		CharPtr		textP;
		FieldPtr	fld;
		
		fld = GetObjectPtr(e->data.fldChanged.fieldID);
		textH = FldGetTextHandle(fld);
		if (textH) {
		    textP = MemHandleLock((VoidHand)(textH));
		    switch (e->data.fldChanged.fieldID) {
			case fCounter1: prefs.counter1 = StrAToI(textP); break;
			case fCounter2: prefs.counter2 = StrAToI(textP); break;
			case fCounter3: prefs.counter3 = StrAToI(textP); break;
			case fCounter4: prefs.counter4 = StrAToI(textP); break;
			default: ErrNonFatalDisplayIf(true, "invalif objectID in increaseField()"); break;
		    }
		    MemHandleUnlock((VoidHand)(textH));
		}
		handled = true;
		break;
    	}
    	default       : ErrNonFatalDisplayIf(true, "invalid fldChangedEvent"); break;
    	}
    	break;
    }

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

static int findSelectionByValue(long value)
{
    switch(value) {
    case  -1: return 0; break;
    case   1: return 1; break;
    case   2: return 2; break;
    case   5: return 3; break;
    case  10: return 4; break;
    case 100: return 5; break;
    default : return 6; break;
    }
}

static Boolean OptionsFormHandleEvent (EventPtr e)
{
    Boolean handled = false;
    const FormPtr frm = FrmGetActiveForm();
    Char tmpChar1[10], tmpChar2[10], tmpChar3[10], tmpChar4[10];
    static Boolean updateMainForm;
    
    
    CALLBACK_PROLOGUE

    switch (e->eType) {
    case frmUpdateEvent:
	CtlEraseControl(GetObjectPtr(e->data.frmUpdate.updateCode));
	switch (e->data.frmUpdate.updateCode) {
	case bCounter1:
	    CtlSetLabel(GetObjectPtr(bCounter1), newprefs.label1);
	    updateMainForm = true;
	    break;
	case bCounter2:
	    CtlSetLabel(GetObjectPtr(bCounter2), newprefs.label2);
	    updateMainForm = true;
	    break;
	case bCounter3:
	    CtlSetLabel(GetObjectPtr(bCounter3), newprefs.label3);
	    updateMainForm = true;
	    break;
	case bCounter4:
	    CtlSetLabel(GetObjectPtr(bCounter4), newprefs.label4);
	    updateMainForm = true;
	    break;
	case lList1:
	    CtlSetLabel(GetObjectPtr(bList1), StrIToA(tmpChar1, newprefs.inc1));
	    break;
	case lList2:
	    CtlSetLabel(GetObjectPtr(bList2), StrIToA(tmpChar1, newprefs.inc2));
	    break;
	case lList3:
	    CtlSetLabel(GetObjectPtr(bList3), StrIToA(tmpChar1, newprefs.inc3));
	    break;
	case lList4:
	    CtlSetLabel(GetObjectPtr(bList4), StrIToA(tmpChar1, newprefs.inc4));
	    break;
	default:
	    ErrNonFatalDisplayIf(true, "frmUpdateEvent: unknown updateCode");
	    break;
	}
	CtlDrawControl(GetObjectPtr(e->data.frmUpdate.updateCode));
	handled = true;
	break;
    case frmOpenEvent:
	updateMainForm = false;
	MemMove(&newprefs, &prefs, sizeof(newprefs));
        LstSetSelection(GetObjectPtr(lList1), findSelectionByValue(newprefs.inc1));
        LstSetSelection(GetObjectPtr(lList2), findSelectionByValue(newprefs.inc2));
        LstSetSelection(GetObjectPtr(lList3), findSelectionByValue(newprefs.inc3));
        LstSetSelection(GetObjectPtr(lList4), findSelectionByValue(newprefs.inc4));

	CtlSetLabel(GetObjectPtr(bList1), StrIToA(tmpChar1, newprefs.inc1));
	CtlSetLabel(GetObjectPtr(bList2), StrIToA(tmpChar2, newprefs.inc2));
	CtlSetLabel(GetObjectPtr(bList3), StrIToA(tmpChar3, newprefs.inc3));
	CtlSetLabel(GetObjectPtr(bList4), StrIToA(tmpChar4, newprefs.inc4));

	CtlSetLabel(GetObjectPtr(bCounter1), newprefs.label1);
	CtlSetLabel(GetObjectPtr(bCounter2), newprefs.label2);
	CtlSetLabel(GetObjectPtr(bCounter3), newprefs.label3);
	CtlSetLabel(GetObjectPtr(bCounter4), newprefs.label4);

	CtlSetValue(GetObjectPtr(cCounter1), false);
	CtlSetValue(GetObjectPtr(cCounter2), false);
	CtlSetValue(GetObjectPtr(cCounter3), false);
	CtlSetValue(GetObjectPtr(cCounter4), false);

	switch (newprefs.hardware) {
	case 1: CtlSetValue(GetObjectPtr(cCounter1), true); break;
	case 2: CtlSetValue(GetObjectPtr(cCounter2), true); break;
	case 3: CtlSetValue(GetObjectPtr(cCounter3), true); break;
	case 4: CtlSetValue(GetObjectPtr(cCounter4), true); break;
	default: ErrNonFatalDisplayIf(true, "invalid newprefs.hardware?"); break;
	}

	FrmDrawForm(frm);
	handled = true;
	break;

    case menuEvent:
	MenuEraseStatus(NULL);

	switch(e->data.menu.itemID) {
	}

    	handled = true;
	break;

    case ctlSelectEvent:
	switch(e->data.ctlSelect.controlID) {
	case bOK:
		if (CtlGetValue(GetObjectPtr(cCounter1)))
		    newprefs.hardware = 1;
		else if (CtlGetValue(GetObjectPtr(cCounter2)))
		    newprefs.hardware = 2;
		else if (CtlGetValue(GetObjectPtr(cCounter3)))
		    newprefs.hardware = 3;
		else if (CtlGetValue(GetObjectPtr(cCounter4)))
		    newprefs.hardware = 4;
		else ErrNonFatalDisplayIf(true, "options: no hardware set?");
		MemMove(&prefs, &newprefs, sizeof(prefs));
		if (updateMainForm) {
		    FrmCloseAllForms();
		    FrmGotoForm(MainForm);
		} else
		    FrmReturnToForm(MainForm);
		handled = true;
		break;
	case bCancel:
		FrmReturnToForm(MainForm);
		handled = true;
		break;
	case bCounter1:
	case bCounter2:
	case bCounter3:
	case bCounter4:
		aktLabel = e->data.ctlSelect.controlID;
		FrmPopupForm(enterNameForm);
		handled = true;
		break;
	}
	break;
    case popSelectEvent:
    	if (e->data.popSelect.selection == 6) {
	   
    	   aktList = e->data.popSelect.listID;

	   FrmPopupForm(enterValueForm);
	   handled = true;

	} else {
	   CharPtr tmpList;
	   
	   switch(e->data.popSelect.listID) {
	   case lList1:
	   	tmpList = LstGetSelectionText(GetObjectPtr(lList1),
	   	                              LstGetSelection(GetObjectPtr(lList1)));
	   	newprefs.inc1 = StrAToI(tmpList);
	   	break;
	   case lList2:
	   	tmpList = LstGetSelectionText(GetObjectPtr(lList2),
	   	                              LstGetSelection(GetObjectPtr(lList2)));
		newprefs.inc2 = StrAToI(tmpList);
		break;
	   case lList3:
	   	tmpList = LstGetSelectionText(GetObjectPtr(lList3),
	   	                              LstGetSelection(GetObjectPtr(lList3)));
		newprefs.inc3 = StrAToI(tmpList);
		break;
	   case lList4:
	   	tmpList = LstGetSelectionText(GetObjectPtr(lList4),
	   	                              LstGetSelection(GetObjectPtr(lList4)));
		newprefs.inc4 = StrAToI(tmpList);
		break;
	   default:     ErrNonFatalDisplayIf(true, "bad listID received by OptionsFormHandleEvent"); break;
	   }	   
	}
    	break;

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

static Boolean enterValueFormHandleEvent (EventPtr e)
{
    Boolean handled = false;
    const FormPtr frm = FrmGetActiveForm();

    
    CALLBACK_PROLOGUE

    switch (e->eType) {
    case frmOpenEvent: {
	FieldPtr fld = GetObjectPtr(fenterValue);
	Char    tmpText[10];

	switch(aktList) {
	case lList1: StrIToA(tmpText, newprefs.inc1); break;
	case lList2: StrIToA(tmpText, newprefs.inc2); break;
	case lList3: StrIToA(tmpText, newprefs.inc3); break;
	case lList4: StrIToA(tmpText, newprefs.inc4); break;
	default:     ErrNonFatalDisplayIf(true, "bad listID received by enterValueFormHandleEvent"); break;
	}
	FldInsert(fld, tmpText, StrLen(tmpText));
	FrmDrawForm(frm);
	FrmSetFocus(frm, FrmGetObjectIndex(frm, fenterValue));
	FldSetInsertionPoint(fld, FldGetTextLength(fld));
	handled = true;
	break;
    }

    case menuEvent:
	MenuEraseStatus(NULL);

	switch(e->data.menu.itemID) {
	}

    	handled = true;
	break;

    case ctlSelectEvent:
	switch(e->data.ctlSelect.controlID) {
	case bOK: {
		FieldPtr fld = GetObjectPtr(fenterValue);
		Handle	textH;
		CharPtr	textP;
		Long	tmpCounter;

		textH = FldGetTextHandle(fld);
		if (textH) {
			textP = MemHandleLock((VoidHand)(textH));
			tmpCounter = StrAToI(textP);
			MemHandleUnlock((VoidHand)(textH));
			switch(aktList) {
			case lList1: newprefs.inc1 = tmpCounter; break;
			case lList2: newprefs.inc2 = tmpCounter; break;
			case lList3: newprefs.inc3 = tmpCounter; break;
			case lList4: newprefs.inc4 = tmpCounter; break;
			default:     ErrNonFatalDisplayIf(true, "bad listID received by enterValueFormHandleEvent"); break;
			}
		}
		FrmUpdateForm(OptionsForm, aktList);
		FrmReturnToForm(OptionsForm);
		handled = true;
		break;
	}
	case bCancel:
		FrmReturnToForm(OptionsForm);
		handled = true;
		break;
	}
	break;

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

static Boolean enterNameFormHandleEvent (EventPtr e)
{
    Boolean handled = false;
    const FormPtr frm = FrmGetActiveForm();

    
    CALLBACK_PROLOGUE

    switch (e->eType) {
    case frmOpenEvent: {
	FieldPtr fld = GetObjectPtr(fenterName);

	switch (aktLabel) {
	case bCounter1: FldInsert(fld, newprefs.label1, StrLen(newprefs.label1)); break;
	case bCounter2: FldInsert(fld, newprefs.label2, StrLen(newprefs.label2)); break;
	case bCounter3: FldInsert(fld, newprefs.label3, StrLen(newprefs.label3)); break;
	case bCounter4: FldInsert(fld, newprefs.label4, StrLen(newprefs.label4)); break;
	default       :
	    ErrNonFatalDisplayIf(true, "aktLabel unknown in enterNameFormHandleEvent");
	    break;
	}
	FrmDrawForm(frm);
	FrmSetFocus(frm, FrmGetObjectIndex(frm, fenterName));
	FldSetInsertionPoint(fld, FldGetTextLength(fld));
	handled = true;
	break;
    }

    case menuEvent:
	MenuEraseStatus(NULL);

	switch(e->data.menu.itemID) {
	}

    	handled = true;
	break;

    case ctlSelectEvent:
	switch(e->data.ctlSelect.controlID) {
	case bOK: {
		FieldPtr fld = GetObjectPtr(fenterName);
		Handle	textH;
		CharPtr	textP;

		textH = FldGetTextHandle(fld);
		if (textH) {
			textP = MemHandleLock((VoidHand)(textH));
			switch (aktLabel) {
			case bCounter1: StrCopy(newprefs.label1, textP); break;
			case bCounter2: StrCopy(newprefs.label2, textP); break;
			case bCounter3: StrCopy(newprefs.label3, textP); break;
			case bCounter4: StrCopy(newprefs.label4, textP); break;
			default       : ErrNonFatalDisplayIf(true, "invalif aktLabel passed to enterNameFormHandleEvent"); break;
			}
			MemHandleUnlock((VoidHand)(textH));
		}
		FrmUpdateForm(OptionsForm, aktLabel);
		FrmReturnToForm(OptionsForm);
		handled = true;
		break;
	}
	case bCancel:
		FrmReturnToForm(OptionsForm);
		handled = true;
		break;
	break;
	}

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}

static Boolean ApplicationHandleEvent(EventPtr e)
{
    FormPtr frm;
    Word    formId;
    Boolean handled = false;

    if (e->eType == frmLoadEvent) {
	formId = e->data.frmLoad.formID;
	frm = FrmInitForm(formId);
	FrmSetActiveForm(frm);

	switch(formId) {
	case MainForm:
/*	    MainFormPtr = frm; */
	    FrmSetEventHandler(frm, MainFormHandleEvent);
	    break;
	case OptionsForm:
	    FrmSetEventHandler(frm, OptionsFormHandleEvent);
	    break;
	case enterValueForm:
	    FrmSetEventHandler(frm, enterValueFormHandleEvent);
	    break;
	case enterNameForm:
	    FrmSetEventHandler(frm, enterNameFormHandleEvent);
	    break;
	}
	handled = true;
    }

    return handled;
}

static Word StartApplication(void)
{
    
    if (PrefGetAppPreferencesV10(APPID, 0, &prefs, sizeof(prefs))) {
    } else {
    }

    FrmGotoForm(MainForm);
    return 0;
}

static void StopApplication(void)
{

    PrefSetAppPreferencesV10(APPID, 0, &prefs, sizeof(prefs));

    FrmSaveAllForms();
    FrmCloseAllForms();
}

static void EventLoop(void)
{
    Word err;
    EventType e;

    do {
	EvtGetEvent(&e, evtWaitForever);
	if (! SysHandleEvent (&e))
	    if (! MenuHandleEvent (NULL, &e, &err))
		if (! ApplicationHandleEvent (&e))
		    FrmDispatchEvent (&e);
    } while (e.eType != appStopEvent);
}

DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
    Word err;

#if ERROR_CHECK_LEVEL == ERROR_CHECK_FULL
#warning compiling with ERROR_CHECK_FULL
#ifdef DEBUG_MEMORY_LEAK
    DWord startFreeDynamicMemory, endFreeDynamicMemory, junk;
    Word crc;
    
    crc = Crc16CalcBlock(NULL, 256, 0);
    MemHeapFreeBytes(0, &startFreeDynamicMemory, &junk);
#endif DEBUG_MEMORY_LEAK
#endif

    if (cmd == sysAppLaunchCmdNormalLaunch) {

	err = StartApplication();
	if (err) return err;

	EventLoop();
	StopApplication();

    } else {
	return sysErrParamErr;
    }

#if (ERROR_CHECK_LEVEL == ERROR_CHECK_FULL) && defined(DEBUG_MEMORY_LEAK)
    ErrNonFatalDisplayIf(crc != Crc16CalcBlock(NULL, 256, 0),
                         "Low mem corrupted");
    MemHeapFreeBytes(0, &endFreeDynamicMemory, &junk);
    ErrNonFatalDisplayIf(startFreeDynamicMemory != endFreeDynamicMemory,
                         "Memory leak!");
#endif

    return 0;
}
