/*
 *  Version 1.3 Copyright (C) 2000 Harakan Software
 *
 *  Copyright (C) 1998 International Computer Science Institute (ICSI)
 *
 *  Author: Vladimir Minenko, minenko@icsi.berkeley.edu
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 *
 */


#include "callback.h"

#include "palmVNC.h"
#include "palmVNCRsc.h"

#include <CharAttr.h>
#include <MemoryMgr.h>
#include <System/KeyMgr.h>
#include <System/SystemMgr.h>

#define MIN_VERSION		0x03000000 // PalmOS 3.0

#define INVALID_PIXEL		0xffffffff

#define COLORMAP_SIZE		256

// switch defines for scrollbars
// we have our own scrollbars because the native scrollbars are too large
#define HORIZ			1
#define HORIZ_UP		2
#define HORIZ_DOWN		3
#define HORIZ_TOP		4
#define HORIZ_BUTTOM	5

#define VERT			6
#define VERT_UP			7
#define VERT_DOWN		8
#define VERT_TOP		9
#define VERT_BUTTOM		10

#define MAIN_PRESSED	11
#define MAIN_MOVED		12
#define MAIN_ONCE		13
#define MAIN_DOUBLE		14

// modifiers and special Windows keys
#define TAB_KEY			0xff09
#define ENTER_KEY		0xff0d
#define ESC_KEY			0xff1b
#define SHIFT_KEY		0xffe1
#define CTRL_KEY		0xffe3
#define ALT_KEY			0xffe9
#define F4_KEY			0xffc1
#define DEL_KEY			0xffff

#define F1_KEY          0xffbe

static unsigned long lastPenTime = 0;
static unsigned short lastX=0, lastY=0;
static unsigned short currentMouseButton = 0;

// off-screen panning window
static WinHandle canvas = 0;

// Off-screen keyboard bitmap window
static WinHandle keyboard = 0;

// Define keyboard state bits
static unsigned short keyboard_state = 0;

#define keyboard_visible 7
#define keyboard_esc 0
#define keyboard_shift 1
#define keyboard_ctrl 2
#define keyboard_alt 3
#define keyboard_fn 4

#define keyboard_keywidth 20
#define keyboard_num_controls 5

// on-screen main window
static WinHandle topLevel = 0;

// data boundaries of the off-screen panning window
static char *screenData = NULL;

// Colour depth option
static unsigned short screenDepth = 0;


static unsigned short osVersion = 0;

// BGR233 codes for minimal greyscale colours
#define BLACK		0x00
#define DARK_GRAY	0x52
#define LIGHT_GRAY	0xA4
#define WHITE		0xFF
#define DARK_RED    0x04
#define DARK_BLUE   0x80

unsigned short BGR233ToIntensityNibble[COLORMAP_SIZE] = {
 0, 1, 3, 4, 5, 7, 8, 9,  1, 2, 3, 5, 6, 7, 9,10,  1, 3, 4, 5, 7, 8, 9,11,  2, 3, 5, 6, 7, 9,10,12, 
 3, 4, 5, 7, 8,10,11,12,  3, 5, 6, 7, 9,10,12,13,  4, 5, 7, 8,10,11,12,14,  5, 6, 8, 9,10,12,13,14,
  
 1, 2, 3, 5, 6, 7, 9,10,  1, 3, 4, 5, 7, 8, 9,11,  2, 3, 5, 6, 7, 9,10,11,  3, 4, 5, 7, 8, 9,11,12, 
 3, 5, 6, 7, 9,10,11,13,  4, 5, 7, 8, 9,11,12,13,  5, 6, 7, 9,10,11,13,14,  5, 7, 8, 9,11,12,13,15, 
 
 1, 2, 4, 5, 6, 8, 9,10,  2, 3, 4, 6, 7, 8,10,11,  2, 4, 5, 6, 8, 9,11,12,  3, 4, 6, 7, 9,10,11,13, 
 4, 5, 6, 8, 9,11,12,13,  4, 6, 7, 9,10,11,13,14,  5, 7, 8, 9,11,12,13,15,  6, 7, 9,10,11,13,14,15, 
 
 2, 3, 4, 6, 7, 8,10,11,  2, 4, 5, 6, 8, 9,10,12,  3, 4, 6, 7, 8,10,11,12,  4, 5, 6, 8, 9,10,12,13, 
 4, 6, 7, 8,10,11,12,14,  5, 6, 8, 9,10,12,13,14,  6, 7, 8,10,11,12,14,15,  6, 8, 9,10,12,13,14,15
};

typedef UInt8 BGR233ColorType;

BGR233ColorType IntensityNibbleToBGR233[16] = {
	0x00,
	0x00,
	0x09,
	0x09,
	0x52,
	0x52,
	0x5B,
	0x5B,
	0xA4,
	0xA4,
	0xB6,
	0xB6,	
	0xFF
};

PalmVNCPreferences pref;
static DmOpenRef prefDB;

// pen event processing
static int downX, downY;
static unsigned short buttonState = 0;
static int viewportX = 0, viewportY = 0;

static pen_zooming = false;

// for error handling
char AppErrStr[AppErrStrLen];

RectangleType vertScrollbarRect = {{VIEWABLE_MAX_W - SCROLLBAR_SIZE, 0},
				{SCROLLBAR_SIZE, VIEWABLE_MAX_H}};
RectangleType horizScrollbarRect = {{0, VIEWABLE_MAX_H - SCROLLBAR_SIZE},
				{VIEWABLE_MAX_H - SCROLLBAR_SIZE, SCROLLBAR_SIZE}};
RectangleType vertSlideRect = {{VIEWABLE_MAX_W - SCROLLBAR_SIZE, 0},
				{SCROLLBAR_SIZE, SCROLLBAR_SIZE}};
RectangleType horizSlideRect = {{0, VIEWABLE_MAX_H - SCROLLBAR_SIZE},
				{SCROLLBAR_SIZE, SCROLLBAR_SIZE}};
RectangleType updateRect = {{0, 0}, {0, 0}};

RectangleType viewableBounds = {{0, 0}, {VIEWABLE_MAX_W, VIEWABLE_MAX_H}};
RectangleType virtualBounds = {{0, 0}, {VIRTUAL_20_MAX_W, VIRTUAL_20_MAX_H}};

Err errno;
Long AppNetTimeout;
Word AppNetRefnum;
static Boolean networkIsOpen = false;

static void InitVirtual (void);
static Err CreateWindows (FormPtr frmPtr);
static Boolean OpenConnection (void);
static Boolean OpenNetwork (void);
static void ShutdownNetwork (Boolean enforcedClose);
static void ShutdownConnection (void);
static void SendKeySequence (unsigned short mod1, unsigned short mod2, unsigned short mod3,
		      unsigned short *keys, int n, Boolean ascii);
static void PositionViewport (void);
static void UpdateScrollbars (void);
static void PositionScrollbars (void);
static void UpdateViewport (void);
static void ProcessPenUpDownEvent (EventPtr e);
static void ProcessPenMoveEvent (EventPtr e);
static void ProcessScrollPenDownEvent (EventPtr e);
static void ProcessScrollPenMoveEvent (EventPtr e);
static Boolean ConsoleFormHandleEvent (EventPtr e);
static DmOpenRef OpenDatabase (PalmVNCPreferences *pref);
static void SaveStatus (DmOpenRef db, PalmVNCPreferences *pref);


/******************************************************************
 * mySetForeColor
 *
 * Sets the foregroung ink to the specified BGR233 color,
 * regardless of mode.
 *****************************************************************/
void mySetForeColor (BGR233ColorType color)
{
	RGBColorType newColor;
	
	if (screenDepth == 8)
	{
		WinSetForeColor(color);
		WinSetTextColor(color);
	}
	else
	{
		newColor.index = 0;
		newColor.r = (BGR233ToIntensityNibble[color] << 4) | BGR233ToIntensityNibble[color];
		newColor.g = newColor.r;
		newColor.b = newColor.r;
		WinSetColors(&newColor, 0, 0, 0);
	}
}


/******************************************************************
 * DrawingBlocked
 *
 * returns true if the menu bar is open or the main form isn't
 * active
 *****************************************************************/
Boolean DrawingBlocked (void)
{
  	if (MenuGetActiveMenu())
    	return (MenuGetActiveMenu()->attr.visible);
 
	/*** Crashes ***/
/*  if (MenuGetActiveMenu () != (MenuBarPtr)NULL)
    return true;*/

  	/*** Erratic ***/
    return MainForm != FrmGetActiveFormID();

/*	return (topLevel != WinGetActiveWindow()); */

}


/***********************************************************************
 *
 * RomVersionCompatible
 *
 * This routine checks that a ROM version meets your
 * minimum requirement and sets the version number in global variable
 ***********************************************************************/
static DWord RomVersionCompatible (DWord requiredVersion, Word launchFlags)
{
  DWord romVersion;

  // See if we have at least the minimum required version of the ROM or later.
  FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);

  if (romVersion < requiredVersion) {
    if ((launchFlags & (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp))
		== (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) {
      FrmAlert (VersionAlert);
		
      // Pilot 1.0 will continuously relaunch this app unless we switch to 
      // another safe one.
      if (romVersion < 0x02000000) {
	Err err;
        AppLaunchWithCommand (sysFileCDefaultApp,
		sysAppLaunchCmdNormalLaunch, NULL);
      }
    }
		
    return sysErrRomIncompatible;
  }

  osVersion = sysGetROMVerMajor(romVersion);

  return 0;
}


/******************************************************************
 * PostProgressMessage
 *
 * writes a message and and draws a simple progress bar
 *****************************************************************/
void PostProgressMessage (char *msgStr, unsigned short curr, unsigned short max)
{
  RGBColorType oldColor_RGB;
  RGBColorType newColor_RGB;
  IndexedColorType oldColor_Index;
  RectangleType r = {{20, 140}, {120, 10}};
  unsigned short length = StrLen (msgStr);

 if (DrawingBlocked ())
   return;

  mySetForeColor (DARK_GRAY);
  	
  WinEraseRectangle (&r, 0);

  WinDrawChars (msgStr, length,
	80 - FntLineWidth (msgStr, length) / 2, 139);

  WinDrawRectangleFrame (simpleFrame, &r);
  
  if (max) {
    r.extent.x = (120 * curr) / max;
    WinInvertRectangle (&r, 0);
  }
  /*WinSetForeColor (&grayScalePixelToColor[BLACK], &oldColor); Should be oldColor */
  mySetForeColor (BLACK);
}


/******************************************************************
 * GetUpdateRectAtPos
 *
 * Re-aasigns the coordinates of the update rect. and returns the
 * pointer to it.
 *****************************************************************/ 
RectanglePtr GetUpdateRectAtPos (unsigned short x, unsigned short y)
{
  updateRect.topLeft.x = x;
  updateRect.topLeft.y = y;
  return &updateRect;
}


/******************************************************************
 * InitVirtual
 *
 * Inits the content of the off-screen window
 *****************************************************************/ 
static void InitVirtual (void)
{
  IndexedColorType oldColor;
  WinHandle prevDrawWin = WinGetDrawWindow ();
  unsigned short hpos = 0;
  int i, j, k;

  WinSetDrawWindow (canvas);

  WinEraseWindow ();
  FntSetFont (largeBoldFont);
  hpos = 50;

  	mySetForeColor (DARK_GRAY);
  WinDrawChars("Palm", 4, hpos, 20);
  hpos += FntLineWidth("Palm", 4);
  
  if (screenDepth == 8)
  {
  	mySetForeColor (DARK_RED);
  }
  else
  {
  	mySetForeColor(BLACK);
  }
  WinDrawChars("VNC", 3, hpos, 20);
  
  mySetForeColor(BLACK);
  WinDrawLine(48, 18, 98, 18);
  WinDrawLine(48, 35, 98, 35);

  if (screenDepth == 8)
  {
  	mySetForeColor (DARK_BLUE);
  }
  else
  {
  	mySetForeColor(BLACK);
  }
  FntSetFont (boldFont);
  WinDrawChars("version 1.3f", 12, 45, 45/*40*/);
  
/*  if (screenDepth == 8)
  {
  	mySetForeColor (DARK_RED);
  }
  else
  {
  	mySetForeColor(BLACK);
  } */
/*  WinDrawChars("BETA - DO NOT DISTRIBUTE!", 25, 10, 55); */
/*  WinDrawChars("ALPHA - DO NOT DISTRIBUTE!", 26, 3, 55); */
/*  WinDrawChars("PRE-RELEASE VERSION", 19, 21, 50); */
/*  WinDrawChars("DO NOT DISTRIBUTE!", 18, 26, 60); */
 	mySetForeColor(BLACK);
  
  FntSetFont (stdFont);
  WinDrawChars("For more help on VNC see:", 25, 0, 105);
  if (screenDepth == 8)
  {
  	mySetForeColor (DARK_BLUE);
  }
  else
  {
  	mySetForeColor(BLACK);
  }
  WinDrawChars("http://www.uk.research.att.com/vnc/", 35, 0, 121);
  
  if (screenDepth == 8)
  {
	mySetForeColor (BLACK);
	WinDrawChars("using 8-bit colour", 18, 42, 72);
	for (i = 0; i < 8; i++)
		for (j = 0; j < 100; j++)
		{
			if (screenData[((81 - i) * VIEWABLE_MAX_W) + 27 + j] == 0) screenData[((81 - i) * VIEWABLE_MAX_W) + 27 + j] = (i << 3) + (7 - i);
		}	
  }
  else if (screenDepth == 4)
  {
	  for (i = 0; i < 7; i++)
	  {
		 mySetForeColor (IntensityNibbleToBGR233[(i*2)]);
		 WinDrawLine(27, 78+i, 127, 78+i);
		 WinDrawLine(27, 78-i, 127, 78-i);
	  }
	  mySetForeColor (BLACK);
	  WinEraseChars("using 4-bit greyscale", 21, 36, 72);
  }
  else if (screenDepth == 2)
  {
	  for (i = 0; i < 7; i++)
	  {
		 mySetForeColor (IntensityNibbleToBGR233[i*2]);
		 WinDrawLine(27, 78+i, 127, 78+i);
		 WinDrawLine(27, 78-i, 127, 78-i);
	  }
	  mySetForeColor (BLACK);
  	  WinEraseChars("using 2-bit greyscale", 21, 36, 72);
  }
  
  mySetForeColor (BLACK);
  PostProgressMessage ("Not Connected", 0, 0);

  FntSetFont (stdFont);
  WinSetDrawWindow (prevDrawWin);

  WinCopyRectangle (canvas, topLevel,
	GetUpdateRectAtPos (palmFB.viewable.x - palmFB.virtual.x,
			    palmFB.viewable.y - palmFB.virtual.y),
	0, 0, scrCopy);

}


/******************************************************************
 * DrawVirtualFilledRectangle
 *
 * Draws a filled rectangle in the virtual window If adjust is true
 * x and y coordinates are translatds to the origin in the virtual
 *****************************************************************/
void DrawVirtualFilledRectangle (unsigned short color,
				short x, short y,
				unsigned short width, unsigned short height,
				int adjust)
{
	short row = 0;
	short column = 0;
  	short start_column = 0;
	short end_column = 0;

  	// align to real coordinates
  	if (adjust) {
	  	x -= palmFB.virtual.x;
	  	y -= palmFB.virtual.y;
  	}

	// Check top-left corner
	if ((x >= palmFB.virtual.w) || (y >= palmFB.virtual.h))
	{
		return;
	}

	// Check top and bottom boundaries
	if (y < 0)
	{
		height += y;
		y = 0;
		if (height < 1)
		{
			return;
		}
	}

	if ((y + height) >= palmFB.virtual.h)
	{
		height = (palmFB.virtual.h - y);
	}

	// Check left and right boundaries
	if (x < 0)
	{
		start_column = -x;
	}
	else
		start_column = 0;

	if ((x + width) >= palmFB.virtual.w)
	{
		end_column = (palmFB.virtual.w - x)-1;
	}
	else
		end_column = width-1;

	if (end_column < start_column)
	{
		return;			
	}

	if (screenDepth == 8)
	{
		/* Do whole region */
		for (row = 0; row < height; row++)
		{
			for (column = start_column; column <= end_column; column ++)
			{
				screenData[(((row + y) * VIEWABLE_MAX_W) + column + x)] = color;			
			}
		}
	}
	else if (screenDepth == 4)
	{
		color = BGR233ToIntensityNibble[color];
	  	/* Do left-hand column if non-aligned */
	  	if (((x+start_column) % 2)) /* 1, 3, 5, ...  */
		{
			for (row = 0; row < height; row++)
			{
				screenData[(((row + y) * VIEWABLE_MAX_W) + start_column + x) / 2] |= (0xF);
				screenData[(((row + y) * VIEWABLE_MAX_W) + start_column + x) / 2] ^= (color);
			}
			start_column++;
		}
	
		/* Check if only 1 column */
		if (end_column < start_column)
			return;
	
	  	/* Do right-hand column if non-aligned */
	  	if (!((x+end_column) % 2)) /* 0, 2, 4, ...  */
		{
			for (row = 0; row < height; row++)
			{
				screenData[(((row + y) * VIEWABLE_MAX_W) + end_column + x) / 2] |= (0xF << 4);
				screenData[(((row + y) * VIEWABLE_MAX_W) + end_column + x) / 2] ^= (color << 4);
			}
			end_column--;
		}
	
		/* Check if only 2 columns */
		if (end_column <= start_column)
			return;
	
		/* Do middle region */
		color = ~(color + (color << 4));
		for (row = 0; row < height; row++)
		{
			for (column = start_column; column < end_column; column += 2)
			{
				screenData[(((row + y) * VIEWABLE_MAX_W) + column + x) / 2] = color;			
			}
		}
	}
	else if (screenDepth == 2)
	{
		/* Very slow implementation for now to sanity-check */
		
		color = (BGR233ToIntensityNibble[color] >> 2) & 0x3;
		
		for (row = 0; row < height; row++)
		{
			for (column = start_column; column <= end_column; column++)
			{
				switch ((x + column) % 4)
				{
					case 0:
						{
							screenData[(((row + y) * VIEWABLE_MAX_W) + column + x) / 4] |= (0x3 << 6);
							screenData[(((row + y) * VIEWABLE_MAX_W) + column + x) / 4] ^= (color << 6);
							break;
						}
					case 1:
						{
							screenData[(((row + y) * VIEWABLE_MAX_W) + column + x) / 4] |= (0x3 << 4);
							screenData[(((row + y) * VIEWABLE_MAX_W) + column + x) / 4] ^= (color << 4);
							break;
						}
					case 2:
						{
							screenData[(((row + y) * VIEWABLE_MAX_W) + column + x) / 4] |= (0x3 << 2);
							screenData[(((row + y) * VIEWABLE_MAX_W) + column + x) / 4] ^= (color << 2);
							break;
						}
					case 3:
						{
							screenData[(((row + y) * VIEWABLE_MAX_W) + column + x) / 4] |= (0x3 << 0);
							screenData[(((row + y) * VIEWABLE_MAX_W) + column + x) / 4] ^= (color << 0);
							break;
						}
				}					
			}
		}
	}
}


/******************************************************************
 * CopyDataToScreen
 *
 * Copies a 8bit pixel array into a correspond 4bit data area of the
 * off-screen window. Probably can be made more efficient, but this
 * part must be replaced by "official" bitmap operations via
 * WinDrawBitmap(). This re-implementation is also required for
 * a support of grayscale representation
 *****************************************************************/ 
void CopyDataToScreen (CARD8 *buf, short x, short y, unsigned short width, unsigned short height)
{
	short row = 0;
	short column = 0;
  	short start_column = 0;
	short end_column = 0;

	short x_off = 0;
	unsigned short color_temp = 0;
	unsigned short color_mask = 0;

  	// align to real coordinates
  	x -= palmFB.virtual.x;
  	y -= palmFB.virtual.y;

	// Check top-left corner
	if ((x >= palmFB.virtual.w) || (y >= palmFB.virtual.h))
	{
		return;
	}

	// Check top and bottom boundaries
	if (y < 0)
	{
		height += y;
		y = 0;
		if (height < 1)
		{
			return;
		}
	}

	if ((y + height) >= palmFB.virtual.h)
	{
		height = (palmFB.virtual.h - y);
	}

	// Check left and right boundaries
	if (x < 0)
	{
		start_column = -x;
	}
	else
		start_column = 0;

	if ((x + width) >= palmFB.virtual.w)
	{
		end_column = (palmFB.virtual.w - x)-1;
	}
	else
		end_column = width-1;

	if (end_column < start_column)
	{
		return;			
	}

	if (screenDepth == 8)
	{
		/* Do middle region */
		for (row = 0; row < height; row++)
		{
			for (column = start_column; column <= end_column; column ++)
			{
				screenData[(((row + y) * VIEWABLE_MAX_W) + column + x)] = buf[(row * width) + column];			
			}
		}
	}
	else if (screenDepth == 4)
	{
	  	/* Do left-hand column if non-aligned */
	  	if (((x+column) % 2)) /* 1, 3, 5, ...  */
		{
			for (row = 0; row < height; row++)
			{
				screenData[(((row + y) * VIEWABLE_MAX_W) + start_column + x) / 2] |= (0xF);
				screenData[(((row + y) * VIEWABLE_MAX_W) + start_column + x) / 2] ^= (BGR233ToIntensityNibble[buf[(row * width) + start_column]]);
			}
			start_column++;
		}
	
		/* Check if only 1 column */
		if (end_column < start_column)
			return;
	
	  	/* Do right-hand column if non-aligned */
	  	if (!((x+end_column) % 2)) /* 0, 2, 4, ...  */
		{
			for (row = 0; row < height; row++)
			{
				screenData[(((row + y) * VIEWABLE_MAX_W) + end_column + x) / 2] |= (0xF << 4);
				screenData[(((row + y) * VIEWABLE_MAX_W) + end_column + x) / 2] ^= (BGR233ToIntensityNibble[buf[(row * width) + end_column]] << 4);
			}
			end_column--;
		}
	
		/* Check if only 2 columns */
		if (end_column <= start_column)
			return;
	
		/* Do middle region */
		for (row = 0; row < height; row++)
		{
			for (column = start_column; column < end_column; column += 2)
			{
				screenData[(((row + y) * VIEWABLE_MAX_W) + column + x) / 2] = ~(BGR233ToIntensityNibble[buf[(row * width) + column + 1]] +
																			   (BGR233ToIntensityNibble[buf[(row * width) + column]] << 4));			
			}
		}
	}
	else if (screenDepth == 2)
	{
		/* Very slow implementation for now to sanity-check */
		
		for (row = 0; row < height; row++)
		{
			for (column = start_column; column <= end_column; column++)
			{
				switch ((x + column) % 4)
				{
					case 0:
						{
							screenData[(((row + y) * VIEWABLE_MAX_W) + column + x) / 4] |= (0x3 << 6);
							screenData[(((row + y) * VIEWABLE_MAX_W) + column + x) / 4] ^= ((BGR233ToIntensityNibble[buf[(row * width) + column]] << 4) & (0x3 << 6));
							break;
						}
					case 1:
						{
							screenData[(((row + y) * VIEWABLE_MAX_W) + column + x) / 4] |= (0x3 << 4);
							screenData[(((row + y) * VIEWABLE_MAX_W) + column + x) / 4] ^= ((BGR233ToIntensityNibble[buf[(row * width) + column]] << 2) & (0x3 << 4));
							break;
						}
					case 2:
						{
							screenData[(((row + y) * VIEWABLE_MAX_W) + column + x) / 4] |= (0x3 << 2);
							screenData[(((row + y) * VIEWABLE_MAX_W) + column + x) / 4] ^= ((BGR233ToIntensityNibble[buf[(row * width) + column]] << 0) & (0x3 << 2));
							break;
						}
					case 3:
						{
							screenData[(((row + y) * VIEWABLE_MAX_W) + column + x) / 4] |= (0x3 << 0);
							screenData[(((row + y) * VIEWABLE_MAX_W) + column + x) / 4] ^= ((BGR233ToIntensityNibble[buf[(row * width) + column]] >> 2) & (0x3 << 0));
							break;
						}
				}					
			}
		}
	}
}

/******************************************************************
 * CreateWindows
 *
 * Creates the off-screen panning window and inits global variables.
 * Called only once wenn the MainFrom is loaded and opened.
 *****************************************************************/ 
static Err CreateWindows (FormPtr frmPtr)
{
  Err error;
  WinPtr canvasWinPtr;
  WinHandle prevDrawWin = WinGetDrawWindow ();
  BitmapType* keyboard_bmp;
  MemHandle keyboard_res;


  // always use BGR233 format! Maybe we will have something else in the future
  useBGR233 = True;

  /* this is a set-up for BRG233 format. We tell the server that we
   * use BRG233 and have 8 bit color depth. We map BRG233 pixels to
   * 4 bit grayscale. All that is because VNC currently does not
   * provide data formats for depths less then 8
   */

  myFormat.bitsPerPixel = 8;
  myFormat.depth = 8;
  myFormat.trueColour = 1;
  myFormat.bigEndian = 0;
  myFormat.redMax = 7;
  myFormat.greenMax = 7;
  myFormat.blueMax = 3;
  myFormat.redShift = 0;
  myFormat.greenShift = 3;
  myFormat.blueShift = 6;

  viewportX = 0;
  viewportY = 0;

  /* set up the size of the update rectangle */
  updateRect.extent.x = palmFB.viewable.w;
  updateRect.extent.y = palmFB.viewable.h;

  /* the display area take the whole main form */
  topLevel = FrmGetWindowHandle (frmPtr);

  if (!canvas)
	/* canvas = WinGetDrawWindow (); */
   canvas = WinCreateOffscreenWindow(VIEWABLE_MAX_W, VIEWABLE_MAX_H, /* palmFB.virtual.w, palmFB.virtual.h, */
		screenFormat, &error);
		
  if (!keyboard)
  {
  	keyboard_res = DmGetResource(bitmapRsc, KeyboardBitmap);
  	keyboard_bmp = MemHandleLock(keyboard_res);
    keyboard = WinCreateOffscreenWindow(keyboard_bmp->width, keyboard_bmp->height, screenFormat, &error);
    WinSetDrawWindow(keyboard);
    WinDrawBitmap(keyboard_bmp, 0, 0);
    MemHandleUnlock(keyboard_res);
    DmReleaseResource(keyboard_res);
  }

  WinSetDrawWindow (canvas);
  WinEraseWindow ();
  WinSetDrawWindow (prevDrawWin);

  canvasWinPtr = WinGetWindowPointer(canvas);

  screenData = canvasWinPtr->displayAddrV20;

  return 0;
}


/******************************************************************
 * OpenNetworkPreferences
 *
 * Opens the network preferences dialog
 *****************************************************************/ 
Boolean OpenNetworkPreferences (void)
{
	Err						err;
	UInt					cardNo;
	LocalID					dbID;
	DmSearchStateType		searchState;

	// Shut down the app, reset video, etc.
//	StopApplication();

	// Get the card number and database id of the Address application.
	err = DmGetNextDatabaseByTypeCreator (true, &searchState, 
		sysFileTPanel, sysFileCNetworkPanel, true, &cardNo, &dbID);
	ErrFatalDisplayIf(err, "Network panel not found");
	if (err) return;

	SysUIAppSwitch(cardNo, dbID, sysAppLaunchCmdPanelCalledFromApp, NULL);
}


/******************************************************************
 * OpenNetwork
 *
 * Loads the lib and connects to the network
 *****************************************************************/ 
Boolean OpenNetwork (void)
{
  Word  ifErrs;
  Err  err;

  printf ("\nOpening the network...");

  err = NetLibOpen (AppNetRefnum, &ifErrs);

  if (err == netErrAlreadyOpen || err == 0)  {
    printf("skipped\n");
    return true;
  } else {

    if (ifErrs == netErrUserCancel) {
      ShutdownNetwork (true);
      return false;
    }

    printf ("\nERROR: opening network: %s\n", ErrToString (err));

    if (ifErrs == netErrPrefNotFound)
      printf ("\nNetwork preference not set\n");
    
    return false;
  }

  printf ("opened\n");

  return true;
}


/******************************************************************
 * OpenConnection
 *
 * Opens a connection to the server
 *****************************************************************/ 
static Boolean OpenConnection (void)
{
  Err  err;
  int port = pref.server.port + VNC_DATA_PORT_NR;

  if (!networkIsOpen) {
    if ((networkIsOpen = OpenNetwork ()) != true) {
      printf ("ERROR: cannot open network\n");
      return false;
    }
  }

  if (rfbsock > 0) {
    printf ("Already connected\n");
    return true;
  }

  printf ("Connecting to \"%s:%d\"...\n",
		pref.server.address, port);

  err = 0;

  viewportX = palmFB.viewable.x = palmFB.virtual.x = pref.server.beginX;
  viewportY = palmFB.viewable.y = palmFB.virtual.y = pref.server.beginY;

  PostProgressMessage ("Connecting...", 0, 0);

  if (!ConnectToRFBServer (pref.server.address, port)) {
    printf ("ERROR: cannot connect to \"%s:%d\"\n", pref.server.address, port);
    err = true;
  } else if (!InitialiseRFBConnection(rfbsock)) {
    printf ("ERROR: cannot initialize a connection to \"%s:%d\"\n", pref.server.address, port);
    err = true;
  } else if (!SetFormatAndEncodings()) {
    err = true;
    printf ("ERROR: cannot send formats\n");
  } else if (pref.server.scaleFactor != 1)
  {
  	if (!SetScaleFactor(pref.server.scaleFactor)) {
		err = true;
		printf ("ERROR: cannot set scale factor\n");
	}
  }
  if (pref.server.scaleFactor == 1)
  {
	  if ((err == false) && (!SendFramebufferUpdateRequest(palmFB.virtual.x, palmFB.virtual.y,
			 palmFB.virtual.w, palmFB.virtual.h, False))) {
	    printf ("ERROR: cannot send buffer update\n");
	    err = true;
	  }
  }

  if (err) {
    PostProgressMessage ("Cannot connect!", 0, 0);
    ShutdownConnection ();
    return false;
  }

  /* set up the size of scrollbar slides */
  horizSlideRect.extent.x =
	(SWord)((palmFB.viewable.w * palmFB.viewable.w) /
		palmFB.remote.w);
  vertSlideRect.extent.y =
	(SWord)((palmFB.viewable.h * palmFB.viewable.h) /
		palmFB.remote.h);

  /* set the pen time tracing, used to catch double-taps */
  lastPenTime = TimGetTicks ();

  printf ("Connected!\n");

  return true;
}


/******************************************************************
 * ShutdownConnection
 *
 * Closes the current connection to the VNC server
 *****************************************************************/ 
void ShutdownConnection ()
{
  if (rfbsock > 0) {
    close (rfbsock);
    rfbsock = 0;

    PrintSentRecvStat ();

    printf ("Connection closed\n\n");

    /* set up the size of scrollbar slides */
    horizSlideRect.extent.x = 0;
    vertSlideRect.extent.y = 0;

    viewportX = palmFB.viewable.x = palmFB.virtual.x = 0;
    viewportY = palmFB.viewable.y = palmFB.virtual.y = 0;

    palmFB.remote.h = palmFB.remote.w = 0;

    /*InitVirtual ();*/

    PositionScrollbars ();
/*  UpdateViewport(); */

    PostProgressMessage ("Server connection closed", 0, 0);
  }
}


/******************************************************************
 * ShutdownNetwork
 *
 * Closes the network. If enforcedClose == True requests the modem
 * to hang up, otherwise closes only ots own API to the Network lib.
 *****************************************************************/ 
void ShutdownNetwork (Boolean enforcedClose)
{
  Err err;

  ShutdownConnection ();

  printf ("Closing NetLib...\n");
  err = NetLibClose (AppNetRefnum, enforcedClose);

  if (err == netErrNotOpen)
    printf ("ERROR: cannot close: %s\n", ErrToString(err));
  else if (err == netErrStillOpen)
    printf ("Network is still in use\n");
  else
    printf ("Network is closed\n");
	
  networkIsOpen = false;
  return;
}


/******************************************************************
 * SendKeySequence
 *
 * Sends a sequence of n characters as a keyboard input to the VNC
 * server. Works with unsigned short data because the server works
 * with 16bit KeySyms from X11. mod1 and mod2 are two available
 * modifiers. If ascii is true, convert the data into a sequence of
 * unsigned short
 *****************************************************************/ 
void SendKeySequence (unsigned short mod1, unsigned short mod2, unsigned short mod3,
		      unsigned short *keys, int n, Boolean ascii)
{
  int i;
  unsigned short key;

  if (keys == 0) {
    return;
  }

  if (mod1 > 0)
    SendKeyEvent(mod1, true); // modifier 1 pressed

  if (mod2 > 0)
    SendKeyEvent(mod2, true); // modifier 2 pressed
    
  if (mod3 > 0)
    SendKeyEvent(mod3, true); // modifier 3 pressed

  for (i = 0; i < n; i++) {
    if (ascii)
      key = ((unsigned char *)keys)[i];
    else
      key = keys[i];

  if (key == 10) key = 13; // Override for carriage return

	if ((keyboard_state & (1 << keyboard_visible)) && (keyboard_state & (1 << keyboard_shift)))
    {
    	if ((key >= 'a') && (key <= 'z')) key -= ('a' - 'A');
    }

	if ((keyboard_state & (1 << keyboard_visible)) && (keyboard_state & (1 << keyboard_fn)))
	{
		if ((key >= '1') && (key <= '9')) key = (key - '1') + F1_KEY;
		if (key == '0') key = F1_KEY + 9;
	}
	

    SendKeyEvent(key, true); // pressed
    SendKeyEvent(key, false); // released
  }

  if (mod3 > 0)
    SendKeyEvent(mod3, false); // released

  if (mod2 > 0)
    SendKeyEvent(mod2, false); // released

  if (mod1 > 0)
    SendKeyEvent(mod1, false); // released
}


/******************************************************************
 * PositionViewport
 *
 * Moves visible area over the virtual panning window. If the
 * the boundaries of the visible area is beyond the panning window,
 * the panning window will be moved and a new portion of the screen
 * data will be requested
 *****************************************************************/ 
static void PositionViewport ()
{
  int deltaX=0, deltaY=0;
  int updateViewport = False;

  if (viewportX < 0)
    viewportX = 0;

  if (viewportX > (palmFB.remote.w - palmFB.viewable.w))
    viewportX = palmFB.remote.w - palmFB.viewable.w;

  if (viewportY < 0)
    viewportY = 0;

  if (viewportY > (palmFB.remote.h - palmFB.viewable.h))
    viewportY = palmFB.remote.h - palmFB.viewable.h;

  if (palmFB.viewable.x != viewportX) {
    if (viewportX < palmFB.virtual.x ||
	viewportX + palmFB.viewable.w > palmFB.virtual.w) {
      deltaX = (int)palmFB.viewable.x - viewportX;
    }
    updateViewport = True;
    palmFB.viewable.x = viewportX;
  }

  if ((int)palmFB.virtual.x - deltaX < 0)
     deltaX = palmFB.virtual.x;

  if (palmFB.viewable.y != viewportY) {
    if (viewportY < palmFB.virtual.y ||
	viewportY + palmFB.viewable.h > palmFB.virtual.h) {
      deltaY = (int)palmFB.viewable.y - viewportY;
    }
    updateViewport = True;
    palmFB.viewable.y = viewportY;
  }

  if ((int)palmFB.virtual.y - deltaY < 0)
    deltaY = palmFB.virtual.y;

  if (deltaX != 0 || deltaY != 0) {
    int obscuredRectX=0, obscuredRectY=0;
    RectangleType r;

    r.topLeft.x = r.topLeft.y = 0;
    r.extent.x = palmFB.virtual.w;
    r.extent.y = palmFB.virtual.h;

	if (deltaX > 0) r.extent.x -= deltaX;
	if (deltaY > 0) r.extent.y -= deltaY;

    WinCopyRectangle (canvas, canvas, &r, deltaX, deltaY, scrCopy);

    if (deltaX < 0)
      obscuredRectX = (int)palmFB.virtual.w + deltaX;

    if (deltaX != 0) {

      DrawVirtualFilledRectangle (LIGHT_GRAY, obscuredRectX, 0,
		abs(deltaX), palmFB.virtual.h, false);

      palmFB.virtual.x -= deltaX;

      (void)SendFramebufferUpdateRequest (palmFB.virtual.x + obscuredRectX,
					palmFB.virtual.y,
					abs(deltaX),
					palmFB.virtual.h,
					False);
    }

    if (deltaY < 0)
      obscuredRectY = (int)palmFB.virtual.h + deltaY;

    if (deltaY != 0) {

      DrawVirtualFilledRectangle (LIGHT_GRAY, 0, obscuredRectY,
		palmFB.virtual.w, abs(deltaY), false);

      palmFB.virtual.y -= deltaY;

      (void)SendFramebufferUpdateRequest (palmFB.virtual.x,
					palmFB.virtual.y + obscuredRectY,
					palmFB.virtual.w,
					abs(deltaY),
					False);
    }

    sendUpdateRequest = true;
    updateViewport = true;
  }

  if (updateViewport)
    UpdateViewport ();
}


/******************************************************************
 * UpdateViewport
 *
 * Copies the viewable portion of the off-screen window into the
 * visible window
 *****************************************************************/ 
static void UpdateViewport (void)
{
  RectangleType rect;
  int i;
  
  if (rfbsock == 0)
	return;

 if (DrawingBlocked ())
   return;

  WinCopyRectangle (canvas, topLevel,
	GetUpdateRectAtPos (palmFB.viewable.x - palmFB.virtual.x,
			    palmFB.viewable.y - palmFB.virtual.y),
	0, 0, scrCopy);

  	if (keyboard_state & (1 << keyboard_visible))
	{
		/* Copy keyboard */
		rect.extent.x = 1 + keyboard_keywidth;
		rect.extent.y = 6;
		for (i = 0; i < keyboard_num_controls; i++)
		{
			rect.topLeft.x = keyboard_keywidth * i;
			if (keyboard_state & (1 << i))
				rect.topLeft.y = 6;
			else
				rect.topLeft.y = 0;
			WinCopyRectangle(keyboard, topLevel, &rect, rect.topLeft.x, 150, scrCopy);
		}
	}
}


/******************************************************************
 * UpdateScrollbars
 *
 * Repaints scrollbars
 *****************************************************************/ 
static void UpdateScrollbars (void)
{
  IndexedColorType oldColor;

  mySetForeColor (LIGHT_GRAY);
  WinDrawRectangle (&horizScrollbarRect, 0);
  WinDrawRectangle (&vertScrollbarRect, 0);

  mySetForeColor (BLACK);
  WinDrawRectangle (&horizSlideRect, 0);
  WinDrawRectangle (&vertSlideRect, 0);
}


/******************************************************************
 * PositionScrollbars
 *
 * re-computes scrollbar rectangles and re-paints scrollbars
 *****************************************************************/ 
static void PositionScrollbars ()
{
  if (palmFB.remote.w == 0)
    horizSlideRect.topLeft.x = 0;
  else
    horizSlideRect.topLeft.x = (((UInt32)viewportX * (UInt32)palmFB.viewable.w) / palmFB.remote.w);

  if (palmFB.remote.h == 0)
    vertSlideRect.topLeft.y = 0;
  else
    vertSlideRect.topLeft.y = ((UInt32)(viewportY * (UInt32)palmFB.viewable.h) / palmFB.remote.h);

  UpdateScrollbars ();
}


/******************************************************************
 * PointInRectangle
 *
 * returns true if the point (x,y) is in the rect. rp otherwise
 * false
 *****************************************************************/ 
Boolean PointInRectangle (unsigned x, unsigned y, RectangleType *rp)
{
  return ((x >= rp->topLeft.x) &&
	(y >= rp->topLeft.y) &&
	((unsigned)(x - rp->topLeft.x) < rp->extent.x) &&
	((unsigned)(y - rp->topLeft.y) < rp->extent.y));
}


/******************************************************************
 * InMainArea
 *
 * returns true if the point (x,y) is in the VNC area otherwise
 * false
 *****************************************************************/ 
Boolean InMainArea (unsigned short x, unsigned short y)
{
  return (x < palmFB.viewable.w &&
	  y < palmFB.viewable.h);
}

/******************************************************************
 * InKeyboardArea
 *
 * returns true if the point (x,y) is in the keyboard area otherwise
 * false
 *****************************************************************/ 
Boolean InKeyboardArea (unsigned short x, unsigned short y)
{
	if (!(keyboard_state & (1 << keyboard_visible))) return false;
	if ((x < (keyboard_num_controls * keyboard_keywidth + 1)) && (y >= 150) && (y <= 155))
		return true;
	return false;
}


/******************************************************************
 * ToggleKeyboardState
 *
 * Toggles the special keyboard on and off
 ******************************************************************/
static void ToggleKeyboardState (void)
{
	keyboard_state ^= (1 << keyboard_visible);
	if (keyboard_state & (1 << keyboard_visible))
	{
   		if (keyboard_state & (1 << keyboard_shift))
	    	SendKeyEvent(SHIFT_KEY, true);
   		if (keyboard_ctrl & (1 << keyboard_ctrl))
	    	SendKeyEvent(CTRL_KEY, true);
   		if (keyboard_alt & (1 << keyboard_alt))
	    	SendKeyEvent(ALT_KEY, true);
	}
	else
	{
   		if (keyboard_state & (1 << keyboard_shift))
	    	SendKeyEvent(SHIFT_KEY, false);
   		if (keyboard_state & (1 << keyboard_ctrl))
	    	SendKeyEvent(CTRL_KEY, false);
   		if (keyboard_state & (1 << keyboard_alt))
	    	SendKeyEvent(ALT_KEY, false);
	}
}


/******************************************************************
 * ProcessKeyboardPenDownEvent
 *
 * Handles events penDown for the keyboard masked area
 ******************************************************************/
static void ProcessKeyboardPenDownEvent (EventPtr e)
{
	unsigned short key;
	if (e->screenX >= 156)
	{
		ToggleKeyboardState();
	}		
	else
	{
		key = (e->screenX / keyboard_keywidth);
		keyboard_state ^= (1 << key);
       	if (key == keyboard_shift)
       	{
       		if (keyboard_state & (1 << keyboard_shift))
		    	SendKeyEvent(SHIFT_KEY, true);
		    else
		    	SendKeyEvent(SHIFT_KEY, false);
		}
       	if (key == keyboard_ctrl)
       	{
       		if (keyboard_state & (1 << keyboard_ctrl))
		    	SendKeyEvent(CTRL_KEY, true);
		    else
		    	SendKeyEvent(CTRL_KEY, false);
		}
       	if (key == keyboard_alt)
       	{
       		if (keyboard_state & (1 << keyboard_alt))
		    	SendKeyEvent(ALT_KEY, true);
		    else
		    	SendKeyEvent(ALT_KEY, false);
		}
	}
	
	UpdateViewport();
}

/******************************************************************
 * ProcessKeyboardPenUpEvent
 *
 * Handles events penUp for the keyboard masked area
 ******************************************************************/
static void ProcessKeyboardPenUpEvent (EventPtr e)
{
	unsigned short key;
	unsigned short oneKey;
	if (e->screenX >= 156)
		return;
	
	key = (e->screenX / keyboard_keywidth);
	
	/* Escape key */
	if (key == 0)
	{
		keyboard_state ^= (1 << key);
		oneKey = ESC_KEY;
		SendKeySequence(0, 0, 0, &oneKey, 1, false);
		UpdateViewport();
	}
}


/******************************************************************
 * InScrollbar
 *
 * returns true if the point (x,y) is in the specified scrollbar
 * area otherwise false
 *****************************************************************/ 
Boolean InScrollbar (unsigned short barType, unsigned short x, unsigned short y)
{
  Boolean ret = false;

  switch (barType) {
    case HORIZ:
      ret = PointInRectangle (x, y, &horizSlideRect);
      break;

    case HORIZ_UP:
      ret = PointInRectangle (x, y, &horizScrollbarRect) &&
		!PointInRectangle (x, y, &horizSlideRect) &&
		(x < horizSlideRect.topLeft.x);
      break;

    case HORIZ_DOWN:
      ret = PointInRectangle (x, y, &horizScrollbarRect) &&
		!PointInRectangle (x, y, &horizSlideRect) &&
		(x > horizSlideRect.topLeft.x + horizSlideRect.extent.x);
      break;

    case HORIZ_TOP:
      ret = false;
      break;

    case HORIZ_BUTTOM:
      ret = false;
      break;

    case VERT:
      ret = PointInRectangle (x, y, &vertSlideRect);
      break;

    case VERT_UP:
      ret = PointInRectangle (x, y, &vertScrollbarRect) &&
		!PointInRectangle (x, y, &vertSlideRect) &&
		(y < vertSlideRect.topLeft.y);
      break;

    case VERT_DOWN:
      ret = PointInRectangle (x, y, &vertScrollbarRect) &&
		!PointInRectangle (x, y, &vertSlideRect) &&
		(y > vertSlideRect.topLeft.y + horizSlideRect.extent.y);
      break;

    case VERT_TOP:
      ret = false;
      break;

    case VERT_BUTTOM:
      ret = false;
      break;

    default:
      return false;
  }

  return ret;
}


/******************************************************************
 * ProcessScrollPenMoveEvent
 *
 * Handles events penMoveEvent in the main form
 *****************************************************************/
static void ProcessScrollPenMoveEvent (EventPtr e)
{
  if (buttonState == HORIZ) {
    viewportX += ((e->screenX - downX) * (int)(palmFB.remote.w
			/ palmFB.viewable.w));
    downX = e->screenX;

  } else if (buttonState == VERT) {
    viewportY += ((e->screenY - downY) * (int)(palmFB.remote.h
			/ palmFB.viewable.h));
    downY = e->screenY;

  } else
    return;

  PositionViewport ();
  PositionScrollbars ();
}


/******************************************************************
 * ProcessScrollKeyEvent
 *
 * Handles events keyEvent for the scrolling hard keys
 ******************************************************************/
static void ProcessScrollKeyEvent (EventPtr e)
{
	DWord keystate = KeyCurrentState();
	if (keystate & keyBitPageUp) viewportY -= 16;
	if (keystate & keyBitPageDown) viewportY += 16;
	if (keystate & keyBitHard3) viewportX -= 16;
	if (keystate & keyBitHard4) viewportX += 16;
	
	PositionViewport();
	PositionScrollbars();	
		
}


/******************************************************************
 * ProcessScrollPenDownEvent
 *
 * Handles events penDownEvent in scrollers
 *****************************************************************/
static void ProcessScrollPenDownEvent (EventPtr e)
{
  if (InScrollbar (HORIZ, e->screenX , e->screenY)) {
    downX =  e->screenX;

    buttonState = HORIZ;
    return;
  } else if (InScrollbar (VERT,  e->screenX ,  e->screenY)) {
    downY =  e->screenY;

    buttonState = VERT;
    return;
  } else if (InScrollbar (HORIZ_UP,  e->screenX ,  e->screenY)) {
    viewportX -= palmFB.viewable.w;

  } else if (InScrollbar (HORIZ_DOWN, e->screenX , e->screenY)) {
    viewportX += palmFB.viewable.w;

  } else if (InScrollbar (VERT_UP, e->screenX , e->screenY)) {
    viewportY -= palmFB.viewable.h;

  } else if (InScrollbar (VERT_DOWN, e->screenX , e->screenY)) {
    viewportY += palmFB.viewable.h;

  } else
    return;

  buttonState = 0;
  PositionViewport ();
  PositionScrollbars ();
}


/******************************************************************
 * ProcessPenMoveEvent
 *
 * Handles events penMoveEvent in the main form
 *****************************************************************/
static void ProcessPenMoveEvent (EventPtr e)
{
  unsigned long penTime = TimGetTicks ();
  unsigned short serverX, serverY;

  if (DrawingBlocked ())
    return;

  if (penTime <= lastPenTime + 100)
    return;

  serverX = e->screenX + palmFB.viewable.x;
  serverY = e->screenY + palmFB.viewable.y;

  if (buttonState == MAIN_PRESSED || buttonState == MAIN_MOVED) {
    (void)SendPointerEvent (serverX, serverY, currentMouseButton);
    buttonState = MAIN_MOVED;

  }
}


/******************************************************************
 * ProcessPenUpDownEvent
 *
 * Handles events penUpEvent and PenDownEvent in the main form
 *****************************************************************/
static void ProcessPenUpDownEvent (EventPtr e)
{
  	unsigned long penTime;
  	unsigned short serverX, serverY;

  	if (DrawingBlocked ())
    	return;

  	penTime = TimGetTicks ();
  	serverX = e->screenX + palmFB.viewable.x;
  	serverY = e->screenY + palmFB.viewable.y;

  	if (lastX == 0)
    	lastX = serverX;
  	if (lastY == 0)
    	lastY = serverY;

  	if (e->penDown)
  	{
  		if (pen_zooming == true)
  		{
  			palmFB.zoomarea_ena = true;
  			palmFB.zoomarea.x = e->screenX;
  			palmFB.zoomarea.y = e->screenY;
  			/* viewportY = palmFB.viewable.y = palmFB.virtual.y = e->screenY; */
  	  		SetScaleFactor(pref.server.scaleFactor);
  	  		pen_zooming = false;
  	  		return;
  		}
  		else
  		{
  			if (buttonState == MAIN_ONCE && penTime <= lastPenTime + 100)
    		{
      			buttonState = MAIN_DOUBLE;
      			(void)SendPointerEvent (lastX, lastY, currentMouseButton);
      			(void)SendPointerEvent (lastX, lastY, 0);
    		}
    		else 
    		{
      			buttonState = MAIN_PRESSED;
      			// sync the intial mouse position first
      			(void)SendPointerEvent (serverX, serverY, 0);
      			(void)SendPointerEvent (serverX, serverY, currentMouseButton);
    		}

    		lastX = serverX;
    		lastY = serverY;
    		return;
  		}
  	}
  	
  	else
  	{ 	// penUp
    	if (buttonState == MAIN_PRESSED) 
    	{
    		(void)SendPointerEvent (lastX, lastY, 0);

    		buttonState = MAIN_ONCE;
    	}
    	else if (buttonState == MAIN_MOVED) 
    	{
    		buttonState = 0;
    		(void)SendPointerEvent (serverX, serverY, 0);
    	}


	    // Allow the masked off key to now send keyDownEvents.
	    KeySetMask(keyBitsAll);
	    currentMouseButton = 1;
  	}

  	lastPenTime = penTime;
}


/******************************************************************
 * PrefFormHandleEvent
 *
 * Handles events in the console form
 *****************************************************************/ 
static Boolean PrefFormHandleEvent (EventPtr e)
{
  Boolean handled, closePref;
  FormPtr frm;
  FormPtr passwdFrm;
  FieldPtr  fieldPtr;
  FieldPtr  pw_Fptr;
  VoidHand  addrH, portH, passwdH, beginXH, beginYH, scaleFactorH, loginNTH, passwNTH;
  char buffer[32];
  char *str;
  Word field;

  CALLBACK_PROLOGUE

	frm = FrmGetActiveForm();
 	handled = false;
	closePref = false;

  switch (e->eType) {
    case frmOpenEvent:
      fieldPtr = FrmGetObjectPtr (frm,
		FrmGetObjectIndex (frm, ServerAddressField));
      addrH = MemHandleNew (32);
      MemMove (MemHandleLock (addrH),
		pref.server.address, StrLen (pref.server.address)+1);
      MemHandleUnlock (addrH);
      FldSetTextHandle (fieldPtr, (Handle)addrH);

      fieldPtr = FrmGetObjectPtr (frm,
		FrmGetObjectIndex (frm, ServerPortField));
      portH = MemHandleNew (6);
      str = StrIToA (buffer, pref.server.port);
      MemMove (MemHandleLock (portH), str, StrLen (str)+1);
      MemHandleUnlock (portH);
      FldSetTextHandle (fieldPtr, (Handle)portH);

	  if (StrLen(pref.server.password) == 0)
	  {
		CtlSetLabel(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, PasswordField)), "-Prompt-");
	  }
	  else
	  {
		CtlSetLabel(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, PasswordField)), "-Assigned-");
	  }

	  if (StrLen(pref.server.passwNT) == 0)
	  {
		CtlSetLabel(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, PasswordNTField)), "-Prompt-");
	  }
	  else
	  {
		CtlSetLabel(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, PasswordNTField)), "-Assigned-");
	  }

      fieldPtr = FrmGetObjectPtr (frm,
		FrmGetObjectIndex (frm, BeginXField));
      beginXH = MemHandleNew (5);
      str = StrIToA (buffer, pref.server.beginX);
      MemMove (MemHandleLock (beginXH), str, StrLen (str)+1);
      MemHandleUnlock (beginXH);
      FldSetTextHandle (fieldPtr, (Handle)beginXH);

      fieldPtr = FrmGetObjectPtr (frm,
		FrmGetObjectIndex (frm, BeginYField));
      beginYH = MemHandleNew (5);
      str = StrIToA (buffer, pref.server.beginY);
      MemMove (MemHandleLock (beginYH), str, StrLen (str)+1);
      MemHandleUnlock (beginYH);
      FldSetTextHandle (fieldPtr, (Handle)beginYH);

      fieldPtr = FrmGetObjectPtr (frm,
		FrmGetObjectIndex (frm, ScaleFactorField));
      scaleFactorH = MemHandleNew (2);
      str = StrIToA (buffer, pref.server.scaleFactor);
      MemMove (MemHandleLock (scaleFactorH), str, StrLen (str)+1);
      MemHandleUnlock (scaleFactorH);
      FldSetTextHandle (fieldPtr, (Handle)scaleFactorH);

      FrmSetControlValue (frm, FrmGetObjectIndex (frm, ShareCheckBox), 
		pref.server.sharedDesktop);

      fieldPtr = FrmGetObjectPtr (frm,
		FrmGetObjectIndex (frm, LoginNameNTField));
      loginNTH = MemHandleNew (32);
      MemMove (MemHandleLock (loginNTH),
		pref.server.loginNameNT, StrLen (pref.server.loginNameNT)+1);
      MemHandleUnlock (loginNTH);
      FldSetTextHandle (fieldPtr, (Handle)loginNTH);

      FrmDrawForm(frm);
      break;

    case frmCloseEvent:
      closePref = true;
      handled = true;
      break;

    case ctlEnterEvent:
      switch(e->data.ctlSelect.controlID) {
	case PrefOkButton:
	  str = FldGetTextPtr (FrmGetObjectPtr (frm,
		FrmGetObjectIndex (frm, ServerAddressField)));
	  MemSet (pref.server.address, 32, 0);
	  MemMove (pref.server.address, str, StrLen (str));

	  str = FldGetTextPtr (FrmGetObjectPtr (frm,
		FrmGetObjectIndex (frm, ServerPortField)));
	  pref.server.port = StrAToI (str);

	  str = FldGetTextPtr (FrmGetObjectPtr (frm,
		FrmGetObjectIndex (frm, BeginXField)));
	  pref.server.beginX = StrAToI (str);

	  str = FldGetTextPtr (FrmGetObjectPtr (frm,
		FrmGetObjectIndex (frm, BeginYField)));
	  pref.server.beginY = StrAToI (str);

	  str = FldGetTextPtr (FrmGetObjectPtr (frm,
		FrmGetObjectIndex (frm, ScaleFactorField)));
	  pref.server.scaleFactor = StrAToI (str);

	  str = FldGetTextPtr (FrmGetObjectPtr (frm,
		FrmGetObjectIndex (frm, LoginNameNTField)));
	  MemSet (pref.server.loginNameNT, 32, 0);
	  MemMove (pref.server.loginNameNT, str, StrLen (str));

	  // do not break here, go through PrefCancleButton to set-up freeing

	case PrefCancelButton:
		handled = true;
        closePref = true;
	  	break;

	default:
	  break;
      }
      break;

    case ctlSelectEvent:
      switch(e->data.ctlSelect.controlID) {
        case ShareCheckBox:
	  pref.server.sharedDesktop = e->data.ctlSelect.on;
	  break;

		case PasswordField:
			passwdFrm = FrmInitForm(PasswordForm);
		    if (FrmDoDialog (passwdFrm) == PassFormOkButton)
			{
		      	pw_Fptr = FrmGetObjectPtr (passwdFrm,
											FrmGetObjectIndex (passwdFrm, PassFormPassField));

				MemSet (&pref.server.password, 32, 0);
				if (FldGetTextPtr (pw_Fptr) != NULL)
					MemMove (&pref.server.password, FldGetTextPtr (pw_Fptr), StrLen (FldGetTextPtr (pw_Fptr)) + 1);
			}
			FrmDeleteForm(passwdFrm);
			  if (StrLen(pref.server.password) == 0)
			  {
				CtlSetLabel(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, PasswordField)), "-Prompt-");
			  }
			  else
			  {
				CtlSetLabel(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, PasswordField)), "-Assigned-");
			  }
			handled = true;
			break;
	
		case PasswordNTField:
			passwdFrm = FrmInitForm(PasswordForm);
		    if (FrmDoDialog (passwdFrm) == PassFormOkButton)
			{
		      	pw_Fptr = FrmGetObjectPtr (passwdFrm,
											FrmGetObjectIndex (passwdFrm, PassFormPassField));

				MemSet (&pref.server.passwNT, 32, 0);
				if (FldGetTextPtr (pw_Fptr) != NULL)
					MemMove (&pref.server.passwNT, FldGetTextPtr (pw_Fptr), StrLen (FldGetTextPtr (pw_Fptr)) + 1);
			}
			FrmDeleteForm(passwdFrm);
			  if (StrLen(pref.server.passwNT) == 0)
			  {
				CtlSetLabel(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, PasswordNTField)), "-Prompt-");
			  }
			  else
			  {
				CtlSetLabel(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, PasswordNTField)), "-Assigned-");
			  }
			handled = true;
			break;

        default:
          break;
      }
      break;

    case menuEvent:
	MenuEraseStatus(NULL);
	field = FrmGetFocus(frm);
	switch(e->data.menu.itemID) {
	  case EditMenuCopyItem:
	    FldCopy (FrmGetObjectPtr (frm, field));
	    break;

	  case EditMenuPasteItem:
	    FldPaste (FrmGetObjectPtr (frm, field));
	    break;

	  default:
	    break;
	}
	break;

    default:
        break;

  }

  if (closePref) {
    //the memory under field text handles is freed by the UI manager
    CALLBACK_EPILOGUE
    FrmReturnToForm (MainForm);
  }

  CALLBACK_EPILOGUE

  return handled;
}

/******************************************************************
 * PasswordFormHandleEvent
 *
 * Handles events in the password entry form
 *****************************************************************/ 
static Boolean PasswordFormHandleEvent (EventPtr e)
{
  Boolean handled, closePassForm;
  FormPtr frm;
  FieldPtr  fieldPtr;
  VoidHand  passwdH;
  char buffer[32];
  char *str;
  Word field;

  CALLBACK_PROLOGUE

	frm = FrmGetActiveForm();
 	handled = false;
	closePassForm = false;

  	switch (e->eType) 
	{
    	case frmOpenEvent:
	      	fieldPtr = FrmGetObjectPtr (frm,
										FrmGetObjectIndex (frm, PassFormPassField));
  	    	passwdH = MemHandleNew (32);
 	     	MemMove (MemHandleLock (passwdH),
									"\0", 1);
 	     	MemHandleUnlock (passwdH);
 	     	FldSetTextHandle (fieldPtr, (Handle)passwdH);
/*			FrmSetFocus(frm, FrmGetObjectIndex (frm, PassFormPassField));*/
 	     	FrmDrawForm(frm);
/*			FldSetInsPtPosition(fieldPtr, 0); */
  	    	break;

  		case frmCloseEvent:
 	    	closePassForm = true;
	      	handled = true;
	      	break;

    	case ctlSelectEvent:
      		switch(e->data.ctlSelect.controlID)
			{
				case PassFormOkButton:

				  	// do not break here, go through PassFormCancleButton to set-up freeing

				case PassFormCancelButton:
					handled = true;
        			closePassForm = true;
	  				break;

				default:
	 				 break;
    		}
      		break;

    	case menuEvent:
			MenuEraseStatus(NULL);
			field = FrmGetFocus(frm);
			switch(e->data.menu.itemID) 
			{
	  			case EditMenuCopyItem:
	    			FldCopy (FrmGetObjectPtr (frm, field));
	    			break;

	 	 		case EditMenuPasteItem:
	    			FldPaste (FrmGetObjectPtr (frm, field));
	    			break;

	  			default:
	    			break;
			}
			break;

    	default:
        	break;
  	}

  	if (closePassForm)
	{
    	// the memory under field text handles is freed by the UI manager
    	CALLBACK_EPILOGUE
    	FrmReturnToForm (PrefForm);
  	}

  	CALLBACK_EPILOGUE
  	return handled;
}

/******************************************************************
 * ConsoleFormHandleEvent
 *
 * Handles events in the Console form
 *****************************************************************/ 
static Boolean ConsoleFormHandleEvent (EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    Word field;

    CALLBACK_PROLOGUE

	handled = false;
	frm = FrmGetActiveForm();

    // Let StdIO handler to it's things first.
    if (StdHandleEvent(e)) {
      CALLBACK_EPILOGUE
      return true;
    }

    switch (e->eType) {
    case frmOpenEvent:
	FrmDrawForm(frm);
	handled = true;
	break;

    case ctlSelectEvent:
	switch(e->data.ctlSelect.controlID) {
	  case ConsoleCloseButton:
	    (void)StdSave();
	    FrmReturnToForm (MainForm);
	    handled = true;
	    break;
	  case ConsoleClearButton:
	    StdClear();
	    handled = true;
	    break;
	  default:
	    handled = true;
	    break;
	}
	break;

    case menuEvent:
	MenuEraseStatus(NULL);
	field = FrmGetFocus(frm);
	switch(e->data.menu.itemID) {
	  case EditMenuCopyItem:
	    FldCopy (FrmGetObjectPtr (frm, field));
	    break;

	  case EditMenuPasteItem:
	    FldPaste (FrmGetObjectPtr (frm, field));
	    break;

	  default:
	    break;
	}
	break;

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}


/******************************************************************
 * MainFormHandleEvent
 *
 * 
 *****************************************************************/ 
static Boolean MainFormHandleEvent (EventPtr e)
{
    Boolean handled;
    FormPtr frm;
    unsigned short oneKey;
    Word clipboardLength;
    VoidHand handle;
    VoidPtr ptr;
	int scale;

	char mypass[32];
  	FormPtr passwdFrm;
  	FieldPtr  pw_Fptr;

    CALLBACK_PROLOGUE

	handled = false;

    switch (e->eType) {
    case frmOpenEvent:
	frm = FrmGetActiveForm();
	FrmDrawForm(frm);
	CreateWindows (frm);

    InitVirtual ();
	UpdateScrollbars ();
	UpdateViewport ();
	handled = true;
	break;

    case frmUpdateEvent:
	UpdateViewport ();
	UpdateScrollbars ();

	handled = true;
	break;

	case nilEvent:
		UpdateViewport ();
		break;

    case menuEvent:
	/*MenuEraseStatus(NULL);*/

	switch(e->data.menu.itemID) {
	  case MainMenuConsoleItem:
	    FrmPopupForm (ConsoleForm);
	    UpdateViewport ();
	    break;

	  case MainMenuPrefItem:
	    (void)FrmPopupForm (PrefForm);
	    UpdateViewport ();
	    break;

	  case MainMenuOpenItem:
	    if (!OpenConnection ())
	      (void)FrmPopupForm (ConsoleForm);
	    else
              PositionScrollbars ();
	    break;

	  case MainMenuCloseItem:
	    ShutdownConnection ();
	    break;

	  case MainMenuHangUp:
	    ShutdownNetwork (true);
	    PostProgressMessage ("Modem disconnected", 0, 0);
	    break;

	  case MainMenuNetworkPrefsItem:
	    OpenNetworkPreferences();
	    break;

	  case MainMenuAboutItem:
	    FrmAlert(AboutAlert);
	    break;

	  case MainMenuHelpItem:
	    FrmHelp(HelpText);
	    break;

          // request a "hard" update of the whole virtual area
	  case MainMenuUpdateItem:
	    (void)SendFramebufferUpdateRequest((int)palmFB.virtual.x,
		(int)palmFB.virtual.y,
		(int)palmFB.virtual.w,
		(int)palmFB.virtual.h,
		false);
	    break;

	  case MainMenuSavePosItem:
            pref.server.beginX = palmFB.viewable.x;
            pref.server.beginY = palmFB.viewable.y;
            break;

	  case MainMenuSendCADItem:
	    oneKey = DEL_KEY;
	    SendKeySequence (CTRL_KEY, ALT_KEY, 0, &oneKey, 1, false);
	    break;

      case MainMenuSpecialKeysItem:
      	ToggleKeyboardState();
      	break;


	  case MainMenuAltF4Item:
	    oneKey = F4_KEY;
	    SendKeySequence (ALT_KEY, 0, 0, &oneKey, 1, false);
	    break;

	  case MainMenuNTLoginItem:
	    // switch to the user name text field
	    oneKey = (unsigned char)'u';
	    SendKeySequence (ALT_KEY, 0, 0, &oneKey, 1, false);

	    // send the user name
	    if (StrLen (pref.server.loginNameNT) > 0) {
	      SendKeySequence (0, 0, 0, (unsigned short *)&pref.server.loginNameNT,
				StrLen (pref.server.loginNameNT), true);

	      // switch to the password text field
	      oneKey = (unsigned char)'p';
	      SendKeySequence (ALT_KEY, 0, 0, &oneKey, 1, false);

	      // send the password
			if (StrLen(pref.server.passwNT) == 0)
			{
				MemSet (&mypass, 32, 0);
				passwdFrm = FrmInitForm(PasswordForm);
			    if (FrmDoDialog (passwdFrm) == PassFormOkButton)
				{
			      	pw_Fptr = FrmGetObjectPtr (passwdFrm,
												FrmGetObjectIndex (passwdFrm, PassFormPassField));
		
					if (FldGetTextPtr (pw_Fptr) != NULL)
						MemMove (&mypass, FldGetTextPtr (pw_Fptr), StrLen (FldGetTextPtr (pw_Fptr)) + 1);
				}
				FrmDeleteForm(passwdFrm);
			}
			else
			{
				StrCopy(mypass, pref.server.passwNT);
			}
	      if (StrLen (mypass) > 0)
	        SendKeySequence (0, 0, 0, (unsigned short *)&mypass,
				StrLen (mypass), true);
	    }

	    // confirm with "enter", even if no user name or password 
	    oneKey = ENTER_KEY;
	    SendKeySequence (0, 0, 0, &oneKey, 1, false);

	    break;

	  case MainMenuNTLogoutItem:
	    oneKey = (unsigned char)'l';
	    SendKeySequence (ALT_KEY, 0, 0, &oneKey, 1, false);
	    oneKey = ENTER_KEY;
	    SendKeySequence (0, 0, 0, &oneKey, 1, false);
	    break;

	  case MainMenuNTLockItem:
	    oneKey = (unsigned char)'w';
	    SendKeySequence (ALT_KEY, 0, 0, &oneKey, 1, false);
	    break;

	  case MainMenuRebootNTItem:
	    oneKey = (unsigned char)'s';
	    SendKeySequence (ALT_KEY, 0, 0, &oneKey, 1, false);
	    oneKey = (unsigned char)'r';
	    SendKeySequence (ALT_KEY, 0, 0, &oneKey, 1, false);
	    oneKey = ENTER_KEY;
	    SendKeySequence (0, 0, 0, &oneKey, 1, false);
	    break;

	  case MainMenuPasteItem:
	    handle = ClipboardGetItem (clipboardText, &clipboardLength);
	    if (handle != 0 && clipboardLength != 0) {
	      ptr = MemHandleLock (handle);
	      SendClientCutText (ptr, clipboardLength);
	      (void)MemHandleUnlock (handle);
	    }
	    break;

	  case MainMenuZoomDesktopItem:
	  case MainMenuZoom1Item:
	  case MainMenuZoom2Item:
	  case MainMenuZoom3Item:
	  case MainMenuZoom4Item:
		palmFB.zoomarea_ena = true;
		palmFB.zoomarea.x = viewportX + (palmFB.viewable.w / 2);
		palmFB.zoomarea.y = viewportY + (palmFB.viewable.h / 2);
		switch (e->data.menu.itemID)
		{
			case MainMenuZoomDesktopItem:
				scale = (palmFB.desktop.w / palmFB.viewable.w);
				if (palmFB.desktop.w > (palmFB.viewable.w * scale)) scale++;
				if ((palmFB.desktop.h / palmFB.viewable.h) > scale) scale = (palmFB.desktop.h / palmFB.viewable.h);
				if (palmFB.desktop.h > (palmFB.viewable.h * scale)) scale++;
				pen_zooming = true;
				break;

			case MainMenuZoom2Item: 	scale = 2; 	break;
			case MainMenuZoom3Item:		scale = 3;	break;
			case MainMenuZoom4Item:		scale = 4;	break;
			default:					scale = 1;	break;
		}

		SetScaleFactor(scale);
		break;

	  default:
	    break;
	}
    	handled = true;
	break;

    case penDownEvent:
		if (rfbsock <= 0)
		  	break;

		if (InKeyboardArea (e->screenX, e->screenY))
		{
			ProcessKeyboardPenDownEvent(e);
			handled = true;
			break;
		}
		else if (!InMainArea (e->screenX, e->screenY))
		{
			ProcessScrollPenDownEvent (e);
			handled = true;
	 		break;
		}

    case penUpEvent:
		if (rfbsock <= 0)
			  break;

		if (InKeyboardArea (e->screenX, e->screenY))
		{
			ProcessKeyboardPenUpEvent(e);
			handled = true;
		}
		else if (InMainArea (e->screenX, e->screenY))
		{
			ProcessPenUpDownEvent (e);
    	  	handled = true;
		}
		else
	  		buttonState = 0;

		break;
		
		
    case penMoveEvent:
		if (rfbsock <= 0)
		  break;
	
		if (buttonState == HORIZ || buttonState == VERT /*!InMainArea (e->screenX, e->screenY)*/) {
		  ProcessScrollPenMoveEvent (e);
		  handled = true;
	          break;
	        }
		else
	  	  if (!(InKeyboardArea (e->screenX, e->screenY)))
		  	ProcessPenMoveEvent (e);
	
	    handled = true;
		break;

    case keyDownEvent:
    
    /* Allow keyboard areas of Graffiti to activate special keys menu */
    if  ((e->data.keyDown.modifiers & commandKeyMask) &&
	    ((e->data.keyDown.chr == vchrKeyboard) || 
     	 (e->data.keyDown.chr == vchrKeyboardNumeric) ||
    	 (e->data.keyDown.chr == vchrKeyboardAlpha)))
    {
		if (rfbsock <= 0)
		  	break;
    	ToggleKeyboardState();
    	handled = true;
    	KeySetMask(~KeyCurrentState());
    	UpdateViewport();
    	break;
    }
   
	// the hardware key 1 (DateBook) used to emulate the middle mouse button
	if (KeyCurrentState() & keyBitHard1)
	{
	    currentMouseButton = 2;
	    handled = true;
  	    // Mask off the key to avoid repeat keys causing clicking sounds
	    KeySetMask(~KeyCurrentState());
	    break;
	}
	// the hardware key 2 (PhoneBook) used to emulate the right mouse button
	if (KeyCurrentState() & keyBitHard2)
	{
	    currentMouseButton = 4;
	    handled = true;
	    // Mask off the key to avoid repeat keys causing clicking sounds
	    KeySetMask(~KeyCurrentState());
	    break;
	}
	
	if (KeyCurrentState() & (keyBitHard3 | keyBitHard4 | keyBitPageUp | keyBitPageDown))
	{
		ProcessScrollKeyEvent(e);
		handled = true;
		break;
	}

    if ((!(e->data.keyDown.modifiers & commandKeyMask)) && IsAscii (e->data.keyDown.chr))
    {
		SendKeySequence (0, 0, 0, &(e->data.keyDown.chr), 1, false);
    }

    	handled = true;
	break;

    default:
        break;
    }

    CALLBACK_EPILOGUE

    return handled;
}


/******************************************************************
 * ApplicationHandleEvent
 *
 * loads the form and sets up event handlers
 *****************************************************************/ 
static Boolean ApplicationHandleEvent (EventPtr e)
{
    FormPtr frm;
    Word    formId;
    Boolean handled = false;

    if (e->eType == frmLoadEvent) {
	formId = e->data.frmLoad.formID;
	frm = FrmInitForm (formId);
	FrmSetActiveForm (frm);

	switch(formId) {
	case MainForm:
	    FrmSetEventHandler(frm, MainFormHandleEvent);
	    break;

	case ConsoleForm:
	    FrmSetEventHandler(frm, ConsoleFormHandleEvent);
	    break;

	case PrefForm:
	    FrmSetEventHandler(frm, PrefFormHandleEvent);
	    break;

	case PasswordForm:
		FrmSetEventHandler(frm, PasswordFormHandleEvent);
		break;
	}

	handled = true;
    }

    return handled;
}


/******************************************************************
 * OpenDatabase
 *
 * Opens the database and copies the saved preferences
 *****************************************************************/ 
 static DmOpenRef OpenDatabase (PalmVNCPreferences *pref)
{
  UInt index = 0;
  VoidHand RecHandle;
  VoidPtr RecPointer;
  DmOpenRef db;
  Word err, oldRecordSize;

  // Create database, if it doesn't exist, and save default status.
  if (!(db = DmOpenDatabaseByTypeCreator ('Data', CREATOR_ID, dmModeReadWrite))) {
    if ((err = DmCreateDatabase (0, DATABASE_NAME, CREATOR_ID, 'Data', false)))
      return 0;

    db = DmOpenDatabaseByTypeCreator ('Data', CREATOR_ID, dmModeReadWrite);

    RecHandle = DmNewRecord(db, &index, sizeof (PalmVNCPreferences));
    DmWrite(MemHandleLock (RecHandle), 0, pref, sizeof (PalmVNCPreferences));
    MemHandleUnlock (RecHandle);
    DmReleaseRecord (db, index, true);
  }

  // check the size
  RecHandle = DmGetRecord (db, 0);
  RecPointer = MemHandleLock (RecHandle);
  oldRecordSize = MemPtrSize (RecPointer);
  MemHandleUnlock (RecHandle);
  DmReleaseRecord(db, 0, true);
  if (oldRecordSize != sizeof (PalmVNCPreferences)) // it is not a record from the current version
  {
  	  printf("Old version detected.\n");
	  DmDeleteRecord(db, 0);
	  RecHandle = DmNewRecord(db, &index, sizeof (PalmVNCPreferences));
      DmWrite(MemHandleLock (RecHandle), 0, pref, sizeof (PalmVNCPreferences));
      MemHandleUnlock (RecHandle);
      DmReleaseRecord (db, index, true);
  }

  // Load a saved status.
  RecHandle = DmGetRecord (db, 0);
  RecPointer = MemHandleLock (RecHandle);
  MemSet (pref, sizeof (PalmVNCPreferences), (Byte)0);
  MemMove (pref, RecPointer, sizeof (PalmVNCPreferences));
  MemHandleUnlock (RecHandle);

  DmReleaseRecord (db, 0, true);

  return db;
}


/******************************************************************
 * SaveStatus
 *
 * Save preferences information.
 *****************************************************************/ 
static void SaveStatus (DmOpenRef db, PalmVNCPreferences *pref)
{
  VoidPtr p = MemHandleLock (DmGetRecord(db, 0));
  DmWrite (p, 0, pref, sizeof (PalmVNCPreferences));
  MemPtrUnlock (p);
  DmReleaseRecord (db, 0, true);
}


/******************************************************************
 * StartApplication
 *
 * Get preferences, open (or create) app database
 *****************************************************************/ 
static Word StartApplication (void)
{
  Err err;
  char *dummy_addr = "host.company.com";
  char *dummy_name = "username";
  char *dummy_passwd = "\0";

  // Init our terminal functions for printing to the screen
  // They allow to use printf () & Co. with the output to
  // a text field in the Console form
  StdInit (ConsoleForm, IOField, ConsoleScrollbar);

  // Get the refNum of the Net Library
  err = SysLibFind ("Net.lib", &AppNetRefnum);
  if (err) {
    ErrDisplay ("\nNet Library not found, exiting...");
    return err;
  }

  // Set default timeout
  AppNetTimeout = 10*sysTicksPerSecond;

  palmFB.viewable.x = palmFB.viewable.y = 0;
  palmFB.viewable.h = VIEWABLE_MAX_H - SCROLLBAR_SIZE;
  palmFB.viewable.w = VIEWABLE_MAX_W - SCROLLBAR_SIZE;
  palmFB.virtual.x = palmFB.virtual.y = 0;
  palmFB.zoomarea_ena = false;

	palmFB.virtual.h = VIRTUAL_30_MAX_H;
	palmFB.virtual.w = VIRTUAL_30_MAX_W;
	virtualBounds.extent.x = VIRTUAL_30_MAX_W;
	virtualBounds.extent.y = VIRTUAL_30_MAX_H;

  palmFB.remote.x = palmFB.remote.y = palmFB.remote.h = palmFB.remote.w = 0;

  // fill default values
  MemMove (pref.server.address, dummy_addr, StrLen (dummy_addr)+1);
  MemMove (pref.server.password, dummy_passwd, StrLen (dummy_passwd)+1);
  pref.server.port = 0;
  pref.server.beginX = 0;
  pref.server.beginY = 0;
  pref.server.scaleFactor = 1;
  pref.server.sharedDesktop = true;
  MemMove (pref.server.loginNameNT, dummy_name, StrLen (dummy_name)+1);
  MemMove (pref.server.passwNT, dummy_passwd, StrLen (dummy_passwd)+1);
  pref.screenDepth = 1;

  // get preferences which overwrite default values if DB exists
  if ((prefDB = OpenDatabase(&pref)) == 0)
    return dmErrCantOpen;

  pen_zooming = false;

  FrmGotoForm(MainForm);
  return 0;
}


/******************************************************************
 * StopApplication
 *
 * Save preferences, close forms, close app database
 *****************************************************************/ 
static void StopApplication (void)
{
  	FrmSaveAllForms();
/*  FrmCloseAllForms(); */
  	ShutdownNetwork (false);
  	StdFree();
  	SaveStatus(prefDB, &pref);
  	DmCloseDatabase(prefDB);

	if (screenDepth == 8)
	{
		WinPalette (winPaletteSetToDefault, 0, 0, 0);
	}
	
	ScrDisplayMode(scrDisplayModeSetToDefaults,
	        	NULL, NULL, NULL, NULL);

}


/******************************************************************
 * EventLoop
 *
 * The main event loop
 *****************************************************************/ 
static void EventLoop (void)
{
  Word err;
  EventType e;
  fd_set fds;
  long eventTimeout;
  Boolean passSystem = false;

  do {
    if (rfbsock)
      eventTimeout = 0; // we sleep in select
    else
      eventTimeout = evtWaitForever;

    EvtGetEvent(&e, eventTimeout);

    // we capture the DateBook and PhoneBook keys to emulate
    // the second and the third mouse buttons
    passSystem = e.eType == keyDownEvent &&
		(e.data.keyDown.chr == hard1Chr ||
		 e.data.keyDown.chr == hard2Chr ||
		 e.data.keyDown.chr == hard3Chr ||
		 e.data.keyDown.chr == hard4Chr ||
		 e.data.keyDown.chr == pageUpChr ||
		 e.data.keyDown.chr == pageDownChr ||
		 e.data.keyDown.chr == vchrKeyboard ||
		 e.data.keyDown.chr == vchrKeyboardAlpha ||
		 e.data.keyDown.chr == vchrKeyboardNumeric) ;

    if (passSystem) {
      if (!ApplicationHandleEvent (&e))
	FrmDispatchEvent (&e);
    }
    else {
      if (!SysHandleEvent (&e))
        if (!MenuHandleEvent (NULL, &e, &err))
	  if (!ApplicationHandleEvent (&e))
	    FrmDispatchEvent (&e);
    }

    if (rfbsock <= 0)
      continue;

    FD_ZERO (&fds);
    FD_SET (sysFileDescStdIn, &fds);      
    FD_SET (rfbsock,&fds);

	if (updateViewportSize) 
	{

	  WinHandle prevDrawWin = WinGetDrawWindow ();
	  WinSetDrawWindow (canvas);
	  WinEraseWindow ();
	  WinSetDrawWindow(prevDrawWin);

	  viewportX = palmFB.virtual.x;
	  viewportY = palmFB.virtual.y;

		/* set up the size of scrollbar slides */
		horizSlideRect.extent.x =
		(SWord)(((UInt32)palmFB.virtual.w * (UInt32)palmFB.virtual.w) /
			palmFB.remote.w);
		vertSlideRect.extent.y =
		(SWord)(((UInt32)palmFB.virtual.h * (UInt32)palmFB.virtual.h) /
			palmFB.remote.h);
		
		/* set the pen time tracing, used to catch double-taps */
		lastPenTime = TimGetTicks ();

	    (void)SendFramebufferUpdateRequest((int)palmFB.virtual.x,
		(int)palmFB.virtual.y,
		(int)palmFB.virtual.w,
		(int)palmFB.virtual.h,
		false);

		UpdateViewport();
		PositionScrollbars();
		updateViewportSize = false;
		sendUpdateRequest = false;
	}

    if (sendUpdateRequest) {
      if (!SendIncrementalFramebufferUpdateRequest())
	ShutdownConnection();
      sendUpdateRequest = false;
    }

// the BSD select() call cannot be compiled with GCC, use the native one 
//   if (select (FD_SETSIZE, &fds, NULL, NULL, NULL) < 0) {

    if (NetLibSelect (AppNetRefnum, rfbsock+1, &fds, NULL, NULL,
			AppNetTimeout, &errno) < 0) {
      printf ("select returns error\n");
      ShutdownConnection ();
    }

    if (FD_ISSET(rfbsock, &fds)) {
      if (!HandleRFBServerMessage()) {
        ShutdownConnection ();
	/* (void)FrmPopupForm (ConsoleForm); */
		PostProgressMessage ("Connection Lost!", 0, 0);
	  }
   	  UpdateViewport ();
    }
#ifdef DEBUG_FLOW
      else if (!FD_ISSET(sysFileDescStdIn, &fds))
	printf ("Select returned nothing, timeout?\n");
#endif

  } while (e.eType != appStopEvent);
}


/******************************************************************
 * PilotMain
 *
 *
 *****************************************************************/ 
DWord PilotMain (Word cmd, Ptr cmdPBP, Word launchFlags)
{
  Word err;
  DWord requiredDepth = 8;
  RGBColorType* col;
  int i;
  
  if (cmd == sysAppLaunchCmdNormalLaunch) {

    err = RomVersionCompatible (MIN_VERSION, launchFlags);
    if (err) return (err);

	/* First, try for 8-bit colour */
    err = ScrDisplayMode(scrDisplayModeSet,
            	NULL, NULL, &requiredDepth, NULL);
    if (err)
    {
	    /* Okay, try for 4-bit greyscale */
	    requiredDepth = 4;
	    err = ScrDisplayMode(scrDisplayModeSet,
	            	NULL, NULL, &requiredDepth, NULL);
	    if (err)
	    {
	    	/* Okay, let's try dropping back to 2-bit mode */
	    	requiredDepth = 2;
		    err = ScrDisplayMode(scrDisplayModeSet,
		            	NULL, NULL, &requiredDepth, NULL);
	    	if (err)
	    	{
	    	    FrmAlert (VersionAlert);
	      		return err;
	      	}
	      	else
	      	{
	      		screenDepth = 2;
	      	}
	    }
	    else
	    {
	    	screenDepth = 4;
	    }
	}
	else
	{
		screenDepth = 8;
	
		col = malloc(sizeof(RGBColorType) * 256);
		/* Set color palette to match BGR233 */
		for (i = 0; i < 256; i++)
		{
			col[i].index = 0;
			col[i].r = (((i >> 0) & 0x7) << 5) | (((i >> 0) & 0x7) << 2) | (((i >> 0) & 0x7) >> 1);
			col[i].g = (((i >> 3) & 0x7) << 5) | (((i >> 3) & 0x7) << 2) | (((i >> 3) & 0x7) >> 1);
			col[i].b = (((i >> 6) & 0x3) << 6) | (((i >> 6) & 0x3) << 4) | (((i >> 6) & 0x3) << 2) | ((i >> 6) & 0x3);
		}
		WinPalette (winPaletteSet, 0, 256, col);
		free(col);
	}
	
    err = StartApplication();
    if (err) return err;

    sendUpdateRequest = true;
    EventLoop();

    StopApplication();

  } else {
    return sysErrParamErr;
  }

  return 0;
}
