/*
 * $Id: Account.c,v 1.1 2000/08/20 15:33:21 jmanning Exp $
 *
 * Copyright (C) 2000 by Jackie Manning (jackie@linuxfan.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <PalmOS.h>
#include "Account.h"
#include "PCashRsc.h"
#include "Utils.h"
#include "DBUtils.h"

static PackedAccount *lAccount, *rAccount;

Err ReadAccount( DmOpenRef db, UInt16 recNo, Account *acct ) {

    MemHandle trnHandle;
    PackedAccount *packedAccount;
    
    trnHandle = DmQueryRecord( db, recNo );
    if( trnHandle == NULL ) {
        return DmGetLastErr();
    }
    packedAccount = MemHandleLock( trnHandle );
    UnpackAccount( acct, packedAccount );
    MemHandleUnlock( trnHandle );
    return 0;
}

UInt16 InsertAccount( Account *acct, DmOpenRef db ) {

    MemHandle newRecord;
    UInt16 newIndex = dmMaxRecordIndex;
    PackedAccount *packedAccount;
    Char buff[30];
    Err err;
    
    newRecord = DmNewRecord( db, &newIndex, getAccountSize( acct ) );
    PackAccount( acct, newRecord );
//    packedAccount = MemHandleLock( newRecord );
//    newIndex = DmFindSortPosition( db, packedAccount, NULL, CompareAccount, 0 );
//    MemHandleUnlock( newRecord );
//    err = DmAttachRecord( db, &newIndex, newRecord, NULL );
    DmReleaseRecord( db, newIndex, false );
    return newIndex;
}

Err UpdateAccount( Account *acct, UInt16 recNo, DmOpenRef db ) {

    MemHandle handle;
    Err err = 0;
    PackedAccount *packedAccount;
    
    handle = DmResizeRecord( db, recNo, getAccountSize( acct ) );
    if( handle == NULL ) {
        err = DmGetLastErr();
    }
    else {
        PackAccount( acct, handle );
        DmReleaseRecord( db, recNo, false );
    }
    return err;
}

PackedAccount *PackAccount( Account *acct, MemHandle acctDBEntry ) {

    UInt16      length = 0;
    void        *s;
    UInt32      offset = 0;

    length = getAccountSize( acct );
    // resize the VoidHand
    if (MemHandleResize(acctDBEntry, length) == 0) {
        // copy the fields
        s = MemHandleLock(acctDBEntry);

        offset = 0;
        WriteObject( s, &offset, &acct->beginningBalance, 
                sizeof(acct->beginningBalance) );
        WriteObject( s, &offset, &acct->lastCheckNo, 
                sizeof(acct->lastCheckNo) );
        WriteObject( s, &offset, acct->name, StrLen(acct->name) + 1 );

        MemHandleUnlock(acctDBEntry);
    }
    return (PackedAccount*)s;
}

UInt16 getAccountSize( Account *acct ) {

    UInt16 length;
    length = 0
            + sizeof( acct->beginningBalance ) 
            + sizeof(acct->lastCheckNo)
            + StrLen(acct->name) + 1
            ;
    return length;
}

void UnpackAccount( Account *acct, PackedAccount *packedAccount ) {

    acct->beginningBalance = packedAccount->beginningBalance;
    acct->lastCheckNo = packedAccount->lastCheckNo;
    acct->name = packedAccount->name;

}

// End -------------------------------------------------------------------------
