/*
 *  $Id: PCash.c,v 1.18 2000/11/21 03:04:19 jmanning Exp $
 *
 * PCash - PalmOS Checking and Budgeting program
 * Copyright (C) 2000 by Jackie Manning <jackie@linuxfan.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <PalmOS.h>
#include <stdarg.h>

#include "grid.h"
#include "PCashRsc.h"
#include "Trans.h"
#include "Account.h"

#define SELECT_ALL 0
#define SELECT_ACTUAL 1

#define GetObjectPtr(f,i) FrmGetObjectPtr((f),FrmGetObjectIndex((f),(i)))

#define     PREFS_BOTTOM_RECORD    (UInt16)1
#define     PREFS_TOP_RECORD    (UInt16)2
#define     PREFS_SORT_ORDER    (UInt16)3
#define     PREFS_ACCOUNT       (UInt16)4

Account     account;
UInt16      acctRec;
UInt16      bottomRecord = 0;
UInt16      topRecord = 0;
UInt16      prefsSize;

UInt16      dbCardNo;
LocalID     dbLocalID;
UInt16      dbVersion = 0;
DmOpenRef   gTransDB;
DmOpenRef   gAccountDB;

Boolean debug = false;
Boolean ovdebug = false;

static GridType grid;
static GridType aGrid;
UInt32 cSecs;
Int16 month;
Int16 day;
Int16 year;
DateType transDate;
extern GridModelType model;
extern GridModelType acctModel;
UInt16 indexCnt = 0;
//MemHandle idxHandle;
MemHandle totalsHandle;
Boolean selectAll = true;
//UInt16 selection = SELECT_ACTUAL;
double postedBal = 0.00;
double currBal = 0.00;
void UpdatePostedBalance( void );
static void RebuildTotals( Int16 fromRow );

static Err SetAccount( UInt16 acctRecNo ) {

    Boolean created = false;
    Err err;
    char dbName[32];


    acctRec = acctRecNo;
    err = ReadAccount( gAccountDB, acctRec, &account );
    StrPrintF( dbName, "PCSH-Tx%u", acctRec );
    gTransDB = NULL;
    err = OpenOrCreateDB( &gTransDB, 'Tran', 'PCSH', dmModeReadWrite, 0,
                          dbName, &created );
    PrefSetAppPreferences( 'PCSH', PREFS_ACCOUNT, 0,
        &acctRecNo, sizeof( acctRecNo ), true );
    return err;
}

void SetFieldText( FieldType *fld, const Char * text ) {
    
    MemHandle handle = MemHandleNew( FldGetMaxChars( fld ) );
    Char * fldText;
    MemHandle oldHdl = FldGetTextHandle( fld );
    
    
    fldText = MemHandleLock( handle );
    StrCopy( fldText, text );
    MemPtrUnlock( fldText );
    FldSetTextHandle( fld, handle );
    if( oldHdl )
        MemHandleFree( oldHdl );
}

void UpdatePostedBalance( void ) {
    
    Int16 pWidth;
    
    char buff[30];

    RebuildTotals( 0 );    
// This is quick and dirty -- needs work
    
    StrCopy( buff, "                " );
    WinDrawChars( buff, StrLen( buff ), 125, 0 );
    doubleToStr( buff, currBal, 2 );
    pWidth = FntCharsWidth( buff, StrLen( buff ) );
    WinDrawChars( buff, StrLen( buff ), 160-pWidth, 0 );
    
    StrCopy( buff, "                " );
    WinDrawChars( buff, StrLen( buff ), 35, 0 );
    doubleToStr( buff, postedBal, 2 );
    WinDrawChars( buff, StrLen( buff ), 35, 0 );
    
}

void DbgPrintF(const Char * fmt, ...) {

    Char buf[256];
    va_list args;

    va_start(args, fmt);
    StrVPrintF(buf, fmt, args);
    va_end(args);
//if( debug && ovdebug) 
    FrmCustomAlert( idDebugAlert, buf, " ", " ");
}

static void MainFormUpdateScrollers(Boolean scrollUp, Boolean scrollDown) {

    UInt16 upIndex, downIndex;
    FormPtr form;

    form = FrmGetActiveForm();
    upIndex = FrmGetObjectIndex(form, idMainUpButton);
    downIndex = FrmGetObjectIndex(form, idMainDownButton);
    FrmUpdateScrollers(form, upIndex, downIndex, scrollUp, scrollDown);
}

static void AcctUpdateScrollers(Boolean scrollUp, Boolean scrollDown) {

    UInt16 upIndex, downIndex;
    FormPtr form;

    form = FrmGetActiveForm();
    upIndex = FrmGetObjectIndex(form, idAcctUpBtn);
    downIndex = FrmGetObjectIndex(form, idAcctDownBtn);
    FrmUpdateScrollers(form, upIndex, downIndex, scrollUp, scrollDown);
}
/*
static void RebuildIndex( Int16 select ) {

    MemHandle handle;
    Trans trans;
    PackedTrans *packedTran;
    UInt32 *index;
    UInt16 recPtr;
    UInt16 num_recs;
    UInt32 recID;
    
    Err err;
    
    num_recs = DmNumRecords( gTransDB );
    err = MemHandleResize( idxHandle, num_recs * sizeof( UInt32 ) );
    
    index = MemHandleLock( idxHandle );
    
    indexCnt = 0;
    recPtr = 0;
    while( recPtr<num_recs ) {

        if( !selectAll ) {
            ReadTrans( gTransDB, recPtr, &trans );
            if( trans.budget ) {
                recPtr++;
                continue;
            }
        }
        DmRecordInfo( gTransDB, recPtr, NULL, &recID, NULL );
        index[recPtr] = recID;
        indexCnt++;
        recPtr++;
    }
    
    MemHandleUnlock( idxHandle );
}

static void DeleteIndex( UInt16 index ) {

    UInt16 numRecs;
    UInt16 dbRecs;
    UInt32 *indexes;
    Err err;
    
    dbRecs = DmNumRecords( gTransDB );
    if( dbRecs < 1 ) {
        numRecs = 50;
    }
    else {
        numRecs = dbRecs;
    }
    indexes = MemHandleLock( idxHandle );
    if( DmNumRecords( gTransDB ) > 0 ) {
        MemMove( &(indexes[index]), &(indexes[index+1]), 
            ( dbRecs - index ) * sizeof( UInt32 ) );
    }
    DmRecordInfo( gTransDB, index, NULL, &(indexes[index]), NULL );
    MemHandleUnlock( idxHandle );
}

static void InsertIndex( UInt16 index ) {

    UInt16 numRecs;
    UInt16 dbRecs;
    UInt32 *indexes;
    Err err;
    
    dbRecs = DmNumRecords( gTransDB );
    if( dbRecs < 1 ) {
        numRecs = 50;
    }
    else {
        numRecs = dbRecs;
    }
    err = MemHandleResize( idxHandle, numRecs * sizeof( UInt32 ) );
    indexes = MemHandleLock( idxHandle );
    if( DmNumRecords( gTransDB ) > 0 ) {
        MemMove( &(indexes[index+1]), &(indexes[index]), 
            ( dbRecs - index - 1) * sizeof( UInt32 ) );
    }
    DmRecordInfo( gTransDB, index, NULL, &(indexes[index]), NULL );
    MemHandleUnlock( idxHandle );
    indexCnt++;
}

static void InitIndex( void ) {

    UInt16 dbSize;
    
    dbSize = DmNumRecords( gTransDB );
    if( dbSize < 1 ) {
        dbSize = 50;
    }
    idxHandle = MemHandleNew( dbSize * sizeof( UInt32 ) );
    RebuildIndex( 0 );
}
*/
static void RebuildTotals( Int16 fromRow ) {

    Trans trans;
    double *totals;
    double total;
    UInt16 recPtr;
    UInt16 recNo;
    UInt16 num_recs;
    DateType currDate;
    UInt32 currDays;
//    UInt32 *indexes;
    
    Err err;

    DateSecondsToDate( TimGetSeconds(), &currDate );
    currDays = DateToDays( currDate );
//    indexes = MemHandleLock( idxHandle );    
    num_recs = DmNumRecords( gTransDB );
    err = MemHandleResize( totalsHandle, num_recs * sizeof( double ) );
    
    totals = MemHandleLock( totalsHandle );
    if( fromRow > 0 ) {
        total = totals[ fromRow - 1 ];
    }
    else {
        total = account.beginningBalance;
        postedBal = total;
        currBal = total;
    }
    recPtr = fromRow;
    while( recPtr<num_recs ) {
    
//        DmFindRecordByID( gTransDB, indexes[recPtr], &recNo );
//        ReadTrans( gTransDB, recNo, &trans );
        ReadTrans( gTransDB, recPtr, &trans );
        if( selectAll || trans.budget == ACTUAL ) {
            total += trans.amount;
            if( fromRow == 0 ) {
                if( trans.budget == ACTUAL ) {
                    currBal += trans.amount;
                }
                if( trans.posted ) {
                    postedBal += trans.amount;
                }
            }
        }
        totals[recPtr] = total;
        recPtr++;
    }
    
    MemHandleUnlock( totalsHandle );
//    MemHandleUnlock( idxHandle );
}

static void InitTotals( void ) {

    UInt16 dbSize;
    
    dbSize = DmNumRecords( gTransDB );
    if( dbSize < 1 ) {
        dbSize = 50;
    }
    totalsHandle = MemHandleNew( dbSize * sizeof( double ) );
    RebuildTotals( 0 );
}

static void LoadFormFromTrans( FormPtr frm, Trans *trans ) {

//    FormPtr form = frm;
//    FormPtr depFrm = frm;
    FieldType *fld;
    UInt16 i;
    ControlPtr ctrl;
//    ListType *lst;
    char dateStr[ dateStringLength ];
    MemHandle fldHandle;
    PackedTrans *packedTrans;
    MemHandle trnHandle;
    MemHandle amtHandle;
    Char buff[80];

    transDate = trans->date;

    month = transDate.month;
    day = transDate.day;
    year = transDate.year + firstYear;
    DateToAscii( month, day, year, dfMDYWithSlashes, dateStr );
    ctrl = GetObjectPtr( frm, idEditDateSel );
    CtlSetLabel( ctrl, dateStr );

    switch( trans->transType ) {
    
    case DEPOSIT_TRANS:
        ctrl = GetObjectPtr( frm, idEditDepositPBtn );
        break;
        
    case WITHDRAWAL_TRANS:
        ctrl = GetObjectPtr( frm, idEditWithdrawalPBtn );
        break;
        
    default:
        ctrl = GetObjectPtr( frm, idEditCheckPBtn );
        break;
        
    }
    CtlSetValue( ctrl, 1 );
    if( trans->transType == DEPOSIT_TRANS ) {
        doubleToStr( buff, trans->amount, 2 );
    }
    else {
        doubleToStr( buff, -trans->amount, 2 );
    }
    SetFieldText( GetObjectPtr( frm, idEditAmountFld ),buff );
    SetFieldText( GetObjectPtr( frm, idEditPartyFld ), trans->party );
    SetFieldText( GetObjectPtr( frm, idEditNumFld ), trans->ref );
    SetFieldText( GetObjectPtr( frm, idEditNoteFld ), trans->note );
//    ctrl = GetObjectPtr( frm, idEditPostedChk );
//    CtlSetValue( ctrl, trans->posted );
//    ctrl = GetObjectPtr( frm, idEditBudgetChk );
//    CtlSetValue( ctrl, trans->budget );
    
    if( trans->posted ) {
        ctrl = GetObjectPtr( frm, idEditPostedPBtn );
    }
    else {
        switch( trans->budget ) {
        
        case PAYABLE:
            ctrl = GetObjectPtr( frm, idEditPayablePBtn );
            break;
        case ESTIMATED:
            ctrl = GetObjectPtr( frm, idEditBudgetPBtn );
            break;

        default:
            ctrl = GetObjectPtr( frm, idEditActualPBtn );
            break;
        }
        
    }
    CtlSetValue( ctrl, 1 );

}

static void LoadTransFromForm( FormPtr frm, UInt16 frmBase, Trans *trans ) {

    char emptyStr[] = " ";
    char wrkStr[30];
    char *str;
    Int16 i;
    FieldPtr fld;
    UInt16 newRow;
//    ListPtr typeLst;
    ControlPtr ctrl;
    ControlPtr chk;
    Int16 sgn = 1;
    short chkVal;
    double dbl;

    ctrl = GetObjectPtr( frm, frmBase + ID_TypeDeposit );
    if( CtlGetValue( ctrl ) == 1 ) {
        trans->transType = 'D';
    }
    else {
        sgn = -1;
        ctrl = GetObjectPtr( frm, frmBase + ID_TypeWithdrawal );
        if( CtlGetValue( ctrl ) == 1 ) {
            trans->transType = 'W';
        }
        else {
            trans->transType = 'C';
        }
    }
    str = NULL;
    fld = GetObjectPtr( frm, frmBase + ID_AmountFld );
    str = FldGetTextPtr( fld );
    if( str ) {
        StringToDouble( str, &trans->amount );
        trans->amount *= sgn;
    } else {
        trans->amount = 0;
    }
    
    ctrl = GetObjectPtr( frm, frmBase + ID_StatusPosted );
    if( CtlGetValue( ctrl ) == 1 ) {
        trans->posted = true;
        trans->budget = ACTUAL;
    }
    else {
        trans->posted = false;
        ctrl = GetObjectPtr( frm, frmBase + ID_StatusActual );
        if( CtlGetValue( ctrl ) == 1 ) {
            trans->budget = ACTUAL;
        }
        else {
            ctrl = GetObjectPtr( frm, frmBase + ID_StatusPayable );
            if( CtlGetValue( ctrl ) == 1 ) {
                trans->budget = PAYABLE;
            }
            else {
                trans->budget = ESTIMATED;
            }
        }
    }
/*    chk = GetObjectPtr( frm, frmBase + ID_PostedChk );
    chkVal = CtlGetValue( chk );
    if( chkVal == 0 ) {
        trans->posted = false;
    } else {
        trans->posted = true;
    }
    
    chk = GetObjectPtr( frm, frmBase + ID_BudgetChk );
    chkVal = CtlGetValue( chk );
    if( chkVal == 0 ) {
        trans->budget = false;
    } else {
        trans->budget = true;
    }
*/
    fld = GetObjectPtr( frm, frmBase + ID_PartyFld );
    str = FldGetTextPtr( fld );
    if( str ) {
        trans->party = str;
    } else {
        trans->party = emptyStr;
    }

    trans->date = transDate;

    fld = GetObjectPtr( frm, frmBase + ID_NoteFld );
    str = FldGetTextPtr( fld );
    if( str ) {
        trans->note = str;
    } else {
        trans->note = emptyStr;
    }

    fld = GetObjectPtr( frm, frmBase + ID_NumFld );
    str = FldGetTextPtr( fld );
    if( str ) {
        trans->ref = str;
    } else {
        trans->ref = emptyStr;
    }

}

static UInt16 CreateTrans( FormPtr frm, UInt16 frmBase ) {

    Trans trans;
    UInt16 newRow;
    Int32 checkNo = 0;

    LoadTransFromForm( frm, frmBase, &trans );
    newRow = InsertTrans( &trans, gTransDB );
    if( trans.transType == 'C' ) {
        checkNo = StrAToI( trans.ref );
        account.lastCheckNo = checkNo;
        UpdateAccount( &account, acctRec, gAccountDB );
    }
    GridMakeIndexVisible( &grid, newRow );
    return newRow;
}

static UInt16 UpdateTransaction( FormPtr frm, UInt16 frmBase, UInt16 newRow ) {

    Trans trans;
    UInt16 nRow;
    
    LoadTransFromForm( frm, frmBase, &trans );
    UpdateTrans( &trans, newRow, gTransDB, &nRow );
    GridMakeIndexVisible( &grid, newRow);
    return nRow;

}

static Boolean TransFormHandleEvent (EventPtr e) {

    char buff[ 80 ];
    Boolean handled = false;
    FormPtr frm;
    ControlPtr lst;
    int checkNo = 0;

//    CALLBACK_PROLOGUE

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        break;

    case menuEvent:
        MenuEraseStatus(NULL);

        switch(e->data.menu.itemID) {}


        handled = true;
        break;

    case ctlSelectEvent:
        frm = FrmGetActiveForm();
        switch(e->data.ctlSelect.controlID) {
        
        case idTrnDateSel:

            month = transDate.month;
            day = transDate.day;
            year = transDate.year + firstYear;
            SelectDay ( selectDayByDay, &month, &day, &year, "Select date");
            DateToAscii( month, day, year, dfMDYWithSlashes, buff );
            lst = GetObjectPtr( frm, idTrnDateSel );
            CtlSetLabel( lst, buff );
            transDate.day = day;
            transDate.month = month;
            transDate.year = year - firstYear;
            handled = true;
            break;
            
        case idTrnCheckPBtn:
            checkNo = account.lastCheckNo + 1;
            StrPrintF( buff, "%u", checkNo );
            SetFieldText( GetObjectPtr( frm, idTrnNumFld ),buff );
            FldSetDirty( GetObjectPtr( frm, idTrnNumFld ), true );
            FldDrawField( GetObjectPtr( frm, idTrnNumFld ) );
            handled = true;
            break;
        }
        break;

    default:
        break;
    }

//    CALLBACK_EPILOGUE

    return handled;
}

static Boolean EditFormHandleEvent (EventPtr e) {

    char dateStr[ dateStringLength ];
    Boolean handled = false;
    FormPtr frm;
    ControlPtr lst;

//    CALLBACK_PROLOGUE

    switch (e->eType) {
    case frmOpenEvent:
        frm = FrmGetActiveForm();
        break;

    case menuEvent:
        MenuEraseStatus(NULL);

        switch(e->data.menu.itemID) {}


        handled = true;
        break;

    case ctlSelectEvent:
        frm = FrmGetActiveForm();
        switch(e->data.ctlSelect.controlID) {
        
        case idEditDateSel:

            month = transDate.month;
            day = transDate.day;
            year = transDate.year + firstYear;
            SelectDay ( selectDayByDay, &month, &day, &year, "Select date");
            DateToAscii( month, day, year, dfMDYWithSlashes, dateStr );
            lst = GetObjectPtr( frm, idEditDateSel );
            CtlSetLabel( lst, dateStr );
            transDate.day = day;
            transDate.month = month;
            transDate.year = year - firstYear;

            break;
            
        case idEditCopyBtn:
            frm = FrmGetActiveForm();
            CreateTrans( frm, idEditTransForm );
            handled = true;
            break;
        }
        break;

    default:
        break;
    }

//    CALLBACK_EPILOGUE

    return handled;
}

void SortTransactions( UInt16 aSortOrder ) {

    UInt16 i;
    Err err;
    
    SetTransSortOrder( aSortOrder );
    err = DmQuickSort( gTransDB, CompareTrans, 0 );
    RebuildTotals( 0 );
    for (i = 0; i < grid.numScreenRows; i++) {
        grid.screenRows[i].redraw = true;
    }
    GridMakeIndexVisible( &grid, 0 );
    GridDraw(&grid);
}

Boolean AcctFormHandleEvent(EventPtr event) {

    Account newAcct;
    char *str;
    FormPtr form;
    FormPtr depFrm;
    FieldType *fld;
    UInt16 i;
    ControlPtr ctrl;
    ListType *lst;
    char dateStr[ dateStringLength ];
    MemHandle fldHandle;
    Trans trans;
    PackedTrans *packedTrans;
    MemHandle trnHandle;
    MemHandle amtHandle;
    Char buff[80];
    UInt16 newRow;
    UInt32 cellID;
    Err err;

    switch (event->eType) {
    case frmOpenEvent:
	form = FrmGetActiveForm();
        ctrl = GetObjectPtr( form, idAcctOpenPBtn );
        CtlSetValue( ctrl, 1 );

	GridInit(&aGrid, form, idAcctGridGadget,
		 idAcctGridCheckbox, &acctModel, 0);
	FrmDrawForm(form);
        GridSetHeaderLook(&aGrid, false, false, false, boldFont);
        GridSetColumnLook(&aGrid, 0);
	GridDraw(&aGrid);
	return true;

    case frmCloseEvent:
	GridFree(&aGrid);
	break;
/*
    case menuEvent:
        switch ( event->data.menu.itemID ) {
        
        case idMainMenuGotoTop:
            GridMakeIndexVisible(&aGrid,0);
            GridDraw(&aGrid);
            return true;
            
        case idMainMenuGotoBottom:
            i = DmNumRecords( gTransDB );
            GridMakeIndexVisible(&aGrid,i-1);
            GridDraw(&aGrid);
            return true;
            
        case idMainMenuAbout:
            FrmAlert( idAboutAlert );
            GridDraw(&aGrid);
            return true;
            
        case idMainMenuSortDate:
            SortTransactions( SORT_BY_DATE );
            return true;
            
        case idMainMenuSortAmount:
            SortTransactions( SORT_BY_AMOUNT );
            return true;
            
        case idMainMenuSortParty:
            SortTransactions( SORT_BY_PARTY );
            return true;
            
        }
        break;
*/        
    case penDownEvent:
	if (GridHandleEvent(&aGrid, event))
	    return true;
	break;

    case gridSelectEvent:

	form = FrmGetActiveForm();
        ctrl = GetObjectPtr( form, idAcctOpenPBtn );
        if( CtlGetValue( ctrl ) ) {
            DmCloseDatabase( gTransDB );    
            err = SetAccount( event->data.tblSelect.row );
            if( err ) {
                DbgPrintF( "Error setaccount %u", err );
            }
            else {
                RebuildTotals( 0 );
                FrmGotoForm( idMainForm );
            }
        }
        else {
            err = ReadAccount( gAccountDB, event->data.tblSelect.row, 
                    &newAcct );
            depFrm = FrmInitForm( idNewAcctForm );
            FrmSetTitle( depFrm, "Edit Account" );
            FrmSetActiveForm( depFrm );
            SetFieldText( GetObjectPtr( depFrm, idNewAcctNameFld ), 
                newAcct.name );
            doubleToStr( buff, newAcct.beginningBalance, 2 );
            SetFieldText( GetObjectPtr( depFrm, idNewAcctBalFld ),buff );
            i = FrmDoDialog( depFrm );
            if( i == idNewAcctDoneBtn ) {
                str = NULL;
                fld = GetObjectPtr( depFrm, idNewAcctNameFld );
                str = FldGetTextPtr( fld );
                if( str ) {
                    newAcct.name = str;
                    str = FldGetTextPtr( GetObjectPtr( depFrm, idNewAcctBalFld ) );
                    if( str ) {
                        StringToDouble( str, &(newAcct.beginningBalance ) );
                    } else {
                        newAcct.beginningBalance = 0.0;
                    }
                    UpdateAccount( &newAcct, 
                            event->data.tblSelect.row, gAccountDB );
                }
                for (i = 0; i < aGrid.numScreenRows; i++) {
                    aGrid.screenRows[i].redraw = true;
                }
            }
            FrmSetActiveForm( form );
            FrmDeleteForm( depFrm );
            GridDraw(&aGrid);
        }
	return true;

/*
        form = FrmGetActiveForm();
        depFrm = FrmInitForm( idEditTransForm );
        FrmSetActiveForm( depFrm );
        FrmSetEventHandler( depFrm, &EditFormHandleEvent );
        err = ReadTrans( gTransDB, event->data.tblSelect.row, &trans );
        LoadFormFromTrans( depFrm, &trans );        
        i = FrmDoDialog( depFrm );
        if( i == idEditDoneBtn ) {
            newRow = UpdateTransaction( 
                    depFrm, idEditTransForm, event->data.tblSelect.row );
        }
        else if( i == idEditDeleteBtn ) {
            newRow = event->data.tblSelect.row - 1;
            if( newRow < 0 ) newRow = 0;
            DmDetachRecord( gTransDB, event->data.tblSelect.row, &fldHandle );
            MemHandleFree( fldHandle );
        }
        else if( i == idEditCopyBtn ) {
            newRow = CreateTrans( depFrm, idEditTransForm );
        }
        FrmSetActiveForm( form );
        FrmDeleteForm( depFrm );
        if( i != idEditCancelBtn ) {
            for (i = 0; i < aGrid.numScreenRows; i++) {
                aGrid.screenRows[i].redraw = true;
            }
            RebuildTotals( 0 );
            GridMakeIndexVisible( &aGrid, newRow );
        }
        GridDraw(&aGrid);
        UpdatePostedBalance();
*/

    case gridUpdateVScrollersEvent:
	AcctUpdateScrollers(event->data.generic.datum[1],
				event->data.generic.datum[2]);
	return true;
        
    case ctlSelectEvent:
    
        switch(event->data.ctlSelect.controlID) {
        
        case idAcctNewBtn:
            form = FrmGetActiveForm();
            depFrm = FrmInitForm( idNewAcctForm );
            FrmSetActiveForm( depFrm );
            SetFieldText( GetObjectPtr( depFrm, idNewAcctNameFld ), "" );
            i = FrmDoDialog( depFrm );
            if( i == idNewAcctDoneBtn ) {
                newAcct.beginningBalance = 0.0;
                newAcct.lastCheckNo = 0;
                str = NULL;
                fld = GetObjectPtr( depFrm, idNewAcctNameFld );
                str = FldGetTextPtr( fld );
                if( str ) {
                    newAcct.name = str;
                    str = FldGetTextPtr( GetObjectPtr( depFrm, idNewAcctBalFld ) );
                    if( str ) {
                        StringToDouble( str, &(newAcct.beginningBalance ) );
                    } else {
                        newAcct.beginningBalance = 0.0;
                    }
                    InsertAccount( &newAcct, gAccountDB );
                }
            }
            FrmSetActiveForm( form );
            FrmDeleteForm( depFrm );
/*            for (i = 0; i < aGrid.numScreenRows; i++) {
                aGrid.screenRows[i].redraw = true;
            }
*/
            GridDraw(&aGrid);

            return true;

/*        
        case idMainNewButton:
        
            form = FrmGetActiveForm();
            depFrm = FrmInitForm( idNewTransForm );
            FrmSetActiveForm( depFrm );
            FrmSetEventHandler( depFrm, &TransFormHandleEvent );
            cSecs = TimGetSeconds();
            DateSecondsToDate( cSecs, (DatePtr) &transDate );
            month = transDate.month;
            day = transDate.day;
            year = transDate.year + firstYear;
            DateToAscii( month, day, year, dfMDYWithSlashes, dateStr );
            ctrl = GetObjectPtr( depFrm, idTrnDateSel );
            CtlSetLabel( ctrl, dateStr );
            ctrl = GetObjectPtr( depFrm, idTrnWithdrawalPBtn );
            CtlSetValue( ctrl, 1 );
            SetFieldText( GetObjectPtr( depFrm, idTrnPartyFld ), "" );
            SetFieldText( GetObjectPtr( depFrm, idTrnNumFld ), "" );
            SetFieldText( GetObjectPtr( depFrm, idTrnNoteFld ), "" );

            i = FrmDoDialog( depFrm );
            if( i == idTrnDoneBtn ) {
                newRow = CreateTrans( depFrm, idNewTransForm );
                RebuildTotals( 0 );
            }
            FrmSetActiveForm( form );
            FrmDeleteForm( depFrm );
            for (i = 0; i < aGrid.numScreenRows; i++) {
                aGrid.screenRows[i].redraw = true;
            }
            GridDraw(&aGrid);
            UpdatePostedBalance();
            return true;
*/
        }
        break;

    case ctlRepeatEvent:
	switch (event->data.ctlRepeat.controlID) {
	case idAcctUpBtn:
	    GridScrollVertical(&aGrid, winUp, aGrid.numScreenRows );
	    break;

	case idAcctDownBtn:
	    GridScrollVertical(&aGrid, winDown, aGrid.numScreenRows );
	    break;

	}
	break;

    case keyDownEvent:
	switch (event->data.keyDown.chr) {
	case pageUpChr:
	    GridScrollVertical(&aGrid, winUp, GridGetNumberOfRows(&aGrid));
	    return true;

	case pageDownChr:
	    GridScrollVertical(&aGrid, winDown, GridGetNumberOfRows(&aGrid));
	    return true;
	}
	break;

    }
    return false;
}

Boolean MainFormHandleEvent(EventPtr event) {

    FormPtr form;
    FormPtr depFrm;
    FieldType *fld;
    UInt16 i;
    ControlPtr ctrl;
    ListType *lst;
    char dateStr[ dateStringLength ];
    MemHandle fldHandle;
    Trans trans;
    PackedTrans *packedTrans;
    MemHandle trnHandle;
    MemHandle amtHandle;
    Char buff[80];
    UInt16 newRow;
    UInt32 cellID;
    Err err;

    switch (event->eType) {
    case frmOpenEvent:
	form = FrmGetActiveForm();

	GridInit(&grid, form, idMainGridGadget,
		 idMainGridCheckbox, &model, 0);
	FrmDrawForm(form);
        ctrl = GetObjectPtr( form, idMainAcctSel );
        CtlSetLabel( ctrl, account.name );
        GridSetHeaderLook(&grid, true, 0, true, boldFont);
        GridSetColumnLook(&grid, 0);
        GridMakeIndexVisible( &grid, bottomRecord );
        GridMakeIndexVisible( &grid, topRecord );
	GridDraw(&grid);
        UpdatePostedBalance();
	return true;

    case frmCloseEvent:
	GridFree(&grid);
	break;

    case menuEvent:
        switch ( event->data.menu.itemID ) {
        
        case idMainMenuAccounts:
            FrmGotoForm( idAcctForm );
            return true;
            
        case idMainMenuGotoTop:
            GridMakeIndexVisible(&grid,0);
            GridDraw(&grid);
            return true;
            
        case idMainMenuGotoBottom:
            i = DmNumRecords( gTransDB );
            GridMakeIndexVisible(&grid,i-1);
            GridDraw(&grid);
            return true;
            
        case idMainMenuAbout:
            FrmAlert( idAboutAlert );
            GridDraw(&grid);
            return true;
            
        case idMainMenuSortDate:
            SortTransactions( SORT_BY_DATE );
            return true;
            
        case idMainMenuSortAmount:
            SortTransactions( SORT_BY_AMOUNT );
            return true;
            
        case idMainMenuSortParty:
            SortTransactions( SORT_BY_PARTY );
            return true;
            
        }
        break;
        
    case penDownEvent:
	if (GridHandleEvent(&grid, event))
	    return true;
	break;

    case gridSelectEvent:

        form = FrmGetActiveForm();
        depFrm = FrmInitForm( idEditTransForm );
        FrmSetActiveForm( depFrm );
        FrmSetEventHandler( depFrm, &EditFormHandleEvent );
        err = ReadTrans( gTransDB, event->data.tblSelect.row, &trans );
        LoadFormFromTrans( depFrm, &trans );        
        i = FrmDoDialog( depFrm );
        if( i == idEditDoneBtn ) {
            newRow = UpdateTransaction( 
                    depFrm, idEditTransForm, event->data.tblSelect.row );
        }
        else if( i == idEditDeleteBtn ) {
            newRow = event->data.tblSelect.row - 1;
            if( newRow < 0 ) newRow = 0;
            DmDetachRecord( gTransDB, event->data.tblSelect.row, &fldHandle );
            MemHandleFree( fldHandle );
        }
        else if( i == idEditCopyBtn ) {
            newRow = CreateTrans( depFrm, idEditTransForm );
        }
/*        if( newRow >= DmNumRecords( gTransDB ) ) {
            newRow = DmNumRecords( gTransDB ) - 1;
        }
*/
        FrmSetActiveForm( form );
        FrmDeleteForm( depFrm );
        if( i != idEditCancelBtn ) {
            for (i = 0; i < grid.numScreenRows; i++) {
                grid.screenRows[i].redraw = true;
            }
            RebuildTotals( 0 );
            GridMakeIndexVisible( &grid, newRow );
        }
        GridDraw(&grid);
        UpdatePostedBalance();
	return true;

    case gridUpdateVScrollersEvent:
	MainFormUpdateScrollers(event->data.generic.datum[1],
				event->data.generic.datum[2]);
	return true;
        
    case ctlSelectEvent:
    
        switch(event->data.ctlSelect.controlID) {
        
        case idMainAcctSel:
            FrmGotoForm( idAcctForm );
            return true;
            
        case idMainNewButton:
        
            form = FrmGetActiveForm();
            depFrm = FrmInitForm( idNewTransForm );
            FrmSetActiveForm( depFrm );
            FrmSetEventHandler( depFrm, &TransFormHandleEvent );
            cSecs = TimGetSeconds();
            DateSecondsToDate( cSecs, (DatePtr) &transDate );
            month = transDate.month;
            day = transDate.day;
            year = transDate.year + firstYear;
            DateToAscii( month, day, year, dfMDYWithSlashes, dateStr );
            ctrl = GetObjectPtr( depFrm, idTrnDateSel );
            CtlSetLabel( ctrl, dateStr );
            ctrl = GetObjectPtr( depFrm, idTrnWithdrawalPBtn );
            CtlSetValue( ctrl, 1 );
            ctrl = GetObjectPtr( depFrm, idTrnActualPBtn );
            CtlSetValue( ctrl, 1 );
            SetFieldText( GetObjectPtr( depFrm, idTrnPartyFld ), "" );
            SetFieldText( GetObjectPtr( depFrm, idTrnNumFld ), "" );
            SetFieldText( GetObjectPtr( depFrm, idTrnNoteFld ), "" );

            i = FrmDoDialog( depFrm );
            if( i == idTrnDoneBtn ) {
                newRow = CreateTrans( depFrm, idNewTransForm );
                RebuildTotals( 0 );
            }
            FrmSetActiveForm( form );
            FrmDeleteForm( depFrm );
            for (i = 0; i < grid.numScreenRows; i++) {
                grid.screenRows[i].redraw = true;
            }
            GridDraw(&grid);
            UpdatePostedBalance();
            return true;
        }
        break;

    case ctlRepeatEvent:
	switch (event->data.ctlRepeat.controlID) {
	case idMainUpButton:
	    GridScrollVertical(&grid, winUp, grid.numScreenRows );
	    break;

	case idMainDownButton:
	    GridScrollVertical(&grid, winDown, grid.numScreenRows );
	    break;

	}
	break;

    case keyDownEvent:
	switch (event->data.keyDown.chr) {
	case pageUpChr:
	    GridScrollVertical(&grid, winUp, GridGetNumberOfRows(&grid));
	    return true;

	case pageDownChr:
	    GridScrollVertical(&grid, winDown, GridGetNumberOfRows(&grid));
	    return true;
	}
	break;

    }

    return false;
}

Boolean AppHandleEvent(EventPtr event) {

    UInt16 formID;
    FormPtr form;

    if (event->eType == frmLoadEvent) {
        formID = event->data.frmLoad.formID;
        form = FrmInitForm(formID);
        FrmSetActiveForm(form);

        switch (formID) {
	case idMainForm:
	    FrmSetEventHandler(form, MainFormHandleEvent);
	    break;

	case idAcctForm:
	    FrmSetEventHandler(form, AcctFormHandleEvent);
	    break;

	default:
            ErrDisplay("unknown form ID");
            return false;
	}

	return true;
    }

    return false;
}

void EventLoop(void) {

    Err err;
    EventType event;

    do {
        EvtGetEvent(&event, evtWaitForever);

        if (!SysHandleEvent(&event))
            if (!MenuHandleEvent( NULL, &event, &err))
                if (!AppHandleEvent(&event))
                    FrmDispatchEvent(&event);
    } while (event.eType != appStopEvent);
}

static void InitAccounts( void ) {

    DmSearchStateType searchState;
    Err         err;
    MemHandle   nameHdl;
    Boolean     newSearch;
    UInt16      cardNo;
    LocalID     locID;
    char *newAcctName;
    char newName[32];
    char dbName[32];
    
    account.beginningBalance = 0.0;
    account.lastCheckNo = 0;
    nameHdl = MemHandleNew( 4 );
    newAcctName = MemHandleLock( nameHdl );
    StrCopy( newAcctName, "New" );
    account.name = newAcctName;
    acctRec = InsertAccount( &account, gAccountDB );
    MemHandleUnlock( nameHdl );
    MemHandleFree( nameHdl );
    PrefSetAppPreferences( 'PCSH', PREFS_ACCOUNT, 0,
        &acctRec, sizeof( acctRec ), true );
    StrPrintF( newName, "PCSH-Tx%u", acctRec );
    err = DmGetNextDatabaseByTypeCreator( true, &searchState,
            'Tran', 'PCSH', true, &cardNo, &locID );
    while( err != dmErrCantFind ) {
        DmDatabaseInfo( cardNo, locID, dbName, NULL, NULL, NULL, 
                NULL, NULL, NULL, NULL, NULL, NULL, NULL );
        if( StrCompare( dbName, newName ) == 0 ) {
            return;
        }
        if( StrCompare( dbName, "PCSH-Trans" ) == 0 ) {
            DmSetDatabaseInfo( cardNo, locID, newName, NULL, NULL, NULL, 
                    NULL, NULL, NULL, NULL, NULL, NULL, NULL );
            return;
        }
        err = DmGetNextDatabaseByTypeCreator( false, &searchState,
                'Tran', 'PCSH', true, &cardNo, &locID );
    }
}

static Err StartApplication(void) {

    Boolean created = false;
    UInt32        mode = dmModeReadWrite;
    Int16       prefRes;
    UInt16      acctRec;
    Err err;

    err = OpenOrCreateDB( &gAccountDB, 'Acct', 'PCSH', dmModeReadWrite, 0,
                          "PCSH-Acct", &created );
    if( err ) {
        return err;
    }
    if( created ) {
        InitAccounts();
    }
    prefsSize = sizeof( acctRec );
    prefRes = PrefGetAppPreferences( 'PCSH', PREFS_ACCOUNT,
                            &acctRec, &prefsSize, true );
    if( prefRes == noPreferenceFound ) {
        acctRec = 0;
    }
    err = SetAccount( acctRec );
    if( err )
        return err;
    prefsSize = sizeof( sortOrder );
    prefRes = PrefGetAppPreferences( 'PCSH', PREFS_SORT_ORDER,
                            &sortOrder, &prefsSize, true );
    if( prefRes == noPreferenceFound ) {
        sortOrder = SORT_BY_DATE;
    }
    SetTransSortOrder( sortOrder );
//    err = DmQuickSort( gTransDB, CompareTrans, 0 );
//    if( err )
//        return err;
    err = DmOpenDatabaseInfo( gTransDB, &dbLocalID, NULL, NULL, &dbCardNo, NULL );
    if( err )
        return err;
    err = DmDatabaseInfo( dbCardNo, dbLocalID, 
            NULL, NULL, &dbVersion, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL );
    if( err )
        return err;
//    InitIndex();
    InitTotals();
    prefsSize = sizeof( bottomRecord );
    prefRes = PrefGetAppPreferences( 'PCSH', PREFS_BOTTOM_RECORD,
                            &bottomRecord, &prefsSize, true );
    if( prefRes == noPreferenceFound ) {
        bottomRecord = 0;
    }
    prefsSize = sizeof( topRecord );
    prefRes = PrefGetAppPreferences( 'PCSH', PREFS_TOP_RECORD,
                            &topRecord, &prefsSize, true );
    if( prefRes == noPreferenceFound ) {
        topRecord = 0;
    }
    FrmGotoForm( idMainForm );
    return err;
}

static void StopApplication(void) {

    PrefSetAppPreferences( 'PCSH', PREFS_ACCOUNT, 0,
        &acctRec, sizeof( acctRec ), true );
    PrefSetAppPreferences( 'PCSH', PREFS_SORT_ORDER, 0,
        &sortOrder, sizeof( sortOrder ), true );
    bottomRecord = grid.topmostRow + grid.numScreenRows - 1;
    PrefSetAppPreferences( 'PCSH', PREFS_BOTTOM_RECORD, 0,
        &bottomRecord, sizeof( bottomRecord ), true );
        
    topRecord = grid.topmostRow;
    PrefSetAppPreferences( 'PCSH', PREFS_TOP_RECORD, 0,
        &topRecord, sizeof( topRecord ), true );
    FrmSaveAllForms();
    FrmCloseAllForms();
    MemHandleFree( totalsHandle );
    DmCloseDatabase( gTransDB );
    DmCloseDatabase( gAccountDB );

}

UInt32 PilotMain(UInt16 cmd, void * cmdPBP, UInt16 launchFlags) {

    Err err;

    if (cmd == sysAppLaunchCmdNormalLaunch) {

        err = StartApplication();
        if (err) return err;

        EventLoop();
        StopApplication();

    } 
    else {
        return sysErrParamErr;
    }
    return 0;
}
