/*
 * $Id: PCashGridModel.c,v 1.8 2000/10/21 01:49:53 jmanning Exp $
 *
 * Copyright (C) 2000 by Jackie Manning (jackie@linuxfan.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <PalmOS.h>
#include <stdarg.h>

#include "grid.h"
#include "PCashRsc.h"
#include "Trans.h"

#define GetObjectPtr(f,i) FrmGetObjectPtr((f),FrmGetObjectIndex((f),(i)))

void UpdatePostedBalance( void );
void ModelGetCellID(void * __unused, UInt16 index, UInt32 * idP);

UInt16 column_widths[5] = { 25, 12, 49, 35, 39 };
UInt16 num_rows = 35;
extern DmOpenRef gTransDB;
//extern MemHandle idxHandle;
extern MemHandle totalsHandle;
extern double postedBal;
extern double currBal;

UInt16 ModelGetColumnCount(void * __unused) {

    return 5;
}

static void ModelGetColumnName(void * __unused, UInt16 col,
			       Char * name, UInt32 len) {
                               
    switch (col) {
    case 0:
	StrNCopy(name, "Date", len);
	break;
    case 1:
	StrNCopy(name, "P", len);
	break;
    case 2:
	StrNCopy(name, "Party", len);
	break;
    case 3:
	StrNCopy(name, "Amt", len);
	break;
    case 4:
	StrNCopy(name, "Bal", len);
	break;
    }
}

UInt16 ModelGetColumnWidth(void * __unused, UInt16 col) {
    return column_widths[col];
}

void ModelSetColumnWidth(void * __unused, UInt16 col, UInt16 width) {
    column_widths[col] = width;
}

void * ModelLockIndex(void * gridModelData, UInt16 index, Boolean writable) {

    static Trans trans;
    UInt16 idx = 0;
    UInt32 id;
    
//    ModelGetCellID( gridModelData, index, &id );
//    DmFindRecordByID( gTransDB, id, &idx );
    ReadTrans( gTransDB, index, &trans );
    return &trans;
}

void ModelUnlockIndex(void * GridModelData, UInt16 index, Boolean writable,
		      void * row_data ) {
                      
}

GridCellType ModelGetCellType(void * __unused, UInt16 index, UInt16 col) {

    char buff[80];
    
    switch (col) {
    case 0:
    case 2:
	return gridCellTypeString;
    case 3:
    case 4:
	return gridCellTypeRString;
    case 1:
	return gridCellTypeBoolean;
    }
}

double getTotal( UInt16 index ) {

    double *totals;
    double total = 0.0;
    
    totals = MemHandleLock( totalsHandle );
    total = totals[index];
    MemHandleUnlock( totalsHandle );
    return total;
}

void getTotalString( Char * data, UInt16 index ) {

    doubleToStr( data, getTotal( index ), 2 );
    
}

void ModelGetCellString(void * __unused, UInt16 index, UInt16 col,
			Char * data, UInt32 len, void * row_data)
{

    Err err;
    Trans *trans;
    Char format[10];

    trans = row_data;
    if( col == 4 ) {
        getTotalString( data, index );
        return;
    }
    
    switch( col ) {
    
    case 0:
        if( trans->date.month < 10 ) {
            StrCopy( format, "0%u/" );
        }
        else {
            StrCopy( format, "%u/" );
        }
        if( trans->date.day < 10 ) {
            StrCat( format, "0%u" );
        }
        else {
            StrCat( format, "%u" );
        }
        StrPrintF( data, format, trans->date.month, trans->date.day );
        break;
        
    case 2:
        if( trans->budget == ESTIMATED ) {
            if( trans->transType == CHECK_TRANS ) {
                StrPrintF( data, "-%s:%s", trans->ref, trans->party );
            }
            else {
                StrPrintF( data, "-%c:%s", trans->transType, trans->party );
            }
        }
        else if( trans->budget == PAYABLE ) {
            if( trans->transType == CHECK_TRANS ) {
                StrPrintF( data, "+%s:%s", trans->ref, trans->party );
            }
            else {
                StrPrintF( data, "+%c:%s", trans->transType, trans->party );
            }
        }
        else {
            if( trans->transType == CHECK_TRANS ) {
                StrPrintF( data, "%s:%s", trans->ref, trans->party );
            }
            else {
                StrPrintF( data, "%c:%s", trans->transType, trans->party );
            }
        }
        break;
        
    case 3:
        if( trans->transType != DEPOSIT_TRANS ) {
            doubleToStr( data, -trans->amount, 2 );
        }
        else {
            doubleToStr( data, trans->amount, 2 );
        }
        break;
        
    case 4:
        doubleToStr( data, 0.0, 2 );
        break;
        
    default:
        StrPrintF(data, "%u,%u", index, col);
    }
}

Boolean ModelGetCellBoolean(void * __unused, UInt16 index, UInt16 col,
			    void * row_data) {

    return ((Trans*)row_data)->posted;        
}

void ModelSetCellBoolean(void * __unused, UInt16 index, UInt16 col, Boolean b,
			 void * row_data) {
                         
    UInt16 newPos;
    Trans *trn = row_data;

    if( trn->posted != b ) {    
        if( trn->posted ) {
            postedBal -= trn->amount;
        }
        else {
            postedBal += trn->amount;
            if( trn->budget != ACTUAL ) {
                currBal += trn->amount;
                trn->budget = ACTUAL;
            }
        }

    //    if( ((Trans*)row_data)->budget ) {
            trn->posted = b;
            UpdateTrans( trn, index, gTransDB, &newPos );
    //    }
        UpdatePostedBalance();
    }
}

void ModelGetCellID(void * __unused, UInt16 index, UInt32 * idP) {

    UInt32 *indexes;
    
//    indexes = MemHandleLock( idxHandle );
//    *idP = indexes[index];
//    MemHandleUnlock( idxHandle );
//    return total;
    
//    DmRecordInfo( gTransDB, index, NULL, idP, NULL );
    *idP = (UInt32) index;
}

Boolean ModelSeek(void * __unused, UInt16 * indexP, Int16 offset,
		  Int16 direction) {
                  
    UInt16 i = *indexP;
    
    num_rows = DmNumRecords( gTransDB );
    switch (direction) {
    case dmSeekForward:
	if (i + offset >= num_rows)
	    return false;
	i += offset;
	break;

    case dmSeekBackward:
	if (i >= num_rows)
	    i = num_rows - 1; /* move to valid index number */
	if (offset > i)
	    return false;
	i -= offset;
	break;
    }

    *indexP = i;
    return true;
}

GridModelType model = {
    GetColumnCount:     ModelGetColumnCount,
    GetColumnName:      ModelGetColumnName,
    GetColumnWidth:     ModelGetColumnWidth,
    SetColumnWidth:     ModelSetColumnWidth,
    LockIndex:          ModelLockIndex,
    UnlockIndex:        ModelUnlockIndex,
    GetCellType:        ModelGetCellType,
    GetCellString:      ModelGetCellString,
    GetCellBoolean:     ModelGetCellBoolean,
    SetCellBoolean:     ModelSetCellBoolean,
    GetCellID:          ModelGetCellID,
    Seek:               ModelSeek,
};

