/*
 * $Id: Trans.c,v 1.5 2000/08/20 15:33:21 jmanning Exp $
 *
 * Copyright (C) 2000 by Jackie Manning (jackie@linuxfan.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <PalmOS.h>
#include "Trans.h"
#include "PCashRsc.h"
#include "Utils.h"
#include "DBUtils.h"

static PackedTrans *lTrans, *rTrans;
static UInt32 lDays, rDays;
static Int16 i16;
DmComparF *CompareTrans;
UInt16 sortOrder;

/*
Boolean ActualTransFilter( MemHandle handle ) {
    
    MemHandle trnHandle;
    PackedTrans *packedTrans;
    Trans trans;
    Boolean result = false;
    
    packedTrans = MemHandleLock( handle );
    result = !packedTrans->budget;
    MemHandleUnlock( handle );
    
    return result;
}

Boolean AllTransFilter( MemHandle handle ) {
    return true;
}
*/

void SetTransSortOrder( UInt16 aSortOrder ) {

    switch( aSortOrder ) {
    case SORT_BY_AMOUNT:
        sortOrder = aSortOrder;
        CompareTrans = CompareTransByAmount;
        break;
    case SORT_BY_PARTY:
        sortOrder = aSortOrder;
        CompareTrans = CompareTransByParty;
        break;
    default:
        sortOrder = SORT_BY_DATE;
        CompareTrans = CompareTransByDate;
    }
}

Int16 CompareTransByAmount( 
        void *rec1, void *rec2, Int16 other, 
        SortRecordInfoPtr rec1SortInfo, 
        SortRecordInfoPtr rec2SortInfo, 
        MemHandle appInfoH ) {
        
    lTrans = (PackedTrans*)rec1;
    rTrans = (PackedTrans*)rec2;
    if( lTrans->amount > rTrans->amount ) return -1;
    if( lTrans->amount < rTrans->amount ) return 1;
    lDays = DateToDays( lTrans->date );
    rDays = DateToDays( rTrans->date );
    if( lDays < rDays ) return -1;
    if( lDays > rDays ) return 1;
    return 0;
}

Int16 CompareTransByParty( 
        void *rec1, void *rec2, Int16 other, 
        SortRecordInfoPtr rec1SortInfo, 
        SortRecordInfoPtr rec2SortInfo, 
        MemHandle appInfoH ) {
        
    lTrans = (PackedTrans*)rec1;
    rTrans = (PackedTrans*)rec2;
    i16 = StrCompare( lTrans->party, rTrans->party );
    if( i16 != 0 ) {
        return i16;
    }
    else {
        lDays = DateToDays( lTrans->date );
        rDays = DateToDays( rTrans->date );
        if( lDays < rDays ) return -1;
        if( lDays > rDays ) return 1;
        if( lTrans->amount > rTrans->amount ) return -1;
        if( lTrans->amount < rTrans->amount ) return 1;
    }
    return 0;
}

Int16 CompareTransByDate( 
        void *rec1, void *rec2, Int16 other, 
        SortRecordInfoPtr rec1SortInfo, 
        SortRecordInfoPtr rec2SortInfo, 
        MemHandle appInfoH ) {
        
    lTrans = (PackedTrans*)rec1;
    rTrans = (PackedTrans*)rec2;
    lDays = DateToDays( lTrans->date );
    rDays = DateToDays( rTrans->date );
    if( lDays < rDays ) return -1;
    if( lDays > rDays ) return 1;
    if( lTrans->amount > rTrans->amount ) return -1;
    if( lTrans->amount < rTrans->amount ) return 1;
    return 0;
}

Err ReadTrans( DmOpenRef db, UInt16 recNo, Trans *trans ) {

    MemHandle trnHandle;
    PackedTrans *packedTrans;
    
    trnHandle = DmQueryRecord( db, recNo );
    if( trnHandle == NULL ) {
        return DmGetLastErr();
    }
    packedTrans = MemHandleLock( trnHandle );
    UnpackTrans( trans, packedTrans );
    MemHandleUnlock( trnHandle );
    return 0;
}

UInt16 InsertTrans( Trans *trans, DmOpenRef db ) {

    MemHandle newRecord;
    UInt16 newIndex = dmMaxRecordIndex;
    PackedTrans *packedTrans;
    Char buff[30];
    Err err;
    
    newRecord = DmNewHandle( db, getTransSize( trans ) );
    PackTrans( trans, newRecord );
    packedTrans = MemHandleLock( newRecord );
    newIndex = DmFindSortPosition( db, packedTrans, NULL, CompareTrans, 0 );
    MemHandleUnlock( newRecord );
    err = DmAttachRecord( db, &newIndex, newRecord, NULL );
    DmReleaseRecord( db, newIndex, false );
    return newIndex;
}

Err UpdateTrans( Trans *trans, UInt16 recNo, DmOpenRef db, UInt16 *newPos ) {

    MemHandle handle;
//    UInt16 newPos;
    Err err = 0;
    PackedTrans *packedTrans;
    
    handle = DmResizeRecord( db, recNo, getTransSize( trans ) );
    if( handle == NULL ) {
        err = DmGetLastErr();
    }
    else {
        PackTrans( trans, handle );
        DmReleaseRecord( db, recNo, false );
        packedTrans = MemHandleLock( handle );
        *newPos = DmFindSortPosition( db, packedTrans, NULL, CompareTrans, 0 );
        if( *newPos != recNo ) {
            err = DmMoveRecord( db, recNo, *newPos );
        }
        if( *newPos > recNo ) {
            *newPos = recNo;
        }
        err = MemHandleUnlock( handle );
    }
    return err;
}
/*
void WriteObject( void *s, UInt32 *offset, const void *srcP, UInt32 length ) {

    DmWrite( s, *offset, srcP, length );
    *offset += length;
}
*/
PackedTrans *PackTrans( Trans *trans, MemHandle transDBEntry ) {

    UInt16      length = 0;
    void        *s;
    UInt32      offset = 0;

    length = getTransSize( trans );
    // resize the VoidHand
    if (MemHandleResize(transDBEntry, length) == 0) {
        // copy the fields
        s = MemHandleLock(transDBEntry);

        if( trans->budget ) {
            trans->posted = false;
        }        
        offset = 0;
        WriteObject( s, &offset, &trans->account, sizeof(trans->account) );
        WriteObject( s, &offset, &trans->amount, sizeof(trans->amount) );
        WriteObject( s, &offset, &trans->date, sizeof(trans->date) );
        WriteObject( s, &offset, &trans->transType, sizeof(trans->transType) );
        WriteObject( s, &offset, &trans->posted, sizeof(trans->posted) );
        WriteObject( s, &offset, &trans->budget, sizeof(trans->budget) );
        WriteObject( s, &offset, trans->party, StrLen(trans->party) + 1 );
        WriteObject( s, &offset, trans->ref, StrLen(trans->ref) + 1 );
        WriteObject( s, &offset, trans->note, StrLen(trans->note) + 1 );

        MemHandleUnlock(transDBEntry);
    }
    return (PackedTrans*)s;
}

UInt16 getTransSize( Trans *trans ) {

    UInt16 length;
    length = 0
            + sizeof( trans->account ) 
            + sizeof(trans->date)
            + sizeof( trans->amount ) 
            + sizeof( trans->transType ) 
            + sizeof( trans->posted ) 
            + sizeof( trans->budget ) 
            + StrLen(trans->party) + 1
            + StrLen(trans->ref) + 1
            + StrLen(trans->note) + 1
            ;
    return length;
}

void UnpackTrans( Trans *trans, PackedTrans *packedTrans ) {

    const char *s = packedTrans->party;
    trans->account = packedTrans->account;
    trans->amount = packedTrans->amount;
    trans->date = packedTrans->date;
    trans->transType = packedTrans->transType;
    trans->posted = packedTrans->posted;
    trans->budget = packedTrans->budget;
    trans->party = s;
    s += StrLen( s ) + 1;
    trans->ref = s;
    s += StrLen( s ) + 1;
    trans->note = s;

}

// End -------------------------------------------------------------------------
