/* Main code for ZLib test app. */

#include <Pilot.h>
#include "callback.h"
#include "zlibtestRsc.h"

#include <stdio.h>
#include "zlib.h"


/* ZLib test functions, declarations and whatever else that's neccessary to
 * test this bugger in a proper manner. */

static z_stream zs;

static Char inbuf[] = "Testing ZLib! He chanted a song of wizardry, "\
"Of piercing, opening, of treachery, Revelaing, uncovering, betraying. "\
"Then sudden Felagund there swaying Sang in answer a song of staying, "\
"Resisting, battling against power, Of secrets kept, strength like a "\
"tower, And trust unbroken, freedom, escape; Of changing and of shifting "\
"shape, Of snares eluded, broken traps, The prison opening, the chain that "\
"snaps.";	/* Source data */

#define BufSize 1024

static Word winBits_g = 10;
static Word	memLevel_g = 3;

static void testZlib(void)
{
	Int r;
	int zlen;
	static char *zbuf;		/* pointer to compressed data */
	static char *outbuf;	/* pointer to decompressed data */

	/* A lame sanity-check */
	if (8 == winBits_g) {
		printf("\nNo, no, no!\n");
		printf("A winBits value of 8 is possible, but for some lame reason it\n");
		printf("will cause a fatal exception.\nDon't use it!\n");
		return;
	}
	
	/* Set up buffers, prepare for compressing */
	zbuf = outbuf = NULL;
	zbuf = MemPtrNew(BufSize);
	if (NULL == zbuf) {
		printf("**ERROR: Couldn't allocate buffer for compressed data!\n");
		goto cleanup;
	}
	outbuf = MemPtrNew(BufSize);
	if (NULL == outbuf) {
		printf("**ERROR: Couldn't allocate buffer for decompressed data!\n");
		goto cleanup;
	}

	MemSet(&zs, sizeof(zs), 0);

	zs.next_in = inbuf;
	zs.avail_in = StrLen(inbuf);

	ErrFatalDisplayIf(zs.avail_in >= BufSize, "inbuf too large!");

	zs.next_out = zbuf;
	zs.avail_out = BufSize;

	/* Let ZLib initialize memory callback functions to its own. */
	zs.zalloc = Z_NULL;
	zs.zfree = Z_NULL;
	zs.opaque = NULL;
	/* This is the same as PGPLib uses. It's waaay overkill - as it stands now,
	 * with windowBits=-13, and memlevel=8, it would use (1<<(13+2))+(1<<(8+9))
	 * = 163840 bytes. It should perhaps be 8 and 3, respectively, which would
	 * require 5120 bytes. If it is 8 and 2, you'll need 3071. With 8 and 1, it
	 * is 2048, and that's as small as it can be. Incidentally, that's as slow
	 * as it can be too. */
	 /* After some testing, I came to the conclusion that 8 and 1 is too small
	  * - the pilot always gets a fatal exception with windowBits = 8, and
	  * any memLevel. Avoid, in other words.
	  * 10 and 3 works quite nicely. */
	r = deflateInit2(&zs, Z_DEFAULT_COMPRESSION,
			Z_DEFLATED, -winBits_g, memLevel_g,
			Z_DEFAULT_STRATEGY);
	printf("deflateInit2() returned %d\n", r);

	ErrFatalDisplayIf(r != Z_OK, "deflateInit2 failed!");

	r = deflate(&zs, Z_FINISH);
	printf("deflate() returned %d\n", r);
	ErrFatalDisplayIf(r != Z_STREAM_END, "deflate failed!");

	r = deflateEnd(&zs);
	printf("deflateEnd() returned %d\n", r);
	ErrFatalDisplayIf(r != Z_OK, "deflateEnd failed!");

	zlen = zs.total_out;

	/* Ok, so now we hopefully has some nice and correctly compressed data in
	 * zbuf, so now we try to decompress it. */

	MemSet(&zs, sizeof(zs), 0);

	zs.next_in = zbuf;
	zs.avail_in = zlen;

	zs.next_out = outbuf;
	zs.avail_out = BufSize;

	/* Let ZLib initialize memory callback functions to its own. */
	zs.zalloc = Z_NULL;
	zs.zfree = Z_NULL;
	zs.opaque = NULL;

	/* See deflateInit2 above for comments regarding the windowBits */
	r = inflateInit2(&zs, -winBits_g);
	printf("inflateInit2() returned %d\n", r);
	ErrFatalDisplayIf(r != Z_OK, "inflateInit2 failed!");

	/* Anything but Z_FINISH is a *very* bad idea on the Pilot! */
	do {
		r = inflate(&zs, Z_FINISH);
	} while (r != Z_STREAM_END);
	printf("inflate() returned %d\n", r);
	ErrFatalDisplayIf(r != Z_STREAM_END, "inflate failed!");

	r = inflateEnd(&zs);
	printf("inflateEnd() returned %d\n", r);
	ErrFatalDisplayIf(r != Z_OK, "inflateEnd failed!");


	/* OK, now dawns the moment of truth! Did we get an correct decompression
	 * or a complete failure? */
	r = StrCompare(inbuf, outbuf);
	printf("StrCompare(in, out) returned %d\n", r);
	outbuf[100]='\0';
	printf("outbuf exerpt: \"%s\"\n", outbuf);
cleanup:
	if (NULL != zbuf)
		MemPtrFree(zbuf);
	if (NULL != outbuf)
		MemPtrFree(outbuf);
	zbuf = outbuf = NULL;
}

/* End of ZLib test... stuff. */

static void UpdateControls(void)
{
	DWord   dw;
	FormPtr pfrm;
	ULong		freebytes, requiredbytes, maxbytes;
	static Char		buf[100];

	pfrm = FrmGetActiveForm();

	dw = FrmGetObjectIndex(pfrm, idWinBits);
	StrIToA(buf, (Long)winBits_g);
	FrmHideObject(pfrm, dw);
	FrmCopyLabel(pfrm, idWinBits, buf);
	FrmShowObject(pfrm, dw);

	dw = FrmGetObjectIndex(pfrm, idMemLevel);
	StrIToA(buf, memLevel_g);
	FrmHideObject(pfrm, dw);
	FrmCopyLabel(pfrm, idMemLevel, buf);
	FrmShowObject(pfrm, dw);

	MemCardInfo(0, NULL, NULL, NULL, NULL, NULL, NULL, &freebytes);
	StrPrintF(buf, "Total free mem: %ld bytes", freebytes);
	WinDrawChars(buf, StrLen(buf), 5, 46);
	/* The next two lines is probably the most horrid piece of anti-compiler-
	 * pessimising I've ever made! Argh! Ints should be 32 bits, ALWAYS! Casting
	 * sux!  -Perm
	 * */
	requiredbytes = ((Long)1) << ((Long)(winBits_g + 2));
	requiredbytes += ((Long)1) << ((Long)(memLevel_g + 9));
	StrPrintF(buf, "Mem required: %ld bytes", requiredbytes);
	WinDrawChars(buf, StrLen(buf), 5, 58);
	freebytes = -1;
	if (MemHeapDynamic(0)) {
		MemHeapCompact(0);
		freebytes = MemHeapFreeBytes(0, &freebytes, &maxbytes);
		StrPrintF(buf, "Available on heap: %ld ", freebytes);
		WinDrawChars(buf, StrLen(buf), 5, 70);
		StrPrintF(buf, "Max chunk on heap: %ld ", maxbytes);
		WinDrawChars(buf, StrLen(buf), 5, 82);
	}
	StrPrintF(buf, "Probable fit: %s  ", (requiredbytes < maxbytes)?"yes":"no");
	WinDrawChars(buf, StrLen(buf), 5, 94);
}

static Boolean MainFormHandleEvent (EventPtr e)
{
	Boolean handled = false;
	FormPtr frm;

	CALLBACK_PROLOGUE;

	switch (e->eType) {
		case frmOpenEvent:
			frm = FrmGetActiveForm();
			FrmDrawForm(frm);
			UpdateControls();
			handled = true;
			break;

		case menuEvent:
			MenuEraseStatus(NULL);

			switch(e->data.menu.itemID) {
			}

			handled = true;
			break;

		case ctlSelectEvent:
			switch(e->data.ctlSelect.controlID) {
				case goBtn:
					testZlib();
					handled = true;
					break;
				case idWinBits:
				case idMemLevel:
					{
						ListPtr     plst;
						Word        iSelect;
						FormPtr     pMe = FrmGetActiveForm();
						if (e->data.ctlSelect.controlID == idWinBits) {
							plst = FrmGetObjectPtr(pMe,
									FrmGetObjectIndex(pMe, idLstWinBits));
							LstSetSelection(plst, winBits_g - 8 );
						} else {
							plst = FrmGetObjectPtr(pMe,
									FrmGetObjectIndex(pMe, idLstMemLevel));
							LstSetSelection(plst, memLevel_g - 1 );
						}
						iSelect =  LstPopupList(plst);
						if (iSelect != noListSelection) {
							if (e->data.ctlSelect.controlID == idWinBits) {
								winBits_g = iSelect + 8;
							} else {
								memLevel_g = iSelect + 1;
							}
						}
						UpdateControls();
					}
					handled = true;
					break;
				default:
					break;
			}

		default:
			break;
	}

	CALLBACK_EPILOGUE;

	return handled;
}

static Boolean ApplicationHandleEvent(EventPtr e)
{
    FormPtr frm;
    Word    formId;
    Boolean handled = false;

    if (e->eType == frmLoadEvent) {
	formId = e->data.frmLoad.formID;
	frm = FrmInitForm(formId);
	FrmSetActiveForm(frm);

	switch(formId) {
	    case MainForm:
		FrmSetEventHandler(frm, MainFormHandleEvent);
		break;
	}
	handled = true;
    }

    return handled;
}

/* Get preferences, open (or create) app database */
static Word StartApplication(void)
{
    FrmGotoForm(MainForm);
    return 0;
}

/* Save preferences, close forms, close app database */
static void StopApplication(void)
{
    FrmSaveAllForms();
    FrmCloseAllForms();
}

/* The main event loop */
static void EventLoop(void)
{
    Word err;
    EventType e;

    do {
	EvtGetEvent(&e, evtWaitForever);
	if (! SysHandleEvent (&e))
	    if (! MenuHandleEvent (NULL, &e, &err))
		if (! ApplicationHandleEvent (&e))
		    FrmDispatchEvent (&e);
    } while (e.eType != appStopEvent);
}

/* Main entry point; it is unlikely you will need to change this except to
   handle other launch command codes */
DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
    Word err;

    if (cmd == sysAppLaunchCmdNormalLaunch) {

	err = StartApplication();
	if (err) return err;

	EventLoop();
	StopApplication();

    } else {
	return sysErrParamErr;
    }

    return 0;
}
