#include <Pilot.h>
#include <ctype.h>
#include "resist.h"


#define AppVersionNum 0x0001
#define AppCreator    CREATOR /* Makefile defines this */


#define RES_4BAND          0x8000
#define RES_INVALID_TOL    0x4000
#define RES_INVALID_OHM    0x2000
#define RES_INVALID_ACC    0x1000
#define RES_INVALID_SYN    0x0800
#define RES_INVALID_MASK   0x7f00
#define RES_TYPE_MASK      0x00ff
#define RES_DEFAULT        (RES_4BAND|3)

#define PRF_BEEP     0x0001
#define PRF_AUTOSEL  0x0002
#define PRF_ENGSTYLE 0x0004
#define PRF_DEFAULT  (PRF_BEEP|PRF_AUTOSEL|PRF_ENGSTYLE)

#define lstID_Band1    5000

#define DIGIT(c)  (((c)>='0')&&((c)<='9'))
#define UNIT(c)   ((c)=='r'||(c)=='R'||(c)=='k'||(c)=='K'||(c)=='m'||(c)=='M')
#define PERIOD(c) ((c)=='.')

#define b1 (Prefs.bands[0])
#define b2 (Prefs.bands[1])
#define b3 (Prefs.bands[2])
#define b4 (Prefs.bands[3])
#define b5 (Prefs.bands[4])

#define NUMEXPONENTS 9
static unsigned int zeroes[NUMEXPONENTS]={2,3,4,5,6,7,8,1,0};

#define NUMTOLERANCES 4
static unsigned int tolerances[NUMTOLERANCES]={1,2,5,10};

#define NUME_SERIES 4
static unsigned int num_values[NUME_SERIES]={3,6,12,24};

static unsigned int preferred_values[24]=
{100,220,470,			                  /* E3 */
 150,330,680,		                   	  /* E6 (delta) */
 120,180,270,390,560,820,	                  /* E12 (delta) */
 110,130,160,200,240,300,360,430,510,620,750,910  /* E24 (delta) */
};


static WinHandle OffscreenBitmapWinH=0;


typedef struct SoloPrefType {
  DWord	    magic;		/* CREATOR */
  Word      res_type;		/* 4/5 bands + Exx type */
  Word      prefs;		/* Preferences */
  char      bands[5];		/* Band indices */
} ResistPrefType;


static ResistPrefType Prefs;


static void
MakeBitmap()
{
  Word error;

  if (!OffscreenBitmapWinH){
    OffscreenBitmapWinH=WinCreateOffscreenWindow(11,11,screenFormat,&error);
    if(OffscreenBitmapWinH) {
      Handle    resH;
      BitmapPtr resP;
      WinHandle currDrawWindow=WinSetDrawWindow(OffscreenBitmapWinH);
      
      resH=(Handle)DmGetResource(bitmapRsc,bmpID_bzzt);
      ErrFatalDisplayIf(!resH,"no bitmap");
      resP=(BitmapPtr)MemHandleLock((VoidHand)resH);
      WinDrawBitmap(resP,0,0);
      MemPtrUnlock(resP);
      DmReleaseResource((VoidHand)resH);
      WinSetDrawWindow(currDrawWindow);
    } else {
      ErrDisplay("unallocated bitmap window");
    }
  }
}



static VoidPtr
GetObjectPtr (Word objectID)
{
  FormPtr frm = FrmGetActiveForm ();
  return (FrmGetObjectPtr(frm,FrmGetObjectIndex(frm,objectID)));
}


void
SetBands()
{
  static void CheckSeries();

  int i;
  for(i=0;i<5;i++){
    ListPtr p=GetObjectPtr(lstID_Band1+i);
    LstSetSelection(p,Prefs.bands[i]);
    LstMakeItemVisible(p,Prefs.bands[i]);
  }
  CheckSeries();
}



#if 0
static void
display(int d)
{
  RectangleType rect;
  char s[10];

  rect.topLeft.x=0;
  rect.topLeft.y=100;
  rect.extent.x=160;
  rect.extent.y=20;
  WinEraseRectangle(&rect,1);
  FntSetFont(2);

  StrPrintF(s,"%d",d);

  WinDrawChars(s,StrLen(s),0,100);
}
#endif



static void 
InterpretAndDisplay()
{
  FieldPtr p=GetObjectPtr(fldID_ohm);
  CharPtr s;
  Handle h0,h;
  FormPtr frm;
  DWord ohms, exp;
  int i;

  h=(Handle)MemHandleNew(20);
  s=MemHandleLock((VoidHand)h);

  exp=zeroes[(int)b4];
  if(Prefs.res_type&RES_4BAND){
    ohms = b1*10 + b2;
  } else {
    ohms = b1*100 + b2*10 + b3;
  }

  StrPrintF(s,"%ld",ohms);
  for(i=2;i<exp;i++)StrCat(s,"0");

  switch(exp){
  case 0:
    if(Prefs.prefs&PRF_ENGSTYLE)StrPrintF(s,"%ldR%02ld",ohms/100,ohms%100);
    else StrPrintF(s,"%ld.%02ld",ohms/100,ohms%100);
    break;
  case 1:
    if(Prefs.prefs&PRF_ENGSTYLE) StrPrintF(s,"%ldR%ld",ohms/10,ohms%10);
    else StrPrintF(s,"%ld.%ld",ohms/10,ohms%10);
    break;
  default:
    ohms=StrAToI(s);
    if(Prefs.prefs&PRF_ENGSTYLE){
      switch(StrLen(s)){
      case 1:
      case 2:
      case 3: StrCat(s,"R"); break;
      case 4: StrPrintF(s,"%ldK%02ld",ohms/1000,ohms%1000); break;
      case 5: StrPrintF(s,"%ldK%ld",ohms/1000,ohms%1000); break;
      case 6: StrPrintF(s,"%ldK",ohms/1000); break;
      case 7: StrPrintF(s,"%ldM%02ld",ohms/1000000,ohms%1000000); break;
      case 8: StrPrintF(s,"%ldM%ld",ohms/1000000,ohms%1000000); break;
      case 9: StrPrintF(s,"%ldM",ohms/1000000); break;
      }
    } else {
      switch(StrLen(s)){
      case 1:
      case 2:
      case 3: break;
      case 4: StrPrintF(s,"%ld.%02ldk",ohms/1000,ohms%1000); break;
      case 5: StrPrintF(s,"%ld.%ldk",ohms/1000,ohms%1000); break;
      case 6: StrPrintF(s,"%ldk",ohms/1000); break;
      case 7: StrPrintF(s,"%ld.%02ldM",ohms/1000000,ohms%1000000); break;
      case 8: StrPrintF(s,"%ld.%ldM",ohms/1000000,ohms%1000000); break;
      case 9: StrPrintF(s,"%ldM",ohms/1000000); break;
      }
    }
  }

  i=StrLen(s)-1;
  if(StrChr(s,'.')||Prefs.prefs&PRF_ENGSTYLE){
    if(((s[i]=='k') || (s[i]=='M')) && (i>0)){
      char unit=s[i];
      for(--i;i>0 && (s[i]=='0'||s[i]=='.');i--)s[i]=unit,s[i+1]='\0';
    } else for(;i>0 && (s[i]=='0'||s[i]=='.');i--)s[i]='\0';
  }
  MemHandleUnlock((VoidHand)h);

  (Handle)h0=FldGetTextHandle(p);
  FldSetTextHandle(p,h);
  if(Prefs.prefs&PRF_AUTOSEL)FldSetSelection(p,0,StrLen(s));
  FldDrawField(p);
  frm=FrmGetActiveForm();
  FrmSetFocus(frm,FrmGetObjectIndex(frm,fldID_ohm));
  if(h0!=NULL)MemHandleFree((VoidHand)h0);
}


static void
UpdateTolerance()
{
  FieldPtr p=GetObjectPtr(fldID_tol);
  CharPtr s;
  Handle h0,h;
  FormPtr frm;

  h=(Handle)MemHandleNew(3);
  s=MemHandleLock((VoidHand)h);

  StrPrintF(s,"%d",tolerances[(int)b5]);
  MemHandleUnlock((VoidHand)h);
  (Handle)h0=FldGetTextHandle(p);
  FldSetTextHandle(p,h);
  if(Prefs.prefs&PRF_AUTOSEL)FldSetSelection(p,0,StrLen(s));
  FldDrawField(p);
  frm=FrmGetActiveForm();
  FrmSetFocus(frm,FrmGetObjectIndex(frm,fldID_tol));
  if(h0!=NULL)MemHandleFree((VoidHand)h0);
}


static void
MakeInvalid(DWord flag)
{
  FormPtr       frm=FrmGetActiveForm();
  ControlPtr    ctl=GetObjectPtr(ctlID_approx);
  SWord         x,y;
  RectangleType rect;
    
  Prefs.res_type|=flag;
    
  if(Prefs.prefs&PRF_BEEP)SndPlaySystemSound(sndError);

  CtlSetUsable(ctl,1);
  CtlDrawControl(ctl);

  rect.extent.x=rect.extent.y=11;
  rect.topLeft.x=rect.topLeft.y=0;

  FrmGetObjectPosition(frm,FrmGetObjectIndex(frm,ctlID_approx),&x,&y);
  WinCopyRectangle(OffscreenBitmapWinH,WinGetDisplayWindow(),
		   &rect,x+1,y+1,scrCopy);
}


static void
MakeValid(DWord flag)
{
  ControlPtr ctl=GetObjectPtr(ctlID_approx);
  Prefs.res_type&=~flag;
  if((Prefs.res_type&RES_INVALID_MASK)==0){
    CtlEraseControl(ctl);
    CtlSetUsable(ctl,0);
  }
}



static void
CheckSeries()
{
  if((Prefs.res_type&RES_TYPE_MASK)==0){
    MakeValid(RES_INVALID_OHM);
    return;
  } else {
    int num=num_values[((Prefs.res_type&RES_TYPE_MASK)-1)%NUME_SERIES];
    int i,j=b1*100+b2*10+(Prefs.res_type&RES_4BAND?0:b3);

    for(i=0;i<num;i++){
      if(preferred_values[i]==j){
	MakeValid(RES_INVALID_OHM);
	return;
      }
    }
  }
  MakeInvalid(RES_INVALID_OHM);
}




#define STATE_LEFT             0
#define STATE_UNIT_OR_PERIOD   1
#define STATE_RIGHT_OF_PERIOD  2
#define STATE_RIGHT_OF_UNIT    3
#define STATE_END_OR_UNIT      4
#define STATE_END              5

static Boolean
ParseResistance(char *s)
{
  DWord num=0, intpart=1, decpart=0;
  char  unit='R', b[4]={48,48,48,48};
  int   state=STATE_LEFT;
  char *cp;

  for(cp=s;*cp;cp++){
  again:
    switch(state){

    case STATE_LEFT:
      if(DIGIT(*cp)){
	if(num==0 && *cp=='0')continue;
	if(intpart==1)b[0]=*cp;
	else if(intpart==10)b[1]=*cp;
	else if(intpart==100)b[2]=*cp;
	else if(intpart==1000)b[3]=*cp;
	num=num*10+(*cp-'0');
	intpart*=10;
	break;
      } else {
	state=STATE_UNIT_OR_PERIOD;
	goto again;
      }

    case STATE_UNIT_OR_PERIOD:
      if(UNIT(*cp)){
	unit=*cp;
	state=STATE_RIGHT_OF_UNIT;
	break;
      } else if(PERIOD(*cp)){
	state=STATE_RIGHT_OF_PERIOD;
	break;
      }
      return false;

    case STATE_RIGHT_OF_PERIOD:
    case STATE_RIGHT_OF_UNIT:
      if(DIGIT(*cp)){
	if(intpart==1)b[0]=*cp;
	else if(intpart==10)b[1]=*cp;
	else if(intpart==100)b[2]=*cp;
	else if(intpart==1000)b[3]=*cp;
	num=num*10+(*cp-'0');
	intpart*=10;
	if(decpart)decpart*=10;
	else decpart=1;
	break;
      } else {
	if(state==STATE_RIGHT_OF_UNIT)state=STATE_END;
	else state=STATE_END_OR_UNIT;
	goto again;
      }

    case STATE_END_OR_UNIT:
      if(UNIT(*cp)){
	unit=*cp;
	state=STATE_END;
	break;
      }
      return false;

    case STATE_END:
      return false;
    }
  }

  {
    //char s[80];
    int decade;
    b1=b[0]-'0';
    b2=b[1]-'0';
    b3=b[2]-'0';

    if(Prefs.res_type&RES_4BAND && b3){
      MakeInvalid(RES_INVALID_ACC);
      return true;
    } else if((Prefs.res_type&RES_4BAND)==0 && b[3]!='0'){
      MakeInvalid(RES_INVALID_ACC);
      return true;
    }

    decade=-1;
    if(Prefs.res_type&RES_4BAND){
      switch(unit){
      case 'r':
      case 'R':
	switch(intpart){
	case 1: decade=0; break;	        /* zero */
	case 10: decade=7; break;   /* single-digit */
	case 100: decade=0; break;	        /* two digits */
	case 1000: decade=1; break;	        /* three digits */
	case 10000: decade=2; break;
	case 100000: decade=3; break;
	case 1000000: decade=4; break;
	}
	break;
      case 'k':
      case 'K':
	switch(intpart/(decpart?decpart*10:1)){
	case 10: decade=2; break;   /* single-digit */
	case 100: decade=3; break;	        /* two digits */
	case 1000: decade=4; break;	        /* three digits */
	case 10000: decade=5; break;
	case 100000: decade=6; break;
	}
	break;
      case 'm':
      case 'M':
	switch(intpart/(decpart?decpart*10:1)){
	case 10: decade=5; break;   /* single-digit */
	case 100: decade=6; break;	        /* two digits */
	}
	break;
      }
    } else {
      switch(unit){
      case 'r':
      case 'R':
	switch(intpart){
	case 1: decade=0; break;	        /* zero */
	case 10: decade=8; break;   /* single-digit */
	case 100: decade=7; break;	        /* two digits */
	case 1000: decade=0; break;	        /* three digits */
	case 10000: decade=1; break;
	case 100000: decade=2; break;
	case 1000000: decade=3; break;
	}
	break;
      case 'k':
      case 'K':
	switch(intpart/(decpart?decpart*10:1)){
	case 10: decade=1; break;   /* single-digit */
	case 100: decade=2; break;	        /* two digits */
	case 1000: decade=3; break;	        /* three digits */
	case 10000: decade=4; break;
	case 100000: decade=5; break;
	}
	break;
      case 'm':
      case 'M':
	switch(intpart/(decpart?decpart*10:1)){
	case 10: decade=4; break;   /* single-digit */
	case 100: decade=5; break;	        /* two digits */
	}
	break;
      }
    }

    if(decade>=0){
      b4=decade;
      MakeValid(RES_INVALID_ACC);
    } else MakeInvalid(RES_INVALID_ACC);

    SetBands();

    //StrPrintF(s,"%ld,%ld,%ld",num,intpart,decpart);
    //display(s);
  }

  return true;
}


static Boolean
ReverseUpdate(EventPtr event)
{
  FormPtr  frm   = FrmGetActiveForm();
  Word     index = FrmGetFocus(frm);
  FieldPtr fld   = FrmGetObjectPtr(frm,index);
  CharPtr  s;

  if(!FldHandleEvent(fld,event) && event->eType!=fldChangedEvent)return false;
  s=FldGetTextPtr(fld); 
  //display(s);

  switch(FrmGetObjectId(frm,index)){
  case fldID_ohm:
    {
      Boolean a=ParseResistance(s);
      if(!a)MakeInvalid(RES_INVALID_SYN);
      else MakeValid(RES_INVALID_SYN);
    }
    break;
  case fldID_tol:
    {
      int i,j=StrAToI(s);

      if(StrLen(s))for(i=0;i<NUMTOLERANCES;i++){
	if(tolerances[i]==j){
	  ListPtr lst=GetObjectPtr(lstID_Band1+4);

	  LstSetSelection(lst,i);
	  LstMakeItemVisible(lst,i);
	  MakeValid(RES_INVALID_TOL);
	  return true;
	}
      }

      MakeInvalid(RES_INVALID_TOL);
    }
  }

  return true;
}


static void
LoadPrefs()
{
  Boolean found;
  
  found=PrefGetAppPreferencesV10(AppCreator,AppVersionNum,&Prefs,sizeof(Prefs));
  if (!found || Prefs.magic!=CREATOR) {
    int i;
    Prefs.magic=CREATOR;
    Prefs.res_type=RES_DEFAULT;
    Prefs.prefs=PRF_DEFAULT;
    for(i=0;i<4;i++)Prefs.bands[i]=0;
    Prefs.bands[4]=2;
  }
}


static void
SavePrefs(void)
{
  Prefs.magic=CREATOR;
  PrefSetAppPreferencesV10(AppCreator,AppVersionNum,&Prefs,sizeof(Prefs));
}


void
SetValues()
{
  CtlSetValue(GetObjectPtr(ctlID_Bands4),Prefs.res_type&RES_4BAND);
  CtlSetValue(GetObjectPtr(ctlID_Bands5),(Prefs.res_type&RES_4BAND)==0);

  LstSetSelection(GetObjectPtr(lstID_Type),Prefs.res_type&RES_TYPE_MASK);
  CtlSetLabel(GetObjectPtr(ctlID_Type),
	      LstGetSelectionText(GetObjectPtr(lstID_Type),
				  Prefs.res_type&RES_TYPE_MASK));

  SetBands();
  /*((ListPtr)GetObjectPtr(lstID_Band1+2))->attr.usable=(Prefs.res_type&RES_4BAND)==0;*/
}


static void
HideShowBand3()
{
  ListPtr p=GetObjectPtr(lstID_Band1+2);
  if(Prefs.res_type&RES_4BAND){
    RectangleType r;
    p->attr.usable=0;
    p->attr.enabled=0;
    p->attr.visible=0;
    r.topLeft.x=p->bounds.topLeft.x-1;
    r.topLeft.y=p->bounds.topLeft.y-1;
    r.extent.x=p->bounds.extent.x+2;
    r.extent.y=p->bounds.extent.y+2;
    WinEraseRectangle(&r,1);
  } else {
    p->attr.usable=1;
    p->attr.enabled=1;
    LstDrawList(p);
  }
  InterpretAndDisplay();
}


static Boolean
BandSelect(EventPtr event)
{
  int id=event->data.lstSelect.listID-lstID_Band1;
  Prefs.bands[id]=event->data.lstSelect.selection;

  if(id>=0 && id<=3)InterpretAndDisplay();
  else if(id==4)UpdateTolerance();

  CheckSeries();

  return true;
}


static void
Approximate()
{
  if(Prefs.res_type&RES_TYPE_MASK){
    int ohms,i;
    int num=num_values[((Prefs.res_type&RES_TYPE_MASK)-1)%NUME_SERIES];
    int minval=100, mindist=10000;

    if(Prefs.res_type&RES_4BAND)ohms=b1*100+b2*10;
    else ohms=b1*100+b2*10+b3;

    for(i=0;i<num;i++){
      int dist=preferred_values[i]-ohms;
      if(dist<0)dist=-dist;
      if(dist<mindist){
	mindist=dist;
	minval=preferred_values[i];
      }
      if(!dist)break;		/* Best we ca do! */
    }

    b1=minval/100;
    b2=(minval%100)/10;
    b3=minval%10;		/* Set third band regardless of 4/4B state */
  }

  /* Now check the tolerance value */
  {
    FieldPtr fld = GetObjectPtr(fldID_tol);
    CharPtr  s;
    int tol;
    int i,minval=0,mindist=10000;

    s=FldGetTextPtr(fld); 
    //tol=tolerances[(int)b5];
    tol=StrAToI(s);
    //display(tol);
    for(i=0;i<NUMTOLERANCES;i++){
      int dist=tolerances[i]-tol;
      if(dist<0)dist=-dist;
      if(dist<mindist){
	mindist=dist;
	minval=i;
      }
      if(!dist)break;		/* Best we can do! */
    }
    b5=minval;
  }

  MakeValid(RES_INVALID_MASK);	/* Make everything valid */
  UpdateTolerance();
  SetBands();
  InterpretAndDisplay();
}


static void
DoPrefs()
{
  FormPtr frm = FrmInitForm(PrefsForm);

  CtlSetValue(FrmGetObjectPtr(frm,FrmGetObjectIndex(frm,chkID_beep)),
	      Prefs.prefs&PRF_BEEP?1:0);
  CtlSetValue(FrmGetObjectPtr(frm,FrmGetObjectIndex(frm,chkID_auto)),
	      Prefs.prefs&PRF_AUTOSEL?1:0);
  LstSetSelection(FrmGetObjectPtr(frm,FrmGetObjectIndex(frm,lstID_disp)),
		  Prefs.prefs&PRF_ENGSTYLE?0:1);
  CtlSetLabel(FrmGetObjectPtr(frm,FrmGetObjectIndex(frm,ctlID_disp)),
	      LstGetSelectionText(FrmGetObjectPtr(frm,FrmGetObjectIndex(frm,lstID_disp)),
				  Prefs.prefs&PRF_ENGSTYLE?0:1));

  if(FrmDoDialog(frm)==btnID_prefsok){
    Prefs.prefs&=~(PRF_BEEP|PRF_AUTOSEL|PRF_ENGSTYLE);
    if(CtlGetValue(FrmGetObjectPtr(frm,FrmGetObjectIndex(frm,chkID_beep))))
      Prefs.prefs|=PRF_BEEP;
    if(CtlGetValue(FrmGetObjectPtr(frm,FrmGetObjectIndex(frm,chkID_auto))))
      Prefs.prefs|=PRF_AUTOSEL;
    if(LstGetSelection(FrmGetObjectPtr(frm,FrmGetObjectIndex(frm,lstID_disp)))==0)
      Prefs.prefs|=PRF_ENGSTYLE;
  }

  FrmDeleteForm(frm);
}


static Boolean
MainFormHandleEvent(EventPtr event)
{
  Boolean handled=false;
  FormPtr frm;
  
  switch (event->eType) {
  case frmOpenEvent:
    frm = FrmGetActiveForm();
    HideShowBand3();
    FrmDrawForm(frm);
    FrmSetFocus(frm,FrmGetObjectIndex(frm,fldID_ohm));
    UpdateTolerance();
    InterpretAndDisplay();
    handled = true;
    break;

  case keyDownEvent:
  case fldChangedEvent:
    handled=ReverseUpdate(event);
    break;

  case lstSelectEvent:
    handled=BandSelect(event);
    break;

  case popSelectEvent:
    if(event->data.popSelect.listID==lstID_Type){
      int i=event->data.popSelect.selection;
      if(i!=(Prefs.res_type&RES_TYPE_MASK)){
	Prefs.res_type=(Prefs.res_type&~RES_TYPE_MASK)|i;
	CheckSeries();
	CtlSetLabel(GetObjectPtr(ctlID_Type),
		    LstGetSelectionText(GetObjectPtr(lstID_Type),
					Prefs.res_type&RES_TYPE_MASK));
      }
      handled=true;
    }
    break;

  case ctlSelectEvent:
    switch(event->data.ctlSelect.controlID){
    case ctlID_Bands4:
      Prefs.res_type = (Prefs.res_type&RES_TYPE_MASK)|RES_4BAND;
      HideShowBand3();
      handled=true;
      break;
    case ctlID_Bands5:
      Prefs.res_type &= RES_TYPE_MASK;
      HideShowBand3();
      handled=true;
      break;
    case ctlID_approx:
      Approximate();
      handled=true;
      break;
    }
    break;

  case menuEvent:
    MenuEraseStatus(MenuGetActiveMenu());
    switch(event->data.menu.itemID){

    case mnuID_About:
      frm = FrmInitForm(AboutForm);
      FrmDoDialog(frm);
      FrmDeleteForm(frm);
      handled = true;
      break;
    case mnuID_Prefs:
      DoPrefs();
      handled = true;
      break;
    case mnuID_Instr:
      FrmHelp(stringID_Instructions);
      handled=true;
      break;
    case mnuID_ESeries:
      FrmHelp(stringID_ESeries);
      handled=true;
      break;
    }
    break;

  default:
    break;			/* To stop spurious warning */
  }

  //FldGrabFocus(GetObjectPtr(fldID_ohm));
  return handled;
}


static Boolean
ApplicationHandleEvent(EventPtr event)
{
  FormPtr  frm;
  Int      formId;
  Boolean  handled=false;

  if (event->eType == frmLoadEvent) {
    formId=event->data.frmLoad.formID;
    frm=FrmInitForm(formId);
    FrmSetActiveForm(frm);
    SetValues();
    
    switch(formId) {
    case MainForm:
      FrmSetEventHandler(frm,MainFormHandleEvent);
      break;
    }
    handled = true;
  }
  return handled;
}


static void
EventLoop()
{
  EventType event;
  Word      error;

  do {
    EvtGetEvent(&event, 2);
    if (!SysHandleEvent(&event)){
      if (!MenuHandleEvent(0, &event, &error)){
	if (!ApplicationHandleEvent(&event)){
	  FrmDispatchEvent(&event);
	}
      }
    }
  } while (event.eType!=appStopEvent);
}


static Boolean
StartApplication()
{
  FrmGotoForm(MainForm);
  MakeBitmap();
  LoadPrefs();

  return false;
}


static void
StopApplication()
{
  FrmCloseAllForms();
  SavePrefs();
}


DWord
PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{

  if(cmd==sysAppLaunchCmdNormalLaunch) {
    if (!StartApplication()) {
      EventLoop();
      StopApplication();
    }
  }
  return 0;
}
