/****************************************************************************
*																			*
*					SCEZ chipcard library - Reader detect					*
*						Copyright Matthias Bruestle 1999					*
*					For licensing, see the file COPYRIGHT					*
*																			*
****************************************************************************/

/* $Id: crdetect.c,v 1.4 2000/09/04 09:41:15 zwiebeltu Exp $ */

/* TODO: Detect GCR400. */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#if !defined(WINDOWS) && !defined(__palmos__)
#include <unistd.h> /* for sleep */
#elif defined(__BORLANDC__)
#include <dos.h>	/* for sleep */
#elif defined(WINDOWS)
#include <windows.h>
#endif

#if defined( WITH_CRYPTLIB )
#include "misc/sio.h"
#elif defined( __palmos__ )
#include "siolight.h"
#else
#include "sio.h"
#endif /* WITH_CRYPTLIB */

#include "scgeneral.h"
#include "screader.h"
#include "scacr20.h"
#include "scctapi.h"
#include "scdumbmouse.h"
#include "screflex60.h"
#include "sctowitoko.h"

int scAcr20Detect( char *param, BYTE *prob, int *major, int *minor )
{
	SIO_INFO *si;
	char *port;
	BYTE buffer[14];

	if( param==NULL ) return( SC_EXIT_BAD_PARAM );

	if( (strcmp(param,"0")==0) || (strcmp(param,"1")==0) ||
		(strcmp(param,"2")==0) || (strcmp(param,"3")==0) ) {
		port=scReaderExpandPort( param );
		if( port==NULL ) return( SC_EXIT_BAD_PARAM );
	} else port=param;

	*prob=0;
	*major=0;
	*minor=0;

	/* Open port */
	si = SIO_Open( port );
	if( si==NULL ) return( SC_EXIT_IO_ERROR );

	/* Set main port parameters */
	SIO_SetSpeed( si, 9600 );
	SIO_SetDataBits( si, 8 );
	SIO_SetParity( si, SIO_PARITY_EVEN );
	SIO_SetStopBits( si, 2 );
	SIO_SetIOMode( si, SIO_IOMODE_DIRECT );
	SIO_WriteSettings( si );

	/* Get unwanted bytes */
#ifdef __palmos__
	SIO_Flush(si);
#else
	while(SIO_ReadChar(si)!=-1);
#endif /* __palmos__ */

	/* Wait for status message. */
	SIO_WaitForData( si, 2000000 );

	if( SIO_ReadBuffer( si, buffer, 14 ) != 14 ) {
		SIO_Close( si );
		return( SC_EXIT_OK );
	}

	SIO_Close( si );

	if( memcmp( buffer, "\x02\x30\x31\x46\x46\x30\x30\x30\x31\x31\x32", 11 ) )
		return( SC_EXIT_OK );

	*major=SC_READER_ACR20;
	*minor=SC_ACR20_SERIAL;
	/* Unless there is a other reader which sends the same message on
	 * power on it is a sure sign, that his reader is present.
	 */
	*prob=250;

	return( SC_EXIT_OK );
}

int scB1Detect( char *param, BYTE *prob, int *major, int *minor )
{
	SIO_INFO *si;
	char *port;
	BYTE buffer[4];

	if( param==NULL ) return( SC_EXIT_BAD_PARAM );

	if( (strcmp(param,"0")==0) || (strcmp(param,"1")==0) ||
		(strcmp(param,"2")==0) || (strcmp(param,"3")==0) ) {
		port=scReaderExpandPort( param );
		if( port==NULL ) return( SC_EXIT_BAD_PARAM );
	} else port=param;

	*prob=0;
	*major=0;
	*minor=0;

	/* Open port */
	si = SIO_Open( port );
	if( si==NULL ) return( SC_EXIT_IO_ERROR );

	/* Set main port parameters */
	SIO_SetSpeed( si, 9600 );
	SIO_SetDataBits( si, 8 );
	SIO_SetParity( si, SIO_PARITY_EVEN );
	SIO_SetStopBits( si, 2 );
	SIO_SetIOMode( si, SIO_IOMODE_DIRECT );

	SIO_WriteSettings( si );

	/* Get unwanted bytes */
#ifdef __palmos__
	SIO_Flush(si);
#else
	while(SIO_ReadChar(si)!=-1);
#endif /* __palmos__ */

	/* Send command */
	if( SIO_WriteBuffer( si, "\x00\xC0\x00\xC0", 4 ) != 4 ) {
		SIO_Close( si );
		return( SC_EXIT_IO_ERROR );
	}

	/* Read answer */
	if( SIO_ReadBuffer( si, buffer, 4 ) != 4 ) {
		SIO_Close( si );
		return( SC_EXIT_OK );
	}

	SIO_Close( si );

	if( memcmp( buffer, "\x00\xE0\x00\xE0", 4 )==0 ) {
		*major=SC_READER_B1;
		*minor=0;
		/* There are more readers with T=1-like protocol.
		 * Even a Dumb Mouse with T=1 card might be mistaken for
		 * this reader. Need more tests.
		 */
		*prob=50;
	}

	return( SC_EXIT_OK );
}

/* Can load a shared library. */
#if defined(HAS_DLOPEN) || defined(__WIN16__) || defined(__WIN32__)

#if defined(HAS_DLOPEN)
#include <dlfcn.h>
#elif defined(__WIN32__)
/* #include <windows.h> */
#endif /* HAS_DLOPEN */

#if defined(HAS_DLOPEN)
#define CTAPI_CALL
#elif defined(__WIN32__)
#define CTAPI_CALL __stdcall
#endif /* HAS_DLOPEN */

/* Function prototypes */
unsigned char CTAPI_CALL CT_init( WORD ctn, WORD pn );
unsigned char CTAPI_CALL CT_data( WORD ctn, BYTE *dad, BYTE *sad, WORD lenc,
	BYTE *command, WORD *lenr, BYTE *response );
unsigned char CTAPI_CALL CT_close( WORD ctn );

/* Global function pointers. */
typedef unsigned char (CTAPI_CALL *CT_INIT)( WORD ctn, WORD pn );
typedef unsigned char (CTAPI_CALL *CT_DATA)( WORD ctn, BYTE *dad, BYTE *sad,
	WORD lenc, BYTE *command, WORD *lenr, BYTE *response );
typedef unsigned char (CTAPI_CALL *CT_CLOSE)( WORD ctn );
static CT_INIT pCT_init = NULL;
static CT_DATA pCT_data = NULL;
static CT_CLOSE pCT_close = NULL;
#define CT_init pCT_init
#define CT_data pCT_data
#define CT_close pCT_close

int scCtapiDetect( char *param, BYTE *prob, int *major, int *minor )
{
	char *libname[]= {
#if defined(HAS_DLOPEN)
#ifdef __FreeBSD__
		"/usr/lib/libctapi.so",
		"/usr/lib/libctapi-t2i.so",
		"/usr/lib/libctapi-mct09.so",
		"/usr/lib/libctapi-towitoko.so",
		"/usr/lib/libix_ix36.so",
		"/usr/lib/liblit_a210.so",
		"/usr/lib/libslb_rf60.so",
		"/usr/lib/libtodos_ag.so",
#else /* !__FreeBSD__ */
		"libctapi.so",
		"libctapi-t2i.so",
		"libctapi-mct09.so",
		"libctapi-towitoko.so",
		"libix_ix36.so",
		"liblit_a210.so",
		"libslb_rf60.so",
		"libtodos_ag.so",
#endif /* __FreeBSD__ */
#elif defined(WINDOWS)
		"ctapiw32.dll",
		"cttwkw32.dll",
		"ct32.dll",
#endif /* HAS_DLOPEN */
		NULL
	};

#if defined(HAS_DLOPEN)
	static void *LibHandle = NULL;
#elif defined(WINDOWS)
#define NULL_HINSTANCE	(HINSTANCE) NULL
	static HINSTANCE LibHandle = NULL_HINSTANCE;
#endif /* HAS_DLOPEN */

	BYTE resetct[]={0x20,0x11,0x00,0x00};
	BYTE getstatus[]={0x20,0x13,0x00,0x80,0x00};
	BYTE response[40];
	BYTE sad=2, dad=1;
	WORD ctn=0, pn, lenc=4, lenr=2;
	unsigned char ret;
	int i=0, tmp;

	*prob=0;
	*major=0;
	*minor=0;

	if( sscanf( param, "%2d", &tmp )!=1 ) return( SC_EXIT_BAD_PARAM );
	if( (tmp<0) && (tmp>15) ) return( SC_EXIT_BAD_PARAM );
	pn=tmp & 0xFFFF;

	while( libname[i]!=NULL ) {
		ctn++;

#ifdef HAS_DLOPEN
		if( LibHandle!=NULL ) {
			dlclose( LibHandle );
			LibHandle=NULL;
		}

		LibHandle = dlopen( libname[i], RTLD_LAZY );
		if( !LibHandle ) { i++; continue; }

		pCT_init = (CT_INIT) dlsym( LibHandle, "CT_init" );
		if( dlerror()!=NULL ) { i++; dlclose( LibHandle ); continue; }
		pCT_data = (CT_DATA) dlsym( LibHandle, "CT_data" );
		if( dlerror()!=NULL ) { i++; dlclose( LibHandle ); continue; }
		pCT_close = (CT_CLOSE) dlsym( LibHandle, "CT_close" );
		if( dlerror()!=NULL ) { i++; dlclose( LibHandle ); continue; }

		/* Init */
		ret = (*CT_init)( ctn, pn );
		if( ret!=SC_CTAPI_EXIT_OK ) { i++; dlclose( LibHandle ); continue; }

		/* Reset CT */
		dad=1; sad=2; lenc=4; lenr=2;
		ret = (*CT_data)( ctn, &dad, &sad, lenc, resetct, &lenr, response );
		if( (ret!=SC_CTAPI_EXIT_OK) || (lenr!=2) || (response[0]!=0x90) )
		{ i++; (*CT_close)( ctn ); dlclose( LibHandle ); continue; }

		/* Get Status ICC */
		dad=1; sad=2; lenc=5; lenr=sizeof(response);
		ret = (*CT_data)( ctn, &dad, &sad, lenc, getstatus, &lenr, response );
		if( (ret!=SC_CTAPI_EXIT_OK) || (lenr<2) || (response[lenr-2]!=0x90) )
		{ i++; (*CT_close)( ctn ); dlclose( LibHandle ); continue; }

		ret = (*CT_close)( ctn );
		if( ret!=SC_CTAPI_EXIT_OK ) { i++; dlclose( LibHandle ); continue; }

		if( LibHandle!=NULL ) {
			dlclose( LibHandle );
			LibHandle=NULL;
		}
#elif defined(WINDOWS)
		if( LibHandle!=NULL_HINSTANCE ) {
			FreeLibrary( LibHandle );
			LibHandle=(HINSTANCE)NULL;
		}

		if( ( LibHandle = LoadLibrary( libname[i] ) ) == NULL_HINSTANCE )
		{ i++; continue; }

		pCT_init = (CT_INIT) GetProcAddress( LibHandle, "CT_init" );
		pCT_data = (CT_DATA) GetProcAddress( LibHandle, "CT_data" );
		pCT_close = (CT_CLOSE) GetProcAddress( LibHandle, "CT_close" );

		/* Make sure we got valid pointers for every card function */
		if( pCT_init == NULL || pCT_data == NULL || pCT_close == NULL )
		{ i++; FreeLibrary( LibHandle ); continue; }

		/* Init */
		ret = (*CT_init)( ctn, pn );
		if( ret!=SC_CTAPI_EXIT_OK ) { i++; FreeLibrary( LibHandle ); continue; }

		/* Reset CT */
		dad=1; sad=2; lenr=2;
		ret = (*CT_data)( ctn, &dad, &sad, lenc, resetct, &lenr, response );
		if( (ret!=SC_CTAPI_EXIT_OK) || (lenr!=2) || (response[0]!=0x90) )
		{ i++; (*CT_close)( ctn ); FreeLibrary( LibHandle ); continue; }

		/* Get Status ICC */
		dad=1; sad=2; lenc=5; lenr=sizeof(response);
		ret = (*CT_data)( ctn, &dad, &sad, lenc, getstatus, &lenr, response );
		if( (ret!=SC_CTAPI_EXIT_OK) || (lenr<2) || (response[lenr-2]!=0x90) )
		{ i++; (*CT_close)( ctn ); FreeLibrary( LibHandle ); continue; }

		ret = (*CT_close)( ctn );
		if( ret!=SC_CTAPI_EXIT_OK ) { i++; FreeLibrary( LibHandle ); continue; }

		if( LibHandle!=NULL_HINSTANCE ) {
			FreeLibrary( LibHandle );
			LibHandle=(HINSTANCE)NULL;
		}
#endif /* HAS_DLOPEN */

		*major=SC_READER_CTAPI;
		*minor=i; /* Some information about the library name. */
		*prob=150;

		return( SC_EXIT_OK );
	}

#ifdef HAS_DLOPEN
	if( LibHandle!=NULL ) dlclose( LibHandle );
#elif defined(WINDOWS)
	if( LibHandle!=NULL_HINSTANCE ) FreeLibrary( LibHandle );
#endif /* HAS_DLOPEN */

	return( SC_EXIT_OK );
}
#else /* !(HAS_DLOPEN || WINDOWS) */
int scCtapiDetect( char *param, BYTE *prob, int *major, int *minor )
{
	return( SC_EXIT_NOT_SUPPORTED );
}
#endif /* HAS_DLOPEN || WINDOWS */

int scDumbmouseDetect( char *param, BYTE *prob, int *major, int *minor )
{
	SIO_INFO *si;
	SC_READER_INFO ri;
	SC_CARD_INFO ci;
	char *port;
	int ret;

	if( param==NULL ) return( SC_EXIT_BAD_PARAM );

	if( (strcmp(param,"0")==0) || (strcmp(param,"1")==0) ||
		(strcmp(param,"2")==0) || (strcmp(param,"3")==0) ) {
		port=scReaderExpandPort( param );
		if( port==NULL ) return( SC_EXIT_BAD_PARAM );
	} else port=param;

	*prob=0;
	*major=0;
	*minor=0;

	/* Open port */
	si = SIO_Open( port );
	if( si==NULL ) return( SC_EXIT_IO_ERROR );

	/* Set main port parameters */
	SIO_SetSpeed( si, 9600 );
	SIO_SetDataBits( si, 8 );
	SIO_SetParity( si, SIO_PARITY_EVEN );
	SIO_SetStopBits( si, 2 );
	SIO_SetIOMode( si, SIO_IOMODE_DIRECT );
	SIO_WriteSettings( si );

	SIO_DropRTS( si );
	SIO_ReadControlState( si );
	if( SIO_GetControlState( si, SIO_CONTROL_CTS )==1 ) {
		SIO_Close( si );
		return( SC_EXIT_OK );
	}

	SIO_RaiseRTS( si ); /* +12V */
	SIO_ReadControlState( si );
	if( SIO_GetControlState( si, SIO_CONTROL_CTS )==0 ) {
		*minor=SC_DUMBMOUSE_UNIPROG;
		/* We have successfully opened a serial port and it is possible,
		 * that someone has attached a UniProg at it.
		 */
		*prob=1;
	} else {
		*minor=SC_DUMBMOUSE_DUMBMOUSE;
		/* We have successfully opened a serial port and RTS is connected
		 * to CTS. The Dumb Mouse and many other devices, e.g. ACR20S,
		 * qualify for this.
		 */
		*prob=16;
	}

	SIO_Close( si );

	ri.si=NULL;
	ri.slot=1;

	ret=scDumbmouseInit( &ri, port );
	if( ret!=SC_EXIT_OK ) {
		*prob=0;
		scDumbmouseShutdown( &ri );
		return( SC_EXIT_OK );
	}

	/* We just try a reset. Maybe we have luck and the user has left a
	 * smart card stuck in the reader. If we get a valid ATR back, the
	 * device has to be at least similar to a Dumb Mouse.
	 */
	ret=scDumbmouseResetCard( &ri, &ci );
	if( ret==SC_EXIT_OK ) *prob=120;

	scDumbmouseShutdown( &ri );

	*major=SC_READER_DUMBMOUSE;

	return( SC_EXIT_OK );
}

int scReflex60Detect( char *param, BYTE *prob, int *major, int *minor )
{
	SC_READER_INFO ri;
	char *port;
	int ret;

	if( param==NULL ) return( SC_EXIT_BAD_PARAM );

	if( (strcmp(param,"0")==0) || (strcmp(param,"1")==0) ||
		(strcmp(param,"2")==0) || (strcmp(param,"3")==0) ) {
		port=scReaderExpandPort( param );
		if( port==NULL ) return( SC_EXIT_BAD_PARAM );
	} else port=param;

	*prob=0;
	*major=0;
	*minor=0;

	ri.si=NULL;
	ri.slot=1;

	ret=scReflex60Init( &ri, param );
	if( ret==SC_EXIT_OK ) {
		*prob=200;
		*major=SC_READER_REFLEX60;
		*minor=ri.minor;
	}

	scReflex60Shutdown( &ri );

	return( SC_EXIT_OK );
}

int scTowitokoDetect( char *param, BYTE *prob, int *major, int *minor )
{
	SIO_INFO *si;
	char *port;
	BYTE buffer[3];
	BYTE cs;
	int i;

	if( param==NULL ) return( SC_EXIT_BAD_PARAM );

	if( (strcmp(param,"0")==0) || (strcmp(param,"1")==0) ||
		(strcmp(param,"2")==0) || (strcmp(param,"3")==0) ) {
		port=scReaderExpandPort( param );
		if( port==NULL ) return( SC_EXIT_BAD_PARAM );
	} else port=param;

	*prob=0;
	*major=0;
	*minor=0;

	/* Open port */
	si = SIO_Open( port );
	if( si==NULL ) return( SC_EXIT_IO_ERROR );

	/* Set main port parameters */
	SIO_SetSpeed( si, 9600 );
	SIO_SetDataBits( si, 8 );
	SIO_SetParity( si, SIO_PARITY_EVEN );
	SIO_SetStopBits( si, 2 );
	SIO_RaiseRTS( si ); /* +12V */
	SIO_RaiseDTR( si ); /* +12V */
	SIO_SetIOMode( si, SIO_IOMODE_DIRECT );

	SIO_WriteSettings( si );

	/* Give it some time to save some electrons. */
#ifdef __palmos__
	sleep(2);
#elif defined(WINDOWS)
	Sleep(1000);
#else
	sleep(1);
#endif /* __palmos__ */

	/* Get unwanted bytes */
#ifdef __palmos__
	SIO_Flush(si);
#else
	while(SIO_ReadChar(si)!=-1);
#endif /* __palmos__ */

	/* Send command */
	if( SIO_WriteBuffer( si, "\x00\x01", 2 ) != 2 ) {
		SIO_Close( si );
		return( SC_EXIT_IO_ERROR );
	}

	/* Read answer */
	if( SIO_ReadBuffer( si, buffer, 3 ) != 3 ) {
		SIO_Close( si );
		return( SC_EXIT_OK );
	}

	SIO_Close( si );

	/* Check checksum */
	cs=0x01;

	for( i=0; i<3-1; i++ ) scTowitokoUpdateCS( cs , buffer[i] );

	if( cs != buffer[2] ) return( SC_EXIT_OK );

	/* Set variables */
	if( (buffer[0]==SC_TOWITOKO_CHIPDRIVE_TWIN) || 
		(buffer[0]==SC_TOWITOKO_CHIPDRIVE_EXTERN) || 
		(buffer[0]==SC_TOWITOKO_CHIPDRIVE_INTERN) || 
		(buffer[0]==SC_TOWITOKO_KARTENZWERG) || 
		(buffer[0]==SC_TOWITOKO_CHIPDRIVE_MICRO) || 
		(buffer[0]==0x64) /* Some other Chipdrive Micro. */ ) {
        if( buffer[0]==0x64 ) buffer[0]=SC_TOWITOKO_CHIPDRIVE_MICRO;

		*major=SC_READER_TOWITOKO;
		*minor=buffer[0];
		/* Compared to random data it has a false identification of at most
		 * 1:13000. In real life false identifications should be much less
		 * common. It could be improved by sending some more commands to
		 * the reader.
		 */
		*prob=200;
	}

	return( SC_EXIT_OK );
}

int main( int argc, char *argv[] )
{
	char *ports[]={"0","1","2","3"};
	int i, ret;
	BYTE prob;
	int major, minor;

	for( i=0; i<4; i++ ) {
		printf( "Port %d:\n", i );

		ret=scAcr20Detect( ports[i], &prob, &major, &minor );
		if( (ret==SC_EXIT_OK) && (prob!=0) ) 
			printf( "  ACR20: Probability %d, Major %d, Minor %d\n", prob,
				major, minor );

		ret=scB1Detect( ports[i], &prob, &major, &minor );
		if( (ret==SC_EXIT_OK) && (prob!=0) ) 
			printf( "  Siemens B1: Probability %d, Major %d, Minor %d\n", prob,
				major, minor );

		ret=scCtapiDetect( ports[i], &prob, &major, &minor );
		if( (ret==SC_EXIT_OK) && (prob!=0) ) 
			printf( "  CT-API: Probability %d, Major %d, Minor %d\n", prob,
				major, minor );

		ret=scDumbmouseDetect( ports[i], &prob, &major, &minor );
		if( (ret==SC_EXIT_OK) && (prob!=0) ) 
			printf( "  Dumb Mouse: Probability %d, Major %d, Minor %d\n", prob,
				major, minor );

		ret=scReflex60Detect( ports[i], &prob, &major, &minor );
		if( (ret==SC_EXIT_OK) && (prob!=0) ) 
			printf( "  Reflex 60: Probability %d, Major %d, Minor %d\n", prob,
				major, minor );

		ret=scTowitokoDetect( ports[i], &prob, &major, &minor );
		if( (ret==SC_EXIT_OK) && (prob!=0) ) 
			printf( "  Towitoko: Probability %d, Major %d, Minor %d\n", prob,
				major, minor );
	}

	return(0);
}



