/****************************************************************************
*																			*
*					SCEZ chipcard library - Cryptoflex routines				*
*						Copyright Matthias Bruestle 1999					*
*					For licensing, see the file COPYRIGHT					*
*																			*
****************************************************************************/

/* $Id: sccryptoflex.h,v 1.7 2000/09/01 15:31:27 zwiebeltu Exp $ */

#ifndef SC_CRYPTOFLEX_H
#define SC_CRYPTOFLEX_H

#ifndef SC_GENERAL_H
#error scgeneral.h needs to be included before this.
#endif

#define SC_CRYPTOFLEX_PIN_SIZE			0x08
#define	SC_CRYPTOFLEX_CHALL_SIZE		0x08
#define	SC_CRYPTOFLEX_MAC_SIZE			0x06
#define	SC_CRYPTOFLEX_DES_SIZE			0x08
#define	SC_CRYPTOFLEX_RSA_SIZE			0x80

/* Usefull defines for command parameters */

#define	SC_CRYPTOFLEX_ALGO_DES			0x00	/* SC_MULTIFLEX_ALGO_DES */
#define	SC_CRYPTOFLEX_ALGO_3DES			0x02

#define	SC_CRYPTOFLEX_MODE_DECRYPT		0x00
#define	SC_CRYPTOFLEX_MODE_ENCRYPT		0x01

#define	SC_CRYPTOFLEX_FILE_TRANSPARENT	0x01
#define	SC_CRYPTOFLEX_FILE_FIXED		0x02
#define	SC_CRYPTOFLEX_FILE_VARIABLE		0x04
#define	SC_CRYPTOFLEX_FILE_CYCLIC		0x06
#define	SC_CRYPTOFLEX_FILE_DIRECTORY	0x38

#define	SC_CRYPTOFLEX_BLOCKED			0x00
#define	SC_CRYPTOFLEX_UNBLOCKED			0x01

#define	SC_CRYPTOFLEX_ACCESS_ALWAYS		0x00
#define	SC_CRYPTOFLEX_ACCESS_PIN		0x01
#define	SC_CRYPTOFLEX_ACCESS_PROT		0x03
#define	SC_CRYPTOFLEX_ACCESS_AUTH		0x04
#define	SC_CRYPTOFLEX_ACCESS_PIN_PROT	0x06
#define	SC_CRYPTOFLEX_ACCESS_PIN_AUTH	0x08
#define	SC_CRYPTOFLEX_ACCESS_NEVER		0x0F

#define	SC_CRYPTOFLEX_RECORD_FIRST		0x00
#define	SC_CRYPTOFLEX_RECORD_LAST		0x01
#define	SC_CRYPTOFLEX_RECORD_NEXT		0x02
#define	SC_CRYPTOFLEX_RECORD_PREV		0x03
#define	SC_CRYPTOFLEX_RECORD_CURR_INDEX	0x04

#define	SC_CRYPTOFLEX_FID_PIN			0x0000
#define	SC_CRYPTOFLEX_FID_INTAUTH		0x0001
#define	SC_CRYPTOFLEX_FID_SERIAL		0x0002
#define	SC_CRYPTOFLEX_FID_EXTAUTH		0x0011
#define	SC_CRYPTOFLEX_FID_SECKEY		0x0012
#define	SC_CRYPTOFLEX_FID_PUBKEY		0x1012
#define	SC_CRYPTOFLEX_FID_MF			0x3F00

#define	SC_CRYPTOFLEX_OFFSET_TRANSPKEY	15

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/* Initialize card function pointer */
int scCryptoflexInit( SC_CARD_INFO *ci );

/* Capabilities */
int scCryptoflexGetCap( SC_CARD_INFO *ci, SC_CARD_CAP *cp );

/* Fill card data in ci */
int scCryptoflexGetCardData( SC_CARD_INFO *ci );

/* Set F and D. */
int scCryptoflexSetFD( SC_READER_INFO *ri, SC_CARD_INFO *ci, LONG fd );

#ifdef WITH_DES

/* Generate DES Auth */
void scCryptoflexGenerateAuth( BYTE *key, BYTE *chall, BYTE *auth,
	BYTE algo );

/* Compare DES Auth */
BOOLEAN scCryptoflexCompareAuth( BYTE *key, BYTE *chall, BYTE *auth,
	BYTE algo );

/* Generates DES MAC */
/* Data has to be:
 * INS | P1 | P2 | Lc (without cryptogram) | Data
 * Overall it can be at most 24 bytes.
 */
void scCryptoflexGenerateMAC( BYTE *key, BYTE *chall, BYTE *data,
	int datalen, BYTE *mac, int *maclen, BYTE algo );

#endif /* WITH_DES */

/* Commands */

int scCryptoflexCmdChangePIN( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *oldpin, BYTE *newpin );

int scCryptoflexCmdCreateFile( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	WORD fid, int flen, BYTE ftype, BYTE init, BYTE status, BYTE reclen,
	BYTE recnum, BYTE *acond, BYTE *akeys );

#ifdef WITH_DES
int scCryptoflexCmdCreateFileMAC( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	WORD fid, int flen, BYTE ftype, BYTE init, BYTE status, BYTE reclen,
	BYTE recnum, BYTE *acond, BYTE *akeys, BYTE *key, BYTE *chall, BYTE algo );
#endif /* WITH_DES */

int scCryptoflexCmdCreateRecord( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *data, int datalen );

#ifdef WITH_DES
int scCryptoflexCmdCreateRecordMAC( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *data, int datalen, BYTE *key, BYTE *chall, BYTE algo );
#endif /* WITH_DES */

int scCryptoflexCmdDecrease( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *amount, BYTE *resp, int *resplen );

#ifdef WITH_DES
int scCryptoflexCmdDecreaseMAC( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *amount, BYTE *key, BYTE *chall, BYTE algo, BYTE *resp, int *resplen );
#endif /* WITH_DES */

int scCryptoflexCmdDeleteFile( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	WORD fid );

#ifdef WITH_DES
int scCryptoflexCmdDeleteFileMAC( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	WORD fid, BYTE *key, BYTE *chall, BYTE algo );
#endif /* WITH_DES */

int scCryptoflexCmdDirectory( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *resp, int *resplen );

#ifdef WITH_DES
int scCryptoflexCmdExtAuth( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE keynum, BYTE *key, BYTE *chall, BYTE algo );
#endif /* WITH_DES */

int scCryptoflexCmdGetChall( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *chall, int *len );

/* sw is used to get the length of the available data. */
int scCryptoflexCmdGetResp( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *resp, int *resplen );

int scCryptoflexCmdIncrease( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *amount, BYTE *resp, int *resplen );

#ifdef WITH_DES
int scCryptoflexCmdIncreaseMAC( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *amount, BYTE *key, BYTE *chall, BYTE algo, BYTE *resp, int *resplen );
#endif /* WITH_DES */

#ifdef WITH_DES
int scCryptoflexCmdIntAuth( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE keynum, BYTE *chall, BYTE *key, BYTE algo );
#endif /* WITH_DES */

int scCryptoflexCmdInvalidate( SC_READER_INFO *ri, SC_CARD_INFO *ci );

#ifdef WITH_DES
int scCryptoflexCmdInvalidateMAC( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *key, BYTE *chall, BYTE algo );
#endif /* WITH_DES */

int scCryptoflexCmdReadBinary( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	int offset, BYTE *data, int *datalen );

int scCryptoflexCmdReadRecord( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE recnum, BYTE mode, BYTE *data, int *datalen );

int scCryptoflexCmdRehabilitate( SC_READER_INFO *ri, SC_CARD_INFO *ci );

#ifdef WITH_DES
int scCryptoflexCmdRehabilitateMAC( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *key, BYTE *chall, BYTE algo );
#endif /* WITH_DES */

int scCryptoflexCmdSeek( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE offset, BYTE mode, BYTE *pattern, BYTE patlen );

int scCryptoflexCmdSelectFile( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	WORD fid, BYTE *resp, int *resplen );

int scCryptoflexCmdUnblockPIN( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *unblock, BYTE *newpin );

int scCryptoflexCmdUpdateBinary( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	int offset, BYTE *data, BYTE datalen );

#ifdef WITH_DES
int scCryptoflexCmdUpdateBinaryMAC( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	int offset, BYTE *data, BYTE datalen, BYTE *key, BYTE *chall, BYTE algo );
#endif /* WITH_DES */

int scCryptoflexCmdUpdateRecord( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE recnum, BYTE mode, BYTE *data, BYTE datalen );

#ifdef WITH_DES
int scCryptoflexCmdUpdateRecordMAC( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE recnum, BYTE mode, BYTE *data, BYTE datalen, BYTE *key, BYTE *chall,
	BYTE algo );
#endif /* WITH_DES */

int scCryptoflexCmdVerifyKey( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE keynum, BYTE *key, BYTE keylen );

int scCryptoflexCmdVerifyPIN( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *pin );

/* IntAuth with RSA Key */
int scCryptoflexCmdRsaSign( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE keynum, BYTE *data, BYTE *resp, int *resplen );

/* Full DES IntAuth */
int scCryptoflexCmdDesCrypt( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE keynum, BYTE mode, BYTE *data, BYTE *resp, int *resplen );

int scCryptoflexCmdLoadCert( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE keynum, BYTE *data );

int scCryptoflexCmdVerifyPubKey( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *data );

int scCryptoflexCmdVerifyData( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE keynum, BYTE *data, BYTE *resp, int *resplen );

/* Waits 4 minutes for key generation and issues than a Get Response. */
int scCryptoflexCmdRsaKeyGen( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE keynum, BYTE *resp, int *resplen );

#ifdef WITH_DES
int scCryptoflexCmdUpdateEnc( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	int offset, BYTE *data, BYTE datalen, BYTE *key, BYTE algo );
#endif /* WITH_DES */

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* SC_CRYPTOFLEX_H */

