/****************************************************************************
*																			*
*					SCEZ chipcard library - GSM SIM routines				*
*						Copyright Matthias Bruestle 1999					*
*					For licensing, see the file COPYRIGHT					*
*																			*
****************************************************************************/

/* $Id: scgsmsim.h,v 1.4 2000/09/01 15:31:27 zwiebeltu Exp $ */

#ifndef SC_GSMSIM_H
#define SC_GSMSIM_H

#ifndef SC_GENERAL_H
#error scgeneral.h needs to be included before this.
#endif

/* Usefull defines for command parameters */

#define	SC_GSMSIM_RECMODE_NEXT	0x02
#define	SC_GSMSIM_RECMODE_PREV	0x03
#define	SC_GSMSIM_RECMODE_ABS	0x04
#define	SC_GSMSIM_RECMODE_CUR	0x04

#define	SC_GSMSIM_RECNUM_CUR	0x00

#define	SC_GSMSIM_SEEKMODE_BEG	0x00
#define	SC_GSMSIM_SEEKMODE_END	0x01
#define	SC_GSMSIM_SEEKMODE_NEXT	0x02
#define	SC_GSMSIM_SEEKMODE_PREV	0x03

#define	SC_GSMSIM_CHV1			0x01
#define	SC_GSMSIM_CHV2			0x02

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/* Initialize card function pointer */
int scGsmsimInit( SC_CARD_INFO *ci );

/* Capabilities */
int scGsmsimGetCap( SC_CARD_INFO *ci, SC_CARD_CAP *cp );

/* Fill card data in ci */
int scGsmsimGetCardData( SC_CARD_INFO *ci );

/* Set F and D. */
int scGsmsimSetFD( SC_READER_INFO *ri, SC_CARD_INFO *ci, LONG fd );

/* Misc */

/* Translate character set. */
char scGsmsimGsmToIso( char c );
char scGsmsimIsoToGsm( char c );

/* Packs string as 7bit characters into octets. */
int scGsmsimPackStr( char *c, int clen, BYTE *b, int *blen );

/* Unpacks 7bit characters. */
int scGsmsimUnpackStr( BYTE *b, char *c, int clen );

/* Commands */

int scGsmsimCmdSelect( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	WORD fid, BYTE *resp, int *resplen );

int scGsmsimCmdStatus( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *resp, int *resplen );

int scGsmsimCmdReadBin( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	WORD offset, BYTE *data, int *datalen );

int scGsmsimCmdUpdateBin( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	WORD offset, BYTE *data, BYTE datalen );

int scGsmsimCmdReadRec( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE recnum, BYTE mode, BYTE *data, int *datalen );

int scGsmsimCmdUpdateRec( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE recnum, BYTE mode, BYTE *data, BYTE datalen );

int scGsmsimCmdSeek( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BOOLEAN resp, BYTE mode, BYTE *pattern, BYTE patlen, BYTE *num );

int scGsmsimCmdIncrease( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *amount, BYTE *resp, int *resplen );

int scGsmsimCmdVerifyChv( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE num, BYTE *pin );

int scGsmsimCmdChangeChv( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE num, BYTE *oldpin, BYTE *newpin );

int scGsmsimCmdDisableChv( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *pin );

int scGsmsimCmdEnableChv( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *pin );

int scGsmsimCmdUnblockChv( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE num, BYTE *unblockpin, BYTE *newpin );

int scGsmsimCmdInvalidate( SC_READER_INFO *ri, SC_CARD_INFO *ci );

int scGsmsimCmdRehabilitate( SC_READER_INFO *ri, SC_CARD_INFO *ci );

int scGsmsimCmdRunGsmAlgo( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *rand, BYTE *sres, BYTE *kc );

int scGsmsimCmdSleep( SC_READER_INFO *ri, SC_CARD_INFO *ci );

/* sw is used to get the length of the available data. */
int scGsmsimCmdGetResp( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *resp, int *resplen );

int scGsmsimCmdTermProf( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *prof, BYTE proflen );

int scGsmsimCmdEnvelope( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *comd, BYTE comdlen, BYTE *resp, int *resplen );

/* sw is used to get the length of the available data. */
int scGsmsimCmdFetch( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *resp, int *resplen );

int scGsmsimCmdTermResp( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *comd, BYTE comdlen );

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* SC_GSMSIM_H */

