/****************************************************************************
*																			*
*					SCEZ chipcard library - Multiflex routines				*
*						Copyright Matthias Bruestle 1999					*
*					For licensing, see the file COPYRIGHT					*
*																			*
****************************************************************************/

/* $Id: scmultiflex.h,v 1.5 2000/09/01 15:31:27 zwiebeltu Exp $ */

#ifndef SC_MULTIFLEX_H
#define SC_MULTIFLEX_H

#ifndef SC_GENERAL_H
#error scgeneral.h needs to be included before this.
#endif

#define SC_MULTIFLEX_PIN_SIZE	8
#define	SC_MULTIFLEX_CHALL_SIZE	8
#define	SC_MULTIFLEX_MAC_SIZE	6

/* Usefull defines for command parameters */

#define	SC_MULTIFLEX_ALGO_DES			0x00	/* SC_CRYPTOFLEX_ALGO_DES */

#define	SC_MULTIFLEX_FILE_TRANSPARENT	0x01
#define	SC_MULTIFLEX_FILE_FIXED			0x02
#define	SC_MULTIFLEX_FILE_VARIABLE		0x04
#define	SC_MULTIFLEX_FILE_CYCLIC		0x06
#define	SC_MULTIFLEX_FILE_DIRECTORY		0x38

#define	SC_MULTIFLEX_BLOCKED			0x00
#define	SC_MULTIFLEX_UNBLOCKED			0x01

#define	SC_MULTIFLEX_ACCESS_ALWAYS		0x00
#define	SC_MULTIFLEX_ACCESS_PIN			0x01
#define	SC_MULTIFLEX_ACCESS_PROT		0x03
#define	SC_MULTIFLEX_ACCESS_AUTH		0x04
#define	SC_MULTIFLEX_ACCESS_PIN_PROT	0x06
#define	SC_MULTIFLEX_ACCESS_PIN_AUTH	0x08
#define	SC_MULTIFLEX_ACCESS_NEVER		0x0F

#define	SC_MULTIFLEX_RECORD_FIRST		0x00
#define	SC_MULTIFLEX_RECORD_LAST		0x01
#define	SC_MULTIFLEX_RECORD_NEXT		0x02
#define	SC_MULTIFLEX_RECORD_PREV		0x03
#define	SC_MULTIFLEX_RECORD_CURR_INDEX	0x04

#define	SC_MULTIFLEX_FID_PIN			0x0000
#define	SC_MULTIFLEX_FID_INTAUTH		0x0001
#define	SC_MULTIFLEX_FID_SERIAL			0x0002
#define	SC_MULTIFLEX_FID_EXTAUTH		0x0011
#define	SC_MULTIFLEX_FID_MF				0x3F00

#define	SC_MULTIFLEX_OFFSET_TRANSPKEY	15

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/* Initialize card function pointer */
int scMultiflexInit( SC_CARD_INFO *ci );

/* Capabilities */
int scMultiflexGetCap( SC_CARD_INFO *ci, SC_CARD_CAP *cp );

/* Fill card data in ci */
int scMultiflexGetCardData( SC_CARD_INFO *ci );

/* Set F and D. */
int scMultiflexSetFD( SC_READER_INFO *ri, SC_CARD_INFO *ci, LONG fd );

#ifdef WITH_DES

/* Generate DES Auth */
void scMultiflexGenerateAuth( BYTE *key, BYTE *chall, BYTE *auth );

/* Compare DES Auth */
BOOLEAN scMultiflexCompareAuth( BYTE *key, BYTE *chall, BYTE *auth );

/* Generates DES MAC */
/* Data has to be:
 * INS | P1 | P2 | Lc (without cryptogram) | Data
 * Overall it can be at most 24 bytes.
 */
void scMultiflexGenerateMAC( BYTE *key, BYTE *chall, BYTE *data,
	int datalen, BYTE *mac, int *maclen );

#endif /* WITH_DES */

/* Commands */

int scMultiflexCmdChangePIN( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *oldpin, BYTE *newpin );

int scMultiflexCmdCreateFile( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	WORD fid, int flen, BYTE ftype, BYTE init, BYTE status,
	BYTE reclen, BYTE recnum, BYTE *acond, BYTE *akeys );

#ifdef WITH_DES
int scMultiflexCmdCreateFileMAC( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	WORD fid, int flen, BYTE ftype, BYTE init, BYTE status, BYTE reclen,
	BYTE recnum, BYTE *acond, BYTE *akeys, BYTE *key, BYTE *chall );
#endif /* WITH_DES */

int scMultiflexCmdCreateRecord( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *data, int datalen );

#ifdef WITH_DES
int scMultiflexCmdCreateRecordMAC( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *data, int datalen, BYTE *key, BYTE *chall );
#endif /* WITH_DES */

int scMultiflexCmdDecrease( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *amount, BYTE *resp, int *resplen );

#ifdef WITH_DES
int scMultiflexCmdDecreaseMAC( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *amount, BYTE *key, BYTE *chall, BYTE *resp, int *resplen );
#endif /* WITH_DES */

int scMultiflexCmdDecreaseSt( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *amount, BYTE *resp, int *resplen );

#ifdef WITH_DES
int scMultiflexCmdDecreaseStMAC( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *amount, BYTE *key, BYTE *chall, BYTE *resp, int *resplen );
#endif /* WITH_DES */

int scMultiflexCmdDeleteFile( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	WORD fid );

#ifdef WITH_DES
int scMultiflexCmdDeleteFileMAC( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	WORD fid, BYTE *key, BYTE *chall );
#endif /* WITH_DES */

int scMultiflexCmdDirectory( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *resp, int *resplen );

#ifdef WITH_DES
int scMultiflexCmdExtAuth( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE keynum, BYTE *key, BYTE *chall );
#endif /* WITH_DES */

int scMultiflexCmdGetChall( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *chall );

int scMultiflexCmdGiveChall( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *chall );

/* sw is used to get the length of the available data. */
int scMultiflexCmdGetResp( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *resp, int *resplen );

int scMultiflexCmdIncrease( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *amount, BYTE *resp, int *resplen );

#ifdef WITH_DES
int scMultiflexCmdIncreaseMAC( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *amount, BYTE *key, BYTE *chall, BYTE *resp, int *resplen );
#endif /* WITH_DES */

int scMultiflexCmdIncreaseSt( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *amount, BYTE *resp, int *resplen );

#ifdef WITH_DES
int scMultiflexCmdIncreaseStMAC( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *amount, BYTE *key, BYTE *chall, BYTE *resp, int *resplen );
#endif /* WITH_DES */

#ifdef WITH_DES
int scMultiflexCmdIntAuth( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE keynum, BYTE *chall, BYTE *key );
#endif /* WITH_DES */

int scMultiflexCmdInvalidate( SC_READER_INFO *ri, SC_CARD_INFO *ci );

#ifdef WITH_DES
int scMultiflexCmdInvalidateMAC( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *key, BYTE *chall );
#endif /* WITH_DES */

int scMultiflexCmdReadBinary( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	int offset, BYTE *data, int *datalen );

int scMultiflexCmdReadRecord( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE recnum, BYTE mode, BYTE *data, int *datalen );

int scMultiflexCmdRehabilitate( SC_READER_INFO *ri, SC_CARD_INFO *ci );

#ifdef WITH_DES
int scMultiflexCmdRehabilitateMAC( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *key, BYTE *chall );
#endif /* WITH_DES */

int scMultiflexCmdSeek( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE offset, BYTE mode, BYTE *pattern, BYTE patlen );

int scMultiflexCmdSelectFile( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	WORD fid, BYTE *resp, int *resplen );

int scMultiflexCmdUnblockPIN( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *unblock, BYTE *newpin );

int scMultiflexCmdUpdateBinary( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	int offset, BYTE *data, BYTE datalen );

#ifdef WITH_DES
int scMultiflexCmdUpdateBinaryMAC( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	int offset, BYTE *data, BYTE datalen, BYTE *key, BYTE *chall );
#endif /* WITH_DES */

int scMultiflexCmdUpdateRecord( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE recnum, BYTE mode, BYTE *data, BYTE datalen );

#ifdef WITH_DES
int scMultiflexCmdUpdateRecordMAC( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE recnum, BYTE mode, BYTE *data, BYTE datalen, BYTE *key, BYTE *chall );
#endif /* WITH_DES */

int scMultiflexCmdVerifyKey( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE keynum, BYTE *key, BYTE keylen );

int scMultiflexCmdVerifyPIN( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *pin );

#ifdef WITH_DES
int scMultiflexCmdDesCrypt( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE keynum, BYTE mode, BYTE *data, BYTE *resp, int *resplen );
#endif /* WITH_DES */

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* SC_MULTIFLEX_H */

