/****************************************************************************
*																			*
*					SCEZ chipcard library - TCOS 2.0 routines				*
*						Copyright Matthias Bruestle 1999					*
*					For licensing, see the file COPYRIGHT					*
*																			*
****************************************************************************/

/* $Id: sctcos.h,v 1.5 2000/09/01 15:31:27 zwiebeltu Exp $ */

#ifndef SC_TCOS_H
#define SC_TCOS_H

#ifndef SC_GENERAL_H
#error scgeneral.h needs to be included before this.
#endif

/* Usefull defines for command parameters */

/* Change password */

#define SC_TCOS_NULLPIN				"\x00\x00\x00\x00\x00\x00"
#define SC_TCOS_NULLPINLEN			6

/* List directory */

#define SC_TCOS_LIST_DF				0x01
#define SC_TCOS_LIST_EF				0x02

/* Manage security environment */

#define SC_TCOS_MSE_APP_SM_APDU		0x11
#define SC_TCOS_MSE_APP_SM_RAPDU	0x21
#define SC_TCOS_MSE_APP_PSO			0x41

#define SC_TCOS_MSE_OP_MAC			0xB4
#define SC_TCOS_MSE_OP_ENCRYPT		0xB8

#define SC_TCOS_MSE_ALGO_IDEA		0x04
#define SC_TCOS_MSE_ALGO_DES		0x08
#define SC_TCOS_MSE_ALGO_DES3		0x0C
#define SC_TCOS_MSE_ALGO_RSA		0x10

#define SC_TCOS_MSE_MODE_ECB		0x00
#define SC_TCOS_MSE_MODE_CBC		0x01

/* Perform security operation */

#define SC_TCOS_PSO_OUT_NONE		0x00	/* Verify hash */
#define SC_TCOS_PSO_OUT_DEC_TXT		0x80	/* Decrypt */
#define SC_TCOS_PSO_OUT_ENC_TXT_TLV	0x84	/* Encrypt */
#define SC_TCOS_PSO_OUT_ENC_TXT		0x86	/* Encrypt */
#define SC_TCOS_PSO_OUT_CKS			0x8E	/* Calculate hash */
#define SC_TCOS_PSO_OUT_SIGN		0x9E	/* Sign */

#define SC_TCOS_PSO_IN_CLEAR_TXT	0x80	/* Calculate hash/Encrypt */
#define	SC_TCOS_PSO_IN_ENC_TXT_TLV	0x84	/* Decrypt */
#define	SC_TCOS_PSO_IN_ENC_TXT		0x86	/* Decrypt */
#define	SC_TCOS_PSO_IN_HASH			0x9A	/* Sign */
#define	SC_TCOS_PSO_IN_VER_TEMPLATE	0xA2	/* Verify hash */

/* Read/Update record */

#define SC_TCOS_RECORD_FIRST		0x00
#define SC_TCOS_RECORD_LAST			0x01
#define SC_TCOS_RECORD_NEXT			0x02
#define SC_TCOS_RECORD_PREV			0x03
#define SC_TCOS_RECORD_ABS			0x04

/* Select */

#define	SC_TCOS_FID_MF				0x3F00

#define SC_TCOS_SELECT_MF			0x00
#define SC_TCOS_SELECT_DF			0x01
#define SC_TCOS_SELECT_EF			0x02
#define SC_TCOS_SELECT_PARENT		0x03
#define SC_TCOS_SELECT_AID			0x04
#define SC_TCOS_SELECT_ABS_PATH		0x08
#define SC_TCOS_SELECT_REL_PATH		0x09

#define SC_TCOS_RDATA_FCI			0x00
#define SC_TCOS_RDATA_FCP			0x04
#define SC_TCOS_RDATA_FMD			0x08
#define SC_TCOS_RDATA_NONE			0x0C

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/* Initialize card function pointer */
int scTcosInit( SC_CARD_INFO *ci );

/* Capabilities */
int scTcosGetCap( SC_CARD_INFO *ci, SC_CARD_CAP *cp );

/* Fill card data in ci */
int scTcosGetCardData( SC_CARD_INFO *ci );

/* Set F and D. */
int scTcosSetFD( SC_READER_INFO *ri, SC_CARD_INFO *ci, LONG fd );

int scTcosCmdAskRandom( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE *rand, int *len );

int scTcosCmdChangePassword( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BOOLEAN local, BYTE num, BYTE *oldpass, BYTE oldlen, BYTE *newpass,
	BYTE newlen );

int scTcosCmdListDir( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE ftype, BYTE start, BYTE *resp, int *resplen );

int scTcosCmdManageSecEnv( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE app, BYTE op, BYTE *data, BYTE datalen );

int scTcosCmdPerformSecOp( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE in, BYTE out, BYTE *data, BYTE datalen, BYTE *resp, int *resplen );

/* offset can be sfid/soffset (highest bit set); */
int scTcosCmdReadBinary( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	WORD offset, BYTE *data, int *datalen );

/* Uses sfid if it is in the correct range. */
int scTcosCmdReadRecord( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE rec, BYTE addr, BYTE sfid, BYTE *data, int *datalen );

int scTcosCmdSelect( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE type, WORD fid, BYTE *aidpath, BYTE aidpathlen, BYTE rtype,
	BYTE *resp, int *resplen );

/* offset can be sfid/soffset (highest bit set); */
int scTcosCmdUpdateBinary( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	WORD offset, BYTE *data,  BYTE datalen );

/* Uses sfid if it is in the correct range. */
int scTcosCmdUpdateRecord( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BYTE rec, BYTE addr, BYTE sfid, BYTE *data, BYTE datalen );

int scTcosCmdVerifyPassword( SC_READER_INFO *ri, SC_CARD_INFO *ci,
	BOOLEAN local, BYTE num, BYTE *pass, BYTE passlen );

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* SC_TCOS_H */


