/*
 * ****************************************************************
 * $Source: /home/CVS/Sieve/main.c,v $
 * $Author: hmueller $
 * $Date: 1999/10/20 13:31:51 $
 * $Revision: 1.1.1.1 $
 * ****************************************************************
 * OS       : Palm OS 1.0
 * Target   : Palm Pilot I, II, III, V
 * Compiler : prc-tools Version 0.5.0 (GCC 2.7, EGCS-2.91)
 * Copyright: Holger Mueller <hmueller@bigfoot.com>
 *            http://hmueller.purespace.de/pilot
 * Notes    : Sieve benchmark
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the license, or (at
 * your option) any later version.
 *    
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ****************************************************************
 */

/* Pilot header files. */
#include <Pilot.h>

/* Local header files. */
#include "Sieve.h"				// ID's of all resources
#include "main.h"				// main def's


#ifndef FALSE
#  define FALSE false
#endif
#ifndef TRUE
#  define TRUE true
#endif
#define SIEVE_BUF_SIZE 8192		// size of sieve buffer


/*
 * Global variables
 */
static Long timeout = -1;		// timeout for EvtGetEvent()
char flags[SIEVE_BUF_SIZE + 1];


/*
 * ===============================================================
 * Function : Sieve
 * ---------------------------------------------------------------
 * Purpose  : Calculate sieve bechmark
 * Call     : time = Sieve();
 * Author   : Holger Mueller
 * ===============================================================
 * Parameter:
 * ---------------------------------------------------------------
 * VAR             I/O TYPE    REMARK
 * ---------------------------------------------------------------
 * time				O  long    time of sieve benchmark
 * ===============================================================
 */
long Sieve(void)
{
  register int i,
	k, n,
	prime,
	count;
  ULong time;
  UInt auto_off_time;
  FormPtr form = FrmGetActiveForm();

  CtlShowControl((ControlPtr)FrmGetObjectPtr(form, \
    FrmGetObjectIndex(form, ID_POPUP)));
  auto_off_time = SysSetAutoOffTime(0);	// let Pilot stay alive
  time = TimGetTicks();

  for (n = 0; n < 100; n++) {
	count = 0;
	for (i = 0; i <= SIEVE_BUF_SIZE; i++)
	  flags[i] = TRUE;
	for (i = 0; i <= SIEVE_BUF_SIZE; i++) {
	  if (flags[i]) {
		prime = i + i + 3;
		for (k = i + prime; k <= SIEVE_BUF_SIZE; k += prime) 
		  flags[k] = FALSE;
		count++;
	  }
	}
  }

  time = TimGetTicks() - time;
  time *= 1000L / (long)SysTicksPerSecond(); // ticks -> ms
  SysSetAutoOffTime(auto_off_time);	// reset old auto off time
  CtlHideControl((ControlPtr)FrmGetObjectPtr(form, \
    FrmGetObjectIndex(form, ID_POPUP)));

  return(time);
} /* Sieve */


/*
 * ===============================================================
 * Function : DisplValue
 * ---------------------------------------------------------------
 * Purpose  : Display all values of the main form
 * Call     : DisplValue(form, label_id, value);
 * Author   : Holger Mueller
 * ===============================================================
 * Parameter:
 * ---------------------------------------------------------------
 * VAR             I/O TYPE    REMARK
 * ---------------------------------------------------------------
 * form             I  FormPtr aktuelles form
 * label_id         I  Word	   ID es Labels
 * value            I  long    neuer Wert
 * ===============================================================
 */
void DisplValue(FormPtr form, Word label_id, long value)
{
  char msg[32];					// help string
  RectangleType r;

  // Ausgabe vorher loeschen!
  FrmGetObjectPosition(form, FrmGetObjectIndex(form, label_id), \
					   &r.topLeft.x, &r.topLeft.y);
  r.extent.x = 5 * FntAverageCharWidth();
  r.extent.y = FntCharHeight();
  WinEraseRectangle(&r, 0);

  StrIToA(msg, value);
  // ACHTUNG: "msg" darf nicht laenger sein, als im rcp-File definiert!
  FrmCopyLabel(form, label_id, msg);

  return;
} /* DisplValue */


/*
 * ===============================================================
 * Function : MainFormHandleEvent
 * ---------------------------------------------------------------
 * Purpose  : Handle "Main" form
 * Call     : handled = MainFormHandleEvent(&e);
 * Author   : Holger Mueller
 * ===============================================================
 * Parameter:
 * ---------------------------------------------------------------
 * VAR             I/O TYPE    REMARK
 * ---------------------------------------------------------------
 * e			   I/O EventPtr event buffer
 * handled          O  Boolean TRUE: was handled by this function
 *                             FALSE: was NOT handled
 * ===============================================================
 */
static Boolean MainFormHandleEvent(EventPtr e)
{
  Boolean handled = FALSE;
  Boolean start_sieve = FALSE;

  switch (e->eType) {
  case frmOpenEvent:
	FrmDrawForm(FrmGetActiveForm());
	handled = TRUE;
	break;
  case ctlSelectEvent:
	switch (e->data.ctlSelect.controlID) {
	case ID_START_BUTTON:
	  start_sieve = TRUE;
	  handled = TRUE;
	  break;
	case ID_ABOUT_BUTTON:
	  FrmHelp(ID_ABOUT_MSG);
	  handled = TRUE;
	  break;
	}
	break;
  case nilEvent:
	break;
  default:
	break;
  }

  if (handled && start_sieve) {
	DisplValue(FrmGetActiveForm(), ID_TIME, Sieve());
	start_sieve = FALSE;
  }
  timeout = -1;

  return(handled);
} /* MainFormHandleEvent */


/*
 * ===============================================================
 * Function : ApplicationHandleEvent
 * ---------------------------------------------------------------
 * Purpose  : Handle all application events and set event handling 
 *            routines for the different forms
 * Call     : handled = ApplicationHandleEvent(&e);
 * Author   : Holger Mueller
 * ===============================================================
 * Parameter:
 * ---------------------------------------------------------------
 * VAR             I/O TYPE    REMARK
 * ---------------------------------------------------------------
 * e			   I/O EventPtr event buffer
 * handled          O  Boolean TRUE: was handled by this function
 *                             FALSE: was NOT handled
 * ===============================================================
 */
static Boolean ApplicationHandleEvent(EventPtr e)
{
  Boolean handled = FALSE;

  if (e->eType == frmLoadEvent) {
    Word frmId = e->data.frmLoad.formID;
    FormPtr frm = FrmInitForm(frmId);
    
    FrmSetActiveForm(frm);
    
    switch (frmId) {
    case ID_FRM_MAIN:
      FrmSetEventHandler(frm, MainFormHandleEvent);
      handled = true;
      break;
    }
  }

  return(handled);
}


/*
 * ===============================================================
 * Function : EventLook
 * ---------------------------------------------------------------
 * Purpose  : Main event lookup loop
 * Call     : ret = EventLook();
 * Author   : Holger Mueller
 * ===============================================================
 * Parameter:
 * ---------------------------------------------------------------
 * VAR             I/O TYPE    REMARK
 * ---------------------------------------------------------------
 * ret              O  Err     error status
 * ===============================================================
 */
Err EventLook(void)
{
  Err err = 0;					// error status
  EventType e;					// event structure

  timeout = -1;
  do {
	EvtGetEvent(&e, timeout);
    if (SysHandleEvent(&e))
      continue;
	if (MenuHandleEvent(NULL, &e, &err))
	  continue;
	if (ApplicationHandleEvent(&e))
	  continue;
	FrmDispatchEvent(&e);
  } while (e.eType != appStopEvent);

  return(err);
} /* EventLook */


/*
 * ===============================================================
 * Function : PilotMain
 * ---------------------------------------------------------------
 * Purpose  : 
 * Call     : by OS
 * Author   : Holger Mueller
 * ===============================================================
 * Parameter:
 * ---------------------------------------------------------------
 * VAR             I/O TYPE    REMARK
 * ---------------------------------------------------------------
 * ===============================================================
 */
DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
  Err err;

  if (cmd == sysAppLaunchCmdNormalLaunch) {
	FrmGotoForm(ID_FRM_MAIN);
	err = EventLook();
	if (err)
	  return(err);
    FrmCloseAllForms();
  }

  return(0);
} /* PilotMain */

#if 0
  StrIToA(msg, i);
  WinDrawChars(msg, StrLen(msg), 130, 20);
#endif
