/* Main code for Suck-O-Meter
 * $Id: suckometer.c,v 1.2 2000/10/15 08:19:34 jaq Exp $
 */


#include <Pilot.h>
#include "callback.h"

#include "suckometer.h"

struct preferences {
    Word mode;
} Preferences;

static Boolean CalcSuckage()
{
    Boolean sucks = false;

    switch (Preferences.mode) {
      case RadioOffMode:
	sucks = true;
	break;
      case RockHardMode:
	sucks = false;
	break;
      default:
	if (SysRandom(0) % 2)
	    sucks = true;
	break;
    }
    return sucks;
}

static void DrawBitmap(Word bid)
{
    VoidHand res = 0;
    BitmapPtr bmp = 0;

    res = DmGetResource(bitmapRsc, bid);
    bmp = MemHandleLock(res);
    WinDrawBitmap(bmp, 0, 20);
    MemPtrUnlock(bmp);
    DmReleaseResource(res);
}

static Boolean SucksHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr frm;

#ifdef __GNUC__
    CALLBACK_PROLOGUE;
#endif

    switch (event->eType) {
      case frmOpenEvent:
	frm = FrmGetActiveForm();
	FrmDrawForm(frm);
	if (CalcSuckage())
	    DrawBitmap(SuckBitmap);
	else
	    DrawBitmap(NoSuckBitmap);
	handled = true;
	break;
      case ctlSelectEvent:
	frm = FrmGetActiveForm();
	switch (event->data.ctlEnter.controlID) {
	  case AgainButton:
	    FrmReturnToForm(MainForm);
	    handled = true;
	    break;
	  default:
	    break;
	}
	break;
      default:
	break;
    }

#ifdef __GNUC__
    CALLBACK_EPILOGUE;
#endif

    return handled;
}

static Boolean PrefsHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr frm;
    ControlPtr ctl;

#ifdef __GNUC__
    CALLBACK_PROLOGUE;
#endif

    switch (event->eType) {
      case frmOpenEvent:
	frm = FrmGetActiveForm();
	FrmDrawForm(frm);
	ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, Preferences.mode));
	CtlSetValue(ctl, true);
	handled = true;
	break;
      case ctlSelectEvent:
	frm = FrmGetActiveForm();
	switch (event->data.ctlEnter.controlID) {
	  case PrefOKButton:
	    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, RadioOffMode));
	    if (CtlGetValue(ctl))
		Preferences.mode = RadioOffMode;
	    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, RockHardMode));
	    if (CtlGetValue(ctl))
		Preferences.mode = RockHardMode;
	    ctl = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, NormalMode));
	    if (CtlGetValue(ctl))
		Preferences.mode = NormalMode;
	    handled = true;
	    /* fall thru */
	  case PrefCanButton:
	    FrmReturnToForm(MainForm);
	    handled = true;
	    break;
	  default:
	    break;
	}
      default:
	break;
    }

#ifdef __GNUC__
    CALLBACK_EPILOGUE;
#endif

    return handled;
}

/*
static Boolean PreHandleEvent(EventPtr event) {
	Boolean handled = false;

#ifdef __GNUC__
	CALLBACK_PROLOGUE;
#endif
	if (event->eType == penUpEvent)
		if ((event->screenY > 168) && (event->screenX < 150) && (event->screenX > 10))
			handled = SucksHandleEvent(event);
	return handled;
	
#ifdef __GNUC__
	CALLBACK_EPILOGUE;
#endif
}
*/

static Boolean MainFormHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr frm;

#ifdef __GNUC__
    CALLBACK_PROLOGUE;
#endif

    switch (event->eType) {
      case frmOpenEvent:
	frm = FrmGetActiveForm();
	FrmDrawForm(frm);
	DrawBitmap(ArrowBitmap);
	handled = true;
	break;
      case menuEvent:
	MenuEraseStatus(NULL);
	switch (event->data.menu.itemID) {
	  case AbootMenu:
	    FrmAlert(AbootAlert);
	    handled = true;
	    break;
	  case PrefsMenu:
	    FrmPopupForm(PrefsForm);
	    handled = true;
	    break;
	  default:
	    break;
	}
	handled = true;
	break;
      case ctlSelectEvent:
	switch (event->data.ctlSelect.controlID) {
	  case SuckButton:
	    FrmPopupForm(SucksForm);
	    handled = true;
	    break;
	  default:
	    break;
	}
	break;
      default:
	break;
    }

#ifdef __GNUC__
    CALLBACK_EPILOGUE;
#endif

    return handled;
}

static Boolean ApplicationHandleEvent(EventPtr e)
{
    FormPtr frm;
    Word formId;
    Boolean handled = false;

    if (e->eType == frmLoadEvent) {
	formId = e->data.frmLoad.formID;
	frm = FrmInitForm(formId);
	FrmSetActiveForm(frm);

	switch (formId) {
	  case MainForm:
	    FrmSetEventHandler(frm, MainFormHandleEvent);
	    break;
	  case PrefsForm:
	    FrmSetEventHandler(frm, PrefsHandleEvent);
	    break;
	  case SucksForm:
	    FrmSetEventHandler(frm, SucksHandleEvent);
	    break;
	  default:
	    break;
	}
	handled = true;
    }
    return handled;
}

/* Get preferences, open (or create) app database */
static Word StartApplication(void)
{
    Word prefSize = sizeof(Preferences);

    Preferences.mode = NormalMode;
    PrefGetAppPreferences('SUKM', 0, &Preferences, &prefSize, false);
    FrmGotoForm(MainForm);
    return 0;
}

/* Save preferences, close forms, close app database */
static void StopApplication(void)
{
    FrmSaveAllForms();
    FrmCloseAllForms();
    PrefSetAppPreferences('SUKM', 0, 0x0001, &Preferences,
			  sizeof(Preferences), false);
}

/* The main event loop */
static void EventLoop(void)
{
    Word err;
    EventType e;

    do {
	EvtGetEvent(&e, evtWaitForever);
	/* if (!PreHandleEvent(&e)) */
	if (!SysHandleEvent(&e))
	    if (!MenuHandleEvent(NULL, &e, &err))
		if (!ApplicationHandleEvent(&e))
		    FrmDispatchEvent(&e);
    } while (e.eType != appStopEvent);
}

/* check the rom version of the palm device
 * taken from Tom Dyas' tictactoe 0.95 code -- thanx, Tom :) */
static Err RomVersionCompat(DWord reqdVer, Word launchFlags)
{
    DWord romVersion;
    Err err;

    FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
    if (romVersion < reqdVer) {
	if (
	    (launchFlags &
	     (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) ==
	    (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) {
	    FrmAlert(BadRomAlert);
	    /* Pilot 1.0 will continuously relaunch this app unless we switch to
	     * another safe one */
	    if (romVersion < 0x02000000)
		AppLaunchWithCommand(sysFileCDefaultApp,
				     sysAppLaunchCmdNormalLaunch, NULL);
	}
	return (sysErrRomIncompatible);
    }
    return (0);
}


/* Main entry point; it is unlikely you will need to change this except to
   handle other launch command codes */
DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
    Word err;

    err =
	RomVersionCompat(sysMakeROMVersion(2, 0, 0, sysROMStageRelease, 0),
			 launchFlags);
    if (err)
	return err;

    if (cmd == sysAppLaunchCmdNormalLaunch) {
	err = StartApplication();
	if (err)
	    return err;
	EventLoop();
	StopApplication();
    } else {
	return sysErrParamErr;
    }
    return 0;
}
