#include "titraxAll.h"

#define numCommands1 6
#define numCommands2 13

static Char comEntry[numCommands2][18] = {"Save All", "Clear All", 
														"Remove All",
														"Order Asc.", "Order Desc.",
														"Orig. Order", 
														"Change Time", "Shift Time", 
														"Clear", "Remove", "Edit Note", 
														"Center Active",
														"Change Properties"};

static CharPtr comEntry1P[numCommands1] = {comEntry[0], comEntry[1], 
														 comEntry[2], comEntry[3],
														 comEntry[4], comEntry[5]};

static CharPtr comEntry2P[numCommands2] = {comEntry[0], comEntry[1], 
														 comEntry[2], comEntry[3], 
														 comEntry[4], comEntry[5],
														 comEntry[6], comEntry[7],
														 comEntry[8], comEntry[9], 
														 comEntry[10], comEntry[11],
														 comEntry[12]};

static Int startRow = 0;
static Boolean inDnD = false;

Int
SlotCompare(void* pFirst, void* pSecond, Long direction)
{
  Int result;
  Int firstSlot;
  Int secondSlot;
  ULong firstElapsed;
  ULong secondElapsed;

  CALLBACK_PROLOGUE
  
  firstSlot = *((Int*)pFirst);
  secondSlot = *((Int*)pSecond);

  firstElapsed = Elapsed(firstSlot);
  secondElapsed = Elapsed(secondSlot);
  
  if ((Ignored(firstSlot)) && (!Ignored(secondSlot)))
	 result = direction;
  else if ((!Ignored(firstSlot)) && (Ignored(secondSlot)))
	 result = -direction;
  else {
	 if (firstElapsed > secondElapsed)
		result = -direction;
	 else if (firstElapsed < secondElapsed)
		result = direction;
	 else if (!TextIsEmpty(firstSlot) && TextIsEmpty(secondSlot))
		result = -direction;
	 else if (TextIsEmpty(firstSlot) && !TextIsEmpty(secondSlot))
		result = direction;
	 else
		result = 0;
  }

  CALLBACK_EPILOGUE

  return result;	 
}

void
SortSlots(Long direction)
{
  SysInsertionSort(RowToSlotAddr(), MAXSLOT, sizeof(Int), 
						 SlotCompare, direction);
}

void
DeActivateSlot(Int slot, Int update)
{
  if (slot == noSlotActive)
	 return;

  if (ActiveSlot() == slot) {
	 SetElapsed(slot, Elapsed(slot) + TimGetSeconds() - StartedAt(slot));
	 SetActiveSlot(noSlotActive);
	 SetStartedAt(slot, 0);
	 
	 if (DateBook(slot))
		// no datebook entries if time less than 3 minutes
		if ((TimGetSeconds() - StartedAt(MAXSLOT)) > 180)
		  // ask if entry lasts more than one month
		  if ((TimGetSeconds() - StartedAt(MAXSLOT)) > tendays*3) {
			 if (FrmAlert(titraxLargeEntryAlert) == 0)
				AddAppointment(slot, StartedAt(MAXSLOT), TimGetSeconds());
		  }
		  else
			 AddAppointment(slot, StartedAt(MAXSLOT), TimGetSeconds());
	 
	 if (Track(slot))
		if ((TimGetSeconds() - StartedAt(MAXSLOT)) > 180)
		  AddTimeStamp(slot, StartedAt(MAXSLOT), TimGetSeconds());

	 SetStartedAt(MAXSLOT, 0);
  }


  if (update) {
	 UpdateProjectListButton();
	 UpdateSlot(slot);
  }
}

void
ActivateSlot(Int slot, Int update)
{
  Int oldActiveSlot = ActiveSlot();

  if (oldActiveSlot == slot)
	 return;

  DeActivateSlot(ActiveSlot(), false);

  if (ActiveSlot() != slot) {
	 SetActiveSlot(slot);
	 SetStartedAt(slot, TimGetSeconds());
	 SetStartedAt(MAXSLOT, TimGetSeconds());
  }

  if (update) {
	 UpdateProjectListButton();
	 UpdateSlot(oldActiveSlot);
	 UpdateSlot(slot);
  }
}

void
UpdateProjectListButton(void)
{
  Int i;
  Int width = 0;
  FontID oldFont = FntSetFont(stdFont);
  Int num = ProjectListEntries();
  Char** list = ProjectListP();
  static Char text[MAXTEXT+1];
  ListPtr listP = GetObjectPtr(mainFormProjectList);

  LstSetListChoices(listP, list, num);
  LstSetHeight(listP, num);

  if ((num == 0) || (ActiveSlot() == noSlotActive)) {
	 CtlSetLabel(GetObjectPtr(mainFormProjectPopup), "None");
	 LstSetSelection(listP, noListSelection);
  }
  else {
	 StrNCopy(text, list[SlotToProjectListNum(ActiveSlot())], MAXTEXT+1);
	 TrimToLength(text, 80);

	 CtlSetLabel(GetObjectPtr(mainFormProjectPopup), text);
	 LstSetSelection(listP, SlotToProjectListNum(ActiveSlot()));
  }

  for(i = 0; i < num; i++)
	 width = max(FntCharsWidth(list[i], StrLen(list[i])), width);

  listP->bounds.extent.x = width + 4;
  listP->bounds.topLeft.x = 160 - (width + 4) - 2;

  FntSetFont(oldFont);
}

void
DrawSumTimeButton(void)
{
  FontID oldFont;
  Char time[MAXTEXT+1];
  ULong sum = SumElapsed();

  oldFont = FntSetFont(boldFont);

  StrNCopy(time, TimeToString(sum, time), MAXTEXT+1);

  ExpandLeftToLength(time, 42);

  CtlSetLabel(GetObjectPtr(mainFormTimeSum), time);
  
  FntSetFont(oldFont);
}

void 
UpdateInfoButtons(void)
{
  Boolean slotIsActive = (ActiveSlot() != noSlotActive);

  CtlSetEnabled(GetObjectPtr(mainFormDatebook), slotIsActive);
  CtlSetEnabled(GetObjectPtr(mainFormTrack), slotIsActive);
  CtlSetEnabled(GetObjectPtr(mainFormIgnore), slotIsActive);
  
  if (!slotIsActive) {
	 CtlSetLabel(GetObjectPtr(mainFormDatebook), "");
	 CtlSetLabel(GetObjectPtr(mainFormTrack), "");
	 CtlSetLabel(GetObjectPtr(mainFormIgnore), "");
  }
  else {
	 if (DateBook(ActiveSlot()))
		CtlSetLabel(GetObjectPtr(mainFormDatebook), "D");
	 else
		CtlSetLabel(GetObjectPtr(mainFormDatebook), "d");
	 
	 if (Track(ActiveSlot()))
		CtlSetLabel(GetObjectPtr(mainFormTrack), "T");
	 else
		CtlSetLabel(GetObjectPtr(mainFormTrack), "t");
	 
	 if (Ignored(ActiveSlot()))
		CtlSetLabel(GetObjectPtr(mainFormIgnore), "I");
	 else
		CtlSetLabel(GetObjectPtr(mainFormIgnore), "i");
  }
}

void UpdateElapsed(Int slot)
{
  ULong t, elapsed, started;

  if (slot == noSlotActive)
	 return;

  if (slot == ActiveSlot()) {
	 t = TimGetSeconds();
	 elapsed = Elapsed(slot) + t;
	 started = StartedAt(slot);
	 /*
	  * We must handle the case, that the user changes the pilot's clock
	  * while titrax is running. When the user turns the clock back,
	  * Elapsed(no) might become negative. We handle this case, by 
	  * resetting the current project.
	  */
	 if (started > elapsed) {
		SetStartedAt(slot, t);
		SetStartedAt(MAXSLOT, t);
		DeActivateSlot(slot, true);
		(void)FrmAlert(titraxTimeChangedAlert);
	 }
	 else {
		SetElapsed(slot, elapsed - started);
		SetStartedAt(slot, t);
	 }
  }
}

void
UpdateSlot(Int slot)
{
  if (slot == noSlotActive)
	 return;

  UpdateElapsed(slot);

  if (FrmGetActiveFormID() != mainForm)
	 return;

  TblMarkRowInvalid(GetObjectPtr(mainFormTable), SlotToRow(slot));
  TblRedrawTable(GetObjectPtr(mainFormTable));

  DrawSumTimeButton();
  UpdateInfoButtons();
}

void
UpdateActiveSlot(void)
{
  UpdateSlot(ActiveSlot());
}

void 
UpdateAllSlots(void)
{
  Int row;
  TablePtr tableP = GetObjectPtr(mainFormTable);

  for(row = FirstRow(); row < FirstRow() + VISSLOT; row++) {
	 UpdateElapsed(RowToSlot(row));
  }

  if (FrmGetActiveFormID() != mainForm)
	 return;

  TblMarkTableInvalid(tableP);
  TblRedrawTable(tableP);
  DrawSumTimeButton();
  UpdateInfoButtons();
}

void
ScrollTo(Int newRow)
{
  Int row;
  TablePtr tableP = GetObjectPtr(mainFormTable);

  if (newRow == FirstRow())
	 return;

  if (newRow < 0)
	 newRow = 0;

  if (newRow > (MAXSLOT - VISSLOT))
	 newRow = MAXSLOT - VISSLOT;

  SetFirstRow(newRow);

  for(row = 0; row < MAXSLOT; row++) {
	 TblSetRowUsable(tableP, row, false);
  }

  for(row = FirstRow(); row < FirstRow() + VISSLOT; row++)
	 TblSetRowUsable(tableP, row, true);

  UpdateAllSlots();
  MainFormCheckScroll();
}

void
ScrollUp(Int amount)
{
  ScrollTo(FirstRow() - amount);
}

void
ScrollDown(Int amount)
{
  ScrollTo(FirstRow() + amount);
}

void
ScrollCenterActive()
{
  if (ActiveSlot() == noSlotActive)
	 return;

  ScrollTo(SlotToRow(ActiveSlot()) - ((VISSLOT - 1) / 2));
}
		
void 
SwapRows(Int firstRow, Int secondRow)
{
  Int firstSlot = RowToSlot(firstRow);
  Int secondSlot = RowToSlot(secondRow);
  TablePtr tableP = GetObjectPtr(mainFormTable);


  if (firstRow == secondRow)
	 return;

  TblSetItemInt(tableP, firstRow, 0, secondSlot);
  SetRowToSlot(firstRow, secondSlot);

  TblSetItemInt(tableP, secondRow, 0, firstSlot);
  SetRowToSlot(secondRow, firstSlot);

  UpdateProjectList();
  UpdateProjectListButton();
  UpdateSlot(firstSlot);
  UpdateSlot(secondSlot);

  if (CenterActive())
	 ScrollCenterActive();
}

void
SortCommand(Long direction)
{
  Int row;
  TablePtr tableP = GetObjectPtr(mainFormTable);

  UpdateActiveSlot();
  SortSlots(direction);

  for(row = 0; row < MAXSLOT; row++)
	 TblSetItemInt(tableP, row, 0, RowToSlot(row));

  UpdateProjectList();
  UpdateProjectListButton();
  UpdateAllSlots();

  if (CenterActive())
	 ScrollCenterActive();
}

void
OrigOrderCommand(void)
{
  Int row;
  TablePtr tableP = GetObjectPtr(mainFormTable);

  for(row = 0; row < MAXSLOT; row++) {
	 SetRowToSlot(row, row);
	 TblSetItemInt(tableP, row, 0, row);
  }

  UpdateProjectList();
  UpdateProjectListButton();
  UpdateAllSlots();

  if (CenterActive())
	 ScrollCenterActive();
}

void
ClearAllCommand(void)
{
  if (FrmAlert(titraxClearAllAlert) == 0) {
	 DeActivateSlot(ActiveSlot(), false);
	 ClearAllTimes();
	 UpdateAllSlots();
	 UpdateProjectListButton();
  }
}

void
SaveAllCommand(void)
{
  if (FrmAlert(titraxSaveAllAlert) == 0) {
	 DeActivateSlot(ActiveSlot(), true);
	 if (SaveMemopad())
		CreateMemopadSummary();
	 if (SaveDatebook())
		CreateDatebookSummary(TimGetSeconds());
  }

  if (SaveAllClearsAll())
	 ClearAllCommand();
}

void
ChangeNameCommand(Int slot)
{
  SetEditedSlot(slot);
  FrmPopupForm(editForm);
}

void
ChangeTimeCommand(void)
{
  if (DateBook(ActiveSlot()) && ShiftWarnings())
	 (void)FrmAlert(titraxMainFormDateBookAlert);
  FrmPopupForm(timeForm);
}

void
ShiftTimeCommand(void)
{
  if (DateBook(ActiveSlot()) && ShiftWarnings())
	 (void)FrmAlert(titraxMainFormDateBookAlert);
  FrmPopupForm(shiftForm);
}

void
ClearCommand(void)
{
  Int slotNo = ActiveSlot();

  if (FrmAlert(titraxClearAlert) == 0) {
	 DeActivateSlot(slotNo, false);
	 ClearTime(slotNo);
	 UpdateProjectListButton();
	 UpdateSlot(slotNo);
  }
}

void
RemoveCommand(void)
{
  Int slotNo = ActiveSlot();

  if (FrmAlert(titraxRemoveAlert) == 0) {
	 DeActivateSlot(slotNo, false);
	 ClearSlot(slotNo);
	 UpdateProjectList();
	 UpdateProjectListButton();
	 UpdateSlot(slotNo);
  }
}

void
RemoveAllCommand(void)
{
  if (FrmAlert(titraxRemoveAllAlert) == 0) {
	 DeActivateSlot(ActiveSlot(), false);
	 ClearAllSlots();
	 UpdateProjectList();
	 UpdateProjectListButton();
	 UpdateAllSlots();
  }
}

void
EditNoteCommand(Int slot)
{
  SetEditedNoteSlot(slot);
  FrmPopupForm(noteForm);
}

void
CreateCommandList(void)
{
  Int i;
  Int width = 0;
  Int height;
  FontID oldFont = FntSetFont(stdFont);
  ListPtr listP = GetObjectPtr(mainFormCommandList);

  if (ActiveSlot() == noSlotActive) {
	 LstSetListChoices(listP, comEntry1P, numCommands1);
	 LstSetHeight(listP, numCommands1);
  }
  else {
	 LstSetListChoices(listP, comEntry2P, numCommands2);
	 LstSetHeight(listP, numCommands2);
  }
  LstSetSelection(listP, noListSelection);

  for(i = 0; i < LstGetNumberOfItems(listP); i++)
	 width = max(FntCharsWidth(comEntry[i], StrLen(comEntry[i])), width);

  FntSetFont(oldFont);

  listP->bounds.extent.x = width + 4;
  height = listP->bounds.extent.y;
  LstSetPosition(listP, 3, 155-height);
}

Int
YPositionToRow(Int y)
{
  Int row;
  RectangleType rect;
  TablePtr tableP = GetObjectPtr(mainFormTable);

  for(row = FirstRow(); row < FirstRow() + VISSLOT; row++) {
	 TblGetItemBounds(tableP, row, 0, &rect);
	 if (RctPtInRectangle(rect.topLeft.x + 1, y, &rect))
		break;
  }

  return row;
}

Int
XPositionToColumn(Int x)
{
  Int column;
  RectangleType rect;
  TablePtr tableP = GetObjectPtr(mainFormTable);

  for(column = 0; column < MAXCOL; column++) {
	 TblGetItemBounds(tableP, FirstRow(), column, &rect);
	 if (x < (rect.topLeft.x + rect.extent.x))
		break;
  }

  return column;
}

void
DrawSlot(VoidPtr tableP, Word row, Word column, RectanglePtr bounds)
{
  Int len;
  Int slot;
  Char text[4*MAXTEXT + 1];
  FontID oldFont;

  CALLBACK_PROLOGUE

  len = 0;
  slot = RowToSlot(row);

  TblUnhighlightSelection(tableP);
  WinEraseRectangle(bounds, 0);

  switch (column) {
  case 0: {
	 oldFont = FntSetFont(stdFont);

	 if (!TextIsEmpty(slot)) {
		GetText(slot, text);

		TrimToLength(text, BUTTONW - 2);
		len = StrLen(text);
	 }
	 else {
		StrCopy(text, " \205");
		len = 2;
	 }
	 break;
  }

  case 1: {
	 oldFont = FntSetFont(stdFont);

	 if (!TextIsEmpty(slot)) {
		TimeToString(Elapsed(slot), text);
		ExpandLeftToLength(text, TIMEW-1);
	 }
	 else
		StrCopy(text, "");

	 len = StrLen(text);
	 break;
  }

  case 2: {
	 oldFont = FntSetFont(symbolFont);

	 if (!TextIsEmpty(slot))
		if (NoteIsEmpty(slot))
		  StrCopy(text, "\013");
		else
		  StrCopy(text, "\014");
	 else
		return;

	 len = 1;
	 break;
  }

  default:
	 return;
  }
  
  WinDrawChars(text, len, bounds->topLeft.x, bounds->topLeft.y);
  FntSetFont(oldFont);

  if ((ActiveSlot() == slot) && ((column == 0) || (column == 1)))
	 WinInvertRectangle(bounds, 0);
  
  CALLBACK_EPILOGUE
}

void
MainFormInit(void)
{
  Int row;
  Int column;
  FormPtr frm = FrmGetActiveForm();
  TablePtr tableP = GetObjectPtr(mainFormTable);
  RectangleType rect1 = {{0, 139}, {160, 2}};
  RectangleType rect2 = {{60, 141}, {2, 19}};
  RectangleType rect3 = {{114, 141}, {2, 19}};

  TblMarkTableInvalid(tableP);

  for(row = 0; row < MAXSLOT; row++) {
	 TblSetRowUsable(tableP, row, false);

	 for(column = 0; column < MAXCOL; column++)
		TblSetItemStyle(tableP, row, column, customTableItem);

	 TblSetItemInt(tableP, row, 0, RowToSlot(row));
  }
  
  for(row = FirstRow(); row < FirstRow() + VISSLOT; row++)
	 TblSetRowUsable(tableP, row, true);

  for(column = 0; column < MAXCOL; column++) {
	 TblSetCustomDrawProcedure(tableP, column, DrawSlot);
	 TblSetColumnUsable(tableP, column, true);
  }

  TblSetColumnSpacing(tableP, 0, 0);
  TblSetColumnSpacing(tableP, 1, SPACE);
  TblSetColumnSpacing(tableP, 2, 0);

  UpdateProjectListButton();

  if (CenterActive())
	 ScrollCenterActive();

  FrmDrawForm(frm);

  WinDrawRectangle(&rect1, 0);
  WinDrawRectangle(&rect2, 0);
  WinDrawRectangle(&rect3, 0);

  DrawSumTimeButton();
  UpdateInfoButtons();

  MainFormCheckScroll();

  UpdateActiveSlot();
}

void
MainFormCheckScroll()
{
  SclSetScrollBar(GetObjectPtr(mainFormScrollBar), FirstRow(), 
						 0, MAXSLOT-VISSLOT, VISSLOT);
}

Boolean
MainFormHandleEvent(EventPtr event)
{
  Boolean handled = false;
  TablePtr tableP = GetObjectPtr(mainFormTable);
  RectangleType rect;

  switch (event->eType) {
  case ctlSelectEvent: {
	 if (event->data.ctlSelect.controlID == mainFormCommand) {
		ListPtr listP = GetObjectPtr(mainFormCommandList);

		CreateCommandList();

		switch (LstPopupList(listP)) {
		case 0:
		  SaveAllCommand(); break;
		case 1:
		  ClearAllCommand(); break;
		case 2:
		  RemoveAllCommand(); break;
		case 3:
		  SortCommand(ascending); break;
		case 4:
		  SortCommand(descending); break;
		case 5:
		  OrigOrderCommand(); break;
		case 6:
		  ChangeTimeCommand(); break;
		case 7:
		  ShiftTimeCommand(); break;
		case 8:
		  ClearCommand(); break;
		case 9:
		  RemoveCommand(); break;
		case 10:
		  EditNoteCommand(ActiveSlot()); break;
		case 11:
		  ScrollCenterActive(); break;
		case 12:
		  ChangeNameCommand(ActiveSlot()); break;
		default:
		}
		
		LstSetSelection(listP, noListSelection);
		handled = true;
	 }
		
	 if (event->data.ctlSelect.controlID == mainFormProjectPopup)
		handled = (ProjectListEntries() == 0);
	 
	 if (event->data.ctlSelect.controlID == mainFormDatebook) {
		SetDateBook(ActiveSlot(), !DateBook(ActiveSlot()));
		UpdateInfoButtons();
		handled = true;
	 }

	 if (event->data.ctlSelect.controlID == mainFormTrack) {
		SetTrack(ActiveSlot(), !Track(ActiveSlot()));
		UpdateInfoButtons();
		handled = true;
	 }

	 if (event->data.ctlSelect.controlID == mainFormIgnore) {
		SetIgnored(ActiveSlot(), !Ignored(ActiveSlot()));
		DrawSumTimeButton();
		UpdateInfoButtons();
		handled = true;
	 }

	 break;
  }

  case penDownEvent: {
	 TblGetBounds(tableP, &rect);
	 if (RctPtInRectangle(event->screenX, event->screenY, &rect)) {
		startRow = YPositionToRow(event->screenY);
		inDnD = true;
		handled = true;
	 }
	 else {
		inDnD = false;
		handled = false;
	 }
	 break;
  }

  case penUpEvent: {
	 if (inDnD) {
		inDnD = false;
		TblGetBounds(tableP, &rect);
		if (RctPtInRectangle(event->screenX, event->screenY, &rect)) {
		  Int endRow = YPositionToRow(event->screenY);
		  if (startRow != endRow)
			 SwapRows(startRow, endRow);
		  else {
			 Int column = XPositionToColumn(event->screenX);
			 Int slot = RowToSlot(endRow);
				
			 if (column < 2) {
				if (ActiveSlot() == slot)
				  DeActivateSlot(slot, true);
				else {
				  if (TextIsEmpty(slot))
					 ChangeNameCommand(slot);
				  else
					 ActivateSlot(slot, true);
				}
			 }
			 
			 if (column == 2) {
				if (!TextIsEmpty(slot))
				  EditNoteCommand(slot);
				else
				  UpdateSlot(slot);
			 }
		  }
		}
		handled = true;
	 }
	 else
		handled = false;

	 break;
  }

  case penMoveEvent: {
	 if (inDnD) {
		TblGetBounds(tableP, &rect);
		if (!RctPtInRectangle(event->screenX, event->screenY, &rect)) {
		  Int xD, yD;
		  Boolean penD;
		  
		  TblUnhighlightSelection(tableP);
		  
		  if (event->screenY < rect.topLeft.y)
			 do {
				ScrollUp(1);
				EvtGetPen(&xD, &yD, &penD);
			 } while (penD && (yD < rect.topLeft.y));
		  if (event->screenY > rect.topLeft.y + rect.extent.y)
			 do {
				ScrollDown(1);
				EvtGetPen(&xD, &yD, &penD);
			 } while (penD && (yD > rect.topLeft.y + rect.extent.y));
		}
		else {
		  Int currentRow = YPositionToRow(event->screenY);
		  TblSelectItem(tableP, currentRow, 0);
		}
		handled = true;
	 }
	 else
		handled = false;

	 break;
  }

  case popSelectEvent:
	 if (event->data.popSelect.listID == mainFormProjectList) {
		Word selection = event->data.popSelect.selection;
	 
		if (selection != noListSelection) {
		  Int slotNo = ProjectListNumToSlot(selection);
		  if (ActiveSlot() == slotNo)
			 DeActivateSlot(slotNo, true);
		  else
			 ActivateSlot(slotNo, true);

		  if (CenterActive())
			 ScrollCenterActive();
		}
		handled = true;
	 }
	 break;

  case frmOpenEvent:
	 MainFormInit();
	 handled = true;
	 break;

  case frmCloseEvent:
	 inDnD = false;
	 handled = false;
	 break;

  case appStopEvent:
	 SetActualForm(mainForm);
	 inDnD = false;
	 handled = false;
	 break;

  case nilEvent:
	 if (FrmGetActiveFormID() == mainForm)
		UpdateActiveSlot();
	 handled = true;
	 break;

  case sclRepeatEvent:
	 ScrollTo(event->data.sclRepeat.newValue);
	 handled = false;
	 break;

  case keyDownEvent:
	 if (event->data.keyDown.chr == pageUpChr) {
		ScrollUp(VISSLOT);
		handled = true;
	 }

	 if (event->data.keyDown.chr == pageDownChr) {
		ScrollDown(VISSLOT);
		handled = true;
	 }
	 break;

  case frmTitleSelectEvent:
	 if (MenuHack()) {
		MenuBarPtr menuP = MenuInit(mainFormMenu);

		MenuSetActiveMenu(menuP);
		MenuDrawMenu(menuP);
		handled = true;
	 }
	 else
		handled = false;
	 break;

  case menuEvent:
	 MenuEraseStatus(0);
	 if (event->data.menu.itemID == mainFormMenuOption) {
		FrmPopupForm(optionForm);
		handled = true;
	 }
	 if (event->data.menu.itemID == mainFormMenuAbout) {
		FormPtr frm = FrmInitForm(aboutForm);

		FrmDoDialog(frm);
		FrmDeleteForm(frm);
		handled = true;
	 }
	 break;

  default:
	 handled = false;
  }

  return handled;
}

