/* 
 * titrax.c
 *
 * A simple titrax clone written from scratch.
 * ("simple" is nowadays a good joke)
 *
 * Author: Bernd Esser <Bernd.Esser@testnetz.detecon.net>
 *
 * $Id: titrax.c,v 3.5 1998/04/26 09:03:29 ga58 Exp $
 *
 * $Log: titrax.c,v $
 * Revision 3.5  1998/04/26 09:03:29  ga58
 *  - finished the conversion to a shared library version
 * CVS:
 * CVS:
 *
 * Revision 3.4  1998/04/19 15:44:29  ga58
 * This is the first branch of the 3.x series of titrax. Most of the new
 * features and code changes are described in the NEWS file.
 * CVS:
 * CVS:
 *
 * Revision 2.7  1998/04/13 09:22:33  ga58
 *  - added a new option: Clear note in datebook mode
 * CVS:
 * CVS:
 *
 * Revision 2.6  1998/04/10 16:54:09  ga58
 * code cleanups in some modules
 * added the list-code to titraxDB
 * modularized mainForm
 * removed Elapsed and StartedAt from titraxDB and integrated them into
 * titraxPref
 * reworked TextIsEmpty and NoteIsEmpty in titraxDB
 * tried to make the code faster
 * CVS:
 * CVS:
 *
 * Revision 2.5  1998/04/05 16:23:55  ga58
 * removed LastCall and SetLastCall
 * CVS:
 * CVS:
 *
 * Revision 2.4  1998/04/04 15:26:34  ga58
 * version 2.4
 * CVS:
 * CVS:
 *
 * Revision 1.1  1998/04/04 15:24:33  ga58
 * first version in CVS
 * CVS:
 *
 * Revision 1.1.1.1  1998/04/04 15:04:07  ga58
 * first version
 *
 * Revision 2.4  1998/04/04 14:41:06  ga58
 * Added an option, to turn of the datebook-warnings
 *
 * Revision 2.3  1998/03/29 15:11:13  ga58
 * implemented dave's changes for UpdateSlot
 *
 * Revision 2.2  1998/03/28 13:08:06  ga58
 * minor changes
 *
 * Revision 2.1  1998/03/24 17:23:32  ga58
 * Added an Option form
 *
 * Revision 2.0  1998/03/24 08:55:56  ga58
 * The changes for datebook mode are nearly complete now. First Beta
 * version.
 *
 * Revision 1.26  1998/03/24 07:50:50  ga58
 * Almost all functionality for the datebook-mode is implemented now. Some
 * features are not completed yet.
 *
 * Revision 1.25  1998/03/22 18:08:19  ga58
 * changed the event-loop code
 *
 * Revision 1.24  1998/03/21 17:58:03  ga58
 * added a row of buttons to toggle the datebook-logging
 *
 * Revision 1.23  1998/03/21 16:06:31  ga58
 * Included the datebook-patch from Dave and modified it a little bit
 *
 * Revision 1.22  1998/02/14 12:16:50  ga58
 * fixed a bug in SaveData. Now uses StrNCat, so titrax won't run on
 * PalmOS 1.x devices
 *
 * Revision 1.21  1998/02/13 15:26:55  ga58
 * Increased the number of available project-slots from 11 to 30
 *
 * Revision 1.20  1998/01/31 15:48:48  ga58
 * Added the option to toggle the display format for times.
 *
 * Revision 1.19  1997/12/14 16:40:01  ga58
 * - changed the mainform to nonmodal
 * - the sumtime display is now in the title area
 * - made the buttons bigger
 *
 * Revision 1.18  1997/11/23 20:08:59  ga58
 * - added an experimental timestamp option
 *
 * Revision 1.17  1997/11/18 19:49:06  ga58
 * - menues can be invoked through the windows title bar
 *
 * Revision 1.16  1997/11/17 21:35:25  ga58
 * - fixed the redraw bug in the notes-dialog
 * - added undo and select all menu options in the notes and edit dialog
 *
 * Revision 1.15  1997/11/16 17:42:58  ga58
 * - removed the scrolling glitch in the note-dialog
 * - new about box
 *
 * Revision 1.14  1997/11/15 17:59:29  ga58
 * *** empty log message ***
 *
 * Revision 1.13  1997/11/14 17:49:10  ga58
 * - seperated the sources into various files
 * - switching between projects with the up and down buttons
 * - removed timestamps
 * - added scrolling for the notes window
 * - interface changes
 *
 * Revision 1.12  1997/11/12 17:42:47  ga58
 * - changed the note icons
 * - added a timestamp menu option
 * - added a timestamp button in the note dialog
 *
 * Revision 1.11  1997/11/09 14:59:18  ga58
 * - added a justification toggle for the project names
 * - all times are displayed right justified
 * - added cut, copy and paste menu items to the edit and note dialogs
 * - added a separate clear all button
 * - tapping in the time area invokes the edit dialog
 * - cleaned up and improved the tip-menu
 *
 * Revision 1.10  1997/11/08 18:24:18  ga58
 * - new layout
 * - pushbuttons for the labels
 * - new redraw code
 *
 * Revision 1.9  1997/11/06 18:10:49  ga58
 * - added utility macro MakeDigits
 * - EditViewInit and NoteViewInit only set the selection, if the string is empt
 *
 * Revision 1.8  1997/11/02 17:32:48  ga58
 * - titrax can now save all data in a memopad note
 *
 * Revision 1.7  1997/11/02 15:23:34  ga58
 * - finished the switch to a databased application
 * - some functionality added
 * - code cleanup
 *
 * Revision 1.6  1997/10/26 17:35:08  ga58
 * complete rewrite, now uses database for the slots
 *
 * v1.4 17.10.1997 be test-only version
 * v1.3 12.10.1997 be introduced a new layout for the change-time part of 
 *                    the edit-form
 * v1.2 11.10.1997 be changed edit-form layout
 * v1.1 03.10.1997 be found an overflow condition in TimeToString, code 
 *                    clean-up.
 * v1.0 25.09.1997 be
 *
 */

#pragma pack(2)

#include "titraxAll.h"

/*
 * Time measuring concept: when a project is started, the variable
 * startedAt is filled with the current time (TimGetSeconds). At each 
 * checkpoint the difference between startedAt and the current time is 
 * added to the variable elpased and startedAt is filled with the 
 * current time. Only the currently active project has running == true.
 */

/*
 * GetObjectPtr helper-function
 */
VoidPtr
GetObjectPtr(Int ObjectID)
{
  FormPtr frm = FrmGetActiveForm();

  return FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ObjectID));
}

/*
 * Eventloop
 */
static void
EventLoop(void)
{
  static EventType event;
  
  do {
	 Word error;

	 EvtGetEvent(&event, 60*SysTicksPerSecond());

	 if (SysHandleEvent(&event)) {
		/*
		 * this is a hack, to detect a switched-on device and to force an
		 * update
		 */
		if (event.eType == keyDownEvent)
		  UpdateActiveSlot();
	 }
	 else {
		if (!MenuHandleEvent(MenuGetActiveMenu(), &event, &error)) {
		  switch(event.eType) {
		  case frmLoadEvent: {
			 FormPtr frm = FrmInitForm(event.data.frmLoad.formID);
			 FrmSetActiveForm(frm);
			 
			 switch (event.data.frmLoad.formID) {
			 case mainForm:
				FrmSetEventHandler(frm, MainFormHandleEvent);
				break;
			 case editForm:
				FrmSetEventHandler(frm, EditFormHandleEvent);
				break;
			 case noteForm:
				FrmSetEventHandler(frm, NoteFormHandleEvent);
				break;
			 case timeForm:
				FrmSetEventHandler(frm, TimeFormHandleEvent);
				break;
			 case shiftForm:
				FrmSetEventHandler(frm, ShiftFormHandleEvent);
				break;
			 case optionForm:
				FrmSetEventHandler(frm, OptionFormHandleEvent);
				break;
			 default:
				break;
			 }
			 break;
		  }
		  
		  default:
			 FrmDispatchEvent(&event);
			 break;
		  }
		}
	 }
  } while (event.eType != appStopEvent);
}

/*
 * Init-routine
 */
static Boolean
StartApplication(void)
{
  DWord version;

  FtrGet(sysFtrCreator, sysFtrNumROMVersion, &version);
  if (sysGetROMVerMajor(version) == 1) {
	 (void)FrmAlert(titraxVersionAlert);
	 return true;
  }

  CreatePrefs();

  if (!ReadPrefs()) {
	 ClearPrefs();
	 CreateDatabase();
  }
  else
	 OpenDatabase();

  CreateProjectList();

  return false;
}

/*
 * Exit-routine
 */
static void
StopApplication(void)
{
  FrmSaveAllForms();
  FrmCloseAllForms();
 
  DeleteProjectList();

  CloseDatabase();

  WritePrefs();
  DeletePrefs();

}

/*
 * Main
 */
DWord 
PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
  DWord err = 0;

  switch(cmd) {
  case sysAppLaunchCmdNormalLaunch:
	 err = StartApplication();
	 if (!err) {
		FrmGotoForm(mainForm);
		/* 
		 * if we were interrupted while not in the main-form, restore the
		 * last active form
		 */
		if (ActualForm() != mainForm)
		  FrmGotoForm(ActualForm());
		EventLoop();
		StopApplication();
	 }
	 break;
		
  default:
	 err = sysErrParamErr;
	 break;
  }
  return err;
}
