/*
 * titraxDB.c
 *
 * Titrax database subroutines
 *
 * $Id: titraxDB.c,v 3.5 1998/04/26 09:03:29 ga58 Exp $
 *
 * $Log: titraxDB.c,v $
 * Revision 3.5  1998/04/26 09:03:29  ga58
 *  - finished the conversion to a shared library version
 * CVS:
 * CVS:
 *
 * Revision 3.4  1998/04/19 15:44:30  ga58
 * This is the first branch of the 3.x series of titrax. Most of the new
 * features and code changes are described in the NEWS file.
 * CVS:
 * CVS:
 *
 * Revision 2.6  1998/04/13 09:22:33  ga58
 *  - added a new option: Clear note in datebook mode
 * CVS:
 * CVS:
 *
 * Revision 2.5  1998/04/10 16:54:10  ga58
 * code cleanups in some modules
 * added the list-code to titraxDB
 * modularized mainForm
 * removed Elapsed and StartedAt from titraxDB and integrated them into
 * titraxPref
 * reworked TextIsEmpty and NoteIsEmpty in titraxDB
 * tried to make the code faster
 * CVS:
 * CVS:
 *
 * Revision 2.4  1998/04/04 15:26:35  ga58
 * version 2.4
 * CVS:
 * CVS:
 *
 * Revision 1.1  1998/04/04 15:24:33  ga58
 * first version in CVS
 * CVS:
 *
 * Revision 1.1.1.1  1998/04/04 15:04:07  ga58
 * first version
 *
 * Revision 1.4  1998/03/21 16:06:31  ga58
 * *** empty log message ***
 *
 * Revision 1.3  1997/11/23 15:30:38  ga58
 * - fixed a off-by-one bug in ReadNote
 *
 * Revision 1.2  1997/11/15 17:59:29  ga58
 * *** empty log message ***
 *
 * Revision 1.1  1997/11/14 17:51:04  ga58
 * Initial revision
 *
 */

#pragma pack(2)

#include "titraxAll.h"

static DmOpenRef TitraxDB;

#define minNoteLen 64

/*
 * function declarations
 */
void CreateRecords(void);

/*
 * Database operations
 * 
 * These routines do NOT perform any error-checking
 */
void
OpenDatabase(void)
{
  TitraxDB = DmOpenDatabaseByTypeCreator(titraxDBType, titraxAppType, dmModeReadWrite);
  ErrFatalDisplayIf(TitraxDB == 0, "Could not open database.");
}

void
CreateDatabase(void)
{
  Err err;
  LocalID dbID;

  /* check for a existing database */
  dbID = DmFindDatabase(0, titraxDBName);
  if (dbID != 0)
	 DmDeleteDatabase(0, dbID);

  err = DmCreateDatabase(0, titraxDBName, titraxAppType, titraxDBType, false);
  ErrFatalDisplayIf(err, "Could not create database.");

  TitraxDB = DmOpenDatabaseByTypeCreator(titraxDBType, titraxAppType, dmModeReadWrite);
  ErrFatalDisplayIf(TitraxDB == 0, "Could not create database.");

  CreateRecords();
}

void
CloseDatabase(void)
{
  DmCloseDatabase(TitraxDB);
}

Int
CreateRecord(UInt idx, Int size)
{
  UInt i;
  Err error;
  VoidPtr recP;
  VoidHand recH;

  i = idx;

  recH = DmNewRecord(TitraxDB, &i, size);
  ErrFatalDisplayIf(recH == 0, "Cannot create new record handle.");
  recP = MemHandleLock(recH);
  ErrFatalDisplayIf(recP == 0, "Cannot lock new record handle.");

  error = DmSet(recP, 0, size, 0);
  ErrFatalDisplayIf(error, "Could not create record.");

  MemHandleUnlock(recH);
  DmReleaseRecord(TitraxDB, i, true);

  return i;
}

void
CreateRecords(void)
{
  UInt i;
  UInt index = 0;

  for(i = 0; i < MAXSLOT; i++) {
	 SetTextRecIdx(i, CreateRecord(index, MAXTEXT+1));
	 index++;
  }

  for(i = 0; i < MAXSLOT; i++) {
	 SetNoteRecIdx(i, CreateRecord(index, minNoteLen));
	 index++;
  }
}

void
WriteRecord(UInt idx, VoidPtr p, Int size, Int cSize)
{
  VoidPtr recP;
  VoidHand recH;

  recH = DmGetRecord(TitraxDB, idx);
  ErrFatalDisplayIf(recH == 0, "Cannot get record handle.");
  recP = MemHandleLock(recH);
  ErrFatalDisplayIf(recP == 0, "Cannot lock record handle.");

  DmSet(recP, 0, cSize, 0);
  DmWrite(recP, 0, p, size);

  MemHandleUnlock(recH);
  DmReleaseRecord(TitraxDB, idx, true);
}

void 
ReadRecord(UInt idx, VoidPtr p, Int size)
{
  VoidPtr recP;
  VoidHand recH;

  recH = DmQueryRecord(TitraxDB, idx);
  recP = MemHandleLock(recH);

  MemMove(p, recP, size);

  MemHandleUnlock(recH);
  DmReleaseRecord(TitraxDB, idx, false);
}

Handle
GetRecordHandle(UInt idx)
{
  return (Handle)DmGetRecord(TitraxDB, idx);
}

void
ReleaseRecordHandle(UInt idx, Int dirty)
{
  DmReleaseRecord(TitraxDB, idx, dirty);
}

void
GetText(Int no, CharPtr charP)
{
  ReadRecord(TextRecIdx(no), charP, MAXTEXT+1);
}

void
SetText(Int no, CharPtr charP)
{
  WriteRecord(TextRecIdx(no), charP, StrLen(charP), MAXTEXT+1);
}

Boolean
EmptySlot(UInt idx)
{
  static Char t[2];

  ReadRecord(idx, t, 2);

  return t[0] == 0;
}

Boolean
TextIsEmpty(Int no)
{
  return EmptySlot(TextRecIdx(no));
}

/*
 * note subroutines are handled different because the notes should have
 * a dynamic size
 */
void 
WriteNoteRecord(UInt idx, CharPtr charP, Int len)
{
  ULong newLen;
  VoidPtr recP;
  VoidHand recH;

  newLen = (len > minNoteLen) ? len : minNoteLen;

  recH = DmResizeRecord(TitraxDB, idx, newLen);
  ErrFatalDisplayIf(recH == 0, "Note record resize failed.");
  recP = MemHandleLock(recH);
  ErrFatalDisplayIf(recP == 0, "Note record lock failed.");

  DmSet(recP, 0, newLen, 0);
  DmStrCopy(recP, 0, charP);

  MemHandleUnlock(recH);
  DmReleaseRecord(TitraxDB, idx, true);
}

VoidHand
ReadNoteRecord(UInt idx)
{
  Int len;
  VoidPtr recP, strP;
  VoidHand recH, strH;

  recH = DmQueryRecord(TitraxDB, idx);
  recP = MemHandleLock(recH);

  len = StrLen(recP) + 1;

  strH = MemHandleNew(len);
  strP = MemHandleLock(strH);

  MemSet(strP, len, 0);
  MemMove(strP, recP, StrLen(recP));
  MemHandleUnlock(strH);

  MemHandleUnlock(recH);
  DmReleaseRecord(TitraxDB, idx, false);

  return strH;
}

Int
NoteStrLen(UInt idx)
{
  Int len;
  VoidPtr recP;
  VoidHand recH;

  recH = DmQueryRecord(TitraxDB, idx);
  recP = MemHandleLock(recH);

  len = StrLen(recP);

  MemHandleUnlock(recH);
  DmReleaseRecord(TitraxDB, idx, false);

  return len;
}

void
SetNote(Int no, CharPtr charP)
{
  WriteNoteRecord(NoteRecIdx(no), charP, StrLen(charP) + 1);
}
  
VoidHand
GetNote(Int no)
{
  return ReadNoteRecord(NoteRecIdx(no));
}

Boolean
NoteIsEmpty(Int no)
{
  Char t[2];

  ReadRecord(NoteRecIdx(no), t, 2);

  return t[0] == 0;
}

void
ClearNote(Int no)
{
  Char str[2];

  StrCopy(str, "");
  WriteNoteRecord(NoteRecIdx(no), str, StrLen(str) + 1);
}

Int 
NoteLen(Int no)
{
  return NoteStrLen(NoteRecIdx(no));
}
