/*
 * Titrax preferences subroutines
 *
 * $Id: titraxPref.c,v 3.7 1998/04/27 06:36:47 ga58 Exp $
 *
 * $Log: titraxPref.c,v $
 * Revision 3.7  1998/04/27 06:36:47  ga58
 *  - the width of the command popup window and the project list window are now
 *    calculated dynamically.
 * CVS:
 * CVS:
 *
 * Revision 3.6  1998/04/26 09:03:29  ga58
 *  - finished the conversion to a shared library version
 * CVS:
 * CVS:
 *
 * Revision 3.5  1998/04/21 05:20:01  ga58
 *  - new functions SumElapsed. This is a preliminary step for implementing
 *    project flags. The first flag will be a ignored-flag. Projects with
 *    this flag set, will be ignored in the sum and in the sorting process.
 * CVS:
 * CVS:
 *
 * Revision 3.4  1998/04/19 15:44:30  ga58
 * This is the first branch of the 3.x series of titrax. Most of the new
 * features and code changes are described in the NEWS file.
 * CVS:
 * CVS:
 *
 * Revision 2.8  1998/04/13 09:22:33  ga58
 *  - added a new option: Clear note in datebook mode
 * CVS:
 * CVS:
 *
 * Revision 2.7  1998/04/12 09:31:09  ga58
 *  - a lot of minor changes in the UI
 *  - a lot of minor changes in the code
 *  - added new options
 * CVS:
 * CVS:
 *
 * Revision 2.6  1998/04/10 16:54:10  ga58
 * code cleanups in some modules
 * added the list-code to titraxDB
 * modularized mainForm
 * removed Elapsed and StartedAt from titraxDB and integrated them into
 * titraxPref
 * reworked TextIsEmpty and NoteIsEmpty in titraxDB
 * tried to make the code faster
 * CVS:
 * CVS:
 *
 * Revision 2.5  1998/04/05 16:23:55  ga58
 * removed LastCall and SetLastCall
 * CVS:
 * CVS:
 *
 * Revision 2.4  1998/04/04 15:26:35  ga58
 * version 2.4
 * CVS:
 * CVS:
 *
 * Revision 1.1  1998/04/04 15:24:33  ga58
 * first version in CVS
 * CVS:
 *
 * Revision 1.1.1.1  1998/04/04 15:04:07  ga58
 * first version
 *
 * Revision 1.9  1998/04/04 14:41:06  ga58
 * *** empty log message ***
 *
 * Revision 1.8  1998/03/24 17:23:32  ga58
 * *** empty log message ***
 *
 * Revision 1.7  1998/03/24 07:50:50  ga58
 * *** empty log message ***
 *
 * Revision 1.6  1998/03/21 16:06:31  ga58
 * *** empty log message ***
 *
 * Revision 1.5  1998/02/13 15:26:55  ga58
 * Increased the number of available project-slots from 11 to 30
 *
 * Revision 1.4  1998/01/31 15:48:48  ga58
 * Added the option to toggle the display format for times.
 *
 * Revision 1.3  1997/11/23 20:08:59  ga58
 * *** empty log message ***
 *
 * Revision 1.2  1997/11/15 17:59:29  ga58
 * *** empty log message ***
 *
 * Revision 1.1  1997/11/14 17:51:04  ga58
 * Initial revision
 *
 */

#pragma pack(2)

#include "titraxAll.h"

struct TitraxFlagType {
  unsigned ignored : 1;
  unsigned unused1 : 1;
  unsigned unused2 : 1;
  unsigned unused3 : 1;
  unsigned unused4 : 1;
  unsigned unused5 : 1;
  unsigned unused6 : 1;
  unsigned unused7 : 1;
  unsigned unused8 : 8;
};
  
typedef struct {
  /* this are the record-indices in the database */
  UInt text[MAXSLOT];
  UInt note[MAXSLOT];
  /* global vars which must be saved */
  Int rowToSlot[MAXSLOT];
  ULong elapsed[MAXSLOT];
  ULong startedAt[MAXSLOT + 1];
  Int activeSlot;
  Int editedSlot;
  Int editedNoteSlot;
  Int actualForm;
  Int shiftSlot[MAXSLOT];
  ULong inc;
  Boolean track[MAXSLOT];
  Boolean dateBook[MAXSLOT];
  Boolean showDays;
  Int firstRow;
  Boolean saveMemopad;
  Boolean saveDatebook;
  Boolean shiftWarnings;
  Boolean centerActive;
  Boolean saveAllClearsAll;
  Boolean clearClearsNote;
  Boolean menuHack;
  Boolean alwaysNewNote;
  struct TitraxFlagType flag[MAXSLOT];
} TitraxPreferenceType;

static VoidHand prefHand;
static TitraxPreferenceType* prefP;

#define titraxVersionNum 39

void 
CreatePrefs(void)
{
  prefHand = MemHandleNew(sizeof(TitraxPreferenceType));
  ErrFatalDisplayIf(prefHand == 0, "Out of memory for preferences.");

  prefP = (TitraxPreferenceType*)MemHandleLock(prefHand);
}

Boolean
ReadPrefs(void)
{
  Word prefSize = sizeof(TitraxPreferenceType);
  
  if (PrefGetAppPreferences((ULong)titraxAppType, 0, prefP, &prefSize, true) == noPreferenceFound)
	 return false;

  return (prefSize == sizeof(TitraxPreferenceType));
}

void
WritePrefs(void)
{
  PrefSetAppPreferences((ULong)titraxAppType, 0, titraxVersionNum, prefP, sizeof(TitraxPreferenceType), true);
}

void
ClearPrefs(void)
{
  Int i;

  prefP->activeSlot = noSlotActive;
  prefP->editedSlot = noSlotActive;
  prefP->editedNoteSlot = noSlotActive;
  prefP->actualForm = mainForm;
  prefP->inc = 60;
  prefP->showDays = false;
  prefP->firstRow = 0;
  prefP->saveDatebook = true;
  prefP->saveMemopad = false;
  prefP->shiftWarnings = true;
  prefP->centerActive = true;
  prefP->saveAllClearsAll = true;
  prefP->clearClearsNote = true;
  prefP->menuHack = true;
  prefP->alwaysNewNote = false;

  for(i = 0; i < MAXSLOT; i++) {
	 prefP->elapsed[i] = 0;
	 prefP->startedAt[i] = 0;
	 prefP->rowToSlot[i] = i;
	 prefP->shiftSlot[i] = noSlotActive;
	 prefP->track[i] = false;
	 prefP->dateBook[i] = true;
	 prefP->flag[i].ignored = 0;
  }

  prefP->startedAt[MAXSLOT] = 0;
}

void
DeletePrefs()
{
  if (MemHandleUnlock(prefHand) != 0)
	 return;
  if (MemHandleFree(prefHand) != 0)
	 return;
}

ULong
Elapsed(Int no)
{
  return prefP->elapsed[no];
}

void
SetElapsed(Int no, ULong val)
{
  prefP->elapsed[no] = val;
}

ULong
SumElapsed(void)
{
  Int slot;
  ULong sum = 0;

  for(slot = 0; slot < MAXSLOT; slot++) {
	 if (!Ignored(slot))
		sum += prefP->elapsed[slot];
  }

  return sum;
}

ULong
StartedAt(Int no)
{
  return prefP->startedAt[no];
}

void
SetStartedAt(Int no, ULong val)
{
  prefP->startedAt[no] = val;
}

Int
EditedSlot(void)
{
  return prefP->editedSlot;
}

void
SetEditedSlot(Int no)
{
  prefP->editedSlot = no;
}

Int
EditedNoteSlot(void)
{
  return prefP->editedNoteSlot;
}

void
SetEditedNoteSlot(Int no)
{
  prefP->editedNoteSlot = no;
}

Int
ActualForm(void)
{
  return prefP->actualForm;
}

void
SetActualForm(Int f)
{
  prefP->actualForm = f;
}

Int
ShiftSlot(Int no)
{
  return prefP->shiftSlot[no];
}

void
SetShiftSlot(Int no, Int val)
{
  prefP->shiftSlot[no] = val;
}

UInt
TextRecIdx(Int no)
{
  return prefP->text[no];
}

void
SetTextRecIdx(Int no, UInt idx)
{
  prefP->text[no] = idx;
}

UInt
NoteRecIdx(Int no)
{
  return prefP->note[no];
}

void
SetNoteRecIdx(Int no, UInt idx)
{
  prefP->note[no] = idx;
}

Int
ActiveSlot(void)
{
  return prefP->activeSlot;
}

void
SetActiveSlot(Int no)
{
  prefP->activeSlot = no;
}

ULong
Increment(void)
{
  return prefP->inc;
}

void
SetIncrement(ULong val)
{
  prefP->inc = val;
}

Boolean
Track(Int no)
{
  return prefP->track[no];
}

void
SetTrack(Int no, Boolean val)
{
  prefP->track[no] = val;
}

Boolean
ShowDays(void)
{
  return prefP->showDays;
}

void
SetShowDays(Boolean val)
{
  prefP->showDays = val;
}

Int
FirstRow(void)
{
  return prefP->firstRow;
}

void
SetFirstRow(Int val)
{
  prefP->firstRow = val;
}

Boolean
DateBook(Int no)
{
  return prefP->dateBook[no];
}

void
SetDateBook(Int no, Boolean val)
{
  prefP->dateBook[no] = val;
}

Boolean
SaveDatebook(void)
{
  return prefP->saveDatebook;
}

void
SetSaveDatebook(Boolean val)
{
  prefP->saveDatebook = val;
}

Boolean
SaveMemopad(void)
{
  return prefP->saveMemopad;
}

void
SetSaveMemopad(Boolean val)
{
  prefP->saveMemopad = val;
}

Boolean
ShiftWarnings(void)
{
  return prefP->shiftWarnings;
}

void
SetShiftWarnings(Boolean val)
{
  prefP->shiftWarnings = val;
}

Boolean
CenterActive(void)
{
  return prefP->centerActive;
}

void
SetCenterActive(Boolean val)
{
  prefP->centerActive = val;
}

Boolean
SaveAllClearsAll(void)
{
  return prefP->saveAllClearsAll;
}

void
SetSaveAllClearsAll(Boolean val)
{
  prefP->saveAllClearsAll = val;
}

Boolean
ClearClearsNote(void)
{
  return prefP->clearClearsNote;
}

void
SetClearClearsNote(Boolean val)
{
  prefP->clearClearsNote = val;
}

Boolean
MenuHack(void)
{
  return prefP->menuHack;
}

void
SetMenuHack(Boolean val)
{
  prefP->menuHack = val;
}

Boolean
AlwaysNewNote(void)
{
  return prefP->alwaysNewNote;
}

void
SetAlwaysNewNote(Boolean val)
{
  prefP->alwaysNewNote = val;
}

Int
RowToSlot(Int row)
{
  return prefP->rowToSlot[row];
}

void
SetRowToSlot(Int row, Int slot)
{
  prefP->rowToSlot[row] = slot;
}

Int
SlotToRow(Int slot)
{
  Int i;

  for(i = 0; i < MAXSLOT; i++)
	 if (prefP->rowToSlot[i] == slot)
		return i;

  return -1;
}

Int*
RowToSlotAddr(void)
{
  return prefP->rowToSlot;
}

Boolean
Ignored(Int slot)
{
  return prefP->flag[slot].ignored;
}

void
SetIgnored(Int slot, Boolean val)
{
  prefP->flag[slot].ignored = val;
}

