/*
 * Vexed v1.3 - Vexed.c "The Main Stuff, i.e. event handlers, loops, etc"
 * Copyright (C) 1999 James McCombe (cybertube@earthling.net)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * */

#pragma pack(2)
#include <Common.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>

#include "Vexed.h"
#include "protos.h"

static int StartApplication(void);
static void EventLoop(void);
static void StopApplication(void);
static Boolean frmMainEventH(EventPtr event);
static Boolean frmAboutEventH(EventPtr event);
static Boolean frmPrefsEventH(EventPtr event);
static Boolean frmSelectLevelEventH(EventPtr event);

struct {
	struct {
		int x, y;
	} begin, end;
} playermove;

extern struct {
	Boolean
		PieceMoveAnim, // whether the movement of a piece by the user is animated
		GravityAnim, // whether the gravity effect is animated
		EliminationAnim, // whether piece elimination is animated
		Sound; // whether sound is on or off
	int CurrentLevel, SolvedLevels; // stores the level the user was last at
} VexedPreferences;

extern int current [9] [11];
extern int stored [9] [11];

//---------------------------------------------------------------------------
DWord  PilotMain (Word cmd, Ptr cmdPBP, Word launchFlags)
{
int error;

if (cmd == sysAppLaunchCmdNormalLaunch)
	{
	error = StartApplication();	// Application start code
	if (error)
	    return error;
	EventLoop();	            	// Event loop
	StopApplication ();	        // Application stop code
	}
	return 0;
}


static int StartApplication(void)
{
LoadBitmapResources();
LoadPreferences();
FrmGotoForm(FormMain);
}


static void EventLoop(void)
{
short err;
int formID;
FormPtr form;
EventType event;

do
	{
	EvtGetEvent(&event, 200);
	if (SysHandleEvent(&event))
	    continue;
	if (MenuHandleEvent((void *)0, &event, &err))
	    continue;
	if (event.eType == frmLoadEvent)
		{
		formID = event.data.frmLoad.formID;
		form = FrmInitForm(formID);
		FrmSetActiveForm(form);
		switch (formID)
			{
			case FormMain:
				FrmSetEventHandler(form, (FormEventHandlerPtr) frmMainEventH);
				break;

			case FormAbout:
				FrmSetEventHandler(form, (FormEventHandlerPtr) frmAboutEventH);
				break;

			case FormPrefs:
				FrmSetEventHandler(form, (FormEventHandlerPtr) frmPrefsEventH);
				break;

			case FormSelectLevel:
				FrmSetEventHandler(form, (FormEventHandlerPtr) frmSelectLevelEventH);
				break;
			}
		}
	FrmDispatchEvent(&event);
	}
	while(event.eType != appStopEvent);
}


static void StopApplication(void)
{
SavePreferences();
UnloadBitmapResources();
}


static Boolean frmMainEventH(EventPtr event)
{
FormPtr   form;
int       handled = 0;
SWord penx, peny;
Boolean bstate;

switch (event->eType)
	{
	case frmOpenEvent:
		form = FrmGetActiveForm();
		FrmDrawForm(form);
		DrawIntro(); // draws the intro screen upon game startup
		while (bstate) EvtGetPen(&penx, &peny, &bstate);
		while (!bstate) EvtGetPen(&penx, &peny, &bstate);
		LoadLevel(VexedPreferences.CurrentLevel);
		handled = 1;
		break;
	case penDownEvent:
		EvtGetPen(&penx, &peny, &bstate);
		if (peny < 144) // the gameplay area
			{
			playermove.begin.x = (int) (penx / 16);
			playermove.begin.y = (int) ((peny / 16) - 1);
			handled = 1;
			break;
			}
		if (peny > 144) // the control icons area
			{
			if (penx > 144) // the restart level icon
				{
				LoadLevel(VexedPreferences.CurrentLevel);
				handled = 1;
				break;
				}
			if ((penx > 128) && (penx < 144)) // the recall icon
				{
				if (stored[0][0] == 0) 
					FrmAlert(AlertNothingStored);
				else
					{
					int x, y;
					for (y = 0; y <= 7; y ++)
						for (x = 0; x <= 9; x ++)
							current[y][x] = stored[y][x]; // move stored => current level array
					DrawLevel();
					}
				}
			if ((penx > 112) && (penx < 128)) // the memorise icon
				{
				int x, y;
				for (y = 0; y <= 7; y ++)
					for (x = 0; x <= 9; x ++)
						stored[y][x] = current[y][x]; // move current => stored level array
				}
			if ((penx < 16) && (VexedPreferences.CurrentLevel)) // the first level icon
				{
				VexedPreferences.CurrentLevel = 0;
				LoadLevel(VexedPreferences.CurrentLevel);
				handled = 1;
				break;
				}
			if ((penx < 32) && (penx > 16) && (VexedPreferences.CurrentLevel)) // the previous level icon
				{
				VexedPreferences.CurrentLevel --;
				LoadLevel(VexedPreferences.CurrentLevel);
				handled = 1;
				break;
				}
			if ((penx < 48) && (penx > 32)) // the select level icon
				{
				FrmPopupForm(FormSelectLevel);
				handled = 1;
				break;
				}
			if ((penx < 64) && (penx > 48)) // the next level icon
				{
				if (!(VexedPreferences.CurrentLevel + 1 > VexedPreferences.SolvedLevels))
					{
					VexedPreferences.CurrentLevel ++;
					LoadLevel(VexedPreferences.CurrentLevel);
					}
				handled = 1;
				break;
				}
			if ((penx < 80) && (penx > 64) && (!(VexedPreferences.CurrentLevel == VexedPreferences.SolvedLevels)))
				{ // the last solved level icon
				VexedPreferences.CurrentLevel = VexedPreferences.SolvedLevels;
				LoadLevel(VexedPreferences.CurrentLevel);
				handled = 1;
				break;
				}
			}
	case penUpEvent:
		EvtGetPen(&penx, &peny, &bstate);
		if (peny < 144) // penUpEvent is only an issue in the gameplay area
			{
			playermove.end.x = (int) (penx / 16);
			playermove.end.y = (int) ((peny / 16) - 1);
			if (!((playermove.begin.x == playermove.end.x) && (playermove.begin.y == playermove.end.y)))
				{
				if (playermove.end.x > playermove.begin.x)
					if (PlayerMakeMove(playermove.begin.x, playermove.begin.y, Right))
						{
						VexedPreferences.CurrentLevel ++;
						if (VexedPreferences.CurrentLevel > VexedPreferences.SolvedLevels)
							VexedPreferences.SolvedLevels ++;
						LoadLevel(VexedPreferences.CurrentLevel); // move onto next level if current is completed
						handled = 1;
						break;
						}
				if (playermove.end.x < playermove.begin.x)
					if (PlayerMakeMove(playermove.begin.x, playermove.begin.y, Left))
						{
						VexedPreferences.CurrentLevel ++;
						if (VexedPreferences.CurrentLevel > VexedPreferences.SolvedLevels)
							VexedPreferences.SolvedLevels ++;
						LoadLevel(VexedPreferences.CurrentLevel); // move onto next level if current is completed
						handled = 1;
						break;
						}
				}
			handled = 1;
			break;
			}
	case menuEvent:
        		if (event->data.menu.itemID == ItemAbout)
                		{
                		FrmPopupForm(FormAbout);
                		handled = 1;
			break;
			}
		if (event->data.menu.itemID == ItemPreferences)
			{
			FrmPopupForm(FormPrefs);
			handled = 1;
			break;
			}
		if (event->data.menu.itemID == ItemHowToPlay)
			{
			FrmHelp(StringInfo);
			handled = 1;
			break;
			}
	case nilEvent:
		handled = 1;
   		break;
	}
return handled;
}

static Boolean frmAboutEventH(EventPtr event)
{
FormPtr   form;
int handled = 0;

switch (event->eType)
        {
        case frmOpenEvent:
                form = FrmGetActiveForm();
                FrmDrawForm(form);
                handled = 1;
                break;
        case ctlSelectEvent:
                if (event->data.ctlEnter.controlID == ButtonOK)
                        {
                        FrmReturnToForm(0);
                        handled = 1;
                        break;
                        }
        }
return handled;
}

static Boolean frmPrefsEventH(EventPtr event)
{
FormPtr   form;
int handled = 0;

switch (event->eType)
	{
	case frmOpenEvent:
		form = FrmGetActiveForm();

		// set the check boxes to the current preferences state
		FrmSetControlValue(form, FrmGetObjectIndex(form, CheckBoxPieceAnim), VexedPreferences.PieceMoveAnim);
		FrmSetControlValue(form, FrmGetObjectIndex(form, CheckBoxGravityAnim), VexedPreferences.GravityAnim);
		FrmSetControlValue(form, FrmGetObjectIndex(form, CheckBoxEliminationAnim), VexedPreferences.EliminationAnim);
		FrmSetControlValue(form, FrmGetObjectIndex(form, CheckBoxSound), VexedPreferences.Sound);

		FrmDrawForm(form);
		handled = 1;
		break;
	case ctlSelectEvent:
		if (event->data.ctlEnter.controlID == ButtonOK2)
			{
			form = FrmGetActiveForm();
			// set the preferences state to the state of the checkboxes on the form
			VexedPreferences.PieceMoveAnim = FrmGetControlValue(form, FrmGetObjectIndex(form, CheckBoxPieceAnim));
			VexedPreferences.GravityAnim = FrmGetControlValue(form, FrmGetObjectIndex(form, CheckBoxGravityAnim));
			VexedPreferences.EliminationAnim = FrmGetControlValue(form, FrmGetObjectIndex(form, CheckBoxEliminationAnim));
			VexedPreferences.Sound = FrmGetControlValue(form, FrmGetObjectIndex(form, CheckBoxSound));

			// leave the preferences form
                        	FrmReturnToForm(0);
                        	handled = 1;
                        	break;
                        	}
		
		if (event->data.ctlEnter.controlID == ButtonReset)
			{
			if (!FrmAlert(AlertResetYesNo))
				{
				VexedPreferences.SolvedLevels = 0;
				VexedPreferences.CurrentLevel = 0;
				FrmReturnToForm(0);
				LoadLevel(VexedPreferences.CurrentLevel);
				handled = 1;
				break;
				}
			}
		
		if (event->data.ctlEnter.controlID == ButtonCancel)
			{
			FrmReturnToForm(0);
			handled = 1;
			break;
			}
        }
return handled;
}

static Boolean frmSelectLevelEventH(EventPtr event)
{
FormPtr   form;
int handled = 0;
int FirstPart, LastPart; // the two digits in the level number
char NewLabel [22];

switch (event->eType)
	{
	case frmOpenEvent:
		form = FrmGetActiveForm();
		// make the select level dialog have the current level selected here's a smart way of doing it
		// be aware that all resource id's in this form run in numerical order as the push buttons do
		FirstPart = VexedPreferences.CurrentLevel / 10; // gets us a floating point value
		LastPart = VexedPreferences.CurrentLevel % 10; // gets the last part
		FrmSetControlValue(form, FrmGetObjectIndex(form, FirstPart + 1076), 1);
		FrmSetControlValue(form, FrmGetObjectIndex(form, LastPart + 1082), 1);
		StrPrintF(NewLabel, "Solved up to level %i", VexedPreferences.SolvedLevels);
		CtlSetLabel(FrmGetObjectPtr(form, FrmGetObjectIndex(form, LabelSolvedLevels)), NewLabel);
		FrmDrawForm(form);
		handled = 1;
		break;
	case ctlSelectEvent:
		if (event->data.ctlEnter.controlID == ButtonSelectLevelOK)
			{
			form = FrmGetActiveForm();
			FirstPart = FrmGetObjectId(form, FrmGetControlGroupSelection(form, 1)) - 1076;
			LastPart = FrmGetObjectId(form, FrmGetControlGroupSelection(form, 2)) - 1082;

			// leave the select level form
                        	FrmReturnToForm(0);
			if (((FirstPart * 10) + LastPart) <= MaxLevels)
				{
				if (!(((FirstPart * 10) + LastPart) > VexedPreferences.SolvedLevels)) // have you solved upto this level?
					{
					VexedPreferences.CurrentLevel = (FirstPart * 10) + LastPart;
					LoadLevel(VexedPreferences.CurrentLevel); // load and draw the level onscreen
					}
				else
					FrmAlert(AlertAccessDenied); // if you are tryed to goto a level you haven't solved before
				}
			else
				FrmAlert(AlertNoLevel); // level ain't typed in yet!!!!

                        	handled = 1;
                        	break;
                        	}
		if (event->data.ctlEnter.controlID == ButtonSelectLevelCancel)
			{
			// leave the select level form
			FrmReturnToForm(0);
			handled = 1;
			break;
			}
        }
return handled;
}
