/*
 * Vexed v1.3 - gfx.c "The graphical functions for Vexed"
 * Copyright (C) 1999 James McCombe (cybertube@earthling.net)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * */

#include <Common.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>

#include "vexed.h"
#include "protos.h"

extern VoidHand Bitmap_Handle1, Bitmap_Handle2, Bitmap_Handle3, Bitmap_Handle4, Bitmap_Handle5, Bitmap_Handle6, Bitmap_Handle7,
	Bitmap_Handle8, Bitmap_Handle9;

extern BitmapPtr Bitmap [10];

extern int current [9] [11];
extern int BlocksRemaining;

extern struct {
	Boolean
		PieceMoveAnim, // whether the movement of a piece by the user is animated
		GravityAnim, // whether the gravity effect is animated
		EliminationAnim, // whether piece elimination is animated
		Sound; // whether sound is on or off
	int CurrentLevel, SolvedLevels; // stores the level the user was last at
} VexedPreferences;

void LoadBitmapResources()
{
Bitmap_Handle1 = DmGet1Resource ('Tbmp', BitmapBlock1);
Bitmap_Handle2 = DmGet1Resource ('Tbmp', BitmapBlock2);
Bitmap_Handle3 = DmGet1Resource ('Tbmp', BitmapBlock3);
Bitmap_Handle4 = DmGet1Resource ('Tbmp', BitmapBlock4);
Bitmap_Handle5 = DmGet1Resource ('Tbmp', BitmapBlock5);
Bitmap_Handle6 = DmGet1Resource ('Tbmp', BitmapBlock6);
Bitmap_Handle7 = DmGet1Resource ('Tbmp', BitmapBlock7);
Bitmap_Handle8 = DmGet1Resource ('Tbmp', BitmapBlock8);
Bitmap_Handle9 = DmGet1Resource ('Tbmp', BitmapBlock9);

Bitmap[1] = MemHandleLock(Bitmap_Handle1);
Bitmap[2] = MemHandleLock(Bitmap_Handle2);
Bitmap[3] = MemHandleLock(Bitmap_Handle3);
Bitmap[4] = MemHandleLock(Bitmap_Handle4);
Bitmap[5] = MemHandleLock(Bitmap_Handle5);
Bitmap[6] = MemHandleLock(Bitmap_Handle6);
Bitmap[7] = MemHandleLock(Bitmap_Handle7);
Bitmap[8] = MemHandleLock(Bitmap_Handle8);
Bitmap[9] = MemHandleLock(Bitmap_Handle9);
}

void UnloadBitmapResources()
{
MemHandleUnlock(Bitmap_Handle1);
MemHandleUnlock(Bitmap_Handle2);
MemHandleUnlock(Bitmap_Handle3);
MemHandleUnlock(Bitmap_Handle4);
MemHandleUnlock(Bitmap_Handle5);
MemHandleUnlock(Bitmap_Handle6);
MemHandleUnlock(Bitmap_Handle7);
MemHandleUnlock(Bitmap_Handle8);
MemHandleUnlock(Bitmap_Handle9);
}

void DrawBlinds()
{
int i, x;
RectangleType r;

r.extent.y = 128;
r.topLeft.y = 16;

for (i = 0; i <= 16; i ++)
	{
	r.extent.x = i;
	for (x = 0; x <= 144; x += 16)
		{
		r.topLeft.x = x;
		WinEraseRectangle(&r, 0);
		}
	SysTaskDelay(1);
	}
}

void DrawLevel()
{
int i, x, y;
RectangleType r;

BlocksRemaining = 0;

DrawBlinds(); // do the cool blinds effect

for (y = 1; y <= 8; y ++)
	for (x = 0; x <= 9; x ++)
		{
		if (!current[y - 1][x] == 0)
			{
			if (!(current[y - 1][x] == 9)) BlocksRemaining ++; // if its a playable block then add one to BlocksRemaining
			WinDrawBitmap(Bitmap[current[y - 1][x]], x * 16, y * 16);
			}
		else
			{
			r.topLeft.x = x * 16;
			r.topLeft.y = y * 16;
			r.extent.x = 16;
			r.extent.y = 16;
			WinEraseRectangle(&r, 0);
			}
		}
UpdateStats(); // display the on screen stats
}

void UpdateStats()
{
char text [30];
RectangleType r = { 55, 1, 105, 11 };

StrPrintF(text, "Level: %d - Blocks Left: %d", VexedPreferences.CurrentLevel, BlocksRemaining); // make up the text string
WinEraseRectangle(&r, 0); // rub out what was there before
WinDrawChars(text, StrLen(text), 55, 1); // draw the text string
}

void DrawIntro()
{
VoidHand Title_Handle;
BitmapPtr Title;
char text [20];

Title_Handle = DmGet1Resource('Tbmp', BitmapTitle); // load bitmap resource and get handle
Title = MemHandleLock(Title_Handle); // lock the bitmap resource into memory and get a pointer to it

WinDrawBitmap(Title, 27, 52); // draw the bitmap
StrPrintF(text, "by James McCombe");
WinDrawChars(text, StrLen(text), 40, 90);

StrPrintF(text, "Tap Screen");
WinDrawChars(text, StrLen(text), 57, 100); // display some text

MemHandleUnlock(Title_Handle); // unload the bitmap from memory (unlock)
}

void EndSequence()
{
char text [15];
int x1, x2, randnum1, randnum2;
SWord penx, peny;
Boolean bstate;
RectangleType r;

while (bstate) EvtGetPen(&penx, &peny, &bstate);

// draw some text of adulation to the rather smart user!
r.topLeft.x = 0;
r.topLeft.y = 16;
r.extent.x = 160;
r.extent.y = 128;
WinEraseRectangle(&r, 0);

FntSetFont(2);
StrPrintF(text, "Well Done!");
WinDrawChars(text, StrLen(text), 50, 75);
FntSetFont(0);
StrPrintF(text, "Game Complete");
WinDrawChars(text, StrLen(text), 46, 90);


// fire some random blocks along the top and bottom of the screen!
while (1)
	{
	randnum1 = (int) ((SysRandom(0) / 4096) + 1);
	randnum2 = (int) ((SysRandom(0) / 4096) + 1);
	SoundEffect(HighBeep);
	for (x1 = 0; x1 <= 160; x1 ++)
		{
		r.topLeft.x = x1 - 1;
		r.topLeft.y = 16;
		r.extent.x = 16;
		r.extent.y = 16;
		WinEraseRectangle(&r, 0);
		WinDrawBitmap(Bitmap[randnum1], x1, 16);

		x2 = (160 - x1);

		r.topLeft.x = x2 + 1;
		r.topLeft.y = 128;
		r.extent.x = 16;
		r.extent.y = 16;
		WinEraseRectangle(&r, 0);
		WinDrawBitmap(Bitmap[randnum2], x2, 128);

		SysTaskDelay(2);
		EvtGetPen(&penx, &peny, &bstate);
		if (bstate) break;
		}
	SoundEffect(LowBeep);
	r.topLeft.x = 0;
	r.topLeft.y = 128;
	r.extent.x = 16;
	r.extent.y = 16;
	WinEraseRectangle(&r, 0);
	if (bstate) break;
	}
SoundEffect(HighSweepLow);
}
