
/*
 * Dialog.java
 *
 *
 * Copyright (c) 2000, David M. Archer
 *
 * See LICENSE file for more information.
 *
 */

package wabadma.ui;

import waba.ui.*;
import waba.fx.*;

import extra.ui.*;

/**
 * A waba control that displays a dialog box with a single OK button.
 *
 * @author David M. Archer <dmarcher@pobox.com>
 *
 * @version 0.3, 2000 Nov 23 - added a postEvent of PRESSED when OK is hit.
 * @version 0.2, 2000 Nov 21 - added String[] constructor.
 * @version 0.1, 2000 Nov 07 - initial version.
 */

public class Dialog extends Container {
	private	boolean displayed = false;
	private	Button okButton;
	private final static int BORDER_SZ = 2; // size of border
	private final static int LEADER_SZ = 5; // leader from edge
	private final static int TITLE_H = 13; 	// height of title
	private final static int BUTTON_H = 15;	// height of button

	public OutputBox out; //an output box containing the text in the dialog.


	/**
	 * Create an empty titled Dialog box of the given metrics using 
	 * the given font.
	 *
	 */
	public Dialog(String title, int x, int y, int w, int h, Font font) {
		this.setRect(x,y,w,h);
		Title t = new Title(title);
		t.setRect(0,0,width,TITLE_H);
		add(t);
		int ow = width - (LEADER_SZ*2);
		int oh = height - (LEADER_SZ*2) - (BUTTON_H + TITLE_H);
		out = new OutputBox(LEADER_SZ, LEADER_SZ+TITLE_H+1, ow, oh, font);
		add(out);
		okButton = new Button("OK");
	        okButton.setRect(64,h-16-LEADER_SZ,32,BUTTON_H);
                add(okButton);
        	displayed = true;
	}


	/**
	 * Create an empty Dialog box of the given metrics with the given 
	 * title in a default font.
	 *
	 */
	public Dialog(String title, int x, int y, int w, int h) {
		this(title,x,y,w,h, new Font("Helvetica", Font.PLAIN, 12));
	}


	/**
	 * Create a Dialog box of the given metrics with the given
	 * title in a default font with the given text centered or not.
	 */
	public Dialog(String title, int x, int y, int w, int h, String text, boolean centered) {
		this(title, x, y, w, h);
		if (centered)
			out.printlnCentered(text);
		else out.println(text);
	}


	/**
	 * Create a Dialog box of the given metrics with the given
	 * title in a default font with the given text.
	 */
	public Dialog(String title, int x, int y, int w, int h, String text) {
		this(title, x, y, w, h);
		out.println(text);
	}


	/**
	 * Create a Dialog box of the given metrics with the given
	 * title in a default font with the pre-formatted text given
	 * as a String[].
	 */
	public Dialog(String title, int x, int y, int w, int h, String[] sa) {
		this(title, x, y, w, h);
		out.printFormattedStringArray(sa);
	}


	/**
	 * Return true if the Dialog is currently displayed.
	 *
	 */
	public boolean isActive() {
		return displayed;
	}


	/**
	 * Draw a border around the box.
	 *
	 */
        private void drawOutline(Graphics g, int x, int y, int w, int h) {
                g.setColor(0,0,0);
                g.fillRect(x,y,w,h);
                g.setColor(255,255,255);
		// this is a bit mutated now that we get the top border from
		// Title
                g.fillRect(x+BORDER_SZ,y,w-(BORDER_SZ*2),h-BORDER_SZ);
                g.setColor(0,0,0);
        }


        private void clearDialog() {
                displayed = false;
		this.getParent().remove(this);
        }


	public void onPaint(Graphics g) {
		if (displayed) {
			//System.err.println("Dialog onPaint");
                	drawOutline(g,0,TITLE_H,width,height-TITLE_H);
		}
	}


	public void onEvent(Event e) {
		if (displayed) {
			//System.err.println("Dialog onEvent");
			if (e.type == ControlEvent.PRESSED) {
				if (e.target == okButton) {
					postEvent(new ControlEvent(ControlEvent.PRESSED, this));
					clearDialog();
				}
			}
		}
	}
}


