///////////////////////////////////////////////////////////////////////////////
// PalmOS Application to view the Caltrain Schedule
//
// Copyright info:
//    This is free software; you can redistribute it and/or modify
//    it as you like.
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
//
//  FileName:
//    Glob_data.h
// 
//  Description:
//    Global data from the caltrain complete schedule web page.
//    Train numbers and times.
//    NOTE: web data is from Sept. 24, 2000 schedule
//    See current schedule at:
//      http://www.transitinfo.org/cgi-bin/sched?C=CT&R=&M=A
//    See new schedule at:
//      http://www.transitinfo.org/cgi-bin/sched?C=CT&R=&M=A&E=New
//
// History:
//    4/14/2000     - Created.  Rob Baynes
//    6/2/2000      - Use binary data.  Mark Wilson
//    9/12/2000     - Updated schedule.  Robert Baynes.
//    9/19/2000    - Updated to new Sept. 24, 2000 schedule.  Rob
///////////////////////////////////////////////////////////////////////////////

// MEW - a simple hack to make the data binary.  Pack train numbers & times 
//  into words.  Use flags to tell 'em apart (there's only one call-back 
//  function) and to denote Express, TIME_AM)m/pm etc
#define EXPRESS_TRAIN   (0x0400)
#define DISCOUNT_TRAIN  (0x0800)
#define GAME_TRAIN      (0x1000)
#define TIME_AM         (0x2000)
#define TIME_PM         (0x4000)
// a macro that packs together a time (hr in the hi byte)
// a N/A time is 0:99am (!)
#define TIME(_hr,_min,_ap) ((_hr << 8) | _min | _ap)


// The number of North & South trains for each of the day choices:
Word        g_TrainCounts[ c_NumDays ] = { 39,      // Mon-Fri trains
                                           16,      // Sat. trains
                                           10 };    // Sun. trains

// The strings for North & South bound train numbers for each day:
Word        g_Trains[ c_NumDirs ][ c_NumDays ][ c_MaxTrains ] = 
            {   // 2 dirs  North,South
                {   // 3 days  Mon-Fri,Sat,Sun
                    { 
                            // 39 Northbound Mon-Fri train numbers
                        21,   
                        23 | EXPRESS_TRAIN,
                        25   ,
                        27 | EXPRESS_TRAIN,
                        29 | EXPRESS_TRAIN,
                        31   ,  
                        33 | EXPRESS_TRAIN,                           
                        35 | EXPRESS_TRAIN,
                        37 | EXPRESS_TRAIN,
                        39   ,
                        41 | EXPRESS_TRAIN,
                        43 | EXPRESS_TRAIN,
                        45   ,
                        47 | EXPRESS_TRAIN,
                        49   ,
                        51 | DISCOUNT_TRAIN,
                        53 | DISCOUNT_TRAIN,
                        55 | DISCOUNT_TRAIN,
                        57 | DISCOUNT_TRAIN,
                        59 | DISCOUNT_TRAIN,
                        61 | DISCOUNT_TRAIN,
                        63 | DISCOUNT_TRAIN,
                        65 | DISCOUNT_TRAIN,
                        67 | DISCOUNT_TRAIN,
                        69 | DISCOUNT_TRAIN,
                        71 | DISCOUNT_TRAIN,
                        73 | DISCOUNT_TRAIN,
                        75   ,
                        77   ,
                        79   ,
                        81   ,
                        83 | EXPRESS_TRAIN,
                        85   ,
                        87 | EXPRESS_TRAIN,
                        89   ,
                        91   ,
                        95   ,
                        97   ,
                        99 | GAME_TRAIN
                    }, // Northbound Mon-Fri trains (39)
                    { 
                            // Northbound Saturday trains (16)
                        127,  
                        139,
                        147,
                        151,
                        155,
                        159,
                        163,
                        167,
                        171,
                        175,
                        179,
                        183,
                        187,
                        191,
                        195,
                        199 | GAME_TRAIN
                    }, // Northbound Saturday trains (16
                    { 
                            // Northbound Sunday trains (10)
                        243,  
                        251,
                        255,
                        259,
                        263,
                        271,
                        279,
                        287,
                        295,
                        299 | GAME_TRAIN
                    } // Northbound Sunday trains (10)
                },
                {
                    { 
                            // Southbound Mon-Fri trains (39)
                        20   ,   
                        22   ,   
                        24   ,   
                        26 | EXPRESS_TRAIN,   
                        28   ,   
                        30 | EXPRESS_TRAIN,   
                        32   ,   
                        34   ,   
                        36   ,   
                        38 | DISCOUNT_TRAIN,   
                        40 | DISCOUNT_TRAIN,   
                        42 | DISCOUNT_TRAIN,   
                        44 | DISCOUNT_TRAIN,   
                        46 | DISCOUNT_TRAIN,   
                        48 | DISCOUNT_TRAIN,   
                        50 | DISCOUNT_TRAIN,   
                        52 | DISCOUNT_TRAIN,   
                        54 | DISCOUNT_TRAIN,   
                        56 | DISCOUNT_TRAIN,   
                        58 | DISCOUNT_TRAIN,   
                        60 | DISCOUNT_TRAIN,   
                        62   ,   
                        64   ,   
                        66   ,   
                        68 | EXPRESS_TRAIN,   
                        70   ,   
                        72 | EXPRESS_TRAIN,   
                        74   ,   
                        76 | EXPRESS_TRAIN,   
                        78 | EXPRESS_TRAIN,   
                        80 | EXPRESS_TRAIN,   
                        82   ,   
                        84 | EXPRESS_TRAIN,   
                        86   ,   
                        88   ,   
                        92   ,   
                        94   ,   
                        96   ,   
                        98   ,
                    }, // Southbound Mon-Fri trains (39)
                    { 
                            // Southbound Saturday trains (16)
                        128,   
                        134,   
                        138,   
                        142,   
                        146,   
                        150,   
                        154,   
                        158,   
                        162,   
                        166,   
                        174,   
                        184,   
                        188,   
                        192,   
                        196,   
                        198,   
                    }, // Southbound Saturday trains (16)
                    { 
                            // Southbound Sunday trains (10)
                        234,   
                        242,   
                        250,   
                        258,   
                        266,   
                        274,   
                        284,   
                        288,   
                        292,   
                        296,   
                    } // Southbound Sunday trains (10)
                }
            };

// The strings for North & South bound train times for each day:
Word        g_Stations[ c_NumDirs ][ c_NumDays ][ c_Stations ][ c_MaxTrains ]= 
            {   // 2 dirs  North,South
                {   // 3 days  Mon-Fri,Sat,Sun
                    {   // 32 stations  0 = SF
                        {   // North M-F San Francisco times
                            TIME(6, 15, TIME_AM),  
                            TIME(6, 43, TIME_AM),  
                            TIME(7, 12, TIME_AM),  
                            TIME(7, 15, TIME_AM),  
                            TIME(7, 33, TIME_AM),  
                            TIME(7, 50, TIME_AM),  
                            TIME(7, 57, TIME_AM),  
                            TIME(8, 1, TIME_AM),  
                            TIME(8, 10, TIME_AM),  
                            TIME(8, 39, TIME_AM),  
                            TIME(8, 45, TIME_AM),  
                            TIME(8, 51, TIME_AM),  
                            TIME(9, 9, TIME_AM),  
                            TIME(9, 15, TIME_AM),  
                            TIME(9, 53, TIME_AM),  
                            TIME(10, 38, TIME_AM),  
                            TIME(11, 15, TIME_AM),  
                            TIME(11, 45, TIME_AM),  
                            TIME(12, 15, TIME_PM),  
                            TIME(12, 45, TIME_PM),  
                            TIME(1, 15, TIME_PM),  
                            TIME(1, 45, TIME_PM),  
                            TIME(2, 15, TIME_PM),  
                            TIME(2, 45, TIME_PM),  
                            TIME(3, 15, TIME_PM),  
                            TIME(3, 45, TIME_PM),  
                            TIME(4, 15, TIME_PM),  
                            TIME(4, 47, TIME_PM),  
                            TIME(5, 16, TIME_PM),  
                            TIME(5, 48, TIME_PM),  
                            TIME(6, 19, TIME_PM),  
                            TIME(6, 38, TIME_PM),  
                            TIME(7, 1, TIME_PM),  
                            TIME(7, 14, TIME_PM),  
                            TIME(7, 52, TIME_PM),  
                            TIME(8, 37, TIME_PM),  
                            TIME(9, 36, TIME_PM),  
                            TIME(10, 35, TIME_PM),  
                            TIME(12, 5, TIME_AM),  
                        },
                        {   // North M-F 22nd Street
                            TIME(6, 4, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 0, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 23, TIME_AM) ,  
                            TIME(7, 40, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 27, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 57, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(9, 41, TIME_AM) ,  
                            TIME(10, 27, TIME_AM),  
                            TIME(11, 4, TIME_AM),  
                            TIME(11, 34, TIME_AM),  
                            TIME(12, 4, TIME_PM),  
                            TIME(12, 34, TIME_PM),  
                            TIME(1, 4, TIME_PM) ,  
                            TIME(1, 34, TIME_PM) ,  
                            TIME(2, 4, TIME_PM) ,  
                            TIME(2, 34, TIME_PM) ,  
                            TIME(3, 4, TIME_PM) ,  
                            TIME(3, 34, TIME_PM) ,  
                            TIME(4, 4, TIME_PM) ,  
                            TIME(4, 36, TIME_PM) ,  
                            TIME(5, 5, TIME_PM) ,  
                            TIME(5, 36, TIME_PM) ,  
                            TIME(6, 7, TIME_PM) ,  
                            TIME(6, 26, TIME_PM) ,  
                            TIME(6, 49, TIME_PM) ,  
                            TIME(7, 3, TIME_PM) ,  
                            TIME(7, 41, TIME_PM) ,  
                            TIME(8, 25, TIME_PM) ,  
                            TIME(9, 25, TIME_PM) ,  
                            TIME(10, 24, TIME_PM),  
                            TIME(11, 54, TIME_PM),  
                        },
                        {   // North M-F Paul Avenue
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(7, 36, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(4, 32, TIME_PM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(6, 3, TIME_PM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                        },
                        {   // North M-F Bayshore
                            TIME(5, 58, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 54, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 17, TIME_AM) ,  
                            TIME(7, 33, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 21, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 51, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(9, 35, TIME_AM) ,  
                            TIME(10, 21, TIME_AM),  
                            TIME(10, 58, TIME_AM),  
                            TIME(11, 28, TIME_AM),  
                            TIME(11, 58, TIME_AM),  
                            TIME(12, 28, TIME_PM),  
                            TIME(12, 58, TIME_PM),  
                            TIME(1, 28, TIME_PM) ,  
                            TIME(1, 58, TIME_PM) ,  
                            TIME(2, 28, TIME_PM) ,  
                            TIME(2, 58, TIME_PM) ,  
                            TIME(3, 28, TIME_PM) ,  
                            TIME(3, 58, TIME_PM) ,  
                            TIME(4, 29, TIME_PM) ,  
                            TIME(4, 59, TIME_PM) ,  
                            TIME(5, 30, TIME_PM) ,  
                            TIME(6, 0, TIME_PM) ,  
                            TIME(6, 20, TIME_PM) ,  
                            TIME(6, 43, TIME_PM) ,  
                            TIME(6, 57, TIME_PM) ,  
                            TIME(7, 35, TIME_PM) ,  
                            TIME(8, 19, TIME_PM) ,  
                            TIME(9, 19, TIME_PM) ,  
                            TIME(10, 18, TIME_PM),  
                            TIME(11, 48, TIME_PM),  
                        },
                        {   // North M-F South SF
                            TIME(5, 53, TIME_AM) ,  
                            TIME(6, 26, TIME_AM) ,  
                            TIME(6, 49, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 12, TIME_AM) ,  
                            TIME(7, 28, TIME_AM) ,  
                            TIME(7, 37, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 16, TIME_AM) ,  
                            TIME(8, 26, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 46, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(9, 30, TIME_AM) ,  
                            TIME(10, 16, TIME_AM),  
                            TIME(10, 53, TIME_AM),  
                            TIME(11, 23, TIME_AM),  
                            TIME(11, 53, TIME_AM),  
                            TIME(12, 23, TIME_PM),  
                            TIME(12, 53, TIME_PM),  
                            TIME(1, 23, TIME_PM) ,  
                            TIME(1, 53, TIME_PM) ,  
                            TIME(2, 23, TIME_PM) ,  
                            TIME(2, 53, TIME_PM) ,  
                            TIME(3, 23, TIME_PM) ,  
                            TIME(3, 53, TIME_PM) ,  
                            TIME(4, 24, TIME_PM) ,  
                            TIME(4, 54, TIME_PM) ,  
                            TIME(5, 25, TIME_PM) ,  
                            TIME(5, 55, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 38, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 30, TIME_PM) ,  
                            TIME(8, 14, TIME_PM) ,  
                            TIME(9, 14, TIME_PM) ,  
                            TIME(10, 13, TIME_PM),  
                            TIME(11, 43, TIME_PM),  
                        },
                        {   // North M-F San Bruno
                            TIME(5, 50, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 46, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 9, TIME_AM) ,  
                            TIME(7, 25, TIME_AM) ,  
                            TIME(7, 34, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 13, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 31, TIME_AM) ,  
                            TIME(8, 43, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(9, 27, TIME_AM) ,  
                            TIME(10, 13, TIME_AM),  
                            TIME(10, 50, TIME_AM),  
                            TIME(11, 20, TIME_AM),  
                            TIME(11, 50, TIME_AM),  
                            TIME(12, 20, TIME_PM),  
                            TIME(12, 50, TIME_PM),  
                            TIME(1, 20, TIME_PM) ,  
                            TIME(1, 50, TIME_PM) ,  
                            TIME(2, 20, TIME_PM) ,  
                            TIME(2, 50, TIME_PM) ,  
                            TIME(3, 20, TIME_PM) ,  
                            TIME(3, 50, TIME_PM) ,  
                            TIME(4, 21, TIME_PM) ,  
                            TIME(4, 51, TIME_PM) ,  
                            TIME(5, 22, TIME_PM) ,  
                            TIME(5, 52, TIME_PM) ,  
                            TIME(6, 12, TIME_PM) ,  
                            TIME(6, 35, TIME_PM) ,  
                            TIME(6, 49, TIME_PM) ,  
                            TIME(7, 27, TIME_PM) ,  
                            TIME(8, 11, TIME_PM) ,  
                            TIME(9, 11, TIME_PM) ,  
                            TIME(10, 10, TIME_PM),  
                            TIME(11, 40, TIME_PM),  
                        },
                        {   // North M-F Millbrae
                            TIME(5, 45, TIME_AM) ,  
                            TIME(6, 19, TIME_AM) ,  
                            TIME(6, 41, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 4, TIME_AM) ,  
                            TIME(7, 20, TIME_AM) ,  
                            TIME(7, 29, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 8, TIME_AM) ,  
                            TIME(8, 19, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 38, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(9, 22, TIME_AM) ,  
                            TIME(10, 8, TIME_AM),  
                            TIME(10, 45, TIME_AM),  
                            TIME(11, 15, TIME_AM),  
                            TIME(11, 45, TIME_AM),  
                            TIME(12, 15, TIME_PM),  
                            TIME(12, 45, TIME_PM),  
                            TIME(1, 15, TIME_PM) ,  
                            TIME(1, 45, TIME_PM) ,  
                            TIME(2, 15, TIME_PM) ,  
                            TIME(2, 45, TIME_PM) ,  
                            TIME(3, 15, TIME_PM) ,  
                            TIME(3, 45, TIME_PM) ,  
                            TIME(4, 16, TIME_PM) ,  
                            TIME(4, 46, TIME_PM) ,  
                            TIME(5, 17, TIME_PM) ,  
                            TIME(5, 47, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 30, TIME_PM) ,  
                            TIME(6, 44, TIME_PM) ,  
                            TIME(7, 22, TIME_PM) ,  
                            TIME(8, 6, TIME_PM) ,  
                            TIME(9, 6, TIME_PM) ,  
                            TIME(10, 5, TIME_PM),  
                            TIME(11, 35, TIME_PM),  
                        },
                        {   // North M-F Broadway
                            TIME(5, 41, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 37, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 0, TIME_AM) ,  
                            TIME(7, 16, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 35, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 4, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 23, TIME_AM) ,  
                            TIME(8, 34, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(9, 18, TIME_AM) ,  
                            TIME(10, 3, TIME_AM),  
                            TIME(10, 40, TIME_AM),  
                            TIME(11, 10, TIME_AM),  
                            TIME(11, 40, TIME_AM),  
                            TIME(12, 10, TIME_PM),  
                            TIME(12, 40, TIME_PM),  
                            TIME(1, 10, TIME_PM) ,  
                            TIME(1, 40, TIME_PM) ,  
                            TIME(2, 10, TIME_PM) ,  
                            TIME(2, 40, TIME_PM) ,  
                            TIME(3, 10, TIME_PM) ,  
                            TIME(3, 40, TIME_PM) ,  
                            TIME(4, 11, TIME_PM) ,  
                            TIME(4, 41, TIME_PM) ,  
                            TIME(5, 12, TIME_PM) ,  
                            TIME(5, 42, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 25, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 19, TIME_PM) ,  
                            TIME(8, 2, TIME_PM) ,  
                            TIME(9, 1, TIME_PM) ,  
                            TIME(10, 1, TIME_PM),  
                            TIME(11, 31, TIME_PM),  
                        },
                        {   // North M-F Burlingame
                            TIME(5, 38, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 34, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 57, TIME_AM) ,  
                            TIME(7, 13, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 32, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 1, TIME_AM) ,  
                            TIME(8, 12, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 32, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(9, 15, TIME_AM) ,  
                            TIME(10, 0, TIME_AM),  
                            TIME(10, 37, TIME_AM),  
                            TIME(11, 7, TIME_AM),  
                            TIME(11, 37, TIME_AM),  
                            TIME(12, 7, TIME_PM),  
                            TIME(12, 37, TIME_PM),  
                            TIME(1, 7, TIME_PM) ,  
                            TIME(1, 37, TIME_PM) ,  
                            TIME(2, 7, TIME_PM) ,  
                            TIME(2, 37, TIME_PM) ,  
                            TIME(3, 7, TIME_PM) ,  
                            TIME(3, 37, TIME_PM) ,  
                            TIME(4, 8, TIME_PM) ,  
                            TIME(4, 38, TIME_PM) ,  
                            TIME(5, 9, TIME_PM) ,  
                            TIME(5, 39, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 22, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 16, TIME_PM) ,  
                            TIME(7, 59, TIME_PM) ,  
                            TIME(8, 58, TIME_PM) ,  
                            TIME(9, 58, TIME_PM) ,  
                            TIME(11, 28, TIME_PM),  
                        },
                        {   // North M-F San Mateo
                            TIME(5, 35, TIME_AM) ,  
                            TIME(6, 12, TIME_AM) ,  
                            TIME(6, 31, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 54, TIME_AM) ,  
                            TIME(7, 10, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 29, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 58, TIME_AM) ,  
                            TIME(8, 9, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 29, TIME_AM) ,  
                            TIME(8, 44, TIME_AM) ,  
                            TIME(9, 12, TIME_AM) ,  
                            TIME(9, 57, TIME_AM) ,  
                            TIME(10, 34, TIME_AM),  
                            TIME(11, 4, TIME_AM),  
                            TIME(11, 34, TIME_AM),  
                            TIME(12, 4, TIME_PM),  
                            TIME(12, 34, TIME_PM),  
                            TIME(1, 4, TIME_PM) ,  
                            TIME(1, 34, TIME_PM) ,  
                            TIME(2, 4, TIME_PM) ,  
                            TIME(2, 34, TIME_PM) ,  
                            TIME(3, 4, TIME_PM) ,  
                            TIME(3, 34, TIME_PM) ,  
                            TIME(4, 5, TIME_PM) ,  
                            TIME(4, 35, TIME_PM) ,  
                            TIME(5, 6, TIME_PM) ,  
                            TIME(5, 36, TIME_PM) ,  
                            TIME(6, 1, TIME_PM) ,  
                            TIME(6, 19, TIME_PM) ,  
                            TIME(6, 37, TIME_PM) ,  
                            TIME(7, 13, TIME_PM) ,  
                            TIME(7, 56, TIME_PM) ,  
                            TIME(8, 55, TIME_PM) ,  
                            TIME(9, 55, TIME_PM) ,  
                            TIME(11, 25, TIME_PM),  
                        },
                        {   // North M-F Hayward Park
                            TIME(5, 33, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 29, TIME_AM) ,  
                            TIME(6, 43, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 7, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 26, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 55, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 17, TIME_AM) ,  
                            TIME(8, 27, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(9, 10, TIME_AM) ,  
                            TIME(9, 55, TIME_AM) ,  
                            TIME(10, 32, TIME_AM),  
                            TIME(11, 2, TIME_AM),  
                            TIME(11, 32, TIME_AM),  
                            TIME(12, 2, TIME_PM),  
                            TIME(12, 32, TIME_PM),  
                            TIME(1, 2, TIME_PM) ,  
                            TIME(1, 32, TIME_PM) ,  
                            TIME(2, 2, TIME_PM) ,  
                            TIME(2, 32, TIME_PM) ,  
                            TIME(3, 2, TIME_PM) ,  
                            TIME(3, 32, TIME_PM) ,  
                            TIME(4, 3, TIME_PM) ,  
                            TIME(4, 33, TIME_PM) ,  
                            TIME(5, 4, TIME_PM) ,  
                            TIME(5, 34, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 17, TIME_PM) ,  
                            TIME(6, 35, TIME_PM) ,  
                            TIME(7, 11, TIME_PM) ,  
                            TIME(7, 54, TIME_PM) ,  
                            TIME(8, 53, TIME_PM) ,  
                            TIME(9, 53, TIME_PM) ,  
                            TIME(11, 23, TIME_PM),  
                        },
                        {   // North M-F Hillsdale
                            TIME(5, 30, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 26, TIME_AM) ,  
                            TIME(6, 41, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 4, TIME_AM) ,  
                            TIME(7, 19, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 38, TIME_AM) ,  
                            TIME(7, 52, TIME_AM) ,  
                            TIME(8, 5, TIME_AM) ,  
                            TIME(8, 14, TIME_AM) ,  
                            TIME(8, 24, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(9, 7, TIME_AM) ,  
                            TIME(9, 52, TIME_AM) ,  
                            TIME(10, 29, TIME_AM),  
                            TIME(10, 59, TIME_AM),  
                            TIME(11, 29, TIME_AM),  
                            TIME(11, 59, TIME_AM),  
                            TIME(12, 29, TIME_PM),  
                            TIME(12, 59, TIME_PM),  
                            TIME(1, 29, TIME_PM) ,  
                            TIME(1, 59, TIME_PM) ,  
                            TIME(2, 29, TIME_PM) ,  
                            TIME(2, 59, TIME_PM) ,  
                            TIME(3, 29, TIME_PM) ,  
                            TIME(4, 0, TIME_PM) ,  
                            TIME(4, 30, TIME_PM) ,  
                            TIME(5, 1, TIME_PM) ,  
                            TIME(5, 31, TIME_PM) ,  
                            TIME(5, 57, TIME_PM) ,  
                            TIME(6, 14, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 8, TIME_PM) ,  
                            TIME(7, 51, TIME_PM) ,  
                            TIME(8, 50, TIME_PM) ,  
                            TIME(9, 50, TIME_PM) ,  
                            TIME(11, 20, TIME_PM),  
                        },
                        {   // North M-F Belmont
                            TIME(5, 27, TIME_AM) ,  
                            TIME(6, 6, TIME_AM) ,  
                            TIME(6, 23, TIME_AM) ,  
                            TIME(6, 38, TIME_AM) ,  
                            TIME(6, 48, TIME_AM) ,  
                            TIME(7, 1, TIME_AM) ,  
                            TIME(7, 16, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 49, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 11, TIME_AM) ,  
                            TIME(8, 21, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(9, 4, TIME_AM) ,  
                            TIME(9, 49, TIME_AM) ,  
                            TIME(10, 26, TIME_AM),  
                            TIME(10, 56, TIME_AM),  
                            TIME(11, 26, TIME_AM),  
                            TIME(11, 56, TIME_AM),  
                            TIME(12, 26, TIME_PM),  
                            TIME(12, 56, TIME_PM),  
                            TIME(1, 26, TIME_PM) ,  
                            TIME(1, 56, TIME_PM) ,  
                            TIME(2, 26, TIME_PM) ,  
                            TIME(2, 56, TIME_PM) ,  
                            TIME(3, 26, TIME_PM) ,  
                            TIME(3, 57, TIME_PM) ,  
                            TIME(4, 27, TIME_PM) ,  
                            TIME(4, 58, TIME_PM) ,  
                            TIME(5, 28, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 11, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 5, TIME_PM) ,  
                            TIME(7, 48, TIME_PM) ,  
                            TIME(8, 47, TIME_PM) ,  
                            TIME(9, 47, TIME_PM) ,  
                            TIME(11, 17, TIME_PM),  
                        },
                        {   // North M-F San Carlos
                            TIME(5, 24, TIME_AM) ,  
                            TIME(6, 3, TIME_AM) ,  
                            TIME(6, 20, TIME_AM) ,  
                            TIME(6, 35, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 58, TIME_AM) ,  
                            TIME(7, 13, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 46, TIME_AM) ,  
                            TIME(8, 0, TIME_AM) ,  
                            TIME(8, 8, TIME_AM) ,  
                            TIME(8, 18, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(9, 1, TIME_AM) ,  
                            TIME(9, 46, TIME_AM) ,  
                            TIME(10, 23, TIME_AM),  
                            TIME(10, 53, TIME_AM),  
                            TIME(11, 23, TIME_AM),  
                            TIME(11, 53, TIME_AM),  
                            TIME(12, 23, TIME_PM),  
                            TIME(12, 53, TIME_PM),  
                            TIME(1, 23, TIME_PM) ,  
                            TIME(1, 53, TIME_PM) ,  
                            TIME(2, 23, TIME_PM) ,  
                            TIME(2, 53, TIME_PM) ,  
                            TIME(3, 23, TIME_PM) ,  
                            TIME(3, 54, TIME_PM) ,  
                            TIME(4, 24, TIME_PM) ,  
                            TIME(4, 55, TIME_PM) ,  
                            TIME(5, 25, TIME_PM) ,  
                            TIME(5, 51, TIME_PM) ,  
                            TIME(6, 8, TIME_PM) ,  
                            TIME(6, 30, TIME_PM) ,  
                            TIME(7, 2, TIME_PM) ,  
                            TIME(7, 45, TIME_PM) ,  
                            TIME(8, 44, TIME_PM) ,  
                            TIME(9, 44, TIME_PM) ,  
                            TIME(11, 14, TIME_PM),  
                        },
                        {   // North M-F Redwood City
                            TIME(5, 20, TIME_AM) ,  
                            TIME(5, 59, TIME_AM) ,  
                            TIME(6, 16, TIME_AM) ,  
                            TIME(6, 31, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 54, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 16, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 42, TIME_AM) ,  
                            TIME(7, 56, TIME_AM) ,  
                            TIME(8, 4, TIME_AM) ,  
                            TIME(8, 14, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 57, TIME_AM) ,  
                            TIME(9, 42, TIME_AM) ,  
                            TIME(10, 19, TIME_AM),  
                            TIME(10, 49, TIME_AM),  
                            TIME(11, 19, TIME_AM),  
                            TIME(11, 49, TIME_AM),  
                            TIME(12, 19, TIME_PM),  
                            TIME(12, 49, TIME_PM),  
                            TIME(1, 19, TIME_PM) ,  
                            TIME(1, 49, TIME_PM) ,  
                            TIME(2, 19, TIME_PM) ,  
                            TIME(2, 49, TIME_PM) ,  
                            TIME(3, 19, TIME_PM) ,  
                            TIME(3, 50, TIME_PM) ,  
                            TIME(4, 20, TIME_PM) ,  
                            TIME(4, 51, TIME_PM) ,  
                            TIME(5, 21, TIME_PM) ,  
                            TIME(5, 47, TIME_PM) ,  
                            TIME(6, 4, TIME_PM) ,  
                            TIME(6, 26, TIME_PM) ,  
                            TIME(6, 58, TIME_PM) ,  
                            TIME(7, 41, TIME_PM) ,  
                            TIME(8, 40, TIME_PM) ,  
                            TIME(9, 40, TIME_PM) ,  
                            TIME(11, 10, TIME_PM),  
                        },
                        {   // North M-F Atherton
                            TIME(5, 14, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 10, TIME_AM) ,  
                            TIME(6, 25, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 47, TIME_AM) ,  
                            TIME(7, 5, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 35, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 7, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 50, TIME_AM) ,  
                            TIME(9, 36, TIME_AM) ,  
                            TIME(10, 13, TIME_AM),  
                            TIME(10, 43, TIME_AM),  
                            TIME(11, 13, TIME_AM),  
                            TIME(11, 43, TIME_AM),  
                            TIME(12, 13, TIME_PM),  
                            TIME(12, 43, TIME_PM),  
                            TIME(1, 13, TIME_PM) ,  
                            TIME(1, 43, TIME_PM) ,  
                            TIME(2, 13, TIME_PM) ,  
                            TIME(2, 43, TIME_PM) ,  
                            TIME(3, 13, TIME_PM) ,  
                            TIME(3, 43, TIME_PM) ,  
                            TIME(4, 13, TIME_PM) ,  
                            TIME(4, 44, TIME_PM) ,  
                            TIME(5, 14, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 57, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 51, TIME_PM) ,  
                            TIME(7, 34, TIME_PM) ,  
                            TIME(8, 34, TIME_PM) ,  
                            TIME(9, 34, TIME_PM) ,  
                            TIME(11, 4, TIME_PM),  
                        },
                        {   // North M-F Menlo Park
                            TIME(5, 12, TIME_AM) ,  
                            TIME(5, 52, TIME_AM) ,  
                            TIME(6, 7, TIME_AM) ,  
                            TIME(6, 22, TIME_AM) ,  
                            TIME(6, 36, TIME_AM) ,  
                            TIME(6, 44, TIME_AM) ,  
                            TIME(7, 2, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 25, TIME_AM) ,  
                            TIME(7, 32, TIME_AM) ,  
                            TIME(7, 49, TIME_AM) ,  
                            TIME(7, 57, TIME_AM) ,  
                            TIME(8, 4, TIME_AM) ,  
                            TIME(8, 31, TIME_AM) ,  
                            TIME(8, 47, TIME_AM) ,  
                            TIME(9, 33, TIME_AM) ,  
                            TIME(10, 10, TIME_AM),  
                            TIME(10, 40, TIME_AM),  
                            TIME(11, 10, TIME_AM),  
                            TIME(11, 40, TIME_AM),  
                            TIME(12, 10, TIME_PM),  
                            TIME(12, 40, TIME_PM),  
                            TIME(1, 10, TIME_PM) ,  
                            TIME(1, 40, TIME_PM) ,  
                            TIME(2, 10, TIME_PM) ,  
                            TIME(2, 40, TIME_PM) ,  
                            TIME(3, 10, TIME_PM) ,  
                            TIME(3, 40, TIME_PM) ,  
                            TIME(4, 10, TIME_PM) ,  
                            TIME(4, 41, TIME_PM) ,  
                            TIME(5, 11, TIME_PM) ,  
                            TIME(5, 40, TIME_PM) ,  
                            TIME(5, 54, TIME_PM) ,  
                            TIME(6, 19, TIME_PM) ,  
                            TIME(6, 48, TIME_PM) ,  
                            TIME(7, 32, TIME_PM) ,  
                            TIME(8, 32, TIME_PM) ,  
                            TIME(9, 32, TIME_PM) ,  
                            TIME(11, 2, TIME_PM),  
                        },
                        {   // North M-F Palo Alto
                            TIME(5, 9, TIME_AM) ,  
                            TIME(5, 49, TIME_AM) ,  
                            TIME(6, 4, TIME_AM) ,  
                            TIME(6, 19, TIME_AM) ,  
                            TIME(6, 33, TIME_AM) ,  
                            TIME(6, 41, TIME_AM) ,  
                            TIME(6, 59, TIME_AM) ,  
                            TIME(7, 8, TIME_AM) ,  
                            TIME(7, 22, TIME_AM) ,  
                            TIME(7, 29, TIME_AM) ,  
                            TIME(7, 46, TIME_AM) ,  
                            TIME(7, 54, TIME_AM) ,  
                            TIME(8, 1, TIME_AM) ,  
                            TIME(8, 28, TIME_AM) ,  
                            TIME(8, 44, TIME_AM) ,  
                            TIME(9, 30, TIME_AM) ,  
                            TIME(10, 7, TIME_AM),  
                            TIME(10, 37, TIME_AM),  
                            TIME(11, 7, TIME_AM),  
                            TIME(11, 37, TIME_AM),  
                            TIME(12, 7, TIME_PM),  
                            TIME(12, 37, TIME_PM),  
                            TIME(1, 7, TIME_PM) ,  
                            TIME(1, 37, TIME_PM) ,  
                            TIME(2, 7, TIME_PM) ,  
                            TIME(2, 37, TIME_PM) ,  
                            TIME(3, 7, TIME_PM) ,  
                            TIME(3, 37, TIME_PM) ,  
                            TIME(4, 7, TIME_PM) ,  
                            TIME(4, 38, TIME_PM) ,  
                            TIME(5, 8, TIME_PM) ,  
                            TIME(5, 37, TIME_PM) ,  
                            TIME(5, 51, TIME_PM) ,  
                            TIME(6, 16, TIME_PM) ,  
                            TIME(6, 45, TIME_PM) ,  
                            TIME(7, 29, TIME_PM) ,  
                            TIME(8, 29, TIME_PM) ,  
                            TIME(9, 29, TIME_PM) ,  
                            TIME(10, 59, TIME_PM),  
                        },
                        {   // North M-F California Av
                            TIME(5, 6, TIME_AM) ,  
                            TIME(5, 46, TIME_AM) ,  
                            TIME(6, 1, TIME_AM) ,  
                            TIME(6, 16, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 38, TIME_AM) ,  
                            TIME(6, 56, TIME_AM) ,  
                            TIME(7, 5, TIME_AM) ,  
                            TIME(7, 19, TIME_AM) ,  
                            TIME(7, 26, TIME_AM) ,  
                            TIME(7, 43, TIME_AM) ,  
                            TIME(7, 51, TIME_AM) ,  
                            TIME(7, 58, TIME_AM) ,  
                            TIME(8, 25, TIME_AM) ,  
                            TIME(8, 41, TIME_AM) ,  
                            TIME(9, 27, TIME_AM) ,  
                            TIME(10, 4, TIME_AM),  
                            TIME(10, 34, TIME_AM),  
                            TIME(11, 4, TIME_AM),  
                            TIME(11, 34, TIME_AM),  
                            TIME(12, 4, TIME_PM),  
                            TIME(12, 34, TIME_PM),  
                            TIME(1, 4, TIME_PM) ,  
                            TIME(1, 34, TIME_PM) ,  
                            TIME(2, 4, TIME_PM) ,  
                            TIME(2, 34, TIME_PM) ,  
                            TIME(3, 4, TIME_PM) ,  
                            TIME(3, 34, TIME_PM) ,  
                            TIME(4, 4, TIME_PM) ,  
                            TIME(4, 35, TIME_PM) ,  
                            TIME(5, 5, TIME_PM) ,  
                            TIME(5, 34, TIME_PM) ,  
                            TIME(5, 48, TIME_PM) ,  
                            TIME(6, 13, TIME_PM) ,  
                            TIME(6, 42, TIME_PM) ,  
                            TIME(7, 26, TIME_PM) ,  
                            TIME(8, 26, TIME_PM) ,  
                            TIME(9, 26, TIME_PM) ,  
                            TIME(10, 56, TIME_PM),  
                        },
                        {   // North M-F San Antonio
                            TIME(5, 3, TIME_AM) ,  
                            TIME(5, 43, TIME_AM) ,  
                            TIME(5, 58, TIME_AM) ,  
                            TIME(6, 13, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 35, TIME_AM) ,  
                            TIME(6, 53, TIME_AM) ,  
                            TIME(7, 2, TIME_AM) ,  
                            TIME(7, 16, TIME_AM) ,  
                            TIME(7, 23, TIME_AM) ,  
                            TIME(7, 40, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 55, TIME_AM) ,  
                            TIME(8, 22, TIME_AM) ,  
                            TIME(8, 38, TIME_AM) ,  
                            TIME(9, 24, TIME_AM) ,  
                            TIME(10, 1, TIME_AM),  
                            TIME(10, 31, TIME_AM),  
                            TIME(11, 1, TIME_AM),  
                            TIME(11, 31, TIME_AM),  
                            TIME(12, 1, TIME_PM),  
                            TIME(12, 31, TIME_PM),  
                            TIME(1, 1, TIME_PM) ,  
                            TIME(1, 31, TIME_PM) ,  
                            TIME(2, 1, TIME_PM) ,  
                            TIME(2, 31, TIME_PM) ,  
                            TIME(3, 1, TIME_PM) ,  
                            TIME(3, 31, TIME_PM) ,  
                            TIME(4, 1, TIME_PM) ,  
                            TIME(4, 32, TIME_PM) ,  
                            TIME(5, 2, TIME_PM) ,  
                            TIME(5, 31, TIME_PM) ,  
                            TIME(5, 45, TIME_PM) ,  
                            TIME(6, 10, TIME_PM) ,  
                            TIME(6, 39, TIME_PM) ,  
                            TIME(7, 23, TIME_PM) ,  
                            TIME(8, 23, TIME_PM) ,  
                            TIME(9, 23, TIME_PM) ,  
                            TIME(10, 53, TIME_PM),  
                        },
                        {   // North M-F Mountain View
                            TIME(4, 59, TIME_AM) ,  
                            TIME(5, 39, TIME_AM) ,  
                            TIME(5, 54, TIME_AM) ,  
                            TIME(6, 9, TIME_AM) ,  
                            TIME(6, 25, TIME_AM) ,  
                            TIME(6, 31, TIME_AM) ,  
                            TIME(6, 49, TIME_AM) ,  
                            TIME(6, 58, TIME_AM) ,  
                            TIME(7, 12, TIME_AM) ,  
                            TIME(7, 19, TIME_AM) ,  
                            TIME(7, 36, TIME_AM) ,  
                            TIME(7, 45, TIME_AM) ,  
                            TIME(7, 51, TIME_AM) ,  
                            TIME(8, 18, TIME_AM) ,  
                            TIME(8, 34, TIME_AM) ,  
                            TIME(9, 20, TIME_AM) ,  
                            TIME(9, 57, TIME_AM) ,  
                            TIME(10, 27, TIME_AM),  
                            TIME(10, 57, TIME_AM),  
                            TIME(11, 27, TIME_AM),  
                            TIME(11, 57, TIME_AM),  
                            TIME(12, 27, TIME_PM),  
                            TIME(12, 57, TIME_PM),  
                            TIME(1, 27, TIME_PM) ,  
                            TIME(1, 57, TIME_PM) ,  
                            TIME(2, 27, TIME_PM) ,  
                            TIME(2, 57, TIME_PM) ,  
                            TIME(3, 27, TIME_PM) ,  
                            TIME(3, 57, TIME_PM) ,  
                            TIME(4, 28, TIME_PM) ,  
                            TIME(4, 58, TIME_PM) ,  
                            TIME(5, 27, TIME_PM) ,  
                            TIME(5, 41, TIME_PM) ,  
                            TIME(6, 6, TIME_PM) ,  
                            TIME(6, 35, TIME_PM) ,  
                            TIME(7, 19, TIME_PM) ,  
                            TIME(8, 19, TIME_PM) ,  
                            TIME(9, 19, TIME_PM) ,  
                            TIME(10, 49, TIME_PM),  
                        },
                        {   // North M-F Sunnyvale
                            TIME(4, 54, TIME_AM) ,  
                            TIME(5, 34, TIME_AM) ,  
                            TIME(5, 49, TIME_AM) ,  
                            TIME(6, 4, TIME_AM) ,  
                            TIME(6, 19, TIME_AM) ,  
                            TIME(6, 26, TIME_AM) ,  
                            TIME(6, 44, TIME_AM) ,  
                            TIME(6, 53, TIME_AM) ,  
                            TIME(7, 6, TIME_AM) ,  
                            TIME(7, 14, TIME_AM) ,  
                            TIME(7, 30, TIME_AM) ,  
                            TIME(7, 39, TIME_AM) ,  
                            TIME(7, 46, TIME_AM) ,  
                            TIME(8, 12, TIME_AM) ,  
                            TIME(8, 29, TIME_AM) ,  
                            TIME(9, 14, TIME_AM) ,  
                            TIME(9, 51, TIME_AM) ,  
                            TIME(10, 21, TIME_AM),  
                            TIME(10, 51, TIME_AM),  
                            TIME(11, 21, TIME_AM),  
                            TIME(11, 51, TIME_AM),  
                            TIME(12, 21, TIME_PM),  
                            TIME(12, 51, TIME_PM),  
                            TIME(1, 21, TIME_PM) ,  
                            TIME(1, 51, TIME_PM) ,  
                            TIME(2, 21, TIME_PM) ,  
                            TIME(2, 51, TIME_PM) ,  
                            TIME(3, 22, TIME_PM) ,  
                            TIME(3, 52, TIME_PM) ,  
                            TIME(4, 22, TIME_PM) ,  
                            TIME(4, 52, TIME_PM) ,  
                            TIME(5, 21, TIME_PM) ,  
                            TIME(5, 35, TIME_PM) ,  
                            TIME(6, 0, TIME_PM) ,  
                            TIME(6, 29, TIME_PM) ,  
                            TIME(7, 14, TIME_PM) ,  
                            TIME(8, 14, TIME_PM) ,  
                            TIME(9, 14, TIME_PM) ,  
                            TIME(10, 44, TIME_PM),  
                        },
                        {   // North M-F Lawrence Expy
                            TIME(4, 50, TIME_AM) ,  
                            TIME(5, 30, TIME_AM) ,  
                            TIME(5, 45, TIME_AM) ,  
                            TIME(6, 0, TIME_AM) ,  
                            TIME(6, 15, TIME_AM) ,  
                            TIME(6, 22, TIME_AM) ,  
                            TIME(6, 40, TIME_AM) ,  
                            TIME(6, 49, TIME_AM) ,  
                            TIME(7, 2, TIME_AM) ,  
                            TIME(7, 10, TIME_AM) ,  
                            TIME(7, 26, TIME_AM) ,  
                            TIME(7, 35, TIME_AM) ,  
                            TIME(7, 42, TIME_AM) ,  
                            TIME(8, 8, TIME_AM) ,  
                            TIME(8, 25, TIME_AM) ,  
                            TIME(9, 10, TIME_AM) ,  
                            TIME(9, 47, TIME_AM) ,  
                            TIME(10, 17, TIME_AM),  
                            TIME(10, 47, TIME_AM),  
                            TIME(11, 17, TIME_AM),  
                            TIME(11, 47, TIME_AM),  
                            TIME(12, 17, TIME_PM),  
                            TIME(12, 47, TIME_PM),  
                            TIME(1, 17, TIME_PM) ,  
                            TIME(1, 47, TIME_PM) ,  
                            TIME(2, 17, TIME_PM) ,  
                            TIME(2, 47, TIME_PM) ,  
                            TIME(3, 18, TIME_PM) ,  
                            TIME(3, 48, TIME_PM) ,  
                            TIME(4, 18, TIME_PM) ,  
                            TIME(4, 48, TIME_PM) ,  
                            TIME(5, 17, TIME_PM) ,  
                            TIME(5, 31, TIME_PM) ,  
                            TIME(5, 56, TIME_PM) ,  
                            TIME(6, 25, TIME_PM) ,  
                            TIME(7, 10, TIME_PM) ,  
                            TIME(8, 10, TIME_PM) ,  
                            TIME(9, 10, TIME_PM) ,  
                            TIME(10, 40, TIME_PM),  
                        },
                        {   // North M-F Santa Clara
                            TIME(4, 45, TIME_AM) ,  
                            TIME(5, 25, TIME_AM) ,  
                            TIME(5, 40, TIME_AM) ,  
                            TIME(5, 55, TIME_AM) ,  
                            TIME(6, 10, TIME_AM) ,  
                            TIME(6, 17, TIME_AM) ,  
                            TIME(6, 35, TIME_AM) ,  
                            TIME(6, 44, TIME_AM) ,  
                            TIME(6, 57, TIME_AM) ,  
                            TIME(7, 5, TIME_AM) ,  
                            TIME(7, 21, TIME_AM) ,  
                            TIME(7, 30, TIME_AM) ,  
                            TIME(7, 37, TIME_AM) ,  
                            TIME(8, 3, TIME_AM) ,  
                            TIME(8, 20, TIME_AM) ,  
                            TIME(9, 5, TIME_AM) ,  
                            TIME(9, 42, TIME_AM) ,  
                            TIME(10, 12, TIME_AM),  
                            TIME(10, 42, TIME_AM),  
                            TIME(11, 12, TIME_AM),  
                            TIME(11, 42, TIME_AM),  
                            TIME(12, 12, TIME_PM),  
                            TIME(12, 42, TIME_PM),  
                            TIME(1, 12, TIME_PM) ,  
                            TIME(1, 42, TIME_PM) ,  
                            TIME(2, 12, TIME_PM) ,  
                            TIME(2, 42, TIME_PM) ,  
                            TIME(3, 13, TIME_PM) ,  
                            TIME(3, 43, TIME_PM) ,  
                            TIME(4, 13, TIME_PM) ,  
                            TIME(4, 43, TIME_PM) ,  
                            TIME(5, 12, TIME_PM) ,  
                            TIME(5, 26, TIME_PM) ,  
                            TIME(5, 51, TIME_PM) ,  
                            TIME(6, 20, TIME_PM) ,  
                            TIME(7, 5, TIME_PM) ,  
                            TIME(8, 5, TIME_PM) ,  
                            TIME(9, 5, TIME_PM) ,  
                            TIME(10, 35, TIME_PM),  
                        },
                        {   // North M-F College Park
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(7, 18, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(8, 0, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(3, 10, TIME_PM),  
                            TIME(3, 40, TIME_PM),  
                            TIME(4, 10, TIME_PM),  
                            TIME(4, 40, TIME_PM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(5, 23, TIME_PM),  
                            TIME(5, 48, TIME_PM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                        },
                        {   // North M-F San Jose
                            TIME(4, 40, TIME_AM) ,  
                            TIME(5, 20, TIME_AM) ,  
                            TIME(5, 35, TIME_AM) ,  
                            TIME(5, 50, TIME_AM) ,  
                            TIME(6, 5, TIME_AM) ,  
                            TIME(6, 12, TIME_AM) ,  
                            TIME(6, 30, TIME_AM) ,  
                            TIME(6, 39, TIME_AM) ,  
                            TIME(6, 52, TIME_AM) ,  
                            TIME(7, 0, TIME_AM) ,  
                            TIME(7, 15, TIME_AM) ,  
                            TIME(7, 25, TIME_AM) ,  
                            TIME(7, 32, TIME_AM) ,  
                            TIME(7, 57, TIME_AM) ,  
                            TIME(8, 15, TIME_AM) ,  
                            TIME(9, 0, TIME_AM) ,  
                            TIME(9, 37, TIME_AM) ,  
                            TIME(10, 7, TIME_AM),  
                            TIME(10, 37, TIME_AM),  
                            TIME(11, 7, TIME_AM),  
                            TIME(11, 37, TIME_AM),  
                            TIME(12, 7, TIME_PM),  
                            TIME(12, 37, TIME_PM),  
                            TIME(1, 7, TIME_PM) ,  
                            TIME(1, 37, TIME_PM) ,  
                            TIME(2, 7, TIME_PM) ,  
                            TIME(2, 37, TIME_PM) ,  
                            TIME(3, 7, TIME_PM) ,  
                            TIME(3, 37, TIME_PM) ,  
                            TIME(4, 7, TIME_PM) ,  
                            TIME(4, 37, TIME_PM) ,  
                            TIME(5, 7, TIME_PM) ,  
                            TIME(5, 20, TIME_PM) ,  
                            TIME(5, 45, TIME_PM) ,  
                            TIME(6, 15, TIME_PM) ,  
                            TIME(7, 0, TIME_PM) ,  
                            TIME(8, 0, TIME_PM) ,  
                            TIME(9, 0, TIME_PM) ,  
                            TIME(10, 30, TIME_PM),  
                        },
                        {   // North M-F Tamien
                            TIME(4, 33, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 28, TIME_AM) ,  
                            TIME(5, 43, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 5, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 32, TIME_AM) ,  
                            TIME(6, 45, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 8, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 25, TIME_AM) ,  
                            TIME(7, 50, TIME_AM) ,  
                            TIME(8, 8, TIME_AM) ,  
                            TIME(8, 53, TIME_AM) ,  
                            TIME(9, 30, TIME_AM) ,  
                            TIME(10, 0, TIME_AM),  
                            TIME(10, 30, TIME_AM),  
                            TIME(11, 0, TIME_AM),  
                            TIME(11, 30, TIME_AM),  
                            TIME(12, 0, TIME_PM),  
                            TIME(12, 30, TIME_PM),  
                            TIME(1, 0, TIME_PM) ,  
                            TIME(1, 30, TIME_PM) ,  
                            TIME(2, 0, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(3, 0, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(4, 0, TIME_PM) ,  
                            TIME(4, 30, TIME_PM) ,  
                            TIME(5, 0, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 38, TIME_PM) ,  
                            TIME(6, 8, TIME_PM) ,  
                            TIME(6, 53, TIME_PM) ,  
                            TIME(7, 53, TIME_PM) ,  
                            TIME(8, 53, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                        },
                        {   // North M-F Capitol
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(5, 57, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(6, 37, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(7, 0, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(7, 42, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                        },
                        {   // North M-F Blossom Hill
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(5, 50, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(6, 30, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(6, 53, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(7, 35, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                        },
                        {   // North M-F Morgan Hill
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(5, 37, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(6, 17, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(6, 40, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(7, 22, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                        },
                        {   // North M-F San Martin
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(5, 30, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(6, 10, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(6, 33, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(7, 15, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                        },
                        {   // North M-F Gilroy
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(5, 20, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(6, 0, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(6, 23, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(7, 5, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                        }
                    }, // Northbound Mon-Fri times (39)
                    { 
                        {   // North Sat times San Francisco
                            TIME(7, 56, TIME_AM),
                            TIME(8, 56, TIME_AM),
                            TIME(9, 56, TIME_AM),
                            TIME(10, 56, TIME_AM),
                            TIME(11, 56, TIME_AM),
                            TIME(12, 56, TIME_PM),
                            TIME(1, 56, TIME_PM),
                            TIME(2, 56, TIME_PM),
                            TIME(3, 56, TIME_PM),
                            TIME(4, 56, TIME_PM),
                            TIME(5, 56, TIME_PM),
                            TIME(6, 56, TIME_PM),
                            TIME(7, 56, TIME_PM),
                            TIME(8, 56, TIME_PM),
                            TIME(9, 56, TIME_PM),
                            TIME(12, 26, TIME_AM),
                        },
                        {   // North Sat 22nd Street
                            TIME(7, 30, TIME_AM) ,
                            TIME(8, 30, TIME_AM) ,
                            TIME(9, 30, TIME_AM) ,
                            TIME(10, 30, TIME_AM),
                            TIME(11, 30, TIME_AM),
                            TIME(12, 30, TIME_PM),
                            TIME(1, 30, TIME_PM) ,
                            TIME(2, 30, TIME_PM) ,
                            TIME(3, 30, TIME_PM) ,
                            TIME(4, 30, TIME_PM) ,
                            TIME(5, 30, TIME_PM) ,
                            TIME(6, 30, TIME_PM) ,
                            TIME(7, 30, TIME_PM) ,
                            TIME(8, 30, TIME_PM) ,
                            TIME(9, 30, TIME_PM) ,
                            TIME(12, 0, TIME_AM),
                        },
                        {   // North Sat Paul Avenue
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        },
                        {   // North Sat Bayshore
                            TIME(7, 24, TIME_AM) ,
                            TIME(8, 24, TIME_AM) ,
                            TIME(9, 24, TIME_AM) ,
                            TIME(10, 24, TIME_AM),
                            TIME(11, 24, TIME_AM),
                            TIME(12, 24, TIME_PM),
                            TIME(1, 24, TIME_PM) ,
                            TIME(2, 24, TIME_PM) ,
                            TIME(3, 24, TIME_PM) ,
                            TIME(4, 24, TIME_PM) ,
                            TIME(5, 24, TIME_PM) ,
                            TIME(6, 24, TIME_PM) ,
                            TIME(7, 24, TIME_PM) ,
                            TIME(8, 24, TIME_PM) ,
                            TIME(9, 24, TIME_PM) ,
                            TIME(11, 54, TIME_PM),
                        },
                        {   // North Sat South SF
                            TIME(7, 19, TIME_AM) ,
                            TIME(8, 19, TIME_AM) ,
                            TIME(9, 19, TIME_AM) ,
                            TIME(10, 19, TIME_AM),
                            TIME(11, 19, TIME_AM),
                            TIME(12, 19, TIME_PM),
                            TIME(1, 19, TIME_PM) ,
                            TIME(2, 19, TIME_PM) ,
                            TIME(3, 19, TIME_PM) ,
                            TIME(4, 19, TIME_PM) ,
                            TIME(5, 19, TIME_PM) ,
                            TIME(6, 19, TIME_PM) ,
                            TIME(7, 19, TIME_PM) ,
                            TIME(8, 19, TIME_PM) ,
                            TIME(9, 19, TIME_PM) ,
                            TIME(11, 49, TIME_PM),
                        },
                        {   // North Sat San Bruno
                            TIME(7, 14, TIME_AM) ,
                            TIME(8, 14, TIME_AM) ,
                            TIME(9, 14, TIME_AM) ,
                            TIME(10, 14, TIME_AM),
                            TIME(11, 14, TIME_AM),
                            TIME(12, 14, TIME_PM),
                            TIME(1, 14, TIME_PM) ,
                            TIME(2, 14, TIME_PM) ,
                            TIME(3, 14, TIME_PM) ,
                            TIME(4, 14, TIME_PM) ,
                            TIME(5, 14, TIME_PM) ,
                            TIME(6, 14, TIME_PM) ,
                            TIME(7, 14, TIME_PM) ,
                            TIME(8, 14, TIME_PM) ,
                            TIME(9, 14, TIME_PM) ,
                            TIME(11, 44, TIME_PM),
                        },
                        {   // North Sat Millbrae
                            TIME(7, 9, TIME_AM) ,
                            TIME(8, 9, TIME_AM) ,
                            TIME(9, 9, TIME_AM) ,
                            TIME(10, 9, TIME_AM),
                            TIME(11, 9, TIME_AM),
                            TIME(12, 9, TIME_PM),
                            TIME(1, 9, TIME_PM) ,
                            TIME(2, 9, TIME_PM) ,
                            TIME(3, 9, TIME_PM) ,
                            TIME(4, 9, TIME_PM) ,
                            TIME(5, 9, TIME_PM) ,
                            TIME(6, 9, TIME_PM) ,
                            TIME(7, 9, TIME_PM) ,
                            TIME(8, 9, TIME_PM) ,
                            TIME(9, 9, TIME_PM) ,
                            TIME(11, 39, TIME_PM),
                        },
                        {   // North Sat Broadway
                            TIME(7, 6, TIME_AM) ,
                            TIME(8, 6, TIME_AM) ,
                            TIME(9, 6, TIME_AM) ,
                            TIME(10, 6, TIME_AM),
                            TIME(11, 6, TIME_AM),
                            TIME(12, 6, TIME_PM),
                            TIME(1, 6, TIME_PM) ,
                            TIME(2, 6, TIME_PM) ,
                            TIME(3, 6, TIME_PM) ,
                            TIME(4, 6, TIME_PM) ,
                            TIME(5, 6, TIME_PM) ,
                            TIME(6, 6, TIME_PM) ,
                            TIME(7, 6, TIME_PM) ,
                            TIME(8, 6, TIME_PM) ,
                            TIME(9, 6, TIME_PM) ,
                            TIME(11, 36, TIME_PM),
                        },
                        {   // North Sat Burlingame
                            TIME(7, 1, TIME_AM) ,
                            TIME(8, 1, TIME_AM) ,
                            TIME(9, 1, TIME_AM) ,
                            TIME(10, 1, TIME_AM),
                            TIME(11, 1, TIME_AM),
                            TIME(12, 1, TIME_PM),
                            TIME(1, 1, TIME_PM) ,
                            TIME(2, 1, TIME_PM) ,
                            TIME(3, 1, TIME_PM) ,
                            TIME(4, 1, TIME_PM) ,
                            TIME(5, 1, TIME_PM) ,
                            TIME(6, 1, TIME_PM) ,
                            TIME(7, 1, TIME_PM) ,
                            TIME(8, 1, TIME_PM) ,
                            TIME(9, 1, TIME_PM) ,
                            TIME(11, 31, TIME_PM),
                        },
                        {   // North Sat San Mateo
                            TIME(6, 58, TIME_AM) ,
                            TIME(7, 58, TIME_AM) ,
                            TIME(8, 58, TIME_AM) ,
                            TIME(9, 58, TIME_AM) ,
                            TIME(10, 58, TIME_AM),
                            TIME(11, 58, TIME_AM),
                            TIME(12, 58, TIME_PM),
                            TIME(1, 58, TIME_PM) ,
                            TIME(2, 58, TIME_PM) ,
                            TIME(3, 58, TIME_PM) ,
                            TIME(4, 58, TIME_PM) ,
                            TIME(5, 58, TIME_PM) ,
                            TIME(6, 58, TIME_PM) ,
                            TIME(7, 58, TIME_PM) ,
                            TIME(8, 58, TIME_PM) ,
                            TIME(11, 28, TIME_PM),
                        },
                        {   // North Sat Hayward Park
                            TIME(6, 56, TIME_AM) ,
                            TIME(7, 56, TIME_AM) ,
                            TIME(8, 56, TIME_AM) ,
                            TIME(9, 56, TIME_AM) ,
                            TIME(10, 56, TIME_AM),
                            TIME(11, 56, TIME_AM),
                            TIME(12, 56, TIME_PM),
                            TIME(1, 56, TIME_PM) ,
                            TIME(2, 56, TIME_PM) ,
                            TIME(3, 56, TIME_PM) ,
                            TIME(4, 56, TIME_PM) ,
                            TIME(5, 56, TIME_PM) ,
                            TIME(6, 56, TIME_PM) ,
                            TIME(7, 56, TIME_PM) ,
                            TIME(8, 56, TIME_PM) ,
                            TIME(11, 26, TIME_PM),
                        },
                        {   // North Sat Hillsdale
                            TIME(6, 53, TIME_AM) ,
                            TIME(7, 53, TIME_AM) ,
                            TIME(8, 53, TIME_AM) ,
                            TIME(9, 53, TIME_AM) ,
                            TIME(10, 53, TIME_AM),
                            TIME(11, 53, TIME_AM),
                            TIME(12, 53, TIME_PM),
                            TIME(1, 53, TIME_PM) ,
                            TIME(2, 53, TIME_PM) ,
                            TIME(3, 53, TIME_PM) ,
                            TIME(4, 53, TIME_PM) ,
                            TIME(5, 53, TIME_PM) ,
                            TIME(6, 53, TIME_PM) ,
                            TIME(7, 53, TIME_PM) ,
                            TIME(8, 53, TIME_PM) ,
                            TIME(11, 23, TIME_PM),
                        },
                        {   // North Sat Belmont
                            TIME(6, 50, TIME_AM) ,
                            TIME(7, 50, TIME_AM) ,
                            TIME(8, 50, TIME_AM) ,
                            TIME(9, 50, TIME_AM) ,
                            TIME(10, 50, TIME_AM),
                            TIME(11, 50, TIME_AM),
                            TIME(12, 50, TIME_PM),
                            TIME(1, 50, TIME_PM) ,
                            TIME(2, 50, TIME_PM) ,
                            TIME(3, 50, TIME_PM) ,
                            TIME(4, 50, TIME_PM) ,
                            TIME(5, 50, TIME_PM) ,
                            TIME(6, 50, TIME_PM) ,
                            TIME(7, 50, TIME_PM) ,
                            TIME(8, 50, TIME_PM) ,
                            TIME(11, 20, TIME_PM),
                        },
                        {   // North Sat San Carlos
                            TIME(6, 47, TIME_AM) ,
                            TIME(7, 47, TIME_AM) ,
                            TIME(8, 47, TIME_AM) ,
                            TIME(9, 47, TIME_AM) ,
                            TIME(10, 47, TIME_AM),
                            TIME(11, 47, TIME_AM),
                            TIME(12, 47, TIME_PM),
                            TIME(1, 47, TIME_PM) ,
                            TIME(2, 47, TIME_PM) ,
                            TIME(3, 47, TIME_PM) ,
                            TIME(4, 47, TIME_PM) ,
                            TIME(5, 47, TIME_PM) ,
                            TIME(6, 47, TIME_PM) ,
                            TIME(7, 47, TIME_PM) ,
                            TIME(8, 47, TIME_PM) ,
                            TIME(11, 17, TIME_PM),
                        },
                        {   // North Sat Redwood City
                            TIME(6, 41, TIME_AM) ,
                            TIME(7, 41, TIME_AM) ,
                            TIME(8, 41, TIME_AM) ,
                            TIME(9, 41, TIME_AM) ,
                            TIME(10, 41, TIME_AM),
                            TIME(11, 41, TIME_AM),
                            TIME(12, 41, TIME_PM),
                            TIME(1, 41, TIME_PM) ,
                            TIME(2, 41, TIME_PM) ,
                            TIME(3, 41, TIME_PM) ,
                            TIME(4, 41, TIME_PM) ,
                            TIME(5, 41, TIME_PM) ,
                            TIME(6, 41, TIME_PM) ,
                            TIME(7, 41, TIME_PM) ,
                            TIME(8, 41, TIME_PM) ,
                            TIME(11, 11, TIME_PM),
                        },
                        {   // North Sat Atherton
                            TIME(6, 35, TIME_AM) ,
                            TIME(7, 35, TIME_AM) ,
                            TIME(8, 35, TIME_AM) ,
                            TIME(9, 35, TIME_AM) ,
                            TIME(10, 35, TIME_AM),
                            TIME(11, 35, TIME_AM),
                            TIME(12, 35, TIME_PM),
                            TIME(1, 35, TIME_PM) ,
                            TIME(2, 35, TIME_PM) ,
                            TIME(3, 35, TIME_PM) ,
                            TIME(4, 35, TIME_PM) ,
                            TIME(5, 35, TIME_PM) ,
                            TIME(6, 35, TIME_PM) ,
                            TIME(7, 35, TIME_PM) ,
                            TIME(8, 35, TIME_PM) ,
                            TIME(11, 5, TIME_PM),
                        },
                        {   // North Sat Menlo Park
                            TIME(6, 33, TIME_AM) ,
                            TIME(7, 33, TIME_AM) ,
                            TIME(8, 33, TIME_AM) ,
                            TIME(9, 33, TIME_AM) ,
                            TIME(10, 33, TIME_AM),
                            TIME(11, 33, TIME_AM),
                            TIME(12, 33, TIME_PM),
                            TIME(1, 33, TIME_PM) ,
                            TIME(2, 33, TIME_PM) ,
                            TIME(3, 33, TIME_PM) ,
                            TIME(4, 33, TIME_PM) ,
                            TIME(5, 33, TIME_PM) ,
                            TIME(6, 33, TIME_PM) ,
                            TIME(7, 33, TIME_PM) ,
                            TIME(8, 33, TIME_PM) ,
                            TIME(11, 3, TIME_PM),
                        },
                        {   // North Sat Palo Alto
                            TIME(6, 30, TIME_AM) ,
                            TIME(7, 30, TIME_AM) ,
                            TIME(8, 30, TIME_AM) ,
                            TIME(9, 30, TIME_AM) ,
                            TIME(10, 30, TIME_AM),
                            TIME(11, 30, TIME_AM),
                            TIME(12, 30, TIME_PM),
                            TIME(1, 30, TIME_PM) ,
                            TIME(2, 30, TIME_PM) ,
                            TIME(3, 30, TIME_PM) ,
                            TIME(4, 30, TIME_PM) ,
                            TIME(5, 30, TIME_PM) ,
                            TIME(6, 30, TIME_PM) ,
                            TIME(7, 30, TIME_PM) ,
                            TIME(8, 30, TIME_PM) ,
                            TIME(11, 0, TIME_PM),
                        },
                        {   // North Sat California Av
                            TIME(6, 27, TIME_AM) ,
                            TIME(7, 27, TIME_AM) ,
                            TIME(8, 27, TIME_AM) ,
                            TIME(9, 27, TIME_AM) ,
                            TIME(10, 27, TIME_AM),
                            TIME(11, 27, TIME_AM),
                            TIME(12, 27, TIME_PM),
                            TIME(1, 27, TIME_PM) ,
                            TIME(2, 27, TIME_PM) ,
                            TIME(3, 27, TIME_PM) ,
                            TIME(4, 27, TIME_PM) ,
                            TIME(5, 27, TIME_PM) ,
                            TIME(6, 27, TIME_PM) ,
                            TIME(7, 27, TIME_PM) ,
                            TIME(8, 27, TIME_PM) ,
                            TIME(10, 57, TIME_PM),
                        },
                        {   // North Sat San Antonio
                            TIME(6, 24, TIME_AM) ,
                            TIME(7, 24, TIME_AM) ,
                            TIME(8, 24, TIME_AM) ,
                            TIME(9, 24, TIME_AM) ,
                            TIME(10, 24, TIME_AM),
                            TIME(11, 24, TIME_AM),
                            TIME(12, 24, TIME_PM),
                            TIME(1, 24, TIME_PM) ,
                            TIME(2, 24, TIME_PM) ,
                            TIME(3, 24, TIME_PM) ,
                            TIME(4, 24, TIME_PM) ,
                            TIME(5, 24, TIME_PM) ,
                            TIME(6, 24, TIME_PM) ,
                            TIME(7, 24, TIME_PM) ,
                            TIME(8, 24, TIME_PM) ,
                            TIME(10, 54, TIME_PM),
                        },
                        {   // North Sat Mountain View
                            TIME(6, 20, TIME_AM) ,
                            TIME(7, 20, TIME_AM) ,
                            TIME(8, 20, TIME_AM) ,
                            TIME(9, 20, TIME_AM) ,
                            TIME(10, 20, TIME_AM),
                            TIME(11, 20, TIME_AM),
                            TIME(12, 20, TIME_PM),
                            TIME(1, 20, TIME_PM) ,
                            TIME(2, 20, TIME_PM) ,
                            TIME(3, 20, TIME_PM) ,
                            TIME(4, 20, TIME_PM) ,
                            TIME(5, 20, TIME_PM) ,
                            TIME(6, 20, TIME_PM) ,
                            TIME(7, 20, TIME_PM) ,
                            TIME(8, 20, TIME_PM) ,
                            TIME(10, 50, TIME_PM),
                        },
                        {   // North Sat Sunnyvale
                            TIME(6, 14, TIME_AM) ,
                            TIME(7, 14, TIME_AM) ,
                            TIME(8, 14, TIME_AM) ,
                            TIME(9, 14, TIME_AM) ,
                            TIME(10, 14, TIME_AM),
                            TIME(11, 14, TIME_AM),
                            TIME(12, 14, TIME_PM),
                            TIME(1, 14, TIME_PM) ,
                            TIME(2, 14, TIME_PM) ,
                            TIME(3, 14, TIME_PM) ,
                            TIME(4, 14, TIME_PM) ,
                            TIME(5, 14, TIME_PM) ,
                            TIME(6, 14, TIME_PM) ,
                            TIME(7, 14, TIME_PM) ,
                            TIME(8, 14, TIME_PM) ,
                            TIME(10, 44, TIME_PM),
                        },
                        {   // North Sat Lawrence Expy
                            TIME(6, 10, TIME_AM) ,
                            TIME(7, 10, TIME_AM) ,
                            TIME(8, 10, TIME_AM) ,
                            TIME(9, 10, TIME_AM) ,
                            TIME(10, 10, TIME_AM),
                            TIME(11, 10, TIME_AM),
                            TIME(12, 10, TIME_PM),
                            TIME(1, 10, TIME_PM) ,
                            TIME(2, 10, TIME_PM) ,
                            TIME(3, 10, TIME_PM) ,
                            TIME(4, 10, TIME_PM) ,
                            TIME(5, 10, TIME_PM) ,
                            TIME(6, 10, TIME_PM) ,
                            TIME(7, 10, TIME_PM) ,
                            TIME(8, 10, TIME_PM) ,
                            TIME(10, 40, TIME_PM),
                        },
                        {   // North Sat Santa Clara
                            TIME(6, 5, TIME_AM) ,
                            TIME(7, 5, TIME_AM) ,
                            TIME(8, 5, TIME_AM) ,
                            TIME(9, 5, TIME_AM) ,
                            TIME(10, 5, TIME_AM),
                            TIME(11, 5, TIME_AM),
                            TIME(12, 5, TIME_PM),
                            TIME(1, 5, TIME_PM) ,
                            TIME(2, 5, TIME_PM) ,
                            TIME(3, 5, TIME_PM) ,
                            TIME(4, 5, TIME_PM) ,
                            TIME(5, 5, TIME_PM) ,
                            TIME(6, 5, TIME_PM) ,
                            TIME(7, 5, TIME_PM) ,
                            TIME(8, 5, TIME_PM) ,
                            TIME(10, 35, TIME_PM),
                        },
                        {   // North Sat College Park
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        },
                        {   // North Sat San Jose
                            TIME(6, 0, TIME_AM) ,
                            TIME(7, 0, TIME_AM) ,
                            TIME(8, 0, TIME_AM) ,
                            TIME(9, 0, TIME_AM) ,
                            TIME(10, 0, TIME_AM),
                            TIME(11, 0, TIME_AM),
                            TIME(12, 0, TIME_PM),
                            TIME(1, 0, TIME_PM) ,
                            TIME(2, 0, TIME_PM) ,
                            TIME(3, 0, TIME_PM) ,
                            TIME(4, 0, TIME_PM) ,
                            TIME(5, 0, TIME_PM) ,
                            TIME(6, 0, TIME_PM) ,
                            TIME(7, 0, TIME_PM) ,
                            TIME(8, 0, TIME_PM) ,
                            TIME(10, 30, TIME_PM),
                        },
                        {   // North Sat Tamien
                            TIME(5, 53, TIME_AM) ,
                            TIME(6, 53, TIME_AM) ,
                            TIME(7, 53, TIME_AM) ,
                            TIME(8, 53, TIME_AM) ,
                            TIME(9, 53, TIME_AM) ,
                            TIME(10, 53, TIME_AM),
                            TIME(11, 53, TIME_AM),
                            TIME(12, 53, TIME_PM),
                            TIME(1, 53, TIME_PM) ,
                            TIME(2, 53, TIME_PM) ,
                            TIME(3, 53, TIME_PM) ,
                            TIME(4, 53, TIME_PM) ,
                            TIME(5, 53, TIME_PM) ,
                            TIME(6, 53, TIME_PM) ,
                            TIME(7, 53, TIME_PM) ,
                            TIME(0, 99, TIME_AM)    ,
                        },
                        {   // North Sat Capitol
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        },
                        {   // North Sat Blossom Hill
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        },
                        {   // North Sat Morgan Hill
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                        },
                        {   // North Sat San Martin
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                            TIME(0, 99, TIME_AM)  ,
                        },
                        {   // North Sat Gilroy
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        }
                    }, // Northbound Saturday times (16)
                    { 
                        {   // North Sun times San Francisco
                            TIME(9, 26, TIME_AM) ,
                            TIME(10, 56, TIME_AM),
                            TIME(11, 56, TIME_AM),
                            TIME(12, 56, TIME_PM),
                            TIME(1, 56, TIME_PM) ,
                            TIME(3, 56, TIME_PM) ,
                            TIME(5, 56, TIME_PM) ,
                            TIME(7, 56, TIME_PM) ,
                            TIME(9, 56, TIME_PM) ,
                            TIME(12, 26, TIME_AM),
                        },
                        {   // North Sun 22nd Street
                            TIME(9, 0, TIME_AM) ,
                            TIME(10, 30, TIME_AM),
                            TIME(11, 30, TIME_AM),
                            TIME(12, 30, TIME_PM),
                            TIME(1, 30, TIME_PM) ,
                            TIME(3, 30, TIME_PM) ,
                            TIME(5, 30, TIME_PM) ,
                            TIME(7, 30, TIME_PM) ,
                            TIME(9, 30, TIME_PM) ,
                            TIME(12, 0, TIME_AM),
                        },
                        {   // North Sun Paul Avenue
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        },
                        {   // North Sun Bayshore
                            TIME(8, 54, TIME_AM) ,
                            TIME(10, 24, TIME_AM),
                            TIME(11, 24, TIME_AM),
                            TIME(12, 24, TIME_PM),
                            TIME(1, 24, TIME_PM) ,
                            TIME(3, 24, TIME_PM) ,
                            TIME(5, 24, TIME_PM) ,
                            TIME(7, 24, TIME_PM) ,
                            TIME(9, 24, TIME_PM) ,
                            TIME(11, 54, TIME_PM),
                        },
                        {   // North Sun South SF
                            TIME(8, 49, TIME_AM) ,
                            TIME(10, 19, TIME_AM),
                            TIME(11, 19, TIME_AM),
                            TIME(12, 19, TIME_PM),
                            TIME(1, 19, TIME_PM) ,
                            TIME(3, 19, TIME_PM) ,
                            TIME(5, 19, TIME_PM) ,
                            TIME(7, 19, TIME_PM) ,
                            TIME(9, 19, TIME_PM) ,
                            TIME(11, 49, TIME_PM),
                        },
                        {   // North Sun San Bruno
                            TIME(8, 44, TIME_AM) ,
                            TIME(10, 14, TIME_AM),
                            TIME(11, 14, TIME_AM),
                            TIME(12, 14, TIME_PM),
                            TIME(1, 14, TIME_PM) ,
                            TIME(3, 14, TIME_PM) ,
                            TIME(5, 14, TIME_PM) ,
                            TIME(7, 14, TIME_PM) ,
                            TIME(9, 14, TIME_PM) ,
                            TIME(11, 44, TIME_PM),
                        },
                        {   // North Sun Millbrae
                            TIME(8, 39, TIME_AM) ,
                            TIME(10, 9, TIME_AM),
                            TIME(11, 9, TIME_AM),
                            TIME(12, 9, TIME_PM),
                            TIME(1, 9, TIME_PM) ,
                            TIME(3, 9, TIME_PM) ,
                            TIME(5, 9, TIME_PM) ,
                            TIME(7, 9, TIME_PM) ,
                            TIME(9, 9, TIME_PM) ,
                            TIME(11, 39, TIME_PM),
                        },
                        {   // North Sun Broadway
                            TIME(8, 36, TIME_AM) ,
                            TIME(10, 6, TIME_AM),
                            TIME(11, 6, TIME_AM),
                            TIME(12, 6, TIME_PM),
                            TIME(1, 6, TIME_PM) ,
                            TIME(3, 6, TIME_PM) ,
                            TIME(5, 6, TIME_PM) ,
                            TIME(7, 6, TIME_PM) ,
                            TIME(9, 6, TIME_PM) ,
                            TIME(11, 36, TIME_PM),
                        },
                        {   // North Sun Burlingame
                            TIME(8, 31, TIME_AM) ,
                            TIME(10, 1, TIME_AM),
                            TIME(11, 1, TIME_AM),
                            TIME(12, 1, TIME_PM),
                            TIME(1, 1, TIME_PM) ,
                            TIME(3, 1, TIME_PM) ,
                            TIME(5, 1, TIME_PM) ,
                            TIME(7, 1, TIME_PM) ,
                            TIME(9, 1, TIME_PM) ,
                            TIME(11, 31, TIME_PM),
                        },
                        {   // North Sun San Mateo
                            TIME(8, 28, TIME_AM) ,
                            TIME(9, 58, TIME_AM) ,
                            TIME(10, 58, TIME_AM),
                            TIME(11, 58, TIME_AM),
                            TIME(12, 58, TIME_PM),
                            TIME(2, 58, TIME_PM) ,
                            TIME(4, 58, TIME_PM) ,
                            TIME(6, 58, TIME_PM) ,
                            TIME(8, 58, TIME_PM) ,
                            TIME(11, 28, TIME_PM),
                        },
                        {   // North Sun Hayward Park
                            TIME(8, 26, TIME_AM) ,
                            TIME(9, 56, TIME_AM) ,
                            TIME(10, 56, TIME_AM),
                            TIME(11, 56, TIME_AM),
                            TIME(12, 56, TIME_PM),
                            TIME(2, 56, TIME_PM) ,
                            TIME(4, 56, TIME_PM) ,
                            TIME(6, 56, TIME_PM) ,
                            TIME(8, 56, TIME_PM) ,
                            TIME(11, 26, TIME_PM),
                        },
                        {   // North Sun Hillsdale
                            TIME(8, 23, TIME_AM) ,
                            TIME(9, 53, TIME_AM) ,
                            TIME(10, 53, TIME_AM),
                            TIME(11, 53, TIME_AM),
                            TIME(12, 53, TIME_PM),
                            TIME(2, 53, TIME_PM) ,
                            TIME(4, 53, TIME_PM) ,
                            TIME(6, 53, TIME_PM) ,
                            TIME(8, 53, TIME_PM) ,
                            TIME(11, 23, TIME_PM),
                        },
                        {   // North Sun Belmont
                            TIME(8, 20, TIME_AM) ,
                            TIME(9, 50, TIME_AM) ,
                            TIME(10, 50, TIME_AM),
                            TIME(11, 50, TIME_AM),
                            TIME(12, 50, TIME_PM),
                            TIME(2, 50, TIME_PM) ,
                            TIME(4, 50, TIME_PM) ,
                            TIME(6, 50, TIME_PM) ,
                            TIME(8, 50, TIME_PM) ,
                            TIME(11, 20, TIME_PM),
                        },
                        {   // North Sun San Carlos
                            TIME(8, 17, TIME_AM) ,
                            TIME(9, 47, TIME_AM) ,
                            TIME(10, 47, TIME_AM),
                            TIME(11, 47, TIME_AM),
                            TIME(12, 47, TIME_PM),
                            TIME(2, 47, TIME_PM) ,
                            TIME(4, 47, TIME_PM) ,
                            TIME(6, 47, TIME_PM) ,
                            TIME(8, 47, TIME_PM) ,
                            TIME(11, 17, TIME_PM),
                        },
                        {   // North Sun Redwood City
                            TIME(8, 11, TIME_AM) ,
                            TIME(9, 41, TIME_AM) ,
                            TIME(10, 41, TIME_AM),
                            TIME(11, 41, TIME_AM),
                            TIME(12, 41, TIME_PM),
                            TIME(2, 41, TIME_PM) ,
                            TIME(4, 41, TIME_PM) ,
                            TIME(6, 41, TIME_PM) ,
                            TIME(8, 41, TIME_PM) ,
                            TIME(11, 11, TIME_PM),
                        },
                        {   // North Sun Atherton
                            TIME(8, 5, TIME_AM) ,
                            TIME(9, 35, TIME_AM) ,
                            TIME(10, 35, TIME_AM),
                            TIME(11, 35, TIME_AM),
                            TIME(12, 35, TIME_PM),
                            TIME(2, 35, TIME_PM) ,
                            TIME(4, 35, TIME_PM) ,
                            TIME(6, 35, TIME_PM) ,
                            TIME(8, 35, TIME_PM) ,
                            TIME(11, 5, TIME_PM),
                        },
                        {   // North Sun Menlo Park
                            TIME(8, 3, TIME_AM) ,
                            TIME(9, 33, TIME_AM) ,
                            TIME(10, 33, TIME_AM),
                            TIME(11, 33, TIME_AM),
                            TIME(12, 33, TIME_PM),
                            TIME(2, 33, TIME_PM) ,
                            TIME(4, 33, TIME_PM) ,
                            TIME(6, 33, TIME_PM) ,
                            TIME(8, 33, TIME_PM) ,
                            TIME(11, 3, TIME_PM),
                        },
                        {   // North Sun Palo Alto
                            TIME(8, 0, TIME_AM) ,
                            TIME(9, 30, TIME_AM) ,
                            TIME(10, 30, TIME_AM),
                            TIME(11, 30, TIME_AM),
                            TIME(12, 30, TIME_PM),
                            TIME(2, 30, TIME_PM) ,
                            TIME(4, 30, TIME_PM) ,
                            TIME(6, 30, TIME_PM) ,
                            TIME(8, 30, TIME_PM) ,
                            TIME(11, 0, TIME_PM),
                        },
                        {   // North Sun California Av
                            TIME(7, 57, TIME_AM) ,
                            TIME(9, 27, TIME_AM) ,
                            TIME(10, 27, TIME_AM),
                            TIME(11, 27, TIME_AM),
                            TIME(12, 27, TIME_PM),
                            TIME(2, 27, TIME_PM) ,
                            TIME(4, 27, TIME_PM) ,
                            TIME(6, 27, TIME_PM) ,
                            TIME(8, 27, TIME_PM) ,
                            TIME(10, 57, TIME_PM),
                        },
                        {   // North Sun San Antonio
                            TIME(7, 54, TIME_AM) ,
                            TIME(9, 24, TIME_AM) ,
                            TIME(10, 24, TIME_AM),
                            TIME(11, 24, TIME_AM),
                            TIME(12, 24, TIME_PM),
                            TIME(2, 24, TIME_PM) ,
                            TIME(4, 24, TIME_PM) ,
                            TIME(6, 24, TIME_PM) ,
                            TIME(8, 24, TIME_PM) ,
                            TIME(10, 54, TIME_PM),
                        },
                        {   // North Sun Mountain View
                            TIME(7, 50, TIME_AM) ,
                            TIME(9, 20, TIME_AM) ,
                            TIME(10, 20, TIME_AM),
                            TIME(11, 20, TIME_AM),
                            TIME(12, 20, TIME_PM),
                            TIME(2, 20, TIME_PM) ,
                            TIME(4, 20, TIME_PM) ,
                            TIME(6, 20, TIME_PM) ,
                            TIME(8, 20, TIME_PM) ,
                            TIME(10, 50, TIME_PM),
                        },
                        {   // North Sun Sunnyvale
                            TIME(7, 44, TIME_AM) ,
                            TIME(9, 14, TIME_AM) ,
                            TIME(10, 14, TIME_AM),
                            TIME(11, 14, TIME_AM),
                            TIME(12, 14, TIME_PM),
                            TIME(2, 14, TIME_PM) ,
                            TIME(4, 14, TIME_PM) ,
                            TIME(6, 14, TIME_PM) ,
                            TIME(8, 14, TIME_PM) ,
                            TIME(10, 44, TIME_PM),
                        },
                        {   // North Sun Lawrence Expy
                            TIME(7, 40, TIME_AM) ,
                            TIME(9, 10, TIME_AM) ,
                            TIME(10, 10, TIME_AM),
                            TIME(11, 10, TIME_AM),
                            TIME(12, 10, TIME_PM),
                            TIME(2, 10, TIME_PM) ,
                            TIME(4, 10, TIME_PM) ,
                            TIME(6, 10, TIME_PM) ,
                            TIME(8, 10, TIME_PM) ,
                            TIME(10, 40, TIME_PM),
                        },
                        {   // North Sun Santa Clara
                            TIME(7, 35, TIME_AM) ,
                            TIME(9, 5, TIME_AM) ,
                            TIME(10, 5, TIME_AM),
                            TIME(11, 5, TIME_AM),
                            TIME(12, 5, TIME_PM),
                            TIME(2, 5, TIME_PM) ,
                            TIME(4, 5, TIME_PM) ,
                            TIME(6, 5, TIME_PM) ,
                            TIME(8, 5, TIME_PM) ,
                            TIME(10, 35, TIME_PM),
                        },
                        {   // North Sun College Park
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        },
                        {   // North Sun San Jose
                            TIME(7, 30, TIME_AM) ,
                            TIME(9, 0, TIME_AM) ,
                            TIME(10, 0, TIME_AM),
                            TIME(11, 0, TIME_AM),
                            TIME(12, 0, TIME_PM),
                            TIME(2, 0, TIME_PM) ,
                            TIME(4, 0, TIME_PM) ,
                            TIME(6, 0, TIME_PM) ,
                            TIME(8, 0, TIME_PM) ,
                            TIME(10, 30, TIME_PM),
                        },
                        {   // North Sun Tamien
                            TIME(7, 23, TIME_AM) ,
                            TIME(8, 53, TIME_AM) ,
                            TIME(9, 53, TIME_AM) ,
                            TIME(10, 53, TIME_AM),
                            TIME(11, 53, TIME_AM),
                            TIME(1, 53, TIME_PM) ,
                            TIME(3, 53, TIME_PM) ,
                            TIME(5, 53, TIME_PM) ,
                            TIME(7, 53, TIME_PM) ,
                            TIME(0, 99, TIME_AM)    ,
                        },
                        {   // North Sun Capitol
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        },
                        {   // North Sun Blossom Hill
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        },
                        {   // North Sun Morgan Hill
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        },
                        {   // North Sun San Martin
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        },
                        {   // North Sun Gilroy
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        }
                    } // Northbound Sunday times (10)
                },
                {
                    { 
                        {   // Southbound Mon-Fri times SF
                            TIME(5, 0, TIME_AM) ,  
                            TIME(6, 0, TIME_AM) ,  
                            TIME(6, 30, TIME_AM) ,  
                            TIME(6, 55, TIME_AM) ,  
                            TIME(7, 0, TIME_AM) ,  
                            TIME(7, 25, TIME_AM) ,  
                            TIME(7, 30, TIME_AM) ,  
                            TIME(8, 0, TIME_AM) ,  
                            TIME(8, 30, TIME_AM) ,  
                            TIME(9, 0, TIME_AM) ,  
                            TIME(9, 30, TIME_AM) ,  
                            TIME(10, 0, TIME_AM),  
                            TIME(10, 30, TIME_AM),  
                            TIME(11, 0, TIME_AM),  
                            TIME(11, 30, TIME_AM),  
                            TIME(12, 0, TIME_PM),  
                            TIME(12, 30, TIME_PM),  
                            TIME(1, 0, TIME_PM) ,  
                            TIME(1, 30, TIME_PM) ,  
                            TIME(2, 0, TIME_PM) ,  
                            TIME(2, 30, TIME_PM) ,  
                            TIME(3, 0, TIME_PM) ,  
                            TIME(3, 30, TIME_PM) ,  
                            TIME(3, 55, TIME_PM) ,  
                            TIME(4, 22, TIME_PM) ,  
                            TIME(4, 26, TIME_PM) ,  
                            TIME(4, 51, TIME_PM) ,  
                            TIME(4, 55, TIME_PM) ,  
                            TIME(5, 20, TIME_PM) ,  
                            TIME(5, 26, TIME_PM) ,  
                            TIME(5, 31, TIME_PM) ,  
                            TIME(5, 36, TIME_PM) ,  
                            TIME(6, 0, TIME_PM) ,  
                            TIME(6, 20, TIME_PM) ,  
                            TIME(7, 0, TIME_PM) ,  
                            TIME(8, 0, TIME_PM) ,  
                            TIME(9, 0, TIME_PM) ,  
                            TIME(10, 0, TIME_PM),  
                            TIME(11, 59, TIME_PM),  
                        },
                        {   // South M-F 22nd Street
                            TIME(5, 5, TIME_AM) ,  
                            TIME(6, 5, TIME_AM) ,  
                            TIME(6, 35, TIME_AM) ,  
                            TIME(7, 0, TIME_AM) ,  
                            TIME(7, 5, TIME_AM) ,  
                            TIME(7, 30, TIME_AM) ,  
                            TIME(7, 35, TIME_AM) ,  
                            TIME(8, 5, TIME_AM) ,  
                            TIME(8, 35, TIME_AM) ,  
                            TIME(9, 5, TIME_AM) ,  
                            TIME(9, 35, TIME_AM) ,  
                            TIME(10, 5, TIME_AM),  
                            TIME(10, 35, TIME_AM),  
                            TIME(11, 5, TIME_AM),  
                            TIME(11, 35, TIME_AM),  
                            TIME(12, 5, TIME_PM),  
                            TIME(12, 35, TIME_PM),  
                            TIME(1, 5, TIME_PM) ,  
                            TIME(1, 35, TIME_PM) ,  
                            TIME(2, 5, TIME_PM) ,  
                            TIME(2, 35, TIME_PM) ,  
                            TIME(3, 5, TIME_PM) ,  
                            TIME(3, 35, TIME_PM) ,  
                            TIME(4, 0, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(4, 31, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 0, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 41, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 25, TIME_PM) ,  
                            TIME(7, 5, TIME_PM) ,  
                            TIME(8, 5, TIME_PM) ,  
                            TIME(9, 5, TIME_PM) ,  
                            TIME(10, 5, TIME_PM),  
                            TIME(12, 4, TIME_AM),  
                        },
                        {   // South M-F Paul Avenue
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(6, 8, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(8, 8, TIME_AM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(3, 38, TIME_PM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(6, 28, TIME_PM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                        },
                        {   // South M-F Bayshore
                            TIME(5, 10, TIME_AM) ,  
                            TIME(6, 11, TIME_AM) ,  
                            TIME(6, 40, TIME_AM) ,  
                            TIME(7, 5, TIME_AM) ,  
                            TIME(7, 10, TIME_AM) ,  
                            TIME(7, 35, TIME_AM) ,  
                            TIME(7, 40, TIME_AM) ,  
                            TIME(8, 11, TIME_AM) ,  
                            TIME(8, 40, TIME_AM) ,  
                            TIME(9, 10, TIME_AM) ,  
                            TIME(9, 40, TIME_AM) ,  
                            TIME(10, 10, TIME_AM),  
                            TIME(10, 40, TIME_AM),  
                            TIME(11, 10, TIME_AM),  
                            TIME(11, 40, TIME_AM),  
                            TIME(12, 10, TIME_PM),  
                            TIME(12, 40, TIME_PM),  
                            TIME(1, 10, TIME_PM) ,  
                            TIME(1, 40, TIME_PM) ,  
                            TIME(2, 10, TIME_PM) ,  
                            TIME(2, 40, TIME_PM) ,  
                            TIME(3, 10, TIME_PM) ,  
                            TIME(3, 41, TIME_PM) ,  
                            TIME(4, 5, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(4, 36, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 5, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 46, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 31, TIME_PM) ,  
                            TIME(7, 10, TIME_PM) ,  
                            TIME(8, 10, TIME_PM) ,  
                            TIME(9, 10, TIME_PM) ,  
                            TIME(10, 10, TIME_PM),  
                            TIME(12, 9, TIME_AM),  
                        },
                        {   // South M-F South SF
                            TIME(5, 16, TIME_AM) ,  
                            TIME(6, 17, TIME_AM) ,  
                            TIME(6, 46, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 16, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 46, TIME_AM) ,  
                            TIME(8, 17, TIME_AM) ,  
                            TIME(8, 46, TIME_AM) ,  
                            TIME(9, 16, TIME_AM) ,  
                            TIME(9, 46, TIME_AM) ,  
                            TIME(10, 16, TIME_AM),  
                            TIME(10, 46, TIME_AM),  
                            TIME(11, 16, TIME_AM),  
                            TIME(11, 46, TIME_AM),  
                            TIME(12, 16, TIME_PM),  
                            TIME(12, 46, TIME_PM),  
                            TIME(1, 16, TIME_PM) ,  
                            TIME(1, 46, TIME_PM) ,  
                            TIME(2, 16, TIME_PM) ,  
                            TIME(2, 46, TIME_PM) ,  
                            TIME(3, 16, TIME_PM) ,  
                            TIME(3, 47, TIME_PM) ,  
                            TIME(4, 11, TIME_PM) ,  
                            TIME(4, 36, TIME_PM) ,  
                            TIME(4, 42, TIME_PM) ,  
                            TIME(5, 5, TIME_PM) ,  
                            TIME(5, 11, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 52, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 37, TIME_PM) ,  
                            TIME(7, 16, TIME_PM) ,  
                            TIME(8, 16, TIME_PM) ,  
                            TIME(9, 16, TIME_PM) ,  
                            TIME(10, 16, TIME_PM),  
                            TIME(12, 15, TIME_AM),  
                        },
                        {   // South M-F San Bruno
                            TIME(5, 19, TIME_AM) ,  
                            TIME(6, 20, TIME_AM) ,  
                            TIME(6, 49, TIME_AM) ,  
                            TIME(7, 11, TIME_AM) ,  
                            TIME(7, 19, TIME_AM) ,  
                            TIME(7, 41, TIME_AM) ,  
                            TIME(7, 49, TIME_AM) ,  
                            TIME(8, 20, TIME_AM) ,  
                            TIME(8, 49, TIME_AM) ,  
                            TIME(9, 19, TIME_AM) ,  
                            TIME(9, 49, TIME_AM) ,  
                            TIME(10, 19, TIME_AM),  
                            TIME(10, 49, TIME_AM),  
                            TIME(11, 19, TIME_AM),  
                            TIME(11, 49, TIME_AM),  
                            TIME(12, 19, TIME_PM),  
                            TIME(12, 49, TIME_PM),  
                            TIME(1, 19, TIME_PM) ,  
                            TIME(1, 49, TIME_PM) ,  
                            TIME(2, 19, TIME_PM) ,  
                            TIME(2, 49, TIME_PM) ,  
                            TIME(3, 19, TIME_PM) ,  
                            TIME(3, 50, TIME_PM) ,  
                            TIME(4, 14, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(4, 45, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 14, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 46, TIME_PM) ,  
                            TIME(5, 55, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 40, TIME_PM) ,  
                            TIME(7, 19, TIME_PM) ,  
                            TIME(8, 19, TIME_PM) ,  
                            TIME(9, 19, TIME_PM) ,  
                            TIME(10, 19, TIME_PM),  
                            TIME(12, 18, TIME_AM),  
                        },
                        {   // South M-F Millbrae
                            TIME(5, 25, TIME_AM) ,  
                            TIME(6, 26, TIME_AM) ,  
                            TIME(6, 55, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 25, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 55, TIME_AM) ,  
                            TIME(8, 26, TIME_AM) ,  
                            TIME(8, 55, TIME_AM) ,  
                            TIME(9, 25, TIME_AM) ,  
                            TIME(9, 55, TIME_AM) ,  
                            TIME(10, 25, TIME_AM),  
                            TIME(10, 55, TIME_AM),  
                            TIME(11, 25, TIME_AM),  
                            TIME(11, 55, TIME_AM),  
                            TIME(12, 26, TIME_PM),  
                            TIME(12, 56, TIME_PM),  
                            TIME(1, 25, TIME_PM) ,  
                            TIME(1, 55, TIME_PM) ,  
                            TIME(2, 26, TIME_PM) ,  
                            TIME(2, 55, TIME_PM) ,  
                            TIME(3, 26, TIME_PM) ,  
                            TIME(3, 56, TIME_PM) ,  
                            TIME(4, 21, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(4, 51, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 20, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 1, TIME_PM) ,  
                            TIME(6, 20, TIME_PM) ,  
                            TIME(6, 46, TIME_PM) ,  
                            TIME(7, 25, TIME_PM) ,  
                            TIME(8, 25, TIME_PM) ,  
                            TIME(9, 25, TIME_PM) ,  
                            TIME(10, 25, TIME_PM),  
                            TIME(12, 24, TIME_AM),  
                        },
                        {   // South M-F Broadway
                            TIME(5, 28, TIME_AM) ,  
                            TIME(6, 29, TIME_AM) ,  
                            TIME(6, 58, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 28, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 58, TIME_AM) ,  
                            TIME(8, 29, TIME_AM) ,  
                            TIME(8, 58, TIME_AM) ,  
                            TIME(9, 28, TIME_AM) ,  
                            TIME(9, 58, TIME_AM) ,  
                            TIME(10, 28, TIME_AM),  
                            TIME(10, 58, TIME_AM),  
                            TIME(11, 28, TIME_AM),  
                            TIME(11, 58, TIME_AM),  
                            TIME(12, 29, TIME_PM),  
                            TIME(12, 59, TIME_PM),  
                            TIME(1, 28, TIME_PM) ,  
                            TIME(1, 58, TIME_PM) ,  
                            TIME(2, 29, TIME_PM) ,  
                            TIME(2, 58, TIME_PM) ,  
                            TIME(3, 29, TIME_PM) ,  
                            TIME(3, 59, TIME_PM) ,  
                            TIME(4, 24, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(4, 54, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 23, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 4, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 49, TIME_PM) ,  
                            TIME(7, 28, TIME_PM) ,  
                            TIME(8, 28, TIME_PM) ,  
                            TIME(9, 28, TIME_PM) ,  
                            TIME(10, 28, TIME_PM),  
                            TIME(12, 27, TIME_AM),  
                        },
                        {   // South M-F Burlingame
                            TIME(5, 30, TIME_AM) ,  
                            TIME(6, 31, TIME_AM) ,  
                            TIME(7, 0, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 31, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 1, TIME_AM) ,  
                            TIME(8, 32, TIME_AM) ,  
                            TIME(9, 0, TIME_AM) ,  
                            TIME(9, 30, TIME_AM) ,  
                            TIME(10, 0, TIME_AM),  
                            TIME(10, 30, TIME_AM),  
                            TIME(11, 0, TIME_AM),  
                            TIME(11, 30, TIME_AM),  
                            TIME(12, 0, TIME_PM),  
                            TIME(12, 31, TIME_PM),  
                            TIME(1, 1, TIME_PM) ,  
                            TIME(1, 30, TIME_PM) ,  
                            TIME(2, 0, TIME_PM) ,  
                            TIME(2, 31, TIME_PM) ,  
                            TIME(3, 0, TIME_PM) ,  
                            TIME(3, 31, TIME_PM) ,  
                            TIME(4, 1, TIME_PM) ,  
                            TIME(4, 27, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(4, 57, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 26, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 55, TIME_PM) ,  
                            TIME(6, 7, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 51, TIME_PM) ,  
                            TIME(7, 30, TIME_PM) ,  
                            TIME(8, 30, TIME_PM) ,  
                            TIME(9, 30, TIME_PM) ,  
                            TIME(10, 30, TIME_PM),  
                            TIME(12, 29, TIME_AM),  
                        },
                        {   // South M-F San Mateo
                            TIME(5, 33, TIME_AM) ,  
                            TIME(6, 34, TIME_AM) ,  
                            TIME(7, 3, TIME_AM) ,  
                            TIME(7, 22, TIME_AM) ,  
                            TIME(7, 34, TIME_AM) ,  
                            TIME(7, 52, TIME_AM) ,  
                            TIME(8, 4, TIME_AM) ,  
                            TIME(8, 35, TIME_AM) ,  
                            TIME(9, 3, TIME_AM) ,  
                            TIME(9, 33, TIME_AM) ,  
                            TIME(10, 3, TIME_AM),  
                            TIME(10, 33, TIME_AM),  
                            TIME(11, 3, TIME_AM),  
                            TIME(11, 33, TIME_AM),  
                            TIME(12, 3, TIME_PM),  
                            TIME(12, 34, TIME_PM),  
                            TIME(1, 4, TIME_PM) ,  
                            TIME(1, 33, TIME_PM) ,  
                            TIME(2, 3, TIME_PM) ,  
                            TIME(2, 34, TIME_PM) ,  
                            TIME(3, 3, TIME_PM) ,  
                            TIME(3, 34, TIME_PM) ,  
                            TIME(4, 4, TIME_PM) ,  
                            TIME(4, 30, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 0, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 29, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 58, TIME_PM) ,  
                            TIME(6, 10, TIME_PM) ,  
                            TIME(6, 26, TIME_PM) ,  
                            TIME(6, 54, TIME_PM) ,  
                            TIME(7, 33, TIME_PM) ,  
                            TIME(8, 33, TIME_PM) ,  
                            TIME(9, 33, TIME_PM) ,  
                            TIME(10, 33, TIME_PM),  
                            TIME(12, 32, TIME_AM),  
                        },
                        {   // South M-F Hayward Park
                            TIME(5, 36, TIME_AM) ,  
                            TIME(6, 37, TIME_AM) ,  
                            TIME(7, 6, TIME_AM) ,  
                            TIME(7, 25, TIME_AM) ,  
                            TIME(7, 37, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 7, TIME_AM) ,  
                            TIME(8, 38, TIME_AM) ,  
                            TIME(9, 6, TIME_AM) ,  
                            TIME(9, 36, TIME_AM) ,  
                            TIME(10, 6, TIME_AM),  
                            TIME(10, 36, TIME_AM),  
                            TIME(11, 6, TIME_AM),  
                            TIME(11, 36, TIME_AM),  
                            TIME(12, 6, TIME_PM),  
                            TIME(12, 37, TIME_PM),  
                            TIME(1, 7, TIME_PM) ,  
                            TIME(1, 36, TIME_PM) ,  
                            TIME(2, 6, TIME_PM) ,  
                            TIME(2, 37, TIME_PM) ,  
                            TIME(3, 6, TIME_PM) ,  
                            TIME(3, 37, TIME_PM) ,  
                            TIME(4, 7, TIME_PM) ,  
                            TIME(4, 33, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 3, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 32, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 50, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 13, TIME_PM) ,  
                            TIME(6, 29, TIME_PM) ,  
                            TIME(6, 57, TIME_PM) ,  
                            TIME(7, 36, TIME_PM) ,  
                            TIME(8, 36, TIME_PM) ,  
                            TIME(9, 36, TIME_PM) ,  
                            TIME(10, 36, TIME_PM),  
                            TIME(12, 35, TIME_AM),  
                        },
                        {   // South M-F Hillsdale
                            TIME(5, 39, TIME_AM) ,  
                            TIME(6, 40, TIME_AM) ,  
                            TIME(7, 9, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 40, TIME_AM) ,  
                            TIME(7, 56, TIME_AM) ,  
                            TIME(8, 10, TIME_AM) ,  
                            TIME(8, 41, TIME_AM) ,  
                            TIME(9, 9, TIME_AM) ,  
                            TIME(9, 39, TIME_AM) ,  
                            TIME(10, 9, TIME_AM),  
                            TIME(10, 39, TIME_AM),  
                            TIME(11, 9, TIME_AM),  
                            TIME(11, 39, TIME_AM),  
                            TIME(12, 9, TIME_PM),  
                            TIME(12, 40, TIME_PM),  
                            TIME(1, 10, TIME_PM) ,  
                            TIME(1, 39, TIME_PM) ,  
                            TIME(2, 9, TIME_PM) ,  
                            TIME(2, 40, TIME_PM) ,  
                            TIME(3, 9, TIME_PM) ,  
                            TIME(3, 40, TIME_PM) ,  
                            TIME(4, 10, TIME_PM) ,  
                            TIME(4, 36, TIME_PM) ,  
                            TIME(4, 49, TIME_PM) ,  
                            TIME(5, 6, TIME_PM) ,  
                            TIME(5, 19, TIME_PM) ,  
                            TIME(5, 35, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 53, TIME_PM) ,  
                            TIME(6, 2, TIME_PM) ,  
                            TIME(6, 16, TIME_PM) ,  
                            TIME(6, 32, TIME_PM) ,  
                            TIME(7, 0, TIME_PM) ,  
                            TIME(7, 39, TIME_PM) ,  
                            TIME(8, 39, TIME_PM) ,  
                            TIME(9, 39, TIME_PM) ,  
                            TIME(10, 39, TIME_PM),  
                            TIME(12, 38, TIME_AM),  
                        },
                        {   // South M-F Belmont
                            TIME(5, 42, TIME_AM) ,  
                            TIME(6, 43, TIME_AM) ,  
                            TIME(7, 12, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 43, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 13, TIME_AM) ,  
                            TIME(8, 44, TIME_AM) ,  
                            TIME(9, 12, TIME_AM) ,  
                            TIME(9, 42, TIME_AM) ,  
                            TIME(10, 12, TIME_AM),  
                            TIME(10, 42, TIME_AM),  
                            TIME(11, 12, TIME_AM),  
                            TIME(11, 42, TIME_AM),  
                            TIME(12, 12, TIME_PM),  
                            TIME(12, 43, TIME_PM),  
                            TIME(1, 13, TIME_PM) ,  
                            TIME(1, 42, TIME_PM) ,  
                            TIME(2, 12, TIME_PM) ,  
                            TIME(2, 43, TIME_PM) ,  
                            TIME(3, 12, TIME_PM) ,  
                            TIME(3, 43, TIME_PM) ,  
                            TIME(4, 13, TIME_PM) ,  
                            TIME(4, 39, TIME_PM) ,  
                            TIME(4, 52, TIME_PM) ,  
                            TIME(5, 9, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 38, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 56, TIME_PM) ,  
                            TIME(6, 5, TIME_PM) ,  
                            TIME(6, 19, TIME_PM) ,  
                            TIME(6, 35, TIME_PM) ,  
                            TIME(7, 3, TIME_PM) ,  
                            TIME(7, 42, TIME_PM) ,  
                            TIME(8, 42, TIME_PM) ,  
                            TIME(9, 42, TIME_PM) ,  
                            TIME(10, 42, TIME_PM),  
                            TIME(12, 41, TIME_AM),  
                        },
                        {   // South M-F San Carlos
                            TIME(5, 45, TIME_AM) ,  
                            TIME(6, 46, TIME_AM) ,  
                            TIME(7, 15, TIME_AM) ,  
                            TIME(7, 33, TIME_AM) ,  
                            TIME(7, 46, TIME_AM) ,  
                            TIME(8, 2, TIME_AM) ,  
                            TIME(8, 16, TIME_AM) ,  
                            TIME(8, 47, TIME_AM) ,  
                            TIME(9, 15, TIME_AM) ,  
                            TIME(9, 45, TIME_AM) ,  
                            TIME(10, 15, TIME_AM),  
                            TIME(10, 45, TIME_AM),  
                            TIME(11, 15, TIME_AM),  
                            TIME(11, 45, TIME_AM),  
                            TIME(12, 15, TIME_PM),  
                            TIME(12, 46, TIME_PM),  
                            TIME(1, 16, TIME_PM) ,  
                            TIME(1, 45, TIME_PM) ,  
                            TIME(2, 15, TIME_PM) ,  
                            TIME(2, 46, TIME_PM) ,  
                            TIME(3, 15, TIME_PM) ,  
                            TIME(3, 46, TIME_PM) ,  
                            TIME(4, 16, TIME_PM) ,  
                            TIME(4, 42, TIME_PM) ,  
                            TIME(4, 55, TIME_PM) ,  
                            TIME(5, 12, TIME_PM) ,  
                            TIME(5, 24, TIME_PM) ,  
                            TIME(5, 41, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 8, TIME_PM) ,  
                            TIME(6, 22, TIME_PM) ,  
                            TIME(6, 38, TIME_PM) ,  
                            TIME(7, 6, TIME_PM) ,  
                            TIME(7, 45, TIME_PM) ,  
                            TIME(8, 45, TIME_PM) ,  
                            TIME(9, 45, TIME_PM) ,  
                            TIME(10, 45, TIME_PM),  
                            TIME(12, 44, TIME_AM),  
                        },
                        {   // South M-F Redwood City
                            TIME(5, 50, TIME_AM) ,  
                            TIME(6, 52, TIME_AM) ,  
                            TIME(7, 21, TIME_AM) ,  
                            TIME(7, 38, TIME_AM) ,  
                            TIME(7, 52, TIME_AM) ,  
                            TIME(8, 7, TIME_AM) ,  
                            TIME(8, 22, TIME_AM) ,  
                            TIME(8, 53, TIME_AM) ,  
                            TIME(9, 21, TIME_AM) ,  
                            TIME(9, 50, TIME_AM) ,  
                            TIME(10, 20, TIME_AM),  
                            TIME(10, 50, TIME_AM),  
                            TIME(11, 20, TIME_AM),  
                            TIME(11, 50, TIME_AM),  
                            TIME(12, 20, TIME_PM),  
                            TIME(12, 51, TIME_PM),  
                            TIME(1, 21, TIME_PM) ,  
                            TIME(1, 50, TIME_PM) ,  
                            TIME(2, 20, TIME_PM) ,  
                            TIME(2, 51, TIME_PM) ,  
                            TIME(3, 20, TIME_PM) ,  
                            TIME(3, 52, TIME_PM) ,  
                            TIME(4, 22, TIME_PM) ,  
                            TIME(4, 49, TIME_PM) ,  
                            TIME(5, 1, TIME_PM) ,  
                            TIME(5, 18, TIME_PM) ,  
                            TIME(5, 29, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 3, TIME_PM) ,  
                            TIME(6, 13, TIME_PM) ,  
                            TIME(6, 28, TIME_PM) ,  
                            TIME(6, 43, TIME_PM) ,  
                            TIME(7, 12, TIME_PM) ,  
                            TIME(7, 51, TIME_PM) ,  
                            TIME(8, 50, TIME_PM) ,  
                            TIME(9, 50, TIME_PM) ,  
                            TIME(10, 50, TIME_PM),  
                            TIME(12, 49, TIME_AM),  
                        },
                        {   // South M-F Atherton
                            TIME(5, 54, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 25, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 56, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 26, TIME_AM) ,  
                            TIME(8, 57, TIME_AM) ,  
                            TIME(9, 25, TIME_AM) ,  
                            TIME(9, 54, TIME_AM) ,  
                            TIME(10, 24, TIME_AM),  
                            TIME(10, 54, TIME_AM),  
                            TIME(11, 24, TIME_AM),  
                            TIME(11, 54, TIME_AM),  
                            TIME(12, 24, TIME_PM),  
                            TIME(12, 55, TIME_PM),  
                            TIME(1, 25, TIME_PM) ,  
                            TIME(1, 54, TIME_PM) ,  
                            TIME(2, 24, TIME_PM) ,  
                            TIME(2, 55, TIME_PM) ,  
                            TIME(3, 24, TIME_PM) ,  
                            TIME(3, 56, TIME_PM) ,  
                            TIME(4, 26, TIME_PM) ,  
                            TIME(4, 53, TIME_PM) ,  
                            TIME(5, 5, TIME_PM) ,  
                            TIME(5, 22, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 48, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 7, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 32, TIME_PM) ,  
                            TIME(6, 47, TIME_PM) ,  
                            TIME(7, 16, TIME_PM) ,  
                            TIME(7, 55, TIME_PM) ,  
                            TIME(8, 54, TIME_PM) ,  
                            TIME(9, 54, TIME_PM) ,  
                            TIME(10, 54, TIME_PM),  
                            TIME(12, 53, TIME_AM),  
                        },
                        {   // South M-F Menlo Park
                            TIME(5, 56, TIME_AM) ,  
                            TIME(6, 59, TIME_AM) ,  
                            TIME(7, 28, TIME_AM) ,  
                            TIME(7, 43, TIME_AM) ,  
                            TIME(7, 59, TIME_AM) ,  
                            TIME(8, 12, TIME_AM) ,  
                            TIME(8, 29, TIME_AM) ,  
                            TIME(9, 0, TIME_AM) ,  
                            TIME(9, 28, TIME_AM) ,  
                            TIME(9, 57, TIME_AM) ,  
                            TIME(10, 27, TIME_AM),  
                            TIME(10, 57, TIME_AM),  
                            TIME(11, 27, TIME_AM),  
                            TIME(11, 57, TIME_AM),  
                            TIME(12, 27, TIME_PM),  
                            TIME(12, 58, TIME_PM),  
                            TIME(1, 28, TIME_PM) ,  
                            TIME(1, 57, TIME_PM) ,  
                            TIME(2, 27, TIME_PM) ,  
                            TIME(2, 58, TIME_PM) ,  
                            TIME(3, 27, TIME_PM) ,  
                            TIME(3, 59, TIME_PM) ,  
                            TIME(4, 29, TIME_PM) ,  
                            TIME(4, 56, TIME_PM) ,  
                            TIME(5, 8, TIME_PM) ,  
                            TIME(5, 25, TIME_PM) ,  
                            TIME(5, 34, TIME_PM) ,  
                            TIME(5, 51, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 10, TIME_PM) ,  
                            TIME(6, 19, TIME_PM) ,  
                            TIME(6, 35, TIME_PM) ,  
                            TIME(6, 50, TIME_PM) ,  
                            TIME(7, 19, TIME_PM) ,  
                            TIME(7, 58, TIME_PM) ,  
                            TIME(8, 57, TIME_PM) ,  
                            TIME(9, 57, TIME_PM) ,  
                            TIME(10, 57, TIME_PM),  
                            TIME(12, 56, TIME_AM),  
                        },
                        {   // South M-F Palo Alto
                            TIME(5, 59, TIME_AM) ,  
                            TIME(7, 2, TIME_AM) ,  
                            TIME(7, 31, TIME_AM) ,  
                            TIME(7, 46, TIME_AM) ,  
                            TIME(8, 2, TIME_AM) ,  
                            TIME(8, 15, TIME_AM) ,  
                            TIME(8, 32, TIME_AM) ,  
                            TIME(9, 3, TIME_AM) ,  
                            TIME(9, 31, TIME_AM) ,  
                            TIME(10, 0, TIME_AM),  
                            TIME(10, 30, TIME_AM),  
                            TIME(11, 0, TIME_AM),  
                            TIME(11, 30, TIME_AM),  
                            TIME(12, 0, TIME_PM),  
                            TIME(12, 30, TIME_PM),  
                            TIME(1, 1, TIME_PM) ,  
                            TIME(1, 31, TIME_PM) ,  
                            TIME(2, 0, TIME_PM) ,  
                            TIME(2, 30, TIME_PM) ,  
                            TIME(3, 1, TIME_PM) ,  
                            TIME(3, 30, TIME_PM) ,  
                            TIME(4, 2, TIME_PM) ,  
                            TIME(4, 32, TIME_PM) ,  
                            TIME(4, 59, TIME_PM) ,  
                            TIME(5, 11, TIME_PM) ,  
                            TIME(5, 28, TIME_PM) ,  
                            TIME(5, 37, TIME_PM) ,  
                            TIME(5, 54, TIME_PM) ,  
                            TIME(6, 1, TIME_PM) ,  
                            TIME(6, 13, TIME_PM) ,  
                            TIME(6, 22, TIME_PM) ,  
                            TIME(6, 38, TIME_PM) ,  
                            TIME(6, 53, TIME_PM) ,  
                            TIME(7, 22, TIME_PM) ,  
                            TIME(8, 1, TIME_PM) ,  
                            TIME(9, 0, TIME_PM) ,  
                            TIME(10, 0, TIME_PM),  
                            TIME(11, 0, TIME_PM),  
                            TIME(12, 59, TIME_AM),  
                        },
                        {   // South M-F California Av
                            TIME(6, 2, TIME_AM) ,  
                            TIME(7, 5, TIME_AM) ,  
                            TIME(7, 34, TIME_AM) ,  
                            TIME(7, 49, TIME_AM) ,  
                            TIME(8, 5, TIME_AM) ,  
                            TIME(8, 18, TIME_AM) ,  
                            TIME(8, 35, TIME_AM) ,  
                            TIME(9, 6, TIME_AM) ,  
                            TIME(9, 34, TIME_AM) ,  
                            TIME(10, 3, TIME_AM),  
                            TIME(10, 33, TIME_AM),  
                            TIME(11, 3, TIME_AM),  
                            TIME(11, 33, TIME_AM),  
                            TIME(12, 3, TIME_PM),  
                            TIME(12, 33, TIME_PM),  
                            TIME(1, 4, TIME_PM) ,  
                            TIME(1, 34, TIME_PM) ,  
                            TIME(2, 3, TIME_PM) ,  
                            TIME(2, 33, TIME_PM) ,  
                            TIME(3, 4, TIME_PM) ,  
                            TIME(3, 33, TIME_PM) ,  
                            TIME(4, 5, TIME_PM) ,  
                            TIME(4, 35, TIME_PM) ,  
                            TIME(5, 2, TIME_PM) ,  
                            TIME(5, 14, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(5, 40, TIME_PM) ,  
                            TIME(5, 57, TIME_PM) ,  
                            TIME(6, 4, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 25, TIME_PM) ,  
                            TIME(6, 41, TIME_PM) ,  
                            TIME(6, 56, TIME_PM) ,  
                            TIME(7, 25, TIME_PM) ,  
                            TIME(8, 4, TIME_PM) ,  
                            TIME(9, 3, TIME_PM) ,  
                            TIME(10, 3, TIME_PM),  
                            TIME(11, 3, TIME_PM),  
                            TIME(1, 2, TIME_AM) ,  
                        },
                        {   // South M-F San Antonio
                            TIME(6, 5, TIME_AM) ,  
                            TIME(7, 8, TIME_AM) ,  
                            TIME(7, 37, TIME_AM) ,  
                            TIME(7, 52, TIME_AM) ,  
                            TIME(8, 8, TIME_AM) ,  
                            TIME(8, 21, TIME_AM) ,  
                            TIME(8, 38, TIME_AM) ,  
                            TIME(9, 9, TIME_AM) ,  
                            TIME(9, 37, TIME_AM) ,  
                            TIME(10, 6, TIME_AM),  
                            TIME(10, 36, TIME_AM),  
                            TIME(11, 6, TIME_AM),  
                            TIME(11, 36, TIME_AM),  
                            TIME(12, 6, TIME_PM),  
                            TIME(12, 36, TIME_PM),  
                            TIME(1, 7, TIME_PM) ,  
                            TIME(1, 37, TIME_PM) ,  
                            TIME(2, 6, TIME_PM) ,  
                            TIME(2, 36, TIME_PM) ,  
                            TIME(3, 7, TIME_PM) ,  
                            TIME(3, 36, TIME_PM) ,  
                            TIME(4, 8, TIME_PM) ,  
                            TIME(4, 38, TIME_PM) ,  
                            TIME(5, 5, TIME_PM) ,  
                            TIME(5, 17, TIME_PM) ,  
                            TIME(5, 33, TIME_PM) ,  
                            TIME(5, 43, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 7, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 28, TIME_PM) ,  
                            TIME(6, 44, TIME_PM) ,  
                            TIME(6, 59, TIME_PM) ,  
                            TIME(7, 28, TIME_PM) ,  
                            TIME(8, 7, TIME_PM) ,  
                            TIME(9, 6, TIME_PM) ,  
                            TIME(10, 6, TIME_PM),  
                            TIME(11, 6, TIME_PM),  
                            TIME(1, 5, TIME_AM) ,  
                        },
                        {   // South M-F Mountain View
                            TIME(6, 10, TIME_AM) ,  
                            TIME(7, 13, TIME_AM) ,  
                            TIME(7, 42, TIME_AM) ,  
                            TIME(7, 57, TIME_AM) ,  
                            TIME(8, 13, TIME_AM) ,  
                            TIME(8, 26, TIME_AM) ,  
                            TIME(8, 43, TIME_AM) ,  
                            TIME(9, 14, TIME_AM) ,  
                            TIME(9, 42, TIME_AM) ,  
                            TIME(10, 11, TIME_AM),  
                            TIME(10, 41, TIME_AM),  
                            TIME(11, 11, TIME_AM),  
                            TIME(11, 41, TIME_AM),  
                            TIME(12, 11, TIME_PM),  
                            TIME(12, 41, TIME_PM),  
                            TIME(1, 13, TIME_PM) ,  
                            TIME(1, 43, TIME_PM) ,  
                            TIME(2, 11, TIME_PM) ,  
                            TIME(2, 41, TIME_PM) ,  
                            TIME(3, 13, TIME_PM) ,  
                            TIME(3, 41, TIME_PM) ,  
                            TIME(4, 14, TIME_PM) ,  
                            TIME(4, 43, TIME_PM) ,  
                            TIME(5, 11, TIME_PM) ,  
                            TIME(5, 22, TIME_PM) ,  
                            TIME(5, 38, TIME_PM) ,  
                            TIME(5, 49, TIME_PM) ,  
                            TIME(6, 4, TIME_PM) ,  
                            TIME(6, 13, TIME_PM) ,  
                            TIME(6, 22, TIME_PM) ,  
                            TIME(6, 34, TIME_PM) ,  
                            TIME(6, 50, TIME_PM) ,  
                            TIME(7, 4, TIME_PM) ,  
                            TIME(7, 33, TIME_PM) ,  
                            TIME(8, 12, TIME_PM) ,  
                            TIME(9, 12, TIME_PM) ,  
                            TIME(10, 12, TIME_PM),  
                            TIME(11, 12, TIME_PM),  
                            TIME(1, 10, TIME_AM) ,  
                        },
                        {   // South M-F Sunnyvale
                            TIME(6, 15, TIME_AM) ,  
                            TIME(7, 17, TIME_AM) ,  
                            TIME(7, 46, TIME_AM) ,  
                            TIME(8, 1, TIME_AM) ,  
                            TIME(8, 17, TIME_AM) ,  
                            TIME(8, 30, TIME_AM) ,  
                            TIME(8, 47, TIME_AM) ,  
                            TIME(9, 18, TIME_AM) ,  
                            TIME(9, 46, TIME_AM) ,  
                            TIME(10, 15, TIME_AM),  
                            TIME(10, 45, TIME_AM),  
                            TIME(11, 15, TIME_AM),  
                            TIME(11, 45, TIME_AM),  
                            TIME(12, 15, TIME_PM),  
                            TIME(12, 45, TIME_PM),  
                            TIME(1, 17, TIME_PM) ,  
                            TIME(1, 47, TIME_PM) ,  
                            TIME(2, 15, TIME_PM) ,  
                            TIME(2, 45, TIME_PM) ,  
                            TIME(3, 17, TIME_PM) ,  
                            TIME(3, 45, TIME_PM) ,  
                            TIME(4, 18, TIME_PM) ,  
                            TIME(4, 47, TIME_PM) ,  
                            TIME(5, 16, TIME_PM) ,  
                            TIME(5, 26, TIME_PM) ,  
                            TIME(5, 42, TIME_PM) ,  
                            TIME(5, 53, TIME_PM) ,  
                            TIME(6, 8, TIME_PM) ,  
                            TIME(6, 17, TIME_PM) ,  
                            TIME(6, 26, TIME_PM) ,  
                            TIME(6, 38, TIME_PM) ,  
                            TIME(6, 54, TIME_PM) ,  
                            TIME(7, 8, TIME_PM) ,  
                            TIME(7, 37, TIME_PM) ,  
                            TIME(8, 16, TIME_PM) ,  
                            TIME(9, 16, TIME_PM) ,  
                            TIME(10, 16, TIME_PM),  
                            TIME(11, 16, TIME_PM),  
                            TIME(1, 14, TIME_AM) ,  
                        },
                        {   // South M-F Lawrence Expy
                            TIME(6, 19, TIME_AM) ,  
                            TIME(7, 21, TIME_AM) ,  
                            TIME(7, 50, TIME_AM) ,  
                            TIME(8, 5, TIME_AM) ,  
                            TIME(8, 21, TIME_AM) ,  
                            TIME(8, 34, TIME_AM) ,  
                            TIME(8, 51, TIME_AM) ,  
                            TIME(9, 22, TIME_AM) ,  
                            TIME(9, 50, TIME_AM) ,  
                            TIME(10, 19, TIME_AM),  
                            TIME(10, 49, TIME_AM),  
                            TIME(11, 19, TIME_AM),  
                            TIME(11, 49, TIME_AM),  
                            TIME(12, 19, TIME_PM),  
                            TIME(12, 49, TIME_PM),  
                            TIME(1, 21, TIME_PM) ,  
                            TIME(1, 51, TIME_PM) ,  
                            TIME(2, 19, TIME_PM) ,  
                            TIME(2, 49, TIME_PM) ,  
                            TIME(3, 21, TIME_PM) ,  
                            TIME(3, 49, TIME_PM) ,  
                            TIME(4, 22, TIME_PM) ,  
                            TIME(4, 51, TIME_PM) ,  
                            TIME(5, 20, TIME_PM) ,  
                            TIME(5, 30, TIME_PM) ,  
                            TIME(5, 46, TIME_PM) ,  
                            TIME(5, 57, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 21, TIME_PM) ,  
                            TIME(6, 30, TIME_PM) ,  
                            TIME(6, 42, TIME_PM) ,  
                            TIME(6, 58, TIME_PM) ,  
                            TIME(7, 12, TIME_PM) ,  
                            TIME(7, 41, TIME_PM) ,  
                            TIME(8, 20, TIME_PM) ,  
                            TIME(9, 20, TIME_PM) ,  
                            TIME(10, 20, TIME_PM),  
                            TIME(11, 20, TIME_PM),  
                            TIME(1, 18, TIME_AM) ,  
                        },
                        {   // South M-F Santa Clara
                            TIME(6, 24, TIME_AM) ,  
                            TIME(7, 26, TIME_AM) ,  
                            TIME(7, 55, TIME_AM) ,  
                            TIME(8, 10, TIME_AM) ,  
                            TIME(8, 26, TIME_AM) ,  
                            TIME(8, 39, TIME_AM) ,  
                            TIME(8, 56, TIME_AM) ,  
                            TIME(9, 27, TIME_AM) ,  
                            TIME(9, 55, TIME_AM) ,  
                            TIME(10, 24, TIME_AM),  
                            TIME(10, 54, TIME_AM),  
                            TIME(11, 24, TIME_AM),  
                            TIME(11, 54, TIME_AM),  
                            TIME(12, 24, TIME_PM),  
                            TIME(12, 54, TIME_PM),  
                            TIME(1, 26, TIME_PM) ,  
                            TIME(1, 56, TIME_PM) ,  
                            TIME(2, 24, TIME_PM) ,  
                            TIME(2, 54, TIME_PM) ,  
                            TIME(3, 26, TIME_PM) ,  
                            TIME(3, 54, TIME_PM) ,  
                            TIME(4, 27, TIME_PM) ,  
                            TIME(4, 56, TIME_PM) ,  
                            TIME(5, 25, TIME_PM) ,  
                            TIME(5, 35, TIME_PM) ,  
                            TIME(5, 51, TIME_PM) ,  
                            TIME(6, 2, TIME_PM) ,  
                            TIME(6, 16, TIME_PM) ,  
                            TIME(6, 26, TIME_PM) ,  
                            TIME(6, 35, TIME_PM) ,  
                            TIME(6, 47, TIME_PM) ,  
                            TIME(7, 3, TIME_PM) ,  
                            TIME(7, 17, TIME_PM) ,  
                            TIME(7, 46, TIME_PM) ,  
                            TIME(8, 25, TIME_PM) ,  
                            TIME(9, 25, TIME_PM) ,  
                            TIME(10, 25, TIME_PM),  
                            TIME(11, 25, TIME_PM),  
                            TIME(1, 23, TIME_AM) ,  
                        },
                        {   // South M-F College Park
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 29, TIME_AM) ,  
                            TIME(7, 58, TIME_AM) ,  
                            TIME(8, 13, TIME_AM) ,  
                            TIME(8, 29, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(4, 30, TIME_PM) ,  
                            TIME(4, 59, TIME_PM) ,  
                            TIME(5, 28, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(0, 99, TIME_AM)    ,  
                        },
                        {   // South M-F San Jose
                            TIME(6, 34, TIME_AM) ,  
                            TIME(7, 39, TIME_AM) ,  
                            TIME(8, 8, TIME_AM) ,  
                            TIME(8, 22, TIME_AM) ,  
                            TIME(8, 39, TIME_AM) ,  
                            TIME(8, 50, TIME_AM) ,  
                            TIME(9, 8, TIME_AM) ,  
                            TIME(9, 38, TIME_AM) ,  
                            TIME(10, 6, TIME_AM),  
                            TIME(10, 34, TIME_AM),  
                            TIME(11, 4, TIME_AM),  
                            TIME(11, 35, TIME_AM),  
                            TIME(12, 5, TIME_PM),  
                            TIME(12, 35, TIME_PM),  
                            TIME(1, 5, TIME_PM) ,  
                            TIME(1, 37, TIME_PM) ,  
                            TIME(2, 7, TIME_PM) ,  
                            TIME(2, 34, TIME_PM) ,  
                            TIME(3, 4, TIME_PM) ,  
                            TIME(3, 37, TIME_PM) ,  
                            TIME(4, 4, TIME_PM) ,  
                            TIME(4, 39, TIME_PM) ,  
                            TIME(5, 8, TIME_PM) ,  
                            TIME(5, 37, TIME_PM) ,  
                            TIME(5, 45, TIME_PM) ,  
                            TIME(6, 2, TIME_PM) ,  
                            TIME(6, 13, TIME_PM) ,  
                            TIME(6, 27, TIME_PM) ,  
                            TIME(6, 39, TIME_PM) ,  
                            TIME(6, 47, TIME_PM) ,  
                            TIME(6, 58, TIME_PM) ,  
                            TIME(7, 13, TIME_PM) ,  
                            TIME(7, 28, TIME_PM) ,  
                            TIME(7, 58, TIME_PM) ,  
                            TIME(8, 37, TIME_PM) ,  
                            TIME(9, 36, TIME_PM) ,  
                            TIME(10, 36, TIME_PM),  
                            TIME(11, 36, TIME_PM),  
                            TIME(1, 34, TIME_AM) ,  
                        },
                        {   // South M-F Tamien
                            TIME(6, 40, TIME_AM) ,  
                            TIME(7, 45, TIME_AM) ,  
                            TIME(8, 14, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(8, 45, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(9, 14, TIME_AM) ,  
                            TIME(9, 44, TIME_AM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(10, 40, TIME_AM),  
                            TIME(11, 10, TIME_AM),  
                            TIME(11, 41, TIME_AM),  
                            TIME(12, 11, TIME_PM),  
                            TIME(12, 41, TIME_PM),  
                            TIME(1, 11, TIME_PM) ,  
                            TIME(1, 43, TIME_PM) ,  
                            TIME(2, 13, TIME_PM) ,  
                            TIME(2, 40, TIME_PM) ,  
                            TIME(3, 10, TIME_PM) ,  
                            TIME(3, 43, TIME_PM) ,  
                            TIME(4, 10, TIME_PM) ,  
                            TIME(4, 46, TIME_PM) ,  
                            TIME(5, 14, TIME_PM) ,  
                            TIME(5, 44, TIME_PM) ,  
                            TIME(5, 51, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(6, 18, TIME_PM) ,  
                            TIME(6, 33, TIME_PM) ,  
                            TIME(6, 46, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 4, TIME_PM) ,  
                            TIME(0, 99, TIME_AM)    ,  
                            TIME(7, 34, TIME_PM) ,  
                            TIME(8, 4, TIME_PM) ,  
                            TIME(8, 43, TIME_PM) ,  
                            TIME(9, 42, TIME_PM) ,  
                            TIME(10, 42, TIME_PM),  
                            TIME(11, 42, TIME_PM),  
                            TIME(0, 99, TIME_AM)    ,  
                        },
                        {   // South M-F Capitol
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(4, 52, TIME_PM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(5, 50, TIME_PM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(6, 24, TIME_PM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(6, 52, TIME_PM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                        },
                        {   // South M-F Blossom Hill
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(4, 58, TIME_PM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(5, 56, TIME_PM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(6, 30, TIME_PM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(6, 58, TIME_PM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                        },
                        {   // South M-F Morgan Hill
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(5, 11, TIME_PM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(6, 9, TIME_PM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(6, 43, TIME_PM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(7, 11, TIME_PM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                        },
                        {   // South M-F San Martin
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(5, 17, TIME_PM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(6, 15, TIME_PM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(6, 49, TIME_PM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(7, 17, TIME_PM),  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                            TIME(0, 99, TIME_AM)   ,  
                        },
                        {   // South M-F Gilroy
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(5, 32, TIME_PM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(6, 30, TIME_PM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(7, 4, TIME_PM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(7, 32, TIME_PM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                            TIME(0, 99, TIME_AM),  
                        }
                    }, // Southbound Mon-Fri times (39)
                    { 
                        {   // Southbound Saturday times SF
                            TIME(7, 0, TIME_AM) ,
                            TIME(8, 0, TIME_AM) ,
                            TIME(9, 0, TIME_AM) ,
                            TIME(10, 0, TIME_AM),
                            TIME(11, 0, TIME_AM),
                            TIME(12, 0, TIME_PM),
                            TIME(1, 0, TIME_PM) ,
                            TIME(2, 0, TIME_PM) ,
                            TIME(3, 0, TIME_PM) ,
                            TIME(4, 0, TIME_PM) ,
                            TIME(5, 0, TIME_PM) ,
                            TIME(6, 0, TIME_PM) ,
                            TIME(7, 0, TIME_PM) ,
                            TIME(8, 0, TIME_PM) ,
                            TIME(10, 0, TIME_PM),
                            TIME(11, 59, TIME_PM),
                        },
                        {   // South Sat 22nd Street
                            TIME(7, 5, TIME_AM) ,
                            TIME(8, 5, TIME_AM) ,
                            TIME(9, 5, TIME_AM) ,
                            TIME(10, 5, TIME_AM),
                            TIME(11, 5, TIME_AM),
                            TIME(12, 5, TIME_PM),
                            TIME(1, 5, TIME_PM) ,
                            TIME(2, 5, TIME_PM) ,
                            TIME(3, 5, TIME_PM) ,
                            TIME(4, 5, TIME_PM) ,
                            TIME(5, 5, TIME_PM) ,
                            TIME(6, 5, TIME_PM) ,
                            TIME(7, 5, TIME_PM) ,
                            TIME(8, 5, TIME_PM) ,
                            TIME(10, 5, TIME_PM),
                            TIME(12, 4, TIME_AM),
                        },
                        {   // South Sat Paul Avenue
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        },
                        {   // South Sat Bayshore
                            TIME(7, 12, TIME_AM) ,
                            TIME(8, 12, TIME_AM) ,
                            TIME(9, 12, TIME_AM) ,
                            TIME(10, 12, TIME_AM),
                            TIME(11, 12, TIME_AM),
                            TIME(12, 12, TIME_PM),
                            TIME(1, 12, TIME_PM) ,
                            TIME(2, 12, TIME_PM) ,
                            TIME(3, 12, TIME_PM) ,
                            TIME(4, 12, TIME_PM) ,
                            TIME(5, 12, TIME_PM) ,
                            TIME(6, 12, TIME_PM) ,
                            TIME(7, 12, TIME_PM) ,
                            TIME(8, 12, TIME_PM) ,
                            TIME(10, 12, TIME_PM),
                            TIME(12, 11, TIME_AM),
                        },
                        {   // South Sat South SF
                            TIME(7, 18, TIME_AM) ,
                            TIME(8, 18, TIME_AM) ,
                            TIME(9, 18, TIME_AM) ,
                            TIME(10, 18, TIME_AM),
                            TIME(11, 18, TIME_AM),
                            TIME(12, 18, TIME_PM),
                            TIME(1, 18, TIME_PM) ,
                            TIME(2, 18, TIME_PM) ,
                            TIME(3, 18, TIME_PM) ,
                            TIME(4, 18, TIME_PM) ,
                            TIME(5, 18, TIME_PM) ,
                            TIME(6, 18, TIME_PM) ,
                            TIME(7, 18, TIME_PM) ,
                            TIME(8, 18, TIME_PM) ,
                            TIME(10, 18, TIME_PM),
                            TIME(12, 17, TIME_AM),
                        },
                        {   // South Sat San Bruno
                            TIME(7, 21, TIME_AM) ,
                            TIME(8, 21, TIME_AM) ,
                            TIME(9, 21, TIME_AM) ,
                            TIME(10, 21, TIME_AM),
                            TIME(11, 21, TIME_AM),
                            TIME(12, 21, TIME_PM),
                            TIME(1, 21, TIME_PM) ,
                            TIME(2, 21, TIME_PM) ,
                            TIME(3, 21, TIME_PM) ,
                            TIME(4, 21, TIME_PM) ,
                            TIME(5, 21, TIME_PM) ,
                            TIME(6, 21, TIME_PM) ,
                            TIME(7, 21, TIME_PM) ,
                            TIME(8, 21, TIME_PM) ,
                            TIME(10, 21, TIME_PM),
                            TIME(12, 20, TIME_AM),
                        },
                        {   // South Sat Millbrae
                            TIME(7, 28, TIME_AM) ,
                            TIME(8, 28, TIME_AM) ,
                            TIME(9, 28, TIME_AM) ,
                            TIME(10, 28, TIME_AM),
                            TIME(11, 28, TIME_AM),
                            TIME(12, 28, TIME_PM),
                            TIME(1, 28, TIME_PM) ,
                            TIME(2, 28, TIME_PM) ,
                            TIME(3, 28, TIME_PM) ,
                            TIME(4, 28, TIME_PM) ,
                            TIME(5, 28, TIME_PM) ,
                            TIME(6, 28, TIME_PM) ,
                            TIME(7, 28, TIME_PM) ,
                            TIME(8, 28, TIME_PM) ,
                            TIME(10, 28, TIME_PM),
                            TIME(12, 27, TIME_AM),
                        },
                        {   // South Sat Broadway
                            TIME(7, 31, TIME_AM) ,
                            TIME(8, 31, TIME_AM) ,
                            TIME(9, 31, TIME_AM) ,
                            TIME(10, 31, TIME_AM),
                            TIME(11, 31, TIME_AM),
                            TIME(12, 31, TIME_PM),
                            TIME(1, 31, TIME_PM) ,
                            TIME(2, 31, TIME_PM) ,
                            TIME(3, 31, TIME_PM) ,
                            TIME(4, 31, TIME_PM) ,
                            TIME(5, 31, TIME_PM) ,
                            TIME(6, 31, TIME_PM) ,
                            TIME(7, 31, TIME_PM) ,
                            TIME(8, 31, TIME_PM) ,
                            TIME(10, 31, TIME_PM),
                            TIME(12, 30, TIME_AM),
                        },
                        {   // South Sat Burlingame
                            TIME(7, 33, TIME_AM) ,
                            TIME(8, 33, TIME_AM) ,
                            TIME(9, 33, TIME_AM) ,
                            TIME(10, 33, TIME_AM),
                            TIME(11, 33, TIME_AM),
                            TIME(12, 33, TIME_PM),
                            TIME(1, 33, TIME_PM) ,
                            TIME(2, 33, TIME_PM) ,
                            TIME(3, 33, TIME_PM) ,
                            TIME(4, 33, TIME_PM) ,
                            TIME(5, 33, TIME_PM) ,
                            TIME(6, 33, TIME_PM) ,
                            TIME(7, 33, TIME_PM) ,
                            TIME(8, 33, TIME_PM) ,
                            TIME(10, 33, TIME_PM),
                            TIME(12, 32, TIME_AM),
                        },
                        {   // South Sat San Mateo
                            TIME(7, 36, TIME_AM) ,
                            TIME(8, 36, TIME_AM) ,
                            TIME(9, 36, TIME_AM) ,
                            TIME(10, 36, TIME_AM),
                            TIME(11, 36, TIME_AM),
                            TIME(12, 36, TIME_PM),
                            TIME(1, 36, TIME_PM) ,
                            TIME(2, 36, TIME_PM) ,
                            TIME(3, 36, TIME_PM) ,
                            TIME(4, 36, TIME_PM) ,
                            TIME(5, 36, TIME_PM) ,
                            TIME(6, 36, TIME_PM) ,
                            TIME(7, 36, TIME_PM) ,
                            TIME(8, 36, TIME_PM) ,
                            TIME(10, 36, TIME_PM),
                            TIME(12, 35, TIME_AM),
                        },
                        {   // South Sat Hayward Park
                            TIME(7, 39, TIME_AM) ,
                            TIME(8, 39, TIME_AM) ,
                            TIME(9, 39, TIME_AM) ,
                            TIME(10, 39, TIME_AM),
                            TIME(11, 39, TIME_AM),
                            TIME(12, 39, TIME_PM),
                            TIME(1, 39, TIME_PM) ,
                            TIME(2, 39, TIME_PM) ,
                            TIME(3, 39, TIME_PM) ,
                            TIME(4, 39, TIME_PM) ,
                            TIME(5, 39, TIME_PM) ,
                            TIME(6, 39, TIME_PM) ,
                            TIME(7, 39, TIME_PM) ,
                            TIME(8, 39, TIME_PM) ,
                            TIME(10, 39, TIME_PM),
                            TIME(12, 38, TIME_AM),
                        },
                        {   // South Sat Hillsdale
                            TIME(7, 42, TIME_AM) ,
                            TIME(8, 42, TIME_AM) ,
                            TIME(9, 42, TIME_AM) ,
                            TIME(10, 42, TIME_AM),
                            TIME(11, 42, TIME_AM),
                            TIME(12, 42, TIME_PM),
                            TIME(1, 42, TIME_PM) ,
                            TIME(2, 42, TIME_PM) ,
                            TIME(3, 42, TIME_PM) ,
                            TIME(4, 42, TIME_PM) ,
                            TIME(5, 42, TIME_PM) ,
                            TIME(6, 42, TIME_PM) ,
                            TIME(7, 42, TIME_PM) ,
                            TIME(8, 42, TIME_PM) ,
                            TIME(10, 42, TIME_PM),
                            TIME(12, 41, TIME_AM),
                        },
                        {   // South Sat Belmont
                            TIME(7, 47, TIME_AM) ,
                            TIME(8, 47, TIME_AM) ,
                            TIME(9, 47, TIME_AM) ,
                            TIME(10, 47, TIME_AM),
                            TIME(11, 47, TIME_AM),
                            TIME(12, 47, TIME_PM),
                            TIME(1, 47, TIME_PM) ,
                            TIME(2, 47, TIME_PM) ,
                            TIME(3, 47, TIME_PM) ,
                            TIME(4, 47, TIME_PM) ,
                            TIME(5, 47, TIME_PM) ,
                            TIME(6, 47, TIME_PM) ,
                            TIME(7, 47, TIME_PM) ,
                            TIME(8, 47, TIME_PM) ,
                            TIME(10, 47, TIME_PM),
                            TIME(12, 46, TIME_AM),
                        },
                        {   // South Sat San Carlos
                            TIME(7, 50, TIME_AM) ,
                            TIME(8, 50, TIME_AM) ,
                            TIME(9, 50, TIME_AM) ,
                            TIME(10, 50, TIME_AM),
                            TIME(11, 50, TIME_AM),
                            TIME(12, 50, TIME_PM),
                            TIME(1, 50, TIME_PM) ,
                            TIME(2, 50, TIME_PM) ,
                            TIME(3, 50, TIME_PM) ,
                            TIME(4, 50, TIME_PM) ,
                            TIME(5, 50, TIME_PM) ,
                            TIME(6, 50, TIME_PM) ,
                            TIME(7, 50, TIME_PM) ,
                            TIME(8, 50, TIME_PM) ,
                            TIME(10, 50, TIME_PM),
                            TIME(12, 49, TIME_AM),
                        },
                        {   // South Sat Redwood City
                            TIME(7, 55, TIME_AM) ,
                            TIME(8, 55, TIME_AM) ,
                            TIME(9, 55, TIME_AM) ,
                            TIME(10, 55, TIME_AM),
                            TIME(11, 55, TIME_AM),
                            TIME(12, 55, TIME_PM),
                            TIME(1, 55, TIME_PM) ,
                            TIME(2, 55, TIME_PM) ,
                            TIME(3, 55, TIME_PM) ,
                            TIME(4, 55, TIME_PM) ,
                            TIME(5, 55, TIME_PM) ,
                            TIME(6, 55, TIME_PM) ,
                            TIME(7, 55, TIME_PM) ,
                            TIME(8, 55, TIME_PM) ,
                            TIME(10, 55, TIME_PM),
                            TIME(12, 54, TIME_AM),
                        },
                        {   // South Sat Atherton
                            TIME(7, 59, TIME_AM) ,
                            TIME(8, 59, TIME_AM) ,
                            TIME(9, 59, TIME_AM) ,
                            TIME(10, 59, TIME_AM),
                            TIME(11, 59, TIME_AM),
                            TIME(12, 59, TIME_PM),
                            TIME(1, 59, TIME_PM) ,
                            TIME(2, 59, TIME_PM) ,
                            TIME(3, 59, TIME_PM) ,
                            TIME(4, 59, TIME_PM) ,
                            TIME(5, 59, TIME_PM) ,
                            TIME(6, 59, TIME_PM) ,
                            TIME(7, 59, TIME_PM) ,
                            TIME(8, 59, TIME_PM) ,
                            TIME(10, 59, TIME_PM),
                            TIME(12, 58, TIME_AM),
                        },
                        {   // South Sat Menlo Park
                            TIME(8, 4, TIME_AM) ,
                            TIME(9, 4, TIME_AM) ,
                            TIME(10, 4, TIME_AM),
                            TIME(11, 4, TIME_AM),
                            TIME(12, 4, TIME_PM),
                            TIME(1, 4, TIME_PM) ,
                            TIME(2, 4, TIME_PM) ,
                            TIME(3, 4, TIME_PM) ,
                            TIME(4, 4, TIME_PM) ,
                            TIME(5, 4, TIME_PM) ,
                            TIME(6, 4, TIME_PM) ,
                            TIME(7, 4, TIME_PM) ,
                            TIME(8, 4, TIME_PM) ,
                            TIME(9, 4, TIME_PM) ,
                            TIME(11, 4, TIME_PM),
                            TIME(1, 3, TIME_AM) ,
                        },
                        {   // South Sat Palo Alto
                            TIME(8, 7, TIME_AM) ,
                            TIME(9, 7, TIME_AM) ,
                            TIME(10, 7, TIME_AM),
                            TIME(11, 7, TIME_AM),
                            TIME(12, 7, TIME_PM),
                            TIME(1, 7, TIME_PM) ,
                            TIME(2, 7, TIME_PM) ,
                            TIME(3, 7, TIME_PM) ,
                            TIME(4, 7, TIME_PM) ,
                            TIME(5, 7, TIME_PM) ,
                            TIME(6, 7, TIME_PM) ,
                            TIME(7, 7, TIME_PM) ,
                            TIME(8, 7, TIME_PM) ,
                            TIME(9, 7, TIME_PM) ,
                            TIME(11, 7, TIME_PM),
                            TIME(1, 6, TIME_AM) ,
                        },
                        {   // South Sat California Av
                            TIME(8, 10, TIME_AM) ,
                            TIME(9, 10, TIME_AM) ,
                            TIME(10, 10, TIME_AM),
                            TIME(11, 10, TIME_AM),
                            TIME(12, 10, TIME_PM),
                            TIME(1, 10, TIME_PM) ,
                            TIME(2, 10, TIME_PM) ,
                            TIME(3, 10, TIME_PM) ,
                            TIME(4, 10, TIME_PM) ,
                            TIME(5, 10, TIME_PM) ,
                            TIME(6, 10, TIME_PM) ,
                            TIME(7, 10, TIME_PM) ,
                            TIME(8, 10, TIME_PM) ,
                            TIME(9, 10, TIME_PM) ,
                            TIME(11, 10, TIME_PM),
                            TIME(1, 9, TIME_AM) ,
                        },
                        {   // South Sat San Antonio
                            TIME(8, 13, TIME_AM) ,
                            TIME(9, 13, TIME_AM) ,
                            TIME(10, 13, TIME_AM),
                            TIME(11, 13, TIME_AM),
                            TIME(12, 13, TIME_PM),
                            TIME(1, 13, TIME_PM) ,
                            TIME(2, 13, TIME_PM) ,
                            TIME(3, 13, TIME_PM) ,
                            TIME(4, 13, TIME_PM) ,
                            TIME(5, 13, TIME_PM) ,
                            TIME(6, 13, TIME_PM) ,
                            TIME(7, 13, TIME_PM) ,
                            TIME(8, 13, TIME_PM) ,
                            TIME(9, 13, TIME_PM) ,
                            TIME(11, 13, TIME_PM),
                            TIME(1, 12, TIME_AM) ,
                        },
                        {   // South Sat Mountain View
                            TIME(8, 19, TIME_AM) ,
                            TIME(9, 19, TIME_AM) ,
                            TIME(10, 19, TIME_AM),
                            TIME(11, 19, TIME_AM),
                            TIME(12, 19, TIME_PM),
                            TIME(1, 19, TIME_PM) ,
                            TIME(2, 19, TIME_PM) ,
                            TIME(3, 19, TIME_PM) ,
                            TIME(4, 19, TIME_PM) ,
                            TIME(5, 19, TIME_PM) ,
                            TIME(6, 19, TIME_PM) ,
                            TIME(7, 19, TIME_PM) ,
                            TIME(8, 19, TIME_PM) ,
                            TIME(9, 19, TIME_PM) ,
                            TIME(11, 19, TIME_PM),
                            TIME(1, 18, TIME_AM) ,
                        },
                        {   // South Sat Sunnyvale
                            TIME(8, 23, TIME_AM) ,
                            TIME(9, 23, TIME_AM) ,
                            TIME(10, 23, TIME_AM),
                            TIME(11, 23, TIME_AM),
                            TIME(12, 23, TIME_PM),
                            TIME(1, 23, TIME_PM) ,
                            TIME(2, 23, TIME_PM) ,
                            TIME(3, 23, TIME_PM) ,
                            TIME(4, 23, TIME_PM) ,
                            TIME(5, 23, TIME_PM) ,
                            TIME(6, 23, TIME_PM) ,
                            TIME(7, 23, TIME_PM) ,
                            TIME(8, 23, TIME_PM) ,
                            TIME(9, 23, TIME_PM) ,
                            TIME(11, 23, TIME_PM),
                            TIME(1, 22, TIME_AM) ,
                        },
                        {   // South Sat Lawrence Expy
                            TIME(8, 27, TIME_AM) ,
                            TIME(9, 27, TIME_AM) ,
                            TIME(10, 27, TIME_AM),
                            TIME(11, 27, TIME_AM),
                            TIME(12, 27, TIME_PM),
                            TIME(1, 27, TIME_PM) ,
                            TIME(2, 27, TIME_PM) ,
                            TIME(3, 27, TIME_PM) ,
                            TIME(4, 27, TIME_PM) ,
                            TIME(5, 27, TIME_PM) ,
                            TIME(6, 27, TIME_PM) ,
                            TIME(7, 27, TIME_PM) ,
                            TIME(8, 27, TIME_PM) ,
                            TIME(9, 27, TIME_PM) ,
                            TIME(11, 27, TIME_PM),
                            TIME(1, 26, TIME_AM) ,
                        },
                        {   // South Sat Santa Clara
                            TIME(8, 32, TIME_AM) ,
                            TIME(9, 32, TIME_AM) ,
                            TIME(10, 32, TIME_AM),
                            TIME(11, 32, TIME_AM),
                            TIME(12, 32, TIME_PM),
                            TIME(1, 32, TIME_PM) ,
                            TIME(2, 32, TIME_PM) ,
                            TIME(3, 32, TIME_PM) ,
                            TIME(4, 32, TIME_PM) ,
                            TIME(5, 32, TIME_PM) ,
                            TIME(6, 32, TIME_PM) ,
                            TIME(7, 32, TIME_PM) ,
                            TIME(8, 32, TIME_PM) ,
                            TIME(9, 32, TIME_PM) ,
                            TIME(11, 32, TIME_PM),
                            TIME(1, 31, TIME_AM) ,
                        },
                        {   // South Sat College Park
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        },
                        {   // South Sat San Jose
                            TIME(8, 40, TIME_AM) ,
                            TIME(9, 40, TIME_AM) ,
                            TIME(10, 40, TIME_AM),
                            TIME(11, 40, TIME_AM),
                            TIME(12, 40, TIME_PM),
                            TIME(1, 40, TIME_PM) ,
                            TIME(2, 40, TIME_PM) ,
                            TIME(3, 40, TIME_PM) ,
                            TIME(4, 40, TIME_PM) ,
                            TIME(5, 40, TIME_PM) ,
                            TIME(6, 40, TIME_PM) ,
                            TIME(7, 40, TIME_PM) ,
                            TIME(8, 40, TIME_PM) ,
                            TIME(9, 40, TIME_PM) ,
                            TIME(11, 40, TIME_PM),
                            TIME(1, 54, TIME_AM) ,
                        },
                        {   // South Sat Tamien
                            TIME(9, 1, TIME_AM) ,
                            TIME(10, 1, TIME_AM),
                            TIME(11, 1, TIME_AM),
                            TIME(12, 1, TIME_PM),
                            TIME(1, 1, TIME_PM) ,
                            TIME(2, 1, TIME_PM) ,
                            TIME(3, 1, TIME_PM) ,
                            TIME(4, 1, TIME_PM) ,
                            TIME(5, 1, TIME_PM) ,
                            TIME(6, 1, TIME_PM) ,
                            TIME(7, 1, TIME_PM) ,
                            TIME(8, 1, TIME_PM) ,
                            TIME(9, 1, TIME_PM) ,
                            TIME(10, 1, TIME_PM),
                            TIME(12, 1, TIME_AM),
                            TIME(0, 99, TIME_AM)    ,
                        },
                        {   // South Sat Capitol
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        },
                        {   // South Sat Blossom Hill
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        },
                        {   // South Sat Morgan Hill
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        },
                        {   // South Sat San Martin
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        },
                        {   // South Sat Gilroy
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                        }
                    }, // Southbound Saturday times (16)
                    { 
                        {   // Southbound Sunday times SF
                            TIME(8, 0, TIME_AM) ,
                            TIME(10, 0, TIME_AM),
                            TIME(12, 0, TIME_PM),
                            TIME(2, 0, TIME_PM) ,
                            TIME(4, 0, TIME_PM) ,
                            TIME(5, 0, TIME_PM) ,
                            TIME(6, 0, TIME_PM) ,
                            TIME(7, 0, TIME_PM) ,
                            TIME(8, 0, TIME_PM) ,
                            TIME(10, 0, TIME_PM),
                        },
                        {   // South Sun 22nd Street
                            TIME(8, 5, TIME_AM) ,
                            TIME(10, 5, TIME_AM),
                            TIME(12, 5, TIME_PM),
                            TIME(2, 5, TIME_PM) ,
                            TIME(4, 5, TIME_PM) ,
                            TIME(5, 5, TIME_PM) ,
                            TIME(6, 5, TIME_PM) ,
                            TIME(7, 5, TIME_PM) ,
                            TIME(8, 5, TIME_PM) ,
                            TIME(10, 5, TIME_PM),
                        },
                        {   // South Sun Paul Avenue
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        },
                        {   // South Sun Bayshore
                            TIME(8, 12, TIME_AM) ,
                            TIME(10, 12, TIME_AM),
                            TIME(12, 12, TIME_PM),
                            TIME(2, 12, TIME_PM) ,
                            TIME(4, 12, TIME_PM) ,
                            TIME(5, 12, TIME_PM) ,
                            TIME(6, 12, TIME_PM) ,
                            TIME(7, 12, TIME_PM) ,
                            TIME(8, 12, TIME_PM) ,
                            TIME(10, 12, TIME_PM),
                        },
                        {   // South Sun South SF
                            TIME(8, 18, TIME_AM) ,
                            TIME(10, 18, TIME_AM),
                            TIME(12, 18, TIME_PM),
                            TIME(2, 18, TIME_PM) ,
                            TIME(4, 18, TIME_PM) ,
                            TIME(5, 18, TIME_PM) ,
                            TIME(6, 18, TIME_PM) ,
                            TIME(7, 18, TIME_PM) ,
                            TIME(8, 18, TIME_PM) ,
                            TIME(10, 18, TIME_PM),
                        },
                        {   // South Sun San Bruno
                            TIME(8, 21, TIME_AM) ,
                            TIME(10, 21, TIME_AM),
                            TIME(12, 21, TIME_PM),
                            TIME(2, 21, TIME_PM) ,
                            TIME(4, 21, TIME_PM) ,
                            TIME(5, 21, TIME_PM) ,
                            TIME(6, 21, TIME_PM) ,
                            TIME(7, 21, TIME_PM) ,
                            TIME(8, 21, TIME_PM) ,
                            TIME(10, 21, TIME_PM),
                        },
                        {   // South Sun Millbrae
                            TIME(8, 28, TIME_AM) ,
                            TIME(10, 28, TIME_AM),
                            TIME(12, 28, TIME_PM),
                            TIME(2, 28, TIME_PM) ,
                            TIME(4, 28, TIME_PM) ,
                            TIME(5, 28, TIME_PM) ,
                            TIME(6, 28, TIME_PM) ,
                            TIME(7, 28, TIME_PM) ,
                            TIME(8, 28, TIME_PM) ,
                            TIME(10, 28, TIME_PM),
                        },
                        {   // South Sun Broadway
                            TIME(8, 31, TIME_AM) ,
                            TIME(10, 31, TIME_AM),
                            TIME(12, 31, TIME_PM),
                            TIME(2, 31, TIME_PM) ,
                            TIME(4, 31, TIME_PM) ,
                            TIME(5, 31, TIME_PM) ,
                            TIME(6, 31, TIME_PM) ,
                            TIME(7, 31, TIME_PM) ,
                            TIME(8, 31, TIME_PM) ,
                            TIME(10, 31, TIME_PM),
                        },
                        {   // South Sun Burlingame
                            TIME(8, 33, TIME_AM) ,
                            TIME(10, 33, TIME_AM),
                            TIME(12, 33, TIME_PM),
                            TIME(2, 33, TIME_PM) ,
                            TIME(4, 33, TIME_PM) ,
                            TIME(5, 33, TIME_PM) ,
                            TIME(6, 33, TIME_PM) ,
                            TIME(7, 33, TIME_PM) ,
                            TIME(8, 33, TIME_PM) ,
                            TIME(10, 33, TIME_PM),
                        },
                        {   // South Sun San Mateo
                            TIME(8, 36, TIME_AM) ,
                            TIME(10, 36, TIME_AM),
                            TIME(12, 36, TIME_PM),
                            TIME(2, 36, TIME_PM) ,
                            TIME(4, 36, TIME_PM) ,
                            TIME(5, 36, TIME_PM) ,
                            TIME(6, 36, TIME_PM) ,
                            TIME(7, 36, TIME_PM) ,
                            TIME(8, 36, TIME_PM) ,
                            TIME(10, 36, TIME_PM),
                        },
                        {   // South Sun Hayward Park
                            TIME(8, 39, TIME_AM) ,
                            TIME(10, 39, TIME_AM),
                            TIME(12, 39, TIME_PM),
                            TIME(2, 39, TIME_PM) ,
                            TIME(4, 39, TIME_PM) ,
                            TIME(5, 39, TIME_PM) ,
                            TIME(6, 39, TIME_PM) ,
                            TIME(7, 39, TIME_PM) ,
                            TIME(8, 39, TIME_PM) ,
                            TIME(10, 39, TIME_PM),
                        },
                        {   // South Sun Hillsdale
                            TIME(8, 42, TIME_AM) ,
                            TIME(10, 42, TIME_AM),
                            TIME(12, 42, TIME_PM),
                            TIME(2, 42, TIME_PM) ,
                            TIME(4, 42, TIME_PM) ,
                            TIME(5, 42, TIME_PM) ,
                            TIME(6, 42, TIME_PM) ,
                            TIME(7, 42, TIME_PM) ,
                            TIME(8, 42, TIME_PM) ,
                            TIME(10, 42, TIME_PM),
                        },
                        {   // South Sun Belmont
                            TIME(8, 47, TIME_AM) ,
                            TIME(10, 47, TIME_AM),
                            TIME(12, 47, TIME_PM),
                            TIME(2, 47, TIME_PM) ,
                            TIME(4, 47, TIME_PM) ,
                            TIME(5, 47, TIME_PM) ,
                            TIME(6, 47, TIME_PM) ,
                            TIME(7, 47, TIME_PM) ,
                            TIME(8, 47, TIME_PM) ,
                            TIME(10, 47, TIME_PM),
                        },
                        {   // South Sun San Carlos
                            TIME(8, 50, TIME_AM) ,
                            TIME(10, 50, TIME_AM),
                            TIME(12, 50, TIME_PM),
                            TIME(2, 50, TIME_PM) ,
                            TIME(4, 50, TIME_PM) ,
                            TIME(5, 50, TIME_PM) ,
                            TIME(6, 50, TIME_PM) ,
                            TIME(7, 50, TIME_PM) ,
                            TIME(8, 50, TIME_PM) ,
                            TIME(10, 50, TIME_PM),
                        },
                        {   // South Sun Redwood City
                            TIME(8, 55, TIME_AM) ,
                            TIME(10, 55, TIME_AM),
                            TIME(12, 55, TIME_PM),
                            TIME(2, 55, TIME_PM) ,
                            TIME(4, 55, TIME_PM) ,
                            TIME(5, 55, TIME_PM) ,
                            TIME(6, 55, TIME_PM) ,
                            TIME(7, 55, TIME_PM) ,
                            TIME(8, 55, TIME_PM) ,
                            TIME(10, 55, TIME_PM),
                        },
                        {   // South Sun Atherton
                            TIME(8, 59, TIME_AM) ,
                            TIME(10, 59, TIME_AM),
                            TIME(12, 59, TIME_PM),
                            TIME(2, 59, TIME_PM) ,
                            TIME(4, 59, TIME_PM) ,
                            TIME(5, 59, TIME_PM) ,
                            TIME(6, 59, TIME_PM) ,
                            TIME(7, 59, TIME_PM) ,
                            TIME(8, 59, TIME_PM) ,
                            TIME(10, 59, TIME_PM),
                        },
                        {   // South Sun Menlo Park
                            TIME(9, 4, TIME_AM) ,
                            TIME(11, 4, TIME_AM),
                            TIME(1, 4, TIME_PM) ,
                            TIME(3, 4, TIME_PM) ,
                            TIME(5, 4, TIME_PM) ,
                            TIME(6, 4, TIME_PM) ,
                            TIME(7, 4, TIME_PM) ,
                            TIME(8, 4, TIME_PM) ,
                            TIME(9, 4, TIME_PM) ,
                            TIME(11, 4, TIME_PM),
                        },
                        {   // South Sun Palo Alto
                            TIME(9, 7, TIME_AM) ,
                            TIME(11, 7, TIME_AM),
                            TIME(1, 7, TIME_PM) ,
                            TIME(3, 7, TIME_PM) ,
                            TIME(5, 7, TIME_PM) ,
                            TIME(6, 7, TIME_PM) ,
                            TIME(7, 7, TIME_PM) ,
                            TIME(8, 7, TIME_PM) ,
                            TIME(9, 7, TIME_PM) ,
                            TIME(11, 7, TIME_PM),
                        },
                        {   // South Sun California Av
                            TIME(9, 10, TIME_AM) ,
                            TIME(11, 10, TIME_AM),
                            TIME(1, 10, TIME_PM) ,
                            TIME(3, 10, TIME_PM) ,
                            TIME(5, 10, TIME_PM) ,
                            TIME(6, 10, TIME_PM) ,
                            TIME(7, 10, TIME_PM) ,
                            TIME(8, 10, TIME_PM) ,
                            TIME(9, 10, TIME_PM) ,
                            TIME(11, 10, TIME_PM),
                        },
                        {   // South Sun San Antonio
                            TIME(9, 13, TIME_AM) ,
                            TIME(11, 13, TIME_AM),
                            TIME(1, 13, TIME_PM) ,
                            TIME(3, 13, TIME_PM) ,
                            TIME(5, 13, TIME_PM) ,
                            TIME(6, 13, TIME_PM) ,
                            TIME(7, 13, TIME_PM) ,
                            TIME(8, 13, TIME_PM) ,
                            TIME(9, 13, TIME_PM) ,
                            TIME(11, 13, TIME_PM),
                        },
                        {   // South Sun Mountain View
                            TIME(9, 19, TIME_AM) ,
                            TIME(11, 19, TIME_AM),
                            TIME(1, 19, TIME_PM) ,
                            TIME(3, 19, TIME_PM) ,
                            TIME(5, 19, TIME_PM) ,
                            TIME(6, 19, TIME_PM) ,
                            TIME(7, 19, TIME_PM) ,
                            TIME(8, 19, TIME_PM) ,
                            TIME(9, 19, TIME_PM) ,
                            TIME(11, 19, TIME_PM),
                        },
                        {   // South Sun Sunnyvale
                            TIME(9, 23, TIME_AM) ,
                            TIME(11, 23, TIME_AM),
                            TIME(1, 23, TIME_PM) ,
                            TIME(3, 23, TIME_PM) ,
                            TIME(5, 23, TIME_PM) ,
                            TIME(6, 23, TIME_PM) ,
                            TIME(7, 23, TIME_PM) ,
                            TIME(8, 23, TIME_PM) ,
                            TIME(9, 23, TIME_PM) ,
                            TIME(11, 23, TIME_PM),
                        },
                        {   // South Sun Lawrence Expy
                            TIME(9, 27, TIME_AM) ,
                            TIME(11, 27, TIME_AM),
                            TIME(1, 27, TIME_PM) ,
                            TIME(3, 27, TIME_PM) ,
                            TIME(5, 27, TIME_PM) ,
                            TIME(6, 27, TIME_PM) ,
                            TIME(7, 27, TIME_PM) ,
                            TIME(8, 27, TIME_PM) ,
                            TIME(9, 27, TIME_PM) ,
                            TIME(11, 27, TIME_PM),
                        },
                        {   // South Sun Santa Clara
                            TIME(9, 32, TIME_AM) ,
                            TIME(11, 32, TIME_AM),
                            TIME(1, 32, TIME_PM) ,
                            TIME(3, 32, TIME_PM) ,
                            TIME(5, 32, TIME_PM) ,
                            TIME(6, 32, TIME_PM) ,
                            TIME(7, 32, TIME_PM) ,
                            TIME(8, 32, TIME_PM) ,
                            TIME(9, 32, TIME_PM) ,
                            TIME(11, 32, TIME_PM),
                        },
                        {   // South Sun College Park
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        },
                        {   // South Sun San Jose
                            TIME(9, 40, TIME_AM) ,
                            TIME(11, 40, TIME_AM),
                            TIME(1, 40, TIME_PM) ,
                            TIME(3, 40, TIME_PM) ,
                            TIME(5, 40, TIME_PM) ,
                            TIME(6, 40, TIME_PM) ,
                            TIME(7, 40, TIME_PM) ,
                            TIME(8, 40, TIME_PM) ,
                            TIME(9, 40, TIME_PM) ,
                            TIME(11, 40, TIME_PM),
                        },
                        {   // South Sun Tamien
                            TIME(10, 1, TIME_AM),
                            TIME(12, 1, TIME_PM),
                            TIME(2, 1, TIME_PM) ,
                            TIME(4, 1, TIME_PM) ,
                            TIME(6, 1, TIME_PM) ,
                            TIME(7, 1, TIME_PM) ,
                            TIME(8, 1, TIME_PM) ,
                            TIME(9, 1, TIME_PM) ,
                            TIME(10, 1, TIME_PM),
                            TIME(12, 1, TIME_AM),
                        },
                        {   // South Sun Capitol
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        },
                        {   // South Sun Blossom Hill
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        },
                        {   // South Sun Morgan Hill
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        },
                        {   // South Sun San Martin
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                            TIME(0, 99, TIME_AM) ,
                        },
                        {   // South Sun Gilroy
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                            TIME(0, 99, TIME_AM),
                        }
                    } // Southbound Sunday times (10)
                }
            };


// end of file